///////////////////////////////////////////////////////////
// This file was auto-generated by cmd/dumpstdlibast.go. //
// https://github.com/google/go-jsonnet#generated-stdlib //
//                                                       //
// --------------- DO NOT EDIT BY HAND! ---------------  //
///////////////////////////////////////////////////////////

package astgen

import (
	"github.com/google/go-jsonnet/ast"
)

var p7Var = "$"
var p7 = &p7Var
var p11Var = "object <anonymous>"
var p11 = &p11Var
var p15Var = "function <anonymous>"
var p15 = &p15Var
var p26Var = "thunk from <function <anonymous>>"
var p26 = &p26Var
var p34Var = "function <anonymous>"
var p34 = &p34Var
var p45Var = "thunk from <function <anonymous>>"
var p45 = &p45Var
var p53Var = "function <anonymous>"
var p53 = &p53Var
var p64Var = "thunk from <function <anonymous>>"
var p64 = &p64Var
var p72Var = "function <anonymous>"
var p72 = &p72Var
var p83Var = "thunk from <function <anonymous>>"
var p83 = &p83Var
var p91Var = "function <anonymous>"
var p91 = &p91Var
var p102Var = "thunk from <function <anonymous>>"
var p102 = &p102Var
var p110Var = "function <anonymous>"
var p110 = &p110Var
var p121Var = "thunk from <function <anonymous>>"
var p121 = &p121Var
var p130Var = "function <anonymous>"
var p130 = &p130Var
var p143Var = "thunk from <function <anonymous>>"
var p143 = &p143Var
var p160Var = "function <anonymous>"
var p160 = &p160Var
var p169Var = "thunk from <function <anonymous>>"
var p169 = &p169Var
var p182Var = "thunk from <function <anonymous>>"
var p182 = &p182Var
var p195Var = "thunk from <function <anonymous>>"
var p195 = &p195Var
var p214Var = "thunk from <function <anonymous>>"
var p214 = &p214Var
var p224Var = "thunk from <thunk from <function <anonymous>>>"
var p224 = &p224Var
var p233Var = "thunk from <thunk from <thunk from <function <anonymous>>>>"
var p233 = &p233Var
var p243Var = "thunk from <thunk from <thunk from <thunk from <function <anonymous>>>>>"
var p243 = &p243Var
var p256Var = "thunk from <thunk from <thunk from <thunk from <thunk from <function <anonymous>>>>>>"
var p256 = &p256Var
var p264Var = "function <anonymous>"
var p264 = &p264Var
var p295Var = "thunk from <function <anonymous>>"
var p295 = &p295Var
var p311Var = "thunk from <function <anonymous>>"
var p311 = &p311Var
var p327Var = "thunk from <function <anonymous>>"
var p327 = &p327Var
var p335Var = "function <anonymous>"
var p335 = &p335Var
var p348Var = "thunk from <function <anonymous>>"
var p348 = &p348Var
var p359Var = "thunk from <function <anonymous>>"
var p359 = &p359Var
var p376Var = "thunk from <function <anonymous>>"
var p376 = &p376Var
var p388Var = "thunk from <thunk from <function <anonymous>>>"
var p388 = &p388Var
var p398Var = "function <anonymous>"
var p398 = &p398Var
var p411Var = "thunk from <function <anonymous>>"
var p411 = &p411Var
var p422Var = "thunk from <function <anonymous>>"
var p422 = &p422Var
var p439Var = "thunk from <function <anonymous>>"
var p439 = &p439Var
var p452Var = "thunk from <thunk from <function <anonymous>>>"
var p452 = &p452Var
var p463Var = "thunk from <thunk from <function <anonymous>>>"
var p463 = &p463Var
var p474Var = "thunk from <thunk from <function <anonymous>>>"
var p474 = &p474Var
var p484Var = "function <anonymous>"
var p484 = &p484Var
var p499Var = "thunk from <function <anonymous>>"
var p499 = &p499Var
var p511Var = "thunk from <function <anonymous>>"
var p511 = &p511Var
var p536Var = "thunk from <function <anonymous>>"
var p536 = &p536Var
var p553Var = "function <anonymous>"
var p553 = &p553Var
var p557Var = "thunk <len> from <function <anonymous>>"
var p557 = &p557Var
var p566Var = "thunk from <thunk <len> from <function <anonymous>>>"
var p566 = &p566Var
var p587Var = "thunk from <function <anonymous>>"
var p587 = &p587Var
var p616Var = "thunk from <function <anonymous>>"
var p616 = &p616Var
var p636Var = "function <anonymous>"
var p636 = &p636Var
var p646Var = "thunk from <function <anonymous>>"
var p646 = &p646Var
var p655Var = "thunk from <thunk from <function <anonymous>>>"
var p655 = &p655Var
var p666Var = "function <anonymous>"
var p666 = &p666Var
var p676Var = "thunk from <function <anonymous>>"
var p676 = &p676Var
var p685Var = "thunk from <thunk from <function <anonymous>>>"
var p685 = &p685Var
var p691Var = "function <anonymous>"
var p691 = &p691Var
var p704Var = "function <anonymous>"
var p704 = &p704Var
var p713Var = "thunk from <function <anonymous>>"
var p713 = &p713Var
var p730Var = "thunk from <function <anonymous>>"
var p730 = &p730Var
var p766Var = "thunk from <function <anonymous>>"
var p766 = &p766Var
var p780Var = "thunk from <function <anonymous>>"
var p780 = &p780Var
var p798Var = "thunk from <function <anonymous>>"
var p798 = &p798Var
var p814Var = "thunk from <function <anonymous>>"
var p814 = &p814Var
var p823Var = "function <anonymous>"
var p823 = &p823Var
var p832Var = "thunk from <function <anonymous>>"
var p832 = &p832Var
var p847Var = "thunk from <function <anonymous>>"
var p847 = &p847Var
var p857Var = "thunk from <function <anonymous>>"
var p857 = &p857Var
var p876Var = "thunk from <function <anonymous>>"
var p876 = &p876Var
var p885Var = "function <anonymous>"
var p885 = &p885Var
var p894Var = "thunk from <function <anonymous>>"
var p894 = &p894Var
var p909Var = "thunk from <function <anonymous>>"
var p909 = &p909Var
var p919Var = "thunk from <function <anonymous>>"
var p919 = &p919Var
var p938Var = "thunk from <function <anonymous>>"
var p938 = &p938Var
var p947Var = "function <anonymous>"
var p947 = &p947Var
var p956Var = "thunk from <function <anonymous>>"
var p956 = &p956Var
var p969Var = "thunk from <function <anonymous>>"
var p969 = &p969Var
var p984Var = "thunk from <function <anonymous>>"
var p984 = &p984Var
var p997Var = "thunk from <function <anonymous>>"
var p997 = &p997Var
var p1017Var = "thunk from <function <anonymous>>"
var p1017 = &p1017Var
var p1033Var = "thunk from <function <anonymous>>"
var p1033 = &p1033Var
var p1049Var = "thunk from <function <anonymous>>"
var p1049 = &p1049Var
var p1058Var = "function <anonymous>"
var p1058 = &p1058Var
var p1067Var = "thunk from <function <anonymous>>"
var p1067 = &p1067Var
var p1080Var = "thunk from <function <anonymous>>"
var p1080 = &p1080Var
var p1095Var = "thunk from <function <anonymous>>"
var p1095 = &p1095Var
var p1109Var = "thunk from <function <anonymous>>"
var p1109 = &p1109Var
var p1116Var = "thunk <aux> from <function <anonymous>>"
var p1116 = &p1116Var
var p1121Var = "function <aux>"
var p1121 = &p1121Var
var p1125Var = "thunk <c> from <function <aux>>"
var p1125 = &p1125Var
var p1136Var = "thunk <i2> from <function <aux>>"
var p1136 = &p1136Var
var p1157Var = "thunk from <function <aux>>"
var p1157 = &p1157Var
var p1168Var = "thunk from <function <aux>>"
var p1168 = &p1168Var
var p1200Var = "thunk from <function <aux>>"
var p1200 = &p1200Var
var p1211Var = "thunk from <function <aux>>"
var p1211 = &p1211Var
var p1225Var = "thunk from <thunk from <function <aux>>>"
var p1225 = &p1225Var
var p1236Var = "thunk from <function <aux>>"
var p1236 = &p1236Var
var p1257Var = "thunk from <function <anonymous>>"
var p1257 = &p1257Var
var p1278Var = "thunk from <function <anonymous>>"
var p1278 = &p1278Var
var p1294Var = "thunk from <function <anonymous>>"
var p1294 = &p1294Var
var p1310Var = "thunk from <function <anonymous>>"
var p1310 = &p1310Var
var p1326Var = "thunk from <function <anonymous>>"
var p1326 = &p1326Var
var p1335Var = "function <anonymous>"
var p1335 = &p1335Var
var p1344Var = "thunk from <function <anonymous>>"
var p1344 = &p1344Var
var p1357Var = "thunk from <function <anonymous>>"
var p1357 = &p1357Var
var p1370Var = "thunk from <function <anonymous>>"
var p1370 = &p1370Var
var p1385Var = "thunk <str_len> from <function <anonymous>>"
var p1385 = &p1385Var
var p1394Var = "thunk from <thunk <str_len> from <function <anonymous>>>"
var p1394 = &p1394Var
var p1401Var = "thunk <from_len> from <function <anonymous>>"
var p1401 = &p1401Var
var p1410Var = "thunk from <thunk <from_len> from <function <anonymous>>>"
var p1410 = &p1410Var
var p1418Var = "thunk <found_at> from <function <anonymous>>"
var p1418 = &p1418Var
var p1422Var = "function <found_at>"
var p1422 = &p1422Var
var p1452Var = "thunk <replace_after> from <function <anonymous>>"
var p1452 = &p1452Var
var p1457Var = "function <replace_after>"
var p1457 = &p1457Var
var p1494Var = "thunk from <function <replace_after>>"
var p1494 = &p1494Var
var p1501Var = "thunk <new_index> from <function <replace_after>>"
var p1501 = &p1501Var
var p1514Var = "thunk from <thunk <new_index> from <function <replace_after>>>"
var p1514 = &p1514Var
var p1523Var = "thunk from <function <replace_after>>"
var p1523 = &p1523Var
var p1558Var = "thunk from <function <replace_after>>"
var p1558 = &p1558Var
var p1587Var = "thunk from <function <anonymous>>"
var p1587 = &p1587Var
var p1598Var = "thunk from <thunk from <function <anonymous>>>"
var p1598 = &p1598Var
var p1610Var = "thunk from <function <anonymous>>"
var p1610 = &p1610Var
var p1627Var = "function <anonymous>"
var p1627 = &p1627Var
var p1631Var = "thunk <cp> from <function <anonymous>>"
var p1631 = &p1631Var
var p1641Var = "thunk <up_letter> from <function <anonymous>>"
var p1641 = &p1641Var
var p1645Var = "function <up_letter>"
var p1645 = &p1645Var
var p1657Var = "thunk from <function <up_letter>>"
var p1657 = &p1657Var
var p1668Var = "thunk from <function <up_letter>>"
var p1668 = &p1668Var
var p1681Var = "thunk from <function <up_letter>>"
var p1681 = &p1681Var
var p1689Var = "thunk from <thunk from <function <up_letter>>>"
var p1689 = &p1689Var
var p1706Var = "thunk from <function <anonymous>>"
var p1706 = &p1706Var
var p1716Var = "thunk from <thunk from <function <anonymous>>>"
var p1716 = &p1716Var
var p1727Var = "thunk from <thunk from <thunk from <function <anonymous>>>>"
var p1727 = &p1727Var
var p1735Var = "function <anonymous>"
var p1735 = &p1735Var
var p1739Var = "thunk <cp> from <function <anonymous>>"
var p1739 = &p1739Var
var p1749Var = "thunk <down_letter> from <function <anonymous>>"
var p1749 = &p1749Var
var p1753Var = "function <down_letter>"
var p1753 = &p1753Var
var p1765Var = "thunk from <function <down_letter>>"
var p1765 = &p1765Var
var p1776Var = "thunk from <function <down_letter>>"
var p1776 = &p1776Var
var p1789Var = "thunk from <function <down_letter>>"
var p1789 = &p1789Var
var p1797Var = "thunk from <thunk from <function <down_letter>>>"
var p1797 = &p1797Var
var p1814Var = "thunk from <function <anonymous>>"
var p1814 = &p1814Var
var p1824Var = "thunk from <thunk from <function <anonymous>>>"
var p1824 = &p1824Var
var p1835Var = "thunk from <thunk from <thunk from <function <anonymous>>>>"
var p1835 = &p1835Var
var p1842Var = "function <anonymous>"
var p1842 = &p1842Var
var p1852Var = "thunk from <function <anonymous>>"
var p1852 = &p1852Var
var p1865Var = "function <anonymous>"
var p1865 = &p1865Var
var p1877Var = "function <anonymous>"
var p1877 = &p1877Var
var p1882Var = "thunk <joiner> from <function <anonymous>>"
var p1882 = &p1882Var
var p1893Var = "thunk from <thunk <joiner> from <function <anonymous>>>"
var p1893 = &p1893Var
var p1908Var = "thunk from <thunk <joiner> from <function <anonymous>>>"
var p1908 = &p1908Var
var p1924Var = "thunk from <function <anonymous>>"
var p1924 = &p1924Var
var p1935Var = "thunk from <thunk from <function <anonymous>>>"
var p1935 = &p1935Var
var p1941Var = "function <anonymous>"
var p1941 = &p1941Var
var p1949Var = "function <anonymous>"
var p1949 = &p1949Var
var p1955Var = "thunk <invar> from <function <anonymous>>"
var p1955 = &p1955Var
var p1960Var = "object <invar>"
var p1960 = &p1960Var
var p1993Var = "thunk from <object <invar>>"
var p1993 = &p1993Var
var p2021Var = "thunk from <object <invar>>"
var p2021 = &p2021Var
var p2033Var = "thunk from <object <invar>>"
var p2033 = &p2033Var
var p2085Var = "thunk from <function <anonymous>>"
var p2085 = &p2085Var
var p2096Var = "thunk from <function <anonymous>>"
var p2096 = &p2096Var
var p2103Var = "thunk <build> from <function <anonymous>>"
var p2103 = &p2103Var
var p2108Var = "function <build>"
var p2108 = &p2108Var
var p2142Var = "thunk from <function <build>>"
var p2142 = &p2142Var
var p2176Var = "thunk from <thunk from <function <build>>>"
var p2176 = &p2176Var
var p2203Var = "thunk from <function <anonymous>>"
var p2203 = &p2203Var
var p2240Var = "thunk from <function <anonymous>>"
var p2240 = &p2240Var
var p2270Var = "thunk from <function <anonymous>>"
var p2270 = &p2270Var
var p2291Var = "function <anonymous>"
var p2291 = &p2291Var
var p2302Var = "thunk from <function <anonymous>>"
var p2302 = &p2302Var
var p2316Var = "thunk from <function <anonymous>>"
var p2316 = &p2316Var
var p2333Var = "thunk from <function <anonymous>>"
var p2333 = &p2333Var
var p2347Var = "thunk from <function <anonymous>>"
var p2347 = &p2347Var
var p2356Var = "thunk from <thunk from <function <anonymous>>>"
var p2356 = &p2356Var
var p2369Var = "function <anonymous>"
var p2369 = &p2369Var
var p2378Var = "thunk from <function <anonymous>>"
var p2378 = &p2378Var
var p2387Var = "thunk from <thunk from <function <anonymous>>>"
var p2387 = &p2387Var
var p2391Var = "function <anonymous>"
var p2391 = &p2391Var
var p2405Var = "function <anonymous>"
var p2405 = &p2405Var
var p2418Var = "thunk from <function <anonymous>>"
var p2418 = &p2418Var
var p2429Var = "thunk from <function <anonymous>>"
var p2429 = &p2429Var
var p2441Var = "thunk from <function <anonymous>>"
var p2441 = &p2441Var
var p2457Var = "thunk from <function <anonymous>>"
var p2457 = &p2457Var
var p2469Var = "thunk from <function <anonymous>>"
var p2469 = &p2469Var
var p2495Var = "thunk from <function <anonymous>>"
var p2495 = &p2495Var
var p2507Var = "thunk from <function <anonymous>>"
var p2507 = &p2507Var
var p2516Var = "function <anonymous>"
var p2516 = &p2516Var
var p2529Var = "thunk from <function <anonymous>>"
var p2529 = &p2529Var
var p2546Var = "thunk from <function <anonymous>>"
var p2546 = &p2546Var
var p2564Var = "thunk from <function <anonymous>>"
var p2564 = &p2564Var
var p2577Var = "thunk from <function <anonymous>>"
var p2577 = &p2577Var
var p2594Var = "thunk from <function <anonymous>>"
var p2594 = &p2594Var
var p2607Var = "thunk from <function <anonymous>>"
var p2607 = &p2607Var
var p2616Var = "thunk from <thunk from <function <anonymous>>>"
var p2616 = &p2616Var
var p2622Var = "function <anonymous>"
var p2622 = &p2622Var
var p2628Var = "thunk from <function <anonymous>>"
var p2628 = &p2628Var
var p2640Var = "function <anonymous>"
var p2640 = &p2640Var
var p2653Var = "thunk from <function <anonymous>>"
var p2653 = &p2653Var
var p2670Var = "thunk from <function <anonymous>>"
var p2670 = &p2670Var
var p2688Var = "thunk from <function <anonymous>>"
var p2688 = &p2688Var
var p2701Var = "thunk from <function <anonymous>>"
var p2701 = &p2701Var
var p2718Var = "thunk from <function <anonymous>>"
var p2718 = &p2718Var
var p2731Var = "thunk from <function <anonymous>>"
var p2731 = &p2731Var
var p2740Var = "thunk from <thunk from <function <anonymous>>>"
var p2740 = &p2740Var
var p2746Var = "function <anonymous>"
var p2746 = &p2746Var
var p2752Var = "thunk from <function <anonymous>>"
var p2752 = &p2752Var
var p2766Var = "function <anonymous>"
var p2766 = &p2766Var
var p2779Var = "thunk from <function <anonymous>>"
var p2779 = &p2779Var
var p2796Var = "thunk from <function <anonymous>>"
var p2796 = &p2796Var
var p2812Var = "thunk from <function <anonymous>>"
var p2812 = &p2812Var
var p2829Var = "thunk from <function <anonymous>>"
var p2829 = &p2829Var
var p2861Var = "object <anonymous>"
var p2861 = &p2861Var
var p2867Var = "thunk from <object <anonymous>>"
var p2867 = &p2867Var
var p2884Var = "thunk from <function <anonymous>>"
var p2884 = &p2884Var
var p2892Var = "function <anonymous>"
var p2892 = &p2892Var
var p2905Var = "thunk from <function <anonymous>>"
var p2905 = &p2905Var
var p2922Var = "thunk from <function <anonymous>>"
var p2922 = &p2922Var
var p2936Var = "thunk from <function <anonymous>>"
var p2936 = &p2936Var
var p2948Var = "thunk from <function <anonymous>>"
var p2948 = &p2948Var
var p2957Var = "thunk from <thunk from <function <anonymous>>>"
var p2957 = &p2957Var
var p2966Var = "thunk from <thunk from <thunk from <function <anonymous>>>>"
var p2966 = &p2966Var
var p2972Var = "function <anonymous>"
var p2972 = &p2972Var
var p2978Var = "thunk from <function <anonymous>>"
var p2978 = &p2978Var
var p2996Var = "thunk from <function <anonymous>>"
var p2996 = &p2996Var
var p3008Var = "thunk from <function <anonymous>>"
var p3008 = &p3008Var
var p3018Var = "thunk from <thunk from <function <anonymous>>>"
var p3018 = &p3018Var
var p3027Var = "thunk from <thunk from <thunk from <function <anonymous>>>>"
var p3027 = &p3027Var
var p3033Var = "function <anonymous>"
var p3033 = &p3033Var
var p3039Var = "thunk from <function <anonymous>>"
var p3039 = &p3039Var
var p3060Var = "thunk from <function <anonymous>>"
var p3060 = &p3060Var
var p3068Var = "function <anonymous>"
var p3068 = &p3068Var
var p3072Var = "thunk <aux> from <function <anonymous>>"
var p3072 = &p3072Var
var p3077Var = "function <aux>"
var p3077 = &p3077Var
var p3092Var = "thunk from <function <aux>>"
var p3092 = &p3092Var
var p3116Var = "thunk from <function <aux>>"
var p3116 = &p3116Var
var p3141Var = "thunk from <function <aux>>"
var p3141 = &p3141Var
var p3156Var = "thunk from <function <aux>>"
var p3156 = &p3156Var
var p3174Var = "thunk from <function <aux>>"
var p3174 = &p3174Var
var p3183Var = "thunk from <thunk from <function <aux>>>"
var p3183 = &p3183Var
var p3196Var = "thunk from <thunk from <function <aux>>>"
var p3196 = &p3196Var
var p3214Var = "thunk from <function <aux>>"
var p3214 = &p3214Var
var p3240Var = "thunk from <function <aux>>"
var p3240 = &p3240Var
var p3276Var = "thunk from <function <anonymous>>"
var p3276 = &p3276Var
var p3293Var = "thunk from <function <anonymous>>"
var p3293 = &p3293Var
var p3307Var = "thunk from <function <anonymous>>"
var p3307 = &p3307Var
var p3316Var = "thunk from <function <anonymous>>"
var p3316 = &p3316Var
var p3333Var = "thunk from <function <anonymous>>"
var p3333 = &p3333Var
var p3342Var = "thunk from <function <anonymous>>"
var p3342 = &p3342Var
var p3363Var = "thunk from <function <anonymous>>"
var p3363 = &p3363Var
var p3370Var = "function <anonymous>"
var p3370 = &p3370Var
var p3380Var = "thunk from <function <anonymous>>"
var p3380 = &p3380Var
var p3388Var = "thunk from <thunk from <function <anonymous>>>"
var p3388 = &p3388Var
var p3395Var = "function <anonymous>"
var p3395 = &p3395Var
var p3406Var = "thunk from <function <anonymous>>"
var p3406 = &p3406Var
var p3423Var = "thunk from <function <anonymous>>"
var p3423 = &p3423Var
var p3435Var = "thunk from <function <anonymous>>"
var p3435 = &p3435Var
var p3451Var = "thunk from <thunk from <function <anonymous>>>"
var p3451 = &p3451Var
var p3460Var = "thunk from <thunk from <thunk from <function <anonymous>>>>"
var p3460 = &p3460Var
var p3486Var = "thunk from <function <anonymous>>"
var p3486 = &p3486Var
var p3497Var = "function <anonymous>"
var p3497 = &p3497Var
var p3501Var = "thunk <try_parse_mapping_key> from <function <anonymous>>"
var p3501 = &p3501Var
var p3507Var = "function <try_parse_mapping_key>"
var p3507 = &p3507Var
var p3520Var = "thunk from <function <try_parse_mapping_key>>"
var p3520 = &p3520Var
var p3527Var = "thunk <c> from <function <try_parse_mapping_key>>"
var p3527 = &p3527Var
var p3546Var = "thunk <consume> from <function <try_parse_mapping_key>>"
var p3546 = &p3546Var
var p3551Var = "function <consume>"
var p3551 = &p3551Var
var p3566Var = "thunk from <function <consume>>"
var p3566 = &p3566Var
var p3577Var = "thunk <c> from <function <consume>>"
var p3577 = &p3577Var
var p3598Var = "thunk from <function <consume>>"
var p3598 = &p3598Var
var p3618Var = "object <anonymous>"
var p3618 = &p3618Var
var p3633Var = "thunk from <function <try_parse_mapping_key>>"
var p3633 = &p3633Var
var p3648Var = "object <anonymous>"
var p3648 = &p3648Var
var p3659Var = "thunk <try_parse_cflags> from <function <anonymous>>"
var p3659 = &p3659Var
var p3664Var = "function <try_parse_cflags>"
var p3664 = &p3664Var
var p3668Var = "thunk <consume> from <function <try_parse_cflags>>"
var p3668 = &p3668Var
var p3674Var = "function <consume>"
var p3674 = &p3674Var
var p3687Var = "thunk from <function <consume>>"
var p3687 = &p3687Var
var p3694Var = "thunk <c> from <function <consume>>"
var p3694 = &p3694Var
var p3715Var = "thunk from <function <consume>>"
var p3715 = &p3715Var
var p3730Var = "object <anonymous>"
var p3730 = &p3730Var
var p3746Var = "thunk from <function <consume>>"
var p3746 = &p3746Var
var p3761Var = "object <anonymous>"
var p3761 = &p3761Var
var p3777Var = "thunk from <function <consume>>"
var p3777 = &p3777Var
var p3792Var = "object <anonymous>"
var p3792 = &p3792Var
var p3808Var = "thunk from <function <consume>>"
var p3808 = &p3808Var
var p3823Var = "object <anonymous>"
var p3823 = &p3823Var
var p3839Var = "thunk from <function <consume>>"
var p3839 = &p3839Var
var p3854Var = "object <anonymous>"
var p3854 = &p3854Var
var p3862Var = "object <anonymous>"
var p3862 = &p3862Var
var p3876Var = "thunk from <function <try_parse_cflags>>"
var p3876 = &p3876Var
var p3884Var = "object <anonymous>"
var p3884 = &p3884Var
var p3898Var = "thunk <try_parse_field_width> from <function <anonymous>>"
var p3898 = &p3898Var
var p3903Var = "function <try_parse_field_width>"
var p3903 = &p3903Var
var p3920Var = "thunk from <function <try_parse_field_width>>"
var p3920 = &p3920Var
var p3937Var = "object <anonymous>"
var p3937 = &p3937Var
var p3950Var = "thunk <consume> from <function <try_parse_field_width>>"
var p3950 = &p3950Var
var p3956Var = "function <consume>"
var p3956 = &p3956Var
var p3969Var = "thunk from <function <consume>>"
var p3969 = &p3969Var
var p3976Var = "thunk <c> from <function <consume>>"
var p3976 = &p3976Var
var p3997Var = "thunk from <function <consume>>"
var p3997 = &p3997Var
var p4027Var = "thunk from <function <consume>>"
var p4027 = &p4027Var
var p4057Var = "thunk from <function <consume>>"
var p4057 = &p4057Var
var p4087Var = "thunk from <function <consume>>"
var p4087 = &p4087Var
var p4117Var = "thunk from <function <consume>>"
var p4117 = &p4117Var
var p4147Var = "thunk from <function <consume>>"
var p4147 = &p4147Var
var p4177Var = "thunk from <function <consume>>"
var p4177 = &p4177Var
var p4207Var = "thunk from <function <consume>>"
var p4207 = &p4207Var
var p4237Var = "thunk from <function <consume>>"
var p4237 = &p4237Var
var p4267Var = "thunk from <function <consume>>"
var p4267 = &p4267Var
var p4289Var = "object <anonymous>"
var p4289 = &p4289Var
var p4303Var = "thunk from <function <try_parse_field_width>>"
var p4303 = &p4303Var
var p4313Var = "thunk <try_parse_precision> from <function <anonymous>>"
var p4313 = &p4313Var
var p4319Var = "function <try_parse_precision>"
var p4319 = &p4319Var
var p4332Var = "thunk from <function <try_parse_precision>>"
var p4332 = &p4332Var
var p4339Var = "thunk <c> from <function <try_parse_precision>>"
var p4339 = &p4339Var
var p4360Var = "thunk from <function <try_parse_precision>>"
var p4360 = &p4360Var
var p4374Var = "object <anonymous>"
var p4374 = &p4374Var
var p4386Var = "thunk <try_parse_length_modifier> from <function <anonymous>>"
var p4386 = &p4386Var
var p4392Var = "function <try_parse_length_modifier>"
var p4392 = &p4392Var
var p4405Var = "thunk from <function <try_parse_length_modifier>>"
var p4405 = &p4405Var
var p4412Var = "thunk <c> from <function <try_parse_length_modifier>>"
var p4412 = &p4412Var
var p4457Var = "thunk <parse_conv_type> from <function <anonymous>>"
var p4457 = &p4457Var
var p4463Var = "function <parse_conv_type>"
var p4463 = &p4463Var
var p4476Var = "thunk from <function <parse_conv_type>>"
var p4476 = &p4476Var
var p4483Var = "thunk <c> from <function <parse_conv_type>>"
var p4483 = &p4483Var
var p4517Var = "object <anonymous>"
var p4517 = &p4517Var
var p4540Var = "object <anonymous>"
var p4540 = &p4540Var
var p4563Var = "object <anonymous>"
var p4563 = &p4563Var
var p4586Var = "object <anonymous>"
var p4586 = &p4586Var
var p4609Var = "object <anonymous>"
var p4609 = &p4609Var
var p4632Var = "object <anonymous>"
var p4632 = &p4632Var
var p4655Var = "object <anonymous>"
var p4655 = &p4655Var
var p4678Var = "object <anonymous>"
var p4678 = &p4678Var
var p4701Var = "object <anonymous>"
var p4701 = &p4701Var
var p4724Var = "object <anonymous>"
var p4724 = &p4724Var
var p4747Var = "object <anonymous>"
var p4747 = &p4747Var
var p4770Var = "object <anonymous>"
var p4770 = &p4770Var
var p4793Var = "object <anonymous>"
var p4793 = &p4793Var
var p4819Var = "thunk <parse_code> from <function <anonymous>>"
var p4819 = &p4819Var
var p4825Var = "function <parse_code>"
var p4825 = &p4825Var
var p4838Var = "thunk from <function <parse_code>>"
var p4838 = &p4838Var
var p4845Var = "thunk <mkey> from <function <parse_code>>"
var p4845 = &p4845Var
var p4851Var = "thunk from <thunk <mkey> from <function <parse_code>>>"
var p4851 = &p4851Var
var p4860Var = "thunk <cflags> from <function <parse_code>>"
var p4860 = &p4860Var
var p4866Var = "thunk from <thunk <cflags> from <function <parse_code>>>"
var p4866 = &p4866Var
var p4878Var = "thunk <fw> from <function <parse_code>>"
var p4878 = &p4878Var
var p4884Var = "thunk from <thunk <fw> from <function <parse_code>>>"
var p4884 = &p4884Var
var p4896Var = "thunk <prec> from <function <parse_code>>"
var p4896 = &p4896Var
var p4902Var = "thunk from <thunk <prec> from <function <parse_code>>>"
var p4902 = &p4902Var
var p4914Var = "thunk <len_mod> from <function <parse_code>>"
var p4914 = &p4914Var
var p4920Var = "thunk from <thunk <len_mod> from <function <parse_code>>>"
var p4920 = &p4920Var
var p4932Var = "thunk <ctype> from <function <parse_code>>"
var p4932 = &p4932Var
var p4938Var = "thunk from <thunk <ctype> from <function <parse_code>>>"
var p4938 = &p4938Var
var p4948Var = "object <anonymous>"
var p4948 = &p4948Var
var p4959Var = "object <anonymous>"
var p4959 = &p4959Var
var p5002Var = "thunk <parse_codes> from <function <anonymous>>"
var p5002 = &p5002Var
var p5007Var = "function <parse_codes>"
var p5007 = &p5007Var
var p5022Var = "thunk from <function <parse_codes>>"
var p5022 = &p5022Var
var p5033Var = "thunk from <function <parse_codes>>"
var p5033 = &p5033Var
var p5041Var = "thunk <c> from <function <parse_codes>>"
var p5041 = &p5041Var
var p5060Var = "thunk <r> from <function <parse_codes>>"
var p5060 = &p5060Var
var p5066Var = "thunk from <thunk <r> from <function <parse_codes>>>"
var p5066 = &p5066Var
var p5080Var = "thunk from <function <parse_codes>>"
var p5080 = &p5080Var
var p5095Var = "thunk from <thunk from <function <parse_codes>>>"
var p5095 = &p5095Var
var p5111Var = "thunk from <function <parse_codes>>"
var p5111 = &p5111Var
var p5131Var = "thunk <codes> from <function <anonymous>>"
var p5131 = &p5131Var
var p5137Var = "thunk from <thunk <codes> from <function <anonymous>>>"
var p5137 = &p5137Var
var p5151Var = "thunk <padding> from <function <anonymous>>"
var p5151 = &p5151Var
var p5155Var = "function <padding>"
var p5155 = &p5155Var
var p5159Var = "thunk <aux> from <function <padding>>"
var p5159 = &p5159Var
var p5164Var = "function <aux>"
var p5164 = &p5164Var
var p5182Var = "thunk from <function <aux>>"
var p5182 = &p5182Var
var p5200Var = "thunk from <function <padding>>"
var p5200 = &p5200Var
var p5209Var = "thunk <pad_left> from <function <anonymous>>"
var p5209 = &p5209Var
var p5213Var = "function <pad_left>"
var p5213 = &p5213Var
var p5222Var = "thunk from <function <pad_left>>"
var p5222 = &p5222Var
var p5235Var = "thunk from <thunk from <function <pad_left>>>"
var p5235 = &p5235Var
var p5247Var = "thunk <pad_right> from <function <anonymous>>"
var p5247 = &p5247Var
var p5251Var = "function <pad_right>"
var p5251 = &p5251Var
var p5262Var = "thunk from <function <pad_right>>"
var p5262 = &p5262Var
var p5275Var = "thunk from <thunk from <function <pad_right>>>"
var p5275 = &p5275Var
var p5297Var = "thunk <render_int> from <function <anonymous>>"
var p5297 = &p5297Var
var p5303Var = "function <render_int>"
var p5303 = &p5303Var
var p5308Var = "thunk <dec> from <function <render_int>>"
var p5308 = &p5308Var
var p5323Var = "thunk <aux> from <thunk <dec> from <function <render_int>>>"
var p5323 = &p5323Var
var p5328Var = "function <aux>"
var p5328 = &p5328Var
var p5348Var = "thunk from <function <aux>>"
var p5348 = &p5348Var
var p5357Var = "thunk from <thunk from <function <aux>>>"
var p5357 = &p5357Var
var p5382Var = "thunk from <thunk <dec> from <function <render_int>>>"
var p5382 = &p5382Var
var p5389Var = "thunk <zp> from <function <render_int>>"
var p5389 = &p5389Var
var p5412Var = "thunk <zp2> from <function <render_int>>"
var p5412 = &p5412Var
var p5421Var = "thunk from <thunk <zp2> from <function <render_int>>>"
var p5421 = &p5421Var
var p5430Var = "thunk <dec2> from <function <render_int>>"
var p5430 = &p5430Var
var p5436Var = "thunk from <thunk <dec2> from <function <render_int>>>"
var p5436 = &p5436Var
var p5467Var = "thunk <render_hex> from <function <anonymous>>"
var p5467 = &p5467Var
var p5472Var = "function <render_hex>"
var p5472 = &p5472Var
var p5476Var = "thunk <numerals> from <function <render_hex>>"
var p5476 = &p5476Var
var p5481Var = "thunk from <thunk <numerals> from <function <render_hex>>>"
var p5481 = &p5481Var
var p5499Var = "thunk from <thunk <numerals> from <function <render_hex>>>"
var p5499 = &p5499Var
var p5509Var = "thunk from <thunk <numerals> from <function <render_hex>>>"
var p5509 = &p5509Var
var p5520Var = "thunk <n_> from <function <render_hex>>"
var p5520 = &p5520Var
var p5529Var = "thunk from <thunk <n_> from <function <render_hex>>>"
var p5529 = &p5529Var
var p5536Var = "thunk <aux> from <function <render_hex>>"
var p5536 = &p5536Var
var p5541Var = "function <aux>"
var p5541 = &p5541Var
var p5560Var = "thunk from <function <aux>>"
var p5560 = &p5560Var
var p5569Var = "thunk from <thunk from <function <aux>>>"
var p5569 = &p5569Var
var p5594Var = "thunk <hex> from <function <render_hex>>"
var p5594 = &p5594Var
var p5607Var = "thunk from <thunk <hex> from <function <render_hex>>>"
var p5607 = &p5607Var
var p5617Var = "thunk from <thunk <hex> from <function <render_hex>>>"
var p5617 = &p5617Var
var p5626Var = "thunk from <thunk from <thunk <hex> from <function <render_hex>>>>"
var p5626 = &p5626Var
var p5633Var = "thunk <neg> from <function <render_hex>>"
var p5633 = &p5633Var
var p5643Var = "thunk <zp> from <function <render_hex>>"
var p5643 = &p5643Var
var p5675Var = "thunk <zp2> from <function <render_hex>>"
var p5675 = &p5675Var
var p5684Var = "thunk from <thunk <zp2> from <function <render_hex>>>"
var p5684 = &p5684Var
var p5693Var = "thunk <hex2> from <function <render_hex>>"
var p5693 = &p5693Var
var p5713Var = "thunk from <thunk <hex2> from <function <render_hex>>>"
var p5713 = &p5713Var
var p5743Var = "thunk <strip_trailing_zero> from <function <anonymous>>"
var p5743 = &p5743Var
var p5748Var = "function <strip_trailing_zero>"
var p5748 = &p5748Var
var p5752Var = "thunk <aux> from <function <strip_trailing_zero>>"
var p5752 = &p5752Var
var p5757Var = "function <aux>"
var p5757 = &p5757Var
var p5786Var = "thunk from <function <aux>>"
var p5786 = &p5786Var
var p5804Var = "thunk from <function <aux>>"
var p5804 = &p5804Var
var p5819Var = "thunk from <function <strip_trailing_zero>>"
var p5819 = &p5819Var
var p5832Var = "thunk from <thunk from <function <strip_trailing_zero>>>"
var p5832 = &p5832Var
var p5841Var = "thunk <render_float_dec> from <function <anonymous>>"
var p5841 = &p5841Var
var p5846Var = "function <render_float_dec>"
var p5846 = &p5846Var
var p5850Var = "thunk <n_> from <function <render_float_dec>>"
var p5850 = &p5850Var
var p5859Var = "thunk from <thunk <n_> from <function <render_float_dec>>>"
var p5859 = &p5859Var
var p5866Var = "thunk <whole> from <function <render_float_dec>>"
var p5866 = &p5866Var
var p5875Var = "thunk from <thunk <whole> from <function <render_float_dec>>>"
var p5875 = &p5875Var
var p5882Var = "thunk <dot_size> from <function <render_float_dec>>"
var p5882 = &p5882Var
var p5902Var = "thunk <zp> from <function <render_float_dec>>"
var p5902 = &p5902Var
var p5917Var = "thunk <str> from <function <render_float_dec>>"
var p5917 = &p5917Var
var p5923Var = "thunk from <thunk <str> from <function <render_float_dec>>>"
var p5923 = &p5923Var
var p5964Var = "thunk <frac> from <function <render_float_dec>>"
var p5964 = &p5964Var
var p5973Var = "thunk from <thunk <frac> from <function <render_float_dec>>>"
var p5973 = &p5973Var
var p5992Var = "thunk from <thunk from <thunk <frac> from <function <render_float_dec>>>>"
var p5992 = &p5992Var
var p6013Var = "thunk <frac_str> from <function <render_float_dec>>"
var p6013 = &p6013Var
var p6019Var = "thunk from <thunk <frac_str> from <function <render_float_dec>>>"
var p6019 = &p6019Var
var p6049Var = "thunk from <function <render_float_dec>>"
var p6049 = &p6049Var
var p6063Var = "thunk <render_float_sci> from <function <anonymous>>"
var p6063 = &p6063Var
var p6068Var = "function <render_float_sci>"
var p6068 = &p6068Var
var p6072Var = "thunk <exponent> from <function <render_float_sci>>"
var p6072 = &p6072Var
var p6089Var = "thunk from <thunk <exponent> from <function <render_float_sci>>>"
var p6089 = &p6089Var
var p6100Var = "thunk from <thunk from <thunk <exponent> from <function <render_float_sci>>>>"
var p6100 = &p6100Var
var p6109Var = "thunk from <thunk from <thunk from <thunk <exponent> from <function <render_float_sci>>>>>"
var p6109 = &p6109Var
var p6120Var = "thunk from <thunk from <thunk <exponent> from <function <render_float_sci>>>>"
var p6120 = &p6120Var
var p6126Var = "thunk <suff> from <function <render_float_sci>>"
var p6126 = &p6126Var
var p6141Var = "thunk from <thunk <suff> from <function <render_float_sci>>>"
var p6141 = &p6141Var
var p6155Var = "thunk from <thunk from <thunk <suff> from <function <render_float_sci>>>>"
var p6155 = &p6155Var
var p6168Var = "thunk <mantissa> from <function <render_float_sci>>"
var p6168 = &p6168Var
var p6195Var = "thunk from <thunk <mantissa> from <function <render_float_sci>>>"
var p6195 = &p6195Var
var p6216Var = "thunk from <thunk <mantissa> from <function <render_float_sci>>>"
var p6216 = &p6216Var
var p6224Var = "thunk <zp2> from <function <render_float_sci>>"
var p6224 = &p6224Var
var p6237Var = "thunk from <thunk <zp2> from <function <render_float_sci>>>"
var p6237 = &p6237Var
var p6248Var = "thunk from <function <render_float_sci>>"
var p6248 = &p6248Var
var p6270Var = "thunk <format_code> from <function <anonymous>>"
var p6270 = &p6270Var
var p6275Var = "function <format_code>"
var p6275 = &p6275Var
var p6279Var = "thunk <cflags> from <function <format_code>>"
var p6279 = &p6279Var
var p6289Var = "thunk <fpprec> from <function <format_code>>"
var p6289 = &p6289Var
var p6304Var = "thunk <iprec> from <function <format_code>>"
var p6304 = &p6304Var
var p6319Var = "thunk <zp> from <function <format_code>>"
var p6319 = &p6319Var
var p6359Var = "thunk from <function <format_code>>"
var p6359 = &p6359Var
var p6386Var = "thunk from <function <format_code>>"
var p6386 = &p6386Var
var p6412Var = "thunk from <function <format_code>>"
var p6412 = &p6412Var
var p6422Var = "thunk from <function <format_code>>"
var p6422 = &p6422Var
var p6437Var = "thunk from <thunk from <function <format_code>>>"
var p6437 = &p6437Var
var p6446Var = "thunk from <thunk from <thunk from <function <format_code>>>>"
var p6446 = &p6446Var
var p6489Var = "thunk from <function <format_code>>"
var p6489 = &p6489Var
var p6515Var = "thunk from <function <format_code>>"
var p6515 = &p6515Var
var p6523Var = "thunk <zero_prefix> from <function <format_code>>"
var p6523 = &p6523Var
var p6539Var = "thunk from <function <format_code>>"
var p6539 = &p6539Var
var p6554Var = "thunk from <thunk from <function <format_code>>>"
var p6554 = &p6554Var
var p6563Var = "thunk from <thunk from <thunk from <function <format_code>>>>"
var p6563 = &p6563Var
var p6607Var = "thunk from <function <format_code>>"
var p6607 = &p6607Var
var p6633Var = "thunk from <function <format_code>>"
var p6633 = &p6633Var
var p6643Var = "thunk from <function <format_code>>"
var p6643 = &p6643Var
var p6652Var = "thunk from <thunk from <function <format_code>>>"
var p6652 = &p6652Var
var p6709Var = "thunk from <function <format_code>>"
var p6709 = &p6709Var
var p6735Var = "thunk from <function <format_code>>"
var p6735 = &p6735Var
var p6745Var = "thunk from <function <format_code>>"
var p6745 = &p6745Var
var p6798Var = "thunk from <function <format_code>>"
var p6798 = &p6798Var
var p6824Var = "thunk from <function <format_code>>"
var p6824 = &p6824Var
var p6834Var = "thunk from <function <format_code>>"
var p6834 = &p6834Var
var p6893Var = "thunk from <function <format_code>>"
var p6893 = &p6893Var
var p6919Var = "thunk from <function <format_code>>"
var p6919 = &p6919Var
var p6927Var = "thunk <exponent> from <function <format_code>>"
var p6927 = &p6927Var
var p6936Var = "thunk from <thunk <exponent> from <function <format_code>>>"
var p6936 = &p6936Var
var p6947Var = "thunk from <thunk from <thunk <exponent> from <function <format_code>>>>"
var p6947 = &p6947Var
var p6956Var = "thunk from <thunk from <thunk from <thunk <exponent> from <function <format_code>>>>>"
var p6956 = &p6956Var
var p6967Var = "thunk from <thunk from <thunk <exponent> from <function <format_code>>>>"
var p6967 = &p6967Var
var p6992Var = "thunk from <function <format_code>>"
var p6992 = &p6992Var
var p7039Var = "thunk <digits_before_pt> from <function <format_code>>"
var p7039 = &p7039Var
var p7048Var = "thunk from <thunk <digits_before_pt> from <function <format_code>>>"
var p7048 = &p7048Var
var p7061Var = "thunk from <function <format_code>>"
var p7061 = &p7061Var
var p7122Var = "thunk from <function <format_code>>"
var p7122 = &p7122Var
var p7135Var = "thunk from <function <format_code>>"
var p7135 = &p7135Var
var p7151Var = "thunk from <function <format_code>>"
var p7151 = &p7151Var
var p7168Var = "thunk from <function <format_code>>"
var p7168 = &p7168Var
var p7190Var = "thunk from <function <format_code>>"
var p7190 = &p7190Var
var p7208Var = "thunk from <function <format_code>>"
var p7208 = &p7208Var
var p7228Var = "thunk <format_codes_arr> from <function <anonymous>>"
var p7228 = &p7228Var
var p7233Var = "function <format_codes_arr>"
var p7233 = &p7233Var
var p7248Var = "thunk from <function <format_codes_arr>>"
var p7248 = &p7248Var
var p7266Var = "thunk from <function <format_codes_arr>>"
var p7266 = &p7266Var
var p7287Var = "thunk from <function <format_codes_arr>>"
var p7287 = &p7287Var
var p7302Var = "thunk <code> from <function <format_codes_arr>>"
var p7302 = &p7302Var
var p7322Var = "thunk from <function <format_codes_arr>>"
var p7322 = &p7322Var
var p7332Var = "thunk from <function <format_codes_arr>>"
var p7332 = &p7332Var
var p7355Var = "thunk <tmp> from <function <format_codes_arr>>"
var p7355 = &p7355Var
var p7370Var = "object <anonymous>"
var p7370 = &p7370Var
var p7391Var = "thunk from <object <anonymous>>"
var p7391 = &p7391Var
var p7412Var = "thunk from <object <anonymous>>"
var p7412 = &p7412Var
var p7430Var = "object <anonymous>"
var p7430 = &p7430Var
var p7443Var = "thunk <tmp2> from <function <format_codes_arr>>"
var p7443 = &p7443Var
var p7458Var = "object <anonymous>"
var p7458 = &p7458Var
var p7485Var = "thunk from <object <anonymous>>"
var p7485 = &p7485Var
var p7506Var = "thunk from <object <anonymous>>"
var p7506 = &p7506Var
var p7530Var = "object <anonymous>"
var p7530 = &p7530Var
var p7546Var = "thunk <j2> from <function <format_codes_arr>>"
var p7546 = &p7546Var
var p7557Var = "thunk <val> from <function <format_codes_arr>>"
var p7557 = &p7557Var
var p7572Var = "thunk from <thunk <val> from <function <format_codes_arr>>>"
var p7572 = &p7572Var
var p7601Var = "thunk from <thunk <val> from <function <format_codes_arr>>>"
var p7601 = &p7601Var
var p7612Var = "thunk <s> from <function <format_codes_arr>>"
var p7612 = &p7612Var
var p7632Var = "thunk from <thunk <s> from <function <format_codes_arr>>>"
var p7632 = &p7632Var
var p7654Var = "thunk <s_padded> from <function <format_codes_arr>>"
var p7654 = &p7654Var
var p7671Var = "thunk from <thunk <s_padded> from <function <format_codes_arr>>>"
var p7671 = &p7671Var
var p7687Var = "thunk from <thunk <s_padded> from <function <format_codes_arr>>>"
var p7687 = &p7687Var
var p7701Var = "thunk <j3> from <function <format_codes_arr>>"
var p7701 = &p7701Var
var p7728Var = "thunk from <function <format_codes_arr>>"
var p7728 = &p7728Var
var p7751Var = "thunk <format_codes_obj> from <function <anonymous>>"
var p7751 = &p7751Var
var p7756Var = "function <format_codes_obj>"
var p7756 = &p7756Var
var p7771Var = "thunk from <function <format_codes_obj>>"
var p7771 = &p7771Var
var p7782Var = "thunk <code> from <function <format_codes_obj>>"
var p7782 = &p7782Var
var p7802Var = "thunk from <function <format_codes_obj>>"
var p7802 = &p7802Var
var p7812Var = "thunk from <function <format_codes_obj>>"
var p7812 = &p7812Var
var p7834Var = "thunk <f> from <function <format_codes_obj>>"
var p7834 = &p7834Var
var p7860Var = "thunk <fw> from <function <format_codes_obj>>"
var p7860 = &p7860Var
var p7886Var = "thunk <prec> from <function <format_codes_obj>>"
var p7886 = &p7886Var
var p7912Var = "thunk <val> from <function <format_codes_obj>>"
var p7912 = &p7912Var
var p7923Var = "thunk from <thunk <val> from <function <format_codes_obj>>>"
var p7923 = &p7923Var
var p7949Var = "thunk <s> from <function <format_codes_obj>>"
var p7949 = &p7949Var
var p7969Var = "thunk from <thunk <s> from <function <format_codes_obj>>>"
var p7969 = &p7969Var
var p7985Var = "thunk <s_padded> from <function <format_codes_obj>>"
var p7985 = &p7985Var
var p8002Var = "thunk from <thunk <s_padded> from <function <format_codes_obj>>>"
var p8002 = &p8002Var
var p8015Var = "thunk from <thunk <s_padded> from <function <format_codes_obj>>>"
var p8015 = &p8015Var
var p8027Var = "thunk from <function <format_codes_obj>>"
var p8027 = &p8027Var
var p8054Var = "thunk from <function <anonymous>>"
var p8054 = &p8054Var
var p8063Var = "thunk from <function <anonymous>>"
var p8063 = &p8063Var
var p8082Var = "thunk from <function <anonymous>>"
var p8082 = &p8082Var
var p8091Var = "thunk from <function <anonymous>>"
var p8091 = &p8091Var
var p8105Var = "thunk from <function <anonymous>>"
var p8105 = &p8105Var
var p8111Var = "thunk from <thunk from <function <anonymous>>>"
var p8111 = &p8111Var
var p8122Var = "function <anonymous>"
var p8122 = &p8122Var
var p8126Var = "thunk <aux> from <function <anonymous>>"
var p8126 = &p8126Var
var p8131Var = "function <aux>"
var p8131 = &p8131Var
var p8149Var = "thunk from <function <aux>>"
var p8149 = &p8149Var
var p8159Var = "thunk from <thunk from <function <aux>>>"
var p8159 = &p8159Var
var p8179Var = "thunk from <function <anonymous>>"
var p8179 = &p8179Var
var p8196Var = "thunk from <thunk from <function <anonymous>>>"
var p8196 = &p8196Var
var p8205Var = "function <anonymous>"
var p8205 = &p8205Var
var p8209Var = "thunk <aux> from <function <anonymous>>"
var p8209 = &p8209Var
var p8214Var = "function <aux>"
var p8214 = &p8214Var
var p8229Var = "thunk from <function <aux>>"
var p8229 = &p8229Var
var p8242Var = "thunk from <function <aux>>"
var p8242 = &p8242Var
var p8252Var = "thunk from <thunk from <function <aux>>>"
var p8252 = &p8252Var
var p8272Var = "thunk from <function <anonymous>>"
var p8272 = &p8272Var
var p8285Var = "function <anonymous>"
var p8285 = &p8285Var
var p8298Var = "thunk from <function <anonymous>>"
var p8298 = &p8298Var
var p8315Var = "thunk from <function <anonymous>>"
var p8315 = &p8315Var
var p8331Var = "thunk from <function <anonymous>>"
var p8331 = &p8331Var
var p8348Var = "thunk from <function <anonymous>>"
var p8348 = &p8348Var
var p8364Var = "thunk from <function <anonymous>>"
var p8364 = &p8364Var
var p8381Var = "thunk from <function <anonymous>>"
var p8381 = &p8381Var
var p8394Var = "thunk from <function <anonymous>>"
var p8394 = &p8394Var
var p8405Var = "thunk from <thunk from <function <anonymous>>>"
var p8405 = &p8405Var
var p8414Var = "function <anonymous>"
var p8414 = &p8414Var
var p8446Var = "function <anonymous>"
var p8446 = &p8446Var
var p8459Var = "thunk from <function <anonymous>>"
var p8459 = &p8459Var
var p8476Var = "thunk from <function <anonymous>>"
var p8476 = &p8476Var
var p8499Var = "function <anonymous>"
var p8499 = &p8499Var
var p8512Var = "thunk from <function <anonymous>>"
var p8512 = &p8512Var
var p8529Var = "thunk from <function <anonymous>>"
var p8529 = &p8529Var
var p8561Var = "function <anonymous>"
var p8561 = &p8561Var
var p8574Var = "thunk from <function <anonymous>>"
var p8574 = &p8574Var
var p8591Var = "thunk from <function <anonymous>>"
var p8591 = &p8591Var
var p8607Var = "thunk from <function <anonymous>>"
var p8607 = &p8607Var
var p8624Var = "thunk from <function <anonymous>>"
var p8624 = &p8624Var
var p8646Var = "function <anonymous>"
var p8646 = &p8646Var
var p8659Var = "thunk from <function <anonymous>>"
var p8659 = &p8659Var
var p8676Var = "thunk from <function <anonymous>>"
var p8676 = &p8676Var
var p8692Var = "thunk from <function <anonymous>>"
var p8692 = &p8692Var
var p8709Var = "thunk from <function <anonymous>>"
var p8709 = &p8709Var
var p8730Var = "function <anonymous>"
var p8730 = &p8730Var
var p8759Var = "function <anonymous>"
var p8759 = &p8759Var
var p8769Var = "thunk from <function <anonymous>>"
var p8769 = &p8769Var
var p8772Var = "function <anonymous>"
var p8772 = &p8772Var
var p8787Var = "function <anonymous>"
var p8787 = &p8787Var
var p8791Var = "thunk <body_lines> from <function <anonymous>>"
var p8791 = &p8791Var
var p8795Var = "function <body_lines>"
var p8795 = &p8795Var
var p8805Var = "thunk from <function <body_lines>>"
var p8805 = &p8805Var
var p8822Var = "thunk from <thunk from <function <body_lines>>>"
var p8822 = &p8822Var
var p8826Var = "thunk <value_or_values> from <thunk from <thunk from <function <body_lines>>>>"
var p8826 = &p8826Var
var p8844Var = "thunk from <thunk from <thunk from <function <body_lines>>>>"
var p8844 = &p8844Var
var p8869Var = "thunk from <thunk from <thunk from <function <body_lines>>>>"
var p8869 = &p8869Var
var p8874Var = "thunk from <thunk from <thunk from <thunk from <function <body_lines>>>>>"
var p8874 = &p8874Var
var p8894Var = "thunk from <thunk from <thunk from <function <body_lines>>>>"
var p8894 = &p8894Var
var p8899Var = "thunk from <thunk from <thunk from <thunk from <function <body_lines>>>>>"
var p8899 = &p8899Var
var p8912Var = "thunk from <thunk from <function <body_lines>>>"
var p8912 = &p8912Var
var p8919Var = "thunk <section_lines> from <function <anonymous>>"
var p8919 = &p8919Var
var p8923Var = "function <section_lines>"
var p8923 = &p8923Var
var p8937Var = "thunk from <function <section_lines>>"
var p8937 = &p8937Var
var p8942Var = "thunk from <thunk from <function <section_lines>>>"
var p8942 = &p8942Var
var p8950Var = "thunk from <function <section_lines>>"
var p8950 = &p8950Var
var p8954Var = "thunk <main_body> from <function <anonymous>>"
var p8954 = &p8954Var
var p8965Var = "thunk from <thunk <main_body> from <function <anonymous>>>"
var p8965 = &p8965Var
var p8974Var = "thunk from <thunk <main_body> from <function <anonymous>>>"
var p8974 = &p8974Var
var p8996Var = "thunk from <thunk <all_sections> from <function <anonymous>>>"
var p8996 = &p8996Var
var p9003Var = "thunk from <thunk from <thunk <all_sections> from <function <anonymous>>>>"
var p9003 = &p9003Var
var p9015Var = "thunk <all_sections> from <function <anonymous>>"
var p9015 = &p9015Var
var p9024Var = "thunk from <thunk <all_sections> from <function <anonymous>>>"
var p9024 = &p9024Var
var p9039Var = "thunk from <function <anonymous>>"
var p9039 = &p9039Var
var p9055Var = "thunk from <thunk from <function <anonymous>>>"
var p9055 = &p9055Var
var p9060Var = "thunk from <thunk from <function <anonymous>>>"
var p9060 = &p9060Var
var p9067Var = "function <anonymous>"
var p9067 = &p9067Var
var p9071Var = "thunk <str> from <function <anonymous>>"
var p9071 = &p9071Var
var p9080Var = "thunk from <thunk <str> from <function <anonymous>>>"
var p9080 = &p9080Var
var p9087Var = "thunk <trans> from <function <anonymous>>"
var p9087 = &p9087Var
var p9092Var = "function <trans>"
var p9092 = &p9092Var
var p9167Var = "thunk <cp> from <function <trans>>"
var p9167 = &p9167Var
var p9176Var = "thunk from <thunk <cp> from <function <trans>>>"
var p9176 = &p9176Var
var p9214Var = "thunk from <function <trans>>"
var p9214 = &p9214Var
var p9239Var = "thunk from <function <anonymous>>"
var p9239 = &p9239Var
var p9255Var = "thunk from <thunk from <function <anonymous>>>"
var p9255 = &p9255Var
var p9261Var = "thunk from <thunk from <thunk from <function <anonymous>>>>"
var p9261 = &p9261Var
var p9272Var = "thunk from <thunk from <function <anonymous>>>"
var p9272 = &p9272Var
var p9279Var = "function <anonymous>"
var p9279 = &p9279Var
var p9289Var = "thunk from <function <anonymous>>"
var p9289 = &p9289Var
var p9297Var = "function <anonymous>"
var p9297 = &p9297Var
var p9301Var = "thunk <str> from <function <anonymous>>"
var p9301 = &p9301Var
var p9310Var = "thunk from <thunk <str> from <function <anonymous>>>"
var p9310 = &p9310Var
var p9317Var = "thunk <trans> from <function <anonymous>>"
var p9317 = &p9317Var
var p9321Var = "function <trans>"
var p9321 = &p9321Var
var p9353Var = "thunk from <function <anonymous>>"
var p9353 = &p9353Var
var p9369Var = "thunk from <thunk from <function <anonymous>>>"
var p9369 = &p9369Var
var p9375Var = "thunk from <thunk from <thunk from <function <anonymous>>>>"
var p9375 = &p9375Var
var p9386Var = "thunk from <thunk from <function <anonymous>>>"
var p9386 = &p9386Var
var p9394Var = "function <anonymous>"
var p9394 = &p9394Var
var p9398Var = "thunk <str> from <function <anonymous>>"
var p9398 = &p9398Var
var p9407Var = "thunk from <thunk <str> from <function <anonymous>>>"
var p9407 = &p9407Var
var p9414Var = "thunk <trans> from <function <anonymous>>"
var p9414 = &p9414Var
var p9418Var = "function <trans>"
var p9418 = &p9418Var
var p9441Var = "thunk from <function <anonymous>>"
var p9441 = &p9441Var
var p9445Var = "function <anonymous>"
var p9445 = &p9445Var
var p9455Var = "thunk from <function <anonymous>>"
var p9455 = &p9455Var
var p9466Var = "thunk from <thunk from <function <anonymous>>>"
var p9466 = &p9466Var
var p9474Var = "function <anonymous>"
var p9474 = &p9474Var
var p9483Var = "thunk from <function <anonymous>>"
var p9483 = &p9483Var
var p9492Var = "function <anonymous>"
var p9492 = &p9492Var
var p9496Var = "thunk <aux> from <function <anonymous>>"
var p9496 = &p9496Var
var p9501Var = "function <aux>"
var p9501 = &p9501Var
var p9542Var = "thunk from <function <aux>>"
var p9542 = &p9542Var
var p9562Var = "thunk from <function <aux>>"
var p9562 = &p9562Var
var p9574Var = "thunk from <function <aux>>"
var p9574 = &p9574Var
var p9588Var = "thunk from <function <aux>>"
var p9588 = &p9588Var
var p9610Var = "thunk from <function <aux>>"
var p9610 = &p9610Var
var p9617Var = "thunk <range> from <function <aux>>"
var p9617 = &p9617Var
var p9626Var = "thunk from <thunk <range> from <function <aux>>>"
var p9626 = &p9626Var
var p9638Var = "thunk from <thunk from <thunk <range> from <function <aux>>>>"
var p9638 = &p9638Var
var p9646Var = "thunk <new_indent> from <function <aux>>"
var p9646 = &p9646Var
var p9657Var = "thunk <lines> from <function <aux>>"
var p9657 = &p9657Var
var p9664Var = "thunk from <thunk <lines> from <function <aux>>>"
var p9664 = &p9664Var
var p9675Var = "thunk from <thunk <lines> from <function <aux>>>"
var p9675 = &p9675Var
var p9678Var = "thunk from <thunk from <thunk <lines> from <function <aux>>>>"
var p9678 = &p9678Var
var p9695Var = "thunk from <thunk from <thunk <lines> from <function <aux>>>>"
var p9695 = &p9695Var
var p9699Var = "thunk from <thunk from <thunk from <thunk <lines> from <function <aux>>>>>"
var p9699 = &p9699Var
var p9709Var = "thunk from <thunk from <thunk from <thunk from <thunk <lines> from <function <aux>>>>>>"
var p9709 = &p9709Var
var p9723Var = "thunk from <thunk from <thunk from <thunk from <thunk from <thunk <lines> from <function <aux>>>>>>>"
var p9723 = &p9723Var
var p9734Var = "thunk from <thunk <lines> from <function <aux>>>"
var p9734 = &p9734Var
var p9752Var = "thunk from <function <aux>>"
var p9752 = &p9752Var
var p9767Var = "thunk from <function <aux>>"
var p9767 = &p9767Var
var p9774Var = "thunk <lines> from <function <aux>>"
var p9774 = &p9774Var
var p9781Var = "thunk from <thunk <lines> from <function <aux>>>"
var p9781 = &p9781Var
var p9792Var = "thunk from <thunk <lines> from <function <aux>>>"
var p9792 = &p9792Var
var p9795Var = "thunk from <thunk from <thunk <lines> from <function <aux>>>>"
var p9795 = &p9795Var
var p9812Var = "thunk from <thunk from <thunk <lines> from <function <aux>>>>"
var p9812 = &p9812Var
var p9816Var = "thunk from <thunk from <thunk from <thunk <lines> from <function <aux>>>>>"
var p9816 = &p9816Var
var p9837Var = "thunk from <thunk from <thunk from <thunk from <thunk <lines> from <function <aux>>>>>>"
var p9837 = &p9837Var
var p9847Var = "thunk from <thunk from <thunk from <thunk from <thunk <lines> from <function <aux>>>>>>"
var p9847 = &p9847Var
var p9861Var = "thunk from <thunk from <thunk from <thunk from <thunk from <thunk <lines> from <function <aux>>>>>>>"
var p9861 = &p9861Var
var p9878Var = "thunk from <thunk from <thunk <lines> from <function <aux>>>>"
var p9878 = &p9878Var
var p9885Var = "thunk from <thunk <lines> from <function <aux>>>"
var p9885 = &p9885Var
var p9903Var = "thunk from <function <aux>>"
var p9903 = &p9903Var
var p9914Var = "thunk from <function <anonymous>>"
var p9914 = &p9914Var
var p9923Var = "function <anonymous>"
var p9923 = &p9923Var
var p9929Var = "thunk <aux> from <function <anonymous>>"
var p9929 = &p9929Var
var p9934Var = "function <aux>"
var p9934 = &p9934Var
var p9975Var = "thunk from <function <aux>>"
var p9975 = &p9975Var
var p9995Var = "thunk from <function <aux>>"
var p9995 = &p9995Var
var p10002Var = "thunk <len> from <function <aux>>"
var p10002 = &p10002Var
var p10011Var = "thunk from <thunk <len> from <function <aux>>>"
var p10011 = &p10011Var
var p10043Var = "thunk <split> from <function <aux>>"
var p10043 = &p10043Var
var p10052Var = "thunk from <thunk <split> from <function <aux>>>"
var p10052 = &p10052Var
var p10065Var = "thunk from <function <aux>>"
var p10065 = &p10065Var
var p10078Var = "thunk from <thunk from <function <aux>>>"
var p10078 = &p10078Var
var p10101Var = "thunk from <thunk from <function <aux>>>"
var p10101 = &p10101Var
var p10116Var = "thunk from <function <aux>>"
var p10116 = &p10116Var
var p10130Var = "thunk from <function <aux>>"
var p10130 = &p10130Var
var p10152Var = "thunk from <function <aux>>"
var p10152 = &p10152Var
var p10168Var = "thunk from <function <aux>>"
var p10168 = &p10168Var
var p10179Var = "thunk <params> from <function <aux>>"
var p10179 = &p10179Var
var p10184Var = "function <params>"
var p10184 = &p10184Var
var p10197Var = "thunk from <function <params>>"
var p10197 = &p10197Var
var p10210Var = "thunk from <function <params>>"
var p10210 = &p10210Var
var p10218Var = "object <anonymous>"
var p10218 = &p10218Var
var p10242Var = "thunk from <function <params>>"
var p10242 = &p10242Var
var p10255Var = "thunk from <function <params>>"
var p10255 = &p10255Var
var p10263Var = "object <anonymous>"
var p10263 = &p10263Var
var p10275Var = "object <anonymous>"
var p10275 = &p10275Var
var p10284Var = "thunk <range> from <function <aux>>"
var p10284 = &p10284Var
var p10293Var = "thunk from <thunk <range> from <function <aux>>>"
var p10293 = &p10293Var
var p10305Var = "thunk from <thunk from <thunk <range> from <function <aux>>>>"
var p10305 = &p10305Var
var p10338Var = "thunk from <thunk <parts> from <function <aux>>>"
var p10338 = &p10338Var
var p10355Var = "thunk from <thunk from <thunk <parts> from <function <aux>>>>"
var p10355 = &p10355Var
var p10369Var = "thunk from <thunk from <thunk from <thunk <parts> from <function <aux>>>>>"
var p10369 = &p10369Var
var p10377Var = "thunk <parts> from <function <aux>>"
var p10377 = &p10377Var
var p10381Var = "thunk from <thunk <parts> from <function <aux>>>"
var p10381 = &p10381Var
var p10387Var = "thunk from <thunk from <thunk <parts> from <function <aux>>>>"
var p10387 = &p10387Var
var p10405Var = "thunk from <function <aux>>"
var p10405 = &p10405Var
var p10424Var = "thunk from <function <aux>>"
var p10424 = &p10424Var
var p10440Var = "thunk from <function <aux>>"
var p10440 = &p10440Var
var p10451Var = "thunk <params> from <function <aux>>"
var p10451 = &p10451Var
var p10456Var = "function <params>"
var p10456 = &p10456Var
var p10469Var = "thunk from <function <params>>"
var p10469 = &p10469Var
var p10482Var = "thunk from <function <params>>"
var p10482 = &p10482Var
var p10490Var = "object <anonymous>"
var p10490 = &p10490Var
var p10520Var = "thunk from <function <params>>"
var p10520 = &p10520Var
var p10533Var = "thunk from <function <params>>"
var p10533 = &p10533Var
var p10541Var = "object <anonymous>"
var p10541 = &p10541Var
var p10557Var = "object <anonymous>"
var p10557 = &p10557Var
var p10591Var = "thunk from <thunk <lines> from <function <aux>>>"
var p10591 = &p10591Var
var p10607Var = "thunk from <thunk from <thunk <lines> from <function <aux>>>>"
var p10607 = &p10607Var
var p10621Var = "thunk from <thunk from <thunk <lines> from <function <aux>>>>"
var p10621 = &p10621Var
var p10635Var = "thunk from <thunk from <thunk from <thunk <lines> from <function <aux>>>>>"
var p10635 = &p10635Var
var p10643Var = "thunk <lines> from <function <aux>>"
var p10643 = &p10643Var
var p10647Var = "thunk from <thunk <lines> from <function <aux>>>"
var p10647 = &p10647Var
var p10653Var = "thunk from <thunk from <thunk <lines> from <function <aux>>>>"
var p10653 = &p10653Var
var p10668Var = "thunk from <thunk <lines> from <function <aux>>>"
var p10668 = &p10668Var
var p10680Var = "thunk from <function <aux>>"
var p10680 = &p10680Var
var p10695Var = "thunk from <function <anonymous>>"
var p10695 = &p10695Var
var p10704Var = "function <anonymous>"
var p10704 = &p10704Var
var p10720Var = "thunk from <function <anonymous>>"
var p10720 = &p10720Var
var p10737Var = "thunk from <function <anonymous>>"
var p10737 = &p10737Var
var p10756Var = "thunk from <function <anonymous>>"
var p10756 = &p10756Var
var p10772Var = "thunk from <thunk from <function <anonymous>>>"
var p10772 = &p10772Var
var p10781Var = "thunk from <thunk from <thunk from <function <anonymous>>>>"
var p10781 = &p10781Var
var p10800Var = "function <anonymous>"
var p10800 = &p10800Var
var p10811Var = "thunk from <function <anonymous>>"
var p10811 = &p10811Var
var p10841Var = "thunk from <thunk <fields> from <function <anonymous>>>"
var p10841 = &p10841Var
var p10846Var = "thunk from <thunk from <thunk <fields> from <function <anonymous>>>>"
var p10846 = &p10846Var
var p10855Var = "thunk from <thunk from <thunk from <thunk <fields> from <function <anonymous>>>>>"
var p10855 = &p10855Var
var p10866Var = "thunk from <thunk from <thunk from <thunk <fields> from <function <anonymous>>>>>"
var p10866 = &p10866Var
var p10873Var = "thunk <fields> from <function <anonymous>>"
var p10873 = &p10873Var
var p10882Var = "thunk from <thunk <fields> from <function <anonymous>>>"
var p10882 = &p10882Var
var p10898Var = "thunk from <function <anonymous>>"
var p10898 = &p10898Var
var p10907Var = "thunk from <thunk from <function <anonymous>>>"
var p10907 = &p10907Var
var p10922Var = "thunk from <function <anonymous>>"
var p10922 = &p10922Var
var p10938Var = "thunk from <function <anonymous>>"
var p10938 = &p10938Var
var p10947Var = "thunk from <thunk from <function <anonymous>>>"
var p10947 = &p10947Var
var p10963Var = "thunk from <thunk from <thunk from <function <anonymous>>>>"
var p10963 = &p10963Var
var p10972Var = "thunk from <thunk from <thunk from <thunk from <function <anonymous>>>>>"
var p10972 = &p10972Var
var p10988Var = "thunk from <function <anonymous>>"
var p10988 = &p10988Var
var p11004Var = "thunk from <function <anonymous>>"
var p11004 = &p11004Var
var p11013Var = "thunk from <thunk from <function <anonymous>>>"
var p11013 = &p11013Var
var p11027Var = "thunk from <function <anonymous>>"
var p11027 = &p11027Var
var p11044Var = "thunk from <function <anonymous>>"
var p11044 = &p11044Var
var p11056Var = "thunk from <function <anonymous>>"
var p11056 = &p11056Var
var p11095Var = "function <anonymous>"
var p11095 = &p11095Var
var p11121Var = "thunk from <thunk <vars> from <function <anonymous>>>"
var p11121 = &p11121Var
var p11126Var = "thunk from <thunk from <thunk <vars> from <function <anonymous>>>>"
var p11126 = &p11126Var
var p11137Var = "thunk from <thunk from <thunk from <thunk <vars> from <function <anonymous>>>>>"
var p11137 = &p11137Var
var p11144Var = "thunk <vars> from <function <anonymous>>"
var p11144 = &p11144Var
var p11153Var = "thunk from <thunk <vars> from <function <anonymous>>>"
var p11153 = &p11153Var
var p11165Var = "thunk from <function <anonymous>>"
var p11165 = &p11165Var
var p11173Var = "thunk from <thunk from <function <anonymous>>>"
var p11173 = &p11173Var
var p11180Var = "function <anonymous>"
var p11180 = &p11180Var
var p11193Var = "thunk from <function <anonymous>>"
var p11193 = &p11193Var
var p11216Var = "thunk from <function <anonymous>>"
var p11216 = &p11216Var
var p11224Var = "thunk <aux> from <function <anonymous>>"
var p11224 = &p11224Var
var p11229Var = "function <aux>"
var p11229 = &p11229Var
var p11240Var = "thunk from <function <aux>>"
var p11240 = &p11240Var
var p11251Var = "thunk <tag> from <function <aux>>"
var p11251 = &p11251Var
var p11261Var = "thunk <has_attrs> from <function <aux>>"
var p11261 = &p11261Var
var p11274Var = "thunk from <thunk <has_attrs> from <function <aux>>>"
var p11274 = &p11274Var
var p11286Var = "thunk from <thunk <has_attrs> from <function <aux>>>"
var p11286 = &p11286Var
var p11296Var = "thunk <attrs> from <function <aux>>"
var p11296 = &p11296Var
var p11311Var = "thunk <children> from <function <aux>>"
var p11311 = &p11311Var
var p11346Var = "thunk <attrs_str> from <function <aux>>"
var p11346 = &p11346Var
var p11356Var = "thunk from <thunk <attrs_str> from <function <aux>>>"
var p11356 = &p11356Var
var p11380Var = "thunk from <thunk from <thunk <attrs_str> from <function <aux>>>>"
var p11380 = &p11380Var
var p11385Var = "thunk from <thunk from <thunk from <thunk <attrs_str> from <function <aux>>>>>"
var p11385 = &p11385Var
var p11402Var = "thunk from <thunk from <thunk <attrs_str> from <function <aux>>>>"
var p11402 = &p11402Var
var p11414Var = "thunk from <function <aux>>"
var p11414 = &p11414Var
var p11418Var = "thunk from <thunk from <function <aux>>>"
var p11418 = &p11418Var
var p11439Var = "thunk from <thunk from <thunk from <function <aux>>>>"
var p11439 = &p11439Var
var p11445Var = "thunk from <thunk from <thunk from <thunk from <function <aux>>>>>"
var p11445 = &p11445Var
var p11460Var = "thunk from <function <anonymous>>"
var p11460 = &p11460Var
var p11468Var = "function <anonymous>"
var p11468 = &p11468Var
var p11473Var = "thunk <bytes> from <function <anonymous>>"
var p11473 = &p11473Var
var p11484Var = "thunk from <thunk <bytes> from <function <anonymous>>>"
var p11484 = &p11484Var
var p11496Var = "thunk from <thunk <bytes> from <function <anonymous>>>"
var p11496 = &p11496Var
var p11500Var = "function <anonymous>"
var p11500 = &p11500Var
var p11509Var = "thunk from <function <anonymous>>"
var p11509 = &p11509Var
var p11522Var = "thunk <aux> from <function <anonymous>>"
var p11522 = &p11522Var
var p11527Var = "function <aux>"
var p11527 = &p11527Var
var p11542Var = "thunk from <function <aux>>"
var p11542 = &p11542Var
var p11566Var = "thunk from <function <aux>>"
var p11566 = &p11566Var
var p11573Var = "thunk <str> from <function <aux>>"
var p11573 = &p11573Var
var p11622Var = "thunk from <function <aux>>"
var p11622 = &p11622Var
var p11654Var = "thunk from <function <aux>>"
var p11654 = &p11654Var
var p11661Var = "thunk <str> from <function <aux>>"
var p11661 = &p11661Var
var p11750Var = "thunk from <function <aux>>"
var p11750 = &p11750Var
var p11769Var = "thunk <str> from <function <aux>>"
var p11769 = &p11769Var
var p11891Var = "thunk from <function <aux>>"
var p11891 = &p11891Var
var p11909Var = "thunk <sanity> from <function <anonymous>>"
var p11909 = &p11909Var
var p11918Var = "thunk from <thunk <sanity> from <function <anonymous>>>"
var p11918 = &p11918Var
var p11921Var = "function <anonymous>"
var p11921 = &p11921Var
var p11951Var = "thunk from <function <anonymous>>"
var p11951 = &p11951Var
var p11961Var = "function <anonymous>"
var p11961 = &p11961Var
var p11982Var = "thunk from <function <anonymous>>"
var p11982 = &p11982Var
var p12006Var = "thunk <aux> from <function <anonymous>>"
var p12006 = &p12006Var
var p12011Var = "function <aux>"
var p12011 = &p12011Var
var p12026Var = "thunk from <function <aux>>"
var p12026 = &p12026Var
var p12038Var = "thunk <n1> from <function <aux>>"
var p12038 = &p12038Var
var p12042Var = "thunk from <thunk <n1> from <function <aux>>>"
var p12042 = &p12042Var
var p12080Var = "thunk <n2> from <function <aux>>"
var p12080 = &p12080Var
var p12100Var = "thunk from <thunk <n2> from <function <aux>>>"
var p12100 = &p12100Var
var p12144Var = "thunk <n3> from <function <aux>>"
var p12144 = &p12144Var
var p12164Var = "thunk from <thunk <n3> from <function <aux>>>"
var p12164 = &p12164Var
var p12205Var = "thunk from <function <aux>>"
var p12205 = &p12205Var
var p12233Var = "thunk from <function <anonymous>>"
var p12233 = &p12233Var
var p12243Var = "function <anonymous>"
var p12243 = &p12243Var
var p12247Var = "thunk <bytes> from <function <anonymous>>"
var p12247 = &p12247Var
var p12256Var = "thunk from <thunk <bytes> from <function <anonymous>>>"
var p12256 = &p12256Var
var p12268Var = "thunk from <function <anonymous>>"
var p12268 = &p12268Var
var p12278Var = "thunk from <thunk from <function <anonymous>>>"
var p12278 = &p12278Var
var p12282Var = "function <anonymous>"
var p12282 = &p12282Var
var p12291Var = "thunk from <function <anonymous>>"
var p12291 = &p12291Var
var p12301Var = "function <anonymous>"
var p12301 = &p12301Var
var p12305Var = "thunk <l> from <function <anonymous>>"
var p12305 = &p12305Var
var p12314Var = "thunk from <thunk <l> from <function <anonymous>>>"
var p12314 = &p12314Var
var p12326Var = "thunk from <function <anonymous>>"
var p12326 = &p12326Var
var p12332Var = "function <anonymous>"
var p12332 = &p12332Var
var p12350Var = "function <anonymous>"
var p12350 = &p12350Var
var p12357Var = "thunk <quickSort> from <function <anonymous>>"
var p12357 = &p12357Var
var p12361Var = "function <quickSort>"
var p12361 = &p12361Var
var p12368Var = "thunk <l> from <function <quickSort>>"
var p12368 = &p12368Var
var p12377Var = "thunk from <thunk <l> from <function <quickSort>>>"
var p12377 = &p12377Var
var p12393Var = "thunk from <function <quickSort>>"
var p12393 = &p12393Var
var p12405Var = "thunk <pos> from <function <quickSort>>"
var p12405 = &p12405Var
var p12411Var = "thunk <pivot> from <function <quickSort>>"
var p12411 = &p12411Var
var p12417Var = "thunk from <thunk <pivot> from <function <quickSort>>>"
var p12417 = &p12417Var
var p12428Var = "thunk <rest> from <function <quickSort>>"
var p12428 = &p12428Var
var p12437Var = "thunk from <thunk <rest> from <function <quickSort>>>"
var p12437 = &p12437Var
var p12446Var = "function <anonymous>"
var p12446 = &p12446Var
var p12474Var = "thunk <left> from <function <quickSort>>"
var p12474 = &p12474Var
var p12483Var = "thunk from <thunk <left> from <function <quickSort>>>"
var p12483 = &p12483Var
var p12487Var = "function <anonymous>"
var p12487 = &p12487Var
var p12495Var = "thunk from <function <anonymous>>"
var p12495 = &p12495Var
var p12506Var = "thunk <right> from <function <quickSort>>"
var p12506 = &p12506Var
var p12515Var = "thunk from <thunk <right> from <function <quickSort>>>"
var p12515 = &p12515Var
var p12519Var = "function <anonymous>"
var p12519 = &p12519Var
var p12527Var = "thunk from <function <anonymous>>"
var p12527 = &p12527Var
var p12544Var = "thunk from <function <quickSort>>"
var p12544 = &p12544Var
var p12552Var = "thunk from <function <quickSort>>"
var p12552 = &p12552Var
var p12564Var = "thunk from <function <quickSort>>"
var p12564 = &p12564Var
var p12573Var = "thunk <merge> from <function <anonymous>>"
var p12573 = &p12573Var
var p12578Var = "function <merge>"
var p12578 = &p12578Var
var p12582Var = "thunk <la> from <function <merge>>"
var p12582 = &p12582Var
var p12591Var = "thunk from <thunk <la> from <function <merge>>>"
var p12591 = &p12591Var
var p12594Var = "thunk <lb> from <function <merge>>"
var p12594 = &p12594Var
var p12603Var = "thunk from <thunk <lb> from <function <merge>>>"
var p12603 = &p12603Var
var p12610Var = "thunk <aux> from <function <merge>>"
var p12610 = &p12610Var
var p12615Var = "function <aux>"
var p12615 = &p12615Var
var p12682Var = "thunk from <function <aux>>"
var p12682 = &p12682Var
var p12694Var = "thunk from <function <aux>>"
var p12694 = &p12694Var
var p12707Var = "thunk from <function <aux>>"
var p12707 = &p12707Var
var p12722Var = "thunk from <thunk from <function <aux>>>"
var p12722 = &p12722Var
var p12736Var = "thunk from <function <aux>>"
var p12736 = &p12736Var
var p12751Var = "thunk from <thunk from <function <aux>>>"
var p12751 = &p12751Var
var p12764Var = "thunk from <function <merge>>"
var p12764 = &p12764Var
var p12772Var = "thunk <l> from <function <anonymous>>"
var p12772 = &p12772Var
var p12781Var = "thunk from <thunk <l> from <function <anonymous>>>"
var p12781 = &p12781Var
var p12797Var = "thunk from <function <anonymous>>"
var p12797 = &p12797Var
var p12807Var = "thunk from <function <anonymous>>"
var p12807 = &p12807Var
var p12818Var = "thunk <mid> from <function <anonymous>>"
var p12818 = &p12818Var
var p12827Var = "thunk from <thunk <mid> from <function <anonymous>>>"
var p12827 = &p12827Var
var p12845Var = "thunk <left> from <function <anonymous>>"
var p12845 = &p12845Var
var p12860Var = "thunk <right> from <function <anonymous>>"
var p12860 = &p12860Var
var p12873Var = "thunk from <function <anonymous>>"
var p12873 = &p12873Var
var p12882Var = "thunk from <thunk from <function <anonymous>>>"
var p12882 = &p12882Var
var p12896Var = "thunk from <thunk from <function <anonymous>>>"
var p12896 = &p12896Var
var p12906Var = "function <anonymous>"
var p12906 = &p12906Var
var p12913Var = "thunk <f> from <function <anonymous>>"
var p12913 = &p12913Var
var p12918Var = "function <f>"
var p12918 = &p12918Var
var p12931Var = "thunk from <function <f>>"
var p12931 = &p12931Var
var p12939Var = "thunk from <function <f>>"
var p12939 = &p12939Var
var p12952Var = "thunk from <function <f>>"
var p12952 = &p12952Var
var p12967Var = "thunk from <thunk from <function <f>>>"
var p12967 = &p12967Var
var p12976Var = "thunk from <function <f>>"
var p12976 = &p12976Var
var p12991Var = "thunk from <function <f>>"
var p12991 = &p12991Var
var p13003Var = "thunk from <function <anonymous>>"
var p13003 = &p13003Var
var p13013Var = "function <anonymous>"
var p13013 = &p13013Var
var p13025Var = "thunk from <function <anonymous>>"
var p13025 = &p13025Var
var p13034Var = "thunk from <thunk from <function <anonymous>>>"
var p13034 = &p13034Var
var p13045Var = "function <anonymous>"
var p13045 = &p13045Var
var p13060Var = "thunk from <function <anonymous>>"
var p13060 = &p13060Var
var p13069Var = "thunk from <thunk from <function <anonymous>>>"
var p13069 = &p13069Var
var p13073Var = "thunk from <thunk from <thunk from <function <anonymous>>>>"
var p13073 = &p13073Var
var p13085Var = "function <anonymous>"
var p13085 = &p13085Var
var p13093Var = "thunk <aux> from <function <anonymous>>"
var p13093 = &p13093Var
var p13098Var = "function <aux>"
var p13098 = &p13098Var
var p13113Var = "thunk from <function <aux>>"
var p13113 = &p13113Var
var p13150Var = "thunk from <function <aux>>"
var p13150 = &p13150Var
var p13177Var = "thunk <ak> from <function <aux>>"
var p13177 = &p13177Var
var p13183Var = "thunk from <thunk <ak> from <function <aux>>>"
var p13183 = &p13183Var
var p13194Var = "thunk <bk> from <function <aux>>"
var p13194 = &p13194Var
var p13200Var = "thunk from <thunk <bk> from <function <aux>>>"
var p13200 = &p13200Var
var p13222Var = "thunk from <function <aux>>"
var p13222 = &p13222Var
var p13244Var = "thunk from <thunk from <function <aux>>>"
var p13244 = &p13244Var
var p13266Var = "thunk from <function <aux>>"
var p13266 = &p13266Var
var p13285Var = "thunk from <thunk from <function <aux>>>"
var p13285 = &p13285Var
var p13299Var = "thunk from <function <aux>>"
var p13299 = &p13299Var
var p13318Var = "thunk from <thunk from <function <aux>>>"
var p13318 = &p13318Var
var p13331Var = "thunk from <function <anonymous>>"
var p13331 = &p13331Var
var p13343Var = "function <anonymous>"
var p13343 = &p13343Var
var p13350Var = "thunk <aux> from <function <anonymous>>"
var p13350 = &p13350Var
var p13355Var = "function <aux>"
var p13355 = &p13355Var
var p13372Var = "thunk from <function <aux>>"
var p13372 = &p13372Var
var p13387Var = "thunk from <function <aux>>"
var p13387 = &p13387Var
var p13404Var = "thunk from <function <aux>>"
var p13404 = &p13404Var
var p13416Var = "thunk from <function <aux>>"
var p13416 = &p13416Var
var p13429Var = "thunk from <function <aux>>"
var p13429 = &p13429Var
var p13451Var = "thunk from <thunk from <function <aux>>>"
var p13451 = &p13451Var
var p13468Var = "thunk from <function <aux>>"
var p13468 = &p13468Var
var p13480Var = "thunk from <function <aux>>"
var p13480 = &p13480Var
var p13493Var = "thunk from <function <aux>>"
var p13493 = &p13493Var
var p13514Var = "thunk from <function <aux>>"
var p13514 = &p13514Var
var p13534Var = "thunk from <function <anonymous>>"
var p13534 = &p13534Var
var p13546Var = "function <anonymous>"
var p13546 = &p13546Var
var p13553Var = "thunk <aux> from <function <anonymous>>"
var p13553 = &p13553Var
var p13558Var = "function <aux>"
var p13558 = &p13558Var
var p13573Var = "thunk from <function <aux>>"
var p13573 = &p13573Var
var p13594Var = "thunk from <function <aux>>"
var p13594 = &p13594Var
var p13627Var = "thunk from <function <aux>>"
var p13627 = &p13627Var
var p13639Var = "thunk from <function <aux>>"
var p13639 = &p13639Var
var p13652Var = "thunk from <function <aux>>"
var p13652 = &p13652Var
var p13679Var = "thunk from <function <aux>>"
var p13679 = &p13679Var
var p13691Var = "thunk from <function <aux>>"
var p13691 = &p13691Var
var p13704Var = "thunk from <function <aux>>"
var p13704 = &p13704Var
var p13723Var = "thunk from <thunk from <function <aux>>>"
var p13723 = &p13723Var
var p13737Var = "thunk from <function <aux>>"
var p13737 = &p13737Var
var p13757Var = "thunk from <function <anonymous>>"
var p13757 = &p13757Var
var p13770Var = "function <anonymous>"
var p13770 = &p13770Var
var p13781Var = "thunk from <function <anonymous>>"
var p13781 = &p13781Var
var p13789Var = "thunk <target_object> from <function <anonymous>>"
var p13789 = &p13789Var
var p13800Var = "thunk from <thunk <target_object> from <function <anonymous>>>"
var p13800 = &p13800Var
var p13811Var = "thunk <target_fields> from <function <anonymous>>"
var p13811 = &p13811Var
var p13822Var = "thunk from <thunk <target_fields> from <function <anonymous>>>"
var p13822 = &p13822Var
var p13833Var = "thunk from <thunk <target_fields> from <function <anonymous>>>"
var p13833 = &p13833Var
var p13854Var = "thunk <null_fields> from <function <anonymous>>"
var p13854 = &p13854Var
var p13867Var = "thunk from <thunk <null_fields> from <function <anonymous>>>"
var p13867 = &p13867Var
var p13879Var = "thunk from <thunk <null_fields> from <function <anonymous>>>"
var p13879 = &p13879Var
var p13886Var = "thunk <both_fields> from <function <anonymous>>"
var p13886 = &p13886Var
var p13895Var = "thunk from <thunk <both_fields> from <function <anonymous>>>"
var p13895 = &p13895Var
var p13906Var = "thunk from <thunk from <thunk <both_fields> from <function <anonymous>>>>"
var p13906 = &p13906Var
var p13938Var = "object <anonymous>"
var p13938 = &p13938Var
var p13951Var = "thunk from <object <anonymous>>"
var p13951 = &p13951Var
var p13976Var = "thunk from <object <anonymous>>"
var p13976 = &p13976Var
var p13990Var = "thunk from <object <anonymous>>"
var p13990 = &p13990Var
var p14008Var = "thunk from <object <anonymous>>"
var p14008 = &p14008Var
var p14029Var = "thunk from <function <anonymous>>"
var p14029 = &p14029Var
var p14042Var = "function <anonymous>"
var p14042 = &p14042Var
var p14052Var = "thunk from <function <anonymous>>"
var p14052 = &p14052Var
var p14060Var = "function <anonymous>"
var p14060 = &p14060Var
var p14070Var = "thunk from <function <anonymous>>"
var p14070 = &p14070Var
var p14078Var = "function <anonymous>"
var p14078 = &p14078Var
var p14088Var = "thunk from <function <anonymous>>"
var p14088 = &p14088Var
var p14098Var = "function <anonymous>"
var p14098 = &p14098Var
var p14108Var = "thunk from <function <anonymous>>"
var p14108 = &p14108Var
var p14119Var = "function <anonymous>"
var p14119 = &p14119Var
var p14123Var = "thunk <ta> from <function <anonymous>>"
var p14123 = &p14123Var
var p14132Var = "thunk from <thunk <ta> from <function <anonymous>>>"
var p14132 = &p14132Var
var p14139Var = "thunk <tb> from <function <anonymous>>"
var p14139 = &p14139Var
var p14148Var = "thunk from <thunk <tb> from <function <anonymous>>>"
var p14148 = &p14148Var
var p14164Var = "thunk from <function <anonymous>>"
var p14164 = &p14164Var
var p14183Var = "thunk from <function <anonymous>>"
var p14183 = &p14183Var
var p14191Var = "thunk <la> from <function <anonymous>>"
var p14191 = &p14191Var
var p14200Var = "thunk from <thunk <la> from <function <anonymous>>>"
var p14200 = &p14200Var
var p14216Var = "thunk from <function <anonymous>>"
var p14216 = &p14216Var
var p14227Var = "thunk from <thunk from <function <anonymous>>>"
var p14227 = &p14227Var
var p14237Var = "thunk <aux> from <function <anonymous>>"
var p14237 = &p14237Var
var p14242Var = "function <aux>"
var p14242 = &p14242Var
var p14279Var = "thunk from <function <aux>>"
var p14279 = &p14279Var
var p14295Var = "thunk from <function <anonymous>>"
var p14295 = &p14295Var
var p14312Var = "thunk from <function <anonymous>>"
var p14312 = &p14312Var
var p14320Var = "thunk <fields> from <function <anonymous>>"
var p14320 = &p14320Var
var p14329Var = "thunk from <thunk <fields> from <function <anonymous>>>"
var p14329 = &p14329Var
var p14336Var = "thunk <lfields> from <function <anonymous>>"
var p14336 = &p14336Var
var p14345Var = "thunk from <thunk <lfields> from <function <anonymous>>>"
var p14345 = &p14345Var
var p14363Var = "thunk from <function <anonymous>>"
var p14363 = &p14363Var
var p14373Var = "thunk <aux> from <function <anonymous>>"
var p14373 = &p14373Var
var p14378Var = "function <aux>"
var p14378 = &p14378Var
var p14395Var = "thunk <f> from <function <aux>>"
var p14395 = &p14395Var
var p14425Var = "thunk from <function <aux>>"
var p14425 = &p14425Var
var p14441Var = "thunk from <function <anonymous>>"
var p14441 = &p14441Var
var p14457Var = "thunk from <function <anonymous>>"
var p14457 = &p14457Var
var p14467Var = "function <anonymous>"
var p14467 = &p14467Var
var p14471Var = "thunk <arr> from <function <anonymous>>"
var p14471 = &p14471Var
var p14480Var = "thunk from <thunk <arr> from <function <anonymous>>>"
var p14480 = &p14480Var
var p14493Var = "thunk from <function <anonymous>>"
var p14493 = &p14493Var
var p14505Var = "thunk from <thunk from <function <anonymous>>>"
var p14505 = &p14505Var
var p14516Var = "thunk from <thunk from <thunk from <function <anonymous>>>>"
var p14516 = &p14516Var
var p14523Var = "function <anonymous>"
var p14523 = &p14523Var
var p14533Var = "thunk from <thunk from <function <anonymous>>>"
var p14533 = &p14533Var
var p14541Var = "function <anonymous>"
var p14541 = &p14541Var
var p14545Var = "thunk <isContent> from <function <anonymous>>"
var p14545 = &p14545Var
var p14550Var = "function <isContent>"
var p14550 = &p14550Var
var p14571Var = "thunk from <function <isContent>>"
var p14571 = &p14571Var
var p14585Var = "thunk from <function <isContent>>"
var p14585 = &p14585Var
var p14600Var = "thunk from <function <isContent>>"
var p14600 = &p14600Var
var p14614Var = "thunk from <function <isContent>>"
var p14614 = &p14614Var
var p14632Var = "thunk from <function <anonymous>>"
var p14632 = &p14632Var
var p14653Var = "thunk from <function <anonymous>>"
var p14653 = &p14653Var
var p14662Var = "thunk from <thunk from <function <anonymous>>>"
var p14662 = &p14662Var
var p14668Var = "thunk from <function <anonymous>>"
var p14668 = &p14668Var
var p14677Var = "thunk from <thunk from <function <anonymous>>>"
var p14677 = &p14677Var
var p14694Var = "thunk from <function <anonymous>>"
var p14694 = &p14694Var
var p14723Var = "thunk from <function <anonymous>>"
var p14723 = &p14723Var
var p14732Var = "thunk from <thunk from <function <anonymous>>>"
var p14732 = &p14732Var
var p14747Var = "object <anonymous>"
var p14747 = &p14747Var
var p14756Var = "thunk from <object <anonymous>>"
var p14756 = &p14756Var
var p14772Var = "thunk from <function <anonymous>>"
var p14772 = &p14772Var
var p14783Var = "function <anonymous>"
var p14783 = &p14783Var
var p14796Var = "thunk from <function <anonymous>>"
var p14796 = &p14796Var
var p14813Var = "thunk from <function <anonymous>>"
var p14813 = &p14813Var
var p14829Var = "thunk from <function <anonymous>>"
var p14829 = &p14829Var
var p14846Var = "thunk from <function <anonymous>>"
var p14846 = &p14846Var
var p14854Var = "thunk <pat_len> from <function <anonymous>>"
var p14854 = &p14854Var
var p14863Var = "thunk from <thunk <pat_len> from <function <anonymous>>>"
var p14863 = &p14863Var
var p14870Var = "thunk <str_len> from <function <anonymous>>"
var p14870 = &p14870Var
var p14879Var = "thunk from <thunk <str_len> from <function <anonymous>>>"
var p14879 = &p14879Var
var p14917Var = "thunk from <function <anonymous>>"
var p14917 = &p14917Var
var p14921Var = "function <anonymous>"
var p14921 = &p14921Var
var p14953Var = "thunk from <thunk from <function <anonymous>>>"
var p14953 = &p14953Var
var p14966Var = "function <anonymous>"
var p14966 = &p14966Var
var p14979Var = "thunk from <function <anonymous>>"
var p14979 = &p14979Var
var p14996Var = "thunk from <function <anonymous>>"
var p14996 = &p14996Var
var p15009Var = "thunk from <function <anonymous>>"
var p15009 = &p15009Var
var p15013Var = "function <anonymous>"
var p15013 = &p15013Var
var p15032Var = "thunk from <thunk from <function <anonymous>>>"
var p15032 = &p15032Var
var p15044Var = "thunk from <thunk from <thunk from <function <anonymous>>>>"
var p15044 = &p15044Var
var p15049Var = "object <anonymous>"
var p15049 = &p15049Var
var p15051Var = "object <anonymous>"
var p15051 = &p15051Var
var p15054Var = "function <anonymous>"
var p15054 = &p15054Var
var p15057Var = "object <anonymous>"
var p15057 = &p15057Var
var p15063Var = "function <anonymous>"
var p15063 = &p15063Var
var p15066Var = "function <anonymous>"
var p15066 = &p15066Var
var p15069Var = "function <anonymous>"
var p15069 = &p15069Var
var p15072Var = "function <anonymous>"
var p15072 = &p15072Var
var p15074Var = "function <anonymous>"
var p15074 = &p15074Var
var p15077Var = "function <anonymous>"
var p15077 = &p15077Var
var p15080Var = "function <anonymous>"
var p15080 = &p15080Var
var p15084Var = "function <anonymous>"
var p15084 = &p15084Var
var p15088Var = "thunk <zero_code> from <function <anonymous>>"
var p15088 = &p15088Var
var p15091Var = "thunk <zero_code> from <function <anonymous>>"
var p15091 = &p15091Var
var p15098Var = "thunk from <thunk <zero_code> from <function <anonymous>>>"
var p15098 = &p15098Var
var p15101Var = "function <anonymous>"
var p15101 = &p15101Var
var p15105Var = "thunk <upper_a_code> from <function <anonymous>>"
var p15105 = &p15105Var
var p15108Var = "thunk <upper_a_code> from <function <anonymous>>"
var p15108 = &p15108Var
var p15115Var = "thunk from <thunk <upper_a_code> from <function <anonymous>>>"
var p15115 = &p15115Var
var p15118Var = "function <anonymous>"
var p15118 = &p15118Var
var p15122Var = "thunk <lower_a_code> from <function <anonymous>>"
var p15122 = &p15122Var
var p15125Var = "thunk <lower_a_code> from <function <anonymous>>"
var p15125 = &p15125Var
var p15132Var = "thunk from <thunk <lower_a_code> from <function <anonymous>>>"
var p15132 = &p15132Var
var p15135Var = "function <anonymous>"
var p15135 = &p15135Var
var p15139Var = "thunk <addDigit> from <function <anonymous>>"
var p15139 = &p15139Var
var p15144Var = "function <addDigit>"
var p15144 = &p15144Var
var p15148Var = "thunk <code> from <function <addDigit>>"
var p15148 = &p15148Var
var p15151Var = "thunk <code> from <function <addDigit>>"
var p15151 = &p15151Var
var p15158Var = "thunk from <thunk <code> from <function <addDigit>>>"
var p15158 = &p15158Var
var p15162Var = "function <addDigit>"
var p15162 = &p15162Var
var p15166Var = "thunk <digit> from <function <addDigit>>"
var p15166 = &p15166Var
var p15169Var = "thunk <digit> from <function <addDigit>>"
var p15169 = &p15169Var
var p15172Var = "thunk <digit> from <function <addDigit>>"
var p15172 = &p15172Var
var p15175Var = "thunk <digit> from <function <addDigit>>"
var p15175 = &p15175Var
var p15178Var = "thunk <digit> from <function <addDigit>>"
var p15178 = &p15178Var
var p15181Var = "thunk <digit> from <function <addDigit>>"
var p15181 = &p15181Var
var p15185Var = "thunk <digit> from <function <addDigit>>"
var p15185 = &p15185Var
var p15188Var = "thunk <digit> from <function <addDigit>>"
var p15188 = &p15188Var
var p15191Var = "thunk <digit> from <function <addDigit>>"
var p15191 = &p15191Var
var p15194Var = "thunk <digit> from <function <addDigit>>"
var p15194 = &p15194Var
var p15197Var = "thunk <digit> from <function <addDigit>>"
var p15197 = &p15197Var
var p15200Var = "thunk <digit> from <function <addDigit>>"
var p15200 = &p15200Var
var p15203Var = "thunk <digit> from <function <addDigit>>"
var p15203 = &p15203Var
var p15206Var = "thunk <digit> from <function <addDigit>>"
var p15206 = &p15206Var
var p15209Var = "thunk <digit> from <function <addDigit>>"
var p15209 = &p15209Var
var p15213Var = "thunk <digit> from <function <addDigit>>"
var p15213 = &p15213Var
var p15216Var = "thunk <digit> from <function <addDigit>>"
var p15216 = &p15216Var
var p15219Var = "thunk <digit> from <function <addDigit>>"
var p15219 = &p15219Var
var p15222Var = "thunk <digit> from <function <addDigit>>"
var p15222 = &p15222Var
var p15226Var = "thunk <digit> from <function <addDigit>>"
var p15226 = &p15226Var
var p15229Var = "thunk <digit> from <function <addDigit>>"
var p15229 = &p15229Var
var p15234Var = "function <addDigit>"
var p15234 = &p15234Var
var p15237Var = "function <addDigit>"
var p15237 = &p15237Var
var p15240Var = "function <addDigit>"
var p15240 = &p15240Var
var p15243Var = "function <addDigit>"
var p15243 = &p15243Var
var p15245Var = "function <addDigit>"
var p15245 = &p15245Var
var p15248Var = "function <addDigit>"
var p15248 = &p15248Var
var p15251Var = "function <addDigit>"
var p15251 = &p15251Var
var p15254Var = "function <addDigit>"
var p15254 = &p15254Var
var p15257Var = "function <addDigit>"
var p15257 = &p15257Var
var p15261Var = "function <addDigit>"
var p15261 = &p15261Var
var p15264Var = "function <addDigit>"
var p15264 = &p15264Var
var p15267Var = "function <addDigit>"
var p15267 = &p15267Var
var p15280Var = "function <addDigit>"
var p15280 = &p15280Var
var p15282Var = "function <addDigit>"
var p15282 = &p15282Var
var p15286Var = "thunk from <function <addDigit>>"
var p15286 = &p15286Var
var p15289Var = "thunk from <function <addDigit>>"
var p15289 = &p15289Var
var p15292Var = "function <anonymous>"
var p15292 = &p15292Var
var p15295Var = "function <anonymous>"
var p15295 = &p15295Var
var p15303Var = "thunk from <function <anonymous>>"
var p15303 = &p15303Var
var p15306Var = "thunk from <function <anonymous>>"
var p15306 = &p15306Var
var p15309Var = "thunk from <function <anonymous>>"
var p15309 = &p15309Var
var p15316Var = "thunk from <thunk from <function <anonymous>>>"
var p15316 = &p15316Var
var p15319Var = "thunk from <function <anonymous>>"
var p15319 = &p15319Var
var p15331Var = "function <anonymous>"
var p15331 = &p15331Var
var p15333Var = "function <anonymous>"
var p15333 = &p15333Var
var p15336Var = "object <anonymous>"
var p15336 = &p15336Var
var p15360Var = "object <anonymous>"
var p15360 = &p15360Var
var p15364Var = "object <anonymous>"
var p15364 = &p15364Var
var p15367Var = "object <anonymous>"
var p15367 = &p15367Var
var p15370Var = "object <anonymous>"
var p15370 = &p15370Var
var p15373Var = "object <anonymous>"
var p15373 = &p15373Var
var p15376Var = "object <anonymous>"
var p15376 = &p15376Var
var p15379Var = "object <anonymous>"
var p15379 = &p15379Var
var p15386Var = "thunk from <object <anonymous>>"
var p15386 = &p15386Var
var p15388Var = "thunk from <object <anonymous>>"
var p15388 = &p15388Var
var p1 = &ast.Source{
	Lines: []string{
		"/*\n",
		"Copyright 2015 Google Inc. All rights reserved.\n",
		"\n",
		"Licensed under the Apache License, Version 2.0 (the \"License\");\n",
		"you may not use this file except in compliance with the License.\n",
		"You may obtain a copy of the License at\n",
		"\n",
		"    http://www.apache.org/licenses/LICENSE-2.0\n",
		"\n",
		"Unless required by applicable law or agreed to in writing, software\n",
		"distributed under the License is distributed on an \"AS IS\" BASIS,\n",
		"WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\n",
		"See the License for the specific language governing permissions and\n",
		"limitations under the License.\n",
		"*/\n",
		"\n",
		"/* This is the Jsonnet standard library, at least the parts of it that are written in Jsonnet.\n",
		" *\n",
		" * There are some native methods as well, which are defined in the interpreter and added to this\n",
		" * file.  It is never necessary to import std.jsonnet, it is embedded into the interpreter at\n",
		" * compile-time and automatically imported into all other Jsonnet programs.\n",
		" */\n",
		"{\n",
		"\n",
		"  local std = self,\n",
		"  local id = function(x) x,\n",
		"\n",
		"  isString(v):: std.type(v) == 'string',\n",
		"  isNumber(v):: std.type(v) == 'number',\n",
		"  isBoolean(v):: std.type(v) == 'boolean',\n",
		"  isObject(v):: std.type(v) == 'object',\n",
		"  isArray(v):: std.type(v) == 'array',\n",
		"  isFunction(v):: std.type(v) == 'function',\n",
		"\n",
		"  toString(a)::\n",
		"    if std.type(a) == 'string' then a else '' + a,\n",
		"\n",
		"  substr(str, from, len)::\n",
		"    assert std.isString(str) : 'substr first parameter should be a string, got ' + std.type(str);\n",
		"    assert std.isNumber(from) : 'substr second parameter should be a string, got ' + std.type(from);\n",
		"    assert std.isNumber(len) : 'substr third parameter should be a string, got ' + std.type(len);\n",
		"    assert len >=0 : 'substr third parameter should be greater than zero, got ' + len;\n",
		"    std.join('', std.makeArray(std.max(0, std.min(len, std.length(str) - from)), function(i) str[i + from])),\n",
		"\n",
		"  startsWith(a, b)::\n",
		"    if std.length(a) < std.length(b) then\n",
		"      false\n",
		"    else\n",
		"      std.substr(a, 0, std.length(b)) == b,\n",
		"\n",
		"  endsWith(a, b)::\n",
		"    if std.length(a) < std.length(b) then\n",
		"      false\n",
		"    else\n",
		"      std.substr(a, std.length(a) - std.length(b), std.length(b)) == b,\n",
		"\n",
		"  lstripChars(str, chars)::\n",
		"    if std.length(str) > 0 && std.member(chars, str[0]) then\n",
		"      std.lstripChars(str[1:], chars)\n",
		"    else\n",
		"      str,\n",
		"\n",
		"  rstripChars(str, chars)::\n",
		"    local len = std.length(str);\n",
		"    if len > 0 && std.member(chars, str[len - 1]) then\n",
		"      std.rstripChars(str[:len - 1], chars)\n",
		"    else\n",
		"      str,\n",
		"\n",
		"  stripChars(str, chars)::\n",
		"    std.lstripChars(std.rstripChars(str, chars), chars),\n",
		"\n",
		"  stringChars(str)::\n",
		"    std.makeArray(std.length(str), function(i) str[i]),\n",
		"\n",
		"  local parse_nat(str, base) =\n",
		"    assert base > 0 && base <= 16 : 'integer base %d invalid' % base;\n",
		"    // These codepoints are in ascending order:\n",
		"    local zero_code = std.codepoint('0');\n",
		"    local upper_a_code = std.codepoint('A');\n",
		"    local lower_a_code = std.codepoint('a');\n",
		"    local addDigit(aggregate, char) =\n",
		"      local code = std.codepoint(char);\n",
		"      local digit = if code >= lower_a_code then\n",
		"        code - lower_a_code + 10\n",
		"      else if code >= upper_a_code then\n",
		"        code - upper_a_code + 10\n",
		"      else\n",
		"        code - zero_code;\n",
		"      assert digit >= 0 && digit < base : '%s is not a base %d integer' % [str, base];\n",
		"      base * aggregate + digit;\n",
		"    std.foldl(addDigit, std.stringChars(str), 0),\n",
		"\n",
		"  parseInt(str)::\n",
		"    assert std.isString(str) : 'Expected string, got ' + std.type(str);\n",
		"    assert std.length(str) > 0 && str != '-' : 'Not an integer: \"%s\"' % [str];\n",
		"    if str[0] == '-' then\n",
		"      -parse_nat(str[1:], 10)\n",
		"    else\n",
		"      parse_nat(str, 10),\n",
		"\n",
		"  parseOctal(str)::\n",
		"    assert std.isString(str) : 'Expected string, got ' + std.type(str);\n",
		"    assert std.length(str) > 0 : 'Not an octal number: \"\"';\n",
		"    parse_nat(str, 8),\n",
		"\n",
		"  parseHex(str)::\n",
		"    assert std.isString(str) : 'Expected string, got ' + std.type(str);\n",
		"    assert std.length(str) > 0 : 'Not hexadecimal: \"\"';\n",
		"    parse_nat(str, 16),\n",
		"\n",
		"  split(str, c)::\n",
		"    assert std.isString(str) : 'std.split first parameter should be a string, got ' + std.type(str);\n",
		"    assert std.isString(c) : 'std.split second parameter should be a string, got ' + std.type(c);\n",
		"    assert std.length(c) == 1 :  'std.split second parameter should have length 1, got ' + std.length(c);\n",
		"    std.splitLimit(str, c, -1),\n",
		"\n",
		"  splitLimit(str, c, maxsplits)::\n",
		"    assert std.isString(str) : 'std.splitLimit first parameter should be a string, got ' + std.type(str);\n",
		"    assert std.isString(c) : 'std.splitLimit second parameter should be a string, got ' + std.type(c);\n",
		"    assert std.length(c) == 1 :  'std.splitLimit second parameter should have length 1, got ' + std.length(c);\n",
		"    assert std.isNumber(maxsplits) : 'std.splitLimit third parameter should be a number, got ' + std.type(maxsplits);\n",
		"    local aux(str, delim, i, arr, v) =\n",
		"      local c = str[i];\n",
		"      local i2 = i + 1;\n",
		"      if i >= std.length(str) then\n",
		"        arr + [v]\n",
		"      else if c == delim && (maxsplits == -1 || std.length(arr) < maxsplits) then\n",
		"        aux(str, delim, i2, arr + [v], '') tailstrict\n",
		"      else\n",
		"        aux(str, delim, i2, arr, v + c) tailstrict;\n",
		"    aux(str, c, 0, [], ''),\n",
		"\n",
		"  strReplace(str, from, to)::\n",
		"    assert std.isString(str);\n",
		"    assert std.isString(from);\n",
		"    assert std.isString(to);\n",
		"    assert from != '' : \"'from' string must not be zero length.\";\n",
		"\n",
		"    // Cache for performance.\n",
		"    local str_len = std.length(str);\n",
		"    local from_len = std.length(from);\n",
		"\n",
		"    // True if from is at str[i].\n",
		"    local found_at(i) = str[i:i + from_len] == from;\n",
		"\n",
		"    // Return the remainder of 'str' starting with 'start_index' where\n",
		"    // all occurrences of 'from' after 'curr_index' are replaced with 'to'.\n",
		"    local replace_after(start_index, curr_index, acc) =\n",
		"      if curr_index > str_len then\n",
		"        acc + str[start_index:curr_index]\n",
		"      else if found_at(curr_index) then\n",
		"        local new_index = curr_index + std.length(from);\n",
		"        replace_after(new_index, new_index, acc + str[start_index:curr_index] + to) tailstrict\n",
		"      else\n",
		"        replace_after(start_index, curr_index + 1, acc) tailstrict;\n",
		"\n",
		"    // if from_len==1, then we replace by splitting and rejoining the\n",
		"    // string which is much faster than recursing on replace_after\n",
		"    if from_len == 1 then\n",
		"      std.join(to, std.split(str, from))\n",
		"    else\n",
		"      replace_after(0, 0, ''),\n",
		"\n",
		"  asciiUpper(str)::\n",
		"    local cp = std.codepoint;\n",
		"    local up_letter(c) = if cp(c) >= 97 && cp(c) < 123 then\n",
		"      std.char(cp(c) - 32)\n",
		"    else\n",
		"      c;\n",
		"    std.join('', std.map(up_letter, std.stringChars(str))),\n",
		"\n",
		"  asciiLower(str)::\n",
		"    local cp = std.codepoint;\n",
		"    local down_letter(c) = if cp(c) >= 65 && cp(c) < 91 then\n",
		"      std.char(cp(c) + 32)\n",
		"    else\n",
		"      c;\n",
		"    std.join('', std.map(down_letter, std.stringChars(str))),\n",
		"\n",
		"  range(from, to)::\n",
		"    std.makeArray(to - from + 1, function(i) i + from),\n",
		"\n",
		"  repeat(what, count)::\n",
		"    local joiner =\n",
		"      if std.isString(what) then \"\"\n",
		"      else if std.isArray(what) then []\n",
		"      else error \"std.repeat first argument must be an array or a string\";\n",
		"    std.join(joiner, std.makeArray(count, function(i) what)),\n",
		"\n",
		"  slice(indexable, index, end, step)::\n",
		"    local invar =\n",
		"      // loop invariant with defaults applied\n",
		"      {\n",
		"        indexable: indexable,\n",
		"        index:\n",
		"          if index == null then 0\n",
		"          else index,\n",
		"        end:\n",
		"          if end == null then std.length(indexable)\n",
		"          else end,\n",
		"        step:\n",
		"          if step == null then 1\n",
		"          else step,\n",
		"        length: std.length(indexable),\n",
		"        type: std.type(indexable),\n",
		"      };\n",
		"    assert invar.index >= 0 && invar.end >= 0 && invar.step >= 0 : 'got [%s:%s:%s] but negative index, end, and steps are not supported' % [invar.index, invar.end, invar.step];\n",
		"    assert step != 0 : 'got %s but step must be greater than 0' % step;\n",
		"    assert std.isString(indexable) || std.isArray(indexable) : 'std.slice accepts a string or an array, but got: %s' % std.type(indexable);\n",
		"    local build(slice, cur) =\n",
		"      if cur >= invar.end || cur >= invar.length then\n",
		"        slice\n",
		"      else\n",
		"        build(\n",
		"          if invar.type == 'string' then\n",
		"            slice + invar.indexable[cur]\n",
		"          else\n",
		"            slice + [invar.indexable[cur]],\n",
		"          cur + invar.step\n",
		"        ) tailstrict;\n",
		"    build(if invar.type == 'string' then '' else [], invar.index),\n",
		"\n",
		"  member(arr, x)::\n",
		"    if std.isArray(arr) then\n",
		"      std.count(arr, x) > 0\n",
		"    else if std.isString(arr) then\n",
		"      std.length(std.findSubstr(x, arr)) > 0\n",
		"    else error \"std.member first argument must be an array or a string\",\n",
		"\n",
		"  count(arr, x):: std.length(std.filter(function(v) v == x, arr)),\n",
		"\n",
		"  mod(a, b)::\n",
		"    if std.isNumber(a) && std.isNumber(b) then\n",
		"      std.modulo(a, b)\n",
		"    else if std.isString(a) then\n",
		"      std.format(a, b)\n",
		"    else\n",
		"      error 'Operator % cannot be used on types ' + std.type(a) + ' and ' + std.type(b) + '.',\n",
		"\n",
		"  map(func, arr)::\n",
		"    if !std.isFunction(func) then\n",
		"      error ('std.map first param must be function, got ' + std.type(func))\n",
		"    else if !std.isArray(arr) && !std.isString(arr) then\n",
		"      error ('std.map second param must be array / string, got ' + std.type(arr))\n",
		"    else\n",
		"      std.makeArray(std.length(arr), function(i) func(arr[i])),\n",
		"\n",
		"  mapWithIndex(func, arr)::\n",
		"    if !std.isFunction(func) then\n",
		"      error ('std.mapWithIndex first param must be function, got ' + std.type(func))\n",
		"    else if !std.isArray(arr) && !std.isString(arr) then\n",
		"      error ('std.mapWithIndex second param must be array, got ' + std.type(arr))\n",
		"    else\n",
		"      std.makeArray(std.length(arr), function(i) func(i, arr[i])),\n",
		"\n",
		"  mapWithKey(func, obj)::\n",
		"    if !std.isFunction(func) then\n",
		"      error ('std.mapWithKey first param must be function, got ' + std.type(func))\n",
		"    else if !std.isObject(obj) then\n",
		"      error ('std.mapWithKey second param must be object, got ' + std.type(obj))\n",
		"    else\n",
		"      { [k]: func(k, obj[k]) for k in std.objectFields(obj) },\n",
		"\n",
		"  flatMap(func, arr)::\n",
		"    if !std.isFunction(func) then\n",
		"      error ('std.flatMap first param must be function, got ' + std.type(func))\n",
		"    else if std.isArray(arr) then\n",
		"      std.flattenArrays(std.makeArray(std.length(arr), function(i) func(arr[i])))\n",
		"    else if std.isString(arr) then\n",
		"      std.join('', std.makeArray(std.length(arr), function(i) func(arr[i])))\n",
		"    else error ('std.flatMap second param must be array / string, got ' + std.type(arr)),\n",
		"\n",
		"  join(sep, arr)::\n",
		"    local aux(arr, i, first, running) =\n",
		"      if i >= std.length(arr) then\n",
		"        running\n",
		"      else if arr[i] == null then\n",
		"        aux(arr, i + 1, first, running) tailstrict\n",
		"      else if std.type(arr[i]) != std.type(sep) then\n",
		"        error 'expected %s but arr[%d] was %s ' % [std.type(sep), i, std.type(arr[i])]\n",
		"      else if first then\n",
		"        aux(arr, i + 1, false, running + arr[i]) tailstrict\n",
		"      else\n",
		"        aux(arr, i + 1, false, running + sep + arr[i]) tailstrict;\n",
		"    if !std.isArray(arr) then\n",
		"      error 'join second parameter should be array, got ' + std.type(arr)\n",
		"    else if std.isString(sep) then\n",
		"      aux(arr, 0, true, '')\n",
		"    else if std.isArray(sep) then\n",
		"      aux(arr, 0, true, [])\n",
		"    else\n",
		"      error 'join first parameter should be string or array, got ' + std.type(sep),\n",
		"\n",
		"  lines(arr)::\n",
		"    std.join('\\n', arr + ['']),\n",
		"\n",
		"  deepJoin(arr)::\n",
		"    if std.isString(arr) then\n",
		"      arr\n",
		"    else if std.isArray(arr) then\n",
		"      std.join('', [std.deepJoin(x) for x in arr])\n",
		"    else\n",
		"      error 'Expected string or array, got %s' % std.type(arr),\n",
		"\n",
		"\n",
		"  format(str, vals)::\n",
		"\n",
		"    /////////////////////////////\n",
		"    // Parse the mini-language //\n",
		"    /////////////////////////////\n",
		"\n",
		"    local try_parse_mapping_key(str, i) =\n",
		"      assert i < std.length(str) : 'Truncated format code.';\n",
		"      local c = str[i];\n",
		"      if c == '(' then\n",
		"        local consume(str, j, v) =\n",
		"          if j >= std.length(str) then\n",
		"            error 'Truncated format code.'\n",
		"          else\n",
		"            local c = str[j];\n",
		"            if c != ')' then\n",
		"              consume(str, j + 1, v + c)\n",
		"            else\n",
		"              { i: j + 1, v: v };\n",
		"        consume(str, i + 1, '')\n",
		"      else\n",
		"        { i: i, v: null };\n",
		"\n",
		"    local try_parse_cflags(str, i) =\n",
		"      local consume(str, j, v) =\n",
		"        assert j < std.length(str) : 'Truncated format code.';\n",
		"        local c = str[j];\n",
		"        if c == '#' then\n",
		"          consume(str, j + 1, v { alt: true })\n",
		"        else if c == '0' then\n",
		"          consume(str, j + 1, v { zero: true })\n",
		"        else if c == '-' then\n",
		"          consume(str, j + 1, v { left: true })\n",
		"        else if c == ' ' then\n",
		"          consume(str, j + 1, v { blank: true })\n",
		"        else if c == '+' then\n",
		"          consume(str, j + 1, v { plus: true })\n",
		"        else\n",
		"          { i: j, v: v };\n",
		"      consume(str, i, { alt: false, zero: false, left: false, blank: false, plus: false });\n",
		"\n",
		"    local try_parse_field_width(str, i) =\n",
		"      if i < std.length(str) && str[i] == '*' then\n",
		"        { i: i + 1, v: '*' }\n",
		"      else\n",
		"        local consume(str, j, v) =\n",
		"          assert j < std.length(str) : 'Truncated format code.';\n",
		"          local c = str[j];\n",
		"          if c == '0' then\n",
		"            consume(str, j + 1, v * 10 + 0)\n",
		"          else if c == '1' then\n",
		"            consume(str, j + 1, v * 10 + 1)\n",
		"          else if c == '2' then\n",
		"            consume(str, j + 1, v * 10 + 2)\n",
		"          else if c == '3' then\n",
		"            consume(str, j + 1, v * 10 + 3)\n",
		"          else if c == '4' then\n",
		"            consume(str, j + 1, v * 10 + 4)\n",
		"          else if c == '5' then\n",
		"            consume(str, j + 1, v * 10 + 5)\n",
		"          else if c == '6' then\n",
		"            consume(str, j + 1, v * 10 + 6)\n",
		"          else if c == '7' then\n",
		"            consume(str, j + 1, v * 10 + 7)\n",
		"          else if c == '8' then\n",
		"            consume(str, j + 1, v * 10 + 8)\n",
		"          else if c == '9' then\n",
		"            consume(str, j + 1, v * 10 + 9)\n",
		"          else\n",
		"            { i: j, v: v };\n",
		"        consume(str, i, 0);\n",
		"\n",
		"    local try_parse_precision(str, i) =\n",
		"      assert i < std.length(str) : 'Truncated format code.';\n",
		"      local c = str[i];\n",
		"      if c == '.' then\n",
		"        try_parse_field_width(str, i + 1)\n",
		"      else\n",
		"        { i: i, v: null };\n",
		"\n",
		"    // Ignored, if it exists.\n",
		"    local try_parse_length_modifier(str, i) =\n",
		"      assert i < std.length(str) : 'Truncated format code.';\n",
		"      local c = str[i];\n",
		"      if c == 'h' || c == 'l' || c == 'L' then\n",
		"        i + 1\n",
		"      else\n",
		"        i;\n",
		"\n",
		"    local parse_conv_type(str, i) =\n",
		"      assert i < std.length(str) : 'Truncated format code.';\n",
		"      local c = str[i];\n",
		"      if c == 'd' || c == 'i' || c == 'u' then\n",
		"        { i: i + 1, v: 'd', caps: false }\n",
		"      else if c == 'o' then\n",
		"        { i: i + 1, v: 'o', caps: false }\n",
		"      else if c == 'x' then\n",
		"        { i: i + 1, v: 'x', caps: false }\n",
		"      else if c == 'X' then\n",
		"        { i: i + 1, v: 'x', caps: true }\n",
		"      else if c == 'e' then\n",
		"        { i: i + 1, v: 'e', caps: false }\n",
		"      else if c == 'E' then\n",
		"        { i: i + 1, v: 'e', caps: true }\n",
		"      else if c == 'f' then\n",
		"        { i: i + 1, v: 'f', caps: false }\n",
		"      else if c == 'F' then\n",
		"        { i: i + 1, v: 'f', caps: true }\n",
		"      else if c == 'g' then\n",
		"        { i: i + 1, v: 'g', caps: false }\n",
		"      else if c == 'G' then\n",
		"        { i: i + 1, v: 'g', caps: true }\n",
		"      else if c == 'c' then\n",
		"        { i: i + 1, v: 'c', caps: false }\n",
		"      else if c == 's' then\n",
		"        { i: i + 1, v: 's', caps: false }\n",
		"      else if c == '%' then\n",
		"        { i: i + 1, v: '%', caps: false }\n",
		"      else\n",
		"        error 'Unrecognised conversion type: ' + c;\n",
		"\n",
		"\n",
		"    // Parsed initial %, now the rest.\n",
		"    local parse_code(str, i) =\n",
		"      assert i < std.length(str) : 'Truncated format code.';\n",
		"      local mkey = try_parse_mapping_key(str, i);\n",
		"      local cflags = try_parse_cflags(str, mkey.i);\n",
		"      local fw = try_parse_field_width(str, cflags.i);\n",
		"      local prec = try_parse_precision(str, fw.i);\n",
		"      local len_mod = try_parse_length_modifier(str, prec.i);\n",
		"      local ctype = parse_conv_type(str, len_mod);\n",
		"      {\n",
		"        i: ctype.i,\n",
		"        code: {\n",
		"          mkey: mkey.v,\n",
		"          cflags: cflags.v,\n",
		"          fw: fw.v,\n",
		"          prec: prec.v,\n",
		"          ctype: ctype.v,\n",
		"          caps: ctype.caps,\n",
		"        },\n",
		"      };\n",
		"\n",
		"    // Parse a format string (containing none or more % format tags).\n",
		"    local parse_codes(str, i, out, cur) =\n",
		"      if i >= std.length(str) then\n",
		"        out + [cur]\n",
		"      else\n",
		"        local c = str[i];\n",
		"        if c == '%' then\n",
		"          local r = parse_code(str, i + 1);\n",
		"          parse_codes(str, r.i, out + [cur, r.code], '') tailstrict\n",
		"        else\n",
		"          parse_codes(str, i + 1, out, cur + c) tailstrict;\n",
		"\n",
		"    local codes = parse_codes(str, 0, [], '');\n",
		"\n",
		"\n",
		"    ///////////////////////\n",
		"    // Format the values //\n",
		"    ///////////////////////\n",
		"\n",
		"    // Useful utilities\n",
		"    local padding(w, s) =\n",
		"      local aux(w, v) =\n",
		"        if w <= 0 then\n",
		"          v\n",
		"        else\n",
		"          aux(w - 1, v + s);\n",
		"      aux(w, '');\n",
		"\n",
		"    // Add s to the left of str so that its length is at least w.\n",
		"    local pad_left(str, w, s) =\n",
		"      padding(w - std.length(str), s) + str;\n",
		"\n",
		"    // Add s to the right of str so that its length is at least w.\n",
		"    local pad_right(str, w, s) =\n",
		"      str + padding(w - std.length(str), s);\n",
		"\n",
		"    // Render a sign & magnitude integer (radix ranges from decimal to binary).\n",
		"    // neg should be a boolean, and when true indicates that we should render a negative number.\n",
		"    // mag must always be a whole number >= 0, it's the magnitude of the integer to render\n",
		"    // min_chars must be a whole number >= 0\n",
		"    //   It is the field width, i.e. std.length() of the result should be >= min_chars\n",
		"    // min_digits must be a whole number >= 0. It's the number of zeroes to pad with.\n",
		"    // blank must be a boolean, if true adds an additional ' ' in front of a positive number, so\n",
		"    // that it is aligned with negative numbers with the same number of digits.\n",
		"    // plus must be a boolean, if true adds a '+' in front of a postive number, so that it is\n",
		"    // aligned with negative numbers with the same number of digits.  This takes precedence over\n",
		"    // blank, if both are true.\n",
		"    // radix must be a whole number >1 and <= 10.  It is the base of the system of numerals.\n",
		"    // zero_prefix is a string prefixed before the sign to all numbers that are not 0.\n",
		"    local render_int(neg, mag, min_chars, min_digits, blank, plus, radix, zero_prefix) =\n",
		"      // dec is the minimal string needed to represent the number as text.\n",
		"      local dec =\n",
		"        if mag == 0 then\n",
		"          '0'\n",
		"        else\n",
		"          local aux(n) =\n",
		"            if n == 0 then\n",
		"              zero_prefix\n",
		"            else\n",
		"              aux(std.floor(n / radix)) + (n % radix);\n",
		"          aux(mag);\n",
		"      local zp = min_chars - (if neg || blank || plus then 1 else 0);\n",
		"      local zp2 = std.max(zp, min_digits);\n",
		"      local dec2 = pad_left(dec, zp2, '0');\n",
		"      (if neg then '-' else if plus then '+' else if blank then ' ' else '') + dec2;\n",
		"\n",
		"    // Render an integer in hexadecimal.\n",
		"    local render_hex(n__, min_chars, min_digits, blank, plus, add_zerox, capitals) =\n",
		"      local numerals = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]\n",
		"                       + if capitals then ['A', 'B', 'C', 'D', 'E', 'F']\n",
		"                       else ['a', 'b', 'c', 'd', 'e', 'f'];\n",
		"      local n_ = std.abs(n__);\n",
		"      local aux(n) =\n",
		"        if n == 0 then\n",
		"          ''\n",
		"        else\n",
		"          aux(std.floor(n / 16)) + numerals[n % 16];\n",
		"      local hex = if std.floor(n_) == 0 then '0' else aux(std.floor(n_));\n",
		"      local neg = n__ < 0;\n",
		"      local zp = min_chars - (if neg || blank || plus then 1 else 0)\n",
		"                 - (if add_zerox then 2 else 0);\n",
		"      local zp2 = std.max(zp, min_digits);\n",
		"      local hex2 = (if add_zerox then (if capitals then '0X' else '0x') else '')\n",
		"                   + pad_left(hex, zp2, '0');\n",
		"      (if neg then '-' else if plus then '+' else if blank then ' ' else '') + hex2;\n",
		"\n",
		"    local strip_trailing_zero(str) =\n",
		"      local aux(str, i) =\n",
		"        if i < 0 then\n",
		"          ''\n",
		"        else\n",
		"          if str[i] == '0' then\n",
		"            aux(str, i - 1)\n",
		"          else\n",
		"            std.substr(str, 0, i + 1);\n",
		"      aux(str, std.length(str) - 1);\n",
		"\n",
		"    // Render floating point in decimal form\n",
		"    local render_float_dec(n__, zero_pad, blank, plus, ensure_pt, trailing, prec) =\n",
		"      local n_ = std.abs(n__);\n",
		"      local whole = std.floor(n_);\n",
		"      local dot_size = if prec == 0 && !ensure_pt then 0 else 1;\n",
		"      local zp = zero_pad - prec - dot_size;\n",
		"      local str = render_int(n__ < 0, whole, zp, 0, blank, plus, 10, '');\n",
		"      if prec == 0 then\n",
		"        str + if ensure_pt then '.' else ''\n",
		"      else\n",
		"        local frac = std.floor((n_ - whole) * std.pow(10, prec) + 0.5);\n",
		"        if trailing || frac > 0 then\n",
		"          local frac_str = render_int(false, frac, prec, 0, false, false, 10, '');\n",
		"          str + '.' + if !trailing then strip_trailing_zero(frac_str) else frac_str\n",
		"        else\n",
		"          str;\n",
		"\n",
		"    // Render floating point in scientific form\n",
		"    local render_float_sci(n__, zero_pad, blank, plus, ensure_pt, trailing, caps, prec) =\n",
		"      local exponent = if n__ == 0 then 0 else std.floor(std.log(std.abs(n__)) / std.log(10));\n",
		"      local suff = (if caps then 'E' else 'e')\n",
		"                   + render_int(exponent < 0, std.abs(exponent), 3, 0, false, true, 10, '');\n",
		"      local mantissa = if exponent == -324 then\n",
		"        // Avoid a rounding error where std.pow(10, -324) is 0\n",
		"        // -324 is the smallest exponent possible.\n",
		"        n__ * 10 / std.pow(10, exponent + 1)\n",
		"      else\n",
		"        n__ / std.pow(10, exponent);\n",
		"      local zp2 = zero_pad - std.length(suff);\n",
		"      render_float_dec(mantissa, zp2, blank, plus, ensure_pt, trailing, prec) + suff;\n",
		"\n",
		"    // Render a value with an arbitrary format code.\n",
		"    local format_code(val, code, fw, prec_or_null, i) =\n",
		"      local cflags = code.cflags;\n",
		"      local fpprec = if prec_or_null != null then prec_or_null else 6;\n",
		"      local iprec = if prec_or_null != null then prec_or_null else 0;\n",
		"      local zp = if cflags.zero && !cflags.left then fw else 0;\n",
		"      if code.ctype == 's' then\n",
		"        std.toString(val)\n",
		"      else if code.ctype == 'd' then\n",
		"        if std.type(val) != 'number' then\n",
		"          error 'Format required number at '\n",
		"                + i + ', got ' + std.type(val)\n",
		"        else\n",
		"          render_int(val <= -1, std.floor(std.abs(val)), zp, iprec, cflags.blank, cflags.plus, 10, '')\n",
		"      else if code.ctype == 'o' then\n",
		"        if std.type(val) != 'number' then\n",
		"          error 'Format required number at '\n",
		"                + i + ', got ' + std.type(val)\n",
		"        else\n",
		"          local zero_prefix = if cflags.alt then '0' else '';\n",
		"          render_int(val <= -1, std.floor(std.abs(val)), zp, iprec, cflags.blank, cflags.plus, 8, zero_prefix)\n",
		"      else if code.ctype == 'x' then\n",
		"        if std.type(val) != 'number' then\n",
		"          error 'Format required number at '\n",
		"                + i + ', got ' + std.type(val)\n",
		"        else\n",
		"          render_hex(std.floor(val),\n",
		"                     zp,\n",
		"                     iprec,\n",
		"                     cflags.blank,\n",
		"                     cflags.plus,\n",
		"                     cflags.alt,\n",
		"                     code.caps)\n",
		"      else if code.ctype == 'f' then\n",
		"        if std.type(val) != 'number' then\n",
		"          error 'Format required number at '\n",
		"                + i + ', got ' + std.type(val)\n",
		"        else\n",
		"          render_float_dec(val,\n",
		"                           zp,\n",
		"                           cflags.blank,\n",
		"                           cflags.plus,\n",
		"                           cflags.alt,\n",
		"                           true,\n",
		"                           fpprec)\n",
		"      else if code.ctype == 'e' then\n",
		"        if std.type(val) != 'number' then\n",
		"          error 'Format required number at '\n",
		"                + i + ', got ' + std.type(val)\n",
		"        else\n",
		"          render_float_sci(val,\n",
		"                           zp,\n",
		"                           cflags.blank,\n",
		"                           cflags.plus,\n",
		"                           cflags.alt,\n",
		"                           true,\n",
		"                           code.caps,\n",
		"                           fpprec)\n",
		"      else if code.ctype == 'g' then\n",
		"        if std.type(val) != 'number' then\n",
		"          error 'Format required number at '\n",
		"                + i + ', got ' + std.type(val)\n",
		"        else\n",
		"          local exponent = std.floor(std.log(std.abs(val)) / std.log(10));\n",
		"          if exponent < -4 || exponent >= fpprec then\n",
		"            render_float_sci(val,\n",
		"                             zp,\n",
		"                             cflags.blank,\n",
		"                             cflags.plus,\n",
		"                             cflags.alt,\n",
		"                             cflags.alt,\n",
		"                             code.caps,\n",
		"                             fpprec - 1)\n",
		"          else\n",
		"            local digits_before_pt = std.max(1, exponent + 1);\n",
		"            render_float_dec(val,\n",
		"                             zp,\n",
		"                             cflags.blank,\n",
		"                             cflags.plus,\n",
		"                             cflags.alt,\n",
		"                             cflags.alt,\n",
		"                             fpprec - digits_before_pt)\n",
		"      else if code.ctype == 'c' then\n",
		"        if std.type(val) == 'number' then\n",
		"          std.char(val)\n",
		"        else if std.type(val) == 'string' then\n",
		"          if std.length(val) == 1 then\n",
		"            val\n",
		"          else\n",
		"            error '%c expected 1-sized string got: ' + std.length(val)\n",
		"        else\n",
		"          error '%c expected number / string, got: ' + std.type(val)\n",
		"      else\n",
		"        error 'Unknown code: ' + code.ctype;\n",
		"\n",
		"    // Render a parsed format string with an array of values.\n",
		"    local format_codes_arr(codes, arr, i, j, v) =\n",
		"      if i >= std.length(codes) then\n",
		"        if j < std.length(arr) then\n",
		"          error ('Too many values to format: ' + std.length(arr) + ', expected ' + j)\n",
		"        else\n",
		"          v\n",
		"      else\n",
		"        local code = codes[i];\n",
		"        if std.type(code) == 'string' then\n",
		"          format_codes_arr(codes, arr, i + 1, j, v + code) tailstrict\n",
		"        else\n",
		"          local tmp = if code.fw == '*' then {\n",
		"            j: j + 1,\n",
		"            fw: if j >= std.length(arr) then\n",
		"              error ('Not enough values to format: ' + std.length(arr) + ', expected at least ' + j)\n",
		"            else\n",
		"              arr[j],\n",
		"          } else {\n",
		"            j: j,\n",
		"            fw: code.fw,\n",
		"          };\n",
		"          local tmp2 = if code.prec == '*' then {\n",
		"            j: tmp.j + 1,\n",
		"            prec: if tmp.j >= std.length(arr) then\n",
		"              error ('Not enough values to format: ' + std.length(arr) + ', expected at least ' + tmp.j)\n",
		"            else\n",
		"              arr[tmp.j],\n",
		"          } else {\n",
		"            j: tmp.j,\n",
		"            prec: code.prec,\n",
		"          };\n",
		"          local j2 = tmp2.j;\n",
		"          local val =\n",
		"            if j2 < std.length(arr) then\n",
		"              arr[j2]\n",
		"            else\n",
		"              error ('Not enough values to format: ' + std.length(arr) + ', expected more than ' + j2);\n",
		"          local s =\n",
		"            if code.ctype == '%' then\n",
		"              '%'\n",
		"            else\n",
		"              format_code(val, code, tmp.fw, tmp2.prec, j2);\n",
		"          local s_padded =\n",
		"            if code.cflags.left then\n",
		"              pad_right(s, tmp.fw, ' ')\n",
		"            else\n",
		"              pad_left(s, tmp.fw, ' ');\n",
		"          local j3 =\n",
		"            if code.ctype == '%' then\n",
		"              j2\n",
		"            else\n",
		"              j2 + 1;\n",
		"          format_codes_arr(codes, arr, i + 1, j3, v + s_padded) tailstrict;\n",
		"\n",
		"    // Render a parsed format string with an object of values.\n",
		"    local format_codes_obj(codes, obj, i, v) =\n",
		"      if i >= std.length(codes) then\n",
		"        v\n",
		"      else\n",
		"        local code = codes[i];\n",
		"        if std.type(code) == 'string' then\n",
		"          format_codes_obj(codes, obj, i + 1, v + code) tailstrict\n",
		"        else\n",
		"          local f =\n",
		"            if code.mkey == null then\n",
		"              error 'Mapping keys required.'\n",
		"            else\n",
		"              code.mkey;\n",
		"          local fw =\n",
		"            if code.fw == '*' then\n",
		"              error 'Cannot use * field width with object.'\n",
		"            else\n",
		"              code.fw;\n",
		"          local prec =\n",
		"            if code.prec == '*' then\n",
		"              error 'Cannot use * precision with object.'\n",
		"            else\n",
		"              code.prec;\n",
		"          local val =\n",
		"            if std.objectHasAll(obj, f) then\n",
		"              obj[f]\n",
		"            else\n",
		"              error 'No such field: ' + f;\n",
		"          local s =\n",
		"            if code.ctype == '%' then\n",
		"              '%'\n",
		"            else\n",
		"              format_code(val, code, fw, prec, f);\n",
		"          local s_padded =\n",
		"            if code.cflags.left then\n",
		"              pad_right(s, fw, ' ')\n",
		"            else\n",
		"              pad_left(s, fw, ' ');\n",
		"          format_codes_obj(codes, obj, i + 1, v + s_padded) tailstrict;\n",
		"\n",
		"    if std.isArray(vals) then\n",
		"      format_codes_arr(codes, vals, 0, 0, '')\n",
		"    else if std.isObject(vals) then\n",
		"      format_codes_obj(codes, vals, 0, '')\n",
		"    else\n",
		"      format_codes_arr(codes, [vals], 0, 0, ''),\n",
		"\n",
		"  foldr(func, arr, init)::\n",
		"    local aux(func, arr, running, idx) =\n",
		"      if idx < 0 then\n",
		"        running\n",
		"      else\n",
		"        aux(func, arr, func(arr[idx], running), idx - 1) tailstrict;\n",
		"    aux(func, arr, init, std.length(arr) - 1),\n",
		"\n",
		"  foldl(func, arr, init)::\n",
		"    local aux(func, arr, running, idx) =\n",
		"      if idx >= std.length(arr) then\n",
		"        running\n",
		"      else\n",
		"        aux(func, arr, func(running, arr[idx]), idx + 1) tailstrict;\n",
		"    aux(func, arr, init, 0),\n",
		"\n",
		"\n",
		"  filterMap(filter_func, map_func, arr)::\n",
		"    if !std.isFunction(filter_func) then\n",
		"      error ('std.filterMap first param must be function, got ' + std.type(filter_func))\n",
		"    else if !std.isFunction(map_func) then\n",
		"      error ('std.filterMap second param must be function, got ' + std.type(map_func))\n",
		"    else if !std.isArray(arr) then\n",
		"      error ('std.filterMap third param must be array, got ' + std.type(arr))\n",
		"    else\n",
		"      std.map(map_func, std.filter(filter_func, arr)),\n",
		"\n",
		"  assertEqual(a, b)::\n",
		"    if a == b then\n",
		"      true\n",
		"    else\n",
		"      error 'Assertion failed. ' + a + ' != ' + b,\n",
		"\n",
		"  abs(n)::\n",
		"    if !std.isNumber(n) then\n",
		"      error 'std.abs expected number, got ' + std.type(n)\n",
		"    else\n",
		"      if n > 0 then n else -n,\n",
		"\n",
		"  sign(n)::\n",
		"    if !std.isNumber(n) then\n",
		"      error 'std.sign expected number, got ' + std.type(n)\n",
		"    else\n",
		"      if n > 0 then\n",
		"        1\n",
		"      else if n < 0 then\n",
		"        -1\n",
		"      else 0,\n",
		"\n",
		"  max(a, b)::\n",
		"    if !std.isNumber(a) then\n",
		"      error 'std.max first param expected number, got ' + std.type(a)\n",
		"    else if !std.isNumber(b) then\n",
		"      error 'std.max second param expected number, got ' + std.type(b)\n",
		"    else\n",
		"      if a > b then a else b,\n",
		"\n",
		"  min(a, b)::\n",
		"    if !std.isNumber(a) then\n",
		"      error 'std.max first param expected number, got ' + std.type(a)\n",
		"    else if !std.isNumber(b) then\n",
		"      error 'std.max second param expected number, got ' + std.type(b)\n",
		"    else\n",
		"      if a < b then a else b,\n",
		"\n",
		"  clamp(x, minVal, maxVal)::\n",
		"    if x  < minVal then minVal\n",
		"    else if x > maxVal then maxVal\n",
		"    else x,\n",
		"\n",
		"  flattenArrays(arrs)::\n",
		"    std.foldl(function(a, b) a + b, arrs, []),\n",
		"\n",
		"  manifestIni(ini)::\n",
		"    local body_lines(body) =\n",
		"      std.join([], [\n",
		"        local value_or_values = body[k];\n",
		"        if std.isArray(value_or_values) then\n",
		"          ['%s = %s' % [k, value] for value in value_or_values]\n",
		"        else\n",
		"          ['%s = %s' % [k, value_or_values]]\n",
		"\n",
		"        for k in std.objectFields(body)\n",
		"      ]);\n",
		"\n",
		"    local section_lines(sname, sbody) = ['[%s]' % [sname]] + body_lines(sbody),\n",
		"          main_body = if std.objectHas(ini, 'main') then body_lines(ini.main) else [],\n",
		"          all_sections = [\n",
		"      section_lines(k, ini.sections[k])\n",
		"      for k in std.objectFields(ini.sections)\n",
		"    ];\n",
		"    std.join('\\n', main_body + std.flattenArrays(all_sections) + ['']),\n",
		"\n",
		"  escapeStringJson(str_)::\n",
		"    local str = std.toString(str_);\n",
		"    local trans(ch) =\n",
		"      if ch == '\"' then\n",
		"        '\\\\\"'\n",
		"      else if ch == '\\\\' then\n",
		"        '\\\\\\\\'\n",
		"      else if ch == '\\b' then\n",
		"        '\\\\b'\n",
		"      else if ch == '\\f' then\n",
		"        '\\\\f'\n",
		"      else if ch == '\\n' then\n",
		"        '\\\\n'\n",
		"      else if ch == '\\r' then\n",
		"        '\\\\r'\n",
		"      else if ch == '\\t' then\n",
		"        '\\\\t'\n",
		"      else\n",
		"        local cp = std.codepoint(ch);\n",
		"        if cp < 32 || (cp >= 127 && cp <= 159) then\n",
		"          '\\\\u%04x' % [cp]\n",
		"        else\n",
		"          ch;\n",
		"    '\"%s\"' % std.join('', [trans(ch) for ch in std.stringChars(str)]),\n",
		"\n",
		"  escapeStringPython(str)::\n",
		"    std.escapeStringJson(str),\n",
		"\n",
		"  escapeStringBash(str_)::\n",
		"    local str = std.toString(str_);\n",
		"    local trans(ch) =\n",
		"      if ch == \"'\" then\n",
		"        \"'\\\"'\\\"'\"\n",
		"      else\n",
		"        ch;\n",
		"    \"'%s'\" % std.join('', [trans(ch) for ch in std.stringChars(str)]),\n",
		"\n",
		"  escapeStringDollars(str_)::\n",
		"    local str = std.toString(str_);\n",
		"    local trans(ch) =\n",
		"      if ch == '$' then\n",
		"        '$$'\n",
		"      else\n",
		"        ch;\n",
		"    std.foldl(function(a, b) a + trans(b), std.stringChars(str), ''),\n",
		"\n",
		"  manifestJson(value):: std.manifestJsonEx(value, '    '),\n",
		"\n",
		"  manifestJsonEx(value, indent)::\n",
		"    local aux(v, path, cindent) =\n",
		"      if v == true then\n",
		"        'true'\n",
		"      else if v == false then\n",
		"        'false'\n",
		"      else if v == null then\n",
		"        'null'\n",
		"      else if std.isNumber(v) then\n",
		"        '' + v\n",
		"      else if std.isString(v) then\n",
		"        std.escapeStringJson(v)\n",
		"      else if std.isFunction(v) then\n",
		"        error 'Tried to manifest function at ' + path\n",
		"      else if std.isArray(v) then\n",
		"        local range = std.range(0, std.length(v) - 1);\n",
		"        local new_indent = cindent + indent;\n",
		"        local lines = ['[\\n']\n",
		"                      + std.join([',\\n'],\n",
		"                                 [\n",
		"                                   [new_indent + aux(v[i], path + [i], new_indent)]\n",
		"                                   for i in range\n",
		"                                 ])\n",
		"                      + ['\\n' + cindent + ']'];\n",
		"        std.join('', lines)\n",
		"      else if std.isObject(v) then\n",
		"        local lines = ['{\\n']\n",
		"                      + std.join([',\\n'],\n",
		"                                 [\n",
		"                                   [cindent + indent + std.escapeStringJson(k) + ': '\n",
		"                                    + aux(v[k], path + [k], cindent + indent)]\n",
		"                                   for k in std.objectFields(v)\n",
		"                                 ])\n",
		"                      + ['\\n' + cindent + '}'];\n",
		"        std.join('', lines);\n",
		"    aux(value, [], ''),\n",
		"\n",
		"  manifestYamlDoc(value, indent_array_in_object=false)::\n",
		"    local aux(v, path, cindent) =\n",
		"      if v == true then\n",
		"        'true'\n",
		"      else if v == false then\n",
		"        'false'\n",
		"      else if v == null then\n",
		"        'null'\n",
		"      else if std.isNumber(v) then\n",
		"        '' + v\n",
		"      else if std.isString(v) then\n",
		"        local len = std.length(v);\n",
		"        if len == 0 then\n",
		"          '\"\"'\n",
		"        else if v[len - 1] == '\\n' then\n",
		"          local split = std.split(v, '\\n');\n",
		"          std.join('\\n' + cindent + '  ', ['|'] + split[0:std.length(split) - 1])\n",
		"        else\n",
		"          std.escapeStringJson(v)\n",
		"      else if std.isFunction(v) then\n",
		"        error 'Tried to manifest function at ' + path\n",
		"      else if std.isArray(v) then\n",
		"        if std.length(v) == 0 then\n",
		"          '[]'\n",
		"        else\n",
		"          local params(value) =\n",
		"            if std.isArray(value) && std.length(value) > 0 then {\n",
		"              // While we could avoid the new line, it yields YAML that is\n",
		"              // hard to read, e.g.:\n",
		"              // - - - 1\n",
		"              //     - 2\n",
		"              //   - - 3\n",
		"              //     - 4\n",
		"              new_indent: cindent + '  ',\n",
		"              space: '\\n' + self.new_indent,\n",
		"            } else if std.isObject(value) && std.length(value) > 0 then {\n",
		"              new_indent: cindent + '  ',\n",
		"              // In this case we can start on the same line as the - because the indentation\n",
		"              // matches up then.  The converse is not true, because fields are not always\n",
		"              // 1 character long.\n",
		"              space: ' ',\n",
		"            } else {\n",
		"              // In this case, new_indent is only used in the case of multi-line strings.\n",
		"              new_indent: cindent,\n",
		"              space: ' ',\n",
		"            };\n",
		"          local range = std.range(0, std.length(v) - 1);\n",
		"          local parts = [\n",
		"            '-' + param.space + aux(v[i], path + [i], param.new_indent)\n",
		"            for i in range\n",
		"            for param in [params(v[i])]\n",
		"          ];\n",
		"          std.join('\\n' + cindent, parts)\n",
		"      else if std.isObject(v) then\n",
		"        if std.length(v) == 0 then\n",
		"          '{}'\n",
		"        else\n",
		"          local params(value) =\n",
		"            if std.isArray(value) && std.length(value) > 0 then {\n",
		"              // Not indenting allows e.g.\n",
		"              // ports:\n",
		"              // - 80\n",
		"              // instead of\n",
		"              // ports:\n",
		"              //   - 80\n",
		"              new_indent: if indent_array_in_object then cindent + '  ' else cindent,\n",
		"              space: '\\n' + self.new_indent,\n",
		"            } else if std.isObject(value) && std.length(value) > 0 then {\n",
		"              new_indent: cindent + '  ',\n",
		"              space: '\\n' + self.new_indent,\n",
		"            } else {\n",
		"              // In this case, new_indent is only used in the case of multi-line strings.\n",
		"              new_indent: cindent,\n",
		"              space: ' ',\n",
		"            };\n",
		"          local lines = [\n",
		"            std.escapeStringJson(k) + ':' + param.space + aux(v[k], path + [k], param.new_indent)\n",
		"            for k in std.objectFields(v)\n",
		"            for param in [params(v[k])]\n",
		"          ];\n",
		"          std.join('\\n' + cindent, lines);\n",
		"    aux(value, [], ''),\n",
		"\n",
		"  manifestYamlStream(value, indent_array_in_object=false, c_document_end=true)::\n",
		"    if !std.isArray(value) then\n",
		"      error 'manifestYamlStream only takes arrays, got ' + std.type(value)\n",
		"    else\n",
		"      '---\\n' + std.join(\n",
		"        '\\n---\\n', [std.manifestYamlDoc(e, indent_array_in_object) for e in value]\n",
		"      ) + if c_document_end then '\\n...\\n' else '\\n',\n",
		"\n",
		"\n",
		"  manifestPython(v)::\n",
		"    if std.isObject(v) then\n",
		"      local fields = [\n",
		"        '%s: %s' % [std.escapeStringPython(k), std.manifestPython(v[k])]\n",
		"        for k in std.objectFields(v)\n",
		"      ];\n",
		"      '{%s}' % [std.join(', ', fields)]\n",
		"    else if std.isArray(v) then\n",
		"      '[%s]' % [std.join(', ', [std.manifestPython(v2) for v2 in v])]\n",
		"    else if std.isString(v) then\n",
		"      '%s' % [std.escapeStringPython(v)]\n",
		"    else if std.isFunction(v) then\n",
		"      error 'cannot manifest function'\n",
		"    else if std.isNumber(v) then\n",
		"      std.toString(v)\n",
		"    else if v == true then\n",
		"      'True'\n",
		"    else if v == false then\n",
		"      'False'\n",
		"    else if v == null then\n",
		"      'None',\n",
		"\n",
		"  manifestPythonVars(conf)::\n",
		"    local vars = ['%s = %s' % [k, std.manifestPython(conf[k])] for k in std.objectFields(conf)];\n",
		"    std.join('\\n', vars + ['']),\n",
		"\n",
		"  manifestXmlJsonml(value)::\n",
		"    if !std.isArray(value) then\n",
		"      error 'Expected a JSONML value (an array), got %s' % std.type(value)\n",
		"    else\n",
		"      local aux(v) =\n",
		"        if std.isString(v) then\n",
		"          v\n",
		"        else\n",
		"          local tag = v[0];\n",
		"          local has_attrs = std.length(v) > 1 && std.isObject(v[1]);\n",
		"          local attrs = if has_attrs then v[1] else {};\n",
		"          local children = if has_attrs then v[2:] else v[1:];\n",
		"          local attrs_str =\n",
		"            std.join('', [' %s=\"%s\"' % [k, attrs[k]] for k in std.objectFields(attrs)]);\n",
		"          std.deepJoin(['<', tag, attrs_str, '>', [aux(x) for x in children], '</', tag, '>']);\n",
		"\n",
		"      aux(value),\n",
		"\n",
		"  local base64_table = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/',\n",
		"  local base64_inv = { [base64_table[i]]: i for i in std.range(0, 63) },\n",
		"\n",
		"  base64(input)::\n",
		"    local bytes =\n",
		"      if std.isString(input) then\n",
		"        std.map(function(c) std.codepoint(c), input)\n",
		"      else\n",
		"        input;\n",
		"\n",
		"    local aux(arr, i, r) =\n",
		"      if i >= std.length(arr) then\n",
		"        r\n",
		"      else if i + 1 >= std.length(arr) then\n",
		"        local str =\n",
		"          // 6 MSB of i\n",
		"          base64_table[(arr[i] & 252) >> 2] +\n",
		"          // 2 LSB of i\n",
		"          base64_table[(arr[i] & 3) << 4] +\n",
		"          '==';\n",
		"        aux(arr, i + 3, r + str) tailstrict\n",
		"      else if i + 2 >= std.length(arr) then\n",
		"        local str =\n",
		"          // 6 MSB of i\n",
		"          base64_table[(arr[i] & 252) >> 2] +\n",
		"          // 2 LSB of i, 4 MSB of i+1\n",
		"          base64_table[(arr[i] & 3) << 4 | (arr[i + 1] & 240) >> 4] +\n",
		"          // 4 LSB of i+1\n",
		"          base64_table[(arr[i + 1] & 15) << 2] +\n",
		"          '=';\n",
		"        aux(arr, i + 3, r + str) tailstrict\n",
		"      else\n",
		"        local str =\n",
		"          // 6 MSB of i\n",
		"          base64_table[(arr[i] & 252) >> 2] +\n",
		"          // 2 LSB of i, 4 MSB of i+1\n",
		"          base64_table[(arr[i] & 3) << 4 | (arr[i + 1] & 240) >> 4] +\n",
		"          // 4 LSB of i+1, 2 MSB of i+2\n",
		"          base64_table[(arr[i + 1] & 15) << 2 | (arr[i + 2] & 192) >> 6] +\n",
		"          // 6 LSB of i+2\n",
		"          base64_table[(arr[i + 2] & 63)];\n",
		"        aux(arr, i + 3, r + str) tailstrict;\n",
		"\n",
		"    local sanity = std.foldl(function(r, a) r && (a < 256), bytes, true);\n",
		"    if !sanity then\n",
		"      error 'Can only base64 encode strings / arrays of single bytes.'\n",
		"    else\n",
		"      aux(bytes, 0, ''),\n",
		"\n",
		"\n",
		"  base64DecodeBytes(str)::\n",
		"    if std.length(str) % 4 != 0 then\n",
		"      error 'Not a base64 encoded string \"%s\"' % str\n",
		"    else\n",
		"      local aux(str, i, r) =\n",
		"        if i >= std.length(str) then\n",
		"          r\n",
		"        else\n",
		"          // all 6 bits of i, 2 MSB of i+1\n",
		"          local n1 = [base64_inv[str[i]] << 2 | (base64_inv[str[i + 1]] >> 4)];\n",
		"          // 4 LSB of i+1, 4MSB of i+2\n",
		"          local n2 =\n",
		"            if str[i + 2] == '=' then []\n",
		"            else [(base64_inv[str[i + 1]] & 15) << 4 | (base64_inv[str[i + 2]] >> 2)];\n",
		"          // 2 LSB of i+2, all 6 bits of i+3\n",
		"          local n3 =\n",
		"            if str[i + 3] == '=' then []\n",
		"            else [(base64_inv[str[i + 2]] & 3) << 6 | base64_inv[str[i + 3]]];\n",
		"          aux(str, i + 4, r + n1 + n2 + n3) tailstrict;\n",
		"      aux(str, 0, []),\n",
		"\n",
		"  base64Decode(str)::\n",
		"    local bytes = std.base64DecodeBytes(str);\n",
		"    std.join('', std.map(function(b) std.char(b), bytes)),\n",
		"\n",
		"  reverse(arr)::\n",
		"    local l = std.length(arr);\n",
		"    std.makeArray(l, function(i) arr[l - i - 1]),\n",
		"\n",
		"  // Merge-sort for long arrays and naive quicksort for shorter ones\n",
		"  sort(arr, keyF=id)::\n",
		"    local quickSort(arr, keyF=id) =\n",
		"      local l = std.length(arr);\n",
		"      if std.length(arr) <= 1 then\n",
		"        arr\n",
		"      else\n",
		"        local pos = 0;\n",
		"        local pivot = keyF(arr[pos]);\n",
		"        local rest = std.makeArray(l - 1, function(i) if i < pos then arr[i] else arr[i + 1]);\n",
		"        local left = std.filter(function(x) keyF(x) < pivot, rest);\n",
		"        local right = std.filter(function(x) keyF(x) >= pivot, rest);\n",
		"        quickSort(left, keyF) + [arr[pos]] + quickSort(right, keyF);\n",
		"\n",
		"    local merge(a, b) =\n",
		"      local la = std.length(a), lb = std.length(b);\n",
		"      local aux(i, j, prefix) =\n",
		"        if i == la then\n",
		"          prefix + b[j:]\n",
		"        else if j == lb then\n",
		"          prefix + a[i:]\n",
		"        else\n",
		"          if keyF(a[i]) <= keyF(b[j]) then\n",
		"            aux(i + 1, j, prefix + [a[i]]) tailstrict\n",
		"          else\n",
		"            aux(i, j + 1, prefix + [b[j]]) tailstrict;\n",
		"      aux(0, 0, []);\n",
		"\n",
		"    local l = std.length(arr);\n",
		"    if std.length(arr) <= 30 then\n",
		"      quickSort(arr, keyF=keyF)\n",
		"    else\n",
		"      local mid = std.floor(l / 2);\n",
		"      local left = arr[:mid], right = arr[mid:];\n",
		"      merge(std.sort(left, keyF=keyF), std.sort(right, keyF=keyF)),\n",
		"\n",
		"  uniq(arr, keyF=id)::\n",
		"    local f(a, b) =\n",
		"      if std.length(a) == 0 then\n",
		"        [b]\n",
		"      else if keyF(a[std.length(a) - 1]) == keyF(b) then\n",
		"        a\n",
		"      else\n",
		"        a + [b];\n",
		"    std.foldl(f, arr, []),\n",
		"\n",
		"  set(arr, keyF=id)::\n",
		"    std.uniq(std.sort(arr, keyF), keyF),\n",
		"\n",
		"  setMember(x, arr, keyF=id)::\n",
		"    // TODO(dcunnin): Binary chop for O(log n) complexity\n",
		"    std.length(std.setInter([x], arr, keyF)) > 0,\n",
		"\n",
		"  setUnion(a, b, keyF=id)::\n",
		"    // NOTE: order matters, values in `a` win\n",
		"    local aux(a, b, i, j, acc) =\n",
		"      if i >= std.length(a) then\n",
		"        acc + b[j:]\n",
		"      else if j >= std.length(b) then\n",
		"        acc + a[i:]\n",
		"      else\n",
		"        local ak = keyF(a[i]);\n",
		"        local bk = keyF(b[j]);\n",
		"        if ak == bk then\n",
		"          aux(a, b, i + 1, j + 1, acc + [a[i]]) tailstrict\n",
		"        else if ak < bk then\n",
		"          aux(a, b, i + 1, j, acc + [a[i]]) tailstrict\n",
		"        else\n",
		"          aux(a, b, i, j + 1, acc + [b[j]]) tailstrict;\n",
		"    aux(a, b, 0, 0, []),\n",
		"\n",
		"  setInter(a, b, keyF=id)::\n",
		"    local aux(a, b, i, j, acc) =\n",
		"      if i >= std.length(a) || j >= std.length(b) then\n",
		"        acc\n",
		"      else\n",
		"        if keyF(a[i]) == keyF(b[j]) then\n",
		"          aux(a, b, i + 1, j + 1, acc + [a[i]]) tailstrict\n",
		"        else if keyF(a[i]) < keyF(b[j]) then\n",
		"          aux(a, b, i + 1, j, acc) tailstrict\n",
		"        else\n",
		"          aux(a, b, i, j + 1, acc) tailstrict;\n",
		"    aux(a, b, 0, 0, []) tailstrict,\n",
		"\n",
		"  setDiff(a, b, keyF=id)::\n",
		"    local aux(a, b, i, j, acc) =\n",
		"      if i >= std.length(a) then\n",
		"        acc\n",
		"      else if j >= std.length(b) then\n",
		"        acc + a[i:]\n",
		"      else\n",
		"        if keyF(a[i]) == keyF(b[j]) then\n",
		"          aux(a, b, i + 1, j + 1, acc) tailstrict\n",
		"        else if keyF(a[i]) < keyF(b[j]) then\n",
		"          aux(a, b, i + 1, j, acc + [a[i]]) tailstrict\n",
		"        else\n",
		"          aux(a, b, i, j + 1, acc) tailstrict;\n",
		"    aux(a, b, 0, 0, []) tailstrict,\n",
		"\n",
		"  mergePatch(target, patch)::\n",
		"    if std.isObject(patch) then\n",
		"      local target_object =\n",
		"        if std.isObject(target) then target else {};\n",
		"\n",
		"      local target_fields =\n",
		"        if std.isObject(target_object) then std.objectFields(target_object) else [];\n",
		"\n",
		"      local null_fields = [k for k in std.objectFields(patch) if patch[k] == null];\n",
		"      local both_fields = std.setUnion(target_fields, std.objectFields(patch));\n",
		"\n",
		"      {\n",
		"        [k]:\n",
		"          if !std.objectHas(patch, k) then\n",
		"            target_object[k]\n",
		"          else if !std.objectHas(target_object, k) then\n",
		"            std.mergePatch(null, patch[k]) tailstrict\n",
		"          else\n",
		"            std.mergePatch(target_object[k], patch[k]) tailstrict\n",
		"        for k in std.setDiff(both_fields, null_fields)\n",
		"      }\n",
		"    else\n",
		"      patch,\n",
		"\n",
		"  objectFields(o)::\n",
		"    std.objectFieldsEx(o, false),\n",
		"\n",
		"  objectFieldsAll(o)::\n",
		"    std.objectFieldsEx(o, true),\n",
		"\n",
		"  objectHas(o, f)::\n",
		"    std.objectHasEx(o, f, false),\n",
		"\n",
		"  objectHasAll(o, f)::\n",
		"    std.objectHasEx(o, f, true),\n",
		"\n",
		"  equals(a, b)::\n",
		"    local ta = std.type(a);\n",
		"    local tb = std.type(b);\n",
		"    if !std.primitiveEquals(ta, tb) then\n",
		"      false\n",
		"    else\n",
		"      if std.primitiveEquals(ta, 'array') then\n",
		"        local la = std.length(a);\n",
		"        if !std.primitiveEquals(la, std.length(b)) then\n",
		"          false\n",
		"        else\n",
		"          local aux(a, b, i) =\n",
		"            if i >= la then\n",
		"              true\n",
		"            else if a[i] != b[i] then\n",
		"              false\n",
		"            else\n",
		"              aux(a, b, i + 1) tailstrict;\n",
		"          aux(a, b, 0)\n",
		"      else if std.primitiveEquals(ta, 'object') then\n",
		"        local fields = std.objectFields(a);\n",
		"        local lfields = std.length(fields);\n",
		"        if fields != std.objectFields(b) then\n",
		"          false\n",
		"        else\n",
		"          local aux(a, b, i) =\n",
		"            if i >= lfields then\n",
		"              true\n",
		"            else if local f = fields[i]; a[f] != b[f] then\n",
		"              false\n",
		"            else\n",
		"              aux(a, b, i + 1) tailstrict;\n",
		"          aux(a, b, 0)\n",
		"      else\n",
		"        std.primitiveEquals(a, b),\n",
		"\n",
		"\n",
		"  resolvePath(f, r)::\n",
		"    local arr = std.split(f, '/');\n",
		"    std.join('/', std.makeArray(std.length(arr) - 1, function(i) arr[i]) + [r]),\n",
		"\n",
		"  prune(a)::\n",
		"    local isContent(b) =\n",
		"      if b == null then\n",
		"        false\n",
		"      else if std.isArray(b) then\n",
		"        std.length(b) > 0\n",
		"      else if std.isObject(b) then\n",
		"        std.length(b) > 0\n",
		"      else\n",
		"        true;\n",
		"    if std.isArray(a) then\n",
		"      [std.prune(x) for x in a if isContent($.prune(x))]\n",
		"    else if std.isObject(a) then {\n",
		"      [x]: $.prune(a[x])\n",
		"      for x in std.objectFields(a)\n",
		"      if isContent(std.prune(a[x]))\n",
		"    } else\n",
		"      a,\n",
		"\n",
		"  findSubstr(pat, str)::\n",
		"    if !std.isString(pat) then\n",
		"      error 'findSubstr first parameter should be a string, got ' + std.type(pat)\n",
		"    else if !std.isString(str) then\n",
		"      error 'findSubstr second parameter should be a string, got ' + std.type(str)\n",
		"    else\n",
		"      local pat_len = std.length(pat);\n",
		"      local str_len = std.length(str);\n",
		"      if pat_len == 0 || str_len == 0 || pat_len > str_len then\n",
		"        []\n",
		"      else\n",
		"        std.filter(function(i) str[i:i + pat_len] == pat, std.range(0, str_len - pat_len)),\n",
		"\n",
		"  find(value, arr)::\n",
		"    if !std.isArray(arr) then\n",
		"      error 'find second parameter should be an array, got ' + std.type(arr)\n",
		"    else\n",
		"      std.filter(function(i) arr[i] == value, std.range(0, std.length(arr) - 1)),\n",
		"}\n",
		"\n",
	},
}

// StdAst is the AST for the standard library.
var StdAst = _StdAst
var _StdAst = &ast.DesugaredObject{
	NodeBase: ast.NodeBase{
		LocRange: ast.LocationRange{
			FileName: "<std>",
			Begin: ast.Location{
				Line: int(23),
				Column: int(1),
			},
			End: ast.Location{
				Line: int(1382),
				Column: int(2),
			},
			File: p1,
		},
		Fodder: ast.Fodder{
			ast.FodderElement{
				Kind: ast.FodderKind(0),
				Blanks: int(0),
				Indent: int(0),
				Comment: []string{},
			},
			ast.FodderElement{
				Kind: ast.FodderKind(2),
				Blanks: int(1),
				Indent: int(0),
				Comment: []string{
					"/*",
					"Copyright 2015 Google Inc. All rights reserved.",
					"",
					"Licensed under the Apache License, Version 2.0 (the \"License\");",
					"you may not use this file except in compliance with the License.",
					"You may obtain a copy of the License at",
					"",
					"    http://www.apache.org/licenses/LICENSE-2.0",
					"",
					"Unless required by applicable law or agreed to in writing, software",
					"distributed under the License is distributed on an \"AS IS\" BASIS,",
					"WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.",
					"See the License for the specific language governing permissions and",
					"limitations under the License.",
					"*/",
				},
			},
			ast.FodderElement{
				Kind: ast.FodderKind(2),
				Blanks: int(0),
				Indent: int(0),
				Comment: []string{
					"/* This is the Jsonnet standard library, at least the parts of it that are written in Jsonnet.",
					" *",
					" * There are some native methods as well, which are defined in the interpreter and added to this",
					" * file.  It is never necessary to import std.jsonnet, it is embedded into the interpreter at",
					" * compile-time and automatically imported into all other Jsonnet programs.",
					" */",
				},
			},
		},
		Ctx: p7,
		FreeVars: nil,
	},
	Asserts: ast.Nodes{},
	Fields: ast.DesugaredObjectFields{
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "isString",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "v",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(28),
								Column: int(12),
							},
							End: ast.Location{
								Line: int(28),
								Column: int(13),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Binary{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(28),
								Column: int(17),
							},
							End: ast.Location{
								Line: int(28),
								Column: int(40),
							},
							File: p1,
						},
						Fodder: ast.Fodder{},
						Ctx: p15,
						FreeVars: ast.Identifiers{
							"std",
							"v",
						},
					},
					Left: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(28),
									Column: int(17),
								},
								End: ast.Location{
									Line: int(28),
									Column: int(28),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p15,
							FreeVars: ast.Identifiers{
								"std",
								"v",
							},
						},
						Target: &ast.Index{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(28),
										Column: int(17),
									},
									End: ast.Location{
										Line: int(28),
										Column: int(25),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p15,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(28),
											Column: int(17),
										},
										End: ast.Location{
											Line: int(28),
											Column: int(20),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Id: "std",
							},
							LeftBracketFodder: ast.Fodder{},
							Index: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Value: "type",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							RightBracketFodder: ast.Fodder{},
							Id: nil,
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(28),
													Column: int(26),
												},
												End: ast.Location{
													Line: int(28),
													Column: int(27),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p26,
											FreeVars: ast.Identifiers{
												"v",
											},
										},
										Id: "v",
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
					OpFodder: ast.Fodder{},
					Op: ast.BinaryOp(12),
					Right: &ast.LiteralString{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(28),
									Column: int(32),
								},
								End: ast.Location{
									Line: int(28),
									Column: int(40),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p15,
							FreeVars: nil,
						},
						Value: "string",
						Kind: ast.LiteralStringKind(1),
						BlockIndent: "",
						BlockTermIndent: "",
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(28),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(28),
					Column: int(40),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "isNumber",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "v",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(29),
								Column: int(12),
							},
							End: ast.Location{
								Line: int(29),
								Column: int(13),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Binary{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(29),
								Column: int(17),
							},
							End: ast.Location{
								Line: int(29),
								Column: int(40),
							},
							File: p1,
						},
						Fodder: ast.Fodder{},
						Ctx: p34,
						FreeVars: ast.Identifiers{
							"std",
							"v",
						},
					},
					Left: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(29),
									Column: int(17),
								},
								End: ast.Location{
									Line: int(29),
									Column: int(28),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p34,
							FreeVars: ast.Identifiers{
								"std",
								"v",
							},
						},
						Target: &ast.Index{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(29),
										Column: int(17),
									},
									End: ast.Location{
										Line: int(29),
										Column: int(25),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p34,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(29),
											Column: int(17),
										},
										End: ast.Location{
											Line: int(29),
											Column: int(20),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Id: "std",
							},
							LeftBracketFodder: ast.Fodder{},
							Index: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Value: "type",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							RightBracketFodder: ast.Fodder{},
							Id: nil,
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(29),
													Column: int(26),
												},
												End: ast.Location{
													Line: int(29),
													Column: int(27),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p45,
											FreeVars: ast.Identifiers{
												"v",
											},
										},
										Id: "v",
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
					OpFodder: ast.Fodder{},
					Op: ast.BinaryOp(12),
					Right: &ast.LiteralString{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(29),
									Column: int(32),
								},
								End: ast.Location{
									Line: int(29),
									Column: int(40),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p34,
							FreeVars: nil,
						},
						Value: "number",
						Kind: ast.LiteralStringKind(1),
						BlockIndent: "",
						BlockTermIndent: "",
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(29),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(29),
					Column: int(40),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "isBoolean",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "v",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(30),
								Column: int(13),
							},
							End: ast.Location{
								Line: int(30),
								Column: int(14),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Binary{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(30),
								Column: int(18),
							},
							End: ast.Location{
								Line: int(30),
								Column: int(42),
							},
							File: p1,
						},
						Fodder: ast.Fodder{},
						Ctx: p53,
						FreeVars: ast.Identifiers{
							"std",
							"v",
						},
					},
					Left: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(30),
									Column: int(18),
								},
								End: ast.Location{
									Line: int(30),
									Column: int(29),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p53,
							FreeVars: ast.Identifiers{
								"std",
								"v",
							},
						},
						Target: &ast.Index{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(30),
										Column: int(18),
									},
									End: ast.Location{
										Line: int(30),
										Column: int(26),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p53,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(30),
											Column: int(18),
										},
										End: ast.Location{
											Line: int(30),
											Column: int(21),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Id: "std",
							},
							LeftBracketFodder: ast.Fodder{},
							Index: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Value: "type",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							RightBracketFodder: ast.Fodder{},
							Id: nil,
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(30),
													Column: int(27),
												},
												End: ast.Location{
													Line: int(30),
													Column: int(28),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p64,
											FreeVars: ast.Identifiers{
												"v",
											},
										},
										Id: "v",
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
					OpFodder: ast.Fodder{},
					Op: ast.BinaryOp(12),
					Right: &ast.LiteralString{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(30),
									Column: int(33),
								},
								End: ast.Location{
									Line: int(30),
									Column: int(42),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p53,
							FreeVars: nil,
						},
						Value: "boolean",
						Kind: ast.LiteralStringKind(1),
						BlockIndent: "",
						BlockTermIndent: "",
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(30),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(30),
					Column: int(42),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "isObject",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "v",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(31),
								Column: int(12),
							},
							End: ast.Location{
								Line: int(31),
								Column: int(13),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Binary{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(31),
								Column: int(17),
							},
							End: ast.Location{
								Line: int(31),
								Column: int(40),
							},
							File: p1,
						},
						Fodder: ast.Fodder{},
						Ctx: p72,
						FreeVars: ast.Identifiers{
							"std",
							"v",
						},
					},
					Left: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(31),
									Column: int(17),
								},
								End: ast.Location{
									Line: int(31),
									Column: int(28),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p72,
							FreeVars: ast.Identifiers{
								"std",
								"v",
							},
						},
						Target: &ast.Index{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(31),
										Column: int(17),
									},
									End: ast.Location{
										Line: int(31),
										Column: int(25),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p72,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(31),
											Column: int(17),
										},
										End: ast.Location{
											Line: int(31),
											Column: int(20),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Id: "std",
							},
							LeftBracketFodder: ast.Fodder{},
							Index: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Value: "type",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							RightBracketFodder: ast.Fodder{},
							Id: nil,
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(31),
													Column: int(26),
												},
												End: ast.Location{
													Line: int(31),
													Column: int(27),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p83,
											FreeVars: ast.Identifiers{
												"v",
											},
										},
										Id: "v",
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
					OpFodder: ast.Fodder{},
					Op: ast.BinaryOp(12),
					Right: &ast.LiteralString{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(31),
									Column: int(32),
								},
								End: ast.Location{
									Line: int(31),
									Column: int(40),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p72,
							FreeVars: nil,
						},
						Value: "object",
						Kind: ast.LiteralStringKind(1),
						BlockIndent: "",
						BlockTermIndent: "",
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(31),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(31),
					Column: int(40),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "isArray",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "v",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(32),
								Column: int(11),
							},
							End: ast.Location{
								Line: int(32),
								Column: int(12),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Binary{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(32),
								Column: int(16),
							},
							End: ast.Location{
								Line: int(32),
								Column: int(38),
							},
							File: p1,
						},
						Fodder: ast.Fodder{},
						Ctx: p91,
						FreeVars: ast.Identifiers{
							"std",
							"v",
						},
					},
					Left: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(32),
									Column: int(16),
								},
								End: ast.Location{
									Line: int(32),
									Column: int(27),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p91,
							FreeVars: ast.Identifiers{
								"std",
								"v",
							},
						},
						Target: &ast.Index{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(32),
										Column: int(16),
									},
									End: ast.Location{
										Line: int(32),
										Column: int(24),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p91,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(32),
											Column: int(16),
										},
										End: ast.Location{
											Line: int(32),
											Column: int(19),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Id: "std",
							},
							LeftBracketFodder: ast.Fodder{},
							Index: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Value: "type",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							RightBracketFodder: ast.Fodder{},
							Id: nil,
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(32),
													Column: int(25),
												},
												End: ast.Location{
													Line: int(32),
													Column: int(26),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p102,
											FreeVars: ast.Identifiers{
												"v",
											},
										},
										Id: "v",
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
					OpFodder: ast.Fodder{},
					Op: ast.BinaryOp(12),
					Right: &ast.LiteralString{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(32),
									Column: int(31),
								},
								End: ast.Location{
									Line: int(32),
									Column: int(38),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p91,
							FreeVars: nil,
						},
						Value: "array",
						Kind: ast.LiteralStringKind(1),
						BlockIndent: "",
						BlockTermIndent: "",
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(32),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(32),
					Column: int(38),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "isFunction",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "v",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(33),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(33),
								Column: int(15),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Binary{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(33),
								Column: int(19),
							},
							End: ast.Location{
								Line: int(33),
								Column: int(44),
							},
							File: p1,
						},
						Fodder: ast.Fodder{},
						Ctx: p110,
						FreeVars: ast.Identifiers{
							"std",
							"v",
						},
					},
					Left: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(33),
									Column: int(19),
								},
								End: ast.Location{
									Line: int(33),
									Column: int(30),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p110,
							FreeVars: ast.Identifiers{
								"std",
								"v",
							},
						},
						Target: &ast.Index{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(33),
										Column: int(19),
									},
									End: ast.Location{
										Line: int(33),
										Column: int(27),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p110,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(33),
											Column: int(19),
										},
										End: ast.Location{
											Line: int(33),
											Column: int(22),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Id: "std",
							},
							LeftBracketFodder: ast.Fodder{},
							Index: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Value: "type",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							RightBracketFodder: ast.Fodder{},
							Id: nil,
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(33),
													Column: int(28),
												},
												End: ast.Location{
													Line: int(33),
													Column: int(29),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p121,
											FreeVars: ast.Identifiers{
												"v",
											},
										},
										Id: "v",
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
					OpFodder: ast.Fodder{},
					Op: ast.BinaryOp(12),
					Right: &ast.LiteralString{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(33),
									Column: int(34),
								},
								End: ast.Location{
									Line: int(33),
									Column: int(44),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p110,
							FreeVars: nil,
						},
						Value: "function",
						Kind: ast.LiteralStringKind(1),
						BlockIndent: "",
						BlockTermIndent: "",
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(33),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(33),
					Column: int(44),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "toString",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "a",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(35),
								Column: int(12),
							},
							End: ast.Location{
								Line: int(35),
								Column: int(13),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(36),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(36),
								Column: int(50),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p130,
						FreeVars: ast.Identifiers{
							"a",
							"std",
						},
					},
					Cond: &ast.Binary{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(36),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(36),
									Column: int(31),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p130,
							FreeVars: ast.Identifiers{
								"a",
								"std",
							},
						},
						Left: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(36),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(36),
										Column: int(19),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p130,
								FreeVars: ast.Identifiers{
									"a",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(36),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(36),
											Column: int(16),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p130,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(36),
												Column: int(8),
											},
											End: ast.Location{
												Line: int(36),
												Column: int(11),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "type",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(36),
														Column: int(17),
													},
													End: ast.Location{
														Line: int(36),
														Column: int(18),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p143,
												FreeVars: ast.Identifiers{
													"a",
												},
											},
											Id: "a",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
						OpFodder: ast.Fodder{},
						Op: ast.BinaryOp(12),
						Right: &ast.LiteralString{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(36),
										Column: int(23),
									},
									End: ast.Location{
										Line: int(36),
										Column: int(31),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p130,
								FreeVars: nil,
							},
							Value: "string",
							Kind: ast.LiteralStringKind(1),
							BlockIndent: "",
							BlockTermIndent: "",
						},
					},
					ThenFodder: ast.Fodder{},
					BranchTrue: &ast.Var{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(36),
									Column: int(37),
								},
								End: ast.Location{
									Line: int(36),
									Column: int(38),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p130,
							FreeVars: ast.Identifiers{
								"a",
							},
						},
						Id: "a",
					},
					ElseFodder: ast.Fodder{},
					BranchFalse: &ast.Binary{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(36),
									Column: int(44),
								},
								End: ast.Location{
									Line: int(36),
									Column: int(50),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p130,
							FreeVars: ast.Identifiers{
								"a",
							},
						},
						Left: &ast.LiteralString{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(36),
										Column: int(44),
									},
									End: ast.Location{
										Line: int(36),
										Column: int(46),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p130,
								FreeVars: nil,
							},
							Value: "",
							Kind: ast.LiteralStringKind(1),
							BlockIndent: "",
							BlockTermIndent: "",
						},
						OpFodder: ast.Fodder{},
						Op: ast.BinaryOp(3),
						Right: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(36),
										Column: int(49),
									},
									End: ast.Location{
										Line: int(36),
										Column: int(50),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p130,
								FreeVars: ast.Identifiers{
									"a",
								},
							},
							Id: "a",
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(35),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(36),
					Column: int(50),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "substr",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(38),
								Column: int(10),
							},
							End: ast.Location{
								Line: int(38),
								Column: int(13),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "from",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(38),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(38),
								Column: int(19),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "len",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(38),
								Column: int(21),
							},
							End: ast.Location{
								Line: int(38),
								Column: int(24),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "",
							Begin: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							End: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							File: nil,
						},
						Fodder: nil,
						Ctx: nil,
						FreeVars: ast.Identifiers{
							"from",
							"len",
							"std",
							"str",
						},
					},
					Cond: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(39),
									Column: int(12),
								},
								End: ast.Location{
									Line: int(39),
									Column: int(29),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p160,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
						},
						Target: &ast.Index{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(39),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(39),
										Column: int(24),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p160,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(39),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(39),
											Column: int(15),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Id: "std",
							},
							LeftBracketFodder: ast.Fodder{},
							Index: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Value: "isString",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							RightBracketFodder: ast.Fodder{},
							Id: nil,
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(39),
													Column: int(25),
												},
												End: ast.Location{
													Line: int(39),
													Column: int(28),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p169,
											FreeVars: ast.Identifiers{
												"str",
											},
										},
										Id: "str",
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
					ThenFodder: nil,
					BranchTrue: &ast.Conditional{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"from",
								"len",
								"std",
								"str",
							},
						},
						Cond: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(40),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(40),
										Column: int(30),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p160,
								FreeVars: ast.Identifiers{
									"from",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(40),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(40),
											Column: int(24),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p160,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(40),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(40),
												Column: int(15),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "isNumber",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(40),
														Column: int(25),
													},
													End: ast.Location{
														Line: int(40),
														Column: int(29),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p182,
												FreeVars: ast.Identifiers{
													"from",
												},
											},
											Id: "from",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
						ThenFodder: nil,
						BranchTrue: &ast.Conditional{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"from",
									"len",
									"std",
									"str",
								},
							},
							Cond: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(41),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(41),
											Column: int(29),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p160,
									FreeVars: ast.Identifiers{
										"len",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(41),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(41),
												Column: int(24),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p160,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(41),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(41),
													Column: int(15),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "isNumber",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(41),
															Column: int(25),
														},
														End: ast.Location{
															Line: int(41),
															Column: int(28),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p195,
													FreeVars: ast.Identifiers{
														"len",
													},
												},
												Id: "len",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
							ThenFodder: nil,
							BranchTrue: &ast.Conditional{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"from",
										"len",
										"std",
										"str",
									},
								},
								Cond: &ast.Binary{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(42),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(42),
												Column: int(19),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p160,
										FreeVars: ast.Identifiers{
											"len",
										},
									},
									Left: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(42),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(42),
													Column: int(15),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p160,
											FreeVars: ast.Identifiers{
												"len",
											},
										},
										Id: "len",
									},
									OpFodder: ast.Fodder{},
									Op: ast.BinaryOp(8),
									Right: &ast.LiteralNumber{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(42),
													Column: int(18),
												},
												End: ast.Location{
													Line: int(42),
													Column: int(19),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p160,
											FreeVars: nil,
										},
										OriginalString: "0",
									},
								},
								ThenFodder: nil,
								BranchTrue: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(43),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(43),
												Column: int(109),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p160,
										FreeVars: ast.Identifiers{
											"from",
											"len",
											"std",
											"str",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(43),
													Column: int(5),
												},
												End: ast.Location{
													Line: int(43),
													Column: int(13),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p160,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(43),
														Column: int(5),
													},
													End: ast.Location{
														Line: int(43),
														Column: int(8),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(4),
														Comment: []string{},
													},
												},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "join",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(43),
																Column: int(14),
															},
															End: ast.Location{
																Line: int(43),
																Column: int(16),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p214,
														FreeVars: nil,
													},
													Value: "",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(43),
																Column: int(18),
															},
															End: ast.Location{
																Line: int(43),
																Column: int(108),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p214,
														FreeVars: ast.Identifiers{
															"from",
															"len",
															"std",
															"str",
														},
													},
													Target: &ast.Index{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(43),
																	Column: int(18),
																},
																End: ast.Location{
																	Line: int(43),
																	Column: int(31),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p214,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(43),
																		Column: int(18),
																	},
																	End: ast.Location{
																		Line: int(43),
																		Column: int(21),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Id: "std",
														},
														LeftBracketFodder: ast.Fodder{},
														Index: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
															Value: "makeArray",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														RightBracketFodder: ast.Fodder{},
														Id: nil,
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Apply{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(43),
																				Column: int(32),
																			},
																			End: ast.Location{
																				Line: int(43),
																				Column: int(80),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p224,
																		FreeVars: ast.Identifiers{
																			"from",
																			"len",
																			"std",
																			"str",
																		},
																	},
																	Target: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(43),
																					Column: int(32),
																				},
																				End: ast.Location{
																					Line: int(43),
																					Column: int(39),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p224,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(43),
																						Column: int(32),
																					},
																					End: ast.Location{
																						Line: int(43),
																						Column: int(35),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Id: "std",
																		},
																		LeftBracketFodder: ast.Fodder{},
																		Index: &ast.LiteralString{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: nil,
																			},
																			Value: "max",
																			Kind: ast.LiteralStringKind(1),
																			BlockIndent: "",
																			BlockTermIndent: "",
																		},
																		RightBracketFodder: ast.Fodder{},
																		Id: nil,
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.LiteralNumber{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(43),
																								Column: int(40),
																							},
																							End: ast.Location{
																								Line: int(43),
																								Column: int(41),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p233,
																						FreeVars: nil,
																					},
																					OriginalString: "0",
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Apply{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(43),
																								Column: int(43),
																							},
																							End: ast.Location{
																								Line: int(43),
																								Column: int(79),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p233,
																						FreeVars: ast.Identifiers{
																							"from",
																							"len",
																							"std",
																							"str",
																						},
																					},
																					Target: &ast.Index{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(43),
																									Column: int(43),
																								},
																								End: ast.Location{
																									Line: int(43),
																									Column: int(50),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p233,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																						},
																						Target: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(43),
																										Column: int(43),
																									},
																									End: ast.Location{
																										Line: int(43),
																										Column: int(46),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: nil,
																								FreeVars: ast.Identifiers{
																									"std",
																								},
																							},
																							Id: "std",
																						},
																						LeftBracketFodder: ast.Fodder{},
																						Index: &ast.LiteralString{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									File: nil,
																								},
																								Fodder: nil,
																								Ctx: nil,
																								FreeVars: nil,
																							},
																							Value: "min",
																							Kind: ast.LiteralStringKind(1),
																							BlockIndent: "",
																							BlockTermIndent: "",
																						},
																						RightBracketFodder: ast.Fodder{},
																						Id: nil,
																					},
																					FodderLeft: ast.Fodder{},
																					Arguments: ast.Arguments{
																						Positional: []ast.CommaSeparatedExpr{
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(43),
																												Column: int(51),
																											},
																											End: ast.Location{
																												Line: int(43),
																												Column: int(54),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p243,
																										FreeVars: ast.Identifiers{
																											"len",
																										},
																									},
																									Id: "len",
																								},
																								CommaFodder: ast.Fodder{},
																							},
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Binary{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(43),
																												Column: int(56),
																											},
																											End: ast.Location{
																												Line: int(43),
																												Column: int(78),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p243,
																										FreeVars: ast.Identifiers{
																											"from",
																											"std",
																											"str",
																										},
																									},
																									Left: &ast.Apply{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(43),
																													Column: int(56),
																												},
																												End: ast.Location{
																													Line: int(43),
																													Column: int(71),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p243,
																											FreeVars: ast.Identifiers{
																												"std",
																												"str",
																											},
																										},
																										Target: &ast.Index{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(43),
																														Column: int(56),
																													},
																													End: ast.Location{
																														Line: int(43),
																														Column: int(66),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p243,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																											},
																											Target: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(43),
																															Column: int(56),
																														},
																														End: ast.Location{
																															Line: int(43),
																															Column: int(59),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																												},
																												Id: "std",
																											},
																											LeftBracketFodder: ast.Fodder{},
																											Index: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: nil,
																												},
																												Value: "length",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																											RightBracketFodder: ast.Fodder{},
																											Id: nil,
																										},
																										FodderLeft: ast.Fodder{},
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(43),
																																	Column: int(67),
																																},
																																End: ast.Location{
																																	Line: int(43),
																																	Column: int(70),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p256,
																															FreeVars: ast.Identifiers{
																																"str",
																															},
																														},
																														Id: "str",
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										TrailingComma: false,
																										TailStrict: false,
																										FodderRight: ast.Fodder{},
																										TailStrictFodder: nil,
																									},
																									OpFodder: ast.Fodder{},
																									Op: ast.BinaryOp(4),
																									Right: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(43),
																													Column: int(74),
																												},
																												End: ast.Location{
																													Line: int(43),
																													Column: int(78),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p243,
																											FreeVars: ast.Identifiers{
																												"from",
																											},
																										},
																										Id: "from",
																									},
																								},
																								CommaFodder: nil,
																							},
																						},
																						Named: nil,
																					},
																					TrailingComma: false,
																					TailStrict: false,
																					FodderRight: ast.Fodder{},
																					TailStrictFodder: nil,
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	TrailingComma: false,
																	TailStrict: false,
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Function{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(43),
																				Column: int(82),
																			},
																			End: ast.Location{
																				Line: int(43),
																				Column: int(107),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p224,
																		FreeVars: ast.Identifiers{
																			"from",
																			"str",
																		},
																	},
																	ParenLeftFodder: ast.Fodder{},
																	Parameters: []ast.Parameter{
																		ast.Parameter{
																			NameFodder: ast.Fodder{},
																			Name: "i",
																			EqFodder: nil,
																			DefaultArg: nil,
																			CommaFodder: nil,
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(43),
																					Column: int(91),
																				},
																				End: ast.Location{
																					Line: int(43),
																					Column: int(92),
																				},
																				File: p1,
																			},
																		},
																	},
																	TrailingComma: false,
																	ParenRightFodder: ast.Fodder{},
																	Body: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(43),
																					Column: int(94),
																				},
																				End: ast.Location{
																					Line: int(43),
																					Column: int(107),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p264,
																			FreeVars: ast.Identifiers{
																				"from",
																				"i",
																				"str",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(43),
																						Column: int(94),
																					},
																					End: ast.Location{
																						Line: int(43),
																						Column: int(97),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p264,
																				FreeVars: ast.Identifiers{
																					"str",
																				},
																			},
																			Id: "str",
																		},
																		LeftBracketFodder: ast.Fodder{},
																		Index: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(43),
																						Column: int(98),
																					},
																					End: ast.Location{
																						Line: int(43),
																						Column: int(106),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p264,
																				FreeVars: ast.Identifiers{
																					"from",
																					"i",
																				},
																			},
																			Left: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(43),
																							Column: int(98),
																						},
																						End: ast.Location{
																							Line: int(43),
																							Column: int(99),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p264,
																					FreeVars: ast.Identifiers{
																						"i",
																					},
																				},
																				Id: "i",
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(3),
																			Right: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(43),
																							Column: int(102),
																						},
																						End: ast.Location{
																							Line: int(43),
																							Column: int(106),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p264,
																					FreeVars: ast.Identifiers{
																						"from",
																					},
																				},
																				Id: "from",
																			},
																		},
																		RightBracketFodder: ast.Fodder{},
																		Id: nil,
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
								ElseFodder: nil,
								BranchFalse: &ast.Error{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"len",
										},
									},
									Expr: &ast.Binary{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(42),
													Column: int(22),
												},
												End: ast.Location{
													Line: int(42),
													Column: int(86),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p160,
											FreeVars: ast.Identifiers{
												"len",
											},
										},
										Left: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(42),
														Column: int(22),
													},
													End: ast.Location{
														Line: int(42),
														Column: int(80),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p160,
												FreeVars: nil,
											},
											Value: "substr third parameter should be greater than zero, got ",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										OpFodder: ast.Fodder{},
										Op: ast.BinaryOp(3),
										Right: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(42),
														Column: int(83),
													},
													End: ast.Location{
														Line: int(42),
														Column: int(86),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p160,
												FreeVars: ast.Identifiers{
													"len",
												},
											},
											Id: "len",
										},
									},
								},
							},
							ElseFodder: nil,
							BranchFalse: &ast.Error{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"len",
										"std",
									},
								},
								Expr: &ast.Binary{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(41),
												Column: int(32),
											},
											End: ast.Location{
												Line: int(41),
												Column: int(97),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p160,
										FreeVars: ast.Identifiers{
											"len",
											"std",
										},
									},
									Left: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(41),
													Column: int(32),
												},
												End: ast.Location{
													Line: int(41),
													Column: int(81),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p160,
											FreeVars: nil,
										},
										Value: "substr third parameter should be a string, got ",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									OpFodder: ast.Fodder{},
									Op: ast.BinaryOp(3),
									Right: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(41),
													Column: int(84),
												},
												End: ast.Location{
													Line: int(41),
													Column: int(97),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p160,
											FreeVars: ast.Identifiers{
												"len",
												"std",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(41),
														Column: int(84),
													},
													End: ast.Location{
														Line: int(41),
														Column: int(92),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p160,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(41),
															Column: int(84),
														},
														End: ast.Location{
															Line: int(41),
															Column: int(87),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: ast.Fodder{},
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "type",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: ast.Fodder{},
											Id: nil,
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(41),
																	Column: int(93),
																},
																End: ast.Location{
																	Line: int(41),
																	Column: int(96),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p295,
															FreeVars: ast.Identifiers{
																"len",
															},
														},
														Id: "len",
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
								},
							},
						},
						ElseFodder: nil,
						BranchFalse: &ast.Error{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"from",
									"std",
								},
							},
							Expr: &ast.Binary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(40),
											Column: int(33),
										},
										End: ast.Location{
											Line: int(40),
											Column: int(100),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p160,
									FreeVars: ast.Identifiers{
										"from",
										"std",
									},
								},
								Left: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(40),
												Column: int(33),
											},
											End: ast.Location{
												Line: int(40),
												Column: int(83),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p160,
										FreeVars: nil,
									},
									Value: "substr second parameter should be a string, got ",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								OpFodder: ast.Fodder{},
								Op: ast.BinaryOp(3),
								Right: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(40),
												Column: int(86),
											},
											End: ast.Location{
												Line: int(40),
												Column: int(100),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p160,
										FreeVars: ast.Identifiers{
											"from",
											"std",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(40),
													Column: int(86),
												},
												End: ast.Location{
													Line: int(40),
													Column: int(94),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p160,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(40),
														Column: int(86),
													},
													End: ast.Location{
														Line: int(40),
														Column: int(89),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "type",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(40),
																Column: int(95),
															},
															End: ast.Location{
																Line: int(40),
																Column: int(99),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p311,
														FreeVars: ast.Identifiers{
															"from",
														},
													},
													Id: "from",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
							},
						},
					},
					ElseFodder: nil,
					BranchFalse: &ast.Error{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
						},
						Expr: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(39),
										Column: int(32),
									},
									End: ast.Location{
										Line: int(39),
										Column: int(97),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p160,
								FreeVars: ast.Identifiers{
									"std",
									"str",
								},
							},
							Left: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(39),
											Column: int(32),
										},
										End: ast.Location{
											Line: int(39),
											Column: int(81),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p160,
									FreeVars: nil,
								},
								Value: "substr first parameter should be a string, got ",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(3),
							Right: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(39),
											Column: int(84),
										},
										End: ast.Location{
											Line: int(39),
											Column: int(97),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p160,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(39),
												Column: int(84),
											},
											End: ast.Location{
												Line: int(39),
												Column: int(92),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p160,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(39),
													Column: int(84),
												},
												End: ast.Location{
													Line: int(39),
													Column: int(87),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "type",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(39),
															Column: int(93),
														},
														End: ast.Location{
															Line: int(39),
															Column: int(96),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p327,
													FreeVars: ast.Identifiers{
														"str",
													},
												},
												Id: "str",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(38),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(43),
					Column: int(109),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "startsWith",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "a",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(45),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(45),
								Column: int(15),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "b",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(45),
								Column: int(17),
							},
							End: ast.Location{
								Line: int(45),
								Column: int(18),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(46),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(49),
								Column: int(43),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p335,
						FreeVars: ast.Identifiers{
							"a",
							"b",
							"std",
						},
					},
					Cond: &ast.Binary{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(46),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(46),
									Column: int(37),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p335,
							FreeVars: ast.Identifiers{
								"a",
								"b",
								"std",
							},
						},
						Left: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(46),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(46),
										Column: int(21),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p335,
								FreeVars: ast.Identifiers{
									"a",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(46),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(46),
											Column: int(18),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p335,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(46),
												Column: int(8),
											},
											End: ast.Location{
												Line: int(46),
												Column: int(11),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "length",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(46),
														Column: int(19),
													},
													End: ast.Location{
														Line: int(46),
														Column: int(20),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p348,
												FreeVars: ast.Identifiers{
													"a",
												},
											},
											Id: "a",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
						OpFodder: ast.Fodder{},
						Op: ast.BinaryOp(9),
						Right: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(46),
										Column: int(24),
									},
									End: ast.Location{
										Line: int(46),
										Column: int(37),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p335,
								FreeVars: ast.Identifiers{
									"b",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(46),
											Column: int(24),
										},
										End: ast.Location{
											Line: int(46),
											Column: int(34),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p335,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(46),
												Column: int(24),
											},
											End: ast.Location{
												Line: int(46),
												Column: int(27),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "length",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(46),
														Column: int(35),
													},
													End: ast.Location{
														Line: int(46),
														Column: int(36),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p359,
												FreeVars: ast.Identifiers{
													"b",
												},
											},
											Id: "b",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
					},
					ThenFodder: ast.Fodder{},
					BranchTrue: &ast.LiteralBoolean{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(47),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(47),
									Column: int(12),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
									Comment: []string{},
								},
							},
							Ctx: p335,
							FreeVars: nil,
						},
						Value: false,
					},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
							Comment: []string{},
						},
					},
					BranchFalse: &ast.Binary{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(49),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(49),
									Column: int(43),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p335,
							FreeVars: ast.Identifiers{
								"a",
								"b",
								"std",
							},
						},
						Left: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(49),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(49),
										Column: int(38),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p335,
								FreeVars: ast.Identifiers{
									"a",
									"b",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(49),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(49),
											Column: int(17),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p335,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(49),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(49),
												Column: int(10),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "substr",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(49),
														Column: int(18),
													},
													End: ast.Location{
														Line: int(49),
														Column: int(19),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p376,
												FreeVars: ast.Identifiers{
													"a",
												},
											},
											Id: "a",
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralNumber{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(49),
														Column: int(21),
													},
													End: ast.Location{
														Line: int(49),
														Column: int(22),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p376,
												FreeVars: nil,
											},
											OriginalString: "0",
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(49),
														Column: int(24),
													},
													End: ast.Location{
														Line: int(49),
														Column: int(37),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p376,
												FreeVars: ast.Identifiers{
													"b",
													"std",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(49),
															Column: int(24),
														},
														End: ast.Location{
															Line: int(49),
															Column: int(34),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p376,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(49),
																Column: int(24),
															},
															End: ast.Location{
																Line: int(49),
																Column: int(27),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "length",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(49),
																		Column: int(35),
																	},
																	End: ast.Location{
																		Line: int(49),
																		Column: int(36),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p388,
																FreeVars: ast.Identifiers{
																	"b",
																},
															},
															Id: "b",
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
						OpFodder: ast.Fodder{},
						Op: ast.BinaryOp(12),
						Right: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(49),
										Column: int(42),
									},
									End: ast.Location{
										Line: int(49),
										Column: int(43),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p335,
								FreeVars: ast.Identifiers{
									"b",
								},
							},
							Id: "b",
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(45),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(49),
					Column: int(43),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "endsWith",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "a",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(51),
								Column: int(12),
							},
							End: ast.Location{
								Line: int(51),
								Column: int(13),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "b",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(51),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(51),
								Column: int(16),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(52),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(55),
								Column: int(71),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p398,
						FreeVars: ast.Identifiers{
							"a",
							"b",
							"std",
						},
					},
					Cond: &ast.Binary{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(52),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(52),
									Column: int(37),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p398,
							FreeVars: ast.Identifiers{
								"a",
								"b",
								"std",
							},
						},
						Left: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(52),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(52),
										Column: int(21),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p398,
								FreeVars: ast.Identifiers{
									"a",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(52),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(52),
											Column: int(18),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p398,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(52),
												Column: int(8),
											},
											End: ast.Location{
												Line: int(52),
												Column: int(11),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "length",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(52),
														Column: int(19),
													},
													End: ast.Location{
														Line: int(52),
														Column: int(20),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p411,
												FreeVars: ast.Identifiers{
													"a",
												},
											},
											Id: "a",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
						OpFodder: ast.Fodder{},
						Op: ast.BinaryOp(9),
						Right: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(52),
										Column: int(24),
									},
									End: ast.Location{
										Line: int(52),
										Column: int(37),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p398,
								FreeVars: ast.Identifiers{
									"b",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(52),
											Column: int(24),
										},
										End: ast.Location{
											Line: int(52),
											Column: int(34),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p398,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(52),
												Column: int(24),
											},
											End: ast.Location{
												Line: int(52),
												Column: int(27),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "length",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(52),
														Column: int(35),
													},
													End: ast.Location{
														Line: int(52),
														Column: int(36),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p422,
												FreeVars: ast.Identifiers{
													"b",
												},
											},
											Id: "b",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
					},
					ThenFodder: ast.Fodder{},
					BranchTrue: &ast.LiteralBoolean{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(53),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(53),
									Column: int(12),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
									Comment: []string{},
								},
							},
							Ctx: p398,
							FreeVars: nil,
						},
						Value: false,
					},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
							Comment: []string{},
						},
					},
					BranchFalse: &ast.Binary{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(55),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(55),
									Column: int(71),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p398,
							FreeVars: ast.Identifiers{
								"a",
								"b",
								"std",
							},
						},
						Left: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(55),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(55),
										Column: int(66),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p398,
								FreeVars: ast.Identifiers{
									"a",
									"b",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(55),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(55),
											Column: int(17),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p398,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(55),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(55),
												Column: int(10),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "substr",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(55),
														Column: int(18),
													},
													End: ast.Location{
														Line: int(55),
														Column: int(19),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p439,
												FreeVars: ast.Identifiers{
													"a",
												},
											},
											Id: "a",
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Binary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(55),
														Column: int(21),
													},
													End: ast.Location{
														Line: int(55),
														Column: int(50),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p439,
												FreeVars: ast.Identifiers{
													"a",
													"b",
													"std",
												},
											},
											Left: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(55),
															Column: int(21),
														},
														End: ast.Location{
															Line: int(55),
															Column: int(34),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p439,
													FreeVars: ast.Identifiers{
														"a",
														"std",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(55),
																Column: int(21),
															},
															End: ast.Location{
																Line: int(55),
																Column: int(31),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p439,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(55),
																	Column: int(21),
																},
																End: ast.Location{
																	Line: int(55),
																	Column: int(24),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: ast.Fodder{},
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "length",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: ast.Fodder{},
													Id: nil,
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(55),
																			Column: int(32),
																		},
																		End: ast.Location{
																			Line: int(55),
																			Column: int(33),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p452,
																	FreeVars: ast.Identifiers{
																		"a",
																	},
																},
																Id: "a",
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
											OpFodder: ast.Fodder{},
											Op: ast.BinaryOp(4),
											Right: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(55),
															Column: int(37),
														},
														End: ast.Location{
															Line: int(55),
															Column: int(50),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p439,
													FreeVars: ast.Identifiers{
														"b",
														"std",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(55),
																Column: int(37),
															},
															End: ast.Location{
																Line: int(55),
																Column: int(47),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p439,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(55),
																	Column: int(37),
																},
																End: ast.Location{
																	Line: int(55),
																	Column: int(40),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: ast.Fodder{},
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "length",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: ast.Fodder{},
													Id: nil,
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(55),
																			Column: int(48),
																		},
																		End: ast.Location{
																			Line: int(55),
																			Column: int(49),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p463,
																	FreeVars: ast.Identifiers{
																		"b",
																	},
																},
																Id: "b",
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(55),
														Column: int(52),
													},
													End: ast.Location{
														Line: int(55),
														Column: int(65),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p439,
												FreeVars: ast.Identifiers{
													"b",
													"std",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(55),
															Column: int(52),
														},
														End: ast.Location{
															Line: int(55),
															Column: int(62),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p439,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(55),
																Column: int(52),
															},
															End: ast.Location{
																Line: int(55),
																Column: int(55),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "length",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(55),
																		Column: int(63),
																	},
																	End: ast.Location{
																		Line: int(55),
																		Column: int(64),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p474,
																FreeVars: ast.Identifiers{
																	"b",
																},
															},
															Id: "b",
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
						OpFodder: ast.Fodder{},
						Op: ast.BinaryOp(12),
						Right: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(55),
										Column: int(70),
									},
									End: ast.Location{
										Line: int(55),
										Column: int(71),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p398,
								FreeVars: ast.Identifiers{
									"b",
								},
							},
							Id: "b",
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(51),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(55),
					Column: int(71),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "lstripChars",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(57),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(57),
								Column: int(18),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "chars",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(57),
								Column: int(20),
							},
							End: ast.Location{
								Line: int(57),
								Column: int(25),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(58),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(61),
								Column: int(10),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p484,
						FreeVars: ast.Identifiers{
							"chars",
							"std",
							"str",
						},
					},
					Cond: &ast.Binary{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(58),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(58),
									Column: int(56),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p484,
							FreeVars: ast.Identifiers{
								"chars",
								"std",
								"str",
							},
						},
						Left: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(58),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(58),
										Column: int(27),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p484,
								FreeVars: ast.Identifiers{
									"std",
									"str",
								},
							},
							Left: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(58),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(58),
											Column: int(23),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p484,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(58),
												Column: int(8),
											},
											End: ast.Location{
												Line: int(58),
												Column: int(18),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p484,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(58),
													Column: int(8),
												},
												End: ast.Location{
													Line: int(58),
													Column: int(11),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "length",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(58),
															Column: int(19),
														},
														End: ast.Location{
															Line: int(58),
															Column: int(22),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p499,
													FreeVars: ast.Identifiers{
														"str",
													},
												},
												Id: "str",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(7),
							Right: &ast.LiteralNumber{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(58),
											Column: int(26),
										},
										End: ast.Location{
											Line: int(58),
											Column: int(27),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p484,
									FreeVars: nil,
								},
								OriginalString: "0",
							},
						},
						OpFodder: ast.Fodder{},
						Op: ast.BinaryOp(17),
						Right: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(58),
										Column: int(31),
									},
									End: ast.Location{
										Line: int(58),
										Column: int(56),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p484,
								FreeVars: ast.Identifiers{
									"chars",
									"std",
									"str",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(58),
											Column: int(31),
										},
										End: ast.Location{
											Line: int(58),
											Column: int(41),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p484,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(58),
												Column: int(31),
											},
											End: ast.Location{
												Line: int(58),
												Column: int(34),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "member",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(58),
														Column: int(42),
													},
													End: ast.Location{
														Line: int(58),
														Column: int(47),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p511,
												FreeVars: ast.Identifiers{
													"chars",
												},
											},
											Id: "chars",
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(58),
														Column: int(49),
													},
													End: ast.Location{
														Line: int(58),
														Column: int(55),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p511,
												FreeVars: ast.Identifiers{
													"str",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(58),
															Column: int(49),
														},
														End: ast.Location{
															Line: int(58),
															Column: int(52),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p511,
													FreeVars: ast.Identifiers{
														"str",
													},
												},
												Id: "str",
											},
											LeftBracketFodder: ast.Fodder{},
											Index: &ast.LiteralNumber{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(58),
															Column: int(53),
														},
														End: ast.Location{
															Line: int(58),
															Column: int(54),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p511,
													FreeVars: nil,
												},
												OriginalString: "0",
											},
											RightBracketFodder: ast.Fodder{},
											Id: nil,
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
					},
					ThenFodder: ast.Fodder{},
					BranchTrue: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(59),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(59),
									Column: int(38),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p484,
							FreeVars: ast.Identifiers{
								"chars",
								"std",
								"str",
							},
						},
						Target: &ast.Index{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(59),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(59),
										Column: int(22),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p484,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(59),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(59),
											Column: int(10),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
											Comment: []string{},
										},
									},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Id: "std",
							},
							LeftBracketFodder: ast.Fodder{},
							Index: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Value: "lstripChars",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							RightBracketFodder: ast.Fodder{},
							Id: nil,
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
												"str",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: nil,
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "slice",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: nil,
											Id: nil,
										},
										FodderLeft: nil,
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(59),
																	Column: int(23),
																},
																End: ast.Location{
																	Line: int(59),
																	Column: int(26),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p536,
															FreeVars: ast.Identifiers{
																"str",
															},
														},
														Id: "str",
													},
													CommaFodder: nil,
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.LiteralNumber{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(59),
																	Column: int(27),
																},
																End: ast.Location{
																	Line: int(59),
																	Column: int(28),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p536,
															FreeVars: nil,
														},
														OriginalString: "1",
													},
													CommaFodder: nil,
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.LiteralNull{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
													},
													CommaFodder: nil,
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.LiteralNull{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: nil,
										TailStrictFodder: nil,
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(59),
													Column: int(32),
												},
												End: ast.Location{
													Line: int(59),
													Column: int(37),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p536,
											FreeVars: ast.Identifiers{
												"chars",
											},
										},
										Id: "chars",
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
							Comment: []string{},
						},
					},
					BranchFalse: &ast.Var{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(61),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(61),
									Column: int(10),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
									Comment: []string{},
								},
							},
							Ctx: p484,
							FreeVars: ast.Identifiers{
								"str",
							},
						},
						Id: "str",
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(57),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(61),
					Column: int(10),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "rstripChars",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(63),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(63),
								Column: int(18),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "chars",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(63),
								Column: int(20),
							},
							End: ast.Location{
								Line: int(63),
								Column: int(25),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(64),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(68),
								Column: int(10),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p553,
						FreeVars: ast.Identifiers{
							"chars",
							"std",
							"str",
						},
					},
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Variable: "len",
							EqFodder: ast.Fodder{},
							Body: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(64),
											Column: int(17),
										},
										End: ast.Location{
											Line: int(64),
											Column: int(32),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p557,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(64),
												Column: int(17),
											},
											End: ast.Location{
												Line: int(64),
												Column: int(27),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p557,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(64),
													Column: int(17),
												},
												End: ast.Location{
													Line: int(64),
													Column: int(20),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "length",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(64),
															Column: int(28),
														},
														End: ast.Location{
															Line: int(64),
															Column: int(31),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p566,
													FreeVars: ast.Identifiers{
														"str",
													},
												},
												Id: "str",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
							Fun: nil,
							CloseFodder: ast.Fodder{},
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(64),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(64),
									Column: int(32),
								},
								File: p1,
							},
						},
					},
					Body: &ast.Conditional{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(65),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(68),
									Column: int(10),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
									Comment: []string{},
								},
							},
							Ctx: p553,
							FreeVars: ast.Identifiers{
								"chars",
								"len",
								"std",
								"str",
							},
						},
						Cond: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(65),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(65),
										Column: int(50),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p553,
								FreeVars: ast.Identifiers{
									"chars",
									"len",
									"std",
									"str",
								},
							},
							Left: &ast.Binary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(65),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(65),
											Column: int(15),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p553,
									FreeVars: ast.Identifiers{
										"len",
									},
								},
								Left: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(65),
												Column: int(8),
											},
											End: ast.Location{
												Line: int(65),
												Column: int(11),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p553,
										FreeVars: ast.Identifiers{
											"len",
										},
									},
									Id: "len",
								},
								OpFodder: ast.Fodder{},
								Op: ast.BinaryOp(7),
								Right: &ast.LiteralNumber{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(65),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(65),
												Column: int(15),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p553,
										FreeVars: nil,
									},
									OriginalString: "0",
								},
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(17),
							Right: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(65),
											Column: int(19),
										},
										End: ast.Location{
											Line: int(65),
											Column: int(50),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p553,
									FreeVars: ast.Identifiers{
										"chars",
										"len",
										"std",
										"str",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(65),
												Column: int(19),
											},
											End: ast.Location{
												Line: int(65),
												Column: int(29),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p553,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(65),
													Column: int(19),
												},
												End: ast.Location{
													Line: int(65),
													Column: int(22),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "member",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(65),
															Column: int(30),
														},
														End: ast.Location{
															Line: int(65),
															Column: int(35),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p587,
													FreeVars: ast.Identifiers{
														"chars",
													},
												},
												Id: "chars",
											},
											CommaFodder: ast.Fodder{},
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(65),
															Column: int(37),
														},
														End: ast.Location{
															Line: int(65),
															Column: int(49),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p587,
													FreeVars: ast.Identifiers{
														"len",
														"str",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(65),
																Column: int(37),
															},
															End: ast.Location{
																Line: int(65),
																Column: int(40),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p587,
														FreeVars: ast.Identifiers{
															"str",
														},
													},
													Id: "str",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.Binary{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(65),
																Column: int(41),
															},
															End: ast.Location{
																Line: int(65),
																Column: int(48),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p587,
														FreeVars: ast.Identifiers{
															"len",
														},
													},
													Left: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(65),
																	Column: int(41),
																},
																End: ast.Location{
																	Line: int(65),
																	Column: int(44),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p587,
															FreeVars: ast.Identifiers{
																"len",
															},
														},
														Id: "len",
													},
													OpFodder: ast.Fodder{},
													Op: ast.BinaryOp(4),
													Right: &ast.LiteralNumber{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(65),
																	Column: int(47),
																},
																End: ast.Location{
																	Line: int(65),
																	Column: int(48),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p587,
															FreeVars: nil,
														},
														OriginalString: "1",
													},
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
						},
						ThenFodder: ast.Fodder{},
						BranchTrue: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(66),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(66),
										Column: int(44),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p553,
								FreeVars: ast.Identifiers{
									"chars",
									"len",
									"std",
									"str",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(66),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(66),
											Column: int(22),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p553,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(66),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(66),
												Column: int(10),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "rstripChars",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"len",
													"std",
													"str",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: nil,
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "slice",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: nil,
												Id: nil,
											},
											FodderLeft: nil,
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(66),
																		Column: int(23),
																	},
																	End: ast.Location{
																		Line: int(66),
																		Column: int(26),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p616,
																FreeVars: ast.Identifiers{
																	"str",
																},
															},
															Id: "str",
														},
														CommaFodder: nil,
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.LiteralNull{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
														},
														CommaFodder: nil,
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Binary{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(66),
																		Column: int(28),
																	},
																	End: ast.Location{
																		Line: int(66),
																		Column: int(35),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p616,
																FreeVars: ast.Identifiers{
																	"len",
																},
															},
															Left: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(66),
																			Column: int(28),
																		},
																		End: ast.Location{
																			Line: int(66),
																			Column: int(31),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p616,
																	FreeVars: ast.Identifiers{
																		"len",
																	},
																},
																Id: "len",
															},
															OpFodder: ast.Fodder{},
															Op: ast.BinaryOp(4),
															Right: &ast.LiteralNumber{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(66),
																			Column: int(34),
																		},
																		End: ast.Location{
																			Line: int(66),
																			Column: int(35),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p616,
																	FreeVars: nil,
																},
																OriginalString: "1",
															},
														},
														CommaFodder: nil,
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.LiteralNull{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: nil,
											TailStrictFodder: nil,
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(66),
														Column: int(38),
													},
													End: ast.Location{
														Line: int(66),
														Column: int(43),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p616,
												FreeVars: ast.Identifiers{
													"chars",
												},
											},
											Id: "chars",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						BranchFalse: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(68),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(68),
										Column: int(10),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
										Comment: []string{},
									},
								},
								Ctx: p553,
								FreeVars: ast.Identifiers{
									"str",
								},
							},
							Id: "str",
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(63),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(68),
					Column: int(10),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "stripChars",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(70),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(70),
								Column: int(17),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "chars",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(70),
								Column: int(19),
							},
							End: ast.Location{
								Line: int(70),
								Column: int(24),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(71),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(71),
								Column: int(56),
							},
							File: p1,
						},
						Fodder: ast.Fodder{},
						Ctx: p636,
						FreeVars: ast.Identifiers{
							"chars",
							"std",
							"str",
						},
					},
					Target: &ast.Index{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(71),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(71),
									Column: int(20),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p636,
							FreeVars: ast.Identifiers{
								"std",
							},
						},
						Target: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(71),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(71),
										Column: int(8),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
										Comment: []string{},
									},
								},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Id: "std",
						},
						LeftBracketFodder: ast.Fodder{},
						Index: &ast.LiteralString{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: nil,
							},
							Value: "lstripChars",
							Kind: ast.LiteralStringKind(1),
							BlockIndent: "",
							BlockTermIndent: "",
						},
						RightBracketFodder: ast.Fodder{},
						Id: nil,
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(71),
												Column: int(21),
											},
											End: ast.Location{
												Line: int(71),
												Column: int(48),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p646,
										FreeVars: ast.Identifiers{
											"chars",
											"std",
											"str",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(71),
													Column: int(21),
												},
												End: ast.Location{
													Line: int(71),
													Column: int(36),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p646,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(71),
														Column: int(21),
													},
													End: ast.Location{
														Line: int(71),
														Column: int(24),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "rstripChars",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(71),
																Column: int(37),
															},
															End: ast.Location{
																Line: int(71),
																Column: int(40),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p655,
														FreeVars: ast.Identifiers{
															"str",
														},
													},
													Id: "str",
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(71),
																Column: int(42),
															},
															End: ast.Location{
																Line: int(71),
																Column: int(47),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p655,
														FreeVars: ast.Identifiers{
															"chars",
														},
													},
													Id: "chars",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(71),
												Column: int(50),
											},
											End: ast.Location{
												Line: int(71),
												Column: int(55),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p646,
										FreeVars: ast.Identifiers{
											"chars",
										},
									},
									Id: "chars",
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					TrailingComma: false,
					TailStrict: false,
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(70),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(71),
					Column: int(56),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "stringChars",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(73),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(73),
								Column: int(18),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(74),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(74),
								Column: int(55),
							},
							File: p1,
						},
						Fodder: ast.Fodder{},
						Ctx: p666,
						FreeVars: ast.Identifiers{
							"std",
							"str",
						},
					},
					Target: &ast.Index{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(74),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(74),
									Column: int(18),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p666,
							FreeVars: ast.Identifiers{
								"std",
							},
						},
						Target: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(74),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(74),
										Column: int(8),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
										Comment: []string{},
									},
								},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Id: "std",
						},
						LeftBracketFodder: ast.Fodder{},
						Index: &ast.LiteralString{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: nil,
							},
							Value: "makeArray",
							Kind: ast.LiteralStringKind(1),
							BlockIndent: "",
							BlockTermIndent: "",
						},
						RightBracketFodder: ast.Fodder{},
						Id: nil,
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(74),
												Column: int(19),
											},
											End: ast.Location{
												Line: int(74),
												Column: int(34),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p676,
										FreeVars: ast.Identifiers{
											"std",
											"str",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(74),
													Column: int(19),
												},
												End: ast.Location{
													Line: int(74),
													Column: int(29),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p676,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(74),
														Column: int(19),
													},
													End: ast.Location{
														Line: int(74),
														Column: int(22),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "length",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(74),
																Column: int(30),
															},
															End: ast.Location{
																Line: int(74),
																Column: int(33),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p685,
														FreeVars: ast.Identifiers{
															"str",
														},
													},
													Id: "str",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.Function{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(74),
												Column: int(36),
											},
											End: ast.Location{
												Line: int(74),
												Column: int(54),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p676,
										FreeVars: ast.Identifiers{
											"str",
										},
									},
									ParenLeftFodder: ast.Fodder{},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "i",
											EqFodder: nil,
											DefaultArg: nil,
											CommaFodder: nil,
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(74),
													Column: int(45),
												},
												End: ast.Location{
													Line: int(74),
													Column: int(46),
												},
												File: p1,
											},
										},
									},
									TrailingComma: false,
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(74),
													Column: int(48),
												},
												End: ast.Location{
													Line: int(74),
													Column: int(54),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p691,
											FreeVars: ast.Identifiers{
												"i",
												"str",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(74),
														Column: int(48),
													},
													End: ast.Location{
														Line: int(74),
														Column: int(51),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p691,
												FreeVars: ast.Identifiers{
													"str",
												},
											},
											Id: "str",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(74),
														Column: int(52),
													},
													End: ast.Location{
														Line: int(74),
														Column: int(53),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p691,
												FreeVars: ast.Identifiers{
													"i",
												},
											},
											Id: "i",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					TrailingComma: false,
					TailStrict: false,
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(73),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(74),
					Column: int(55),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "parseInt",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"parse_nat",
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(94),
								Column: int(12),
							},
							End: ast.Location{
								Line: int(94),
								Column: int(15),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "",
							Begin: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							End: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							File: nil,
						},
						Fodder: nil,
						Ctx: nil,
						FreeVars: ast.Identifiers{
							"parse_nat",
							"std",
							"str",
						},
					},
					Cond: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(95),
									Column: int(12),
								},
								End: ast.Location{
									Line: int(95),
									Column: int(29),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p704,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
						},
						Target: &ast.Index{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(95),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(95),
										Column: int(24),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p704,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(95),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(95),
											Column: int(15),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Id: "std",
							},
							LeftBracketFodder: ast.Fodder{},
							Index: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Value: "isString",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							RightBracketFodder: ast.Fodder{},
							Id: nil,
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(95),
													Column: int(25),
												},
												End: ast.Location{
													Line: int(95),
													Column: int(28),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p713,
											FreeVars: ast.Identifiers{
												"str",
											},
										},
										Id: "str",
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
					ThenFodder: nil,
					BranchTrue: &ast.Conditional{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"parse_nat",
								"std",
								"str",
							},
						},
						Cond: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(96),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(96),
										Column: int(45),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p704,
								FreeVars: ast.Identifiers{
									"std",
									"str",
								},
							},
							Left: &ast.Binary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(96),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(96),
											Column: int(31),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p704,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
								},
								Left: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(96),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(96),
												Column: int(27),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p704,
										FreeVars: ast.Identifiers{
											"std",
											"str",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(96),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(96),
													Column: int(22),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p704,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(96),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(96),
														Column: int(15),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "length",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(96),
																Column: int(23),
															},
															End: ast.Location{
																Line: int(96),
																Column: int(26),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p730,
														FreeVars: ast.Identifiers{
															"str",
														},
													},
													Id: "str",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
								OpFodder: ast.Fodder{},
								Op: ast.BinaryOp(7),
								Right: &ast.LiteralNumber{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(96),
												Column: int(30),
											},
											End: ast.Location{
												Line: int(96),
												Column: int(31),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p704,
										FreeVars: nil,
									},
									OriginalString: "0",
								},
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(17),
							Right: &ast.Binary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(96),
											Column: int(35),
										},
										End: ast.Location{
											Line: int(96),
											Column: int(45),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p704,
									FreeVars: ast.Identifiers{
										"str",
									},
								},
								Left: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(96),
												Column: int(35),
											},
											End: ast.Location{
												Line: int(96),
												Column: int(38),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p704,
										FreeVars: ast.Identifiers{
											"str",
										},
									},
									Id: "str",
								},
								OpFodder: ast.Fodder{},
								Op: ast.BinaryOp(13),
								Right: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(96),
												Column: int(42),
											},
											End: ast.Location{
												Line: int(96),
												Column: int(45),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p704,
										FreeVars: nil,
									},
									Value: "-",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
							},
						},
						ThenFodder: nil,
						BranchTrue: &ast.Conditional{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(97),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(100),
										Column: int(25),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
										Comment: []string{},
									},
								},
								Ctx: p704,
								FreeVars: ast.Identifiers{
									"parse_nat",
									"std",
									"str",
								},
							},
							Cond: &ast.Binary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(97),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(97),
											Column: int(21),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p704,
									FreeVars: ast.Identifiers{
										"str",
									},
								},
								Left: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(97),
												Column: int(8),
											},
											End: ast.Location{
												Line: int(97),
												Column: int(14),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p704,
										FreeVars: ast.Identifiers{
											"str",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(97),
													Column: int(8),
												},
												End: ast.Location{
													Line: int(97),
													Column: int(11),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p704,
											FreeVars: ast.Identifiers{
												"str",
											},
										},
										Id: "str",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralNumber{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(97),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(97),
													Column: int(13),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p704,
											FreeVars: nil,
										},
										OriginalString: "0",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								OpFodder: ast.Fodder{},
								Op: ast.BinaryOp(12),
								Right: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(97),
												Column: int(18),
											},
											End: ast.Location{
												Line: int(97),
												Column: int(21),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p704,
										FreeVars: nil,
									},
									Value: "-",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
							},
							ThenFodder: ast.Fodder{},
							BranchTrue: &ast.Unary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(98),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(98),
											Column: int(30),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
											Comment: []string{},
										},
									},
									Ctx: p704,
									FreeVars: ast.Identifiers{
										"parse_nat",
										"std",
										"str",
									},
								},
								Op: ast.UnaryOp(3),
								Expr: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(98),
												Column: int(8),
											},
											End: ast.Location{
												Line: int(98),
												Column: int(30),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p704,
										FreeVars: ast.Identifiers{
											"parse_nat",
											"std",
											"str",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(98),
													Column: int(8),
												},
												End: ast.Location{
													Line: int(98),
													Column: int(17),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p704,
											FreeVars: ast.Identifiers{
												"parse_nat",
											},
										},
										Id: "parse_nat",
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
															"str",
														},
													},
													Target: &ast.Index{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Id: "std",
														},
														LeftBracketFodder: nil,
														Index: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
															Value: "slice",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														RightBracketFodder: nil,
														Id: nil,
													},
													FodderLeft: nil,
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(98),
																				Column: int(18),
																			},
																			End: ast.Location{
																				Line: int(98),
																				Column: int(21),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p766,
																		FreeVars: ast.Identifiers{
																			"str",
																		},
																	},
																	Id: "str",
																},
																CommaFodder: nil,
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.LiteralNumber{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(98),
																				Column: int(22),
																			},
																			End: ast.Location{
																				Line: int(98),
																				Column: int(23),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p766,
																		FreeVars: nil,
																	},
																	OriginalString: "1",
																},
																CommaFodder: nil,
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.LiteralNull{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: nil,
																	},
																},
																CommaFodder: nil,
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.LiteralNull{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: nil,
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: nil,
													TailStrictFodder: nil,
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.LiteralNumber{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(98),
																Column: int(27),
															},
															End: ast.Location{
																Line: int(98),
																Column: int(29),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p766,
														FreeVars: nil,
													},
													OriginalString: "10",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
							},
							ElseFodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
									Comment: []string{},
								},
							},
							BranchFalse: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(100),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(100),
											Column: int(25),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p704,
									FreeVars: ast.Identifiers{
										"parse_nat",
										"str",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(100),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(100),
												Column: int(16),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										Ctx: p704,
										FreeVars: ast.Identifiers{
											"parse_nat",
										},
									},
									Id: "parse_nat",
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(100),
															Column: int(17),
														},
														End: ast.Location{
															Line: int(100),
															Column: int(20),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p780,
													FreeVars: ast.Identifiers{
														"str",
													},
												},
												Id: "str",
											},
											CommaFodder: ast.Fodder{},
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.LiteralNumber{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(100),
															Column: int(22),
														},
														End: ast.Location{
															Line: int(100),
															Column: int(24),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p780,
													FreeVars: nil,
												},
												OriginalString: "10",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
						},
						ElseFodder: nil,
						BranchFalse: &ast.Error{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
									"str",
								},
							},
							Expr: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: nil,
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "mod",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: nil,
									Id: nil,
								},
								FodderLeft: nil,
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(96),
															Column: int(48),
														},
														End: ast.Location{
															Line: int(96),
															Column: int(70),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p704,
													FreeVars: nil,
												},
												Value: "Not an integer: \"%s\"",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											CommaFodder: nil,
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.Array{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(96),
															Column: int(73),
														},
														End: ast.Location{
															Line: int(96),
															Column: int(78),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p704,
													FreeVars: ast.Identifiers{
														"str",
													},
												},
												Elements: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(96),
																		Column: int(74),
																	},
																	End: ast.Location{
																		Line: int(96),
																		Column: int(77),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p798,
																FreeVars: ast.Identifiers{
																	"str",
																},
															},
															Id: "str",
														},
														CommaFodder: nil,
													},
												},
												TrailingComma: false,
												CloseFodder: ast.Fodder{},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: nil,
								TailStrictFodder: nil,
							},
						},
					},
					ElseFodder: nil,
					BranchFalse: &ast.Error{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
						},
						Expr: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(95),
										Column: int(32),
									},
									End: ast.Location{
										Line: int(95),
										Column: int(71),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p704,
								FreeVars: ast.Identifiers{
									"std",
									"str",
								},
							},
							Left: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(95),
											Column: int(32),
										},
										End: ast.Location{
											Line: int(95),
											Column: int(55),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p704,
									FreeVars: nil,
								},
								Value: "Expected string, got ",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(3),
							Right: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(95),
											Column: int(58),
										},
										End: ast.Location{
											Line: int(95),
											Column: int(71),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p704,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(95),
												Column: int(58),
											},
											End: ast.Location{
												Line: int(95),
												Column: int(66),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p704,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(95),
													Column: int(58),
												},
												End: ast.Location{
													Line: int(95),
													Column: int(61),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "type",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(95),
															Column: int(67),
														},
														End: ast.Location{
															Line: int(95),
															Column: int(70),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p814,
													FreeVars: ast.Identifiers{
														"str",
													},
												},
												Id: "str",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(94),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(100),
					Column: int(25),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "parseOctal",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"parse_nat",
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(102),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(102),
								Column: int(17),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "",
							Begin: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							End: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							File: nil,
						},
						Fodder: nil,
						Ctx: nil,
						FreeVars: ast.Identifiers{
							"parse_nat",
							"std",
							"str",
						},
					},
					Cond: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(103),
									Column: int(12),
								},
								End: ast.Location{
									Line: int(103),
									Column: int(29),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p823,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
						},
						Target: &ast.Index{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(103),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(103),
										Column: int(24),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p823,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(103),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(103),
											Column: int(15),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Id: "std",
							},
							LeftBracketFodder: ast.Fodder{},
							Index: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Value: "isString",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							RightBracketFodder: ast.Fodder{},
							Id: nil,
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(103),
													Column: int(25),
												},
												End: ast.Location{
													Line: int(103),
													Column: int(28),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p832,
											FreeVars: ast.Identifiers{
												"str",
											},
										},
										Id: "str",
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
					ThenFodder: nil,
					BranchTrue: &ast.Conditional{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"parse_nat",
								"std",
								"str",
							},
						},
						Cond: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(104),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(104),
										Column: int(31),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p823,
								FreeVars: ast.Identifiers{
									"std",
									"str",
								},
							},
							Left: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(104),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(104),
											Column: int(27),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p823,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(104),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(104),
												Column: int(22),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p823,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(104),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(104),
													Column: int(15),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "length",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(104),
															Column: int(23),
														},
														End: ast.Location{
															Line: int(104),
															Column: int(26),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p847,
													FreeVars: ast.Identifiers{
														"str",
													},
												},
												Id: "str",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(7),
							Right: &ast.LiteralNumber{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(104),
											Column: int(30),
										},
										End: ast.Location{
											Line: int(104),
											Column: int(31),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p823,
									FreeVars: nil,
								},
								OriginalString: "0",
							},
						},
						ThenFodder: nil,
						BranchTrue: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(105),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(105),
										Column: int(22),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p823,
								FreeVars: ast.Identifiers{
									"parse_nat",
									"str",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(105),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(105),
											Column: int(14),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
											Comment: []string{},
										},
									},
									Ctx: p823,
									FreeVars: ast.Identifiers{
										"parse_nat",
									},
								},
								Id: "parse_nat",
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(105),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(105),
														Column: int(18),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p857,
												FreeVars: ast.Identifiers{
													"str",
												},
											},
											Id: "str",
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralNumber{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(105),
														Column: int(20),
													},
													End: ast.Location{
														Line: int(105),
														Column: int(21),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p857,
												FreeVars: nil,
											},
											OriginalString: "8",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
						ElseFodder: nil,
						BranchFalse: &ast.Error{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: nil,
							},
							Expr: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(104),
											Column: int(34),
										},
										End: ast.Location{
											Line: int(104),
											Column: int(59),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p823,
									FreeVars: nil,
								},
								Value: "Not an octal number: \"\"",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
						},
					},
					ElseFodder: nil,
					BranchFalse: &ast.Error{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
						},
						Expr: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(103),
										Column: int(32),
									},
									End: ast.Location{
										Line: int(103),
										Column: int(71),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p823,
								FreeVars: ast.Identifiers{
									"std",
									"str",
								},
							},
							Left: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(103),
											Column: int(32),
										},
										End: ast.Location{
											Line: int(103),
											Column: int(55),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p823,
									FreeVars: nil,
								},
								Value: "Expected string, got ",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(3),
							Right: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(103),
											Column: int(58),
										},
										End: ast.Location{
											Line: int(103),
											Column: int(71),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p823,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(103),
												Column: int(58),
											},
											End: ast.Location{
												Line: int(103),
												Column: int(66),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p823,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(103),
													Column: int(58),
												},
												End: ast.Location{
													Line: int(103),
													Column: int(61),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "type",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(103),
															Column: int(67),
														},
														End: ast.Location{
															Line: int(103),
															Column: int(70),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p876,
													FreeVars: ast.Identifiers{
														"str",
													},
												},
												Id: "str",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(102),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(105),
					Column: int(22),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "parseHex",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"parse_nat",
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(107),
								Column: int(12),
							},
							End: ast.Location{
								Line: int(107),
								Column: int(15),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "",
							Begin: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							End: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							File: nil,
						},
						Fodder: nil,
						Ctx: nil,
						FreeVars: ast.Identifiers{
							"parse_nat",
							"std",
							"str",
						},
					},
					Cond: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(108),
									Column: int(12),
								},
								End: ast.Location{
									Line: int(108),
									Column: int(29),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p885,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
						},
						Target: &ast.Index{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(108),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(108),
										Column: int(24),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p885,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(108),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(108),
											Column: int(15),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Id: "std",
							},
							LeftBracketFodder: ast.Fodder{},
							Index: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Value: "isString",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							RightBracketFodder: ast.Fodder{},
							Id: nil,
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(108),
													Column: int(25),
												},
												End: ast.Location{
													Line: int(108),
													Column: int(28),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p894,
											FreeVars: ast.Identifiers{
												"str",
											},
										},
										Id: "str",
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
					ThenFodder: nil,
					BranchTrue: &ast.Conditional{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"parse_nat",
								"std",
								"str",
							},
						},
						Cond: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(109),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(109),
										Column: int(31),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p885,
								FreeVars: ast.Identifiers{
									"std",
									"str",
								},
							},
							Left: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(109),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(109),
											Column: int(27),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p885,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(109),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(109),
												Column: int(22),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p885,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(109),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(109),
													Column: int(15),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "length",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(109),
															Column: int(23),
														},
														End: ast.Location{
															Line: int(109),
															Column: int(26),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p909,
													FreeVars: ast.Identifiers{
														"str",
													},
												},
												Id: "str",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(7),
							Right: &ast.LiteralNumber{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(109),
											Column: int(30),
										},
										End: ast.Location{
											Line: int(109),
											Column: int(31),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p885,
									FreeVars: nil,
								},
								OriginalString: "0",
							},
						},
						ThenFodder: nil,
						BranchTrue: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(110),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(110),
										Column: int(23),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p885,
								FreeVars: ast.Identifiers{
									"parse_nat",
									"str",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(110),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(110),
											Column: int(14),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
											Comment: []string{},
										},
									},
									Ctx: p885,
									FreeVars: ast.Identifiers{
										"parse_nat",
									},
								},
								Id: "parse_nat",
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(110),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(110),
														Column: int(18),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p919,
												FreeVars: ast.Identifiers{
													"str",
												},
											},
											Id: "str",
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralNumber{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(110),
														Column: int(20),
													},
													End: ast.Location{
														Line: int(110),
														Column: int(22),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p919,
												FreeVars: nil,
											},
											OriginalString: "16",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
						ElseFodder: nil,
						BranchFalse: &ast.Error{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: nil,
							},
							Expr: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(109),
											Column: int(34),
										},
										End: ast.Location{
											Line: int(109),
											Column: int(55),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p885,
									FreeVars: nil,
								},
								Value: "Not hexadecimal: \"\"",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
						},
					},
					ElseFodder: nil,
					BranchFalse: &ast.Error{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
						},
						Expr: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(108),
										Column: int(32),
									},
									End: ast.Location{
										Line: int(108),
										Column: int(71),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p885,
								FreeVars: ast.Identifiers{
									"std",
									"str",
								},
							},
							Left: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(108),
											Column: int(32),
										},
										End: ast.Location{
											Line: int(108),
											Column: int(55),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p885,
									FreeVars: nil,
								},
								Value: "Expected string, got ",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(3),
							Right: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(108),
											Column: int(58),
										},
										End: ast.Location{
											Line: int(108),
											Column: int(71),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p885,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(108),
												Column: int(58),
											},
											End: ast.Location{
												Line: int(108),
												Column: int(66),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p885,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(108),
													Column: int(58),
												},
												End: ast.Location{
													Line: int(108),
													Column: int(61),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "type",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(108),
															Column: int(67),
														},
														End: ast.Location{
															Line: int(108),
															Column: int(70),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p938,
													FreeVars: ast.Identifiers{
														"str",
													},
												},
												Id: "str",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(107),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(110),
					Column: int(23),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "split",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(112),
								Column: int(9),
							},
							End: ast.Location{
								Line: int(112),
								Column: int(12),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "c",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(112),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(112),
								Column: int(15),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "",
							Begin: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							End: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							File: nil,
						},
						Fodder: nil,
						Ctx: nil,
						FreeVars: ast.Identifiers{
							"c",
							"std",
							"str",
						},
					},
					Cond: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(113),
									Column: int(12),
								},
								End: ast.Location{
									Line: int(113),
									Column: int(29),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p947,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
						},
						Target: &ast.Index{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(113),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(113),
										Column: int(24),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p947,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(113),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(113),
											Column: int(15),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Id: "std",
							},
							LeftBracketFodder: ast.Fodder{},
							Index: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Value: "isString",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							RightBracketFodder: ast.Fodder{},
							Id: nil,
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(113),
													Column: int(25),
												},
												End: ast.Location{
													Line: int(113),
													Column: int(28),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p956,
											FreeVars: ast.Identifiers{
												"str",
											},
										},
										Id: "str",
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
					ThenFodder: nil,
					BranchTrue: &ast.Conditional{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"c",
								"std",
								"str",
							},
						},
						Cond: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(114),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(114),
										Column: int(27),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p947,
								FreeVars: ast.Identifiers{
									"c",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(114),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(114),
											Column: int(24),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p947,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(114),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(114),
												Column: int(15),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "isString",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(114),
														Column: int(25),
													},
													End: ast.Location{
														Line: int(114),
														Column: int(26),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p969,
												FreeVars: ast.Identifiers{
													"c",
												},
											},
											Id: "c",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
						ThenFodder: nil,
						BranchTrue: &ast.Conditional{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"c",
									"std",
									"str",
								},
							},
							Cond: &ast.Binary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(115),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(115),
											Column: int(30),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p947,
									FreeVars: ast.Identifiers{
										"c",
										"std",
									},
								},
								Left: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(115),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(115),
												Column: int(25),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p947,
										FreeVars: ast.Identifiers{
											"c",
											"std",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(115),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(115),
													Column: int(22),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p947,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(115),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(115),
														Column: int(15),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "length",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(115),
																Column: int(23),
															},
															End: ast.Location{
																Line: int(115),
																Column: int(24),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p984,
														FreeVars: ast.Identifiers{
															"c",
														},
													},
													Id: "c",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
								OpFodder: ast.Fodder{},
								Op: ast.BinaryOp(12),
								Right: &ast.LiteralNumber{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(115),
												Column: int(29),
											},
											End: ast.Location{
												Line: int(115),
												Column: int(30),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p947,
										FreeVars: nil,
									},
									OriginalString: "1",
								},
							},
							ThenFodder: nil,
							BranchTrue: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(116),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(116),
											Column: int(31),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p947,
									FreeVars: ast.Identifiers{
										"c",
										"std",
										"str",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(116),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(116),
												Column: int(19),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p947,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(116),
													Column: int(5),
												},
												End: ast.Location{
													Line: int(116),
													Column: int(8),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(4),
													Comment: []string{},
												},
											},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "splitLimit",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(116),
															Column: int(20),
														},
														End: ast.Location{
															Line: int(116),
															Column: int(23),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p997,
													FreeVars: ast.Identifiers{
														"str",
													},
												},
												Id: "str",
											},
											CommaFodder: ast.Fodder{},
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(116),
															Column: int(25),
														},
														End: ast.Location{
															Line: int(116),
															Column: int(26),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p997,
													FreeVars: ast.Identifiers{
														"c",
													},
												},
												Id: "c",
											},
											CommaFodder: ast.Fodder{},
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.Unary{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(116),
															Column: int(28),
														},
														End: ast.Location{
															Line: int(116),
															Column: int(30),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p997,
													FreeVars: nil,
												},
												Op: ast.UnaryOp(3),
												Expr: &ast.LiteralNumber{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(116),
																Column: int(29),
															},
															End: ast.Location{
																Line: int(116),
																Column: int(30),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p997,
														FreeVars: nil,
													},
													OriginalString: "1",
												},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
							ElseFodder: nil,
							BranchFalse: &ast.Error{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"c",
										"std",
									},
								},
								Expr: &ast.Binary{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(115),
												Column: int(34),
											},
											End: ast.Location{
												Line: int(115),
												Column: int(105),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p947,
										FreeVars: ast.Identifiers{
											"c",
											"std",
										},
									},
									Left: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(115),
													Column: int(34),
												},
												End: ast.Location{
													Line: int(115),
													Column: int(89),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p947,
											FreeVars: nil,
										},
										Value: "std.split second parameter should have length 1, got ",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									OpFodder: ast.Fodder{},
									Op: ast.BinaryOp(3),
									Right: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(115),
													Column: int(92),
												},
												End: ast.Location{
													Line: int(115),
													Column: int(105),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p947,
											FreeVars: ast.Identifiers{
												"c",
												"std",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(115),
														Column: int(92),
													},
													End: ast.Location{
														Line: int(115),
														Column: int(102),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p947,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(115),
															Column: int(92),
														},
														End: ast.Location{
															Line: int(115),
															Column: int(95),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: ast.Fodder{},
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "length",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: ast.Fodder{},
											Id: nil,
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(115),
																	Column: int(103),
																},
																End: ast.Location{
																	Line: int(115),
																	Column: int(104),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p1017,
															FreeVars: ast.Identifiers{
																"c",
															},
														},
														Id: "c",
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
								},
							},
						},
						ElseFodder: nil,
						BranchFalse: &ast.Error{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"c",
									"std",
								},
							},
							Expr: &ast.Binary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(114),
											Column: int(30),
										},
										End: ast.Location{
											Line: int(114),
											Column: int(97),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p947,
									FreeVars: ast.Identifiers{
										"c",
										"std",
									},
								},
								Left: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(114),
												Column: int(30),
											},
											End: ast.Location{
												Line: int(114),
												Column: int(83),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p947,
										FreeVars: nil,
									},
									Value: "std.split second parameter should be a string, got ",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								OpFodder: ast.Fodder{},
								Op: ast.BinaryOp(3),
								Right: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(114),
												Column: int(86),
											},
											End: ast.Location{
												Line: int(114),
												Column: int(97),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p947,
										FreeVars: ast.Identifiers{
											"c",
											"std",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(114),
													Column: int(86),
												},
												End: ast.Location{
													Line: int(114),
													Column: int(94),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p947,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(114),
														Column: int(86),
													},
													End: ast.Location{
														Line: int(114),
														Column: int(89),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "type",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(114),
																Column: int(95),
															},
															End: ast.Location{
																Line: int(114),
																Column: int(96),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p1033,
														FreeVars: ast.Identifiers{
															"c",
														},
													},
													Id: "c",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
							},
						},
					},
					ElseFodder: nil,
					BranchFalse: &ast.Error{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
						},
						Expr: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(113),
										Column: int(32),
									},
									End: ast.Location{
										Line: int(113),
										Column: int(100),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p947,
								FreeVars: ast.Identifiers{
									"std",
									"str",
								},
							},
							Left: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(113),
											Column: int(32),
										},
										End: ast.Location{
											Line: int(113),
											Column: int(84),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p947,
									FreeVars: nil,
								},
								Value: "std.split first parameter should be a string, got ",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(3),
							Right: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(113),
											Column: int(87),
										},
										End: ast.Location{
											Line: int(113),
											Column: int(100),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p947,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(113),
												Column: int(87),
											},
											End: ast.Location{
												Line: int(113),
												Column: int(95),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p947,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(113),
													Column: int(87),
												},
												End: ast.Location{
													Line: int(113),
													Column: int(90),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "type",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(113),
															Column: int(96),
														},
														End: ast.Location{
															Line: int(113),
															Column: int(99),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p1049,
													FreeVars: ast.Identifiers{
														"str",
													},
												},
												Id: "str",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(112),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(116),
					Column: int(31),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "splitLimit",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(118),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(118),
								Column: int(17),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "c",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(118),
								Column: int(19),
							},
							End: ast.Location{
								Line: int(118),
								Column: int(20),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "maxsplits",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(118),
								Column: int(22),
							},
							End: ast.Location{
								Line: int(118),
								Column: int(31),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "",
							Begin: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							End: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							File: nil,
						},
						Fodder: nil,
						Ctx: nil,
						FreeVars: ast.Identifiers{
							"c",
							"maxsplits",
							"std",
							"str",
						},
					},
					Cond: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(119),
									Column: int(12),
								},
								End: ast.Location{
									Line: int(119),
									Column: int(29),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p1058,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
						},
						Target: &ast.Index{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(119),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(119),
										Column: int(24),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p1058,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(119),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(119),
											Column: int(15),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Id: "std",
							},
							LeftBracketFodder: ast.Fodder{},
							Index: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Value: "isString",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							RightBracketFodder: ast.Fodder{},
							Id: nil,
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(119),
													Column: int(25),
												},
												End: ast.Location{
													Line: int(119),
													Column: int(28),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1067,
											FreeVars: ast.Identifiers{
												"str",
											},
										},
										Id: "str",
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
					ThenFodder: nil,
					BranchTrue: &ast.Conditional{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"c",
								"maxsplits",
								"std",
								"str",
							},
						},
						Cond: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(120),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(120),
										Column: int(27),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p1058,
								FreeVars: ast.Identifiers{
									"c",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(120),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(120),
											Column: int(24),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p1058,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(120),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(120),
												Column: int(15),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "isString",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(120),
														Column: int(25),
													},
													End: ast.Location{
														Line: int(120),
														Column: int(26),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p1080,
												FreeVars: ast.Identifiers{
													"c",
												},
											},
											Id: "c",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
						ThenFodder: nil,
						BranchTrue: &ast.Conditional{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"c",
									"maxsplits",
									"std",
									"str",
								},
							},
							Cond: &ast.Binary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(121),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(121),
											Column: int(30),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p1058,
									FreeVars: ast.Identifiers{
										"c",
										"std",
									},
								},
								Left: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(121),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(121),
												Column: int(25),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p1058,
										FreeVars: ast.Identifiers{
											"c",
											"std",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(121),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(121),
													Column: int(22),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1058,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(121),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(121),
														Column: int(15),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "length",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(121),
																Column: int(23),
															},
															End: ast.Location{
																Line: int(121),
																Column: int(24),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p1095,
														FreeVars: ast.Identifiers{
															"c",
														},
													},
													Id: "c",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
								OpFodder: ast.Fodder{},
								Op: ast.BinaryOp(12),
								Right: &ast.LiteralNumber{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(121),
												Column: int(29),
											},
											End: ast.Location{
												Line: int(121),
												Column: int(30),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p1058,
										FreeVars: nil,
									},
									OriginalString: "1",
								},
							},
							ThenFodder: nil,
							BranchTrue: &ast.Conditional{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"c",
										"maxsplits",
										"std",
										"str",
									},
								},
								Cond: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(122),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(122),
												Column: int(35),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p1058,
										FreeVars: ast.Identifiers{
											"maxsplits",
											"std",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(122),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(122),
													Column: int(24),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1058,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(122),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(122),
														Column: int(15),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "isNumber",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(122),
																Column: int(25),
															},
															End: ast.Location{
																Line: int(122),
																Column: int(34),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p1109,
														FreeVars: ast.Identifiers{
															"maxsplits",
														},
													},
													Id: "maxsplits",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
								ThenFodder: nil,
								BranchTrue: &ast.Local{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(123),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(132),
												Column: int(27),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(4),
												Comment: []string{},
											},
										},
										Ctx: p1058,
										FreeVars: ast.Identifiers{
											"c",
											"maxsplits",
											"std",
											"str",
										},
									},
									Binds: ast.LocalBinds{
										ast.LocalBind{
											VarFodder: nil,
											Variable: "aux",
											EqFodder: nil,
											Body: &ast.Function{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(123),
															Column: int(11),
														},
														End: ast.Location{
															Line: int(131),
															Column: int(40),
														},
														File: p1,
													},
													Fodder: nil,
													Ctx: p1116,
													FreeVars: ast.Identifiers{
														"aux",
														"maxsplits",
														"std",
													},
												},
												ParenLeftFodder: ast.Fodder{},
												Parameters: []ast.Parameter{
													ast.Parameter{
														NameFodder: ast.Fodder{},
														Name: "str",
														EqFodder: nil,
														DefaultArg: nil,
														CommaFodder: ast.Fodder{},
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(123),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(123),
																Column: int(18),
															},
															File: p1,
														},
													},
													ast.Parameter{
														NameFodder: ast.Fodder{},
														Name: "delim",
														EqFodder: nil,
														DefaultArg: nil,
														CommaFodder: ast.Fodder{},
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(123),
																Column: int(20),
															},
															End: ast.Location{
																Line: int(123),
																Column: int(25),
															},
															File: p1,
														},
													},
													ast.Parameter{
														NameFodder: ast.Fodder{},
														Name: "i",
														EqFodder: nil,
														DefaultArg: nil,
														CommaFodder: ast.Fodder{},
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(123),
																Column: int(27),
															},
															End: ast.Location{
																Line: int(123),
																Column: int(28),
															},
															File: p1,
														},
													},
													ast.Parameter{
														NameFodder: ast.Fodder{},
														Name: "arr",
														EqFodder: nil,
														DefaultArg: nil,
														CommaFodder: ast.Fodder{},
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(123),
																Column: int(30),
															},
															End: ast.Location{
																Line: int(123),
																Column: int(33),
															},
															File: p1,
														},
													},
													ast.Parameter{
														NameFodder: ast.Fodder{},
														Name: "v",
														EqFodder: nil,
														DefaultArg: nil,
														CommaFodder: nil,
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(123),
																Column: int(35),
															},
															End: ast.Location{
																Line: int(123),
																Column: int(36),
															},
															File: p1,
														},
													},
												},
												TrailingComma: false,
												ParenRightFodder: ast.Fodder{},
												Body: &ast.Local{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(124),
																Column: int(7),
															},
															End: ast.Location{
																Line: int(131),
																Column: int(40),
															},
															File: p1,
														},
														Fodder: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(6),
																Comment: []string{},
															},
														},
														Ctx: p1121,
														FreeVars: ast.Identifiers{
															"arr",
															"aux",
															"delim",
															"i",
															"maxsplits",
															"std",
															"str",
															"v",
														},
													},
													Binds: ast.LocalBinds{
														ast.LocalBind{
															VarFodder: ast.Fodder{},
															Variable: "c",
															EqFodder: ast.Fodder{},
															Body: &ast.Index{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(124),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(124),
																			Column: int(23),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p1125,
																	FreeVars: ast.Identifiers{
																		"i",
																		"str",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(124),
																				Column: int(17),
																			},
																			End: ast.Location{
																				Line: int(124),
																				Column: int(20),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p1125,
																		FreeVars: ast.Identifiers{
																			"str",
																		},
																	},
																	Id: "str",
																},
																LeftBracketFodder: ast.Fodder{},
																Index: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(124),
																				Column: int(21),
																			},
																			End: ast.Location{
																				Line: int(124),
																				Column: int(22),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p1125,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																	},
																	Id: "i",
																},
																RightBracketFodder: ast.Fodder{},
																Id: nil,
															},
															Fun: nil,
															CloseFodder: ast.Fodder{},
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(124),
																	Column: int(13),
																},
																End: ast.Location{
																	Line: int(124),
																	Column: int(23),
																},
																File: p1,
															},
														},
													},
													Body: &ast.Local{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(125),
																	Column: int(7),
																},
																End: ast.Location{
																	Line: int(131),
																	Column: int(40),
																},
																File: p1,
															},
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(6),
																	Comment: []string{},
																},
															},
															Ctx: p1121,
															FreeVars: ast.Identifiers{
																"arr",
																"aux",
																"c",
																"delim",
																"i",
																"maxsplits",
																"std",
																"str",
																"v",
															},
														},
														Binds: ast.LocalBinds{
															ast.LocalBind{
																VarFodder: ast.Fodder{},
																Variable: "i2",
																EqFodder: ast.Fodder{},
																Body: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(125),
																				Column: int(18),
																			},
																			End: ast.Location{
																				Line: int(125),
																				Column: int(23),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p1136,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																	},
																	Left: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(125),
																					Column: int(18),
																				},
																				End: ast.Location{
																					Line: int(125),
																					Column: int(19),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p1136,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																		},
																		Id: "i",
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(3),
																	Right: &ast.LiteralNumber{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(125),
																					Column: int(22),
																				},
																				End: ast.Location{
																					Line: int(125),
																					Column: int(23),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p1136,
																			FreeVars: nil,
																		},
																		OriginalString: "1",
																	},
																},
																Fun: nil,
																CloseFodder: ast.Fodder{},
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(125),
																		Column: int(13),
																	},
																	End: ast.Location{
																		Line: int(125),
																		Column: int(23),
																	},
																	File: p1,
																},
															},
														},
														Body: &ast.Conditional{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(126),
																		Column: int(7),
																	},
																	End: ast.Location{
																		Line: int(131),
																		Column: int(40),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(6),
																		Comment: []string{},
																	},
																},
																Ctx: p1121,
																FreeVars: ast.Identifiers{
																	"arr",
																	"aux",
																	"c",
																	"delim",
																	"i",
																	"i2",
																	"maxsplits",
																	"std",
																	"str",
																	"v",
																},
															},
															Cond: &ast.Binary{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(126),
																			Column: int(10),
																		},
																		End: ast.Location{
																			Line: int(126),
																			Column: int(30),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p1121,
																	FreeVars: ast.Identifiers{
																		"i",
																		"std",
																		"str",
																	},
																},
																Left: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(126),
																				Column: int(10),
																			},
																			End: ast.Location{
																				Line: int(126),
																				Column: int(11),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p1121,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																	},
																	Id: "i",
																},
																OpFodder: ast.Fodder{},
																Op: ast.BinaryOp(8),
																Right: &ast.Apply{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(126),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(126),
																				Column: int(30),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p1121,
																		FreeVars: ast.Identifiers{
																			"std",
																			"str",
																		},
																	},
																	Target: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(126),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(126),
																					Column: int(25),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p1121,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(126),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(126),
																						Column: int(18),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Id: "std",
																		},
																		LeftBracketFodder: ast.Fodder{},
																		Index: &ast.LiteralString{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: nil,
																			},
																			Value: "length",
																			Kind: ast.LiteralStringKind(1),
																			BlockIndent: "",
																			BlockTermIndent: "",
																		},
																		RightBracketFodder: ast.Fodder{},
																		Id: nil,
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(126),
																								Column: int(26),
																							},
																							End: ast.Location{
																								Line: int(126),
																								Column: int(29),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p1157,
																						FreeVars: ast.Identifiers{
																							"str",
																						},
																					},
																					Id: "str",
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	TrailingComma: false,
																	TailStrict: false,
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																},
															},
															ThenFodder: ast.Fodder{},
															BranchTrue: &ast.Binary{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(127),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(127),
																			Column: int(18),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p1121,
																	FreeVars: ast.Identifiers{
																		"arr",
																		"v",
																	},
																},
																Left: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(127),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(127),
																				Column: int(12),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(8),
																				Comment: []string{},
																			},
																		},
																		Ctx: p1121,
																		FreeVars: ast.Identifiers{
																			"arr",
																		},
																	},
																	Id: "arr",
																},
																OpFodder: ast.Fodder{},
																Op: ast.BinaryOp(3),
																Right: &ast.Array{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(127),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(127),
																				Column: int(18),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p1121,
																		FreeVars: ast.Identifiers{
																			"v",
																		},
																	},
																	Elements: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(127),
																							Column: int(16),
																						},
																						End: ast.Location{
																							Line: int(127),
																							Column: int(17),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p1168,
																					FreeVars: ast.Identifiers{
																						"v",
																					},
																				},
																				Id: "v",
																			},
																			CommaFodder: nil,
																		},
																	},
																	TrailingComma: false,
																	CloseFodder: ast.Fodder{},
																},
															},
															ElseFodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(6),
																	Comment: []string{},
																},
															},
															BranchFalse: &ast.Conditional{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(128),
																			Column: int(12),
																		},
																		End: ast.Location{
																			Line: int(131),
																			Column: int(40),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p1121,
																	FreeVars: ast.Identifiers{
																		"arr",
																		"aux",
																		"c",
																		"delim",
																		"i2",
																		"maxsplits",
																		"std",
																		"str",
																		"v",
																	},
																},
																Cond: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(128),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(128),
																				Column: int(77),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p1121,
																		FreeVars: ast.Identifiers{
																			"arr",
																			"c",
																			"delim",
																			"maxsplits",
																			"std",
																		},
																	},
																	Left: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(128),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(128),
																					Column: int(25),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p1121,
																			FreeVars: ast.Identifiers{
																				"c",
																				"delim",
																			},
																		},
																		Left: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(128),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(128),
																						Column: int(16),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p1121,
																				FreeVars: ast.Identifiers{
																					"c",
																				},
																			},
																			Id: "c",
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(12),
																		Right: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(128),
																						Column: int(20),
																					},
																					End: ast.Location{
																						Line: int(128),
																						Column: int(25),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p1121,
																				FreeVars: ast.Identifiers{
																					"delim",
																				},
																			},
																			Id: "delim",
																		},
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(17),
																	Right: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(128),
																					Column: int(30),
																				},
																				End: ast.Location{
																					Line: int(128),
																					Column: int(76),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p1121,
																			FreeVars: ast.Identifiers{
																				"arr",
																				"maxsplits",
																				"std",
																			},
																		},
																		Left: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(128),
																						Column: int(30),
																					},
																					End: ast.Location{
																						Line: int(128),
																						Column: int(45),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p1121,
																				FreeVars: ast.Identifiers{
																					"maxsplits",
																				},
																			},
																			Left: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(128),
																							Column: int(30),
																						},
																						End: ast.Location{
																							Line: int(128),
																							Column: int(39),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p1121,
																					FreeVars: ast.Identifiers{
																						"maxsplits",
																					},
																				},
																				Id: "maxsplits",
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(12),
																			Right: &ast.Unary{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(128),
																							Column: int(43),
																						},
																						End: ast.Location{
																							Line: int(128),
																							Column: int(45),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p1121,
																					FreeVars: nil,
																				},
																				Op: ast.UnaryOp(3),
																				Expr: &ast.LiteralNumber{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(128),
																								Column: int(44),
																							},
																							End: ast.Location{
																								Line: int(128),
																								Column: int(45),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p1121,
																						FreeVars: nil,
																					},
																					OriginalString: "1",
																				},
																			},
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(18),
																		Right: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(128),
																						Column: int(49),
																					},
																					End: ast.Location{
																						Line: int(128),
																						Column: int(76),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p1121,
																				FreeVars: ast.Identifiers{
																					"arr",
																					"maxsplits",
																					"std",
																				},
																			},
																			Left: &ast.Apply{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(128),
																							Column: int(49),
																						},
																						End: ast.Location{
																							Line: int(128),
																							Column: int(64),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p1121,
																					FreeVars: ast.Identifiers{
																						"arr",
																						"std",
																					},
																				},
																				Target: &ast.Index{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(128),
																								Column: int(49),
																							},
																							End: ast.Location{
																								Line: int(128),
																								Column: int(59),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p1121,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																					},
																					Target: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(128),
																									Column: int(49),
																								},
																								End: ast.Location{
																									Line: int(128),
																									Column: int(52),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: nil,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																						},
																						Id: "std",
																					},
																					LeftBracketFodder: ast.Fodder{},
																					Index: &ast.LiteralString{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								File: nil,
																							},
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: nil,
																						},
																						Value: "length",
																						Kind: ast.LiteralStringKind(1),
																						BlockIndent: "",
																						BlockTermIndent: "",
																					},
																					RightBracketFodder: ast.Fodder{},
																					Id: nil,
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(128),
																											Column: int(60),
																										},
																										End: ast.Location{
																											Line: int(128),
																											Column: int(63),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p1200,
																									FreeVars: ast.Identifiers{
																										"arr",
																									},
																								},
																								Id: "arr",
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				TrailingComma: false,
																				TailStrict: false,
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: nil,
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(9),
																			Right: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(128),
																							Column: int(67),
																						},
																						End: ast.Location{
																							Line: int(128),
																							Column: int(76),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p1121,
																					FreeVars: ast.Identifiers{
																						"maxsplits",
																					},
																				},
																				Id: "maxsplits",
																			},
																		},
																	},
																},
																ThenFodder: ast.Fodder{},
																BranchTrue: &ast.Apply{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(129),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(129),
																				Column: int(43),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p1121,
																		FreeVars: ast.Identifiers{
																			"arr",
																			"aux",
																			"delim",
																			"i2",
																			"str",
																			"v",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(129),
																					Column: int(9),
																				},
																				End: ast.Location{
																					Line: int(129),
																					Column: int(12),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(8),
																					Comment: []string{},
																				},
																			},
																			Ctx: p1121,
																			FreeVars: ast.Identifiers{
																				"aux",
																			},
																		},
																		Id: "aux",
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(129),
																								Column: int(13),
																							},
																							End: ast.Location{
																								Line: int(129),
																								Column: int(16),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p1211,
																						FreeVars: ast.Identifiers{
																							"str",
																						},
																					},
																					Id: "str",
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(129),
																								Column: int(18),
																							},
																							End: ast.Location{
																								Line: int(129),
																								Column: int(23),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p1211,
																						FreeVars: ast.Identifiers{
																							"delim",
																						},
																					},
																					Id: "delim",
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(129),
																								Column: int(25),
																							},
																							End: ast.Location{
																								Line: int(129),
																								Column: int(27),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p1211,
																						FreeVars: ast.Identifiers{
																							"i2",
																						},
																					},
																					Id: "i2",
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Binary{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(129),
																								Column: int(29),
																							},
																							End: ast.Location{
																								Line: int(129),
																								Column: int(38),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p1211,
																						FreeVars: ast.Identifiers{
																							"arr",
																							"v",
																						},
																					},
																					Left: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(129),
																									Column: int(29),
																								},
																								End: ast.Location{
																									Line: int(129),
																									Column: int(32),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p1211,
																							FreeVars: ast.Identifiers{
																								"arr",
																							},
																						},
																						Id: "arr",
																					},
																					OpFodder: ast.Fodder{},
																					Op: ast.BinaryOp(3),
																					Right: &ast.Array{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(129),
																									Column: int(35),
																								},
																								End: ast.Location{
																									Line: int(129),
																									Column: int(38),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p1211,
																							FreeVars: ast.Identifiers{
																								"v",
																							},
																						},
																						Elements: []ast.CommaSeparatedExpr{
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(129),
																												Column: int(36),
																											},
																											End: ast.Location{
																												Line: int(129),
																												Column: int(37),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p1225,
																										FreeVars: ast.Identifiers{
																											"v",
																										},
																									},
																									Id: "v",
																								},
																								CommaFodder: nil,
																							},
																						},
																						TrailingComma: false,
																						CloseFodder: ast.Fodder{},
																					},
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.LiteralString{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(129),
																								Column: int(40),
																							},
																							End: ast.Location{
																								Line: int(129),
																								Column: int(42),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p1211,
																						FreeVars: nil,
																					},
																					Value: "",
																					Kind: ast.LiteralStringKind(1),
																					BlockIndent: "",
																					BlockTermIndent: "",
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	TrailingComma: false,
																	TailStrict: true,
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: ast.Fodder{},
																},
																ElseFodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(6),
																		Comment: []string{},
																	},
																},
																BranchFalse: &ast.Apply{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(131),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(131),
																				Column: int(40),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p1121,
																		FreeVars: ast.Identifiers{
																			"arr",
																			"aux",
																			"c",
																			"delim",
																			"i2",
																			"str",
																			"v",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(131),
																					Column: int(9),
																				},
																				End: ast.Location{
																					Line: int(131),
																					Column: int(12),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(8),
																					Comment: []string{},
																				},
																			},
																			Ctx: p1121,
																			FreeVars: ast.Identifiers{
																				"aux",
																			},
																		},
																		Id: "aux",
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(131),
																								Column: int(13),
																							},
																							End: ast.Location{
																								Line: int(131),
																								Column: int(16),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p1236,
																						FreeVars: ast.Identifiers{
																							"str",
																						},
																					},
																					Id: "str",
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(131),
																								Column: int(18),
																							},
																							End: ast.Location{
																								Line: int(131),
																								Column: int(23),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p1236,
																						FreeVars: ast.Identifiers{
																							"delim",
																						},
																					},
																					Id: "delim",
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(131),
																								Column: int(25),
																							},
																							End: ast.Location{
																								Line: int(131),
																								Column: int(27),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p1236,
																						FreeVars: ast.Identifiers{
																							"i2",
																						},
																					},
																					Id: "i2",
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(131),
																								Column: int(29),
																							},
																							End: ast.Location{
																								Line: int(131),
																								Column: int(32),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p1236,
																						FreeVars: ast.Identifiers{
																							"arr",
																						},
																					},
																					Id: "arr",
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Binary{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(131),
																								Column: int(34),
																							},
																							End: ast.Location{
																								Line: int(131),
																								Column: int(39),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p1236,
																						FreeVars: ast.Identifiers{
																							"c",
																							"v",
																						},
																					},
																					Left: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(131),
																									Column: int(34),
																								},
																								End: ast.Location{
																									Line: int(131),
																									Column: int(35),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p1236,
																							FreeVars: ast.Identifiers{
																								"v",
																							},
																						},
																						Id: "v",
																					},
																					OpFodder: ast.Fodder{},
																					Op: ast.BinaryOp(3),
																					Right: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(131),
																									Column: int(38),
																								},
																								End: ast.Location{
																									Line: int(131),
																									Column: int(39),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p1236,
																							FreeVars: ast.Identifiers{
																								"c",
																							},
																						},
																						Id: "c",
																					},
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	TrailingComma: false,
																	TailStrict: true,
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: ast.Fodder{},
																},
															},
														},
													},
												},
											},
											Fun: nil,
											CloseFodder: nil,
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
										},
									},
									Body: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(132),
													Column: int(5),
												},
												End: ast.Location{
													Line: int(132),
													Column: int(27),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1058,
											FreeVars: ast.Identifiers{
												"aux",
												"c",
												"str",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(132),
														Column: int(5),
													},
													End: ast.Location{
														Line: int(132),
														Column: int(8),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(4),
														Comment: []string{},
													},
												},
												Ctx: p1058,
												FreeVars: ast.Identifiers{
													"aux",
												},
											},
											Id: "aux",
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(132),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(132),
																	Column: int(12),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p1257,
															FreeVars: ast.Identifiers{
																"str",
															},
														},
														Id: "str",
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(132),
																	Column: int(14),
																},
																End: ast.Location{
																	Line: int(132),
																	Column: int(15),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p1257,
															FreeVars: ast.Identifiers{
																"c",
															},
														},
														Id: "c",
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.LiteralNumber{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(132),
																	Column: int(17),
																},
																End: ast.Location{
																	Line: int(132),
																	Column: int(18),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p1257,
															FreeVars: nil,
														},
														OriginalString: "0",
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Array{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(132),
																	Column: int(20),
																},
																End: ast.Location{
																	Line: int(132),
																	Column: int(22),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p1257,
															FreeVars: nil,
														},
														Elements: nil,
														TrailingComma: false,
														CloseFodder: ast.Fodder{},
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(132),
																	Column: int(24),
																},
																End: ast.Location{
																	Line: int(132),
																	Column: int(26),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p1257,
															FreeVars: nil,
														},
														Value: "",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
								},
								ElseFodder: nil,
								BranchFalse: &ast.Error{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"maxsplits",
											"std",
										},
									},
									Expr: &ast.Binary{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(122),
													Column: int(38),
												},
												End: ast.Location{
													Line: int(122),
													Column: int(117),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1058,
											FreeVars: ast.Identifiers{
												"maxsplits",
												"std",
											},
										},
										Left: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(122),
														Column: int(38),
													},
													End: ast.Location{
														Line: int(122),
														Column: int(95),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p1058,
												FreeVars: nil,
											},
											Value: "std.splitLimit third parameter should be a number, got ",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										OpFodder: ast.Fodder{},
										Op: ast.BinaryOp(3),
										Right: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(122),
														Column: int(98),
													},
													End: ast.Location{
														Line: int(122),
														Column: int(117),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p1058,
												FreeVars: ast.Identifiers{
													"maxsplits",
													"std",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(122),
															Column: int(98),
														},
														End: ast.Location{
															Line: int(122),
															Column: int(106),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p1058,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(122),
																Column: int(98),
															},
															End: ast.Location{
																Line: int(122),
																Column: int(101),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "type",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(122),
																		Column: int(107),
																	},
																	End: ast.Location{
																		Line: int(122),
																		Column: int(116),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p1278,
																FreeVars: ast.Identifiers{
																	"maxsplits",
																},
															},
															Id: "maxsplits",
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
									},
								},
							},
							ElseFodder: nil,
							BranchFalse: &ast.Error{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"c",
										"std",
									},
								},
								Expr: &ast.Binary{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(121),
												Column: int(34),
											},
											End: ast.Location{
												Line: int(121),
												Column: int(110),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p1058,
										FreeVars: ast.Identifiers{
											"c",
											"std",
										},
									},
									Left: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(121),
													Column: int(34),
												},
												End: ast.Location{
													Line: int(121),
													Column: int(94),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1058,
											FreeVars: nil,
										},
										Value: "std.splitLimit second parameter should have length 1, got ",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									OpFodder: ast.Fodder{},
									Op: ast.BinaryOp(3),
									Right: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(121),
													Column: int(97),
												},
												End: ast.Location{
													Line: int(121),
													Column: int(110),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1058,
											FreeVars: ast.Identifiers{
												"c",
												"std",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(121),
														Column: int(97),
													},
													End: ast.Location{
														Line: int(121),
														Column: int(107),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p1058,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(121),
															Column: int(97),
														},
														End: ast.Location{
															Line: int(121),
															Column: int(100),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: ast.Fodder{},
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "length",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: ast.Fodder{},
											Id: nil,
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(121),
																	Column: int(108),
																},
																End: ast.Location{
																	Line: int(121),
																	Column: int(109),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p1294,
															FreeVars: ast.Identifiers{
																"c",
															},
														},
														Id: "c",
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
								},
							},
						},
						ElseFodder: nil,
						BranchFalse: &ast.Error{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"c",
									"std",
								},
							},
							Expr: &ast.Binary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(120),
											Column: int(30),
										},
										End: ast.Location{
											Line: int(120),
											Column: int(102),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p1058,
									FreeVars: ast.Identifiers{
										"c",
										"std",
									},
								},
								Left: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(120),
												Column: int(30),
											},
											End: ast.Location{
												Line: int(120),
												Column: int(88),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p1058,
										FreeVars: nil,
									},
									Value: "std.splitLimit second parameter should be a string, got ",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								OpFodder: ast.Fodder{},
								Op: ast.BinaryOp(3),
								Right: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(120),
												Column: int(91),
											},
											End: ast.Location{
												Line: int(120),
												Column: int(102),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p1058,
										FreeVars: ast.Identifiers{
											"c",
											"std",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(120),
													Column: int(91),
												},
												End: ast.Location{
													Line: int(120),
													Column: int(99),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1058,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(120),
														Column: int(91),
													},
													End: ast.Location{
														Line: int(120),
														Column: int(94),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "type",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(120),
																Column: int(100),
															},
															End: ast.Location{
																Line: int(120),
																Column: int(101),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p1310,
														FreeVars: ast.Identifiers{
															"c",
														},
													},
													Id: "c",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
							},
						},
					},
					ElseFodder: nil,
					BranchFalse: &ast.Error{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
						},
						Expr: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(119),
										Column: int(32),
									},
									End: ast.Location{
										Line: int(119),
										Column: int(105),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p1058,
								FreeVars: ast.Identifiers{
									"std",
									"str",
								},
							},
							Left: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(119),
											Column: int(32),
										},
										End: ast.Location{
											Line: int(119),
											Column: int(89),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p1058,
									FreeVars: nil,
								},
								Value: "std.splitLimit first parameter should be a string, got ",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(3),
							Right: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(119),
											Column: int(92),
										},
										End: ast.Location{
											Line: int(119),
											Column: int(105),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p1058,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(119),
												Column: int(92),
											},
											End: ast.Location{
												Line: int(119),
												Column: int(100),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p1058,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(119),
													Column: int(92),
												},
												End: ast.Location{
													Line: int(119),
													Column: int(95),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "type",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(119),
															Column: int(101),
														},
														End: ast.Location{
															Line: int(119),
															Column: int(104),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p1326,
													FreeVars: ast.Identifiers{
														"str",
													},
												},
												Id: "str",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(118),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(132),
					Column: int(27),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "strReplace",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(134),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(134),
								Column: int(17),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "from",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(134),
								Column: int(19),
							},
							End: ast.Location{
								Line: int(134),
								Column: int(23),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "to",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(134),
								Column: int(25),
							},
							End: ast.Location{
								Line: int(134),
								Column: int(27),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "",
							Begin: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							End: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							File: nil,
						},
						Fodder: nil,
						Ctx: nil,
						FreeVars: ast.Identifiers{
							"from",
							"std",
							"str",
							"to",
						},
					},
					Cond: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(135),
									Column: int(12),
								},
								End: ast.Location{
									Line: int(135),
									Column: int(29),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p1335,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
						},
						Target: &ast.Index{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(135),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(135),
										Column: int(24),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p1335,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(135),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(135),
											Column: int(15),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Id: "std",
							},
							LeftBracketFodder: ast.Fodder{},
							Index: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Value: "isString",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							RightBracketFodder: ast.Fodder{},
							Id: nil,
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(135),
													Column: int(25),
												},
												End: ast.Location{
													Line: int(135),
													Column: int(28),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1344,
											FreeVars: ast.Identifiers{
												"str",
											},
										},
										Id: "str",
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
					ThenFodder: nil,
					BranchTrue: &ast.Conditional{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"from",
								"std",
								"str",
								"to",
							},
						},
						Cond: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(136),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(136),
										Column: int(30),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p1335,
								FreeVars: ast.Identifiers{
									"from",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(136),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(136),
											Column: int(24),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p1335,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(136),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(136),
												Column: int(15),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "isString",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(136),
														Column: int(25),
													},
													End: ast.Location{
														Line: int(136),
														Column: int(29),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p1357,
												FreeVars: ast.Identifiers{
													"from",
												},
											},
											Id: "from",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
						ThenFodder: nil,
						BranchTrue: &ast.Conditional{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"from",
									"std",
									"str",
									"to",
								},
							},
							Cond: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(137),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(137),
											Column: int(28),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p1335,
									FreeVars: ast.Identifiers{
										"std",
										"to",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(137),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(137),
												Column: int(24),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p1335,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(137),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(137),
													Column: int(15),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "isString",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(137),
															Column: int(25),
														},
														End: ast.Location{
															Line: int(137),
															Column: int(27),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p1370,
													FreeVars: ast.Identifiers{
														"to",
													},
												},
												Id: "to",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
							ThenFodder: nil,
							BranchTrue: &ast.Conditional{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"from",
										"std",
										"str",
										"to",
									},
								},
								Cond: &ast.Binary{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(138),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(138),
												Column: int(22),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p1335,
										FreeVars: ast.Identifiers{
											"from",
										},
									},
									Left: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(138),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(138),
													Column: int(16),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1335,
											FreeVars: ast.Identifiers{
												"from",
											},
										},
										Id: "from",
									},
									OpFodder: ast.Fodder{},
									Op: ast.BinaryOp(13),
									Right: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(138),
													Column: int(20),
												},
												End: ast.Location{
													Line: int(138),
													Column: int(22),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1335,
											FreeVars: nil,
										},
										Value: "",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
								},
								ThenFodder: nil,
								BranchTrue: &ast.Local{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(141),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(163),
												Column: int(30),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(1),
												Indent: int(4),
												Comment: []string{},
											},
											ast.FodderElement{
												Kind: ast.FodderKind(2),
												Blanks: int(0),
												Indent: int(4),
												Comment: []string{
													"// Cache for performance.",
												},
											},
										},
										Ctx: p1335,
										FreeVars: ast.Identifiers{
											"from",
											"std",
											"str",
											"to",
										},
									},
									Binds: ast.LocalBinds{
										ast.LocalBind{
											VarFodder: ast.Fodder{},
											Variable: "str_len",
											EqFodder: ast.Fodder{},
											Body: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(141),
															Column: int(21),
														},
														End: ast.Location{
															Line: int(141),
															Column: int(36),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p1385,
													FreeVars: ast.Identifiers{
														"std",
														"str",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(141),
																Column: int(21),
															},
															End: ast.Location{
																Line: int(141),
																Column: int(31),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p1385,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(141),
																	Column: int(21),
																},
																End: ast.Location{
																	Line: int(141),
																	Column: int(24),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: ast.Fodder{},
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "length",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: ast.Fodder{},
													Id: nil,
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(141),
																			Column: int(32),
																		},
																		End: ast.Location{
																			Line: int(141),
																			Column: int(35),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p1394,
																	FreeVars: ast.Identifiers{
																		"str",
																	},
																},
																Id: "str",
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
											Fun: nil,
											CloseFodder: ast.Fodder{},
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(141),
													Column: int(11),
												},
												End: ast.Location{
													Line: int(141),
													Column: int(36),
												},
												File: p1,
											},
										},
									},
									Body: &ast.Local{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(142),
													Column: int(5),
												},
												End: ast.Location{
													Line: int(163),
													Column: int(30),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(4),
													Comment: []string{},
												},
											},
											Ctx: p1335,
											FreeVars: ast.Identifiers{
												"from",
												"std",
												"str",
												"str_len",
												"to",
											},
										},
										Binds: ast.LocalBinds{
											ast.LocalBind{
												VarFodder: ast.Fodder{},
												Variable: "from_len",
												EqFodder: ast.Fodder{},
												Body: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(142),
																Column: int(22),
															},
															End: ast.Location{
																Line: int(142),
																Column: int(38),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p1401,
														FreeVars: ast.Identifiers{
															"from",
															"std",
														},
													},
													Target: &ast.Index{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(142),
																	Column: int(22),
																},
																End: ast.Location{
																	Line: int(142),
																	Column: int(32),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p1401,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(142),
																		Column: int(22),
																	},
																	End: ast.Location{
																		Line: int(142),
																		Column: int(25),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Id: "std",
														},
														LeftBracketFodder: ast.Fodder{},
														Index: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
															Value: "length",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														RightBracketFodder: ast.Fodder{},
														Id: nil,
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(142),
																				Column: int(33),
																			},
																			End: ast.Location{
																				Line: int(142),
																				Column: int(37),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p1410,
																		FreeVars: ast.Identifiers{
																			"from",
																		},
																	},
																	Id: "from",
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
												},
												Fun: nil,
												CloseFodder: ast.Fodder{},
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(142),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(142),
														Column: int(38),
													},
													File: p1,
												},
											},
										},
										Body: &ast.Local{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(145),
														Column: int(5),
													},
													End: ast.Location{
														Line: int(163),
														Column: int(30),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(1),
														Indent: int(4),
														Comment: []string{},
													},
													ast.FodderElement{
														Kind: ast.FodderKind(2),
														Blanks: int(0),
														Indent: int(4),
														Comment: []string{
															"// True if from is at str[i].",
														},
													},
												},
												Ctx: p1335,
												FreeVars: ast.Identifiers{
													"from",
													"from_len",
													"std",
													"str",
													"str_len",
													"to",
												},
											},
											Binds: ast.LocalBinds{
												ast.LocalBind{
													VarFodder: nil,
													Variable: "found_at",
													EqFodder: nil,
													Body: &ast.Function{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(145),
																	Column: int(11),
																},
																End: ast.Location{
																	Line: int(145),
																	Column: int(52),
																},
																File: p1,
															},
															Fodder: nil,
															Ctx: p1418,
															FreeVars: ast.Identifiers{
																"from",
																"from_len",
																"std",
																"str",
															},
														},
														ParenLeftFodder: ast.Fodder{},
														Parameters: []ast.Parameter{
															ast.Parameter{
																NameFodder: ast.Fodder{},
																Name: "i",
																EqFodder: nil,
																DefaultArg: nil,
																CommaFodder: nil,
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(145),
																		Column: int(20),
																	},
																	End: ast.Location{
																		Line: int(145),
																		Column: int(21),
																	},
																	File: p1,
																},
															},
														},
														TrailingComma: false,
														ParenRightFodder: ast.Fodder{},
														Body: &ast.Binary{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(145),
																		Column: int(25),
																	},
																	End: ast.Location{
																		Line: int(145),
																		Column: int(52),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p1422,
																FreeVars: ast.Identifiers{
																	"from",
																	"from_len",
																	"i",
																	"std",
																	"str",
																},
															},
															Left: &ast.Apply{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"from_len",
																		"i",
																		"std",
																		"str",
																	},
																},
																Target: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Id: "std",
																	},
																	LeftBracketFodder: nil,
																	Index: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: nil,
																		},
																		Value: "slice",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	RightBracketFodder: nil,
																	Id: nil,
																},
																FodderLeft: nil,
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(145),
																							Column: int(25),
																						},
																						End: ast.Location{
																							Line: int(145),
																							Column: int(28),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p1422,
																					FreeVars: ast.Identifiers{
																						"str",
																					},
																				},
																				Id: "str",
																			},
																			CommaFodder: nil,
																		},
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(145),
																							Column: int(29),
																						},
																						End: ast.Location{
																							Line: int(145),
																							Column: int(30),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p1422,
																					FreeVars: ast.Identifiers{
																						"i",
																					},
																				},
																				Id: "i",
																			},
																			CommaFodder: nil,
																		},
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Binary{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(145),
																							Column: int(31),
																						},
																						End: ast.Location{
																							Line: int(145),
																							Column: int(43),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p1422,
																					FreeVars: ast.Identifiers{
																						"from_len",
																						"i",
																					},
																				},
																				Left: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(145),
																								Column: int(31),
																							},
																							End: ast.Location{
																								Line: int(145),
																								Column: int(32),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p1422,
																						FreeVars: ast.Identifiers{
																							"i",
																						},
																					},
																					Id: "i",
																				},
																				OpFodder: ast.Fodder{},
																				Op: ast.BinaryOp(3),
																				Right: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(145),
																								Column: int(35),
																							},
																							End: ast.Location{
																								Line: int(145),
																								Column: int(43),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p1422,
																						FreeVars: ast.Identifiers{
																							"from_len",
																						},
																					},
																					Id: "from_len",
																				},
																			},
																			CommaFodder: nil,
																		},
																		ast.CommaSeparatedExpr{
																			Expr: &ast.LiteralNull{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: nil,
																				},
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																TrailingComma: false,
																TailStrict: false,
																FodderRight: nil,
																TailStrictFodder: nil,
															},
															OpFodder: ast.Fodder{},
															Op: ast.BinaryOp(12),
															Right: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(145),
																			Column: int(48),
																		},
																		End: ast.Location{
																			Line: int(145),
																			Column: int(52),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p1422,
																	FreeVars: ast.Identifiers{
																		"from",
																	},
																},
																Id: "from",
															},
														},
													},
													Fun: nil,
													CloseFodder: nil,
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
												},
											},
											Body: &ast.Local{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(149),
															Column: int(5),
														},
														End: ast.Location{
															Line: int(163),
															Column: int(30),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(1),
															Indent: int(4),
															Comment: []string{},
														},
														ast.FodderElement{
															Kind: ast.FodderKind(2),
															Blanks: int(0),
															Indent: int(4),
															Comment: []string{
																"// Return the remainder of 'str' starting with 'start_index' where",
															},
														},
														ast.FodderElement{
															Kind: ast.FodderKind(2),
															Blanks: int(0),
															Indent: int(4),
															Comment: []string{
																"// all occurrences of 'from' after 'curr_index' are replaced with 'to'.",
															},
														},
													},
													Ctx: p1335,
													FreeVars: ast.Identifiers{
														"found_at",
														"from",
														"from_len",
														"std",
														"str",
														"str_len",
														"to",
													},
												},
												Binds: ast.LocalBinds{
													ast.LocalBind{
														VarFodder: nil,
														Variable: "replace_after",
														EqFodder: nil,
														Body: &ast.Function{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(149),
																		Column: int(11),
																	},
																	End: ast.Location{
																		Line: int(156),
																		Column: int(56),
																	},
																	File: p1,
																},
																Fodder: nil,
																Ctx: p1452,
																FreeVars: ast.Identifiers{
																	"found_at",
																	"from",
																	"replace_after",
																	"std",
																	"str",
																	"str_len",
																	"to",
																},
															},
															ParenLeftFodder: ast.Fodder{},
															Parameters: []ast.Parameter{
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "start_index",
																	EqFodder: nil,
																	DefaultArg: nil,
																	CommaFodder: ast.Fodder{},
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(149),
																			Column: int(25),
																		},
																		End: ast.Location{
																			Line: int(149),
																			Column: int(36),
																		},
																		File: p1,
																	},
																},
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "curr_index",
																	EqFodder: nil,
																	DefaultArg: nil,
																	CommaFodder: ast.Fodder{},
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(149),
																			Column: int(38),
																		},
																		End: ast.Location{
																			Line: int(149),
																			Column: int(48),
																		},
																		File: p1,
																	},
																},
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "acc",
																	EqFodder: nil,
																	DefaultArg: nil,
																	CommaFodder: nil,
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(149),
																			Column: int(50),
																		},
																		End: ast.Location{
																			Line: int(149),
																			Column: int(53),
																		},
																		File: p1,
																	},
																},
															},
															TrailingComma: false,
															ParenRightFodder: ast.Fodder{},
															Body: &ast.Conditional{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(150),
																			Column: int(7),
																		},
																		End: ast.Location{
																			Line: int(156),
																			Column: int(56),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(6),
																			Comment: []string{},
																		},
																	},
																	Ctx: p1457,
																	FreeVars: ast.Identifiers{
																		"acc",
																		"curr_index",
																		"found_at",
																		"from",
																		"replace_after",
																		"start_index",
																		"std",
																		"str",
																		"str_len",
																		"to",
																	},
																},
																Cond: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(150),
																				Column: int(10),
																			},
																			End: ast.Location{
																				Line: int(150),
																				Column: int(30),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p1457,
																		FreeVars: ast.Identifiers{
																			"curr_index",
																			"str_len",
																		},
																	},
																	Left: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(150),
																					Column: int(10),
																				},
																				End: ast.Location{
																					Line: int(150),
																					Column: int(20),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p1457,
																			FreeVars: ast.Identifiers{
																				"curr_index",
																			},
																		},
																		Id: "curr_index",
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(7),
																	Right: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(150),
																					Column: int(23),
																				},
																				End: ast.Location{
																					Line: int(150),
																					Column: int(30),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p1457,
																			FreeVars: ast.Identifiers{
																				"str_len",
																			},
																		},
																		Id: "str_len",
																	},
																},
																ThenFodder: ast.Fodder{},
																BranchTrue: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(151),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(151),
																				Column: int(42),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p1457,
																		FreeVars: ast.Identifiers{
																			"acc",
																			"curr_index",
																			"start_index",
																			"std",
																			"str",
																		},
																	},
																	Left: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(151),
																					Column: int(9),
																				},
																				End: ast.Location{
																					Line: int(151),
																					Column: int(12),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(8),
																					Comment: []string{},
																				},
																			},
																			Ctx: p1457,
																			FreeVars: ast.Identifiers{
																				"acc",
																			},
																		},
																		Id: "acc",
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(3),
																	Right: &ast.Apply{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"curr_index",
																				"start_index",
																				"std",
																				"str",
																			},
																		},
																		Target: &ast.Index{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																				},
																				Id: "std",
																			},
																			LeftBracketFodder: nil,
																			Index: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: nil,
																				},
																				Value: "slice",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																			RightBracketFodder: nil,
																			Id: nil,
																		},
																		FodderLeft: nil,
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(151),
																									Column: int(15),
																								},
																								End: ast.Location{
																									Line: int(151),
																									Column: int(18),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p1457,
																							FreeVars: ast.Identifiers{
																								"str",
																							},
																						},
																						Id: "str",
																					},
																					CommaFodder: nil,
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(151),
																									Column: int(19),
																								},
																								End: ast.Location{
																									Line: int(151),
																									Column: int(30),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p1457,
																							FreeVars: ast.Identifiers{
																								"start_index",
																							},
																						},
																						Id: "start_index",
																					},
																					CommaFodder: nil,
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(151),
																									Column: int(31),
																								},
																								End: ast.Location{
																									Line: int(151),
																									Column: int(41),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p1457,
																							FreeVars: ast.Identifiers{
																								"curr_index",
																							},
																						},
																						Id: "curr_index",
																					},
																					CommaFodder: nil,
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.LiteralNull{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								File: nil,
																							},
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: nil,
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		TrailingComma: false,
																		TailStrict: false,
																		FodderRight: nil,
																		TailStrictFodder: nil,
																	},
																},
																ElseFodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(6),
																		Comment: []string{},
																	},
																},
																BranchFalse: &ast.Conditional{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(152),
																				Column: int(12),
																			},
																			End: ast.Location{
																				Line: int(156),
																				Column: int(56),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p1457,
																		FreeVars: ast.Identifiers{
																			"acc",
																			"curr_index",
																			"found_at",
																			"from",
																			"replace_after",
																			"start_index",
																			"std",
																			"str",
																			"to",
																		},
																	},
																	Cond: &ast.Apply{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(152),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(152),
																					Column: int(35),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p1457,
																			FreeVars: ast.Identifiers{
																				"curr_index",
																				"found_at",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(152),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(152),
																						Column: int(23),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p1457,
																				FreeVars: ast.Identifiers{
																					"found_at",
																				},
																			},
																			Id: "found_at",
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(152),
																									Column: int(24),
																								},
																								End: ast.Location{
																									Line: int(152),
																									Column: int(34),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p1494,
																							FreeVars: ast.Identifiers{
																								"curr_index",
																							},
																						},
																						Id: "curr_index",
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		TrailingComma: false,
																		TailStrict: false,
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																	},
																	ThenFodder: ast.Fodder{},
																	BranchTrue: &ast.Local{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(153),
																					Column: int(9),
																				},
																				End: ast.Location{
																					Line: int(154),
																					Column: int(84),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(8),
																					Comment: []string{},
																				},
																			},
																			Ctx: p1457,
																			FreeVars: ast.Identifiers{
																				"acc",
																				"curr_index",
																				"from",
																				"replace_after",
																				"start_index",
																				"std",
																				"str",
																				"to",
																			},
																		},
																		Binds: ast.LocalBinds{
																			ast.LocalBind{
																				VarFodder: ast.Fodder{},
																				Variable: "new_index",
																				EqFodder: ast.Fodder{},
																				Body: &ast.Binary{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(153),
																								Column: int(27),
																							},
																							End: ast.Location{
																								Line: int(153),
																								Column: int(56),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p1501,
																						FreeVars: ast.Identifiers{
																							"curr_index",
																							"from",
																							"std",
																						},
																					},
																					Left: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(153),
																									Column: int(27),
																								},
																								End: ast.Location{
																									Line: int(153),
																									Column: int(37),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p1501,
																							FreeVars: ast.Identifiers{
																								"curr_index",
																							},
																						},
																						Id: "curr_index",
																					},
																					OpFodder: ast.Fodder{},
																					Op: ast.BinaryOp(3),
																					Right: &ast.Apply{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(153),
																									Column: int(40),
																								},
																								End: ast.Location{
																									Line: int(153),
																									Column: int(56),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p1501,
																							FreeVars: ast.Identifiers{
																								"from",
																								"std",
																							},
																						},
																						Target: &ast.Index{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(153),
																										Column: int(40),
																									},
																									End: ast.Location{
																										Line: int(153),
																										Column: int(50),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p1501,
																								FreeVars: ast.Identifiers{
																									"std",
																								},
																							},
																							Target: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(153),
																											Column: int(40),
																										},
																										End: ast.Location{
																											Line: int(153),
																											Column: int(43),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: nil,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																								},
																								Id: "std",
																							},
																							LeftBracketFodder: ast.Fodder{},
																							Index: &ast.LiteralString{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										File: nil,
																									},
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: nil,
																								},
																								Value: "length",
																								Kind: ast.LiteralStringKind(1),
																								BlockIndent: "",
																								BlockTermIndent: "",
																							},
																							RightBracketFodder: ast.Fodder{},
																							Id: nil,
																						},
																						FodderLeft: ast.Fodder{},
																						Arguments: ast.Arguments{
																							Positional: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(153),
																													Column: int(51),
																												},
																												End: ast.Location{
																													Line: int(153),
																													Column: int(55),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p1514,
																											FreeVars: ast.Identifiers{
																												"from",
																											},
																										},
																										Id: "from",
																									},
																									CommaFodder: nil,
																								},
																							},
																							Named: nil,
																						},
																						TrailingComma: false,
																						TailStrict: false,
																						FodderRight: ast.Fodder{},
																						TailStrictFodder: nil,
																					},
																				},
																				Fun: nil,
																				CloseFodder: ast.Fodder{},
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(153),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(153),
																						Column: int(56),
																					},
																					File: p1,
																				},
																			},
																		},
																		Body: &ast.Apply{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(154),
																						Column: int(9),
																					},
																					End: ast.Location{
																						Line: int(154),
																						Column: int(84),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p1457,
																				FreeVars: ast.Identifiers{
																					"acc",
																					"curr_index",
																					"new_index",
																					"replace_after",
																					"start_index",
																					"std",
																					"str",
																					"to",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(154),
																							Column: int(9),
																						},
																						End: ast.Location{
																							Line: int(154),
																							Column: int(22),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(8),
																							Comment: []string{},
																						},
																					},
																					Ctx: p1457,
																					FreeVars: ast.Identifiers{
																						"replace_after",
																					},
																				},
																				Id: "replace_after",
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(154),
																										Column: int(23),
																									},
																									End: ast.Location{
																										Line: int(154),
																										Column: int(32),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p1523,
																								FreeVars: ast.Identifiers{
																									"new_index",
																								},
																							},
																							Id: "new_index",
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(154),
																										Column: int(34),
																									},
																									End: ast.Location{
																										Line: int(154),
																										Column: int(43),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p1523,
																								FreeVars: ast.Identifiers{
																									"new_index",
																								},
																							},
																							Id: "new_index",
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Binary{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(154),
																										Column: int(45),
																									},
																									End: ast.Location{
																										Line: int(154),
																										Column: int(83),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p1523,
																								FreeVars: ast.Identifiers{
																									"acc",
																									"curr_index",
																									"start_index",
																									"std",
																									"str",
																									"to",
																								},
																							},
																							Left: &ast.Binary{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(154),
																											Column: int(45),
																										},
																										End: ast.Location{
																											Line: int(154),
																											Column: int(78),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p1523,
																									FreeVars: ast.Identifiers{
																										"acc",
																										"curr_index",
																										"start_index",
																										"std",
																										"str",
																									},
																								},
																								Left: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(154),
																												Column: int(45),
																											},
																											End: ast.Location{
																												Line: int(154),
																												Column: int(48),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p1523,
																										FreeVars: ast.Identifiers{
																											"acc",
																										},
																									},
																									Id: "acc",
																								},
																								OpFodder: ast.Fodder{},
																								Op: ast.BinaryOp(3),
																								Right: &ast.Apply{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											File: nil,
																										},
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: ast.Identifiers{
																											"curr_index",
																											"start_index",
																											"std",
																											"str",
																										},
																									},
																									Target: &ast.Index{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												File: nil,
																											},
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"std",
																											},
																										},
																										Target: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													File: nil,
																												},
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																											},
																											Id: "std",
																										},
																										LeftBracketFodder: nil,
																										Index: &ast.LiteralString{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													File: nil,
																												},
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: nil,
																											},
																											Value: "slice",
																											Kind: ast.LiteralStringKind(1),
																											BlockIndent: "",
																											BlockTermIndent: "",
																										},
																										RightBracketFodder: nil,
																										Id: nil,
																									},
																									FodderLeft: nil,
																									Arguments: ast.Arguments{
																										Positional: []ast.CommaSeparatedExpr{
																											ast.CommaSeparatedExpr{
																												Expr: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(154),
																																Column: int(51),
																															},
																															End: ast.Location{
																																Line: int(154),
																																Column: int(54),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p1523,
																														FreeVars: ast.Identifiers{
																															"str",
																														},
																													},
																													Id: "str",
																												},
																												CommaFodder: nil,
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(154),
																																Column: int(55),
																															},
																															End: ast.Location{
																																Line: int(154),
																																Column: int(66),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p1523,
																														FreeVars: ast.Identifiers{
																															"start_index",
																														},
																													},
																													Id: "start_index",
																												},
																												CommaFodder: nil,
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(154),
																																Column: int(67),
																															},
																															End: ast.Location{
																																Line: int(154),
																																Column: int(77),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p1523,
																														FreeVars: ast.Identifiers{
																															"curr_index",
																														},
																													},
																													Id: "curr_index",
																												},
																												CommaFodder: nil,
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.LiteralNull{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															File: nil,
																														},
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: nil,
																													},
																												},
																												CommaFodder: nil,
																											},
																										},
																										Named: nil,
																									},
																									TrailingComma: false,
																									TailStrict: false,
																									FodderRight: nil,
																									TailStrictFodder: nil,
																								},
																							},
																							OpFodder: ast.Fodder{},
																							Op: ast.BinaryOp(3),
																							Right: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(154),
																											Column: int(81),
																										},
																										End: ast.Location{
																											Line: int(154),
																											Column: int(83),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p1523,
																									FreeVars: ast.Identifiers{
																										"to",
																									},
																								},
																								Id: "to",
																							},
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			TrailingComma: false,
																			TailStrict: true,
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: ast.Fodder{},
																		},
																	},
																	ElseFodder: ast.Fodder{
																		ast.FodderElement{
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(6),
																			Comment: []string{},
																		},
																	},
																	BranchFalse: &ast.Apply{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(156),
																					Column: int(9),
																				},
																				End: ast.Location{
																					Line: int(156),
																					Column: int(56),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p1457,
																			FreeVars: ast.Identifiers{
																				"acc",
																				"curr_index",
																				"replace_after",
																				"start_index",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(156),
																						Column: int(9),
																					},
																					End: ast.Location{
																						Line: int(156),
																						Column: int(22),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(8),
																						Comment: []string{},
																					},
																				},
																				Ctx: p1457,
																				FreeVars: ast.Identifiers{
																					"replace_after",
																				},
																			},
																			Id: "replace_after",
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(156),
																									Column: int(23),
																								},
																								End: ast.Location{
																									Line: int(156),
																									Column: int(34),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p1558,
																							FreeVars: ast.Identifiers{
																								"start_index",
																							},
																						},
																						Id: "start_index",
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Binary{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(156),
																									Column: int(36),
																								},
																								End: ast.Location{
																									Line: int(156),
																									Column: int(50),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p1558,
																							FreeVars: ast.Identifiers{
																								"curr_index",
																							},
																						},
																						Left: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(156),
																										Column: int(36),
																									},
																									End: ast.Location{
																										Line: int(156),
																										Column: int(46),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p1558,
																								FreeVars: ast.Identifiers{
																									"curr_index",
																								},
																							},
																							Id: "curr_index",
																						},
																						OpFodder: ast.Fodder{},
																						Op: ast.BinaryOp(3),
																						Right: &ast.LiteralNumber{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(156),
																										Column: int(49),
																									},
																									End: ast.Location{
																										Line: int(156),
																										Column: int(50),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p1558,
																								FreeVars: nil,
																							},
																							OriginalString: "1",
																						},
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(156),
																									Column: int(52),
																								},
																								End: ast.Location{
																									Line: int(156),
																									Column: int(55),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p1558,
																							FreeVars: ast.Identifiers{
																								"acc",
																							},
																						},
																						Id: "acc",
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		TrailingComma: false,
																		TailStrict: true,
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: ast.Fodder{},
																	},
																},
															},
														},
														Fun: nil,
														CloseFodder: nil,
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
													},
												},
												Body: &ast.Conditional{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(160),
																Column: int(5),
															},
															End: ast.Location{
																Line: int(163),
																Column: int(30),
															},
															File: p1,
														},
														Fodder: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(1),
																Indent: int(4),
																Comment: []string{},
															},
															ast.FodderElement{
																Kind: ast.FodderKind(2),
																Blanks: int(0),
																Indent: int(4),
																Comment: []string{
																	"// if from_len==1, then we replace by splitting and rejoining the",
																},
															},
															ast.FodderElement{
																Kind: ast.FodderKind(2),
																Blanks: int(0),
																Indent: int(4),
																Comment: []string{
																	"// string which is much faster than recursing on replace_after",
																},
															},
														},
														Ctx: p1335,
														FreeVars: ast.Identifiers{
															"from",
															"from_len",
															"replace_after",
															"std",
															"str",
															"to",
														},
													},
													Cond: &ast.Binary{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(160),
																	Column: int(8),
																},
																End: ast.Location{
																	Line: int(160),
																	Column: int(21),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p1335,
															FreeVars: ast.Identifiers{
																"from_len",
															},
														},
														Left: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(160),
																		Column: int(8),
																	},
																	End: ast.Location{
																		Line: int(160),
																		Column: int(16),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p1335,
																FreeVars: ast.Identifiers{
																	"from_len",
																},
															},
															Id: "from_len",
														},
														OpFodder: ast.Fodder{},
														Op: ast.BinaryOp(12),
														Right: &ast.LiteralNumber{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(160),
																		Column: int(20),
																	},
																	End: ast.Location{
																		Line: int(160),
																		Column: int(21),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p1335,
																FreeVars: nil,
															},
															OriginalString: "1",
														},
													},
													ThenFodder: ast.Fodder{},
													BranchTrue: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(161),
																	Column: int(7),
																},
																End: ast.Location{
																	Line: int(161),
																	Column: int(41),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p1335,
															FreeVars: ast.Identifiers{
																"from",
																"std",
																"str",
																"to",
															},
														},
														Target: &ast.Index{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(161),
																		Column: int(7),
																	},
																	End: ast.Location{
																		Line: int(161),
																		Column: int(15),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p1335,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Target: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(161),
																			Column: int(7),
																		},
																		End: ast.Location{
																			Line: int(161),
																			Column: int(10),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(6),
																			Comment: []string{},
																		},
																	},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Id: "std",
															},
															LeftBracketFodder: ast.Fodder{},
															Index: &ast.LiteralString{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: nil,
																},
																Value: "join",
																Kind: ast.LiteralStringKind(1),
																BlockIndent: "",
																BlockTermIndent: "",
															},
															RightBracketFodder: ast.Fodder{},
															Id: nil,
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(161),
																					Column: int(16),
																				},
																				End: ast.Location{
																					Line: int(161),
																					Column: int(18),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p1587,
																			FreeVars: ast.Identifiers{
																				"to",
																			},
																		},
																		Id: "to",
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Apply{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(161),
																					Column: int(20),
																				},
																				End: ast.Location{
																					Line: int(161),
																					Column: int(40),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p1587,
																			FreeVars: ast.Identifiers{
																				"from",
																				"std",
																				"str",
																			},
																		},
																		Target: &ast.Index{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(161),
																						Column: int(20),
																					},
																					End: ast.Location{
																						Line: int(161),
																						Column: int(29),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p1587,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(161),
																							Column: int(20),
																						},
																						End: ast.Location{
																							Line: int(161),
																							Column: int(23),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																				},
																				Id: "std",
																			},
																			LeftBracketFodder: ast.Fodder{},
																			Index: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: nil,
																				},
																				Value: "split",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																			RightBracketFodder: ast.Fodder{},
																			Id: nil,
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(161),
																									Column: int(30),
																								},
																								End: ast.Location{
																									Line: int(161),
																									Column: int(33),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p1598,
																							FreeVars: ast.Identifiers{
																								"str",
																							},
																						},
																						Id: "str",
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(161),
																									Column: int(35),
																								},
																								End: ast.Location{
																									Line: int(161),
																									Column: int(39),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p1598,
																							FreeVars: ast.Identifiers{
																								"from",
																							},
																						},
																						Id: "from",
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		TrailingComma: false,
																		TailStrict: false,
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														TrailingComma: false,
														TailStrict: false,
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
													},
													ElseFodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(4),
															Comment: []string{},
														},
													},
													BranchFalse: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(163),
																	Column: int(7),
																},
																End: ast.Location{
																	Line: int(163),
																	Column: int(30),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p1335,
															FreeVars: ast.Identifiers{
																"replace_after",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(163),
																		Column: int(7),
																	},
																	End: ast.Location{
																		Line: int(163),
																		Column: int(20),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(6),
																		Comment: []string{},
																	},
																},
																Ctx: p1335,
																FreeVars: ast.Identifiers{
																	"replace_after",
																},
															},
															Id: "replace_after",
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.LiteralNumber{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(163),
																					Column: int(21),
																				},
																				End: ast.Location{
																					Line: int(163),
																					Column: int(22),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p1610,
																			FreeVars: nil,
																		},
																		OriginalString: "0",
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.LiteralNumber{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(163),
																					Column: int(24),
																				},
																				End: ast.Location{
																					Line: int(163),
																					Column: int(25),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p1610,
																			FreeVars: nil,
																		},
																		OriginalString: "0",
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(163),
																					Column: int(27),
																				},
																				End: ast.Location{
																					Line: int(163),
																					Column: int(29),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p1610,
																			FreeVars: nil,
																		},
																		Value: "",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														TrailingComma: false,
														TailStrict: false,
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
													},
												},
											},
										},
									},
								},
								ElseFodder: nil,
								BranchFalse: &ast.Error{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Expr: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(138),
													Column: int(25),
												},
												End: ast.Location{
													Line: int(138),
													Column: int(65),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1335,
											FreeVars: nil,
										},
										Value: "'from' string must not be zero length.",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
								},
							},
							ElseFodder: nil,
							BranchFalse: &ast.Error{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Expr: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "Assertion failed",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
							},
						},
						ElseFodder: nil,
						BranchFalse: &ast.Error{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: nil,
							},
							Expr: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Value: "Assertion failed",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
						},
					},
					ElseFodder: nil,
					BranchFalse: &ast.Error{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
							Fodder: nil,
							Ctx: nil,
							FreeVars: nil,
						},
						Expr: &ast.LiteralString{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: nil,
							},
							Value: "Assertion failed",
							Kind: ast.LiteralStringKind(1),
							BlockIndent: "",
							BlockTermIndent: "",
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(134),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(163),
					Column: int(30),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "asciiUpper",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(165),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(165),
								Column: int(17),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(166),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(171),
								Column: int(59),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p1627,
						FreeVars: ast.Identifiers{
							"std",
							"str",
						},
					},
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Variable: "cp",
							EqFodder: ast.Fodder{},
							Body: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(166),
											Column: int(16),
										},
										End: ast.Location{
											Line: int(166),
											Column: int(29),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p1631,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(166),
												Column: int(16),
											},
											End: ast.Location{
												Line: int(166),
												Column: int(19),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "codepoint",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							Fun: nil,
							CloseFodder: ast.Fodder{},
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(166),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(166),
									Column: int(29),
								},
								File: p1,
							},
						},
					},
					Body: &ast.Local{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(167),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(171),
									Column: int(59),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
									Comment: []string{},
								},
							},
							Ctx: p1627,
							FreeVars: ast.Identifiers{
								"cp",
								"std",
								"str",
							},
						},
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: nil,
								Variable: "up_letter",
								EqFodder: nil,
								Body: &ast.Function{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(167),
												Column: int(11),
											},
											End: ast.Location{
												Line: int(170),
												Column: int(8),
											},
											File: p1,
										},
										Fodder: nil,
										Ctx: p1641,
										FreeVars: ast.Identifiers{
											"cp",
											"std",
										},
									},
									ParenLeftFodder: ast.Fodder{},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "c",
											EqFodder: nil,
											DefaultArg: nil,
											CommaFodder: nil,
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(167),
													Column: int(21),
												},
												End: ast.Location{
													Line: int(167),
													Column: int(22),
												},
												File: p1,
											},
										},
									},
									TrailingComma: false,
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Conditional{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(167),
													Column: int(26),
												},
												End: ast.Location{
													Line: int(170),
													Column: int(8),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1645,
											FreeVars: ast.Identifiers{
												"c",
												"cp",
												"std",
											},
										},
										Cond: &ast.Binary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(167),
														Column: int(29),
													},
													End: ast.Location{
														Line: int(167),
														Column: int(55),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p1645,
												FreeVars: ast.Identifiers{
													"c",
													"cp",
												},
											},
											Left: &ast.Binary{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(167),
															Column: int(29),
														},
														End: ast.Location{
															Line: int(167),
															Column: int(40),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p1645,
													FreeVars: ast.Identifiers{
														"c",
														"cp",
													},
												},
												Left: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(167),
																Column: int(29),
															},
															End: ast.Location{
																Line: int(167),
																Column: int(34),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p1645,
														FreeVars: ast.Identifiers{
															"c",
															"cp",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(167),
																	Column: int(29),
																},
																End: ast.Location{
																	Line: int(167),
																	Column: int(31),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p1645,
															FreeVars: ast.Identifiers{
																"cp",
															},
														},
														Id: "cp",
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(167),
																				Column: int(32),
																			},
																			End: ast.Location{
																				Line: int(167),
																				Column: int(33),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p1657,
																		FreeVars: ast.Identifiers{
																			"c",
																		},
																	},
																	Id: "c",
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
												},
												OpFodder: ast.Fodder{},
												Op: ast.BinaryOp(8),
												Right: &ast.LiteralNumber{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(167),
																Column: int(38),
															},
															End: ast.Location{
																Line: int(167),
																Column: int(40),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p1645,
														FreeVars: nil,
													},
													OriginalString: "97",
												},
											},
											OpFodder: ast.Fodder{},
											Op: ast.BinaryOp(17),
											Right: &ast.Binary{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(167),
															Column: int(44),
														},
														End: ast.Location{
															Line: int(167),
															Column: int(55),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p1645,
													FreeVars: ast.Identifiers{
														"c",
														"cp",
													},
												},
												Left: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(167),
																Column: int(44),
															},
															End: ast.Location{
																Line: int(167),
																Column: int(49),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p1645,
														FreeVars: ast.Identifiers{
															"c",
															"cp",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(167),
																	Column: int(44),
																},
																End: ast.Location{
																	Line: int(167),
																	Column: int(46),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p1645,
															FreeVars: ast.Identifiers{
																"cp",
															},
														},
														Id: "cp",
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(167),
																				Column: int(47),
																			},
																			End: ast.Location{
																				Line: int(167),
																				Column: int(48),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p1668,
																		FreeVars: ast.Identifiers{
																			"c",
																		},
																	},
																	Id: "c",
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
												},
												OpFodder: ast.Fodder{},
												Op: ast.BinaryOp(9),
												Right: &ast.LiteralNumber{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(167),
																Column: int(52),
															},
															End: ast.Location{
																Line: int(167),
																Column: int(55),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p1645,
														FreeVars: nil,
													},
													OriginalString: "123",
												},
											},
										},
										ThenFodder: ast.Fodder{},
										BranchTrue: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(168),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(168),
														Column: int(27),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p1645,
												FreeVars: ast.Identifiers{
													"c",
													"cp",
													"std",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(168),
															Column: int(7),
														},
														End: ast.Location{
															Line: int(168),
															Column: int(15),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p1645,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(168),
																Column: int(7),
															},
															End: ast.Location{
																Line: int(168),
																Column: int(10),
															},
															File: p1,
														},
														Fodder: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(6),
																Comment: []string{},
															},
														},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "char",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Binary{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(168),
																		Column: int(16),
																	},
																	End: ast.Location{
																		Line: int(168),
																		Column: int(26),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p1681,
																FreeVars: ast.Identifiers{
																	"c",
																	"cp",
																},
															},
															Left: &ast.Apply{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(168),
																			Column: int(16),
																		},
																		End: ast.Location{
																			Line: int(168),
																			Column: int(21),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p1681,
																	FreeVars: ast.Identifiers{
																		"c",
																		"cp",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(168),
																				Column: int(16),
																			},
																			End: ast.Location{
																				Line: int(168),
																				Column: int(18),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p1681,
																		FreeVars: ast.Identifiers{
																			"cp",
																		},
																	},
																	Id: "cp",
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(168),
																							Column: int(19),
																						},
																						End: ast.Location{
																							Line: int(168),
																							Column: int(20),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p1689,
																					FreeVars: ast.Identifiers{
																						"c",
																					},
																				},
																				Id: "c",
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																TrailingComma: false,
																TailStrict: false,
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
															},
															OpFodder: ast.Fodder{},
															Op: ast.BinaryOp(4),
															Right: &ast.LiteralNumber{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(168),
																			Column: int(24),
																		},
																		End: ast.Location{
																			Line: int(168),
																			Column: int(26),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p1681,
																	FreeVars: nil,
																},
																OriginalString: "32",
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(4),
												Comment: []string{},
											},
										},
										BranchFalse: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(170),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(170),
														Column: int(8),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
														Comment: []string{},
													},
												},
												Ctx: p1645,
												FreeVars: ast.Identifiers{
													"c",
												},
											},
											Id: "c",
										},
									},
								},
								Fun: nil,
								CloseFodder: nil,
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
							},
						},
						Body: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(171),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(171),
										Column: int(59),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p1627,
								FreeVars: ast.Identifiers{
									"std",
									"str",
									"up_letter",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(171),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(171),
											Column: int(13),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p1627,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(171),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(171),
												Column: int(8),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(4),
												Comment: []string{},
											},
										},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "join",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(171),
														Column: int(14),
													},
													End: ast.Location{
														Line: int(171),
														Column: int(16),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p1706,
												FreeVars: nil,
											},
											Value: "",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(171),
														Column: int(18),
													},
													End: ast.Location{
														Line: int(171),
														Column: int(58),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p1706,
												FreeVars: ast.Identifiers{
													"std",
													"str",
													"up_letter",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(171),
															Column: int(18),
														},
														End: ast.Location{
															Line: int(171),
															Column: int(25),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p1706,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(171),
																Column: int(18),
															},
															End: ast.Location{
																Line: int(171),
																Column: int(21),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "map",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(171),
																		Column: int(26),
																	},
																	End: ast.Location{
																		Line: int(171),
																		Column: int(35),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p1716,
																FreeVars: ast.Identifiers{
																	"up_letter",
																},
															},
															Id: "up_letter",
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Apply{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(171),
																		Column: int(37),
																	},
																	End: ast.Location{
																		Line: int(171),
																		Column: int(57),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p1716,
																FreeVars: ast.Identifiers{
																	"std",
																	"str",
																},
															},
															Target: &ast.Index{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(171),
																			Column: int(37),
																		},
																		End: ast.Location{
																			Line: int(171),
																			Column: int(52),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p1716,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(171),
																				Column: int(37),
																			},
																			End: ast.Location{
																				Line: int(171),
																				Column: int(40),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Id: "std",
																},
																LeftBracketFodder: ast.Fodder{},
																Index: &ast.LiteralString{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: nil,
																	},
																	Value: "stringChars",
																	Kind: ast.LiteralStringKind(1),
																	BlockIndent: "",
																	BlockTermIndent: "",
																},
																RightBracketFodder: ast.Fodder{},
																Id: nil,
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(171),
																						Column: int(53),
																					},
																					End: ast.Location{
																						Line: int(171),
																						Column: int(56),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p1727,
																				FreeVars: ast.Identifiers{
																					"str",
																				},
																			},
																			Id: "str",
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															TrailingComma: false,
															TailStrict: false,
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(165),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(171),
					Column: int(59),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "asciiLower",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(173),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(173),
								Column: int(17),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(174),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(179),
								Column: int(61),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p1735,
						FreeVars: ast.Identifiers{
							"std",
							"str",
						},
					},
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Variable: "cp",
							EqFodder: ast.Fodder{},
							Body: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(174),
											Column: int(16),
										},
										End: ast.Location{
											Line: int(174),
											Column: int(29),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p1739,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(174),
												Column: int(16),
											},
											End: ast.Location{
												Line: int(174),
												Column: int(19),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "codepoint",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							Fun: nil,
							CloseFodder: ast.Fodder{},
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(174),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(174),
									Column: int(29),
								},
								File: p1,
							},
						},
					},
					Body: &ast.Local{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(175),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(179),
									Column: int(61),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
									Comment: []string{},
								},
							},
							Ctx: p1735,
							FreeVars: ast.Identifiers{
								"cp",
								"std",
								"str",
							},
						},
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: nil,
								Variable: "down_letter",
								EqFodder: nil,
								Body: &ast.Function{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(175),
												Column: int(11),
											},
											End: ast.Location{
												Line: int(178),
												Column: int(8),
											},
											File: p1,
										},
										Fodder: nil,
										Ctx: p1749,
										FreeVars: ast.Identifiers{
											"cp",
											"std",
										},
									},
									ParenLeftFodder: ast.Fodder{},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "c",
											EqFodder: nil,
											DefaultArg: nil,
											CommaFodder: nil,
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(175),
													Column: int(23),
												},
												End: ast.Location{
													Line: int(175),
													Column: int(24),
												},
												File: p1,
											},
										},
									},
									TrailingComma: false,
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Conditional{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(175),
													Column: int(28),
												},
												End: ast.Location{
													Line: int(178),
													Column: int(8),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1753,
											FreeVars: ast.Identifiers{
												"c",
												"cp",
												"std",
											},
										},
										Cond: &ast.Binary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(175),
														Column: int(31),
													},
													End: ast.Location{
														Line: int(175),
														Column: int(56),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p1753,
												FreeVars: ast.Identifiers{
													"c",
													"cp",
												},
											},
											Left: &ast.Binary{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(175),
															Column: int(31),
														},
														End: ast.Location{
															Line: int(175),
															Column: int(42),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p1753,
													FreeVars: ast.Identifiers{
														"c",
														"cp",
													},
												},
												Left: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(175),
																Column: int(31),
															},
															End: ast.Location{
																Line: int(175),
																Column: int(36),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p1753,
														FreeVars: ast.Identifiers{
															"c",
															"cp",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(175),
																	Column: int(31),
																},
																End: ast.Location{
																	Line: int(175),
																	Column: int(33),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p1753,
															FreeVars: ast.Identifiers{
																"cp",
															},
														},
														Id: "cp",
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(175),
																				Column: int(34),
																			},
																			End: ast.Location{
																				Line: int(175),
																				Column: int(35),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p1765,
																		FreeVars: ast.Identifiers{
																			"c",
																		},
																	},
																	Id: "c",
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
												},
												OpFodder: ast.Fodder{},
												Op: ast.BinaryOp(8),
												Right: &ast.LiteralNumber{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(175),
																Column: int(40),
															},
															End: ast.Location{
																Line: int(175),
																Column: int(42),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p1753,
														FreeVars: nil,
													},
													OriginalString: "65",
												},
											},
											OpFodder: ast.Fodder{},
											Op: ast.BinaryOp(17),
											Right: &ast.Binary{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(175),
															Column: int(46),
														},
														End: ast.Location{
															Line: int(175),
															Column: int(56),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p1753,
													FreeVars: ast.Identifiers{
														"c",
														"cp",
													},
												},
												Left: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(175),
																Column: int(46),
															},
															End: ast.Location{
																Line: int(175),
																Column: int(51),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p1753,
														FreeVars: ast.Identifiers{
															"c",
															"cp",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(175),
																	Column: int(46),
																},
																End: ast.Location{
																	Line: int(175),
																	Column: int(48),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p1753,
															FreeVars: ast.Identifiers{
																"cp",
															},
														},
														Id: "cp",
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(175),
																				Column: int(49),
																			},
																			End: ast.Location{
																				Line: int(175),
																				Column: int(50),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p1776,
																		FreeVars: ast.Identifiers{
																			"c",
																		},
																	},
																	Id: "c",
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
												},
												OpFodder: ast.Fodder{},
												Op: ast.BinaryOp(9),
												Right: &ast.LiteralNumber{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(175),
																Column: int(54),
															},
															End: ast.Location{
																Line: int(175),
																Column: int(56),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p1753,
														FreeVars: nil,
													},
													OriginalString: "91",
												},
											},
										},
										ThenFodder: ast.Fodder{},
										BranchTrue: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(176),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(176),
														Column: int(27),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p1753,
												FreeVars: ast.Identifiers{
													"c",
													"cp",
													"std",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(176),
															Column: int(7),
														},
														End: ast.Location{
															Line: int(176),
															Column: int(15),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p1753,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(176),
																Column: int(7),
															},
															End: ast.Location{
																Line: int(176),
																Column: int(10),
															},
															File: p1,
														},
														Fodder: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(6),
																Comment: []string{},
															},
														},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "char",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Binary{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(176),
																		Column: int(16),
																	},
																	End: ast.Location{
																		Line: int(176),
																		Column: int(26),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p1789,
																FreeVars: ast.Identifiers{
																	"c",
																	"cp",
																},
															},
															Left: &ast.Apply{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(176),
																			Column: int(16),
																		},
																		End: ast.Location{
																			Line: int(176),
																			Column: int(21),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p1789,
																	FreeVars: ast.Identifiers{
																		"c",
																		"cp",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(176),
																				Column: int(16),
																			},
																			End: ast.Location{
																				Line: int(176),
																				Column: int(18),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p1789,
																		FreeVars: ast.Identifiers{
																			"cp",
																		},
																	},
																	Id: "cp",
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(176),
																							Column: int(19),
																						},
																						End: ast.Location{
																							Line: int(176),
																							Column: int(20),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p1797,
																					FreeVars: ast.Identifiers{
																						"c",
																					},
																				},
																				Id: "c",
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																TrailingComma: false,
																TailStrict: false,
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
															},
															OpFodder: ast.Fodder{},
															Op: ast.BinaryOp(3),
															Right: &ast.LiteralNumber{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(176),
																			Column: int(24),
																		},
																		End: ast.Location{
																			Line: int(176),
																			Column: int(26),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p1789,
																	FreeVars: nil,
																},
																OriginalString: "32",
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(4),
												Comment: []string{},
											},
										},
										BranchFalse: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(178),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(178),
														Column: int(8),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
														Comment: []string{},
													},
												},
												Ctx: p1753,
												FreeVars: ast.Identifiers{
													"c",
												},
											},
											Id: "c",
										},
									},
								},
								Fun: nil,
								CloseFodder: nil,
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
							},
						},
						Body: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(179),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(179),
										Column: int(61),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p1735,
								FreeVars: ast.Identifiers{
									"down_letter",
									"std",
									"str",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(179),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(179),
											Column: int(13),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p1735,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(179),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(179),
												Column: int(8),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(4),
												Comment: []string{},
											},
										},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "join",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(179),
														Column: int(14),
													},
													End: ast.Location{
														Line: int(179),
														Column: int(16),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p1814,
												FreeVars: nil,
											},
											Value: "",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(179),
														Column: int(18),
													},
													End: ast.Location{
														Line: int(179),
														Column: int(60),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p1814,
												FreeVars: ast.Identifiers{
													"down_letter",
													"std",
													"str",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(179),
															Column: int(18),
														},
														End: ast.Location{
															Line: int(179),
															Column: int(25),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p1814,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(179),
																Column: int(18),
															},
															End: ast.Location{
																Line: int(179),
																Column: int(21),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "map",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(179),
																		Column: int(26),
																	},
																	End: ast.Location{
																		Line: int(179),
																		Column: int(37),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p1824,
																FreeVars: ast.Identifiers{
																	"down_letter",
																},
															},
															Id: "down_letter",
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Apply{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(179),
																		Column: int(39),
																	},
																	End: ast.Location{
																		Line: int(179),
																		Column: int(59),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p1824,
																FreeVars: ast.Identifiers{
																	"std",
																	"str",
																},
															},
															Target: &ast.Index{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(179),
																			Column: int(39),
																		},
																		End: ast.Location{
																			Line: int(179),
																			Column: int(54),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p1824,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(179),
																				Column: int(39),
																			},
																			End: ast.Location{
																				Line: int(179),
																				Column: int(42),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Id: "std",
																},
																LeftBracketFodder: ast.Fodder{},
																Index: &ast.LiteralString{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: nil,
																	},
																	Value: "stringChars",
																	Kind: ast.LiteralStringKind(1),
																	BlockIndent: "",
																	BlockTermIndent: "",
																},
																RightBracketFodder: ast.Fodder{},
																Id: nil,
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(179),
																						Column: int(55),
																					},
																					End: ast.Location{
																						Line: int(179),
																						Column: int(58),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p1835,
																				FreeVars: ast.Identifiers{
																					"str",
																				},
																			},
																			Id: "str",
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															TrailingComma: false,
															TailStrict: false,
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(173),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(179),
					Column: int(61),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "range",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "from",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(181),
								Column: int(9),
							},
							End: ast.Location{
								Line: int(181),
								Column: int(13),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "to",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(181),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(181),
								Column: int(17),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(182),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(182),
								Column: int(55),
							},
							File: p1,
						},
						Fodder: ast.Fodder{},
						Ctx: p1842,
						FreeVars: ast.Identifiers{
							"from",
							"std",
							"to",
						},
					},
					Target: &ast.Index{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(182),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(182),
									Column: int(18),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p1842,
							FreeVars: ast.Identifiers{
								"std",
							},
						},
						Target: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(182),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(182),
										Column: int(8),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
										Comment: []string{},
									},
								},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Id: "std",
						},
						LeftBracketFodder: ast.Fodder{},
						Index: &ast.LiteralString{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: nil,
							},
							Value: "makeArray",
							Kind: ast.LiteralStringKind(1),
							BlockIndent: "",
							BlockTermIndent: "",
						},
						RightBracketFodder: ast.Fodder{},
						Id: nil,
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Binary{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(182),
												Column: int(19),
											},
											End: ast.Location{
												Line: int(182),
												Column: int(32),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p1852,
										FreeVars: ast.Identifiers{
											"from",
											"to",
										},
									},
									Left: &ast.Binary{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(182),
													Column: int(19),
												},
												End: ast.Location{
													Line: int(182),
													Column: int(28),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1852,
											FreeVars: ast.Identifiers{
												"from",
												"to",
											},
										},
										Left: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(182),
														Column: int(19),
													},
													End: ast.Location{
														Line: int(182),
														Column: int(21),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p1852,
												FreeVars: ast.Identifiers{
													"to",
												},
											},
											Id: "to",
										},
										OpFodder: ast.Fodder{},
										Op: ast.BinaryOp(4),
										Right: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(182),
														Column: int(24),
													},
													End: ast.Location{
														Line: int(182),
														Column: int(28),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p1852,
												FreeVars: ast.Identifiers{
													"from",
												},
											},
											Id: "from",
										},
									},
									OpFodder: ast.Fodder{},
									Op: ast.BinaryOp(3),
									Right: &ast.LiteralNumber{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(182),
													Column: int(31),
												},
												End: ast.Location{
													Line: int(182),
													Column: int(32),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1852,
											FreeVars: nil,
										},
										OriginalString: "1",
									},
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.Function{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(182),
												Column: int(34),
											},
											End: ast.Location{
												Line: int(182),
												Column: int(54),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p1852,
										FreeVars: ast.Identifiers{
											"from",
										},
									},
									ParenLeftFodder: ast.Fodder{},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "i",
											EqFodder: nil,
											DefaultArg: nil,
											CommaFodder: nil,
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(182),
													Column: int(43),
												},
												End: ast.Location{
													Line: int(182),
													Column: int(44),
												},
												File: p1,
											},
										},
									},
									TrailingComma: false,
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Binary{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(182),
													Column: int(46),
												},
												End: ast.Location{
													Line: int(182),
													Column: int(54),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1865,
											FreeVars: ast.Identifiers{
												"from",
												"i",
											},
										},
										Left: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(182),
														Column: int(46),
													},
													End: ast.Location{
														Line: int(182),
														Column: int(47),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p1865,
												FreeVars: ast.Identifiers{
													"i",
												},
											},
											Id: "i",
										},
										OpFodder: ast.Fodder{},
										Op: ast.BinaryOp(3),
										Right: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(182),
														Column: int(50),
													},
													End: ast.Location{
														Line: int(182),
														Column: int(54),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p1865,
												FreeVars: ast.Identifiers{
													"from",
												},
											},
											Id: "from",
										},
									},
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					TrailingComma: false,
					TailStrict: false,
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(181),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(182),
					Column: int(55),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "repeat",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "what",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(184),
								Column: int(10),
							},
							End: ast.Location{
								Line: int(184),
								Column: int(14),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "count",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(184),
								Column: int(16),
							},
							End: ast.Location{
								Line: int(184),
								Column: int(21),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(185),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(189),
								Column: int(61),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p1877,
						FreeVars: ast.Identifiers{
							"count",
							"std",
							"what",
						},
					},
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Variable: "joiner",
							EqFodder: ast.Fodder{},
							Body: &ast.Conditional{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(186),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(188),
											Column: int(74),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
											Comment: []string{},
										},
									},
									Ctx: p1882,
									FreeVars: ast.Identifiers{
										"std",
										"what",
									},
								},
								Cond: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(186),
												Column: int(10),
											},
											End: ast.Location{
												Line: int(186),
												Column: int(28),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p1882,
										FreeVars: ast.Identifiers{
											"std",
											"what",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(186),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(186),
													Column: int(22),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1882,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(186),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(186),
														Column: int(13),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "isString",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(186),
																Column: int(23),
															},
															End: ast.Location{
																Line: int(186),
																Column: int(27),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p1893,
														FreeVars: ast.Identifiers{
															"what",
														},
													},
													Id: "what",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
								ThenFodder: ast.Fodder{},
								BranchTrue: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(186),
												Column: int(34),
											},
											End: ast.Location{
												Line: int(186),
												Column: int(36),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p1882,
										FreeVars: nil,
									},
									Value: "",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								ElseFodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
										Comment: []string{},
									},
								},
								BranchFalse: &ast.Conditional{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(187),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(188),
												Column: int(74),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p1882,
										FreeVars: ast.Identifiers{
											"std",
											"what",
										},
									},
									Cond: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(187),
													Column: int(15),
												},
												End: ast.Location{
													Line: int(187),
													Column: int(32),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1882,
											FreeVars: ast.Identifiers{
												"std",
												"what",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(187),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(187),
														Column: int(26),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p1882,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(187),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(187),
															Column: int(18),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: ast.Fodder{},
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "isArray",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: ast.Fodder{},
											Id: nil,
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(187),
																	Column: int(27),
																},
																End: ast.Location{
																	Line: int(187),
																	Column: int(31),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p1908,
															FreeVars: ast.Identifiers{
																"what",
															},
														},
														Id: "what",
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
									ThenFodder: ast.Fodder{},
									BranchTrue: &ast.Array{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(187),
													Column: int(38),
												},
												End: ast.Location{
													Line: int(187),
													Column: int(40),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1882,
											FreeVars: nil,
										},
										Elements: nil,
										TrailingComma: false,
										CloseFodder: ast.Fodder{},
									},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
											Comment: []string{},
										},
									},
									BranchFalse: &ast.Error{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(188),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(188),
													Column: int(74),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1882,
											FreeVars: nil,
										},
										Expr: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(188),
														Column: int(18),
													},
													End: ast.Location{
														Line: int(188),
														Column: int(74),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p1882,
												FreeVars: nil,
											},
											Value: "std.repeat first argument must be an array or a string",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
									},
								},
							},
							Fun: nil,
							CloseFodder: ast.Fodder{},
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(185),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(188),
									Column: int(74),
								},
								File: p1,
							},
						},
					},
					Body: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(189),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(189),
									Column: int(61),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p1877,
							FreeVars: ast.Identifiers{
								"count",
								"joiner",
								"std",
								"what",
							},
						},
						Target: &ast.Index{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(189),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(189),
										Column: int(13),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p1877,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(189),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(189),
											Column: int(8),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
											Comment: []string{},
										},
									},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Id: "std",
							},
							LeftBracketFodder: ast.Fodder{},
							Index: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Value: "join",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							RightBracketFodder: ast.Fodder{},
							Id: nil,
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(189),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(189),
													Column: int(20),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1924,
											FreeVars: ast.Identifiers{
												"joiner",
											},
										},
										Id: "joiner",
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(189),
													Column: int(22),
												},
												End: ast.Location{
													Line: int(189),
													Column: int(60),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1924,
											FreeVars: ast.Identifiers{
												"count",
												"std",
												"what",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(189),
														Column: int(22),
													},
													End: ast.Location{
														Line: int(189),
														Column: int(35),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p1924,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(189),
															Column: int(22),
														},
														End: ast.Location{
															Line: int(189),
															Column: int(25),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: ast.Fodder{},
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "makeArray",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: ast.Fodder{},
											Id: nil,
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(189),
																	Column: int(36),
																},
																End: ast.Location{
																	Line: int(189),
																	Column: int(41),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p1935,
															FreeVars: ast.Identifiers{
																"count",
															},
														},
														Id: "count",
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Function{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(189),
																	Column: int(43),
																},
																End: ast.Location{
																	Line: int(189),
																	Column: int(59),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p1935,
															FreeVars: ast.Identifiers{
																"what",
															},
														},
														ParenLeftFodder: ast.Fodder{},
														Parameters: []ast.Parameter{
															ast.Parameter{
																NameFodder: ast.Fodder{},
																Name: "i",
																EqFodder: nil,
																DefaultArg: nil,
																CommaFodder: nil,
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(189),
																		Column: int(52),
																	},
																	End: ast.Location{
																		Line: int(189),
																		Column: int(53),
																	},
																	File: p1,
																},
															},
														},
														TrailingComma: false,
														ParenRightFodder: ast.Fodder{},
														Body: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(189),
																		Column: int(55),
																	},
																	End: ast.Location{
																		Line: int(189),
																		Column: int(59),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p1941,
																FreeVars: ast.Identifiers{
																	"what",
																},
															},
															Id: "what",
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(184),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(189),
					Column: int(61),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "slice",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "indexable",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(191),
								Column: int(9),
							},
							End: ast.Location{
								Line: int(191),
								Column: int(18),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "index",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(191),
								Column: int(20),
							},
							End: ast.Location{
								Line: int(191),
								Column: int(25),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "end",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(191),
								Column: int(27),
							},
							End: ast.Location{
								Line: int(191),
								Column: int(30),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "step",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(191),
								Column: int(32),
							},
							End: ast.Location{
								Line: int(191),
								Column: int(36),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(192),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(222),
								Column: int(66),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p1949,
						FreeVars: ast.Identifiers{
							"end",
							"index",
							"indexable",
							"std",
							"step",
						},
					},
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Variable: "invar",
							EqFodder: ast.Fodder{},
							Body: &ast.DesugaredObject{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(194),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(207),
											Column: int(8),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
											Comment: []string{},
										},
										ast.FodderElement{
											Kind: ast.FodderKind(2),
											Blanks: int(0),
											Indent: int(6),
											Comment: []string{
												"// loop invariant with defaults applied",
											},
										},
									},
									Ctx: p1955,
									FreeVars: ast.Identifiers{
										"end",
										"index",
										"indexable",
										"std",
										"step",
									},
								},
								Asserts: ast.Nodes{},
								Fields: ast.DesugaredObjectFields{
									ast.DesugaredObjectField{
										Hide: ast.ObjectFieldHide(1),
										Name: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "indexable",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										Body: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(195),
														Column: int(20),
													},
													End: ast.Location{
														Line: int(195),
														Column: int(29),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p1960,
												FreeVars: ast.Identifiers{
													"indexable",
												},
											},
											Id: "indexable",
										},
										PlusSuper: false,
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(195),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(195),
												Column: int(29),
											},
											File: p1,
										},
									},
									ast.DesugaredObjectField{
										Hide: ast.ObjectFieldHide(1),
										Name: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "index",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										Body: &ast.Conditional{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(197),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(198),
														Column: int(21),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(10),
														Comment: []string{},
													},
												},
												Ctx: p1960,
												FreeVars: ast.Identifiers{
													"index",
												},
											},
											Cond: &ast.Binary{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(197),
															Column: int(14),
														},
														End: ast.Location{
															Line: int(197),
															Column: int(27),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p1960,
													FreeVars: ast.Identifiers{
														"index",
													},
												},
												Left: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(197),
																Column: int(14),
															},
															End: ast.Location{
																Line: int(197),
																Column: int(19),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p1960,
														FreeVars: ast.Identifiers{
															"index",
														},
													},
													Id: "index",
												},
												OpFodder: ast.Fodder{},
												Op: ast.BinaryOp(12),
												Right: &ast.LiteralNull{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(197),
																Column: int(23),
															},
															End: ast.Location{
																Line: int(197),
																Column: int(27),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p1960,
														FreeVars: nil,
													},
												},
											},
											ThenFodder: ast.Fodder{},
											BranchTrue: &ast.LiteralNumber{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(197),
															Column: int(33),
														},
														End: ast.Location{
															Line: int(197),
															Column: int(34),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p1960,
													FreeVars: nil,
												},
												OriginalString: "0",
											},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(10),
													Comment: []string{},
												},
											},
											BranchFalse: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(198),
															Column: int(16),
														},
														End: ast.Location{
															Line: int(198),
															Column: int(21),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p1960,
													FreeVars: ast.Identifiers{
														"index",
													},
												},
												Id: "index",
											},
										},
										PlusSuper: false,
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(196),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(198),
												Column: int(21),
											},
											File: p1,
										},
									},
									ast.DesugaredObjectField{
										Hide: ast.ObjectFieldHide(1),
										Name: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "end",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										Body: &ast.Conditional{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(200),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(201),
														Column: int(19),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(10),
														Comment: []string{},
													},
												},
												Ctx: p1960,
												FreeVars: ast.Identifiers{
													"end",
													"indexable",
													"std",
												},
											},
											Cond: &ast.Binary{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(200),
															Column: int(14),
														},
														End: ast.Location{
															Line: int(200),
															Column: int(25),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p1960,
													FreeVars: ast.Identifiers{
														"end",
													},
												},
												Left: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(200),
																Column: int(14),
															},
															End: ast.Location{
																Line: int(200),
																Column: int(17),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p1960,
														FreeVars: ast.Identifiers{
															"end",
														},
													},
													Id: "end",
												},
												OpFodder: ast.Fodder{},
												Op: ast.BinaryOp(12),
												Right: &ast.LiteralNull{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(200),
																Column: int(21),
															},
															End: ast.Location{
																Line: int(200),
																Column: int(25),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p1960,
														FreeVars: nil,
													},
												},
											},
											ThenFodder: ast.Fodder{},
											BranchTrue: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(200),
															Column: int(31),
														},
														End: ast.Location{
															Line: int(200),
															Column: int(52),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p1960,
													FreeVars: ast.Identifiers{
														"indexable",
														"std",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(200),
																Column: int(31),
															},
															End: ast.Location{
																Line: int(200),
																Column: int(41),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p1960,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(200),
																	Column: int(31),
																},
																End: ast.Location{
																	Line: int(200),
																	Column: int(34),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: ast.Fodder{},
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "length",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: ast.Fodder{},
													Id: nil,
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(200),
																			Column: int(42),
																		},
																		End: ast.Location{
																			Line: int(200),
																			Column: int(51),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p1993,
																	FreeVars: ast.Identifiers{
																		"indexable",
																	},
																},
																Id: "indexable",
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(10),
													Comment: []string{},
												},
											},
											BranchFalse: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(201),
															Column: int(16),
														},
														End: ast.Location{
															Line: int(201),
															Column: int(19),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p1960,
													FreeVars: ast.Identifiers{
														"end",
													},
												},
												Id: "end",
											},
										},
										PlusSuper: false,
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(199),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(201),
												Column: int(19),
											},
											File: p1,
										},
									},
									ast.DesugaredObjectField{
										Hide: ast.ObjectFieldHide(1),
										Name: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "step",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										Body: &ast.Conditional{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(203),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(204),
														Column: int(20),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(10),
														Comment: []string{},
													},
												},
												Ctx: p1960,
												FreeVars: ast.Identifiers{
													"step",
												},
											},
											Cond: &ast.Binary{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(203),
															Column: int(14),
														},
														End: ast.Location{
															Line: int(203),
															Column: int(26),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p1960,
													FreeVars: ast.Identifiers{
														"step",
													},
												},
												Left: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(203),
																Column: int(14),
															},
															End: ast.Location{
																Line: int(203),
																Column: int(18),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p1960,
														FreeVars: ast.Identifiers{
															"step",
														},
													},
													Id: "step",
												},
												OpFodder: ast.Fodder{},
												Op: ast.BinaryOp(12),
												Right: &ast.LiteralNull{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(203),
																Column: int(22),
															},
															End: ast.Location{
																Line: int(203),
																Column: int(26),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p1960,
														FreeVars: nil,
													},
												},
											},
											ThenFodder: ast.Fodder{},
											BranchTrue: &ast.LiteralNumber{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(203),
															Column: int(32),
														},
														End: ast.Location{
															Line: int(203),
															Column: int(33),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p1960,
													FreeVars: nil,
												},
												OriginalString: "1",
											},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(10),
													Comment: []string{},
												},
											},
											BranchFalse: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(204),
															Column: int(16),
														},
														End: ast.Location{
															Line: int(204),
															Column: int(20),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p1960,
													FreeVars: ast.Identifiers{
														"step",
													},
												},
												Id: "step",
											},
										},
										PlusSuper: false,
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(202),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(204),
												Column: int(20),
											},
											File: p1,
										},
									},
									ast.DesugaredObjectField{
										Hide: ast.ObjectFieldHide(1),
										Name: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "length",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										Body: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(205),
														Column: int(17),
													},
													End: ast.Location{
														Line: int(205),
														Column: int(38),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p1960,
												FreeVars: ast.Identifiers{
													"indexable",
													"std",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(205),
															Column: int(17),
														},
														End: ast.Location{
															Line: int(205),
															Column: int(27),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p1960,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(205),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(205),
																Column: int(20),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "length",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(205),
																		Column: int(28),
																	},
																	End: ast.Location{
																		Line: int(205),
																		Column: int(37),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p2021,
																FreeVars: ast.Identifiers{
																	"indexable",
																},
															},
															Id: "indexable",
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
										PlusSuper: false,
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(205),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(205),
												Column: int(38),
											},
											File: p1,
										},
									},
									ast.DesugaredObjectField{
										Hide: ast.ObjectFieldHide(1),
										Name: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "type",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										Body: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(206),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(206),
														Column: int(34),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p1960,
												FreeVars: ast.Identifiers{
													"indexable",
													"std",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(206),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(206),
															Column: int(23),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p1960,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(206),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(206),
																Column: int(18),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "type",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(206),
																		Column: int(24),
																	},
																	End: ast.Location{
																		Line: int(206),
																		Column: int(33),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p2033,
																FreeVars: ast.Identifiers{
																	"indexable",
																},
															},
															Id: "indexable",
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
										PlusSuper: false,
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(206),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(206),
												Column: int(34),
											},
											File: p1,
										},
									},
								},
								Locals: ast.LocalBinds{},
							},
							Fun: nil,
							CloseFodder: ast.Fodder{},
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(192),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(207),
									Column: int(8),
								},
								File: p1,
							},
						},
					},
					Body: &ast.Conditional{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"indexable",
								"invar",
								"std",
								"step",
							},
						},
						Cond: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(208),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(208),
										Column: int(65),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p1949,
								FreeVars: ast.Identifiers{
									"invar",
								},
							},
							Left: &ast.Binary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(208),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(208),
											Column: int(46),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p1949,
									FreeVars: ast.Identifiers{
										"invar",
									},
								},
								Left: &ast.Binary{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(208),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(208),
												Column: int(28),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p1949,
										FreeVars: ast.Identifiers{
											"invar",
										},
									},
									Left: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(208),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(208),
													Column: int(23),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1949,
											FreeVars: ast.Identifiers{
												"invar",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(208),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(208),
														Column: int(17),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"invar",
												},
											},
											Id: "invar",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "index",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									OpFodder: ast.Fodder{},
									Op: ast.BinaryOp(8),
									Right: &ast.LiteralNumber{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(208),
													Column: int(27),
												},
												End: ast.Location{
													Line: int(208),
													Column: int(28),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1949,
											FreeVars: nil,
										},
										OriginalString: "0",
									},
								},
								OpFodder: ast.Fodder{},
								Op: ast.BinaryOp(17),
								Right: &ast.Binary{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(208),
												Column: int(32),
											},
											End: ast.Location{
												Line: int(208),
												Column: int(46),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p1949,
										FreeVars: ast.Identifiers{
											"invar",
										},
									},
									Left: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(208),
													Column: int(32),
												},
												End: ast.Location{
													Line: int(208),
													Column: int(41),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1949,
											FreeVars: ast.Identifiers{
												"invar",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(208),
														Column: int(32),
													},
													End: ast.Location{
														Line: int(208),
														Column: int(37),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"invar",
												},
											},
											Id: "invar",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "end",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									OpFodder: ast.Fodder{},
									Op: ast.BinaryOp(8),
									Right: &ast.LiteralNumber{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(208),
													Column: int(45),
												},
												End: ast.Location{
													Line: int(208),
													Column: int(46),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1949,
											FreeVars: nil,
										},
										OriginalString: "0",
									},
								},
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(17),
							Right: &ast.Binary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(208),
											Column: int(50),
										},
										End: ast.Location{
											Line: int(208),
											Column: int(65),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p1949,
									FreeVars: ast.Identifiers{
										"invar",
									},
								},
								Left: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(208),
												Column: int(50),
											},
											End: ast.Location{
												Line: int(208),
												Column: int(60),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p1949,
										FreeVars: ast.Identifiers{
											"invar",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(208),
													Column: int(50),
												},
												End: ast.Location{
													Line: int(208),
													Column: int(55),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"invar",
											},
										},
										Id: "invar",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "step",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								OpFodder: ast.Fodder{},
								Op: ast.BinaryOp(8),
								Right: &ast.LiteralNumber{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(208),
												Column: int(64),
											},
											End: ast.Location{
												Line: int(208),
												Column: int(65),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p1949,
										FreeVars: nil,
									},
									OriginalString: "0",
								},
							},
						},
						ThenFodder: nil,
						BranchTrue: &ast.Conditional{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"indexable",
									"invar",
									"std",
									"step",
								},
							},
							Cond: &ast.Binary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(209),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(209),
											Column: int(21),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p1949,
									FreeVars: ast.Identifiers{
										"step",
									},
								},
								Left: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(209),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(209),
												Column: int(16),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p1949,
										FreeVars: ast.Identifiers{
											"step",
										},
									},
									Id: "step",
								},
								OpFodder: ast.Fodder{},
								Op: ast.BinaryOp(13),
								Right: &ast.LiteralNumber{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(209),
												Column: int(20),
											},
											End: ast.Location{
												Line: int(209),
												Column: int(21),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p1949,
										FreeVars: nil,
									},
									OriginalString: "0",
								},
							},
							ThenFodder: nil,
							BranchTrue: &ast.Conditional{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"indexable",
										"invar",
										"std",
									},
								},
								Cond: &ast.Binary{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(210),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(210),
												Column: int(61),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p1949,
										FreeVars: ast.Identifiers{
											"indexable",
											"std",
										},
									},
									Left: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(210),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(210),
													Column: int(35),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1949,
											FreeVars: ast.Identifiers{
												"indexable",
												"std",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(210),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(210),
														Column: int(24),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p1949,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(210),
															Column: int(12),
														},
														End: ast.Location{
															Line: int(210),
															Column: int(15),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: ast.Fodder{},
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "isString",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: ast.Fodder{},
											Id: nil,
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(210),
																	Column: int(25),
																},
																End: ast.Location{
																	Line: int(210),
																	Column: int(34),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p2085,
															FreeVars: ast.Identifiers{
																"indexable",
															},
														},
														Id: "indexable",
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
									OpFodder: ast.Fodder{},
									Op: ast.BinaryOp(18),
									Right: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(210),
													Column: int(39),
												},
												End: ast.Location{
													Line: int(210),
													Column: int(61),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1949,
											FreeVars: ast.Identifiers{
												"indexable",
												"std",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(210),
														Column: int(39),
													},
													End: ast.Location{
														Line: int(210),
														Column: int(50),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p1949,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(210),
															Column: int(39),
														},
														End: ast.Location{
															Line: int(210),
															Column: int(42),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: ast.Fodder{},
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "isArray",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: ast.Fodder{},
											Id: nil,
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(210),
																	Column: int(51),
																},
																End: ast.Location{
																	Line: int(210),
																	Column: int(60),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p2096,
															FreeVars: ast.Identifiers{
																"indexable",
															},
														},
														Id: "indexable",
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
								},
								ThenFodder: nil,
								BranchTrue: &ast.Local{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(211),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(222),
												Column: int(66),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(4),
												Comment: []string{},
											},
										},
										Ctx: p1949,
										FreeVars: ast.Identifiers{
											"invar",
										},
									},
									Binds: ast.LocalBinds{
										ast.LocalBind{
											VarFodder: nil,
											Variable: "build",
											EqFodder: nil,
											Body: &ast.Function{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(211),
															Column: int(11),
														},
														End: ast.Location{
															Line: int(221),
															Column: int(10),
														},
														File: p1,
													},
													Fodder: nil,
													Ctx: p2103,
													FreeVars: ast.Identifiers{
														"build",
														"invar",
													},
												},
												ParenLeftFodder: ast.Fodder{},
												Parameters: []ast.Parameter{
													ast.Parameter{
														NameFodder: ast.Fodder{},
														Name: "slice",
														EqFodder: nil,
														DefaultArg: nil,
														CommaFodder: ast.Fodder{},
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(211),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(211),
																Column: int(22),
															},
															File: p1,
														},
													},
													ast.Parameter{
														NameFodder: ast.Fodder{},
														Name: "cur",
														EqFodder: nil,
														DefaultArg: nil,
														CommaFodder: nil,
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(211),
																Column: int(24),
															},
															End: ast.Location{
																Line: int(211),
																Column: int(27),
															},
															File: p1,
														},
													},
												},
												TrailingComma: false,
												ParenRightFodder: ast.Fodder{},
												Body: &ast.Conditional{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(212),
																Column: int(7),
															},
															End: ast.Location{
																Line: int(221),
																Column: int(10),
															},
															File: p1,
														},
														Fodder: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(6),
																Comment: []string{},
															},
														},
														Ctx: p2108,
														FreeVars: ast.Identifiers{
															"build",
															"cur",
															"invar",
															"slice",
														},
													},
													Cond: &ast.Binary{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(212),
																	Column: int(10),
																},
																End: ast.Location{
																	Line: int(212),
																	Column: int(49),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p2108,
															FreeVars: ast.Identifiers{
																"cur",
																"invar",
															},
														},
														Left: &ast.Binary{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(212),
																		Column: int(10),
																	},
																	End: ast.Location{
																		Line: int(212),
																		Column: int(26),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p2108,
																FreeVars: ast.Identifiers{
																	"cur",
																	"invar",
																},
															},
															Left: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(212),
																			Column: int(10),
																		},
																		End: ast.Location{
																			Line: int(212),
																			Column: int(13),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p2108,
																	FreeVars: ast.Identifiers{
																		"cur",
																	},
																},
																Id: "cur",
															},
															OpFodder: ast.Fodder{},
															Op: ast.BinaryOp(8),
															Right: &ast.Index{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(212),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(212),
																			Column: int(26),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p2108,
																	FreeVars: ast.Identifiers{
																		"invar",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(212),
																				Column: int(17),
																			},
																			End: ast.Location{
																				Line: int(212),
																				Column: int(22),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"invar",
																		},
																	},
																	Id: "invar",
																},
																LeftBracketFodder: ast.Fodder{},
																Index: &ast.LiteralString{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: nil,
																	},
																	Value: "end",
																	Kind: ast.LiteralStringKind(1),
																	BlockIndent: "",
																	BlockTermIndent: "",
																},
																RightBracketFodder: ast.Fodder{},
																Id: nil,
															},
														},
														OpFodder: ast.Fodder{},
														Op: ast.BinaryOp(18),
														Right: &ast.Binary{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(212),
																		Column: int(30),
																	},
																	End: ast.Location{
																		Line: int(212),
																		Column: int(49),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p2108,
																FreeVars: ast.Identifiers{
																	"cur",
																	"invar",
																},
															},
															Left: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(212),
																			Column: int(30),
																		},
																		End: ast.Location{
																			Line: int(212),
																			Column: int(33),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p2108,
																	FreeVars: ast.Identifiers{
																		"cur",
																	},
																},
																Id: "cur",
															},
															OpFodder: ast.Fodder{},
															Op: ast.BinaryOp(8),
															Right: &ast.Index{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(212),
																			Column: int(37),
																		},
																		End: ast.Location{
																			Line: int(212),
																			Column: int(49),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p2108,
																	FreeVars: ast.Identifiers{
																		"invar",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(212),
																				Column: int(37),
																			},
																			End: ast.Location{
																				Line: int(212),
																				Column: int(42),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"invar",
																		},
																	},
																	Id: "invar",
																},
																LeftBracketFodder: ast.Fodder{},
																Index: &ast.LiteralString{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: nil,
																	},
																	Value: "length",
																	Kind: ast.LiteralStringKind(1),
																	BlockIndent: "",
																	BlockTermIndent: "",
																},
																RightBracketFodder: ast.Fodder{},
																Id: nil,
															},
														},
													},
													ThenFodder: ast.Fodder{},
													BranchTrue: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(213),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(213),
																	Column: int(14),
																},
																File: p1,
															},
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																	Comment: []string{},
																},
															},
															Ctx: p2108,
															FreeVars: ast.Identifiers{
																"slice",
															},
														},
														Id: "slice",
													},
													ElseFodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(6),
															Comment: []string{},
														},
													},
													BranchFalse: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(215),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(221),
																	Column: int(10),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p2108,
															FreeVars: ast.Identifiers{
																"build",
																"cur",
																"invar",
																"slice",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(215),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(215),
																		Column: int(14),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																		Comment: []string{},
																	},
																},
																Ctx: p2108,
																FreeVars: ast.Identifiers{
																	"build",
																},
															},
															Id: "build",
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Conditional{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(216),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(219),
																					Column: int(43),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(10),
																					Comment: []string{},
																				},
																			},
																			Ctx: p2142,
																			FreeVars: ast.Identifiers{
																				"cur",
																				"invar",
																				"slice",
																			},
																		},
																		Cond: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(216),
																						Column: int(14),
																					},
																					End: ast.Location{
																						Line: int(216),
																						Column: int(36),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p2142,
																				FreeVars: ast.Identifiers{
																					"invar",
																				},
																			},
																			Left: &ast.Index{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(216),
																							Column: int(14),
																						},
																						End: ast.Location{
																							Line: int(216),
																							Column: int(24),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p2142,
																					FreeVars: ast.Identifiers{
																						"invar",
																					},
																				},
																				Target: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(216),
																								Column: int(14),
																							},
																							End: ast.Location{
																								Line: int(216),
																								Column: int(19),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: nil,
																						FreeVars: ast.Identifiers{
																							"invar",
																						},
																					},
																					Id: "invar",
																				},
																				LeftBracketFodder: ast.Fodder{},
																				Index: &ast.LiteralString{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							File: nil,
																						},
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: nil,
																					},
																					Value: "type",
																					Kind: ast.LiteralStringKind(1),
																					BlockIndent: "",
																					BlockTermIndent: "",
																				},
																				RightBracketFodder: ast.Fodder{},
																				Id: nil,
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(12),
																			Right: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(216),
																							Column: int(28),
																						},
																						End: ast.Location{
																							Line: int(216),
																							Column: int(36),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p2142,
																					FreeVars: nil,
																				},
																				Value: "string",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																		},
																		ThenFodder: ast.Fodder{},
																		BranchTrue: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(217),
																						Column: int(13),
																					},
																					End: ast.Location{
																						Line: int(217),
																						Column: int(41),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p2142,
																				FreeVars: ast.Identifiers{
																					"cur",
																					"invar",
																					"slice",
																				},
																			},
																			Left: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(217),
																							Column: int(13),
																						},
																						End: ast.Location{
																							Line: int(217),
																							Column: int(18),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(12),
																							Comment: []string{},
																						},
																					},
																					Ctx: p2142,
																					FreeVars: ast.Identifiers{
																						"slice",
																					},
																				},
																				Id: "slice",
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(3),
																			Right: &ast.Index{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(217),
																							Column: int(21),
																						},
																						End: ast.Location{
																							Line: int(217),
																							Column: int(41),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p2142,
																					FreeVars: ast.Identifiers{
																						"cur",
																						"invar",
																					},
																				},
																				Target: &ast.Index{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(217),
																								Column: int(21),
																							},
																							End: ast.Location{
																								Line: int(217),
																								Column: int(36),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p2142,
																						FreeVars: ast.Identifiers{
																							"invar",
																						},
																					},
																					Target: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(217),
																									Column: int(21),
																								},
																								End: ast.Location{
																									Line: int(217),
																									Column: int(26),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: nil,
																							FreeVars: ast.Identifiers{
																								"invar",
																							},
																						},
																						Id: "invar",
																					},
																					LeftBracketFodder: ast.Fodder{},
																					Index: &ast.LiteralString{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								File: nil,
																							},
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: nil,
																						},
																						Value: "indexable",
																						Kind: ast.LiteralStringKind(1),
																						BlockIndent: "",
																						BlockTermIndent: "",
																					},
																					RightBracketFodder: ast.Fodder{},
																					Id: nil,
																				},
																				LeftBracketFodder: ast.Fodder{},
																				Index: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(217),
																								Column: int(37),
																							},
																							End: ast.Location{
																								Line: int(217),
																								Column: int(40),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p2142,
																						FreeVars: ast.Identifiers{
																							"cur",
																						},
																					},
																					Id: "cur",
																				},
																				RightBracketFodder: ast.Fodder{},
																				Id: nil,
																			},
																		},
																		ElseFodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(10),
																				Comment: []string{},
																			},
																		},
																		BranchFalse: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(219),
																						Column: int(13),
																					},
																					End: ast.Location{
																						Line: int(219),
																						Column: int(43),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p2142,
																				FreeVars: ast.Identifiers{
																					"cur",
																					"invar",
																					"slice",
																				},
																			},
																			Left: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(219),
																							Column: int(13),
																						},
																						End: ast.Location{
																							Line: int(219),
																							Column: int(18),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(12),
																							Comment: []string{},
																						},
																					},
																					Ctx: p2142,
																					FreeVars: ast.Identifiers{
																						"slice",
																					},
																				},
																				Id: "slice",
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(3),
																			Right: &ast.Array{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(219),
																							Column: int(21),
																						},
																						End: ast.Location{
																							Line: int(219),
																							Column: int(43),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p2142,
																					FreeVars: ast.Identifiers{
																						"cur",
																						"invar",
																					},
																				},
																				Elements: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Index{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(219),
																										Column: int(22),
																									},
																									End: ast.Location{
																										Line: int(219),
																										Column: int(42),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p2176,
																								FreeVars: ast.Identifiers{
																									"cur",
																									"invar",
																								},
																							},
																							Target: &ast.Index{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(219),
																											Column: int(22),
																										},
																										End: ast.Location{
																											Line: int(219),
																											Column: int(37),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p2176,
																									FreeVars: ast.Identifiers{
																										"invar",
																									},
																								},
																								Target: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(219),
																												Column: int(22),
																											},
																											End: ast.Location{
																												Line: int(219),
																												Column: int(27),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: nil,
																										FreeVars: ast.Identifiers{
																											"invar",
																										},
																									},
																									Id: "invar",
																								},
																								LeftBracketFodder: ast.Fodder{},
																								Index: &ast.LiteralString{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											File: nil,
																										},
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: nil,
																									},
																									Value: "indexable",
																									Kind: ast.LiteralStringKind(1),
																									BlockIndent: "",
																									BlockTermIndent: "",
																								},
																								RightBracketFodder: ast.Fodder{},
																								Id: nil,
																							},
																							LeftBracketFodder: ast.Fodder{},
																							Index: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(219),
																											Column: int(38),
																										},
																										End: ast.Location{
																											Line: int(219),
																											Column: int(41),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p2176,
																									FreeVars: ast.Identifiers{
																										"cur",
																									},
																								},
																								Id: "cur",
																							},
																							RightBracketFodder: ast.Fodder{},
																							Id: nil,
																						},
																						CommaFodder: nil,
																					},
																				},
																				TrailingComma: false,
																				CloseFodder: ast.Fodder{},
																			},
																		},
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(220),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(220),
																					Column: int(27),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p2142,
																			FreeVars: ast.Identifiers{
																				"cur",
																				"invar",
																			},
																		},
																		Left: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(220),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(220),
																						Column: int(14),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(10),
																						Comment: []string{},
																					},
																				},
																				Ctx: p2142,
																				FreeVars: ast.Identifiers{
																					"cur",
																				},
																			},
																			Id: "cur",
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(3),
																		Right: &ast.Index{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(220),
																						Column: int(17),
																					},
																					End: ast.Location{
																						Line: int(220),
																						Column: int(27),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p2142,
																				FreeVars: ast.Identifiers{
																					"invar",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(220),
																							Column: int(17),
																						},
																						End: ast.Location{
																							Line: int(220),
																							Column: int(22),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"invar",
																					},
																				},
																				Id: "invar",
																			},
																			LeftBracketFodder: ast.Fodder{},
																			Index: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: nil,
																				},
																				Value: "step",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																			RightBracketFodder: ast.Fodder{},
																			Id: nil,
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														TrailingComma: false,
														TailStrict: true,
														FodderRight: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(8),
																Comment: []string{},
															},
														},
														TailStrictFodder: ast.Fodder{},
													},
												},
											},
											Fun: nil,
											CloseFodder: nil,
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
										},
									},
									Body: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(222),
													Column: int(5),
												},
												End: ast.Location{
													Line: int(222),
													Column: int(66),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p1949,
											FreeVars: ast.Identifiers{
												"build",
												"invar",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(222),
														Column: int(5),
													},
													End: ast.Location{
														Line: int(222),
														Column: int(10),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(4),
														Comment: []string{},
													},
												},
												Ctx: p1949,
												FreeVars: ast.Identifiers{
													"build",
												},
											},
											Id: "build",
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Conditional{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(222),
																	Column: int(11),
																},
																End: ast.Location{
																	Line: int(222),
																	Column: int(52),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p2203,
															FreeVars: ast.Identifiers{
																"invar",
															},
														},
														Cond: &ast.Binary{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(222),
																		Column: int(14),
																	},
																	End: ast.Location{
																		Line: int(222),
																		Column: int(36),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p2203,
																FreeVars: ast.Identifiers{
																	"invar",
																},
															},
															Left: &ast.Index{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(222),
																			Column: int(14),
																		},
																		End: ast.Location{
																			Line: int(222),
																			Column: int(24),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p2203,
																	FreeVars: ast.Identifiers{
																		"invar",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(222),
																				Column: int(14),
																			},
																			End: ast.Location{
																				Line: int(222),
																				Column: int(19),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"invar",
																		},
																	},
																	Id: "invar",
																},
																LeftBracketFodder: ast.Fodder{},
																Index: &ast.LiteralString{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: nil,
																	},
																	Value: "type",
																	Kind: ast.LiteralStringKind(1),
																	BlockIndent: "",
																	BlockTermIndent: "",
																},
																RightBracketFodder: ast.Fodder{},
																Id: nil,
															},
															OpFodder: ast.Fodder{},
															Op: ast.BinaryOp(12),
															Right: &ast.LiteralString{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(222),
																			Column: int(28),
																		},
																		End: ast.Location{
																			Line: int(222),
																			Column: int(36),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p2203,
																	FreeVars: nil,
																},
																Value: "string",
																Kind: ast.LiteralStringKind(1),
																BlockIndent: "",
																BlockTermIndent: "",
															},
														},
														ThenFodder: ast.Fodder{},
														BranchTrue: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(222),
																		Column: int(42),
																	},
																	End: ast.Location{
																		Line: int(222),
																		Column: int(44),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p2203,
																FreeVars: nil,
															},
															Value: "",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														ElseFodder: ast.Fodder{},
														BranchFalse: &ast.Array{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(222),
																		Column: int(50),
																	},
																	End: ast.Location{
																		Line: int(222),
																		Column: int(52),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p2203,
																FreeVars: nil,
															},
															Elements: nil,
															TrailingComma: false,
															CloseFodder: ast.Fodder{},
														},
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Index{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(222),
																	Column: int(54),
																},
																End: ast.Location{
																	Line: int(222),
																	Column: int(65),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p2203,
															FreeVars: ast.Identifiers{
																"invar",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(222),
																		Column: int(54),
																	},
																	End: ast.Location{
																		Line: int(222),
																		Column: int(59),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"invar",
																},
															},
															Id: "invar",
														},
														LeftBracketFodder: ast.Fodder{},
														Index: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
															Value: "index",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														RightBracketFodder: ast.Fodder{},
														Id: nil,
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
								},
								ElseFodder: nil,
								BranchFalse: &ast.Error{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"indexable",
											"std",
										},
									},
									Expr: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"indexable",
												"std",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: nil,
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "mod",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: nil,
											Id: nil,
										},
										FodderLeft: nil,
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(210),
																	Column: int(64),
																},
																End: ast.Location{
																	Line: int(210),
																	Column: int(117),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p1949,
															FreeVars: nil,
														},
														Value: "std.slice accepts a string or an array, but got: %s",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													CommaFodder: nil,
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(210),
																	Column: int(120),
																},
																End: ast.Location{
																	Line: int(210),
																	Column: int(139),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p1949,
															FreeVars: ast.Identifiers{
																"indexable",
																"std",
															},
														},
														Target: &ast.Index{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(210),
																		Column: int(120),
																	},
																	End: ast.Location{
																		Line: int(210),
																		Column: int(128),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p1949,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Target: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(210),
																			Column: int(120),
																		},
																		End: ast.Location{
																			Line: int(210),
																			Column: int(123),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Id: "std",
															},
															LeftBracketFodder: ast.Fodder{},
															Index: &ast.LiteralString{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: nil,
																},
																Value: "type",
																Kind: ast.LiteralStringKind(1),
																BlockIndent: "",
																BlockTermIndent: "",
															},
															RightBracketFodder: ast.Fodder{},
															Id: nil,
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(210),
																					Column: int(129),
																				},
																				End: ast.Location{
																					Line: int(210),
																					Column: int(138),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p2240,
																			FreeVars: ast.Identifiers{
																				"indexable",
																			},
																		},
																		Id: "indexable",
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														TrailingComma: false,
														TailStrict: false,
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: nil,
										TailStrictFodder: nil,
									},
								},
							},
							ElseFodder: nil,
							BranchFalse: &ast.Error{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
										"step",
									},
								},
								Expr: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
											"step",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: nil,
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "mod",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: nil,
										Id: nil,
									},
									FodderLeft: nil,
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(209),
																Column: int(24),
															},
															End: ast.Location{
																Line: int(209),
																Column: int(64),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p1949,
														FreeVars: nil,
													},
													Value: "got %s but step must be greater than 0",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												CommaFodder: nil,
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(209),
																Column: int(67),
															},
															End: ast.Location{
																Line: int(209),
																Column: int(71),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p1949,
														FreeVars: ast.Identifiers{
															"step",
														},
													},
													Id: "step",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: nil,
									TailStrictFodder: nil,
								},
							},
						},
						ElseFodder: nil,
						BranchFalse: &ast.Error{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"invar",
									"std",
								},
							},
							Expr: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"invar",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: nil,
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "mod",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: nil,
									Id: nil,
								},
								FodderLeft: nil,
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(208),
															Column: int(68),
														},
														End: ast.Location{
															Line: int(208),
															Column: int(137),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p1949,
													FreeVars: nil,
												},
												Value: "got [%s:%s:%s] but negative index, end, and steps are not supported",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											CommaFodder: nil,
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.Array{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(208),
															Column: int(140),
														},
														End: ast.Location{
															Line: int(208),
															Column: int(176),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p1949,
													FreeVars: ast.Identifiers{
														"invar",
													},
												},
												Elements: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Index{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(208),
																		Column: int(141),
																	},
																	End: ast.Location{
																		Line: int(208),
																		Column: int(152),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p2270,
																FreeVars: ast.Identifiers{
																	"invar",
																},
															},
															Target: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(208),
																			Column: int(141),
																		},
																		End: ast.Location{
																			Line: int(208),
																			Column: int(146),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"invar",
																	},
																},
																Id: "invar",
															},
															LeftBracketFodder: ast.Fodder{},
															Index: &ast.LiteralString{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: nil,
																},
																Value: "index",
																Kind: ast.LiteralStringKind(1),
																BlockIndent: "",
																BlockTermIndent: "",
															},
															RightBracketFodder: ast.Fodder{},
															Id: nil,
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Index{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(208),
																		Column: int(154),
																	},
																	End: ast.Location{
																		Line: int(208),
																		Column: int(163),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p2270,
																FreeVars: ast.Identifiers{
																	"invar",
																},
															},
															Target: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(208),
																			Column: int(154),
																		},
																		End: ast.Location{
																			Line: int(208),
																			Column: int(159),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"invar",
																	},
																},
																Id: "invar",
															},
															LeftBracketFodder: ast.Fodder{},
															Index: &ast.LiteralString{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: nil,
																},
																Value: "end",
																Kind: ast.LiteralStringKind(1),
																BlockIndent: "",
																BlockTermIndent: "",
															},
															RightBracketFodder: ast.Fodder{},
															Id: nil,
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Index{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(208),
																		Column: int(165),
																	},
																	End: ast.Location{
																		Line: int(208),
																		Column: int(175),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p2270,
																FreeVars: ast.Identifiers{
																	"invar",
																},
															},
															Target: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(208),
																			Column: int(165),
																		},
																		End: ast.Location{
																			Line: int(208),
																			Column: int(170),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"invar",
																	},
																},
																Id: "invar",
															},
															LeftBracketFodder: ast.Fodder{},
															Index: &ast.LiteralString{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: nil,
																},
																Value: "step",
																Kind: ast.LiteralStringKind(1),
																BlockIndent: "",
																BlockTermIndent: "",
															},
															RightBracketFodder: ast.Fodder{},
															Id: nil,
														},
														CommaFodder: nil,
													},
												},
												TrailingComma: false,
												CloseFodder: ast.Fodder{},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: nil,
								TailStrictFodder: nil,
							},
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(191),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(222),
					Column: int(66),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "member",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(224),
								Column: int(10),
							},
							End: ast.Location{
								Line: int(224),
								Column: int(13),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "x",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(224),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(224),
								Column: int(16),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(225),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(229),
								Column: int(72),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p2291,
						FreeVars: ast.Identifiers{
							"arr",
							"std",
							"x",
						},
					},
					Cond: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(225),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(225),
									Column: int(24),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p2291,
							FreeVars: ast.Identifiers{
								"arr",
								"std",
							},
						},
						Target: &ast.Index{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(225),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(225),
										Column: int(19),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p2291,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(225),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(225),
											Column: int(11),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Id: "std",
							},
							LeftBracketFodder: ast.Fodder{},
							Index: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Value: "isArray",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							RightBracketFodder: ast.Fodder{},
							Id: nil,
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(225),
													Column: int(20),
												},
												End: ast.Location{
													Line: int(225),
													Column: int(23),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p2302,
											FreeVars: ast.Identifiers{
												"arr",
											},
										},
										Id: "arr",
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
					ThenFodder: ast.Fodder{},
					BranchTrue: &ast.Binary{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(226),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(226),
									Column: int(28),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p2291,
							FreeVars: ast.Identifiers{
								"arr",
								"std",
								"x",
							},
						},
						Left: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(226),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(226),
										Column: int(24),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p2291,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
									"x",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(226),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(226),
											Column: int(16),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2291,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(226),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(226),
												Column: int(10),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "count",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(226),
														Column: int(17),
													},
													End: ast.Location{
														Line: int(226),
														Column: int(20),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p2316,
												FreeVars: ast.Identifiers{
													"arr",
												},
											},
											Id: "arr",
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(226),
														Column: int(22),
													},
													End: ast.Location{
														Line: int(226),
														Column: int(23),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p2316,
												FreeVars: ast.Identifiers{
													"x",
												},
											},
											Id: "x",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
						OpFodder: ast.Fodder{},
						Op: ast.BinaryOp(7),
						Right: &ast.LiteralNumber{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(226),
										Column: int(27),
									},
									End: ast.Location{
										Line: int(226),
										Column: int(28),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p2291,
								FreeVars: nil,
							},
							OriginalString: "0",
						},
					},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
							Comment: []string{},
						},
					},
					BranchFalse: &ast.Conditional{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(227),
									Column: int(10),
								},
								End: ast.Location{
									Line: int(229),
									Column: int(72),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p2291,
							FreeVars: ast.Identifiers{
								"arr",
								"std",
								"x",
							},
						},
						Cond: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(227),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(227),
										Column: int(30),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p2291,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(227),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(227),
											Column: int(25),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2291,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(227),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(227),
												Column: int(16),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "isString",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(227),
														Column: int(26),
													},
													End: ast.Location{
														Line: int(227),
														Column: int(29),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p2333,
												FreeVars: ast.Identifiers{
													"arr",
												},
											},
											Id: "arr",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
						ThenFodder: ast.Fodder{},
						BranchTrue: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(228),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(228),
										Column: int(45),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p2291,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
									"x",
								},
							},
							Left: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(228),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(228),
											Column: int(41),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2291,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
										"x",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(228),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(228),
												Column: int(17),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p2291,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(228),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(228),
													Column: int(10),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
													Comment: []string{},
												},
											},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "length",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(228),
															Column: int(18),
														},
														End: ast.Location{
															Line: int(228),
															Column: int(40),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p2347,
													FreeVars: ast.Identifiers{
														"arr",
														"std",
														"x",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(228),
																Column: int(18),
															},
															End: ast.Location{
																Line: int(228),
																Column: int(32),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p2347,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(228),
																	Column: int(18),
																},
																End: ast.Location{
																	Line: int(228),
																	Column: int(21),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: ast.Fodder{},
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "findSubstr",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: ast.Fodder{},
													Id: nil,
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(228),
																			Column: int(33),
																		},
																		End: ast.Location{
																			Line: int(228),
																			Column: int(34),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p2356,
																	FreeVars: ast.Identifiers{
																		"x",
																	},
																},
																Id: "x",
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(228),
																			Column: int(36),
																		},
																		End: ast.Location{
																			Line: int(228),
																			Column: int(39),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p2356,
																	FreeVars: ast.Identifiers{
																		"arr",
																	},
																},
																Id: "arr",
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(7),
							Right: &ast.LiteralNumber{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(228),
											Column: int(44),
										},
										End: ast.Location{
											Line: int(228),
											Column: int(45),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2291,
									FreeVars: nil,
								},
								OriginalString: "0",
							},
						},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						BranchFalse: &ast.Error{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(229),
										Column: int(10),
									},
									End: ast.Location{
										Line: int(229),
										Column: int(72),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p2291,
								FreeVars: nil,
							},
							Expr: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(229),
											Column: int(16),
										},
										End: ast.Location{
											Line: int(229),
											Column: int(72),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2291,
									FreeVars: nil,
								},
								Value: "std.member first argument must be an array or a string",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(224),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(229),
					Column: int(72),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "count",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(231),
								Column: int(9),
							},
							End: ast.Location{
								Line: int(231),
								Column: int(12),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "x",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(231),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(231),
								Column: int(15),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(231),
								Column: int(19),
							},
							End: ast.Location{
								Line: int(231),
								Column: int(66),
							},
							File: p1,
						},
						Fodder: ast.Fodder{},
						Ctx: p2369,
						FreeVars: ast.Identifiers{
							"arr",
							"std",
							"x",
						},
					},
					Target: &ast.Index{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(231),
									Column: int(19),
								},
								End: ast.Location{
									Line: int(231),
									Column: int(29),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p2369,
							FreeVars: ast.Identifiers{
								"std",
							},
						},
						Target: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(231),
										Column: int(19),
									},
									End: ast.Location{
										Line: int(231),
										Column: int(22),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Id: "std",
						},
						LeftBracketFodder: ast.Fodder{},
						Index: &ast.LiteralString{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: nil,
							},
							Value: "length",
							Kind: ast.LiteralStringKind(1),
							BlockIndent: "",
							BlockTermIndent: "",
						},
						RightBracketFodder: ast.Fodder{},
						Id: nil,
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(231),
												Column: int(30),
											},
											End: ast.Location{
												Line: int(231),
												Column: int(65),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p2378,
										FreeVars: ast.Identifiers{
											"arr",
											"std",
											"x",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(231),
													Column: int(30),
												},
												End: ast.Location{
													Line: int(231),
													Column: int(40),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p2378,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(231),
														Column: int(30),
													},
													End: ast.Location{
														Line: int(231),
														Column: int(33),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "filter",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Function{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(231),
																Column: int(41),
															},
															End: ast.Location{
																Line: int(231),
																Column: int(59),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p2387,
														FreeVars: ast.Identifiers{
															"x",
														},
													},
													ParenLeftFodder: ast.Fodder{},
													Parameters: []ast.Parameter{
														ast.Parameter{
															NameFodder: ast.Fodder{},
															Name: "v",
															EqFodder: nil,
															DefaultArg: nil,
															CommaFodder: nil,
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(231),
																	Column: int(50),
																},
																End: ast.Location{
																	Line: int(231),
																	Column: int(51),
																},
																File: p1,
															},
														},
													},
													TrailingComma: false,
													ParenRightFodder: ast.Fodder{},
													Body: &ast.Binary{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(231),
																	Column: int(53),
																},
																End: ast.Location{
																	Line: int(231),
																	Column: int(59),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p2391,
															FreeVars: ast.Identifiers{
																"v",
																"x",
															},
														},
														Left: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(231),
																		Column: int(53),
																	},
																	End: ast.Location{
																		Line: int(231),
																		Column: int(54),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p2391,
																FreeVars: ast.Identifiers{
																	"v",
																},
															},
															Id: "v",
														},
														OpFodder: ast.Fodder{},
														Op: ast.BinaryOp(12),
														Right: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(231),
																		Column: int(58),
																	},
																	End: ast.Location{
																		Line: int(231),
																		Column: int(59),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p2391,
																FreeVars: ast.Identifiers{
																	"x",
																},
															},
															Id: "x",
														},
													},
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(231),
																Column: int(61),
															},
															End: ast.Location{
																Line: int(231),
																Column: int(64),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p2387,
														FreeVars: ast.Identifiers{
															"arr",
														},
													},
													Id: "arr",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					TrailingComma: false,
					TailStrict: false,
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(231),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(231),
					Column: int(66),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "mod",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "a",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(233),
								Column: int(7),
							},
							End: ast.Location{
								Line: int(233),
								Column: int(8),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "b",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(233),
								Column: int(10),
							},
							End: ast.Location{
								Line: int(233),
								Column: int(11),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(234),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(239),
								Column: int(94),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p2405,
						FreeVars: ast.Identifiers{
							"a",
							"b",
							"std",
						},
					},
					Cond: &ast.Binary{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(234),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(234),
									Column: int(42),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p2405,
							FreeVars: ast.Identifiers{
								"a",
								"b",
								"std",
							},
						},
						Left: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(234),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(234),
										Column: int(23),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p2405,
								FreeVars: ast.Identifiers{
									"a",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(234),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(234),
											Column: int(20),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2405,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(234),
												Column: int(8),
											},
											End: ast.Location{
												Line: int(234),
												Column: int(11),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "isNumber",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(234),
														Column: int(21),
													},
													End: ast.Location{
														Line: int(234),
														Column: int(22),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p2418,
												FreeVars: ast.Identifiers{
													"a",
												},
											},
											Id: "a",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
						OpFodder: ast.Fodder{},
						Op: ast.BinaryOp(17),
						Right: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(234),
										Column: int(27),
									},
									End: ast.Location{
										Line: int(234),
										Column: int(42),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p2405,
								FreeVars: ast.Identifiers{
									"b",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(234),
											Column: int(27),
										},
										End: ast.Location{
											Line: int(234),
											Column: int(39),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2405,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(234),
												Column: int(27),
											},
											End: ast.Location{
												Line: int(234),
												Column: int(30),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "isNumber",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(234),
														Column: int(40),
													},
													End: ast.Location{
														Line: int(234),
														Column: int(41),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p2429,
												FreeVars: ast.Identifiers{
													"b",
												},
											},
											Id: "b",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
					},
					ThenFodder: ast.Fodder{},
					BranchTrue: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(235),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(235),
									Column: int(23),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p2405,
							FreeVars: ast.Identifiers{
								"a",
								"b",
								"std",
							},
						},
						Target: &ast.Index{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(235),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(235),
										Column: int(17),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p2405,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(235),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(235),
											Column: int(10),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
											Comment: []string{},
										},
									},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Id: "std",
							},
							LeftBracketFodder: ast.Fodder{},
							Index: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Value: "modulo",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							RightBracketFodder: ast.Fodder{},
							Id: nil,
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(235),
													Column: int(18),
												},
												End: ast.Location{
													Line: int(235),
													Column: int(19),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p2441,
											FreeVars: ast.Identifiers{
												"a",
											},
										},
										Id: "a",
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(235),
													Column: int(21),
												},
												End: ast.Location{
													Line: int(235),
													Column: int(22),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p2441,
											FreeVars: ast.Identifiers{
												"b",
											},
										},
										Id: "b",
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
							Comment: []string{},
						},
					},
					BranchFalse: &ast.Conditional{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(236),
									Column: int(10),
								},
								End: ast.Location{
									Line: int(239),
									Column: int(94),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p2405,
							FreeVars: ast.Identifiers{
								"a",
								"b",
								"std",
							},
						},
						Cond: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(236),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(236),
										Column: int(28),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p2405,
								FreeVars: ast.Identifiers{
									"a",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(236),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(236),
											Column: int(25),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2405,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(236),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(236),
												Column: int(16),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "isString",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(236),
														Column: int(26),
													},
													End: ast.Location{
														Line: int(236),
														Column: int(27),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p2457,
												FreeVars: ast.Identifiers{
													"a",
												},
											},
											Id: "a",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
						ThenFodder: ast.Fodder{},
						BranchTrue: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(237),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(237),
										Column: int(23),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p2405,
								FreeVars: ast.Identifiers{
									"a",
									"b",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(237),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(237),
											Column: int(17),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2405,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(237),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(237),
												Column: int(10),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "format",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(237),
														Column: int(18),
													},
													End: ast.Location{
														Line: int(237),
														Column: int(19),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p2469,
												FreeVars: ast.Identifiers{
													"a",
												},
											},
											Id: "a",
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(237),
														Column: int(21),
													},
													End: ast.Location{
														Line: int(237),
														Column: int(22),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p2469,
												FreeVars: ast.Identifiers{
													"b",
												},
											},
											Id: "b",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						BranchFalse: &ast.Error{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(239),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(239),
										Column: int(94),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
										Comment: []string{},
									},
								},
								Ctx: p2405,
								FreeVars: ast.Identifiers{
									"a",
									"b",
									"std",
								},
							},
							Expr: &ast.Binary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(239),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(239),
											Column: int(94),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2405,
									FreeVars: ast.Identifiers{
										"a",
										"b",
										"std",
									},
								},
								Left: &ast.Binary{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(239),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(239),
												Column: int(88),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p2405,
										FreeVars: ast.Identifiers{
											"a",
											"b",
											"std",
										},
									},
									Left: &ast.Binary{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(239),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(239),
													Column: int(74),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p2405,
											FreeVars: ast.Identifiers{
												"a",
												"std",
											},
										},
										Left: &ast.Binary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(239),
														Column: int(13),
													},
													End: ast.Location{
														Line: int(239),
														Column: int(64),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p2405,
												FreeVars: ast.Identifiers{
													"a",
													"std",
												},
											},
											Left: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(239),
															Column: int(13),
														},
														End: ast.Location{
															Line: int(239),
															Column: int(50),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p2405,
													FreeVars: nil,
												},
												Value: "Operator % cannot be used on types ",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											OpFodder: ast.Fodder{},
											Op: ast.BinaryOp(3),
											Right: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(239),
															Column: int(53),
														},
														End: ast.Location{
															Line: int(239),
															Column: int(64),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p2405,
													FreeVars: ast.Identifiers{
														"a",
														"std",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(239),
																Column: int(53),
															},
															End: ast.Location{
																Line: int(239),
																Column: int(61),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p2405,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(239),
																	Column: int(53),
																},
																End: ast.Location{
																	Line: int(239),
																	Column: int(56),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: ast.Fodder{},
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "type",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: ast.Fodder{},
													Id: nil,
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(239),
																			Column: int(62),
																		},
																		End: ast.Location{
																			Line: int(239),
																			Column: int(63),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p2495,
																	FreeVars: ast.Identifiers{
																		"a",
																	},
																},
																Id: "a",
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
										},
										OpFodder: ast.Fodder{},
										Op: ast.BinaryOp(3),
										Right: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(239),
														Column: int(67),
													},
													End: ast.Location{
														Line: int(239),
														Column: int(74),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p2405,
												FreeVars: nil,
											},
											Value: " and ",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
									},
									OpFodder: ast.Fodder{},
									Op: ast.BinaryOp(3),
									Right: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(239),
													Column: int(77),
												},
												End: ast.Location{
													Line: int(239),
													Column: int(88),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p2405,
											FreeVars: ast.Identifiers{
												"b",
												"std",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(239),
														Column: int(77),
													},
													End: ast.Location{
														Line: int(239),
														Column: int(85),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p2405,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(239),
															Column: int(77),
														},
														End: ast.Location{
															Line: int(239),
															Column: int(80),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: ast.Fodder{},
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "type",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: ast.Fodder{},
											Id: nil,
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(239),
																	Column: int(86),
																},
																End: ast.Location{
																	Line: int(239),
																	Column: int(87),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p2507,
															FreeVars: ast.Identifiers{
																"b",
															},
														},
														Id: "b",
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
								},
								OpFodder: ast.Fodder{},
								Op: ast.BinaryOp(3),
								Right: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(239),
												Column: int(91),
											},
											End: ast.Location{
												Line: int(239),
												Column: int(94),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p2405,
										FreeVars: nil,
									},
									Value: ".",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
							},
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(233),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(239),
					Column: int(94),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "map",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "func",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(241),
								Column: int(7),
							},
							End: ast.Location{
								Line: int(241),
								Column: int(11),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(241),
								Column: int(13),
							},
							End: ast.Location{
								Line: int(241),
								Column: int(16),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(242),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(247),
								Column: int(63),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p2516,
						FreeVars: ast.Identifiers{
							"arr",
							"func",
							"std",
						},
					},
					Cond: &ast.Unary{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(242),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(242),
									Column: int(29),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p2516,
							FreeVars: ast.Identifiers{
								"func",
								"std",
							},
						},
						Op: ast.UnaryOp(0),
						Expr: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(242),
										Column: int(9),
									},
									End: ast.Location{
										Line: int(242),
										Column: int(29),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p2516,
								FreeVars: ast.Identifiers{
									"func",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(242),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(242),
											Column: int(23),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2516,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(242),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(242),
												Column: int(12),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "isFunction",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(242),
														Column: int(24),
													},
													End: ast.Location{
														Line: int(242),
														Column: int(28),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p2529,
												FreeVars: ast.Identifiers{
													"func",
												},
											},
											Id: "func",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
					},
					ThenFodder: ast.Fodder{},
					BranchTrue: &ast.Error{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(243),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(243),
									Column: int(76),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
									Comment: []string{},
								},
							},
							Ctx: p2516,
							FreeVars: ast.Identifiers{
								"func",
								"std",
							},
						},
						Expr: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(243),
										Column: int(14),
									},
									End: ast.Location{
										Line: int(243),
										Column: int(75),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p2516,
								FreeVars: ast.Identifiers{
									"func",
									"std",
								},
							},
							Left: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(243),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(243),
											Column: int(58),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2516,
									FreeVars: nil,
								},
								Value: "std.map first param must be function, got ",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(3),
							Right: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(243),
											Column: int(61),
										},
										End: ast.Location{
											Line: int(243),
											Column: int(75),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2516,
									FreeVars: ast.Identifiers{
										"func",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(243),
												Column: int(61),
											},
											End: ast.Location{
												Line: int(243),
												Column: int(69),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p2516,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(243),
													Column: int(61),
												},
												End: ast.Location{
													Line: int(243),
													Column: int(64),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "type",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(243),
															Column: int(70),
														},
														End: ast.Location{
															Line: int(243),
															Column: int(74),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p2546,
													FreeVars: ast.Identifiers{
														"func",
													},
												},
												Id: "func",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
						},
					},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
							Comment: []string{},
						},
					},
					BranchFalse: &ast.Conditional{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(244),
									Column: int(10),
								},
								End: ast.Location{
									Line: int(247),
									Column: int(63),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p2516,
							FreeVars: ast.Identifiers{
								"arr",
								"func",
								"std",
							},
						},
						Cond: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(244),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(244),
										Column: int(52),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p2516,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
								},
							},
							Left: &ast.Unary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(244),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(244),
											Column: int(30),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2516,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
								},
								Op: ast.UnaryOp(0),
								Expr: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(244),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(244),
												Column: int(30),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p2516,
										FreeVars: ast.Identifiers{
											"arr",
											"std",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(244),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(244),
													Column: int(25),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p2516,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(244),
														Column: int(14),
													},
													End: ast.Location{
														Line: int(244),
														Column: int(17),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "isArray",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(244),
																Column: int(26),
															},
															End: ast.Location{
																Line: int(244),
																Column: int(29),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p2564,
														FreeVars: ast.Identifiers{
															"arr",
														},
													},
													Id: "arr",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(17),
							Right: &ast.Unary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(244),
											Column: int(34),
										},
										End: ast.Location{
											Line: int(244),
											Column: int(52),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2516,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
								},
								Op: ast.UnaryOp(0),
								Expr: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(244),
												Column: int(35),
											},
											End: ast.Location{
												Line: int(244),
												Column: int(52),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p2516,
										FreeVars: ast.Identifiers{
											"arr",
											"std",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(244),
													Column: int(35),
												},
												End: ast.Location{
													Line: int(244),
													Column: int(47),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p2516,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(244),
														Column: int(35),
													},
													End: ast.Location{
														Line: int(244),
														Column: int(38),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "isString",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(244),
																Column: int(48),
															},
															End: ast.Location{
																Line: int(244),
																Column: int(51),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p2577,
														FreeVars: ast.Identifiers{
															"arr",
														},
													},
													Id: "arr",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
							},
						},
						ThenFodder: ast.Fodder{},
						BranchTrue: &ast.Error{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(245),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(245),
										Column: int(82),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
										Comment: []string{},
									},
								},
								Ctx: p2516,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
								},
							},
							Expr: &ast.Binary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(245),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(245),
											Column: int(81),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2516,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
								},
								Left: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(245),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(245),
												Column: int(65),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p2516,
										FreeVars: nil,
									},
									Value: "std.map second param must be array / string, got ",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								OpFodder: ast.Fodder{},
								Op: ast.BinaryOp(3),
								Right: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(245),
												Column: int(68),
											},
											End: ast.Location{
												Line: int(245),
												Column: int(81),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p2516,
										FreeVars: ast.Identifiers{
											"arr",
											"std",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(245),
													Column: int(68),
												},
												End: ast.Location{
													Line: int(245),
													Column: int(76),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p2516,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(245),
														Column: int(68),
													},
													End: ast.Location{
														Line: int(245),
														Column: int(71),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "type",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(245),
																Column: int(77),
															},
															End: ast.Location{
																Line: int(245),
																Column: int(80),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p2594,
														FreeVars: ast.Identifiers{
															"arr",
														},
													},
													Id: "arr",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
							},
						},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						BranchFalse: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(247),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(247),
										Column: int(63),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p2516,
								FreeVars: ast.Identifiers{
									"arr",
									"func",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(247),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(247),
											Column: int(20),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2516,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(247),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(247),
												Column: int(10),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "makeArray",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(247),
														Column: int(21),
													},
													End: ast.Location{
														Line: int(247),
														Column: int(36),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p2607,
												FreeVars: ast.Identifiers{
													"arr",
													"std",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(247),
															Column: int(21),
														},
														End: ast.Location{
															Line: int(247),
															Column: int(31),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p2607,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(247),
																Column: int(21),
															},
															End: ast.Location{
																Line: int(247),
																Column: int(24),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "length",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(247),
																		Column: int(32),
																	},
																	End: ast.Location{
																		Line: int(247),
																		Column: int(35),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p2616,
																FreeVars: ast.Identifiers{
																	"arr",
																},
															},
															Id: "arr",
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Function{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(247),
														Column: int(38),
													},
													End: ast.Location{
														Line: int(247),
														Column: int(62),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p2607,
												FreeVars: ast.Identifiers{
													"arr",
													"func",
												},
											},
											ParenLeftFodder: ast.Fodder{},
											Parameters: []ast.Parameter{
												ast.Parameter{
													NameFodder: ast.Fodder{},
													Name: "i",
													EqFodder: nil,
													DefaultArg: nil,
													CommaFodder: nil,
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(247),
															Column: int(47),
														},
														End: ast.Location{
															Line: int(247),
															Column: int(48),
														},
														File: p1,
													},
												},
											},
											TrailingComma: false,
											ParenRightFodder: ast.Fodder{},
											Body: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(247),
															Column: int(50),
														},
														End: ast.Location{
															Line: int(247),
															Column: int(62),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p2622,
													FreeVars: ast.Identifiers{
														"arr",
														"func",
														"i",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(247),
																Column: int(50),
															},
															End: ast.Location{
																Line: int(247),
																Column: int(54),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p2622,
														FreeVars: ast.Identifiers{
															"func",
														},
													},
													Id: "func",
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Index{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(247),
																			Column: int(55),
																		},
																		End: ast.Location{
																			Line: int(247),
																			Column: int(61),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p2628,
																	FreeVars: ast.Identifiers{
																		"arr",
																		"i",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(247),
																				Column: int(55),
																			},
																			End: ast.Location{
																				Line: int(247),
																				Column: int(58),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p2628,
																		FreeVars: ast.Identifiers{
																			"arr",
																		},
																	},
																	Id: "arr",
																},
																LeftBracketFodder: ast.Fodder{},
																Index: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(247),
																				Column: int(59),
																			},
																			End: ast.Location{
																				Line: int(247),
																				Column: int(60),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p2628,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																	},
																	Id: "i",
																},
																RightBracketFodder: ast.Fodder{},
																Id: nil,
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(241),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(247),
					Column: int(63),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "mapWithIndex",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "func",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(249),
								Column: int(16),
							},
							End: ast.Location{
								Line: int(249),
								Column: int(20),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(249),
								Column: int(22),
							},
							End: ast.Location{
								Line: int(249),
								Column: int(25),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(250),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(255),
								Column: int(66),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p2640,
						FreeVars: ast.Identifiers{
							"arr",
							"func",
							"std",
						},
					},
					Cond: &ast.Unary{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(250),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(250),
									Column: int(29),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p2640,
							FreeVars: ast.Identifiers{
								"func",
								"std",
							},
						},
						Op: ast.UnaryOp(0),
						Expr: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(250),
										Column: int(9),
									},
									End: ast.Location{
										Line: int(250),
										Column: int(29),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p2640,
								FreeVars: ast.Identifiers{
									"func",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(250),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(250),
											Column: int(23),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2640,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(250),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(250),
												Column: int(12),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "isFunction",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(250),
														Column: int(24),
													},
													End: ast.Location{
														Line: int(250),
														Column: int(28),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p2653,
												FreeVars: ast.Identifiers{
													"func",
												},
											},
											Id: "func",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
					},
					ThenFodder: ast.Fodder{},
					BranchTrue: &ast.Error{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(251),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(251),
									Column: int(85),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
									Comment: []string{},
								},
							},
							Ctx: p2640,
							FreeVars: ast.Identifiers{
								"func",
								"std",
							},
						},
						Expr: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(251),
										Column: int(14),
									},
									End: ast.Location{
										Line: int(251),
										Column: int(84),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p2640,
								FreeVars: ast.Identifiers{
									"func",
									"std",
								},
							},
							Left: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(251),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(251),
											Column: int(67),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2640,
									FreeVars: nil,
								},
								Value: "std.mapWithIndex first param must be function, got ",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(3),
							Right: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(251),
											Column: int(70),
										},
										End: ast.Location{
											Line: int(251),
											Column: int(84),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2640,
									FreeVars: ast.Identifiers{
										"func",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(251),
												Column: int(70),
											},
											End: ast.Location{
												Line: int(251),
												Column: int(78),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p2640,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(251),
													Column: int(70),
												},
												End: ast.Location{
													Line: int(251),
													Column: int(73),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "type",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(251),
															Column: int(79),
														},
														End: ast.Location{
															Line: int(251),
															Column: int(83),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p2670,
													FreeVars: ast.Identifiers{
														"func",
													},
												},
												Id: "func",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
						},
					},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
							Comment: []string{},
						},
					},
					BranchFalse: &ast.Conditional{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(252),
									Column: int(10),
								},
								End: ast.Location{
									Line: int(255),
									Column: int(66),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p2640,
							FreeVars: ast.Identifiers{
								"arr",
								"func",
								"std",
							},
						},
						Cond: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(252),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(252),
										Column: int(52),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p2640,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
								},
							},
							Left: &ast.Unary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(252),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(252),
											Column: int(30),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2640,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
								},
								Op: ast.UnaryOp(0),
								Expr: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(252),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(252),
												Column: int(30),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p2640,
										FreeVars: ast.Identifiers{
											"arr",
											"std",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(252),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(252),
													Column: int(25),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p2640,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(252),
														Column: int(14),
													},
													End: ast.Location{
														Line: int(252),
														Column: int(17),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "isArray",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(252),
																Column: int(26),
															},
															End: ast.Location{
																Line: int(252),
																Column: int(29),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p2688,
														FreeVars: ast.Identifiers{
															"arr",
														},
													},
													Id: "arr",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(17),
							Right: &ast.Unary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(252),
											Column: int(34),
										},
										End: ast.Location{
											Line: int(252),
											Column: int(52),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2640,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
								},
								Op: ast.UnaryOp(0),
								Expr: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(252),
												Column: int(35),
											},
											End: ast.Location{
												Line: int(252),
												Column: int(52),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p2640,
										FreeVars: ast.Identifiers{
											"arr",
											"std",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(252),
													Column: int(35),
												},
												End: ast.Location{
													Line: int(252),
													Column: int(47),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p2640,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(252),
														Column: int(35),
													},
													End: ast.Location{
														Line: int(252),
														Column: int(38),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "isString",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(252),
																Column: int(48),
															},
															End: ast.Location{
																Line: int(252),
																Column: int(51),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p2701,
														FreeVars: ast.Identifiers{
															"arr",
														},
													},
													Id: "arr",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
							},
						},
						ThenFodder: ast.Fodder{},
						BranchTrue: &ast.Error{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(253),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(253),
										Column: int(82),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
										Comment: []string{},
									},
								},
								Ctx: p2640,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
								},
							},
							Expr: &ast.Binary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(253),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(253),
											Column: int(81),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2640,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
								},
								Left: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(253),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(253),
												Column: int(65),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p2640,
										FreeVars: nil,
									},
									Value: "std.mapWithIndex second param must be array, got ",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								OpFodder: ast.Fodder{},
								Op: ast.BinaryOp(3),
								Right: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(253),
												Column: int(68),
											},
											End: ast.Location{
												Line: int(253),
												Column: int(81),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p2640,
										FreeVars: ast.Identifiers{
											"arr",
											"std",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(253),
													Column: int(68),
												},
												End: ast.Location{
													Line: int(253),
													Column: int(76),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p2640,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(253),
														Column: int(68),
													},
													End: ast.Location{
														Line: int(253),
														Column: int(71),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "type",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(253),
																Column: int(77),
															},
															End: ast.Location{
																Line: int(253),
																Column: int(80),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p2718,
														FreeVars: ast.Identifiers{
															"arr",
														},
													},
													Id: "arr",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
							},
						},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						BranchFalse: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(255),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(255),
										Column: int(66),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p2640,
								FreeVars: ast.Identifiers{
									"arr",
									"func",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(255),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(255),
											Column: int(20),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2640,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(255),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(255),
												Column: int(10),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "makeArray",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(255),
														Column: int(21),
													},
													End: ast.Location{
														Line: int(255),
														Column: int(36),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p2731,
												FreeVars: ast.Identifiers{
													"arr",
													"std",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(255),
															Column: int(21),
														},
														End: ast.Location{
															Line: int(255),
															Column: int(31),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p2731,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(255),
																Column: int(21),
															},
															End: ast.Location{
																Line: int(255),
																Column: int(24),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "length",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(255),
																		Column: int(32),
																	},
																	End: ast.Location{
																		Line: int(255),
																		Column: int(35),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p2740,
																FreeVars: ast.Identifiers{
																	"arr",
																},
															},
															Id: "arr",
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Function{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(255),
														Column: int(38),
													},
													End: ast.Location{
														Line: int(255),
														Column: int(65),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p2731,
												FreeVars: ast.Identifiers{
													"arr",
													"func",
												},
											},
											ParenLeftFodder: ast.Fodder{},
											Parameters: []ast.Parameter{
												ast.Parameter{
													NameFodder: ast.Fodder{},
													Name: "i",
													EqFodder: nil,
													DefaultArg: nil,
													CommaFodder: nil,
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(255),
															Column: int(47),
														},
														End: ast.Location{
															Line: int(255),
															Column: int(48),
														},
														File: p1,
													},
												},
											},
											TrailingComma: false,
											ParenRightFodder: ast.Fodder{},
											Body: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(255),
															Column: int(50),
														},
														End: ast.Location{
															Line: int(255),
															Column: int(65),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p2746,
													FreeVars: ast.Identifiers{
														"arr",
														"func",
														"i",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(255),
																Column: int(50),
															},
															End: ast.Location{
																Line: int(255),
																Column: int(54),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p2746,
														FreeVars: ast.Identifiers{
															"func",
														},
													},
													Id: "func",
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(255),
																			Column: int(55),
																		},
																		End: ast.Location{
																			Line: int(255),
																			Column: int(56),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p2752,
																	FreeVars: ast.Identifiers{
																		"i",
																	},
																},
																Id: "i",
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Index{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(255),
																			Column: int(58),
																		},
																		End: ast.Location{
																			Line: int(255),
																			Column: int(64),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p2752,
																	FreeVars: ast.Identifiers{
																		"arr",
																		"i",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(255),
																				Column: int(58),
																			},
																			End: ast.Location{
																				Line: int(255),
																				Column: int(61),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p2752,
																		FreeVars: ast.Identifiers{
																			"arr",
																		},
																	},
																	Id: "arr",
																},
																LeftBracketFodder: ast.Fodder{},
																Index: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(255),
																				Column: int(62),
																			},
																			End: ast.Location{
																				Line: int(255),
																				Column: int(63),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p2752,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																	},
																	Id: "i",
																},
																RightBracketFodder: ast.Fodder{},
																Id: nil,
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(249),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(255),
					Column: int(66),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "mapWithKey",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "func",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(257),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(257),
								Column: int(18),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "obj",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(257),
								Column: int(20),
							},
							End: ast.Location{
								Line: int(257),
								Column: int(23),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(258),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(263),
								Column: int(62),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p2766,
						FreeVars: ast.Identifiers{
							"func",
							"obj",
							"std",
						},
					},
					Cond: &ast.Unary{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(258),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(258),
									Column: int(29),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p2766,
							FreeVars: ast.Identifiers{
								"func",
								"std",
							},
						},
						Op: ast.UnaryOp(0),
						Expr: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(258),
										Column: int(9),
									},
									End: ast.Location{
										Line: int(258),
										Column: int(29),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p2766,
								FreeVars: ast.Identifiers{
									"func",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(258),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(258),
											Column: int(23),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2766,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(258),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(258),
												Column: int(12),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "isFunction",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(258),
														Column: int(24),
													},
													End: ast.Location{
														Line: int(258),
														Column: int(28),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p2779,
												FreeVars: ast.Identifiers{
													"func",
												},
											},
											Id: "func",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
					},
					ThenFodder: ast.Fodder{},
					BranchTrue: &ast.Error{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(259),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(259),
									Column: int(83),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
									Comment: []string{},
								},
							},
							Ctx: p2766,
							FreeVars: ast.Identifiers{
								"func",
								"std",
							},
						},
						Expr: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(259),
										Column: int(14),
									},
									End: ast.Location{
										Line: int(259),
										Column: int(82),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p2766,
								FreeVars: ast.Identifiers{
									"func",
									"std",
								},
							},
							Left: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(259),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(259),
											Column: int(65),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2766,
									FreeVars: nil,
								},
								Value: "std.mapWithKey first param must be function, got ",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(3),
							Right: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(259),
											Column: int(68),
										},
										End: ast.Location{
											Line: int(259),
											Column: int(82),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2766,
									FreeVars: ast.Identifiers{
										"func",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(259),
												Column: int(68),
											},
											End: ast.Location{
												Line: int(259),
												Column: int(76),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p2766,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(259),
													Column: int(68),
												},
												End: ast.Location{
													Line: int(259),
													Column: int(71),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "type",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(259),
															Column: int(77),
														},
														End: ast.Location{
															Line: int(259),
															Column: int(81),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p2796,
													FreeVars: ast.Identifiers{
														"func",
													},
												},
												Id: "func",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
						},
					},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
							Comment: []string{},
						},
					},
					BranchFalse: &ast.Conditional{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(260),
									Column: int(10),
								},
								End: ast.Location{
									Line: int(263),
									Column: int(62),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p2766,
							FreeVars: ast.Identifiers{
								"func",
								"obj",
								"std",
							},
						},
						Cond: &ast.Unary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(260),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(260),
										Column: int(31),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p2766,
								FreeVars: ast.Identifiers{
									"obj",
									"std",
								},
							},
							Op: ast.UnaryOp(0),
							Expr: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(260),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(260),
											Column: int(31),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2766,
									FreeVars: ast.Identifiers{
										"obj",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(260),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(260),
												Column: int(26),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p2766,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(260),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(260),
													Column: int(17),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "isObject",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(260),
															Column: int(27),
														},
														End: ast.Location{
															Line: int(260),
															Column: int(30),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p2812,
													FreeVars: ast.Identifiers{
														"obj",
													},
												},
												Id: "obj",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
						},
						ThenFodder: ast.Fodder{},
						BranchTrue: &ast.Error{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(261),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(261),
										Column: int(81),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
										Comment: []string{},
									},
								},
								Ctx: p2766,
								FreeVars: ast.Identifiers{
									"obj",
									"std",
								},
							},
							Expr: &ast.Binary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(261),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(261),
											Column: int(80),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2766,
									FreeVars: ast.Identifiers{
										"obj",
										"std",
									},
								},
								Left: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(261),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(261),
												Column: int(64),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p2766,
										FreeVars: nil,
									},
									Value: "std.mapWithKey second param must be object, got ",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								OpFodder: ast.Fodder{},
								Op: ast.BinaryOp(3),
								Right: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(261),
												Column: int(67),
											},
											End: ast.Location{
												Line: int(261),
												Column: int(80),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p2766,
										FreeVars: ast.Identifiers{
											"obj",
											"std",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(261),
													Column: int(67),
												},
												End: ast.Location{
													Line: int(261),
													Column: int(75),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p2766,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(261),
														Column: int(67),
													},
													End: ast.Location{
														Line: int(261),
														Column: int(70),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "type",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(261),
																Column: int(76),
															},
															End: ast.Location{
																Line: int(261),
																Column: int(79),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p2829,
														FreeVars: ast.Identifiers{
															"obj",
														},
													},
													Id: "obj",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
							},
						},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						BranchFalse: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"func",
									"obj",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: nil,
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "$objectFlatMerge",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: nil,
								Id: nil,
							},
							FodderLeft: nil,
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"func",
													"obj",
													"std",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: nil,
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "flatMap",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: nil,
												Id: nil,
											},
											FodderLeft: nil,
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Function{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"func",
																	"obj",
																},
															},
															ParenLeftFodder: nil,
															Parameters: []ast.Parameter{
																ast.Parameter{
																	NameFodder: nil,
																	Name: "k",
																	EqFodder: nil,
																	DefaultArg: nil,
																	CommaFodder: nil,
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																},
															},
															TrailingComma: false,
															ParenRightFodder: nil,
															Body: &ast.Array{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"func",
																		"k",
																		"obj",
																	},
																},
																Elements: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.DesugaredObject{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(263),
																						Column: int(7),
																					},
																					End: ast.Location{
																						Line: int(263),
																						Column: int(62),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(6),
																						Comment: []string{},
																					},
																				},
																				Ctx: p2766,
																				FreeVars: ast.Identifiers{
																					"func",
																					"k",
																					"obj",
																				},
																			},
																			Asserts: ast.Nodes{},
																			Fields: ast.DesugaredObjectFields{
																				ast.DesugaredObjectField{
																					Hide: ast.ObjectFieldHide(1),
																					Name: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(263),
																									Column: int(10),
																								},
																								End: ast.Location{
																									Line: int(263),
																									Column: int(11),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p2766,
																							FreeVars: ast.Identifiers{
																								"k",
																							},
																						},
																						Id: "k",
																					},
																					Body: &ast.Apply{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(263),
																									Column: int(14),
																								},
																								End: ast.Location{
																									Line: int(263),
																									Column: int(29),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p2861,
																							FreeVars: ast.Identifiers{
																								"func",
																								"k",
																								"obj",
																							},
																						},
																						Target: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(263),
																										Column: int(14),
																									},
																									End: ast.Location{
																										Line: int(263),
																										Column: int(18),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p2861,
																								FreeVars: ast.Identifiers{
																									"func",
																								},
																							},
																							Id: "func",
																						},
																						FodderLeft: ast.Fodder{},
																						Arguments: ast.Arguments{
																							Positional: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(263),
																													Column: int(19),
																												},
																												End: ast.Location{
																													Line: int(263),
																													Column: int(20),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p2867,
																											FreeVars: ast.Identifiers{
																												"k",
																											},
																										},
																										Id: "k",
																									},
																									CommaFodder: ast.Fodder{},
																								},
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Index{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(263),
																													Column: int(22),
																												},
																												End: ast.Location{
																													Line: int(263),
																													Column: int(28),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p2867,
																											FreeVars: ast.Identifiers{
																												"k",
																												"obj",
																											},
																										},
																										Target: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(263),
																														Column: int(22),
																													},
																													End: ast.Location{
																														Line: int(263),
																														Column: int(25),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p2867,
																												FreeVars: ast.Identifiers{
																													"obj",
																												},
																											},
																											Id: "obj",
																										},
																										LeftBracketFodder: ast.Fodder{},
																										Index: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(263),
																														Column: int(26),
																													},
																													End: ast.Location{
																														Line: int(263),
																														Column: int(27),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p2867,
																												FreeVars: ast.Identifiers{
																													"k",
																												},
																											},
																											Id: "k",
																										},
																										RightBracketFodder: ast.Fodder{},
																										Id: nil,
																									},
																									CommaFodder: nil,
																								},
																							},
																							Named: nil,
																						},
																						TrailingComma: false,
																						TailStrict: false,
																						FodderRight: ast.Fodder{},
																						TailStrictFodder: nil,
																					},
																					PlusSuper: false,
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(263),
																							Column: int(9),
																						},
																						End: ast.Location{
																							Line: int(263),
																							Column: int(29),
																						},
																						File: p1,
																					},
																				},
																			},
																			Locals: ast.LocalBinds{},
																		},
																		CommaFodder: nil,
																	},
																},
																TrailingComma: false,
																CloseFodder: nil,
															},
														},
														CommaFodder: nil,
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Apply{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(263),
																		Column: int(39),
																	},
																	End: ast.Location{
																		Line: int(263),
																		Column: int(60),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p2766,
																FreeVars: ast.Identifiers{
																	"obj",
																	"std",
																},
															},
															Target: &ast.Index{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(263),
																			Column: int(39),
																		},
																		End: ast.Location{
																			Line: int(263),
																			Column: int(55),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p2766,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(263),
																				Column: int(39),
																			},
																			End: ast.Location{
																				Line: int(263),
																				Column: int(42),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Id: "std",
																},
																LeftBracketFodder: ast.Fodder{},
																Index: &ast.LiteralString{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: nil,
																	},
																	Value: "objectFields",
																	Kind: ast.LiteralStringKind(1),
																	BlockIndent: "",
																	BlockTermIndent: "",
																},
																RightBracketFodder: ast.Fodder{},
																Id: nil,
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(263),
																						Column: int(56),
																					},
																					End: ast.Location{
																						Line: int(263),
																						Column: int(59),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p2884,
																				FreeVars: ast.Identifiers{
																					"obj",
																				},
																			},
																			Id: "obj",
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															TrailingComma: false,
															TailStrict: false,
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: nil,
											TailStrictFodder: nil,
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: nil,
							TailStrictFodder: nil,
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(257),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(263),
					Column: int(62),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "flatMap",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "func",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(265),
								Column: int(11),
							},
							End: ast.Location{
								Line: int(265),
								Column: int(15),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(265),
								Column: int(17),
							},
							End: ast.Location{
								Line: int(265),
								Column: int(20),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(266),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(272),
								Column: int(89),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p2892,
						FreeVars: ast.Identifiers{
							"arr",
							"func",
							"std",
						},
					},
					Cond: &ast.Unary{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(266),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(266),
									Column: int(29),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p2892,
							FreeVars: ast.Identifiers{
								"func",
								"std",
							},
						},
						Op: ast.UnaryOp(0),
						Expr: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(266),
										Column: int(9),
									},
									End: ast.Location{
										Line: int(266),
										Column: int(29),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p2892,
								FreeVars: ast.Identifiers{
									"func",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(266),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(266),
											Column: int(23),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2892,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(266),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(266),
												Column: int(12),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "isFunction",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(266),
														Column: int(24),
													},
													End: ast.Location{
														Line: int(266),
														Column: int(28),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p2905,
												FreeVars: ast.Identifiers{
													"func",
												},
											},
											Id: "func",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
					},
					ThenFodder: ast.Fodder{},
					BranchTrue: &ast.Error{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(267),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(267),
									Column: int(80),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
									Comment: []string{},
								},
							},
							Ctx: p2892,
							FreeVars: ast.Identifiers{
								"func",
								"std",
							},
						},
						Expr: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(267),
										Column: int(14),
									},
									End: ast.Location{
										Line: int(267),
										Column: int(79),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p2892,
								FreeVars: ast.Identifiers{
									"func",
									"std",
								},
							},
							Left: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(267),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(267),
											Column: int(62),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2892,
									FreeVars: nil,
								},
								Value: "std.flatMap first param must be function, got ",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(3),
							Right: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(267),
											Column: int(65),
										},
										End: ast.Location{
											Line: int(267),
											Column: int(79),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2892,
									FreeVars: ast.Identifiers{
										"func",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(267),
												Column: int(65),
											},
											End: ast.Location{
												Line: int(267),
												Column: int(73),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p2892,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(267),
													Column: int(65),
												},
												End: ast.Location{
													Line: int(267),
													Column: int(68),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "type",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(267),
															Column: int(74),
														},
														End: ast.Location{
															Line: int(267),
															Column: int(78),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p2922,
													FreeVars: ast.Identifiers{
														"func",
													},
												},
												Id: "func",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
						},
					},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
							Comment: []string{},
						},
					},
					BranchFalse: &ast.Conditional{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(268),
									Column: int(10),
								},
								End: ast.Location{
									Line: int(272),
									Column: int(89),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p2892,
							FreeVars: ast.Identifiers{
								"arr",
								"func",
								"std",
							},
						},
						Cond: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(268),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(268),
										Column: int(29),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p2892,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(268),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(268),
											Column: int(24),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2892,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(268),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(268),
												Column: int(16),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "isArray",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(268),
														Column: int(25),
													},
													End: ast.Location{
														Line: int(268),
														Column: int(28),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p2936,
												FreeVars: ast.Identifiers{
													"arr",
												},
											},
											Id: "arr",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
						ThenFodder: ast.Fodder{},
						BranchTrue: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(269),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(269),
										Column: int(82),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p2892,
								FreeVars: ast.Identifiers{
									"arr",
									"func",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(269),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(269),
											Column: int(24),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2892,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(269),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(269),
												Column: int(10),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "flattenArrays",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(269),
														Column: int(25),
													},
													End: ast.Location{
														Line: int(269),
														Column: int(81),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p2948,
												FreeVars: ast.Identifiers{
													"arr",
													"func",
													"std",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(269),
															Column: int(25),
														},
														End: ast.Location{
															Line: int(269),
															Column: int(38),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p2948,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(269),
																Column: int(25),
															},
															End: ast.Location{
																Line: int(269),
																Column: int(28),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "makeArray",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Apply{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(269),
																		Column: int(39),
																	},
																	End: ast.Location{
																		Line: int(269),
																		Column: int(54),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p2957,
																FreeVars: ast.Identifiers{
																	"arr",
																	"std",
																},
															},
															Target: &ast.Index{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(269),
																			Column: int(39),
																		},
																		End: ast.Location{
																			Line: int(269),
																			Column: int(49),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p2957,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(269),
																				Column: int(39),
																			},
																			End: ast.Location{
																				Line: int(269),
																				Column: int(42),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Id: "std",
																},
																LeftBracketFodder: ast.Fodder{},
																Index: &ast.LiteralString{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: nil,
																	},
																	Value: "length",
																	Kind: ast.LiteralStringKind(1),
																	BlockIndent: "",
																	BlockTermIndent: "",
																},
																RightBracketFodder: ast.Fodder{},
																Id: nil,
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(269),
																						Column: int(50),
																					},
																					End: ast.Location{
																						Line: int(269),
																						Column: int(53),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p2966,
																				FreeVars: ast.Identifiers{
																					"arr",
																				},
																			},
																			Id: "arr",
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															TrailingComma: false,
															TailStrict: false,
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Function{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(269),
																		Column: int(56),
																	},
																	End: ast.Location{
																		Line: int(269),
																		Column: int(80),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p2957,
																FreeVars: ast.Identifiers{
																	"arr",
																	"func",
																},
															},
															ParenLeftFodder: ast.Fodder{},
															Parameters: []ast.Parameter{
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "i",
																	EqFodder: nil,
																	DefaultArg: nil,
																	CommaFodder: nil,
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(269),
																			Column: int(65),
																		},
																		End: ast.Location{
																			Line: int(269),
																			Column: int(66),
																		},
																		File: p1,
																	},
																},
															},
															TrailingComma: false,
															ParenRightFodder: ast.Fodder{},
															Body: &ast.Apply{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(269),
																			Column: int(68),
																		},
																		End: ast.Location{
																			Line: int(269),
																			Column: int(80),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p2972,
																	FreeVars: ast.Identifiers{
																		"arr",
																		"func",
																		"i",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(269),
																				Column: int(68),
																			},
																			End: ast.Location{
																				Line: int(269),
																				Column: int(72),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p2972,
																		FreeVars: ast.Identifiers{
																			"func",
																		},
																	},
																	Id: "func",
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Index{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(269),
																							Column: int(73),
																						},
																						End: ast.Location{
																							Line: int(269),
																							Column: int(79),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p2978,
																					FreeVars: ast.Identifiers{
																						"arr",
																						"i",
																					},
																				},
																				Target: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(269),
																								Column: int(73),
																							},
																							End: ast.Location{
																								Line: int(269),
																								Column: int(76),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p2978,
																						FreeVars: ast.Identifiers{
																							"arr",
																						},
																					},
																					Id: "arr",
																				},
																				LeftBracketFodder: ast.Fodder{},
																				Index: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(269),
																								Column: int(77),
																							},
																							End: ast.Location{
																								Line: int(269),
																								Column: int(78),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p2978,
																						FreeVars: ast.Identifiers{
																							"i",
																						},
																					},
																					Id: "i",
																				},
																				RightBracketFodder: ast.Fodder{},
																				Id: nil,
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																TrailingComma: false,
																TailStrict: false,
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						BranchFalse: &ast.Conditional{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(270),
										Column: int(10),
									},
									End: ast.Location{
										Line: int(272),
										Column: int(89),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p2892,
								FreeVars: ast.Identifiers{
									"arr",
									"func",
									"std",
								},
							},
							Cond: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(270),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(270),
											Column: int(30),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2892,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(270),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(270),
												Column: int(25),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p2892,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(270),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(270),
													Column: int(16),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "isString",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(270),
															Column: int(26),
														},
														End: ast.Location{
															Line: int(270),
															Column: int(29),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p2996,
													FreeVars: ast.Identifiers{
														"arr",
													},
												},
												Id: "arr",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
							ThenFodder: ast.Fodder{},
							BranchTrue: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(271),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(271),
											Column: int(77),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2892,
									FreeVars: ast.Identifiers{
										"arr",
										"func",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(271),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(271),
												Column: int(15),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p2892,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(271),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(271),
													Column: int(10),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
													Comment: []string{},
												},
											},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "join",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(271),
															Column: int(16),
														},
														End: ast.Location{
															Line: int(271),
															Column: int(18),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p3008,
													FreeVars: nil,
												},
												Value: "",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											CommaFodder: ast.Fodder{},
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(271),
															Column: int(20),
														},
														End: ast.Location{
															Line: int(271),
															Column: int(76),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p3008,
													FreeVars: ast.Identifiers{
														"arr",
														"func",
														"std",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(271),
																Column: int(20),
															},
															End: ast.Location{
																Line: int(271),
																Column: int(33),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p3008,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(271),
																	Column: int(20),
																},
																End: ast.Location{
																	Line: int(271),
																	Column: int(23),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: ast.Fodder{},
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "makeArray",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: ast.Fodder{},
													Id: nil,
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Apply{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(271),
																			Column: int(34),
																		},
																		End: ast.Location{
																			Line: int(271),
																			Column: int(49),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p3018,
																	FreeVars: ast.Identifiers{
																		"arr",
																		"std",
																	},
																},
																Target: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(271),
																				Column: int(34),
																			},
																			End: ast.Location{
																				Line: int(271),
																				Column: int(44),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p3018,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(271),
																					Column: int(34),
																				},
																				End: ast.Location{
																					Line: int(271),
																					Column: int(37),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Id: "std",
																	},
																	LeftBracketFodder: ast.Fodder{},
																	Index: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: nil,
																		},
																		Value: "length",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	RightBracketFodder: ast.Fodder{},
																	Id: nil,
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(271),
																							Column: int(45),
																						},
																						End: ast.Location{
																							Line: int(271),
																							Column: int(48),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p3027,
																					FreeVars: ast.Identifiers{
																						"arr",
																					},
																				},
																				Id: "arr",
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																TrailingComma: false,
																TailStrict: false,
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Function{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(271),
																			Column: int(51),
																		},
																		End: ast.Location{
																			Line: int(271),
																			Column: int(75),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p3018,
																	FreeVars: ast.Identifiers{
																		"arr",
																		"func",
																	},
																},
																ParenLeftFodder: ast.Fodder{},
																Parameters: []ast.Parameter{
																	ast.Parameter{
																		NameFodder: ast.Fodder{},
																		Name: "i",
																		EqFodder: nil,
																		DefaultArg: nil,
																		CommaFodder: nil,
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(271),
																				Column: int(60),
																			},
																			End: ast.Location{
																				Line: int(271),
																				Column: int(61),
																			},
																			File: p1,
																		},
																	},
																},
																TrailingComma: false,
																ParenRightFodder: ast.Fodder{},
																Body: &ast.Apply{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(271),
																				Column: int(63),
																			},
																			End: ast.Location{
																				Line: int(271),
																				Column: int(75),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p3033,
																		FreeVars: ast.Identifiers{
																			"arr",
																			"func",
																			"i",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(271),
																					Column: int(63),
																				},
																				End: ast.Location{
																					Line: int(271),
																					Column: int(67),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3033,
																			FreeVars: ast.Identifiers{
																				"func",
																			},
																		},
																		Id: "func",
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Index{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(271),
																								Column: int(68),
																							},
																							End: ast.Location{
																								Line: int(271),
																								Column: int(74),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p3039,
																						FreeVars: ast.Identifiers{
																							"arr",
																							"i",
																						},
																					},
																					Target: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(271),
																									Column: int(68),
																								},
																								End: ast.Location{
																									Line: int(271),
																									Column: int(71),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p3039,
																							FreeVars: ast.Identifiers{
																								"arr",
																							},
																						},
																						Id: "arr",
																					},
																					LeftBracketFodder: ast.Fodder{},
																					Index: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(271),
																									Column: int(72),
																								},
																								End: ast.Location{
																									Line: int(271),
																									Column: int(73),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p3039,
																							FreeVars: ast.Identifiers{
																								"i",
																							},
																						},
																						Id: "i",
																					},
																					RightBracketFodder: ast.Fodder{},
																					Id: nil,
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	TrailingComma: false,
																	TailStrict: false,
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
							ElseFodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
									Comment: []string{},
								},
							},
							BranchFalse: &ast.Error{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(272),
											Column: int(10),
										},
										End: ast.Location{
											Line: int(272),
											Column: int(89),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p2892,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
								},
								Expr: &ast.Binary{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(272),
												Column: int(17),
											},
											End: ast.Location{
												Line: int(272),
												Column: int(88),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p2892,
										FreeVars: ast.Identifiers{
											"arr",
											"std",
										},
									},
									Left: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(272),
													Column: int(17),
												},
												End: ast.Location{
													Line: int(272),
													Column: int(72),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p2892,
											FreeVars: nil,
										},
										Value: "std.flatMap second param must be array / string, got ",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									OpFodder: ast.Fodder{},
									Op: ast.BinaryOp(3),
									Right: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(272),
													Column: int(75),
												},
												End: ast.Location{
													Line: int(272),
													Column: int(88),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p2892,
											FreeVars: ast.Identifiers{
												"arr",
												"std",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(272),
														Column: int(75),
													},
													End: ast.Location{
														Line: int(272),
														Column: int(83),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p2892,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(272),
															Column: int(75),
														},
														End: ast.Location{
															Line: int(272),
															Column: int(78),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: ast.Fodder{},
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "type",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: ast.Fodder{},
											Id: nil,
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(272),
																	Column: int(84),
																},
																End: ast.Location{
																	Line: int(272),
																	Column: int(87),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p3060,
															FreeVars: ast.Identifiers{
																"arr",
															},
														},
														Id: "arr",
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
								},
							},
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(265),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(272),
					Column: int(89),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "join",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "sep",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(274),
								Column: int(8),
							},
							End: ast.Location{
								Line: int(274),
								Column: int(11),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(274),
								Column: int(13),
							},
							End: ast.Location{
								Line: int(274),
								Column: int(16),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(275),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(293),
								Column: int(83),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p3068,
						FreeVars: ast.Identifiers{
							"arr",
							"sep",
							"std",
						},
					},
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: nil,
							Variable: "aux",
							EqFodder: nil,
							Body: &ast.Function{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(275),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(285),
											Column: int(55),
										},
										File: p1,
									},
									Fodder: nil,
									Ctx: p3072,
									FreeVars: ast.Identifiers{
										"aux",
										"sep",
										"std",
									},
								},
								ParenLeftFodder: ast.Fodder{},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "arr",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(275),
												Column: int(15),
											},
											End: ast.Location{
												Line: int(275),
												Column: int(18),
											},
											File: p1,
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "i",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(275),
												Column: int(20),
											},
											End: ast.Location{
												Line: int(275),
												Column: int(21),
											},
											File: p1,
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "first",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(275),
												Column: int(23),
											},
											End: ast.Location{
												Line: int(275),
												Column: int(28),
											},
											File: p1,
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "running",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: nil,
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(275),
												Column: int(30),
											},
											End: ast.Location{
												Line: int(275),
												Column: int(37),
											},
											File: p1,
										},
									},
								},
								TrailingComma: false,
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Conditional{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(276),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(285),
												Column: int(55),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										Ctx: p3077,
										FreeVars: ast.Identifiers{
											"arr",
											"aux",
											"first",
											"i",
											"running",
											"sep",
											"std",
										},
									},
									Cond: &ast.Binary{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(276),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(276),
													Column: int(30),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p3077,
											FreeVars: ast.Identifiers{
												"arr",
												"i",
												"std",
											},
										},
										Left: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(276),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(276),
														Column: int(11),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p3077,
												FreeVars: ast.Identifiers{
													"i",
												},
											},
											Id: "i",
										},
										OpFodder: ast.Fodder{},
										Op: ast.BinaryOp(8),
										Right: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(276),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(276),
														Column: int(30),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p3077,
												FreeVars: ast.Identifiers{
													"arr",
													"std",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(276),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(276),
															Column: int(25),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p3077,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(276),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(276),
																Column: int(18),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "length",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(276),
																		Column: int(26),
																	},
																	End: ast.Location{
																		Line: int(276),
																		Column: int(29),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p3092,
																FreeVars: ast.Identifiers{
																	"arr",
																},
															},
															Id: "arr",
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
									},
									ThenFodder: ast.Fodder{},
									BranchTrue: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(277),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(277),
													Column: int(16),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
													Comment: []string{},
												},
											},
											Ctx: p3077,
											FreeVars: ast.Identifiers{
												"running",
											},
										},
										Id: "running",
									},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
											Comment: []string{},
										},
									},
									BranchFalse: &ast.Conditional{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(278),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(285),
													Column: int(55),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p3077,
											FreeVars: ast.Identifiers{
												"arr",
												"aux",
												"first",
												"i",
												"running",
												"sep",
												"std",
											},
										},
										Cond: &ast.Binary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(278),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(278),
														Column: int(29),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p3077,
												FreeVars: ast.Identifiers{
													"arr",
													"i",
												},
											},
											Left: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(278),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(278),
															Column: int(21),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p3077,
													FreeVars: ast.Identifiers{
														"arr",
														"i",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(278),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(278),
																Column: int(18),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p3077,
														FreeVars: ast.Identifiers{
															"arr",
														},
													},
													Id: "arr",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(278),
																Column: int(19),
															},
															End: ast.Location{
																Line: int(278),
																Column: int(20),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p3077,
														FreeVars: ast.Identifiers{
															"i",
														},
													},
													Id: "i",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											OpFodder: ast.Fodder{},
											Op: ast.BinaryOp(12),
											Right: &ast.LiteralNull{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(278),
															Column: int(25),
														},
														End: ast.Location{
															Line: int(278),
															Column: int(29),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p3077,
													FreeVars: nil,
												},
											},
										},
										ThenFodder: ast.Fodder{},
										BranchTrue: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(279),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(279),
														Column: int(40),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p3077,
												FreeVars: ast.Identifiers{
													"arr",
													"aux",
													"first",
													"i",
													"running",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(279),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(279),
															Column: int(12),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
															Comment: []string{},
														},
													},
													Ctx: p3077,
													FreeVars: ast.Identifiers{
														"aux",
													},
												},
												Id: "aux",
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(279),
																		Column: int(13),
																	},
																	End: ast.Location{
																		Line: int(279),
																		Column: int(16),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p3116,
																FreeVars: ast.Identifiers{
																	"arr",
																},
															},
															Id: "arr",
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Binary{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(279),
																		Column: int(18),
																	},
																	End: ast.Location{
																		Line: int(279),
																		Column: int(23),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p3116,
																FreeVars: ast.Identifiers{
																	"i",
																},
															},
															Left: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(279),
																			Column: int(18),
																		},
																		End: ast.Location{
																			Line: int(279),
																			Column: int(19),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p3116,
																	FreeVars: ast.Identifiers{
																		"i",
																	},
																},
																Id: "i",
															},
															OpFodder: ast.Fodder{},
															Op: ast.BinaryOp(3),
															Right: &ast.LiteralNumber{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(279),
																			Column: int(22),
																		},
																		End: ast.Location{
																			Line: int(279),
																			Column: int(23),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p3116,
																	FreeVars: nil,
																},
																OriginalString: "1",
															},
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(279),
																		Column: int(25),
																	},
																	End: ast.Location{
																		Line: int(279),
																		Column: int(30),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p3116,
																FreeVars: ast.Identifiers{
																	"first",
																},
															},
															Id: "first",
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(279),
																		Column: int(32),
																	},
																	End: ast.Location{
																		Line: int(279),
																		Column: int(39),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p3116,
																FreeVars: ast.Identifiers{
																	"running",
																},
															},
															Id: "running",
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: true,
											FodderRight: ast.Fodder{},
											TailStrictFodder: ast.Fodder{},
										},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										BranchFalse: &ast.Conditional{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(280),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(285),
														Column: int(55),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p3077,
												FreeVars: ast.Identifiers{
													"arr",
													"aux",
													"first",
													"i",
													"running",
													"sep",
													"std",
												},
											},
											Cond: &ast.Binary{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(280),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(280),
															Column: int(48),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p3077,
													FreeVars: ast.Identifiers{
														"arr",
														"i",
														"sep",
														"std",
													},
												},
												Left: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(280),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(280),
																Column: int(31),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p3077,
														FreeVars: ast.Identifiers{
															"arr",
															"i",
															"std",
														},
													},
													Target: &ast.Index{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(280),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(280),
																	Column: int(23),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p3077,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(280),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(280),
																		Column: int(18),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Id: "std",
														},
														LeftBracketFodder: ast.Fodder{},
														Index: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
															Value: "type",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														RightBracketFodder: ast.Fodder{},
														Id: nil,
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(280),
																				Column: int(24),
																			},
																			End: ast.Location{
																				Line: int(280),
																				Column: int(30),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p3141,
																		FreeVars: ast.Identifiers{
																			"arr",
																			"i",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(280),
																					Column: int(24),
																				},
																				End: ast.Location{
																					Line: int(280),
																					Column: int(27),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3141,
																			FreeVars: ast.Identifiers{
																				"arr",
																			},
																		},
																		Id: "arr",
																	},
																	LeftBracketFodder: ast.Fodder{},
																	Index: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(280),
																					Column: int(28),
																				},
																				End: ast.Location{
																					Line: int(280),
																					Column: int(29),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3141,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																		},
																		Id: "i",
																	},
																	RightBracketFodder: ast.Fodder{},
																	Id: nil,
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
												},
												OpFodder: ast.Fodder{},
												Op: ast.BinaryOp(13),
												Right: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(280),
																Column: int(35),
															},
															End: ast.Location{
																Line: int(280),
																Column: int(48),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p3077,
														FreeVars: ast.Identifiers{
															"sep",
															"std",
														},
													},
													Target: &ast.Index{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(280),
																	Column: int(35),
																},
																End: ast.Location{
																	Line: int(280),
																	Column: int(43),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p3077,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(280),
																		Column: int(35),
																	},
																	End: ast.Location{
																		Line: int(280),
																		Column: int(38),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Id: "std",
														},
														LeftBracketFodder: ast.Fodder{},
														Index: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
															Value: "type",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														RightBracketFodder: ast.Fodder{},
														Id: nil,
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(280),
																				Column: int(44),
																			},
																			End: ast.Location{
																				Line: int(280),
																				Column: int(47),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p3156,
																		FreeVars: ast.Identifiers{
																			"sep",
																		},
																	},
																	Id: "sep",
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
												},
											},
											ThenFodder: ast.Fodder{},
											BranchTrue: &ast.Error{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(281),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(281),
															Column: int(87),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
															Comment: []string{},
														},
													},
													Ctx: p3077,
													FreeVars: ast.Identifiers{
														"arr",
														"i",
														"sep",
														"std",
													},
												},
												Expr: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"arr",
															"i",
															"sep",
															"std",
														},
													},
													Target: &ast.Index{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Id: "std",
														},
														LeftBracketFodder: nil,
														Index: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
															Value: "mod",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														RightBracketFodder: nil,
														Id: nil,
													},
													FodderLeft: nil,
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.LiteralString{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(281),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(281),
																				Column: int(48),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p3077,
																		FreeVars: nil,
																	},
																	Value: "expected %s but arr[%d] was %s ",
																	Kind: ast.LiteralStringKind(1),
																	BlockIndent: "",
																	BlockTermIndent: "",
																},
																CommaFodder: nil,
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Array{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(281),
																				Column: int(51),
																			},
																			End: ast.Location{
																				Line: int(281),
																				Column: int(87),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p3077,
																		FreeVars: ast.Identifiers{
																			"arr",
																			"i",
																			"sep",
																			"std",
																		},
																	},
																	Elements: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Apply{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(281),
																							Column: int(52),
																						},
																						End: ast.Location{
																							Line: int(281),
																							Column: int(65),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p3174,
																					FreeVars: ast.Identifiers{
																						"sep",
																						"std",
																					},
																				},
																				Target: &ast.Index{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(281),
																								Column: int(52),
																							},
																							End: ast.Location{
																								Line: int(281),
																								Column: int(60),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p3174,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																					},
																					Target: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(281),
																									Column: int(52),
																								},
																								End: ast.Location{
																									Line: int(281),
																									Column: int(55),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: nil,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																						},
																						Id: "std",
																					},
																					LeftBracketFodder: ast.Fodder{},
																					Index: &ast.LiteralString{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								File: nil,
																							},
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: nil,
																						},
																						Value: "type",
																						Kind: ast.LiteralStringKind(1),
																						BlockIndent: "",
																						BlockTermIndent: "",
																					},
																					RightBracketFodder: ast.Fodder{},
																					Id: nil,
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(281),
																											Column: int(61),
																										},
																										End: ast.Location{
																											Line: int(281),
																											Column: int(64),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p3183,
																									FreeVars: ast.Identifiers{
																										"sep",
																									},
																								},
																								Id: "sep",
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				TrailingComma: false,
																				TailStrict: false,
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: nil,
																			},
																			CommaFodder: ast.Fodder{},
																		},
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(281),
																							Column: int(67),
																						},
																						End: ast.Location{
																							Line: int(281),
																							Column: int(68),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p3174,
																					FreeVars: ast.Identifiers{
																						"i",
																					},
																				},
																				Id: "i",
																			},
																			CommaFodder: ast.Fodder{},
																		},
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Apply{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(281),
																							Column: int(70),
																						},
																						End: ast.Location{
																							Line: int(281),
																							Column: int(86),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p3174,
																					FreeVars: ast.Identifiers{
																						"arr",
																						"i",
																						"std",
																					},
																				},
																				Target: &ast.Index{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(281),
																								Column: int(70),
																							},
																							End: ast.Location{
																								Line: int(281),
																								Column: int(78),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p3174,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																					},
																					Target: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(281),
																									Column: int(70),
																								},
																								End: ast.Location{
																									Line: int(281),
																									Column: int(73),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: nil,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																						},
																						Id: "std",
																					},
																					LeftBracketFodder: ast.Fodder{},
																					Index: &ast.LiteralString{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								File: nil,
																							},
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: nil,
																						},
																						Value: "type",
																						Kind: ast.LiteralStringKind(1),
																						BlockIndent: "",
																						BlockTermIndent: "",
																					},
																					RightBracketFodder: ast.Fodder{},
																					Id: nil,
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Index{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(281),
																											Column: int(79),
																										},
																										End: ast.Location{
																											Line: int(281),
																											Column: int(85),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p3196,
																									FreeVars: ast.Identifiers{
																										"arr",
																										"i",
																									},
																								},
																								Target: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(281),
																												Column: int(79),
																											},
																											End: ast.Location{
																												Line: int(281),
																												Column: int(82),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p3196,
																										FreeVars: ast.Identifiers{
																											"arr",
																										},
																									},
																									Id: "arr",
																								},
																								LeftBracketFodder: ast.Fodder{},
																								Index: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(281),
																												Column: int(83),
																											},
																											End: ast.Location{
																												Line: int(281),
																												Column: int(84),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p3196,
																										FreeVars: ast.Identifiers{
																											"i",
																										},
																									},
																									Id: "i",
																								},
																								RightBracketFodder: ast.Fodder{},
																								Id: nil,
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				TrailingComma: false,
																				TailStrict: false,
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: nil,
																			},
																			CommaFodder: nil,
																		},
																	},
																	TrailingComma: false,
																	CloseFodder: ast.Fodder{},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: nil,
													TailStrictFodder: nil,
												},
											},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
													Comment: []string{},
												},
											},
											BranchFalse: &ast.Conditional{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(282),
															Column: int(12),
														},
														End: ast.Location{
															Line: int(285),
															Column: int(55),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p3077,
													FreeVars: ast.Identifiers{
														"arr",
														"aux",
														"first",
														"i",
														"running",
														"sep",
													},
												},
												Cond: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(282),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(282),
																Column: int(20),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p3077,
														FreeVars: ast.Identifiers{
															"first",
														},
													},
													Id: "first",
												},
												ThenFodder: ast.Fodder{},
												BranchTrue: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(283),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(283),
																Column: int(49),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p3077,
														FreeVars: ast.Identifiers{
															"arr",
															"aux",
															"i",
															"running",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(283),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(283),
																	Column: int(12),
																},
																File: p1,
															},
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																	Comment: []string{},
																},
															},
															Ctx: p3077,
															FreeVars: ast.Identifiers{
																"aux",
															},
														},
														Id: "aux",
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(283),
																				Column: int(13),
																			},
																			End: ast.Location{
																				Line: int(283),
																				Column: int(16),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p3214,
																		FreeVars: ast.Identifiers{
																			"arr",
																		},
																	},
																	Id: "arr",
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(283),
																				Column: int(18),
																			},
																			End: ast.Location{
																				Line: int(283),
																				Column: int(23),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p3214,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																	},
																	Left: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(283),
																					Column: int(18),
																				},
																				End: ast.Location{
																					Line: int(283),
																					Column: int(19),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3214,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																		},
																		Id: "i",
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(3),
																	Right: &ast.LiteralNumber{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(283),
																					Column: int(22),
																				},
																				End: ast.Location{
																					Line: int(283),
																					Column: int(23),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3214,
																			FreeVars: nil,
																		},
																		OriginalString: "1",
																	},
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.LiteralBoolean{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(283),
																				Column: int(25),
																			},
																			End: ast.Location{
																				Line: int(283),
																				Column: int(30),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p3214,
																		FreeVars: nil,
																	},
																	Value: false,
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(283),
																				Column: int(32),
																			},
																			End: ast.Location{
																				Line: int(283),
																				Column: int(48),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p3214,
																		FreeVars: ast.Identifiers{
																			"arr",
																			"i",
																			"running",
																		},
																	},
																	Left: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(283),
																					Column: int(32),
																				},
																				End: ast.Location{
																					Line: int(283),
																					Column: int(39),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3214,
																			FreeVars: ast.Identifiers{
																				"running",
																			},
																		},
																		Id: "running",
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(3),
																	Right: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(283),
																					Column: int(42),
																				},
																				End: ast.Location{
																					Line: int(283),
																					Column: int(48),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3214,
																			FreeVars: ast.Identifiers{
																				"arr",
																				"i",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(283),
																						Column: int(42),
																					},
																					End: ast.Location{
																						Line: int(283),
																						Column: int(45),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p3214,
																				FreeVars: ast.Identifiers{
																					"arr",
																				},
																			},
																			Id: "arr",
																		},
																		LeftBracketFodder: ast.Fodder{},
																		Index: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(283),
																						Column: int(46),
																					},
																					End: ast.Location{
																						Line: int(283),
																						Column: int(47),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p3214,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																			},
																			Id: "i",
																		},
																		RightBracketFodder: ast.Fodder{},
																		Id: nil,
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: true,
													FodderRight: ast.Fodder{},
													TailStrictFodder: ast.Fodder{},
												},
												ElseFodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
														Comment: []string{},
													},
												},
												BranchFalse: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(285),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(285),
																Column: int(55),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p3077,
														FreeVars: ast.Identifiers{
															"arr",
															"aux",
															"i",
															"running",
															"sep",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(285),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(285),
																	Column: int(12),
																},
																File: p1,
															},
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																	Comment: []string{},
																},
															},
															Ctx: p3077,
															FreeVars: ast.Identifiers{
																"aux",
															},
														},
														Id: "aux",
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(285),
																				Column: int(13),
																			},
																			End: ast.Location{
																				Line: int(285),
																				Column: int(16),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p3240,
																		FreeVars: ast.Identifiers{
																			"arr",
																		},
																	},
																	Id: "arr",
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(285),
																				Column: int(18),
																			},
																			End: ast.Location{
																				Line: int(285),
																				Column: int(23),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p3240,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																	},
																	Left: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(285),
																					Column: int(18),
																				},
																				End: ast.Location{
																					Line: int(285),
																					Column: int(19),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3240,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																		},
																		Id: "i",
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(3),
																	Right: &ast.LiteralNumber{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(285),
																					Column: int(22),
																				},
																				End: ast.Location{
																					Line: int(285),
																					Column: int(23),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3240,
																			FreeVars: nil,
																		},
																		OriginalString: "1",
																	},
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.LiteralBoolean{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(285),
																				Column: int(25),
																			},
																			End: ast.Location{
																				Line: int(285),
																				Column: int(30),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p3240,
																		FreeVars: nil,
																	},
																	Value: false,
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(285),
																				Column: int(32),
																			},
																			End: ast.Location{
																				Line: int(285),
																				Column: int(54),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p3240,
																		FreeVars: ast.Identifiers{
																			"arr",
																			"i",
																			"running",
																			"sep",
																		},
																	},
																	Left: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(285),
																					Column: int(32),
																				},
																				End: ast.Location{
																					Line: int(285),
																					Column: int(45),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3240,
																			FreeVars: ast.Identifiers{
																				"running",
																				"sep",
																			},
																		},
																		Left: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(285),
																						Column: int(32),
																					},
																					End: ast.Location{
																						Line: int(285),
																						Column: int(39),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p3240,
																				FreeVars: ast.Identifiers{
																					"running",
																				},
																			},
																			Id: "running",
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(3),
																		Right: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(285),
																						Column: int(42),
																					},
																					End: ast.Location{
																						Line: int(285),
																						Column: int(45),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p3240,
																				FreeVars: ast.Identifiers{
																					"sep",
																				},
																			},
																			Id: "sep",
																		},
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(3),
																	Right: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(285),
																					Column: int(48),
																				},
																				End: ast.Location{
																					Line: int(285),
																					Column: int(54),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3240,
																			FreeVars: ast.Identifiers{
																				"arr",
																				"i",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(285),
																						Column: int(48),
																					},
																					End: ast.Location{
																						Line: int(285),
																						Column: int(51),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p3240,
																				FreeVars: ast.Identifiers{
																					"arr",
																				},
																			},
																			Id: "arr",
																		},
																		LeftBracketFodder: ast.Fodder{},
																		Index: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(285),
																						Column: int(52),
																					},
																					End: ast.Location{
																						Line: int(285),
																						Column: int(53),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p3240,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																			},
																			Id: "i",
																		},
																		RightBracketFodder: ast.Fodder{},
																		Id: nil,
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: true,
													FodderRight: ast.Fodder{},
													TailStrictFodder: ast.Fodder{},
												},
											},
										},
									},
								},
							},
							Fun: nil,
							CloseFodder: nil,
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
						},
					},
					Body: &ast.Conditional{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(286),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(293),
									Column: int(83),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
									Comment: []string{},
								},
							},
							Ctx: p3068,
							FreeVars: ast.Identifiers{
								"arr",
								"aux",
								"sep",
								"std",
							},
						},
						Cond: &ast.Unary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(286),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(286),
										Column: int(25),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p3068,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
								},
							},
							Op: ast.UnaryOp(0),
							Expr: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(286),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(286),
											Column: int(25),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p3068,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(286),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(286),
												Column: int(20),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p3068,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(286),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(286),
													Column: int(12),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "isArray",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(286),
															Column: int(21),
														},
														End: ast.Location{
															Line: int(286),
															Column: int(24),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p3276,
													FreeVars: ast.Identifiers{
														"arr",
													},
												},
												Id: "arr",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
						},
						ThenFodder: ast.Fodder{},
						BranchTrue: &ast.Error{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(287),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(287),
										Column: int(74),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
										Comment: []string{},
									},
								},
								Ctx: p3068,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
								},
							},
							Expr: &ast.Binary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(287),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(287),
											Column: int(74),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p3068,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
								},
								Left: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(287),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(287),
												Column: int(58),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p3068,
										FreeVars: nil,
									},
									Value: "join second parameter should be array, got ",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								OpFodder: ast.Fodder{},
								Op: ast.BinaryOp(3),
								Right: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(287),
												Column: int(61),
											},
											End: ast.Location{
												Line: int(287),
												Column: int(74),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p3068,
										FreeVars: ast.Identifiers{
											"arr",
											"std",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(287),
													Column: int(61),
												},
												End: ast.Location{
													Line: int(287),
													Column: int(69),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p3068,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(287),
														Column: int(61),
													},
													End: ast.Location{
														Line: int(287),
														Column: int(64),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "type",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(287),
																Column: int(70),
															},
															End: ast.Location{
																Line: int(287),
																Column: int(73),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p3293,
														FreeVars: ast.Identifiers{
															"arr",
														},
													},
													Id: "arr",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
							},
						},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						BranchFalse: &ast.Conditional{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(288),
										Column: int(10),
									},
									End: ast.Location{
										Line: int(293),
										Column: int(83),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p3068,
								FreeVars: ast.Identifiers{
									"arr",
									"aux",
									"sep",
									"std",
								},
							},
							Cond: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(288),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(288),
											Column: int(30),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p3068,
									FreeVars: ast.Identifiers{
										"sep",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(288),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(288),
												Column: int(25),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p3068,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(288),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(288),
													Column: int(16),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "isString",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(288),
															Column: int(26),
														},
														End: ast.Location{
															Line: int(288),
															Column: int(29),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p3307,
													FreeVars: ast.Identifiers{
														"sep",
													},
												},
												Id: "sep",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
							ThenFodder: ast.Fodder{},
							BranchTrue: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(289),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(289),
											Column: int(28),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p3068,
									FreeVars: ast.Identifiers{
										"arr",
										"aux",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(289),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(289),
												Column: int(10),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										Ctx: p3068,
										FreeVars: ast.Identifiers{
											"aux",
										},
									},
									Id: "aux",
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(289),
															Column: int(11),
														},
														End: ast.Location{
															Line: int(289),
															Column: int(14),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p3316,
													FreeVars: ast.Identifiers{
														"arr",
													},
												},
												Id: "arr",
											},
											CommaFodder: ast.Fodder{},
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.LiteralNumber{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(289),
															Column: int(16),
														},
														End: ast.Location{
															Line: int(289),
															Column: int(17),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p3316,
													FreeVars: nil,
												},
												OriginalString: "0",
											},
											CommaFodder: ast.Fodder{},
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.LiteralBoolean{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(289),
															Column: int(19),
														},
														End: ast.Location{
															Line: int(289),
															Column: int(23),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p3316,
													FreeVars: nil,
												},
												Value: true,
											},
											CommaFodder: ast.Fodder{},
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(289),
															Column: int(25),
														},
														End: ast.Location{
															Line: int(289),
															Column: int(27),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p3316,
													FreeVars: nil,
												},
												Value: "",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
							ElseFodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
									Comment: []string{},
								},
							},
							BranchFalse: &ast.Conditional{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(290),
											Column: int(10),
										},
										End: ast.Location{
											Line: int(293),
											Column: int(83),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p3068,
									FreeVars: ast.Identifiers{
										"arr",
										"aux",
										"sep",
										"std",
									},
								},
								Cond: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(290),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(290),
												Column: int(29),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p3068,
										FreeVars: ast.Identifiers{
											"sep",
											"std",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(290),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(290),
													Column: int(24),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p3068,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(290),
														Column: int(13),
													},
													End: ast.Location{
														Line: int(290),
														Column: int(16),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "isArray",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(290),
																Column: int(25),
															},
															End: ast.Location{
																Line: int(290),
																Column: int(28),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p3333,
														FreeVars: ast.Identifiers{
															"sep",
														},
													},
													Id: "sep",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
								ThenFodder: ast.Fodder{},
								BranchTrue: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(291),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(291),
												Column: int(28),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p3068,
										FreeVars: ast.Identifiers{
											"arr",
											"aux",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(291),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(291),
													Column: int(10),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
													Comment: []string{},
												},
											},
											Ctx: p3068,
											FreeVars: ast.Identifiers{
												"aux",
											},
										},
										Id: "aux",
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(291),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(291),
																Column: int(14),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p3342,
														FreeVars: ast.Identifiers{
															"arr",
														},
													},
													Id: "arr",
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.LiteralNumber{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(291),
																Column: int(16),
															},
															End: ast.Location{
																Line: int(291),
																Column: int(17),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p3342,
														FreeVars: nil,
													},
													OriginalString: "0",
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.LiteralBoolean{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(291),
																Column: int(19),
															},
															End: ast.Location{
																Line: int(291),
																Column: int(23),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p3342,
														FreeVars: nil,
													},
													Value: true,
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.Array{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(291),
																Column: int(25),
															},
															End: ast.Location{
																Line: int(291),
																Column: int(27),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p3342,
														FreeVars: nil,
													},
													Elements: nil,
													TrailingComma: false,
													CloseFodder: ast.Fodder{},
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
								ElseFodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
										Comment: []string{},
									},
								},
								BranchFalse: &ast.Error{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(293),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(293),
												Column: int(83),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										Ctx: p3068,
										FreeVars: ast.Identifiers{
											"sep",
											"std",
										},
									},
									Expr: &ast.Binary{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(293),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(293),
													Column: int(83),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p3068,
											FreeVars: ast.Identifiers{
												"sep",
												"std",
											},
										},
										Left: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(293),
														Column: int(13),
													},
													End: ast.Location{
														Line: int(293),
														Column: int(67),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p3068,
												FreeVars: nil,
											},
											Value: "join first parameter should be string or array, got ",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										OpFodder: ast.Fodder{},
										Op: ast.BinaryOp(3),
										Right: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(293),
														Column: int(70),
													},
													End: ast.Location{
														Line: int(293),
														Column: int(83),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p3068,
												FreeVars: ast.Identifiers{
													"sep",
													"std",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(293),
															Column: int(70),
														},
														End: ast.Location{
															Line: int(293),
															Column: int(78),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p3068,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(293),
																Column: int(70),
															},
															End: ast.Location{
																Line: int(293),
																Column: int(73),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "type",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(293),
																		Column: int(79),
																	},
																	End: ast.Location{
																		Line: int(293),
																		Column: int(82),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p3363,
																FreeVars: ast.Identifiers{
																	"sep",
																},
															},
															Id: "sep",
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
									},
								},
							},
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(274),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(293),
					Column: int(83),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "lines",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(295),
								Column: int(9),
							},
							End: ast.Location{
								Line: int(295),
								Column: int(12),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(296),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(296),
								Column: int(31),
							},
							File: p1,
						},
						Fodder: ast.Fodder{},
						Ctx: p3370,
						FreeVars: ast.Identifiers{
							"arr",
							"std",
						},
					},
					Target: &ast.Index{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(296),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(296),
									Column: int(13),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p3370,
							FreeVars: ast.Identifiers{
								"std",
							},
						},
						Target: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(296),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(296),
										Column: int(8),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
										Comment: []string{},
									},
								},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Id: "std",
						},
						LeftBracketFodder: ast.Fodder{},
						Index: &ast.LiteralString{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: nil,
							},
							Value: "join",
							Kind: ast.LiteralStringKind(1),
							BlockIndent: "",
							BlockTermIndent: "",
						},
						RightBracketFodder: ast.Fodder{},
						Id: nil,
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(296),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(296),
												Column: int(18),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p3380,
										FreeVars: nil,
									},
									Value: "\n",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.Binary{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(296),
												Column: int(20),
											},
											End: ast.Location{
												Line: int(296),
												Column: int(30),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p3380,
										FreeVars: ast.Identifiers{
											"arr",
										},
									},
									Left: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(296),
													Column: int(20),
												},
												End: ast.Location{
													Line: int(296),
													Column: int(23),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p3380,
											FreeVars: ast.Identifiers{
												"arr",
											},
										},
										Id: "arr",
									},
									OpFodder: ast.Fodder{},
									Op: ast.BinaryOp(3),
									Right: &ast.Array{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(296),
													Column: int(26),
												},
												End: ast.Location{
													Line: int(296),
													Column: int(30),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p3380,
											FreeVars: nil,
										},
										Elements: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(296),
																Column: int(27),
															},
															End: ast.Location{
																Line: int(296),
																Column: int(29),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p3388,
														FreeVars: nil,
													},
													Value: "",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												CommaFodder: nil,
											},
										},
										TrailingComma: false,
										CloseFodder: ast.Fodder{},
									},
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					TrailingComma: false,
					TailStrict: false,
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(295),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(296),
					Column: int(31),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "deepJoin",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(298),
								Column: int(12),
							},
							End: ast.Location{
								Line: int(298),
								Column: int(15),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(299),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(304),
								Column: int(63),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p3395,
						FreeVars: ast.Identifiers{
							"arr",
							"std",
						},
					},
					Cond: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(299),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(299),
									Column: int(25),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p3395,
							FreeVars: ast.Identifiers{
								"arr",
								"std",
							},
						},
						Target: &ast.Index{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(299),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(299),
										Column: int(20),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p3395,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(299),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(299),
											Column: int(11),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Id: "std",
							},
							LeftBracketFodder: ast.Fodder{},
							Index: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Value: "isString",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							RightBracketFodder: ast.Fodder{},
							Id: nil,
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(299),
													Column: int(21),
												},
												End: ast.Location{
													Line: int(299),
													Column: int(24),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p3406,
											FreeVars: ast.Identifiers{
												"arr",
											},
										},
										Id: "arr",
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
					ThenFodder: ast.Fodder{},
					BranchTrue: &ast.Var{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(300),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(300),
									Column: int(10),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
									Comment: []string{},
								},
							},
							Ctx: p3395,
							FreeVars: ast.Identifiers{
								"arr",
							},
						},
						Id: "arr",
					},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
							Comment: []string{},
						},
					},
					BranchFalse: &ast.Conditional{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(301),
									Column: int(10),
								},
								End: ast.Location{
									Line: int(304),
									Column: int(63),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p3395,
							FreeVars: ast.Identifiers{
								"arr",
								"std",
							},
						},
						Cond: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(301),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(301),
										Column: int(29),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p3395,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(301),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(301),
											Column: int(24),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p3395,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(301),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(301),
												Column: int(16),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "isArray",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(301),
														Column: int(25),
													},
													End: ast.Location{
														Line: int(301),
														Column: int(28),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p3423,
												FreeVars: ast.Identifiers{
													"arr",
												},
											},
											Id: "arr",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
						ThenFodder: ast.Fodder{},
						BranchTrue: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(302),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(302),
										Column: int(51),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p3395,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(302),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(302),
											Column: int(15),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p3395,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(302),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(302),
												Column: int(10),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "join",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(302),
														Column: int(16),
													},
													End: ast.Location{
														Line: int(302),
														Column: int(18),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p3435,
												FreeVars: nil,
											},
											Value: "",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"arr",
													"std",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: nil,
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "flatMap",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: nil,
												Id: nil,
											},
											FodderLeft: nil,
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Function{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															ParenLeftFodder: nil,
															Parameters: []ast.Parameter{
																ast.Parameter{
																	NameFodder: nil,
																	Name: "x",
																	EqFodder: nil,
																	DefaultArg: nil,
																	CommaFodder: nil,
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																},
															},
															TrailingComma: false,
															ParenRightFodder: nil,
															Body: &ast.Array{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																		"x",
																	},
																},
																Elements: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Apply{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(302),
																						Column: int(21),
																					},
																					End: ast.Location{
																						Line: int(302),
																						Column: int(36),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p3451,
																				FreeVars: ast.Identifiers{
																					"std",
																					"x",
																				},
																			},
																			Target: &ast.Index{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(302),
																							Column: int(21),
																						},
																						End: ast.Location{
																							Line: int(302),
																							Column: int(33),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p3451,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																				},
																				Target: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(302),
																								Column: int(21),
																							},
																							End: ast.Location{
																								Line: int(302),
																								Column: int(24),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: nil,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																					},
																					Id: "std",
																				},
																				LeftBracketFodder: ast.Fodder{},
																				Index: &ast.LiteralString{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							File: nil,
																						},
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: nil,
																					},
																					Value: "deepJoin",
																					Kind: ast.LiteralStringKind(1),
																					BlockIndent: "",
																					BlockTermIndent: "",
																				},
																				RightBracketFodder: ast.Fodder{},
																				Id: nil,
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(302),
																										Column: int(34),
																									},
																									End: ast.Location{
																										Line: int(302),
																										Column: int(35),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p3460,
																								FreeVars: ast.Identifiers{
																									"x",
																								},
																							},
																							Id: "x",
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			TrailingComma: false,
																			TailStrict: false,
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																		},
																		CommaFodder: nil,
																	},
																},
																TrailingComma: false,
																CloseFodder: nil,
															},
														},
														CommaFodder: nil,
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(302),
																		Column: int(46),
																	},
																	End: ast.Location{
																		Line: int(302),
																		Column: int(49),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p3435,
																FreeVars: ast.Identifiers{
																	"arr",
																},
															},
															Id: "arr",
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: nil,
											TailStrictFodder: nil,
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						BranchFalse: &ast.Error{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(304),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(304),
										Column: int(63),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
										Comment: []string{},
									},
								},
								Ctx: p3395,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
								},
							},
							Expr: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: nil,
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "mod",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: nil,
									Id: nil,
								},
								FodderLeft: nil,
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(304),
															Column: int(13),
														},
														End: ast.Location{
															Line: int(304),
															Column: int(47),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p3395,
													FreeVars: nil,
												},
												Value: "Expected string or array, got %s",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											CommaFodder: nil,
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(304),
															Column: int(50),
														},
														End: ast.Location{
															Line: int(304),
															Column: int(63),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p3395,
													FreeVars: ast.Identifiers{
														"arr",
														"std",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(304),
																Column: int(50),
															},
															End: ast.Location{
																Line: int(304),
																Column: int(58),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p3395,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(304),
																	Column: int(50),
																},
																End: ast.Location{
																	Line: int(304),
																	Column: int(53),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: ast.Fodder{},
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "type",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: ast.Fodder{},
													Id: nil,
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(304),
																			Column: int(59),
																		},
																		End: ast.Location{
																			Line: int(304),
																			Column: int(62),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p3486,
																	FreeVars: ast.Identifiers{
																		"arr",
																	},
																},
																Id: "arr",
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: nil,
								TailStrictFodder: nil,
							},
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(298),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(304),
					Column: int(63),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "format",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(307),
								Column: int(10),
							},
							End: ast.Location{
								Line: int(307),
								Column: int(13),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "vals",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(307),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(307),
								Column: int(19),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(313),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(774),
								Column: int(48),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(1),
								Indent: int(4),
								Comment: []string{},
							},
							ast.FodderElement{
								Kind: ast.FodderKind(2),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{
									"/////////////////////////////",
								},
							},
							ast.FodderElement{
								Kind: ast.FodderKind(2),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{
									"// Parse the mini-language //",
								},
							},
							ast.FodderElement{
								Kind: ast.FodderKind(2),
								Blanks: int(1),
								Indent: int(4),
								Comment: []string{
									"/////////////////////////////",
								},
							},
						},
						Ctx: p3497,
						FreeVars: ast.Identifiers{
							"std",
							"str",
							"vals",
						},
					},
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: nil,
							Variable: "try_parse_mapping_key",
							EqFodder: nil,
							Body: &ast.Function{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(313),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(328),
											Column: int(26),
										},
										File: p1,
									},
									Fodder: nil,
									Ctx: p3501,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								ParenLeftFodder: ast.Fodder{},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "str",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(313),
												Column: int(33),
											},
											End: ast.Location{
												Line: int(313),
												Column: int(36),
											},
											File: p1,
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "i",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: nil,
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(313),
												Column: int(38),
											},
											End: ast.Location{
												Line: int(313),
												Column: int(39),
											},
											File: p1,
										},
									},
								},
								TrailingComma: false,
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Conditional{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"i",
											"std",
											"str",
										},
									},
									Cond: &ast.Binary{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(314),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(314),
													Column: int(33),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p3507,
											FreeVars: ast.Identifiers{
												"i",
												"std",
												"str",
											},
										},
										Left: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(314),
														Column: int(14),
													},
													End: ast.Location{
														Line: int(314),
														Column: int(15),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p3507,
												FreeVars: ast.Identifiers{
													"i",
												},
											},
											Id: "i",
										},
										OpFodder: ast.Fodder{},
										Op: ast.BinaryOp(9),
										Right: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(314),
														Column: int(18),
													},
													End: ast.Location{
														Line: int(314),
														Column: int(33),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p3507,
												FreeVars: ast.Identifiers{
													"std",
													"str",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(314),
															Column: int(18),
														},
														End: ast.Location{
															Line: int(314),
															Column: int(28),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p3507,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(314),
																Column: int(18),
															},
															End: ast.Location{
																Line: int(314),
																Column: int(21),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "length",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(314),
																		Column: int(29),
																	},
																	End: ast.Location{
																		Line: int(314),
																		Column: int(32),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p3520,
																FreeVars: ast.Identifiers{
																	"str",
																},
															},
															Id: "str",
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
									},
									ThenFodder: nil,
									BranchTrue: &ast.Local{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(315),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(328),
													Column: int(26),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
													Comment: []string{},
												},
											},
											Ctx: p3507,
											FreeVars: ast.Identifiers{
												"i",
												"std",
												"str",
											},
										},
										Binds: ast.LocalBinds{
											ast.LocalBind{
												VarFodder: ast.Fodder{},
												Variable: "c",
												EqFodder: ast.Fodder{},
												Body: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(315),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(315),
																Column: int(23),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p3527,
														FreeVars: ast.Identifiers{
															"i",
															"str",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(315),
																	Column: int(17),
																},
																End: ast.Location{
																	Line: int(315),
																	Column: int(20),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p3527,
															FreeVars: ast.Identifiers{
																"str",
															},
														},
														Id: "str",
													},
													LeftBracketFodder: ast.Fodder{},
													Index: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(315),
																	Column: int(21),
																},
																End: ast.Location{
																	Line: int(315),
																	Column: int(22),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p3527,
															FreeVars: ast.Identifiers{
																"i",
															},
														},
														Id: "i",
													},
													RightBracketFodder: ast.Fodder{},
													Id: nil,
												},
												Fun: nil,
												CloseFodder: ast.Fodder{},
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(315),
														Column: int(13),
													},
													End: ast.Location{
														Line: int(315),
														Column: int(23),
													},
													File: p1,
												},
											},
										},
										Body: &ast.Conditional{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(316),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(328),
														Column: int(26),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
														Comment: []string{},
													},
												},
												Ctx: p3507,
												FreeVars: ast.Identifiers{
													"c",
													"i",
													"std",
													"str",
												},
											},
											Cond: &ast.Binary{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(316),
															Column: int(10),
														},
														End: ast.Location{
															Line: int(316),
															Column: int(18),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p3507,
													FreeVars: ast.Identifiers{
														"c",
													},
												},
												Left: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(316),
																Column: int(10),
															},
															End: ast.Location{
																Line: int(316),
																Column: int(11),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p3507,
														FreeVars: ast.Identifiers{
															"c",
														},
													},
													Id: "c",
												},
												OpFodder: ast.Fodder{},
												Op: ast.BinaryOp(12),
												Right: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(316),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(316),
																Column: int(18),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p3507,
														FreeVars: nil,
													},
													Value: "(",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
											},
											ThenFodder: ast.Fodder{},
											BranchTrue: &ast.Local{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(317),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(326),
															Column: int(32),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
															Comment: []string{},
														},
													},
													Ctx: p3507,
													FreeVars: ast.Identifiers{
														"i",
														"std",
														"str",
													},
												},
												Binds: ast.LocalBinds{
													ast.LocalBind{
														VarFodder: nil,
														Variable: "consume",
														EqFodder: nil,
														Body: &ast.Function{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(317),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(325),
																		Column: int(33),
																	},
																	File: p1,
																},
																Fodder: nil,
																Ctx: p3546,
																FreeVars: ast.Identifiers{
																	"consume",
																	"std",
																},
															},
															ParenLeftFodder: ast.Fodder{},
															Parameters: []ast.Parameter{
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "str",
																	EqFodder: nil,
																	DefaultArg: nil,
																	CommaFodder: ast.Fodder{},
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(317),
																			Column: int(23),
																		},
																		End: ast.Location{
																			Line: int(317),
																			Column: int(26),
																		},
																		File: p1,
																	},
																},
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "j",
																	EqFodder: nil,
																	DefaultArg: nil,
																	CommaFodder: ast.Fodder{},
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(317),
																			Column: int(28),
																		},
																		End: ast.Location{
																			Line: int(317),
																			Column: int(29),
																		},
																		File: p1,
																	},
																},
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "v",
																	EqFodder: nil,
																	DefaultArg: nil,
																	CommaFodder: nil,
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(317),
																			Column: int(31),
																		},
																		End: ast.Location{
																			Line: int(317),
																			Column: int(32),
																		},
																		File: p1,
																	},
																},
															},
															TrailingComma: false,
															ParenRightFodder: ast.Fodder{},
															Body: &ast.Conditional{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(318),
																			Column: int(11),
																		},
																		End: ast.Location{
																			Line: int(325),
																			Column: int(33),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(10),
																			Comment: []string{},
																		},
																	},
																	Ctx: p3551,
																	FreeVars: ast.Identifiers{
																		"consume",
																		"j",
																		"std",
																		"str",
																		"v",
																	},
																},
																Cond: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(318),
																				Column: int(14),
																			},
																			End: ast.Location{
																				Line: int(318),
																				Column: int(34),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p3551,
																		FreeVars: ast.Identifiers{
																			"j",
																			"std",
																			"str",
																		},
																	},
																	Left: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(318),
																					Column: int(14),
																				},
																				End: ast.Location{
																					Line: int(318),
																					Column: int(15),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3551,
																			FreeVars: ast.Identifiers{
																				"j",
																			},
																		},
																		Id: "j",
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(8),
																	Right: &ast.Apply{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(318),
																					Column: int(19),
																				},
																				End: ast.Location{
																					Line: int(318),
																					Column: int(34),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3551,
																			FreeVars: ast.Identifiers{
																				"std",
																				"str",
																			},
																		},
																		Target: &ast.Index{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(318),
																						Column: int(19),
																					},
																					End: ast.Location{
																						Line: int(318),
																						Column: int(29),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p3551,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(318),
																							Column: int(19),
																						},
																						End: ast.Location{
																							Line: int(318),
																							Column: int(22),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																				},
																				Id: "std",
																			},
																			LeftBracketFodder: ast.Fodder{},
																			Index: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: nil,
																				},
																				Value: "length",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																			RightBracketFodder: ast.Fodder{},
																			Id: nil,
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(318),
																									Column: int(30),
																								},
																								End: ast.Location{
																									Line: int(318),
																									Column: int(33),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p3566,
																							FreeVars: ast.Identifiers{
																								"str",
																							},
																						},
																						Id: "str",
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		TrailingComma: false,
																		TailStrict: false,
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																	},
																},
																ThenFodder: ast.Fodder{},
																BranchTrue: &ast.Error{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(319),
																				Column: int(13),
																			},
																			End: ast.Location{
																				Line: int(319),
																				Column: int(43),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(12),
																				Comment: []string{},
																			},
																		},
																		Ctx: p3551,
																		FreeVars: nil,
																	},
																	Expr: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(319),
																					Column: int(19),
																				},
																				End: ast.Location{
																					Line: int(319),
																					Column: int(43),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3551,
																			FreeVars: nil,
																		},
																		Value: "Truncated format code.",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																},
																ElseFodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(10),
																		Comment: []string{},
																	},
																},
																BranchFalse: &ast.Local{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(321),
																				Column: int(13),
																			},
																			End: ast.Location{
																				Line: int(325),
																				Column: int(33),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(12),
																				Comment: []string{},
																			},
																		},
																		Ctx: p3551,
																		FreeVars: ast.Identifiers{
																			"consume",
																			"j",
																			"str",
																			"v",
																		},
																	},
																	Binds: ast.LocalBinds{
																		ast.LocalBind{
																			VarFodder: ast.Fodder{},
																			Variable: "c",
																			EqFodder: ast.Fodder{},
																			Body: &ast.Index{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(321),
																							Column: int(23),
																						},
																						End: ast.Location{
																							Line: int(321),
																							Column: int(29),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p3577,
																					FreeVars: ast.Identifiers{
																						"j",
																						"str",
																					},
																				},
																				Target: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(321),
																								Column: int(23),
																							},
																							End: ast.Location{
																								Line: int(321),
																								Column: int(26),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p3577,
																						FreeVars: ast.Identifiers{
																							"str",
																						},
																					},
																					Id: "str",
																				},
																				LeftBracketFodder: ast.Fodder{},
																				Index: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(321),
																								Column: int(27),
																							},
																							End: ast.Location{
																								Line: int(321),
																								Column: int(28),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p3577,
																						FreeVars: ast.Identifiers{
																							"j",
																						},
																					},
																					Id: "j",
																				},
																				RightBracketFodder: ast.Fodder{},
																				Id: nil,
																			},
																			Fun: nil,
																			CloseFodder: ast.Fodder{},
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(321),
																					Column: int(19),
																				},
																				End: ast.Location{
																					Line: int(321),
																					Column: int(29),
																				},
																				File: p1,
																			},
																		},
																	},
																	Body: &ast.Conditional{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(322),
																					Column: int(13),
																				},
																				End: ast.Location{
																					Line: int(325),
																					Column: int(33),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(12),
																					Comment: []string{},
																				},
																			},
																			Ctx: p3551,
																			FreeVars: ast.Identifiers{
																				"c",
																				"consume",
																				"j",
																				"str",
																				"v",
																			},
																		},
																		Cond: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(322),
																						Column: int(16),
																					},
																					End: ast.Location{
																						Line: int(322),
																						Column: int(24),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p3551,
																				FreeVars: ast.Identifiers{
																					"c",
																				},
																			},
																			Left: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(322),
																							Column: int(16),
																						},
																						End: ast.Location{
																							Line: int(322),
																							Column: int(17),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p3551,
																					FreeVars: ast.Identifiers{
																						"c",
																					},
																				},
																				Id: "c",
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(13),
																			Right: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(322),
																							Column: int(21),
																						},
																						End: ast.Location{
																							Line: int(322),
																							Column: int(24),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p3551,
																					FreeVars: nil,
																				},
																				Value: ")",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																		},
																		ThenFodder: ast.Fodder{},
																		BranchTrue: &ast.Apply{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(323),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(323),
																						Column: int(41),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p3551,
																				FreeVars: ast.Identifiers{
																					"c",
																					"consume",
																					"j",
																					"str",
																					"v",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(323),
																							Column: int(15),
																						},
																						End: ast.Location{
																							Line: int(323),
																							Column: int(22),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(14),
																							Comment: []string{},
																						},
																					},
																					Ctx: p3551,
																					FreeVars: ast.Identifiers{
																						"consume",
																					},
																				},
																				Id: "consume",
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(323),
																										Column: int(23),
																									},
																									End: ast.Location{
																										Line: int(323),
																										Column: int(26),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p3598,
																								FreeVars: ast.Identifiers{
																									"str",
																								},
																							},
																							Id: "str",
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Binary{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(323),
																										Column: int(28),
																									},
																									End: ast.Location{
																										Line: int(323),
																										Column: int(33),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p3598,
																								FreeVars: ast.Identifiers{
																									"j",
																								},
																							},
																							Left: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(323),
																											Column: int(28),
																										},
																										End: ast.Location{
																											Line: int(323),
																											Column: int(29),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p3598,
																									FreeVars: ast.Identifiers{
																										"j",
																									},
																								},
																								Id: "j",
																							},
																							OpFodder: ast.Fodder{},
																							Op: ast.BinaryOp(3),
																							Right: &ast.LiteralNumber{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(323),
																											Column: int(32),
																										},
																										End: ast.Location{
																											Line: int(323),
																											Column: int(33),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p3598,
																									FreeVars: nil,
																								},
																								OriginalString: "1",
																							},
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Binary{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(323),
																										Column: int(35),
																									},
																									End: ast.Location{
																										Line: int(323),
																										Column: int(40),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p3598,
																								FreeVars: ast.Identifiers{
																									"c",
																									"v",
																								},
																							},
																							Left: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(323),
																											Column: int(35),
																										},
																										End: ast.Location{
																											Line: int(323),
																											Column: int(36),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p3598,
																									FreeVars: ast.Identifiers{
																										"v",
																									},
																								},
																								Id: "v",
																							},
																							OpFodder: ast.Fodder{},
																							Op: ast.BinaryOp(3),
																							Right: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(323),
																											Column: int(39),
																										},
																										End: ast.Location{
																											Line: int(323),
																											Column: int(40),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p3598,
																									FreeVars: ast.Identifiers{
																										"c",
																									},
																								},
																								Id: "c",
																							},
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			TrailingComma: false,
																			TailStrict: false,
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																		},
																		ElseFodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(12),
																				Comment: []string{},
																			},
																		},
																		BranchFalse: &ast.DesugaredObject{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(325),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(325),
																						Column: int(33),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(14),
																						Comment: []string{},
																					},
																				},
																				Ctx: p3551,
																				FreeVars: ast.Identifiers{
																					"j",
																					"v",
																				},
																			},
																			Asserts: ast.Nodes{},
																			Fields: ast.DesugaredObjectFields{
																				ast.DesugaredObjectField{
																					Hide: ast.ObjectFieldHide(1),
																					Name: &ast.LiteralString{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								File: nil,
																							},
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: nil,
																						},
																						Value: "i",
																						Kind: ast.LiteralStringKind(1),
																						BlockIndent: "",
																						BlockTermIndent: "",
																					},
																					Body: &ast.Binary{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(325),
																									Column: int(20),
																								},
																								End: ast.Location{
																									Line: int(325),
																									Column: int(25),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p3618,
																							FreeVars: ast.Identifiers{
																								"j",
																							},
																						},
																						Left: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(325),
																										Column: int(20),
																									},
																									End: ast.Location{
																										Line: int(325),
																										Column: int(21),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p3618,
																								FreeVars: ast.Identifiers{
																									"j",
																								},
																							},
																							Id: "j",
																						},
																						OpFodder: ast.Fodder{},
																						Op: ast.BinaryOp(3),
																						Right: &ast.LiteralNumber{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(325),
																										Column: int(24),
																									},
																									End: ast.Location{
																										Line: int(325),
																										Column: int(25),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p3618,
																								FreeVars: nil,
																							},
																							OriginalString: "1",
																						},
																					},
																					PlusSuper: false,
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(325),
																							Column: int(17),
																						},
																						End: ast.Location{
																							Line: int(325),
																							Column: int(25),
																						},
																						File: p1,
																					},
																				},
																				ast.DesugaredObjectField{
																					Hide: ast.ObjectFieldHide(1),
																					Name: &ast.LiteralString{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								File: nil,
																							},
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: nil,
																						},
																						Value: "v",
																						Kind: ast.LiteralStringKind(1),
																						BlockIndent: "",
																						BlockTermIndent: "",
																					},
																					Body: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(325),
																									Column: int(30),
																								},
																								End: ast.Location{
																									Line: int(325),
																									Column: int(31),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p3618,
																							FreeVars: ast.Identifiers{
																								"v",
																							},
																						},
																						Id: "v",
																					},
																					PlusSuper: false,
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(325),
																							Column: int(27),
																						},
																						End: ast.Location{
																							Line: int(325),
																							Column: int(31),
																						},
																						File: p1,
																					},
																				},
																			},
																			Locals: ast.LocalBinds{},
																		},
																	},
																},
															},
														},
														Fun: nil,
														CloseFodder: nil,
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
													},
												},
												Body: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(326),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(326),
																Column: int(32),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p3507,
														FreeVars: ast.Identifiers{
															"consume",
															"i",
															"str",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(326),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(326),
																	Column: int(16),
																},
																File: p1,
															},
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																	Comment: []string{},
																},
															},
															Ctx: p3507,
															FreeVars: ast.Identifiers{
																"consume",
															},
														},
														Id: "consume",
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(326),
																				Column: int(17),
																			},
																			End: ast.Location{
																				Line: int(326),
																				Column: int(20),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p3633,
																		FreeVars: ast.Identifiers{
																			"str",
																		},
																	},
																	Id: "str",
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(326),
																				Column: int(22),
																			},
																			End: ast.Location{
																				Line: int(326),
																				Column: int(27),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p3633,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																	},
																	Left: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(326),
																					Column: int(22),
																				},
																				End: ast.Location{
																					Line: int(326),
																					Column: int(23),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3633,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																		},
																		Id: "i",
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(3),
																	Right: &ast.LiteralNumber{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(326),
																					Column: int(26),
																				},
																				End: ast.Location{
																					Line: int(326),
																					Column: int(27),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3633,
																			FreeVars: nil,
																		},
																		OriginalString: "1",
																	},
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.LiteralString{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(326),
																				Column: int(29),
																			},
																			End: ast.Location{
																				Line: int(326),
																				Column: int(31),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p3633,
																		FreeVars: nil,
																	},
																	Value: "",
																	Kind: ast.LiteralStringKind(1),
																	BlockIndent: "",
																	BlockTermIndent: "",
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
												},
											},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
													Comment: []string{},
												},
											},
											BranchFalse: &ast.DesugaredObject{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(328),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(328),
															Column: int(26),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
															Comment: []string{},
														},
													},
													Ctx: p3507,
													FreeVars: ast.Identifiers{
														"i",
													},
												},
												Asserts: ast.Nodes{},
												Fields: ast.DesugaredObjectFields{
													ast.DesugaredObjectField{
														Hide: ast.ObjectFieldHide(1),
														Name: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
															Value: "i",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														Body: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(328),
																		Column: int(14),
																	},
																	End: ast.Location{
																		Line: int(328),
																		Column: int(15),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p3648,
																FreeVars: ast.Identifiers{
																	"i",
																},
															},
															Id: "i",
														},
														PlusSuper: false,
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(328),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(328),
																Column: int(15),
															},
															File: p1,
														},
													},
													ast.DesugaredObjectField{
														Hide: ast.ObjectFieldHide(1),
														Name: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
															Value: "v",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														Body: &ast.LiteralNull{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(328),
																		Column: int(20),
																	},
																	End: ast.Location{
																		Line: int(328),
																		Column: int(24),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p3648,
																FreeVars: nil,
															},
														},
														PlusSuper: false,
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(328),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(328),
																Column: int(24),
															},
															File: p1,
														},
													},
												},
												Locals: ast.LocalBinds{},
											},
										},
									},
									ElseFodder: nil,
									BranchFalse: &ast.Error{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Expr: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(314),
														Column: int(36),
													},
													End: ast.Location{
														Line: int(314),
														Column: int(60),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p3507,
												FreeVars: nil,
											},
											Value: "Truncated format code.",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
									},
								},
							},
							Fun: nil,
							CloseFodder: nil,
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
						},
					},
					Body: &ast.Local{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(330),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(774),
									Column: int(48),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(1),
									Indent: int(4),
									Comment: []string{},
								},
							},
							Ctx: p3497,
							FreeVars: ast.Identifiers{
								"std",
								"str",
								"try_parse_mapping_key",
								"vals",
							},
						},
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: nil,
								Variable: "try_parse_cflags",
								EqFodder: nil,
								Body: &ast.Function{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(330),
												Column: int(11),
											},
											End: ast.Location{
												Line: int(346),
												Column: int(91),
											},
											File: p1,
										},
										Fodder: nil,
										Ctx: p3659,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									ParenLeftFodder: ast.Fodder{},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "str",
											EqFodder: nil,
											DefaultArg: nil,
											CommaFodder: ast.Fodder{},
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(330),
													Column: int(28),
												},
												End: ast.Location{
													Line: int(330),
													Column: int(31),
												},
												File: p1,
											},
										},
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "i",
											EqFodder: nil,
											DefaultArg: nil,
											CommaFodder: nil,
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(330),
													Column: int(33),
												},
												End: ast.Location{
													Line: int(330),
													Column: int(34),
												},
												File: p1,
											},
										},
									},
									TrailingComma: false,
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Local{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(331),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(346),
													Column: int(91),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
													Comment: []string{},
												},
											},
											Ctx: p3664,
											FreeVars: ast.Identifiers{
												"i",
												"std",
												"str",
											},
										},
										Binds: ast.LocalBinds{
											ast.LocalBind{
												VarFodder: nil,
												Variable: "consume",
												EqFodder: nil,
												Body: &ast.Function{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(331),
																Column: int(13),
															},
															End: ast.Location{
																Line: int(345),
																Column: int(25),
															},
															File: p1,
														},
														Fodder: nil,
														Ctx: p3668,
														FreeVars: ast.Identifiers{
															"consume",
															"std",
														},
													},
													ParenLeftFodder: ast.Fodder{},
													Parameters: []ast.Parameter{
														ast.Parameter{
															NameFodder: ast.Fodder{},
															Name: "str",
															EqFodder: nil,
															DefaultArg: nil,
															CommaFodder: ast.Fodder{},
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(331),
																	Column: int(21),
																},
																End: ast.Location{
																	Line: int(331),
																	Column: int(24),
																},
																File: p1,
															},
														},
														ast.Parameter{
															NameFodder: ast.Fodder{},
															Name: "j",
															EqFodder: nil,
															DefaultArg: nil,
															CommaFodder: ast.Fodder{},
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(331),
																	Column: int(26),
																},
																End: ast.Location{
																	Line: int(331),
																	Column: int(27),
																},
																File: p1,
															},
														},
														ast.Parameter{
															NameFodder: ast.Fodder{},
															Name: "v",
															EqFodder: nil,
															DefaultArg: nil,
															CommaFodder: nil,
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(331),
																	Column: int(29),
																},
																End: ast.Location{
																	Line: int(331),
																	Column: int(30),
																},
																File: p1,
															},
														},
													},
													TrailingComma: false,
													ParenRightFodder: ast.Fodder{},
													Body: &ast.Conditional{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"consume",
																"j",
																"std",
																"str",
																"v",
															},
														},
														Cond: &ast.Binary{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(332),
																		Column: int(16),
																	},
																	End: ast.Location{
																		Line: int(332),
																		Column: int(35),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p3674,
																FreeVars: ast.Identifiers{
																	"j",
																	"std",
																	"str",
																},
															},
															Left: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(332),
																			Column: int(16),
																		},
																		End: ast.Location{
																			Line: int(332),
																			Column: int(17),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p3674,
																	FreeVars: ast.Identifiers{
																		"j",
																	},
																},
																Id: "j",
															},
															OpFodder: ast.Fodder{},
															Op: ast.BinaryOp(9),
															Right: &ast.Apply{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(332),
																			Column: int(20),
																		},
																		End: ast.Location{
																			Line: int(332),
																			Column: int(35),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p3674,
																	FreeVars: ast.Identifiers{
																		"std",
																		"str",
																	},
																},
																Target: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(332),
																				Column: int(20),
																			},
																			End: ast.Location{
																				Line: int(332),
																				Column: int(30),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p3674,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(332),
																					Column: int(20),
																				},
																				End: ast.Location{
																					Line: int(332),
																					Column: int(23),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Id: "std",
																	},
																	LeftBracketFodder: ast.Fodder{},
																	Index: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: nil,
																		},
																		Value: "length",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	RightBracketFodder: ast.Fodder{},
																	Id: nil,
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(332),
																							Column: int(31),
																						},
																						End: ast.Location{
																							Line: int(332),
																							Column: int(34),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p3687,
																					FreeVars: ast.Identifiers{
																						"str",
																					},
																				},
																				Id: "str",
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																TrailingComma: false,
																TailStrict: false,
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
															},
														},
														ThenFodder: nil,
														BranchTrue: &ast.Local{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(333),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(345),
																		Column: int(25),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																		Comment: []string{},
																	},
																},
																Ctx: p3674,
																FreeVars: ast.Identifiers{
																	"consume",
																	"j",
																	"str",
																	"v",
																},
															},
															Binds: ast.LocalBinds{
																ast.LocalBind{
																	VarFodder: ast.Fodder{},
																	Variable: "c",
																	EqFodder: ast.Fodder{},
																	Body: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(333),
																					Column: int(19),
																				},
																				End: ast.Location{
																					Line: int(333),
																					Column: int(25),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3694,
																			FreeVars: ast.Identifiers{
																				"j",
																				"str",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(333),
																						Column: int(19),
																					},
																					End: ast.Location{
																						Line: int(333),
																						Column: int(22),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p3694,
																				FreeVars: ast.Identifiers{
																					"str",
																				},
																			},
																			Id: "str",
																		},
																		LeftBracketFodder: ast.Fodder{},
																		Index: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(333),
																						Column: int(23),
																					},
																					End: ast.Location{
																						Line: int(333),
																						Column: int(24),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p3694,
																				FreeVars: ast.Identifiers{
																					"j",
																				},
																			},
																			Id: "j",
																		},
																		RightBracketFodder: ast.Fodder{},
																		Id: nil,
																	},
																	Fun: nil,
																	CloseFodder: ast.Fodder{},
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(333),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(333),
																			Column: int(25),
																		},
																		File: p1,
																	},
																},
															},
															Body: &ast.Conditional{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(334),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(345),
																			Column: int(25),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																			Comment: []string{},
																		},
																	},
																	Ctx: p3674,
																	FreeVars: ast.Identifiers{
																		"c",
																		"consume",
																		"j",
																		"str",
																		"v",
																	},
																},
																Cond: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(334),
																				Column: int(12),
																			},
																			End: ast.Location{
																				Line: int(334),
																				Column: int(20),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p3674,
																		FreeVars: ast.Identifiers{
																			"c",
																		},
																	},
																	Left: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(334),
																					Column: int(12),
																				},
																				End: ast.Location{
																					Line: int(334),
																					Column: int(13),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3674,
																			FreeVars: ast.Identifiers{
																				"c",
																			},
																		},
																		Id: "c",
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(12),
																	Right: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(334),
																					Column: int(17),
																				},
																				End: ast.Location{
																					Line: int(334),
																					Column: int(20),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3674,
																			FreeVars: nil,
																		},
																		Value: "#",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																},
																ThenFodder: ast.Fodder{},
																BranchTrue: &ast.Apply{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(335),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(335),
																				Column: int(47),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p3674,
																		FreeVars: ast.Identifiers{
																			"consume",
																			"j",
																			"str",
																			"v",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(335),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(335),
																					Column: int(18),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(10),
																					Comment: []string{},
																				},
																			},
																			Ctx: p3674,
																			FreeVars: ast.Identifiers{
																				"consume",
																			},
																		},
																		Id: "consume",
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(335),
																								Column: int(19),
																							},
																							End: ast.Location{
																								Line: int(335),
																								Column: int(22),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p3715,
																						FreeVars: ast.Identifiers{
																							"str",
																						},
																					},
																					Id: "str",
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Binary{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(335),
																								Column: int(24),
																							},
																							End: ast.Location{
																								Line: int(335),
																								Column: int(29),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p3715,
																						FreeVars: ast.Identifiers{
																							"j",
																						},
																					},
																					Left: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(335),
																									Column: int(24),
																								},
																								End: ast.Location{
																									Line: int(335),
																									Column: int(25),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p3715,
																							FreeVars: ast.Identifiers{
																								"j",
																							},
																						},
																						Id: "j",
																					},
																					OpFodder: ast.Fodder{},
																					Op: ast.BinaryOp(3),
																					Right: &ast.LiteralNumber{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(335),
																									Column: int(28),
																								},
																								End: ast.Location{
																									Line: int(335),
																									Column: int(29),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p3715,
																							FreeVars: nil,
																						},
																						OriginalString: "1",
																					},
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Binary{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(335),
																								Column: int(31),
																							},
																							End: ast.Location{
																								Line: int(335),
																								Column: int(46),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p3715,
																						FreeVars: ast.Identifiers{
																							"v",
																						},
																					},
																					Left: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(335),
																									Column: int(31),
																								},
																								End: ast.Location{
																									Line: int(335),
																									Column: int(32),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p3715,
																							FreeVars: ast.Identifiers{
																								"v",
																							},
																						},
																						Id: "v",
																					},
																					OpFodder: nil,
																					Op: ast.BinaryOp(3),
																					Right: &ast.DesugaredObject{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(335),
																									Column: int(33),
																								},
																								End: ast.Location{
																									Line: int(335),
																									Column: int(46),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p3715,
																							FreeVars: nil,
																						},
																						Asserts: ast.Nodes{},
																						Fields: ast.DesugaredObjectFields{
																							ast.DesugaredObjectField{
																								Hide: ast.ObjectFieldHide(1),
																								Name: &ast.LiteralString{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											File: nil,
																										},
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: nil,
																									},
																									Value: "alt",
																									Kind: ast.LiteralStringKind(1),
																									BlockIndent: "",
																									BlockTermIndent: "",
																								},
																								Body: &ast.LiteralBoolean{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(335),
																												Column: int(40),
																											},
																											End: ast.Location{
																												Line: int(335),
																												Column: int(44),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p3730,
																										FreeVars: nil,
																									},
																									Value: true,
																								},
																								PlusSuper: false,
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(335),
																										Column: int(35),
																									},
																									End: ast.Location{
																										Line: int(335),
																										Column: int(44),
																									},
																									File: p1,
																								},
																							},
																						},
																						Locals: ast.LocalBinds{},
																					},
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	TrailingComma: false,
																	TailStrict: false,
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																},
																ElseFodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																		Comment: []string{},
																	},
																},
																BranchFalse: &ast.Conditional{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(336),
																				Column: int(14),
																			},
																			End: ast.Location{
																				Line: int(345),
																				Column: int(25),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p3674,
																		FreeVars: ast.Identifiers{
																			"c",
																			"consume",
																			"j",
																			"str",
																			"v",
																		},
																	},
																	Cond: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(336),
																					Column: int(17),
																				},
																				End: ast.Location{
																					Line: int(336),
																					Column: int(25),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3674,
																			FreeVars: ast.Identifiers{
																				"c",
																			},
																		},
																		Left: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(336),
																						Column: int(17),
																					},
																					End: ast.Location{
																						Line: int(336),
																						Column: int(18),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p3674,
																				FreeVars: ast.Identifiers{
																					"c",
																				},
																			},
																			Id: "c",
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(12),
																		Right: &ast.LiteralString{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(336),
																						Column: int(22),
																					},
																					End: ast.Location{
																						Line: int(336),
																						Column: int(25),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p3674,
																				FreeVars: nil,
																			},
																			Value: "0",
																			Kind: ast.LiteralStringKind(1),
																			BlockIndent: "",
																			BlockTermIndent: "",
																		},
																	},
																	ThenFodder: ast.Fodder{},
																	BranchTrue: &ast.Apply{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(337),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(337),
																					Column: int(48),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3674,
																			FreeVars: ast.Identifiers{
																				"consume",
																				"j",
																				"str",
																				"v",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(337),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(337),
																						Column: int(18),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(10),
																						Comment: []string{},
																					},
																				},
																				Ctx: p3674,
																				FreeVars: ast.Identifiers{
																					"consume",
																				},
																			},
																			Id: "consume",
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(337),
																									Column: int(19),
																								},
																								End: ast.Location{
																									Line: int(337),
																									Column: int(22),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p3746,
																							FreeVars: ast.Identifiers{
																								"str",
																							},
																						},
																						Id: "str",
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Binary{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(337),
																									Column: int(24),
																								},
																								End: ast.Location{
																									Line: int(337),
																									Column: int(29),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p3746,
																							FreeVars: ast.Identifiers{
																								"j",
																							},
																						},
																						Left: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(337),
																										Column: int(24),
																									},
																									End: ast.Location{
																										Line: int(337),
																										Column: int(25),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p3746,
																								FreeVars: ast.Identifiers{
																									"j",
																								},
																							},
																							Id: "j",
																						},
																						OpFodder: ast.Fodder{},
																						Op: ast.BinaryOp(3),
																						Right: &ast.LiteralNumber{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(337),
																										Column: int(28),
																									},
																									End: ast.Location{
																										Line: int(337),
																										Column: int(29),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p3746,
																								FreeVars: nil,
																							},
																							OriginalString: "1",
																						},
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Binary{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(337),
																									Column: int(31),
																								},
																								End: ast.Location{
																									Line: int(337),
																									Column: int(47),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p3746,
																							FreeVars: ast.Identifiers{
																								"v",
																							},
																						},
																						Left: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(337),
																										Column: int(31),
																									},
																									End: ast.Location{
																										Line: int(337),
																										Column: int(32),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p3746,
																								FreeVars: ast.Identifiers{
																									"v",
																								},
																							},
																							Id: "v",
																						},
																						OpFodder: nil,
																						Op: ast.BinaryOp(3),
																						Right: &ast.DesugaredObject{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(337),
																										Column: int(33),
																									},
																									End: ast.Location{
																										Line: int(337),
																										Column: int(47),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p3746,
																								FreeVars: nil,
																							},
																							Asserts: ast.Nodes{},
																							Fields: ast.DesugaredObjectFields{
																								ast.DesugaredObjectField{
																									Hide: ast.ObjectFieldHide(1),
																									Name: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												File: nil,
																											},
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: nil,
																										},
																										Value: "zero",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																									Body: &ast.LiteralBoolean{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(337),
																													Column: int(41),
																												},
																												End: ast.Location{
																													Line: int(337),
																													Column: int(45),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p3761,
																											FreeVars: nil,
																										},
																										Value: true,
																									},
																									PlusSuper: false,
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(337),
																											Column: int(35),
																										},
																										End: ast.Location{
																											Line: int(337),
																											Column: int(45),
																										},
																										File: p1,
																									},
																								},
																							},
																							Locals: ast.LocalBinds{},
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		TrailingComma: false,
																		TailStrict: false,
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																	},
																	ElseFodder: ast.Fodder{
																		ast.FodderElement{
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																			Comment: []string{},
																		},
																	},
																	BranchFalse: &ast.Conditional{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(338),
																					Column: int(14),
																				},
																				End: ast.Location{
																					Line: int(345),
																					Column: int(25),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3674,
																			FreeVars: ast.Identifiers{
																				"c",
																				"consume",
																				"j",
																				"str",
																				"v",
																			},
																		},
																		Cond: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(338),
																						Column: int(17),
																					},
																					End: ast.Location{
																						Line: int(338),
																						Column: int(25),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p3674,
																				FreeVars: ast.Identifiers{
																					"c",
																				},
																			},
																			Left: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(338),
																							Column: int(17),
																						},
																						End: ast.Location{
																							Line: int(338),
																							Column: int(18),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p3674,
																					FreeVars: ast.Identifiers{
																						"c",
																					},
																				},
																				Id: "c",
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(12),
																			Right: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(338),
																							Column: int(22),
																						},
																						End: ast.Location{
																							Line: int(338),
																							Column: int(25),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p3674,
																					FreeVars: nil,
																				},
																				Value: "-",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																		},
																		ThenFodder: ast.Fodder{},
																		BranchTrue: &ast.Apply{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(339),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(339),
																						Column: int(48),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p3674,
																				FreeVars: ast.Identifiers{
																					"consume",
																					"j",
																					"str",
																					"v",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(339),
																							Column: int(11),
																						},
																						End: ast.Location{
																							Line: int(339),
																							Column: int(18),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(10),
																							Comment: []string{},
																						},
																					},
																					Ctx: p3674,
																					FreeVars: ast.Identifiers{
																						"consume",
																					},
																				},
																				Id: "consume",
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(339),
																										Column: int(19),
																									},
																									End: ast.Location{
																										Line: int(339),
																										Column: int(22),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p3777,
																								FreeVars: ast.Identifiers{
																									"str",
																								},
																							},
																							Id: "str",
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Binary{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(339),
																										Column: int(24),
																									},
																									End: ast.Location{
																										Line: int(339),
																										Column: int(29),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p3777,
																								FreeVars: ast.Identifiers{
																									"j",
																								},
																							},
																							Left: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(339),
																											Column: int(24),
																										},
																										End: ast.Location{
																											Line: int(339),
																											Column: int(25),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p3777,
																									FreeVars: ast.Identifiers{
																										"j",
																									},
																								},
																								Id: "j",
																							},
																							OpFodder: ast.Fodder{},
																							Op: ast.BinaryOp(3),
																							Right: &ast.LiteralNumber{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(339),
																											Column: int(28),
																										},
																										End: ast.Location{
																											Line: int(339),
																											Column: int(29),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p3777,
																									FreeVars: nil,
																								},
																								OriginalString: "1",
																							},
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Binary{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(339),
																										Column: int(31),
																									},
																									End: ast.Location{
																										Line: int(339),
																										Column: int(47),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p3777,
																								FreeVars: ast.Identifiers{
																									"v",
																								},
																							},
																							Left: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(339),
																											Column: int(31),
																										},
																										End: ast.Location{
																											Line: int(339),
																											Column: int(32),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p3777,
																									FreeVars: ast.Identifiers{
																										"v",
																									},
																								},
																								Id: "v",
																							},
																							OpFodder: nil,
																							Op: ast.BinaryOp(3),
																							Right: &ast.DesugaredObject{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(339),
																											Column: int(33),
																										},
																										End: ast.Location{
																											Line: int(339),
																											Column: int(47),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p3777,
																									FreeVars: nil,
																								},
																								Asserts: ast.Nodes{},
																								Fields: ast.DesugaredObjectFields{
																									ast.DesugaredObjectField{
																										Hide: ast.ObjectFieldHide(1),
																										Name: &ast.LiteralString{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													File: nil,
																												},
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: nil,
																											},
																											Value: "left",
																											Kind: ast.LiteralStringKind(1),
																											BlockIndent: "",
																											BlockTermIndent: "",
																										},
																										Body: &ast.LiteralBoolean{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(339),
																														Column: int(41),
																													},
																													End: ast.Location{
																														Line: int(339),
																														Column: int(45),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p3792,
																												FreeVars: nil,
																											},
																											Value: true,
																										},
																										PlusSuper: false,
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(339),
																												Column: int(35),
																											},
																											End: ast.Location{
																												Line: int(339),
																												Column: int(45),
																											},
																											File: p1,
																										},
																									},
																								},
																								Locals: ast.LocalBinds{},
																							},
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			TrailingComma: false,
																			TailStrict: false,
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																		},
																		ElseFodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(8),
																				Comment: []string{},
																			},
																		},
																		BranchFalse: &ast.Conditional{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(340),
																						Column: int(14),
																					},
																					End: ast.Location{
																						Line: int(345),
																						Column: int(25),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p3674,
																				FreeVars: ast.Identifiers{
																					"c",
																					"consume",
																					"j",
																					"str",
																					"v",
																				},
																			},
																			Cond: &ast.Binary{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(340),
																							Column: int(17),
																						},
																						End: ast.Location{
																							Line: int(340),
																							Column: int(25),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p3674,
																					FreeVars: ast.Identifiers{
																						"c",
																					},
																				},
																				Left: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(340),
																								Column: int(17),
																							},
																							End: ast.Location{
																								Line: int(340),
																								Column: int(18),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p3674,
																						FreeVars: ast.Identifiers{
																							"c",
																						},
																					},
																					Id: "c",
																				},
																				OpFodder: ast.Fodder{},
																				Op: ast.BinaryOp(12),
																				Right: &ast.LiteralString{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(340),
																								Column: int(22),
																							},
																							End: ast.Location{
																								Line: int(340),
																								Column: int(25),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p3674,
																						FreeVars: nil,
																					},
																					Value: " ",
																					Kind: ast.LiteralStringKind(1),
																					BlockIndent: "",
																					BlockTermIndent: "",
																				},
																			},
																			ThenFodder: ast.Fodder{},
																			BranchTrue: &ast.Apply{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(341),
																							Column: int(11),
																						},
																						End: ast.Location{
																							Line: int(341),
																							Column: int(49),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p3674,
																					FreeVars: ast.Identifiers{
																						"consume",
																						"j",
																						"str",
																						"v",
																					},
																				},
																				Target: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(341),
																								Column: int(11),
																							},
																							End: ast.Location{
																								Line: int(341),
																								Column: int(18),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(10),
																								Comment: []string{},
																							},
																						},
																						Ctx: p3674,
																						FreeVars: ast.Identifiers{
																							"consume",
																						},
																					},
																					Id: "consume",
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(341),
																											Column: int(19),
																										},
																										End: ast.Location{
																											Line: int(341),
																											Column: int(22),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p3808,
																									FreeVars: ast.Identifiers{
																										"str",
																									},
																								},
																								Id: "str",
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Binary{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(341),
																											Column: int(24),
																										},
																										End: ast.Location{
																											Line: int(341),
																											Column: int(29),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p3808,
																									FreeVars: ast.Identifiers{
																										"j",
																									},
																								},
																								Left: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(341),
																												Column: int(24),
																											},
																											End: ast.Location{
																												Line: int(341),
																												Column: int(25),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p3808,
																										FreeVars: ast.Identifiers{
																											"j",
																										},
																									},
																									Id: "j",
																								},
																								OpFodder: ast.Fodder{},
																								Op: ast.BinaryOp(3),
																								Right: &ast.LiteralNumber{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(341),
																												Column: int(28),
																											},
																											End: ast.Location{
																												Line: int(341),
																												Column: int(29),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p3808,
																										FreeVars: nil,
																									},
																									OriginalString: "1",
																								},
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Binary{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(341),
																											Column: int(31),
																										},
																										End: ast.Location{
																											Line: int(341),
																											Column: int(48),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p3808,
																									FreeVars: ast.Identifiers{
																										"v",
																									},
																								},
																								Left: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(341),
																												Column: int(31),
																											},
																											End: ast.Location{
																												Line: int(341),
																												Column: int(32),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p3808,
																										FreeVars: ast.Identifiers{
																											"v",
																										},
																									},
																									Id: "v",
																								},
																								OpFodder: nil,
																								Op: ast.BinaryOp(3),
																								Right: &ast.DesugaredObject{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(341),
																												Column: int(33),
																											},
																											End: ast.Location{
																												Line: int(341),
																												Column: int(48),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p3808,
																										FreeVars: nil,
																									},
																									Asserts: ast.Nodes{},
																									Fields: ast.DesugaredObjectFields{
																										ast.DesugaredObjectField{
																											Hide: ast.ObjectFieldHide(1),
																											Name: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: nil,
																												},
																												Value: "blank",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																											Body: &ast.LiteralBoolean{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(341),
																															Column: int(42),
																														},
																														End: ast.Location{
																															Line: int(341),
																															Column: int(46),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p3823,
																													FreeVars: nil,
																												},
																												Value: true,
																											},
																											PlusSuper: false,
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(341),
																													Column: int(35),
																												},
																												End: ast.Location{
																													Line: int(341),
																													Column: int(46),
																												},
																												File: p1,
																											},
																										},
																									},
																									Locals: ast.LocalBinds{},
																								},
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				TrailingComma: false,
																				TailStrict: false,
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: nil,
																			},
																			ElseFodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(8),
																					Comment: []string{},
																				},
																			},
																			BranchFalse: &ast.Conditional{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(342),
																							Column: int(14),
																						},
																						End: ast.Location{
																							Line: int(345),
																							Column: int(25),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p3674,
																					FreeVars: ast.Identifiers{
																						"c",
																						"consume",
																						"j",
																						"str",
																						"v",
																					},
																				},
																				Cond: &ast.Binary{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(342),
																								Column: int(17),
																							},
																							End: ast.Location{
																								Line: int(342),
																								Column: int(25),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p3674,
																						FreeVars: ast.Identifiers{
																							"c",
																						},
																					},
																					Left: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(342),
																									Column: int(17),
																								},
																								End: ast.Location{
																									Line: int(342),
																									Column: int(18),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p3674,
																							FreeVars: ast.Identifiers{
																								"c",
																							},
																						},
																						Id: "c",
																					},
																					OpFodder: ast.Fodder{},
																					Op: ast.BinaryOp(12),
																					Right: &ast.LiteralString{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(342),
																									Column: int(22),
																								},
																								End: ast.Location{
																									Line: int(342),
																									Column: int(25),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p3674,
																							FreeVars: nil,
																						},
																						Value: "+",
																						Kind: ast.LiteralStringKind(1),
																						BlockIndent: "",
																						BlockTermIndent: "",
																					},
																				},
																				ThenFodder: ast.Fodder{},
																				BranchTrue: &ast.Apply{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(343),
																								Column: int(11),
																							},
																							End: ast.Location{
																								Line: int(343),
																								Column: int(48),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p3674,
																						FreeVars: ast.Identifiers{
																							"consume",
																							"j",
																							"str",
																							"v",
																						},
																					},
																					Target: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(343),
																									Column: int(11),
																								},
																								End: ast.Location{
																									Line: int(343),
																									Column: int(18),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{
																								ast.FodderElement{
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(10),
																									Comment: []string{},
																								},
																							},
																							Ctx: p3674,
																							FreeVars: ast.Identifiers{
																								"consume",
																							},
																						},
																						Id: "consume",
																					},
																					FodderLeft: ast.Fodder{},
																					Arguments: ast.Arguments{
																						Positional: []ast.CommaSeparatedExpr{
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(343),
																												Column: int(19),
																											},
																											End: ast.Location{
																												Line: int(343),
																												Column: int(22),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p3839,
																										FreeVars: ast.Identifiers{
																											"str",
																										},
																									},
																									Id: "str",
																								},
																								CommaFodder: ast.Fodder{},
																							},
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Binary{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(343),
																												Column: int(24),
																											},
																											End: ast.Location{
																												Line: int(343),
																												Column: int(29),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p3839,
																										FreeVars: ast.Identifiers{
																											"j",
																										},
																									},
																									Left: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(343),
																													Column: int(24),
																												},
																												End: ast.Location{
																													Line: int(343),
																													Column: int(25),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p3839,
																											FreeVars: ast.Identifiers{
																												"j",
																											},
																										},
																										Id: "j",
																									},
																									OpFodder: ast.Fodder{},
																									Op: ast.BinaryOp(3),
																									Right: &ast.LiteralNumber{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(343),
																													Column: int(28),
																												},
																												End: ast.Location{
																													Line: int(343),
																													Column: int(29),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p3839,
																											FreeVars: nil,
																										},
																										OriginalString: "1",
																									},
																								},
																								CommaFodder: ast.Fodder{},
																							},
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Binary{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(343),
																												Column: int(31),
																											},
																											End: ast.Location{
																												Line: int(343),
																												Column: int(47),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p3839,
																										FreeVars: ast.Identifiers{
																											"v",
																										},
																									},
																									Left: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(343),
																													Column: int(31),
																												},
																												End: ast.Location{
																													Line: int(343),
																													Column: int(32),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p3839,
																											FreeVars: ast.Identifiers{
																												"v",
																											},
																										},
																										Id: "v",
																									},
																									OpFodder: nil,
																									Op: ast.BinaryOp(3),
																									Right: &ast.DesugaredObject{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(343),
																													Column: int(33),
																												},
																												End: ast.Location{
																													Line: int(343),
																													Column: int(47),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p3839,
																											FreeVars: nil,
																										},
																										Asserts: ast.Nodes{},
																										Fields: ast.DesugaredObjectFields{
																											ast.DesugaredObjectField{
																												Hide: ast.ObjectFieldHide(1),
																												Name: &ast.LiteralString{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															File: nil,
																														},
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: nil,
																													},
																													Value: "plus",
																													Kind: ast.LiteralStringKind(1),
																													BlockIndent: "",
																													BlockTermIndent: "",
																												},
																												Body: &ast.LiteralBoolean{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(343),
																																Column: int(41),
																															},
																															End: ast.Location{
																																Line: int(343),
																																Column: int(45),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p3854,
																														FreeVars: nil,
																													},
																													Value: true,
																												},
																												PlusSuper: false,
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(343),
																														Column: int(35),
																													},
																													End: ast.Location{
																														Line: int(343),
																														Column: int(45),
																													},
																													File: p1,
																												},
																											},
																										},
																										Locals: ast.LocalBinds{},
																									},
																								},
																								CommaFodder: nil,
																							},
																						},
																						Named: nil,
																					},
																					TrailingComma: false,
																					TailStrict: false,
																					FodderRight: ast.Fodder{},
																					TailStrictFodder: nil,
																				},
																				ElseFodder: ast.Fodder{
																					ast.FodderElement{
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(8),
																						Comment: []string{},
																					},
																				},
																				BranchFalse: &ast.DesugaredObject{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(345),
																								Column: int(11),
																							},
																							End: ast.Location{
																								Line: int(345),
																								Column: int(25),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(10),
																								Comment: []string{},
																							},
																						},
																						Ctx: p3674,
																						FreeVars: ast.Identifiers{
																							"j",
																							"v",
																						},
																					},
																					Asserts: ast.Nodes{},
																					Fields: ast.DesugaredObjectFields{
																						ast.DesugaredObjectField{
																							Hide: ast.ObjectFieldHide(1),
																							Name: &ast.LiteralString{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										File: nil,
																									},
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: nil,
																								},
																								Value: "i",
																								Kind: ast.LiteralStringKind(1),
																								BlockIndent: "",
																								BlockTermIndent: "",
																							},
																							Body: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(345),
																											Column: int(16),
																										},
																										End: ast.Location{
																											Line: int(345),
																											Column: int(17),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p3862,
																									FreeVars: ast.Identifiers{
																										"j",
																									},
																								},
																								Id: "j",
																							},
																							PlusSuper: false,
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(345),
																									Column: int(13),
																								},
																								End: ast.Location{
																									Line: int(345),
																									Column: int(17),
																								},
																								File: p1,
																							},
																						},
																						ast.DesugaredObjectField{
																							Hide: ast.ObjectFieldHide(1),
																							Name: &ast.LiteralString{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										File: nil,
																									},
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: nil,
																								},
																								Value: "v",
																								Kind: ast.LiteralStringKind(1),
																								BlockIndent: "",
																								BlockTermIndent: "",
																							},
																							Body: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(345),
																											Column: int(22),
																										},
																										End: ast.Location{
																											Line: int(345),
																											Column: int(23),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p3862,
																									FreeVars: ast.Identifiers{
																										"v",
																									},
																								},
																								Id: "v",
																							},
																							PlusSuper: false,
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(345),
																									Column: int(19),
																								},
																								End: ast.Location{
																									Line: int(345),
																									Column: int(23),
																								},
																								File: p1,
																							},
																						},
																					},
																					Locals: ast.LocalBinds{},
																				},
																			},
																		},
																	},
																},
															},
														},
														ElseFodder: nil,
														BranchFalse: &ast.Error{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
															Expr: &ast.LiteralString{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(332),
																			Column: int(38),
																		},
																		End: ast.Location{
																			Line: int(332),
																			Column: int(62),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p3674,
																	FreeVars: nil,
																},
																Value: "Truncated format code.",
																Kind: ast.LiteralStringKind(1),
																BlockIndent: "",
																BlockTermIndent: "",
															},
														},
													},
												},
												Fun: nil,
												CloseFodder: nil,
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
											},
										},
										Body: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(346),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(346),
														Column: int(91),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p3664,
												FreeVars: ast.Identifiers{
													"consume",
													"i",
													"str",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(346),
															Column: int(7),
														},
														End: ast.Location{
															Line: int(346),
															Column: int(14),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(6),
															Comment: []string{},
														},
													},
													Ctx: p3664,
													FreeVars: ast.Identifiers{
														"consume",
													},
												},
												Id: "consume",
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(346),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(346),
																		Column: int(18),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p3876,
																FreeVars: ast.Identifiers{
																	"str",
																},
															},
															Id: "str",
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(346),
																		Column: int(20),
																	},
																	End: ast.Location{
																		Line: int(346),
																		Column: int(21),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p3876,
																FreeVars: ast.Identifiers{
																	"i",
																},
															},
															Id: "i",
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.DesugaredObject{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(346),
																		Column: int(23),
																	},
																	End: ast.Location{
																		Line: int(346),
																		Column: int(90),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p3876,
																FreeVars: nil,
															},
															Asserts: ast.Nodes{},
															Fields: ast.DesugaredObjectFields{
																ast.DesugaredObjectField{
																	Hide: ast.ObjectFieldHide(1),
																	Name: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: nil,
																		},
																		Value: "alt",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	Body: &ast.LiteralBoolean{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(346),
																					Column: int(30),
																				},
																				End: ast.Location{
																					Line: int(346),
																					Column: int(35),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3884,
																			FreeVars: nil,
																		},
																		Value: false,
																	},
																	PlusSuper: false,
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(346),
																			Column: int(25),
																		},
																		End: ast.Location{
																			Line: int(346),
																			Column: int(35),
																		},
																		File: p1,
																	},
																},
																ast.DesugaredObjectField{
																	Hide: ast.ObjectFieldHide(1),
																	Name: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: nil,
																		},
																		Value: "zero",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	Body: &ast.LiteralBoolean{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(346),
																					Column: int(43),
																				},
																				End: ast.Location{
																					Line: int(346),
																					Column: int(48),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3884,
																			FreeVars: nil,
																		},
																		Value: false,
																	},
																	PlusSuper: false,
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(346),
																			Column: int(37),
																		},
																		End: ast.Location{
																			Line: int(346),
																			Column: int(48),
																		},
																		File: p1,
																	},
																},
																ast.DesugaredObjectField{
																	Hide: ast.ObjectFieldHide(1),
																	Name: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: nil,
																		},
																		Value: "left",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	Body: &ast.LiteralBoolean{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(346),
																					Column: int(56),
																				},
																				End: ast.Location{
																					Line: int(346),
																					Column: int(61),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3884,
																			FreeVars: nil,
																		},
																		Value: false,
																	},
																	PlusSuper: false,
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(346),
																			Column: int(50),
																		},
																		End: ast.Location{
																			Line: int(346),
																			Column: int(61),
																		},
																		File: p1,
																	},
																},
																ast.DesugaredObjectField{
																	Hide: ast.ObjectFieldHide(1),
																	Name: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: nil,
																		},
																		Value: "blank",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	Body: &ast.LiteralBoolean{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(346),
																					Column: int(70),
																				},
																				End: ast.Location{
																					Line: int(346),
																					Column: int(75),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3884,
																			FreeVars: nil,
																		},
																		Value: false,
																	},
																	PlusSuper: false,
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(346),
																			Column: int(63),
																		},
																		End: ast.Location{
																			Line: int(346),
																			Column: int(75),
																		},
																		File: p1,
																	},
																},
																ast.DesugaredObjectField{
																	Hide: ast.ObjectFieldHide(1),
																	Name: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: nil,
																		},
																		Value: "plus",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	Body: &ast.LiteralBoolean{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(346),
																					Column: int(83),
																				},
																				End: ast.Location{
																					Line: int(346),
																					Column: int(88),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3884,
																			FreeVars: nil,
																		},
																		Value: false,
																	},
																	PlusSuper: false,
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(346),
																			Column: int(77),
																		},
																		End: ast.Location{
																			Line: int(346),
																			Column: int(88),
																		},
																		File: p1,
																	},
																},
															},
															Locals: ast.LocalBinds{},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
									},
								},
								Fun: nil,
								CloseFodder: nil,
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
							},
						},
						Body: &ast.Local{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(348),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(774),
										Column: int(48),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(1),
										Indent: int(4),
										Comment: []string{},
									},
								},
								Ctx: p3497,
								FreeVars: ast.Identifiers{
									"std",
									"str",
									"try_parse_cflags",
									"try_parse_mapping_key",
									"vals",
								},
							},
							Binds: ast.LocalBinds{
								ast.LocalBind{
									VarFodder: nil,
									Variable: "try_parse_field_width",
									EqFodder: nil,
									Body: &ast.Function{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(348),
													Column: int(11),
												},
												End: ast.Location{
													Line: int(377),
													Column: int(27),
												},
												File: p1,
											},
											Fodder: nil,
											Ctx: p3898,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										ParenLeftFodder: ast.Fodder{},
										Parameters: []ast.Parameter{
											ast.Parameter{
												NameFodder: ast.Fodder{},
												Name: "str",
												EqFodder: nil,
												DefaultArg: nil,
												CommaFodder: ast.Fodder{},
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(348),
														Column: int(33),
													},
													End: ast.Location{
														Line: int(348),
														Column: int(36),
													},
													File: p1,
												},
											},
											ast.Parameter{
												NameFodder: ast.Fodder{},
												Name: "i",
												EqFodder: nil,
												DefaultArg: nil,
												CommaFodder: nil,
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(348),
														Column: int(38),
													},
													End: ast.Location{
														Line: int(348),
														Column: int(39),
													},
													File: p1,
												},
											},
										},
										TrailingComma: false,
										ParenRightFodder: ast.Fodder{},
										Body: &ast.Conditional{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(349),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(377),
														Column: int(27),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
														Comment: []string{},
													},
												},
												Ctx: p3903,
												FreeVars: ast.Identifiers{
													"i",
													"std",
													"str",
												},
											},
											Cond: &ast.Binary{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(349),
															Column: int(10),
														},
														End: ast.Location{
															Line: int(349),
															Column: int(46),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p3903,
													FreeVars: ast.Identifiers{
														"i",
														"std",
														"str",
													},
												},
												Left: &ast.Binary{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(349),
																Column: int(10),
															},
															End: ast.Location{
																Line: int(349),
																Column: int(29),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p3903,
														FreeVars: ast.Identifiers{
															"i",
															"std",
															"str",
														},
													},
													Left: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(349),
																	Column: int(10),
																},
																End: ast.Location{
																	Line: int(349),
																	Column: int(11),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p3903,
															FreeVars: ast.Identifiers{
																"i",
															},
														},
														Id: "i",
													},
													OpFodder: ast.Fodder{},
													Op: ast.BinaryOp(9),
													Right: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(349),
																	Column: int(14),
																},
																End: ast.Location{
																	Line: int(349),
																	Column: int(29),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p3903,
															FreeVars: ast.Identifiers{
																"std",
																"str",
															},
														},
														Target: &ast.Index{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(349),
																		Column: int(14),
																	},
																	End: ast.Location{
																		Line: int(349),
																		Column: int(24),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p3903,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Target: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(349),
																			Column: int(14),
																		},
																		End: ast.Location{
																			Line: int(349),
																			Column: int(17),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Id: "std",
															},
															LeftBracketFodder: ast.Fodder{},
															Index: &ast.LiteralString{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: nil,
																},
																Value: "length",
																Kind: ast.LiteralStringKind(1),
																BlockIndent: "",
																BlockTermIndent: "",
															},
															RightBracketFodder: ast.Fodder{},
															Id: nil,
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(349),
																					Column: int(25),
																				},
																				End: ast.Location{
																					Line: int(349),
																					Column: int(28),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3920,
																			FreeVars: ast.Identifiers{
																				"str",
																			},
																		},
																		Id: "str",
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														TrailingComma: false,
														TailStrict: false,
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
													},
												},
												OpFodder: ast.Fodder{},
												Op: ast.BinaryOp(17),
												Right: &ast.Binary{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(349),
																Column: int(33),
															},
															End: ast.Location{
																Line: int(349),
																Column: int(46),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p3903,
														FreeVars: ast.Identifiers{
															"i",
															"str",
														},
													},
													Left: &ast.Index{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(349),
																	Column: int(33),
																},
																End: ast.Location{
																	Line: int(349),
																	Column: int(39),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p3903,
															FreeVars: ast.Identifiers{
																"i",
																"str",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(349),
																		Column: int(33),
																	},
																	End: ast.Location{
																		Line: int(349),
																		Column: int(36),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p3903,
																FreeVars: ast.Identifiers{
																	"str",
																},
															},
															Id: "str",
														},
														LeftBracketFodder: ast.Fodder{},
														Index: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(349),
																		Column: int(37),
																	},
																	End: ast.Location{
																		Line: int(349),
																		Column: int(38),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p3903,
																FreeVars: ast.Identifiers{
																	"i",
																},
															},
															Id: "i",
														},
														RightBracketFodder: ast.Fodder{},
														Id: nil,
													},
													OpFodder: ast.Fodder{},
													Op: ast.BinaryOp(12),
													Right: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(349),
																	Column: int(43),
																},
																End: ast.Location{
																	Line: int(349),
																	Column: int(46),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p3903,
															FreeVars: nil,
														},
														Value: "*",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
												},
											},
											ThenFodder: ast.Fodder{},
											BranchTrue: &ast.DesugaredObject{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(350),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(350),
															Column: int(29),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
															Comment: []string{},
														},
													},
													Ctx: p3903,
													FreeVars: ast.Identifiers{
														"i",
													},
												},
												Asserts: ast.Nodes{},
												Fields: ast.DesugaredObjectFields{
													ast.DesugaredObjectField{
														Hide: ast.ObjectFieldHide(1),
														Name: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
															Value: "i",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														Body: &ast.Binary{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(350),
																		Column: int(14),
																	},
																	End: ast.Location{
																		Line: int(350),
																		Column: int(19),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p3937,
																FreeVars: ast.Identifiers{
																	"i",
																},
															},
															Left: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(350),
																			Column: int(14),
																		},
																		End: ast.Location{
																			Line: int(350),
																			Column: int(15),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p3937,
																	FreeVars: ast.Identifiers{
																		"i",
																	},
																},
																Id: "i",
															},
															OpFodder: ast.Fodder{},
															Op: ast.BinaryOp(3),
															Right: &ast.LiteralNumber{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(350),
																			Column: int(18),
																		},
																		End: ast.Location{
																			Line: int(350),
																			Column: int(19),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p3937,
																	FreeVars: nil,
																},
																OriginalString: "1",
															},
														},
														PlusSuper: false,
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(350),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(350),
																Column: int(19),
															},
															File: p1,
														},
													},
													ast.DesugaredObjectField{
														Hide: ast.ObjectFieldHide(1),
														Name: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
															Value: "v",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														Body: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(350),
																		Column: int(24),
																	},
																	End: ast.Location{
																		Line: int(350),
																		Column: int(27),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p3937,
																FreeVars: nil,
															},
															Value: "*",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														PlusSuper: false,
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(350),
																Column: int(21),
															},
															End: ast.Location{
																Line: int(350),
																Column: int(27),
															},
															File: p1,
														},
													},
												},
												Locals: ast.LocalBinds{},
											},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
													Comment: []string{},
												},
											},
											BranchFalse: &ast.Local{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(352),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(377),
															Column: int(27),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
															Comment: []string{},
														},
													},
													Ctx: p3903,
													FreeVars: ast.Identifiers{
														"i",
														"std",
														"str",
													},
												},
												Binds: ast.LocalBinds{
													ast.LocalBind{
														VarFodder: nil,
														Variable: "consume",
														EqFodder: nil,
														Body: &ast.Function{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(352),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(376),
																		Column: int(27),
																	},
																	File: p1,
																},
																Fodder: nil,
																Ctx: p3950,
																FreeVars: ast.Identifiers{
																	"consume",
																	"std",
																},
															},
															ParenLeftFodder: ast.Fodder{},
															Parameters: []ast.Parameter{
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "str",
																	EqFodder: nil,
																	DefaultArg: nil,
																	CommaFodder: ast.Fodder{},
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(352),
																			Column: int(23),
																		},
																		End: ast.Location{
																			Line: int(352),
																			Column: int(26),
																		},
																		File: p1,
																	},
																},
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "j",
																	EqFodder: nil,
																	DefaultArg: nil,
																	CommaFodder: ast.Fodder{},
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(352),
																			Column: int(28),
																		},
																		End: ast.Location{
																			Line: int(352),
																			Column: int(29),
																		},
																		File: p1,
																	},
																},
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "v",
																	EqFodder: nil,
																	DefaultArg: nil,
																	CommaFodder: nil,
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(352),
																			Column: int(31),
																		},
																		End: ast.Location{
																			Line: int(352),
																			Column: int(32),
																		},
																		File: p1,
																	},
																},
															},
															TrailingComma: false,
															ParenRightFodder: ast.Fodder{},
															Body: &ast.Conditional{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"consume",
																		"j",
																		"std",
																		"str",
																		"v",
																	},
																},
																Cond: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(353),
																				Column: int(18),
																			},
																			End: ast.Location{
																				Line: int(353),
																				Column: int(37),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p3956,
																		FreeVars: ast.Identifiers{
																			"j",
																			"std",
																			"str",
																		},
																	},
																	Left: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(353),
																					Column: int(18),
																				},
																				End: ast.Location{
																					Line: int(353),
																					Column: int(19),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3956,
																			FreeVars: ast.Identifiers{
																				"j",
																			},
																		},
																		Id: "j",
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(9),
																	Right: &ast.Apply{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(353),
																					Column: int(22),
																				},
																				End: ast.Location{
																					Line: int(353),
																					Column: int(37),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3956,
																			FreeVars: ast.Identifiers{
																				"std",
																				"str",
																			},
																		},
																		Target: &ast.Index{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(353),
																						Column: int(22),
																					},
																					End: ast.Location{
																						Line: int(353),
																						Column: int(32),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p3956,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(353),
																							Column: int(22),
																						},
																						End: ast.Location{
																							Line: int(353),
																							Column: int(25),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																				},
																				Id: "std",
																			},
																			LeftBracketFodder: ast.Fodder{},
																			Index: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: nil,
																				},
																				Value: "length",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																			RightBracketFodder: ast.Fodder{},
																			Id: nil,
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(353),
																									Column: int(33),
																								},
																								End: ast.Location{
																									Line: int(353),
																									Column: int(36),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p3969,
																							FreeVars: ast.Identifiers{
																								"str",
																							},
																						},
																						Id: "str",
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		TrailingComma: false,
																		TailStrict: false,
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																	},
																},
																ThenFodder: nil,
																BranchTrue: &ast.Local{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(354),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(376),
																				Column: int(27),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(10),
																				Comment: []string{},
																			},
																		},
																		Ctx: p3956,
																		FreeVars: ast.Identifiers{
																			"consume",
																			"j",
																			"str",
																			"v",
																		},
																	},
																	Binds: ast.LocalBinds{
																		ast.LocalBind{
																			VarFodder: ast.Fodder{},
																			Variable: "c",
																			EqFodder: ast.Fodder{},
																			Body: &ast.Index{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(354),
																							Column: int(21),
																						},
																						End: ast.Location{
																							Line: int(354),
																							Column: int(27),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p3976,
																					FreeVars: ast.Identifiers{
																						"j",
																						"str",
																					},
																				},
																				Target: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(354),
																								Column: int(21),
																							},
																							End: ast.Location{
																								Line: int(354),
																								Column: int(24),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p3976,
																						FreeVars: ast.Identifiers{
																							"str",
																						},
																					},
																					Id: "str",
																				},
																				LeftBracketFodder: ast.Fodder{},
																				Index: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(354),
																								Column: int(25),
																							},
																							End: ast.Location{
																								Line: int(354),
																								Column: int(26),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p3976,
																						FreeVars: ast.Identifiers{
																							"j",
																						},
																					},
																					Id: "j",
																				},
																				RightBracketFodder: ast.Fodder{},
																				Id: nil,
																			},
																			Fun: nil,
																			CloseFodder: ast.Fodder{},
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(354),
																					Column: int(17),
																				},
																				End: ast.Location{
																					Line: int(354),
																					Column: int(27),
																				},
																				File: p1,
																			},
																		},
																	},
																	Body: &ast.Conditional{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(355),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(376),
																					Column: int(27),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(10),
																					Comment: []string{},
																				},
																			},
																			Ctx: p3956,
																			FreeVars: ast.Identifiers{
																				"c",
																				"consume",
																				"j",
																				"str",
																				"v",
																			},
																		},
																		Cond: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(355),
																						Column: int(14),
																					},
																					End: ast.Location{
																						Line: int(355),
																						Column: int(22),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p3956,
																				FreeVars: ast.Identifiers{
																					"c",
																				},
																			},
																			Left: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(355),
																							Column: int(14),
																						},
																						End: ast.Location{
																							Line: int(355),
																							Column: int(15),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p3956,
																					FreeVars: ast.Identifiers{
																						"c",
																					},
																				},
																				Id: "c",
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(12),
																			Right: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(355),
																							Column: int(19),
																						},
																						End: ast.Location{
																							Line: int(355),
																							Column: int(22),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p3956,
																					FreeVars: nil,
																				},
																				Value: "0",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																		},
																		ThenFodder: ast.Fodder{},
																		BranchTrue: &ast.Apply{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(356),
																						Column: int(13),
																					},
																					End: ast.Location{
																						Line: int(356),
																						Column: int(44),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p3956,
																				FreeVars: ast.Identifiers{
																					"consume",
																					"j",
																					"str",
																					"v",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(356),
																							Column: int(13),
																						},
																						End: ast.Location{
																							Line: int(356),
																							Column: int(20),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(12),
																							Comment: []string{},
																						},
																					},
																					Ctx: p3956,
																					FreeVars: ast.Identifiers{
																						"consume",
																					},
																				},
																				Id: "consume",
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(356),
																										Column: int(21),
																									},
																									End: ast.Location{
																										Line: int(356),
																										Column: int(24),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p3997,
																								FreeVars: ast.Identifiers{
																									"str",
																								},
																							},
																							Id: "str",
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Binary{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(356),
																										Column: int(26),
																									},
																									End: ast.Location{
																										Line: int(356),
																										Column: int(31),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p3997,
																								FreeVars: ast.Identifiers{
																									"j",
																								},
																							},
																							Left: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(356),
																											Column: int(26),
																										},
																										End: ast.Location{
																											Line: int(356),
																											Column: int(27),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p3997,
																									FreeVars: ast.Identifiers{
																										"j",
																									},
																								},
																								Id: "j",
																							},
																							OpFodder: ast.Fodder{},
																							Op: ast.BinaryOp(3),
																							Right: &ast.LiteralNumber{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(356),
																											Column: int(30),
																										},
																										End: ast.Location{
																											Line: int(356),
																											Column: int(31),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p3997,
																									FreeVars: nil,
																								},
																								OriginalString: "1",
																							},
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Binary{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(356),
																										Column: int(33),
																									},
																									End: ast.Location{
																										Line: int(356),
																										Column: int(43),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p3997,
																								FreeVars: ast.Identifiers{
																									"v",
																								},
																							},
																							Left: &ast.Binary{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(356),
																											Column: int(33),
																										},
																										End: ast.Location{
																											Line: int(356),
																											Column: int(39),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p3997,
																									FreeVars: ast.Identifiers{
																										"v",
																									},
																								},
																								Left: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(356),
																												Column: int(33),
																											},
																											End: ast.Location{
																												Line: int(356),
																												Column: int(34),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p3997,
																										FreeVars: ast.Identifiers{
																											"v",
																										},
																									},
																									Id: "v",
																								},
																								OpFodder: ast.Fodder{},
																								Op: ast.BinaryOp(0),
																								Right: &ast.LiteralNumber{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(356),
																												Column: int(37),
																											},
																											End: ast.Location{
																												Line: int(356),
																												Column: int(39),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p3997,
																										FreeVars: nil,
																									},
																									OriginalString: "10",
																								},
																							},
																							OpFodder: ast.Fodder{},
																							Op: ast.BinaryOp(3),
																							Right: &ast.LiteralNumber{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(356),
																											Column: int(42),
																										},
																										End: ast.Location{
																											Line: int(356),
																											Column: int(43),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p3997,
																									FreeVars: nil,
																								},
																								OriginalString: "0",
																							},
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			TrailingComma: false,
																			TailStrict: false,
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																		},
																		ElseFodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(10),
																				Comment: []string{},
																			},
																		},
																		BranchFalse: &ast.Conditional{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(357),
																						Column: int(16),
																					},
																					End: ast.Location{
																						Line: int(376),
																						Column: int(27),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p3956,
																				FreeVars: ast.Identifiers{
																					"c",
																					"consume",
																					"j",
																					"str",
																					"v",
																				},
																			},
																			Cond: &ast.Binary{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(357),
																							Column: int(19),
																						},
																						End: ast.Location{
																							Line: int(357),
																							Column: int(27),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p3956,
																					FreeVars: ast.Identifiers{
																						"c",
																					},
																				},
																				Left: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(357),
																								Column: int(19),
																							},
																							End: ast.Location{
																								Line: int(357),
																								Column: int(20),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p3956,
																						FreeVars: ast.Identifiers{
																							"c",
																						},
																					},
																					Id: "c",
																				},
																				OpFodder: ast.Fodder{},
																				Op: ast.BinaryOp(12),
																				Right: &ast.LiteralString{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(357),
																								Column: int(24),
																							},
																							End: ast.Location{
																								Line: int(357),
																								Column: int(27),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p3956,
																						FreeVars: nil,
																					},
																					Value: "1",
																					Kind: ast.LiteralStringKind(1),
																					BlockIndent: "",
																					BlockTermIndent: "",
																				},
																			},
																			ThenFodder: ast.Fodder{},
																			BranchTrue: &ast.Apply{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(358),
																							Column: int(13),
																						},
																						End: ast.Location{
																							Line: int(358),
																							Column: int(44),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p3956,
																					FreeVars: ast.Identifiers{
																						"consume",
																						"j",
																						"str",
																						"v",
																					},
																				},
																				Target: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(358),
																								Column: int(13),
																							},
																							End: ast.Location{
																								Line: int(358),
																								Column: int(20),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(12),
																								Comment: []string{},
																							},
																						},
																						Ctx: p3956,
																						FreeVars: ast.Identifiers{
																							"consume",
																						},
																					},
																					Id: "consume",
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(358),
																											Column: int(21),
																										},
																										End: ast.Location{
																											Line: int(358),
																											Column: int(24),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p4027,
																									FreeVars: ast.Identifiers{
																										"str",
																									},
																								},
																								Id: "str",
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Binary{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(358),
																											Column: int(26),
																										},
																										End: ast.Location{
																											Line: int(358),
																											Column: int(31),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p4027,
																									FreeVars: ast.Identifiers{
																										"j",
																									},
																								},
																								Left: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(358),
																												Column: int(26),
																											},
																											End: ast.Location{
																												Line: int(358),
																												Column: int(27),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p4027,
																										FreeVars: ast.Identifiers{
																											"j",
																										},
																									},
																									Id: "j",
																								},
																								OpFodder: ast.Fodder{},
																								Op: ast.BinaryOp(3),
																								Right: &ast.LiteralNumber{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(358),
																												Column: int(30),
																											},
																											End: ast.Location{
																												Line: int(358),
																												Column: int(31),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p4027,
																										FreeVars: nil,
																									},
																									OriginalString: "1",
																								},
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Binary{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(358),
																											Column: int(33),
																										},
																										End: ast.Location{
																											Line: int(358),
																											Column: int(43),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p4027,
																									FreeVars: ast.Identifiers{
																										"v",
																									},
																								},
																								Left: &ast.Binary{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(358),
																												Column: int(33),
																											},
																											End: ast.Location{
																												Line: int(358),
																												Column: int(39),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p4027,
																										FreeVars: ast.Identifiers{
																											"v",
																										},
																									},
																									Left: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(358),
																													Column: int(33),
																												},
																												End: ast.Location{
																													Line: int(358),
																													Column: int(34),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p4027,
																											FreeVars: ast.Identifiers{
																												"v",
																											},
																										},
																										Id: "v",
																									},
																									OpFodder: ast.Fodder{},
																									Op: ast.BinaryOp(0),
																									Right: &ast.LiteralNumber{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(358),
																													Column: int(37),
																												},
																												End: ast.Location{
																													Line: int(358),
																													Column: int(39),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p4027,
																											FreeVars: nil,
																										},
																										OriginalString: "10",
																									},
																								},
																								OpFodder: ast.Fodder{},
																								Op: ast.BinaryOp(3),
																								Right: &ast.LiteralNumber{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(358),
																												Column: int(42),
																											},
																											End: ast.Location{
																												Line: int(358),
																												Column: int(43),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p4027,
																										FreeVars: nil,
																									},
																									OriginalString: "1",
																								},
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				TrailingComma: false,
																				TailStrict: false,
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: nil,
																			},
																			ElseFodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(10),
																					Comment: []string{},
																				},
																			},
																			BranchFalse: &ast.Conditional{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(359),
																							Column: int(16),
																						},
																						End: ast.Location{
																							Line: int(376),
																							Column: int(27),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p3956,
																					FreeVars: ast.Identifiers{
																						"c",
																						"consume",
																						"j",
																						"str",
																						"v",
																					},
																				},
																				Cond: &ast.Binary{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(359),
																								Column: int(19),
																							},
																							End: ast.Location{
																								Line: int(359),
																								Column: int(27),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p3956,
																						FreeVars: ast.Identifiers{
																							"c",
																						},
																					},
																					Left: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(359),
																									Column: int(19),
																								},
																								End: ast.Location{
																									Line: int(359),
																									Column: int(20),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p3956,
																							FreeVars: ast.Identifiers{
																								"c",
																							},
																						},
																						Id: "c",
																					},
																					OpFodder: ast.Fodder{},
																					Op: ast.BinaryOp(12),
																					Right: &ast.LiteralString{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(359),
																									Column: int(24),
																								},
																								End: ast.Location{
																									Line: int(359),
																									Column: int(27),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p3956,
																							FreeVars: nil,
																						},
																						Value: "2",
																						Kind: ast.LiteralStringKind(1),
																						BlockIndent: "",
																						BlockTermIndent: "",
																					},
																				},
																				ThenFodder: ast.Fodder{},
																				BranchTrue: &ast.Apply{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(360),
																								Column: int(13),
																							},
																							End: ast.Location{
																								Line: int(360),
																								Column: int(44),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p3956,
																						FreeVars: ast.Identifiers{
																							"consume",
																							"j",
																							"str",
																							"v",
																						},
																					},
																					Target: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(360),
																									Column: int(13),
																								},
																								End: ast.Location{
																									Line: int(360),
																									Column: int(20),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{
																								ast.FodderElement{
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(12),
																									Comment: []string{},
																								},
																							},
																							Ctx: p3956,
																							FreeVars: ast.Identifiers{
																								"consume",
																							},
																						},
																						Id: "consume",
																					},
																					FodderLeft: ast.Fodder{},
																					Arguments: ast.Arguments{
																						Positional: []ast.CommaSeparatedExpr{
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(360),
																												Column: int(21),
																											},
																											End: ast.Location{
																												Line: int(360),
																												Column: int(24),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p4057,
																										FreeVars: ast.Identifiers{
																											"str",
																										},
																									},
																									Id: "str",
																								},
																								CommaFodder: ast.Fodder{},
																							},
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Binary{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(360),
																												Column: int(26),
																											},
																											End: ast.Location{
																												Line: int(360),
																												Column: int(31),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p4057,
																										FreeVars: ast.Identifiers{
																											"j",
																										},
																									},
																									Left: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(360),
																													Column: int(26),
																												},
																												End: ast.Location{
																													Line: int(360),
																													Column: int(27),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p4057,
																											FreeVars: ast.Identifiers{
																												"j",
																											},
																										},
																										Id: "j",
																									},
																									OpFodder: ast.Fodder{},
																									Op: ast.BinaryOp(3),
																									Right: &ast.LiteralNumber{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(360),
																													Column: int(30),
																												},
																												End: ast.Location{
																													Line: int(360),
																													Column: int(31),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p4057,
																											FreeVars: nil,
																										},
																										OriginalString: "1",
																									},
																								},
																								CommaFodder: ast.Fodder{},
																							},
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Binary{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(360),
																												Column: int(33),
																											},
																											End: ast.Location{
																												Line: int(360),
																												Column: int(43),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p4057,
																										FreeVars: ast.Identifiers{
																											"v",
																										},
																									},
																									Left: &ast.Binary{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(360),
																													Column: int(33),
																												},
																												End: ast.Location{
																													Line: int(360),
																													Column: int(39),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p4057,
																											FreeVars: ast.Identifiers{
																												"v",
																											},
																										},
																										Left: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(360),
																														Column: int(33),
																													},
																													End: ast.Location{
																														Line: int(360),
																														Column: int(34),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p4057,
																												FreeVars: ast.Identifiers{
																													"v",
																												},
																											},
																											Id: "v",
																										},
																										OpFodder: ast.Fodder{},
																										Op: ast.BinaryOp(0),
																										Right: &ast.LiteralNumber{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(360),
																														Column: int(37),
																													},
																													End: ast.Location{
																														Line: int(360),
																														Column: int(39),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p4057,
																												FreeVars: nil,
																											},
																											OriginalString: "10",
																										},
																									},
																									OpFodder: ast.Fodder{},
																									Op: ast.BinaryOp(3),
																									Right: &ast.LiteralNumber{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(360),
																													Column: int(42),
																												},
																												End: ast.Location{
																													Line: int(360),
																													Column: int(43),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p4057,
																											FreeVars: nil,
																										},
																										OriginalString: "2",
																									},
																								},
																								CommaFodder: nil,
																							},
																						},
																						Named: nil,
																					},
																					TrailingComma: false,
																					TailStrict: false,
																					FodderRight: ast.Fodder{},
																					TailStrictFodder: nil,
																				},
																				ElseFodder: ast.Fodder{
																					ast.FodderElement{
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(10),
																						Comment: []string{},
																					},
																				},
																				BranchFalse: &ast.Conditional{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(361),
																								Column: int(16),
																							},
																							End: ast.Location{
																								Line: int(376),
																								Column: int(27),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p3956,
																						FreeVars: ast.Identifiers{
																							"c",
																							"consume",
																							"j",
																							"str",
																							"v",
																						},
																					},
																					Cond: &ast.Binary{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(361),
																									Column: int(19),
																								},
																								End: ast.Location{
																									Line: int(361),
																									Column: int(27),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p3956,
																							FreeVars: ast.Identifiers{
																								"c",
																							},
																						},
																						Left: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(361),
																										Column: int(19),
																									},
																									End: ast.Location{
																										Line: int(361),
																										Column: int(20),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p3956,
																								FreeVars: ast.Identifiers{
																									"c",
																								},
																							},
																							Id: "c",
																						},
																						OpFodder: ast.Fodder{},
																						Op: ast.BinaryOp(12),
																						Right: &ast.LiteralString{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(361),
																										Column: int(24),
																									},
																									End: ast.Location{
																										Line: int(361),
																										Column: int(27),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p3956,
																								FreeVars: nil,
																							},
																							Value: "3",
																							Kind: ast.LiteralStringKind(1),
																							BlockIndent: "",
																							BlockTermIndent: "",
																						},
																					},
																					ThenFodder: ast.Fodder{},
																					BranchTrue: &ast.Apply{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(362),
																									Column: int(13),
																								},
																								End: ast.Location{
																									Line: int(362),
																									Column: int(44),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p3956,
																							FreeVars: ast.Identifiers{
																								"consume",
																								"j",
																								"str",
																								"v",
																							},
																						},
																						Target: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(362),
																										Column: int(13),
																									},
																									End: ast.Location{
																										Line: int(362),
																										Column: int(20),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{
																									ast.FodderElement{
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(12),
																										Comment: []string{},
																									},
																								},
																								Ctx: p3956,
																								FreeVars: ast.Identifiers{
																									"consume",
																								},
																							},
																							Id: "consume",
																						},
																						FodderLeft: ast.Fodder{},
																						Arguments: ast.Arguments{
																							Positional: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(362),
																													Column: int(21),
																												},
																												End: ast.Location{
																													Line: int(362),
																													Column: int(24),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p4087,
																											FreeVars: ast.Identifiers{
																												"str",
																											},
																										},
																										Id: "str",
																									},
																									CommaFodder: ast.Fodder{},
																								},
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Binary{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(362),
																													Column: int(26),
																												},
																												End: ast.Location{
																													Line: int(362),
																													Column: int(31),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p4087,
																											FreeVars: ast.Identifiers{
																												"j",
																											},
																										},
																										Left: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(362),
																														Column: int(26),
																													},
																													End: ast.Location{
																														Line: int(362),
																														Column: int(27),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p4087,
																												FreeVars: ast.Identifiers{
																													"j",
																												},
																											},
																											Id: "j",
																										},
																										OpFodder: ast.Fodder{},
																										Op: ast.BinaryOp(3),
																										Right: &ast.LiteralNumber{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(362),
																														Column: int(30),
																													},
																													End: ast.Location{
																														Line: int(362),
																														Column: int(31),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p4087,
																												FreeVars: nil,
																											},
																											OriginalString: "1",
																										},
																									},
																									CommaFodder: ast.Fodder{},
																								},
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Binary{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(362),
																													Column: int(33),
																												},
																												End: ast.Location{
																													Line: int(362),
																													Column: int(43),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p4087,
																											FreeVars: ast.Identifiers{
																												"v",
																											},
																										},
																										Left: &ast.Binary{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(362),
																														Column: int(33),
																													},
																													End: ast.Location{
																														Line: int(362),
																														Column: int(39),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p4087,
																												FreeVars: ast.Identifiers{
																													"v",
																												},
																											},
																											Left: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(362),
																															Column: int(33),
																														},
																														End: ast.Location{
																															Line: int(362),
																															Column: int(34),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p4087,
																													FreeVars: ast.Identifiers{
																														"v",
																													},
																												},
																												Id: "v",
																											},
																											OpFodder: ast.Fodder{},
																											Op: ast.BinaryOp(0),
																											Right: &ast.LiteralNumber{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(362),
																															Column: int(37),
																														},
																														End: ast.Location{
																															Line: int(362),
																															Column: int(39),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p4087,
																													FreeVars: nil,
																												},
																												OriginalString: "10",
																											},
																										},
																										OpFodder: ast.Fodder{},
																										Op: ast.BinaryOp(3),
																										Right: &ast.LiteralNumber{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(362),
																														Column: int(42),
																													},
																													End: ast.Location{
																														Line: int(362),
																														Column: int(43),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p4087,
																												FreeVars: nil,
																											},
																											OriginalString: "3",
																										},
																									},
																									CommaFodder: nil,
																								},
																							},
																							Named: nil,
																						},
																						TrailingComma: false,
																						TailStrict: false,
																						FodderRight: ast.Fodder{},
																						TailStrictFodder: nil,
																					},
																					ElseFodder: ast.Fodder{
																						ast.FodderElement{
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(10),
																							Comment: []string{},
																						},
																					},
																					BranchFalse: &ast.Conditional{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(363),
																									Column: int(16),
																								},
																								End: ast.Location{
																									Line: int(376),
																									Column: int(27),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p3956,
																							FreeVars: ast.Identifiers{
																								"c",
																								"consume",
																								"j",
																								"str",
																								"v",
																							},
																						},
																						Cond: &ast.Binary{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(363),
																										Column: int(19),
																									},
																									End: ast.Location{
																										Line: int(363),
																										Column: int(27),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p3956,
																								FreeVars: ast.Identifiers{
																									"c",
																								},
																							},
																							Left: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(363),
																											Column: int(19),
																										},
																										End: ast.Location{
																											Line: int(363),
																											Column: int(20),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p3956,
																									FreeVars: ast.Identifiers{
																										"c",
																									},
																								},
																								Id: "c",
																							},
																							OpFodder: ast.Fodder{},
																							Op: ast.BinaryOp(12),
																							Right: &ast.LiteralString{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(363),
																											Column: int(24),
																										},
																										End: ast.Location{
																											Line: int(363),
																											Column: int(27),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p3956,
																									FreeVars: nil,
																								},
																								Value: "4",
																								Kind: ast.LiteralStringKind(1),
																								BlockIndent: "",
																								BlockTermIndent: "",
																							},
																						},
																						ThenFodder: ast.Fodder{},
																						BranchTrue: &ast.Apply{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(364),
																										Column: int(13),
																									},
																									End: ast.Location{
																										Line: int(364),
																										Column: int(44),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p3956,
																								FreeVars: ast.Identifiers{
																									"consume",
																									"j",
																									"str",
																									"v",
																								},
																							},
																							Target: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(364),
																											Column: int(13),
																										},
																										End: ast.Location{
																											Line: int(364),
																											Column: int(20),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{
																										ast.FodderElement{
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(12),
																											Comment: []string{},
																										},
																									},
																									Ctx: p3956,
																									FreeVars: ast.Identifiers{
																										"consume",
																									},
																								},
																								Id: "consume",
																							},
																							FodderLeft: ast.Fodder{},
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(364),
																														Column: int(21),
																													},
																													End: ast.Location{
																														Line: int(364),
																														Column: int(24),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p4117,
																												FreeVars: ast.Identifiers{
																													"str",
																												},
																											},
																											Id: "str",
																										},
																										CommaFodder: ast.Fodder{},
																									},
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Binary{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(364),
																														Column: int(26),
																													},
																													End: ast.Location{
																														Line: int(364),
																														Column: int(31),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p4117,
																												FreeVars: ast.Identifiers{
																													"j",
																												},
																											},
																											Left: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(364),
																															Column: int(26),
																														},
																														End: ast.Location{
																															Line: int(364),
																															Column: int(27),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p4117,
																													FreeVars: ast.Identifiers{
																														"j",
																													},
																												},
																												Id: "j",
																											},
																											OpFodder: ast.Fodder{},
																											Op: ast.BinaryOp(3),
																											Right: &ast.LiteralNumber{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(364),
																															Column: int(30),
																														},
																														End: ast.Location{
																															Line: int(364),
																															Column: int(31),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p4117,
																													FreeVars: nil,
																												},
																												OriginalString: "1",
																											},
																										},
																										CommaFodder: ast.Fodder{},
																									},
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Binary{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(364),
																														Column: int(33),
																													},
																													End: ast.Location{
																														Line: int(364),
																														Column: int(43),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p4117,
																												FreeVars: ast.Identifiers{
																													"v",
																												},
																											},
																											Left: &ast.Binary{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(364),
																															Column: int(33),
																														},
																														End: ast.Location{
																															Line: int(364),
																															Column: int(39),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p4117,
																													FreeVars: ast.Identifiers{
																														"v",
																													},
																												},
																												Left: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(364),
																																Column: int(33),
																															},
																															End: ast.Location{
																																Line: int(364),
																																Column: int(34),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p4117,
																														FreeVars: ast.Identifiers{
																															"v",
																														},
																													},
																													Id: "v",
																												},
																												OpFodder: ast.Fodder{},
																												Op: ast.BinaryOp(0),
																												Right: &ast.LiteralNumber{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(364),
																																Column: int(37),
																															},
																															End: ast.Location{
																																Line: int(364),
																																Column: int(39),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p4117,
																														FreeVars: nil,
																													},
																													OriginalString: "10",
																												},
																											},
																											OpFodder: ast.Fodder{},
																											Op: ast.BinaryOp(3),
																											Right: &ast.LiteralNumber{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(364),
																															Column: int(42),
																														},
																														End: ast.Location{
																															Line: int(364),
																															Column: int(43),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p4117,
																													FreeVars: nil,
																												},
																												OriginalString: "4",
																											},
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							TrailingComma: false,
																							TailStrict: false,
																							FodderRight: ast.Fodder{},
																							TailStrictFodder: nil,
																						},
																						ElseFodder: ast.Fodder{
																							ast.FodderElement{
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(10),
																								Comment: []string{},
																							},
																						},
																						BranchFalse: &ast.Conditional{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(365),
																										Column: int(16),
																									},
																									End: ast.Location{
																										Line: int(376),
																										Column: int(27),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p3956,
																								FreeVars: ast.Identifiers{
																									"c",
																									"consume",
																									"j",
																									"str",
																									"v",
																								},
																							},
																							Cond: &ast.Binary{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(365),
																											Column: int(19),
																										},
																										End: ast.Location{
																											Line: int(365),
																											Column: int(27),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p3956,
																									FreeVars: ast.Identifiers{
																										"c",
																									},
																								},
																								Left: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(365),
																												Column: int(19),
																											},
																											End: ast.Location{
																												Line: int(365),
																												Column: int(20),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p3956,
																										FreeVars: ast.Identifiers{
																											"c",
																										},
																									},
																									Id: "c",
																								},
																								OpFodder: ast.Fodder{},
																								Op: ast.BinaryOp(12),
																								Right: &ast.LiteralString{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(365),
																												Column: int(24),
																											},
																											End: ast.Location{
																												Line: int(365),
																												Column: int(27),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p3956,
																										FreeVars: nil,
																									},
																									Value: "5",
																									Kind: ast.LiteralStringKind(1),
																									BlockIndent: "",
																									BlockTermIndent: "",
																								},
																							},
																							ThenFodder: ast.Fodder{},
																							BranchTrue: &ast.Apply{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(366),
																											Column: int(13),
																										},
																										End: ast.Location{
																											Line: int(366),
																											Column: int(44),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p3956,
																									FreeVars: ast.Identifiers{
																										"consume",
																										"j",
																										"str",
																										"v",
																									},
																								},
																								Target: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(366),
																												Column: int(13),
																											},
																											End: ast.Location{
																												Line: int(366),
																												Column: int(20),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{
																											ast.FodderElement{
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(12),
																												Comment: []string{},
																											},
																										},
																										Ctx: p3956,
																										FreeVars: ast.Identifiers{
																											"consume",
																										},
																									},
																									Id: "consume",
																								},
																								FodderLeft: ast.Fodder{},
																								Arguments: ast.Arguments{
																									Positional: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(366),
																															Column: int(21),
																														},
																														End: ast.Location{
																															Line: int(366),
																															Column: int(24),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p4147,
																													FreeVars: ast.Identifiers{
																														"str",
																													},
																												},
																												Id: "str",
																											},
																											CommaFodder: ast.Fodder{},
																										},
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Binary{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(366),
																															Column: int(26),
																														},
																														End: ast.Location{
																															Line: int(366),
																															Column: int(31),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p4147,
																													FreeVars: ast.Identifiers{
																														"j",
																													},
																												},
																												Left: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(366),
																																Column: int(26),
																															},
																															End: ast.Location{
																																Line: int(366),
																																Column: int(27),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p4147,
																														FreeVars: ast.Identifiers{
																															"j",
																														},
																													},
																													Id: "j",
																												},
																												OpFodder: ast.Fodder{},
																												Op: ast.BinaryOp(3),
																												Right: &ast.LiteralNumber{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(366),
																																Column: int(30),
																															},
																															End: ast.Location{
																																Line: int(366),
																																Column: int(31),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p4147,
																														FreeVars: nil,
																													},
																													OriginalString: "1",
																												},
																											},
																											CommaFodder: ast.Fodder{},
																										},
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Binary{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(366),
																															Column: int(33),
																														},
																														End: ast.Location{
																															Line: int(366),
																															Column: int(43),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p4147,
																													FreeVars: ast.Identifiers{
																														"v",
																													},
																												},
																												Left: &ast.Binary{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(366),
																																Column: int(33),
																															},
																															End: ast.Location{
																																Line: int(366),
																																Column: int(39),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p4147,
																														FreeVars: ast.Identifiers{
																															"v",
																														},
																													},
																													Left: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(366),
																																	Column: int(33),
																																},
																																End: ast.Location{
																																	Line: int(366),
																																	Column: int(34),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p4147,
																															FreeVars: ast.Identifiers{
																																"v",
																															},
																														},
																														Id: "v",
																													},
																													OpFodder: ast.Fodder{},
																													Op: ast.BinaryOp(0),
																													Right: &ast.LiteralNumber{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(366),
																																	Column: int(37),
																																},
																																End: ast.Location{
																																	Line: int(366),
																																	Column: int(39),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p4147,
																															FreeVars: nil,
																														},
																														OriginalString: "10",
																													},
																												},
																												OpFodder: ast.Fodder{},
																												Op: ast.BinaryOp(3),
																												Right: &ast.LiteralNumber{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(366),
																																Column: int(42),
																															},
																															End: ast.Location{
																																Line: int(366),
																																Column: int(43),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p4147,
																														FreeVars: nil,
																													},
																													OriginalString: "5",
																												},
																											},
																											CommaFodder: nil,
																										},
																									},
																									Named: nil,
																								},
																								TrailingComma: false,
																								TailStrict: false,
																								FodderRight: ast.Fodder{},
																								TailStrictFodder: nil,
																							},
																							ElseFodder: ast.Fodder{
																								ast.FodderElement{
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(10),
																									Comment: []string{},
																								},
																							},
																							BranchFalse: &ast.Conditional{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(367),
																											Column: int(16),
																										},
																										End: ast.Location{
																											Line: int(376),
																											Column: int(27),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p3956,
																									FreeVars: ast.Identifiers{
																										"c",
																										"consume",
																										"j",
																										"str",
																										"v",
																									},
																								},
																								Cond: &ast.Binary{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(367),
																												Column: int(19),
																											},
																											End: ast.Location{
																												Line: int(367),
																												Column: int(27),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p3956,
																										FreeVars: ast.Identifiers{
																											"c",
																										},
																									},
																									Left: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(367),
																													Column: int(19),
																												},
																												End: ast.Location{
																													Line: int(367),
																													Column: int(20),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p3956,
																											FreeVars: ast.Identifiers{
																												"c",
																											},
																										},
																										Id: "c",
																									},
																									OpFodder: ast.Fodder{},
																									Op: ast.BinaryOp(12),
																									Right: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(367),
																													Column: int(24),
																												},
																												End: ast.Location{
																													Line: int(367),
																													Column: int(27),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p3956,
																											FreeVars: nil,
																										},
																										Value: "6",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																								},
																								ThenFodder: ast.Fodder{},
																								BranchTrue: &ast.Apply{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(368),
																												Column: int(13),
																											},
																											End: ast.Location{
																												Line: int(368),
																												Column: int(44),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p3956,
																										FreeVars: ast.Identifiers{
																											"consume",
																											"j",
																											"str",
																											"v",
																										},
																									},
																									Target: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(368),
																													Column: int(13),
																												},
																												End: ast.Location{
																													Line: int(368),
																													Column: int(20),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{
																												ast.FodderElement{
																													Kind: ast.FodderKind(0),
																													Blanks: int(0),
																													Indent: int(12),
																													Comment: []string{},
																												},
																											},
																											Ctx: p3956,
																											FreeVars: ast.Identifiers{
																												"consume",
																											},
																										},
																										Id: "consume",
																									},
																									FodderLeft: ast.Fodder{},
																									Arguments: ast.Arguments{
																										Positional: []ast.CommaSeparatedExpr{
																											ast.CommaSeparatedExpr{
																												Expr: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(368),
																																Column: int(21),
																															},
																															End: ast.Location{
																																Line: int(368),
																																Column: int(24),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p4177,
																														FreeVars: ast.Identifiers{
																															"str",
																														},
																													},
																													Id: "str",
																												},
																												CommaFodder: ast.Fodder{},
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.Binary{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(368),
																																Column: int(26),
																															},
																															End: ast.Location{
																																Line: int(368),
																																Column: int(31),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p4177,
																														FreeVars: ast.Identifiers{
																															"j",
																														},
																													},
																													Left: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(368),
																																	Column: int(26),
																																},
																																End: ast.Location{
																																	Line: int(368),
																																	Column: int(27),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p4177,
																															FreeVars: ast.Identifiers{
																																"j",
																															},
																														},
																														Id: "j",
																													},
																													OpFodder: ast.Fodder{},
																													Op: ast.BinaryOp(3),
																													Right: &ast.LiteralNumber{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(368),
																																	Column: int(30),
																																},
																																End: ast.Location{
																																	Line: int(368),
																																	Column: int(31),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p4177,
																															FreeVars: nil,
																														},
																														OriginalString: "1",
																													},
																												},
																												CommaFodder: ast.Fodder{},
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.Binary{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(368),
																																Column: int(33),
																															},
																															End: ast.Location{
																																Line: int(368),
																																Column: int(43),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p4177,
																														FreeVars: ast.Identifiers{
																															"v",
																														},
																													},
																													Left: &ast.Binary{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(368),
																																	Column: int(33),
																																},
																																End: ast.Location{
																																	Line: int(368),
																																	Column: int(39),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p4177,
																															FreeVars: ast.Identifiers{
																																"v",
																															},
																														},
																														Left: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(368),
																																		Column: int(33),
																																	},
																																	End: ast.Location{
																																		Line: int(368),
																																		Column: int(34),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p4177,
																																FreeVars: ast.Identifiers{
																																	"v",
																																},
																															},
																															Id: "v",
																														},
																														OpFodder: ast.Fodder{},
																														Op: ast.BinaryOp(0),
																														Right: &ast.LiteralNumber{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(368),
																																		Column: int(37),
																																	},
																																	End: ast.Location{
																																		Line: int(368),
																																		Column: int(39),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p4177,
																																FreeVars: nil,
																															},
																															OriginalString: "10",
																														},
																													},
																													OpFodder: ast.Fodder{},
																													Op: ast.BinaryOp(3),
																													Right: &ast.LiteralNumber{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(368),
																																	Column: int(42),
																																},
																																End: ast.Location{
																																	Line: int(368),
																																	Column: int(43),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p4177,
																															FreeVars: nil,
																														},
																														OriginalString: "6",
																													},
																												},
																												CommaFodder: nil,
																											},
																										},
																										Named: nil,
																									},
																									TrailingComma: false,
																									TailStrict: false,
																									FodderRight: ast.Fodder{},
																									TailStrictFodder: nil,
																								},
																								ElseFodder: ast.Fodder{
																									ast.FodderElement{
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(10),
																										Comment: []string{},
																									},
																								},
																								BranchFalse: &ast.Conditional{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(369),
																												Column: int(16),
																											},
																											End: ast.Location{
																												Line: int(376),
																												Column: int(27),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p3956,
																										FreeVars: ast.Identifiers{
																											"c",
																											"consume",
																											"j",
																											"str",
																											"v",
																										},
																									},
																									Cond: &ast.Binary{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(369),
																													Column: int(19),
																												},
																												End: ast.Location{
																													Line: int(369),
																													Column: int(27),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p3956,
																											FreeVars: ast.Identifiers{
																												"c",
																											},
																										},
																										Left: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(369),
																														Column: int(19),
																													},
																													End: ast.Location{
																														Line: int(369),
																														Column: int(20),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p3956,
																												FreeVars: ast.Identifiers{
																													"c",
																												},
																											},
																											Id: "c",
																										},
																										OpFodder: ast.Fodder{},
																										Op: ast.BinaryOp(12),
																										Right: &ast.LiteralString{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(369),
																														Column: int(24),
																													},
																													End: ast.Location{
																														Line: int(369),
																														Column: int(27),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p3956,
																												FreeVars: nil,
																											},
																											Value: "7",
																											Kind: ast.LiteralStringKind(1),
																											BlockIndent: "",
																											BlockTermIndent: "",
																										},
																									},
																									ThenFodder: ast.Fodder{},
																									BranchTrue: &ast.Apply{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(370),
																													Column: int(13),
																												},
																												End: ast.Location{
																													Line: int(370),
																													Column: int(44),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p3956,
																											FreeVars: ast.Identifiers{
																												"consume",
																												"j",
																												"str",
																												"v",
																											},
																										},
																										Target: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(370),
																														Column: int(13),
																													},
																													End: ast.Location{
																														Line: int(370),
																														Column: int(20),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{
																													ast.FodderElement{
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(12),
																														Comment: []string{},
																													},
																												},
																												Ctx: p3956,
																												FreeVars: ast.Identifiers{
																													"consume",
																												},
																											},
																											Id: "consume",
																										},
																										FodderLeft: ast.Fodder{},
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(370),
																																	Column: int(21),
																																},
																																End: ast.Location{
																																	Line: int(370),
																																	Column: int(24),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p4207,
																															FreeVars: ast.Identifiers{
																																"str",
																															},
																														},
																														Id: "str",
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Binary{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(370),
																																	Column: int(26),
																																},
																																End: ast.Location{
																																	Line: int(370),
																																	Column: int(31),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p4207,
																															FreeVars: ast.Identifiers{
																																"j",
																															},
																														},
																														Left: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(370),
																																		Column: int(26),
																																	},
																																	End: ast.Location{
																																		Line: int(370),
																																		Column: int(27),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p4207,
																																FreeVars: ast.Identifiers{
																																	"j",
																																},
																															},
																															Id: "j",
																														},
																														OpFodder: ast.Fodder{},
																														Op: ast.BinaryOp(3),
																														Right: &ast.LiteralNumber{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(370),
																																		Column: int(30),
																																	},
																																	End: ast.Location{
																																		Line: int(370),
																																		Column: int(31),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p4207,
																																FreeVars: nil,
																															},
																															OriginalString: "1",
																														},
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Binary{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(370),
																																	Column: int(33),
																																},
																																End: ast.Location{
																																	Line: int(370),
																																	Column: int(43),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p4207,
																															FreeVars: ast.Identifiers{
																																"v",
																															},
																														},
																														Left: &ast.Binary{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(370),
																																		Column: int(33),
																																	},
																																	End: ast.Location{
																																		Line: int(370),
																																		Column: int(39),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p4207,
																																FreeVars: ast.Identifiers{
																																	"v",
																																},
																															},
																															Left: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(370),
																																			Column: int(33),
																																		},
																																		End: ast.Location{
																																			Line: int(370),
																																			Column: int(34),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p4207,
																																	FreeVars: ast.Identifiers{
																																		"v",
																																	},
																																},
																																Id: "v",
																															},
																															OpFodder: ast.Fodder{},
																															Op: ast.BinaryOp(0),
																															Right: &ast.LiteralNumber{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(370),
																																			Column: int(37),
																																		},
																																		End: ast.Location{
																																			Line: int(370),
																																			Column: int(39),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p4207,
																																	FreeVars: nil,
																																},
																																OriginalString: "10",
																															},
																														},
																														OpFodder: ast.Fodder{},
																														Op: ast.BinaryOp(3),
																														Right: &ast.LiteralNumber{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(370),
																																		Column: int(42),
																																	},
																																	End: ast.Location{
																																		Line: int(370),
																																		Column: int(43),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p4207,
																																FreeVars: nil,
																															},
																															OriginalString: "7",
																														},
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										TrailingComma: false,
																										TailStrict: false,
																										FodderRight: ast.Fodder{},
																										TailStrictFodder: nil,
																									},
																									ElseFodder: ast.Fodder{
																										ast.FodderElement{
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(10),
																											Comment: []string{},
																										},
																									},
																									BranchFalse: &ast.Conditional{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(371),
																													Column: int(16),
																												},
																												End: ast.Location{
																													Line: int(376),
																													Column: int(27),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p3956,
																											FreeVars: ast.Identifiers{
																												"c",
																												"consume",
																												"j",
																												"str",
																												"v",
																											},
																										},
																										Cond: &ast.Binary{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(371),
																														Column: int(19),
																													},
																													End: ast.Location{
																														Line: int(371),
																														Column: int(27),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p3956,
																												FreeVars: ast.Identifiers{
																													"c",
																												},
																											},
																											Left: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(371),
																															Column: int(19),
																														},
																														End: ast.Location{
																															Line: int(371),
																															Column: int(20),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p3956,
																													FreeVars: ast.Identifiers{
																														"c",
																													},
																												},
																												Id: "c",
																											},
																											OpFodder: ast.Fodder{},
																											Op: ast.BinaryOp(12),
																											Right: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(371),
																															Column: int(24),
																														},
																														End: ast.Location{
																															Line: int(371),
																															Column: int(27),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p3956,
																													FreeVars: nil,
																												},
																												Value: "8",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																										},
																										ThenFodder: ast.Fodder{},
																										BranchTrue: &ast.Apply{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(372),
																														Column: int(13),
																													},
																													End: ast.Location{
																														Line: int(372),
																														Column: int(44),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p3956,
																												FreeVars: ast.Identifiers{
																													"consume",
																													"j",
																													"str",
																													"v",
																												},
																											},
																											Target: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(372),
																															Column: int(13),
																														},
																														End: ast.Location{
																															Line: int(372),
																															Column: int(20),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(12),
																															Comment: []string{},
																														},
																													},
																													Ctx: p3956,
																													FreeVars: ast.Identifiers{
																														"consume",
																													},
																												},
																												Id: "consume",
																											},
																											FodderLeft: ast.Fodder{},
																											Arguments: ast.Arguments{
																												Positional: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(372),
																																		Column: int(21),
																																	},
																																	End: ast.Location{
																																		Line: int(372),
																																		Column: int(24),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p4237,
																																FreeVars: ast.Identifiers{
																																	"str",
																																},
																															},
																															Id: "str",
																														},
																														CommaFodder: ast.Fodder{},
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Binary{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(372),
																																		Column: int(26),
																																	},
																																	End: ast.Location{
																																		Line: int(372),
																																		Column: int(31),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p4237,
																																FreeVars: ast.Identifiers{
																																	"j",
																																},
																															},
																															Left: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(372),
																																			Column: int(26),
																																		},
																																		End: ast.Location{
																																			Line: int(372),
																																			Column: int(27),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p4237,
																																	FreeVars: ast.Identifiers{
																																		"j",
																																	},
																																},
																																Id: "j",
																															},
																															OpFodder: ast.Fodder{},
																															Op: ast.BinaryOp(3),
																															Right: &ast.LiteralNumber{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(372),
																																			Column: int(30),
																																		},
																																		End: ast.Location{
																																			Line: int(372),
																																			Column: int(31),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p4237,
																																	FreeVars: nil,
																																},
																																OriginalString: "1",
																															},
																														},
																														CommaFodder: ast.Fodder{},
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Binary{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(372),
																																		Column: int(33),
																																	},
																																	End: ast.Location{
																																		Line: int(372),
																																		Column: int(43),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p4237,
																																FreeVars: ast.Identifiers{
																																	"v",
																																},
																															},
																															Left: &ast.Binary{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(372),
																																			Column: int(33),
																																		},
																																		End: ast.Location{
																																			Line: int(372),
																																			Column: int(39),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p4237,
																																	FreeVars: ast.Identifiers{
																																		"v",
																																	},
																																},
																																Left: &ast.Var{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(372),
																																				Column: int(33),
																																			},
																																			End: ast.Location{
																																				Line: int(372),
																																				Column: int(34),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p4237,
																																		FreeVars: ast.Identifiers{
																																			"v",
																																		},
																																	},
																																	Id: "v",
																																},
																																OpFodder: ast.Fodder{},
																																Op: ast.BinaryOp(0),
																																Right: &ast.LiteralNumber{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(372),
																																				Column: int(37),
																																			},
																																			End: ast.Location{
																																				Line: int(372),
																																				Column: int(39),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p4237,
																																		FreeVars: nil,
																																	},
																																	OriginalString: "10",
																																},
																															},
																															OpFodder: ast.Fodder{},
																															Op: ast.BinaryOp(3),
																															Right: &ast.LiteralNumber{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(372),
																																			Column: int(42),
																																		},
																																		End: ast.Location{
																																			Line: int(372),
																																			Column: int(43),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p4237,
																																	FreeVars: nil,
																																},
																																OriginalString: "8",
																															},
																														},
																														CommaFodder: nil,
																													},
																												},
																												Named: nil,
																											},
																											TrailingComma: false,
																											TailStrict: false,
																											FodderRight: ast.Fodder{},
																											TailStrictFodder: nil,
																										},
																										ElseFodder: ast.Fodder{
																											ast.FodderElement{
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(10),
																												Comment: []string{},
																											},
																										},
																										BranchFalse: &ast.Conditional{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(373),
																														Column: int(16),
																													},
																													End: ast.Location{
																														Line: int(376),
																														Column: int(27),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p3956,
																												FreeVars: ast.Identifiers{
																													"c",
																													"consume",
																													"j",
																													"str",
																													"v",
																												},
																											},
																											Cond: &ast.Binary{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(373),
																															Column: int(19),
																														},
																														End: ast.Location{
																															Line: int(373),
																															Column: int(27),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p3956,
																													FreeVars: ast.Identifiers{
																														"c",
																													},
																												},
																												Left: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(373),
																																Column: int(19),
																															},
																															End: ast.Location{
																																Line: int(373),
																																Column: int(20),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p3956,
																														FreeVars: ast.Identifiers{
																															"c",
																														},
																													},
																													Id: "c",
																												},
																												OpFodder: ast.Fodder{},
																												Op: ast.BinaryOp(12),
																												Right: &ast.LiteralString{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(373),
																																Column: int(24),
																															},
																															End: ast.Location{
																																Line: int(373),
																																Column: int(27),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p3956,
																														FreeVars: nil,
																													},
																													Value: "9",
																													Kind: ast.LiteralStringKind(1),
																													BlockIndent: "",
																													BlockTermIndent: "",
																												},
																											},
																											ThenFodder: ast.Fodder{},
																											BranchTrue: &ast.Apply{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(374),
																															Column: int(13),
																														},
																														End: ast.Location{
																															Line: int(374),
																															Column: int(44),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p3956,
																													FreeVars: ast.Identifiers{
																														"consume",
																														"j",
																														"str",
																														"v",
																													},
																												},
																												Target: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(374),
																																Column: int(13),
																															},
																															End: ast.Location{
																																Line: int(374),
																																Column: int(20),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(12),
																																Comment: []string{},
																															},
																														},
																														Ctx: p3956,
																														FreeVars: ast.Identifiers{
																															"consume",
																														},
																													},
																													Id: "consume",
																												},
																												FodderLeft: ast.Fodder{},
																												Arguments: ast.Arguments{
																													Positional: []ast.CommaSeparatedExpr{
																														ast.CommaSeparatedExpr{
																															Expr: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(374),
																																			Column: int(21),
																																		},
																																		End: ast.Location{
																																			Line: int(374),
																																			Column: int(24),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p4267,
																																	FreeVars: ast.Identifiers{
																																		"str",
																																	},
																																},
																																Id: "str",
																															},
																															CommaFodder: ast.Fodder{},
																														},
																														ast.CommaSeparatedExpr{
																															Expr: &ast.Binary{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(374),
																																			Column: int(26),
																																		},
																																		End: ast.Location{
																																			Line: int(374),
																																			Column: int(31),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p4267,
																																	FreeVars: ast.Identifiers{
																																		"j",
																																	},
																																},
																																Left: &ast.Var{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(374),
																																				Column: int(26),
																																			},
																																			End: ast.Location{
																																				Line: int(374),
																																				Column: int(27),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p4267,
																																		FreeVars: ast.Identifiers{
																																			"j",
																																		},
																																	},
																																	Id: "j",
																																},
																																OpFodder: ast.Fodder{},
																																Op: ast.BinaryOp(3),
																																Right: &ast.LiteralNumber{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(374),
																																				Column: int(30),
																																			},
																																			End: ast.Location{
																																				Line: int(374),
																																				Column: int(31),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p4267,
																																		FreeVars: nil,
																																	},
																																	OriginalString: "1",
																																},
																															},
																															CommaFodder: ast.Fodder{},
																														},
																														ast.CommaSeparatedExpr{
																															Expr: &ast.Binary{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(374),
																																			Column: int(33),
																																		},
																																		End: ast.Location{
																																			Line: int(374),
																																			Column: int(43),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p4267,
																																	FreeVars: ast.Identifiers{
																																		"v",
																																	},
																																},
																																Left: &ast.Binary{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(374),
																																				Column: int(33),
																																			},
																																			End: ast.Location{
																																				Line: int(374),
																																				Column: int(39),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p4267,
																																		FreeVars: ast.Identifiers{
																																			"v",
																																		},
																																	},
																																	Left: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(374),
																																					Column: int(33),
																																				},
																																				End: ast.Location{
																																					Line: int(374),
																																					Column: int(34),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p4267,
																																			FreeVars: ast.Identifiers{
																																				"v",
																																			},
																																		},
																																		Id: "v",
																																	},
																																	OpFodder: ast.Fodder{},
																																	Op: ast.BinaryOp(0),
																																	Right: &ast.LiteralNumber{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(374),
																																					Column: int(37),
																																				},
																																				End: ast.Location{
																																					Line: int(374),
																																					Column: int(39),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p4267,
																																			FreeVars: nil,
																																		},
																																		OriginalString: "10",
																																	},
																																},
																																OpFodder: ast.Fodder{},
																																Op: ast.BinaryOp(3),
																																Right: &ast.LiteralNumber{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(374),
																																				Column: int(42),
																																			},
																																			End: ast.Location{
																																				Line: int(374),
																																				Column: int(43),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p4267,
																																		FreeVars: nil,
																																	},
																																	OriginalString: "9",
																																},
																															},
																															CommaFodder: nil,
																														},
																													},
																													Named: nil,
																												},
																												TrailingComma: false,
																												TailStrict: false,
																												FodderRight: ast.Fodder{},
																												TailStrictFodder: nil,
																											},
																											ElseFodder: ast.Fodder{
																												ast.FodderElement{
																													Kind: ast.FodderKind(0),
																													Blanks: int(0),
																													Indent: int(10),
																													Comment: []string{},
																												},
																											},
																											BranchFalse: &ast.DesugaredObject{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(376),
																															Column: int(13),
																														},
																														End: ast.Location{
																															Line: int(376),
																															Column: int(27),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(12),
																															Comment: []string{},
																														},
																													},
																													Ctx: p3956,
																													FreeVars: ast.Identifiers{
																														"j",
																														"v",
																													},
																												},
																												Asserts: ast.Nodes{},
																												Fields: ast.DesugaredObjectFields{
																													ast.DesugaredObjectField{
																														Hide: ast.ObjectFieldHide(1),
																														Name: &ast.LiteralString{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	File: nil,
																																},
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: nil,
																															},
																															Value: "i",
																															Kind: ast.LiteralStringKind(1),
																															BlockIndent: "",
																															BlockTermIndent: "",
																														},
																														Body: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(376),
																																		Column: int(18),
																																	},
																																	End: ast.Location{
																																		Line: int(376),
																																		Column: int(19),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p4289,
																																FreeVars: ast.Identifiers{
																																	"j",
																																},
																															},
																															Id: "j",
																														},
																														PlusSuper: false,
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(376),
																																Column: int(15),
																															},
																															End: ast.Location{
																																Line: int(376),
																																Column: int(19),
																															},
																															File: p1,
																														},
																													},
																													ast.DesugaredObjectField{
																														Hide: ast.ObjectFieldHide(1),
																														Name: &ast.LiteralString{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	File: nil,
																																},
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: nil,
																															},
																															Value: "v",
																															Kind: ast.LiteralStringKind(1),
																															BlockIndent: "",
																															BlockTermIndent: "",
																														},
																														Body: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(376),
																																		Column: int(24),
																																	},
																																	End: ast.Location{
																																		Line: int(376),
																																		Column: int(25),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p4289,
																																FreeVars: ast.Identifiers{
																																	"v",
																																},
																															},
																															Id: "v",
																														},
																														PlusSuper: false,
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(376),
																																Column: int(21),
																															},
																															End: ast.Location{
																																Line: int(376),
																																Column: int(25),
																															},
																															File: p1,
																														},
																													},
																												},
																												Locals: ast.LocalBinds{},
																											},
																										},
																									},
																								},
																							},
																						},
																					},
																				},
																			},
																		},
																	},
																},
																ElseFodder: nil,
																BranchFalse: &ast.Error{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: nil,
																	},
																	Expr: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(353),
																					Column: int(40),
																				},
																				End: ast.Location{
																					Line: int(353),
																					Column: int(64),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p3956,
																			FreeVars: nil,
																		},
																		Value: "Truncated format code.",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																},
															},
														},
														Fun: nil,
														CloseFodder: nil,
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
													},
												},
												Body: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(377),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(377),
																Column: int(27),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p3903,
														FreeVars: ast.Identifiers{
															"consume",
															"i",
															"str",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(377),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(377),
																	Column: int(16),
																},
																File: p1,
															},
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																	Comment: []string{},
																},
															},
															Ctx: p3903,
															FreeVars: ast.Identifiers{
																"consume",
															},
														},
														Id: "consume",
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(377),
																				Column: int(17),
																			},
																			End: ast.Location{
																				Line: int(377),
																				Column: int(20),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p4303,
																		FreeVars: ast.Identifiers{
																			"str",
																		},
																	},
																	Id: "str",
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(377),
																				Column: int(22),
																			},
																			End: ast.Location{
																				Line: int(377),
																				Column: int(23),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p4303,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																	},
																	Id: "i",
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.LiteralNumber{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(377),
																				Column: int(25),
																			},
																			End: ast.Location{
																				Line: int(377),
																				Column: int(26),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p4303,
																		FreeVars: nil,
																	},
																	OriginalString: "0",
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
												},
											},
										},
									},
									Fun: nil,
									CloseFodder: nil,
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
								},
							},
							Body: &ast.Local{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(379),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(774),
											Column: int(48),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(1),
											Indent: int(4),
											Comment: []string{},
										},
									},
									Ctx: p3497,
									FreeVars: ast.Identifiers{
										"std",
										"str",
										"try_parse_cflags",
										"try_parse_field_width",
										"try_parse_mapping_key",
										"vals",
									},
								},
								Binds: ast.LocalBinds{
									ast.LocalBind{
										VarFodder: nil,
										Variable: "try_parse_precision",
										EqFodder: nil,
										Body: &ast.Function{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(379),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(385),
														Column: int(26),
													},
													File: p1,
												},
												Fodder: nil,
												Ctx: p4313,
												FreeVars: ast.Identifiers{
													"std",
													"try_parse_field_width",
												},
											},
											ParenLeftFodder: ast.Fodder{},
											Parameters: []ast.Parameter{
												ast.Parameter{
													NameFodder: ast.Fodder{},
													Name: "str",
													EqFodder: nil,
													DefaultArg: nil,
													CommaFodder: ast.Fodder{},
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(379),
															Column: int(31),
														},
														End: ast.Location{
															Line: int(379),
															Column: int(34),
														},
														File: p1,
													},
												},
												ast.Parameter{
													NameFodder: ast.Fodder{},
													Name: "i",
													EqFodder: nil,
													DefaultArg: nil,
													CommaFodder: nil,
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(379),
															Column: int(36),
														},
														End: ast.Location{
															Line: int(379),
															Column: int(37),
														},
														File: p1,
													},
												},
											},
											TrailingComma: false,
											ParenRightFodder: ast.Fodder{},
											Body: &ast.Conditional{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"i",
														"std",
														"str",
														"try_parse_field_width",
													},
												},
												Cond: &ast.Binary{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(380),
																Column: int(14),
															},
															End: ast.Location{
																Line: int(380),
																Column: int(33),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p4319,
														FreeVars: ast.Identifiers{
															"i",
															"std",
															"str",
														},
													},
													Left: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(380),
																	Column: int(14),
																},
																End: ast.Location{
																	Line: int(380),
																	Column: int(15),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p4319,
															FreeVars: ast.Identifiers{
																"i",
															},
														},
														Id: "i",
													},
													OpFodder: ast.Fodder{},
													Op: ast.BinaryOp(9),
													Right: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(380),
																	Column: int(18),
																},
																End: ast.Location{
																	Line: int(380),
																	Column: int(33),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p4319,
															FreeVars: ast.Identifiers{
																"std",
																"str",
															},
														},
														Target: &ast.Index{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(380),
																		Column: int(18),
																	},
																	End: ast.Location{
																		Line: int(380),
																		Column: int(28),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p4319,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Target: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(380),
																			Column: int(18),
																		},
																		End: ast.Location{
																			Line: int(380),
																			Column: int(21),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Id: "std",
															},
															LeftBracketFodder: ast.Fodder{},
															Index: &ast.LiteralString{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: nil,
																},
																Value: "length",
																Kind: ast.LiteralStringKind(1),
																BlockIndent: "",
																BlockTermIndent: "",
															},
															RightBracketFodder: ast.Fodder{},
															Id: nil,
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(380),
																					Column: int(29),
																				},
																				End: ast.Location{
																					Line: int(380),
																					Column: int(32),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p4332,
																			FreeVars: ast.Identifiers{
																				"str",
																			},
																		},
																		Id: "str",
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														TrailingComma: false,
														TailStrict: false,
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
													},
												},
												ThenFodder: nil,
												BranchTrue: &ast.Local{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(381),
																Column: int(7),
															},
															End: ast.Location{
																Line: int(385),
																Column: int(26),
															},
															File: p1,
														},
														Fodder: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(6),
																Comment: []string{},
															},
														},
														Ctx: p4319,
														FreeVars: ast.Identifiers{
															"i",
															"str",
															"try_parse_field_width",
														},
													},
													Binds: ast.LocalBinds{
														ast.LocalBind{
															VarFodder: ast.Fodder{},
															Variable: "c",
															EqFodder: ast.Fodder{},
															Body: &ast.Index{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(381),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(381),
																			Column: int(23),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p4339,
																	FreeVars: ast.Identifiers{
																		"i",
																		"str",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(381),
																				Column: int(17),
																			},
																			End: ast.Location{
																				Line: int(381),
																				Column: int(20),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p4339,
																		FreeVars: ast.Identifiers{
																			"str",
																		},
																	},
																	Id: "str",
																},
																LeftBracketFodder: ast.Fodder{},
																Index: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(381),
																				Column: int(21),
																			},
																			End: ast.Location{
																				Line: int(381),
																				Column: int(22),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p4339,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																	},
																	Id: "i",
																},
																RightBracketFodder: ast.Fodder{},
																Id: nil,
															},
															Fun: nil,
															CloseFodder: ast.Fodder{},
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(381),
																	Column: int(13),
																},
																End: ast.Location{
																	Line: int(381),
																	Column: int(23),
																},
																File: p1,
															},
														},
													},
													Body: &ast.Conditional{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(382),
																	Column: int(7),
																},
																End: ast.Location{
																	Line: int(385),
																	Column: int(26),
																},
																File: p1,
															},
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(6),
																	Comment: []string{},
																},
															},
															Ctx: p4319,
															FreeVars: ast.Identifiers{
																"c",
																"i",
																"str",
																"try_parse_field_width",
															},
														},
														Cond: &ast.Binary{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(382),
																		Column: int(10),
																	},
																	End: ast.Location{
																		Line: int(382),
																		Column: int(18),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p4319,
																FreeVars: ast.Identifiers{
																	"c",
																},
															},
															Left: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(382),
																			Column: int(10),
																		},
																		End: ast.Location{
																			Line: int(382),
																			Column: int(11),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p4319,
																	FreeVars: ast.Identifiers{
																		"c",
																	},
																},
																Id: "c",
															},
															OpFodder: ast.Fodder{},
															Op: ast.BinaryOp(12),
															Right: &ast.LiteralString{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(382),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(382),
																			Column: int(18),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p4319,
																	FreeVars: nil,
																},
																Value: ".",
																Kind: ast.LiteralStringKind(1),
																BlockIndent: "",
																BlockTermIndent: "",
															},
														},
														ThenFodder: ast.Fodder{},
														BranchTrue: &ast.Apply{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(383),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(383),
																		Column: int(42),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p4319,
																FreeVars: ast.Identifiers{
																	"i",
																	"str",
																	"try_parse_field_width",
																},
															},
															Target: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(383),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(383),
																			Column: int(30),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																			Comment: []string{},
																		},
																	},
																	Ctx: p4319,
																	FreeVars: ast.Identifiers{
																		"try_parse_field_width",
																	},
																},
																Id: "try_parse_field_width",
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(383),
																						Column: int(31),
																					},
																					End: ast.Location{
																						Line: int(383),
																						Column: int(34),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p4360,
																				FreeVars: ast.Identifiers{
																					"str",
																				},
																			},
																			Id: "str",
																		},
																		CommaFodder: ast.Fodder{},
																	},
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(383),
																						Column: int(36),
																					},
																					End: ast.Location{
																						Line: int(383),
																						Column: int(41),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p4360,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																			},
																			Left: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(383),
																							Column: int(36),
																						},
																						End: ast.Location{
																							Line: int(383),
																							Column: int(37),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p4360,
																					FreeVars: ast.Identifiers{
																						"i",
																					},
																				},
																				Id: "i",
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(3),
																			Right: &ast.LiteralNumber{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(383),
																							Column: int(40),
																						},
																						End: ast.Location{
																							Line: int(383),
																							Column: int(41),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p4360,
																					FreeVars: nil,
																				},
																				OriginalString: "1",
																			},
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															TrailingComma: false,
															TailStrict: false,
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
														},
														ElseFodder: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(6),
																Comment: []string{},
															},
														},
														BranchFalse: &ast.DesugaredObject{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(385),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(385),
																		Column: int(26),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																		Comment: []string{},
																	},
																},
																Ctx: p4319,
																FreeVars: ast.Identifiers{
																	"i",
																},
															},
															Asserts: ast.Nodes{},
															Fields: ast.DesugaredObjectFields{
																ast.DesugaredObjectField{
																	Hide: ast.ObjectFieldHide(1),
																	Name: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: nil,
																		},
																		Value: "i",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	Body: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(385),
																					Column: int(14),
																				},
																				End: ast.Location{
																					Line: int(385),
																					Column: int(15),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p4374,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																		},
																		Id: "i",
																	},
																	PlusSuper: false,
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(385),
																			Column: int(11),
																		},
																		End: ast.Location{
																			Line: int(385),
																			Column: int(15),
																		},
																		File: p1,
																	},
																},
																ast.DesugaredObjectField{
																	Hide: ast.ObjectFieldHide(1),
																	Name: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: nil,
																		},
																		Value: "v",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	Body: &ast.LiteralNull{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(385),
																					Column: int(20),
																				},
																				End: ast.Location{
																					Line: int(385),
																					Column: int(24),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p4374,
																			FreeVars: nil,
																		},
																	},
																	PlusSuper: false,
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(385),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(385),
																			Column: int(24),
																		},
																		File: p1,
																	},
																},
															},
															Locals: ast.LocalBinds{},
														},
													},
												},
												ElseFodder: nil,
												BranchFalse: &ast.Error{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Expr: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(380),
																	Column: int(36),
																},
																End: ast.Location{
																	Line: int(380),
																	Column: int(60),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p4319,
															FreeVars: nil,
														},
														Value: "Truncated format code.",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
												},
											},
										},
										Fun: nil,
										CloseFodder: nil,
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
									},
								},
								Body: &ast.Local{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(388),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(774),
												Column: int(48),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(1),
												Indent: int(4),
												Comment: []string{},
											},
											ast.FodderElement{
												Kind: ast.FodderKind(2),
												Blanks: int(0),
												Indent: int(4),
												Comment: []string{
													"// Ignored, if it exists.",
												},
											},
										},
										Ctx: p3497,
										FreeVars: ast.Identifiers{
											"std",
											"str",
											"try_parse_cflags",
											"try_parse_field_width",
											"try_parse_mapping_key",
											"try_parse_precision",
											"vals",
										},
									},
									Binds: ast.LocalBinds{
										ast.LocalBind{
											VarFodder: nil,
											Variable: "try_parse_length_modifier",
											EqFodder: nil,
											Body: &ast.Function{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(388),
															Column: int(11),
														},
														End: ast.Location{
															Line: int(394),
															Column: int(10),
														},
														File: p1,
													},
													Fodder: nil,
													Ctx: p4386,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												ParenLeftFodder: ast.Fodder{},
												Parameters: []ast.Parameter{
													ast.Parameter{
														NameFodder: ast.Fodder{},
														Name: "str",
														EqFodder: nil,
														DefaultArg: nil,
														CommaFodder: ast.Fodder{},
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(388),
																Column: int(37),
															},
															End: ast.Location{
																Line: int(388),
																Column: int(40),
															},
															File: p1,
														},
													},
													ast.Parameter{
														NameFodder: ast.Fodder{},
														Name: "i",
														EqFodder: nil,
														DefaultArg: nil,
														CommaFodder: nil,
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(388),
																Column: int(42),
															},
															End: ast.Location{
																Line: int(388),
																Column: int(43),
															},
															File: p1,
														},
													},
												},
												TrailingComma: false,
												ParenRightFodder: ast.Fodder{},
												Body: &ast.Conditional{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"i",
															"std",
															"str",
														},
													},
													Cond: &ast.Binary{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(389),
																	Column: int(14),
																},
																End: ast.Location{
																	Line: int(389),
																	Column: int(33),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p4392,
															FreeVars: ast.Identifiers{
																"i",
																"std",
																"str",
															},
														},
														Left: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(389),
																		Column: int(14),
																	},
																	End: ast.Location{
																		Line: int(389),
																		Column: int(15),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p4392,
																FreeVars: ast.Identifiers{
																	"i",
																},
															},
															Id: "i",
														},
														OpFodder: ast.Fodder{},
														Op: ast.BinaryOp(9),
														Right: &ast.Apply{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(389),
																		Column: int(18),
																	},
																	End: ast.Location{
																		Line: int(389),
																		Column: int(33),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p4392,
																FreeVars: ast.Identifiers{
																	"std",
																	"str",
																},
															},
															Target: &ast.Index{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(389),
																			Column: int(18),
																		},
																		End: ast.Location{
																			Line: int(389),
																			Column: int(28),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p4392,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(389),
																				Column: int(18),
																			},
																			End: ast.Location{
																				Line: int(389),
																				Column: int(21),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Id: "std",
																},
																LeftBracketFodder: ast.Fodder{},
																Index: &ast.LiteralString{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: nil,
																	},
																	Value: "length",
																	Kind: ast.LiteralStringKind(1),
																	BlockIndent: "",
																	BlockTermIndent: "",
																},
																RightBracketFodder: ast.Fodder{},
																Id: nil,
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(389),
																						Column: int(29),
																					},
																					End: ast.Location{
																						Line: int(389),
																						Column: int(32),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p4405,
																				FreeVars: ast.Identifiers{
																					"str",
																				},
																			},
																			Id: "str",
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															TrailingComma: false,
															TailStrict: false,
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
														},
													},
													ThenFodder: nil,
													BranchTrue: &ast.Local{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(390),
																	Column: int(7),
																},
																End: ast.Location{
																	Line: int(394),
																	Column: int(10),
																},
																File: p1,
															},
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(6),
																	Comment: []string{},
																},
															},
															Ctx: p4392,
															FreeVars: ast.Identifiers{
																"i",
																"str",
															},
														},
														Binds: ast.LocalBinds{
															ast.LocalBind{
																VarFodder: ast.Fodder{},
																Variable: "c",
																EqFodder: ast.Fodder{},
																Body: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(390),
																				Column: int(17),
																			},
																			End: ast.Location{
																				Line: int(390),
																				Column: int(23),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p4412,
																		FreeVars: ast.Identifiers{
																			"i",
																			"str",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(390),
																					Column: int(17),
																				},
																				End: ast.Location{
																					Line: int(390),
																					Column: int(20),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p4412,
																			FreeVars: ast.Identifiers{
																				"str",
																			},
																		},
																		Id: "str",
																	},
																	LeftBracketFodder: ast.Fodder{},
																	Index: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(390),
																					Column: int(21),
																				},
																				End: ast.Location{
																					Line: int(390),
																					Column: int(22),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p4412,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																		},
																		Id: "i",
																	},
																	RightBracketFodder: ast.Fodder{},
																	Id: nil,
																},
																Fun: nil,
																CloseFodder: ast.Fodder{},
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(390),
																		Column: int(13),
																	},
																	End: ast.Location{
																		Line: int(390),
																		Column: int(23),
																	},
																	File: p1,
																},
															},
														},
														Body: &ast.Conditional{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(391),
																		Column: int(7),
																	},
																	End: ast.Location{
																		Line: int(394),
																		Column: int(10),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(6),
																		Comment: []string{},
																	},
																},
																Ctx: p4392,
																FreeVars: ast.Identifiers{
																	"c",
																	"i",
																},
															},
															Cond: &ast.Binary{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(391),
																			Column: int(10),
																		},
																		End: ast.Location{
																			Line: int(391),
																			Column: int(42),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p4392,
																	FreeVars: ast.Identifiers{
																		"c",
																	},
																},
																Left: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(391),
																				Column: int(10),
																			},
																			End: ast.Location{
																				Line: int(391),
																				Column: int(30),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p4392,
																		FreeVars: ast.Identifiers{
																			"c",
																		},
																	},
																	Left: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(391),
																					Column: int(10),
																				},
																				End: ast.Location{
																					Line: int(391),
																					Column: int(18),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p4392,
																			FreeVars: ast.Identifiers{
																				"c",
																			},
																		},
																		Left: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(391),
																						Column: int(10),
																					},
																					End: ast.Location{
																						Line: int(391),
																						Column: int(11),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p4392,
																				FreeVars: ast.Identifiers{
																					"c",
																				},
																			},
																			Id: "c",
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(12),
																		Right: &ast.LiteralString{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(391),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(391),
																						Column: int(18),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p4392,
																				FreeVars: nil,
																			},
																			Value: "h",
																			Kind: ast.LiteralStringKind(1),
																			BlockIndent: "",
																			BlockTermIndent: "",
																		},
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(18),
																	Right: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(391),
																					Column: int(22),
																				},
																				End: ast.Location{
																					Line: int(391),
																					Column: int(30),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p4392,
																			FreeVars: ast.Identifiers{
																				"c",
																			},
																		},
																		Left: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(391),
																						Column: int(22),
																					},
																					End: ast.Location{
																						Line: int(391),
																						Column: int(23),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p4392,
																				FreeVars: ast.Identifiers{
																					"c",
																				},
																			},
																			Id: "c",
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(12),
																		Right: &ast.LiteralString{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(391),
																						Column: int(27),
																					},
																					End: ast.Location{
																						Line: int(391),
																						Column: int(30),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p4392,
																				FreeVars: nil,
																			},
																			Value: "l",
																			Kind: ast.LiteralStringKind(1),
																			BlockIndent: "",
																			BlockTermIndent: "",
																		},
																	},
																},
																OpFodder: ast.Fodder{},
																Op: ast.BinaryOp(18),
																Right: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(391),
																				Column: int(34),
																			},
																			End: ast.Location{
																				Line: int(391),
																				Column: int(42),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p4392,
																		FreeVars: ast.Identifiers{
																			"c",
																		},
																	},
																	Left: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(391),
																					Column: int(34),
																				},
																				End: ast.Location{
																					Line: int(391),
																					Column: int(35),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p4392,
																			FreeVars: ast.Identifiers{
																				"c",
																			},
																		},
																		Id: "c",
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(12),
																	Right: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(391),
																					Column: int(39),
																				},
																				End: ast.Location{
																					Line: int(391),
																					Column: int(42),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p4392,
																			FreeVars: nil,
																		},
																		Value: "L",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																},
															},
															ThenFodder: ast.Fodder{},
															BranchTrue: &ast.Binary{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(392),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(392),
																			Column: int(14),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p4392,
																	FreeVars: ast.Identifiers{
																		"i",
																	},
																},
																Left: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(392),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(392),
																				Column: int(10),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(8),
																				Comment: []string{},
																			},
																		},
																		Ctx: p4392,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																	},
																	Id: "i",
																},
																OpFodder: ast.Fodder{},
																Op: ast.BinaryOp(3),
																Right: &ast.LiteralNumber{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(392),
																				Column: int(13),
																			},
																			End: ast.Location{
																				Line: int(392),
																				Column: int(14),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p4392,
																		FreeVars: nil,
																	},
																	OriginalString: "1",
																},
															},
															ElseFodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(6),
																	Comment: []string{},
																},
															},
															BranchFalse: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(394),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(394),
																			Column: int(10),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																			Comment: []string{},
																		},
																	},
																	Ctx: p4392,
																	FreeVars: ast.Identifiers{
																		"i",
																	},
																},
																Id: "i",
															},
														},
													},
													ElseFodder: nil,
													BranchFalse: &ast.Error{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Expr: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(389),
																		Column: int(36),
																	},
																	End: ast.Location{
																		Line: int(389),
																		Column: int(60),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p4392,
																FreeVars: nil,
															},
															Value: "Truncated format code.",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
													},
												},
											},
											Fun: nil,
											CloseFodder: nil,
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
										},
									},
									Body: &ast.Local{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(396),
													Column: int(5),
												},
												End: ast.Location{
													Line: int(774),
													Column: int(48),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(1),
													Indent: int(4),
													Comment: []string{},
												},
											},
											Ctx: p3497,
											FreeVars: ast.Identifiers{
												"std",
												"str",
												"try_parse_cflags",
												"try_parse_field_width",
												"try_parse_length_modifier",
												"try_parse_mapping_key",
												"try_parse_precision",
												"vals",
											},
										},
										Binds: ast.LocalBinds{
											ast.LocalBind{
												VarFodder: nil,
												Variable: "parse_conv_type",
												EqFodder: nil,
												Body: &ast.Function{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(396),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(426),
																Column: int(51),
															},
															File: p1,
														},
														Fodder: nil,
														Ctx: p4457,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													ParenLeftFodder: ast.Fodder{},
													Parameters: []ast.Parameter{
														ast.Parameter{
															NameFodder: ast.Fodder{},
															Name: "str",
															EqFodder: nil,
															DefaultArg: nil,
															CommaFodder: ast.Fodder{},
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(396),
																	Column: int(27),
																},
																End: ast.Location{
																	Line: int(396),
																	Column: int(30),
																},
																File: p1,
															},
														},
														ast.Parameter{
															NameFodder: ast.Fodder{},
															Name: "i",
															EqFodder: nil,
															DefaultArg: nil,
															CommaFodder: nil,
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(396),
																	Column: int(32),
																},
																End: ast.Location{
																	Line: int(396),
																	Column: int(33),
																},
																File: p1,
															},
														},
													},
													TrailingComma: false,
													ParenRightFodder: ast.Fodder{},
													Body: &ast.Conditional{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"i",
																"std",
																"str",
															},
														},
														Cond: &ast.Binary{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(397),
																		Column: int(14),
																	},
																	End: ast.Location{
																		Line: int(397),
																		Column: int(33),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p4463,
																FreeVars: ast.Identifiers{
																	"i",
																	"std",
																	"str",
																},
															},
															Left: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(397),
																			Column: int(14),
																		},
																		End: ast.Location{
																			Line: int(397),
																			Column: int(15),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p4463,
																	FreeVars: ast.Identifiers{
																		"i",
																	},
																},
																Id: "i",
															},
															OpFodder: ast.Fodder{},
															Op: ast.BinaryOp(9),
															Right: &ast.Apply{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(397),
																			Column: int(18),
																		},
																		End: ast.Location{
																			Line: int(397),
																			Column: int(33),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p4463,
																	FreeVars: ast.Identifiers{
																		"std",
																		"str",
																	},
																},
																Target: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(397),
																				Column: int(18),
																			},
																			End: ast.Location{
																				Line: int(397),
																				Column: int(28),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p4463,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(397),
																					Column: int(18),
																				},
																				End: ast.Location{
																					Line: int(397),
																					Column: int(21),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Id: "std",
																	},
																	LeftBracketFodder: ast.Fodder{},
																	Index: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: nil,
																		},
																		Value: "length",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	RightBracketFodder: ast.Fodder{},
																	Id: nil,
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(397),
																							Column: int(29),
																						},
																						End: ast.Location{
																							Line: int(397),
																							Column: int(32),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p4476,
																					FreeVars: ast.Identifiers{
																						"str",
																					},
																				},
																				Id: "str",
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																TrailingComma: false,
																TailStrict: false,
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
															},
														},
														ThenFodder: nil,
														BranchTrue: &ast.Local{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(398),
																		Column: int(7),
																	},
																	End: ast.Location{
																		Line: int(426),
																		Column: int(51),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(6),
																		Comment: []string{},
																	},
																},
																Ctx: p4463,
																FreeVars: ast.Identifiers{
																	"i",
																	"str",
																},
															},
															Binds: ast.LocalBinds{
																ast.LocalBind{
																	VarFodder: ast.Fodder{},
																	Variable: "c",
																	EqFodder: ast.Fodder{},
																	Body: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(398),
																					Column: int(17),
																				},
																				End: ast.Location{
																					Line: int(398),
																					Column: int(23),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p4483,
																			FreeVars: ast.Identifiers{
																				"i",
																				"str",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(398),
																						Column: int(17),
																					},
																					End: ast.Location{
																						Line: int(398),
																						Column: int(20),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p4483,
																				FreeVars: ast.Identifiers{
																					"str",
																				},
																			},
																			Id: "str",
																		},
																		LeftBracketFodder: ast.Fodder{},
																		Index: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(398),
																						Column: int(21),
																					},
																					End: ast.Location{
																						Line: int(398),
																						Column: int(22),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p4483,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																			},
																			Id: "i",
																		},
																		RightBracketFodder: ast.Fodder{},
																		Id: nil,
																	},
																	Fun: nil,
																	CloseFodder: ast.Fodder{},
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(398),
																			Column: int(13),
																		},
																		End: ast.Location{
																			Line: int(398),
																			Column: int(23),
																		},
																		File: p1,
																	},
																},
															},
															Body: &ast.Conditional{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(399),
																			Column: int(7),
																		},
																		End: ast.Location{
																			Line: int(426),
																			Column: int(51),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(6),
																			Comment: []string{},
																		},
																	},
																	Ctx: p4463,
																	FreeVars: ast.Identifiers{
																		"c",
																		"i",
																	},
																},
																Cond: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(399),
																				Column: int(10),
																			},
																			End: ast.Location{
																				Line: int(399),
																				Column: int(42),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p4463,
																		FreeVars: ast.Identifiers{
																			"c",
																		},
																	},
																	Left: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(399),
																					Column: int(10),
																				},
																				End: ast.Location{
																					Line: int(399),
																					Column: int(30),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p4463,
																			FreeVars: ast.Identifiers{
																				"c",
																			},
																		},
																		Left: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(399),
																						Column: int(10),
																					},
																					End: ast.Location{
																						Line: int(399),
																						Column: int(18),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p4463,
																				FreeVars: ast.Identifiers{
																					"c",
																				},
																			},
																			Left: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(399),
																							Column: int(10),
																						},
																						End: ast.Location{
																							Line: int(399),
																							Column: int(11),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p4463,
																					FreeVars: ast.Identifiers{
																						"c",
																					},
																				},
																				Id: "c",
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(12),
																			Right: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(399),
																							Column: int(15),
																						},
																						End: ast.Location{
																							Line: int(399),
																							Column: int(18),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p4463,
																					FreeVars: nil,
																				},
																				Value: "d",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(18),
																		Right: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(399),
																						Column: int(22),
																					},
																					End: ast.Location{
																						Line: int(399),
																						Column: int(30),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p4463,
																				FreeVars: ast.Identifiers{
																					"c",
																				},
																			},
																			Left: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(399),
																							Column: int(22),
																						},
																						End: ast.Location{
																							Line: int(399),
																							Column: int(23),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p4463,
																					FreeVars: ast.Identifiers{
																						"c",
																					},
																				},
																				Id: "c",
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(12),
																			Right: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(399),
																							Column: int(27),
																						},
																						End: ast.Location{
																							Line: int(399),
																							Column: int(30),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p4463,
																					FreeVars: nil,
																				},
																				Value: "i",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																		},
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(18),
																	Right: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(399),
																					Column: int(34),
																				},
																				End: ast.Location{
																					Line: int(399),
																					Column: int(42),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p4463,
																			FreeVars: ast.Identifiers{
																				"c",
																			},
																		},
																		Left: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(399),
																						Column: int(34),
																					},
																					End: ast.Location{
																						Line: int(399),
																						Column: int(35),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p4463,
																				FreeVars: ast.Identifiers{
																					"c",
																				},
																			},
																			Id: "c",
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(12),
																		Right: &ast.LiteralString{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(399),
																						Column: int(39),
																					},
																					End: ast.Location{
																						Line: int(399),
																						Column: int(42),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p4463,
																				FreeVars: nil,
																			},
																			Value: "u",
																			Kind: ast.LiteralStringKind(1),
																			BlockIndent: "",
																			BlockTermIndent: "",
																		},
																	},
																},
																ThenFodder: ast.Fodder{},
																BranchTrue: &ast.DesugaredObject{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(400),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(400),
																				Column: int(42),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(8),
																				Comment: []string{},
																			},
																		},
																		Ctx: p4463,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																	},
																	Asserts: ast.Nodes{},
																	Fields: ast.DesugaredObjectFields{
																		ast.DesugaredObjectField{
																			Hide: ast.ObjectFieldHide(1),
																			Name: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: nil,
																				},
																				Value: "i",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																			Body: &ast.Binary{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(400),
																							Column: int(14),
																						},
																						End: ast.Location{
																							Line: int(400),
																							Column: int(19),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p4517,
																					FreeVars: ast.Identifiers{
																						"i",
																					},
																				},
																				Left: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(400),
																								Column: int(14),
																							},
																							End: ast.Location{
																								Line: int(400),
																								Column: int(15),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p4517,
																						FreeVars: ast.Identifiers{
																							"i",
																						},
																					},
																					Id: "i",
																				},
																				OpFodder: ast.Fodder{},
																				Op: ast.BinaryOp(3),
																				Right: &ast.LiteralNumber{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(400),
																								Column: int(18),
																							},
																							End: ast.Location{
																								Line: int(400),
																								Column: int(19),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p4517,
																						FreeVars: nil,
																					},
																					OriginalString: "1",
																				},
																			},
																			PlusSuper: false,
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(400),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(400),
																					Column: int(19),
																				},
																				File: p1,
																			},
																		},
																		ast.DesugaredObjectField{
																			Hide: ast.ObjectFieldHide(1),
																			Name: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: nil,
																				},
																				Value: "v",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																			Body: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(400),
																							Column: int(24),
																						},
																						End: ast.Location{
																							Line: int(400),
																							Column: int(27),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p4517,
																					FreeVars: nil,
																				},
																				Value: "d",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																			PlusSuper: false,
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(400),
																					Column: int(21),
																				},
																				End: ast.Location{
																					Line: int(400),
																					Column: int(27),
																				},
																				File: p1,
																			},
																		},
																		ast.DesugaredObjectField{
																			Hide: ast.ObjectFieldHide(1),
																			Name: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: nil,
																				},
																				Value: "caps",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																			Body: &ast.LiteralBoolean{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(400),
																							Column: int(35),
																						},
																						End: ast.Location{
																							Line: int(400),
																							Column: int(40),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p4517,
																					FreeVars: nil,
																				},
																				Value: false,
																			},
																			PlusSuper: false,
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(400),
																					Column: int(29),
																				},
																				End: ast.Location{
																					Line: int(400),
																					Column: int(40),
																				},
																				File: p1,
																			},
																		},
																	},
																	Locals: ast.LocalBinds{},
																},
																ElseFodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(6),
																		Comment: []string{},
																	},
																},
																BranchFalse: &ast.Conditional{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(401),
																				Column: int(12),
																			},
																			End: ast.Location{
																				Line: int(426),
																				Column: int(51),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p4463,
																		FreeVars: ast.Identifiers{
																			"c",
																			"i",
																		},
																	},
																	Cond: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(401),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(401),
																					Column: int(23),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p4463,
																			FreeVars: ast.Identifiers{
																				"c",
																			},
																		},
																		Left: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(401),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(401),
																						Column: int(16),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p4463,
																				FreeVars: ast.Identifiers{
																					"c",
																				},
																			},
																			Id: "c",
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(12),
																		Right: &ast.LiteralString{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(401),
																						Column: int(20),
																					},
																					End: ast.Location{
																						Line: int(401),
																						Column: int(23),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p4463,
																				FreeVars: nil,
																			},
																			Value: "o",
																			Kind: ast.LiteralStringKind(1),
																			BlockIndent: "",
																			BlockTermIndent: "",
																		},
																	},
																	ThenFodder: ast.Fodder{},
																	BranchTrue: &ast.DesugaredObject{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(402),
																					Column: int(9),
																				},
																				End: ast.Location{
																					Line: int(402),
																					Column: int(42),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(8),
																					Comment: []string{},
																				},
																			},
																			Ctx: p4463,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																		},
																		Asserts: ast.Nodes{},
																		Fields: ast.DesugaredObjectFields{
																			ast.DesugaredObjectField{
																				Hide: ast.ObjectFieldHide(1),
																				Name: &ast.LiteralString{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							File: nil,
																						},
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: nil,
																					},
																					Value: "i",
																					Kind: ast.LiteralStringKind(1),
																					BlockIndent: "",
																					BlockTermIndent: "",
																				},
																				Body: &ast.Binary{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(402),
																								Column: int(14),
																							},
																							End: ast.Location{
																								Line: int(402),
																								Column: int(19),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p4540,
																						FreeVars: ast.Identifiers{
																							"i",
																						},
																					},
																					Left: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(402),
																									Column: int(14),
																								},
																								End: ast.Location{
																									Line: int(402),
																									Column: int(15),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p4540,
																							FreeVars: ast.Identifiers{
																								"i",
																							},
																						},
																						Id: "i",
																					},
																					OpFodder: ast.Fodder{},
																					Op: ast.BinaryOp(3),
																					Right: &ast.LiteralNumber{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(402),
																									Column: int(18),
																								},
																								End: ast.Location{
																									Line: int(402),
																									Column: int(19),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p4540,
																							FreeVars: nil,
																						},
																						OriginalString: "1",
																					},
																				},
																				PlusSuper: false,
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(402),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(402),
																						Column: int(19),
																					},
																					File: p1,
																				},
																			},
																			ast.DesugaredObjectField{
																				Hide: ast.ObjectFieldHide(1),
																				Name: &ast.LiteralString{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							File: nil,
																						},
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: nil,
																					},
																					Value: "v",
																					Kind: ast.LiteralStringKind(1),
																					BlockIndent: "",
																					BlockTermIndent: "",
																				},
																				Body: &ast.LiteralString{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(402),
																								Column: int(24),
																							},
																							End: ast.Location{
																								Line: int(402),
																								Column: int(27),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p4540,
																						FreeVars: nil,
																					},
																					Value: "o",
																					Kind: ast.LiteralStringKind(1),
																					BlockIndent: "",
																					BlockTermIndent: "",
																				},
																				PlusSuper: false,
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(402),
																						Column: int(21),
																					},
																					End: ast.Location{
																						Line: int(402),
																						Column: int(27),
																					},
																					File: p1,
																				},
																			},
																			ast.DesugaredObjectField{
																				Hide: ast.ObjectFieldHide(1),
																				Name: &ast.LiteralString{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							File: nil,
																						},
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: nil,
																					},
																					Value: "caps",
																					Kind: ast.LiteralStringKind(1),
																					BlockIndent: "",
																					BlockTermIndent: "",
																				},
																				Body: &ast.LiteralBoolean{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(402),
																								Column: int(35),
																							},
																							End: ast.Location{
																								Line: int(402),
																								Column: int(40),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p4540,
																						FreeVars: nil,
																					},
																					Value: false,
																				},
																				PlusSuper: false,
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(402),
																						Column: int(29),
																					},
																					End: ast.Location{
																						Line: int(402),
																						Column: int(40),
																					},
																					File: p1,
																				},
																			},
																		},
																		Locals: ast.LocalBinds{},
																	},
																	ElseFodder: ast.Fodder{
																		ast.FodderElement{
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(6),
																			Comment: []string{},
																		},
																	},
																	BranchFalse: &ast.Conditional{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(403),
																					Column: int(12),
																				},
																				End: ast.Location{
																					Line: int(426),
																					Column: int(51),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p4463,
																			FreeVars: ast.Identifiers{
																				"c",
																				"i",
																			},
																		},
																		Cond: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(403),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(403),
																						Column: int(23),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p4463,
																				FreeVars: ast.Identifiers{
																					"c",
																				},
																			},
																			Left: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(403),
																							Column: int(15),
																						},
																						End: ast.Location{
																							Line: int(403),
																							Column: int(16),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p4463,
																					FreeVars: ast.Identifiers{
																						"c",
																					},
																				},
																				Id: "c",
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(12),
																			Right: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(403),
																							Column: int(20),
																						},
																						End: ast.Location{
																							Line: int(403),
																							Column: int(23),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p4463,
																					FreeVars: nil,
																				},
																				Value: "x",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																		},
																		ThenFodder: ast.Fodder{},
																		BranchTrue: &ast.DesugaredObject{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(404),
																						Column: int(9),
																					},
																					End: ast.Location{
																						Line: int(404),
																						Column: int(42),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(8),
																						Comment: []string{},
																					},
																				},
																				Ctx: p4463,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																			},
																			Asserts: ast.Nodes{},
																			Fields: ast.DesugaredObjectFields{
																				ast.DesugaredObjectField{
																					Hide: ast.ObjectFieldHide(1),
																					Name: &ast.LiteralString{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								File: nil,
																							},
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: nil,
																						},
																						Value: "i",
																						Kind: ast.LiteralStringKind(1),
																						BlockIndent: "",
																						BlockTermIndent: "",
																					},
																					Body: &ast.Binary{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(404),
																									Column: int(14),
																								},
																								End: ast.Location{
																									Line: int(404),
																									Column: int(19),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p4563,
																							FreeVars: ast.Identifiers{
																								"i",
																							},
																						},
																						Left: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(404),
																										Column: int(14),
																									},
																									End: ast.Location{
																										Line: int(404),
																										Column: int(15),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p4563,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																							},
																							Id: "i",
																						},
																						OpFodder: ast.Fodder{},
																						Op: ast.BinaryOp(3),
																						Right: &ast.LiteralNumber{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(404),
																										Column: int(18),
																									},
																									End: ast.Location{
																										Line: int(404),
																										Column: int(19),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p4563,
																								FreeVars: nil,
																							},
																							OriginalString: "1",
																						},
																					},
																					PlusSuper: false,
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(404),
																							Column: int(11),
																						},
																						End: ast.Location{
																							Line: int(404),
																							Column: int(19),
																						},
																						File: p1,
																					},
																				},
																				ast.DesugaredObjectField{
																					Hide: ast.ObjectFieldHide(1),
																					Name: &ast.LiteralString{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								File: nil,
																							},
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: nil,
																						},
																						Value: "v",
																						Kind: ast.LiteralStringKind(1),
																						BlockIndent: "",
																						BlockTermIndent: "",
																					},
																					Body: &ast.LiteralString{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(404),
																									Column: int(24),
																								},
																								End: ast.Location{
																									Line: int(404),
																									Column: int(27),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p4563,
																							FreeVars: nil,
																						},
																						Value: "x",
																						Kind: ast.LiteralStringKind(1),
																						BlockIndent: "",
																						BlockTermIndent: "",
																					},
																					PlusSuper: false,
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(404),
																							Column: int(21),
																						},
																						End: ast.Location{
																							Line: int(404),
																							Column: int(27),
																						},
																						File: p1,
																					},
																				},
																				ast.DesugaredObjectField{
																					Hide: ast.ObjectFieldHide(1),
																					Name: &ast.LiteralString{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								File: nil,
																							},
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: nil,
																						},
																						Value: "caps",
																						Kind: ast.LiteralStringKind(1),
																						BlockIndent: "",
																						BlockTermIndent: "",
																					},
																					Body: &ast.LiteralBoolean{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(404),
																									Column: int(35),
																								},
																								End: ast.Location{
																									Line: int(404),
																									Column: int(40),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p4563,
																							FreeVars: nil,
																						},
																						Value: false,
																					},
																					PlusSuper: false,
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(404),
																							Column: int(29),
																						},
																						End: ast.Location{
																							Line: int(404),
																							Column: int(40),
																						},
																						File: p1,
																					},
																				},
																			},
																			Locals: ast.LocalBinds{},
																		},
																		ElseFodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(6),
																				Comment: []string{},
																			},
																		},
																		BranchFalse: &ast.Conditional{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(405),
																						Column: int(12),
																					},
																					End: ast.Location{
																						Line: int(426),
																						Column: int(51),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p4463,
																				FreeVars: ast.Identifiers{
																					"c",
																					"i",
																				},
																			},
																			Cond: &ast.Binary{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(405),
																							Column: int(15),
																						},
																						End: ast.Location{
																							Line: int(405),
																							Column: int(23),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p4463,
																					FreeVars: ast.Identifiers{
																						"c",
																					},
																				},
																				Left: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(405),
																								Column: int(15),
																							},
																							End: ast.Location{
																								Line: int(405),
																								Column: int(16),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p4463,
																						FreeVars: ast.Identifiers{
																							"c",
																						},
																					},
																					Id: "c",
																				},
																				OpFodder: ast.Fodder{},
																				Op: ast.BinaryOp(12),
																				Right: &ast.LiteralString{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(405),
																								Column: int(20),
																							},
																							End: ast.Location{
																								Line: int(405),
																								Column: int(23),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p4463,
																						FreeVars: nil,
																					},
																					Value: "X",
																					Kind: ast.LiteralStringKind(1),
																					BlockIndent: "",
																					BlockTermIndent: "",
																				},
																			},
																			ThenFodder: ast.Fodder{},
																			BranchTrue: &ast.DesugaredObject{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(406),
																							Column: int(9),
																						},
																						End: ast.Location{
																							Line: int(406),
																							Column: int(41),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(8),
																							Comment: []string{},
																						},
																					},
																					Ctx: p4463,
																					FreeVars: ast.Identifiers{
																						"i",
																					},
																				},
																				Asserts: ast.Nodes{},
																				Fields: ast.DesugaredObjectFields{
																					ast.DesugaredObjectField{
																						Hide: ast.ObjectFieldHide(1),
																						Name: &ast.LiteralString{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									File: nil,
																								},
																								Fodder: nil,
																								Ctx: nil,
																								FreeVars: nil,
																							},
																							Value: "i",
																							Kind: ast.LiteralStringKind(1),
																							BlockIndent: "",
																							BlockTermIndent: "",
																						},
																						Body: &ast.Binary{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(406),
																										Column: int(14),
																									},
																									End: ast.Location{
																										Line: int(406),
																										Column: int(19),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p4586,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																							},
																							Left: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(406),
																											Column: int(14),
																										},
																										End: ast.Location{
																											Line: int(406),
																											Column: int(15),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p4586,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																								},
																								Id: "i",
																							},
																							OpFodder: ast.Fodder{},
																							Op: ast.BinaryOp(3),
																							Right: &ast.LiteralNumber{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(406),
																											Column: int(18),
																										},
																										End: ast.Location{
																											Line: int(406),
																											Column: int(19),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p4586,
																									FreeVars: nil,
																								},
																								OriginalString: "1",
																							},
																						},
																						PlusSuper: false,
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(406),
																								Column: int(11),
																							},
																							End: ast.Location{
																								Line: int(406),
																								Column: int(19),
																							},
																							File: p1,
																						},
																					},
																					ast.DesugaredObjectField{
																						Hide: ast.ObjectFieldHide(1),
																						Name: &ast.LiteralString{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									File: nil,
																								},
																								Fodder: nil,
																								Ctx: nil,
																								FreeVars: nil,
																							},
																							Value: "v",
																							Kind: ast.LiteralStringKind(1),
																							BlockIndent: "",
																							BlockTermIndent: "",
																						},
																						Body: &ast.LiteralString{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(406),
																										Column: int(24),
																									},
																									End: ast.Location{
																										Line: int(406),
																										Column: int(27),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p4586,
																								FreeVars: nil,
																							},
																							Value: "x",
																							Kind: ast.LiteralStringKind(1),
																							BlockIndent: "",
																							BlockTermIndent: "",
																						},
																						PlusSuper: false,
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(406),
																								Column: int(21),
																							},
																							End: ast.Location{
																								Line: int(406),
																								Column: int(27),
																							},
																							File: p1,
																						},
																					},
																					ast.DesugaredObjectField{
																						Hide: ast.ObjectFieldHide(1),
																						Name: &ast.LiteralString{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									File: nil,
																								},
																								Fodder: nil,
																								Ctx: nil,
																								FreeVars: nil,
																							},
																							Value: "caps",
																							Kind: ast.LiteralStringKind(1),
																							BlockIndent: "",
																							BlockTermIndent: "",
																						},
																						Body: &ast.LiteralBoolean{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(406),
																										Column: int(35),
																									},
																									End: ast.Location{
																										Line: int(406),
																										Column: int(39),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p4586,
																								FreeVars: nil,
																							},
																							Value: true,
																						},
																						PlusSuper: false,
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(406),
																								Column: int(29),
																							},
																							End: ast.Location{
																								Line: int(406),
																								Column: int(39),
																							},
																							File: p1,
																						},
																					},
																				},
																				Locals: ast.LocalBinds{},
																			},
																			ElseFodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(6),
																					Comment: []string{},
																				},
																			},
																			BranchFalse: &ast.Conditional{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(407),
																							Column: int(12),
																						},
																						End: ast.Location{
																							Line: int(426),
																							Column: int(51),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p4463,
																					FreeVars: ast.Identifiers{
																						"c",
																						"i",
																					},
																				},
																				Cond: &ast.Binary{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(407),
																								Column: int(15),
																							},
																							End: ast.Location{
																								Line: int(407),
																								Column: int(23),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p4463,
																						FreeVars: ast.Identifiers{
																							"c",
																						},
																					},
																					Left: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(407),
																									Column: int(15),
																								},
																								End: ast.Location{
																									Line: int(407),
																									Column: int(16),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p4463,
																							FreeVars: ast.Identifiers{
																								"c",
																							},
																						},
																						Id: "c",
																					},
																					OpFodder: ast.Fodder{},
																					Op: ast.BinaryOp(12),
																					Right: &ast.LiteralString{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(407),
																									Column: int(20),
																								},
																								End: ast.Location{
																									Line: int(407),
																									Column: int(23),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p4463,
																							FreeVars: nil,
																						},
																						Value: "e",
																						Kind: ast.LiteralStringKind(1),
																						BlockIndent: "",
																						BlockTermIndent: "",
																					},
																				},
																				ThenFodder: ast.Fodder{},
																				BranchTrue: &ast.DesugaredObject{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(408),
																								Column: int(9),
																							},
																							End: ast.Location{
																								Line: int(408),
																								Column: int(42),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(8),
																								Comment: []string{},
																							},
																						},
																						Ctx: p4463,
																						FreeVars: ast.Identifiers{
																							"i",
																						},
																					},
																					Asserts: ast.Nodes{},
																					Fields: ast.DesugaredObjectFields{
																						ast.DesugaredObjectField{
																							Hide: ast.ObjectFieldHide(1),
																							Name: &ast.LiteralString{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										File: nil,
																									},
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: nil,
																								},
																								Value: "i",
																								Kind: ast.LiteralStringKind(1),
																								BlockIndent: "",
																								BlockTermIndent: "",
																							},
																							Body: &ast.Binary{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(408),
																											Column: int(14),
																										},
																										End: ast.Location{
																											Line: int(408),
																											Column: int(19),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p4609,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																								},
																								Left: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(408),
																												Column: int(14),
																											},
																											End: ast.Location{
																												Line: int(408),
																												Column: int(15),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p4609,
																										FreeVars: ast.Identifiers{
																											"i",
																										},
																									},
																									Id: "i",
																								},
																								OpFodder: ast.Fodder{},
																								Op: ast.BinaryOp(3),
																								Right: &ast.LiteralNumber{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(408),
																												Column: int(18),
																											},
																											End: ast.Location{
																												Line: int(408),
																												Column: int(19),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p4609,
																										FreeVars: nil,
																									},
																									OriginalString: "1",
																								},
																							},
																							PlusSuper: false,
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(408),
																									Column: int(11),
																								},
																								End: ast.Location{
																									Line: int(408),
																									Column: int(19),
																								},
																								File: p1,
																							},
																						},
																						ast.DesugaredObjectField{
																							Hide: ast.ObjectFieldHide(1),
																							Name: &ast.LiteralString{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										File: nil,
																									},
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: nil,
																								},
																								Value: "v",
																								Kind: ast.LiteralStringKind(1),
																								BlockIndent: "",
																								BlockTermIndent: "",
																							},
																							Body: &ast.LiteralString{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(408),
																											Column: int(24),
																										},
																										End: ast.Location{
																											Line: int(408),
																											Column: int(27),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p4609,
																									FreeVars: nil,
																								},
																								Value: "e",
																								Kind: ast.LiteralStringKind(1),
																								BlockIndent: "",
																								BlockTermIndent: "",
																							},
																							PlusSuper: false,
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(408),
																									Column: int(21),
																								},
																								End: ast.Location{
																									Line: int(408),
																									Column: int(27),
																								},
																								File: p1,
																							},
																						},
																						ast.DesugaredObjectField{
																							Hide: ast.ObjectFieldHide(1),
																							Name: &ast.LiteralString{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										File: nil,
																									},
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: nil,
																								},
																								Value: "caps",
																								Kind: ast.LiteralStringKind(1),
																								BlockIndent: "",
																								BlockTermIndent: "",
																							},
																							Body: &ast.LiteralBoolean{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(408),
																											Column: int(35),
																										},
																										End: ast.Location{
																											Line: int(408),
																											Column: int(40),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p4609,
																									FreeVars: nil,
																								},
																								Value: false,
																							},
																							PlusSuper: false,
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(408),
																									Column: int(29),
																								},
																								End: ast.Location{
																									Line: int(408),
																									Column: int(40),
																								},
																								File: p1,
																							},
																						},
																					},
																					Locals: ast.LocalBinds{},
																				},
																				ElseFodder: ast.Fodder{
																					ast.FodderElement{
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(6),
																						Comment: []string{},
																					},
																				},
																				BranchFalse: &ast.Conditional{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(409),
																								Column: int(12),
																							},
																							End: ast.Location{
																								Line: int(426),
																								Column: int(51),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p4463,
																						FreeVars: ast.Identifiers{
																							"c",
																							"i",
																						},
																					},
																					Cond: &ast.Binary{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(409),
																									Column: int(15),
																								},
																								End: ast.Location{
																									Line: int(409),
																									Column: int(23),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p4463,
																							FreeVars: ast.Identifiers{
																								"c",
																							},
																						},
																						Left: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(409),
																										Column: int(15),
																									},
																									End: ast.Location{
																										Line: int(409),
																										Column: int(16),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p4463,
																								FreeVars: ast.Identifiers{
																									"c",
																								},
																							},
																							Id: "c",
																						},
																						OpFodder: ast.Fodder{},
																						Op: ast.BinaryOp(12),
																						Right: &ast.LiteralString{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(409),
																										Column: int(20),
																									},
																									End: ast.Location{
																										Line: int(409),
																										Column: int(23),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p4463,
																								FreeVars: nil,
																							},
																							Value: "E",
																							Kind: ast.LiteralStringKind(1),
																							BlockIndent: "",
																							BlockTermIndent: "",
																						},
																					},
																					ThenFodder: ast.Fodder{},
																					BranchTrue: &ast.DesugaredObject{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(410),
																									Column: int(9),
																								},
																								End: ast.Location{
																									Line: int(410),
																									Column: int(41),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{
																								ast.FodderElement{
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(8),
																									Comment: []string{},
																								},
																							},
																							Ctx: p4463,
																							FreeVars: ast.Identifiers{
																								"i",
																							},
																						},
																						Asserts: ast.Nodes{},
																						Fields: ast.DesugaredObjectFields{
																							ast.DesugaredObjectField{
																								Hide: ast.ObjectFieldHide(1),
																								Name: &ast.LiteralString{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											File: nil,
																										},
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: nil,
																									},
																									Value: "i",
																									Kind: ast.LiteralStringKind(1),
																									BlockIndent: "",
																									BlockTermIndent: "",
																								},
																								Body: &ast.Binary{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(410),
																												Column: int(14),
																											},
																											End: ast.Location{
																												Line: int(410),
																												Column: int(19),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p4632,
																										FreeVars: ast.Identifiers{
																											"i",
																										},
																									},
																									Left: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(410),
																													Column: int(14),
																												},
																												End: ast.Location{
																													Line: int(410),
																													Column: int(15),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p4632,
																											FreeVars: ast.Identifiers{
																												"i",
																											},
																										},
																										Id: "i",
																									},
																									OpFodder: ast.Fodder{},
																									Op: ast.BinaryOp(3),
																									Right: &ast.LiteralNumber{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(410),
																													Column: int(18),
																												},
																												End: ast.Location{
																													Line: int(410),
																													Column: int(19),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p4632,
																											FreeVars: nil,
																										},
																										OriginalString: "1",
																									},
																								},
																								PlusSuper: false,
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(410),
																										Column: int(11),
																									},
																									End: ast.Location{
																										Line: int(410),
																										Column: int(19),
																									},
																									File: p1,
																								},
																							},
																							ast.DesugaredObjectField{
																								Hide: ast.ObjectFieldHide(1),
																								Name: &ast.LiteralString{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											File: nil,
																										},
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: nil,
																									},
																									Value: "v",
																									Kind: ast.LiteralStringKind(1),
																									BlockIndent: "",
																									BlockTermIndent: "",
																								},
																								Body: &ast.LiteralString{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(410),
																												Column: int(24),
																											},
																											End: ast.Location{
																												Line: int(410),
																												Column: int(27),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p4632,
																										FreeVars: nil,
																									},
																									Value: "e",
																									Kind: ast.LiteralStringKind(1),
																									BlockIndent: "",
																									BlockTermIndent: "",
																								},
																								PlusSuper: false,
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(410),
																										Column: int(21),
																									},
																									End: ast.Location{
																										Line: int(410),
																										Column: int(27),
																									},
																									File: p1,
																								},
																							},
																							ast.DesugaredObjectField{
																								Hide: ast.ObjectFieldHide(1),
																								Name: &ast.LiteralString{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											File: nil,
																										},
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: nil,
																									},
																									Value: "caps",
																									Kind: ast.LiteralStringKind(1),
																									BlockIndent: "",
																									BlockTermIndent: "",
																								},
																								Body: &ast.LiteralBoolean{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(410),
																												Column: int(35),
																											},
																											End: ast.Location{
																												Line: int(410),
																												Column: int(39),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p4632,
																										FreeVars: nil,
																									},
																									Value: true,
																								},
																								PlusSuper: false,
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(410),
																										Column: int(29),
																									},
																									End: ast.Location{
																										Line: int(410),
																										Column: int(39),
																									},
																									File: p1,
																								},
																							},
																						},
																						Locals: ast.LocalBinds{},
																					},
																					ElseFodder: ast.Fodder{
																						ast.FodderElement{
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(6),
																							Comment: []string{},
																						},
																					},
																					BranchFalse: &ast.Conditional{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(411),
																									Column: int(12),
																								},
																								End: ast.Location{
																									Line: int(426),
																									Column: int(51),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p4463,
																							FreeVars: ast.Identifiers{
																								"c",
																								"i",
																							},
																						},
																						Cond: &ast.Binary{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(411),
																										Column: int(15),
																									},
																									End: ast.Location{
																										Line: int(411),
																										Column: int(23),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p4463,
																								FreeVars: ast.Identifiers{
																									"c",
																								},
																							},
																							Left: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(411),
																											Column: int(15),
																										},
																										End: ast.Location{
																											Line: int(411),
																											Column: int(16),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p4463,
																									FreeVars: ast.Identifiers{
																										"c",
																									},
																								},
																								Id: "c",
																							},
																							OpFodder: ast.Fodder{},
																							Op: ast.BinaryOp(12),
																							Right: &ast.LiteralString{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(411),
																											Column: int(20),
																										},
																										End: ast.Location{
																											Line: int(411),
																											Column: int(23),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p4463,
																									FreeVars: nil,
																								},
																								Value: "f",
																								Kind: ast.LiteralStringKind(1),
																								BlockIndent: "",
																								BlockTermIndent: "",
																							},
																						},
																						ThenFodder: ast.Fodder{},
																						BranchTrue: &ast.DesugaredObject{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(412),
																										Column: int(9),
																									},
																									End: ast.Location{
																										Line: int(412),
																										Column: int(42),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{
																									ast.FodderElement{
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(8),
																										Comment: []string{},
																									},
																								},
																								Ctx: p4463,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																							},
																							Asserts: ast.Nodes{},
																							Fields: ast.DesugaredObjectFields{
																								ast.DesugaredObjectField{
																									Hide: ast.ObjectFieldHide(1),
																									Name: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												File: nil,
																											},
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: nil,
																										},
																										Value: "i",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																									Body: &ast.Binary{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(412),
																													Column: int(14),
																												},
																												End: ast.Location{
																													Line: int(412),
																													Column: int(19),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p4655,
																											FreeVars: ast.Identifiers{
																												"i",
																											},
																										},
																										Left: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(412),
																														Column: int(14),
																													},
																													End: ast.Location{
																														Line: int(412),
																														Column: int(15),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p4655,
																												FreeVars: ast.Identifiers{
																													"i",
																												},
																											},
																											Id: "i",
																										},
																										OpFodder: ast.Fodder{},
																										Op: ast.BinaryOp(3),
																										Right: &ast.LiteralNumber{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(412),
																														Column: int(18),
																													},
																													End: ast.Location{
																														Line: int(412),
																														Column: int(19),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p4655,
																												FreeVars: nil,
																											},
																											OriginalString: "1",
																										},
																									},
																									PlusSuper: false,
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(412),
																											Column: int(11),
																										},
																										End: ast.Location{
																											Line: int(412),
																											Column: int(19),
																										},
																										File: p1,
																									},
																								},
																								ast.DesugaredObjectField{
																									Hide: ast.ObjectFieldHide(1),
																									Name: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												File: nil,
																											},
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: nil,
																										},
																										Value: "v",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																									Body: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(412),
																													Column: int(24),
																												},
																												End: ast.Location{
																													Line: int(412),
																													Column: int(27),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p4655,
																											FreeVars: nil,
																										},
																										Value: "f",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																									PlusSuper: false,
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(412),
																											Column: int(21),
																										},
																										End: ast.Location{
																											Line: int(412),
																											Column: int(27),
																										},
																										File: p1,
																									},
																								},
																								ast.DesugaredObjectField{
																									Hide: ast.ObjectFieldHide(1),
																									Name: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												File: nil,
																											},
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: nil,
																										},
																										Value: "caps",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																									Body: &ast.LiteralBoolean{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(412),
																													Column: int(35),
																												},
																												End: ast.Location{
																													Line: int(412),
																													Column: int(40),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p4655,
																											FreeVars: nil,
																										},
																										Value: false,
																									},
																									PlusSuper: false,
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(412),
																											Column: int(29),
																										},
																										End: ast.Location{
																											Line: int(412),
																											Column: int(40),
																										},
																										File: p1,
																									},
																								},
																							},
																							Locals: ast.LocalBinds{},
																						},
																						ElseFodder: ast.Fodder{
																							ast.FodderElement{
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(6),
																								Comment: []string{},
																							},
																						},
																						BranchFalse: &ast.Conditional{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(413),
																										Column: int(12),
																									},
																									End: ast.Location{
																										Line: int(426),
																										Column: int(51),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p4463,
																								FreeVars: ast.Identifiers{
																									"c",
																									"i",
																								},
																							},
																							Cond: &ast.Binary{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(413),
																											Column: int(15),
																										},
																										End: ast.Location{
																											Line: int(413),
																											Column: int(23),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p4463,
																									FreeVars: ast.Identifiers{
																										"c",
																									},
																								},
																								Left: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(413),
																												Column: int(15),
																											},
																											End: ast.Location{
																												Line: int(413),
																												Column: int(16),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p4463,
																										FreeVars: ast.Identifiers{
																											"c",
																										},
																									},
																									Id: "c",
																								},
																								OpFodder: ast.Fodder{},
																								Op: ast.BinaryOp(12),
																								Right: &ast.LiteralString{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(413),
																												Column: int(20),
																											},
																											End: ast.Location{
																												Line: int(413),
																												Column: int(23),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p4463,
																										FreeVars: nil,
																									},
																									Value: "F",
																									Kind: ast.LiteralStringKind(1),
																									BlockIndent: "",
																									BlockTermIndent: "",
																								},
																							},
																							ThenFodder: ast.Fodder{},
																							BranchTrue: &ast.DesugaredObject{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(414),
																											Column: int(9),
																										},
																										End: ast.Location{
																											Line: int(414),
																											Column: int(41),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{
																										ast.FodderElement{
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(8),
																											Comment: []string{},
																										},
																									},
																									Ctx: p4463,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																								},
																								Asserts: ast.Nodes{},
																								Fields: ast.DesugaredObjectFields{
																									ast.DesugaredObjectField{
																										Hide: ast.ObjectFieldHide(1),
																										Name: &ast.LiteralString{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													File: nil,
																												},
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: nil,
																											},
																											Value: "i",
																											Kind: ast.LiteralStringKind(1),
																											BlockIndent: "",
																											BlockTermIndent: "",
																										},
																										Body: &ast.Binary{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(414),
																														Column: int(14),
																													},
																													End: ast.Location{
																														Line: int(414),
																														Column: int(19),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p4678,
																												FreeVars: ast.Identifiers{
																													"i",
																												},
																											},
																											Left: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(414),
																															Column: int(14),
																														},
																														End: ast.Location{
																															Line: int(414),
																															Column: int(15),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p4678,
																													FreeVars: ast.Identifiers{
																														"i",
																													},
																												},
																												Id: "i",
																											},
																											OpFodder: ast.Fodder{},
																											Op: ast.BinaryOp(3),
																											Right: &ast.LiteralNumber{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(414),
																															Column: int(18),
																														},
																														End: ast.Location{
																															Line: int(414),
																															Column: int(19),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p4678,
																													FreeVars: nil,
																												},
																												OriginalString: "1",
																											},
																										},
																										PlusSuper: false,
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(414),
																												Column: int(11),
																											},
																											End: ast.Location{
																												Line: int(414),
																												Column: int(19),
																											},
																											File: p1,
																										},
																									},
																									ast.DesugaredObjectField{
																										Hide: ast.ObjectFieldHide(1),
																										Name: &ast.LiteralString{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													File: nil,
																												},
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: nil,
																											},
																											Value: "v",
																											Kind: ast.LiteralStringKind(1),
																											BlockIndent: "",
																											BlockTermIndent: "",
																										},
																										Body: &ast.LiteralString{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(414),
																														Column: int(24),
																													},
																													End: ast.Location{
																														Line: int(414),
																														Column: int(27),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p4678,
																												FreeVars: nil,
																											},
																											Value: "f",
																											Kind: ast.LiteralStringKind(1),
																											BlockIndent: "",
																											BlockTermIndent: "",
																										},
																										PlusSuper: false,
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(414),
																												Column: int(21),
																											},
																											End: ast.Location{
																												Line: int(414),
																												Column: int(27),
																											},
																											File: p1,
																										},
																									},
																									ast.DesugaredObjectField{
																										Hide: ast.ObjectFieldHide(1),
																										Name: &ast.LiteralString{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													File: nil,
																												},
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: nil,
																											},
																											Value: "caps",
																											Kind: ast.LiteralStringKind(1),
																											BlockIndent: "",
																											BlockTermIndent: "",
																										},
																										Body: &ast.LiteralBoolean{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(414),
																														Column: int(35),
																													},
																													End: ast.Location{
																														Line: int(414),
																														Column: int(39),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p4678,
																												FreeVars: nil,
																											},
																											Value: true,
																										},
																										PlusSuper: false,
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(414),
																												Column: int(29),
																											},
																											End: ast.Location{
																												Line: int(414),
																												Column: int(39),
																											},
																											File: p1,
																										},
																									},
																								},
																								Locals: ast.LocalBinds{},
																							},
																							ElseFodder: ast.Fodder{
																								ast.FodderElement{
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(6),
																									Comment: []string{},
																								},
																							},
																							BranchFalse: &ast.Conditional{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(415),
																											Column: int(12),
																										},
																										End: ast.Location{
																											Line: int(426),
																											Column: int(51),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p4463,
																									FreeVars: ast.Identifiers{
																										"c",
																										"i",
																									},
																								},
																								Cond: &ast.Binary{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(415),
																												Column: int(15),
																											},
																											End: ast.Location{
																												Line: int(415),
																												Column: int(23),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p4463,
																										FreeVars: ast.Identifiers{
																											"c",
																										},
																									},
																									Left: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(415),
																													Column: int(15),
																												},
																												End: ast.Location{
																													Line: int(415),
																													Column: int(16),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p4463,
																											FreeVars: ast.Identifiers{
																												"c",
																											},
																										},
																										Id: "c",
																									},
																									OpFodder: ast.Fodder{},
																									Op: ast.BinaryOp(12),
																									Right: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(415),
																													Column: int(20),
																												},
																												End: ast.Location{
																													Line: int(415),
																													Column: int(23),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p4463,
																											FreeVars: nil,
																										},
																										Value: "g",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																								},
																								ThenFodder: ast.Fodder{},
																								BranchTrue: &ast.DesugaredObject{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(416),
																												Column: int(9),
																											},
																											End: ast.Location{
																												Line: int(416),
																												Column: int(42),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{
																											ast.FodderElement{
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(8),
																												Comment: []string{},
																											},
																										},
																										Ctx: p4463,
																										FreeVars: ast.Identifiers{
																											"i",
																										},
																									},
																									Asserts: ast.Nodes{},
																									Fields: ast.DesugaredObjectFields{
																										ast.DesugaredObjectField{
																											Hide: ast.ObjectFieldHide(1),
																											Name: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: nil,
																												},
																												Value: "i",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																											Body: &ast.Binary{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(416),
																															Column: int(14),
																														},
																														End: ast.Location{
																															Line: int(416),
																															Column: int(19),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p4701,
																													FreeVars: ast.Identifiers{
																														"i",
																													},
																												},
																												Left: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(416),
																																Column: int(14),
																															},
																															End: ast.Location{
																																Line: int(416),
																																Column: int(15),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p4701,
																														FreeVars: ast.Identifiers{
																															"i",
																														},
																													},
																													Id: "i",
																												},
																												OpFodder: ast.Fodder{},
																												Op: ast.BinaryOp(3),
																												Right: &ast.LiteralNumber{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(416),
																																Column: int(18),
																															},
																															End: ast.Location{
																																Line: int(416),
																																Column: int(19),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p4701,
																														FreeVars: nil,
																													},
																													OriginalString: "1",
																												},
																											},
																											PlusSuper: false,
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(416),
																													Column: int(11),
																												},
																												End: ast.Location{
																													Line: int(416),
																													Column: int(19),
																												},
																												File: p1,
																											},
																										},
																										ast.DesugaredObjectField{
																											Hide: ast.ObjectFieldHide(1),
																											Name: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: nil,
																												},
																												Value: "v",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																											Body: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(416),
																															Column: int(24),
																														},
																														End: ast.Location{
																															Line: int(416),
																															Column: int(27),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p4701,
																													FreeVars: nil,
																												},
																												Value: "g",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																											PlusSuper: false,
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(416),
																													Column: int(21),
																												},
																												End: ast.Location{
																													Line: int(416),
																													Column: int(27),
																												},
																												File: p1,
																											},
																										},
																										ast.DesugaredObjectField{
																											Hide: ast.ObjectFieldHide(1),
																											Name: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: nil,
																												},
																												Value: "caps",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																											Body: &ast.LiteralBoolean{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(416),
																															Column: int(35),
																														},
																														End: ast.Location{
																															Line: int(416),
																															Column: int(40),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p4701,
																													FreeVars: nil,
																												},
																												Value: false,
																											},
																											PlusSuper: false,
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(416),
																													Column: int(29),
																												},
																												End: ast.Location{
																													Line: int(416),
																													Column: int(40),
																												},
																												File: p1,
																											},
																										},
																									},
																									Locals: ast.LocalBinds{},
																								},
																								ElseFodder: ast.Fodder{
																									ast.FodderElement{
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(6),
																										Comment: []string{},
																									},
																								},
																								BranchFalse: &ast.Conditional{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(417),
																												Column: int(12),
																											},
																											End: ast.Location{
																												Line: int(426),
																												Column: int(51),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p4463,
																										FreeVars: ast.Identifiers{
																											"c",
																											"i",
																										},
																									},
																									Cond: &ast.Binary{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(417),
																													Column: int(15),
																												},
																												End: ast.Location{
																													Line: int(417),
																													Column: int(23),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p4463,
																											FreeVars: ast.Identifiers{
																												"c",
																											},
																										},
																										Left: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(417),
																														Column: int(15),
																													},
																													End: ast.Location{
																														Line: int(417),
																														Column: int(16),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p4463,
																												FreeVars: ast.Identifiers{
																													"c",
																												},
																											},
																											Id: "c",
																										},
																										OpFodder: ast.Fodder{},
																										Op: ast.BinaryOp(12),
																										Right: &ast.LiteralString{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(417),
																														Column: int(20),
																													},
																													End: ast.Location{
																														Line: int(417),
																														Column: int(23),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p4463,
																												FreeVars: nil,
																											},
																											Value: "G",
																											Kind: ast.LiteralStringKind(1),
																											BlockIndent: "",
																											BlockTermIndent: "",
																										},
																									},
																									ThenFodder: ast.Fodder{},
																									BranchTrue: &ast.DesugaredObject{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(418),
																													Column: int(9),
																												},
																												End: ast.Location{
																													Line: int(418),
																													Column: int(41),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{
																												ast.FodderElement{
																													Kind: ast.FodderKind(0),
																													Blanks: int(0),
																													Indent: int(8),
																													Comment: []string{},
																												},
																											},
																											Ctx: p4463,
																											FreeVars: ast.Identifiers{
																												"i",
																											},
																										},
																										Asserts: ast.Nodes{},
																										Fields: ast.DesugaredObjectFields{
																											ast.DesugaredObjectField{
																												Hide: ast.ObjectFieldHide(1),
																												Name: &ast.LiteralString{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															File: nil,
																														},
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: nil,
																													},
																													Value: "i",
																													Kind: ast.LiteralStringKind(1),
																													BlockIndent: "",
																													BlockTermIndent: "",
																												},
																												Body: &ast.Binary{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(418),
																																Column: int(14),
																															},
																															End: ast.Location{
																																Line: int(418),
																																Column: int(19),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p4724,
																														FreeVars: ast.Identifiers{
																															"i",
																														},
																													},
																													Left: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(418),
																																	Column: int(14),
																																},
																																End: ast.Location{
																																	Line: int(418),
																																	Column: int(15),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p4724,
																															FreeVars: ast.Identifiers{
																																"i",
																															},
																														},
																														Id: "i",
																													},
																													OpFodder: ast.Fodder{},
																													Op: ast.BinaryOp(3),
																													Right: &ast.LiteralNumber{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(418),
																																	Column: int(18),
																																},
																																End: ast.Location{
																																	Line: int(418),
																																	Column: int(19),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p4724,
																															FreeVars: nil,
																														},
																														OriginalString: "1",
																													},
																												},
																												PlusSuper: false,
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(418),
																														Column: int(11),
																													},
																													End: ast.Location{
																														Line: int(418),
																														Column: int(19),
																													},
																													File: p1,
																												},
																											},
																											ast.DesugaredObjectField{
																												Hide: ast.ObjectFieldHide(1),
																												Name: &ast.LiteralString{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															File: nil,
																														},
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: nil,
																													},
																													Value: "v",
																													Kind: ast.LiteralStringKind(1),
																													BlockIndent: "",
																													BlockTermIndent: "",
																												},
																												Body: &ast.LiteralString{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(418),
																																Column: int(24),
																															},
																															End: ast.Location{
																																Line: int(418),
																																Column: int(27),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p4724,
																														FreeVars: nil,
																													},
																													Value: "g",
																													Kind: ast.LiteralStringKind(1),
																													BlockIndent: "",
																													BlockTermIndent: "",
																												},
																												PlusSuper: false,
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(418),
																														Column: int(21),
																													},
																													End: ast.Location{
																														Line: int(418),
																														Column: int(27),
																													},
																													File: p1,
																												},
																											},
																											ast.DesugaredObjectField{
																												Hide: ast.ObjectFieldHide(1),
																												Name: &ast.LiteralString{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															File: nil,
																														},
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: nil,
																													},
																													Value: "caps",
																													Kind: ast.LiteralStringKind(1),
																													BlockIndent: "",
																													BlockTermIndent: "",
																												},
																												Body: &ast.LiteralBoolean{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(418),
																																Column: int(35),
																															},
																															End: ast.Location{
																																Line: int(418),
																																Column: int(39),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p4724,
																														FreeVars: nil,
																													},
																													Value: true,
																												},
																												PlusSuper: false,
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(418),
																														Column: int(29),
																													},
																													End: ast.Location{
																														Line: int(418),
																														Column: int(39),
																													},
																													File: p1,
																												},
																											},
																										},
																										Locals: ast.LocalBinds{},
																									},
																									ElseFodder: ast.Fodder{
																										ast.FodderElement{
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(6),
																											Comment: []string{},
																										},
																									},
																									BranchFalse: &ast.Conditional{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(419),
																													Column: int(12),
																												},
																												End: ast.Location{
																													Line: int(426),
																													Column: int(51),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p4463,
																											FreeVars: ast.Identifiers{
																												"c",
																												"i",
																											},
																										},
																										Cond: &ast.Binary{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(419),
																														Column: int(15),
																													},
																													End: ast.Location{
																														Line: int(419),
																														Column: int(23),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p4463,
																												FreeVars: ast.Identifiers{
																													"c",
																												},
																											},
																											Left: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(419),
																															Column: int(15),
																														},
																														End: ast.Location{
																															Line: int(419),
																															Column: int(16),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p4463,
																													FreeVars: ast.Identifiers{
																														"c",
																													},
																												},
																												Id: "c",
																											},
																											OpFodder: ast.Fodder{},
																											Op: ast.BinaryOp(12),
																											Right: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(419),
																															Column: int(20),
																														},
																														End: ast.Location{
																															Line: int(419),
																															Column: int(23),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p4463,
																													FreeVars: nil,
																												},
																												Value: "c",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																										},
																										ThenFodder: ast.Fodder{},
																										BranchTrue: &ast.DesugaredObject{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(420),
																														Column: int(9),
																													},
																													End: ast.Location{
																														Line: int(420),
																														Column: int(42),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{
																													ast.FodderElement{
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(8),
																														Comment: []string{},
																													},
																												},
																												Ctx: p4463,
																												FreeVars: ast.Identifiers{
																													"i",
																												},
																											},
																											Asserts: ast.Nodes{},
																											Fields: ast.DesugaredObjectFields{
																												ast.DesugaredObjectField{
																													Hide: ast.ObjectFieldHide(1),
																													Name: &ast.LiteralString{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																End: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																File: nil,
																															},
																															Fodder: nil,
																															Ctx: nil,
																															FreeVars: nil,
																														},
																														Value: "i",
																														Kind: ast.LiteralStringKind(1),
																														BlockIndent: "",
																														BlockTermIndent: "",
																													},
																													Body: &ast.Binary{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(420),
																																	Column: int(14),
																																},
																																End: ast.Location{
																																	Line: int(420),
																																	Column: int(19),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p4747,
																															FreeVars: ast.Identifiers{
																																"i",
																															},
																														},
																														Left: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(420),
																																		Column: int(14),
																																	},
																																	End: ast.Location{
																																		Line: int(420),
																																		Column: int(15),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p4747,
																																FreeVars: ast.Identifiers{
																																	"i",
																																},
																															},
																															Id: "i",
																														},
																														OpFodder: ast.Fodder{},
																														Op: ast.BinaryOp(3),
																														Right: &ast.LiteralNumber{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(420),
																																		Column: int(18),
																																	},
																																	End: ast.Location{
																																		Line: int(420),
																																		Column: int(19),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p4747,
																																FreeVars: nil,
																															},
																															OriginalString: "1",
																														},
																													},
																													PlusSuper: false,
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(420),
																															Column: int(11),
																														},
																														End: ast.Location{
																															Line: int(420),
																															Column: int(19),
																														},
																														File: p1,
																													},
																												},
																												ast.DesugaredObjectField{
																													Hide: ast.ObjectFieldHide(1),
																													Name: &ast.LiteralString{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																End: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																File: nil,
																															},
																															Fodder: nil,
																															Ctx: nil,
																															FreeVars: nil,
																														},
																														Value: "v",
																														Kind: ast.LiteralStringKind(1),
																														BlockIndent: "",
																														BlockTermIndent: "",
																													},
																													Body: &ast.LiteralString{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(420),
																																	Column: int(24),
																																},
																																End: ast.Location{
																																	Line: int(420),
																																	Column: int(27),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p4747,
																															FreeVars: nil,
																														},
																														Value: "c",
																														Kind: ast.LiteralStringKind(1),
																														BlockIndent: "",
																														BlockTermIndent: "",
																													},
																													PlusSuper: false,
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(420),
																															Column: int(21),
																														},
																														End: ast.Location{
																															Line: int(420),
																															Column: int(27),
																														},
																														File: p1,
																													},
																												},
																												ast.DesugaredObjectField{
																													Hide: ast.ObjectFieldHide(1),
																													Name: &ast.LiteralString{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																End: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																File: nil,
																															},
																															Fodder: nil,
																															Ctx: nil,
																															FreeVars: nil,
																														},
																														Value: "caps",
																														Kind: ast.LiteralStringKind(1),
																														BlockIndent: "",
																														BlockTermIndent: "",
																													},
																													Body: &ast.LiteralBoolean{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(420),
																																	Column: int(35),
																																},
																																End: ast.Location{
																																	Line: int(420),
																																	Column: int(40),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p4747,
																															FreeVars: nil,
																														},
																														Value: false,
																													},
																													PlusSuper: false,
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(420),
																															Column: int(29),
																														},
																														End: ast.Location{
																															Line: int(420),
																															Column: int(40),
																														},
																														File: p1,
																													},
																												},
																											},
																											Locals: ast.LocalBinds{},
																										},
																										ElseFodder: ast.Fodder{
																											ast.FodderElement{
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(6),
																												Comment: []string{},
																											},
																										},
																										BranchFalse: &ast.Conditional{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(421),
																														Column: int(12),
																													},
																													End: ast.Location{
																														Line: int(426),
																														Column: int(51),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p4463,
																												FreeVars: ast.Identifiers{
																													"c",
																													"i",
																												},
																											},
																											Cond: &ast.Binary{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(421),
																															Column: int(15),
																														},
																														End: ast.Location{
																															Line: int(421),
																															Column: int(23),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p4463,
																													FreeVars: ast.Identifiers{
																														"c",
																													},
																												},
																												Left: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(421),
																																Column: int(15),
																															},
																															End: ast.Location{
																																Line: int(421),
																																Column: int(16),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p4463,
																														FreeVars: ast.Identifiers{
																															"c",
																														},
																													},
																													Id: "c",
																												},
																												OpFodder: ast.Fodder{},
																												Op: ast.BinaryOp(12),
																												Right: &ast.LiteralString{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(421),
																																Column: int(20),
																															},
																															End: ast.Location{
																																Line: int(421),
																																Column: int(23),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p4463,
																														FreeVars: nil,
																													},
																													Value: "s",
																													Kind: ast.LiteralStringKind(1),
																													BlockIndent: "",
																													BlockTermIndent: "",
																												},
																											},
																											ThenFodder: ast.Fodder{},
																											BranchTrue: &ast.DesugaredObject{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(422),
																															Column: int(9),
																														},
																														End: ast.Location{
																															Line: int(422),
																															Column: int(42),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(8),
																															Comment: []string{},
																														},
																													},
																													Ctx: p4463,
																													FreeVars: ast.Identifiers{
																														"i",
																													},
																												},
																												Asserts: ast.Nodes{},
																												Fields: ast.DesugaredObjectFields{
																													ast.DesugaredObjectField{
																														Hide: ast.ObjectFieldHide(1),
																														Name: &ast.LiteralString{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	File: nil,
																																},
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: nil,
																															},
																															Value: "i",
																															Kind: ast.LiteralStringKind(1),
																															BlockIndent: "",
																															BlockTermIndent: "",
																														},
																														Body: &ast.Binary{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(422),
																																		Column: int(14),
																																	},
																																	End: ast.Location{
																																		Line: int(422),
																																		Column: int(19),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p4770,
																																FreeVars: ast.Identifiers{
																																	"i",
																																},
																															},
																															Left: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(422),
																																			Column: int(14),
																																		},
																																		End: ast.Location{
																																			Line: int(422),
																																			Column: int(15),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p4770,
																																	FreeVars: ast.Identifiers{
																																		"i",
																																	},
																																},
																																Id: "i",
																															},
																															OpFodder: ast.Fodder{},
																															Op: ast.BinaryOp(3),
																															Right: &ast.LiteralNumber{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(422),
																																			Column: int(18),
																																		},
																																		End: ast.Location{
																																			Line: int(422),
																																			Column: int(19),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p4770,
																																	FreeVars: nil,
																																},
																																OriginalString: "1",
																															},
																														},
																														PlusSuper: false,
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(422),
																																Column: int(11),
																															},
																															End: ast.Location{
																																Line: int(422),
																																Column: int(19),
																															},
																															File: p1,
																														},
																													},
																													ast.DesugaredObjectField{
																														Hide: ast.ObjectFieldHide(1),
																														Name: &ast.LiteralString{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	File: nil,
																																},
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: nil,
																															},
																															Value: "v",
																															Kind: ast.LiteralStringKind(1),
																															BlockIndent: "",
																															BlockTermIndent: "",
																														},
																														Body: &ast.LiteralString{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(422),
																																		Column: int(24),
																																	},
																																	End: ast.Location{
																																		Line: int(422),
																																		Column: int(27),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p4770,
																																FreeVars: nil,
																															},
																															Value: "s",
																															Kind: ast.LiteralStringKind(1),
																															BlockIndent: "",
																															BlockTermIndent: "",
																														},
																														PlusSuper: false,
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(422),
																																Column: int(21),
																															},
																															End: ast.Location{
																																Line: int(422),
																																Column: int(27),
																															},
																															File: p1,
																														},
																													},
																													ast.DesugaredObjectField{
																														Hide: ast.ObjectFieldHide(1),
																														Name: &ast.LiteralString{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	File: nil,
																																},
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: nil,
																															},
																															Value: "caps",
																															Kind: ast.LiteralStringKind(1),
																															BlockIndent: "",
																															BlockTermIndent: "",
																														},
																														Body: &ast.LiteralBoolean{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(422),
																																		Column: int(35),
																																	},
																																	End: ast.Location{
																																		Line: int(422),
																																		Column: int(40),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p4770,
																																FreeVars: nil,
																															},
																															Value: false,
																														},
																														PlusSuper: false,
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(422),
																																Column: int(29),
																															},
																															End: ast.Location{
																																Line: int(422),
																																Column: int(40),
																															},
																															File: p1,
																														},
																													},
																												},
																												Locals: ast.LocalBinds{},
																											},
																											ElseFodder: ast.Fodder{
																												ast.FodderElement{
																													Kind: ast.FodderKind(0),
																													Blanks: int(0),
																													Indent: int(6),
																													Comment: []string{},
																												},
																											},
																											BranchFalse: &ast.Conditional{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(423),
																															Column: int(12),
																														},
																														End: ast.Location{
																															Line: int(426),
																															Column: int(51),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p4463,
																													FreeVars: ast.Identifiers{
																														"c",
																														"i",
																													},
																												},
																												Cond: &ast.Binary{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(423),
																																Column: int(15),
																															},
																															End: ast.Location{
																																Line: int(423),
																																Column: int(23),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p4463,
																														FreeVars: ast.Identifiers{
																															"c",
																														},
																													},
																													Left: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(423),
																																	Column: int(15),
																																},
																																End: ast.Location{
																																	Line: int(423),
																																	Column: int(16),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p4463,
																															FreeVars: ast.Identifiers{
																																"c",
																															},
																														},
																														Id: "c",
																													},
																													OpFodder: ast.Fodder{},
																													Op: ast.BinaryOp(12),
																													Right: &ast.LiteralString{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(423),
																																	Column: int(20),
																																},
																																End: ast.Location{
																																	Line: int(423),
																																	Column: int(23),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p4463,
																															FreeVars: nil,
																														},
																														Value: "%",
																														Kind: ast.LiteralStringKind(1),
																														BlockIndent: "",
																														BlockTermIndent: "",
																													},
																												},
																												ThenFodder: ast.Fodder{},
																												BranchTrue: &ast.DesugaredObject{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(424),
																																Column: int(9),
																															},
																															End: ast.Location{
																																Line: int(424),
																																Column: int(42),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(8),
																																Comment: []string{},
																															},
																														},
																														Ctx: p4463,
																														FreeVars: ast.Identifiers{
																															"i",
																														},
																													},
																													Asserts: ast.Nodes{},
																													Fields: ast.DesugaredObjectFields{
																														ast.DesugaredObjectField{
																															Hide: ast.ObjectFieldHide(1),
																															Name: &ast.LiteralString{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		File: nil,
																																	},
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: nil,
																																},
																																Value: "i",
																																Kind: ast.LiteralStringKind(1),
																																BlockIndent: "",
																																BlockTermIndent: "",
																															},
																															Body: &ast.Binary{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(424),
																																			Column: int(14),
																																		},
																																		End: ast.Location{
																																			Line: int(424),
																																			Column: int(19),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p4793,
																																	FreeVars: ast.Identifiers{
																																		"i",
																																	},
																																},
																																Left: &ast.Var{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(424),
																																				Column: int(14),
																																			},
																																			End: ast.Location{
																																				Line: int(424),
																																				Column: int(15),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p4793,
																																		FreeVars: ast.Identifiers{
																																			"i",
																																		},
																																	},
																																	Id: "i",
																																},
																																OpFodder: ast.Fodder{},
																																Op: ast.BinaryOp(3),
																																Right: &ast.LiteralNumber{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(424),
																																				Column: int(18),
																																			},
																																			End: ast.Location{
																																				Line: int(424),
																																				Column: int(19),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p4793,
																																		FreeVars: nil,
																																	},
																																	OriginalString: "1",
																																},
																															},
																															PlusSuper: false,
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(424),
																																	Column: int(11),
																																},
																																End: ast.Location{
																																	Line: int(424),
																																	Column: int(19),
																																},
																																File: p1,
																															},
																														},
																														ast.DesugaredObjectField{
																															Hide: ast.ObjectFieldHide(1),
																															Name: &ast.LiteralString{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		File: nil,
																																	},
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: nil,
																																},
																																Value: "v",
																																Kind: ast.LiteralStringKind(1),
																																BlockIndent: "",
																																BlockTermIndent: "",
																															},
																															Body: &ast.LiteralString{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(424),
																																			Column: int(24),
																																		},
																																		End: ast.Location{
																																			Line: int(424),
																																			Column: int(27),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p4793,
																																	FreeVars: nil,
																																},
																																Value: "%",
																																Kind: ast.LiteralStringKind(1),
																																BlockIndent: "",
																																BlockTermIndent: "",
																															},
																															PlusSuper: false,
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(424),
																																	Column: int(21),
																																},
																																End: ast.Location{
																																	Line: int(424),
																																	Column: int(27),
																																},
																																File: p1,
																															},
																														},
																														ast.DesugaredObjectField{
																															Hide: ast.ObjectFieldHide(1),
																															Name: &ast.LiteralString{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		File: nil,
																																	},
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: nil,
																																},
																																Value: "caps",
																																Kind: ast.LiteralStringKind(1),
																																BlockIndent: "",
																																BlockTermIndent: "",
																															},
																															Body: &ast.LiteralBoolean{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(424),
																																			Column: int(35),
																																		},
																																		End: ast.Location{
																																			Line: int(424),
																																			Column: int(40),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p4793,
																																	FreeVars: nil,
																																},
																																Value: false,
																															},
																															PlusSuper: false,
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(424),
																																	Column: int(29),
																																},
																																End: ast.Location{
																																	Line: int(424),
																																	Column: int(40),
																																},
																																File: p1,
																															},
																														},
																													},
																													Locals: ast.LocalBinds{},
																												},
																												ElseFodder: ast.Fodder{
																													ast.FodderElement{
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(6),
																														Comment: []string{},
																													},
																												},
																												BranchFalse: &ast.Error{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(426),
																																Column: int(9),
																															},
																															End: ast.Location{
																																Line: int(426),
																																Column: int(51),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(8),
																																Comment: []string{},
																															},
																														},
																														Ctx: p4463,
																														FreeVars: ast.Identifiers{
																															"c",
																														},
																													},
																													Expr: &ast.Binary{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(426),
																																	Column: int(15),
																																},
																																End: ast.Location{
																																	Line: int(426),
																																	Column: int(51),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p4463,
																															FreeVars: ast.Identifiers{
																																"c",
																															},
																														},
																														Left: &ast.LiteralString{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(426),
																																		Column: int(15),
																																	},
																																	End: ast.Location{
																																		Line: int(426),
																																		Column: int(47),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p4463,
																																FreeVars: nil,
																															},
																															Value: "Unrecognised conversion type: ",
																															Kind: ast.LiteralStringKind(1),
																															BlockIndent: "",
																															BlockTermIndent: "",
																														},
																														OpFodder: ast.Fodder{},
																														Op: ast.BinaryOp(3),
																														Right: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(426),
																																		Column: int(50),
																																	},
																																	End: ast.Location{
																																		Line: int(426),
																																		Column: int(51),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p4463,
																																FreeVars: ast.Identifiers{
																																	"c",
																																},
																															},
																															Id: "c",
																														},
																													},
																												},
																											},
																										},
																									},
																								},
																							},
																						},
																					},
																				},
																			},
																		},
																	},
																},
															},
														},
														ElseFodder: nil,
														BranchFalse: &ast.Error{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
															Expr: &ast.LiteralString{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(397),
																			Column: int(36),
																		},
																		End: ast.Location{
																			Line: int(397),
																			Column: int(60),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p4463,
																	FreeVars: nil,
																},
																Value: "Truncated format code.",
																Kind: ast.LiteralStringKind(1),
																BlockIndent: "",
																BlockTermIndent: "",
															},
														},
													},
												},
												Fun: nil,
												CloseFodder: nil,
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
											},
										},
										Body: &ast.Local{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(430),
														Column: int(5),
													},
													End: ast.Location{
														Line: int(774),
														Column: int(48),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(2),
														Indent: int(4),
														Comment: []string{},
													},
													ast.FodderElement{
														Kind: ast.FodderKind(2),
														Blanks: int(0),
														Indent: int(4),
														Comment: []string{
															"// Parsed initial %, now the rest.",
														},
													},
												},
												Ctx: p3497,
												FreeVars: ast.Identifiers{
													"parse_conv_type",
													"std",
													"str",
													"try_parse_cflags",
													"try_parse_field_width",
													"try_parse_length_modifier",
													"try_parse_mapping_key",
													"try_parse_precision",
													"vals",
												},
											},
											Binds: ast.LocalBinds{
												ast.LocalBind{
													VarFodder: nil,
													Variable: "parse_code",
													EqFodder: nil,
													Body: &ast.Function{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(430),
																	Column: int(11),
																},
																End: ast.Location{
																	Line: int(448),
																	Column: int(8),
																},
																File: p1,
															},
															Fodder: nil,
															Ctx: p4819,
															FreeVars: ast.Identifiers{
																"parse_conv_type",
																"std",
																"try_parse_cflags",
																"try_parse_field_width",
																"try_parse_length_modifier",
																"try_parse_mapping_key",
																"try_parse_precision",
															},
														},
														ParenLeftFodder: ast.Fodder{},
														Parameters: []ast.Parameter{
															ast.Parameter{
																NameFodder: ast.Fodder{},
																Name: "str",
																EqFodder: nil,
																DefaultArg: nil,
																CommaFodder: ast.Fodder{},
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(430),
																		Column: int(22),
																	},
																	End: ast.Location{
																		Line: int(430),
																		Column: int(25),
																	},
																	File: p1,
																},
															},
															ast.Parameter{
																NameFodder: ast.Fodder{},
																Name: "i",
																EqFodder: nil,
																DefaultArg: nil,
																CommaFodder: nil,
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(430),
																		Column: int(27),
																	},
																	End: ast.Location{
																		Line: int(430),
																		Column: int(28),
																	},
																	File: p1,
																},
															},
														},
														TrailingComma: false,
														ParenRightFodder: ast.Fodder{},
														Body: &ast.Conditional{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"i",
																	"parse_conv_type",
																	"std",
																	"str",
																	"try_parse_cflags",
																	"try_parse_field_width",
																	"try_parse_length_modifier",
																	"try_parse_mapping_key",
																	"try_parse_precision",
																},
															},
															Cond: &ast.Binary{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(431),
																			Column: int(14),
																		},
																		End: ast.Location{
																			Line: int(431),
																			Column: int(33),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p4825,
																	FreeVars: ast.Identifiers{
																		"i",
																		"std",
																		"str",
																	},
																},
																Left: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(431),
																				Column: int(14),
																			},
																			End: ast.Location{
																				Line: int(431),
																				Column: int(15),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p4825,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																	},
																	Id: "i",
																},
																OpFodder: ast.Fodder{},
																Op: ast.BinaryOp(9),
																Right: &ast.Apply{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(431),
																				Column: int(18),
																			},
																			End: ast.Location{
																				Line: int(431),
																				Column: int(33),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p4825,
																		FreeVars: ast.Identifiers{
																			"std",
																			"str",
																		},
																	},
																	Target: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(431),
																					Column: int(18),
																				},
																				End: ast.Location{
																					Line: int(431),
																					Column: int(28),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p4825,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(431),
																						Column: int(18),
																					},
																					End: ast.Location{
																						Line: int(431),
																						Column: int(21),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Id: "std",
																		},
																		LeftBracketFodder: ast.Fodder{},
																		Index: &ast.LiteralString{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: nil,
																			},
																			Value: "length",
																			Kind: ast.LiteralStringKind(1),
																			BlockIndent: "",
																			BlockTermIndent: "",
																		},
																		RightBracketFodder: ast.Fodder{},
																		Id: nil,
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(431),
																								Column: int(29),
																							},
																							End: ast.Location{
																								Line: int(431),
																								Column: int(32),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p4838,
																						FreeVars: ast.Identifiers{
																							"str",
																						},
																					},
																					Id: "str",
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	TrailingComma: false,
																	TailStrict: false,
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																},
															},
															ThenFodder: nil,
															BranchTrue: &ast.Local{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(432),
																			Column: int(7),
																		},
																		End: ast.Location{
																			Line: int(448),
																			Column: int(8),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(6),
																			Comment: []string{},
																		},
																	},
																	Ctx: p4825,
																	FreeVars: ast.Identifiers{
																		"i",
																		"parse_conv_type",
																		"str",
																		"try_parse_cflags",
																		"try_parse_field_width",
																		"try_parse_length_modifier",
																		"try_parse_mapping_key",
																		"try_parse_precision",
																	},
																},
																Binds: ast.LocalBinds{
																	ast.LocalBind{
																		VarFodder: ast.Fodder{},
																		Variable: "mkey",
																		EqFodder: ast.Fodder{},
																		Body: &ast.Apply{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(432),
																						Column: int(20),
																					},
																					End: ast.Location{
																						Line: int(432),
																						Column: int(49),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p4845,
																				FreeVars: ast.Identifiers{
																					"i",
																					"str",
																					"try_parse_mapping_key",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(432),
																							Column: int(20),
																						},
																						End: ast.Location{
																							Line: int(432),
																							Column: int(41),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p4845,
																					FreeVars: ast.Identifiers{
																						"try_parse_mapping_key",
																					},
																				},
																				Id: "try_parse_mapping_key",
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(432),
																										Column: int(42),
																									},
																									End: ast.Location{
																										Line: int(432),
																										Column: int(45),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p4851,
																								FreeVars: ast.Identifiers{
																									"str",
																								},
																							},
																							Id: "str",
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(432),
																										Column: int(47),
																									},
																									End: ast.Location{
																										Line: int(432),
																										Column: int(48),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p4851,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																							},
																							Id: "i",
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			TrailingComma: false,
																			TailStrict: false,
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																		},
																		Fun: nil,
																		CloseFodder: ast.Fodder{},
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(432),
																				Column: int(13),
																			},
																			End: ast.Location{
																				Line: int(432),
																				Column: int(49),
																			},
																			File: p1,
																		},
																	},
																},
																Body: &ast.Local{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(433),
																				Column: int(7),
																			},
																			End: ast.Location{
																				Line: int(448),
																				Column: int(8),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(6),
																				Comment: []string{},
																			},
																		},
																		Ctx: p4825,
																		FreeVars: ast.Identifiers{
																			"mkey",
																			"parse_conv_type",
																			"str",
																			"try_parse_cflags",
																			"try_parse_field_width",
																			"try_parse_length_modifier",
																			"try_parse_precision",
																		},
																	},
																	Binds: ast.LocalBinds{
																		ast.LocalBind{
																			VarFodder: ast.Fodder{},
																			Variable: "cflags",
																			EqFodder: ast.Fodder{},
																			Body: &ast.Apply{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(433),
																							Column: int(22),
																						},
																						End: ast.Location{
																							Line: int(433),
																							Column: int(51),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p4860,
																					FreeVars: ast.Identifiers{
																						"mkey",
																						"str",
																						"try_parse_cflags",
																					},
																				},
																				Target: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(433),
																								Column: int(22),
																							},
																							End: ast.Location{
																								Line: int(433),
																								Column: int(38),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p4860,
																						FreeVars: ast.Identifiers{
																							"try_parse_cflags",
																						},
																					},
																					Id: "try_parse_cflags",
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(433),
																											Column: int(39),
																										},
																										End: ast.Location{
																											Line: int(433),
																											Column: int(42),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p4866,
																									FreeVars: ast.Identifiers{
																										"str",
																									},
																								},
																								Id: "str",
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Index{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(433),
																											Column: int(44),
																										},
																										End: ast.Location{
																											Line: int(433),
																											Column: int(50),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p4866,
																									FreeVars: ast.Identifiers{
																										"mkey",
																									},
																								},
																								Target: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(433),
																												Column: int(44),
																											},
																											End: ast.Location{
																												Line: int(433),
																												Column: int(48),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: nil,
																										FreeVars: ast.Identifiers{
																											"mkey",
																										},
																									},
																									Id: "mkey",
																								},
																								LeftBracketFodder: ast.Fodder{},
																								Index: &ast.LiteralString{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											File: nil,
																										},
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: nil,
																									},
																									Value: "i",
																									Kind: ast.LiteralStringKind(1),
																									BlockIndent: "",
																									BlockTermIndent: "",
																								},
																								RightBracketFodder: ast.Fodder{},
																								Id: nil,
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				TrailingComma: false,
																				TailStrict: false,
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: nil,
																			},
																			Fun: nil,
																			CloseFodder: ast.Fodder{},
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(433),
																					Column: int(13),
																				},
																				End: ast.Location{
																					Line: int(433),
																					Column: int(51),
																				},
																				File: p1,
																			},
																		},
																	},
																	Body: &ast.Local{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(434),
																					Column: int(7),
																				},
																				End: ast.Location{
																					Line: int(448),
																					Column: int(8),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(6),
																					Comment: []string{},
																				},
																			},
																			Ctx: p4825,
																			FreeVars: ast.Identifiers{
																				"cflags",
																				"mkey",
																				"parse_conv_type",
																				"str",
																				"try_parse_field_width",
																				"try_parse_length_modifier",
																				"try_parse_precision",
																			},
																		},
																		Binds: ast.LocalBinds{
																			ast.LocalBind{
																				VarFodder: ast.Fodder{},
																				Variable: "fw",
																				EqFodder: ast.Fodder{},
																				Body: &ast.Apply{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(434),
																								Column: int(18),
																							},
																							End: ast.Location{
																								Line: int(434),
																								Column: int(54),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p4878,
																						FreeVars: ast.Identifiers{
																							"cflags",
																							"str",
																							"try_parse_field_width",
																						},
																					},
																					Target: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(434),
																									Column: int(18),
																								},
																								End: ast.Location{
																									Line: int(434),
																									Column: int(39),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p4878,
																							FreeVars: ast.Identifiers{
																								"try_parse_field_width",
																							},
																						},
																						Id: "try_parse_field_width",
																					},
																					FodderLeft: ast.Fodder{},
																					Arguments: ast.Arguments{
																						Positional: []ast.CommaSeparatedExpr{
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(434),
																												Column: int(40),
																											},
																											End: ast.Location{
																												Line: int(434),
																												Column: int(43),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p4884,
																										FreeVars: ast.Identifiers{
																											"str",
																										},
																									},
																									Id: "str",
																								},
																								CommaFodder: ast.Fodder{},
																							},
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Index{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(434),
																												Column: int(45),
																											},
																											End: ast.Location{
																												Line: int(434),
																												Column: int(53),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p4884,
																										FreeVars: ast.Identifiers{
																											"cflags",
																										},
																									},
																									Target: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(434),
																													Column: int(45),
																												},
																												End: ast.Location{
																													Line: int(434),
																													Column: int(51),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"cflags",
																											},
																										},
																										Id: "cflags",
																									},
																									LeftBracketFodder: ast.Fodder{},
																									Index: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												File: nil,
																											},
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: nil,
																										},
																										Value: "i",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																									RightBracketFodder: ast.Fodder{},
																									Id: nil,
																								},
																								CommaFodder: nil,
																							},
																						},
																						Named: nil,
																					},
																					TrailingComma: false,
																					TailStrict: false,
																					FodderRight: ast.Fodder{},
																					TailStrictFodder: nil,
																				},
																				Fun: nil,
																				CloseFodder: ast.Fodder{},
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(434),
																						Column: int(13),
																					},
																					End: ast.Location{
																						Line: int(434),
																						Column: int(54),
																					},
																					File: p1,
																				},
																			},
																		},
																		Body: &ast.Local{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(435),
																						Column: int(7),
																					},
																					End: ast.Location{
																						Line: int(448),
																						Column: int(8),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(6),
																						Comment: []string{},
																					},
																				},
																				Ctx: p4825,
																				FreeVars: ast.Identifiers{
																					"cflags",
																					"fw",
																					"mkey",
																					"parse_conv_type",
																					"str",
																					"try_parse_length_modifier",
																					"try_parse_precision",
																				},
																			},
																			Binds: ast.LocalBinds{
																				ast.LocalBind{
																					VarFodder: ast.Fodder{},
																					Variable: "prec",
																					EqFodder: ast.Fodder{},
																					Body: &ast.Apply{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(435),
																									Column: int(20),
																								},
																								End: ast.Location{
																									Line: int(435),
																									Column: int(50),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p4896,
																							FreeVars: ast.Identifiers{
																								"fw",
																								"str",
																								"try_parse_precision",
																							},
																						},
																						Target: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(435),
																										Column: int(20),
																									},
																									End: ast.Location{
																										Line: int(435),
																										Column: int(39),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p4896,
																								FreeVars: ast.Identifiers{
																									"try_parse_precision",
																								},
																							},
																							Id: "try_parse_precision",
																						},
																						FodderLeft: ast.Fodder{},
																						Arguments: ast.Arguments{
																							Positional: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(435),
																													Column: int(40),
																												},
																												End: ast.Location{
																													Line: int(435),
																													Column: int(43),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p4902,
																											FreeVars: ast.Identifiers{
																												"str",
																											},
																										},
																										Id: "str",
																									},
																									CommaFodder: ast.Fodder{},
																								},
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Index{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(435),
																													Column: int(45),
																												},
																												End: ast.Location{
																													Line: int(435),
																													Column: int(49),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p4902,
																											FreeVars: ast.Identifiers{
																												"fw",
																											},
																										},
																										Target: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(435),
																														Column: int(45),
																													},
																													End: ast.Location{
																														Line: int(435),
																														Column: int(47),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"fw",
																												},
																											},
																											Id: "fw",
																										},
																										LeftBracketFodder: ast.Fodder{},
																										Index: &ast.LiteralString{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													File: nil,
																												},
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: nil,
																											},
																											Value: "i",
																											Kind: ast.LiteralStringKind(1),
																											BlockIndent: "",
																											BlockTermIndent: "",
																										},
																										RightBracketFodder: ast.Fodder{},
																										Id: nil,
																									},
																									CommaFodder: nil,
																								},
																							},
																							Named: nil,
																						},
																						TrailingComma: false,
																						TailStrict: false,
																						FodderRight: ast.Fodder{},
																						TailStrictFodder: nil,
																					},
																					Fun: nil,
																					CloseFodder: ast.Fodder{},
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(435),
																							Column: int(13),
																						},
																						End: ast.Location{
																							Line: int(435),
																							Column: int(50),
																						},
																						File: p1,
																					},
																				},
																			},
																			Body: &ast.Local{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(436),
																							Column: int(7),
																						},
																						End: ast.Location{
																							Line: int(448),
																							Column: int(8),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(6),
																							Comment: []string{},
																						},
																					},
																					Ctx: p4825,
																					FreeVars: ast.Identifiers{
																						"cflags",
																						"fw",
																						"mkey",
																						"parse_conv_type",
																						"prec",
																						"str",
																						"try_parse_length_modifier",
																					},
																				},
																				Binds: ast.LocalBinds{
																					ast.LocalBind{
																						VarFodder: ast.Fodder{},
																						Variable: "len_mod",
																						EqFodder: ast.Fodder{},
																						Body: &ast.Apply{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(436),
																										Column: int(23),
																									},
																									End: ast.Location{
																										Line: int(436),
																										Column: int(61),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p4914,
																								FreeVars: ast.Identifiers{
																									"prec",
																									"str",
																									"try_parse_length_modifier",
																								},
																							},
																							Target: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(436),
																											Column: int(23),
																										},
																										End: ast.Location{
																											Line: int(436),
																											Column: int(48),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p4914,
																									FreeVars: ast.Identifiers{
																										"try_parse_length_modifier",
																									},
																								},
																								Id: "try_parse_length_modifier",
																							},
																							FodderLeft: ast.Fodder{},
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(436),
																														Column: int(49),
																													},
																													End: ast.Location{
																														Line: int(436),
																														Column: int(52),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p4920,
																												FreeVars: ast.Identifiers{
																													"str",
																												},
																											},
																											Id: "str",
																										},
																										CommaFodder: ast.Fodder{},
																									},
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Index{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(436),
																														Column: int(54),
																													},
																													End: ast.Location{
																														Line: int(436),
																														Column: int(60),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p4920,
																												FreeVars: ast.Identifiers{
																													"prec",
																												},
																											},
																											Target: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(436),
																															Column: int(54),
																														},
																														End: ast.Location{
																															Line: int(436),
																															Column: int(58),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"prec",
																													},
																												},
																												Id: "prec",
																											},
																											LeftBracketFodder: ast.Fodder{},
																											Index: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: nil,
																												},
																												Value: "i",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																											RightBracketFodder: ast.Fodder{},
																											Id: nil,
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							TrailingComma: false,
																							TailStrict: false,
																							FodderRight: ast.Fodder{},
																							TailStrictFodder: nil,
																						},
																						Fun: nil,
																						CloseFodder: ast.Fodder{},
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(436),
																								Column: int(13),
																							},
																							End: ast.Location{
																								Line: int(436),
																								Column: int(61),
																							},
																							File: p1,
																						},
																					},
																				},
																				Body: &ast.Local{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(437),
																								Column: int(7),
																							},
																							End: ast.Location{
																								Line: int(448),
																								Column: int(8),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(6),
																								Comment: []string{},
																							},
																						},
																						Ctx: p4825,
																						FreeVars: ast.Identifiers{
																							"cflags",
																							"fw",
																							"len_mod",
																							"mkey",
																							"parse_conv_type",
																							"prec",
																							"str",
																						},
																					},
																					Binds: ast.LocalBinds{
																						ast.LocalBind{
																							VarFodder: ast.Fodder{},
																							Variable: "ctype",
																							EqFodder: ast.Fodder{},
																							Body: &ast.Apply{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(437),
																											Column: int(21),
																										},
																										End: ast.Location{
																											Line: int(437),
																											Column: int(50),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p4932,
																									FreeVars: ast.Identifiers{
																										"len_mod",
																										"parse_conv_type",
																										"str",
																									},
																								},
																								Target: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(437),
																												Column: int(21),
																											},
																											End: ast.Location{
																												Line: int(437),
																												Column: int(36),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p4932,
																										FreeVars: ast.Identifiers{
																											"parse_conv_type",
																										},
																									},
																									Id: "parse_conv_type",
																								},
																								FodderLeft: ast.Fodder{},
																								Arguments: ast.Arguments{
																									Positional: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(437),
																															Column: int(37),
																														},
																														End: ast.Location{
																															Line: int(437),
																															Column: int(40),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p4938,
																													FreeVars: ast.Identifiers{
																														"str",
																													},
																												},
																												Id: "str",
																											},
																											CommaFodder: ast.Fodder{},
																										},
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(437),
																															Column: int(42),
																														},
																														End: ast.Location{
																															Line: int(437),
																															Column: int(49),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p4938,
																													FreeVars: ast.Identifiers{
																														"len_mod",
																													},
																												},
																												Id: "len_mod",
																											},
																											CommaFodder: nil,
																										},
																									},
																									Named: nil,
																								},
																								TrailingComma: false,
																								TailStrict: false,
																								FodderRight: ast.Fodder{},
																								TailStrictFodder: nil,
																							},
																							Fun: nil,
																							CloseFodder: ast.Fodder{},
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(437),
																									Column: int(13),
																								},
																								End: ast.Location{
																									Line: int(437),
																									Column: int(50),
																								},
																								File: p1,
																							},
																						},
																					},
																					Body: &ast.DesugaredObject{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(438),
																									Column: int(7),
																								},
																								End: ast.Location{
																									Line: int(448),
																									Column: int(8),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{
																								ast.FodderElement{
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(6),
																									Comment: []string{},
																								},
																							},
																							Ctx: p4825,
																							FreeVars: ast.Identifiers{
																								"cflags",
																								"ctype",
																								"fw",
																								"mkey",
																								"prec",
																							},
																						},
																						Asserts: ast.Nodes{},
																						Fields: ast.DesugaredObjectFields{
																							ast.DesugaredObjectField{
																								Hide: ast.ObjectFieldHide(1),
																								Name: &ast.LiteralString{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											File: nil,
																										},
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: nil,
																									},
																									Value: "i",
																									Kind: ast.LiteralStringKind(1),
																									BlockIndent: "",
																									BlockTermIndent: "",
																								},
																								Body: &ast.Index{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(439),
																												Column: int(12),
																											},
																											End: ast.Location{
																												Line: int(439),
																												Column: int(19),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p4948,
																										FreeVars: ast.Identifiers{
																											"ctype",
																										},
																									},
																									Target: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(439),
																													Column: int(12),
																												},
																												End: ast.Location{
																													Line: int(439),
																													Column: int(17),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"ctype",
																											},
																										},
																										Id: "ctype",
																									},
																									LeftBracketFodder: ast.Fodder{},
																									Index: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												File: nil,
																											},
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: nil,
																										},
																										Value: "i",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																									RightBracketFodder: ast.Fodder{},
																									Id: nil,
																								},
																								PlusSuper: false,
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(439),
																										Column: int(9),
																									},
																									End: ast.Location{
																										Line: int(439),
																										Column: int(19),
																									},
																									File: p1,
																								},
																							},
																							ast.DesugaredObjectField{
																								Hide: ast.ObjectFieldHide(1),
																								Name: &ast.LiteralString{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											File: nil,
																										},
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: nil,
																									},
																									Value: "code",
																									Kind: ast.LiteralStringKind(1),
																									BlockIndent: "",
																									BlockTermIndent: "",
																								},
																								Body: &ast.DesugaredObject{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(440),
																												Column: int(15),
																											},
																											End: ast.Location{
																												Line: int(447),
																												Column: int(10),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p4948,
																										FreeVars: ast.Identifiers{
																											"cflags",
																											"ctype",
																											"fw",
																											"mkey",
																											"prec",
																										},
																									},
																									Asserts: ast.Nodes{},
																									Fields: ast.DesugaredObjectFields{
																										ast.DesugaredObjectField{
																											Hide: ast.ObjectFieldHide(1),
																											Name: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: nil,
																												},
																												Value: "mkey",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																											Body: &ast.Index{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(441),
																															Column: int(17),
																														},
																														End: ast.Location{
																															Line: int(441),
																															Column: int(23),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p4959,
																													FreeVars: ast.Identifiers{
																														"mkey",
																													},
																												},
																												Target: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(441),
																																Column: int(17),
																															},
																															End: ast.Location{
																																Line: int(441),
																																Column: int(21),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"mkey",
																														},
																													},
																													Id: "mkey",
																												},
																												LeftBracketFodder: ast.Fodder{},
																												Index: &ast.LiteralString{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															File: nil,
																														},
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: nil,
																													},
																													Value: "v",
																													Kind: ast.LiteralStringKind(1),
																													BlockIndent: "",
																													BlockTermIndent: "",
																												},
																												RightBracketFodder: ast.Fodder{},
																												Id: nil,
																											},
																											PlusSuper: false,
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(441),
																													Column: int(11),
																												},
																												End: ast.Location{
																													Line: int(441),
																													Column: int(23),
																												},
																												File: p1,
																											},
																										},
																										ast.DesugaredObjectField{
																											Hide: ast.ObjectFieldHide(1),
																											Name: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: nil,
																												},
																												Value: "cflags",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																											Body: &ast.Index{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(442),
																															Column: int(19),
																														},
																														End: ast.Location{
																															Line: int(442),
																															Column: int(27),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p4959,
																													FreeVars: ast.Identifiers{
																														"cflags",
																													},
																												},
																												Target: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(442),
																																Column: int(19),
																															},
																															End: ast.Location{
																																Line: int(442),
																																Column: int(25),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"cflags",
																														},
																													},
																													Id: "cflags",
																												},
																												LeftBracketFodder: ast.Fodder{},
																												Index: &ast.LiteralString{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															File: nil,
																														},
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: nil,
																													},
																													Value: "v",
																													Kind: ast.LiteralStringKind(1),
																													BlockIndent: "",
																													BlockTermIndent: "",
																												},
																												RightBracketFodder: ast.Fodder{},
																												Id: nil,
																											},
																											PlusSuper: false,
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(442),
																													Column: int(11),
																												},
																												End: ast.Location{
																													Line: int(442),
																													Column: int(27),
																												},
																												File: p1,
																											},
																										},
																										ast.DesugaredObjectField{
																											Hide: ast.ObjectFieldHide(1),
																											Name: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: nil,
																												},
																												Value: "fw",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																											Body: &ast.Index{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(443),
																															Column: int(15),
																														},
																														End: ast.Location{
																															Line: int(443),
																															Column: int(19),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p4959,
																													FreeVars: ast.Identifiers{
																														"fw",
																													},
																												},
																												Target: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(443),
																																Column: int(15),
																															},
																															End: ast.Location{
																																Line: int(443),
																																Column: int(17),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"fw",
																														},
																													},
																													Id: "fw",
																												},
																												LeftBracketFodder: ast.Fodder{},
																												Index: &ast.LiteralString{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															File: nil,
																														},
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: nil,
																													},
																													Value: "v",
																													Kind: ast.LiteralStringKind(1),
																													BlockIndent: "",
																													BlockTermIndent: "",
																												},
																												RightBracketFodder: ast.Fodder{},
																												Id: nil,
																											},
																											PlusSuper: false,
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(443),
																													Column: int(11),
																												},
																												End: ast.Location{
																													Line: int(443),
																													Column: int(19),
																												},
																												File: p1,
																											},
																										},
																										ast.DesugaredObjectField{
																											Hide: ast.ObjectFieldHide(1),
																											Name: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: nil,
																												},
																												Value: "prec",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																											Body: &ast.Index{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(444),
																															Column: int(17),
																														},
																														End: ast.Location{
																															Line: int(444),
																															Column: int(23),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p4959,
																													FreeVars: ast.Identifiers{
																														"prec",
																													},
																												},
																												Target: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(444),
																																Column: int(17),
																															},
																															End: ast.Location{
																																Line: int(444),
																																Column: int(21),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"prec",
																														},
																													},
																													Id: "prec",
																												},
																												LeftBracketFodder: ast.Fodder{},
																												Index: &ast.LiteralString{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															File: nil,
																														},
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: nil,
																													},
																													Value: "v",
																													Kind: ast.LiteralStringKind(1),
																													BlockIndent: "",
																													BlockTermIndent: "",
																												},
																												RightBracketFodder: ast.Fodder{},
																												Id: nil,
																											},
																											PlusSuper: false,
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(444),
																													Column: int(11),
																												},
																												End: ast.Location{
																													Line: int(444),
																													Column: int(23),
																												},
																												File: p1,
																											},
																										},
																										ast.DesugaredObjectField{
																											Hide: ast.ObjectFieldHide(1),
																											Name: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: nil,
																												},
																												Value: "ctype",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																											Body: &ast.Index{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(445),
																															Column: int(18),
																														},
																														End: ast.Location{
																															Line: int(445),
																															Column: int(25),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p4959,
																													FreeVars: ast.Identifiers{
																														"ctype",
																													},
																												},
																												Target: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(445),
																																Column: int(18),
																															},
																															End: ast.Location{
																																Line: int(445),
																																Column: int(23),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"ctype",
																														},
																													},
																													Id: "ctype",
																												},
																												LeftBracketFodder: ast.Fodder{},
																												Index: &ast.LiteralString{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															File: nil,
																														},
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: nil,
																													},
																													Value: "v",
																													Kind: ast.LiteralStringKind(1),
																													BlockIndent: "",
																													BlockTermIndent: "",
																												},
																												RightBracketFodder: ast.Fodder{},
																												Id: nil,
																											},
																											PlusSuper: false,
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(445),
																													Column: int(11),
																												},
																												End: ast.Location{
																													Line: int(445),
																													Column: int(25),
																												},
																												File: p1,
																											},
																										},
																										ast.DesugaredObjectField{
																											Hide: ast.ObjectFieldHide(1),
																											Name: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: nil,
																												},
																												Value: "caps",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																											Body: &ast.Index{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(446),
																															Column: int(17),
																														},
																														End: ast.Location{
																															Line: int(446),
																															Column: int(27),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p4959,
																													FreeVars: ast.Identifiers{
																														"ctype",
																													},
																												},
																												Target: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(446),
																																Column: int(17),
																															},
																															End: ast.Location{
																																Line: int(446),
																																Column: int(22),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"ctype",
																														},
																													},
																													Id: "ctype",
																												},
																												LeftBracketFodder: ast.Fodder{},
																												Index: &ast.LiteralString{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															File: nil,
																														},
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: nil,
																													},
																													Value: "caps",
																													Kind: ast.LiteralStringKind(1),
																													BlockIndent: "",
																													BlockTermIndent: "",
																												},
																												RightBracketFodder: ast.Fodder{},
																												Id: nil,
																											},
																											PlusSuper: false,
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(446),
																													Column: int(11),
																												},
																												End: ast.Location{
																													Line: int(446),
																													Column: int(27),
																												},
																												File: p1,
																											},
																										},
																									},
																									Locals: ast.LocalBinds{},
																								},
																								PlusSuper: false,
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(440),
																										Column: int(9),
																									},
																									End: ast.Location{
																										Line: int(447),
																										Column: int(10),
																									},
																									File: p1,
																								},
																							},
																						},
																						Locals: ast.LocalBinds{},
																					},
																				},
																			},
																		},
																	},
																},
															},
															ElseFodder: nil,
															BranchFalse: &ast.Error{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: nil,
																},
																Expr: &ast.LiteralString{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(431),
																				Column: int(36),
																			},
																			End: ast.Location{
																				Line: int(431),
																				Column: int(60),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p4825,
																		FreeVars: nil,
																	},
																	Value: "Truncated format code.",
																	Kind: ast.LiteralStringKind(1),
																	BlockIndent: "",
																	BlockTermIndent: "",
																},
															},
														},
													},
													Fun: nil,
													CloseFodder: nil,
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
												},
											},
											Body: &ast.Local{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(451),
															Column: int(5),
														},
														End: ast.Location{
															Line: int(774),
															Column: int(48),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(1),
															Indent: int(4),
															Comment: []string{},
														},
														ast.FodderElement{
															Kind: ast.FodderKind(2),
															Blanks: int(0),
															Indent: int(4),
															Comment: []string{
																"// Parse a format string (containing none or more % format tags).",
															},
														},
													},
													Ctx: p3497,
													FreeVars: ast.Identifiers{
														"parse_code",
														"std",
														"str",
														"vals",
													},
												},
												Binds: ast.LocalBinds{
													ast.LocalBind{
														VarFodder: nil,
														Variable: "parse_codes",
														EqFodder: nil,
														Body: &ast.Function{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(451),
																		Column: int(11),
																	},
																	End: ast.Location{
																		Line: int(460),
																		Column: int(48),
																	},
																	File: p1,
																},
																Fodder: nil,
																Ctx: p5002,
																FreeVars: ast.Identifiers{
																	"parse_code",
																	"parse_codes",
																	"std",
																},
															},
															ParenLeftFodder: ast.Fodder{},
															Parameters: []ast.Parameter{
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "str",
																	EqFodder: nil,
																	DefaultArg: nil,
																	CommaFodder: ast.Fodder{},
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(451),
																			Column: int(23),
																		},
																		End: ast.Location{
																			Line: int(451),
																			Column: int(26),
																		},
																		File: p1,
																	},
																},
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "i",
																	EqFodder: nil,
																	DefaultArg: nil,
																	CommaFodder: ast.Fodder{},
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(451),
																			Column: int(28),
																		},
																		End: ast.Location{
																			Line: int(451),
																			Column: int(29),
																		},
																		File: p1,
																	},
																},
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "out",
																	EqFodder: nil,
																	DefaultArg: nil,
																	CommaFodder: ast.Fodder{},
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(451),
																			Column: int(31),
																		},
																		End: ast.Location{
																			Line: int(451),
																			Column: int(34),
																		},
																		File: p1,
																	},
																},
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "cur",
																	EqFodder: nil,
																	DefaultArg: nil,
																	CommaFodder: nil,
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(451),
																			Column: int(36),
																		},
																		End: ast.Location{
																			Line: int(451),
																			Column: int(39),
																		},
																		File: p1,
																	},
																},
															},
															TrailingComma: false,
															ParenRightFodder: ast.Fodder{},
															Body: &ast.Conditional{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(452),
																			Column: int(7),
																		},
																		End: ast.Location{
																			Line: int(460),
																			Column: int(48),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(6),
																			Comment: []string{},
																		},
																	},
																	Ctx: p5007,
																	FreeVars: ast.Identifiers{
																		"cur",
																		"i",
																		"out",
																		"parse_code",
																		"parse_codes",
																		"std",
																		"str",
																	},
																},
																Cond: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(452),
																				Column: int(10),
																			},
																			End: ast.Location{
																				Line: int(452),
																				Column: int(30),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p5007,
																		FreeVars: ast.Identifiers{
																			"i",
																			"std",
																			"str",
																		},
																	},
																	Left: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(452),
																					Column: int(10),
																				},
																				End: ast.Location{
																					Line: int(452),
																					Column: int(11),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p5007,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																		},
																		Id: "i",
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(8),
																	Right: &ast.Apply{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(452),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(452),
																					Column: int(30),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p5007,
																			FreeVars: ast.Identifiers{
																				"std",
																				"str",
																			},
																		},
																		Target: &ast.Index{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(452),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(452),
																						Column: int(25),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p5007,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(452),
																							Column: int(15),
																						},
																						End: ast.Location{
																							Line: int(452),
																							Column: int(18),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																				},
																				Id: "std",
																			},
																			LeftBracketFodder: ast.Fodder{},
																			Index: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: nil,
																				},
																				Value: "length",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																			RightBracketFodder: ast.Fodder{},
																			Id: nil,
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(452),
																									Column: int(26),
																								},
																								End: ast.Location{
																									Line: int(452),
																									Column: int(29),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p5022,
																							FreeVars: ast.Identifiers{
																								"str",
																							},
																						},
																						Id: "str",
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		TrailingComma: false,
																		TailStrict: false,
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																	},
																},
																ThenFodder: ast.Fodder{},
																BranchTrue: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(453),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(453),
																				Column: int(20),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p5007,
																		FreeVars: ast.Identifiers{
																			"cur",
																			"out",
																		},
																	},
																	Left: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(453),
																					Column: int(9),
																				},
																				End: ast.Location{
																					Line: int(453),
																					Column: int(12),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(8),
																					Comment: []string{},
																				},
																			},
																			Ctx: p5007,
																			FreeVars: ast.Identifiers{
																				"out",
																			},
																		},
																		Id: "out",
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(3),
																	Right: &ast.Array{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(453),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(453),
																					Column: int(20),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p5007,
																			FreeVars: ast.Identifiers{
																				"cur",
																			},
																		},
																		Elements: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(453),
																								Column: int(16),
																							},
																							End: ast.Location{
																								Line: int(453),
																								Column: int(19),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p5033,
																						FreeVars: ast.Identifiers{
																							"cur",
																						},
																					},
																					Id: "cur",
																				},
																				CommaFodder: nil,
																			},
																		},
																		TrailingComma: false,
																		CloseFodder: ast.Fodder{},
																	},
																},
																ElseFodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(6),
																		Comment: []string{},
																	},
																},
																BranchFalse: &ast.Local{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(455),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(460),
																				Column: int(48),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(8),
																				Comment: []string{},
																			},
																		},
																		Ctx: p5007,
																		FreeVars: ast.Identifiers{
																			"cur",
																			"i",
																			"out",
																			"parse_code",
																			"parse_codes",
																			"str",
																		},
																	},
																	Binds: ast.LocalBinds{
																		ast.LocalBind{
																			VarFodder: ast.Fodder{},
																			Variable: "c",
																			EqFodder: ast.Fodder{},
																			Body: &ast.Index{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(455),
																							Column: int(19),
																						},
																						End: ast.Location{
																							Line: int(455),
																							Column: int(25),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p5041,
																					FreeVars: ast.Identifiers{
																						"i",
																						"str",
																					},
																				},
																				Target: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(455),
																								Column: int(19),
																							},
																							End: ast.Location{
																								Line: int(455),
																								Column: int(22),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p5041,
																						FreeVars: ast.Identifiers{
																							"str",
																						},
																					},
																					Id: "str",
																				},
																				LeftBracketFodder: ast.Fodder{},
																				Index: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(455),
																								Column: int(23),
																							},
																							End: ast.Location{
																								Line: int(455),
																								Column: int(24),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p5041,
																						FreeVars: ast.Identifiers{
																							"i",
																						},
																					},
																					Id: "i",
																				},
																				RightBracketFodder: ast.Fodder{},
																				Id: nil,
																			},
																			Fun: nil,
																			CloseFodder: ast.Fodder{},
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(455),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(455),
																					Column: int(25),
																				},
																				File: p1,
																			},
																		},
																	},
																	Body: &ast.Conditional{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(456),
																					Column: int(9),
																				},
																				End: ast.Location{
																					Line: int(460),
																					Column: int(48),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(8),
																					Comment: []string{},
																				},
																			},
																			Ctx: p5007,
																			FreeVars: ast.Identifiers{
																				"c",
																				"cur",
																				"i",
																				"out",
																				"parse_code",
																				"parse_codes",
																				"str",
																			},
																		},
																		Cond: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(456),
																						Column: int(12),
																					},
																					End: ast.Location{
																						Line: int(456),
																						Column: int(20),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p5007,
																				FreeVars: ast.Identifiers{
																					"c",
																				},
																			},
																			Left: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(456),
																							Column: int(12),
																						},
																						End: ast.Location{
																							Line: int(456),
																							Column: int(13),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p5007,
																					FreeVars: ast.Identifiers{
																						"c",
																					},
																				},
																				Id: "c",
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(12),
																			Right: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(456),
																							Column: int(17),
																						},
																						End: ast.Location{
																							Line: int(456),
																							Column: int(20),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p5007,
																					FreeVars: nil,
																				},
																				Value: "%",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																		},
																		ThenFodder: ast.Fodder{},
																		BranchTrue: &ast.Local{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(457),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(458),
																						Column: int(57),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(10),
																						Comment: []string{},
																					},
																				},
																				Ctx: p5007,
																				FreeVars: ast.Identifiers{
																					"cur",
																					"i",
																					"out",
																					"parse_code",
																					"parse_codes",
																					"str",
																				},
																			},
																			Binds: ast.LocalBinds{
																				ast.LocalBind{
																					VarFodder: ast.Fodder{},
																					Variable: "r",
																					EqFodder: ast.Fodder{},
																					Body: &ast.Apply{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(457),
																									Column: int(21),
																								},
																								End: ast.Location{
																									Line: int(457),
																									Column: int(43),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p5060,
																							FreeVars: ast.Identifiers{
																								"i",
																								"parse_code",
																								"str",
																							},
																						},
																						Target: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(457),
																										Column: int(21),
																									},
																									End: ast.Location{
																										Line: int(457),
																										Column: int(31),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p5060,
																								FreeVars: ast.Identifiers{
																									"parse_code",
																								},
																							},
																							Id: "parse_code",
																						},
																						FodderLeft: ast.Fodder{},
																						Arguments: ast.Arguments{
																							Positional: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(457),
																													Column: int(32),
																												},
																												End: ast.Location{
																													Line: int(457),
																													Column: int(35),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p5066,
																											FreeVars: ast.Identifiers{
																												"str",
																											},
																										},
																										Id: "str",
																									},
																									CommaFodder: ast.Fodder{},
																								},
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Binary{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(457),
																													Column: int(37),
																												},
																												End: ast.Location{
																													Line: int(457),
																													Column: int(42),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p5066,
																											FreeVars: ast.Identifiers{
																												"i",
																											},
																										},
																										Left: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(457),
																														Column: int(37),
																													},
																													End: ast.Location{
																														Line: int(457),
																														Column: int(38),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p5066,
																												FreeVars: ast.Identifiers{
																													"i",
																												},
																											},
																											Id: "i",
																										},
																										OpFodder: ast.Fodder{},
																										Op: ast.BinaryOp(3),
																										Right: &ast.LiteralNumber{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(457),
																														Column: int(41),
																													},
																													End: ast.Location{
																														Line: int(457),
																														Column: int(42),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p5066,
																												FreeVars: nil,
																											},
																											OriginalString: "1",
																										},
																									},
																									CommaFodder: nil,
																								},
																							},
																							Named: nil,
																						},
																						TrailingComma: false,
																						TailStrict: false,
																						FodderRight: ast.Fodder{},
																						TailStrictFodder: nil,
																					},
																					Fun: nil,
																					CloseFodder: ast.Fodder{},
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(457),
																							Column: int(17),
																						},
																						End: ast.Location{
																							Line: int(457),
																							Column: int(43),
																						},
																						File: p1,
																					},
																				},
																			},
																			Body: &ast.Apply{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(458),
																							Column: int(11),
																						},
																						End: ast.Location{
																							Line: int(458),
																							Column: int(57),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p5007,
																					FreeVars: ast.Identifiers{
																						"cur",
																						"out",
																						"parse_codes",
																						"r",
																						"str",
																					},
																				},
																				Target: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(458),
																								Column: int(11),
																							},
																							End: ast.Location{
																								Line: int(458),
																								Column: int(22),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(10),
																								Comment: []string{},
																							},
																						},
																						Ctx: p5007,
																						FreeVars: ast.Identifiers{
																							"parse_codes",
																						},
																					},
																					Id: "parse_codes",
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(458),
																											Column: int(23),
																										},
																										End: ast.Location{
																											Line: int(458),
																											Column: int(26),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p5080,
																									FreeVars: ast.Identifiers{
																										"str",
																									},
																								},
																								Id: "str",
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Index{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(458),
																											Column: int(28),
																										},
																										End: ast.Location{
																											Line: int(458),
																											Column: int(31),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p5080,
																									FreeVars: ast.Identifiers{
																										"r",
																									},
																								},
																								Target: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(458),
																												Column: int(28),
																											},
																											End: ast.Location{
																												Line: int(458),
																												Column: int(29),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: nil,
																										FreeVars: ast.Identifiers{
																											"r",
																										},
																									},
																									Id: "r",
																								},
																								LeftBracketFodder: ast.Fodder{},
																								Index: &ast.LiteralString{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											File: nil,
																										},
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: nil,
																									},
																									Value: "i",
																									Kind: ast.LiteralStringKind(1),
																									BlockIndent: "",
																									BlockTermIndent: "",
																								},
																								RightBracketFodder: ast.Fodder{},
																								Id: nil,
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Binary{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(458),
																											Column: int(33),
																										},
																										End: ast.Location{
																											Line: int(458),
																											Column: int(52),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p5080,
																									FreeVars: ast.Identifiers{
																										"cur",
																										"out",
																										"r",
																									},
																								},
																								Left: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(458),
																												Column: int(33),
																											},
																											End: ast.Location{
																												Line: int(458),
																												Column: int(36),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p5080,
																										FreeVars: ast.Identifiers{
																											"out",
																										},
																									},
																									Id: "out",
																								},
																								OpFodder: ast.Fodder{},
																								Op: ast.BinaryOp(3),
																								Right: &ast.Array{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(458),
																												Column: int(39),
																											},
																											End: ast.Location{
																												Line: int(458),
																												Column: int(52),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p5080,
																										FreeVars: ast.Identifiers{
																											"cur",
																											"r",
																										},
																									},
																									Elements: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(458),
																															Column: int(40),
																														},
																														End: ast.Location{
																															Line: int(458),
																															Column: int(43),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p5095,
																													FreeVars: ast.Identifiers{
																														"cur",
																													},
																												},
																												Id: "cur",
																											},
																											CommaFodder: ast.Fodder{},
																										},
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Index{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(458),
																															Column: int(45),
																														},
																														End: ast.Location{
																															Line: int(458),
																															Column: int(51),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p5095,
																													FreeVars: ast.Identifiers{
																														"r",
																													},
																												},
																												Target: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(458),
																																Column: int(45),
																															},
																															End: ast.Location{
																																Line: int(458),
																																Column: int(46),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"r",
																														},
																													},
																													Id: "r",
																												},
																												LeftBracketFodder: ast.Fodder{},
																												Index: &ast.LiteralString{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															File: nil,
																														},
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: nil,
																													},
																													Value: "code",
																													Kind: ast.LiteralStringKind(1),
																													BlockIndent: "",
																													BlockTermIndent: "",
																												},
																												RightBracketFodder: ast.Fodder{},
																												Id: nil,
																											},
																											CommaFodder: nil,
																										},
																									},
																									TrailingComma: false,
																									CloseFodder: ast.Fodder{},
																								},
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.LiteralString{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(458),
																											Column: int(54),
																										},
																										End: ast.Location{
																											Line: int(458),
																											Column: int(56),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p5080,
																									FreeVars: nil,
																								},
																								Value: "",
																								Kind: ast.LiteralStringKind(1),
																								BlockIndent: "",
																								BlockTermIndent: "",
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				TrailingComma: false,
																				TailStrict: true,
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: ast.Fodder{},
																			},
																		},
																		ElseFodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(8),
																				Comment: []string{},
																			},
																		},
																		BranchFalse: &ast.Apply{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(460),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(460),
																						Column: int(48),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p5007,
																				FreeVars: ast.Identifiers{
																					"c",
																					"cur",
																					"i",
																					"out",
																					"parse_codes",
																					"str",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(460),
																							Column: int(11),
																						},
																						End: ast.Location{
																							Line: int(460),
																							Column: int(22),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(10),
																							Comment: []string{},
																						},
																					},
																					Ctx: p5007,
																					FreeVars: ast.Identifiers{
																						"parse_codes",
																					},
																				},
																				Id: "parse_codes",
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(460),
																										Column: int(23),
																									},
																									End: ast.Location{
																										Line: int(460),
																										Column: int(26),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p5111,
																								FreeVars: ast.Identifiers{
																									"str",
																								},
																							},
																							Id: "str",
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Binary{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(460),
																										Column: int(28),
																									},
																									End: ast.Location{
																										Line: int(460),
																										Column: int(33),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p5111,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																							},
																							Left: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(460),
																											Column: int(28),
																										},
																										End: ast.Location{
																											Line: int(460),
																											Column: int(29),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p5111,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																								},
																								Id: "i",
																							},
																							OpFodder: ast.Fodder{},
																							Op: ast.BinaryOp(3),
																							Right: &ast.LiteralNumber{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(460),
																											Column: int(32),
																										},
																										End: ast.Location{
																											Line: int(460),
																											Column: int(33),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p5111,
																									FreeVars: nil,
																								},
																								OriginalString: "1",
																							},
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(460),
																										Column: int(35),
																									},
																									End: ast.Location{
																										Line: int(460),
																										Column: int(38),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p5111,
																								FreeVars: ast.Identifiers{
																									"out",
																								},
																							},
																							Id: "out",
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Binary{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(460),
																										Column: int(40),
																									},
																									End: ast.Location{
																										Line: int(460),
																										Column: int(47),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p5111,
																								FreeVars: ast.Identifiers{
																									"c",
																									"cur",
																								},
																							},
																							Left: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(460),
																											Column: int(40),
																										},
																										End: ast.Location{
																											Line: int(460),
																											Column: int(43),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p5111,
																									FreeVars: ast.Identifiers{
																										"cur",
																									},
																								},
																								Id: "cur",
																							},
																							OpFodder: ast.Fodder{},
																							Op: ast.BinaryOp(3),
																							Right: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(460),
																											Column: int(46),
																										},
																										End: ast.Location{
																											Line: int(460),
																											Column: int(47),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p5111,
																									FreeVars: ast.Identifiers{
																										"c",
																									},
																								},
																								Id: "c",
																							},
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			TrailingComma: false,
																			TailStrict: true,
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: ast.Fodder{},
																		},
																	},
																},
															},
														},
														Fun: nil,
														CloseFodder: nil,
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
													},
												},
												Body: &ast.Local{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(462),
																Column: int(5),
															},
															End: ast.Location{
																Line: int(774),
																Column: int(48),
															},
															File: p1,
														},
														Fodder: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(1),
																Indent: int(4),
																Comment: []string{},
															},
														},
														Ctx: p3497,
														FreeVars: ast.Identifiers{
															"parse_codes",
															"std",
															"str",
															"vals",
														},
													},
													Binds: ast.LocalBinds{
														ast.LocalBind{
															VarFodder: ast.Fodder{},
															Variable: "codes",
															EqFodder: ast.Fodder{},
															Body: &ast.Apply{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(462),
																			Column: int(19),
																		},
																		End: ast.Location{
																			Line: int(462),
																			Column: int(46),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p5131,
																	FreeVars: ast.Identifiers{
																		"parse_codes",
																		"str",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(462),
																				Column: int(19),
																			},
																			End: ast.Location{
																				Line: int(462),
																				Column: int(30),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p5131,
																		FreeVars: ast.Identifiers{
																			"parse_codes",
																		},
																	},
																	Id: "parse_codes",
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(462),
																							Column: int(31),
																						},
																						End: ast.Location{
																							Line: int(462),
																							Column: int(34),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p5137,
																					FreeVars: ast.Identifiers{
																						"str",
																					},
																				},
																				Id: "str",
																			},
																			CommaFodder: ast.Fodder{},
																		},
																		ast.CommaSeparatedExpr{
																			Expr: &ast.LiteralNumber{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(462),
																							Column: int(36),
																						},
																						End: ast.Location{
																							Line: int(462),
																							Column: int(37),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p5137,
																					FreeVars: nil,
																				},
																				OriginalString: "0",
																			},
																			CommaFodder: ast.Fodder{},
																		},
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Array{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(462),
																							Column: int(39),
																						},
																						End: ast.Location{
																							Line: int(462),
																							Column: int(41),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p5137,
																					FreeVars: nil,
																				},
																				Elements: nil,
																				TrailingComma: false,
																				CloseFodder: ast.Fodder{},
																			},
																			CommaFodder: ast.Fodder{},
																		},
																		ast.CommaSeparatedExpr{
																			Expr: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(462),
																							Column: int(43),
																						},
																						End: ast.Location{
																							Line: int(462),
																							Column: int(45),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p5137,
																					FreeVars: nil,
																				},
																				Value: "",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																TrailingComma: false,
																TailStrict: false,
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
															},
															Fun: nil,
															CloseFodder: ast.Fodder{},
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(462),
																	Column: int(11),
																},
																End: ast.Location{
																	Line: int(462),
																	Column: int(46),
																},
																File: p1,
															},
														},
													},
													Body: &ast.Local{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(470),
																	Column: int(5),
																},
																End: ast.Location{
																	Line: int(774),
																	Column: int(48),
																},
																File: p1,
															},
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(2),
																	Indent: int(4),
																	Comment: []string{},
																},
																ast.FodderElement{
																	Kind: ast.FodderKind(2),
																	Blanks: int(0),
																	Indent: int(4),
																	Comment: []string{
																		"///////////////////////",
																	},
																},
																ast.FodderElement{
																	Kind: ast.FodderKind(2),
																	Blanks: int(0),
																	Indent: int(4),
																	Comment: []string{
																		"// Format the values //",
																	},
																},
																ast.FodderElement{
																	Kind: ast.FodderKind(2),
																	Blanks: int(1),
																	Indent: int(4),
																	Comment: []string{
																		"///////////////////////",
																	},
																},
																ast.FodderElement{
																	Kind: ast.FodderKind(2),
																	Blanks: int(0),
																	Indent: int(4),
																	Comment: []string{
																		"// Useful utilities",
																	},
																},
															},
															Ctx: p3497,
															FreeVars: ast.Identifiers{
																"codes",
																"std",
																"vals",
															},
														},
														Binds: ast.LocalBinds{
															ast.LocalBind{
																VarFodder: nil,
																Variable: "padding",
																EqFodder: nil,
																Body: &ast.Function{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(470),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(476),
																				Column: int(17),
																			},
																			File: p1,
																		},
																		Fodder: nil,
																		Ctx: p5151,
																		FreeVars: nil,
																	},
																	ParenLeftFodder: ast.Fodder{},
																	Parameters: []ast.Parameter{
																		ast.Parameter{
																			NameFodder: ast.Fodder{},
																			Name: "w",
																			EqFodder: nil,
																			DefaultArg: nil,
																			CommaFodder: ast.Fodder{},
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(470),
																					Column: int(19),
																				},
																				End: ast.Location{
																					Line: int(470),
																					Column: int(20),
																				},
																				File: p1,
																			},
																		},
																		ast.Parameter{
																			NameFodder: ast.Fodder{},
																			Name: "s",
																			EqFodder: nil,
																			DefaultArg: nil,
																			CommaFodder: nil,
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(470),
																					Column: int(22),
																				},
																				End: ast.Location{
																					Line: int(470),
																					Column: int(23),
																				},
																				File: p1,
																			},
																		},
																	},
																	TrailingComma: false,
																	ParenRightFodder: ast.Fodder{},
																	Body: &ast.Local{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(471),
																					Column: int(7),
																				},
																				End: ast.Location{
																					Line: int(476),
																					Column: int(17),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(6),
																					Comment: []string{},
																				},
																			},
																			Ctx: p5155,
																			FreeVars: ast.Identifiers{
																				"s",
																				"w",
																			},
																		},
																		Binds: ast.LocalBinds{
																			ast.LocalBind{
																				VarFodder: nil,
																				Variable: "aux",
																				EqFodder: nil,
																				Body: &ast.Function{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(471),
																								Column: int(13),
																							},
																							End: ast.Location{
																								Line: int(475),
																								Column: int(28),
																							},
																							File: p1,
																						},
																						Fodder: nil,
																						Ctx: p5159,
																						FreeVars: ast.Identifiers{
																							"aux",
																							"s",
																						},
																					},
																					ParenLeftFodder: ast.Fodder{},
																					Parameters: []ast.Parameter{
																						ast.Parameter{
																							NameFodder: ast.Fodder{},
																							Name: "w",
																							EqFodder: nil,
																							DefaultArg: nil,
																							CommaFodder: ast.Fodder{},
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(471),
																									Column: int(17),
																								},
																								End: ast.Location{
																									Line: int(471),
																									Column: int(18),
																								},
																								File: p1,
																							},
																						},
																						ast.Parameter{
																							NameFodder: ast.Fodder{},
																							Name: "v",
																							EqFodder: nil,
																							DefaultArg: nil,
																							CommaFodder: nil,
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(471),
																									Column: int(20),
																								},
																								End: ast.Location{
																									Line: int(471),
																									Column: int(21),
																								},
																								File: p1,
																							},
																						},
																					},
																					TrailingComma: false,
																					ParenRightFodder: ast.Fodder{},
																					Body: &ast.Conditional{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(472),
																									Column: int(9),
																								},
																								End: ast.Location{
																									Line: int(475),
																									Column: int(28),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{
																								ast.FodderElement{
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(8),
																									Comment: []string{},
																								},
																							},
																							Ctx: p5164,
																							FreeVars: ast.Identifiers{
																								"aux",
																								"s",
																								"v",
																								"w",
																							},
																						},
																						Cond: &ast.Binary{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(472),
																										Column: int(12),
																									},
																									End: ast.Location{
																										Line: int(472),
																										Column: int(18),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p5164,
																								FreeVars: ast.Identifiers{
																									"w",
																								},
																							},
																							Left: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(472),
																											Column: int(12),
																										},
																										End: ast.Location{
																											Line: int(472),
																											Column: int(13),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p5164,
																									FreeVars: ast.Identifiers{
																										"w",
																									},
																								},
																								Id: "w",
																							},
																							OpFodder: ast.Fodder{},
																							Op: ast.BinaryOp(10),
																							Right: &ast.LiteralNumber{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(472),
																											Column: int(17),
																										},
																										End: ast.Location{
																											Line: int(472),
																											Column: int(18),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p5164,
																									FreeVars: nil,
																								},
																								OriginalString: "0",
																							},
																						},
																						ThenFodder: ast.Fodder{},
																						BranchTrue: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(473),
																										Column: int(11),
																									},
																									End: ast.Location{
																										Line: int(473),
																										Column: int(12),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{
																									ast.FodderElement{
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(10),
																										Comment: []string{},
																									},
																								},
																								Ctx: p5164,
																								FreeVars: ast.Identifiers{
																									"v",
																								},
																							},
																							Id: "v",
																						},
																						ElseFodder: ast.Fodder{
																							ast.FodderElement{
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(8),
																								Comment: []string{},
																							},
																						},
																						BranchFalse: &ast.Apply{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(475),
																										Column: int(11),
																									},
																									End: ast.Location{
																										Line: int(475),
																										Column: int(28),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p5164,
																								FreeVars: ast.Identifiers{
																									"aux",
																									"s",
																									"v",
																									"w",
																								},
																							},
																							Target: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(475),
																											Column: int(11),
																										},
																										End: ast.Location{
																											Line: int(475),
																											Column: int(14),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{
																										ast.FodderElement{
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(10),
																											Comment: []string{},
																										},
																									},
																									Ctx: p5164,
																									FreeVars: ast.Identifiers{
																										"aux",
																									},
																								},
																								Id: "aux",
																							},
																							FodderLeft: ast.Fodder{},
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Binary{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(475),
																														Column: int(15),
																													},
																													End: ast.Location{
																														Line: int(475),
																														Column: int(20),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p5182,
																												FreeVars: ast.Identifiers{
																													"w",
																												},
																											},
																											Left: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(475),
																															Column: int(15),
																														},
																														End: ast.Location{
																															Line: int(475),
																															Column: int(16),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p5182,
																													FreeVars: ast.Identifiers{
																														"w",
																													},
																												},
																												Id: "w",
																											},
																											OpFodder: ast.Fodder{},
																											Op: ast.BinaryOp(4),
																											Right: &ast.LiteralNumber{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(475),
																															Column: int(19),
																														},
																														End: ast.Location{
																															Line: int(475),
																															Column: int(20),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p5182,
																													FreeVars: nil,
																												},
																												OriginalString: "1",
																											},
																										},
																										CommaFodder: ast.Fodder{},
																									},
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Binary{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(475),
																														Column: int(22),
																													},
																													End: ast.Location{
																														Line: int(475),
																														Column: int(27),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p5182,
																												FreeVars: ast.Identifiers{
																													"s",
																													"v",
																												},
																											},
																											Left: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(475),
																															Column: int(22),
																														},
																														End: ast.Location{
																															Line: int(475),
																															Column: int(23),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p5182,
																													FreeVars: ast.Identifiers{
																														"v",
																													},
																												},
																												Id: "v",
																											},
																											OpFodder: ast.Fodder{},
																											Op: ast.BinaryOp(3),
																											Right: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(475),
																															Column: int(26),
																														},
																														End: ast.Location{
																															Line: int(475),
																															Column: int(27),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p5182,
																													FreeVars: ast.Identifiers{
																														"s",
																													},
																												},
																												Id: "s",
																											},
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							TrailingComma: false,
																							TailStrict: false,
																							FodderRight: ast.Fodder{},
																							TailStrictFodder: nil,
																						},
																					},
																				},
																				Fun: nil,
																				CloseFodder: nil,
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																			},
																		},
																		Body: &ast.Apply{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(476),
																						Column: int(7),
																					},
																					End: ast.Location{
																						Line: int(476),
																						Column: int(17),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p5155,
																				FreeVars: ast.Identifiers{
																					"aux",
																					"w",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(476),
																							Column: int(7),
																						},
																						End: ast.Location{
																							Line: int(476),
																							Column: int(10),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(6),
																							Comment: []string{},
																						},
																					},
																					Ctx: p5155,
																					FreeVars: ast.Identifiers{
																						"aux",
																					},
																				},
																				Id: "aux",
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(476),
																										Column: int(11),
																									},
																									End: ast.Location{
																										Line: int(476),
																										Column: int(12),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p5200,
																								FreeVars: ast.Identifiers{
																									"w",
																								},
																							},
																							Id: "w",
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.LiteralString{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(476),
																										Column: int(14),
																									},
																									End: ast.Location{
																										Line: int(476),
																										Column: int(16),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p5200,
																								FreeVars: nil,
																							},
																							Value: "",
																							Kind: ast.LiteralStringKind(1),
																							BlockIndent: "",
																							BlockTermIndent: "",
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			TrailingComma: false,
																			TailStrict: false,
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																		},
																	},
																},
																Fun: nil,
																CloseFodder: nil,
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
															},
														},
														Body: &ast.Local{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(479),
																		Column: int(5),
																	},
																	End: ast.Location{
																		Line: int(774),
																		Column: int(48),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(1),
																		Indent: int(4),
																		Comment: []string{},
																	},
																	ast.FodderElement{
																		Kind: ast.FodderKind(2),
																		Blanks: int(0),
																		Indent: int(4),
																		Comment: []string{
																			"// Add s to the left of str so that its length is at least w.",
																		},
																	},
																},
																Ctx: p3497,
																FreeVars: ast.Identifiers{
																	"codes",
																	"padding",
																	"std",
																	"vals",
																},
															},
															Binds: ast.LocalBinds{
																ast.LocalBind{
																	VarFodder: nil,
																	Variable: "pad_left",
																	EqFodder: nil,
																	Body: &ast.Function{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(479),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(480),
																					Column: int(44),
																				},
																				File: p1,
																			},
																			Fodder: nil,
																			Ctx: p5209,
																			FreeVars: ast.Identifiers{
																				"padding",
																				"std",
																			},
																		},
																		ParenLeftFodder: ast.Fodder{},
																		Parameters: []ast.Parameter{
																			ast.Parameter{
																				NameFodder: ast.Fodder{},
																				Name: "str",
																				EqFodder: nil,
																				DefaultArg: nil,
																				CommaFodder: ast.Fodder{},
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(479),
																						Column: int(20),
																					},
																					End: ast.Location{
																						Line: int(479),
																						Column: int(23),
																					},
																					File: p1,
																				},
																			},
																			ast.Parameter{
																				NameFodder: ast.Fodder{},
																				Name: "w",
																				EqFodder: nil,
																				DefaultArg: nil,
																				CommaFodder: ast.Fodder{},
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(479),
																						Column: int(25),
																					},
																					End: ast.Location{
																						Line: int(479),
																						Column: int(26),
																					},
																					File: p1,
																				},
																			},
																			ast.Parameter{
																				NameFodder: ast.Fodder{},
																				Name: "s",
																				EqFodder: nil,
																				DefaultArg: nil,
																				CommaFodder: nil,
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(479),
																						Column: int(28),
																					},
																					End: ast.Location{
																						Line: int(479),
																						Column: int(29),
																					},
																					File: p1,
																				},
																			},
																		},
																		TrailingComma: false,
																		ParenRightFodder: ast.Fodder{},
																		Body: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(480),
																						Column: int(7),
																					},
																					End: ast.Location{
																						Line: int(480),
																						Column: int(44),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p5213,
																				FreeVars: ast.Identifiers{
																					"padding",
																					"s",
																					"std",
																					"str",
																					"w",
																				},
																			},
																			Left: &ast.Apply{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(480),
																							Column: int(7),
																						},
																						End: ast.Location{
																							Line: int(480),
																							Column: int(38),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p5213,
																					FreeVars: ast.Identifiers{
																						"padding",
																						"s",
																						"std",
																						"str",
																						"w",
																					},
																				},
																				Target: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(480),
																								Column: int(7),
																							},
																							End: ast.Location{
																								Line: int(480),
																								Column: int(14),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(6),
																								Comment: []string{},
																							},
																						},
																						Ctx: p5213,
																						FreeVars: ast.Identifiers{
																							"padding",
																						},
																					},
																					Id: "padding",
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Binary{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(480),
																											Column: int(15),
																										},
																										End: ast.Location{
																											Line: int(480),
																											Column: int(34),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p5222,
																									FreeVars: ast.Identifiers{
																										"std",
																										"str",
																										"w",
																									},
																								},
																								Left: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(480),
																												Column: int(15),
																											},
																											End: ast.Location{
																												Line: int(480),
																												Column: int(16),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p5222,
																										FreeVars: ast.Identifiers{
																											"w",
																										},
																									},
																									Id: "w",
																								},
																								OpFodder: ast.Fodder{},
																								Op: ast.BinaryOp(4),
																								Right: &ast.Apply{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(480),
																												Column: int(19),
																											},
																											End: ast.Location{
																												Line: int(480),
																												Column: int(34),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p5222,
																										FreeVars: ast.Identifiers{
																											"std",
																											"str",
																										},
																									},
																									Target: &ast.Index{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(480),
																													Column: int(19),
																												},
																												End: ast.Location{
																													Line: int(480),
																													Column: int(29),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p5222,
																											FreeVars: ast.Identifiers{
																												"std",
																											},
																										},
																										Target: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(480),
																														Column: int(19),
																													},
																													End: ast.Location{
																														Line: int(480),
																														Column: int(22),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																											},
																											Id: "std",
																										},
																										LeftBracketFodder: ast.Fodder{},
																										Index: &ast.LiteralString{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													File: nil,
																												},
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: nil,
																											},
																											Value: "length",
																											Kind: ast.LiteralStringKind(1),
																											BlockIndent: "",
																											BlockTermIndent: "",
																										},
																										RightBracketFodder: ast.Fodder{},
																										Id: nil,
																									},
																									FodderLeft: ast.Fodder{},
																									Arguments: ast.Arguments{
																										Positional: []ast.CommaSeparatedExpr{
																											ast.CommaSeparatedExpr{
																												Expr: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(480),
																																Column: int(30),
																															},
																															End: ast.Location{
																																Line: int(480),
																																Column: int(33),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5235,
																														FreeVars: ast.Identifiers{
																															"str",
																														},
																													},
																													Id: "str",
																												},
																												CommaFodder: nil,
																											},
																										},
																										Named: nil,
																									},
																									TrailingComma: false,
																									TailStrict: false,
																									FodderRight: ast.Fodder{},
																									TailStrictFodder: nil,
																								},
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(480),
																											Column: int(36),
																										},
																										End: ast.Location{
																											Line: int(480),
																											Column: int(37),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p5222,
																									FreeVars: ast.Identifiers{
																										"s",
																									},
																								},
																								Id: "s",
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				TrailingComma: false,
																				TailStrict: false,
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: nil,
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(3),
																			Right: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(480),
																							Column: int(41),
																						},
																						End: ast.Location{
																							Line: int(480),
																							Column: int(44),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p5213,
																					FreeVars: ast.Identifiers{
																						"str",
																					},
																				},
																				Id: "str",
																			},
																		},
																	},
																	Fun: nil,
																	CloseFodder: nil,
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																},
															},
															Body: &ast.Local{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(483),
																			Column: int(5),
																		},
																		End: ast.Location{
																			Line: int(774),
																			Column: int(48),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Kind: ast.FodderKind(0),
																			Blanks: int(1),
																			Indent: int(4),
																			Comment: []string{},
																		},
																		ast.FodderElement{
																			Kind: ast.FodderKind(2),
																			Blanks: int(0),
																			Indent: int(4),
																			Comment: []string{
																				"// Add s to the right of str so that its length is at least w.",
																			},
																		},
																	},
																	Ctx: p3497,
																	FreeVars: ast.Identifiers{
																		"codes",
																		"pad_left",
																		"padding",
																		"std",
																		"vals",
																	},
																},
																Binds: ast.LocalBinds{
																	ast.LocalBind{
																		VarFodder: nil,
																		Variable: "pad_right",
																		EqFodder: nil,
																		Body: &ast.Function{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(483),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(484),
																						Column: int(44),
																					},
																					File: p1,
																				},
																				Fodder: nil,
																				Ctx: p5247,
																				FreeVars: ast.Identifiers{
																					"padding",
																					"std",
																				},
																			},
																			ParenLeftFodder: ast.Fodder{},
																			Parameters: []ast.Parameter{
																				ast.Parameter{
																					NameFodder: ast.Fodder{},
																					Name: "str",
																					EqFodder: nil,
																					DefaultArg: nil,
																					CommaFodder: ast.Fodder{},
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(483),
																							Column: int(21),
																						},
																						End: ast.Location{
																							Line: int(483),
																							Column: int(24),
																						},
																						File: p1,
																					},
																				},
																				ast.Parameter{
																					NameFodder: ast.Fodder{},
																					Name: "w",
																					EqFodder: nil,
																					DefaultArg: nil,
																					CommaFodder: ast.Fodder{},
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(483),
																							Column: int(26),
																						},
																						End: ast.Location{
																							Line: int(483),
																							Column: int(27),
																						},
																						File: p1,
																					},
																				},
																				ast.Parameter{
																					NameFodder: ast.Fodder{},
																					Name: "s",
																					EqFodder: nil,
																					DefaultArg: nil,
																					CommaFodder: nil,
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(483),
																							Column: int(29),
																						},
																						End: ast.Location{
																							Line: int(483),
																							Column: int(30),
																						},
																						File: p1,
																					},
																				},
																			},
																			TrailingComma: false,
																			ParenRightFodder: ast.Fodder{},
																			Body: &ast.Binary{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(484),
																							Column: int(7),
																						},
																						End: ast.Location{
																							Line: int(484),
																							Column: int(44),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p5251,
																					FreeVars: ast.Identifiers{
																						"padding",
																						"s",
																						"std",
																						"str",
																						"w",
																					},
																				},
																				Left: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(484),
																								Column: int(7),
																							},
																							End: ast.Location{
																								Line: int(484),
																								Column: int(10),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(6),
																								Comment: []string{},
																							},
																						},
																						Ctx: p5251,
																						FreeVars: ast.Identifiers{
																							"str",
																						},
																					},
																					Id: "str",
																				},
																				OpFodder: ast.Fodder{},
																				Op: ast.BinaryOp(3),
																				Right: &ast.Apply{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(484),
																								Column: int(13),
																							},
																							End: ast.Location{
																								Line: int(484),
																								Column: int(44),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p5251,
																						FreeVars: ast.Identifiers{
																							"padding",
																							"s",
																							"std",
																							"str",
																							"w",
																						},
																					},
																					Target: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(484),
																									Column: int(13),
																								},
																								End: ast.Location{
																									Line: int(484),
																									Column: int(20),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p5251,
																							FreeVars: ast.Identifiers{
																								"padding",
																							},
																						},
																						Id: "padding",
																					},
																					FodderLeft: ast.Fodder{},
																					Arguments: ast.Arguments{
																						Positional: []ast.CommaSeparatedExpr{
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Binary{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(484),
																												Column: int(21),
																											},
																											End: ast.Location{
																												Line: int(484),
																												Column: int(40),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p5262,
																										FreeVars: ast.Identifiers{
																											"std",
																											"str",
																											"w",
																										},
																									},
																									Left: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(484),
																													Column: int(21),
																												},
																												End: ast.Location{
																													Line: int(484),
																													Column: int(22),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p5262,
																											FreeVars: ast.Identifiers{
																												"w",
																											},
																										},
																										Id: "w",
																									},
																									OpFodder: ast.Fodder{},
																									Op: ast.BinaryOp(4),
																									Right: &ast.Apply{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(484),
																													Column: int(25),
																												},
																												End: ast.Location{
																													Line: int(484),
																													Column: int(40),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p5262,
																											FreeVars: ast.Identifiers{
																												"std",
																												"str",
																											},
																										},
																										Target: &ast.Index{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(484),
																														Column: int(25),
																													},
																													End: ast.Location{
																														Line: int(484),
																														Column: int(35),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p5262,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																											},
																											Target: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(484),
																															Column: int(25),
																														},
																														End: ast.Location{
																															Line: int(484),
																															Column: int(28),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																												},
																												Id: "std",
																											},
																											LeftBracketFodder: ast.Fodder{},
																											Index: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: nil,
																												},
																												Value: "length",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																											RightBracketFodder: ast.Fodder{},
																											Id: nil,
																										},
																										FodderLeft: ast.Fodder{},
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(484),
																																	Column: int(36),
																																},
																																End: ast.Location{
																																	Line: int(484),
																																	Column: int(39),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5275,
																															FreeVars: ast.Identifiers{
																																"str",
																															},
																														},
																														Id: "str",
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										TrailingComma: false,
																										TailStrict: false,
																										FodderRight: ast.Fodder{},
																										TailStrictFodder: nil,
																									},
																								},
																								CommaFodder: ast.Fodder{},
																							},
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(484),
																												Column: int(42),
																											},
																											End: ast.Location{
																												Line: int(484),
																												Column: int(43),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p5262,
																										FreeVars: ast.Identifiers{
																											"s",
																										},
																									},
																									Id: "s",
																								},
																								CommaFodder: nil,
																							},
																						},
																						Named: nil,
																					},
																					TrailingComma: false,
																					TailStrict: false,
																					FodderRight: ast.Fodder{},
																					TailStrictFodder: nil,
																				},
																			},
																		},
																		Fun: nil,
																		CloseFodder: nil,
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																	},
																},
																Body: &ast.Local{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(499),
																				Column: int(5),
																			},
																			End: ast.Location{
																				Line: int(774),
																				Column: int(48),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(1),
																				Indent: int(4),
																				Comment: []string{},
																			},
																			ast.FodderElement{
																				Kind: ast.FodderKind(2),
																				Blanks: int(0),
																				Indent: int(4),
																				Comment: []string{
																					"// Render a sign & magnitude integer (radix ranges from decimal to binary).",
																				},
																			},
																			ast.FodderElement{
																				Kind: ast.FodderKind(2),
																				Blanks: int(0),
																				Indent: int(4),
																				Comment: []string{
																					"// neg should be a boolean, and when true indicates that we should render a negative number.",
																				},
																			},
																			ast.FodderElement{
																				Kind: ast.FodderKind(2),
																				Blanks: int(0),
																				Indent: int(4),
																				Comment: []string{
																					"// mag must always be a whole number >= 0, it's the magnitude of the integer to render",
																				},
																			},
																			ast.FodderElement{
																				Kind: ast.FodderKind(2),
																				Blanks: int(0),
																				Indent: int(4),
																				Comment: []string{
																					"// min_chars must be a whole number >= 0",
																				},
																			},
																			ast.FodderElement{
																				Kind: ast.FodderKind(2),
																				Blanks: int(0),
																				Indent: int(4),
																				Comment: []string{
																					"//   It is the field width, i.e. std.length() of the result should be >= min_chars",
																				},
																			},
																			ast.FodderElement{
																				Kind: ast.FodderKind(2),
																				Blanks: int(0),
																				Indent: int(4),
																				Comment: []string{
																					"// min_digits must be a whole number >= 0. It's the number of zeroes to pad with.",
																				},
																			},
																			ast.FodderElement{
																				Kind: ast.FodderKind(2),
																				Blanks: int(0),
																				Indent: int(4),
																				Comment: []string{
																					"// blank must be a boolean, if true adds an additional ' ' in front of a positive number, so",
																				},
																			},
																			ast.FodderElement{
																				Kind: ast.FodderKind(2),
																				Blanks: int(0),
																				Indent: int(4),
																				Comment: []string{
																					"// that it is aligned with negative numbers with the same number of digits.",
																				},
																			},
																			ast.FodderElement{
																				Kind: ast.FodderKind(2),
																				Blanks: int(0),
																				Indent: int(4),
																				Comment: []string{
																					"// plus must be a boolean, if true adds a '+' in front of a postive number, so that it is",
																				},
																			},
																			ast.FodderElement{
																				Kind: ast.FodderKind(2),
																				Blanks: int(0),
																				Indent: int(4),
																				Comment: []string{
																					"// aligned with negative numbers with the same number of digits.  This takes precedence over",
																				},
																			},
																			ast.FodderElement{
																				Kind: ast.FodderKind(2),
																				Blanks: int(0),
																				Indent: int(4),
																				Comment: []string{
																					"// blank, if both are true.",
																				},
																			},
																			ast.FodderElement{
																				Kind: ast.FodderKind(2),
																				Blanks: int(0),
																				Indent: int(4),
																				Comment: []string{
																					"// radix must be a whole number >1 and <= 10.  It is the base of the system of numerals.",
																				},
																			},
																			ast.FodderElement{
																				Kind: ast.FodderKind(2),
																				Blanks: int(0),
																				Indent: int(4),
																				Comment: []string{
																					"// zero_prefix is a string prefixed before the sign to all numbers that are not 0.",
																				},
																			},
																		},
																		Ctx: p3497,
																		FreeVars: ast.Identifiers{
																			"codes",
																			"pad_left",
																			"pad_right",
																			"std",
																			"vals",
																		},
																	},
																	Binds: ast.LocalBinds{
																		ast.LocalBind{
																			VarFodder: nil,
																			Variable: "render_int",
																			EqFodder: nil,
																			Body: &ast.Function{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(499),
																							Column: int(11),
																						},
																						End: ast.Location{
																							Line: int(514),
																							Column: int(84),
																						},
																						File: p1,
																					},
																					Fodder: nil,
																					Ctx: p5297,
																					FreeVars: ast.Identifiers{
																						"pad_left",
																						"std",
																					},
																				},
																				ParenLeftFodder: ast.Fodder{},
																				Parameters: []ast.Parameter{
																					ast.Parameter{
																						NameFodder: ast.Fodder{},
																						Name: "neg",
																						EqFodder: nil,
																						DefaultArg: nil,
																						CommaFodder: ast.Fodder{},
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(499),
																								Column: int(22),
																							},
																							End: ast.Location{
																								Line: int(499),
																								Column: int(25),
																							},
																							File: p1,
																						},
																					},
																					ast.Parameter{
																						NameFodder: ast.Fodder{},
																						Name: "mag",
																						EqFodder: nil,
																						DefaultArg: nil,
																						CommaFodder: ast.Fodder{},
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(499),
																								Column: int(27),
																							},
																							End: ast.Location{
																								Line: int(499),
																								Column: int(30),
																							},
																							File: p1,
																						},
																					},
																					ast.Parameter{
																						NameFodder: ast.Fodder{},
																						Name: "min_chars",
																						EqFodder: nil,
																						DefaultArg: nil,
																						CommaFodder: ast.Fodder{},
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(499),
																								Column: int(32),
																							},
																							End: ast.Location{
																								Line: int(499),
																								Column: int(41),
																							},
																							File: p1,
																						},
																					},
																					ast.Parameter{
																						NameFodder: ast.Fodder{},
																						Name: "min_digits",
																						EqFodder: nil,
																						DefaultArg: nil,
																						CommaFodder: ast.Fodder{},
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(499),
																								Column: int(43),
																							},
																							End: ast.Location{
																								Line: int(499),
																								Column: int(53),
																							},
																							File: p1,
																						},
																					},
																					ast.Parameter{
																						NameFodder: ast.Fodder{},
																						Name: "blank",
																						EqFodder: nil,
																						DefaultArg: nil,
																						CommaFodder: ast.Fodder{},
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(499),
																								Column: int(55),
																							},
																							End: ast.Location{
																								Line: int(499),
																								Column: int(60),
																							},
																							File: p1,
																						},
																					},
																					ast.Parameter{
																						NameFodder: ast.Fodder{},
																						Name: "plus",
																						EqFodder: nil,
																						DefaultArg: nil,
																						CommaFodder: ast.Fodder{},
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(499),
																								Column: int(62),
																							},
																							End: ast.Location{
																								Line: int(499),
																								Column: int(66),
																							},
																							File: p1,
																						},
																					},
																					ast.Parameter{
																						NameFodder: ast.Fodder{},
																						Name: "radix",
																						EqFodder: nil,
																						DefaultArg: nil,
																						CommaFodder: ast.Fodder{},
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(499),
																								Column: int(68),
																							},
																							End: ast.Location{
																								Line: int(499),
																								Column: int(73),
																							},
																							File: p1,
																						},
																					},
																					ast.Parameter{
																						NameFodder: ast.Fodder{},
																						Name: "zero_prefix",
																						EqFodder: nil,
																						DefaultArg: nil,
																						CommaFodder: nil,
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(499),
																								Column: int(75),
																							},
																							End: ast.Location{
																								Line: int(499),
																								Column: int(86),
																							},
																							File: p1,
																						},
																					},
																				},
																				TrailingComma: false,
																				ParenRightFodder: ast.Fodder{},
																				Body: &ast.Local{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(501),
																								Column: int(7),
																							},
																							End: ast.Location{
																								Line: int(514),
																								Column: int(84),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(6),
																								Comment: []string{},
																							},
																							ast.FodderElement{
																								Kind: ast.FodderKind(2),
																								Blanks: int(0),
																								Indent: int(6),
																								Comment: []string{
																									"// dec is the minimal string needed to represent the number as text.",
																								},
																							},
																						},
																						Ctx: p5303,
																						FreeVars: ast.Identifiers{
																							"blank",
																							"mag",
																							"min_chars",
																							"min_digits",
																							"neg",
																							"pad_left",
																							"plus",
																							"radix",
																							"std",
																							"zero_prefix",
																						},
																					},
																					Binds: ast.LocalBinds{
																						ast.LocalBind{
																							VarFodder: ast.Fodder{},
																							Variable: "dec",
																							EqFodder: ast.Fodder{},
																							Body: &ast.Conditional{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(502),
																											Column: int(9),
																										},
																										End: ast.Location{
																											Line: int(510),
																											Column: int(19),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{
																										ast.FodderElement{
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(8),
																											Comment: []string{},
																										},
																									},
																									Ctx: p5308,
																									FreeVars: ast.Identifiers{
																										"mag",
																										"radix",
																										"std",
																										"zero_prefix",
																									},
																								},
																								Cond: &ast.Binary{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(502),
																												Column: int(12),
																											},
																											End: ast.Location{
																												Line: int(502),
																												Column: int(20),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p5308,
																										FreeVars: ast.Identifiers{
																											"mag",
																										},
																									},
																									Left: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(502),
																													Column: int(12),
																												},
																												End: ast.Location{
																													Line: int(502),
																													Column: int(15),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p5308,
																											FreeVars: ast.Identifiers{
																												"mag",
																											},
																										},
																										Id: "mag",
																									},
																									OpFodder: ast.Fodder{},
																									Op: ast.BinaryOp(12),
																									Right: &ast.LiteralNumber{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(502),
																													Column: int(19),
																												},
																												End: ast.Location{
																													Line: int(502),
																													Column: int(20),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p5308,
																											FreeVars: nil,
																										},
																										OriginalString: "0",
																									},
																								},
																								ThenFodder: ast.Fodder{},
																								BranchTrue: &ast.LiteralString{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(503),
																												Column: int(11),
																											},
																											End: ast.Location{
																												Line: int(503),
																												Column: int(14),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{
																											ast.FodderElement{
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(10),
																												Comment: []string{},
																											},
																										},
																										Ctx: p5308,
																										FreeVars: nil,
																									},
																									Value: "0",
																									Kind: ast.LiteralStringKind(1),
																									BlockIndent: "",
																									BlockTermIndent: "",
																								},
																								ElseFodder: ast.Fodder{
																									ast.FodderElement{
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(8),
																										Comment: []string{},
																									},
																								},
																								BranchFalse: &ast.Local{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(505),
																												Column: int(11),
																											},
																											End: ast.Location{
																												Line: int(510),
																												Column: int(19),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{
																											ast.FodderElement{
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(10),
																												Comment: []string{},
																											},
																										},
																										Ctx: p5308,
																										FreeVars: ast.Identifiers{
																											"mag",
																											"radix",
																											"std",
																											"zero_prefix",
																										},
																									},
																									Binds: ast.LocalBinds{
																										ast.LocalBind{
																											VarFodder: nil,
																											Variable: "aux",
																											EqFodder: nil,
																											Body: &ast.Function{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(505),
																															Column: int(17),
																														},
																														End: ast.Location{
																															Line: int(509),
																															Column: int(54),
																														},
																														File: p1,
																													},
																													Fodder: nil,
																													Ctx: p5323,
																													FreeVars: ast.Identifiers{
																														"aux",
																														"radix",
																														"std",
																														"zero_prefix",
																													},
																												},
																												ParenLeftFodder: ast.Fodder{},
																												Parameters: []ast.Parameter{
																													ast.Parameter{
																														NameFodder: ast.Fodder{},
																														Name: "n",
																														EqFodder: nil,
																														DefaultArg: nil,
																														CommaFodder: nil,
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(505),
																																Column: int(21),
																															},
																															End: ast.Location{
																																Line: int(505),
																																Column: int(22),
																															},
																															File: p1,
																														},
																													},
																												},
																												TrailingComma: false,
																												ParenRightFodder: ast.Fodder{},
																												Body: &ast.Conditional{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(506),
																																Column: int(13),
																															},
																															End: ast.Location{
																																Line: int(509),
																																Column: int(54),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(12),
																																Comment: []string{},
																															},
																														},
																														Ctx: p5328,
																														FreeVars: ast.Identifiers{
																															"aux",
																															"n",
																															"radix",
																															"std",
																															"zero_prefix",
																														},
																													},
																													Cond: &ast.Binary{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(506),
																																	Column: int(16),
																																},
																																End: ast.Location{
																																	Line: int(506),
																																	Column: int(22),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5328,
																															FreeVars: ast.Identifiers{
																																"n",
																															},
																														},
																														Left: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(506),
																																		Column: int(16),
																																	},
																																	End: ast.Location{
																																		Line: int(506),
																																		Column: int(17),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p5328,
																																FreeVars: ast.Identifiers{
																																	"n",
																																},
																															},
																															Id: "n",
																														},
																														OpFodder: ast.Fodder{},
																														Op: ast.BinaryOp(12),
																														Right: &ast.LiteralNumber{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(506),
																																		Column: int(21),
																																	},
																																	End: ast.Location{
																																		Line: int(506),
																																		Column: int(22),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p5328,
																																FreeVars: nil,
																															},
																															OriginalString: "0",
																														},
																													},
																													ThenFodder: ast.Fodder{},
																													BranchTrue: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(507),
																																	Column: int(15),
																																},
																																End: ast.Location{
																																	Line: int(507),
																																	Column: int(26),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{
																																ast.FodderElement{
																																	Kind: ast.FodderKind(0),
																																	Blanks: int(0),
																																	Indent: int(14),
																																	Comment: []string{},
																																},
																															},
																															Ctx: p5328,
																															FreeVars: ast.Identifiers{
																																"zero_prefix",
																															},
																														},
																														Id: "zero_prefix",
																													},
																													ElseFodder: ast.Fodder{
																														ast.FodderElement{
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(12),
																															Comment: []string{},
																														},
																													},
																													BranchFalse: &ast.Binary{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(509),
																																	Column: int(15),
																																},
																																End: ast.Location{
																																	Line: int(509),
																																	Column: int(54),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5328,
																															FreeVars: ast.Identifiers{
																																"aux",
																																"n",
																																"radix",
																																"std",
																															},
																														},
																														Left: &ast.Apply{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(509),
																																		Column: int(15),
																																	},
																																	End: ast.Location{
																																		Line: int(509),
																																		Column: int(40),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p5328,
																																FreeVars: ast.Identifiers{
																																	"aux",
																																	"n",
																																	"radix",
																																	"std",
																																},
																															},
																															Target: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(509),
																																			Column: int(15),
																																		},
																																		End: ast.Location{
																																			Line: int(509),
																																			Column: int(18),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{
																																		ast.FodderElement{
																																			Kind: ast.FodderKind(0),
																																			Blanks: int(0),
																																			Indent: int(14),
																																			Comment: []string{},
																																		},
																																	},
																																	Ctx: p5328,
																																	FreeVars: ast.Identifiers{
																																		"aux",
																																	},
																																},
																																Id: "aux",
																															},
																															FodderLeft: ast.Fodder{},
																															Arguments: ast.Arguments{
																																Positional: []ast.CommaSeparatedExpr{
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Apply{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(509),
																																						Column: int(19),
																																					},
																																					End: ast.Location{
																																						Line: int(509),
																																						Column: int(39),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p5348,
																																				FreeVars: ast.Identifiers{
																																					"n",
																																					"radix",
																																					"std",
																																				},
																																			},
																																			Target: &ast.Index{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(509),
																																							Column: int(19),
																																						},
																																						End: ast.Location{
																																							Line: int(509),
																																							Column: int(28),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p5348,
																																					FreeVars: ast.Identifiers{
																																						"std",
																																					},
																																				},
																																				Target: &ast.Var{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(509),
																																								Column: int(19),
																																							},
																																							End: ast.Location{
																																								Line: int(509),
																																								Column: int(22),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{
																																							"std",
																																						},
																																					},
																																					Id: "std",
																																				},
																																				LeftBracketFodder: ast.Fodder{},
																																				Index: &ast.LiteralString{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							End: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							File: nil,
																																						},
																																						Fodder: nil,
																																						Ctx: nil,
																																						FreeVars: nil,
																																					},
																																					Value: "floor",
																																					Kind: ast.LiteralStringKind(1),
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																				},
																																				RightBracketFodder: ast.Fodder{},
																																				Id: nil,
																																			},
																																			FodderLeft: ast.Fodder{},
																																			Arguments: ast.Arguments{
																																				Positional: []ast.CommaSeparatedExpr{
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Binary{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(509),
																																										Column: int(29),
																																									},
																																									End: ast.Location{
																																										Line: int(509),
																																										Column: int(38),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p5357,
																																								FreeVars: ast.Identifiers{
																																									"n",
																																									"radix",
																																								},
																																							},
																																							Left: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(509),
																																											Column: int(29),
																																										},
																																										End: ast.Location{
																																											Line: int(509),
																																											Column: int(30),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p5357,
																																									FreeVars: ast.Identifiers{
																																										"n",
																																									},
																																								},
																																								Id: "n",
																																							},
																																							OpFodder: ast.Fodder{},
																																							Op: ast.BinaryOp(1),
																																							Right: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(509),
																																											Column: int(33),
																																										},
																																										End: ast.Location{
																																											Line: int(509),
																																											Column: int(38),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p5357,
																																									FreeVars: ast.Identifiers{
																																										"radix",
																																									},
																																								},
																																								Id: "radix",
																																							},
																																						},
																																						CommaFodder: nil,
																																					},
																																				},
																																				Named: nil,
																																			},
																																			TrailingComma: false,
																																			TailStrict: false,
																																			FodderRight: ast.Fodder{},
																																			TailStrictFodder: nil,
																																		},
																																		CommaFodder: nil,
																																	},
																																},
																																Named: nil,
																															},
																															TrailingComma: false,
																															TailStrict: false,
																															FodderRight: ast.Fodder{},
																															TailStrictFodder: nil,
																														},
																														OpFodder: ast.Fodder{},
																														Op: ast.BinaryOp(3),
																														Right: &ast.Apply{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	File: nil,
																																},
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"n",
																																	"radix",
																																	"std",
																																},
																															},
																															Target: &ast.Index{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		File: nil,
																																	},
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																	},
																																},
																																Target: &ast.Var{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			End: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			File: nil,
																																		},
																																		Fodder: nil,
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{
																																			"std",
																																		},
																																	},
																																	Id: "std",
																																},
																																LeftBracketFodder: nil,
																																Index: &ast.LiteralString{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			End: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			File: nil,
																																		},
																																		Fodder: nil,
																																		Ctx: nil,
																																		FreeVars: nil,
																																	},
																																	Value: "mod",
																																	Kind: ast.LiteralStringKind(1),
																																	BlockIndent: "",
																																	BlockTermIndent: "",
																																},
																																RightBracketFodder: nil,
																																Id: nil,
																															},
																															FodderLeft: nil,
																															Arguments: ast.Arguments{
																																Positional: []ast.CommaSeparatedExpr{
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(509),
																																						Column: int(44),
																																					},
																																					End: ast.Location{
																																						Line: int(509),
																																						Column: int(45),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p5328,
																																				FreeVars: ast.Identifiers{
																																					"n",
																																				},
																																			},
																																			Id: "n",
																																		},
																																		CommaFodder: nil,
																																	},
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(509),
																																						Column: int(48),
																																					},
																																					End: ast.Location{
																																						Line: int(509),
																																						Column: int(53),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p5328,
																																				FreeVars: ast.Identifiers{
																																					"radix",
																																				},
																																			},
																																			Id: "radix",
																																		},
																																		CommaFodder: nil,
																																	},
																																},
																																Named: nil,
																															},
																															TrailingComma: false,
																															TailStrict: false,
																															FodderRight: nil,
																															TailStrictFodder: nil,
																														},
																													},
																												},
																											},
																											Fun: nil,
																											CloseFodder: nil,
																											LocRange: ast.LocationRange{
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												File: nil,
																											},
																										},
																									},
																									Body: &ast.Apply{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(510),
																													Column: int(11),
																												},
																												End: ast.Location{
																													Line: int(510),
																													Column: int(19),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p5308,
																											FreeVars: ast.Identifiers{
																												"aux",
																												"mag",
																											},
																										},
																										Target: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(510),
																														Column: int(11),
																													},
																													End: ast.Location{
																														Line: int(510),
																														Column: int(14),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{
																													ast.FodderElement{
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(10),
																														Comment: []string{},
																													},
																												},
																												Ctx: p5308,
																												FreeVars: ast.Identifiers{
																													"aux",
																												},
																											},
																											Id: "aux",
																										},
																										FodderLeft: ast.Fodder{},
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(510),
																																	Column: int(15),
																																},
																																End: ast.Location{
																																	Line: int(510),
																																	Column: int(18),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5382,
																															FreeVars: ast.Identifiers{
																																"mag",
																															},
																														},
																														Id: "mag",
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										TrailingComma: false,
																										TailStrict: false,
																										FodderRight: ast.Fodder{},
																										TailStrictFodder: nil,
																									},
																								},
																							},
																							Fun: nil,
																							CloseFodder: ast.Fodder{},
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(501),
																									Column: int(13),
																								},
																								End: ast.Location{
																									Line: int(510),
																									Column: int(19),
																								},
																								File: p1,
																							},
																						},
																					},
																					Body: &ast.Local{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(511),
																									Column: int(7),
																								},
																								End: ast.Location{
																									Line: int(514),
																									Column: int(84),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{
																								ast.FodderElement{
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(6),
																									Comment: []string{},
																								},
																							},
																							Ctx: p5303,
																							FreeVars: ast.Identifiers{
																								"blank",
																								"dec",
																								"min_chars",
																								"min_digits",
																								"neg",
																								"pad_left",
																								"plus",
																								"std",
																							},
																						},
																						Binds: ast.LocalBinds{
																							ast.LocalBind{
																								VarFodder: ast.Fodder{},
																								Variable: "zp",
																								EqFodder: ast.Fodder{},
																								Body: &ast.Binary{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(511),
																												Column: int(18),
																											},
																											End: ast.Location{
																												Line: int(511),
																												Column: int(69),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p5389,
																										FreeVars: ast.Identifiers{
																											"blank",
																											"min_chars",
																											"neg",
																											"plus",
																										},
																									},
																									Left: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(511),
																													Column: int(18),
																												},
																												End: ast.Location{
																													Line: int(511),
																													Column: int(27),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p5389,
																											FreeVars: ast.Identifiers{
																												"min_chars",
																											},
																										},
																										Id: "min_chars",
																									},
																									OpFodder: ast.Fodder{},
																									Op: ast.BinaryOp(4),
																									Right: &ast.Conditional{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(511),
																													Column: int(31),
																												},
																												End: ast.Location{
																													Line: int(511),
																													Column: int(68),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p5389,
																											FreeVars: ast.Identifiers{
																												"blank",
																												"neg",
																												"plus",
																											},
																										},
																										Cond: &ast.Binary{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(511),
																														Column: int(34),
																													},
																													End: ast.Location{
																														Line: int(511),
																														Column: int(54),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p5389,
																												FreeVars: ast.Identifiers{
																													"blank",
																													"neg",
																													"plus",
																												},
																											},
																											Left: &ast.Binary{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(511),
																															Column: int(34),
																														},
																														End: ast.Location{
																															Line: int(511),
																															Column: int(46),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p5389,
																													FreeVars: ast.Identifiers{
																														"blank",
																														"neg",
																													},
																												},
																												Left: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(511),
																																Column: int(34),
																															},
																															End: ast.Location{
																																Line: int(511),
																																Column: int(37),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5389,
																														FreeVars: ast.Identifiers{
																															"neg",
																														},
																													},
																													Id: "neg",
																												},
																												OpFodder: ast.Fodder{},
																												Op: ast.BinaryOp(18),
																												Right: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(511),
																																Column: int(41),
																															},
																															End: ast.Location{
																																Line: int(511),
																																Column: int(46),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5389,
																														FreeVars: ast.Identifiers{
																															"blank",
																														},
																													},
																													Id: "blank",
																												},
																											},
																											OpFodder: ast.Fodder{},
																											Op: ast.BinaryOp(18),
																											Right: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(511),
																															Column: int(50),
																														},
																														End: ast.Location{
																															Line: int(511),
																															Column: int(54),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p5389,
																													FreeVars: ast.Identifiers{
																														"plus",
																													},
																												},
																												Id: "plus",
																											},
																										},
																										ThenFodder: ast.Fodder{},
																										BranchTrue: &ast.LiteralNumber{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(511),
																														Column: int(60),
																													},
																													End: ast.Location{
																														Line: int(511),
																														Column: int(61),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p5389,
																												FreeVars: nil,
																											},
																											OriginalString: "1",
																										},
																										ElseFodder: ast.Fodder{},
																										BranchFalse: &ast.LiteralNumber{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(511),
																														Column: int(67),
																													},
																													End: ast.Location{
																														Line: int(511),
																														Column: int(68),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p5389,
																												FreeVars: nil,
																											},
																											OriginalString: "0",
																										},
																									},
																								},
																								Fun: nil,
																								CloseFodder: ast.Fodder{},
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(511),
																										Column: int(13),
																									},
																									End: ast.Location{
																										Line: int(511),
																										Column: int(69),
																									},
																									File: p1,
																								},
																							},
																						},
																						Body: &ast.Local{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(512),
																										Column: int(7),
																									},
																									End: ast.Location{
																										Line: int(514),
																										Column: int(84),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{
																									ast.FodderElement{
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(6),
																										Comment: []string{},
																									},
																								},
																								Ctx: p5303,
																								FreeVars: ast.Identifiers{
																									"blank",
																									"dec",
																									"min_digits",
																									"neg",
																									"pad_left",
																									"plus",
																									"std",
																									"zp",
																								},
																							},
																							Binds: ast.LocalBinds{
																								ast.LocalBind{
																									VarFodder: ast.Fodder{},
																									Variable: "zp2",
																									EqFodder: ast.Fodder{},
																									Body: &ast.Apply{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(512),
																													Column: int(19),
																												},
																												End: ast.Location{
																													Line: int(512),
																													Column: int(42),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p5412,
																											FreeVars: ast.Identifiers{
																												"min_digits",
																												"std",
																												"zp",
																											},
																										},
																										Target: &ast.Index{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(512),
																														Column: int(19),
																													},
																													End: ast.Location{
																														Line: int(512),
																														Column: int(26),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p5412,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																											},
																											Target: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(512),
																															Column: int(19),
																														},
																														End: ast.Location{
																															Line: int(512),
																															Column: int(22),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																												},
																												Id: "std",
																											},
																											LeftBracketFodder: ast.Fodder{},
																											Index: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: nil,
																												},
																												Value: "max",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																											RightBracketFodder: ast.Fodder{},
																											Id: nil,
																										},
																										FodderLeft: ast.Fodder{},
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(512),
																																	Column: int(27),
																																},
																																End: ast.Location{
																																	Line: int(512),
																																	Column: int(29),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5421,
																															FreeVars: ast.Identifiers{
																																"zp",
																															},
																														},
																														Id: "zp",
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(512),
																																	Column: int(31),
																																},
																																End: ast.Location{
																																	Line: int(512),
																																	Column: int(41),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5421,
																															FreeVars: ast.Identifiers{
																																"min_digits",
																															},
																														},
																														Id: "min_digits",
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										TrailingComma: false,
																										TailStrict: false,
																										FodderRight: ast.Fodder{},
																										TailStrictFodder: nil,
																									},
																									Fun: nil,
																									CloseFodder: ast.Fodder{},
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(512),
																											Column: int(13),
																										},
																										End: ast.Location{
																											Line: int(512),
																											Column: int(42),
																										},
																										File: p1,
																									},
																								},
																							},
																							Body: &ast.Local{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(513),
																											Column: int(7),
																										},
																										End: ast.Location{
																											Line: int(514),
																											Column: int(84),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{
																										ast.FodderElement{
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(6),
																											Comment: []string{},
																										},
																									},
																									Ctx: p5303,
																									FreeVars: ast.Identifiers{
																										"blank",
																										"dec",
																										"neg",
																										"pad_left",
																										"plus",
																										"zp2",
																									},
																								},
																								Binds: ast.LocalBinds{
																									ast.LocalBind{
																										VarFodder: ast.Fodder{},
																										Variable: "dec2",
																										EqFodder: ast.Fodder{},
																										Body: &ast.Apply{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(513),
																														Column: int(20),
																													},
																													End: ast.Location{
																														Line: int(513),
																														Column: int(43),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p5430,
																												FreeVars: ast.Identifiers{
																													"dec",
																													"pad_left",
																													"zp2",
																												},
																											},
																											Target: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(513),
																															Column: int(20),
																														},
																														End: ast.Location{
																															Line: int(513),
																															Column: int(28),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p5430,
																													FreeVars: ast.Identifiers{
																														"pad_left",
																													},
																												},
																												Id: "pad_left",
																											},
																											FodderLeft: ast.Fodder{},
																											Arguments: ast.Arguments{
																												Positional: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(513),
																																		Column: int(29),
																																	},
																																	End: ast.Location{
																																		Line: int(513),
																																		Column: int(32),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p5436,
																																FreeVars: ast.Identifiers{
																																	"dec",
																																},
																															},
																															Id: "dec",
																														},
																														CommaFodder: ast.Fodder{},
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(513),
																																		Column: int(34),
																																	},
																																	End: ast.Location{
																																		Line: int(513),
																																		Column: int(37),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p5436,
																																FreeVars: ast.Identifiers{
																																	"zp2",
																																},
																															},
																															Id: "zp2",
																														},
																														CommaFodder: ast.Fodder{},
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.LiteralString{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(513),
																																		Column: int(39),
																																	},
																																	End: ast.Location{
																																		Line: int(513),
																																		Column: int(42),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p5436,
																																FreeVars: nil,
																															},
																															Value: "0",
																															Kind: ast.LiteralStringKind(1),
																															BlockIndent: "",
																															BlockTermIndent: "",
																														},
																														CommaFodder: nil,
																													},
																												},
																												Named: nil,
																											},
																											TrailingComma: false,
																											TailStrict: false,
																											FodderRight: ast.Fodder{},
																											TailStrictFodder: nil,
																										},
																										Fun: nil,
																										CloseFodder: ast.Fodder{},
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(513),
																												Column: int(13),
																											},
																											End: ast.Location{
																												Line: int(513),
																												Column: int(43),
																											},
																											File: p1,
																										},
																									},
																								},
																								Body: &ast.Binary{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(514),
																												Column: int(7),
																											},
																											End: ast.Location{
																												Line: int(514),
																												Column: int(84),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p5303,
																										FreeVars: ast.Identifiers{
																											"blank",
																											"dec2",
																											"neg",
																											"plus",
																										},
																									},
																									Left: &ast.Conditional{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(514),
																													Column: int(8),
																												},
																												End: ast.Location{
																													Line: int(514),
																													Column: int(76),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p5303,
																											FreeVars: ast.Identifiers{
																												"blank",
																												"neg",
																												"plus",
																											},
																										},
																										Cond: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(514),
																														Column: int(11),
																													},
																													End: ast.Location{
																														Line: int(514),
																														Column: int(14),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p5303,
																												FreeVars: ast.Identifiers{
																													"neg",
																												},
																											},
																											Id: "neg",
																										},
																										ThenFodder: ast.Fodder{},
																										BranchTrue: &ast.LiteralString{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(514),
																														Column: int(20),
																													},
																													End: ast.Location{
																														Line: int(514),
																														Column: int(23),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p5303,
																												FreeVars: nil,
																											},
																											Value: "-",
																											Kind: ast.LiteralStringKind(1),
																											BlockIndent: "",
																											BlockTermIndent: "",
																										},
																										ElseFodder: ast.Fodder{},
																										BranchFalse: &ast.Conditional{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(514),
																														Column: int(29),
																													},
																													End: ast.Location{
																														Line: int(514),
																														Column: int(76),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p5303,
																												FreeVars: ast.Identifiers{
																													"blank",
																													"plus",
																												},
																											},
																											Cond: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(514),
																															Column: int(32),
																														},
																														End: ast.Location{
																															Line: int(514),
																															Column: int(36),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p5303,
																													FreeVars: ast.Identifiers{
																														"plus",
																													},
																												},
																												Id: "plus",
																											},
																											ThenFodder: ast.Fodder{},
																											BranchTrue: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(514),
																															Column: int(42),
																														},
																														End: ast.Location{
																															Line: int(514),
																															Column: int(45),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p5303,
																													FreeVars: nil,
																												},
																												Value: "+",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																											ElseFodder: ast.Fodder{},
																											BranchFalse: &ast.Conditional{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(514),
																															Column: int(51),
																														},
																														End: ast.Location{
																															Line: int(514),
																															Column: int(76),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p5303,
																													FreeVars: ast.Identifiers{
																														"blank",
																													},
																												},
																												Cond: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(514),
																																Column: int(54),
																															},
																															End: ast.Location{
																																Line: int(514),
																																Column: int(59),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5303,
																														FreeVars: ast.Identifiers{
																															"blank",
																														},
																													},
																													Id: "blank",
																												},
																												ThenFodder: ast.Fodder{},
																												BranchTrue: &ast.LiteralString{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(514),
																																Column: int(65),
																															},
																															End: ast.Location{
																																Line: int(514),
																																Column: int(68),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5303,
																														FreeVars: nil,
																													},
																													Value: " ",
																													Kind: ast.LiteralStringKind(1),
																													BlockIndent: "",
																													BlockTermIndent: "",
																												},
																												ElseFodder: ast.Fodder{},
																												BranchFalse: &ast.LiteralString{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(514),
																																Column: int(74),
																															},
																															End: ast.Location{
																																Line: int(514),
																																Column: int(76),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5303,
																														FreeVars: nil,
																													},
																													Value: "",
																													Kind: ast.LiteralStringKind(1),
																													BlockIndent: "",
																													BlockTermIndent: "",
																												},
																											},
																										},
																									},
																									OpFodder: ast.Fodder{},
																									Op: ast.BinaryOp(3),
																									Right: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(514),
																													Column: int(80),
																												},
																												End: ast.Location{
																													Line: int(514),
																													Column: int(84),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p5303,
																											FreeVars: ast.Identifiers{
																												"dec2",
																											},
																										},
																										Id: "dec2",
																									},
																								},
																							},
																						},
																					},
																				},
																			},
																			Fun: nil,
																			CloseFodder: nil,
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																		},
																	},
																	Body: &ast.Local{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(517),
																					Column: int(5),
																				},
																				End: ast.Location{
																					Line: int(774),
																					Column: int(48),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(1),
																					Indent: int(4),
																					Comment: []string{},
																				},
																				ast.FodderElement{
																					Kind: ast.FodderKind(2),
																					Blanks: int(0),
																					Indent: int(4),
																					Comment: []string{
																						"// Render an integer in hexadecimal.",
																					},
																				},
																			},
																			Ctx: p3497,
																			FreeVars: ast.Identifiers{
																				"codes",
																				"pad_left",
																				"pad_right",
																				"render_int",
																				"std",
																				"vals",
																			},
																		},
																		Binds: ast.LocalBinds{
																			ast.LocalBind{
																				VarFodder: nil,
																				Variable: "render_hex",
																				EqFodder: nil,
																				Body: &ast.Function{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(517),
																								Column: int(11),
																							},
																							End: ast.Location{
																								Line: int(534),
																								Column: int(84),
																							},
																							File: p1,
																						},
																						Fodder: nil,
																						Ctx: p5467,
																						FreeVars: ast.Identifiers{
																							"pad_left",
																							"std",
																						},
																					},
																					ParenLeftFodder: ast.Fodder{},
																					Parameters: []ast.Parameter{
																						ast.Parameter{
																							NameFodder: ast.Fodder{},
																							Name: "n__",
																							EqFodder: nil,
																							DefaultArg: nil,
																							CommaFodder: ast.Fodder{},
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(517),
																									Column: int(22),
																								},
																								End: ast.Location{
																									Line: int(517),
																									Column: int(25),
																								},
																								File: p1,
																							},
																						},
																						ast.Parameter{
																							NameFodder: ast.Fodder{},
																							Name: "min_chars",
																							EqFodder: nil,
																							DefaultArg: nil,
																							CommaFodder: ast.Fodder{},
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(517),
																									Column: int(27),
																								},
																								End: ast.Location{
																									Line: int(517),
																									Column: int(36),
																								},
																								File: p1,
																							},
																						},
																						ast.Parameter{
																							NameFodder: ast.Fodder{},
																							Name: "min_digits",
																							EqFodder: nil,
																							DefaultArg: nil,
																							CommaFodder: ast.Fodder{},
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(517),
																									Column: int(38),
																								},
																								End: ast.Location{
																									Line: int(517),
																									Column: int(48),
																								},
																								File: p1,
																							},
																						},
																						ast.Parameter{
																							NameFodder: ast.Fodder{},
																							Name: "blank",
																							EqFodder: nil,
																							DefaultArg: nil,
																							CommaFodder: ast.Fodder{},
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(517),
																									Column: int(50),
																								},
																								End: ast.Location{
																									Line: int(517),
																									Column: int(55),
																								},
																								File: p1,
																							},
																						},
																						ast.Parameter{
																							NameFodder: ast.Fodder{},
																							Name: "plus",
																							EqFodder: nil,
																							DefaultArg: nil,
																							CommaFodder: ast.Fodder{},
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(517),
																									Column: int(57),
																								},
																								End: ast.Location{
																									Line: int(517),
																									Column: int(61),
																								},
																								File: p1,
																							},
																						},
																						ast.Parameter{
																							NameFodder: ast.Fodder{},
																							Name: "add_zerox",
																							EqFodder: nil,
																							DefaultArg: nil,
																							CommaFodder: ast.Fodder{},
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(517),
																									Column: int(63),
																								},
																								End: ast.Location{
																									Line: int(517),
																									Column: int(72),
																								},
																								File: p1,
																							},
																						},
																						ast.Parameter{
																							NameFodder: ast.Fodder{},
																							Name: "capitals",
																							EqFodder: nil,
																							DefaultArg: nil,
																							CommaFodder: nil,
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(517),
																									Column: int(74),
																								},
																								End: ast.Location{
																									Line: int(517),
																									Column: int(82),
																								},
																								File: p1,
																							},
																						},
																					},
																					TrailingComma: false,
																					ParenRightFodder: ast.Fodder{},
																					Body: &ast.Local{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(518),
																									Column: int(7),
																								},
																								End: ast.Location{
																									Line: int(534),
																									Column: int(84),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{
																								ast.FodderElement{
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(6),
																									Comment: []string{},
																								},
																							},
																							Ctx: p5472,
																							FreeVars: ast.Identifiers{
																								"add_zerox",
																								"blank",
																								"capitals",
																								"min_chars",
																								"min_digits",
																								"n__",
																								"pad_left",
																								"plus",
																								"std",
																							},
																						},
																						Binds: ast.LocalBinds{
																							ast.LocalBind{
																								VarFodder: ast.Fodder{},
																								Variable: "numerals",
																								EqFodder: ast.Fodder{},
																								Body: &ast.Binary{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(518),
																												Column: int(24),
																											},
																											End: ast.Location{
																												Line: int(520),
																												Column: int(59),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p5476,
																										FreeVars: ast.Identifiers{
																											"capitals",
																										},
																									},
																									Left: &ast.Array{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(518),
																													Column: int(24),
																												},
																												End: ast.Location{
																													Line: int(518),
																													Column: int(54),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p5476,
																											FreeVars: nil,
																										},
																										Elements: []ast.CommaSeparatedExpr{
																											ast.CommaSeparatedExpr{
																												Expr: &ast.LiteralNumber{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(518),
																																Column: int(25),
																															},
																															End: ast.Location{
																																Line: int(518),
																																Column: int(26),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5481,
																														FreeVars: nil,
																													},
																													OriginalString: "0",
																												},
																												CommaFodder: ast.Fodder{},
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.LiteralNumber{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(518),
																																Column: int(28),
																															},
																															End: ast.Location{
																																Line: int(518),
																																Column: int(29),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5481,
																														FreeVars: nil,
																													},
																													OriginalString: "1",
																												},
																												CommaFodder: ast.Fodder{},
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.LiteralNumber{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(518),
																																Column: int(31),
																															},
																															End: ast.Location{
																																Line: int(518),
																																Column: int(32),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5481,
																														FreeVars: nil,
																													},
																													OriginalString: "2",
																												},
																												CommaFodder: ast.Fodder{},
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.LiteralNumber{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(518),
																																Column: int(34),
																															},
																															End: ast.Location{
																																Line: int(518),
																																Column: int(35),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5481,
																														FreeVars: nil,
																													},
																													OriginalString: "3",
																												},
																												CommaFodder: ast.Fodder{},
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.LiteralNumber{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(518),
																																Column: int(37),
																															},
																															End: ast.Location{
																																Line: int(518),
																																Column: int(38),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5481,
																														FreeVars: nil,
																													},
																													OriginalString: "4",
																												},
																												CommaFodder: ast.Fodder{},
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.LiteralNumber{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(518),
																																Column: int(40),
																															},
																															End: ast.Location{
																																Line: int(518),
																																Column: int(41),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5481,
																														FreeVars: nil,
																													},
																													OriginalString: "5",
																												},
																												CommaFodder: ast.Fodder{},
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.LiteralNumber{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(518),
																																Column: int(43),
																															},
																															End: ast.Location{
																																Line: int(518),
																																Column: int(44),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5481,
																														FreeVars: nil,
																													},
																													OriginalString: "6",
																												},
																												CommaFodder: ast.Fodder{},
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.LiteralNumber{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(518),
																																Column: int(46),
																															},
																															End: ast.Location{
																																Line: int(518),
																																Column: int(47),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5481,
																														FreeVars: nil,
																													},
																													OriginalString: "7",
																												},
																												CommaFodder: ast.Fodder{},
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.LiteralNumber{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(518),
																																Column: int(49),
																															},
																															End: ast.Location{
																																Line: int(518),
																																Column: int(50),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5481,
																														FreeVars: nil,
																													},
																													OriginalString: "8",
																												},
																												CommaFodder: ast.Fodder{},
																											},
																											ast.CommaSeparatedExpr{
																												Expr: &ast.LiteralNumber{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(518),
																																Column: int(52),
																															},
																															End: ast.Location{
																																Line: int(518),
																																Column: int(53),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5481,
																														FreeVars: nil,
																													},
																													OriginalString: "9",
																												},
																												CommaFodder: nil,
																											},
																										},
																										TrailingComma: false,
																										CloseFodder: ast.Fodder{},
																									},
																									OpFodder: ast.Fodder{
																										ast.FodderElement{
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(23),
																											Comment: []string{},
																										},
																									},
																									Op: ast.BinaryOp(3),
																									Right: &ast.Conditional{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(519),
																													Column: int(26),
																												},
																												End: ast.Location{
																													Line: int(520),
																													Column: int(59),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p5476,
																											FreeVars: ast.Identifiers{
																												"capitals",
																											},
																										},
																										Cond: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(519),
																														Column: int(29),
																													},
																													End: ast.Location{
																														Line: int(519),
																														Column: int(37),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p5476,
																												FreeVars: ast.Identifiers{
																													"capitals",
																												},
																											},
																											Id: "capitals",
																										},
																										ThenFodder: ast.Fodder{},
																										BranchTrue: &ast.Array{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(519),
																														Column: int(43),
																													},
																													End: ast.Location{
																														Line: int(519),
																														Column: int(73),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p5476,
																												FreeVars: nil,
																											},
																											Elements: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(519),
																																	Column: int(44),
																																},
																																End: ast.Location{
																																	Line: int(519),
																																	Column: int(47),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5499,
																															FreeVars: nil,
																														},
																														Value: "A",
																														Kind: ast.LiteralStringKind(1),
																														BlockIndent: "",
																														BlockTermIndent: "",
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(519),
																																	Column: int(49),
																																},
																																End: ast.Location{
																																	Line: int(519),
																																	Column: int(52),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5499,
																															FreeVars: nil,
																														},
																														Value: "B",
																														Kind: ast.LiteralStringKind(1),
																														BlockIndent: "",
																														BlockTermIndent: "",
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(519),
																																	Column: int(54),
																																},
																																End: ast.Location{
																																	Line: int(519),
																																	Column: int(57),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5499,
																															FreeVars: nil,
																														},
																														Value: "C",
																														Kind: ast.LiteralStringKind(1),
																														BlockIndent: "",
																														BlockTermIndent: "",
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(519),
																																	Column: int(59),
																																},
																																End: ast.Location{
																																	Line: int(519),
																																	Column: int(62),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5499,
																															FreeVars: nil,
																														},
																														Value: "D",
																														Kind: ast.LiteralStringKind(1),
																														BlockIndent: "",
																														BlockTermIndent: "",
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(519),
																																	Column: int(64),
																																},
																																End: ast.Location{
																																	Line: int(519),
																																	Column: int(67),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5499,
																															FreeVars: nil,
																														},
																														Value: "E",
																														Kind: ast.LiteralStringKind(1),
																														BlockIndent: "",
																														BlockTermIndent: "",
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(519),
																																	Column: int(69),
																																},
																																End: ast.Location{
																																	Line: int(519),
																																	Column: int(72),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5499,
																															FreeVars: nil,
																														},
																														Value: "F",
																														Kind: ast.LiteralStringKind(1),
																														BlockIndent: "",
																														BlockTermIndent: "",
																													},
																													CommaFodder: nil,
																												},
																											},
																											TrailingComma: false,
																											CloseFodder: ast.Fodder{},
																										},
																										ElseFodder: ast.Fodder{
																											ast.FodderElement{
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(23),
																												Comment: []string{},
																											},
																										},
																										BranchFalse: &ast.Array{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(520),
																														Column: int(29),
																													},
																													End: ast.Location{
																														Line: int(520),
																														Column: int(59),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p5476,
																												FreeVars: nil,
																											},
																											Elements: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(520),
																																	Column: int(30),
																																},
																																End: ast.Location{
																																	Line: int(520),
																																	Column: int(33),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5509,
																															FreeVars: nil,
																														},
																														Value: "a",
																														Kind: ast.LiteralStringKind(1),
																														BlockIndent: "",
																														BlockTermIndent: "",
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(520),
																																	Column: int(35),
																																},
																																End: ast.Location{
																																	Line: int(520),
																																	Column: int(38),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5509,
																															FreeVars: nil,
																														},
																														Value: "b",
																														Kind: ast.LiteralStringKind(1),
																														BlockIndent: "",
																														BlockTermIndent: "",
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(520),
																																	Column: int(40),
																																},
																																End: ast.Location{
																																	Line: int(520),
																																	Column: int(43),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5509,
																															FreeVars: nil,
																														},
																														Value: "c",
																														Kind: ast.LiteralStringKind(1),
																														BlockIndent: "",
																														BlockTermIndent: "",
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(520),
																																	Column: int(45),
																																},
																																End: ast.Location{
																																	Line: int(520),
																																	Column: int(48),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5509,
																															FreeVars: nil,
																														},
																														Value: "d",
																														Kind: ast.LiteralStringKind(1),
																														BlockIndent: "",
																														BlockTermIndent: "",
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(520),
																																	Column: int(50),
																																},
																																End: ast.Location{
																																	Line: int(520),
																																	Column: int(53),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5509,
																															FreeVars: nil,
																														},
																														Value: "e",
																														Kind: ast.LiteralStringKind(1),
																														BlockIndent: "",
																														BlockTermIndent: "",
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(520),
																																	Column: int(55),
																																},
																																End: ast.Location{
																																	Line: int(520),
																																	Column: int(58),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5509,
																															FreeVars: nil,
																														},
																														Value: "f",
																														Kind: ast.LiteralStringKind(1),
																														BlockIndent: "",
																														BlockTermIndent: "",
																													},
																													CommaFodder: nil,
																												},
																											},
																											TrailingComma: false,
																											CloseFodder: ast.Fodder{},
																										},
																									},
																								},
																								Fun: nil,
																								CloseFodder: ast.Fodder{},
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(518),
																										Column: int(13),
																									},
																									End: ast.Location{
																										Line: int(520),
																										Column: int(59),
																									},
																									File: p1,
																								},
																							},
																						},
																						Body: &ast.Local{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(521),
																										Column: int(7),
																									},
																									End: ast.Location{
																										Line: int(534),
																										Column: int(84),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{
																									ast.FodderElement{
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(6),
																										Comment: []string{},
																									},
																								},
																								Ctx: p5472,
																								FreeVars: ast.Identifiers{
																									"add_zerox",
																									"blank",
																									"capitals",
																									"min_chars",
																									"min_digits",
																									"n__",
																									"numerals",
																									"pad_left",
																									"plus",
																									"std",
																								},
																							},
																							Binds: ast.LocalBinds{
																								ast.LocalBind{
																									VarFodder: ast.Fodder{},
																									Variable: "n_",
																									EqFodder: ast.Fodder{},
																									Body: &ast.Apply{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(521),
																													Column: int(18),
																												},
																												End: ast.Location{
																													Line: int(521),
																													Column: int(30),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p5520,
																											FreeVars: ast.Identifiers{
																												"n__",
																												"std",
																											},
																										},
																										Target: &ast.Index{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(521),
																														Column: int(18),
																													},
																													End: ast.Location{
																														Line: int(521),
																														Column: int(25),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p5520,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																											},
																											Target: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(521),
																															Column: int(18),
																														},
																														End: ast.Location{
																															Line: int(521),
																															Column: int(21),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																												},
																												Id: "std",
																											},
																											LeftBracketFodder: ast.Fodder{},
																											Index: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: nil,
																												},
																												Value: "abs",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																											RightBracketFodder: ast.Fodder{},
																											Id: nil,
																										},
																										FodderLeft: ast.Fodder{},
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(521),
																																	Column: int(26),
																																},
																																End: ast.Location{
																																	Line: int(521),
																																	Column: int(29),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5529,
																															FreeVars: ast.Identifiers{
																																"n__",
																															},
																														},
																														Id: "n__",
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										TrailingComma: false,
																										TailStrict: false,
																										FodderRight: ast.Fodder{},
																										TailStrictFodder: nil,
																									},
																									Fun: nil,
																									CloseFodder: ast.Fodder{},
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(521),
																											Column: int(13),
																										},
																										End: ast.Location{
																											Line: int(521),
																											Column: int(30),
																										},
																										File: p1,
																									},
																								},
																							},
																							Body: &ast.Local{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(522),
																											Column: int(7),
																										},
																										End: ast.Location{
																											Line: int(534),
																											Column: int(84),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{
																										ast.FodderElement{
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(6),
																											Comment: []string{},
																										},
																									},
																									Ctx: p5472,
																									FreeVars: ast.Identifiers{
																										"add_zerox",
																										"blank",
																										"capitals",
																										"min_chars",
																										"min_digits",
																										"n_",
																										"n__",
																										"numerals",
																										"pad_left",
																										"plus",
																										"std",
																									},
																								},
																								Binds: ast.LocalBinds{
																									ast.LocalBind{
																										VarFodder: nil,
																										Variable: "aux",
																										EqFodder: nil,
																										Body: &ast.Function{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(522),
																														Column: int(13),
																													},
																													End: ast.Location{
																														Line: int(526),
																														Column: int(52),
																													},
																													File: p1,
																												},
																												Fodder: nil,
																												Ctx: p5536,
																												FreeVars: ast.Identifiers{
																													"aux",
																													"numerals",
																													"std",
																												},
																											},
																											ParenLeftFodder: ast.Fodder{},
																											Parameters: []ast.Parameter{
																												ast.Parameter{
																													NameFodder: ast.Fodder{},
																													Name: "n",
																													EqFodder: nil,
																													DefaultArg: nil,
																													CommaFodder: nil,
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(522),
																															Column: int(17),
																														},
																														End: ast.Location{
																															Line: int(522),
																															Column: int(18),
																														},
																														File: p1,
																													},
																												},
																											},
																											TrailingComma: false,
																											ParenRightFodder: ast.Fodder{},
																											Body: &ast.Conditional{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(523),
																															Column: int(9),
																														},
																														End: ast.Location{
																															Line: int(526),
																															Column: int(52),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(8),
																															Comment: []string{},
																														},
																													},
																													Ctx: p5541,
																													FreeVars: ast.Identifiers{
																														"aux",
																														"n",
																														"numerals",
																														"std",
																													},
																												},
																												Cond: &ast.Binary{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(523),
																																Column: int(12),
																															},
																															End: ast.Location{
																																Line: int(523),
																																Column: int(18),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5541,
																														FreeVars: ast.Identifiers{
																															"n",
																														},
																													},
																													Left: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(523),
																																	Column: int(12),
																																},
																																End: ast.Location{
																																	Line: int(523),
																																	Column: int(13),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5541,
																															FreeVars: ast.Identifiers{
																																"n",
																															},
																														},
																														Id: "n",
																													},
																													OpFodder: ast.Fodder{},
																													Op: ast.BinaryOp(12),
																													Right: &ast.LiteralNumber{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(523),
																																	Column: int(17),
																																},
																																End: ast.Location{
																																	Line: int(523),
																																	Column: int(18),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5541,
																															FreeVars: nil,
																														},
																														OriginalString: "0",
																													},
																												},
																												ThenFodder: ast.Fodder{},
																												BranchTrue: &ast.LiteralString{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(524),
																																Column: int(11),
																															},
																															End: ast.Location{
																																Line: int(524),
																																Column: int(13),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(10),
																																Comment: []string{},
																															},
																														},
																														Ctx: p5541,
																														FreeVars: nil,
																													},
																													Value: "",
																													Kind: ast.LiteralStringKind(1),
																													BlockIndent: "",
																													BlockTermIndent: "",
																												},
																												ElseFodder: ast.Fodder{
																													ast.FodderElement{
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(8),
																														Comment: []string{},
																													},
																												},
																												BranchFalse: &ast.Binary{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(526),
																																Column: int(11),
																															},
																															End: ast.Location{
																																Line: int(526),
																																Column: int(52),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5541,
																														FreeVars: ast.Identifiers{
																															"aux",
																															"n",
																															"numerals",
																															"std",
																														},
																													},
																													Left: &ast.Apply{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(526),
																																	Column: int(11),
																																},
																																End: ast.Location{
																																	Line: int(526),
																																	Column: int(33),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5541,
																															FreeVars: ast.Identifiers{
																																"aux",
																																"n",
																																"std",
																															},
																														},
																														Target: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(526),
																																		Column: int(11),
																																	},
																																	End: ast.Location{
																																		Line: int(526),
																																		Column: int(14),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{
																																	ast.FodderElement{
																																		Kind: ast.FodderKind(0),
																																		Blanks: int(0),
																																		Indent: int(10),
																																		Comment: []string{},
																																	},
																																},
																																Ctx: p5541,
																																FreeVars: ast.Identifiers{
																																	"aux",
																																},
																															},
																															Id: "aux",
																														},
																														FodderLeft: ast.Fodder{},
																														Arguments: ast.Arguments{
																															Positional: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Apply{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(526),
																																					Column: int(15),
																																				},
																																				End: ast.Location{
																																					Line: int(526),
																																					Column: int(32),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p5560,
																																			FreeVars: ast.Identifiers{
																																				"n",
																																				"std",
																																			},
																																		},
																																		Target: &ast.Index{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(526),
																																						Column: int(15),
																																					},
																																					End: ast.Location{
																																						Line: int(526),
																																						Column: int(24),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p5560,
																																				FreeVars: ast.Identifiers{
																																					"std",
																																				},
																																			},
																																			Target: &ast.Var{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(526),
																																							Column: int(15),
																																						},
																																						End: ast.Location{
																																							Line: int(526),
																																							Column: int(18),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{
																																						"std",
																																					},
																																				},
																																				Id: "std",
																																			},
																																			LeftBracketFodder: ast.Fodder{},
																																			Index: &ast.LiteralString{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						End: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						File: nil,
																																					},
																																					Fodder: nil,
																																					Ctx: nil,
																																					FreeVars: nil,
																																				},
																																				Value: "floor",
																																				Kind: ast.LiteralStringKind(1),
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																			},
																																			RightBracketFodder: ast.Fodder{},
																																			Id: nil,
																																		},
																																		FodderLeft: ast.Fodder{},
																																		Arguments: ast.Arguments{
																																			Positional: []ast.CommaSeparatedExpr{
																																				ast.CommaSeparatedExpr{
																																					Expr: &ast.Binary{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(526),
																																									Column: int(25),
																																								},
																																								End: ast.Location{
																																									Line: int(526),
																																									Column: int(31),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p5569,
																																							FreeVars: ast.Identifiers{
																																								"n",
																																							},
																																						},
																																						Left: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(526),
																																										Column: int(25),
																																									},
																																									End: ast.Location{
																																										Line: int(526),
																																										Column: int(26),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p5569,
																																								FreeVars: ast.Identifiers{
																																									"n",
																																								},
																																							},
																																							Id: "n",
																																						},
																																						OpFodder: ast.Fodder{},
																																						Op: ast.BinaryOp(1),
																																						Right: &ast.LiteralNumber{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(526),
																																										Column: int(29),
																																									},
																																									End: ast.Location{
																																										Line: int(526),
																																										Column: int(31),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p5569,
																																								FreeVars: nil,
																																							},
																																							OriginalString: "16",
																																						},
																																					},
																																					CommaFodder: nil,
																																				},
																																			},
																																			Named: nil,
																																		},
																																		TrailingComma: false,
																																		TailStrict: false,
																																		FodderRight: ast.Fodder{},
																																		TailStrictFodder: nil,
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															Named: nil,
																														},
																														TrailingComma: false,
																														TailStrict: false,
																														FodderRight: ast.Fodder{},
																														TailStrictFodder: nil,
																													},
																													OpFodder: ast.Fodder{},
																													Op: ast.BinaryOp(3),
																													Right: &ast.Index{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(526),
																																	Column: int(36),
																																},
																																End: ast.Location{
																																	Line: int(526),
																																	Column: int(52),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5541,
																															FreeVars: ast.Identifiers{
																																"n",
																																"numerals",
																																"std",
																															},
																														},
																														Target: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(526),
																																		Column: int(36),
																																	},
																																	End: ast.Location{
																																		Line: int(526),
																																		Column: int(44),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p5541,
																																FreeVars: ast.Identifiers{
																																	"numerals",
																																},
																															},
																															Id: "numerals",
																														},
																														LeftBracketFodder: ast.Fodder{},
																														Index: &ast.Apply{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	File: nil,
																																},
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"n",
																																	"std",
																																},
																															},
																															Target: &ast.Index{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		File: nil,
																																	},
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																	},
																																},
																																Target: &ast.Var{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			End: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			File: nil,
																																		},
																																		Fodder: nil,
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{
																																			"std",
																																		},
																																	},
																																	Id: "std",
																																},
																																LeftBracketFodder: nil,
																																Index: &ast.LiteralString{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			End: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			File: nil,
																																		},
																																		Fodder: nil,
																																		Ctx: nil,
																																		FreeVars: nil,
																																	},
																																	Value: "mod",
																																	Kind: ast.LiteralStringKind(1),
																																	BlockIndent: "",
																																	BlockTermIndent: "",
																																},
																																RightBracketFodder: nil,
																																Id: nil,
																															},
																															FodderLeft: nil,
																															Arguments: ast.Arguments{
																																Positional: []ast.CommaSeparatedExpr{
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(526),
																																						Column: int(45),
																																					},
																																					End: ast.Location{
																																						Line: int(526),
																																						Column: int(46),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p5541,
																																				FreeVars: ast.Identifiers{
																																					"n",
																																				},
																																			},
																																			Id: "n",
																																		},
																																		CommaFodder: nil,
																																	},
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.LiteralNumber{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(526),
																																						Column: int(49),
																																					},
																																					End: ast.Location{
																																						Line: int(526),
																																						Column: int(51),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p5541,
																																				FreeVars: nil,
																																			},
																																			OriginalString: "16",
																																		},
																																		CommaFodder: nil,
																																	},
																																},
																																Named: nil,
																															},
																															TrailingComma: false,
																															TailStrict: false,
																															FodderRight: nil,
																															TailStrictFodder: nil,
																														},
																														RightBracketFodder: ast.Fodder{},
																														Id: nil,
																													},
																												},
																											},
																										},
																										Fun: nil,
																										CloseFodder: nil,
																										LocRange: ast.LocationRange{
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											File: nil,
																										},
																									},
																								},
																								Body: &ast.Local{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(527),
																												Column: int(7),
																											},
																											End: ast.Location{
																												Line: int(534),
																												Column: int(84),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{
																											ast.FodderElement{
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(6),
																												Comment: []string{},
																											},
																										},
																										Ctx: p5472,
																										FreeVars: ast.Identifiers{
																											"add_zerox",
																											"aux",
																											"blank",
																											"capitals",
																											"min_chars",
																											"min_digits",
																											"n_",
																											"n__",
																											"pad_left",
																											"plus",
																											"std",
																										},
																									},
																									Binds: ast.LocalBinds{
																										ast.LocalBind{
																											VarFodder: ast.Fodder{},
																											Variable: "hex",
																											EqFodder: ast.Fodder{},
																											Body: &ast.Conditional{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(527),
																															Column: int(19),
																														},
																														End: ast.Location{
																															Line: int(527),
																															Column: int(73),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p5594,
																													FreeVars: ast.Identifiers{
																														"aux",
																														"n_",
																														"std",
																													},
																												},
																												Cond: &ast.Binary{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(527),
																																Column: int(22),
																															},
																															End: ast.Location{
																																Line: int(527),
																																Column: int(40),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5594,
																														FreeVars: ast.Identifiers{
																															"n_",
																															"std",
																														},
																													},
																													Left: &ast.Apply{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(527),
																																	Column: int(22),
																																},
																																End: ast.Location{
																																	Line: int(527),
																																	Column: int(35),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5594,
																															FreeVars: ast.Identifiers{
																																"n_",
																																"std",
																															},
																														},
																														Target: &ast.Index{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(527),
																																		Column: int(22),
																																	},
																																	End: ast.Location{
																																		Line: int(527),
																																		Column: int(31),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p5594,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																															},
																															Target: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(527),
																																			Column: int(22),
																																		},
																																		End: ast.Location{
																																			Line: int(527),
																																			Column: int(25),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																	},
																																},
																																Id: "std",
																															},
																															LeftBracketFodder: ast.Fodder{},
																															Index: &ast.LiteralString{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		File: nil,
																																	},
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: nil,
																																},
																																Value: "floor",
																																Kind: ast.LiteralStringKind(1),
																																BlockIndent: "",
																																BlockTermIndent: "",
																															},
																															RightBracketFodder: ast.Fodder{},
																															Id: nil,
																														},
																														FodderLeft: ast.Fodder{},
																														Arguments: ast.Arguments{
																															Positional: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(527),
																																					Column: int(32),
																																				},
																																				End: ast.Location{
																																					Line: int(527),
																																					Column: int(34),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p5607,
																																			FreeVars: ast.Identifiers{
																																				"n_",
																																			},
																																		},
																																		Id: "n_",
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															Named: nil,
																														},
																														TrailingComma: false,
																														TailStrict: false,
																														FodderRight: ast.Fodder{},
																														TailStrictFodder: nil,
																													},
																													OpFodder: ast.Fodder{},
																													Op: ast.BinaryOp(12),
																													Right: &ast.LiteralNumber{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(527),
																																	Column: int(39),
																																},
																																End: ast.Location{
																																	Line: int(527),
																																	Column: int(40),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5594,
																															FreeVars: nil,
																														},
																														OriginalString: "0",
																													},
																												},
																												ThenFodder: ast.Fodder{},
																												BranchTrue: &ast.LiteralString{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(527),
																																Column: int(46),
																															},
																															End: ast.Location{
																																Line: int(527),
																																Column: int(49),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5594,
																														FreeVars: nil,
																													},
																													Value: "0",
																													Kind: ast.LiteralStringKind(1),
																													BlockIndent: "",
																													BlockTermIndent: "",
																												},
																												ElseFodder: ast.Fodder{},
																												BranchFalse: &ast.Apply{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(527),
																																Column: int(55),
																															},
																															End: ast.Location{
																																Line: int(527),
																																Column: int(73),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5594,
																														FreeVars: ast.Identifiers{
																															"aux",
																															"n_",
																															"std",
																														},
																													},
																													Target: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(527),
																																	Column: int(55),
																																},
																																End: ast.Location{
																																	Line: int(527),
																																	Column: int(58),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5594,
																															FreeVars: ast.Identifiers{
																																"aux",
																															},
																														},
																														Id: "aux",
																													},
																													FodderLeft: ast.Fodder{},
																													Arguments: ast.Arguments{
																														Positional: []ast.CommaSeparatedExpr{
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Apply{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(527),
																																				Column: int(59),
																																			},
																																			End: ast.Location{
																																				Line: int(527),
																																				Column: int(72),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p5617,
																																		FreeVars: ast.Identifiers{
																																			"n_",
																																			"std",
																																		},
																																	},
																																	Target: &ast.Index{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(527),
																																					Column: int(59),
																																				},
																																				End: ast.Location{
																																					Line: int(527),
																																					Column: int(68),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p5617,
																																			FreeVars: ast.Identifiers{
																																				"std",
																																			},
																																		},
																																		Target: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(527),
																																						Column: int(59),
																																					},
																																					End: ast.Location{
																																						Line: int(527),
																																						Column: int(62),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: nil,
																																				FreeVars: ast.Identifiers{
																																					"std",
																																				},
																																			},
																																			Id: "std",
																																		},
																																		LeftBracketFodder: ast.Fodder{},
																																		Index: &ast.LiteralString{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(0),
																																						Column: int(0),
																																					},
																																					End: ast.Location{
																																						Line: int(0),
																																						Column: int(0),
																																					},
																																					File: nil,
																																				},
																																				Fodder: nil,
																																				Ctx: nil,
																																				FreeVars: nil,
																																			},
																																			Value: "floor",
																																			Kind: ast.LiteralStringKind(1),
																																			BlockIndent: "",
																																			BlockTermIndent: "",
																																		},
																																		RightBracketFodder: ast.Fodder{},
																																		Id: nil,
																																	},
																																	FodderLeft: ast.Fodder{},
																																	Arguments: ast.Arguments{
																																		Positional: []ast.CommaSeparatedExpr{
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Var{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(527),
																																								Column: int(69),
																																							},
																																							End: ast.Location{
																																								Line: int(527),
																																								Column: int(71),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p5626,
																																						FreeVars: ast.Identifiers{
																																							"n_",
																																						},
																																					},
																																					Id: "n_",
																																				},
																																				CommaFodder: nil,
																																			},
																																		},
																																		Named: nil,
																																	},
																																	TrailingComma: false,
																																	TailStrict: false,
																																	FodderRight: ast.Fodder{},
																																	TailStrictFodder: nil,
																																},
																																CommaFodder: nil,
																															},
																														},
																														Named: nil,
																													},
																													TrailingComma: false,
																													TailStrict: false,
																													FodderRight: ast.Fodder{},
																													TailStrictFodder: nil,
																												},
																											},
																											Fun: nil,
																											CloseFodder: ast.Fodder{},
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(527),
																													Column: int(13),
																												},
																												End: ast.Location{
																													Line: int(527),
																													Column: int(73),
																												},
																												File: p1,
																											},
																										},
																									},
																									Body: &ast.Local{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(528),
																													Column: int(7),
																												},
																												End: ast.Location{
																													Line: int(534),
																													Column: int(84),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{
																												ast.FodderElement{
																													Kind: ast.FodderKind(0),
																													Blanks: int(0),
																													Indent: int(6),
																													Comment: []string{},
																												},
																											},
																											Ctx: p5472,
																											FreeVars: ast.Identifiers{
																												"add_zerox",
																												"blank",
																												"capitals",
																												"hex",
																												"min_chars",
																												"min_digits",
																												"n__",
																												"pad_left",
																												"plus",
																												"std",
																											},
																										},
																										Binds: ast.LocalBinds{
																											ast.LocalBind{
																												VarFodder: ast.Fodder{},
																												Variable: "neg",
																												EqFodder: ast.Fodder{},
																												Body: &ast.Binary{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(528),
																																Column: int(19),
																															},
																															End: ast.Location{
																																Line: int(528),
																																Column: int(26),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5633,
																														FreeVars: ast.Identifiers{
																															"n__",
																														},
																													},
																													Left: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(528),
																																	Column: int(19),
																																},
																																End: ast.Location{
																																	Line: int(528),
																																	Column: int(22),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5633,
																															FreeVars: ast.Identifiers{
																																"n__",
																															},
																														},
																														Id: "n__",
																													},
																													OpFodder: ast.Fodder{},
																													Op: ast.BinaryOp(9),
																													Right: &ast.LiteralNumber{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(528),
																																	Column: int(25),
																																},
																																End: ast.Location{
																																	Line: int(528),
																																	Column: int(26),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5633,
																															FreeVars: nil,
																														},
																														OriginalString: "0",
																													},
																												},
																												Fun: nil,
																												CloseFodder: ast.Fodder{},
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(528),
																														Column: int(13),
																													},
																													End: ast.Location{
																														Line: int(528),
																														Column: int(26),
																													},
																													File: p1,
																												},
																											},
																										},
																										Body: &ast.Local{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(529),
																														Column: int(7),
																													},
																													End: ast.Location{
																														Line: int(534),
																														Column: int(84),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{
																													ast.FodderElement{
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(6),
																														Comment: []string{},
																													},
																												},
																												Ctx: p5472,
																												FreeVars: ast.Identifiers{
																													"add_zerox",
																													"blank",
																													"capitals",
																													"hex",
																													"min_chars",
																													"min_digits",
																													"neg",
																													"pad_left",
																													"plus",
																													"std",
																												},
																											},
																											Binds: ast.LocalBinds{
																												ast.LocalBind{
																													VarFodder: ast.Fodder{},
																													Variable: "zp",
																													EqFodder: ast.Fodder{},
																													Body: &ast.Binary{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(529),
																																	Column: int(18),
																																},
																																End: ast.Location{
																																	Line: int(530),
																																	Column: int(48),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5643,
																															FreeVars: ast.Identifiers{
																																"add_zerox",
																																"blank",
																																"min_chars",
																																"neg",
																																"plus",
																															},
																														},
																														Left: &ast.Binary{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(529),
																																		Column: int(18),
																																	},
																																	End: ast.Location{
																																		Line: int(529),
																																		Column: int(69),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p5643,
																																FreeVars: ast.Identifiers{
																																	"blank",
																																	"min_chars",
																																	"neg",
																																	"plus",
																																},
																															},
																															Left: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(529),
																																			Column: int(18),
																																		},
																																		End: ast.Location{
																																			Line: int(529),
																																			Column: int(27),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p5643,
																																	FreeVars: ast.Identifiers{
																																		"min_chars",
																																	},
																																},
																																Id: "min_chars",
																															},
																															OpFodder: ast.Fodder{},
																															Op: ast.BinaryOp(4),
																															Right: &ast.Conditional{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(529),
																																			Column: int(31),
																																		},
																																		End: ast.Location{
																																			Line: int(529),
																																			Column: int(68),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p5643,
																																	FreeVars: ast.Identifiers{
																																		"blank",
																																		"neg",
																																		"plus",
																																	},
																																},
																																Cond: &ast.Binary{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(529),
																																				Column: int(34),
																																			},
																																			End: ast.Location{
																																				Line: int(529),
																																				Column: int(54),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p5643,
																																		FreeVars: ast.Identifiers{
																																			"blank",
																																			"neg",
																																			"plus",
																																		},
																																	},
																																	Left: &ast.Binary{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(529),
																																					Column: int(34),
																																				},
																																				End: ast.Location{
																																					Line: int(529),
																																					Column: int(46),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p5643,
																																			FreeVars: ast.Identifiers{
																																				"blank",
																																				"neg",
																																			},
																																		},
																																		Left: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(529),
																																						Column: int(34),
																																					},
																																					End: ast.Location{
																																						Line: int(529),
																																						Column: int(37),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p5643,
																																				FreeVars: ast.Identifiers{
																																					"neg",
																																				},
																																			},
																																			Id: "neg",
																																		},
																																		OpFodder: ast.Fodder{},
																																		Op: ast.BinaryOp(18),
																																		Right: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(529),
																																						Column: int(41),
																																					},
																																					End: ast.Location{
																																						Line: int(529),
																																						Column: int(46),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p5643,
																																				FreeVars: ast.Identifiers{
																																					"blank",
																																				},
																																			},
																																			Id: "blank",
																																		},
																																	},
																																	OpFodder: ast.Fodder{},
																																	Op: ast.BinaryOp(18),
																																	Right: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(529),
																																					Column: int(50),
																																				},
																																				End: ast.Location{
																																					Line: int(529),
																																					Column: int(54),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p5643,
																																			FreeVars: ast.Identifiers{
																																				"plus",
																																			},
																																		},
																																		Id: "plus",
																																	},
																																},
																																ThenFodder: ast.Fodder{},
																																BranchTrue: &ast.LiteralNumber{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(529),
																																				Column: int(60),
																																			},
																																			End: ast.Location{
																																				Line: int(529),
																																				Column: int(61),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p5643,
																																		FreeVars: nil,
																																	},
																																	OriginalString: "1",
																																},
																																ElseFodder: ast.Fodder{},
																																BranchFalse: &ast.LiteralNumber{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(529),
																																				Column: int(67),
																																			},
																																			End: ast.Location{
																																				Line: int(529),
																																				Column: int(68),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p5643,
																																		FreeVars: nil,
																																	},
																																	OriginalString: "0",
																																},
																															},
																														},
																														OpFodder: ast.Fodder{
																															ast.FodderElement{
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(17),
																																Comment: []string{},
																															},
																														},
																														Op: ast.BinaryOp(4),
																														Right: &ast.Conditional{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(530),
																																		Column: int(21),
																																	},
																																	End: ast.Location{
																																		Line: int(530),
																																		Column: int(47),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p5643,
																																FreeVars: ast.Identifiers{
																																	"add_zerox",
																																},
																															},
																															Cond: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(530),
																																			Column: int(24),
																																		},
																																		End: ast.Location{
																																			Line: int(530),
																																			Column: int(33),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p5643,
																																	FreeVars: ast.Identifiers{
																																		"add_zerox",
																																	},
																																},
																																Id: "add_zerox",
																															},
																															ThenFodder: ast.Fodder{},
																															BranchTrue: &ast.LiteralNumber{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(530),
																																			Column: int(39),
																																		},
																																		End: ast.Location{
																																			Line: int(530),
																																			Column: int(40),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p5643,
																																	FreeVars: nil,
																																},
																																OriginalString: "2",
																															},
																															ElseFodder: ast.Fodder{},
																															BranchFalse: &ast.LiteralNumber{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(530),
																																			Column: int(46),
																																		},
																																		End: ast.Location{
																																			Line: int(530),
																																			Column: int(47),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p5643,
																																	FreeVars: nil,
																																},
																																OriginalString: "0",
																															},
																														},
																													},
																													Fun: nil,
																													CloseFodder: ast.Fodder{},
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(529),
																															Column: int(13),
																														},
																														End: ast.Location{
																															Line: int(530),
																															Column: int(48),
																														},
																														File: p1,
																													},
																												},
																											},
																											Body: &ast.Local{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(531),
																															Column: int(7),
																														},
																														End: ast.Location{
																															Line: int(534),
																															Column: int(84),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(6),
																															Comment: []string{},
																														},
																													},
																													Ctx: p5472,
																													FreeVars: ast.Identifiers{
																														"add_zerox",
																														"blank",
																														"capitals",
																														"hex",
																														"min_digits",
																														"neg",
																														"pad_left",
																														"plus",
																														"std",
																														"zp",
																													},
																												},
																												Binds: ast.LocalBinds{
																													ast.LocalBind{
																														VarFodder: ast.Fodder{},
																														Variable: "zp2",
																														EqFodder: ast.Fodder{},
																														Body: &ast.Apply{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(531),
																																		Column: int(19),
																																	},
																																	End: ast.Location{
																																		Line: int(531),
																																		Column: int(42),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p5675,
																																FreeVars: ast.Identifiers{
																																	"min_digits",
																																	"std",
																																	"zp",
																																},
																															},
																															Target: &ast.Index{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(531),
																																			Column: int(19),
																																		},
																																		End: ast.Location{
																																			Line: int(531),
																																			Column: int(26),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p5675,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																	},
																																},
																																Target: &ast.Var{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(531),
																																				Column: int(19),
																																			},
																																			End: ast.Location{
																																				Line: int(531),
																																				Column: int(22),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{
																																			"std",
																																		},
																																	},
																																	Id: "std",
																																},
																																LeftBracketFodder: ast.Fodder{},
																																Index: &ast.LiteralString{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			End: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			File: nil,
																																		},
																																		Fodder: nil,
																																		Ctx: nil,
																																		FreeVars: nil,
																																	},
																																	Value: "max",
																																	Kind: ast.LiteralStringKind(1),
																																	BlockIndent: "",
																																	BlockTermIndent: "",
																																},
																																RightBracketFodder: ast.Fodder{},
																																Id: nil,
																															},
																															FodderLeft: ast.Fodder{},
																															Arguments: ast.Arguments{
																																Positional: []ast.CommaSeparatedExpr{
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(531),
																																						Column: int(27),
																																					},
																																					End: ast.Location{
																																						Line: int(531),
																																						Column: int(29),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p5684,
																																				FreeVars: ast.Identifiers{
																																					"zp",
																																				},
																																			},
																																			Id: "zp",
																																		},
																																		CommaFodder: ast.Fodder{},
																																	},
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(531),
																																						Column: int(31),
																																					},
																																					End: ast.Location{
																																						Line: int(531),
																																						Column: int(41),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p5684,
																																				FreeVars: ast.Identifiers{
																																					"min_digits",
																																				},
																																			},
																																			Id: "min_digits",
																																		},
																																		CommaFodder: nil,
																																	},
																																},
																																Named: nil,
																															},
																															TrailingComma: false,
																															TailStrict: false,
																															FodderRight: ast.Fodder{},
																															TailStrictFodder: nil,
																														},
																														Fun: nil,
																														CloseFodder: ast.Fodder{},
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(531),
																																Column: int(13),
																															},
																															End: ast.Location{
																																Line: int(531),
																																Column: int(42),
																															},
																															File: p1,
																														},
																													},
																												},
																												Body: &ast.Local{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(532),
																																Column: int(7),
																															},
																															End: ast.Location{
																																Line: int(534),
																																Column: int(84),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(6),
																																Comment: []string{},
																															},
																														},
																														Ctx: p5472,
																														FreeVars: ast.Identifiers{
																															"add_zerox",
																															"blank",
																															"capitals",
																															"hex",
																															"neg",
																															"pad_left",
																															"plus",
																															"zp2",
																														},
																													},
																													Binds: ast.LocalBinds{
																														ast.LocalBind{
																															VarFodder: ast.Fodder{},
																															Variable: "hex2",
																															EqFodder: ast.Fodder{},
																															Body: &ast.Binary{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(532),
																																			Column: int(20),
																																		},
																																		End: ast.Location{
																																			Line: int(533),
																																			Column: int(45),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p5693,
																																	FreeVars: ast.Identifiers{
																																		"add_zerox",
																																		"capitals",
																																		"hex",
																																		"pad_left",
																																		"zp2",
																																	},
																																},
																																Left: &ast.Conditional{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(532),
																																				Column: int(21),
																																			},
																																			End: ast.Location{
																																				Line: int(532),
																																				Column: int(80),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p5693,
																																		FreeVars: ast.Identifiers{
																																			"add_zerox",
																																			"capitals",
																																		},
																																	},
																																	Cond: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(532),
																																					Column: int(24),
																																				},
																																				End: ast.Location{
																																					Line: int(532),
																																					Column: int(33),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p5693,
																																			FreeVars: ast.Identifiers{
																																				"add_zerox",
																																			},
																																		},
																																		Id: "add_zerox",
																																	},
																																	ThenFodder: ast.Fodder{},
																																	BranchTrue: &ast.Conditional{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(532),
																																					Column: int(40),
																																				},
																																				End: ast.Location{
																																					Line: int(532),
																																					Column: int(71),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p5693,
																																			FreeVars: ast.Identifiers{
																																				"capitals",
																																			},
																																		},
																																		Cond: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(532),
																																						Column: int(43),
																																					},
																																					End: ast.Location{
																																						Line: int(532),
																																						Column: int(51),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p5693,
																																				FreeVars: ast.Identifiers{
																																					"capitals",
																																				},
																																			},
																																			Id: "capitals",
																																		},
																																		ThenFodder: ast.Fodder{},
																																		BranchTrue: &ast.LiteralString{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(532),
																																						Column: int(57),
																																					},
																																					End: ast.Location{
																																						Line: int(532),
																																						Column: int(61),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p5693,
																																				FreeVars: nil,
																																			},
																																			Value: "0X",
																																			Kind: ast.LiteralStringKind(1),
																																			BlockIndent: "",
																																			BlockTermIndent: "",
																																		},
																																		ElseFodder: ast.Fodder{},
																																		BranchFalse: &ast.LiteralString{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(532),
																																						Column: int(67),
																																					},
																																					End: ast.Location{
																																						Line: int(532),
																																						Column: int(71),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p5693,
																																				FreeVars: nil,
																																			},
																																			Value: "0x",
																																			Kind: ast.LiteralStringKind(1),
																																			BlockIndent: "",
																																			BlockTermIndent: "",
																																		},
																																	},
																																	ElseFodder: ast.Fodder{},
																																	BranchFalse: &ast.LiteralString{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(532),
																																					Column: int(78),
																																				},
																																				End: ast.Location{
																																					Line: int(532),
																																					Column: int(80),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p5693,
																																			FreeVars: nil,
																																		},
																																		Value: "",
																																		Kind: ast.LiteralStringKind(1),
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																	},
																																},
																																OpFodder: ast.Fodder{
																																	ast.FodderElement{
																																		Kind: ast.FodderKind(0),
																																		Blanks: int(0),
																																		Indent: int(19),
																																		Comment: []string{},
																																	},
																																},
																																Op: ast.BinaryOp(3),
																																Right: &ast.Apply{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(533),
																																				Column: int(22),
																																			},
																																			End: ast.Location{
																																				Line: int(533),
																																				Column: int(45),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p5693,
																																		FreeVars: ast.Identifiers{
																																			"hex",
																																			"pad_left",
																																			"zp2",
																																		},
																																	},
																																	Target: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(533),
																																					Column: int(22),
																																				},
																																				End: ast.Location{
																																					Line: int(533),
																																					Column: int(30),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p5693,
																																			FreeVars: ast.Identifiers{
																																				"pad_left",
																																			},
																																		},
																																		Id: "pad_left",
																																	},
																																	FodderLeft: ast.Fodder{},
																																	Arguments: ast.Arguments{
																																		Positional: []ast.CommaSeparatedExpr{
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Var{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(533),
																																								Column: int(31),
																																							},
																																							End: ast.Location{
																																								Line: int(533),
																																								Column: int(34),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p5713,
																																						FreeVars: ast.Identifiers{
																																							"hex",
																																						},
																																					},
																																					Id: "hex",
																																				},
																																				CommaFodder: ast.Fodder{},
																																			},
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Var{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(533),
																																								Column: int(36),
																																							},
																																							End: ast.Location{
																																								Line: int(533),
																																								Column: int(39),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p5713,
																																						FreeVars: ast.Identifiers{
																																							"zp2",
																																						},
																																					},
																																					Id: "zp2",
																																				},
																																				CommaFodder: ast.Fodder{},
																																			},
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.LiteralString{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(533),
																																								Column: int(41),
																																							},
																																							End: ast.Location{
																																								Line: int(533),
																																								Column: int(44),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p5713,
																																						FreeVars: nil,
																																					},
																																					Value: "0",
																																					Kind: ast.LiteralStringKind(1),
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																				},
																																				CommaFodder: nil,
																																			},
																																		},
																																		Named: nil,
																																	},
																																	TrailingComma: false,
																																	TailStrict: false,
																																	FodderRight: ast.Fodder{},
																																	TailStrictFodder: nil,
																																},
																															},
																															Fun: nil,
																															CloseFodder: ast.Fodder{},
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(532),
																																	Column: int(13),
																																},
																																End: ast.Location{
																																	Line: int(533),
																																	Column: int(45),
																																},
																																File: p1,
																															},
																														},
																													},
																													Body: &ast.Binary{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(534),
																																	Column: int(7),
																																},
																																End: ast.Location{
																																	Line: int(534),
																																	Column: int(84),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5472,
																															FreeVars: ast.Identifiers{
																																"blank",
																																"hex2",
																																"neg",
																																"plus",
																															},
																														},
																														Left: &ast.Conditional{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(534),
																																		Column: int(8),
																																	},
																																	End: ast.Location{
																																		Line: int(534),
																																		Column: int(76),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p5472,
																																FreeVars: ast.Identifiers{
																																	"blank",
																																	"neg",
																																	"plus",
																																},
																															},
																															Cond: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(534),
																																			Column: int(11),
																																		},
																																		End: ast.Location{
																																			Line: int(534),
																																			Column: int(14),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p5472,
																																	FreeVars: ast.Identifiers{
																																		"neg",
																																	},
																																},
																																Id: "neg",
																															},
																															ThenFodder: ast.Fodder{},
																															BranchTrue: &ast.LiteralString{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(534),
																																			Column: int(20),
																																		},
																																		End: ast.Location{
																																			Line: int(534),
																																			Column: int(23),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p5472,
																																	FreeVars: nil,
																																},
																																Value: "-",
																																Kind: ast.LiteralStringKind(1),
																																BlockIndent: "",
																																BlockTermIndent: "",
																															},
																															ElseFodder: ast.Fodder{},
																															BranchFalse: &ast.Conditional{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(534),
																																			Column: int(29),
																																		},
																																		End: ast.Location{
																																			Line: int(534),
																																			Column: int(76),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p5472,
																																	FreeVars: ast.Identifiers{
																																		"blank",
																																		"plus",
																																	},
																																},
																																Cond: &ast.Var{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(534),
																																				Column: int(32),
																																			},
																																			End: ast.Location{
																																				Line: int(534),
																																				Column: int(36),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p5472,
																																		FreeVars: ast.Identifiers{
																																			"plus",
																																		},
																																	},
																																	Id: "plus",
																																},
																																ThenFodder: ast.Fodder{},
																																BranchTrue: &ast.LiteralString{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(534),
																																				Column: int(42),
																																			},
																																			End: ast.Location{
																																				Line: int(534),
																																				Column: int(45),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p5472,
																																		FreeVars: nil,
																																	},
																																	Value: "+",
																																	Kind: ast.LiteralStringKind(1),
																																	BlockIndent: "",
																																	BlockTermIndent: "",
																																},
																																ElseFodder: ast.Fodder{},
																																BranchFalse: &ast.Conditional{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(534),
																																				Column: int(51),
																																			},
																																			End: ast.Location{
																																				Line: int(534),
																																				Column: int(76),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p5472,
																																		FreeVars: ast.Identifiers{
																																			"blank",
																																		},
																																	},
																																	Cond: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(534),
																																					Column: int(54),
																																				},
																																				End: ast.Location{
																																					Line: int(534),
																																					Column: int(59),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p5472,
																																			FreeVars: ast.Identifiers{
																																				"blank",
																																			},
																																		},
																																		Id: "blank",
																																	},
																																	ThenFodder: ast.Fodder{},
																																	BranchTrue: &ast.LiteralString{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(534),
																																					Column: int(65),
																																				},
																																				End: ast.Location{
																																					Line: int(534),
																																					Column: int(68),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p5472,
																																			FreeVars: nil,
																																		},
																																		Value: " ",
																																		Kind: ast.LiteralStringKind(1),
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																	},
																																	ElseFodder: ast.Fodder{},
																																	BranchFalse: &ast.LiteralString{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(534),
																																					Column: int(74),
																																				},
																																				End: ast.Location{
																																					Line: int(534),
																																					Column: int(76),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p5472,
																																			FreeVars: nil,
																																		},
																																		Value: "",
																																		Kind: ast.LiteralStringKind(1),
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																	},
																																},
																															},
																														},
																														OpFodder: ast.Fodder{},
																														Op: ast.BinaryOp(3),
																														Right: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(534),
																																		Column: int(80),
																																	},
																																	End: ast.Location{
																																		Line: int(534),
																																		Column: int(84),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p5472,
																																FreeVars: ast.Identifiers{
																																	"hex2",
																																},
																															},
																															Id: "hex2",
																														},
																													},
																												},
																											},
																										},
																									},
																								},
																							},
																						},
																					},
																				},
																				Fun: nil,
																				CloseFodder: nil,
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																			},
																		},
																		Body: &ast.Local{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(536),
																						Column: int(5),
																					},
																					End: ast.Location{
																						Line: int(774),
																						Column: int(48),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Kind: ast.FodderKind(0),
																						Blanks: int(1),
																						Indent: int(4),
																						Comment: []string{},
																					},
																				},
																				Ctx: p3497,
																				FreeVars: ast.Identifiers{
																					"codes",
																					"pad_left",
																					"pad_right",
																					"render_hex",
																					"render_int",
																					"std",
																					"vals",
																				},
																			},
																			Binds: ast.LocalBinds{
																				ast.LocalBind{
																					VarFodder: nil,
																					Variable: "strip_trailing_zero",
																					EqFodder: nil,
																					Body: &ast.Function{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(536),
																									Column: int(11),
																								},
																								End: ast.Location{
																									Line: int(545),
																									Column: int(36),
																								},
																								File: p1,
																							},
																							Fodder: nil,
																							Ctx: p5743,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																						},
																						ParenLeftFodder: ast.Fodder{},
																						Parameters: []ast.Parameter{
																							ast.Parameter{
																								NameFodder: ast.Fodder{},
																								Name: "str",
																								EqFodder: nil,
																								DefaultArg: nil,
																								CommaFodder: nil,
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(536),
																										Column: int(31),
																									},
																									End: ast.Location{
																										Line: int(536),
																										Column: int(34),
																									},
																									File: p1,
																								},
																							},
																						},
																						TrailingComma: false,
																						ParenRightFodder: ast.Fodder{},
																						Body: &ast.Local{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(537),
																										Column: int(7),
																									},
																									End: ast.Location{
																										Line: int(545),
																										Column: int(36),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{
																									ast.FodderElement{
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(6),
																										Comment: []string{},
																									},
																								},
																								Ctx: p5748,
																								FreeVars: ast.Identifiers{
																									"std",
																									"str",
																								},
																							},
																							Binds: ast.LocalBinds{
																								ast.LocalBind{
																									VarFodder: nil,
																									Variable: "aux",
																									EqFodder: nil,
																									Body: &ast.Function{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(537),
																													Column: int(13),
																												},
																												End: ast.Location{
																													Line: int(544),
																													Column: int(38),
																												},
																												File: p1,
																											},
																											Fodder: nil,
																											Ctx: p5752,
																											FreeVars: ast.Identifiers{
																												"aux",
																												"std",
																											},
																										},
																										ParenLeftFodder: ast.Fodder{},
																										Parameters: []ast.Parameter{
																											ast.Parameter{
																												NameFodder: ast.Fodder{},
																												Name: "str",
																												EqFodder: nil,
																												DefaultArg: nil,
																												CommaFodder: ast.Fodder{},
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(537),
																														Column: int(17),
																													},
																													End: ast.Location{
																														Line: int(537),
																														Column: int(20),
																													},
																													File: p1,
																												},
																											},
																											ast.Parameter{
																												NameFodder: ast.Fodder{},
																												Name: "i",
																												EqFodder: nil,
																												DefaultArg: nil,
																												CommaFodder: nil,
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(537),
																														Column: int(22),
																													},
																													End: ast.Location{
																														Line: int(537),
																														Column: int(23),
																													},
																													File: p1,
																												},
																											},
																										},
																										TrailingComma: false,
																										ParenRightFodder: ast.Fodder{},
																										Body: &ast.Conditional{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(538),
																														Column: int(9),
																													},
																													End: ast.Location{
																														Line: int(544),
																														Column: int(38),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{
																													ast.FodderElement{
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(8),
																														Comment: []string{},
																													},
																												},
																												Ctx: p5757,
																												FreeVars: ast.Identifiers{
																													"aux",
																													"i",
																													"std",
																													"str",
																												},
																											},
																											Cond: &ast.Binary{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(538),
																															Column: int(12),
																														},
																														End: ast.Location{
																															Line: int(538),
																															Column: int(17),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p5757,
																													FreeVars: ast.Identifiers{
																														"i",
																													},
																												},
																												Left: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(538),
																																Column: int(12),
																															},
																															End: ast.Location{
																																Line: int(538),
																																Column: int(13),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5757,
																														FreeVars: ast.Identifiers{
																															"i",
																														},
																													},
																													Id: "i",
																												},
																												OpFodder: ast.Fodder{},
																												Op: ast.BinaryOp(9),
																												Right: &ast.LiteralNumber{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(538),
																																Column: int(16),
																															},
																															End: ast.Location{
																																Line: int(538),
																																Column: int(17),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5757,
																														FreeVars: nil,
																													},
																													OriginalString: "0",
																												},
																											},
																											ThenFodder: ast.Fodder{},
																											BranchTrue: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(539),
																															Column: int(11),
																														},
																														End: ast.Location{
																															Line: int(539),
																															Column: int(13),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(10),
																															Comment: []string{},
																														},
																													},
																													Ctx: p5757,
																													FreeVars: nil,
																												},
																												Value: "",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																											ElseFodder: ast.Fodder{
																												ast.FodderElement{
																													Kind: ast.FodderKind(0),
																													Blanks: int(0),
																													Indent: int(8),
																													Comment: []string{},
																												},
																											},
																											BranchFalse: &ast.Conditional{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(541),
																															Column: int(11),
																														},
																														End: ast.Location{
																															Line: int(544),
																															Column: int(38),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(10),
																															Comment: []string{},
																														},
																													},
																													Ctx: p5757,
																													FreeVars: ast.Identifiers{
																														"aux",
																														"i",
																														"std",
																														"str",
																													},
																												},
																												Cond: &ast.Binary{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(541),
																																Column: int(14),
																															},
																															End: ast.Location{
																																Line: int(541),
																																Column: int(27),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5757,
																														FreeVars: ast.Identifiers{
																															"i",
																															"str",
																														},
																													},
																													Left: &ast.Index{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(541),
																																	Column: int(14),
																																},
																																End: ast.Location{
																																	Line: int(541),
																																	Column: int(20),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5757,
																															FreeVars: ast.Identifiers{
																																"i",
																																"str",
																															},
																														},
																														Target: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(541),
																																		Column: int(14),
																																	},
																																	End: ast.Location{
																																		Line: int(541),
																																		Column: int(17),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p5757,
																																FreeVars: ast.Identifiers{
																																	"str",
																																},
																															},
																															Id: "str",
																														},
																														LeftBracketFodder: ast.Fodder{},
																														Index: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(541),
																																		Column: int(18),
																																	},
																																	End: ast.Location{
																																		Line: int(541),
																																		Column: int(19),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p5757,
																																FreeVars: ast.Identifiers{
																																	"i",
																																},
																															},
																															Id: "i",
																														},
																														RightBracketFodder: ast.Fodder{},
																														Id: nil,
																													},
																													OpFodder: ast.Fodder{},
																													Op: ast.BinaryOp(12),
																													Right: &ast.LiteralString{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(541),
																																	Column: int(24),
																																},
																																End: ast.Location{
																																	Line: int(541),
																																	Column: int(27),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5757,
																															FreeVars: nil,
																														},
																														Value: "0",
																														Kind: ast.LiteralStringKind(1),
																														BlockIndent: "",
																														BlockTermIndent: "",
																													},
																												},
																												ThenFodder: ast.Fodder{},
																												BranchTrue: &ast.Apply{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(542),
																																Column: int(13),
																															},
																															End: ast.Location{
																																Line: int(542),
																																Column: int(28),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5757,
																														FreeVars: ast.Identifiers{
																															"aux",
																															"i",
																															"str",
																														},
																													},
																													Target: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(542),
																																	Column: int(13),
																																},
																																End: ast.Location{
																																	Line: int(542),
																																	Column: int(16),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{
																																ast.FodderElement{
																																	Kind: ast.FodderKind(0),
																																	Blanks: int(0),
																																	Indent: int(12),
																																	Comment: []string{},
																																},
																															},
																															Ctx: p5757,
																															FreeVars: ast.Identifiers{
																																"aux",
																															},
																														},
																														Id: "aux",
																													},
																													FodderLeft: ast.Fodder{},
																													Arguments: ast.Arguments{
																														Positional: []ast.CommaSeparatedExpr{
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Var{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(542),
																																				Column: int(17),
																																			},
																																			End: ast.Location{
																																				Line: int(542),
																																				Column: int(20),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p5786,
																																		FreeVars: ast.Identifiers{
																																			"str",
																																		},
																																	},
																																	Id: "str",
																																},
																																CommaFodder: ast.Fodder{},
																															},
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Binary{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(542),
																																				Column: int(22),
																																			},
																																			End: ast.Location{
																																				Line: int(542),
																																				Column: int(27),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p5786,
																																		FreeVars: ast.Identifiers{
																																			"i",
																																		},
																																	},
																																	Left: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(542),
																																					Column: int(22),
																																				},
																																				End: ast.Location{
																																					Line: int(542),
																																					Column: int(23),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p5786,
																																			FreeVars: ast.Identifiers{
																																				"i",
																																			},
																																		},
																																		Id: "i",
																																	},
																																	OpFodder: ast.Fodder{},
																																	Op: ast.BinaryOp(4),
																																	Right: &ast.LiteralNumber{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(542),
																																					Column: int(26),
																																				},
																																				End: ast.Location{
																																					Line: int(542),
																																					Column: int(27),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p5786,
																																			FreeVars: nil,
																																		},
																																		OriginalString: "1",
																																	},
																																},
																																CommaFodder: nil,
																															},
																														},
																														Named: nil,
																													},
																													TrailingComma: false,
																													TailStrict: false,
																													FodderRight: ast.Fodder{},
																													TailStrictFodder: nil,
																												},
																												ElseFodder: ast.Fodder{
																													ast.FodderElement{
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(10),
																														Comment: []string{},
																													},
																												},
																												BranchFalse: &ast.Apply{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(544),
																																Column: int(13),
																															},
																															End: ast.Location{
																																Line: int(544),
																																Column: int(38),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5757,
																														FreeVars: ast.Identifiers{
																															"i",
																															"std",
																															"str",
																														},
																													},
																													Target: &ast.Index{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(544),
																																	Column: int(13),
																																},
																																End: ast.Location{
																																	Line: int(544),
																																	Column: int(23),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5757,
																															FreeVars: ast.Identifiers{
																																"std",
																															},
																														},
																														Target: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(544),
																																		Column: int(13),
																																	},
																																	End: ast.Location{
																																		Line: int(544),
																																		Column: int(16),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{
																																	ast.FodderElement{
																																		Kind: ast.FodderKind(0),
																																		Blanks: int(0),
																																		Indent: int(12),
																																		Comment: []string{},
																																	},
																																},
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																															},
																															Id: "std",
																														},
																														LeftBracketFodder: ast.Fodder{},
																														Index: &ast.LiteralString{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	File: nil,
																																},
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: nil,
																															},
																															Value: "substr",
																															Kind: ast.LiteralStringKind(1),
																															BlockIndent: "",
																															BlockTermIndent: "",
																														},
																														RightBracketFodder: ast.Fodder{},
																														Id: nil,
																													},
																													FodderLeft: ast.Fodder{},
																													Arguments: ast.Arguments{
																														Positional: []ast.CommaSeparatedExpr{
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Var{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(544),
																																				Column: int(24),
																																			},
																																			End: ast.Location{
																																				Line: int(544),
																																				Column: int(27),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p5804,
																																		FreeVars: ast.Identifiers{
																																			"str",
																																		},
																																	},
																																	Id: "str",
																																},
																																CommaFodder: ast.Fodder{},
																															},
																															ast.CommaSeparatedExpr{
																																Expr: &ast.LiteralNumber{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(544),
																																				Column: int(29),
																																			},
																																			End: ast.Location{
																																				Line: int(544),
																																				Column: int(30),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p5804,
																																		FreeVars: nil,
																																	},
																																	OriginalString: "0",
																																},
																																CommaFodder: ast.Fodder{},
																															},
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Binary{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(544),
																																				Column: int(32),
																																			},
																																			End: ast.Location{
																																				Line: int(544),
																																				Column: int(37),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p5804,
																																		FreeVars: ast.Identifiers{
																																			"i",
																																		},
																																	},
																																	Left: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(544),
																																					Column: int(32),
																																				},
																																				End: ast.Location{
																																					Line: int(544),
																																					Column: int(33),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p5804,
																																			FreeVars: ast.Identifiers{
																																				"i",
																																			},
																																		},
																																		Id: "i",
																																	},
																																	OpFodder: ast.Fodder{},
																																	Op: ast.BinaryOp(3),
																																	Right: &ast.LiteralNumber{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(544),
																																					Column: int(36),
																																				},
																																				End: ast.Location{
																																					Line: int(544),
																																					Column: int(37),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p5804,
																																			FreeVars: nil,
																																		},
																																		OriginalString: "1",
																																	},
																																},
																																CommaFodder: nil,
																															},
																														},
																														Named: nil,
																													},
																													TrailingComma: false,
																													TailStrict: false,
																													FodderRight: ast.Fodder{},
																													TailStrictFodder: nil,
																												},
																											},
																										},
																									},
																									Fun: nil,
																									CloseFodder: nil,
																									LocRange: ast.LocationRange{
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										File: nil,
																									},
																								},
																							},
																							Body: &ast.Apply{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(545),
																											Column: int(7),
																										},
																										End: ast.Location{
																											Line: int(545),
																											Column: int(36),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p5748,
																									FreeVars: ast.Identifiers{
																										"aux",
																										"std",
																										"str",
																									},
																								},
																								Target: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(545),
																												Column: int(7),
																											},
																											End: ast.Location{
																												Line: int(545),
																												Column: int(10),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{
																											ast.FodderElement{
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(6),
																												Comment: []string{},
																											},
																										},
																										Ctx: p5748,
																										FreeVars: ast.Identifiers{
																											"aux",
																										},
																									},
																									Id: "aux",
																								},
																								FodderLeft: ast.Fodder{},
																								Arguments: ast.Arguments{
																									Positional: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(545),
																															Column: int(11),
																														},
																														End: ast.Location{
																															Line: int(545),
																															Column: int(14),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p5819,
																													FreeVars: ast.Identifiers{
																														"str",
																													},
																												},
																												Id: "str",
																											},
																											CommaFodder: ast.Fodder{},
																										},
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Binary{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(545),
																															Column: int(16),
																														},
																														End: ast.Location{
																															Line: int(545),
																															Column: int(35),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p5819,
																													FreeVars: ast.Identifiers{
																														"std",
																														"str",
																													},
																												},
																												Left: &ast.Apply{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(545),
																																Column: int(16),
																															},
																															End: ast.Location{
																																Line: int(545),
																																Column: int(31),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5819,
																														FreeVars: ast.Identifiers{
																															"std",
																															"str",
																														},
																													},
																													Target: &ast.Index{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(545),
																																	Column: int(16),
																																},
																																End: ast.Location{
																																	Line: int(545),
																																	Column: int(26),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5819,
																															FreeVars: ast.Identifiers{
																																"std",
																															},
																														},
																														Target: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(545),
																																		Column: int(16),
																																	},
																																	End: ast.Location{
																																		Line: int(545),
																																		Column: int(19),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																															},
																															Id: "std",
																														},
																														LeftBracketFodder: ast.Fodder{},
																														Index: &ast.LiteralString{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	File: nil,
																																},
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: nil,
																															},
																															Value: "length",
																															Kind: ast.LiteralStringKind(1),
																															BlockIndent: "",
																															BlockTermIndent: "",
																														},
																														RightBracketFodder: ast.Fodder{},
																														Id: nil,
																													},
																													FodderLeft: ast.Fodder{},
																													Arguments: ast.Arguments{
																														Positional: []ast.CommaSeparatedExpr{
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Var{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(545),
																																				Column: int(27),
																																			},
																																			End: ast.Location{
																																				Line: int(545),
																																				Column: int(30),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p5832,
																																		FreeVars: ast.Identifiers{
																																			"str",
																																		},
																																	},
																																	Id: "str",
																																},
																																CommaFodder: nil,
																															},
																														},
																														Named: nil,
																													},
																													TrailingComma: false,
																													TailStrict: false,
																													FodderRight: ast.Fodder{},
																													TailStrictFodder: nil,
																												},
																												OpFodder: ast.Fodder{},
																												Op: ast.BinaryOp(4),
																												Right: &ast.LiteralNumber{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(545),
																																Column: int(34),
																															},
																															End: ast.Location{
																																Line: int(545),
																																Column: int(35),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5819,
																														FreeVars: nil,
																													},
																													OriginalString: "1",
																												},
																											},
																											CommaFodder: nil,
																										},
																									},
																									Named: nil,
																								},
																								TrailingComma: false,
																								TailStrict: false,
																								FodderRight: ast.Fodder{},
																								TailStrictFodder: nil,
																							},
																						},
																					},
																					Fun: nil,
																					CloseFodder: nil,
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																				},
																			},
																			Body: &ast.Local{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(548),
																							Column: int(5),
																						},
																						End: ast.Location{
																							Line: int(774),
																							Column: int(48),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Kind: ast.FodderKind(0),
																							Blanks: int(1),
																							Indent: int(4),
																							Comment: []string{},
																						},
																						ast.FodderElement{
																							Kind: ast.FodderKind(2),
																							Blanks: int(0),
																							Indent: int(4),
																							Comment: []string{
																								"// Render floating point in decimal form",
																							},
																						},
																					},
																					Ctx: p3497,
																					FreeVars: ast.Identifiers{
																						"codes",
																						"pad_left",
																						"pad_right",
																						"render_hex",
																						"render_int",
																						"std",
																						"strip_trailing_zero",
																						"vals",
																					},
																				},
																				Binds: ast.LocalBinds{
																					ast.LocalBind{
																						VarFodder: nil,
																						Variable: "render_float_dec",
																						EqFodder: nil,
																						Body: &ast.Function{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(548),
																										Column: int(11),
																									},
																									End: ast.Location{
																										Line: int(562),
																										Column: int(14),
																									},
																									File: p1,
																								},
																								Fodder: nil,
																								Ctx: p5841,
																								FreeVars: ast.Identifiers{
																									"render_int",
																									"std",
																									"strip_trailing_zero",
																								},
																							},
																							ParenLeftFodder: ast.Fodder{},
																							Parameters: []ast.Parameter{
																								ast.Parameter{
																									NameFodder: ast.Fodder{},
																									Name: "n__",
																									EqFodder: nil,
																									DefaultArg: nil,
																									CommaFodder: ast.Fodder{},
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(548),
																											Column: int(28),
																										},
																										End: ast.Location{
																											Line: int(548),
																											Column: int(31),
																										},
																										File: p1,
																									},
																								},
																								ast.Parameter{
																									NameFodder: ast.Fodder{},
																									Name: "zero_pad",
																									EqFodder: nil,
																									DefaultArg: nil,
																									CommaFodder: ast.Fodder{},
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(548),
																											Column: int(33),
																										},
																										End: ast.Location{
																											Line: int(548),
																											Column: int(41),
																										},
																										File: p1,
																									},
																								},
																								ast.Parameter{
																									NameFodder: ast.Fodder{},
																									Name: "blank",
																									EqFodder: nil,
																									DefaultArg: nil,
																									CommaFodder: ast.Fodder{},
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(548),
																											Column: int(43),
																										},
																										End: ast.Location{
																											Line: int(548),
																											Column: int(48),
																										},
																										File: p1,
																									},
																								},
																								ast.Parameter{
																									NameFodder: ast.Fodder{},
																									Name: "plus",
																									EqFodder: nil,
																									DefaultArg: nil,
																									CommaFodder: ast.Fodder{},
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(548),
																											Column: int(50),
																										},
																										End: ast.Location{
																											Line: int(548),
																											Column: int(54),
																										},
																										File: p1,
																									},
																								},
																								ast.Parameter{
																									NameFodder: ast.Fodder{},
																									Name: "ensure_pt",
																									EqFodder: nil,
																									DefaultArg: nil,
																									CommaFodder: ast.Fodder{},
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(548),
																											Column: int(56),
																										},
																										End: ast.Location{
																											Line: int(548),
																											Column: int(65),
																										},
																										File: p1,
																									},
																								},
																								ast.Parameter{
																									NameFodder: ast.Fodder{},
																									Name: "trailing",
																									EqFodder: nil,
																									DefaultArg: nil,
																									CommaFodder: ast.Fodder{},
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(548),
																											Column: int(67),
																										},
																										End: ast.Location{
																											Line: int(548),
																											Column: int(75),
																										},
																										File: p1,
																									},
																								},
																								ast.Parameter{
																									NameFodder: ast.Fodder{},
																									Name: "prec",
																									EqFodder: nil,
																									DefaultArg: nil,
																									CommaFodder: nil,
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(548),
																											Column: int(77),
																										},
																										End: ast.Location{
																											Line: int(548),
																											Column: int(81),
																										},
																										File: p1,
																									},
																								},
																							},
																							TrailingComma: false,
																							ParenRightFodder: ast.Fodder{},
																							Body: &ast.Local{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(549),
																											Column: int(7),
																										},
																										End: ast.Location{
																											Line: int(562),
																											Column: int(14),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{
																										ast.FodderElement{
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(6),
																											Comment: []string{},
																										},
																									},
																									Ctx: p5846,
																									FreeVars: ast.Identifiers{
																										"blank",
																										"ensure_pt",
																										"n__",
																										"plus",
																										"prec",
																										"render_int",
																										"std",
																										"strip_trailing_zero",
																										"trailing",
																										"zero_pad",
																									},
																								},
																								Binds: ast.LocalBinds{
																									ast.LocalBind{
																										VarFodder: ast.Fodder{},
																										Variable: "n_",
																										EqFodder: ast.Fodder{},
																										Body: &ast.Apply{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(549),
																														Column: int(18),
																													},
																													End: ast.Location{
																														Line: int(549),
																														Column: int(30),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p5850,
																												FreeVars: ast.Identifiers{
																													"n__",
																													"std",
																												},
																											},
																											Target: &ast.Index{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(549),
																															Column: int(18),
																														},
																														End: ast.Location{
																															Line: int(549),
																															Column: int(25),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p5850,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																												},
																												Target: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(549),
																																Column: int(18),
																															},
																															End: ast.Location{
																																Line: int(549),
																																Column: int(21),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"std",
																														},
																													},
																													Id: "std",
																												},
																												LeftBracketFodder: ast.Fodder{},
																												Index: &ast.LiteralString{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															File: nil,
																														},
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: nil,
																													},
																													Value: "abs",
																													Kind: ast.LiteralStringKind(1),
																													BlockIndent: "",
																													BlockTermIndent: "",
																												},
																												RightBracketFodder: ast.Fodder{},
																												Id: nil,
																											},
																											FodderLeft: ast.Fodder{},
																											Arguments: ast.Arguments{
																												Positional: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(549),
																																		Column: int(26),
																																	},
																																	End: ast.Location{
																																		Line: int(549),
																																		Column: int(29),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p5859,
																																FreeVars: ast.Identifiers{
																																	"n__",
																																},
																															},
																															Id: "n__",
																														},
																														CommaFodder: nil,
																													},
																												},
																												Named: nil,
																											},
																											TrailingComma: false,
																											TailStrict: false,
																											FodderRight: ast.Fodder{},
																											TailStrictFodder: nil,
																										},
																										Fun: nil,
																										CloseFodder: ast.Fodder{},
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(549),
																												Column: int(13),
																											},
																											End: ast.Location{
																												Line: int(549),
																												Column: int(30),
																											},
																											File: p1,
																										},
																									},
																								},
																								Body: &ast.Local{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(550),
																												Column: int(7),
																											},
																											End: ast.Location{
																												Line: int(562),
																												Column: int(14),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{
																											ast.FodderElement{
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(6),
																												Comment: []string{},
																											},
																										},
																										Ctx: p5846,
																										FreeVars: ast.Identifiers{
																											"blank",
																											"ensure_pt",
																											"n_",
																											"n__",
																											"plus",
																											"prec",
																											"render_int",
																											"std",
																											"strip_trailing_zero",
																											"trailing",
																											"zero_pad",
																										},
																									},
																									Binds: ast.LocalBinds{
																										ast.LocalBind{
																											VarFodder: ast.Fodder{},
																											Variable: "whole",
																											EqFodder: ast.Fodder{},
																											Body: &ast.Apply{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(550),
																															Column: int(21),
																														},
																														End: ast.Location{
																															Line: int(550),
																															Column: int(34),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p5866,
																													FreeVars: ast.Identifiers{
																														"n_",
																														"std",
																													},
																												},
																												Target: &ast.Index{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(550),
																																Column: int(21),
																															},
																															End: ast.Location{
																																Line: int(550),
																																Column: int(30),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5866,
																														FreeVars: ast.Identifiers{
																															"std",
																														},
																													},
																													Target: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(550),
																																	Column: int(21),
																																},
																																End: ast.Location{
																																	Line: int(550),
																																	Column: int(24),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: nil,
																															FreeVars: ast.Identifiers{
																																"std",
																															},
																														},
																														Id: "std",
																													},
																													LeftBracketFodder: ast.Fodder{},
																													Index: &ast.LiteralString{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																End: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																File: nil,
																															},
																															Fodder: nil,
																															Ctx: nil,
																															FreeVars: nil,
																														},
																														Value: "floor",
																														Kind: ast.LiteralStringKind(1),
																														BlockIndent: "",
																														BlockTermIndent: "",
																													},
																													RightBracketFodder: ast.Fodder{},
																													Id: nil,
																												},
																												FodderLeft: ast.Fodder{},
																												Arguments: ast.Arguments{
																													Positional: []ast.CommaSeparatedExpr{
																														ast.CommaSeparatedExpr{
																															Expr: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(550),
																																			Column: int(31),
																																		},
																																		End: ast.Location{
																																			Line: int(550),
																																			Column: int(33),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p5875,
																																	FreeVars: ast.Identifiers{
																																		"n_",
																																	},
																																},
																																Id: "n_",
																															},
																															CommaFodder: nil,
																														},
																													},
																													Named: nil,
																												},
																												TrailingComma: false,
																												TailStrict: false,
																												FodderRight: ast.Fodder{},
																												TailStrictFodder: nil,
																											},
																											Fun: nil,
																											CloseFodder: ast.Fodder{},
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(550),
																													Column: int(13),
																												},
																												End: ast.Location{
																													Line: int(550),
																													Column: int(34),
																												},
																												File: p1,
																											},
																										},
																									},
																									Body: &ast.Local{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(551),
																													Column: int(7),
																												},
																												End: ast.Location{
																													Line: int(562),
																													Column: int(14),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{
																												ast.FodderElement{
																													Kind: ast.FodderKind(0),
																													Blanks: int(0),
																													Indent: int(6),
																													Comment: []string{},
																												},
																											},
																											Ctx: p5846,
																											FreeVars: ast.Identifiers{
																												"blank",
																												"ensure_pt",
																												"n_",
																												"n__",
																												"plus",
																												"prec",
																												"render_int",
																												"std",
																												"strip_trailing_zero",
																												"trailing",
																												"whole",
																												"zero_pad",
																											},
																										},
																										Binds: ast.LocalBinds{
																											ast.LocalBind{
																												VarFodder: ast.Fodder{},
																												Variable: "dot_size",
																												EqFodder: ast.Fodder{},
																												Body: &ast.Conditional{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(551),
																																Column: int(24),
																															},
																															End: ast.Location{
																																Line: int(551),
																																Column: int(64),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p5882,
																														FreeVars: ast.Identifiers{
																															"ensure_pt",
																															"prec",
																														},
																													},
																													Cond: &ast.Binary{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(551),
																																	Column: int(27),
																																},
																																End: ast.Location{
																																	Line: int(551),
																																	Column: int(50),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5882,
																															FreeVars: ast.Identifiers{
																																"ensure_pt",
																																"prec",
																															},
																														},
																														Left: &ast.Binary{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(551),
																																		Column: int(27),
																																	},
																																	End: ast.Location{
																																		Line: int(551),
																																		Column: int(36),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p5882,
																																FreeVars: ast.Identifiers{
																																	"prec",
																																},
																															},
																															Left: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(551),
																																			Column: int(27),
																																		},
																																		End: ast.Location{
																																			Line: int(551),
																																			Column: int(31),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p5882,
																																	FreeVars: ast.Identifiers{
																																		"prec",
																																	},
																																},
																																Id: "prec",
																															},
																															OpFodder: ast.Fodder{},
																															Op: ast.BinaryOp(12),
																															Right: &ast.LiteralNumber{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(551),
																																			Column: int(35),
																																		},
																																		End: ast.Location{
																																			Line: int(551),
																																			Column: int(36),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p5882,
																																	FreeVars: nil,
																																},
																																OriginalString: "0",
																															},
																														},
																														OpFodder: ast.Fodder{},
																														Op: ast.BinaryOp(17),
																														Right: &ast.Unary{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(551),
																																		Column: int(40),
																																	},
																																	End: ast.Location{
																																		Line: int(551),
																																		Column: int(50),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p5882,
																																FreeVars: ast.Identifiers{
																																	"ensure_pt",
																																},
																															},
																															Op: ast.UnaryOp(0),
																															Expr: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(551),
																																			Column: int(41),
																																		},
																																		End: ast.Location{
																																			Line: int(551),
																																			Column: int(50),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p5882,
																																	FreeVars: ast.Identifiers{
																																		"ensure_pt",
																																	},
																																},
																																Id: "ensure_pt",
																															},
																														},
																													},
																													ThenFodder: ast.Fodder{},
																													BranchTrue: &ast.LiteralNumber{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(551),
																																	Column: int(56),
																																},
																																End: ast.Location{
																																	Line: int(551),
																																	Column: int(57),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5882,
																															FreeVars: nil,
																														},
																														OriginalString: "0",
																													},
																													ElseFodder: ast.Fodder{},
																													BranchFalse: &ast.LiteralNumber{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(551),
																																	Column: int(63),
																																},
																																End: ast.Location{
																																	Line: int(551),
																																	Column: int(64),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5882,
																															FreeVars: nil,
																														},
																														OriginalString: "1",
																													},
																												},
																												Fun: nil,
																												CloseFodder: ast.Fodder{},
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(551),
																														Column: int(13),
																													},
																													End: ast.Location{
																														Line: int(551),
																														Column: int(64),
																													},
																													File: p1,
																												},
																											},
																										},
																										Body: &ast.Local{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(552),
																														Column: int(7),
																													},
																													End: ast.Location{
																														Line: int(562),
																														Column: int(14),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{
																													ast.FodderElement{
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(6),
																														Comment: []string{},
																													},
																												},
																												Ctx: p5846,
																												FreeVars: ast.Identifiers{
																													"blank",
																													"dot_size",
																													"ensure_pt",
																													"n_",
																													"n__",
																													"plus",
																													"prec",
																													"render_int",
																													"std",
																													"strip_trailing_zero",
																													"trailing",
																													"whole",
																													"zero_pad",
																												},
																											},
																											Binds: ast.LocalBinds{
																												ast.LocalBind{
																													VarFodder: ast.Fodder{},
																													Variable: "zp",
																													EqFodder: ast.Fodder{},
																													Body: &ast.Binary{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(552),
																																	Column: int(18),
																																},
																																End: ast.Location{
																																	Line: int(552),
																																	Column: int(44),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5902,
																															FreeVars: ast.Identifiers{
																																"dot_size",
																																"prec",
																																"zero_pad",
																															},
																														},
																														Left: &ast.Binary{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(552),
																																		Column: int(18),
																																	},
																																	End: ast.Location{
																																		Line: int(552),
																																		Column: int(33),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p5902,
																																FreeVars: ast.Identifiers{
																																	"prec",
																																	"zero_pad",
																																},
																															},
																															Left: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(552),
																																			Column: int(18),
																																		},
																																		End: ast.Location{
																																			Line: int(552),
																																			Column: int(26),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p5902,
																																	FreeVars: ast.Identifiers{
																																		"zero_pad",
																																	},
																																},
																																Id: "zero_pad",
																															},
																															OpFodder: ast.Fodder{},
																															Op: ast.BinaryOp(4),
																															Right: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(552),
																																			Column: int(29),
																																		},
																																		End: ast.Location{
																																			Line: int(552),
																																			Column: int(33),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p5902,
																																	FreeVars: ast.Identifiers{
																																		"prec",
																																	},
																																},
																																Id: "prec",
																															},
																														},
																														OpFodder: ast.Fodder{},
																														Op: ast.BinaryOp(4),
																														Right: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(552),
																																		Column: int(36),
																																	},
																																	End: ast.Location{
																																		Line: int(552),
																																		Column: int(44),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p5902,
																																FreeVars: ast.Identifiers{
																																	"dot_size",
																																},
																															},
																															Id: "dot_size",
																														},
																													},
																													Fun: nil,
																													CloseFodder: ast.Fodder{},
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(552),
																															Column: int(13),
																														},
																														End: ast.Location{
																															Line: int(552),
																															Column: int(44),
																														},
																														File: p1,
																													},
																												},
																											},
																											Body: &ast.Local{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(553),
																															Column: int(7),
																														},
																														End: ast.Location{
																															Line: int(562),
																															Column: int(14),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(6),
																															Comment: []string{},
																														},
																													},
																													Ctx: p5846,
																													FreeVars: ast.Identifiers{
																														"blank",
																														"ensure_pt",
																														"n_",
																														"n__",
																														"plus",
																														"prec",
																														"render_int",
																														"std",
																														"strip_trailing_zero",
																														"trailing",
																														"whole",
																														"zp",
																													},
																												},
																												Binds: ast.LocalBinds{
																													ast.LocalBind{
																														VarFodder: ast.Fodder{},
																														Variable: "str",
																														EqFodder: ast.Fodder{},
																														Body: &ast.Apply{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(553),
																																		Column: int(19),
																																	},
																																	End: ast.Location{
																																		Line: int(553),
																																		Column: int(73),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p5917,
																																FreeVars: ast.Identifiers{
																																	"blank",
																																	"n__",
																																	"plus",
																																	"render_int",
																																	"whole",
																																	"zp",
																																},
																															},
																															Target: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(553),
																																			Column: int(19),
																																		},
																																		End: ast.Location{
																																			Line: int(553),
																																			Column: int(29),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p5917,
																																	FreeVars: ast.Identifiers{
																																		"render_int",
																																	},
																																},
																																Id: "render_int",
																															},
																															FodderLeft: ast.Fodder{},
																															Arguments: ast.Arguments{
																																Positional: []ast.CommaSeparatedExpr{
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Binary{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(553),
																																						Column: int(30),
																																					},
																																					End: ast.Location{
																																						Line: int(553),
																																						Column: int(37),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p5923,
																																				FreeVars: ast.Identifiers{
																																					"n__",
																																				},
																																			},
																																			Left: &ast.Var{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(553),
																																							Column: int(30),
																																						},
																																						End: ast.Location{
																																							Line: int(553),
																																							Column: int(33),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p5923,
																																					FreeVars: ast.Identifiers{
																																						"n__",
																																					},
																																				},
																																				Id: "n__",
																																			},
																																			OpFodder: ast.Fodder{},
																																			Op: ast.BinaryOp(9),
																																			Right: &ast.LiteralNumber{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(553),
																																							Column: int(36),
																																						},
																																						End: ast.Location{
																																							Line: int(553),
																																							Column: int(37),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p5923,
																																					FreeVars: nil,
																																				},
																																				OriginalString: "0",
																																			},
																																		},
																																		CommaFodder: ast.Fodder{},
																																	},
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(553),
																																						Column: int(39),
																																					},
																																					End: ast.Location{
																																						Line: int(553),
																																						Column: int(44),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p5923,
																																				FreeVars: ast.Identifiers{
																																					"whole",
																																				},
																																			},
																																			Id: "whole",
																																		},
																																		CommaFodder: ast.Fodder{},
																																	},
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(553),
																																						Column: int(46),
																																					},
																																					End: ast.Location{
																																						Line: int(553),
																																						Column: int(48),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p5923,
																																				FreeVars: ast.Identifiers{
																																					"zp",
																																				},
																																			},
																																			Id: "zp",
																																		},
																																		CommaFodder: ast.Fodder{},
																																	},
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.LiteralNumber{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(553),
																																						Column: int(50),
																																					},
																																					End: ast.Location{
																																						Line: int(553),
																																						Column: int(51),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p5923,
																																				FreeVars: nil,
																																			},
																																			OriginalString: "0",
																																		},
																																		CommaFodder: ast.Fodder{},
																																	},
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(553),
																																						Column: int(53),
																																					},
																																					End: ast.Location{
																																						Line: int(553),
																																						Column: int(58),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p5923,
																																				FreeVars: ast.Identifiers{
																																					"blank",
																																				},
																																			},
																																			Id: "blank",
																																		},
																																		CommaFodder: ast.Fodder{},
																																	},
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(553),
																																						Column: int(60),
																																					},
																																					End: ast.Location{
																																						Line: int(553),
																																						Column: int(64),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p5923,
																																				FreeVars: ast.Identifiers{
																																					"plus",
																																				},
																																			},
																																			Id: "plus",
																																		},
																																		CommaFodder: ast.Fodder{},
																																	},
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.LiteralNumber{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(553),
																																						Column: int(66),
																																					},
																																					End: ast.Location{
																																						Line: int(553),
																																						Column: int(68),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p5923,
																																				FreeVars: nil,
																																			},
																																			OriginalString: "10",
																																		},
																																		CommaFodder: ast.Fodder{},
																																	},
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.LiteralString{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(553),
																																						Column: int(70),
																																					},
																																					End: ast.Location{
																																						Line: int(553),
																																						Column: int(72),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p5923,
																																				FreeVars: nil,
																																			},
																																			Value: "",
																																			Kind: ast.LiteralStringKind(1),
																																			BlockIndent: "",
																																			BlockTermIndent: "",
																																		},
																																		CommaFodder: nil,
																																	},
																																},
																																Named: nil,
																															},
																															TrailingComma: false,
																															TailStrict: false,
																															FodderRight: ast.Fodder{},
																															TailStrictFodder: nil,
																														},
																														Fun: nil,
																														CloseFodder: ast.Fodder{},
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(553),
																																Column: int(13),
																															},
																															End: ast.Location{
																																Line: int(553),
																																Column: int(73),
																															},
																															File: p1,
																														},
																													},
																												},
																												Body: &ast.Conditional{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(554),
																																Column: int(7),
																															},
																															End: ast.Location{
																																Line: int(562),
																																Column: int(14),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(6),
																																Comment: []string{},
																															},
																														},
																														Ctx: p5846,
																														FreeVars: ast.Identifiers{
																															"ensure_pt",
																															"n_",
																															"prec",
																															"render_int",
																															"std",
																															"str",
																															"strip_trailing_zero",
																															"trailing",
																															"whole",
																														},
																													},
																													Cond: &ast.Binary{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(554),
																																	Column: int(10),
																																},
																																End: ast.Location{
																																	Line: int(554),
																																	Column: int(19),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5846,
																															FreeVars: ast.Identifiers{
																																"prec",
																															},
																														},
																														Left: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(554),
																																		Column: int(10),
																																	},
																																	End: ast.Location{
																																		Line: int(554),
																																		Column: int(14),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p5846,
																																FreeVars: ast.Identifiers{
																																	"prec",
																																},
																															},
																															Id: "prec",
																														},
																														OpFodder: ast.Fodder{},
																														Op: ast.BinaryOp(12),
																														Right: &ast.LiteralNumber{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(554),
																																		Column: int(18),
																																	},
																																	End: ast.Location{
																																		Line: int(554),
																																		Column: int(19),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p5846,
																																FreeVars: nil,
																															},
																															OriginalString: "0",
																														},
																													},
																													ThenFodder: ast.Fodder{},
																													BranchTrue: &ast.Binary{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(555),
																																	Column: int(9),
																																},
																																End: ast.Location{
																																	Line: int(555),
																																	Column: int(44),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p5846,
																															FreeVars: ast.Identifiers{
																																"ensure_pt",
																																"str",
																															},
																														},
																														Left: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(555),
																																		Column: int(9),
																																	},
																																	End: ast.Location{
																																		Line: int(555),
																																		Column: int(12),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{
																																	ast.FodderElement{
																																		Kind: ast.FodderKind(0),
																																		Blanks: int(0),
																																		Indent: int(8),
																																		Comment: []string{},
																																	},
																																},
																																Ctx: p5846,
																																FreeVars: ast.Identifiers{
																																	"str",
																																},
																															},
																															Id: "str",
																														},
																														OpFodder: ast.Fodder{},
																														Op: ast.BinaryOp(3),
																														Right: &ast.Conditional{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(555),
																																		Column: int(15),
																																	},
																																	End: ast.Location{
																																		Line: int(555),
																																		Column: int(44),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p5846,
																																FreeVars: ast.Identifiers{
																																	"ensure_pt",
																																},
																															},
																															Cond: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(555),
																																			Column: int(18),
																																		},
																																		End: ast.Location{
																																			Line: int(555),
																																			Column: int(27),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p5846,
																																	FreeVars: ast.Identifiers{
																																		"ensure_pt",
																																	},
																																},
																																Id: "ensure_pt",
																															},
																															ThenFodder: ast.Fodder{},
																															BranchTrue: &ast.LiteralString{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(555),
																																			Column: int(33),
																																		},
																																		End: ast.Location{
																																			Line: int(555),
																																			Column: int(36),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p5846,
																																	FreeVars: nil,
																																},
																																Value: ".",
																																Kind: ast.LiteralStringKind(1),
																																BlockIndent: "",
																																BlockTermIndent: "",
																															},
																															ElseFodder: ast.Fodder{},
																															BranchFalse: &ast.LiteralString{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(555),
																																			Column: int(42),
																																		},
																																		End: ast.Location{
																																			Line: int(555),
																																			Column: int(44),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p5846,
																																	FreeVars: nil,
																																},
																																Value: "",
																																Kind: ast.LiteralStringKind(1),
																																BlockIndent: "",
																																BlockTermIndent: "",
																															},
																														},
																													},
																													ElseFodder: ast.Fodder{
																														ast.FodderElement{
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(6),
																															Comment: []string{},
																														},
																													},
																													BranchFalse: &ast.Local{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(557),
																																	Column: int(9),
																																},
																																End: ast.Location{
																																	Line: int(562),
																																	Column: int(14),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{
																																ast.FodderElement{
																																	Kind: ast.FodderKind(0),
																																	Blanks: int(0),
																																	Indent: int(8),
																																	Comment: []string{},
																																},
																															},
																															Ctx: p5846,
																															FreeVars: ast.Identifiers{
																																"n_",
																																"prec",
																																"render_int",
																																"std",
																																"str",
																																"strip_trailing_zero",
																																"trailing",
																																"whole",
																															},
																														},
																														Binds: ast.LocalBinds{
																															ast.LocalBind{
																																VarFodder: ast.Fodder{},
																																Variable: "frac",
																																EqFodder: ast.Fodder{},
																																Body: &ast.Apply{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(557),
																																				Column: int(22),
																																			},
																																			End: ast.Location{
																																				Line: int(557),
																																				Column: int(71),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p5964,
																																		FreeVars: ast.Identifiers{
																																			"n_",
																																			"prec",
																																			"std",
																																			"whole",
																																		},
																																	},
																																	Target: &ast.Index{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(557),
																																					Column: int(22),
																																				},
																																				End: ast.Location{
																																					Line: int(557),
																																					Column: int(31),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p5964,
																																			FreeVars: ast.Identifiers{
																																				"std",
																																			},
																																		},
																																		Target: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(557),
																																						Column: int(22),
																																					},
																																					End: ast.Location{
																																						Line: int(557),
																																						Column: int(25),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: nil,
																																				FreeVars: ast.Identifiers{
																																					"std",
																																				},
																																			},
																																			Id: "std",
																																		},
																																		LeftBracketFodder: ast.Fodder{},
																																		Index: &ast.LiteralString{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(0),
																																						Column: int(0),
																																					},
																																					End: ast.Location{
																																						Line: int(0),
																																						Column: int(0),
																																					},
																																					File: nil,
																																				},
																																				Fodder: nil,
																																				Ctx: nil,
																																				FreeVars: nil,
																																			},
																																			Value: "floor",
																																			Kind: ast.LiteralStringKind(1),
																																			BlockIndent: "",
																																			BlockTermIndent: "",
																																		},
																																		RightBracketFodder: ast.Fodder{},
																																		Id: nil,
																																	},
																																	FodderLeft: ast.Fodder{},
																																	Arguments: ast.Arguments{
																																		Positional: []ast.CommaSeparatedExpr{
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Binary{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(557),
																																								Column: int(32),
																																							},
																																							End: ast.Location{
																																								Line: int(557),
																																								Column: int(70),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p5973,
																																						FreeVars: ast.Identifiers{
																																							"n_",
																																							"prec",
																																							"std",
																																							"whole",
																																						},
																																					},
																																					Left: &ast.Binary{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(557),
																																									Column: int(32),
																																								},
																																								End: ast.Location{
																																									Line: int(557),
																																									Column: int(64),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p5973,
																																							FreeVars: ast.Identifiers{
																																								"n_",
																																								"prec",
																																								"std",
																																								"whole",
																																							},
																																						},
																																						Left: &ast.Binary{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(557),
																																										Column: int(33),
																																									},
																																									End: ast.Location{
																																										Line: int(557),
																																										Column: int(43),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p5973,
																																								FreeVars: ast.Identifiers{
																																									"n_",
																																									"whole",
																																								},
																																							},
																																							Left: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(557),
																																											Column: int(33),
																																										},
																																										End: ast.Location{
																																											Line: int(557),
																																											Column: int(35),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p5973,
																																									FreeVars: ast.Identifiers{
																																										"n_",
																																									},
																																								},
																																								Id: "n_",
																																							},
																																							OpFodder: ast.Fodder{},
																																							Op: ast.BinaryOp(4),
																																							Right: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(557),
																																											Column: int(38),
																																										},
																																										End: ast.Location{
																																											Line: int(557),
																																											Column: int(43),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p5973,
																																									FreeVars: ast.Identifiers{
																																										"whole",
																																									},
																																								},
																																								Id: "whole",
																																							},
																																						},
																																						OpFodder: ast.Fodder{},
																																						Op: ast.BinaryOp(0),
																																						Right: &ast.Apply{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(557),
																																										Column: int(47),
																																									},
																																									End: ast.Location{
																																										Line: int(557),
																																										Column: int(64),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p5973,
																																								FreeVars: ast.Identifiers{
																																									"prec",
																																									"std",
																																								},
																																							},
																																							Target: &ast.Index{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(557),
																																											Column: int(47),
																																										},
																																										End: ast.Location{
																																											Line: int(557),
																																											Column: int(54),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p5973,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																									},
																																								},
																																								Target: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(557),
																																												Column: int(47),
																																											},
																																											End: ast.Location{
																																												Line: int(557),
																																												Column: int(50),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																										},
																																									},
																																									Id: "std",
																																								},
																																								LeftBracketFodder: ast.Fodder{},
																																								Index: &ast.LiteralString{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											File: nil,
																																										},
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: nil,
																																									},
																																									Value: "pow",
																																									Kind: ast.LiteralStringKind(1),
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								Id: nil,
																																							},
																																							FodderLeft: ast.Fodder{},
																																							Arguments: ast.Arguments{
																																								Positional: []ast.CommaSeparatedExpr{
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.LiteralNumber{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(557),
																																														Column: int(55),
																																													},
																																													End: ast.Location{
																																														Line: int(557),
																																														Column: int(57),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p5992,
																																												FreeVars: nil,
																																											},
																																											OriginalString: "10",
																																										},
																																										CommaFodder: ast.Fodder{},
																																									},
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(557),
																																														Column: int(59),
																																													},
																																													End: ast.Location{
																																														Line: int(557),
																																														Column: int(63),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p5992,
																																												FreeVars: ast.Identifiers{
																																													"prec",
																																												},
																																											},
																																											Id: "prec",
																																										},
																																										CommaFodder: nil,
																																									},
																																								},
																																								Named: nil,
																																							},
																																							TrailingComma: false,
																																							TailStrict: false,
																																							FodderRight: ast.Fodder{},
																																							TailStrictFodder: nil,
																																						},
																																					},
																																					OpFodder: ast.Fodder{},
																																					Op: ast.BinaryOp(3),
																																					Right: &ast.LiteralNumber{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(557),
																																									Column: int(67),
																																								},
																																								End: ast.Location{
																																									Line: int(557),
																																									Column: int(70),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p5973,
																																							FreeVars: nil,
																																						},
																																						OriginalString: "0.5",
																																					},
																																				},
																																				CommaFodder: nil,
																																			},
																																		},
																																		Named: nil,
																																	},
																																	TrailingComma: false,
																																	TailStrict: false,
																																	FodderRight: ast.Fodder{},
																																	TailStrictFodder: nil,
																																},
																																Fun: nil,
																																CloseFodder: ast.Fodder{},
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(557),
																																		Column: int(15),
																																	},
																																	End: ast.Location{
																																		Line: int(557),
																																		Column: int(71),
																																	},
																																	File: p1,
																																},
																															},
																														},
																														Body: &ast.Conditional{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(558),
																																		Column: int(9),
																																	},
																																	End: ast.Location{
																																		Line: int(562),
																																		Column: int(14),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{
																																	ast.FodderElement{
																																		Kind: ast.FodderKind(0),
																																		Blanks: int(0),
																																		Indent: int(8),
																																		Comment: []string{},
																																	},
																																},
																																Ctx: p5846,
																																FreeVars: ast.Identifiers{
																																	"frac",
																																	"prec",
																																	"render_int",
																																	"str",
																																	"strip_trailing_zero",
																																	"trailing",
																																},
																															},
																															Cond: &ast.Binary{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(558),
																																			Column: int(12),
																																		},
																																		End: ast.Location{
																																			Line: int(558),
																																			Column: int(32),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p5846,
																																	FreeVars: ast.Identifiers{
																																		"frac",
																																		"trailing",
																																	},
																																},
																																Left: &ast.Var{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(558),
																																				Column: int(12),
																																			},
																																			End: ast.Location{
																																				Line: int(558),
																																				Column: int(20),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p5846,
																																		FreeVars: ast.Identifiers{
																																			"trailing",
																																		},
																																	},
																																	Id: "trailing",
																																},
																																OpFodder: ast.Fodder{},
																																Op: ast.BinaryOp(18),
																																Right: &ast.Binary{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(558),
																																				Column: int(24),
																																			},
																																			End: ast.Location{
																																				Line: int(558),
																																				Column: int(32),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p5846,
																																		FreeVars: ast.Identifiers{
																																			"frac",
																																		},
																																	},
																																	Left: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(558),
																																					Column: int(24),
																																				},
																																				End: ast.Location{
																																					Line: int(558),
																																					Column: int(28),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p5846,
																																			FreeVars: ast.Identifiers{
																																				"frac",
																																			},
																																		},
																																		Id: "frac",
																																	},
																																	OpFodder: ast.Fodder{},
																																	Op: ast.BinaryOp(7),
																																	Right: &ast.LiteralNumber{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(558),
																																					Column: int(31),
																																				},
																																				End: ast.Location{
																																					Line: int(558),
																																					Column: int(32),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p5846,
																																			FreeVars: nil,
																																		},
																																		OriginalString: "0",
																																	},
																																},
																															},
																															ThenFodder: ast.Fodder{},
																															BranchTrue: &ast.Local{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(559),
																																			Column: int(11),
																																		},
																																		End: ast.Location{
																																			Line: int(560),
																																			Column: int(84),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{
																																		ast.FodderElement{
																																			Kind: ast.FodderKind(0),
																																			Blanks: int(0),
																																			Indent: int(10),
																																			Comment: []string{},
																																		},
																																	},
																																	Ctx: p5846,
																																	FreeVars: ast.Identifiers{
																																		"frac",
																																		"prec",
																																		"render_int",
																																		"str",
																																		"strip_trailing_zero",
																																		"trailing",
																																	},
																																},
																																Binds: ast.LocalBinds{
																																	ast.LocalBind{
																																		VarFodder: ast.Fodder{},
																																		Variable: "frac_str",
																																		EqFodder: ast.Fodder{},
																																		Body: &ast.Apply{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(559),
																																						Column: int(28),
																																					},
																																					End: ast.Location{
																																						Line: int(559),
																																						Column: int(82),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p6013,
																																				FreeVars: ast.Identifiers{
																																					"frac",
																																					"prec",
																																					"render_int",
																																				},
																																			},
																																			Target: &ast.Var{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(559),
																																							Column: int(28),
																																						},
																																						End: ast.Location{
																																							Line: int(559),
																																							Column: int(38),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p6013,
																																					FreeVars: ast.Identifiers{
																																						"render_int",
																																					},
																																				},
																																				Id: "render_int",
																																			},
																																			FodderLeft: ast.Fodder{},
																																			Arguments: ast.Arguments{
																																				Positional: []ast.CommaSeparatedExpr{
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.LiteralBoolean{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(559),
																																										Column: int(39),
																																									},
																																									End: ast.Location{
																																										Line: int(559),
																																										Column: int(44),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6019,
																																								FreeVars: nil,
																																							},
																																							Value: false,
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(559),
																																										Column: int(46),
																																									},
																																									End: ast.Location{
																																										Line: int(559),
																																										Column: int(50),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6019,
																																								FreeVars: ast.Identifiers{
																																									"frac",
																																								},
																																							},
																																							Id: "frac",
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(559),
																																										Column: int(52),
																																									},
																																									End: ast.Location{
																																										Line: int(559),
																																										Column: int(56),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6019,
																																								FreeVars: ast.Identifiers{
																																									"prec",
																																								},
																																							},
																																							Id: "prec",
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.LiteralNumber{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(559),
																																										Column: int(58),
																																									},
																																									End: ast.Location{
																																										Line: int(559),
																																										Column: int(59),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6019,
																																								FreeVars: nil,
																																							},
																																							OriginalString: "0",
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.LiteralBoolean{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(559),
																																										Column: int(61),
																																									},
																																									End: ast.Location{
																																										Line: int(559),
																																										Column: int(66),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6019,
																																								FreeVars: nil,
																																							},
																																							Value: false,
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.LiteralBoolean{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(559),
																																										Column: int(68),
																																									},
																																									End: ast.Location{
																																										Line: int(559),
																																										Column: int(73),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6019,
																																								FreeVars: nil,
																																							},
																																							Value: false,
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.LiteralNumber{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(559),
																																										Column: int(75),
																																									},
																																									End: ast.Location{
																																										Line: int(559),
																																										Column: int(77),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6019,
																																								FreeVars: nil,
																																							},
																																							OriginalString: "10",
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.LiteralString{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(559),
																																										Column: int(79),
																																									},
																																									End: ast.Location{
																																										Line: int(559),
																																										Column: int(81),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6019,
																																								FreeVars: nil,
																																							},
																																							Value: "",
																																							Kind: ast.LiteralStringKind(1),
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																						},
																																						CommaFodder: nil,
																																					},
																																				},
																																				Named: nil,
																																			},
																																			TrailingComma: false,
																																			TailStrict: false,
																																			FodderRight: ast.Fodder{},
																																			TailStrictFodder: nil,
																																		},
																																		Fun: nil,
																																		CloseFodder: ast.Fodder{},
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(559),
																																				Column: int(17),
																																			},
																																			End: ast.Location{
																																				Line: int(559),
																																				Column: int(82),
																																			},
																																			File: p1,
																																		},
																																	},
																																},
																																Body: &ast.Binary{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(560),
																																				Column: int(11),
																																			},
																																			End: ast.Location{
																																				Line: int(560),
																																				Column: int(84),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p5846,
																																		FreeVars: ast.Identifiers{
																																			"frac_str",
																																			"str",
																																			"strip_trailing_zero",
																																			"trailing",
																																		},
																																	},
																																	Left: &ast.Binary{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(560),
																																					Column: int(11),
																																				},
																																				End: ast.Location{
																																					Line: int(560),
																																					Column: int(20),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p5846,
																																			FreeVars: ast.Identifiers{
																																				"str",
																																			},
																																		},
																																		Left: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(560),
																																						Column: int(11),
																																					},
																																					End: ast.Location{
																																						Line: int(560),
																																						Column: int(14),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{
																																					ast.FodderElement{
																																						Kind: ast.FodderKind(0),
																																						Blanks: int(0),
																																						Indent: int(10),
																																						Comment: []string{},
																																					},
																																				},
																																				Ctx: p5846,
																																				FreeVars: ast.Identifiers{
																																					"str",
																																				},
																																			},
																																			Id: "str",
																																		},
																																		OpFodder: ast.Fodder{},
																																		Op: ast.BinaryOp(3),
																																		Right: &ast.LiteralString{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(560),
																																						Column: int(17),
																																					},
																																					End: ast.Location{
																																						Line: int(560),
																																						Column: int(20),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p5846,
																																				FreeVars: nil,
																																			},
																																			Value: ".",
																																			Kind: ast.LiteralStringKind(1),
																																			BlockIndent: "",
																																			BlockTermIndent: "",
																																		},
																																	},
																																	OpFodder: ast.Fodder{},
																																	Op: ast.BinaryOp(3),
																																	Right: &ast.Conditional{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(560),
																																					Column: int(23),
																																				},
																																				End: ast.Location{
																																					Line: int(560),
																																					Column: int(84),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p5846,
																																			FreeVars: ast.Identifiers{
																																				"frac_str",
																																				"strip_trailing_zero",
																																				"trailing",
																																			},
																																		},
																																		Cond: &ast.Unary{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(560),
																																						Column: int(26),
																																					},
																																					End: ast.Location{
																																						Line: int(560),
																																						Column: int(35),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p5846,
																																				FreeVars: ast.Identifiers{
																																					"trailing",
																																				},
																																			},
																																			Op: ast.UnaryOp(0),
																																			Expr: &ast.Var{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(560),
																																							Column: int(27),
																																						},
																																						End: ast.Location{
																																							Line: int(560),
																																							Column: int(35),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p5846,
																																					FreeVars: ast.Identifiers{
																																						"trailing",
																																					},
																																				},
																																				Id: "trailing",
																																			},
																																		},
																																		ThenFodder: ast.Fodder{},
																																		BranchTrue: &ast.Apply{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(560),
																																						Column: int(41),
																																					},
																																					End: ast.Location{
																																						Line: int(560),
																																						Column: int(70),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p5846,
																																				FreeVars: ast.Identifiers{
																																					"frac_str",
																																					"strip_trailing_zero",
																																				},
																																			},
																																			Target: &ast.Var{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(560),
																																							Column: int(41),
																																						},
																																						End: ast.Location{
																																							Line: int(560),
																																							Column: int(60),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p5846,
																																					FreeVars: ast.Identifiers{
																																						"strip_trailing_zero",
																																					},
																																				},
																																				Id: "strip_trailing_zero",
																																			},
																																			FodderLeft: ast.Fodder{},
																																			Arguments: ast.Arguments{
																																				Positional: []ast.CommaSeparatedExpr{
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(560),
																																										Column: int(61),
																																									},
																																									End: ast.Location{
																																										Line: int(560),
																																										Column: int(69),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6049,
																																								FreeVars: ast.Identifiers{
																																									"frac_str",
																																								},
																																							},
																																							Id: "frac_str",
																																						},
																																						CommaFodder: nil,
																																					},
																																				},
																																				Named: nil,
																																			},
																																			TrailingComma: false,
																																			TailStrict: false,
																																			FodderRight: ast.Fodder{},
																																			TailStrictFodder: nil,
																																		},
																																		ElseFodder: ast.Fodder{},
																																		BranchFalse: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(560),
																																						Column: int(76),
																																					},
																																					End: ast.Location{
																																						Line: int(560),
																																						Column: int(84),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p5846,
																																				FreeVars: ast.Identifiers{
																																					"frac_str",
																																				},
																																			},
																																			Id: "frac_str",
																																		},
																																	},
																																},
																															},
																															ElseFodder: ast.Fodder{
																																ast.FodderElement{
																																	Kind: ast.FodderKind(0),
																																	Blanks: int(0),
																																	Indent: int(8),
																																	Comment: []string{},
																																},
																															},
																															BranchFalse: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(562),
																																			Column: int(11),
																																		},
																																		End: ast.Location{
																																			Line: int(562),
																																			Column: int(14),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{
																																		ast.FodderElement{
																																			Kind: ast.FodderKind(0),
																																			Blanks: int(0),
																																			Indent: int(10),
																																			Comment: []string{},
																																		},
																																	},
																																	Ctx: p5846,
																																	FreeVars: ast.Identifiers{
																																		"str",
																																	},
																																},
																																Id: "str",
																															},
																														},
																													},
																												},
																											},
																										},
																									},
																								},
																							},
																						},
																						Fun: nil,
																						CloseFodder: nil,
																						LocRange: ast.LocationRange{
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							File: nil,
																						},
																					},
																				},
																				Body: &ast.Local{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(565),
																								Column: int(5),
																							},
																							End: ast.Location{
																								Line: int(774),
																								Column: int(48),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Kind: ast.FodderKind(0),
																								Blanks: int(1),
																								Indent: int(4),
																								Comment: []string{},
																							},
																							ast.FodderElement{
																								Kind: ast.FodderKind(2),
																								Blanks: int(0),
																								Indent: int(4),
																								Comment: []string{
																									"// Render floating point in scientific form",
																								},
																							},
																						},
																						Ctx: p3497,
																						FreeVars: ast.Identifiers{
																							"codes",
																							"pad_left",
																							"pad_right",
																							"render_float_dec",
																							"render_hex",
																							"render_int",
																							"std",
																							"vals",
																						},
																					},
																					Binds: ast.LocalBinds{
																						ast.LocalBind{
																							VarFodder: nil,
																							Variable: "render_float_sci",
																							EqFodder: nil,
																							Body: &ast.Function{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(565),
																											Column: int(11),
																										},
																										End: ast.Location{
																											Line: int(576),
																											Column: int(85),
																										},
																										File: p1,
																									},
																									Fodder: nil,
																									Ctx: p6063,
																									FreeVars: ast.Identifiers{
																										"render_float_dec",
																										"render_int",
																										"std",
																									},
																								},
																								ParenLeftFodder: ast.Fodder{},
																								Parameters: []ast.Parameter{
																									ast.Parameter{
																										NameFodder: ast.Fodder{},
																										Name: "n__",
																										EqFodder: nil,
																										DefaultArg: nil,
																										CommaFodder: ast.Fodder{},
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(565),
																												Column: int(28),
																											},
																											End: ast.Location{
																												Line: int(565),
																												Column: int(31),
																											},
																											File: p1,
																										},
																									},
																									ast.Parameter{
																										NameFodder: ast.Fodder{},
																										Name: "zero_pad",
																										EqFodder: nil,
																										DefaultArg: nil,
																										CommaFodder: ast.Fodder{},
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(565),
																												Column: int(33),
																											},
																											End: ast.Location{
																												Line: int(565),
																												Column: int(41),
																											},
																											File: p1,
																										},
																									},
																									ast.Parameter{
																										NameFodder: ast.Fodder{},
																										Name: "blank",
																										EqFodder: nil,
																										DefaultArg: nil,
																										CommaFodder: ast.Fodder{},
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(565),
																												Column: int(43),
																											},
																											End: ast.Location{
																												Line: int(565),
																												Column: int(48),
																											},
																											File: p1,
																										},
																									},
																									ast.Parameter{
																										NameFodder: ast.Fodder{},
																										Name: "plus",
																										EqFodder: nil,
																										DefaultArg: nil,
																										CommaFodder: ast.Fodder{},
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(565),
																												Column: int(50),
																											},
																											End: ast.Location{
																												Line: int(565),
																												Column: int(54),
																											},
																											File: p1,
																										},
																									},
																									ast.Parameter{
																										NameFodder: ast.Fodder{},
																										Name: "ensure_pt",
																										EqFodder: nil,
																										DefaultArg: nil,
																										CommaFodder: ast.Fodder{},
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(565),
																												Column: int(56),
																											},
																											End: ast.Location{
																												Line: int(565),
																												Column: int(65),
																											},
																											File: p1,
																										},
																									},
																									ast.Parameter{
																										NameFodder: ast.Fodder{},
																										Name: "trailing",
																										EqFodder: nil,
																										DefaultArg: nil,
																										CommaFodder: ast.Fodder{},
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(565),
																												Column: int(67),
																											},
																											End: ast.Location{
																												Line: int(565),
																												Column: int(75),
																											},
																											File: p1,
																										},
																									},
																									ast.Parameter{
																										NameFodder: ast.Fodder{},
																										Name: "caps",
																										EqFodder: nil,
																										DefaultArg: nil,
																										CommaFodder: ast.Fodder{},
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(565),
																												Column: int(77),
																											},
																											End: ast.Location{
																												Line: int(565),
																												Column: int(81),
																											},
																											File: p1,
																										},
																									},
																									ast.Parameter{
																										NameFodder: ast.Fodder{},
																										Name: "prec",
																										EqFodder: nil,
																										DefaultArg: nil,
																										CommaFodder: nil,
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(565),
																												Column: int(83),
																											},
																											End: ast.Location{
																												Line: int(565),
																												Column: int(87),
																											},
																											File: p1,
																										},
																									},
																								},
																								TrailingComma: false,
																								ParenRightFodder: ast.Fodder{},
																								Body: &ast.Local{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(566),
																												Column: int(7),
																											},
																											End: ast.Location{
																												Line: int(576),
																												Column: int(85),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{
																											ast.FodderElement{
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(6),
																												Comment: []string{},
																											},
																										},
																										Ctx: p6068,
																										FreeVars: ast.Identifiers{
																											"blank",
																											"caps",
																											"ensure_pt",
																											"n__",
																											"plus",
																											"prec",
																											"render_float_dec",
																											"render_int",
																											"std",
																											"trailing",
																											"zero_pad",
																										},
																									},
																									Binds: ast.LocalBinds{
																										ast.LocalBind{
																											VarFodder: ast.Fodder{},
																											Variable: "exponent",
																											EqFodder: ast.Fodder{},
																											Body: &ast.Conditional{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(566),
																															Column: int(24),
																														},
																														End: ast.Location{
																															Line: int(566),
																															Column: int(94),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p6072,
																													FreeVars: ast.Identifiers{
																														"n__",
																														"std",
																													},
																												},
																												Cond: &ast.Binary{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(566),
																																Column: int(27),
																															},
																															End: ast.Location{
																																Line: int(566),
																																Column: int(35),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p6072,
																														FreeVars: ast.Identifiers{
																															"n__",
																														},
																													},
																													Left: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(566),
																																	Column: int(27),
																																},
																																End: ast.Location{
																																	Line: int(566),
																																	Column: int(30),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p6072,
																															FreeVars: ast.Identifiers{
																																"n__",
																															},
																														},
																														Id: "n__",
																													},
																													OpFodder: ast.Fodder{},
																													Op: ast.BinaryOp(12),
																													Right: &ast.LiteralNumber{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(566),
																																	Column: int(34),
																																},
																																End: ast.Location{
																																	Line: int(566),
																																	Column: int(35),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p6072,
																															FreeVars: nil,
																														},
																														OriginalString: "0",
																													},
																												},
																												ThenFodder: ast.Fodder{},
																												BranchTrue: &ast.LiteralNumber{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(566),
																																Column: int(41),
																															},
																															End: ast.Location{
																																Line: int(566),
																																Column: int(42),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p6072,
																														FreeVars: nil,
																													},
																													OriginalString: "0",
																												},
																												ElseFodder: ast.Fodder{},
																												BranchFalse: &ast.Apply{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(566),
																																Column: int(48),
																															},
																															End: ast.Location{
																																Line: int(566),
																																Column: int(94),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p6072,
																														FreeVars: ast.Identifiers{
																															"n__",
																															"std",
																														},
																													},
																													Target: &ast.Index{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(566),
																																	Column: int(48),
																																},
																																End: ast.Location{
																																	Line: int(566),
																																	Column: int(57),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p6072,
																															FreeVars: ast.Identifiers{
																																"std",
																															},
																														},
																														Target: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(566),
																																		Column: int(48),
																																	},
																																	End: ast.Location{
																																		Line: int(566),
																																		Column: int(51),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																															},
																															Id: "std",
																														},
																														LeftBracketFodder: ast.Fodder{},
																														Index: &ast.LiteralString{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	File: nil,
																																},
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: nil,
																															},
																															Value: "floor",
																															Kind: ast.LiteralStringKind(1),
																															BlockIndent: "",
																															BlockTermIndent: "",
																														},
																														RightBracketFodder: ast.Fodder{},
																														Id: nil,
																													},
																													FodderLeft: ast.Fodder{},
																													Arguments: ast.Arguments{
																														Positional: []ast.CommaSeparatedExpr{
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Binary{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(566),
																																				Column: int(58),
																																			},
																																			End: ast.Location{
																																				Line: int(566),
																																				Column: int(93),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p6089,
																																		FreeVars: ast.Identifiers{
																																			"n__",
																																			"std",
																																		},
																																	},
																																	Left: &ast.Apply{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(566),
																																					Column: int(58),
																																				},
																																				End: ast.Location{
																																					Line: int(566),
																																					Column: int(79),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p6089,
																																			FreeVars: ast.Identifiers{
																																				"n__",
																																				"std",
																																			},
																																		},
																																		Target: &ast.Index{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(566),
																																						Column: int(58),
																																					},
																																					End: ast.Location{
																																						Line: int(566),
																																						Column: int(65),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p6089,
																																				FreeVars: ast.Identifiers{
																																					"std",
																																				},
																																			},
																																			Target: &ast.Var{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(566),
																																							Column: int(58),
																																						},
																																						End: ast.Location{
																																							Line: int(566),
																																							Column: int(61),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{
																																						"std",
																																					},
																																				},
																																				Id: "std",
																																			},
																																			LeftBracketFodder: ast.Fodder{},
																																			Index: &ast.LiteralString{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						End: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						File: nil,
																																					},
																																					Fodder: nil,
																																					Ctx: nil,
																																					FreeVars: nil,
																																				},
																																				Value: "log",
																																				Kind: ast.LiteralStringKind(1),
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																			},
																																			RightBracketFodder: ast.Fodder{},
																																			Id: nil,
																																		},
																																		FodderLeft: ast.Fodder{},
																																		Arguments: ast.Arguments{
																																			Positional: []ast.CommaSeparatedExpr{
																																				ast.CommaSeparatedExpr{
																																					Expr: &ast.Apply{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(566),
																																									Column: int(66),
																																								},
																																								End: ast.Location{
																																									Line: int(566),
																																									Column: int(78),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p6100,
																																							FreeVars: ast.Identifiers{
																																								"n__",
																																								"std",
																																							},
																																						},
																																						Target: &ast.Index{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(566),
																																										Column: int(66),
																																									},
																																									End: ast.Location{
																																										Line: int(566),
																																										Column: int(73),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6100,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																								},
																																							},
																																							Target: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(566),
																																											Column: int(66),
																																										},
																																										End: ast.Location{
																																											Line: int(566),
																																											Column: int(69),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																									},
																																								},
																																								Id: "std",
																																							},
																																							LeftBracketFodder: ast.Fodder{},
																																							Index: &ast.LiteralString{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										File: nil,
																																									},
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: nil,
																																								},
																																								Value: "abs",
																																								Kind: ast.LiteralStringKind(1),
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							Id: nil,
																																						},
																																						FodderLeft: ast.Fodder{},
																																						Arguments: ast.Arguments{
																																							Positional: []ast.CommaSeparatedExpr{
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(566),
																																													Column: int(74),
																																												},
																																												End: ast.Location{
																																													Line: int(566),
																																													Column: int(77),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p6109,
																																											FreeVars: ast.Identifiers{
																																												"n__",
																																											},
																																										},
																																										Id: "n__",
																																									},
																																									CommaFodder: nil,
																																								},
																																							},
																																							Named: nil,
																																						},
																																						TrailingComma: false,
																																						TailStrict: false,
																																						FodderRight: ast.Fodder{},
																																						TailStrictFodder: nil,
																																					},
																																					CommaFodder: nil,
																																				},
																																			},
																																			Named: nil,
																																		},
																																		TrailingComma: false,
																																		TailStrict: false,
																																		FodderRight: ast.Fodder{},
																																		TailStrictFodder: nil,
																																	},
																																	OpFodder: ast.Fodder{},
																																	Op: ast.BinaryOp(1),
																																	Right: &ast.Apply{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(566),
																																					Column: int(82),
																																				},
																																				End: ast.Location{
																																					Line: int(566),
																																					Column: int(93),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p6089,
																																			FreeVars: ast.Identifiers{
																																				"std",
																																			},
																																		},
																																		Target: &ast.Index{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(566),
																																						Column: int(82),
																																					},
																																					End: ast.Location{
																																						Line: int(566),
																																						Column: int(89),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p6089,
																																				FreeVars: ast.Identifiers{
																																					"std",
																																				},
																																			},
																																			Target: &ast.Var{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(566),
																																							Column: int(82),
																																						},
																																						End: ast.Location{
																																							Line: int(566),
																																							Column: int(85),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{
																																						"std",
																																					},
																																				},
																																				Id: "std",
																																			},
																																			LeftBracketFodder: ast.Fodder{},
																																			Index: &ast.LiteralString{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						End: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						File: nil,
																																					},
																																					Fodder: nil,
																																					Ctx: nil,
																																					FreeVars: nil,
																																				},
																																				Value: "log",
																																				Kind: ast.LiteralStringKind(1),
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																			},
																																			RightBracketFodder: ast.Fodder{},
																																			Id: nil,
																																		},
																																		FodderLeft: ast.Fodder{},
																																		Arguments: ast.Arguments{
																																			Positional: []ast.CommaSeparatedExpr{
																																				ast.CommaSeparatedExpr{
																																					Expr: &ast.LiteralNumber{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(566),
																																									Column: int(90),
																																								},
																																								End: ast.Location{
																																									Line: int(566),
																																									Column: int(92),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p6120,
																																							FreeVars: nil,
																																						},
																																						OriginalString: "10",
																																					},
																																					CommaFodder: nil,
																																				},
																																			},
																																			Named: nil,
																																		},
																																		TrailingComma: false,
																																		TailStrict: false,
																																		FodderRight: ast.Fodder{},
																																		TailStrictFodder: nil,
																																	},
																																},
																																CommaFodder: nil,
																															},
																														},
																														Named: nil,
																													},
																													TrailingComma: false,
																													TailStrict: false,
																													FodderRight: ast.Fodder{},
																													TailStrictFodder: nil,
																												},
																											},
																											Fun: nil,
																											CloseFodder: ast.Fodder{},
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(566),
																													Column: int(13),
																												},
																												End: ast.Location{
																													Line: int(566),
																													Column: int(94),
																												},
																												File: p1,
																											},
																										},
																									},
																									Body: &ast.Local{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(567),
																													Column: int(7),
																												},
																												End: ast.Location{
																													Line: int(576),
																													Column: int(85),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{
																												ast.FodderElement{
																													Kind: ast.FodderKind(0),
																													Blanks: int(0),
																													Indent: int(6),
																													Comment: []string{},
																												},
																											},
																											Ctx: p6068,
																											FreeVars: ast.Identifiers{
																												"blank",
																												"caps",
																												"ensure_pt",
																												"exponent",
																												"n__",
																												"plus",
																												"prec",
																												"render_float_dec",
																												"render_int",
																												"std",
																												"trailing",
																												"zero_pad",
																											},
																										},
																										Binds: ast.LocalBinds{
																											ast.LocalBind{
																												VarFodder: ast.Fodder{},
																												Variable: "suff",
																												EqFodder: ast.Fodder{},
																												Body: &ast.Binary{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(567),
																																Column: int(20),
																															},
																															End: ast.Location{
																																Line: int(568),
																																Column: int(92),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p6126,
																														FreeVars: ast.Identifiers{
																															"caps",
																															"exponent",
																															"render_int",
																															"std",
																														},
																													},
																													Left: &ast.Conditional{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(567),
																																	Column: int(21),
																																},
																																End: ast.Location{
																																	Line: int(567),
																																	Column: int(46),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p6126,
																															FreeVars: ast.Identifiers{
																																"caps",
																															},
																														},
																														Cond: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(567),
																																		Column: int(24),
																																	},
																																	End: ast.Location{
																																		Line: int(567),
																																		Column: int(28),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p6126,
																																FreeVars: ast.Identifiers{
																																	"caps",
																																},
																															},
																															Id: "caps",
																														},
																														ThenFodder: ast.Fodder{},
																														BranchTrue: &ast.LiteralString{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(567),
																																		Column: int(34),
																																	},
																																	End: ast.Location{
																																		Line: int(567),
																																		Column: int(37),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p6126,
																																FreeVars: nil,
																															},
																															Value: "E",
																															Kind: ast.LiteralStringKind(1),
																															BlockIndent: "",
																															BlockTermIndent: "",
																														},
																														ElseFodder: ast.Fodder{},
																														BranchFalse: &ast.LiteralString{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(567),
																																		Column: int(43),
																																	},
																																	End: ast.Location{
																																		Line: int(567),
																																		Column: int(46),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p6126,
																																FreeVars: nil,
																															},
																															Value: "e",
																															Kind: ast.LiteralStringKind(1),
																															BlockIndent: "",
																															BlockTermIndent: "",
																														},
																													},
																													OpFodder: ast.Fodder{
																														ast.FodderElement{
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(19),
																															Comment: []string{},
																														},
																													},
																													Op: ast.BinaryOp(3),
																													Right: &ast.Apply{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(568),
																																	Column: int(22),
																																},
																																End: ast.Location{
																																	Line: int(568),
																																	Column: int(92),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p6126,
																															FreeVars: ast.Identifiers{
																																"exponent",
																																"render_int",
																																"std",
																															},
																														},
																														Target: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(568),
																																		Column: int(22),
																																	},
																																	End: ast.Location{
																																		Line: int(568),
																																		Column: int(32),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p6126,
																																FreeVars: ast.Identifiers{
																																	"render_int",
																																},
																															},
																															Id: "render_int",
																														},
																														FodderLeft: ast.Fodder{},
																														Arguments: ast.Arguments{
																															Positional: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Binary{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(568),
																																					Column: int(33),
																																				},
																																				End: ast.Location{
																																					Line: int(568),
																																					Column: int(45),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p6141,
																																			FreeVars: ast.Identifiers{
																																				"exponent",
																																			},
																																		},
																																		Left: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(568),
																																						Column: int(33),
																																					},
																																					End: ast.Location{
																																						Line: int(568),
																																						Column: int(41),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p6141,
																																				FreeVars: ast.Identifiers{
																																					"exponent",
																																				},
																																			},
																																			Id: "exponent",
																																		},
																																		OpFodder: ast.Fodder{},
																																		Op: ast.BinaryOp(9),
																																		Right: &ast.LiteralNumber{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(568),
																																						Column: int(44),
																																					},
																																					End: ast.Location{
																																						Line: int(568),
																																						Column: int(45),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p6141,
																																				FreeVars: nil,
																																			},
																																			OriginalString: "0",
																																		},
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Apply{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(568),
																																					Column: int(47),
																																				},
																																				End: ast.Location{
																																					Line: int(568),
																																					Column: int(64),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p6141,
																																			FreeVars: ast.Identifiers{
																																				"exponent",
																																				"std",
																																			},
																																		},
																																		Target: &ast.Index{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(568),
																																						Column: int(47),
																																					},
																																					End: ast.Location{
																																						Line: int(568),
																																						Column: int(54),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p6141,
																																				FreeVars: ast.Identifiers{
																																					"std",
																																				},
																																			},
																																			Target: &ast.Var{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(568),
																																							Column: int(47),
																																						},
																																						End: ast.Location{
																																							Line: int(568),
																																							Column: int(50),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{
																																						"std",
																																					},
																																				},
																																				Id: "std",
																																			},
																																			LeftBracketFodder: ast.Fodder{},
																																			Index: &ast.LiteralString{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						End: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						File: nil,
																																					},
																																					Fodder: nil,
																																					Ctx: nil,
																																					FreeVars: nil,
																																				},
																																				Value: "abs",
																																				Kind: ast.LiteralStringKind(1),
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																			},
																																			RightBracketFodder: ast.Fodder{},
																																			Id: nil,
																																		},
																																		FodderLeft: ast.Fodder{},
																																		Arguments: ast.Arguments{
																																			Positional: []ast.CommaSeparatedExpr{
																																				ast.CommaSeparatedExpr{
																																					Expr: &ast.Var{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(568),
																																									Column: int(55),
																																								},
																																								End: ast.Location{
																																									Line: int(568),
																																									Column: int(63),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p6155,
																																							FreeVars: ast.Identifiers{
																																								"exponent",
																																							},
																																						},
																																						Id: "exponent",
																																					},
																																					CommaFodder: nil,
																																				},
																																			},
																																			Named: nil,
																																		},
																																		TrailingComma: false,
																																		TailStrict: false,
																																		FodderRight: ast.Fodder{},
																																		TailStrictFodder: nil,
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.LiteralNumber{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(568),
																																					Column: int(66),
																																				},
																																				End: ast.Location{
																																					Line: int(568),
																																					Column: int(67),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p6141,
																																			FreeVars: nil,
																																		},
																																		OriginalString: "3",
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.LiteralNumber{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(568),
																																					Column: int(69),
																																				},
																																				End: ast.Location{
																																					Line: int(568),
																																					Column: int(70),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p6141,
																																			FreeVars: nil,
																																		},
																																		OriginalString: "0",
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.LiteralBoolean{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(568),
																																					Column: int(72),
																																				},
																																				End: ast.Location{
																																					Line: int(568),
																																					Column: int(77),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p6141,
																																			FreeVars: nil,
																																		},
																																		Value: false,
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.LiteralBoolean{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(568),
																																					Column: int(79),
																																				},
																																				End: ast.Location{
																																					Line: int(568),
																																					Column: int(83),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p6141,
																																			FreeVars: nil,
																																		},
																																		Value: true,
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.LiteralNumber{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(568),
																																					Column: int(85),
																																				},
																																				End: ast.Location{
																																					Line: int(568),
																																					Column: int(87),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p6141,
																																			FreeVars: nil,
																																		},
																																		OriginalString: "10",
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.LiteralString{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(568),
																																					Column: int(89),
																																				},
																																				End: ast.Location{
																																					Line: int(568),
																																					Column: int(91),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p6141,
																																			FreeVars: nil,
																																		},
																																		Value: "",
																																		Kind: ast.LiteralStringKind(1),
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															Named: nil,
																														},
																														TrailingComma: false,
																														TailStrict: false,
																														FodderRight: ast.Fodder{},
																														TailStrictFodder: nil,
																													},
																												},
																												Fun: nil,
																												CloseFodder: ast.Fodder{},
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(567),
																														Column: int(13),
																													},
																													End: ast.Location{
																														Line: int(568),
																														Column: int(92),
																													},
																													File: p1,
																												},
																											},
																										},
																										Body: &ast.Local{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(569),
																														Column: int(7),
																													},
																													End: ast.Location{
																														Line: int(576),
																														Column: int(85),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{
																													ast.FodderElement{
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(6),
																														Comment: []string{},
																													},
																												},
																												Ctx: p6068,
																												FreeVars: ast.Identifiers{
																													"blank",
																													"ensure_pt",
																													"exponent",
																													"n__",
																													"plus",
																													"prec",
																													"render_float_dec",
																													"std",
																													"suff",
																													"trailing",
																													"zero_pad",
																												},
																											},
																											Binds: ast.LocalBinds{
																												ast.LocalBind{
																													VarFodder: ast.Fodder{},
																													Variable: "mantissa",
																													EqFodder: ast.Fodder{},
																													Body: &ast.Conditional{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(569),
																																	Column: int(24),
																																},
																																End: ast.Location{
																																	Line: int(574),
																																	Column: int(36),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p6168,
																															FreeVars: ast.Identifiers{
																																"exponent",
																																"n__",
																																"std",
																															},
																														},
																														Cond: &ast.Binary{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(569),
																																		Column: int(27),
																																	},
																																	End: ast.Location{
																																		Line: int(569),
																																		Column: int(43),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p6168,
																																FreeVars: ast.Identifiers{
																																	"exponent",
																																},
																															},
																															Left: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(569),
																																			Column: int(27),
																																		},
																																		End: ast.Location{
																																			Line: int(569),
																																			Column: int(35),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p6168,
																																	FreeVars: ast.Identifiers{
																																		"exponent",
																																	},
																																},
																																Id: "exponent",
																															},
																															OpFodder: ast.Fodder{},
																															Op: ast.BinaryOp(12),
																															Right: &ast.Unary{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(569),
																																			Column: int(39),
																																		},
																																		End: ast.Location{
																																			Line: int(569),
																																			Column: int(43),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p6168,
																																	FreeVars: nil,
																																},
																																Op: ast.UnaryOp(3),
																																Expr: &ast.LiteralNumber{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(569),
																																				Column: int(40),
																																			},
																																			End: ast.Location{
																																				Line: int(569),
																																				Column: int(43),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p6168,
																																		FreeVars: nil,
																																	},
																																	OriginalString: "324",
																																},
																															},
																														},
																														ThenFodder: ast.Fodder{},
																														BranchTrue: &ast.Binary{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(572),
																																		Column: int(9),
																																	},
																																	End: ast.Location{
																																		Line: int(572),
																																		Column: int(45),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p6168,
																																FreeVars: ast.Identifiers{
																																	"exponent",
																																	"n__",
																																	"std",
																																},
																															},
																															Left: &ast.Binary{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(572),
																																			Column: int(9),
																																		},
																																		End: ast.Location{
																																			Line: int(572),
																																			Column: int(17),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p6168,
																																	FreeVars: ast.Identifiers{
																																		"n__",
																																	},
																																},
																																Left: &ast.Var{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(572),
																																				Column: int(9),
																																			},
																																			End: ast.Location{
																																				Line: int(572),
																																				Column: int(12),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{
																																			ast.FodderElement{
																																				Kind: ast.FodderKind(0),
																																				Blanks: int(0),
																																				Indent: int(8),
																																				Comment: []string{},
																																			},
																																			ast.FodderElement{
																																				Kind: ast.FodderKind(2),
																																				Blanks: int(0),
																																				Indent: int(8),
																																				Comment: []string{
																																					"// Avoid a rounding error where std.pow(10, -324) is 0",
																																				},
																																			},
																																			ast.FodderElement{
																																				Kind: ast.FodderKind(2),
																																				Blanks: int(0),
																																				Indent: int(8),
																																				Comment: []string{
																																					"// -324 is the smallest exponent possible.",
																																				},
																																			},
																																		},
																																		Ctx: p6168,
																																		FreeVars: ast.Identifiers{
																																			"n__",
																																		},
																																	},
																																	Id: "n__",
																																},
																																OpFodder: ast.Fodder{},
																																Op: ast.BinaryOp(0),
																																Right: &ast.LiteralNumber{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(572),
																																				Column: int(15),
																																			},
																																			End: ast.Location{
																																				Line: int(572),
																																				Column: int(17),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p6168,
																																		FreeVars: nil,
																																	},
																																	OriginalString: "10",
																																},
																															},
																															OpFodder: ast.Fodder{},
																															Op: ast.BinaryOp(1),
																															Right: &ast.Apply{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(572),
																																			Column: int(20),
																																		},
																																		End: ast.Location{
																																			Line: int(572),
																																			Column: int(45),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p6168,
																																	FreeVars: ast.Identifiers{
																																		"exponent",
																																		"std",
																																	},
																																},
																																Target: &ast.Index{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(572),
																																				Column: int(20),
																																			},
																																			End: ast.Location{
																																				Line: int(572),
																																				Column: int(27),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p6168,
																																		FreeVars: ast.Identifiers{
																																			"std",
																																		},
																																	},
																																	Target: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(572),
																																					Column: int(20),
																																				},
																																				End: ast.Location{
																																					Line: int(572),
																																					Column: int(23),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{
																																				"std",
																																			},
																																		},
																																		Id: "std",
																																	},
																																	LeftBracketFodder: ast.Fodder{},
																																	Index: &ast.LiteralString{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				End: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				File: nil,
																																			},
																																			Fodder: nil,
																																			Ctx: nil,
																																			FreeVars: nil,
																																		},
																																		Value: "pow",
																																		Kind: ast.LiteralStringKind(1),
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																	},
																																	RightBracketFodder: ast.Fodder{},
																																	Id: nil,
																																},
																																FodderLeft: ast.Fodder{},
																																Arguments: ast.Arguments{
																																	Positional: []ast.CommaSeparatedExpr{
																																		ast.CommaSeparatedExpr{
																																			Expr: &ast.LiteralNumber{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(572),
																																							Column: int(28),
																																						},
																																						End: ast.Location{
																																							Line: int(572),
																																							Column: int(30),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p6195,
																																					FreeVars: nil,
																																				},
																																				OriginalString: "10",
																																			},
																																			CommaFodder: ast.Fodder{},
																																		},
																																		ast.CommaSeparatedExpr{
																																			Expr: &ast.Binary{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(572),
																																							Column: int(32),
																																						},
																																						End: ast.Location{
																																							Line: int(572),
																																							Column: int(44),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p6195,
																																					FreeVars: ast.Identifiers{
																																						"exponent",
																																					},
																																				},
																																				Left: &ast.Var{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(572),
																																								Column: int(32),
																																							},
																																							End: ast.Location{
																																								Line: int(572),
																																								Column: int(40),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p6195,
																																						FreeVars: ast.Identifiers{
																																							"exponent",
																																						},
																																					},
																																					Id: "exponent",
																																				},
																																				OpFodder: ast.Fodder{},
																																				Op: ast.BinaryOp(3),
																																				Right: &ast.LiteralNumber{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(572),
																																								Column: int(43),
																																							},
																																							End: ast.Location{
																																								Line: int(572),
																																								Column: int(44),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p6195,
																																						FreeVars: nil,
																																					},
																																					OriginalString: "1",
																																				},
																																			},
																																			CommaFodder: nil,
																																		},
																																	},
																																	Named: nil,
																																},
																																TrailingComma: false,
																																TailStrict: false,
																																FodderRight: ast.Fodder{},
																																TailStrictFodder: nil,
																															},
																														},
																														ElseFodder: ast.Fodder{
																															ast.FodderElement{
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(6),
																																Comment: []string{},
																															},
																														},
																														BranchFalse: &ast.Binary{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(574),
																																		Column: int(9),
																																	},
																																	End: ast.Location{
																																		Line: int(574),
																																		Column: int(36),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p6168,
																																FreeVars: ast.Identifiers{
																																	"exponent",
																																	"n__",
																																	"std",
																																},
																															},
																															Left: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(574),
																																			Column: int(9),
																																		},
																																		End: ast.Location{
																																			Line: int(574),
																																			Column: int(12),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{
																																		ast.FodderElement{
																																			Kind: ast.FodderKind(0),
																																			Blanks: int(0),
																																			Indent: int(8),
																																			Comment: []string{},
																																		},
																																	},
																																	Ctx: p6168,
																																	FreeVars: ast.Identifiers{
																																		"n__",
																																	},
																																},
																																Id: "n__",
																															},
																															OpFodder: ast.Fodder{},
																															Op: ast.BinaryOp(1),
																															Right: &ast.Apply{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(574),
																																			Column: int(15),
																																		},
																																		End: ast.Location{
																																			Line: int(574),
																																			Column: int(36),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p6168,
																																	FreeVars: ast.Identifiers{
																																		"exponent",
																																		"std",
																																	},
																																},
																																Target: &ast.Index{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(574),
																																				Column: int(15),
																																			},
																																			End: ast.Location{
																																				Line: int(574),
																																				Column: int(22),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p6168,
																																		FreeVars: ast.Identifiers{
																																			"std",
																																		},
																																	},
																																	Target: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(574),
																																					Column: int(15),
																																				},
																																				End: ast.Location{
																																					Line: int(574),
																																					Column: int(18),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{
																																				"std",
																																			},
																																		},
																																		Id: "std",
																																	},
																																	LeftBracketFodder: ast.Fodder{},
																																	Index: &ast.LiteralString{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				End: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				File: nil,
																																			},
																																			Fodder: nil,
																																			Ctx: nil,
																																			FreeVars: nil,
																																		},
																																		Value: "pow",
																																		Kind: ast.LiteralStringKind(1),
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																	},
																																	RightBracketFodder: ast.Fodder{},
																																	Id: nil,
																																},
																																FodderLeft: ast.Fodder{},
																																Arguments: ast.Arguments{
																																	Positional: []ast.CommaSeparatedExpr{
																																		ast.CommaSeparatedExpr{
																																			Expr: &ast.LiteralNumber{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(574),
																																							Column: int(23),
																																						},
																																						End: ast.Location{
																																							Line: int(574),
																																							Column: int(25),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p6216,
																																					FreeVars: nil,
																																				},
																																				OriginalString: "10",
																																			},
																																			CommaFodder: ast.Fodder{},
																																		},
																																		ast.CommaSeparatedExpr{
																																			Expr: &ast.Var{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(574),
																																							Column: int(27),
																																						},
																																						End: ast.Location{
																																							Line: int(574),
																																							Column: int(35),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p6216,
																																					FreeVars: ast.Identifiers{
																																						"exponent",
																																					},
																																				},
																																				Id: "exponent",
																																			},
																																			CommaFodder: nil,
																																		},
																																	},
																																	Named: nil,
																																},
																																TrailingComma: false,
																																TailStrict: false,
																																FodderRight: ast.Fodder{},
																																TailStrictFodder: nil,
																															},
																														},
																													},
																													Fun: nil,
																													CloseFodder: ast.Fodder{},
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(569),
																															Column: int(13),
																														},
																														End: ast.Location{
																															Line: int(574),
																															Column: int(36),
																														},
																														File: p1,
																													},
																												},
																											},
																											Body: &ast.Local{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(575),
																															Column: int(7),
																														},
																														End: ast.Location{
																															Line: int(576),
																															Column: int(85),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(6),
																															Comment: []string{},
																														},
																													},
																													Ctx: p6068,
																													FreeVars: ast.Identifiers{
																														"blank",
																														"ensure_pt",
																														"mantissa",
																														"plus",
																														"prec",
																														"render_float_dec",
																														"std",
																														"suff",
																														"trailing",
																														"zero_pad",
																													},
																												},
																												Binds: ast.LocalBinds{
																													ast.LocalBind{
																														VarFodder: ast.Fodder{},
																														Variable: "zp2",
																														EqFodder: ast.Fodder{},
																														Body: &ast.Binary{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(575),
																																		Column: int(19),
																																	},
																																	End: ast.Location{
																																		Line: int(575),
																																		Column: int(46),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p6224,
																																FreeVars: ast.Identifiers{
																																	"std",
																																	"suff",
																																	"zero_pad",
																																},
																															},
																															Left: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(575),
																																			Column: int(19),
																																		},
																																		End: ast.Location{
																																			Line: int(575),
																																			Column: int(27),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p6224,
																																	FreeVars: ast.Identifiers{
																																		"zero_pad",
																																	},
																																},
																																Id: "zero_pad",
																															},
																															OpFodder: ast.Fodder{},
																															Op: ast.BinaryOp(4),
																															Right: &ast.Apply{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(575),
																																			Column: int(30),
																																		},
																																		End: ast.Location{
																																			Line: int(575),
																																			Column: int(46),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p6224,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																		"suff",
																																	},
																																},
																																Target: &ast.Index{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(575),
																																				Column: int(30),
																																			},
																																			End: ast.Location{
																																				Line: int(575),
																																				Column: int(40),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p6224,
																																		FreeVars: ast.Identifiers{
																																			"std",
																																		},
																																	},
																																	Target: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(575),
																																					Column: int(30),
																																				},
																																				End: ast.Location{
																																					Line: int(575),
																																					Column: int(33),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{
																																				"std",
																																			},
																																		},
																																		Id: "std",
																																	},
																																	LeftBracketFodder: ast.Fodder{},
																																	Index: &ast.LiteralString{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				End: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				File: nil,
																																			},
																																			Fodder: nil,
																																			Ctx: nil,
																																			FreeVars: nil,
																																		},
																																		Value: "length",
																																		Kind: ast.LiteralStringKind(1),
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																	},
																																	RightBracketFodder: ast.Fodder{},
																																	Id: nil,
																																},
																																FodderLeft: ast.Fodder{},
																																Arguments: ast.Arguments{
																																	Positional: []ast.CommaSeparatedExpr{
																																		ast.CommaSeparatedExpr{
																																			Expr: &ast.Var{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(575),
																																							Column: int(41),
																																						},
																																						End: ast.Location{
																																							Line: int(575),
																																							Column: int(45),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p6237,
																																					FreeVars: ast.Identifiers{
																																						"suff",
																																					},
																																				},
																																				Id: "suff",
																																			},
																																			CommaFodder: nil,
																																		},
																																	},
																																	Named: nil,
																																},
																																TrailingComma: false,
																																TailStrict: false,
																																FodderRight: ast.Fodder{},
																																TailStrictFodder: nil,
																															},
																														},
																														Fun: nil,
																														CloseFodder: ast.Fodder{},
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(575),
																																Column: int(13),
																															},
																															End: ast.Location{
																																Line: int(575),
																																Column: int(46),
																															},
																															File: p1,
																														},
																													},
																												},
																												Body: &ast.Binary{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(576),
																																Column: int(7),
																															},
																															End: ast.Location{
																																Line: int(576),
																																Column: int(85),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p6068,
																														FreeVars: ast.Identifiers{
																															"blank",
																															"ensure_pt",
																															"mantissa",
																															"plus",
																															"prec",
																															"render_float_dec",
																															"suff",
																															"trailing",
																															"zp2",
																														},
																													},
																													Left: &ast.Apply{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(576),
																																	Column: int(7),
																																},
																																End: ast.Location{
																																	Line: int(576),
																																	Column: int(78),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p6068,
																															FreeVars: ast.Identifiers{
																																"blank",
																																"ensure_pt",
																																"mantissa",
																																"plus",
																																"prec",
																																"render_float_dec",
																																"trailing",
																																"zp2",
																															},
																														},
																														Target: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(576),
																																		Column: int(7),
																																	},
																																	End: ast.Location{
																																		Line: int(576),
																																		Column: int(23),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{
																																	ast.FodderElement{
																																		Kind: ast.FodderKind(0),
																																		Blanks: int(0),
																																		Indent: int(6),
																																		Comment: []string{},
																																	},
																																},
																																Ctx: p6068,
																																FreeVars: ast.Identifiers{
																																	"render_float_dec",
																																},
																															},
																															Id: "render_float_dec",
																														},
																														FodderLeft: ast.Fodder{},
																														Arguments: ast.Arguments{
																															Positional: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(576),
																																					Column: int(24),
																																				},
																																				End: ast.Location{
																																					Line: int(576),
																																					Column: int(32),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p6248,
																																			FreeVars: ast.Identifiers{
																																				"mantissa",
																																			},
																																		},
																																		Id: "mantissa",
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(576),
																																					Column: int(34),
																																				},
																																				End: ast.Location{
																																					Line: int(576),
																																					Column: int(37),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p6248,
																																			FreeVars: ast.Identifiers{
																																				"zp2",
																																			},
																																		},
																																		Id: "zp2",
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(576),
																																					Column: int(39),
																																				},
																																				End: ast.Location{
																																					Line: int(576),
																																					Column: int(44),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p6248,
																																			FreeVars: ast.Identifiers{
																																				"blank",
																																			},
																																		},
																																		Id: "blank",
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(576),
																																					Column: int(46),
																																				},
																																				End: ast.Location{
																																					Line: int(576),
																																					Column: int(50),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p6248,
																																			FreeVars: ast.Identifiers{
																																				"plus",
																																			},
																																		},
																																		Id: "plus",
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(576),
																																					Column: int(52),
																																				},
																																				End: ast.Location{
																																					Line: int(576),
																																					Column: int(61),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p6248,
																																			FreeVars: ast.Identifiers{
																																				"ensure_pt",
																																			},
																																		},
																																		Id: "ensure_pt",
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(576),
																																					Column: int(63),
																																				},
																																				End: ast.Location{
																																					Line: int(576),
																																					Column: int(71),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p6248,
																																			FreeVars: ast.Identifiers{
																																				"trailing",
																																			},
																																		},
																																		Id: "trailing",
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(576),
																																					Column: int(73),
																																				},
																																				End: ast.Location{
																																					Line: int(576),
																																					Column: int(77),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p6248,
																																			FreeVars: ast.Identifiers{
																																				"prec",
																																			},
																																		},
																																		Id: "prec",
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															Named: nil,
																														},
																														TrailingComma: false,
																														TailStrict: false,
																														FodderRight: ast.Fodder{},
																														TailStrictFodder: nil,
																													},
																													OpFodder: ast.Fodder{},
																													Op: ast.BinaryOp(3),
																													Right: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(576),
																																	Column: int(81),
																																},
																																End: ast.Location{
																																	Line: int(576),
																																	Column: int(85),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p6068,
																															FreeVars: ast.Identifiers{
																																"suff",
																															},
																														},
																														Id: "suff",
																													},
																												},
																											},
																										},
																									},
																								},
																							},
																							Fun: nil,
																							CloseFodder: nil,
																							LocRange: ast.LocationRange{
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								File: nil,
																							},
																						},
																					},
																					Body: &ast.Local{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(579),
																									Column: int(5),
																								},
																								End: ast.Location{
																									Line: int(774),
																									Column: int(48),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{
																								ast.FodderElement{
																									Kind: ast.FodderKind(0),
																									Blanks: int(1),
																									Indent: int(4),
																									Comment: []string{},
																								},
																								ast.FodderElement{
																									Kind: ast.FodderKind(2),
																									Blanks: int(0),
																									Indent: int(4),
																									Comment: []string{
																										"// Render a value with an arbitrary format code.",
																									},
																								},
																							},
																							Ctx: p3497,
																							FreeVars: ast.Identifiers{
																								"codes",
																								"pad_left",
																								"pad_right",
																								"render_float_dec",
																								"render_float_sci",
																								"render_hex",
																								"render_int",
																								"std",
																								"vals",
																							},
																						},
																						Binds: ast.LocalBinds{
																							ast.LocalBind{
																								VarFodder: nil,
																								Variable: "format_code",
																								EqFodder: nil,
																								Body: &ast.Function{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(579),
																												Column: int(11),
																											},
																											End: ast.Location{
																												Line: int(671),
																												Column: int(44),
																											},
																											File: p1,
																										},
																										Fodder: nil,
																										Ctx: p6270,
																										FreeVars: ast.Identifiers{
																											"render_float_dec",
																											"render_float_sci",
																											"render_hex",
																											"render_int",
																											"std",
																										},
																									},
																									ParenLeftFodder: ast.Fodder{},
																									Parameters: []ast.Parameter{
																										ast.Parameter{
																											NameFodder: ast.Fodder{},
																											Name: "val",
																											EqFodder: nil,
																											DefaultArg: nil,
																											CommaFodder: ast.Fodder{},
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(579),
																													Column: int(23),
																												},
																												End: ast.Location{
																													Line: int(579),
																													Column: int(26),
																												},
																												File: p1,
																											},
																										},
																										ast.Parameter{
																											NameFodder: ast.Fodder{},
																											Name: "code",
																											EqFodder: nil,
																											DefaultArg: nil,
																											CommaFodder: ast.Fodder{},
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(579),
																													Column: int(28),
																												},
																												End: ast.Location{
																													Line: int(579),
																													Column: int(32),
																												},
																												File: p1,
																											},
																										},
																										ast.Parameter{
																											NameFodder: ast.Fodder{},
																											Name: "fw",
																											EqFodder: nil,
																											DefaultArg: nil,
																											CommaFodder: ast.Fodder{},
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(579),
																													Column: int(34),
																												},
																												End: ast.Location{
																													Line: int(579),
																													Column: int(36),
																												},
																												File: p1,
																											},
																										},
																										ast.Parameter{
																											NameFodder: ast.Fodder{},
																											Name: "prec_or_null",
																											EqFodder: nil,
																											DefaultArg: nil,
																											CommaFodder: ast.Fodder{},
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(579),
																													Column: int(38),
																												},
																												End: ast.Location{
																													Line: int(579),
																													Column: int(50),
																												},
																												File: p1,
																											},
																										},
																										ast.Parameter{
																											NameFodder: ast.Fodder{},
																											Name: "i",
																											EqFodder: nil,
																											DefaultArg: nil,
																											CommaFodder: nil,
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(579),
																													Column: int(52),
																												},
																												End: ast.Location{
																													Line: int(579),
																													Column: int(53),
																												},
																												File: p1,
																											},
																										},
																									},
																									TrailingComma: false,
																									ParenRightFodder: ast.Fodder{},
																									Body: &ast.Local{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(580),
																													Column: int(7),
																												},
																												End: ast.Location{
																													Line: int(671),
																													Column: int(44),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{
																												ast.FodderElement{
																													Kind: ast.FodderKind(0),
																													Blanks: int(0),
																													Indent: int(6),
																													Comment: []string{},
																												},
																											},
																											Ctx: p6275,
																											FreeVars: ast.Identifiers{
																												"code",
																												"fw",
																												"i",
																												"prec_or_null",
																												"render_float_dec",
																												"render_float_sci",
																												"render_hex",
																												"render_int",
																												"std",
																												"val",
																											},
																										},
																										Binds: ast.LocalBinds{
																											ast.LocalBind{
																												VarFodder: ast.Fodder{},
																												Variable: "cflags",
																												EqFodder: ast.Fodder{},
																												Body: &ast.Index{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(580),
																																Column: int(22),
																															},
																															End: ast.Location{
																																Line: int(580),
																																Column: int(33),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p6279,
																														FreeVars: ast.Identifiers{
																															"code",
																														},
																													},
																													Target: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(580),
																																	Column: int(22),
																																},
																																End: ast.Location{
																																	Line: int(580),
																																	Column: int(26),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: nil,
																															FreeVars: ast.Identifiers{
																																"code",
																															},
																														},
																														Id: "code",
																													},
																													LeftBracketFodder: ast.Fodder{},
																													Index: &ast.LiteralString{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																End: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																File: nil,
																															},
																															Fodder: nil,
																															Ctx: nil,
																															FreeVars: nil,
																														},
																														Value: "cflags",
																														Kind: ast.LiteralStringKind(1),
																														BlockIndent: "",
																														BlockTermIndent: "",
																													},
																													RightBracketFodder: ast.Fodder{},
																													Id: nil,
																												},
																												Fun: nil,
																												CloseFodder: ast.Fodder{},
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(580),
																														Column: int(13),
																													},
																													End: ast.Location{
																														Line: int(580),
																														Column: int(33),
																													},
																													File: p1,
																												},
																											},
																										},
																										Body: &ast.Local{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(581),
																														Column: int(7),
																													},
																													End: ast.Location{
																														Line: int(671),
																														Column: int(44),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{
																													ast.FodderElement{
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(6),
																														Comment: []string{},
																													},
																												},
																												Ctx: p6275,
																												FreeVars: ast.Identifiers{
																													"cflags",
																													"code",
																													"fw",
																													"i",
																													"prec_or_null",
																													"render_float_dec",
																													"render_float_sci",
																													"render_hex",
																													"render_int",
																													"std",
																													"val",
																												},
																											},
																											Binds: ast.LocalBinds{
																												ast.LocalBind{
																													VarFodder: ast.Fodder{},
																													Variable: "fpprec",
																													EqFodder: ast.Fodder{},
																													Body: &ast.Conditional{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(581),
																																	Column: int(22),
																																},
																																End: ast.Location{
																																	Line: int(581),
																																	Column: int(70),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p6289,
																															FreeVars: ast.Identifiers{
																																"prec_or_null",
																															},
																														},
																														Cond: &ast.Binary{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(581),
																																		Column: int(25),
																																	},
																																	End: ast.Location{
																																		Line: int(581),
																																		Column: int(45),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p6289,
																																FreeVars: ast.Identifiers{
																																	"prec_or_null",
																																},
																															},
																															Left: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(581),
																																			Column: int(25),
																																		},
																																		End: ast.Location{
																																			Line: int(581),
																																			Column: int(37),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p6289,
																																	FreeVars: ast.Identifiers{
																																		"prec_or_null",
																																	},
																																},
																																Id: "prec_or_null",
																															},
																															OpFodder: ast.Fodder{},
																															Op: ast.BinaryOp(13),
																															Right: &ast.LiteralNull{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(581),
																																			Column: int(41),
																																		},
																																		End: ast.Location{
																																			Line: int(581),
																																			Column: int(45),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p6289,
																																	FreeVars: nil,
																																},
																															},
																														},
																														ThenFodder: ast.Fodder{},
																														BranchTrue: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(581),
																																		Column: int(51),
																																	},
																																	End: ast.Location{
																																		Line: int(581),
																																		Column: int(63),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p6289,
																																FreeVars: ast.Identifiers{
																																	"prec_or_null",
																																},
																															},
																															Id: "prec_or_null",
																														},
																														ElseFodder: ast.Fodder{},
																														BranchFalse: &ast.LiteralNumber{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(581),
																																		Column: int(69),
																																	},
																																	End: ast.Location{
																																		Line: int(581),
																																		Column: int(70),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p6289,
																																FreeVars: nil,
																															},
																															OriginalString: "6",
																														},
																													},
																													Fun: nil,
																													CloseFodder: ast.Fodder{},
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(581),
																															Column: int(13),
																														},
																														End: ast.Location{
																															Line: int(581),
																															Column: int(70),
																														},
																														File: p1,
																													},
																												},
																											},
																											Body: &ast.Local{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(582),
																															Column: int(7),
																														},
																														End: ast.Location{
																															Line: int(671),
																															Column: int(44),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(6),
																															Comment: []string{},
																														},
																													},
																													Ctx: p6275,
																													FreeVars: ast.Identifiers{
																														"cflags",
																														"code",
																														"fpprec",
																														"fw",
																														"i",
																														"prec_or_null",
																														"render_float_dec",
																														"render_float_sci",
																														"render_hex",
																														"render_int",
																														"std",
																														"val",
																													},
																												},
																												Binds: ast.LocalBinds{
																													ast.LocalBind{
																														VarFodder: ast.Fodder{},
																														Variable: "iprec",
																														EqFodder: ast.Fodder{},
																														Body: &ast.Conditional{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(582),
																																		Column: int(21),
																																	},
																																	End: ast.Location{
																																		Line: int(582),
																																		Column: int(69),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p6304,
																																FreeVars: ast.Identifiers{
																																	"prec_or_null",
																																},
																															},
																															Cond: &ast.Binary{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(582),
																																			Column: int(24),
																																		},
																																		End: ast.Location{
																																			Line: int(582),
																																			Column: int(44),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p6304,
																																	FreeVars: ast.Identifiers{
																																		"prec_or_null",
																																	},
																																},
																																Left: &ast.Var{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(582),
																																				Column: int(24),
																																			},
																																			End: ast.Location{
																																				Line: int(582),
																																				Column: int(36),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p6304,
																																		FreeVars: ast.Identifiers{
																																			"prec_or_null",
																																		},
																																	},
																																	Id: "prec_or_null",
																																},
																																OpFodder: ast.Fodder{},
																																Op: ast.BinaryOp(13),
																																Right: &ast.LiteralNull{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(582),
																																				Column: int(40),
																																			},
																																			End: ast.Location{
																																				Line: int(582),
																																				Column: int(44),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p6304,
																																		FreeVars: nil,
																																	},
																																},
																															},
																															ThenFodder: ast.Fodder{},
																															BranchTrue: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(582),
																																			Column: int(50),
																																		},
																																		End: ast.Location{
																																			Line: int(582),
																																			Column: int(62),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p6304,
																																	FreeVars: ast.Identifiers{
																																		"prec_or_null",
																																	},
																																},
																																Id: "prec_or_null",
																															},
																															ElseFodder: ast.Fodder{},
																															BranchFalse: &ast.LiteralNumber{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(582),
																																			Column: int(68),
																																		},
																																		End: ast.Location{
																																			Line: int(582),
																																			Column: int(69),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p6304,
																																	FreeVars: nil,
																																},
																																OriginalString: "0",
																															},
																														},
																														Fun: nil,
																														CloseFodder: ast.Fodder{},
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(582),
																																Column: int(13),
																															},
																															End: ast.Location{
																																Line: int(582),
																																Column: int(69),
																															},
																															File: p1,
																														},
																													},
																												},
																												Body: &ast.Local{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(583),
																																Column: int(7),
																															},
																															End: ast.Location{
																																Line: int(671),
																																Column: int(44),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(6),
																																Comment: []string{},
																															},
																														},
																														Ctx: p6275,
																														FreeVars: ast.Identifiers{
																															"cflags",
																															"code",
																															"fpprec",
																															"fw",
																															"i",
																															"iprec",
																															"render_float_dec",
																															"render_float_sci",
																															"render_hex",
																															"render_int",
																															"std",
																															"val",
																														},
																													},
																													Binds: ast.LocalBinds{
																														ast.LocalBind{
																															VarFodder: ast.Fodder{},
																															Variable: "zp",
																															EqFodder: ast.Fodder{},
																															Body: &ast.Conditional{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(583),
																																			Column: int(18),
																																		},
																																		End: ast.Location{
																																			Line: int(583),
																																			Column: int(63),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p6319,
																																	FreeVars: ast.Identifiers{
																																		"cflags",
																																		"fw",
																																	},
																																},
																																Cond: &ast.Binary{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(583),
																																				Column: int(21),
																																			},
																																			End: ast.Location{
																																				Line: int(583),
																																				Column: int(48),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p6319,
																																		FreeVars: ast.Identifiers{
																																			"cflags",
																																		},
																																	},
																																	Left: &ast.Index{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(583),
																																					Column: int(21),
																																				},
																																				End: ast.Location{
																																					Line: int(583),
																																					Column: int(32),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p6319,
																																			FreeVars: ast.Identifiers{
																																				"cflags",
																																			},
																																		},
																																		Target: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(583),
																																						Column: int(21),
																																					},
																																					End: ast.Location{
																																						Line: int(583),
																																						Column: int(27),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: nil,
																																				FreeVars: ast.Identifiers{
																																					"cflags",
																																				},
																																			},
																																			Id: "cflags",
																																		},
																																		LeftBracketFodder: ast.Fodder{},
																																		Index: &ast.LiteralString{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(0),
																																						Column: int(0),
																																					},
																																					End: ast.Location{
																																						Line: int(0),
																																						Column: int(0),
																																					},
																																					File: nil,
																																				},
																																				Fodder: nil,
																																				Ctx: nil,
																																				FreeVars: nil,
																																			},
																																			Value: "zero",
																																			Kind: ast.LiteralStringKind(1),
																																			BlockIndent: "",
																																			BlockTermIndent: "",
																																		},
																																		RightBracketFodder: ast.Fodder{},
																																		Id: nil,
																																	},
																																	OpFodder: ast.Fodder{},
																																	Op: ast.BinaryOp(17),
																																	Right: &ast.Unary{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(583),
																																					Column: int(36),
																																				},
																																				End: ast.Location{
																																					Line: int(583),
																																					Column: int(48),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p6319,
																																			FreeVars: ast.Identifiers{
																																				"cflags",
																																			},
																																		},
																																		Op: ast.UnaryOp(0),
																																		Expr: &ast.Index{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(583),
																																						Column: int(37),
																																					},
																																					End: ast.Location{
																																						Line: int(583),
																																						Column: int(48),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p6319,
																																				FreeVars: ast.Identifiers{
																																					"cflags",
																																				},
																																			},
																																			Target: &ast.Var{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(583),
																																							Column: int(37),
																																						},
																																						End: ast.Location{
																																							Line: int(583),
																																							Column: int(43),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{
																																						"cflags",
																																					},
																																				},
																																				Id: "cflags",
																																			},
																																			LeftBracketFodder: ast.Fodder{},
																																			Index: &ast.LiteralString{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						End: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						File: nil,
																																					},
																																					Fodder: nil,
																																					Ctx: nil,
																																					FreeVars: nil,
																																				},
																																				Value: "left",
																																				Kind: ast.LiteralStringKind(1),
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																			},
																																			RightBracketFodder: ast.Fodder{},
																																			Id: nil,
																																		},
																																	},
																																},
																																ThenFodder: ast.Fodder{},
																																BranchTrue: &ast.Var{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(583),
																																				Column: int(54),
																																			},
																																			End: ast.Location{
																																				Line: int(583),
																																				Column: int(56),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p6319,
																																		FreeVars: ast.Identifiers{
																																			"fw",
																																		},
																																	},
																																	Id: "fw",
																																},
																																ElseFodder: ast.Fodder{},
																																BranchFalse: &ast.LiteralNumber{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(583),
																																				Column: int(62),
																																			},
																																			End: ast.Location{
																																				Line: int(583),
																																				Column: int(63),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p6319,
																																		FreeVars: nil,
																																	},
																																	OriginalString: "0",
																																},
																															},
																															Fun: nil,
																															CloseFodder: ast.Fodder{},
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(583),
																																	Column: int(13),
																																},
																																End: ast.Location{
																																	Line: int(583),
																																	Column: int(63),
																																},
																																File: p1,
																															},
																														},
																													},
																													Body: &ast.Conditional{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(584),
																																	Column: int(7),
																																},
																																End: ast.Location{
																																	Line: int(671),
																																	Column: int(44),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{
																																ast.FodderElement{
																																	Kind: ast.FodderKind(0),
																																	Blanks: int(0),
																																	Indent: int(6),
																																	Comment: []string{},
																																},
																															},
																															Ctx: p6275,
																															FreeVars: ast.Identifiers{
																																"cflags",
																																"code",
																																"fpprec",
																																"i",
																																"iprec",
																																"render_float_dec",
																																"render_float_sci",
																																"render_hex",
																																"render_int",
																																"std",
																																"val",
																																"zp",
																															},
																														},
																														Cond: &ast.Binary{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(584),
																																		Column: int(10),
																																	},
																																	End: ast.Location{
																																		Line: int(584),
																																		Column: int(27),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p6275,
																																FreeVars: ast.Identifiers{
																																	"code",
																																},
																															},
																															Left: &ast.Index{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(584),
																																			Column: int(10),
																																		},
																																		End: ast.Location{
																																			Line: int(584),
																																			Column: int(20),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p6275,
																																	FreeVars: ast.Identifiers{
																																		"code",
																																	},
																																},
																																Target: &ast.Var{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(584),
																																				Column: int(10),
																																			},
																																			End: ast.Location{
																																				Line: int(584),
																																				Column: int(14),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{
																																			"code",
																																		},
																																	},
																																	Id: "code",
																																},
																																LeftBracketFodder: ast.Fodder{},
																																Index: &ast.LiteralString{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			End: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			File: nil,
																																		},
																																		Fodder: nil,
																																		Ctx: nil,
																																		FreeVars: nil,
																																	},
																																	Value: "ctype",
																																	Kind: ast.LiteralStringKind(1),
																																	BlockIndent: "",
																																	BlockTermIndent: "",
																																},
																																RightBracketFodder: ast.Fodder{},
																																Id: nil,
																															},
																															OpFodder: ast.Fodder{},
																															Op: ast.BinaryOp(12),
																															Right: &ast.LiteralString{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(584),
																																			Column: int(24),
																																		},
																																		End: ast.Location{
																																			Line: int(584),
																																			Column: int(27),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p6275,
																																	FreeVars: nil,
																																},
																																Value: "s",
																																Kind: ast.LiteralStringKind(1),
																																BlockIndent: "",
																																BlockTermIndent: "",
																															},
																														},
																														ThenFodder: ast.Fodder{},
																														BranchTrue: &ast.Apply{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(585),
																																		Column: int(9),
																																	},
																																	End: ast.Location{
																																		Line: int(585),
																																		Column: int(26),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p6275,
																																FreeVars: ast.Identifiers{
																																	"std",
																																	"val",
																																},
																															},
																															Target: &ast.Index{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(585),
																																			Column: int(9),
																																		},
																																		End: ast.Location{
																																			Line: int(585),
																																			Column: int(21),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p6275,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																	},
																																},
																																Target: &ast.Var{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(585),
																																				Column: int(9),
																																			},
																																			End: ast.Location{
																																				Line: int(585),
																																				Column: int(12),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{
																																			ast.FodderElement{
																																				Kind: ast.FodderKind(0),
																																				Blanks: int(0),
																																				Indent: int(8),
																																				Comment: []string{},
																																			},
																																		},
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{
																																			"std",
																																		},
																																	},
																																	Id: "std",
																																},
																																LeftBracketFodder: ast.Fodder{},
																																Index: &ast.LiteralString{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			End: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			File: nil,
																																		},
																																		Fodder: nil,
																																		Ctx: nil,
																																		FreeVars: nil,
																																	},
																																	Value: "toString",
																																	Kind: ast.LiteralStringKind(1),
																																	BlockIndent: "",
																																	BlockTermIndent: "",
																																},
																																RightBracketFodder: ast.Fodder{},
																																Id: nil,
																															},
																															FodderLeft: ast.Fodder{},
																															Arguments: ast.Arguments{
																																Positional: []ast.CommaSeparatedExpr{
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(585),
																																						Column: int(22),
																																					},
																																					End: ast.Location{
																																						Line: int(585),
																																						Column: int(25),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p6359,
																																				FreeVars: ast.Identifiers{
																																					"val",
																																				},
																																			},
																																			Id: "val",
																																		},
																																		CommaFodder: nil,
																																	},
																																},
																																Named: nil,
																															},
																															TrailingComma: false,
																															TailStrict: false,
																															FodderRight: ast.Fodder{},
																															TailStrictFodder: nil,
																														},
																														ElseFodder: ast.Fodder{
																															ast.FodderElement{
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(6),
																																Comment: []string{},
																															},
																														},
																														BranchFalse: &ast.Conditional{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(586),
																																		Column: int(12),
																																	},
																																	End: ast.Location{
																																		Line: int(671),
																																		Column: int(44),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p6275,
																																FreeVars: ast.Identifiers{
																																	"cflags",
																																	"code",
																																	"fpprec",
																																	"i",
																																	"iprec",
																																	"render_float_dec",
																																	"render_float_sci",
																																	"render_hex",
																																	"render_int",
																																	"std",
																																	"val",
																																	"zp",
																																},
																															},
																															Cond: &ast.Binary{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(586),
																																			Column: int(15),
																																		},
																																		End: ast.Location{
																																			Line: int(586),
																																			Column: int(32),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p6275,
																																	FreeVars: ast.Identifiers{
																																		"code",
																																	},
																																},
																																Left: &ast.Index{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(586),
																																				Column: int(15),
																																			},
																																			End: ast.Location{
																																				Line: int(586),
																																				Column: int(25),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p6275,
																																		FreeVars: ast.Identifiers{
																																			"code",
																																		},
																																	},
																																	Target: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(586),
																																					Column: int(15),
																																				},
																																				End: ast.Location{
																																					Line: int(586),
																																					Column: int(19),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{
																																				"code",
																																			},
																																		},
																																		Id: "code",
																																	},
																																	LeftBracketFodder: ast.Fodder{},
																																	Index: &ast.LiteralString{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				End: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				File: nil,
																																			},
																																			Fodder: nil,
																																			Ctx: nil,
																																			FreeVars: nil,
																																		},
																																		Value: "ctype",
																																		Kind: ast.LiteralStringKind(1),
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																	},
																																	RightBracketFodder: ast.Fodder{},
																																	Id: nil,
																																},
																																OpFodder: ast.Fodder{},
																																Op: ast.BinaryOp(12),
																																Right: &ast.LiteralString{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(586),
																																				Column: int(29),
																																			},
																																			End: ast.Location{
																																				Line: int(586),
																																				Column: int(32),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p6275,
																																		FreeVars: nil,
																																	},
																																	Value: "d",
																																	Kind: ast.LiteralStringKind(1),
																																	BlockIndent: "",
																																	BlockTermIndent: "",
																																},
																															},
																															ThenFodder: ast.Fodder{},
																															BranchTrue: &ast.Conditional{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(587),
																																			Column: int(9),
																																		},
																																		End: ast.Location{
																																			Line: int(591),
																																			Column: int(103),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{
																																		ast.FodderElement{
																																			Kind: ast.FodderKind(0),
																																			Blanks: int(0),
																																			Indent: int(8),
																																			Comment: []string{},
																																		},
																																	},
																																	Ctx: p6275,
																																	FreeVars: ast.Identifiers{
																																		"cflags",
																																		"i",
																																		"iprec",
																																		"render_int",
																																		"std",
																																		"val",
																																		"zp",
																																	},
																																},
																																Cond: &ast.Binary{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(587),
																																				Column: int(12),
																																			},
																																			End: ast.Location{
																																				Line: int(587),
																																				Column: int(37),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p6275,
																																		FreeVars: ast.Identifiers{
																																			"std",
																																			"val",
																																		},
																																	},
																																	Left: &ast.Apply{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(587),
																																					Column: int(12),
																																				},
																																				End: ast.Location{
																																					Line: int(587),
																																					Column: int(25),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p6275,
																																			FreeVars: ast.Identifiers{
																																				"std",
																																				"val",
																																			},
																																		},
																																		Target: &ast.Index{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(587),
																																						Column: int(12),
																																					},
																																					End: ast.Location{
																																						Line: int(587),
																																						Column: int(20),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p6275,
																																				FreeVars: ast.Identifiers{
																																					"std",
																																				},
																																			},
																																			Target: &ast.Var{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(587),
																																							Column: int(12),
																																						},
																																						End: ast.Location{
																																							Line: int(587),
																																							Column: int(15),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{
																																						"std",
																																					},
																																				},
																																				Id: "std",
																																			},
																																			LeftBracketFodder: ast.Fodder{},
																																			Index: &ast.LiteralString{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						End: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						File: nil,
																																					},
																																					Fodder: nil,
																																					Ctx: nil,
																																					FreeVars: nil,
																																				},
																																				Value: "type",
																																				Kind: ast.LiteralStringKind(1),
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																			},
																																			RightBracketFodder: ast.Fodder{},
																																			Id: nil,
																																		},
																																		FodderLeft: ast.Fodder{},
																																		Arguments: ast.Arguments{
																																			Positional: []ast.CommaSeparatedExpr{
																																				ast.CommaSeparatedExpr{
																																					Expr: &ast.Var{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(587),
																																									Column: int(21),
																																								},
																																								End: ast.Location{
																																									Line: int(587),
																																									Column: int(24),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p6386,
																																							FreeVars: ast.Identifiers{
																																								"val",
																																							},
																																						},
																																						Id: "val",
																																					},
																																					CommaFodder: nil,
																																				},
																																			},
																																			Named: nil,
																																		},
																																		TrailingComma: false,
																																		TailStrict: false,
																																		FodderRight: ast.Fodder{},
																																		TailStrictFodder: nil,
																																	},
																																	OpFodder: ast.Fodder{},
																																	Op: ast.BinaryOp(13),
																																	Right: &ast.LiteralString{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(587),
																																					Column: int(29),
																																				},
																																				End: ast.Location{
																																					Line: int(587),
																																					Column: int(37),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p6275,
																																			FreeVars: nil,
																																		},
																																		Value: "number",
																																		Kind: ast.LiteralStringKind(1),
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																	},
																																},
																																ThenFodder: ast.Fodder{},
																																BranchTrue: &ast.Error{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(588),
																																				Column: int(11),
																																			},
																																			End: ast.Location{
																																				Line: int(589),
																																				Column: int(47),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{
																																			ast.FodderElement{
																																				Kind: ast.FodderKind(0),
																																				Blanks: int(0),
																																				Indent: int(10),
																																				Comment: []string{},
																																			},
																																		},
																																		Ctx: p6275,
																																		FreeVars: ast.Identifiers{
																																			"i",
																																			"std",
																																			"val",
																																		},
																																	},
																																	Expr: &ast.Binary{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(588),
																																					Column: int(17),
																																				},
																																				End: ast.Location{
																																					Line: int(589),
																																					Column: int(47),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p6275,
																																			FreeVars: ast.Identifiers{
																																				"i",
																																				"std",
																																				"val",
																																			},
																																		},
																																		Left: &ast.Binary{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(588),
																																						Column: int(17),
																																					},
																																					End: ast.Location{
																																						Line: int(589),
																																						Column: int(31),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p6275,
																																				FreeVars: ast.Identifiers{
																																					"i",
																																				},
																																			},
																																			Left: &ast.Binary{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(588),
																																							Column: int(17),
																																						},
																																						End: ast.Location{
																																							Line: int(589),
																																							Column: int(20),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p6275,
																																					FreeVars: ast.Identifiers{
																																						"i",
																																					},
																																				},
																																				Left: &ast.LiteralString{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(588),
																																								Column: int(17),
																																							},
																																							End: ast.Location{
																																								Line: int(588),
																																								Column: int(45),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p6275,
																																						FreeVars: nil,
																																					},
																																					Value: "Format required number at ",
																																					Kind: ast.LiteralStringKind(1),
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																				},
																																				OpFodder: ast.Fodder{
																																					ast.FodderElement{
																																						Kind: ast.FodderKind(0),
																																						Blanks: int(0),
																																						Indent: int(16),
																																						Comment: []string{},
																																					},
																																				},
																																				Op: ast.BinaryOp(3),
																																				Right: &ast.Var{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(589),
																																								Column: int(19),
																																							},
																																							End: ast.Location{
																																								Line: int(589),
																																								Column: int(20),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p6275,
																																						FreeVars: ast.Identifiers{
																																							"i",
																																						},
																																					},
																																					Id: "i",
																																				},
																																			},
																																			OpFodder: ast.Fodder{},
																																			Op: ast.BinaryOp(3),
																																			Right: &ast.LiteralString{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(589),
																																							Column: int(23),
																																						},
																																						End: ast.Location{
																																							Line: int(589),
																																							Column: int(31),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p6275,
																																					FreeVars: nil,
																																				},
																																				Value: ", got ",
																																				Kind: ast.LiteralStringKind(1),
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																			},
																																		},
																																		OpFodder: ast.Fodder{},
																																		Op: ast.BinaryOp(3),
																																		Right: &ast.Apply{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(589),
																																						Column: int(34),
																																					},
																																					End: ast.Location{
																																						Line: int(589),
																																						Column: int(47),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p6275,
																																				FreeVars: ast.Identifiers{
																																					"std",
																																					"val",
																																				},
																																			},
																																			Target: &ast.Index{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(589),
																																							Column: int(34),
																																						},
																																						End: ast.Location{
																																							Line: int(589),
																																							Column: int(42),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p6275,
																																					FreeVars: ast.Identifiers{
																																						"std",
																																					},
																																				},
																																				Target: &ast.Var{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(589),
																																								Column: int(34),
																																							},
																																							End: ast.Location{
																																								Line: int(589),
																																								Column: int(37),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{
																																							"std",
																																						},
																																					},
																																					Id: "std",
																																				},
																																				LeftBracketFodder: ast.Fodder{},
																																				Index: &ast.LiteralString{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							End: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							File: nil,
																																						},
																																						Fodder: nil,
																																						Ctx: nil,
																																						FreeVars: nil,
																																					},
																																					Value: "type",
																																					Kind: ast.LiteralStringKind(1),
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																				},
																																				RightBracketFodder: ast.Fodder{},
																																				Id: nil,
																																			},
																																			FodderLeft: ast.Fodder{},
																																			Arguments: ast.Arguments{
																																				Positional: []ast.CommaSeparatedExpr{
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(589),
																																										Column: int(43),
																																									},
																																									End: ast.Location{
																																										Line: int(589),
																																										Column: int(46),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6412,
																																								FreeVars: ast.Identifiers{
																																									"val",
																																								},
																																							},
																																							Id: "val",
																																						},
																																						CommaFodder: nil,
																																					},
																																				},
																																				Named: nil,
																																			},
																																			TrailingComma: false,
																																			TailStrict: false,
																																			FodderRight: ast.Fodder{},
																																			TailStrictFodder: nil,
																																		},
																																	},
																																},
																																ElseFodder: ast.Fodder{
																																	ast.FodderElement{
																																		Kind: ast.FodderKind(0),
																																		Blanks: int(0),
																																		Indent: int(8),
																																		Comment: []string{},
																																	},
																																},
																																BranchFalse: &ast.Apply{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(591),
																																				Column: int(11),
																																			},
																																			End: ast.Location{
																																				Line: int(591),
																																				Column: int(103),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p6275,
																																		FreeVars: ast.Identifiers{
																																			"cflags",
																																			"iprec",
																																			"render_int",
																																			"std",
																																			"val",
																																			"zp",
																																		},
																																	},
																																	Target: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(591),
																																					Column: int(11),
																																				},
																																				End: ast.Location{
																																					Line: int(591),
																																					Column: int(21),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{
																																				ast.FodderElement{
																																					Kind: ast.FodderKind(0),
																																					Blanks: int(0),
																																					Indent: int(10),
																																					Comment: []string{},
																																				},
																																			},
																																			Ctx: p6275,
																																			FreeVars: ast.Identifiers{
																																				"render_int",
																																			},
																																		},
																																		Id: "render_int",
																																	},
																																	FodderLeft: ast.Fodder{},
																																	Arguments: ast.Arguments{
																																		Positional: []ast.CommaSeparatedExpr{
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Binary{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(591),
																																								Column: int(22),
																																							},
																																							End: ast.Location{
																																								Line: int(591),
																																								Column: int(31),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p6422,
																																						FreeVars: ast.Identifiers{
																																							"val",
																																						},
																																					},
																																					Left: &ast.Var{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(591),
																																									Column: int(22),
																																								},
																																								End: ast.Location{
																																									Line: int(591),
																																									Column: int(25),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p6422,
																																							FreeVars: ast.Identifiers{
																																								"val",
																																							},
																																						},
																																						Id: "val",
																																					},
																																					OpFodder: ast.Fodder{},
																																					Op: ast.BinaryOp(10),
																																					Right: &ast.Unary{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(591),
																																									Column: int(29),
																																								},
																																								End: ast.Location{
																																									Line: int(591),
																																									Column: int(31),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p6422,
																																							FreeVars: nil,
																																						},
																																						Op: ast.UnaryOp(3),
																																						Expr: &ast.LiteralNumber{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(591),
																																										Column: int(30),
																																									},
																																									End: ast.Location{
																																										Line: int(591),
																																										Column: int(31),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6422,
																																								FreeVars: nil,
																																							},
																																							OriginalString: "1",
																																						},
																																					},
																																				},
																																				CommaFodder: ast.Fodder{},
																																			},
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Apply{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(591),
																																								Column: int(33),
																																							},
																																							End: ast.Location{
																																								Line: int(591),
																																								Column: int(56),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p6422,
																																						FreeVars: ast.Identifiers{
																																							"std",
																																							"val",
																																						},
																																					},
																																					Target: &ast.Index{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(591),
																																									Column: int(33),
																																								},
																																								End: ast.Location{
																																									Line: int(591),
																																									Column: int(42),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p6422,
																																							FreeVars: ast.Identifiers{
																																								"std",
																																							},
																																						},
																																						Target: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(591),
																																										Column: int(33),
																																									},
																																									End: ast.Location{
																																										Line: int(591),
																																										Column: int(36),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: nil,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																								},
																																							},
																																							Id: "std",
																																						},
																																						LeftBracketFodder: ast.Fodder{},
																																						Index: &ast.LiteralString{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																									End: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																									File: nil,
																																								},
																																								Fodder: nil,
																																								Ctx: nil,
																																								FreeVars: nil,
																																							},
																																							Value: "floor",
																																							Kind: ast.LiteralStringKind(1),
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																						},
																																						RightBracketFodder: ast.Fodder{},
																																						Id: nil,
																																					},
																																					FodderLeft: ast.Fodder{},
																																					Arguments: ast.Arguments{
																																						Positional: []ast.CommaSeparatedExpr{
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Apply{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(591),
																																												Column: int(43),
																																											},
																																											End: ast.Location{
																																												Line: int(591),
																																												Column: int(55),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p6437,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																											"val",
																																										},
																																									},
																																									Target: &ast.Index{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(591),
																																													Column: int(43),
																																												},
																																												End: ast.Location{
																																													Line: int(591),
																																													Column: int(50),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p6437,
																																											FreeVars: ast.Identifiers{
																																												"std",
																																											},
																																										},
																																										Target: &ast.Var{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(591),
																																														Column: int(43),
																																													},
																																													End: ast.Location{
																																														Line: int(591),
																																														Column: int(46),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: nil,
																																												FreeVars: ast.Identifiers{
																																													"std",
																																												},
																																											},
																																											Id: "std",
																																										},
																																										LeftBracketFodder: ast.Fodder{},
																																										Index: &ast.LiteralString{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(0),
																																														Column: int(0),
																																													},
																																													End: ast.Location{
																																														Line: int(0),
																																														Column: int(0),
																																													},
																																													File: nil,
																																												},
																																												Fodder: nil,
																																												Ctx: nil,
																																												FreeVars: nil,
																																											},
																																											Value: "abs",
																																											Kind: ast.LiteralStringKind(1),
																																											BlockIndent: "",
																																											BlockTermIndent: "",
																																										},
																																										RightBracketFodder: ast.Fodder{},
																																										Id: nil,
																																									},
																																									FodderLeft: ast.Fodder{},
																																									Arguments: ast.Arguments{
																																										Positional: []ast.CommaSeparatedExpr{
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Var{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "<std>",
																																															Begin: ast.Location{
																																																Line: int(591),
																																																Column: int(51),
																																															},
																																															End: ast.Location{
																																																Line: int(591),
																																																Column: int(54),
																																															},
																																															File: p1,
																																														},
																																														Fodder: ast.Fodder{},
																																														Ctx: p6446,
																																														FreeVars: ast.Identifiers{
																																															"val",
																																														},
																																													},
																																													Id: "val",
																																												},
																																												CommaFodder: nil,
																																											},
																																										},
																																										Named: nil,
																																									},
																																									TrailingComma: false,
																																									TailStrict: false,
																																									FodderRight: ast.Fodder{},
																																									TailStrictFodder: nil,
																																								},
																																								CommaFodder: nil,
																																							},
																																						},
																																						Named: nil,
																																					},
																																					TrailingComma: false,
																																					TailStrict: false,
																																					FodderRight: ast.Fodder{},
																																					TailStrictFodder: nil,
																																				},
																																				CommaFodder: ast.Fodder{},
																																			},
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Var{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(591),
																																								Column: int(58),
																																							},
																																							End: ast.Location{
																																								Line: int(591),
																																								Column: int(60),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p6422,
																																						FreeVars: ast.Identifiers{
																																							"zp",
																																						},
																																					},
																																					Id: "zp",
																																				},
																																				CommaFodder: ast.Fodder{},
																																			},
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Var{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(591),
																																								Column: int(62),
																																							},
																																							End: ast.Location{
																																								Line: int(591),
																																								Column: int(67),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p6422,
																																						FreeVars: ast.Identifiers{
																																							"iprec",
																																						},
																																					},
																																					Id: "iprec",
																																				},
																																				CommaFodder: ast.Fodder{},
																																			},
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Index{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(591),
																																								Column: int(69),
																																							},
																																							End: ast.Location{
																																								Line: int(591),
																																								Column: int(81),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p6422,
																																						FreeVars: ast.Identifiers{
																																							"cflags",
																																						},
																																					},
																																					Target: &ast.Var{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(591),
																																									Column: int(69),
																																								},
																																								End: ast.Location{
																																									Line: int(591),
																																									Column: int(75),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{
																																								"cflags",
																																							},
																																						},
																																						Id: "cflags",
																																					},
																																					LeftBracketFodder: ast.Fodder{},
																																					Index: &ast.LiteralString{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								End: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								File: nil,
																																							},
																																							Fodder: nil,
																																							Ctx: nil,
																																							FreeVars: nil,
																																						},
																																						Value: "blank",
																																						Kind: ast.LiteralStringKind(1),
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																					},
																																					RightBracketFodder: ast.Fodder{},
																																					Id: nil,
																																				},
																																				CommaFodder: ast.Fodder{},
																																			},
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Index{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(591),
																																								Column: int(83),
																																							},
																																							End: ast.Location{
																																								Line: int(591),
																																								Column: int(94),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p6422,
																																						FreeVars: ast.Identifiers{
																																							"cflags",
																																						},
																																					},
																																					Target: &ast.Var{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(591),
																																									Column: int(83),
																																								},
																																								End: ast.Location{
																																									Line: int(591),
																																									Column: int(89),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{
																																								"cflags",
																																							},
																																						},
																																						Id: "cflags",
																																					},
																																					LeftBracketFodder: ast.Fodder{},
																																					Index: &ast.LiteralString{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								End: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								File: nil,
																																							},
																																							Fodder: nil,
																																							Ctx: nil,
																																							FreeVars: nil,
																																						},
																																						Value: "plus",
																																						Kind: ast.LiteralStringKind(1),
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																					},
																																					RightBracketFodder: ast.Fodder{},
																																					Id: nil,
																																				},
																																				CommaFodder: ast.Fodder{},
																																			},
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.LiteralNumber{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(591),
																																								Column: int(96),
																																							},
																																							End: ast.Location{
																																								Line: int(591),
																																								Column: int(98),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p6422,
																																						FreeVars: nil,
																																					},
																																					OriginalString: "10",
																																				},
																																				CommaFodder: ast.Fodder{},
																																			},
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.LiteralString{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(591),
																																								Column: int(100),
																																							},
																																							End: ast.Location{
																																								Line: int(591),
																																								Column: int(102),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p6422,
																																						FreeVars: nil,
																																					},
																																					Value: "",
																																					Kind: ast.LiteralStringKind(1),
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																				},
																																				CommaFodder: nil,
																																			},
																																		},
																																		Named: nil,
																																	},
																																	TrailingComma: false,
																																	TailStrict: false,
																																	FodderRight: ast.Fodder{},
																																	TailStrictFodder: nil,
																																},
																															},
																															ElseFodder: ast.Fodder{
																																ast.FodderElement{
																																	Kind: ast.FodderKind(0),
																																	Blanks: int(0),
																																	Indent: int(6),
																																	Comment: []string{},
																																},
																															},
																															BranchFalse: &ast.Conditional{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(592),
																																			Column: int(12),
																																		},
																																		End: ast.Location{
																																			Line: int(671),
																																			Column: int(44),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p6275,
																																	FreeVars: ast.Identifiers{
																																		"cflags",
																																		"code",
																																		"fpprec",
																																		"i",
																																		"iprec",
																																		"render_float_dec",
																																		"render_float_sci",
																																		"render_hex",
																																		"render_int",
																																		"std",
																																		"val",
																																		"zp",
																																	},
																																},
																																Cond: &ast.Binary{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(592),
																																				Column: int(15),
																																			},
																																			End: ast.Location{
																																				Line: int(592),
																																				Column: int(32),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p6275,
																																		FreeVars: ast.Identifiers{
																																			"code",
																																		},
																																	},
																																	Left: &ast.Index{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(592),
																																					Column: int(15),
																																				},
																																				End: ast.Location{
																																					Line: int(592),
																																					Column: int(25),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p6275,
																																			FreeVars: ast.Identifiers{
																																				"code",
																																			},
																																		},
																																		Target: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(592),
																																						Column: int(15),
																																					},
																																					End: ast.Location{
																																						Line: int(592),
																																						Column: int(19),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: nil,
																																				FreeVars: ast.Identifiers{
																																					"code",
																																				},
																																			},
																																			Id: "code",
																																		},
																																		LeftBracketFodder: ast.Fodder{},
																																		Index: &ast.LiteralString{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(0),
																																						Column: int(0),
																																					},
																																					End: ast.Location{
																																						Line: int(0),
																																						Column: int(0),
																																					},
																																					File: nil,
																																				},
																																				Fodder: nil,
																																				Ctx: nil,
																																				FreeVars: nil,
																																			},
																																			Value: "ctype",
																																			Kind: ast.LiteralStringKind(1),
																																			BlockIndent: "",
																																			BlockTermIndent: "",
																																		},
																																		RightBracketFodder: ast.Fodder{},
																																		Id: nil,
																																	},
																																	OpFodder: ast.Fodder{},
																																	Op: ast.BinaryOp(12),
																																	Right: &ast.LiteralString{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(592),
																																					Column: int(29),
																																				},
																																				End: ast.Location{
																																					Line: int(592),
																																					Column: int(32),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p6275,
																																			FreeVars: nil,
																																		},
																																		Value: "o",
																																		Kind: ast.LiteralStringKind(1),
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																	},
																																},
																																ThenFodder: ast.Fodder{},
																																BranchTrue: &ast.Conditional{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(593),
																																				Column: int(9),
																																			},
																																			End: ast.Location{
																																				Line: int(598),
																																				Column: int(111),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{
																																			ast.FodderElement{
																																				Kind: ast.FodderKind(0),
																																				Blanks: int(0),
																																				Indent: int(8),
																																				Comment: []string{},
																																			},
																																		},
																																		Ctx: p6275,
																																		FreeVars: ast.Identifiers{
																																			"cflags",
																																			"i",
																																			"iprec",
																																			"render_int",
																																			"std",
																																			"val",
																																			"zp",
																																		},
																																	},
																																	Cond: &ast.Binary{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(593),
																																					Column: int(12),
																																				},
																																				End: ast.Location{
																																					Line: int(593),
																																					Column: int(37),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p6275,
																																			FreeVars: ast.Identifiers{
																																				"std",
																																				"val",
																																			},
																																		},
																																		Left: &ast.Apply{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(593),
																																						Column: int(12),
																																					},
																																					End: ast.Location{
																																						Line: int(593),
																																						Column: int(25),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p6275,
																																				FreeVars: ast.Identifiers{
																																					"std",
																																					"val",
																																				},
																																			},
																																			Target: &ast.Index{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(593),
																																							Column: int(12),
																																						},
																																						End: ast.Location{
																																							Line: int(593),
																																							Column: int(20),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p6275,
																																					FreeVars: ast.Identifiers{
																																						"std",
																																					},
																																				},
																																				Target: &ast.Var{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(593),
																																								Column: int(12),
																																							},
																																							End: ast.Location{
																																								Line: int(593),
																																								Column: int(15),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{
																																							"std",
																																						},
																																					},
																																					Id: "std",
																																				},
																																				LeftBracketFodder: ast.Fodder{},
																																				Index: &ast.LiteralString{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							End: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							File: nil,
																																						},
																																						Fodder: nil,
																																						Ctx: nil,
																																						FreeVars: nil,
																																					},
																																					Value: "type",
																																					Kind: ast.LiteralStringKind(1),
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																				},
																																				RightBracketFodder: ast.Fodder{},
																																				Id: nil,
																																			},
																																			FodderLeft: ast.Fodder{},
																																			Arguments: ast.Arguments{
																																				Positional: []ast.CommaSeparatedExpr{
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(593),
																																										Column: int(21),
																																									},
																																									End: ast.Location{
																																										Line: int(593),
																																										Column: int(24),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6489,
																																								FreeVars: ast.Identifiers{
																																									"val",
																																								},
																																							},
																																							Id: "val",
																																						},
																																						CommaFodder: nil,
																																					},
																																				},
																																				Named: nil,
																																			},
																																			TrailingComma: false,
																																			TailStrict: false,
																																			FodderRight: ast.Fodder{},
																																			TailStrictFodder: nil,
																																		},
																																		OpFodder: ast.Fodder{},
																																		Op: ast.BinaryOp(13),
																																		Right: &ast.LiteralString{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(593),
																																						Column: int(29),
																																					},
																																					End: ast.Location{
																																						Line: int(593),
																																						Column: int(37),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p6275,
																																				FreeVars: nil,
																																			},
																																			Value: "number",
																																			Kind: ast.LiteralStringKind(1),
																																			BlockIndent: "",
																																			BlockTermIndent: "",
																																		},
																																	},
																																	ThenFodder: ast.Fodder{},
																																	BranchTrue: &ast.Error{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(594),
																																					Column: int(11),
																																				},
																																				End: ast.Location{
																																					Line: int(595),
																																					Column: int(47),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{
																																				ast.FodderElement{
																																					Kind: ast.FodderKind(0),
																																					Blanks: int(0),
																																					Indent: int(10),
																																					Comment: []string{},
																																				},
																																			},
																																			Ctx: p6275,
																																			FreeVars: ast.Identifiers{
																																				"i",
																																				"std",
																																				"val",
																																			},
																																		},
																																		Expr: &ast.Binary{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(594),
																																						Column: int(17),
																																					},
																																					End: ast.Location{
																																						Line: int(595),
																																						Column: int(47),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p6275,
																																				FreeVars: ast.Identifiers{
																																					"i",
																																					"std",
																																					"val",
																																				},
																																			},
																																			Left: &ast.Binary{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(594),
																																							Column: int(17),
																																						},
																																						End: ast.Location{
																																							Line: int(595),
																																							Column: int(31),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p6275,
																																					FreeVars: ast.Identifiers{
																																						"i",
																																					},
																																				},
																																				Left: &ast.Binary{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(594),
																																								Column: int(17),
																																							},
																																							End: ast.Location{
																																								Line: int(595),
																																								Column: int(20),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p6275,
																																						FreeVars: ast.Identifiers{
																																							"i",
																																						},
																																					},
																																					Left: &ast.LiteralString{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(594),
																																									Column: int(17),
																																								},
																																								End: ast.Location{
																																									Line: int(594),
																																									Column: int(45),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p6275,
																																							FreeVars: nil,
																																						},
																																						Value: "Format required number at ",
																																						Kind: ast.LiteralStringKind(1),
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																					},
																																					OpFodder: ast.Fodder{
																																						ast.FodderElement{
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(16),
																																							Comment: []string{},
																																						},
																																					},
																																					Op: ast.BinaryOp(3),
																																					Right: &ast.Var{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(595),
																																									Column: int(19),
																																								},
																																								End: ast.Location{
																																									Line: int(595),
																																									Column: int(20),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p6275,
																																							FreeVars: ast.Identifiers{
																																								"i",
																																							},
																																						},
																																						Id: "i",
																																					},
																																				},
																																				OpFodder: ast.Fodder{},
																																				Op: ast.BinaryOp(3),
																																				Right: &ast.LiteralString{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(595),
																																								Column: int(23),
																																							},
																																							End: ast.Location{
																																								Line: int(595),
																																								Column: int(31),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p6275,
																																						FreeVars: nil,
																																					},
																																					Value: ", got ",
																																					Kind: ast.LiteralStringKind(1),
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																				},
																																			},
																																			OpFodder: ast.Fodder{},
																																			Op: ast.BinaryOp(3),
																																			Right: &ast.Apply{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(595),
																																							Column: int(34),
																																						},
																																						End: ast.Location{
																																							Line: int(595),
																																							Column: int(47),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p6275,
																																					FreeVars: ast.Identifiers{
																																						"std",
																																						"val",
																																					},
																																				},
																																				Target: &ast.Index{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(595),
																																								Column: int(34),
																																							},
																																							End: ast.Location{
																																								Line: int(595),
																																								Column: int(42),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p6275,
																																						FreeVars: ast.Identifiers{
																																							"std",
																																						},
																																					},
																																					Target: &ast.Var{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(595),
																																									Column: int(34),
																																								},
																																								End: ast.Location{
																																									Line: int(595),
																																									Column: int(37),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{
																																								"std",
																																							},
																																						},
																																						Id: "std",
																																					},
																																					LeftBracketFodder: ast.Fodder{},
																																					Index: &ast.LiteralString{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								End: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								File: nil,
																																							},
																																							Fodder: nil,
																																							Ctx: nil,
																																							FreeVars: nil,
																																						},
																																						Value: "type",
																																						Kind: ast.LiteralStringKind(1),
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																					},
																																					RightBracketFodder: ast.Fodder{},
																																					Id: nil,
																																				},
																																				FodderLeft: ast.Fodder{},
																																				Arguments: ast.Arguments{
																																					Positional: []ast.CommaSeparatedExpr{
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(595),
																																											Column: int(43),
																																										},
																																										End: ast.Location{
																																											Line: int(595),
																																											Column: int(46),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p6515,
																																									FreeVars: ast.Identifiers{
																																										"val",
																																									},
																																								},
																																								Id: "val",
																																							},
																																							CommaFodder: nil,
																																						},
																																					},
																																					Named: nil,
																																				},
																																				TrailingComma: false,
																																				TailStrict: false,
																																				FodderRight: ast.Fodder{},
																																				TailStrictFodder: nil,
																																			},
																																		},
																																	},
																																	ElseFodder: ast.Fodder{
																																		ast.FodderElement{
																																			Kind: ast.FodderKind(0),
																																			Blanks: int(0),
																																			Indent: int(8),
																																			Comment: []string{},
																																		},
																																	},
																																	BranchFalse: &ast.Local{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(597),
																																					Column: int(11),
																																				},
																																				End: ast.Location{
																																					Line: int(598),
																																					Column: int(111),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{
																																				ast.FodderElement{
																																					Kind: ast.FodderKind(0),
																																					Blanks: int(0),
																																					Indent: int(10),
																																					Comment: []string{},
																																				},
																																			},
																																			Ctx: p6275,
																																			FreeVars: ast.Identifiers{
																																				"cflags",
																																				"iprec",
																																				"render_int",
																																				"std",
																																				"val",
																																				"zp",
																																			},
																																		},
																																		Binds: ast.LocalBinds{
																																			ast.LocalBind{
																																				VarFodder: ast.Fodder{},
																																				Variable: "zero_prefix",
																																				EqFodder: ast.Fodder{},
																																				Body: &ast.Conditional{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(597),
																																								Column: int(31),
																																							},
																																							End: ast.Location{
																																								Line: int(597),
																																								Column: int(61),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p6523,
																																						FreeVars: ast.Identifiers{
																																							"cflags",
																																						},
																																					},
																																					Cond: &ast.Index{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(597),
																																									Column: int(34),
																																								},
																																								End: ast.Location{
																																									Line: int(597),
																																									Column: int(44),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p6523,
																																							FreeVars: ast.Identifiers{
																																								"cflags",
																																							},
																																						},
																																						Target: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(597),
																																										Column: int(34),
																																									},
																																									End: ast.Location{
																																										Line: int(597),
																																										Column: int(40),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: nil,
																																								FreeVars: ast.Identifiers{
																																									"cflags",
																																								},
																																							},
																																							Id: "cflags",
																																						},
																																						LeftBracketFodder: ast.Fodder{},
																																						Index: &ast.LiteralString{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																									End: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																									File: nil,
																																								},
																																								Fodder: nil,
																																								Ctx: nil,
																																								FreeVars: nil,
																																							},
																																							Value: "alt",
																																							Kind: ast.LiteralStringKind(1),
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																						},
																																						RightBracketFodder: ast.Fodder{},
																																						Id: nil,
																																					},
																																					ThenFodder: ast.Fodder{},
																																					BranchTrue: &ast.LiteralString{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(597),
																																									Column: int(50),
																																								},
																																								End: ast.Location{
																																									Line: int(597),
																																									Column: int(53),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p6523,
																																							FreeVars: nil,
																																						},
																																						Value: "0",
																																						Kind: ast.LiteralStringKind(1),
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																					},
																																					ElseFodder: ast.Fodder{},
																																					BranchFalse: &ast.LiteralString{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(597),
																																									Column: int(59),
																																								},
																																								End: ast.Location{
																																									Line: int(597),
																																									Column: int(61),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p6523,
																																							FreeVars: nil,
																																						},
																																						Value: "",
																																						Kind: ast.LiteralStringKind(1),
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																					},
																																				},
																																				Fun: nil,
																																				CloseFodder: ast.Fodder{},
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(597),
																																						Column: int(17),
																																					},
																																					End: ast.Location{
																																						Line: int(597),
																																						Column: int(61),
																																					},
																																					File: p1,
																																				},
																																			},
																																		},
																																		Body: &ast.Apply{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(598),
																																						Column: int(11),
																																					},
																																					End: ast.Location{
																																						Line: int(598),
																																						Column: int(111),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p6275,
																																				FreeVars: ast.Identifiers{
																																					"cflags",
																																					"iprec",
																																					"render_int",
																																					"std",
																																					"val",
																																					"zero_prefix",
																																					"zp",
																																				},
																																			},
																																			Target: &ast.Var{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(598),
																																							Column: int(11),
																																						},
																																						End: ast.Location{
																																							Line: int(598),
																																							Column: int(21),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{
																																						ast.FodderElement{
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(10),
																																							Comment: []string{},
																																						},
																																					},
																																					Ctx: p6275,
																																					FreeVars: ast.Identifiers{
																																						"render_int",
																																					},
																																				},
																																				Id: "render_int",
																																			},
																																			FodderLeft: ast.Fodder{},
																																			Arguments: ast.Arguments{
																																				Positional: []ast.CommaSeparatedExpr{
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Binary{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(598),
																																										Column: int(22),
																																									},
																																									End: ast.Location{
																																										Line: int(598),
																																										Column: int(31),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6539,
																																								FreeVars: ast.Identifiers{
																																									"val",
																																								},
																																							},
																																							Left: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(598),
																																											Column: int(22),
																																										},
																																										End: ast.Location{
																																											Line: int(598),
																																											Column: int(25),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p6539,
																																									FreeVars: ast.Identifiers{
																																										"val",
																																									},
																																								},
																																								Id: "val",
																																							},
																																							OpFodder: ast.Fodder{},
																																							Op: ast.BinaryOp(10),
																																							Right: &ast.Unary{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(598),
																																											Column: int(29),
																																										},
																																										End: ast.Location{
																																											Line: int(598),
																																											Column: int(31),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p6539,
																																									FreeVars: nil,
																																								},
																																								Op: ast.UnaryOp(3),
																																								Expr: &ast.LiteralNumber{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(598),
																																												Column: int(30),
																																											},
																																											End: ast.Location{
																																												Line: int(598),
																																												Column: int(31),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p6539,
																																										FreeVars: nil,
																																									},
																																									OriginalString: "1",
																																								},
																																							},
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Apply{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(598),
																																										Column: int(33),
																																									},
																																									End: ast.Location{
																																										Line: int(598),
																																										Column: int(56),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6539,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																									"val",
																																								},
																																							},
																																							Target: &ast.Index{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(598),
																																											Column: int(33),
																																										},
																																										End: ast.Location{
																																											Line: int(598),
																																											Column: int(42),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p6539,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																									},
																																								},
																																								Target: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(598),
																																												Column: int(33),
																																											},
																																											End: ast.Location{
																																												Line: int(598),
																																												Column: int(36),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																										},
																																									},
																																									Id: "std",
																																								},
																																								LeftBracketFodder: ast.Fodder{},
																																								Index: &ast.LiteralString{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											File: nil,
																																										},
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: nil,
																																									},
																																									Value: "floor",
																																									Kind: ast.LiteralStringKind(1),
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								Id: nil,
																																							},
																																							FodderLeft: ast.Fodder{},
																																							Arguments: ast.Arguments{
																																								Positional: []ast.CommaSeparatedExpr{
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Apply{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(598),
																																														Column: int(43),
																																													},
																																													End: ast.Location{
																																														Line: int(598),
																																														Column: int(55),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p6554,
																																												FreeVars: ast.Identifiers{
																																													"std",
																																													"val",
																																												},
																																											},
																																											Target: &ast.Index{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(598),
																																															Column: int(43),
																																														},
																																														End: ast.Location{
																																															Line: int(598),
																																															Column: int(50),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: p6554,
																																													FreeVars: ast.Identifiers{
																																														"std",
																																													},
																																												},
																																												Target: &ast.Var{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "<std>",
																																															Begin: ast.Location{
																																																Line: int(598),
																																																Column: int(43),
																																															},
																																															End: ast.Location{
																																																Line: int(598),
																																																Column: int(46),
																																															},
																																															File: p1,
																																														},
																																														Fodder: ast.Fodder{},
																																														Ctx: nil,
																																														FreeVars: ast.Identifiers{
																																															"std",
																																														},
																																													},
																																													Id: "std",
																																												},
																																												LeftBracketFodder: ast.Fodder{},
																																												Index: &ast.LiteralString{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																															End: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																															File: nil,
																																														},
																																														Fodder: nil,
																																														Ctx: nil,
																																														FreeVars: nil,
																																													},
																																													Value: "abs",
																																													Kind: ast.LiteralStringKind(1),
																																													BlockIndent: "",
																																													BlockTermIndent: "",
																																												},
																																												RightBracketFodder: ast.Fodder{},
																																												Id: nil,
																																											},
																																											FodderLeft: ast.Fodder{},
																																											Arguments: ast.Arguments{
																																												Positional: []ast.CommaSeparatedExpr{
																																													ast.CommaSeparatedExpr{
																																														Expr: &ast.Var{
																																															NodeBase: ast.NodeBase{
																																																LocRange: ast.LocationRange{
																																																	FileName: "<std>",
																																																	Begin: ast.Location{
																																																		Line: int(598),
																																																		Column: int(51),
																																																	},
																																																	End: ast.Location{
																																																		Line: int(598),
																																																		Column: int(54),
																																																	},
																																																	File: p1,
																																																},
																																																Fodder: ast.Fodder{},
																																																Ctx: p6563,
																																																FreeVars: ast.Identifiers{
																																																	"val",
																																																},
																																															},
																																															Id: "val",
																																														},
																																														CommaFodder: nil,
																																													},
																																												},
																																												Named: nil,
																																											},
																																											TrailingComma: false,
																																											TailStrict: false,
																																											FodderRight: ast.Fodder{},
																																											TailStrictFodder: nil,
																																										},
																																										CommaFodder: nil,
																																									},
																																								},
																																								Named: nil,
																																							},
																																							TrailingComma: false,
																																							TailStrict: false,
																																							FodderRight: ast.Fodder{},
																																							TailStrictFodder: nil,
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(598),
																																										Column: int(58),
																																									},
																																									End: ast.Location{
																																										Line: int(598),
																																										Column: int(60),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6539,
																																								FreeVars: ast.Identifiers{
																																									"zp",
																																								},
																																							},
																																							Id: "zp",
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(598),
																																										Column: int(62),
																																									},
																																									End: ast.Location{
																																										Line: int(598),
																																										Column: int(67),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6539,
																																								FreeVars: ast.Identifiers{
																																									"iprec",
																																								},
																																							},
																																							Id: "iprec",
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Index{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(598),
																																										Column: int(69),
																																									},
																																									End: ast.Location{
																																										Line: int(598),
																																										Column: int(81),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6539,
																																								FreeVars: ast.Identifiers{
																																									"cflags",
																																								},
																																							},
																																							Target: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(598),
																																											Column: int(69),
																																										},
																																										End: ast.Location{
																																											Line: int(598),
																																											Column: int(75),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"cflags",
																																									},
																																								},
																																								Id: "cflags",
																																							},
																																							LeftBracketFodder: ast.Fodder{},
																																							Index: &ast.LiteralString{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										File: nil,
																																									},
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: nil,
																																								},
																																								Value: "blank",
																																								Kind: ast.LiteralStringKind(1),
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							Id: nil,
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Index{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(598),
																																										Column: int(83),
																																									},
																																									End: ast.Location{
																																										Line: int(598),
																																										Column: int(94),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6539,
																																								FreeVars: ast.Identifiers{
																																									"cflags",
																																								},
																																							},
																																							Target: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(598),
																																											Column: int(83),
																																										},
																																										End: ast.Location{
																																											Line: int(598),
																																											Column: int(89),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"cflags",
																																									},
																																								},
																																								Id: "cflags",
																																							},
																																							LeftBracketFodder: ast.Fodder{},
																																							Index: &ast.LiteralString{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										File: nil,
																																									},
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: nil,
																																								},
																																								Value: "plus",
																																								Kind: ast.LiteralStringKind(1),
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							Id: nil,
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.LiteralNumber{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(598),
																																										Column: int(96),
																																									},
																																									End: ast.Location{
																																										Line: int(598),
																																										Column: int(97),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6539,
																																								FreeVars: nil,
																																							},
																																							OriginalString: "8",
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(598),
																																										Column: int(99),
																																									},
																																									End: ast.Location{
																																										Line: int(598),
																																										Column: int(110),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6539,
																																								FreeVars: ast.Identifiers{
																																									"zero_prefix",
																																								},
																																							},
																																							Id: "zero_prefix",
																																						},
																																						CommaFodder: nil,
																																					},
																																				},
																																				Named: nil,
																																			},
																																			TrailingComma: false,
																																			TailStrict: false,
																																			FodderRight: ast.Fodder{},
																																			TailStrictFodder: nil,
																																		},
																																	},
																																},
																																ElseFodder: ast.Fodder{
																																	ast.FodderElement{
																																		Kind: ast.FodderKind(0),
																																		Blanks: int(0),
																																		Indent: int(6),
																																		Comment: []string{},
																																	},
																																},
																																BranchFalse: &ast.Conditional{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(599),
																																				Column: int(12),
																																			},
																																			End: ast.Location{
																																				Line: int(671),
																																				Column: int(44),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p6275,
																																		FreeVars: ast.Identifiers{
																																			"cflags",
																																			"code",
																																			"fpprec",
																																			"i",
																																			"iprec",
																																			"render_float_dec",
																																			"render_float_sci",
																																			"render_hex",
																																			"std",
																																			"val",
																																			"zp",
																																		},
																																	},
																																	Cond: &ast.Binary{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(599),
																																					Column: int(15),
																																				},
																																				End: ast.Location{
																																					Line: int(599),
																																					Column: int(32),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p6275,
																																			FreeVars: ast.Identifiers{
																																				"code",
																																			},
																																		},
																																		Left: &ast.Index{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(599),
																																						Column: int(15),
																																					},
																																					End: ast.Location{
																																						Line: int(599),
																																						Column: int(25),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p6275,
																																				FreeVars: ast.Identifiers{
																																					"code",
																																				},
																																			},
																																			Target: &ast.Var{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(599),
																																							Column: int(15),
																																						},
																																						End: ast.Location{
																																							Line: int(599),
																																							Column: int(19),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{
																																						"code",
																																					},
																																				},
																																				Id: "code",
																																			},
																																			LeftBracketFodder: ast.Fodder{},
																																			Index: &ast.LiteralString{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						End: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						File: nil,
																																					},
																																					Fodder: nil,
																																					Ctx: nil,
																																					FreeVars: nil,
																																				},
																																				Value: "ctype",
																																				Kind: ast.LiteralStringKind(1),
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																			},
																																			RightBracketFodder: ast.Fodder{},
																																			Id: nil,
																																		},
																																		OpFodder: ast.Fodder{},
																																		Op: ast.BinaryOp(12),
																																		Right: &ast.LiteralString{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(599),
																																						Column: int(29),
																																					},
																																					End: ast.Location{
																																						Line: int(599),
																																						Column: int(32),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p6275,
																																				FreeVars: nil,
																																			},
																																			Value: "x",
																																			Kind: ast.LiteralStringKind(1),
																																			BlockIndent: "",
																																			BlockTermIndent: "",
																																		},
																																	},
																																	ThenFodder: ast.Fodder{},
																																	BranchTrue: &ast.Conditional{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(600),
																																					Column: int(9),
																																				},
																																				End: ast.Location{
																																					Line: int(610),
																																					Column: int(32),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{
																																				ast.FodderElement{
																																					Kind: ast.FodderKind(0),
																																					Blanks: int(0),
																																					Indent: int(8),
																																					Comment: []string{},
																																				},
																																			},
																																			Ctx: p6275,
																																			FreeVars: ast.Identifiers{
																																				"cflags",
																																				"code",
																																				"i",
																																				"iprec",
																																				"render_hex",
																																				"std",
																																				"val",
																																				"zp",
																																			},
																																		},
																																		Cond: &ast.Binary{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(600),
																																						Column: int(12),
																																					},
																																					End: ast.Location{
																																						Line: int(600),
																																						Column: int(37),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p6275,
																																				FreeVars: ast.Identifiers{
																																					"std",
																																					"val",
																																				},
																																			},
																																			Left: &ast.Apply{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(600),
																																							Column: int(12),
																																						},
																																						End: ast.Location{
																																							Line: int(600),
																																							Column: int(25),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p6275,
																																					FreeVars: ast.Identifiers{
																																						"std",
																																						"val",
																																					},
																																				},
																																				Target: &ast.Index{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(600),
																																								Column: int(12),
																																							},
																																							End: ast.Location{
																																								Line: int(600),
																																								Column: int(20),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p6275,
																																						FreeVars: ast.Identifiers{
																																							"std",
																																						},
																																					},
																																					Target: &ast.Var{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(600),
																																									Column: int(12),
																																								},
																																								End: ast.Location{
																																									Line: int(600),
																																									Column: int(15),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{
																																								"std",
																																							},
																																						},
																																						Id: "std",
																																					},
																																					LeftBracketFodder: ast.Fodder{},
																																					Index: &ast.LiteralString{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								End: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								File: nil,
																																							},
																																							Fodder: nil,
																																							Ctx: nil,
																																							FreeVars: nil,
																																						},
																																						Value: "type",
																																						Kind: ast.LiteralStringKind(1),
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																					},
																																					RightBracketFodder: ast.Fodder{},
																																					Id: nil,
																																				},
																																				FodderLeft: ast.Fodder{},
																																				Arguments: ast.Arguments{
																																					Positional: []ast.CommaSeparatedExpr{
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(600),
																																											Column: int(21),
																																										},
																																										End: ast.Location{
																																											Line: int(600),
																																											Column: int(24),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p6607,
																																									FreeVars: ast.Identifiers{
																																										"val",
																																									},
																																								},
																																								Id: "val",
																																							},
																																							CommaFodder: nil,
																																						},
																																					},
																																					Named: nil,
																																				},
																																				TrailingComma: false,
																																				TailStrict: false,
																																				FodderRight: ast.Fodder{},
																																				TailStrictFodder: nil,
																																			},
																																			OpFodder: ast.Fodder{},
																																			Op: ast.BinaryOp(13),
																																			Right: &ast.LiteralString{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(600),
																																							Column: int(29),
																																						},
																																						End: ast.Location{
																																							Line: int(600),
																																							Column: int(37),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p6275,
																																					FreeVars: nil,
																																				},
																																				Value: "number",
																																				Kind: ast.LiteralStringKind(1),
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																			},
																																		},
																																		ThenFodder: ast.Fodder{},
																																		BranchTrue: &ast.Error{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(601),
																																						Column: int(11),
																																					},
																																					End: ast.Location{
																																						Line: int(602),
																																						Column: int(47),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{
																																					ast.FodderElement{
																																						Kind: ast.FodderKind(0),
																																						Blanks: int(0),
																																						Indent: int(10),
																																						Comment: []string{},
																																					},
																																				},
																																				Ctx: p6275,
																																				FreeVars: ast.Identifiers{
																																					"i",
																																					"std",
																																					"val",
																																				},
																																			},
																																			Expr: &ast.Binary{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(601),
																																							Column: int(17),
																																						},
																																						End: ast.Location{
																																							Line: int(602),
																																							Column: int(47),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p6275,
																																					FreeVars: ast.Identifiers{
																																						"i",
																																						"std",
																																						"val",
																																					},
																																				},
																																				Left: &ast.Binary{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(601),
																																								Column: int(17),
																																							},
																																							End: ast.Location{
																																								Line: int(602),
																																								Column: int(31),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p6275,
																																						FreeVars: ast.Identifiers{
																																							"i",
																																						},
																																					},
																																					Left: &ast.Binary{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(601),
																																									Column: int(17),
																																								},
																																								End: ast.Location{
																																									Line: int(602),
																																									Column: int(20),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p6275,
																																							FreeVars: ast.Identifiers{
																																								"i",
																																							},
																																						},
																																						Left: &ast.LiteralString{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(601),
																																										Column: int(17),
																																									},
																																									End: ast.Location{
																																										Line: int(601),
																																										Column: int(45),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6275,
																																								FreeVars: nil,
																																							},
																																							Value: "Format required number at ",
																																							Kind: ast.LiteralStringKind(1),
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																						},
																																						OpFodder: ast.Fodder{
																																							ast.FodderElement{
																																								Kind: ast.FodderKind(0),
																																								Blanks: int(0),
																																								Indent: int(16),
																																								Comment: []string{},
																																							},
																																						},
																																						Op: ast.BinaryOp(3),
																																						Right: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(602),
																																										Column: int(19),
																																									},
																																									End: ast.Location{
																																										Line: int(602),
																																										Column: int(20),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6275,
																																								FreeVars: ast.Identifiers{
																																									"i",
																																								},
																																							},
																																							Id: "i",
																																						},
																																					},
																																					OpFodder: ast.Fodder{},
																																					Op: ast.BinaryOp(3),
																																					Right: &ast.LiteralString{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(602),
																																									Column: int(23),
																																								},
																																								End: ast.Location{
																																									Line: int(602),
																																									Column: int(31),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p6275,
																																							FreeVars: nil,
																																						},
																																						Value: ", got ",
																																						Kind: ast.LiteralStringKind(1),
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																					},
																																				},
																																				OpFodder: ast.Fodder{},
																																				Op: ast.BinaryOp(3),
																																				Right: &ast.Apply{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(602),
																																								Column: int(34),
																																							},
																																							End: ast.Location{
																																								Line: int(602),
																																								Column: int(47),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p6275,
																																						FreeVars: ast.Identifiers{
																																							"std",
																																							"val",
																																						},
																																					},
																																					Target: &ast.Index{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(602),
																																									Column: int(34),
																																								},
																																								End: ast.Location{
																																									Line: int(602),
																																									Column: int(42),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p6275,
																																							FreeVars: ast.Identifiers{
																																								"std",
																																							},
																																						},
																																						Target: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(602),
																																										Column: int(34),
																																									},
																																									End: ast.Location{
																																										Line: int(602),
																																										Column: int(37),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: nil,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																								},
																																							},
																																							Id: "std",
																																						},
																																						LeftBracketFodder: ast.Fodder{},
																																						Index: &ast.LiteralString{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																									End: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																									File: nil,
																																								},
																																								Fodder: nil,
																																								Ctx: nil,
																																								FreeVars: nil,
																																							},
																																							Value: "type",
																																							Kind: ast.LiteralStringKind(1),
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																						},
																																						RightBracketFodder: ast.Fodder{},
																																						Id: nil,
																																					},
																																					FodderLeft: ast.Fodder{},
																																					Arguments: ast.Arguments{
																																						Positional: []ast.CommaSeparatedExpr{
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(602),
																																												Column: int(43),
																																											},
																																											End: ast.Location{
																																												Line: int(602),
																																												Column: int(46),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p6633,
																																										FreeVars: ast.Identifiers{
																																											"val",
																																										},
																																									},
																																									Id: "val",
																																								},
																																								CommaFodder: nil,
																																							},
																																						},
																																						Named: nil,
																																					},
																																					TrailingComma: false,
																																					TailStrict: false,
																																					FodderRight: ast.Fodder{},
																																					TailStrictFodder: nil,
																																				},
																																			},
																																		},
																																		ElseFodder: ast.Fodder{
																																			ast.FodderElement{
																																				Kind: ast.FodderKind(0),
																																				Blanks: int(0),
																																				Indent: int(8),
																																				Comment: []string{},
																																			},
																																		},
																																		BranchFalse: &ast.Apply{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(604),
																																						Column: int(11),
																																					},
																																					End: ast.Location{
																																						Line: int(610),
																																						Column: int(32),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p6275,
																																				FreeVars: ast.Identifiers{
																																					"cflags",
																																					"code",
																																					"iprec",
																																					"render_hex",
																																					"std",
																																					"val",
																																					"zp",
																																				},
																																			},
																																			Target: &ast.Var{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(604),
																																							Column: int(11),
																																						},
																																						End: ast.Location{
																																							Line: int(604),
																																							Column: int(21),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{
																																						ast.FodderElement{
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(10),
																																							Comment: []string{},
																																						},
																																					},
																																					Ctx: p6275,
																																					FreeVars: ast.Identifiers{
																																						"render_hex",
																																					},
																																				},
																																				Id: "render_hex",
																																			},
																																			FodderLeft: ast.Fodder{},
																																			Arguments: ast.Arguments{
																																				Positional: []ast.CommaSeparatedExpr{
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Apply{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(604),
																																										Column: int(22),
																																									},
																																									End: ast.Location{
																																										Line: int(604),
																																										Column: int(36),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6643,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																									"val",
																																								},
																																							},
																																							Target: &ast.Index{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(604),
																																											Column: int(22),
																																										},
																																										End: ast.Location{
																																											Line: int(604),
																																											Column: int(31),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p6643,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																									},
																																								},
																																								Target: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(604),
																																												Column: int(22),
																																											},
																																											End: ast.Location{
																																												Line: int(604),
																																												Column: int(25),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																										},
																																									},
																																									Id: "std",
																																								},
																																								LeftBracketFodder: ast.Fodder{},
																																								Index: &ast.LiteralString{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											File: nil,
																																										},
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: nil,
																																									},
																																									Value: "floor",
																																									Kind: ast.LiteralStringKind(1),
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								Id: nil,
																																							},
																																							FodderLeft: ast.Fodder{},
																																							Arguments: ast.Arguments{
																																								Positional: []ast.CommaSeparatedExpr{
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(604),
																																														Column: int(32),
																																													},
																																													End: ast.Location{
																																														Line: int(604),
																																														Column: int(35),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p6652,
																																												FreeVars: ast.Identifiers{
																																													"val",
																																												},
																																											},
																																											Id: "val",
																																										},
																																										CommaFodder: nil,
																																									},
																																								},
																																								Named: nil,
																																							},
																																							TrailingComma: false,
																																							TailStrict: false,
																																							FodderRight: ast.Fodder{},
																																							TailStrictFodder: nil,
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(605),
																																										Column: int(22),
																																									},
																																									End: ast.Location{
																																										Line: int(605),
																																										Column: int(24),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{
																																									ast.FodderElement{
																																										Kind: ast.FodderKind(0),
																																										Blanks: int(0),
																																										Indent: int(21),
																																										Comment: []string{},
																																									},
																																								},
																																								Ctx: p6643,
																																								FreeVars: ast.Identifiers{
																																									"zp",
																																								},
																																							},
																																							Id: "zp",
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(606),
																																										Column: int(22),
																																									},
																																									End: ast.Location{
																																										Line: int(606),
																																										Column: int(27),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{
																																									ast.FodderElement{
																																										Kind: ast.FodderKind(0),
																																										Blanks: int(0),
																																										Indent: int(21),
																																										Comment: []string{},
																																									},
																																								},
																																								Ctx: p6643,
																																								FreeVars: ast.Identifiers{
																																									"iprec",
																																								},
																																							},
																																							Id: "iprec",
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Index{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(607),
																																										Column: int(22),
																																									},
																																									End: ast.Location{
																																										Line: int(607),
																																										Column: int(34),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6643,
																																								FreeVars: ast.Identifiers{
																																									"cflags",
																																								},
																																							},
																																							Target: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(607),
																																											Column: int(22),
																																										},
																																										End: ast.Location{
																																											Line: int(607),
																																											Column: int(28),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(21),
																																											Comment: []string{},
																																										},
																																									},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"cflags",
																																									},
																																								},
																																								Id: "cflags",
																																							},
																																							LeftBracketFodder: ast.Fodder{},
																																							Index: &ast.LiteralString{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										File: nil,
																																									},
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: nil,
																																								},
																																								Value: "blank",
																																								Kind: ast.LiteralStringKind(1),
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							Id: nil,
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Index{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(608),
																																										Column: int(22),
																																									},
																																									End: ast.Location{
																																										Line: int(608),
																																										Column: int(33),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6643,
																																								FreeVars: ast.Identifiers{
																																									"cflags",
																																								},
																																							},
																																							Target: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(608),
																																											Column: int(22),
																																										},
																																										End: ast.Location{
																																											Line: int(608),
																																											Column: int(28),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(21),
																																											Comment: []string{},
																																										},
																																									},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"cflags",
																																									},
																																								},
																																								Id: "cflags",
																																							},
																																							LeftBracketFodder: ast.Fodder{},
																																							Index: &ast.LiteralString{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										File: nil,
																																									},
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: nil,
																																								},
																																								Value: "plus",
																																								Kind: ast.LiteralStringKind(1),
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							Id: nil,
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Index{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(609),
																																										Column: int(22),
																																									},
																																									End: ast.Location{
																																										Line: int(609),
																																										Column: int(32),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6643,
																																								FreeVars: ast.Identifiers{
																																									"cflags",
																																								},
																																							},
																																							Target: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(609),
																																											Column: int(22),
																																										},
																																										End: ast.Location{
																																											Line: int(609),
																																											Column: int(28),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(21),
																																											Comment: []string{},
																																										},
																																									},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"cflags",
																																									},
																																								},
																																								Id: "cflags",
																																							},
																																							LeftBracketFodder: ast.Fodder{},
																																							Index: &ast.LiteralString{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										File: nil,
																																									},
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: nil,
																																								},
																																								Value: "alt",
																																								Kind: ast.LiteralStringKind(1),
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							Id: nil,
																																						},
																																						CommaFodder: ast.Fodder{},
																																					},
																																					ast.CommaSeparatedExpr{
																																						Expr: &ast.Index{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(610),
																																										Column: int(22),
																																									},
																																									End: ast.Location{
																																										Line: int(610),
																																										Column: int(31),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6643,
																																								FreeVars: ast.Identifiers{
																																									"code",
																																								},
																																							},
																																							Target: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(610),
																																											Column: int(22),
																																										},
																																										End: ast.Location{
																																											Line: int(610),
																																											Column: int(26),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(21),
																																											Comment: []string{},
																																										},
																																									},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"code",
																																									},
																																								},
																																								Id: "code",
																																							},
																																							LeftBracketFodder: ast.Fodder{},
																																							Index: &ast.LiteralString{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										File: nil,
																																									},
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: nil,
																																								},
																																								Value: "caps",
																																								Kind: ast.LiteralStringKind(1),
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							Id: nil,
																																						},
																																						CommaFodder: nil,
																																					},
																																				},
																																				Named: nil,
																																			},
																																			TrailingComma: false,
																																			TailStrict: false,
																																			FodderRight: ast.Fodder{},
																																			TailStrictFodder: nil,
																																		},
																																	},
																																	ElseFodder: ast.Fodder{
																																		ast.FodderElement{
																																			Kind: ast.FodderKind(0),
																																			Blanks: int(0),
																																			Indent: int(6),
																																			Comment: []string{},
																																		},
																																	},
																																	BranchFalse: &ast.Conditional{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(611),
																																					Column: int(12),
																																				},
																																				End: ast.Location{
																																					Line: int(671),
																																					Column: int(44),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p6275,
																																			FreeVars: ast.Identifiers{
																																				"cflags",
																																				"code",
																																				"fpprec",
																																				"i",
																																				"render_float_dec",
																																				"render_float_sci",
																																				"std",
																																				"val",
																																				"zp",
																																			},
																																		},
																																		Cond: &ast.Binary{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(611),
																																						Column: int(15),
																																					},
																																					End: ast.Location{
																																						Line: int(611),
																																						Column: int(32),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p6275,
																																				FreeVars: ast.Identifiers{
																																					"code",
																																				},
																																			},
																																			Left: &ast.Index{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(611),
																																							Column: int(15),
																																						},
																																						End: ast.Location{
																																							Line: int(611),
																																							Column: int(25),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p6275,
																																					FreeVars: ast.Identifiers{
																																						"code",
																																					},
																																				},
																																				Target: &ast.Var{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(611),
																																								Column: int(15),
																																							},
																																							End: ast.Location{
																																								Line: int(611),
																																								Column: int(19),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{
																																							"code",
																																						},
																																					},
																																					Id: "code",
																																				},
																																				LeftBracketFodder: ast.Fodder{},
																																				Index: &ast.LiteralString{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							End: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							File: nil,
																																						},
																																						Fodder: nil,
																																						Ctx: nil,
																																						FreeVars: nil,
																																					},
																																					Value: "ctype",
																																					Kind: ast.LiteralStringKind(1),
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																				},
																																				RightBracketFodder: ast.Fodder{},
																																				Id: nil,
																																			},
																																			OpFodder: ast.Fodder{},
																																			Op: ast.BinaryOp(12),
																																			Right: &ast.LiteralString{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(611),
																																							Column: int(29),
																																						},
																																						End: ast.Location{
																																							Line: int(611),
																																							Column: int(32),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p6275,
																																					FreeVars: nil,
																																				},
																																				Value: "f",
																																				Kind: ast.LiteralStringKind(1),
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																			},
																																		},
																																		ThenFodder: ast.Fodder{},
																																		BranchTrue: &ast.Conditional{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(612),
																																						Column: int(9),
																																					},
																																					End: ast.Location{
																																						Line: int(622),
																																						Column: int(35),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{
																																					ast.FodderElement{
																																						Kind: ast.FodderKind(0),
																																						Blanks: int(0),
																																						Indent: int(8),
																																						Comment: []string{},
																																					},
																																				},
																																				Ctx: p6275,
																																				FreeVars: ast.Identifiers{
																																					"cflags",
																																					"fpprec",
																																					"i",
																																					"render_float_dec",
																																					"std",
																																					"val",
																																					"zp",
																																				},
																																			},
																																			Cond: &ast.Binary{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(612),
																																							Column: int(12),
																																						},
																																						End: ast.Location{
																																							Line: int(612),
																																							Column: int(37),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p6275,
																																					FreeVars: ast.Identifiers{
																																						"std",
																																						"val",
																																					},
																																				},
																																				Left: &ast.Apply{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(612),
																																								Column: int(12),
																																							},
																																							End: ast.Location{
																																								Line: int(612),
																																								Column: int(25),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p6275,
																																						FreeVars: ast.Identifiers{
																																							"std",
																																							"val",
																																						},
																																					},
																																					Target: &ast.Index{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(612),
																																									Column: int(12),
																																								},
																																								End: ast.Location{
																																									Line: int(612),
																																									Column: int(20),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p6275,
																																							FreeVars: ast.Identifiers{
																																								"std",
																																							},
																																						},
																																						Target: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(612),
																																										Column: int(12),
																																									},
																																									End: ast.Location{
																																										Line: int(612),
																																										Column: int(15),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: nil,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																								},
																																							},
																																							Id: "std",
																																						},
																																						LeftBracketFodder: ast.Fodder{},
																																						Index: &ast.LiteralString{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																									End: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																									File: nil,
																																								},
																																								Fodder: nil,
																																								Ctx: nil,
																																								FreeVars: nil,
																																							},
																																							Value: "type",
																																							Kind: ast.LiteralStringKind(1),
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																						},
																																						RightBracketFodder: ast.Fodder{},
																																						Id: nil,
																																					},
																																					FodderLeft: ast.Fodder{},
																																					Arguments: ast.Arguments{
																																						Positional: []ast.CommaSeparatedExpr{
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(612),
																																												Column: int(21),
																																											},
																																											End: ast.Location{
																																												Line: int(612),
																																												Column: int(24),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p6709,
																																										FreeVars: ast.Identifiers{
																																											"val",
																																										},
																																									},
																																									Id: "val",
																																								},
																																								CommaFodder: nil,
																																							},
																																						},
																																						Named: nil,
																																					},
																																					TrailingComma: false,
																																					TailStrict: false,
																																					FodderRight: ast.Fodder{},
																																					TailStrictFodder: nil,
																																				},
																																				OpFodder: ast.Fodder{},
																																				Op: ast.BinaryOp(13),
																																				Right: &ast.LiteralString{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(612),
																																								Column: int(29),
																																							},
																																							End: ast.Location{
																																								Line: int(612),
																																								Column: int(37),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p6275,
																																						FreeVars: nil,
																																					},
																																					Value: "number",
																																					Kind: ast.LiteralStringKind(1),
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																				},
																																			},
																																			ThenFodder: ast.Fodder{},
																																			BranchTrue: &ast.Error{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(613),
																																							Column: int(11),
																																						},
																																						End: ast.Location{
																																							Line: int(614),
																																							Column: int(47),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{
																																						ast.FodderElement{
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(10),
																																							Comment: []string{},
																																						},
																																					},
																																					Ctx: p6275,
																																					FreeVars: ast.Identifiers{
																																						"i",
																																						"std",
																																						"val",
																																					},
																																				},
																																				Expr: &ast.Binary{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(613),
																																								Column: int(17),
																																							},
																																							End: ast.Location{
																																								Line: int(614),
																																								Column: int(47),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p6275,
																																						FreeVars: ast.Identifiers{
																																							"i",
																																							"std",
																																							"val",
																																						},
																																					},
																																					Left: &ast.Binary{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(613),
																																									Column: int(17),
																																								},
																																								End: ast.Location{
																																									Line: int(614),
																																									Column: int(31),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p6275,
																																							FreeVars: ast.Identifiers{
																																								"i",
																																							},
																																						},
																																						Left: &ast.Binary{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(613),
																																										Column: int(17),
																																									},
																																									End: ast.Location{
																																										Line: int(614),
																																										Column: int(20),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6275,
																																								FreeVars: ast.Identifiers{
																																									"i",
																																								},
																																							},
																																							Left: &ast.LiteralString{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(613),
																																											Column: int(17),
																																										},
																																										End: ast.Location{
																																											Line: int(613),
																																											Column: int(45),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p6275,
																																									FreeVars: nil,
																																								},
																																								Value: "Format required number at ",
																																								Kind: ast.LiteralStringKind(1),
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																							},
																																							OpFodder: ast.Fodder{
																																								ast.FodderElement{
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(16),
																																									Comment: []string{},
																																								},
																																							},
																																							Op: ast.BinaryOp(3),
																																							Right: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(614),
																																											Column: int(19),
																																										},
																																										End: ast.Location{
																																											Line: int(614),
																																											Column: int(20),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p6275,
																																									FreeVars: ast.Identifiers{
																																										"i",
																																									},
																																								},
																																								Id: "i",
																																							},
																																						},
																																						OpFodder: ast.Fodder{},
																																						Op: ast.BinaryOp(3),
																																						Right: &ast.LiteralString{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(614),
																																										Column: int(23),
																																									},
																																									End: ast.Location{
																																										Line: int(614),
																																										Column: int(31),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6275,
																																								FreeVars: nil,
																																							},
																																							Value: ", got ",
																																							Kind: ast.LiteralStringKind(1),
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																						},
																																					},
																																					OpFodder: ast.Fodder{},
																																					Op: ast.BinaryOp(3),
																																					Right: &ast.Apply{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(614),
																																									Column: int(34),
																																								},
																																								End: ast.Location{
																																									Line: int(614),
																																									Column: int(47),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p6275,
																																							FreeVars: ast.Identifiers{
																																								"std",
																																								"val",
																																							},
																																						},
																																						Target: &ast.Index{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(614),
																																										Column: int(34),
																																									},
																																									End: ast.Location{
																																										Line: int(614),
																																										Column: int(42),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6275,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																								},
																																							},
																																							Target: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(614),
																																											Column: int(34),
																																										},
																																										End: ast.Location{
																																											Line: int(614),
																																											Column: int(37),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																									},
																																								},
																																								Id: "std",
																																							},
																																							LeftBracketFodder: ast.Fodder{},
																																							Index: &ast.LiteralString{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										File: nil,
																																									},
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: nil,
																																								},
																																								Value: "type",
																																								Kind: ast.LiteralStringKind(1),
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							Id: nil,
																																						},
																																						FodderLeft: ast.Fodder{},
																																						Arguments: ast.Arguments{
																																							Positional: []ast.CommaSeparatedExpr{
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(614),
																																													Column: int(43),
																																												},
																																												End: ast.Location{
																																													Line: int(614),
																																													Column: int(46),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p6735,
																																											FreeVars: ast.Identifiers{
																																												"val",
																																											},
																																										},
																																										Id: "val",
																																									},
																																									CommaFodder: nil,
																																								},
																																							},
																																							Named: nil,
																																						},
																																						TrailingComma: false,
																																						TailStrict: false,
																																						FodderRight: ast.Fodder{},
																																						TailStrictFodder: nil,
																																					},
																																				},
																																			},
																																			ElseFodder: ast.Fodder{
																																				ast.FodderElement{
																																					Kind: ast.FodderKind(0),
																																					Blanks: int(0),
																																					Indent: int(8),
																																					Comment: []string{},
																																				},
																																			},
																																			BranchFalse: &ast.Apply{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(616),
																																							Column: int(11),
																																						},
																																						End: ast.Location{
																																							Line: int(622),
																																							Column: int(35),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p6275,
																																					FreeVars: ast.Identifiers{
																																						"cflags",
																																						"fpprec",
																																						"render_float_dec",
																																						"val",
																																						"zp",
																																					},
																																				},
																																				Target: &ast.Var{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(616),
																																								Column: int(11),
																																							},
																																							End: ast.Location{
																																								Line: int(616),
																																								Column: int(27),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{
																																							ast.FodderElement{
																																								Kind: ast.FodderKind(0),
																																								Blanks: int(0),
																																								Indent: int(10),
																																								Comment: []string{},
																																							},
																																						},
																																						Ctx: p6275,
																																						FreeVars: ast.Identifiers{
																																							"render_float_dec",
																																						},
																																					},
																																					Id: "render_float_dec",
																																				},
																																				FodderLeft: ast.Fodder{},
																																				Arguments: ast.Arguments{
																																					Positional: []ast.CommaSeparatedExpr{
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(616),
																																											Column: int(28),
																																										},
																																										End: ast.Location{
																																											Line: int(616),
																																											Column: int(31),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p6745,
																																									FreeVars: ast.Identifiers{
																																										"val",
																																									},
																																								},
																																								Id: "val",
																																							},
																																							CommaFodder: ast.Fodder{},
																																						},
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(617),
																																											Column: int(28),
																																										},
																																										End: ast.Location{
																																											Line: int(617),
																																											Column: int(30),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(27),
																																											Comment: []string{},
																																										},
																																									},
																																									Ctx: p6745,
																																									FreeVars: ast.Identifiers{
																																										"zp",
																																									},
																																								},
																																								Id: "zp",
																																							},
																																							CommaFodder: ast.Fodder{},
																																						},
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.Index{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(618),
																																											Column: int(28),
																																										},
																																										End: ast.Location{
																																											Line: int(618),
																																											Column: int(40),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p6745,
																																									FreeVars: ast.Identifiers{
																																										"cflags",
																																									},
																																								},
																																								Target: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(618),
																																												Column: int(28),
																																											},
																																											End: ast.Location{
																																												Line: int(618),
																																												Column: int(34),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{
																																											ast.FodderElement{
																																												Kind: ast.FodderKind(0),
																																												Blanks: int(0),
																																												Indent: int(27),
																																												Comment: []string{},
																																											},
																																										},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"cflags",
																																										},
																																									},
																																									Id: "cflags",
																																								},
																																								LeftBracketFodder: ast.Fodder{},
																																								Index: &ast.LiteralString{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											File: nil,
																																										},
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: nil,
																																									},
																																									Value: "blank",
																																									Kind: ast.LiteralStringKind(1),
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								Id: nil,
																																							},
																																							CommaFodder: ast.Fodder{},
																																						},
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.Index{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(619),
																																											Column: int(28),
																																										},
																																										End: ast.Location{
																																											Line: int(619),
																																											Column: int(39),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p6745,
																																									FreeVars: ast.Identifiers{
																																										"cflags",
																																									},
																																								},
																																								Target: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(619),
																																												Column: int(28),
																																											},
																																											End: ast.Location{
																																												Line: int(619),
																																												Column: int(34),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{
																																											ast.FodderElement{
																																												Kind: ast.FodderKind(0),
																																												Blanks: int(0),
																																												Indent: int(27),
																																												Comment: []string{},
																																											},
																																										},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"cflags",
																																										},
																																									},
																																									Id: "cflags",
																																								},
																																								LeftBracketFodder: ast.Fodder{},
																																								Index: &ast.LiteralString{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											File: nil,
																																										},
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: nil,
																																									},
																																									Value: "plus",
																																									Kind: ast.LiteralStringKind(1),
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								Id: nil,
																																							},
																																							CommaFodder: ast.Fodder{},
																																						},
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.Index{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(620),
																																											Column: int(28),
																																										},
																																										End: ast.Location{
																																											Line: int(620),
																																											Column: int(38),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p6745,
																																									FreeVars: ast.Identifiers{
																																										"cflags",
																																									},
																																								},
																																								Target: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(620),
																																												Column: int(28),
																																											},
																																											End: ast.Location{
																																												Line: int(620),
																																												Column: int(34),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{
																																											ast.FodderElement{
																																												Kind: ast.FodderKind(0),
																																												Blanks: int(0),
																																												Indent: int(27),
																																												Comment: []string{},
																																											},
																																										},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"cflags",
																																										},
																																									},
																																									Id: "cflags",
																																								},
																																								LeftBracketFodder: ast.Fodder{},
																																								Index: &ast.LiteralString{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											File: nil,
																																										},
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: nil,
																																									},
																																									Value: "alt",
																																									Kind: ast.LiteralStringKind(1),
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								Id: nil,
																																							},
																																							CommaFodder: ast.Fodder{},
																																						},
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.LiteralBoolean{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(621),
																																											Column: int(28),
																																										},
																																										End: ast.Location{
																																											Line: int(621),
																																											Column: int(32),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(27),
																																											Comment: []string{},
																																										},
																																									},
																																									Ctx: p6745,
																																									FreeVars: nil,
																																								},
																																								Value: true,
																																							},
																																							CommaFodder: ast.Fodder{},
																																						},
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(622),
																																											Column: int(28),
																																										},
																																										End: ast.Location{
																																											Line: int(622),
																																											Column: int(34),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(27),
																																											Comment: []string{},
																																										},
																																									},
																																									Ctx: p6745,
																																									FreeVars: ast.Identifiers{
																																										"fpprec",
																																									},
																																								},
																																								Id: "fpprec",
																																							},
																																							CommaFodder: nil,
																																						},
																																					},
																																					Named: nil,
																																				},
																																				TrailingComma: false,
																																				TailStrict: false,
																																				FodderRight: ast.Fodder{},
																																				TailStrictFodder: nil,
																																			},
																																		},
																																		ElseFodder: ast.Fodder{
																																			ast.FodderElement{
																																				Kind: ast.FodderKind(0),
																																				Blanks: int(0),
																																				Indent: int(6),
																																				Comment: []string{},
																																			},
																																		},
																																		BranchFalse: &ast.Conditional{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(623),
																																						Column: int(12),
																																					},
																																					End: ast.Location{
																																						Line: int(671),
																																						Column: int(44),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p6275,
																																				FreeVars: ast.Identifiers{
																																					"cflags",
																																					"code",
																																					"fpprec",
																																					"i",
																																					"render_float_dec",
																																					"render_float_sci",
																																					"std",
																																					"val",
																																					"zp",
																																				},
																																			},
																																			Cond: &ast.Binary{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(623),
																																							Column: int(15),
																																						},
																																						End: ast.Location{
																																							Line: int(623),
																																							Column: int(32),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p6275,
																																					FreeVars: ast.Identifiers{
																																						"code",
																																					},
																																				},
																																				Left: &ast.Index{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(623),
																																								Column: int(15),
																																							},
																																							End: ast.Location{
																																								Line: int(623),
																																								Column: int(25),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p6275,
																																						FreeVars: ast.Identifiers{
																																							"code",
																																						},
																																					},
																																					Target: &ast.Var{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(623),
																																									Column: int(15),
																																								},
																																								End: ast.Location{
																																									Line: int(623),
																																									Column: int(19),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{
																																								"code",
																																							},
																																						},
																																						Id: "code",
																																					},
																																					LeftBracketFodder: ast.Fodder{},
																																					Index: &ast.LiteralString{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								End: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								File: nil,
																																							},
																																							Fodder: nil,
																																							Ctx: nil,
																																							FreeVars: nil,
																																						},
																																						Value: "ctype",
																																						Kind: ast.LiteralStringKind(1),
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																					},
																																					RightBracketFodder: ast.Fodder{},
																																					Id: nil,
																																				},
																																				OpFodder: ast.Fodder{},
																																				Op: ast.BinaryOp(12),
																																				Right: &ast.LiteralString{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(623),
																																								Column: int(29),
																																							},
																																							End: ast.Location{
																																								Line: int(623),
																																								Column: int(32),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p6275,
																																						FreeVars: nil,
																																					},
																																					Value: "e",
																																					Kind: ast.LiteralStringKind(1),
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																				},
																																			},
																																			ThenFodder: ast.Fodder{},
																																			BranchTrue: &ast.Conditional{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(624),
																																							Column: int(9),
																																						},
																																						End: ast.Location{
																																							Line: int(635),
																																							Column: int(35),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{
																																						ast.FodderElement{
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(8),
																																							Comment: []string{},
																																						},
																																					},
																																					Ctx: p6275,
																																					FreeVars: ast.Identifiers{
																																						"cflags",
																																						"code",
																																						"fpprec",
																																						"i",
																																						"render_float_sci",
																																						"std",
																																						"val",
																																						"zp",
																																					},
																																				},
																																				Cond: &ast.Binary{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(624),
																																								Column: int(12),
																																							},
																																							End: ast.Location{
																																								Line: int(624),
																																								Column: int(37),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p6275,
																																						FreeVars: ast.Identifiers{
																																							"std",
																																							"val",
																																						},
																																					},
																																					Left: &ast.Apply{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(624),
																																									Column: int(12),
																																								},
																																								End: ast.Location{
																																									Line: int(624),
																																									Column: int(25),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p6275,
																																							FreeVars: ast.Identifiers{
																																								"std",
																																								"val",
																																							},
																																						},
																																						Target: &ast.Index{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(624),
																																										Column: int(12),
																																									},
																																									End: ast.Location{
																																										Line: int(624),
																																										Column: int(20),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6275,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																								},
																																							},
																																							Target: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(624),
																																											Column: int(12),
																																										},
																																										End: ast.Location{
																																											Line: int(624),
																																											Column: int(15),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																									},
																																								},
																																								Id: "std",
																																							},
																																							LeftBracketFodder: ast.Fodder{},
																																							Index: &ast.LiteralString{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										File: nil,
																																									},
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: nil,
																																								},
																																								Value: "type",
																																								Kind: ast.LiteralStringKind(1),
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							Id: nil,
																																						},
																																						FodderLeft: ast.Fodder{},
																																						Arguments: ast.Arguments{
																																							Positional: []ast.CommaSeparatedExpr{
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(624),
																																													Column: int(21),
																																												},
																																												End: ast.Location{
																																													Line: int(624),
																																													Column: int(24),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p6798,
																																											FreeVars: ast.Identifiers{
																																												"val",
																																											},
																																										},
																																										Id: "val",
																																									},
																																									CommaFodder: nil,
																																								},
																																							},
																																							Named: nil,
																																						},
																																						TrailingComma: false,
																																						TailStrict: false,
																																						FodderRight: ast.Fodder{},
																																						TailStrictFodder: nil,
																																					},
																																					OpFodder: ast.Fodder{},
																																					Op: ast.BinaryOp(13),
																																					Right: &ast.LiteralString{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(624),
																																									Column: int(29),
																																								},
																																								End: ast.Location{
																																									Line: int(624),
																																									Column: int(37),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p6275,
																																							FreeVars: nil,
																																						},
																																						Value: "number",
																																						Kind: ast.LiteralStringKind(1),
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																					},
																																				},
																																				ThenFodder: ast.Fodder{},
																																				BranchTrue: &ast.Error{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(625),
																																								Column: int(11),
																																							},
																																							End: ast.Location{
																																								Line: int(626),
																																								Column: int(47),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{
																																							ast.FodderElement{
																																								Kind: ast.FodderKind(0),
																																								Blanks: int(0),
																																								Indent: int(10),
																																								Comment: []string{},
																																							},
																																						},
																																						Ctx: p6275,
																																						FreeVars: ast.Identifiers{
																																							"i",
																																							"std",
																																							"val",
																																						},
																																					},
																																					Expr: &ast.Binary{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(625),
																																									Column: int(17),
																																								},
																																								End: ast.Location{
																																									Line: int(626),
																																									Column: int(47),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p6275,
																																							FreeVars: ast.Identifiers{
																																								"i",
																																								"std",
																																								"val",
																																							},
																																						},
																																						Left: &ast.Binary{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(625),
																																										Column: int(17),
																																									},
																																									End: ast.Location{
																																										Line: int(626),
																																										Column: int(31),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6275,
																																								FreeVars: ast.Identifiers{
																																									"i",
																																								},
																																							},
																																							Left: &ast.Binary{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(625),
																																											Column: int(17),
																																										},
																																										End: ast.Location{
																																											Line: int(626),
																																											Column: int(20),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p6275,
																																									FreeVars: ast.Identifiers{
																																										"i",
																																									},
																																								},
																																								Left: &ast.LiteralString{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(625),
																																												Column: int(17),
																																											},
																																											End: ast.Location{
																																												Line: int(625),
																																												Column: int(45),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p6275,
																																										FreeVars: nil,
																																									},
																																									Value: "Format required number at ",
																																									Kind: ast.LiteralStringKind(1),
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																								},
																																								OpFodder: ast.Fodder{
																																									ast.FodderElement{
																																										Kind: ast.FodderKind(0),
																																										Blanks: int(0),
																																										Indent: int(16),
																																										Comment: []string{},
																																									},
																																								},
																																								Op: ast.BinaryOp(3),
																																								Right: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(626),
																																												Column: int(19),
																																											},
																																											End: ast.Location{
																																												Line: int(626),
																																												Column: int(20),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p6275,
																																										FreeVars: ast.Identifiers{
																																											"i",
																																										},
																																									},
																																									Id: "i",
																																								},
																																							},
																																							OpFodder: ast.Fodder{},
																																							Op: ast.BinaryOp(3),
																																							Right: &ast.LiteralString{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(626),
																																											Column: int(23),
																																										},
																																										End: ast.Location{
																																											Line: int(626),
																																											Column: int(31),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p6275,
																																									FreeVars: nil,
																																								},
																																								Value: ", got ",
																																								Kind: ast.LiteralStringKind(1),
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																							},
																																						},
																																						OpFodder: ast.Fodder{},
																																						Op: ast.BinaryOp(3),
																																						Right: &ast.Apply{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(626),
																																										Column: int(34),
																																									},
																																									End: ast.Location{
																																										Line: int(626),
																																										Column: int(47),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6275,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																									"val",
																																								},
																																							},
																																							Target: &ast.Index{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(626),
																																											Column: int(34),
																																										},
																																										End: ast.Location{
																																											Line: int(626),
																																											Column: int(42),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p6275,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																									},
																																								},
																																								Target: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(626),
																																												Column: int(34),
																																											},
																																											End: ast.Location{
																																												Line: int(626),
																																												Column: int(37),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																										},
																																									},
																																									Id: "std",
																																								},
																																								LeftBracketFodder: ast.Fodder{},
																																								Index: &ast.LiteralString{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											File: nil,
																																										},
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: nil,
																																									},
																																									Value: "type",
																																									Kind: ast.LiteralStringKind(1),
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								Id: nil,
																																							},
																																							FodderLeft: ast.Fodder{},
																																							Arguments: ast.Arguments{
																																								Positional: []ast.CommaSeparatedExpr{
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(626),
																																														Column: int(43),
																																													},
																																													End: ast.Location{
																																														Line: int(626),
																																														Column: int(46),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p6824,
																																												FreeVars: ast.Identifiers{
																																													"val",
																																												},
																																											},
																																											Id: "val",
																																										},
																																										CommaFodder: nil,
																																									},
																																								},
																																								Named: nil,
																																							},
																																							TrailingComma: false,
																																							TailStrict: false,
																																							FodderRight: ast.Fodder{},
																																							TailStrictFodder: nil,
																																						},
																																					},
																																				},
																																				ElseFodder: ast.Fodder{
																																					ast.FodderElement{
																																						Kind: ast.FodderKind(0),
																																						Blanks: int(0),
																																						Indent: int(8),
																																						Comment: []string{},
																																					},
																																				},
																																				BranchFalse: &ast.Apply{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(628),
																																								Column: int(11),
																																							},
																																							End: ast.Location{
																																								Line: int(635),
																																								Column: int(35),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p6275,
																																						FreeVars: ast.Identifiers{
																																							"cflags",
																																							"code",
																																							"fpprec",
																																							"render_float_sci",
																																							"val",
																																							"zp",
																																						},
																																					},
																																					Target: &ast.Var{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(628),
																																									Column: int(11),
																																								},
																																								End: ast.Location{
																																									Line: int(628),
																																									Column: int(27),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(10),
																																									Comment: []string{},
																																								},
																																							},
																																							Ctx: p6275,
																																							FreeVars: ast.Identifiers{
																																								"render_float_sci",
																																							},
																																						},
																																						Id: "render_float_sci",
																																					},
																																					FodderLeft: ast.Fodder{},
																																					Arguments: ast.Arguments{
																																						Positional: []ast.CommaSeparatedExpr{
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(628),
																																												Column: int(28),
																																											},
																																											End: ast.Location{
																																												Line: int(628),
																																												Column: int(31),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p6834,
																																										FreeVars: ast.Identifiers{
																																											"val",
																																										},
																																									},
																																									Id: "val",
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(629),
																																												Column: int(28),
																																											},
																																											End: ast.Location{
																																												Line: int(629),
																																												Column: int(30),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{
																																											ast.FodderElement{
																																												Kind: ast.FodderKind(0),
																																												Blanks: int(0),
																																												Indent: int(27),
																																												Comment: []string{},
																																											},
																																										},
																																										Ctx: p6834,
																																										FreeVars: ast.Identifiers{
																																											"zp",
																																										},
																																									},
																																									Id: "zp",
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Index{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(630),
																																												Column: int(28),
																																											},
																																											End: ast.Location{
																																												Line: int(630),
																																												Column: int(40),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p6834,
																																										FreeVars: ast.Identifiers{
																																											"cflags",
																																										},
																																									},
																																									Target: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(630),
																																													Column: int(28),
																																												},
																																												End: ast.Location{
																																													Line: int(630),
																																													Column: int(34),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{
																																												ast.FodderElement{
																																													Kind: ast.FodderKind(0),
																																													Blanks: int(0),
																																													Indent: int(27),
																																													Comment: []string{},
																																												},
																																											},
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{
																																												"cflags",
																																											},
																																										},
																																										Id: "cflags",
																																									},
																																									LeftBracketFodder: ast.Fodder{},
																																									Index: &ast.LiteralString{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												End: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												File: nil,
																																											},
																																											Fodder: nil,
																																											Ctx: nil,
																																											FreeVars: nil,
																																										},
																																										Value: "blank",
																																										Kind: ast.LiteralStringKind(1),
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																									},
																																									RightBracketFodder: ast.Fodder{},
																																									Id: nil,
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Index{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(631),
																																												Column: int(28),
																																											},
																																											End: ast.Location{
																																												Line: int(631),
																																												Column: int(39),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p6834,
																																										FreeVars: ast.Identifiers{
																																											"cflags",
																																										},
																																									},
																																									Target: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(631),
																																													Column: int(28),
																																												},
																																												End: ast.Location{
																																													Line: int(631),
																																													Column: int(34),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{
																																												ast.FodderElement{
																																													Kind: ast.FodderKind(0),
																																													Blanks: int(0),
																																													Indent: int(27),
																																													Comment: []string{},
																																												},
																																											},
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{
																																												"cflags",
																																											},
																																										},
																																										Id: "cflags",
																																									},
																																									LeftBracketFodder: ast.Fodder{},
																																									Index: &ast.LiteralString{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												End: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												File: nil,
																																											},
																																											Fodder: nil,
																																											Ctx: nil,
																																											FreeVars: nil,
																																										},
																																										Value: "plus",
																																										Kind: ast.LiteralStringKind(1),
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																									},
																																									RightBracketFodder: ast.Fodder{},
																																									Id: nil,
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Index{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(632),
																																												Column: int(28),
																																											},
																																											End: ast.Location{
																																												Line: int(632),
																																												Column: int(38),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p6834,
																																										FreeVars: ast.Identifiers{
																																											"cflags",
																																										},
																																									},
																																									Target: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(632),
																																													Column: int(28),
																																												},
																																												End: ast.Location{
																																													Line: int(632),
																																													Column: int(34),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{
																																												ast.FodderElement{
																																													Kind: ast.FodderKind(0),
																																													Blanks: int(0),
																																													Indent: int(27),
																																													Comment: []string{},
																																												},
																																											},
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{
																																												"cflags",
																																											},
																																										},
																																										Id: "cflags",
																																									},
																																									LeftBracketFodder: ast.Fodder{},
																																									Index: &ast.LiteralString{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												End: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												File: nil,
																																											},
																																											Fodder: nil,
																																											Ctx: nil,
																																											FreeVars: nil,
																																										},
																																										Value: "alt",
																																										Kind: ast.LiteralStringKind(1),
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																									},
																																									RightBracketFodder: ast.Fodder{},
																																									Id: nil,
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.LiteralBoolean{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(633),
																																												Column: int(28),
																																											},
																																											End: ast.Location{
																																												Line: int(633),
																																												Column: int(32),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{
																																											ast.FodderElement{
																																												Kind: ast.FodderKind(0),
																																												Blanks: int(0),
																																												Indent: int(27),
																																												Comment: []string{},
																																											},
																																										},
																																										Ctx: p6834,
																																										FreeVars: nil,
																																									},
																																									Value: true,
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Index{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(634),
																																												Column: int(28),
																																											},
																																											End: ast.Location{
																																												Line: int(634),
																																												Column: int(37),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p6834,
																																										FreeVars: ast.Identifiers{
																																											"code",
																																										},
																																									},
																																									Target: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(634),
																																													Column: int(28),
																																												},
																																												End: ast.Location{
																																													Line: int(634),
																																													Column: int(32),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{
																																												ast.FodderElement{
																																													Kind: ast.FodderKind(0),
																																													Blanks: int(0),
																																													Indent: int(27),
																																													Comment: []string{},
																																												},
																																											},
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{
																																												"code",
																																											},
																																										},
																																										Id: "code",
																																									},
																																									LeftBracketFodder: ast.Fodder{},
																																									Index: &ast.LiteralString{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												End: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												File: nil,
																																											},
																																											Fodder: nil,
																																											Ctx: nil,
																																											FreeVars: nil,
																																										},
																																										Value: "caps",
																																										Kind: ast.LiteralStringKind(1),
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																									},
																																									RightBracketFodder: ast.Fodder{},
																																									Id: nil,
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(635),
																																												Column: int(28),
																																											},
																																											End: ast.Location{
																																												Line: int(635),
																																												Column: int(34),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{
																																											ast.FodderElement{
																																												Kind: ast.FodderKind(0),
																																												Blanks: int(0),
																																												Indent: int(27),
																																												Comment: []string{},
																																											},
																																										},
																																										Ctx: p6834,
																																										FreeVars: ast.Identifiers{
																																											"fpprec",
																																										},
																																									},
																																									Id: "fpprec",
																																								},
																																								CommaFodder: nil,
																																							},
																																						},
																																						Named: nil,
																																					},
																																					TrailingComma: false,
																																					TailStrict: false,
																																					FodderRight: ast.Fodder{},
																																					TailStrictFodder: nil,
																																				},
																																			},
																																			ElseFodder: ast.Fodder{
																																				ast.FodderElement{
																																					Kind: ast.FodderKind(0),
																																					Blanks: int(0),
																																					Indent: int(6),
																																					Comment: []string{},
																																				},
																																			},
																																			BranchFalse: &ast.Conditional{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(636),
																																							Column: int(12),
																																						},
																																						End: ast.Location{
																																							Line: int(671),
																																							Column: int(44),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p6275,
																																					FreeVars: ast.Identifiers{
																																						"cflags",
																																						"code",
																																						"fpprec",
																																						"i",
																																						"render_float_dec",
																																						"render_float_sci",
																																						"std",
																																						"val",
																																						"zp",
																																					},
																																				},
																																				Cond: &ast.Binary{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(636),
																																								Column: int(15),
																																							},
																																							End: ast.Location{
																																								Line: int(636),
																																								Column: int(32),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p6275,
																																						FreeVars: ast.Identifiers{
																																							"code",
																																						},
																																					},
																																					Left: &ast.Index{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(636),
																																									Column: int(15),
																																								},
																																								End: ast.Location{
																																									Line: int(636),
																																									Column: int(25),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p6275,
																																							FreeVars: ast.Identifiers{
																																								"code",
																																							},
																																						},
																																						Target: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(636),
																																										Column: int(15),
																																									},
																																									End: ast.Location{
																																										Line: int(636),
																																										Column: int(19),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: nil,
																																								FreeVars: ast.Identifiers{
																																									"code",
																																								},
																																							},
																																							Id: "code",
																																						},
																																						LeftBracketFodder: ast.Fodder{},
																																						Index: &ast.LiteralString{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																									End: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																									File: nil,
																																								},
																																								Fodder: nil,
																																								Ctx: nil,
																																								FreeVars: nil,
																																							},
																																							Value: "ctype",
																																							Kind: ast.LiteralStringKind(1),
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																						},
																																						RightBracketFodder: ast.Fodder{},
																																						Id: nil,
																																					},
																																					OpFodder: ast.Fodder{},
																																					Op: ast.BinaryOp(12),
																																					Right: &ast.LiteralString{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(636),
																																									Column: int(29),
																																								},
																																								End: ast.Location{
																																									Line: int(636),
																																									Column: int(32),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p6275,
																																							FreeVars: nil,
																																						},
																																						Value: "g",
																																						Kind: ast.LiteralStringKind(1),
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																					},
																																				},
																																				ThenFodder: ast.Fodder{},
																																				BranchTrue: &ast.Conditional{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(637),
																																								Column: int(9),
																																							},
																																							End: ast.Location{
																																								Line: int(659),
																																								Column: int(56),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{
																																							ast.FodderElement{
																																								Kind: ast.FodderKind(0),
																																								Blanks: int(0),
																																								Indent: int(8),
																																								Comment: []string{},
																																							},
																																						},
																																						Ctx: p6275,
																																						FreeVars: ast.Identifiers{
																																							"cflags",
																																							"code",
																																							"fpprec",
																																							"i",
																																							"render_float_dec",
																																							"render_float_sci",
																																							"std",
																																							"val",
																																							"zp",
																																						},
																																					},
																																					Cond: &ast.Binary{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(637),
																																									Column: int(12),
																																								},
																																								End: ast.Location{
																																									Line: int(637),
																																									Column: int(37),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p6275,
																																							FreeVars: ast.Identifiers{
																																								"std",
																																								"val",
																																							},
																																						},
																																						Left: &ast.Apply{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(637),
																																										Column: int(12),
																																									},
																																									End: ast.Location{
																																										Line: int(637),
																																										Column: int(25),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6275,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																									"val",
																																								},
																																							},
																																							Target: &ast.Index{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(637),
																																											Column: int(12),
																																										},
																																										End: ast.Location{
																																											Line: int(637),
																																											Column: int(20),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p6275,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																									},
																																								},
																																								Target: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(637),
																																												Column: int(12),
																																											},
																																											End: ast.Location{
																																												Line: int(637),
																																												Column: int(15),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																										},
																																									},
																																									Id: "std",
																																								},
																																								LeftBracketFodder: ast.Fodder{},
																																								Index: &ast.LiteralString{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											File: nil,
																																										},
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: nil,
																																									},
																																									Value: "type",
																																									Kind: ast.LiteralStringKind(1),
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								Id: nil,
																																							},
																																							FodderLeft: ast.Fodder{},
																																							Arguments: ast.Arguments{
																																								Positional: []ast.CommaSeparatedExpr{
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(637),
																																														Column: int(21),
																																													},
																																													End: ast.Location{
																																														Line: int(637),
																																														Column: int(24),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p6893,
																																												FreeVars: ast.Identifiers{
																																													"val",
																																												},
																																											},
																																											Id: "val",
																																										},
																																										CommaFodder: nil,
																																									},
																																								},
																																								Named: nil,
																																							},
																																							TrailingComma: false,
																																							TailStrict: false,
																																							FodderRight: ast.Fodder{},
																																							TailStrictFodder: nil,
																																						},
																																						OpFodder: ast.Fodder{},
																																						Op: ast.BinaryOp(13),
																																						Right: &ast.LiteralString{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(637),
																																										Column: int(29),
																																									},
																																									End: ast.Location{
																																										Line: int(637),
																																										Column: int(37),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6275,
																																								FreeVars: nil,
																																							},
																																							Value: "number",
																																							Kind: ast.LiteralStringKind(1),
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																						},
																																					},
																																					ThenFodder: ast.Fodder{},
																																					BranchTrue: &ast.Error{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(638),
																																									Column: int(11),
																																								},
																																								End: ast.Location{
																																									Line: int(639),
																																									Column: int(47),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(10),
																																									Comment: []string{},
																																								},
																																							},
																																							Ctx: p6275,
																																							FreeVars: ast.Identifiers{
																																								"i",
																																								"std",
																																								"val",
																																							},
																																						},
																																						Expr: &ast.Binary{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(638),
																																										Column: int(17),
																																									},
																																									End: ast.Location{
																																										Line: int(639),
																																										Column: int(47),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6275,
																																								FreeVars: ast.Identifiers{
																																									"i",
																																									"std",
																																									"val",
																																								},
																																							},
																																							Left: &ast.Binary{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(638),
																																											Column: int(17),
																																										},
																																										End: ast.Location{
																																											Line: int(639),
																																											Column: int(31),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p6275,
																																									FreeVars: ast.Identifiers{
																																										"i",
																																									},
																																								},
																																								Left: &ast.Binary{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(638),
																																												Column: int(17),
																																											},
																																											End: ast.Location{
																																												Line: int(639),
																																												Column: int(20),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p6275,
																																										FreeVars: ast.Identifiers{
																																											"i",
																																										},
																																									},
																																									Left: &ast.LiteralString{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(638),
																																													Column: int(17),
																																												},
																																												End: ast.Location{
																																													Line: int(638),
																																													Column: int(45),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p6275,
																																											FreeVars: nil,
																																										},
																																										Value: "Format required number at ",
																																										Kind: ast.LiteralStringKind(1),
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																									},
																																									OpFodder: ast.Fodder{
																																										ast.FodderElement{
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(16),
																																											Comment: []string{},
																																										},
																																									},
																																									Op: ast.BinaryOp(3),
																																									Right: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(639),
																																													Column: int(19),
																																												},
																																												End: ast.Location{
																																													Line: int(639),
																																													Column: int(20),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p6275,
																																											FreeVars: ast.Identifiers{
																																												"i",
																																											},
																																										},
																																										Id: "i",
																																									},
																																								},
																																								OpFodder: ast.Fodder{},
																																								Op: ast.BinaryOp(3),
																																								Right: &ast.LiteralString{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(639),
																																												Column: int(23),
																																											},
																																											End: ast.Location{
																																												Line: int(639),
																																												Column: int(31),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p6275,
																																										FreeVars: nil,
																																									},
																																									Value: ", got ",
																																									Kind: ast.LiteralStringKind(1),
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																								},
																																							},
																																							OpFodder: ast.Fodder{},
																																							Op: ast.BinaryOp(3),
																																							Right: &ast.Apply{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(639),
																																											Column: int(34),
																																										},
																																										End: ast.Location{
																																											Line: int(639),
																																											Column: int(47),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p6275,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																										"val",
																																									},
																																								},
																																								Target: &ast.Index{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(639),
																																												Column: int(34),
																																											},
																																											End: ast.Location{
																																												Line: int(639),
																																												Column: int(42),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p6275,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																										},
																																									},
																																									Target: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(639),
																																													Column: int(34),
																																												},
																																												End: ast.Location{
																																													Line: int(639),
																																													Column: int(37),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{
																																												"std",
																																											},
																																										},
																																										Id: "std",
																																									},
																																									LeftBracketFodder: ast.Fodder{},
																																									Index: &ast.LiteralString{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												End: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												File: nil,
																																											},
																																											Fodder: nil,
																																											Ctx: nil,
																																											FreeVars: nil,
																																										},
																																										Value: "type",
																																										Kind: ast.LiteralStringKind(1),
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																									},
																																									RightBracketFodder: ast.Fodder{},
																																									Id: nil,
																																								},
																																								FodderLeft: ast.Fodder{},
																																								Arguments: ast.Arguments{
																																									Positional: []ast.CommaSeparatedExpr{
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Var{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(639),
																																															Column: int(43),
																																														},
																																														End: ast.Location{
																																															Line: int(639),
																																															Column: int(46),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: p6919,
																																													FreeVars: ast.Identifiers{
																																														"val",
																																													},
																																												},
																																												Id: "val",
																																											},
																																											CommaFodder: nil,
																																										},
																																									},
																																									Named: nil,
																																								},
																																								TrailingComma: false,
																																								TailStrict: false,
																																								FodderRight: ast.Fodder{},
																																								TailStrictFodder: nil,
																																							},
																																						},
																																					},
																																					ElseFodder: ast.Fodder{
																																						ast.FodderElement{
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(8),
																																							Comment: []string{},
																																						},
																																					},
																																					BranchFalse: &ast.Local{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(641),
																																									Column: int(11),
																																								},
																																								End: ast.Location{
																																									Line: int(659),
																																									Column: int(56),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(10),
																																									Comment: []string{},
																																								},
																																							},
																																							Ctx: p6275,
																																							FreeVars: ast.Identifiers{
																																								"cflags",
																																								"code",
																																								"fpprec",
																																								"render_float_dec",
																																								"render_float_sci",
																																								"std",
																																								"val",
																																								"zp",
																																							},
																																						},
																																						Binds: ast.LocalBinds{
																																							ast.LocalBind{
																																								VarFodder: ast.Fodder{},
																																								Variable: "exponent",
																																								EqFodder: ast.Fodder{},
																																								Body: &ast.Apply{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(641),
																																												Column: int(28),
																																											},
																																											End: ast.Location{
																																												Line: int(641),
																																												Column: int(74),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p6927,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																											"val",
																																										},
																																									},
																																									Target: &ast.Index{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(641),
																																													Column: int(28),
																																												},
																																												End: ast.Location{
																																													Line: int(641),
																																													Column: int(37),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p6927,
																																											FreeVars: ast.Identifiers{
																																												"std",
																																											},
																																										},
																																										Target: &ast.Var{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(641),
																																														Column: int(28),
																																													},
																																													End: ast.Location{
																																														Line: int(641),
																																														Column: int(31),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: nil,
																																												FreeVars: ast.Identifiers{
																																													"std",
																																												},
																																											},
																																											Id: "std",
																																										},
																																										LeftBracketFodder: ast.Fodder{},
																																										Index: &ast.LiteralString{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(0),
																																														Column: int(0),
																																													},
																																													End: ast.Location{
																																														Line: int(0),
																																														Column: int(0),
																																													},
																																													File: nil,
																																												},
																																												Fodder: nil,
																																												Ctx: nil,
																																												FreeVars: nil,
																																											},
																																											Value: "floor",
																																											Kind: ast.LiteralStringKind(1),
																																											BlockIndent: "",
																																											BlockTermIndent: "",
																																										},
																																										RightBracketFodder: ast.Fodder{},
																																										Id: nil,
																																									},
																																									FodderLeft: ast.Fodder{},
																																									Arguments: ast.Arguments{
																																										Positional: []ast.CommaSeparatedExpr{
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Binary{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "<std>",
																																															Begin: ast.Location{
																																																Line: int(641),
																																																Column: int(38),
																																															},
																																															End: ast.Location{
																																																Line: int(641),
																																																Column: int(73),
																																															},
																																															File: p1,
																																														},
																																														Fodder: ast.Fodder{},
																																														Ctx: p6936,
																																														FreeVars: ast.Identifiers{
																																															"std",
																																															"val",
																																														},
																																													},
																																													Left: &ast.Apply{
																																														NodeBase: ast.NodeBase{
																																															LocRange: ast.LocationRange{
																																																FileName: "<std>",
																																																Begin: ast.Location{
																																																	Line: int(641),
																																																	Column: int(38),
																																																},
																																																End: ast.Location{
																																																	Line: int(641),
																																																	Column: int(59),
																																																},
																																																File: p1,
																																															},
																																															Fodder: ast.Fodder{},
																																															Ctx: p6936,
																																															FreeVars: ast.Identifiers{
																																																"std",
																																																"val",
																																															},
																																														},
																																														Target: &ast.Index{
																																															NodeBase: ast.NodeBase{
																																																LocRange: ast.LocationRange{
																																																	FileName: "<std>",
																																																	Begin: ast.Location{
																																																		Line: int(641),
																																																		Column: int(38),
																																																	},
																																																	End: ast.Location{
																																																		Line: int(641),
																																																		Column: int(45),
																																																	},
																																																	File: p1,
																																																},
																																																Fodder: ast.Fodder{},
																																																Ctx: p6936,
																																																FreeVars: ast.Identifiers{
																																																	"std",
																																																},
																																															},
																																															Target: &ast.Var{
																																																NodeBase: ast.NodeBase{
																																																	LocRange: ast.LocationRange{
																																																		FileName: "<std>",
																																																		Begin: ast.Location{
																																																			Line: int(641),
																																																			Column: int(38),
																																																		},
																																																		End: ast.Location{
																																																			Line: int(641),
																																																			Column: int(41),
																																																		},
																																																		File: p1,
																																																	},
																																																	Fodder: ast.Fodder{},
																																																	Ctx: nil,
																																																	FreeVars: ast.Identifiers{
																																																		"std",
																																																	},
																																																},
																																																Id: "std",
																																															},
																																															LeftBracketFodder: ast.Fodder{},
																																															Index: &ast.LiteralString{
																																																NodeBase: ast.NodeBase{
																																																	LocRange: ast.LocationRange{
																																																		FileName: "",
																																																		Begin: ast.Location{
																																																			Line: int(0),
																																																			Column: int(0),
																																																		},
																																																		End: ast.Location{
																																																			Line: int(0),
																																																			Column: int(0),
																																																		},
																																																		File: nil,
																																																	},
																																																	Fodder: nil,
																																																	Ctx: nil,
																																																	FreeVars: nil,
																																																},
																																																Value: "log",
																																																Kind: ast.LiteralStringKind(1),
																																																BlockIndent: "",
																																																BlockTermIndent: "",
																																															},
																																															RightBracketFodder: ast.Fodder{},
																																															Id: nil,
																																														},
																																														FodderLeft: ast.Fodder{},
																																														Arguments: ast.Arguments{
																																															Positional: []ast.CommaSeparatedExpr{
																																																ast.CommaSeparatedExpr{
																																																	Expr: &ast.Apply{
																																																		NodeBase: ast.NodeBase{
																																																			LocRange: ast.LocationRange{
																																																				FileName: "<std>",
																																																				Begin: ast.Location{
																																																					Line: int(641),
																																																					Column: int(46),
																																																				},
																																																				End: ast.Location{
																																																					Line: int(641),
																																																					Column: int(58),
																																																				},
																																																				File: p1,
																																																			},
																																																			Fodder: ast.Fodder{},
																																																			Ctx: p6947,
																																																			FreeVars: ast.Identifiers{
																																																				"std",
																																																				"val",
																																																			},
																																																		},
																																																		Target: &ast.Index{
																																																			NodeBase: ast.NodeBase{
																																																				LocRange: ast.LocationRange{
																																																					FileName: "<std>",
																																																					Begin: ast.Location{
																																																						Line: int(641),
																																																						Column: int(46),
																																																					},
																																																					End: ast.Location{
																																																						Line: int(641),
																																																						Column: int(53),
																																																					},
																																																					File: p1,
																																																				},
																																																				Fodder: ast.Fodder{},
																																																				Ctx: p6947,
																																																				FreeVars: ast.Identifiers{
																																																					"std",
																																																				},
																																																			},
																																																			Target: &ast.Var{
																																																				NodeBase: ast.NodeBase{
																																																					LocRange: ast.LocationRange{
																																																						FileName: "<std>",
																																																						Begin: ast.Location{
																																																							Line: int(641),
																																																							Column: int(46),
																																																						},
																																																						End: ast.Location{
																																																							Line: int(641),
																																																							Column: int(49),
																																																						},
																																																						File: p1,
																																																					},
																																																					Fodder: ast.Fodder{},
																																																					Ctx: nil,
																																																					FreeVars: ast.Identifiers{
																																																						"std",
																																																					},
																																																				},
																																																				Id: "std",
																																																			},
																																																			LeftBracketFodder: ast.Fodder{},
																																																			Index: &ast.LiteralString{
																																																				NodeBase: ast.NodeBase{
																																																					LocRange: ast.LocationRange{
																																																						FileName: "",
																																																						Begin: ast.Location{
																																																							Line: int(0),
																																																							Column: int(0),
																																																						},
																																																						End: ast.Location{
																																																							Line: int(0),
																																																							Column: int(0),
																																																						},
																																																						File: nil,
																																																					},
																																																					Fodder: nil,
																																																					Ctx: nil,
																																																					FreeVars: nil,
																																																				},
																																																				Value: "abs",
																																																				Kind: ast.LiteralStringKind(1),
																																																				BlockIndent: "",
																																																				BlockTermIndent: "",
																																																			},
																																																			RightBracketFodder: ast.Fodder{},
																																																			Id: nil,
																																																		},
																																																		FodderLeft: ast.Fodder{},
																																																		Arguments: ast.Arguments{
																																																			Positional: []ast.CommaSeparatedExpr{
																																																				ast.CommaSeparatedExpr{
																																																					Expr: &ast.Var{
																																																						NodeBase: ast.NodeBase{
																																																							LocRange: ast.LocationRange{
																																																								FileName: "<std>",
																																																								Begin: ast.Location{
																																																									Line: int(641),
																																																									Column: int(54),
																																																								},
																																																								End: ast.Location{
																																																									Line: int(641),
																																																									Column: int(57),
																																																								},
																																																								File: p1,
																																																							},
																																																							Fodder: ast.Fodder{},
																																																							Ctx: p6956,
																																																							FreeVars: ast.Identifiers{
																																																								"val",
																																																							},
																																																						},
																																																						Id: "val",
																																																					},
																																																					CommaFodder: nil,
																																																				},
																																																			},
																																																			Named: nil,
																																																		},
																																																		TrailingComma: false,
																																																		TailStrict: false,
																																																		FodderRight: ast.Fodder{},
																																																		TailStrictFodder: nil,
																																																	},
																																																	CommaFodder: nil,
																																																},
																																															},
																																															Named: nil,
																																														},
																																														TrailingComma: false,
																																														TailStrict: false,
																																														FodderRight: ast.Fodder{},
																																														TailStrictFodder: nil,
																																													},
																																													OpFodder: ast.Fodder{},
																																													Op: ast.BinaryOp(1),
																																													Right: &ast.Apply{
																																														NodeBase: ast.NodeBase{
																																															LocRange: ast.LocationRange{
																																																FileName: "<std>",
																																																Begin: ast.Location{
																																																	Line: int(641),
																																																	Column: int(62),
																																																},
																																																End: ast.Location{
																																																	Line: int(641),
																																																	Column: int(73),
																																																},
																																																File: p1,
																																															},
																																															Fodder: ast.Fodder{},
																																															Ctx: p6936,
																																															FreeVars: ast.Identifiers{
																																																"std",
																																															},
																																														},
																																														Target: &ast.Index{
																																															NodeBase: ast.NodeBase{
																																																LocRange: ast.LocationRange{
																																																	FileName: "<std>",
																																																	Begin: ast.Location{
																																																		Line: int(641),
																																																		Column: int(62),
																																																	},
																																																	End: ast.Location{
																																																		Line: int(641),
																																																		Column: int(69),
																																																	},
																																																	File: p1,
																																																},
																																																Fodder: ast.Fodder{},
																																																Ctx: p6936,
																																																FreeVars: ast.Identifiers{
																																																	"std",
																																																},
																																															},
																																															Target: &ast.Var{
																																																NodeBase: ast.NodeBase{
																																																	LocRange: ast.LocationRange{
																																																		FileName: "<std>",
																																																		Begin: ast.Location{
																																																			Line: int(641),
																																																			Column: int(62),
																																																		},
																																																		End: ast.Location{
																																																			Line: int(641),
																																																			Column: int(65),
																																																		},
																																																		File: p1,
																																																	},
																																																	Fodder: ast.Fodder{},
																																																	Ctx: nil,
																																																	FreeVars: ast.Identifiers{
																																																		"std",
																																																	},
																																																},
																																																Id: "std",
																																															},
																																															LeftBracketFodder: ast.Fodder{},
																																															Index: &ast.LiteralString{
																																																NodeBase: ast.NodeBase{
																																																	LocRange: ast.LocationRange{
																																																		FileName: "",
																																																		Begin: ast.Location{
																																																			Line: int(0),
																																																			Column: int(0),
																																																		},
																																																		End: ast.Location{
																																																			Line: int(0),
																																																			Column: int(0),
																																																		},
																																																		File: nil,
																																																	},
																																																	Fodder: nil,
																																																	Ctx: nil,
																																																	FreeVars: nil,
																																																},
																																																Value: "log",
																																																Kind: ast.LiteralStringKind(1),
																																																BlockIndent: "",
																																																BlockTermIndent: "",
																																															},
																																															RightBracketFodder: ast.Fodder{},
																																															Id: nil,
																																														},
																																														FodderLeft: ast.Fodder{},
																																														Arguments: ast.Arguments{
																																															Positional: []ast.CommaSeparatedExpr{
																																																ast.CommaSeparatedExpr{
																																																	Expr: &ast.LiteralNumber{
																																																		NodeBase: ast.NodeBase{
																																																			LocRange: ast.LocationRange{
																																																				FileName: "<std>",
																																																				Begin: ast.Location{
																																																					Line: int(641),
																																																					Column: int(70),
																																																				},
																																																				End: ast.Location{
																																																					Line: int(641),
																																																					Column: int(72),
																																																				},
																																																				File: p1,
																																																			},
																																																			Fodder: ast.Fodder{},
																																																			Ctx: p6967,
																																																			FreeVars: nil,
																																																		},
																																																		OriginalString: "10",
																																																	},
																																																	CommaFodder: nil,
																																																},
																																															},
																																															Named: nil,
																																														},
																																														TrailingComma: false,
																																														TailStrict: false,
																																														FodderRight: ast.Fodder{},
																																														TailStrictFodder: nil,
																																													},
																																												},
																																												CommaFodder: nil,
																																											},
																																										},
																																										Named: nil,
																																									},
																																									TrailingComma: false,
																																									TailStrict: false,
																																									FodderRight: ast.Fodder{},
																																									TailStrictFodder: nil,
																																								},
																																								Fun: nil,
																																								CloseFodder: ast.Fodder{},
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(641),
																																										Column: int(17),
																																									},
																																									End: ast.Location{
																																										Line: int(641),
																																										Column: int(74),
																																									},
																																									File: p1,
																																								},
																																							},
																																						},
																																						Body: &ast.Conditional{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(642),
																																										Column: int(11),
																																									},
																																									End: ast.Location{
																																										Line: int(659),
																																										Column: int(56),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{
																																									ast.FodderElement{
																																										Kind: ast.FodderKind(0),
																																										Blanks: int(0),
																																										Indent: int(10),
																																										Comment: []string{},
																																									},
																																								},
																																								Ctx: p6275,
																																								FreeVars: ast.Identifiers{
																																									"cflags",
																																									"code",
																																									"exponent",
																																									"fpprec",
																																									"render_float_dec",
																																									"render_float_sci",
																																									"std",
																																									"val",
																																									"zp",
																																								},
																																							},
																																							Cond: &ast.Binary{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(642),
																																											Column: int(14),
																																										},
																																										End: ast.Location{
																																											Line: int(642),
																																											Column: int(49),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p6275,
																																									FreeVars: ast.Identifiers{
																																										"exponent",
																																										"fpprec",
																																									},
																																								},
																																								Left: &ast.Binary{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(642),
																																												Column: int(14),
																																											},
																																											End: ast.Location{
																																												Line: int(642),
																																												Column: int(27),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p6275,
																																										FreeVars: ast.Identifiers{
																																											"exponent",
																																										},
																																									},
																																									Left: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(642),
																																													Column: int(14),
																																												},
																																												End: ast.Location{
																																													Line: int(642),
																																													Column: int(22),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p6275,
																																											FreeVars: ast.Identifiers{
																																												"exponent",
																																											},
																																										},
																																										Id: "exponent",
																																									},
																																									OpFodder: ast.Fodder{},
																																									Op: ast.BinaryOp(9),
																																									Right: &ast.Unary{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(642),
																																													Column: int(25),
																																												},
																																												End: ast.Location{
																																													Line: int(642),
																																													Column: int(27),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p6275,
																																											FreeVars: nil,
																																										},
																																										Op: ast.UnaryOp(3),
																																										Expr: &ast.LiteralNumber{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(642),
																																														Column: int(26),
																																													},
																																													End: ast.Location{
																																														Line: int(642),
																																														Column: int(27),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p6275,
																																												FreeVars: nil,
																																											},
																																											OriginalString: "4",
																																										},
																																									},
																																								},
																																								OpFodder: ast.Fodder{},
																																								Op: ast.BinaryOp(18),
																																								Right: &ast.Binary{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(642),
																																												Column: int(31),
																																											},
																																											End: ast.Location{
																																												Line: int(642),
																																												Column: int(49),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p6275,
																																										FreeVars: ast.Identifiers{
																																											"exponent",
																																											"fpprec",
																																										},
																																									},
																																									Left: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(642),
																																													Column: int(31),
																																												},
																																												End: ast.Location{
																																													Line: int(642),
																																													Column: int(39),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p6275,
																																											FreeVars: ast.Identifiers{
																																												"exponent",
																																											},
																																										},
																																										Id: "exponent",
																																									},
																																									OpFodder: ast.Fodder{},
																																									Op: ast.BinaryOp(8),
																																									Right: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(642),
																																													Column: int(43),
																																												},
																																												End: ast.Location{
																																													Line: int(642),
																																													Column: int(49),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p6275,
																																											FreeVars: ast.Identifiers{
																																												"fpprec",
																																											},
																																										},
																																										Id: "fpprec",
																																									},
																																								},
																																							},
																																							ThenFodder: ast.Fodder{},
																																							BranchTrue: &ast.Apply{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(643),
																																											Column: int(13),
																																										},
																																										End: ast.Location{
																																											Line: int(650),
																																											Column: int(41),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p6275,
																																									FreeVars: ast.Identifiers{
																																										"cflags",
																																										"code",
																																										"fpprec",
																																										"render_float_sci",
																																										"val",
																																										"zp",
																																									},
																																								},
																																								Target: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(643),
																																												Column: int(13),
																																											},
																																											End: ast.Location{
																																												Line: int(643),
																																												Column: int(29),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{
																																											ast.FodderElement{
																																												Kind: ast.FodderKind(0),
																																												Blanks: int(0),
																																												Indent: int(12),
																																												Comment: []string{},
																																											},
																																										},
																																										Ctx: p6275,
																																										FreeVars: ast.Identifiers{
																																											"render_float_sci",
																																										},
																																									},
																																									Id: "render_float_sci",
																																								},
																																								FodderLeft: ast.Fodder{},
																																								Arguments: ast.Arguments{
																																									Positional: []ast.CommaSeparatedExpr{
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Var{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(643),
																																															Column: int(30),
																																														},
																																														End: ast.Location{
																																															Line: int(643),
																																															Column: int(33),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: p6992,
																																													FreeVars: ast.Identifiers{
																																														"val",
																																													},
																																												},
																																												Id: "val",
																																											},
																																											CommaFodder: ast.Fodder{},
																																										},
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Var{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(644),
																																															Column: int(30),
																																														},
																																														End: ast.Location{
																																															Line: int(644),
																																															Column: int(32),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{
																																														ast.FodderElement{
																																															Kind: ast.FodderKind(0),
																																															Blanks: int(0),
																																															Indent: int(29),
																																															Comment: []string{},
																																														},
																																													},
																																													Ctx: p6992,
																																													FreeVars: ast.Identifiers{
																																														"zp",
																																													},
																																												},
																																												Id: "zp",
																																											},
																																											CommaFodder: ast.Fodder{},
																																										},
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Index{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(645),
																																															Column: int(30),
																																														},
																																														End: ast.Location{
																																															Line: int(645),
																																															Column: int(42),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: p6992,
																																													FreeVars: ast.Identifiers{
																																														"cflags",
																																													},
																																												},
																																												Target: &ast.Var{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "<std>",
																																															Begin: ast.Location{
																																																Line: int(645),
																																																Column: int(30),
																																															},
																																															End: ast.Location{
																																																Line: int(645),
																																																Column: int(36),
																																															},
																																															File: p1,
																																														},
																																														Fodder: ast.Fodder{
																																															ast.FodderElement{
																																																Kind: ast.FodderKind(0),
																																																Blanks: int(0),
																																																Indent: int(29),
																																																Comment: []string{},
																																															},
																																														},
																																														Ctx: nil,
																																														FreeVars: ast.Identifiers{
																																															"cflags",
																																														},
																																													},
																																													Id: "cflags",
																																												},
																																												LeftBracketFodder: ast.Fodder{},
																																												Index: &ast.LiteralString{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																															End: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																															File: nil,
																																														},
																																														Fodder: nil,
																																														Ctx: nil,
																																														FreeVars: nil,
																																													},
																																													Value: "blank",
																																													Kind: ast.LiteralStringKind(1),
																																													BlockIndent: "",
																																													BlockTermIndent: "",
																																												},
																																												RightBracketFodder: ast.Fodder{},
																																												Id: nil,
																																											},
																																											CommaFodder: ast.Fodder{},
																																										},
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Index{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(646),
																																															Column: int(30),
																																														},
																																														End: ast.Location{
																																															Line: int(646),
																																															Column: int(41),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: p6992,
																																													FreeVars: ast.Identifiers{
																																														"cflags",
																																													},
																																												},
																																												Target: &ast.Var{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "<std>",
																																															Begin: ast.Location{
																																																Line: int(646),
																																																Column: int(30),
																																															},
																																															End: ast.Location{
																																																Line: int(646),
																																																Column: int(36),
																																															},
																																															File: p1,
																																														},
																																														Fodder: ast.Fodder{
																																															ast.FodderElement{
																																																Kind: ast.FodderKind(0),
																																																Blanks: int(0),
																																																Indent: int(29),
																																																Comment: []string{},
																																															},
																																														},
																																														Ctx: nil,
																																														FreeVars: ast.Identifiers{
																																															"cflags",
																																														},
																																													},
																																													Id: "cflags",
																																												},
																																												LeftBracketFodder: ast.Fodder{},
																																												Index: &ast.LiteralString{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																															End: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																															File: nil,
																																														},
																																														Fodder: nil,
																																														Ctx: nil,
																																														FreeVars: nil,
																																													},
																																													Value: "plus",
																																													Kind: ast.LiteralStringKind(1),
																																													BlockIndent: "",
																																													BlockTermIndent: "",
																																												},
																																												RightBracketFodder: ast.Fodder{},
																																												Id: nil,
																																											},
																																											CommaFodder: ast.Fodder{},
																																										},
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Index{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(647),
																																															Column: int(30),
																																														},
																																														End: ast.Location{
																																															Line: int(647),
																																															Column: int(40),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: p6992,
																																													FreeVars: ast.Identifiers{
																																														"cflags",
																																													},
																																												},
																																												Target: &ast.Var{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "<std>",
																																															Begin: ast.Location{
																																																Line: int(647),
																																																Column: int(30),
																																															},
																																															End: ast.Location{
																																																Line: int(647),
																																																Column: int(36),
																																															},
																																															File: p1,
																																														},
																																														Fodder: ast.Fodder{
																																															ast.FodderElement{
																																																Kind: ast.FodderKind(0),
																																																Blanks: int(0),
																																																Indent: int(29),
																																																Comment: []string{},
																																															},
																																														},
																																														Ctx: nil,
																																														FreeVars: ast.Identifiers{
																																															"cflags",
																																														},
																																													},
																																													Id: "cflags",
																																												},
																																												LeftBracketFodder: ast.Fodder{},
																																												Index: &ast.LiteralString{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																															End: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																															File: nil,
																																														},
																																														Fodder: nil,
																																														Ctx: nil,
																																														FreeVars: nil,
																																													},
																																													Value: "alt",
																																													Kind: ast.LiteralStringKind(1),
																																													BlockIndent: "",
																																													BlockTermIndent: "",
																																												},
																																												RightBracketFodder: ast.Fodder{},
																																												Id: nil,
																																											},
																																											CommaFodder: ast.Fodder{},
																																										},
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Index{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(648),
																																															Column: int(30),
																																														},
																																														End: ast.Location{
																																															Line: int(648),
																																															Column: int(40),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: p6992,
																																													FreeVars: ast.Identifiers{
																																														"cflags",
																																													},
																																												},
																																												Target: &ast.Var{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "<std>",
																																															Begin: ast.Location{
																																																Line: int(648),
																																																Column: int(30),
																																															},
																																															End: ast.Location{
																																																Line: int(648),
																																																Column: int(36),
																																															},
																																															File: p1,
																																														},
																																														Fodder: ast.Fodder{
																																															ast.FodderElement{
																																																Kind: ast.FodderKind(0),
																																																Blanks: int(0),
																																																Indent: int(29),
																																																Comment: []string{},
																																															},
																																														},
																																														Ctx: nil,
																																														FreeVars: ast.Identifiers{
																																															"cflags",
																																														},
																																													},
																																													Id: "cflags",
																																												},
																																												LeftBracketFodder: ast.Fodder{},
																																												Index: &ast.LiteralString{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																															End: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																															File: nil,
																																														},
																																														Fodder: nil,
																																														Ctx: nil,
																																														FreeVars: nil,
																																													},
																																													Value: "alt",
																																													Kind: ast.LiteralStringKind(1),
																																													BlockIndent: "",
																																													BlockTermIndent: "",
																																												},
																																												RightBracketFodder: ast.Fodder{},
																																												Id: nil,
																																											},
																																											CommaFodder: ast.Fodder{},
																																										},
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Index{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(649),
																																															Column: int(30),
																																														},
																																														End: ast.Location{
																																															Line: int(649),
																																															Column: int(39),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: p6992,
																																													FreeVars: ast.Identifiers{
																																														"code",
																																													},
																																												},
																																												Target: &ast.Var{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "<std>",
																																															Begin: ast.Location{
																																																Line: int(649),
																																																Column: int(30),
																																															},
																																															End: ast.Location{
																																																Line: int(649),
																																																Column: int(34),
																																															},
																																															File: p1,
																																														},
																																														Fodder: ast.Fodder{
																																															ast.FodderElement{
																																																Kind: ast.FodderKind(0),
																																																Blanks: int(0),
																																																Indent: int(29),
																																																Comment: []string{},
																																															},
																																														},
																																														Ctx: nil,
																																														FreeVars: ast.Identifiers{
																																															"code",
																																														},
																																													},
																																													Id: "code",
																																												},
																																												LeftBracketFodder: ast.Fodder{},
																																												Index: &ast.LiteralString{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																															End: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																															File: nil,
																																														},
																																														Fodder: nil,
																																														Ctx: nil,
																																														FreeVars: nil,
																																													},
																																													Value: "caps",
																																													Kind: ast.LiteralStringKind(1),
																																													BlockIndent: "",
																																													BlockTermIndent: "",
																																												},
																																												RightBracketFodder: ast.Fodder{},
																																												Id: nil,
																																											},
																																											CommaFodder: ast.Fodder{},
																																										},
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Binary{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(650),
																																															Column: int(30),
																																														},
																																														End: ast.Location{
																																															Line: int(650),
																																															Column: int(40),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: p6992,
																																													FreeVars: ast.Identifiers{
																																														"fpprec",
																																													},
																																												},
																																												Left: &ast.Var{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "<std>",
																																															Begin: ast.Location{
																																																Line: int(650),
																																																Column: int(30),
																																															},
																																															End: ast.Location{
																																																Line: int(650),
																																																Column: int(36),
																																															},
																																															File: p1,
																																														},
																																														Fodder: ast.Fodder{
																																															ast.FodderElement{
																																																Kind: ast.FodderKind(0),
																																																Blanks: int(0),
																																																Indent: int(29),
																																																Comment: []string{},
																																															},
																																														},
																																														Ctx: p6992,
																																														FreeVars: ast.Identifiers{
																																															"fpprec",
																																														},
																																													},
																																													Id: "fpprec",
																																												},
																																												OpFodder: ast.Fodder{},
																																												Op: ast.BinaryOp(4),
																																												Right: &ast.LiteralNumber{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "<std>",
																																															Begin: ast.Location{
																																																Line: int(650),
																																																Column: int(39),
																																															},
																																															End: ast.Location{
																																																Line: int(650),
																																																Column: int(40),
																																															},
																																															File: p1,
																																														},
																																														Fodder: ast.Fodder{},
																																														Ctx: p6992,
																																														FreeVars: nil,
																																													},
																																													OriginalString: "1",
																																												},
																																											},
																																											CommaFodder: nil,
																																										},
																																									},
																																									Named: nil,
																																								},
																																								TrailingComma: false,
																																								TailStrict: false,
																																								FodderRight: ast.Fodder{},
																																								TailStrictFodder: nil,
																																							},
																																							ElseFodder: ast.Fodder{
																																								ast.FodderElement{
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(10),
																																									Comment: []string{},
																																								},
																																							},
																																							BranchFalse: &ast.Local{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(652),
																																											Column: int(13),
																																										},
																																										End: ast.Location{
																																											Line: int(659),
																																											Column: int(56),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(12),
																																											Comment: []string{},
																																										},
																																									},
																																									Ctx: p6275,
																																									FreeVars: ast.Identifiers{
																																										"cflags",
																																										"exponent",
																																										"fpprec",
																																										"render_float_dec",
																																										"std",
																																										"val",
																																										"zp",
																																									},
																																								},
																																								Binds: ast.LocalBinds{
																																									ast.LocalBind{
																																										VarFodder: ast.Fodder{},
																																										Variable: "digits_before_pt",
																																										EqFodder: ast.Fodder{},
																																										Body: &ast.Apply{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(652),
																																														Column: int(38),
																																													},
																																													End: ast.Location{
																																														Line: int(652),
																																														Column: int(62),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p7039,
																																												FreeVars: ast.Identifiers{
																																													"exponent",
																																													"std",
																																												},
																																											},
																																											Target: &ast.Index{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(652),
																																															Column: int(38),
																																														},
																																														End: ast.Location{
																																															Line: int(652),
																																															Column: int(45),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: p7039,
																																													FreeVars: ast.Identifiers{
																																														"std",
																																													},
																																												},
																																												Target: &ast.Var{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "<std>",
																																															Begin: ast.Location{
																																																Line: int(652),
																																																Column: int(38),
																																															},
																																															End: ast.Location{
																																																Line: int(652),
																																																Column: int(41),
																																															},
																																															File: p1,
																																														},
																																														Fodder: ast.Fodder{},
																																														Ctx: nil,
																																														FreeVars: ast.Identifiers{
																																															"std",
																																														},
																																													},
																																													Id: "std",
																																												},
																																												LeftBracketFodder: ast.Fodder{},
																																												Index: &ast.LiteralString{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																															End: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																															File: nil,
																																														},
																																														Fodder: nil,
																																														Ctx: nil,
																																														FreeVars: nil,
																																													},
																																													Value: "max",
																																													Kind: ast.LiteralStringKind(1),
																																													BlockIndent: "",
																																													BlockTermIndent: "",
																																												},
																																												RightBracketFodder: ast.Fodder{},
																																												Id: nil,
																																											},
																																											FodderLeft: ast.Fodder{},
																																											Arguments: ast.Arguments{
																																												Positional: []ast.CommaSeparatedExpr{
																																													ast.CommaSeparatedExpr{
																																														Expr: &ast.LiteralNumber{
																																															NodeBase: ast.NodeBase{
																																																LocRange: ast.LocationRange{
																																																	FileName: "<std>",
																																																	Begin: ast.Location{
																																																		Line: int(652),
																																																		Column: int(46),
																																																	},
																																																	End: ast.Location{
																																																		Line: int(652),
																																																		Column: int(47),
																																																	},
																																																	File: p1,
																																																},
																																																Fodder: ast.Fodder{},
																																																Ctx: p7048,
																																																FreeVars: nil,
																																															},
																																															OriginalString: "1",
																																														},
																																														CommaFodder: ast.Fodder{},
																																													},
																																													ast.CommaSeparatedExpr{
																																														Expr: &ast.Binary{
																																															NodeBase: ast.NodeBase{
																																																LocRange: ast.LocationRange{
																																																	FileName: "<std>",
																																																	Begin: ast.Location{
																																																		Line: int(652),
																																																		Column: int(49),
																																																	},
																																																	End: ast.Location{
																																																		Line: int(652),
																																																		Column: int(61),
																																																	},
																																																	File: p1,
																																																},
																																																Fodder: ast.Fodder{},
																																																Ctx: p7048,
																																																FreeVars: ast.Identifiers{
																																																	"exponent",
																																																},
																																															},
																																															Left: &ast.Var{
																																																NodeBase: ast.NodeBase{
																																																	LocRange: ast.LocationRange{
																																																		FileName: "<std>",
																																																		Begin: ast.Location{
																																																			Line: int(652),
																																																			Column: int(49),
																																																		},
																																																		End: ast.Location{
																																																			Line: int(652),
																																																			Column: int(57),
																																																		},
																																																		File: p1,
																																																	},
																																																	Fodder: ast.Fodder{},
																																																	Ctx: p7048,
																																																	FreeVars: ast.Identifiers{
																																																		"exponent",
																																																	},
																																																},
																																																Id: "exponent",
																																															},
																																															OpFodder: ast.Fodder{},
																																															Op: ast.BinaryOp(3),
																																															Right: &ast.LiteralNumber{
																																																NodeBase: ast.NodeBase{
																																																	LocRange: ast.LocationRange{
																																																		FileName: "<std>",
																																																		Begin: ast.Location{
																																																			Line: int(652),
																																																			Column: int(60),
																																																		},
																																																		End: ast.Location{
																																																			Line: int(652),
																																																			Column: int(61),
																																																		},
																																																		File: p1,
																																																	},
																																																	Fodder: ast.Fodder{},
																																																	Ctx: p7048,
																																																	FreeVars: nil,
																																																},
																																																OriginalString: "1",
																																															},
																																														},
																																														CommaFodder: nil,
																																													},
																																												},
																																												Named: nil,
																																											},
																																											TrailingComma: false,
																																											TailStrict: false,
																																											FodderRight: ast.Fodder{},
																																											TailStrictFodder: nil,
																																										},
																																										Fun: nil,
																																										CloseFodder: ast.Fodder{},
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(652),
																																												Column: int(19),
																																											},
																																											End: ast.Location{
																																												Line: int(652),
																																												Column: int(62),
																																											},
																																											File: p1,
																																										},
																																									},
																																								},
																																								Body: &ast.Apply{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(653),
																																												Column: int(13),
																																											},
																																											End: ast.Location{
																																												Line: int(659),
																																												Column: int(56),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p6275,
																																										FreeVars: ast.Identifiers{
																																											"cflags",
																																											"digits_before_pt",
																																											"fpprec",
																																											"render_float_dec",
																																											"val",
																																											"zp",
																																										},
																																									},
																																									Target: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(653),
																																													Column: int(13),
																																												},
																																												End: ast.Location{
																																													Line: int(653),
																																													Column: int(29),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{
																																												ast.FodderElement{
																																													Kind: ast.FodderKind(0),
																																													Blanks: int(0),
																																													Indent: int(12),
																																													Comment: []string{},
																																												},
																																											},
																																											Ctx: p6275,
																																											FreeVars: ast.Identifiers{
																																												"render_float_dec",
																																											},
																																										},
																																										Id: "render_float_dec",
																																									},
																																									FodderLeft: ast.Fodder{},
																																									Arguments: ast.Arguments{
																																										Positional: []ast.CommaSeparatedExpr{
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Var{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "<std>",
																																															Begin: ast.Location{
																																																Line: int(653),
																																																Column: int(30),
																																															},
																																															End: ast.Location{
																																																Line: int(653),
																																																Column: int(33),
																																															},
																																															File: p1,
																																														},
																																														Fodder: ast.Fodder{},
																																														Ctx: p7061,
																																														FreeVars: ast.Identifiers{
																																															"val",
																																														},
																																													},
																																													Id: "val",
																																												},
																																												CommaFodder: ast.Fodder{},
																																											},
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Var{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "<std>",
																																															Begin: ast.Location{
																																																Line: int(654),
																																																Column: int(30),
																																															},
																																															End: ast.Location{
																																																Line: int(654),
																																																Column: int(32),
																																															},
																																															File: p1,
																																														},
																																														Fodder: ast.Fodder{
																																															ast.FodderElement{
																																																Kind: ast.FodderKind(0),
																																																Blanks: int(0),
																																																Indent: int(29),
																																																Comment: []string{},
																																															},
																																														},
																																														Ctx: p7061,
																																														FreeVars: ast.Identifiers{
																																															"zp",
																																														},
																																													},
																																													Id: "zp",
																																												},
																																												CommaFodder: ast.Fodder{},
																																											},
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Index{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "<std>",
																																															Begin: ast.Location{
																																																Line: int(655),
																																																Column: int(30),
																																															},
																																															End: ast.Location{
																																																Line: int(655),
																																																Column: int(42),
																																															},
																																															File: p1,
																																														},
																																														Fodder: ast.Fodder{},
																																														Ctx: p7061,
																																														FreeVars: ast.Identifiers{
																																															"cflags",
																																														},
																																													},
																																													Target: &ast.Var{
																																														NodeBase: ast.NodeBase{
																																															LocRange: ast.LocationRange{
																																																FileName: "<std>",
																																																Begin: ast.Location{
																																																	Line: int(655),
																																																	Column: int(30),
																																																},
																																																End: ast.Location{
																																																	Line: int(655),
																																																	Column: int(36),
																																																},
																																																File: p1,
																																															},
																																															Fodder: ast.Fodder{
																																																ast.FodderElement{
																																																	Kind: ast.FodderKind(0),
																																																	Blanks: int(0),
																																																	Indent: int(29),
																																																	Comment: []string{},
																																																},
																																															},
																																															Ctx: nil,
																																															FreeVars: ast.Identifiers{
																																																"cflags",
																																															},
																																														},
																																														Id: "cflags",
																																													},
																																													LeftBracketFodder: ast.Fodder{},
																																													Index: &ast.LiteralString{
																																														NodeBase: ast.NodeBase{
																																															LocRange: ast.LocationRange{
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(0),
																																																	Column: int(0),
																																																},
																																																End: ast.Location{
																																																	Line: int(0),
																																																	Column: int(0),
																																																},
																																																File: nil,
																																															},
																																															Fodder: nil,
																																															Ctx: nil,
																																															FreeVars: nil,
																																														},
																																														Value: "blank",
																																														Kind: ast.LiteralStringKind(1),
																																														BlockIndent: "",
																																														BlockTermIndent: "",
																																													},
																																													RightBracketFodder: ast.Fodder{},
																																													Id: nil,
																																												},
																																												CommaFodder: ast.Fodder{},
																																											},
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Index{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "<std>",
																																															Begin: ast.Location{
																																																Line: int(656),
																																																Column: int(30),
																																															},
																																															End: ast.Location{
																																																Line: int(656),
																																																Column: int(41),
																																															},
																																															File: p1,
																																														},
																																														Fodder: ast.Fodder{},
																																														Ctx: p7061,
																																														FreeVars: ast.Identifiers{
																																															"cflags",
																																														},
																																													},
																																													Target: &ast.Var{
																																														NodeBase: ast.NodeBase{
																																															LocRange: ast.LocationRange{
																																																FileName: "<std>",
																																																Begin: ast.Location{
																																																	Line: int(656),
																																																	Column: int(30),
																																																},
																																																End: ast.Location{
																																																	Line: int(656),
																																																	Column: int(36),
																																																},
																																																File: p1,
																																															},
																																															Fodder: ast.Fodder{
																																																ast.FodderElement{
																																																	Kind: ast.FodderKind(0),
																																																	Blanks: int(0),
																																																	Indent: int(29),
																																																	Comment: []string{},
																																																},
																																															},
																																															Ctx: nil,
																																															FreeVars: ast.Identifiers{
																																																"cflags",
																																															},
																																														},
																																														Id: "cflags",
																																													},
																																													LeftBracketFodder: ast.Fodder{},
																																													Index: &ast.LiteralString{
																																														NodeBase: ast.NodeBase{
																																															LocRange: ast.LocationRange{
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(0),
																																																	Column: int(0),
																																																},
																																																End: ast.Location{
																																																	Line: int(0),
																																																	Column: int(0),
																																																},
																																																File: nil,
																																															},
																																															Fodder: nil,
																																															Ctx: nil,
																																															FreeVars: nil,
																																														},
																																														Value: "plus",
																																														Kind: ast.LiteralStringKind(1),
																																														BlockIndent: "",
																																														BlockTermIndent: "",
																																													},
																																													RightBracketFodder: ast.Fodder{},
																																													Id: nil,
																																												},
																																												CommaFodder: ast.Fodder{},
																																											},
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Index{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "<std>",
																																															Begin: ast.Location{
																																																Line: int(657),
																																																Column: int(30),
																																															},
																																															End: ast.Location{
																																																Line: int(657),
																																																Column: int(40),
																																															},
																																															File: p1,
																																														},
																																														Fodder: ast.Fodder{},
																																														Ctx: p7061,
																																														FreeVars: ast.Identifiers{
																																															"cflags",
																																														},
																																													},
																																													Target: &ast.Var{
																																														NodeBase: ast.NodeBase{
																																															LocRange: ast.LocationRange{
																																																FileName: "<std>",
																																																Begin: ast.Location{
																																																	Line: int(657),
																																																	Column: int(30),
																																																},
																																																End: ast.Location{
																																																	Line: int(657),
																																																	Column: int(36),
																																																},
																																																File: p1,
																																															},
																																															Fodder: ast.Fodder{
																																																ast.FodderElement{
																																																	Kind: ast.FodderKind(0),
																																																	Blanks: int(0),
																																																	Indent: int(29),
																																																	Comment: []string{},
																																																},
																																															},
																																															Ctx: nil,
																																															FreeVars: ast.Identifiers{
																																																"cflags",
																																															},
																																														},
																																														Id: "cflags",
																																													},
																																													LeftBracketFodder: ast.Fodder{},
																																													Index: &ast.LiteralString{
																																														NodeBase: ast.NodeBase{
																																															LocRange: ast.LocationRange{
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(0),
																																																	Column: int(0),
																																																},
																																																End: ast.Location{
																																																	Line: int(0),
																																																	Column: int(0),
																																																},
																																																File: nil,
																																															},
																																															Fodder: nil,
																																															Ctx: nil,
																																															FreeVars: nil,
																																														},
																																														Value: "alt",
																																														Kind: ast.LiteralStringKind(1),
																																														BlockIndent: "",
																																														BlockTermIndent: "",
																																													},
																																													RightBracketFodder: ast.Fodder{},
																																													Id: nil,
																																												},
																																												CommaFodder: ast.Fodder{},
																																											},
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Index{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "<std>",
																																															Begin: ast.Location{
																																																Line: int(658),
																																																Column: int(30),
																																															},
																																															End: ast.Location{
																																																Line: int(658),
																																																Column: int(40),
																																															},
																																															File: p1,
																																														},
																																														Fodder: ast.Fodder{},
																																														Ctx: p7061,
																																														FreeVars: ast.Identifiers{
																																															"cflags",
																																														},
																																													},
																																													Target: &ast.Var{
																																														NodeBase: ast.NodeBase{
																																															LocRange: ast.LocationRange{
																																																FileName: "<std>",
																																																Begin: ast.Location{
																																																	Line: int(658),
																																																	Column: int(30),
																																																},
																																																End: ast.Location{
																																																	Line: int(658),
																																																	Column: int(36),
																																																},
																																																File: p1,
																																															},
																																															Fodder: ast.Fodder{
																																																ast.FodderElement{
																																																	Kind: ast.FodderKind(0),
																																																	Blanks: int(0),
																																																	Indent: int(29),
																																																	Comment: []string{},
																																																},
																																															},
																																															Ctx: nil,
																																															FreeVars: ast.Identifiers{
																																																"cflags",
																																															},
																																														},
																																														Id: "cflags",
																																													},
																																													LeftBracketFodder: ast.Fodder{},
																																													Index: &ast.LiteralString{
																																														NodeBase: ast.NodeBase{
																																															LocRange: ast.LocationRange{
																																																FileName: "",
																																																Begin: ast.Location{
																																																	Line: int(0),
																																																	Column: int(0),
																																																},
																																																End: ast.Location{
																																																	Line: int(0),
																																																	Column: int(0),
																																																},
																																																File: nil,
																																															},
																																															Fodder: nil,
																																															Ctx: nil,
																																															FreeVars: nil,
																																														},
																																														Value: "alt",
																																														Kind: ast.LiteralStringKind(1),
																																														BlockIndent: "",
																																														BlockTermIndent: "",
																																													},
																																													RightBracketFodder: ast.Fodder{},
																																													Id: nil,
																																												},
																																												CommaFodder: ast.Fodder{},
																																											},
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Binary{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "<std>",
																																															Begin: ast.Location{
																																																Line: int(659),
																																																Column: int(30),
																																															},
																																															End: ast.Location{
																																																Line: int(659),
																																																Column: int(55),
																																															},
																																															File: p1,
																																														},
																																														Fodder: ast.Fodder{},
																																														Ctx: p7061,
																																														FreeVars: ast.Identifiers{
																																															"digits_before_pt",
																																															"fpprec",
																																														},
																																													},
																																													Left: &ast.Var{
																																														NodeBase: ast.NodeBase{
																																															LocRange: ast.LocationRange{
																																																FileName: "<std>",
																																																Begin: ast.Location{
																																																	Line: int(659),
																																																	Column: int(30),
																																																},
																																																End: ast.Location{
																																																	Line: int(659),
																																																	Column: int(36),
																																																},
																																																File: p1,
																																															},
																																															Fodder: ast.Fodder{
																																																ast.FodderElement{
																																																	Kind: ast.FodderKind(0),
																																																	Blanks: int(0),
																																																	Indent: int(29),
																																																	Comment: []string{},
																																																},
																																															},
																																															Ctx: p7061,
																																															FreeVars: ast.Identifiers{
																																																"fpprec",
																																															},
																																														},
																																														Id: "fpprec",
																																													},
																																													OpFodder: ast.Fodder{},
																																													Op: ast.BinaryOp(4),
																																													Right: &ast.Var{
																																														NodeBase: ast.NodeBase{
																																															LocRange: ast.LocationRange{
																																																FileName: "<std>",
																																																Begin: ast.Location{
																																																	Line: int(659),
																																																	Column: int(39),
																																																},
																																																End: ast.Location{
																																																	Line: int(659),
																																																	Column: int(55),
																																																},
																																																File: p1,
																																															},
																																															Fodder: ast.Fodder{},
																																															Ctx: p7061,
																																															FreeVars: ast.Identifiers{
																																																"digits_before_pt",
																																															},
																																														},
																																														Id: "digits_before_pt",
																																													},
																																												},
																																												CommaFodder: nil,
																																											},
																																										},
																																										Named: nil,
																																									},
																																									TrailingComma: false,
																																									TailStrict: false,
																																									FodderRight: ast.Fodder{},
																																									TailStrictFodder: nil,
																																								},
																																							},
																																						},
																																					},
																																				},
																																				ElseFodder: ast.Fodder{
																																					ast.FodderElement{
																																						Kind: ast.FodderKind(0),
																																						Blanks: int(0),
																																						Indent: int(6),
																																						Comment: []string{},
																																					},
																																				},
																																				BranchFalse: &ast.Conditional{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(660),
																																								Column: int(12),
																																							},
																																							End: ast.Location{
																																								Line: int(671),
																																								Column: int(44),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p6275,
																																						FreeVars: ast.Identifiers{
																																							"code",
																																							"std",
																																							"val",
																																						},
																																					},
																																					Cond: &ast.Binary{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(660),
																																									Column: int(15),
																																								},
																																								End: ast.Location{
																																									Line: int(660),
																																									Column: int(32),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p6275,
																																							FreeVars: ast.Identifiers{
																																								"code",
																																							},
																																						},
																																						Left: &ast.Index{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(660),
																																										Column: int(15),
																																									},
																																									End: ast.Location{
																																										Line: int(660),
																																										Column: int(25),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6275,
																																								FreeVars: ast.Identifiers{
																																									"code",
																																								},
																																							},
																																							Target: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(660),
																																											Column: int(15),
																																										},
																																										End: ast.Location{
																																											Line: int(660),
																																											Column: int(19),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"code",
																																									},
																																								},
																																								Id: "code",
																																							},
																																							LeftBracketFodder: ast.Fodder{},
																																							Index: &ast.LiteralString{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										File: nil,
																																									},
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: nil,
																																								},
																																								Value: "ctype",
																																								Kind: ast.LiteralStringKind(1),
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							Id: nil,
																																						},
																																						OpFodder: ast.Fodder{},
																																						Op: ast.BinaryOp(12),
																																						Right: &ast.LiteralString{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(660),
																																										Column: int(29),
																																									},
																																									End: ast.Location{
																																										Line: int(660),
																																										Column: int(32),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6275,
																																								FreeVars: nil,
																																							},
																																							Value: "c",
																																							Kind: ast.LiteralStringKind(1),
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																						},
																																					},
																																					ThenFodder: ast.Fodder{},
																																					BranchTrue: &ast.Conditional{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(661),
																																									Column: int(9),
																																								},
																																								End: ast.Location{
																																									Line: int(669),
																																									Column: int(69),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(8),
																																									Comment: []string{},
																																								},
																																							},
																																							Ctx: p6275,
																																							FreeVars: ast.Identifiers{
																																								"std",
																																								"val",
																																							},
																																						},
																																						Cond: &ast.Binary{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(661),
																																										Column: int(12),
																																									},
																																									End: ast.Location{
																																										Line: int(661),
																																										Column: int(37),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6275,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																									"val",
																																								},
																																							},
																																							Left: &ast.Apply{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(661),
																																											Column: int(12),
																																										},
																																										End: ast.Location{
																																											Line: int(661),
																																											Column: int(25),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p6275,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																										"val",
																																									},
																																								},
																																								Target: &ast.Index{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(661),
																																												Column: int(12),
																																											},
																																											End: ast.Location{
																																												Line: int(661),
																																												Column: int(20),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p6275,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																										},
																																									},
																																									Target: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(661),
																																													Column: int(12),
																																												},
																																												End: ast.Location{
																																													Line: int(661),
																																													Column: int(15),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{
																																												"std",
																																											},
																																										},
																																										Id: "std",
																																									},
																																									LeftBracketFodder: ast.Fodder{},
																																									Index: &ast.LiteralString{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												End: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												File: nil,
																																											},
																																											Fodder: nil,
																																											Ctx: nil,
																																											FreeVars: nil,
																																										},
																																										Value: "type",
																																										Kind: ast.LiteralStringKind(1),
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																									},
																																									RightBracketFodder: ast.Fodder{},
																																									Id: nil,
																																								},
																																								FodderLeft: ast.Fodder{},
																																								Arguments: ast.Arguments{
																																									Positional: []ast.CommaSeparatedExpr{
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Var{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(661),
																																															Column: int(21),
																																														},
																																														End: ast.Location{
																																															Line: int(661),
																																															Column: int(24),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: p7122,
																																													FreeVars: ast.Identifiers{
																																														"val",
																																													},
																																												},
																																												Id: "val",
																																											},
																																											CommaFodder: nil,
																																										},
																																									},
																																									Named: nil,
																																								},
																																								TrailingComma: false,
																																								TailStrict: false,
																																								FodderRight: ast.Fodder{},
																																								TailStrictFodder: nil,
																																							},
																																							OpFodder: ast.Fodder{},
																																							Op: ast.BinaryOp(12),
																																							Right: &ast.LiteralString{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(661),
																																											Column: int(29),
																																										},
																																										End: ast.Location{
																																											Line: int(661),
																																											Column: int(37),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p6275,
																																									FreeVars: nil,
																																								},
																																								Value: "number",
																																								Kind: ast.LiteralStringKind(1),
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																							},
																																						},
																																						ThenFodder: ast.Fodder{},
																																						BranchTrue: &ast.Apply{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(662),
																																										Column: int(11),
																																									},
																																									End: ast.Location{
																																										Line: int(662),
																																										Column: int(24),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6275,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																									"val",
																																								},
																																							},
																																							Target: &ast.Index{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(662),
																																											Column: int(11),
																																										},
																																										End: ast.Location{
																																											Line: int(662),
																																											Column: int(19),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p6275,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																									},
																																								},
																																								Target: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(662),
																																												Column: int(11),
																																											},
																																											End: ast.Location{
																																												Line: int(662),
																																												Column: int(14),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{
																																											ast.FodderElement{
																																												Kind: ast.FodderKind(0),
																																												Blanks: int(0),
																																												Indent: int(10),
																																												Comment: []string{},
																																											},
																																										},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																										},
																																									},
																																									Id: "std",
																																								},
																																								LeftBracketFodder: ast.Fodder{},
																																								Index: &ast.LiteralString{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											File: nil,
																																										},
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: nil,
																																									},
																																									Value: "char",
																																									Kind: ast.LiteralStringKind(1),
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								Id: nil,
																																							},
																																							FodderLeft: ast.Fodder{},
																																							Arguments: ast.Arguments{
																																								Positional: []ast.CommaSeparatedExpr{
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(662),
																																														Column: int(20),
																																													},
																																													End: ast.Location{
																																														Line: int(662),
																																														Column: int(23),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p7135,
																																												FreeVars: ast.Identifiers{
																																													"val",
																																												},
																																											},
																																											Id: "val",
																																										},
																																										CommaFodder: nil,
																																									},
																																								},
																																								Named: nil,
																																							},
																																							TrailingComma: false,
																																							TailStrict: false,
																																							FodderRight: ast.Fodder{},
																																							TailStrictFodder: nil,
																																						},
																																						ElseFodder: ast.Fodder{
																																							ast.FodderElement{
																																								Kind: ast.FodderKind(0),
																																								Blanks: int(0),
																																								Indent: int(8),
																																								Comment: []string{},
																																							},
																																						},
																																						BranchFalse: &ast.Conditional{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(663),
																																										Column: int(14),
																																									},
																																									End: ast.Location{
																																										Line: int(669),
																																										Column: int(69),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6275,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																									"val",
																																								},
																																							},
																																							Cond: &ast.Binary{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(663),
																																											Column: int(17),
																																										},
																																										End: ast.Location{
																																											Line: int(663),
																																											Column: int(42),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p6275,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																										"val",
																																									},
																																								},
																																								Left: &ast.Apply{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(663),
																																												Column: int(17),
																																											},
																																											End: ast.Location{
																																												Line: int(663),
																																												Column: int(30),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p6275,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																											"val",
																																										},
																																									},
																																									Target: &ast.Index{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(663),
																																													Column: int(17),
																																												},
																																												End: ast.Location{
																																													Line: int(663),
																																													Column: int(25),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p6275,
																																											FreeVars: ast.Identifiers{
																																												"std",
																																											},
																																										},
																																										Target: &ast.Var{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(663),
																																														Column: int(17),
																																													},
																																													End: ast.Location{
																																														Line: int(663),
																																														Column: int(20),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: nil,
																																												FreeVars: ast.Identifiers{
																																													"std",
																																												},
																																											},
																																											Id: "std",
																																										},
																																										LeftBracketFodder: ast.Fodder{},
																																										Index: &ast.LiteralString{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(0),
																																														Column: int(0),
																																													},
																																													End: ast.Location{
																																														Line: int(0),
																																														Column: int(0),
																																													},
																																													File: nil,
																																												},
																																												Fodder: nil,
																																												Ctx: nil,
																																												FreeVars: nil,
																																											},
																																											Value: "type",
																																											Kind: ast.LiteralStringKind(1),
																																											BlockIndent: "",
																																											BlockTermIndent: "",
																																										},
																																										RightBracketFodder: ast.Fodder{},
																																										Id: nil,
																																									},
																																									FodderLeft: ast.Fodder{},
																																									Arguments: ast.Arguments{
																																										Positional: []ast.CommaSeparatedExpr{
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Var{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "<std>",
																																															Begin: ast.Location{
																																																Line: int(663),
																																																Column: int(26),
																																															},
																																															End: ast.Location{
																																																Line: int(663),
																																																Column: int(29),
																																															},
																																															File: p1,
																																														},
																																														Fodder: ast.Fodder{},
																																														Ctx: p7151,
																																														FreeVars: ast.Identifiers{
																																															"val",
																																														},
																																													},
																																													Id: "val",
																																												},
																																												CommaFodder: nil,
																																											},
																																										},
																																										Named: nil,
																																									},
																																									TrailingComma: false,
																																									TailStrict: false,
																																									FodderRight: ast.Fodder{},
																																									TailStrictFodder: nil,
																																								},
																																								OpFodder: ast.Fodder{},
																																								Op: ast.BinaryOp(12),
																																								Right: &ast.LiteralString{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(663),
																																												Column: int(34),
																																											},
																																											End: ast.Location{
																																												Line: int(663),
																																												Column: int(42),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p6275,
																																										FreeVars: nil,
																																									},
																																									Value: "string",
																																									Kind: ast.LiteralStringKind(1),
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																								},
																																							},
																																							ThenFodder: ast.Fodder{},
																																							BranchTrue: &ast.Conditional{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(664),
																																											Column: int(11),
																																										},
																																										End: ast.Location{
																																											Line: int(667),
																																											Column: int(71),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(10),
																																											Comment: []string{},
																																										},
																																									},
																																									Ctx: p6275,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																										"val",
																																									},
																																								},
																																								Cond: &ast.Binary{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(664),
																																												Column: int(14),
																																											},
																																											End: ast.Location{
																																												Line: int(664),
																																												Column: int(34),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p6275,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																											"val",
																																										},
																																									},
																																									Left: &ast.Apply{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(664),
																																													Column: int(14),
																																												},
																																												End: ast.Location{
																																													Line: int(664),
																																													Column: int(29),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p6275,
																																											FreeVars: ast.Identifiers{
																																												"std",
																																												"val",
																																											},
																																										},
																																										Target: &ast.Index{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(664),
																																														Column: int(14),
																																													},
																																													End: ast.Location{
																																														Line: int(664),
																																														Column: int(24),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p6275,
																																												FreeVars: ast.Identifiers{
																																													"std",
																																												},
																																											},
																																											Target: &ast.Var{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(664),
																																															Column: int(14),
																																														},
																																														End: ast.Location{
																																															Line: int(664),
																																															Column: int(17),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: nil,
																																													FreeVars: ast.Identifiers{
																																														"std",
																																													},
																																												},
																																												Id: "std",
																																											},
																																											LeftBracketFodder: ast.Fodder{},
																																											Index: &ast.LiteralString{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(0),
																																															Column: int(0),
																																														},
																																														End: ast.Location{
																																															Line: int(0),
																																															Column: int(0),
																																														},
																																														File: nil,
																																													},
																																													Fodder: nil,
																																													Ctx: nil,
																																													FreeVars: nil,
																																												},
																																												Value: "length",
																																												Kind: ast.LiteralStringKind(1),
																																												BlockIndent: "",
																																												BlockTermIndent: "",
																																											},
																																											RightBracketFodder: ast.Fodder{},
																																											Id: nil,
																																										},
																																										FodderLeft: ast.Fodder{},
																																										Arguments: ast.Arguments{
																																											Positional: []ast.CommaSeparatedExpr{
																																												ast.CommaSeparatedExpr{
																																													Expr: &ast.Var{
																																														NodeBase: ast.NodeBase{
																																															LocRange: ast.LocationRange{
																																																FileName: "<std>",
																																																Begin: ast.Location{
																																																	Line: int(664),
																																																	Column: int(25),
																																																},
																																																End: ast.Location{
																																																	Line: int(664),
																																																	Column: int(28),
																																																},
																																																File: p1,
																																															},
																																															Fodder: ast.Fodder{},
																																															Ctx: p7168,
																																															FreeVars: ast.Identifiers{
																																																"val",
																																															},
																																														},
																																														Id: "val",
																																													},
																																													CommaFodder: nil,
																																												},
																																											},
																																											Named: nil,
																																										},
																																										TrailingComma: false,
																																										TailStrict: false,
																																										FodderRight: ast.Fodder{},
																																										TailStrictFodder: nil,
																																									},
																																									OpFodder: ast.Fodder{},
																																									Op: ast.BinaryOp(12),
																																									Right: &ast.LiteralNumber{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(664),
																																													Column: int(33),
																																												},
																																												End: ast.Location{
																																													Line: int(664),
																																													Column: int(34),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p6275,
																																											FreeVars: nil,
																																										},
																																										OriginalString: "1",
																																									},
																																								},
																																								ThenFodder: ast.Fodder{},
																																								BranchTrue: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(665),
																																												Column: int(13),
																																											},
																																											End: ast.Location{
																																												Line: int(665),
																																												Column: int(16),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{
																																											ast.FodderElement{
																																												Kind: ast.FodderKind(0),
																																												Blanks: int(0),
																																												Indent: int(12),
																																												Comment: []string{},
																																											},
																																										},
																																										Ctx: p6275,
																																										FreeVars: ast.Identifiers{
																																											"val",
																																										},
																																									},
																																									Id: "val",
																																								},
																																								ElseFodder: ast.Fodder{
																																									ast.FodderElement{
																																										Kind: ast.FodderKind(0),
																																										Blanks: int(0),
																																										Indent: int(10),
																																										Comment: []string{},
																																									},
																																								},
																																								BranchFalse: &ast.Error{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(667),
																																												Column: int(13),
																																											},
																																											End: ast.Location{
																																												Line: int(667),
																																												Column: int(71),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{
																																											ast.FodderElement{
																																												Kind: ast.FodderKind(0),
																																												Blanks: int(0),
																																												Indent: int(12),
																																												Comment: []string{},
																																											},
																																										},
																																										Ctx: p6275,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																											"val",
																																										},
																																									},
																																									Expr: &ast.Binary{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(667),
																																													Column: int(19),
																																												},
																																												End: ast.Location{
																																													Line: int(667),
																																													Column: int(71),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p6275,
																																											FreeVars: ast.Identifiers{
																																												"std",
																																												"val",
																																											},
																																										},
																																										Left: &ast.LiteralString{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(667),
																																														Column: int(19),
																																													},
																																													End: ast.Location{
																																														Line: int(667),
																																														Column: int(53),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p6275,
																																												FreeVars: nil,
																																											},
																																											Value: "%c expected 1-sized string got: ",
																																											Kind: ast.LiteralStringKind(1),
																																											BlockIndent: "",
																																											BlockTermIndent: "",
																																										},
																																										OpFodder: ast.Fodder{},
																																										Op: ast.BinaryOp(3),
																																										Right: &ast.Apply{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(667),
																																														Column: int(56),
																																													},
																																													End: ast.Location{
																																														Line: int(667),
																																														Column: int(71),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p6275,
																																												FreeVars: ast.Identifiers{
																																													"std",
																																													"val",
																																												},
																																											},
																																											Target: &ast.Index{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(667),
																																															Column: int(56),
																																														},
																																														End: ast.Location{
																																															Line: int(667),
																																															Column: int(66),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: p6275,
																																													FreeVars: ast.Identifiers{
																																														"std",
																																													},
																																												},
																																												Target: &ast.Var{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "<std>",
																																															Begin: ast.Location{
																																																Line: int(667),
																																																Column: int(56),
																																															},
																																															End: ast.Location{
																																																Line: int(667),
																																																Column: int(59),
																																															},
																																															File: p1,
																																														},
																																														Fodder: ast.Fodder{},
																																														Ctx: nil,
																																														FreeVars: ast.Identifiers{
																																															"std",
																																														},
																																													},
																																													Id: "std",
																																												},
																																												LeftBracketFodder: ast.Fodder{},
																																												Index: &ast.LiteralString{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																															End: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																															File: nil,
																																														},
																																														Fodder: nil,
																																														Ctx: nil,
																																														FreeVars: nil,
																																													},
																																													Value: "length",
																																													Kind: ast.LiteralStringKind(1),
																																													BlockIndent: "",
																																													BlockTermIndent: "",
																																												},
																																												RightBracketFodder: ast.Fodder{},
																																												Id: nil,
																																											},
																																											FodderLeft: ast.Fodder{},
																																											Arguments: ast.Arguments{
																																												Positional: []ast.CommaSeparatedExpr{
																																													ast.CommaSeparatedExpr{
																																														Expr: &ast.Var{
																																															NodeBase: ast.NodeBase{
																																																LocRange: ast.LocationRange{
																																																	FileName: "<std>",
																																																	Begin: ast.Location{
																																																		Line: int(667),
																																																		Column: int(67),
																																																	},
																																																	End: ast.Location{
																																																		Line: int(667),
																																																		Column: int(70),
																																																	},
																																																	File: p1,
																																																},
																																																Fodder: ast.Fodder{},
																																																Ctx: p7190,
																																																FreeVars: ast.Identifiers{
																																																	"val",
																																																},
																																															},
																																															Id: "val",
																																														},
																																														CommaFodder: nil,
																																													},
																																												},
																																												Named: nil,
																																											},
																																											TrailingComma: false,
																																											TailStrict: false,
																																											FodderRight: ast.Fodder{},
																																											TailStrictFodder: nil,
																																										},
																																									},
																																								},
																																							},
																																							ElseFodder: ast.Fodder{
																																								ast.FodderElement{
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(8),
																																									Comment: []string{},
																																								},
																																							},
																																							BranchFalse: &ast.Error{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(669),
																																											Column: int(11),
																																										},
																																										End: ast.Location{
																																											Line: int(669),
																																											Column: int(69),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(10),
																																											Comment: []string{},
																																										},
																																									},
																																									Ctx: p6275,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																										"val",
																																									},
																																								},
																																								Expr: &ast.Binary{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(669),
																																												Column: int(17),
																																											},
																																											End: ast.Location{
																																												Line: int(669),
																																												Column: int(69),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p6275,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																											"val",
																																										},
																																									},
																																									Left: &ast.LiteralString{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(669),
																																													Column: int(17),
																																												},
																																												End: ast.Location{
																																													Line: int(669),
																																													Column: int(53),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p6275,
																																											FreeVars: nil,
																																										},
																																										Value: "%c expected number / string, got: ",
																																										Kind: ast.LiteralStringKind(1),
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																									},
																																									OpFodder: ast.Fodder{},
																																									Op: ast.BinaryOp(3),
																																									Right: &ast.Apply{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(669),
																																													Column: int(56),
																																												},
																																												End: ast.Location{
																																													Line: int(669),
																																													Column: int(69),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p6275,
																																											FreeVars: ast.Identifiers{
																																												"std",
																																												"val",
																																											},
																																										},
																																										Target: &ast.Index{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(669),
																																														Column: int(56),
																																													},
																																													End: ast.Location{
																																														Line: int(669),
																																														Column: int(64),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p6275,
																																												FreeVars: ast.Identifiers{
																																													"std",
																																												},
																																											},
																																											Target: &ast.Var{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(669),
																																															Column: int(56),
																																														},
																																														End: ast.Location{
																																															Line: int(669),
																																															Column: int(59),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: nil,
																																													FreeVars: ast.Identifiers{
																																														"std",
																																													},
																																												},
																																												Id: "std",
																																											},
																																											LeftBracketFodder: ast.Fodder{},
																																											Index: &ast.LiteralString{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(0),
																																															Column: int(0),
																																														},
																																														End: ast.Location{
																																															Line: int(0),
																																															Column: int(0),
																																														},
																																														File: nil,
																																													},
																																													Fodder: nil,
																																													Ctx: nil,
																																													FreeVars: nil,
																																												},
																																												Value: "type",
																																												Kind: ast.LiteralStringKind(1),
																																												BlockIndent: "",
																																												BlockTermIndent: "",
																																											},
																																											RightBracketFodder: ast.Fodder{},
																																											Id: nil,
																																										},
																																										FodderLeft: ast.Fodder{},
																																										Arguments: ast.Arguments{
																																											Positional: []ast.CommaSeparatedExpr{
																																												ast.CommaSeparatedExpr{
																																													Expr: &ast.Var{
																																														NodeBase: ast.NodeBase{
																																															LocRange: ast.LocationRange{
																																																FileName: "<std>",
																																																Begin: ast.Location{
																																																	Line: int(669),
																																																	Column: int(65),
																																																},
																																																End: ast.Location{
																																																	Line: int(669),
																																																	Column: int(68),
																																																},
																																																File: p1,
																																															},
																																															Fodder: ast.Fodder{},
																																															Ctx: p7208,
																																															FreeVars: ast.Identifiers{
																																																"val",
																																															},
																																														},
																																														Id: "val",
																																													},
																																													CommaFodder: nil,
																																												},
																																											},
																																											Named: nil,
																																										},
																																										TrailingComma: false,
																																										TailStrict: false,
																																										FodderRight: ast.Fodder{},
																																										TailStrictFodder: nil,
																																									},
																																								},
																																							},
																																						},
																																					},
																																					ElseFodder: ast.Fodder{
																																						ast.FodderElement{
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(6),
																																							Comment: []string{},
																																						},
																																					},
																																					BranchFalse: &ast.Error{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(671),
																																									Column: int(9),
																																								},
																																								End: ast.Location{
																																									Line: int(671),
																																									Column: int(44),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(8),
																																									Comment: []string{},
																																								},
																																							},
																																							Ctx: p6275,
																																							FreeVars: ast.Identifiers{
																																								"code",
																																							},
																																						},
																																						Expr: &ast.Binary{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(671),
																																										Column: int(15),
																																									},
																																									End: ast.Location{
																																										Line: int(671),
																																										Column: int(44),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p6275,
																																								FreeVars: ast.Identifiers{
																																									"code",
																																								},
																																							},
																																							Left: &ast.LiteralString{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(671),
																																											Column: int(15),
																																										},
																																										End: ast.Location{
																																											Line: int(671),
																																											Column: int(31),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p6275,
																																									FreeVars: nil,
																																								},
																																								Value: "Unknown code: ",
																																								Kind: ast.LiteralStringKind(1),
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																							},
																																							OpFodder: ast.Fodder{},
																																							Op: ast.BinaryOp(3),
																																							Right: &ast.Index{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(671),
																																											Column: int(34),
																																										},
																																										End: ast.Location{
																																											Line: int(671),
																																											Column: int(44),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p6275,
																																									FreeVars: ast.Identifiers{
																																										"code",
																																									},
																																								},
																																								Target: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(671),
																																												Column: int(34),
																																											},
																																											End: ast.Location{
																																												Line: int(671),
																																												Column: int(38),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"code",
																																										},
																																									},
																																									Id: "code",
																																								},
																																								LeftBracketFodder: ast.Fodder{},
																																								Index: &ast.LiteralString{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											File: nil,
																																										},
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: nil,
																																									},
																																									Value: "ctype",
																																									Kind: ast.LiteralStringKind(1),
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								Id: nil,
																																							},
																																						},
																																					},
																																				},
																																			},
																																		},
																																	},
																																},
																															},
																														},
																													},
																												},
																											},
																										},
																									},
																								},
																								Fun: nil,
																								CloseFodder: nil,
																								LocRange: ast.LocationRange{
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									File: nil,
																								},
																							},
																						},
																						Body: &ast.Local{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(674),
																										Column: int(5),
																									},
																									End: ast.Location{
																										Line: int(774),
																										Column: int(48),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{
																									ast.FodderElement{
																										Kind: ast.FodderKind(0),
																										Blanks: int(1),
																										Indent: int(4),
																										Comment: []string{},
																									},
																									ast.FodderElement{
																										Kind: ast.FodderKind(2),
																										Blanks: int(0),
																										Indent: int(4),
																										Comment: []string{
																											"// Render a parsed format string with an array of values.",
																										},
																									},
																								},
																								Ctx: p3497,
																								FreeVars: ast.Identifiers{
																									"codes",
																									"format_code",
																									"pad_left",
																									"pad_right",
																									"std",
																									"vals",
																								},
																							},
																							Binds: ast.LocalBinds{
																								ast.LocalBind{
																									VarFodder: nil,
																									Variable: "format_codes_arr",
																									EqFodder: nil,
																									Body: &ast.Function{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(674),
																													Column: int(11),
																												},
																												End: ast.Location{
																													Line: int(726),
																													Column: int(64),
																												},
																												File: p1,
																											},
																											Fodder: nil,
																											Ctx: p7228,
																											FreeVars: ast.Identifiers{
																												"format_code",
																												"format_codes_arr",
																												"pad_left",
																												"pad_right",
																												"std",
																											},
																										},
																										ParenLeftFodder: ast.Fodder{},
																										Parameters: []ast.Parameter{
																											ast.Parameter{
																												NameFodder: ast.Fodder{},
																												Name: "codes",
																												EqFodder: nil,
																												DefaultArg: nil,
																												CommaFodder: ast.Fodder{},
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(674),
																														Column: int(28),
																													},
																													End: ast.Location{
																														Line: int(674),
																														Column: int(33),
																													},
																													File: p1,
																												},
																											},
																											ast.Parameter{
																												NameFodder: ast.Fodder{},
																												Name: "arr",
																												EqFodder: nil,
																												DefaultArg: nil,
																												CommaFodder: ast.Fodder{},
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(674),
																														Column: int(35),
																													},
																													End: ast.Location{
																														Line: int(674),
																														Column: int(38),
																													},
																													File: p1,
																												},
																											},
																											ast.Parameter{
																												NameFodder: ast.Fodder{},
																												Name: "i",
																												EqFodder: nil,
																												DefaultArg: nil,
																												CommaFodder: ast.Fodder{},
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(674),
																														Column: int(40),
																													},
																													End: ast.Location{
																														Line: int(674),
																														Column: int(41),
																													},
																													File: p1,
																												},
																											},
																											ast.Parameter{
																												NameFodder: ast.Fodder{},
																												Name: "j",
																												EqFodder: nil,
																												DefaultArg: nil,
																												CommaFodder: ast.Fodder{},
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(674),
																														Column: int(43),
																													},
																													End: ast.Location{
																														Line: int(674),
																														Column: int(44),
																													},
																													File: p1,
																												},
																											},
																											ast.Parameter{
																												NameFodder: ast.Fodder{},
																												Name: "v",
																												EqFodder: nil,
																												DefaultArg: nil,
																												CommaFodder: nil,
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(674),
																														Column: int(46),
																													},
																													End: ast.Location{
																														Line: int(674),
																														Column: int(47),
																													},
																													File: p1,
																												},
																											},
																										},
																										TrailingComma: false,
																										ParenRightFodder: ast.Fodder{},
																										Body: &ast.Conditional{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(675),
																														Column: int(7),
																													},
																													End: ast.Location{
																														Line: int(726),
																														Column: int(64),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{
																													ast.FodderElement{
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(6),
																														Comment: []string{},
																													},
																												},
																												Ctx: p7233,
																												FreeVars: ast.Identifiers{
																													"arr",
																													"codes",
																													"format_code",
																													"format_codes_arr",
																													"i",
																													"j",
																													"pad_left",
																													"pad_right",
																													"std",
																													"v",
																												},
																											},
																											Cond: &ast.Binary{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(675),
																															Column: int(10),
																														},
																														End: ast.Location{
																															Line: int(675),
																															Column: int(32),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p7233,
																													FreeVars: ast.Identifiers{
																														"codes",
																														"i",
																														"std",
																													},
																												},
																												Left: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(675),
																																Column: int(10),
																															},
																															End: ast.Location{
																																Line: int(675),
																																Column: int(11),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p7233,
																														FreeVars: ast.Identifiers{
																															"i",
																														},
																													},
																													Id: "i",
																												},
																												OpFodder: ast.Fodder{},
																												Op: ast.BinaryOp(8),
																												Right: &ast.Apply{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(675),
																																Column: int(15),
																															},
																															End: ast.Location{
																																Line: int(675),
																																Column: int(32),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p7233,
																														FreeVars: ast.Identifiers{
																															"codes",
																															"std",
																														},
																													},
																													Target: &ast.Index{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(675),
																																	Column: int(15),
																																},
																																End: ast.Location{
																																	Line: int(675),
																																	Column: int(25),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p7233,
																															FreeVars: ast.Identifiers{
																																"std",
																															},
																														},
																														Target: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(675),
																																		Column: int(15),
																																	},
																																	End: ast.Location{
																																		Line: int(675),
																																		Column: int(18),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																															},
																															Id: "std",
																														},
																														LeftBracketFodder: ast.Fodder{},
																														Index: &ast.LiteralString{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	File: nil,
																																},
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: nil,
																															},
																															Value: "length",
																															Kind: ast.LiteralStringKind(1),
																															BlockIndent: "",
																															BlockTermIndent: "",
																														},
																														RightBracketFodder: ast.Fodder{},
																														Id: nil,
																													},
																													FodderLeft: ast.Fodder{},
																													Arguments: ast.Arguments{
																														Positional: []ast.CommaSeparatedExpr{
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Var{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(675),
																																				Column: int(26),
																																			},
																																			End: ast.Location{
																																				Line: int(675),
																																				Column: int(31),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p7248,
																																		FreeVars: ast.Identifiers{
																																			"codes",
																																		},
																																	},
																																	Id: "codes",
																																},
																																CommaFodder: nil,
																															},
																														},
																														Named: nil,
																													},
																													TrailingComma: false,
																													TailStrict: false,
																													FodderRight: ast.Fodder{},
																													TailStrictFodder: nil,
																												},
																											},
																											ThenFodder: ast.Fodder{},
																											BranchTrue: &ast.Conditional{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(676),
																															Column: int(9),
																														},
																														End: ast.Location{
																															Line: int(679),
																															Column: int(12),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(8),
																															Comment: []string{},
																														},
																													},
																													Ctx: p7233,
																													FreeVars: ast.Identifiers{
																														"arr",
																														"j",
																														"std",
																														"v",
																													},
																												},
																												Cond: &ast.Binary{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(676),
																																Column: int(12),
																															},
																															End: ast.Location{
																																Line: int(676),
																																Column: int(31),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p7233,
																														FreeVars: ast.Identifiers{
																															"arr",
																															"j",
																															"std",
																														},
																													},
																													Left: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(676),
																																	Column: int(12),
																																},
																																End: ast.Location{
																																	Line: int(676),
																																	Column: int(13),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p7233,
																															FreeVars: ast.Identifiers{
																																"j",
																															},
																														},
																														Id: "j",
																													},
																													OpFodder: ast.Fodder{},
																													Op: ast.BinaryOp(9),
																													Right: &ast.Apply{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(676),
																																	Column: int(16),
																																},
																																End: ast.Location{
																																	Line: int(676),
																																	Column: int(31),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p7233,
																															FreeVars: ast.Identifiers{
																																"arr",
																																"std",
																															},
																														},
																														Target: &ast.Index{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(676),
																																		Column: int(16),
																																	},
																																	End: ast.Location{
																																		Line: int(676),
																																		Column: int(26),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p7233,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																															},
																															Target: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(676),
																																			Column: int(16),
																																		},
																																		End: ast.Location{
																																			Line: int(676),
																																			Column: int(19),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																	},
																																},
																																Id: "std",
																															},
																															LeftBracketFodder: ast.Fodder{},
																															Index: &ast.LiteralString{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		File: nil,
																																	},
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: nil,
																																},
																																Value: "length",
																																Kind: ast.LiteralStringKind(1),
																																BlockIndent: "",
																																BlockTermIndent: "",
																															},
																															RightBracketFodder: ast.Fodder{},
																															Id: nil,
																														},
																														FodderLeft: ast.Fodder{},
																														Arguments: ast.Arguments{
																															Positional: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(676),
																																					Column: int(27),
																																				},
																																				End: ast.Location{
																																					Line: int(676),
																																					Column: int(30),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p7266,
																																			FreeVars: ast.Identifiers{
																																				"arr",
																																			},
																																		},
																																		Id: "arr",
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															Named: nil,
																														},
																														TrailingComma: false,
																														TailStrict: false,
																														FodderRight: ast.Fodder{},
																														TailStrictFodder: nil,
																													},
																												},
																												ThenFodder: ast.Fodder{},
																												BranchTrue: &ast.Error{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(677),
																																Column: int(11),
																															},
																															End: ast.Location{
																																Line: int(677),
																																Column: int(86),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(10),
																																Comment: []string{},
																															},
																														},
																														Ctx: p7233,
																														FreeVars: ast.Identifiers{
																															"arr",
																															"j",
																															"std",
																														},
																													},
																													Expr: &ast.Binary{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(677),
																																	Column: int(18),
																																},
																																End: ast.Location{
																																	Line: int(677),
																																	Column: int(85),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p7233,
																															FreeVars: ast.Identifiers{
																																"arr",
																																"j",
																																"std",
																															},
																														},
																														Left: &ast.Binary{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(677),
																																		Column: int(18),
																																	},
																																	End: ast.Location{
																																		Line: int(677),
																																		Column: int(81),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p7233,
																																FreeVars: ast.Identifiers{
																																	"arr",
																																	"std",
																																},
																															},
																															Left: &ast.Binary{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(677),
																																			Column: int(18),
																																		},
																																		End: ast.Location{
																																			Line: int(677),
																																			Column: int(65),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p7233,
																																	FreeVars: ast.Identifiers{
																																		"arr",
																																		"std",
																																	},
																																},
																																Left: &ast.LiteralString{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(677),
																																				Column: int(18),
																																			},
																																			End: ast.Location{
																																				Line: int(677),
																																				Column: int(47),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p7233,
																																		FreeVars: nil,
																																	},
																																	Value: "Too many values to format: ",
																																	Kind: ast.LiteralStringKind(1),
																																	BlockIndent: "",
																																	BlockTermIndent: "",
																																},
																																OpFodder: ast.Fodder{},
																																Op: ast.BinaryOp(3),
																																Right: &ast.Apply{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(677),
																																				Column: int(50),
																																			},
																																			End: ast.Location{
																																				Line: int(677),
																																				Column: int(65),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p7233,
																																		FreeVars: ast.Identifiers{
																																			"arr",
																																			"std",
																																		},
																																	},
																																	Target: &ast.Index{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(677),
																																					Column: int(50),
																																				},
																																				End: ast.Location{
																																					Line: int(677),
																																					Column: int(60),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p7233,
																																			FreeVars: ast.Identifiers{
																																				"std",
																																			},
																																		},
																																		Target: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(677),
																																						Column: int(50),
																																					},
																																					End: ast.Location{
																																						Line: int(677),
																																						Column: int(53),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: nil,
																																				FreeVars: ast.Identifiers{
																																					"std",
																																				},
																																			},
																																			Id: "std",
																																		},
																																		LeftBracketFodder: ast.Fodder{},
																																		Index: &ast.LiteralString{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "",
																																					Begin: ast.Location{
																																						Line: int(0),
																																						Column: int(0),
																																					},
																																					End: ast.Location{
																																						Line: int(0),
																																						Column: int(0),
																																					},
																																					File: nil,
																																				},
																																				Fodder: nil,
																																				Ctx: nil,
																																				FreeVars: nil,
																																			},
																																			Value: "length",
																																			Kind: ast.LiteralStringKind(1),
																																			BlockIndent: "",
																																			BlockTermIndent: "",
																																		},
																																		RightBracketFodder: ast.Fodder{},
																																		Id: nil,
																																	},
																																	FodderLeft: ast.Fodder{},
																																	Arguments: ast.Arguments{
																																		Positional: []ast.CommaSeparatedExpr{
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Var{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(677),
																																								Column: int(61),
																																							},
																																							End: ast.Location{
																																								Line: int(677),
																																								Column: int(64),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p7287,
																																						FreeVars: ast.Identifiers{
																																							"arr",
																																						},
																																					},
																																					Id: "arr",
																																				},
																																				CommaFodder: nil,
																																			},
																																		},
																																		Named: nil,
																																	},
																																	TrailingComma: false,
																																	TailStrict: false,
																																	FodderRight: ast.Fodder{},
																																	TailStrictFodder: nil,
																																},
																															},
																															OpFodder: ast.Fodder{},
																															Op: ast.BinaryOp(3),
																															Right: &ast.LiteralString{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(677),
																																			Column: int(68),
																																		},
																																		End: ast.Location{
																																			Line: int(677),
																																			Column: int(81),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p7233,
																																	FreeVars: nil,
																																},
																																Value: ", expected ",
																																Kind: ast.LiteralStringKind(1),
																																BlockIndent: "",
																																BlockTermIndent: "",
																															},
																														},
																														OpFodder: ast.Fodder{},
																														Op: ast.BinaryOp(3),
																														Right: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(677),
																																		Column: int(84),
																																	},
																																	End: ast.Location{
																																		Line: int(677),
																																		Column: int(85),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p7233,
																																FreeVars: ast.Identifiers{
																																	"j",
																																},
																															},
																															Id: "j",
																														},
																													},
																												},
																												ElseFodder: ast.Fodder{
																													ast.FodderElement{
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(8),
																														Comment: []string{},
																													},
																												},
																												BranchFalse: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(679),
																																Column: int(11),
																															},
																															End: ast.Location{
																																Line: int(679),
																																Column: int(12),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(10),
																																Comment: []string{},
																															},
																														},
																														Ctx: p7233,
																														FreeVars: ast.Identifiers{
																															"v",
																														},
																													},
																													Id: "v",
																												},
																											},
																											ElseFodder: ast.Fodder{
																												ast.FodderElement{
																													Kind: ast.FodderKind(0),
																													Blanks: int(0),
																													Indent: int(6),
																													Comment: []string{},
																												},
																											},
																											BranchFalse: &ast.Local{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(681),
																															Column: int(9),
																														},
																														End: ast.Location{
																															Line: int(726),
																															Column: int(64),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(8),
																															Comment: []string{},
																														},
																													},
																													Ctx: p7233,
																													FreeVars: ast.Identifiers{
																														"arr",
																														"codes",
																														"format_code",
																														"format_codes_arr",
																														"i",
																														"j",
																														"pad_left",
																														"pad_right",
																														"std",
																														"v",
																													},
																												},
																												Binds: ast.LocalBinds{
																													ast.LocalBind{
																														VarFodder: ast.Fodder{},
																														Variable: "code",
																														EqFodder: ast.Fodder{},
																														Body: &ast.Index{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(681),
																																		Column: int(22),
																																	},
																																	End: ast.Location{
																																		Line: int(681),
																																		Column: int(30),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p7302,
																																FreeVars: ast.Identifiers{
																																	"codes",
																																	"i",
																																},
																															},
																															Target: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(681),
																																			Column: int(22),
																																		},
																																		End: ast.Location{
																																			Line: int(681),
																																			Column: int(27),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p7302,
																																	FreeVars: ast.Identifiers{
																																		"codes",
																																	},
																																},
																																Id: "codes",
																															},
																															LeftBracketFodder: ast.Fodder{},
																															Index: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(681),
																																			Column: int(28),
																																		},
																																		End: ast.Location{
																																			Line: int(681),
																																			Column: int(29),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p7302,
																																	FreeVars: ast.Identifiers{
																																		"i",
																																	},
																																},
																																Id: "i",
																															},
																															RightBracketFodder: ast.Fodder{},
																															Id: nil,
																														},
																														Fun: nil,
																														CloseFodder: ast.Fodder{},
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(681),
																																Column: int(15),
																															},
																															End: ast.Location{
																																Line: int(681),
																																Column: int(30),
																															},
																															File: p1,
																														},
																													},
																												},
																												Body: &ast.Conditional{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(682),
																																Column: int(9),
																															},
																															End: ast.Location{
																																Line: int(726),
																																Column: int(64),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(8),
																																Comment: []string{},
																															},
																														},
																														Ctx: p7233,
																														FreeVars: ast.Identifiers{
																															"arr",
																															"code",
																															"codes",
																															"format_code",
																															"format_codes_arr",
																															"i",
																															"j",
																															"pad_left",
																															"pad_right",
																															"std",
																															"v",
																														},
																													},
																													Cond: &ast.Binary{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(682),
																																	Column: int(12),
																																},
																																End: ast.Location{
																																	Line: int(682),
																																	Column: int(38),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p7233,
																															FreeVars: ast.Identifiers{
																																"code",
																																"std",
																															},
																														},
																														Left: &ast.Apply{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(682),
																																		Column: int(12),
																																	},
																																	End: ast.Location{
																																		Line: int(682),
																																		Column: int(26),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p7233,
																																FreeVars: ast.Identifiers{
																																	"code",
																																	"std",
																																},
																															},
																															Target: &ast.Index{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(682),
																																			Column: int(12),
																																		},
																																		End: ast.Location{
																																			Line: int(682),
																																			Column: int(20),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p7233,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																	},
																																},
																																Target: &ast.Var{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(682),
																																				Column: int(12),
																																			},
																																			End: ast.Location{
																																				Line: int(682),
																																				Column: int(15),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{
																																			"std",
																																		},
																																	},
																																	Id: "std",
																																},
																																LeftBracketFodder: ast.Fodder{},
																																Index: &ast.LiteralString{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			End: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			File: nil,
																																		},
																																		Fodder: nil,
																																		Ctx: nil,
																																		FreeVars: nil,
																																	},
																																	Value: "type",
																																	Kind: ast.LiteralStringKind(1),
																																	BlockIndent: "",
																																	BlockTermIndent: "",
																																},
																																RightBracketFodder: ast.Fodder{},
																																Id: nil,
																															},
																															FodderLeft: ast.Fodder{},
																															Arguments: ast.Arguments{
																																Positional: []ast.CommaSeparatedExpr{
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(682),
																																						Column: int(21),
																																					},
																																					End: ast.Location{
																																						Line: int(682),
																																						Column: int(25),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p7322,
																																				FreeVars: ast.Identifiers{
																																					"code",
																																				},
																																			},
																																			Id: "code",
																																		},
																																		CommaFodder: nil,
																																	},
																																},
																																Named: nil,
																															},
																															TrailingComma: false,
																															TailStrict: false,
																															FodderRight: ast.Fodder{},
																															TailStrictFodder: nil,
																														},
																														OpFodder: ast.Fodder{},
																														Op: ast.BinaryOp(12),
																														Right: &ast.LiteralString{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(682),
																																		Column: int(30),
																																	},
																																	End: ast.Location{
																																		Line: int(682),
																																		Column: int(38),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p7233,
																																FreeVars: nil,
																															},
																															Value: "string",
																															Kind: ast.LiteralStringKind(1),
																															BlockIndent: "",
																															BlockTermIndent: "",
																														},
																													},
																													ThenFodder: ast.Fodder{},
																													BranchTrue: &ast.Apply{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(683),
																																	Column: int(11),
																																},
																																End: ast.Location{
																																	Line: int(683),
																																	Column: int(59),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p7233,
																															FreeVars: ast.Identifiers{
																																"arr",
																																"code",
																																"codes",
																																"format_codes_arr",
																																"i",
																																"j",
																																"v",
																															},
																														},
																														Target: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(683),
																																		Column: int(11),
																																	},
																																	End: ast.Location{
																																		Line: int(683),
																																		Column: int(27),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{
																																	ast.FodderElement{
																																		Kind: ast.FodderKind(0),
																																		Blanks: int(0),
																																		Indent: int(10),
																																		Comment: []string{},
																																	},
																																},
																																Ctx: p7233,
																																FreeVars: ast.Identifiers{
																																	"format_codes_arr",
																																},
																															},
																															Id: "format_codes_arr",
																														},
																														FodderLeft: ast.Fodder{},
																														Arguments: ast.Arguments{
																															Positional: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(683),
																																					Column: int(28),
																																				},
																																				End: ast.Location{
																																					Line: int(683),
																																					Column: int(33),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p7332,
																																			FreeVars: ast.Identifiers{
																																				"codes",
																																			},
																																		},
																																		Id: "codes",
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(683),
																																					Column: int(35),
																																				},
																																				End: ast.Location{
																																					Line: int(683),
																																					Column: int(38),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p7332,
																																			FreeVars: ast.Identifiers{
																																				"arr",
																																			},
																																		},
																																		Id: "arr",
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Binary{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(683),
																																					Column: int(40),
																																				},
																																				End: ast.Location{
																																					Line: int(683),
																																					Column: int(45),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p7332,
																																			FreeVars: ast.Identifiers{
																																				"i",
																																			},
																																		},
																																		Left: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(683),
																																						Column: int(40),
																																					},
																																					End: ast.Location{
																																						Line: int(683),
																																						Column: int(41),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p7332,
																																				FreeVars: ast.Identifiers{
																																					"i",
																																				},
																																			},
																																			Id: "i",
																																		},
																																		OpFodder: ast.Fodder{},
																																		Op: ast.BinaryOp(3),
																																		Right: &ast.LiteralNumber{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(683),
																																						Column: int(44),
																																					},
																																					End: ast.Location{
																																						Line: int(683),
																																						Column: int(45),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p7332,
																																				FreeVars: nil,
																																			},
																																			OriginalString: "1",
																																		},
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(683),
																																					Column: int(47),
																																				},
																																				End: ast.Location{
																																					Line: int(683),
																																					Column: int(48),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p7332,
																																			FreeVars: ast.Identifiers{
																																				"j",
																																			},
																																		},
																																		Id: "j",
																																	},
																																	CommaFodder: ast.Fodder{},
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Binary{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(683),
																																					Column: int(50),
																																				},
																																				End: ast.Location{
																																					Line: int(683),
																																					Column: int(58),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p7332,
																																			FreeVars: ast.Identifiers{
																																				"code",
																																				"v",
																																			},
																																		},
																																		Left: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(683),
																																						Column: int(50),
																																					},
																																					End: ast.Location{
																																						Line: int(683),
																																						Column: int(51),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p7332,
																																				FreeVars: ast.Identifiers{
																																					"v",
																																				},
																																			},
																																			Id: "v",
																																		},
																																		OpFodder: ast.Fodder{},
																																		Op: ast.BinaryOp(3),
																																		Right: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(683),
																																						Column: int(54),
																																					},
																																					End: ast.Location{
																																						Line: int(683),
																																						Column: int(58),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p7332,
																																				FreeVars: ast.Identifiers{
																																					"code",
																																				},
																																			},
																																			Id: "code",
																																		},
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															Named: nil,
																														},
																														TrailingComma: false,
																														TailStrict: true,
																														FodderRight: ast.Fodder{},
																														TailStrictFodder: ast.Fodder{},
																													},
																													ElseFodder: ast.Fodder{
																														ast.FodderElement{
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(8),
																															Comment: []string{},
																														},
																													},
																													BranchFalse: &ast.Local{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(685),
																																	Column: int(11),
																																},
																																End: ast.Location{
																																	Line: int(726),
																																	Column: int(64),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{
																																ast.FodderElement{
																																	Kind: ast.FodderKind(0),
																																	Blanks: int(0),
																																	Indent: int(10),
																																	Comment: []string{},
																																},
																															},
																															Ctx: p7233,
																															FreeVars: ast.Identifiers{
																																"arr",
																																"code",
																																"codes",
																																"format_code",
																																"format_codes_arr",
																																"i",
																																"j",
																																"pad_left",
																																"pad_right",
																																"std",
																																"v",
																															},
																														},
																														Binds: ast.LocalBinds{
																															ast.LocalBind{
																																VarFodder: ast.Fodder{},
																																Variable: "tmp",
																																EqFodder: ast.Fodder{},
																																Body: &ast.Conditional{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(685),
																																				Column: int(23),
																																			},
																																			End: ast.Location{
																																				Line: int(694),
																																				Column: int(12),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p7355,
																																		FreeVars: ast.Identifiers{
																																			"arr",
																																			"code",
																																			"j",
																																			"std",
																																		},
																																	},
																																	Cond: &ast.Binary{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(685),
																																					Column: int(26),
																																				},
																																				End: ast.Location{
																																					Line: int(685),
																																					Column: int(40),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p7355,
																																			FreeVars: ast.Identifiers{
																																				"code",
																																			},
																																		},
																																		Left: &ast.Index{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(685),
																																						Column: int(26),
																																					},
																																					End: ast.Location{
																																						Line: int(685),
																																						Column: int(33),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p7355,
																																				FreeVars: ast.Identifiers{
																																					"code",
																																				},
																																			},
																																			Target: &ast.Var{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(685),
																																							Column: int(26),
																																						},
																																						End: ast.Location{
																																							Line: int(685),
																																							Column: int(30),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{
																																						"code",
																																					},
																																				},
																																				Id: "code",
																																			},
																																			LeftBracketFodder: ast.Fodder{},
																																			Index: &ast.LiteralString{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						End: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						File: nil,
																																					},
																																					Fodder: nil,
																																					Ctx: nil,
																																					FreeVars: nil,
																																				},
																																				Value: "fw",
																																				Kind: ast.LiteralStringKind(1),
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																			},
																																			RightBracketFodder: ast.Fodder{},
																																			Id: nil,
																																		},
																																		OpFodder: ast.Fodder{},
																																		Op: ast.BinaryOp(12),
																																		Right: &ast.LiteralString{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(685),
																																						Column: int(37),
																																					},
																																					End: ast.Location{
																																						Line: int(685),
																																						Column: int(40),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p7355,
																																				FreeVars: nil,
																																			},
																																			Value: "*",
																																			Kind: ast.LiteralStringKind(1),
																																			BlockIndent: "",
																																			BlockTermIndent: "",
																																		},
																																	},
																																	ThenFodder: ast.Fodder{},
																																	BranchTrue: &ast.DesugaredObject{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(685),
																																					Column: int(46),
																																				},
																																				End: ast.Location{
																																					Line: int(691),
																																					Column: int(12),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p7355,
																																			FreeVars: ast.Identifiers{
																																				"arr",
																																				"j",
																																				"std",
																																			},
																																		},
																																		Asserts: ast.Nodes{},
																																		Fields: ast.DesugaredObjectFields{
																																			ast.DesugaredObjectField{
																																				Hide: ast.ObjectFieldHide(1),
																																				Name: &ast.LiteralString{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							End: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							File: nil,
																																						},
																																						Fodder: nil,
																																						Ctx: nil,
																																						FreeVars: nil,
																																					},
																																					Value: "j",
																																					Kind: ast.LiteralStringKind(1),
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																				},
																																				Body: &ast.Binary{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(686),
																																								Column: int(16),
																																							},
																																							End: ast.Location{
																																								Line: int(686),
																																								Column: int(21),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p7370,
																																						FreeVars: ast.Identifiers{
																																							"j",
																																						},
																																					},
																																					Left: &ast.Var{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(686),
																																									Column: int(16),
																																								},
																																								End: ast.Location{
																																									Line: int(686),
																																									Column: int(17),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p7370,
																																							FreeVars: ast.Identifiers{
																																								"j",
																																							},
																																						},
																																						Id: "j",
																																					},
																																					OpFodder: ast.Fodder{},
																																					Op: ast.BinaryOp(3),
																																					Right: &ast.LiteralNumber{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(686),
																																									Column: int(20),
																																								},
																																								End: ast.Location{
																																									Line: int(686),
																																									Column: int(21),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p7370,
																																							FreeVars: nil,
																																						},
																																						OriginalString: "1",
																																					},
																																				},
																																				PlusSuper: false,
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(686),
																																						Column: int(13),
																																					},
																																					End: ast.Location{
																																						Line: int(686),
																																						Column: int(21),
																																					},
																																					File: p1,
																																				},
																																			},
																																			ast.DesugaredObjectField{
																																				Hide: ast.ObjectFieldHide(1),
																																				Name: &ast.LiteralString{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							End: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							File: nil,
																																						},
																																						Fodder: nil,
																																						Ctx: nil,
																																						FreeVars: nil,
																																					},
																																					Value: "fw",
																																					Kind: ast.LiteralStringKind(1),
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																				},
																																				Body: &ast.Conditional{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(687),
																																								Column: int(17),
																																							},
																																							End: ast.Location{
																																								Line: int(690),
																																								Column: int(21),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p7370,
																																						FreeVars: ast.Identifiers{
																																							"arr",
																																							"j",
																																							"std",
																																						},
																																					},
																																					Cond: &ast.Binary{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(687),
																																									Column: int(20),
																																								},
																																								End: ast.Location{
																																									Line: int(687),
																																									Column: int(40),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p7370,
																																							FreeVars: ast.Identifiers{
																																								"arr",
																																								"j",
																																								"std",
																																							},
																																						},
																																						Left: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(687),
																																										Column: int(20),
																																									},
																																									End: ast.Location{
																																										Line: int(687),
																																										Column: int(21),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p7370,
																																								FreeVars: ast.Identifiers{
																																									"j",
																																								},
																																							},
																																							Id: "j",
																																						},
																																						OpFodder: ast.Fodder{},
																																						Op: ast.BinaryOp(8),
																																						Right: &ast.Apply{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(687),
																																										Column: int(25),
																																									},
																																									End: ast.Location{
																																										Line: int(687),
																																										Column: int(40),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p7370,
																																								FreeVars: ast.Identifiers{
																																									"arr",
																																									"std",
																																								},
																																							},
																																							Target: &ast.Index{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(687),
																																											Column: int(25),
																																										},
																																										End: ast.Location{
																																											Line: int(687),
																																											Column: int(35),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p7370,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																									},
																																								},
																																								Target: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(687),
																																												Column: int(25),
																																											},
																																											End: ast.Location{
																																												Line: int(687),
																																												Column: int(28),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																										},
																																									},
																																									Id: "std",
																																								},
																																								LeftBracketFodder: ast.Fodder{},
																																								Index: &ast.LiteralString{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											File: nil,
																																										},
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: nil,
																																									},
																																									Value: "length",
																																									Kind: ast.LiteralStringKind(1),
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								Id: nil,
																																							},
																																							FodderLeft: ast.Fodder{},
																																							Arguments: ast.Arguments{
																																								Positional: []ast.CommaSeparatedExpr{
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(687),
																																														Column: int(36),
																																													},
																																													End: ast.Location{
																																														Line: int(687),
																																														Column: int(39),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p7391,
																																												FreeVars: ast.Identifiers{
																																													"arr",
																																												},
																																											},
																																											Id: "arr",
																																										},
																																										CommaFodder: nil,
																																									},
																																								},
																																								Named: nil,
																																							},
																																							TrailingComma: false,
																																							TailStrict: false,
																																							FodderRight: ast.Fodder{},
																																							TailStrictFodder: nil,
																																						},
																																					},
																																					ThenFodder: ast.Fodder{},
																																					BranchTrue: &ast.Error{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(688),
																																									Column: int(15),
																																								},
																																								End: ast.Location{
																																									Line: int(688),
																																									Column: int(101),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(14),
																																									Comment: []string{},
																																								},
																																							},
																																							Ctx: p7370,
																																							FreeVars: ast.Identifiers{
																																								"arr",
																																								"j",
																																								"std",
																																							},
																																						},
																																						Expr: &ast.Binary{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(688),
																																										Column: int(22),
																																									},
																																									End: ast.Location{
																																										Line: int(688),
																																										Column: int(100),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p7370,
																																								FreeVars: ast.Identifiers{
																																									"arr",
																																									"j",
																																									"std",
																																								},
																																							},
																																							Left: &ast.Binary{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(688),
																																											Column: int(22),
																																										},
																																										End: ast.Location{
																																											Line: int(688),
																																											Column: int(96),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p7370,
																																									FreeVars: ast.Identifiers{
																																										"arr",
																																										"std",
																																									},
																																								},
																																								Left: &ast.Binary{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(688),
																																												Column: int(22),
																																											},
																																											End: ast.Location{
																																												Line: int(688),
																																												Column: int(71),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p7370,
																																										FreeVars: ast.Identifiers{
																																											"arr",
																																											"std",
																																										},
																																									},
																																									Left: &ast.LiteralString{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(688),
																																													Column: int(22),
																																												},
																																												End: ast.Location{
																																													Line: int(688),
																																													Column: int(53),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p7370,
																																											FreeVars: nil,
																																										},
																																										Value: "Not enough values to format: ",
																																										Kind: ast.LiteralStringKind(1),
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																									},
																																									OpFodder: ast.Fodder{},
																																									Op: ast.BinaryOp(3),
																																									Right: &ast.Apply{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(688),
																																													Column: int(56),
																																												},
																																												End: ast.Location{
																																													Line: int(688),
																																													Column: int(71),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p7370,
																																											FreeVars: ast.Identifiers{
																																												"arr",
																																												"std",
																																											},
																																										},
																																										Target: &ast.Index{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(688),
																																														Column: int(56),
																																													},
																																													End: ast.Location{
																																														Line: int(688),
																																														Column: int(66),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p7370,
																																												FreeVars: ast.Identifiers{
																																													"std",
																																												},
																																											},
																																											Target: &ast.Var{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(688),
																																															Column: int(56),
																																														},
																																														End: ast.Location{
																																															Line: int(688),
																																															Column: int(59),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: nil,
																																													FreeVars: ast.Identifiers{
																																														"std",
																																													},
																																												},
																																												Id: "std",
																																											},
																																											LeftBracketFodder: ast.Fodder{},
																																											Index: &ast.LiteralString{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(0),
																																															Column: int(0),
																																														},
																																														End: ast.Location{
																																															Line: int(0),
																																															Column: int(0),
																																														},
																																														File: nil,
																																													},
																																													Fodder: nil,
																																													Ctx: nil,
																																													FreeVars: nil,
																																												},
																																												Value: "length",
																																												Kind: ast.LiteralStringKind(1),
																																												BlockIndent: "",
																																												BlockTermIndent: "",
																																											},
																																											RightBracketFodder: ast.Fodder{},
																																											Id: nil,
																																										},
																																										FodderLeft: ast.Fodder{},
																																										Arguments: ast.Arguments{
																																											Positional: []ast.CommaSeparatedExpr{
																																												ast.CommaSeparatedExpr{
																																													Expr: &ast.Var{
																																														NodeBase: ast.NodeBase{
																																															LocRange: ast.LocationRange{
																																																FileName: "<std>",
																																																Begin: ast.Location{
																																																	Line: int(688),
																																																	Column: int(67),
																																																},
																																																End: ast.Location{
																																																	Line: int(688),
																																																	Column: int(70),
																																																},
																																																File: p1,
																																															},
																																															Fodder: ast.Fodder{},
																																															Ctx: p7412,
																																															FreeVars: ast.Identifiers{
																																																"arr",
																																															},
																																														},
																																														Id: "arr",
																																													},
																																													CommaFodder: nil,
																																												},
																																											},
																																											Named: nil,
																																										},
																																										TrailingComma: false,
																																										TailStrict: false,
																																										FodderRight: ast.Fodder{},
																																										TailStrictFodder: nil,
																																									},
																																								},
																																								OpFodder: ast.Fodder{},
																																								Op: ast.BinaryOp(3),
																																								Right: &ast.LiteralString{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(688),
																																												Column: int(74),
																																											},
																																											End: ast.Location{
																																												Line: int(688),
																																												Column: int(96),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p7370,
																																										FreeVars: nil,
																																									},
																																									Value: ", expected at least ",
																																									Kind: ast.LiteralStringKind(1),
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																								},
																																							},
																																							OpFodder: ast.Fodder{},
																																							Op: ast.BinaryOp(3),
																																							Right: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(688),
																																											Column: int(99),
																																										},
																																										End: ast.Location{
																																											Line: int(688),
																																											Column: int(100),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p7370,
																																									FreeVars: ast.Identifiers{
																																										"j",
																																									},
																																								},
																																								Id: "j",
																																							},
																																						},
																																					},
																																					ElseFodder: ast.Fodder{
																																						ast.FodderElement{
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(12),
																																							Comment: []string{},
																																						},
																																					},
																																					BranchFalse: &ast.Index{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(690),
																																									Column: int(15),
																																								},
																																								End: ast.Location{
																																									Line: int(690),
																																									Column: int(21),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p7370,
																																							FreeVars: ast.Identifiers{
																																								"arr",
																																								"j",
																																							},
																																						},
																																						Target: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(690),
																																										Column: int(15),
																																									},
																																									End: ast.Location{
																																										Line: int(690),
																																										Column: int(18),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{
																																									ast.FodderElement{
																																										Kind: ast.FodderKind(0),
																																										Blanks: int(0),
																																										Indent: int(14),
																																										Comment: []string{},
																																									},
																																								},
																																								Ctx: p7370,
																																								FreeVars: ast.Identifiers{
																																									"arr",
																																								},
																																							},
																																							Id: "arr",
																																						},
																																						LeftBracketFodder: ast.Fodder{},
																																						Index: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(690),
																																										Column: int(19),
																																									},
																																									End: ast.Location{
																																										Line: int(690),
																																										Column: int(20),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p7370,
																																								FreeVars: ast.Identifiers{
																																									"j",
																																								},
																																							},
																																							Id: "j",
																																						},
																																						RightBracketFodder: ast.Fodder{},
																																						Id: nil,
																																					},
																																				},
																																				PlusSuper: false,
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(687),
																																						Column: int(13),
																																					},
																																					End: ast.Location{
																																						Line: int(690),
																																						Column: int(21),
																																					},
																																					File: p1,
																																				},
																																			},
																																		},
																																		Locals: ast.LocalBinds{},
																																	},
																																	ElseFodder: ast.Fodder{},
																																	BranchFalse: &ast.DesugaredObject{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(691),
																																					Column: int(18),
																																				},
																																				End: ast.Location{
																																					Line: int(694),
																																					Column: int(12),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p7355,
																																			FreeVars: ast.Identifiers{
																																				"code",
																																				"j",
																																			},
																																		},
																																		Asserts: ast.Nodes{},
																																		Fields: ast.DesugaredObjectFields{
																																			ast.DesugaredObjectField{
																																				Hide: ast.ObjectFieldHide(1),
																																				Name: &ast.LiteralString{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							End: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							File: nil,
																																						},
																																						Fodder: nil,
																																						Ctx: nil,
																																						FreeVars: nil,
																																					},
																																					Value: "j",
																																					Kind: ast.LiteralStringKind(1),
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																				},
																																				Body: &ast.Var{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(692),
																																								Column: int(16),
																																							},
																																							End: ast.Location{
																																								Line: int(692),
																																								Column: int(17),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p7430,
																																						FreeVars: ast.Identifiers{
																																							"j",
																																						},
																																					},
																																					Id: "j",
																																				},
																																				PlusSuper: false,
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(692),
																																						Column: int(13),
																																					},
																																					End: ast.Location{
																																						Line: int(692),
																																						Column: int(17),
																																					},
																																					File: p1,
																																				},
																																			},
																																			ast.DesugaredObjectField{
																																				Hide: ast.ObjectFieldHide(1),
																																				Name: &ast.LiteralString{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							End: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							File: nil,
																																						},
																																						Fodder: nil,
																																						Ctx: nil,
																																						FreeVars: nil,
																																					},
																																					Value: "fw",
																																					Kind: ast.LiteralStringKind(1),
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																				},
																																				Body: &ast.Index{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(693),
																																								Column: int(17),
																																							},
																																							End: ast.Location{
																																								Line: int(693),
																																								Column: int(24),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p7430,
																																						FreeVars: ast.Identifiers{
																																							"code",
																																						},
																																					},
																																					Target: &ast.Var{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(693),
																																									Column: int(17),
																																								},
																																								End: ast.Location{
																																									Line: int(693),
																																									Column: int(21),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{
																																								"code",
																																							},
																																						},
																																						Id: "code",
																																					},
																																					LeftBracketFodder: ast.Fodder{},
																																					Index: &ast.LiteralString{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								End: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								File: nil,
																																							},
																																							Fodder: nil,
																																							Ctx: nil,
																																							FreeVars: nil,
																																						},
																																						Value: "fw",
																																						Kind: ast.LiteralStringKind(1),
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																					},
																																					RightBracketFodder: ast.Fodder{},
																																					Id: nil,
																																				},
																																				PlusSuper: false,
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(693),
																																						Column: int(13),
																																					},
																																					End: ast.Location{
																																						Line: int(693),
																																						Column: int(24),
																																					},
																																					File: p1,
																																				},
																																			},
																																		},
																																		Locals: ast.LocalBinds{},
																																	},
																																},
																																Fun: nil,
																																CloseFodder: ast.Fodder{},
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(685),
																																		Column: int(17),
																																	},
																																	End: ast.Location{
																																		Line: int(694),
																																		Column: int(12),
																																	},
																																	File: p1,
																																},
																															},
																														},
																														Body: &ast.Local{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(695),
																																		Column: int(11),
																																	},
																																	End: ast.Location{
																																		Line: int(726),
																																		Column: int(64),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{
																																	ast.FodderElement{
																																		Kind: ast.FodderKind(0),
																																		Blanks: int(0),
																																		Indent: int(10),
																																		Comment: []string{},
																																	},
																																},
																																Ctx: p7233,
																																FreeVars: ast.Identifiers{
																																	"arr",
																																	"code",
																																	"codes",
																																	"format_code",
																																	"format_codes_arr",
																																	"i",
																																	"pad_left",
																																	"pad_right",
																																	"std",
																																	"tmp",
																																	"v",
																																},
																															},
																															Binds: ast.LocalBinds{
																																ast.LocalBind{
																																	VarFodder: ast.Fodder{},
																																	Variable: "tmp2",
																																	EqFodder: ast.Fodder{},
																																	Body: &ast.Conditional{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(695),
																																					Column: int(24),
																																				},
																																				End: ast.Location{
																																					Line: int(704),
																																					Column: int(12),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p7443,
																																			FreeVars: ast.Identifiers{
																																				"arr",
																																				"code",
																																				"std",
																																				"tmp",
																																			},
																																		},
																																		Cond: &ast.Binary{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(695),
																																						Column: int(27),
																																					},
																																					End: ast.Location{
																																						Line: int(695),
																																						Column: int(43),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p7443,
																																				FreeVars: ast.Identifiers{
																																					"code",
																																				},
																																			},
																																			Left: &ast.Index{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(695),
																																							Column: int(27),
																																						},
																																						End: ast.Location{
																																							Line: int(695),
																																							Column: int(36),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p7443,
																																					FreeVars: ast.Identifiers{
																																						"code",
																																					},
																																				},
																																				Target: &ast.Var{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(695),
																																								Column: int(27),
																																							},
																																							End: ast.Location{
																																								Line: int(695),
																																								Column: int(31),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{
																																							"code",
																																						},
																																					},
																																					Id: "code",
																																				},
																																				LeftBracketFodder: ast.Fodder{},
																																				Index: &ast.LiteralString{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							End: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							File: nil,
																																						},
																																						Fodder: nil,
																																						Ctx: nil,
																																						FreeVars: nil,
																																					},
																																					Value: "prec",
																																					Kind: ast.LiteralStringKind(1),
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																				},
																																				RightBracketFodder: ast.Fodder{},
																																				Id: nil,
																																			},
																																			OpFodder: ast.Fodder{},
																																			Op: ast.BinaryOp(12),
																																			Right: &ast.LiteralString{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(695),
																																							Column: int(40),
																																						},
																																						End: ast.Location{
																																							Line: int(695),
																																							Column: int(43),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p7443,
																																					FreeVars: nil,
																																				},
																																				Value: "*",
																																				Kind: ast.LiteralStringKind(1),
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																			},
																																		},
																																		ThenFodder: ast.Fodder{},
																																		BranchTrue: &ast.DesugaredObject{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(695),
																																						Column: int(49),
																																					},
																																					End: ast.Location{
																																						Line: int(701),
																																						Column: int(12),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p7443,
																																				FreeVars: ast.Identifiers{
																																					"arr",
																																					"std",
																																					"tmp",
																																				},
																																			},
																																			Asserts: ast.Nodes{},
																																			Fields: ast.DesugaredObjectFields{
																																				ast.DesugaredObjectField{
																																					Hide: ast.ObjectFieldHide(1),
																																					Name: &ast.LiteralString{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								End: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								File: nil,
																																							},
																																							Fodder: nil,
																																							Ctx: nil,
																																							FreeVars: nil,
																																						},
																																						Value: "j",
																																						Kind: ast.LiteralStringKind(1),
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																					},
																																					Body: &ast.Binary{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(696),
																																									Column: int(16),
																																								},
																																								End: ast.Location{
																																									Line: int(696),
																																									Column: int(25),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p7458,
																																							FreeVars: ast.Identifiers{
																																								"tmp",
																																							},
																																						},
																																						Left: &ast.Index{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(696),
																																										Column: int(16),
																																									},
																																									End: ast.Location{
																																										Line: int(696),
																																										Column: int(21),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p7458,
																																								FreeVars: ast.Identifiers{
																																									"tmp",
																																								},
																																							},
																																							Target: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(696),
																																											Column: int(16),
																																										},
																																										End: ast.Location{
																																											Line: int(696),
																																											Column: int(19),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"tmp",
																																									},
																																								},
																																								Id: "tmp",
																																							},
																																							LeftBracketFodder: ast.Fodder{},
																																							Index: &ast.LiteralString{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										File: nil,
																																									},
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: nil,
																																								},
																																								Value: "j",
																																								Kind: ast.LiteralStringKind(1),
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							Id: nil,
																																						},
																																						OpFodder: ast.Fodder{},
																																						Op: ast.BinaryOp(3),
																																						Right: &ast.LiteralNumber{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(696),
																																										Column: int(24),
																																									},
																																									End: ast.Location{
																																										Line: int(696),
																																										Column: int(25),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p7458,
																																								FreeVars: nil,
																																							},
																																							OriginalString: "1",
																																						},
																																					},
																																					PlusSuper: false,
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(696),
																																							Column: int(13),
																																						},
																																						End: ast.Location{
																																							Line: int(696),
																																							Column: int(25),
																																						},
																																						File: p1,
																																					},
																																				},
																																				ast.DesugaredObjectField{
																																					Hide: ast.ObjectFieldHide(1),
																																					Name: &ast.LiteralString{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								End: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								File: nil,
																																							},
																																							Fodder: nil,
																																							Ctx: nil,
																																							FreeVars: nil,
																																						},
																																						Value: "prec",
																																						Kind: ast.LiteralStringKind(1),
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																					},
																																					Body: &ast.Conditional{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(697),
																																									Column: int(19),
																																								},
																																								End: ast.Location{
																																									Line: int(700),
																																									Column: int(25),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p7458,
																																							FreeVars: ast.Identifiers{
																																								"arr",
																																								"std",
																																								"tmp",
																																							},
																																						},
																																						Cond: &ast.Binary{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(697),
																																										Column: int(22),
																																									},
																																									End: ast.Location{
																																										Line: int(697),
																																										Column: int(46),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p7458,
																																								FreeVars: ast.Identifiers{
																																									"arr",
																																									"std",
																																									"tmp",
																																								},
																																							},
																																							Left: &ast.Index{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(697),
																																											Column: int(22),
																																										},
																																										End: ast.Location{
																																											Line: int(697),
																																											Column: int(27),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p7458,
																																									FreeVars: ast.Identifiers{
																																										"tmp",
																																									},
																																								},
																																								Target: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(697),
																																												Column: int(22),
																																											},
																																											End: ast.Location{
																																												Line: int(697),
																																												Column: int(25),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"tmp",
																																										},
																																									},
																																									Id: "tmp",
																																								},
																																								LeftBracketFodder: ast.Fodder{},
																																								Index: &ast.LiteralString{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											File: nil,
																																										},
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: nil,
																																									},
																																									Value: "j",
																																									Kind: ast.LiteralStringKind(1),
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								Id: nil,
																																							},
																																							OpFodder: ast.Fodder{},
																																							Op: ast.BinaryOp(8),
																																							Right: &ast.Apply{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(697),
																																											Column: int(31),
																																										},
																																										End: ast.Location{
																																											Line: int(697),
																																											Column: int(46),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p7458,
																																									FreeVars: ast.Identifiers{
																																										"arr",
																																										"std",
																																									},
																																								},
																																								Target: &ast.Index{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(697),
																																												Column: int(31),
																																											},
																																											End: ast.Location{
																																												Line: int(697),
																																												Column: int(41),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p7458,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																										},
																																									},
																																									Target: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(697),
																																													Column: int(31),
																																												},
																																												End: ast.Location{
																																													Line: int(697),
																																													Column: int(34),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{
																																												"std",
																																											},
																																										},
																																										Id: "std",
																																									},
																																									LeftBracketFodder: ast.Fodder{},
																																									Index: &ast.LiteralString{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												End: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												File: nil,
																																											},
																																											Fodder: nil,
																																											Ctx: nil,
																																											FreeVars: nil,
																																										},
																																										Value: "length",
																																										Kind: ast.LiteralStringKind(1),
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																									},
																																									RightBracketFodder: ast.Fodder{},
																																									Id: nil,
																																								},
																																								FodderLeft: ast.Fodder{},
																																								Arguments: ast.Arguments{
																																									Positional: []ast.CommaSeparatedExpr{
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Var{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(697),
																																															Column: int(42),
																																														},
																																														End: ast.Location{
																																															Line: int(697),
																																															Column: int(45),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: p7485,
																																													FreeVars: ast.Identifiers{
																																														"arr",
																																													},
																																												},
																																												Id: "arr",
																																											},
																																											CommaFodder: nil,
																																										},
																																									},
																																									Named: nil,
																																								},
																																								TrailingComma: false,
																																								TailStrict: false,
																																								FodderRight: ast.Fodder{},
																																								TailStrictFodder: nil,
																																							},
																																						},
																																						ThenFodder: ast.Fodder{},
																																						BranchTrue: &ast.Error{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(698),
																																										Column: int(15),
																																									},
																																									End: ast.Location{
																																										Line: int(698),
																																										Column: int(105),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{
																																									ast.FodderElement{
																																										Kind: ast.FodderKind(0),
																																										Blanks: int(0),
																																										Indent: int(14),
																																										Comment: []string{},
																																									},
																																								},
																																								Ctx: p7458,
																																								FreeVars: ast.Identifiers{
																																									"arr",
																																									"std",
																																									"tmp",
																																								},
																																							},
																																							Expr: &ast.Binary{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(698),
																																											Column: int(22),
																																										},
																																										End: ast.Location{
																																											Line: int(698),
																																											Column: int(104),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p7458,
																																									FreeVars: ast.Identifiers{
																																										"arr",
																																										"std",
																																										"tmp",
																																									},
																																								},
																																								Left: &ast.Binary{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(698),
																																												Column: int(22),
																																											},
																																											End: ast.Location{
																																												Line: int(698),
																																												Column: int(96),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p7458,
																																										FreeVars: ast.Identifiers{
																																											"arr",
																																											"std",
																																										},
																																									},
																																									Left: &ast.Binary{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(698),
																																													Column: int(22),
																																												},
																																												End: ast.Location{
																																													Line: int(698),
																																													Column: int(71),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p7458,
																																											FreeVars: ast.Identifiers{
																																												"arr",
																																												"std",
																																											},
																																										},
																																										Left: &ast.LiteralString{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(698),
																																														Column: int(22),
																																													},
																																													End: ast.Location{
																																														Line: int(698),
																																														Column: int(53),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p7458,
																																												FreeVars: nil,
																																											},
																																											Value: "Not enough values to format: ",
																																											Kind: ast.LiteralStringKind(1),
																																											BlockIndent: "",
																																											BlockTermIndent: "",
																																										},
																																										OpFodder: ast.Fodder{},
																																										Op: ast.BinaryOp(3),
																																										Right: &ast.Apply{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(698),
																																														Column: int(56),
																																													},
																																													End: ast.Location{
																																														Line: int(698),
																																														Column: int(71),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p7458,
																																												FreeVars: ast.Identifiers{
																																													"arr",
																																													"std",
																																												},
																																											},
																																											Target: &ast.Index{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(698),
																																															Column: int(56),
																																														},
																																														End: ast.Location{
																																															Line: int(698),
																																															Column: int(66),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: p7458,
																																													FreeVars: ast.Identifiers{
																																														"std",
																																													},
																																												},
																																												Target: &ast.Var{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "<std>",
																																															Begin: ast.Location{
																																																Line: int(698),
																																																Column: int(56),
																																															},
																																															End: ast.Location{
																																																Line: int(698),
																																																Column: int(59),
																																															},
																																															File: p1,
																																														},
																																														Fodder: ast.Fodder{},
																																														Ctx: nil,
																																														FreeVars: ast.Identifiers{
																																															"std",
																																														},
																																													},
																																													Id: "std",
																																												},
																																												LeftBracketFodder: ast.Fodder{},
																																												Index: &ast.LiteralString{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "",
																																															Begin: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																															End: ast.Location{
																																																Line: int(0),
																																																Column: int(0),
																																															},
																																															File: nil,
																																														},
																																														Fodder: nil,
																																														Ctx: nil,
																																														FreeVars: nil,
																																													},
																																													Value: "length",
																																													Kind: ast.LiteralStringKind(1),
																																													BlockIndent: "",
																																													BlockTermIndent: "",
																																												},
																																												RightBracketFodder: ast.Fodder{},
																																												Id: nil,
																																											},
																																											FodderLeft: ast.Fodder{},
																																											Arguments: ast.Arguments{
																																												Positional: []ast.CommaSeparatedExpr{
																																													ast.CommaSeparatedExpr{
																																														Expr: &ast.Var{
																																															NodeBase: ast.NodeBase{
																																																LocRange: ast.LocationRange{
																																																	FileName: "<std>",
																																																	Begin: ast.Location{
																																																		Line: int(698),
																																																		Column: int(67),
																																																	},
																																																	End: ast.Location{
																																																		Line: int(698),
																																																		Column: int(70),
																																																	},
																																																	File: p1,
																																																},
																																																Fodder: ast.Fodder{},
																																																Ctx: p7506,
																																																FreeVars: ast.Identifiers{
																																																	"arr",
																																																},
																																															},
																																															Id: "arr",
																																														},
																																														CommaFodder: nil,
																																													},
																																												},
																																												Named: nil,
																																											},
																																											TrailingComma: false,
																																											TailStrict: false,
																																											FodderRight: ast.Fodder{},
																																											TailStrictFodder: nil,
																																										},
																																									},
																																									OpFodder: ast.Fodder{},
																																									Op: ast.BinaryOp(3),
																																									Right: &ast.LiteralString{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(698),
																																													Column: int(74),
																																												},
																																												End: ast.Location{
																																													Line: int(698),
																																													Column: int(96),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p7458,
																																											FreeVars: nil,
																																										},
																																										Value: ", expected at least ",
																																										Kind: ast.LiteralStringKind(1),
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																									},
																																								},
																																								OpFodder: ast.Fodder{},
																																								Op: ast.BinaryOp(3),
																																								Right: &ast.Index{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(698),
																																												Column: int(99),
																																											},
																																											End: ast.Location{
																																												Line: int(698),
																																												Column: int(104),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p7458,
																																										FreeVars: ast.Identifiers{
																																											"tmp",
																																										},
																																									},
																																									Target: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(698),
																																													Column: int(99),
																																												},
																																												End: ast.Location{
																																													Line: int(698),
																																													Column: int(102),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{
																																												"tmp",
																																											},
																																										},
																																										Id: "tmp",
																																									},
																																									LeftBracketFodder: ast.Fodder{},
																																									Index: &ast.LiteralString{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												End: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												File: nil,
																																											},
																																											Fodder: nil,
																																											Ctx: nil,
																																											FreeVars: nil,
																																										},
																																										Value: "j",
																																										Kind: ast.LiteralStringKind(1),
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																									},
																																									RightBracketFodder: ast.Fodder{},
																																									Id: nil,
																																								},
																																							},
																																						},
																																						ElseFodder: ast.Fodder{
																																							ast.FodderElement{
																																								Kind: ast.FodderKind(0),
																																								Blanks: int(0),
																																								Indent: int(12),
																																								Comment: []string{},
																																							},
																																						},
																																						BranchFalse: &ast.Index{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(700),
																																										Column: int(15),
																																									},
																																									End: ast.Location{
																																										Line: int(700),
																																										Column: int(25),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p7458,
																																								FreeVars: ast.Identifiers{
																																									"arr",
																																									"tmp",
																																								},
																																							},
																																							Target: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(700),
																																											Column: int(15),
																																										},
																																										End: ast.Location{
																																											Line: int(700),
																																											Column: int(18),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(14),
																																											Comment: []string{},
																																										},
																																									},
																																									Ctx: p7458,
																																									FreeVars: ast.Identifiers{
																																										"arr",
																																									},
																																								},
																																								Id: "arr",
																																							},
																																							LeftBracketFodder: ast.Fodder{},
																																							Index: &ast.Index{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(700),
																																											Column: int(19),
																																										},
																																										End: ast.Location{
																																											Line: int(700),
																																											Column: int(24),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p7458,
																																									FreeVars: ast.Identifiers{
																																										"tmp",
																																									},
																																								},
																																								Target: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(700),
																																												Column: int(19),
																																											},
																																											End: ast.Location{
																																												Line: int(700),
																																												Column: int(22),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"tmp",
																																										},
																																									},
																																									Id: "tmp",
																																								},
																																								LeftBracketFodder: ast.Fodder{},
																																								Index: &ast.LiteralString{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											File: nil,
																																										},
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: nil,
																																									},
																																									Value: "j",
																																									Kind: ast.LiteralStringKind(1),
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								Id: nil,
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							Id: nil,
																																						},
																																					},
																																					PlusSuper: false,
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(697),
																																							Column: int(13),
																																						},
																																						End: ast.Location{
																																							Line: int(700),
																																							Column: int(25),
																																						},
																																						File: p1,
																																					},
																																				},
																																			},
																																			Locals: ast.LocalBinds{},
																																		},
																																		ElseFodder: ast.Fodder{},
																																		BranchFalse: &ast.DesugaredObject{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(701),
																																						Column: int(18),
																																					},
																																					End: ast.Location{
																																						Line: int(704),
																																						Column: int(12),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p7443,
																																				FreeVars: ast.Identifiers{
																																					"code",
																																					"tmp",
																																				},
																																			},
																																			Asserts: ast.Nodes{},
																																			Fields: ast.DesugaredObjectFields{
																																				ast.DesugaredObjectField{
																																					Hide: ast.ObjectFieldHide(1),
																																					Name: &ast.LiteralString{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								End: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								File: nil,
																																							},
																																							Fodder: nil,
																																							Ctx: nil,
																																							FreeVars: nil,
																																						},
																																						Value: "j",
																																						Kind: ast.LiteralStringKind(1),
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																					},
																																					Body: &ast.Index{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(702),
																																									Column: int(16),
																																								},
																																								End: ast.Location{
																																									Line: int(702),
																																									Column: int(21),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p7530,
																																							FreeVars: ast.Identifiers{
																																								"tmp",
																																							},
																																						},
																																						Target: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(702),
																																										Column: int(16),
																																									},
																																									End: ast.Location{
																																										Line: int(702),
																																										Column: int(19),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: nil,
																																								FreeVars: ast.Identifiers{
																																									"tmp",
																																								},
																																							},
																																							Id: "tmp",
																																						},
																																						LeftBracketFodder: ast.Fodder{},
																																						Index: &ast.LiteralString{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																									End: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																									File: nil,
																																								},
																																								Fodder: nil,
																																								Ctx: nil,
																																								FreeVars: nil,
																																							},
																																							Value: "j",
																																							Kind: ast.LiteralStringKind(1),
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																						},
																																						RightBracketFodder: ast.Fodder{},
																																						Id: nil,
																																					},
																																					PlusSuper: false,
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(702),
																																							Column: int(13),
																																						},
																																						End: ast.Location{
																																							Line: int(702),
																																							Column: int(21),
																																						},
																																						File: p1,
																																					},
																																				},
																																				ast.DesugaredObjectField{
																																					Hide: ast.ObjectFieldHide(1),
																																					Name: &ast.LiteralString{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								End: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								File: nil,
																																							},
																																							Fodder: nil,
																																							Ctx: nil,
																																							FreeVars: nil,
																																						},
																																						Value: "prec",
																																						Kind: ast.LiteralStringKind(1),
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																					},
																																					Body: &ast.Index{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(703),
																																									Column: int(19),
																																								},
																																								End: ast.Location{
																																									Line: int(703),
																																									Column: int(28),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p7530,
																																							FreeVars: ast.Identifiers{
																																								"code",
																																							},
																																						},
																																						Target: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(703),
																																										Column: int(19),
																																									},
																																									End: ast.Location{
																																										Line: int(703),
																																										Column: int(23),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: nil,
																																								FreeVars: ast.Identifiers{
																																									"code",
																																								},
																																							},
																																							Id: "code",
																																						},
																																						LeftBracketFodder: ast.Fodder{},
																																						Index: &ast.LiteralString{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																									End: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																									File: nil,
																																								},
																																								Fodder: nil,
																																								Ctx: nil,
																																								FreeVars: nil,
																																							},
																																							Value: "prec",
																																							Kind: ast.LiteralStringKind(1),
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																						},
																																						RightBracketFodder: ast.Fodder{},
																																						Id: nil,
																																					},
																																					PlusSuper: false,
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(703),
																																							Column: int(13),
																																						},
																																						End: ast.Location{
																																							Line: int(703),
																																							Column: int(28),
																																						},
																																						File: p1,
																																					},
																																				},
																																			},
																																			Locals: ast.LocalBinds{},
																																		},
																																	},
																																	Fun: nil,
																																	CloseFodder: ast.Fodder{},
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(695),
																																			Column: int(17),
																																		},
																																		End: ast.Location{
																																			Line: int(704),
																																			Column: int(12),
																																		},
																																		File: p1,
																																	},
																																},
																															},
																															Body: &ast.Local{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(705),
																																			Column: int(11),
																																		},
																																		End: ast.Location{
																																			Line: int(726),
																																			Column: int(64),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{
																																		ast.FodderElement{
																																			Kind: ast.FodderKind(0),
																																			Blanks: int(0),
																																			Indent: int(10),
																																			Comment: []string{},
																																		},
																																	},
																																	Ctx: p7233,
																																	FreeVars: ast.Identifiers{
																																		"arr",
																																		"code",
																																		"codes",
																																		"format_code",
																																		"format_codes_arr",
																																		"i",
																																		"pad_left",
																																		"pad_right",
																																		"std",
																																		"tmp",
																																		"tmp2",
																																		"v",
																																	},
																																},
																																Binds: ast.LocalBinds{
																																	ast.LocalBind{
																																		VarFodder: ast.Fodder{},
																																		Variable: "j2",
																																		EqFodder: ast.Fodder{},
																																		Body: &ast.Index{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(705),
																																						Column: int(22),
																																					},
																																					End: ast.Location{
																																						Line: int(705),
																																						Column: int(28),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p7546,
																																				FreeVars: ast.Identifiers{
																																					"tmp2",
																																				},
																																			},
																																			Target: &ast.Var{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(705),
																																							Column: int(22),
																																						},
																																						End: ast.Location{
																																							Line: int(705),
																																							Column: int(26),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{
																																						"tmp2",
																																					},
																																				},
																																				Id: "tmp2",
																																			},
																																			LeftBracketFodder: ast.Fodder{},
																																			Index: &ast.LiteralString{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						End: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						File: nil,
																																					},
																																					Fodder: nil,
																																					Ctx: nil,
																																					FreeVars: nil,
																																				},
																																				Value: "j",
																																				Kind: ast.LiteralStringKind(1),
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																			},
																																			RightBracketFodder: ast.Fodder{},
																																			Id: nil,
																																		},
																																		Fun: nil,
																																		CloseFodder: ast.Fodder{},
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(705),
																																				Column: int(17),
																																			},
																																			End: ast.Location{
																																				Line: int(705),
																																				Column: int(28),
																																			},
																																			File: p1,
																																		},
																																	},
																																},
																																Body: &ast.Local{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(706),
																																				Column: int(11),
																																			},
																																			End: ast.Location{
																																				Line: int(726),
																																				Column: int(64),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{
																																			ast.FodderElement{
																																				Kind: ast.FodderKind(0),
																																				Blanks: int(0),
																																				Indent: int(10),
																																				Comment: []string{},
																																			},
																																		},
																																		Ctx: p7233,
																																		FreeVars: ast.Identifiers{
																																			"arr",
																																			"code",
																																			"codes",
																																			"format_code",
																																			"format_codes_arr",
																																			"i",
																																			"j2",
																																			"pad_left",
																																			"pad_right",
																																			"std",
																																			"tmp",
																																			"tmp2",
																																			"v",
																																		},
																																	},
																																	Binds: ast.LocalBinds{
																																		ast.LocalBind{
																																			VarFodder: ast.Fodder{},
																																			Variable: "val",
																																			EqFodder: ast.Fodder{},
																																			Body: &ast.Conditional{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(707),
																																							Column: int(13),
																																						},
																																						End: ast.Location{
																																							Line: int(710),
																																							Column: int(103),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{
																																						ast.FodderElement{
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(12),
																																							Comment: []string{},
																																						},
																																					},
																																					Ctx: p7557,
																																					FreeVars: ast.Identifiers{
																																						"arr",
																																						"j2",
																																						"std",
																																					},
																																				},
																																				Cond: &ast.Binary{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(707),
																																								Column: int(16),
																																							},
																																							End: ast.Location{
																																								Line: int(707),
																																								Column: int(36),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p7557,
																																						FreeVars: ast.Identifiers{
																																							"arr",
																																							"j2",
																																							"std",
																																						},
																																					},
																																					Left: &ast.Var{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(707),
																																									Column: int(16),
																																								},
																																								End: ast.Location{
																																									Line: int(707),
																																									Column: int(18),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p7557,
																																							FreeVars: ast.Identifiers{
																																								"j2",
																																							},
																																						},
																																						Id: "j2",
																																					},
																																					OpFodder: ast.Fodder{},
																																					Op: ast.BinaryOp(9),
																																					Right: &ast.Apply{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(707),
																																									Column: int(21),
																																								},
																																								End: ast.Location{
																																									Line: int(707),
																																									Column: int(36),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p7557,
																																							FreeVars: ast.Identifiers{
																																								"arr",
																																								"std",
																																							},
																																						},
																																						Target: &ast.Index{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(707),
																																										Column: int(21),
																																									},
																																									End: ast.Location{
																																										Line: int(707),
																																										Column: int(31),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p7557,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																								},
																																							},
																																							Target: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(707),
																																											Column: int(21),
																																										},
																																										End: ast.Location{
																																											Line: int(707),
																																											Column: int(24),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																									},
																																								},
																																								Id: "std",
																																							},
																																							LeftBracketFodder: ast.Fodder{},
																																							Index: &ast.LiteralString{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										File: nil,
																																									},
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: nil,
																																								},
																																								Value: "length",
																																								Kind: ast.LiteralStringKind(1),
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							Id: nil,
																																						},
																																						FodderLeft: ast.Fodder{},
																																						Arguments: ast.Arguments{
																																							Positional: []ast.CommaSeparatedExpr{
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(707),
																																													Column: int(32),
																																												},
																																												End: ast.Location{
																																													Line: int(707),
																																													Column: int(35),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p7572,
																																											FreeVars: ast.Identifiers{
																																												"arr",
																																											},
																																										},
																																										Id: "arr",
																																									},
																																									CommaFodder: nil,
																																								},
																																							},
																																							Named: nil,
																																						},
																																						TrailingComma: false,
																																						TailStrict: false,
																																						FodderRight: ast.Fodder{},
																																						TailStrictFodder: nil,
																																					},
																																				},
																																				ThenFodder: ast.Fodder{},
																																				BranchTrue: &ast.Index{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(708),
																																								Column: int(15),
																																							},
																																							End: ast.Location{
																																								Line: int(708),
																																								Column: int(22),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p7557,
																																						FreeVars: ast.Identifiers{
																																							"arr",
																																							"j2",
																																						},
																																					},
																																					Target: &ast.Var{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(708),
																																									Column: int(15),
																																								},
																																								End: ast.Location{
																																									Line: int(708),
																																									Column: int(18),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(14),
																																									Comment: []string{},
																																								},
																																							},
																																							Ctx: p7557,
																																							FreeVars: ast.Identifiers{
																																								"arr",
																																							},
																																						},
																																						Id: "arr",
																																					},
																																					LeftBracketFodder: ast.Fodder{},
																																					Index: &ast.Var{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(708),
																																									Column: int(19),
																																								},
																																								End: ast.Location{
																																									Line: int(708),
																																									Column: int(21),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p7557,
																																							FreeVars: ast.Identifiers{
																																								"j2",
																																							},
																																						},
																																						Id: "j2",
																																					},
																																					RightBracketFodder: ast.Fodder{},
																																					Id: nil,
																																				},
																																				ElseFodder: ast.Fodder{
																																					ast.FodderElement{
																																						Kind: ast.FodderKind(0),
																																						Blanks: int(0),
																																						Indent: int(12),
																																						Comment: []string{},
																																					},
																																				},
																																				BranchFalse: &ast.Error{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(710),
																																								Column: int(15),
																																							},
																																							End: ast.Location{
																																								Line: int(710),
																																								Column: int(103),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{
																																							ast.FodderElement{
																																								Kind: ast.FodderKind(0),
																																								Blanks: int(0),
																																								Indent: int(14),
																																								Comment: []string{},
																																							},
																																						},
																																						Ctx: p7557,
																																						FreeVars: ast.Identifiers{
																																							"arr",
																																							"j2",
																																							"std",
																																						},
																																					},
																																					Expr: &ast.Binary{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(710),
																																									Column: int(22),
																																								},
																																								End: ast.Location{
																																									Line: int(710),
																																									Column: int(102),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p7557,
																																							FreeVars: ast.Identifiers{
																																								"arr",
																																								"j2",
																																								"std",
																																							},
																																						},
																																						Left: &ast.Binary{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(710),
																																										Column: int(22),
																																									},
																																									End: ast.Location{
																																										Line: int(710),
																																										Column: int(97),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p7557,
																																								FreeVars: ast.Identifiers{
																																									"arr",
																																									"std",
																																								},
																																							},
																																							Left: &ast.Binary{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(710),
																																											Column: int(22),
																																										},
																																										End: ast.Location{
																																											Line: int(710),
																																											Column: int(71),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p7557,
																																									FreeVars: ast.Identifiers{
																																										"arr",
																																										"std",
																																									},
																																								},
																																								Left: &ast.LiteralString{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(710),
																																												Column: int(22),
																																											},
																																											End: ast.Location{
																																												Line: int(710),
																																												Column: int(53),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p7557,
																																										FreeVars: nil,
																																									},
																																									Value: "Not enough values to format: ",
																																									Kind: ast.LiteralStringKind(1),
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																								},
																																								OpFodder: ast.Fodder{},
																																								Op: ast.BinaryOp(3),
																																								Right: &ast.Apply{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(710),
																																												Column: int(56),
																																											},
																																											End: ast.Location{
																																												Line: int(710),
																																												Column: int(71),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p7557,
																																										FreeVars: ast.Identifiers{
																																											"arr",
																																											"std",
																																										},
																																									},
																																									Target: &ast.Index{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(710),
																																													Column: int(56),
																																												},
																																												End: ast.Location{
																																													Line: int(710),
																																													Column: int(66),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p7557,
																																											FreeVars: ast.Identifiers{
																																												"std",
																																											},
																																										},
																																										Target: &ast.Var{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(710),
																																														Column: int(56),
																																													},
																																													End: ast.Location{
																																														Line: int(710),
																																														Column: int(59),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: nil,
																																												FreeVars: ast.Identifiers{
																																													"std",
																																												},
																																											},
																																											Id: "std",
																																										},
																																										LeftBracketFodder: ast.Fodder{},
																																										Index: &ast.LiteralString{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(0),
																																														Column: int(0),
																																													},
																																													End: ast.Location{
																																														Line: int(0),
																																														Column: int(0),
																																													},
																																													File: nil,
																																												},
																																												Fodder: nil,
																																												Ctx: nil,
																																												FreeVars: nil,
																																											},
																																											Value: "length",
																																											Kind: ast.LiteralStringKind(1),
																																											BlockIndent: "",
																																											BlockTermIndent: "",
																																										},
																																										RightBracketFodder: ast.Fodder{},
																																										Id: nil,
																																									},
																																									FodderLeft: ast.Fodder{},
																																									Arguments: ast.Arguments{
																																										Positional: []ast.CommaSeparatedExpr{
																																											ast.CommaSeparatedExpr{
																																												Expr: &ast.Var{
																																													NodeBase: ast.NodeBase{
																																														LocRange: ast.LocationRange{
																																															FileName: "<std>",
																																															Begin: ast.Location{
																																																Line: int(710),
																																																Column: int(67),
																																															},
																																															End: ast.Location{
																																																Line: int(710),
																																																Column: int(70),
																																															},
																																															File: p1,
																																														},
																																														Fodder: ast.Fodder{},
																																														Ctx: p7601,
																																														FreeVars: ast.Identifiers{
																																															"arr",
																																														},
																																													},
																																													Id: "arr",
																																												},
																																												CommaFodder: nil,
																																											},
																																										},
																																										Named: nil,
																																									},
																																									TrailingComma: false,
																																									TailStrict: false,
																																									FodderRight: ast.Fodder{},
																																									TailStrictFodder: nil,
																																								},
																																							},
																																							OpFodder: ast.Fodder{},
																																							Op: ast.BinaryOp(3),
																																							Right: &ast.LiteralString{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(710),
																																											Column: int(74),
																																										},
																																										End: ast.Location{
																																											Line: int(710),
																																											Column: int(97),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p7557,
																																									FreeVars: nil,
																																								},
																																								Value: ", expected more than ",
																																								Kind: ast.LiteralStringKind(1),
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																							},
																																						},
																																						OpFodder: ast.Fodder{},
																																						Op: ast.BinaryOp(3),
																																						Right: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(710),
																																										Column: int(100),
																																									},
																																									End: ast.Location{
																																										Line: int(710),
																																										Column: int(102),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p7557,
																																								FreeVars: ast.Identifiers{
																																									"j2",
																																								},
																																							},
																																							Id: "j2",
																																						},
																																					},
																																				},
																																			},
																																			Fun: nil,
																																			CloseFodder: ast.Fodder{},
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(706),
																																					Column: int(17),
																																				},
																																				End: ast.Location{
																																					Line: int(710),
																																					Column: int(103),
																																				},
																																				File: p1,
																																			},
																																		},
																																	},
																																	Body: &ast.Local{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(711),
																																					Column: int(11),
																																				},
																																				End: ast.Location{
																																					Line: int(726),
																																					Column: int(64),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{
																																				ast.FodderElement{
																																					Kind: ast.FodderKind(0),
																																					Blanks: int(0),
																																					Indent: int(10),
																																					Comment: []string{},
																																				},
																																			},
																																			Ctx: p7233,
																																			FreeVars: ast.Identifiers{
																																				"arr",
																																				"code",
																																				"codes",
																																				"format_code",
																																				"format_codes_arr",
																																				"i",
																																				"j2",
																																				"pad_left",
																																				"pad_right",
																																				"tmp",
																																				"tmp2",
																																				"v",
																																				"val",
																																			},
																																		},
																																		Binds: ast.LocalBinds{
																																			ast.LocalBind{
																																				VarFodder: ast.Fodder{},
																																				Variable: "s",
																																				EqFodder: ast.Fodder{},
																																				Body: &ast.Conditional{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(712),
																																								Column: int(13),
																																							},
																																							End: ast.Location{
																																								Line: int(715),
																																								Column: int(60),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{
																																							ast.FodderElement{
																																								Kind: ast.FodderKind(0),
																																								Blanks: int(0),
																																								Indent: int(12),
																																								Comment: []string{},
																																							},
																																						},
																																						Ctx: p7612,
																																						FreeVars: ast.Identifiers{
																																							"code",
																																							"format_code",
																																							"j2",
																																							"tmp",
																																							"tmp2",
																																							"val",
																																						},
																																					},
																																					Cond: &ast.Binary{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(712),
																																									Column: int(16),
																																								},
																																								End: ast.Location{
																																									Line: int(712),
																																									Column: int(33),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p7612,
																																							FreeVars: ast.Identifiers{
																																								"code",
																																							},
																																						},
																																						Left: &ast.Index{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(712),
																																										Column: int(16),
																																									},
																																									End: ast.Location{
																																										Line: int(712),
																																										Column: int(26),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p7612,
																																								FreeVars: ast.Identifiers{
																																									"code",
																																								},
																																							},
																																							Target: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(712),
																																											Column: int(16),
																																										},
																																										End: ast.Location{
																																											Line: int(712),
																																											Column: int(20),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"code",
																																									},
																																								},
																																								Id: "code",
																																							},
																																							LeftBracketFodder: ast.Fodder{},
																																							Index: &ast.LiteralString{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										File: nil,
																																									},
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: nil,
																																								},
																																								Value: "ctype",
																																								Kind: ast.LiteralStringKind(1),
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							Id: nil,
																																						},
																																						OpFodder: ast.Fodder{},
																																						Op: ast.BinaryOp(12),
																																						Right: &ast.LiteralString{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(712),
																																										Column: int(30),
																																									},
																																									End: ast.Location{
																																										Line: int(712),
																																										Column: int(33),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p7612,
																																								FreeVars: nil,
																																							},
																																							Value: "%",
																																							Kind: ast.LiteralStringKind(1),
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																						},
																																					},
																																					ThenFodder: ast.Fodder{},
																																					BranchTrue: &ast.LiteralString{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(713),
																																									Column: int(15),
																																								},
																																								End: ast.Location{
																																									Line: int(713),
																																									Column: int(18),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(14),
																																									Comment: []string{},
																																								},
																																							},
																																							Ctx: p7612,
																																							FreeVars: nil,
																																						},
																																						Value: "%",
																																						Kind: ast.LiteralStringKind(1),
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																					},
																																					ElseFodder: ast.Fodder{
																																						ast.FodderElement{
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(12),
																																							Comment: []string{},
																																						},
																																					},
																																					BranchFalse: &ast.Apply{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(715),
																																									Column: int(15),
																																								},
																																								End: ast.Location{
																																									Line: int(715),
																																									Column: int(60),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p7612,
																																							FreeVars: ast.Identifiers{
																																								"code",
																																								"format_code",
																																								"j2",
																																								"tmp",
																																								"tmp2",
																																								"val",
																																							},
																																						},
																																						Target: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(715),
																																										Column: int(15),
																																									},
																																									End: ast.Location{
																																										Line: int(715),
																																										Column: int(26),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{
																																									ast.FodderElement{
																																										Kind: ast.FodderKind(0),
																																										Blanks: int(0),
																																										Indent: int(14),
																																										Comment: []string{},
																																									},
																																								},
																																								Ctx: p7612,
																																								FreeVars: ast.Identifiers{
																																									"format_code",
																																								},
																																							},
																																							Id: "format_code",
																																						},
																																						FodderLeft: ast.Fodder{},
																																						Arguments: ast.Arguments{
																																							Positional: []ast.CommaSeparatedExpr{
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(715),
																																													Column: int(27),
																																												},
																																												End: ast.Location{
																																													Line: int(715),
																																													Column: int(30),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p7632,
																																											FreeVars: ast.Identifiers{
																																												"val",
																																											},
																																										},
																																										Id: "val",
																																									},
																																									CommaFodder: ast.Fodder{},
																																								},
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(715),
																																													Column: int(32),
																																												},
																																												End: ast.Location{
																																													Line: int(715),
																																													Column: int(36),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p7632,
																																											FreeVars: ast.Identifiers{
																																												"code",
																																											},
																																										},
																																										Id: "code",
																																									},
																																									CommaFodder: ast.Fodder{},
																																								},
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Index{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(715),
																																													Column: int(38),
																																												},
																																												End: ast.Location{
																																													Line: int(715),
																																													Column: int(44),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p7632,
																																											FreeVars: ast.Identifiers{
																																												"tmp",
																																											},
																																										},
																																										Target: &ast.Var{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(715),
																																														Column: int(38),
																																													},
																																													End: ast.Location{
																																														Line: int(715),
																																														Column: int(41),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: nil,
																																												FreeVars: ast.Identifiers{
																																													"tmp",
																																												},
																																											},
																																											Id: "tmp",
																																										},
																																										LeftBracketFodder: ast.Fodder{},
																																										Index: &ast.LiteralString{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(0),
																																														Column: int(0),
																																													},
																																													End: ast.Location{
																																														Line: int(0),
																																														Column: int(0),
																																													},
																																													File: nil,
																																												},
																																												Fodder: nil,
																																												Ctx: nil,
																																												FreeVars: nil,
																																											},
																																											Value: "fw",
																																											Kind: ast.LiteralStringKind(1),
																																											BlockIndent: "",
																																											BlockTermIndent: "",
																																										},
																																										RightBracketFodder: ast.Fodder{},
																																										Id: nil,
																																									},
																																									CommaFodder: ast.Fodder{},
																																								},
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Index{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(715),
																																													Column: int(46),
																																												},
																																												End: ast.Location{
																																													Line: int(715),
																																													Column: int(55),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p7632,
																																											FreeVars: ast.Identifiers{
																																												"tmp2",
																																											},
																																										},
																																										Target: &ast.Var{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(715),
																																														Column: int(46),
																																													},
																																													End: ast.Location{
																																														Line: int(715),
																																														Column: int(50),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: nil,
																																												FreeVars: ast.Identifiers{
																																													"tmp2",
																																												},
																																											},
																																											Id: "tmp2",
																																										},
																																										LeftBracketFodder: ast.Fodder{},
																																										Index: &ast.LiteralString{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "",
																																													Begin: ast.Location{
																																														Line: int(0),
																																														Column: int(0),
																																													},
																																													End: ast.Location{
																																														Line: int(0),
																																														Column: int(0),
																																													},
																																													File: nil,
																																												},
																																												Fodder: nil,
																																												Ctx: nil,
																																												FreeVars: nil,
																																											},
																																											Value: "prec",
																																											Kind: ast.LiteralStringKind(1),
																																											BlockIndent: "",
																																											BlockTermIndent: "",
																																										},
																																										RightBracketFodder: ast.Fodder{},
																																										Id: nil,
																																									},
																																									CommaFodder: ast.Fodder{},
																																								},
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(715),
																																													Column: int(57),
																																												},
																																												End: ast.Location{
																																													Line: int(715),
																																													Column: int(59),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p7632,
																																											FreeVars: ast.Identifiers{
																																												"j2",
																																											},
																																										},
																																										Id: "j2",
																																									},
																																									CommaFodder: nil,
																																								},
																																							},
																																							Named: nil,
																																						},
																																						TrailingComma: false,
																																						TailStrict: false,
																																						FodderRight: ast.Fodder{},
																																						TailStrictFodder: nil,
																																					},
																																				},
																																				Fun: nil,
																																				CloseFodder: ast.Fodder{},
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(711),
																																						Column: int(17),
																																					},
																																					End: ast.Location{
																																						Line: int(715),
																																						Column: int(60),
																																					},
																																					File: p1,
																																				},
																																			},
																																		},
																																		Body: &ast.Local{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(716),
																																						Column: int(11),
																																					},
																																					End: ast.Location{
																																						Line: int(726),
																																						Column: int(64),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{
																																					ast.FodderElement{
																																						Kind: ast.FodderKind(0),
																																						Blanks: int(0),
																																						Indent: int(10),
																																						Comment: []string{},
																																					},
																																				},
																																				Ctx: p7233,
																																				FreeVars: ast.Identifiers{
																																					"arr",
																																					"code",
																																					"codes",
																																					"format_codes_arr",
																																					"i",
																																					"j2",
																																					"pad_left",
																																					"pad_right",
																																					"s",
																																					"tmp",
																																					"v",
																																				},
																																			},
																																			Binds: ast.LocalBinds{
																																				ast.LocalBind{
																																					VarFodder: ast.Fodder{},
																																					Variable: "s_padded",
																																					EqFodder: ast.Fodder{},
																																					Body: &ast.Conditional{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(717),
																																									Column: int(13),
																																								},
																																								End: ast.Location{
																																									Line: int(720),
																																									Column: int(39),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(12),
																																									Comment: []string{},
																																								},
																																							},
																																							Ctx: p7654,
																																							FreeVars: ast.Identifiers{
																																								"code",
																																								"pad_left",
																																								"pad_right",
																																								"s",
																																								"tmp",
																																							},
																																						},
																																						Cond: &ast.Index{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(717),
																																										Column: int(16),
																																									},
																																									End: ast.Location{
																																										Line: int(717),
																																										Column: int(32),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p7654,
																																								FreeVars: ast.Identifiers{
																																									"code",
																																								},
																																							},
																																							Target: &ast.Index{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(717),
																																											Column: int(16),
																																										},
																																										End: ast.Location{
																																											Line: int(717),
																																											Column: int(27),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"code",
																																									},
																																								},
																																								Target: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(717),
																																												Column: int(16),
																																											},
																																											End: ast.Location{
																																												Line: int(717),
																																												Column: int(20),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"code",
																																										},
																																									},
																																									Id: "code",
																																								},
																																								LeftBracketFodder: ast.Fodder{},
																																								Index: &ast.LiteralString{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											File: nil,
																																										},
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: nil,
																																									},
																																									Value: "cflags",
																																									Kind: ast.LiteralStringKind(1),
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								Id: nil,
																																							},
																																							LeftBracketFodder: ast.Fodder{},
																																							Index: &ast.LiteralString{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										File: nil,
																																									},
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: nil,
																																								},
																																								Value: "left",
																																								Kind: ast.LiteralStringKind(1),
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							Id: nil,
																																						},
																																						ThenFodder: ast.Fodder{},
																																						BranchTrue: &ast.Apply{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(718),
																																										Column: int(15),
																																									},
																																									End: ast.Location{
																																										Line: int(718),
																																										Column: int(40),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p7654,
																																								FreeVars: ast.Identifiers{
																																									"pad_right",
																																									"s",
																																									"tmp",
																																								},
																																							},
																																							Target: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(718),
																																											Column: int(15),
																																										},
																																										End: ast.Location{
																																											Line: int(718),
																																											Column: int(24),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(14),
																																											Comment: []string{},
																																										},
																																									},
																																									Ctx: p7654,
																																									FreeVars: ast.Identifiers{
																																										"pad_right",
																																									},
																																								},
																																								Id: "pad_right",
																																							},
																																							FodderLeft: ast.Fodder{},
																																							Arguments: ast.Arguments{
																																								Positional: []ast.CommaSeparatedExpr{
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(718),
																																														Column: int(25),
																																													},
																																													End: ast.Location{
																																														Line: int(718),
																																														Column: int(26),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p7671,
																																												FreeVars: ast.Identifiers{
																																													"s",
																																												},
																																											},
																																											Id: "s",
																																										},
																																										CommaFodder: ast.Fodder{},
																																									},
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Index{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(718),
																																														Column: int(28),
																																													},
																																													End: ast.Location{
																																														Line: int(718),
																																														Column: int(34),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p7671,
																																												FreeVars: ast.Identifiers{
																																													"tmp",
																																												},
																																											},
																																											Target: &ast.Var{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(718),
																																															Column: int(28),
																																														},
																																														End: ast.Location{
																																															Line: int(718),
																																															Column: int(31),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: nil,
																																													FreeVars: ast.Identifiers{
																																														"tmp",
																																													},
																																												},
																																												Id: "tmp",
																																											},
																																											LeftBracketFodder: ast.Fodder{},
																																											Index: &ast.LiteralString{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(0),
																																															Column: int(0),
																																														},
																																														End: ast.Location{
																																															Line: int(0),
																																															Column: int(0),
																																														},
																																														File: nil,
																																													},
																																													Fodder: nil,
																																													Ctx: nil,
																																													FreeVars: nil,
																																												},
																																												Value: "fw",
																																												Kind: ast.LiteralStringKind(1),
																																												BlockIndent: "",
																																												BlockTermIndent: "",
																																											},
																																											RightBracketFodder: ast.Fodder{},
																																											Id: nil,
																																										},
																																										CommaFodder: ast.Fodder{},
																																									},
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.LiteralString{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(718),
																																														Column: int(36),
																																													},
																																													End: ast.Location{
																																														Line: int(718),
																																														Column: int(39),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p7671,
																																												FreeVars: nil,
																																											},
																																											Value: " ",
																																											Kind: ast.LiteralStringKind(1),
																																											BlockIndent: "",
																																											BlockTermIndent: "",
																																										},
																																										CommaFodder: nil,
																																									},
																																								},
																																								Named: nil,
																																							},
																																							TrailingComma: false,
																																							TailStrict: false,
																																							FodderRight: ast.Fodder{},
																																							TailStrictFodder: nil,
																																						},
																																						ElseFodder: ast.Fodder{
																																							ast.FodderElement{
																																								Kind: ast.FodderKind(0),
																																								Blanks: int(0),
																																								Indent: int(12),
																																								Comment: []string{},
																																							},
																																						},
																																						BranchFalse: &ast.Apply{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(720),
																																										Column: int(15),
																																									},
																																									End: ast.Location{
																																										Line: int(720),
																																										Column: int(39),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p7654,
																																								FreeVars: ast.Identifiers{
																																									"pad_left",
																																									"s",
																																									"tmp",
																																								},
																																							},
																																							Target: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(720),
																																											Column: int(15),
																																										},
																																										End: ast.Location{
																																											Line: int(720),
																																											Column: int(23),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(14),
																																											Comment: []string{},
																																										},
																																									},
																																									Ctx: p7654,
																																									FreeVars: ast.Identifiers{
																																										"pad_left",
																																									},
																																								},
																																								Id: "pad_left",
																																							},
																																							FodderLeft: ast.Fodder{},
																																							Arguments: ast.Arguments{
																																								Positional: []ast.CommaSeparatedExpr{
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(720),
																																														Column: int(24),
																																													},
																																													End: ast.Location{
																																														Line: int(720),
																																														Column: int(25),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p7687,
																																												FreeVars: ast.Identifiers{
																																													"s",
																																												},
																																											},
																																											Id: "s",
																																										},
																																										CommaFodder: ast.Fodder{},
																																									},
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Index{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(720),
																																														Column: int(27),
																																													},
																																													End: ast.Location{
																																														Line: int(720),
																																														Column: int(33),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p7687,
																																												FreeVars: ast.Identifiers{
																																													"tmp",
																																												},
																																											},
																																											Target: &ast.Var{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(720),
																																															Column: int(27),
																																														},
																																														End: ast.Location{
																																															Line: int(720),
																																															Column: int(30),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: nil,
																																													FreeVars: ast.Identifiers{
																																														"tmp",
																																													},
																																												},
																																												Id: "tmp",
																																											},
																																											LeftBracketFodder: ast.Fodder{},
																																											Index: &ast.LiteralString{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "",
																																														Begin: ast.Location{
																																															Line: int(0),
																																															Column: int(0),
																																														},
																																														End: ast.Location{
																																															Line: int(0),
																																															Column: int(0),
																																														},
																																														File: nil,
																																													},
																																													Fodder: nil,
																																													Ctx: nil,
																																													FreeVars: nil,
																																												},
																																												Value: "fw",
																																												Kind: ast.LiteralStringKind(1),
																																												BlockIndent: "",
																																												BlockTermIndent: "",
																																											},
																																											RightBracketFodder: ast.Fodder{},
																																											Id: nil,
																																										},
																																										CommaFodder: ast.Fodder{},
																																									},
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.LiteralString{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(720),
																																														Column: int(35),
																																													},
																																													End: ast.Location{
																																														Line: int(720),
																																														Column: int(38),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p7687,
																																												FreeVars: nil,
																																											},
																																											Value: " ",
																																											Kind: ast.LiteralStringKind(1),
																																											BlockIndent: "",
																																											BlockTermIndent: "",
																																										},
																																										CommaFodder: nil,
																																									},
																																								},
																																								Named: nil,
																																							},
																																							TrailingComma: false,
																																							TailStrict: false,
																																							FodderRight: ast.Fodder{},
																																							TailStrictFodder: nil,
																																						},
																																					},
																																					Fun: nil,
																																					CloseFodder: ast.Fodder{},
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(716),
																																							Column: int(17),
																																						},
																																						End: ast.Location{
																																							Line: int(720),
																																							Column: int(39),
																																						},
																																						File: p1,
																																					},
																																				},
																																			},
																																			Body: &ast.Local{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(721),
																																							Column: int(11),
																																						},
																																						End: ast.Location{
																																							Line: int(726),
																																							Column: int(64),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{
																																						ast.FodderElement{
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(10),
																																							Comment: []string{},
																																						},
																																					},
																																					Ctx: p7233,
																																					FreeVars: ast.Identifiers{
																																						"arr",
																																						"code",
																																						"codes",
																																						"format_codes_arr",
																																						"i",
																																						"j2",
																																						"s_padded",
																																						"v",
																																					},
																																				},
																																				Binds: ast.LocalBinds{
																																					ast.LocalBind{
																																						VarFodder: ast.Fodder{},
																																						Variable: "j3",
																																						EqFodder: ast.Fodder{},
																																						Body: &ast.Conditional{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(722),
																																										Column: int(13),
																																									},
																																									End: ast.Location{
																																										Line: int(725),
																																										Column: int(21),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{
																																									ast.FodderElement{
																																										Kind: ast.FodderKind(0),
																																										Blanks: int(0),
																																										Indent: int(12),
																																										Comment: []string{},
																																									},
																																								},
																																								Ctx: p7701,
																																								FreeVars: ast.Identifiers{
																																									"code",
																																									"j2",
																																								},
																																							},
																																							Cond: &ast.Binary{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(722),
																																											Column: int(16),
																																										},
																																										End: ast.Location{
																																											Line: int(722),
																																											Column: int(33),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p7701,
																																									FreeVars: ast.Identifiers{
																																										"code",
																																									},
																																								},
																																								Left: &ast.Index{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(722),
																																												Column: int(16),
																																											},
																																											End: ast.Location{
																																												Line: int(722),
																																												Column: int(26),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p7701,
																																										FreeVars: ast.Identifiers{
																																											"code",
																																										},
																																									},
																																									Target: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(722),
																																													Column: int(16),
																																												},
																																												End: ast.Location{
																																													Line: int(722),
																																													Column: int(20),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{
																																												"code",
																																											},
																																										},
																																										Id: "code",
																																									},
																																									LeftBracketFodder: ast.Fodder{},
																																									Index: &ast.LiteralString{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												End: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												File: nil,
																																											},
																																											Fodder: nil,
																																											Ctx: nil,
																																											FreeVars: nil,
																																										},
																																										Value: "ctype",
																																										Kind: ast.LiteralStringKind(1),
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																									},
																																									RightBracketFodder: ast.Fodder{},
																																									Id: nil,
																																								},
																																								OpFodder: ast.Fodder{},
																																								Op: ast.BinaryOp(12),
																																								Right: &ast.LiteralString{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(722),
																																												Column: int(30),
																																											},
																																											End: ast.Location{
																																												Line: int(722),
																																												Column: int(33),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p7701,
																																										FreeVars: nil,
																																									},
																																									Value: "%",
																																									Kind: ast.LiteralStringKind(1),
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																								},
																																							},
																																							ThenFodder: ast.Fodder{},
																																							BranchTrue: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(723),
																																											Column: int(15),
																																										},
																																										End: ast.Location{
																																											Line: int(723),
																																											Column: int(17),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(14),
																																											Comment: []string{},
																																										},
																																									},
																																									Ctx: p7701,
																																									FreeVars: ast.Identifiers{
																																										"j2",
																																									},
																																								},
																																								Id: "j2",
																																							},
																																							ElseFodder: ast.Fodder{
																																								ast.FodderElement{
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(12),
																																									Comment: []string{},
																																								},
																																							},
																																							BranchFalse: &ast.Binary{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(725),
																																											Column: int(15),
																																										},
																																										End: ast.Location{
																																											Line: int(725),
																																											Column: int(21),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p7701,
																																									FreeVars: ast.Identifiers{
																																										"j2",
																																									},
																																								},
																																								Left: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(725),
																																												Column: int(15),
																																											},
																																											End: ast.Location{
																																												Line: int(725),
																																												Column: int(17),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{
																																											ast.FodderElement{
																																												Kind: ast.FodderKind(0),
																																												Blanks: int(0),
																																												Indent: int(14),
																																												Comment: []string{},
																																											},
																																										},
																																										Ctx: p7701,
																																										FreeVars: ast.Identifiers{
																																											"j2",
																																										},
																																									},
																																									Id: "j2",
																																								},
																																								OpFodder: ast.Fodder{},
																																								Op: ast.BinaryOp(3),
																																								Right: &ast.LiteralNumber{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(725),
																																												Column: int(20),
																																											},
																																											End: ast.Location{
																																												Line: int(725),
																																												Column: int(21),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p7701,
																																										FreeVars: nil,
																																									},
																																									OriginalString: "1",
																																								},
																																							},
																																						},
																																						Fun: nil,
																																						CloseFodder: ast.Fodder{},
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(721),
																																								Column: int(17),
																																							},
																																							End: ast.Location{
																																								Line: int(725),
																																								Column: int(21),
																																							},
																																							File: p1,
																																						},
																																					},
																																				},
																																				Body: &ast.Apply{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(726),
																																								Column: int(11),
																																							},
																																							End: ast.Location{
																																								Line: int(726),
																																								Column: int(64),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p7233,
																																						FreeVars: ast.Identifiers{
																																							"arr",
																																							"codes",
																																							"format_codes_arr",
																																							"i",
																																							"j3",
																																							"s_padded",
																																							"v",
																																						},
																																					},
																																					Target: &ast.Var{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(726),
																																									Column: int(11),
																																								},
																																								End: ast.Location{
																																									Line: int(726),
																																									Column: int(27),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(10),
																																									Comment: []string{},
																																								},
																																							},
																																							Ctx: p7233,
																																							FreeVars: ast.Identifiers{
																																								"format_codes_arr",
																																							},
																																						},
																																						Id: "format_codes_arr",
																																					},
																																					FodderLeft: ast.Fodder{},
																																					Arguments: ast.Arguments{
																																						Positional: []ast.CommaSeparatedExpr{
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(726),
																																												Column: int(28),
																																											},
																																											End: ast.Location{
																																												Line: int(726),
																																												Column: int(33),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p7728,
																																										FreeVars: ast.Identifiers{
																																											"codes",
																																										},
																																									},
																																									Id: "codes",
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(726),
																																												Column: int(35),
																																											},
																																											End: ast.Location{
																																												Line: int(726),
																																												Column: int(38),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p7728,
																																										FreeVars: ast.Identifiers{
																																											"arr",
																																										},
																																									},
																																									Id: "arr",
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Binary{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(726),
																																												Column: int(40),
																																											},
																																											End: ast.Location{
																																												Line: int(726),
																																												Column: int(45),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p7728,
																																										FreeVars: ast.Identifiers{
																																											"i",
																																										},
																																									},
																																									Left: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(726),
																																													Column: int(40),
																																												},
																																												End: ast.Location{
																																													Line: int(726),
																																													Column: int(41),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p7728,
																																											FreeVars: ast.Identifiers{
																																												"i",
																																											},
																																										},
																																										Id: "i",
																																									},
																																									OpFodder: ast.Fodder{},
																																									Op: ast.BinaryOp(3),
																																									Right: &ast.LiteralNumber{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(726),
																																													Column: int(44),
																																												},
																																												End: ast.Location{
																																													Line: int(726),
																																													Column: int(45),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p7728,
																																											FreeVars: nil,
																																										},
																																										OriginalString: "1",
																																									},
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(726),
																																												Column: int(47),
																																											},
																																											End: ast.Location{
																																												Line: int(726),
																																												Column: int(49),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p7728,
																																										FreeVars: ast.Identifiers{
																																											"j3",
																																										},
																																									},
																																									Id: "j3",
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Binary{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(726),
																																												Column: int(51),
																																											},
																																											End: ast.Location{
																																												Line: int(726),
																																												Column: int(63),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p7728,
																																										FreeVars: ast.Identifiers{
																																											"s_padded",
																																											"v",
																																										},
																																									},
																																									Left: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(726),
																																													Column: int(51),
																																												},
																																												End: ast.Location{
																																													Line: int(726),
																																													Column: int(52),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p7728,
																																											FreeVars: ast.Identifiers{
																																												"v",
																																											},
																																										},
																																										Id: "v",
																																									},
																																									OpFodder: ast.Fodder{},
																																									Op: ast.BinaryOp(3),
																																									Right: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(726),
																																													Column: int(55),
																																												},
																																												End: ast.Location{
																																													Line: int(726),
																																													Column: int(63),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p7728,
																																											FreeVars: ast.Identifiers{
																																												"s_padded",
																																											},
																																										},
																																										Id: "s_padded",
																																									},
																																								},
																																								CommaFodder: nil,
																																							},
																																						},
																																						Named: nil,
																																					},
																																					TrailingComma: false,
																																					TailStrict: true,
																																					FodderRight: ast.Fodder{},
																																					TailStrictFodder: ast.Fodder{},
																																				},
																																			},
																																		},
																																	},
																																},
																															},
																														},
																													},
																												},
																											},
																										},
																									},
																									Fun: nil,
																									CloseFodder: nil,
																									LocRange: ast.LocationRange{
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										File: nil,
																									},
																								},
																							},
																							Body: &ast.Local{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(729),
																											Column: int(5),
																										},
																										End: ast.Location{
																											Line: int(774),
																											Column: int(48),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{
																										ast.FodderElement{
																											Kind: ast.FodderKind(0),
																											Blanks: int(1),
																											Indent: int(4),
																											Comment: []string{},
																										},
																										ast.FodderElement{
																											Kind: ast.FodderKind(2),
																											Blanks: int(0),
																											Indent: int(4),
																											Comment: []string{
																												"// Render a parsed format string with an object of values.",
																											},
																										},
																									},
																									Ctx: p3497,
																									FreeVars: ast.Identifiers{
																										"codes",
																										"format_code",
																										"format_codes_arr",
																										"pad_left",
																										"pad_right",
																										"std",
																										"vals",
																									},
																								},
																								Binds: ast.LocalBinds{
																									ast.LocalBind{
																										VarFodder: nil,
																										Variable: "format_codes_obj",
																										EqFodder: nil,
																										Body: &ast.Function{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(729),
																														Column: int(11),
																													},
																													End: ast.Location{
																														Line: int(767),
																														Column: int(60),
																													},
																													File: p1,
																												},
																												Fodder: nil,
																												Ctx: p7751,
																												FreeVars: ast.Identifiers{
																													"format_code",
																													"format_codes_obj",
																													"pad_left",
																													"pad_right",
																													"std",
																												},
																											},
																											ParenLeftFodder: ast.Fodder{},
																											Parameters: []ast.Parameter{
																												ast.Parameter{
																													NameFodder: ast.Fodder{},
																													Name: "codes",
																													EqFodder: nil,
																													DefaultArg: nil,
																													CommaFodder: ast.Fodder{},
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(729),
																															Column: int(28),
																														},
																														End: ast.Location{
																															Line: int(729),
																															Column: int(33),
																														},
																														File: p1,
																													},
																												},
																												ast.Parameter{
																													NameFodder: ast.Fodder{},
																													Name: "obj",
																													EqFodder: nil,
																													DefaultArg: nil,
																													CommaFodder: ast.Fodder{},
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(729),
																															Column: int(35),
																														},
																														End: ast.Location{
																															Line: int(729),
																															Column: int(38),
																														},
																														File: p1,
																													},
																												},
																												ast.Parameter{
																													NameFodder: ast.Fodder{},
																													Name: "i",
																													EqFodder: nil,
																													DefaultArg: nil,
																													CommaFodder: ast.Fodder{},
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(729),
																															Column: int(40),
																														},
																														End: ast.Location{
																															Line: int(729),
																															Column: int(41),
																														},
																														File: p1,
																													},
																												},
																												ast.Parameter{
																													NameFodder: ast.Fodder{},
																													Name: "v",
																													EqFodder: nil,
																													DefaultArg: nil,
																													CommaFodder: nil,
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(729),
																															Column: int(43),
																														},
																														End: ast.Location{
																															Line: int(729),
																															Column: int(44),
																														},
																														File: p1,
																													},
																												},
																											},
																											TrailingComma: false,
																											ParenRightFodder: ast.Fodder{},
																											Body: &ast.Conditional{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(730),
																															Column: int(7),
																														},
																														End: ast.Location{
																															Line: int(767),
																															Column: int(60),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(6),
																															Comment: []string{},
																														},
																													},
																													Ctx: p7756,
																													FreeVars: ast.Identifiers{
																														"codes",
																														"format_code",
																														"format_codes_obj",
																														"i",
																														"obj",
																														"pad_left",
																														"pad_right",
																														"std",
																														"v",
																													},
																												},
																												Cond: &ast.Binary{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(730),
																																Column: int(10),
																															},
																															End: ast.Location{
																																Line: int(730),
																																Column: int(32),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p7756,
																														FreeVars: ast.Identifiers{
																															"codes",
																															"i",
																															"std",
																														},
																													},
																													Left: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(730),
																																	Column: int(10),
																																},
																																End: ast.Location{
																																	Line: int(730),
																																	Column: int(11),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p7756,
																															FreeVars: ast.Identifiers{
																																"i",
																															},
																														},
																														Id: "i",
																													},
																													OpFodder: ast.Fodder{},
																													Op: ast.BinaryOp(8),
																													Right: &ast.Apply{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(730),
																																	Column: int(15),
																																},
																																End: ast.Location{
																																	Line: int(730),
																																	Column: int(32),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p7756,
																															FreeVars: ast.Identifiers{
																																"codes",
																																"std",
																															},
																														},
																														Target: &ast.Index{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(730),
																																		Column: int(15),
																																	},
																																	End: ast.Location{
																																		Line: int(730),
																																		Column: int(25),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p7756,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																															},
																															Target: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(730),
																																			Column: int(15),
																																		},
																																		End: ast.Location{
																																			Line: int(730),
																																			Column: int(18),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																	},
																																},
																																Id: "std",
																															},
																															LeftBracketFodder: ast.Fodder{},
																															Index: &ast.LiteralString{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		File: nil,
																																	},
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: nil,
																																},
																																Value: "length",
																																Kind: ast.LiteralStringKind(1),
																																BlockIndent: "",
																																BlockTermIndent: "",
																															},
																															RightBracketFodder: ast.Fodder{},
																															Id: nil,
																														},
																														FodderLeft: ast.Fodder{},
																														Arguments: ast.Arguments{
																															Positional: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(730),
																																					Column: int(26),
																																				},
																																				End: ast.Location{
																																					Line: int(730),
																																					Column: int(31),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p7771,
																																			FreeVars: ast.Identifiers{
																																				"codes",
																																			},
																																		},
																																		Id: "codes",
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															Named: nil,
																														},
																														TrailingComma: false,
																														TailStrict: false,
																														FodderRight: ast.Fodder{},
																														TailStrictFodder: nil,
																													},
																												},
																												ThenFodder: ast.Fodder{},
																												BranchTrue: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(731),
																																Column: int(9),
																															},
																															End: ast.Location{
																																Line: int(731),
																																Column: int(10),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(8),
																																Comment: []string{},
																															},
																														},
																														Ctx: p7756,
																														FreeVars: ast.Identifiers{
																															"v",
																														},
																													},
																													Id: "v",
																												},
																												ElseFodder: ast.Fodder{
																													ast.FodderElement{
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(6),
																														Comment: []string{},
																													},
																												},
																												BranchFalse: &ast.Local{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(733),
																																Column: int(9),
																															},
																															End: ast.Location{
																																Line: int(767),
																																Column: int(60),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(8),
																																Comment: []string{},
																															},
																														},
																														Ctx: p7756,
																														FreeVars: ast.Identifiers{
																															"codes",
																															"format_code",
																															"format_codes_obj",
																															"i",
																															"obj",
																															"pad_left",
																															"pad_right",
																															"std",
																															"v",
																														},
																													},
																													Binds: ast.LocalBinds{
																														ast.LocalBind{
																															VarFodder: ast.Fodder{},
																															Variable: "code",
																															EqFodder: ast.Fodder{},
																															Body: &ast.Index{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(733),
																																			Column: int(22),
																																		},
																																		End: ast.Location{
																																			Line: int(733),
																																			Column: int(30),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p7782,
																																	FreeVars: ast.Identifiers{
																																		"codes",
																																		"i",
																																	},
																																},
																																Target: &ast.Var{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(733),
																																				Column: int(22),
																																			},
																																			End: ast.Location{
																																				Line: int(733),
																																				Column: int(27),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p7782,
																																		FreeVars: ast.Identifiers{
																																			"codes",
																																		},
																																	},
																																	Id: "codes",
																																},
																																LeftBracketFodder: ast.Fodder{},
																																Index: &ast.Var{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(733),
																																				Column: int(28),
																																			},
																																			End: ast.Location{
																																				Line: int(733),
																																				Column: int(29),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p7782,
																																		FreeVars: ast.Identifiers{
																																			"i",
																																		},
																																	},
																																	Id: "i",
																																},
																																RightBracketFodder: ast.Fodder{},
																																Id: nil,
																															},
																															Fun: nil,
																															CloseFodder: ast.Fodder{},
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(733),
																																	Column: int(15),
																																},
																																End: ast.Location{
																																	Line: int(733),
																																	Column: int(30),
																																},
																																File: p1,
																															},
																														},
																													},
																													Body: &ast.Conditional{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(734),
																																	Column: int(9),
																																},
																																End: ast.Location{
																																	Line: int(767),
																																	Column: int(60),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{
																																ast.FodderElement{
																																	Kind: ast.FodderKind(0),
																																	Blanks: int(0),
																																	Indent: int(8),
																																	Comment: []string{},
																																},
																															},
																															Ctx: p7756,
																															FreeVars: ast.Identifiers{
																																"code",
																																"codes",
																																"format_code",
																																"format_codes_obj",
																																"i",
																																"obj",
																																"pad_left",
																																"pad_right",
																																"std",
																																"v",
																															},
																														},
																														Cond: &ast.Binary{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(734),
																																		Column: int(12),
																																	},
																																	End: ast.Location{
																																		Line: int(734),
																																		Column: int(38),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p7756,
																																FreeVars: ast.Identifiers{
																																	"code",
																																	"std",
																																},
																															},
																															Left: &ast.Apply{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(734),
																																			Column: int(12),
																																		},
																																		End: ast.Location{
																																			Line: int(734),
																																			Column: int(26),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p7756,
																																	FreeVars: ast.Identifiers{
																																		"code",
																																		"std",
																																	},
																																},
																																Target: &ast.Index{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(734),
																																				Column: int(12),
																																			},
																																			End: ast.Location{
																																				Line: int(734),
																																				Column: int(20),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p7756,
																																		FreeVars: ast.Identifiers{
																																			"std",
																																		},
																																	},
																																	Target: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(734),
																																					Column: int(12),
																																				},
																																				End: ast.Location{
																																					Line: int(734),
																																					Column: int(15),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: nil,
																																			FreeVars: ast.Identifiers{
																																				"std",
																																			},
																																		},
																																		Id: "std",
																																	},
																																	LeftBracketFodder: ast.Fodder{},
																																	Index: &ast.LiteralString{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "",
																																				Begin: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				End: ast.Location{
																																					Line: int(0),
																																					Column: int(0),
																																				},
																																				File: nil,
																																			},
																																			Fodder: nil,
																																			Ctx: nil,
																																			FreeVars: nil,
																																		},
																																		Value: "type",
																																		Kind: ast.LiteralStringKind(1),
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																	},
																																	RightBracketFodder: ast.Fodder{},
																																	Id: nil,
																																},
																																FodderLeft: ast.Fodder{},
																																Arguments: ast.Arguments{
																																	Positional: []ast.CommaSeparatedExpr{
																																		ast.CommaSeparatedExpr{
																																			Expr: &ast.Var{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(734),
																																							Column: int(21),
																																						},
																																						End: ast.Location{
																																							Line: int(734),
																																							Column: int(25),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p7802,
																																					FreeVars: ast.Identifiers{
																																						"code",
																																					},
																																				},
																																				Id: "code",
																																			},
																																			CommaFodder: nil,
																																		},
																																	},
																																	Named: nil,
																																},
																																TrailingComma: false,
																																TailStrict: false,
																																FodderRight: ast.Fodder{},
																																TailStrictFodder: nil,
																															},
																															OpFodder: ast.Fodder{},
																															Op: ast.BinaryOp(12),
																															Right: &ast.LiteralString{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(734),
																																			Column: int(30),
																																		},
																																		End: ast.Location{
																																			Line: int(734),
																																			Column: int(38),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p7756,
																																	FreeVars: nil,
																																},
																																Value: "string",
																																Kind: ast.LiteralStringKind(1),
																																BlockIndent: "",
																																BlockTermIndent: "",
																															},
																														},
																														ThenFodder: ast.Fodder{},
																														BranchTrue: &ast.Apply{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(735),
																																		Column: int(11),
																																	},
																																	End: ast.Location{
																																		Line: int(735),
																																		Column: int(56),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p7756,
																																FreeVars: ast.Identifiers{
																																	"code",
																																	"codes",
																																	"format_codes_obj",
																																	"i",
																																	"obj",
																																	"v",
																																},
																															},
																															Target: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(735),
																																			Column: int(11),
																																		},
																																		End: ast.Location{
																																			Line: int(735),
																																			Column: int(27),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{
																																		ast.FodderElement{
																																			Kind: ast.FodderKind(0),
																																			Blanks: int(0),
																																			Indent: int(10),
																																			Comment: []string{},
																																		},
																																	},
																																	Ctx: p7756,
																																	FreeVars: ast.Identifiers{
																																		"format_codes_obj",
																																	},
																																},
																																Id: "format_codes_obj",
																															},
																															FodderLeft: ast.Fodder{},
																															Arguments: ast.Arguments{
																																Positional: []ast.CommaSeparatedExpr{
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(735),
																																						Column: int(28),
																																					},
																																					End: ast.Location{
																																						Line: int(735),
																																						Column: int(33),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p7812,
																																				FreeVars: ast.Identifiers{
																																					"codes",
																																				},
																																			},
																																			Id: "codes",
																																		},
																																		CommaFodder: ast.Fodder{},
																																	},
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(735),
																																						Column: int(35),
																																					},
																																					End: ast.Location{
																																						Line: int(735),
																																						Column: int(38),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p7812,
																																				FreeVars: ast.Identifiers{
																																					"obj",
																																				},
																																			},
																																			Id: "obj",
																																		},
																																		CommaFodder: ast.Fodder{},
																																	},
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Binary{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(735),
																																						Column: int(40),
																																					},
																																					End: ast.Location{
																																						Line: int(735),
																																						Column: int(45),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p7812,
																																				FreeVars: ast.Identifiers{
																																					"i",
																																				},
																																			},
																																			Left: &ast.Var{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(735),
																																							Column: int(40),
																																						},
																																						End: ast.Location{
																																							Line: int(735),
																																							Column: int(41),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p7812,
																																					FreeVars: ast.Identifiers{
																																						"i",
																																					},
																																				},
																																				Id: "i",
																																			},
																																			OpFodder: ast.Fodder{},
																																			Op: ast.BinaryOp(3),
																																			Right: &ast.LiteralNumber{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(735),
																																							Column: int(44),
																																						},
																																						End: ast.Location{
																																							Line: int(735),
																																							Column: int(45),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p7812,
																																					FreeVars: nil,
																																				},
																																				OriginalString: "1",
																																			},
																																		},
																																		CommaFodder: ast.Fodder{},
																																	},
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Binary{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(735),
																																						Column: int(47),
																																					},
																																					End: ast.Location{
																																						Line: int(735),
																																						Column: int(55),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p7812,
																																				FreeVars: ast.Identifiers{
																																					"code",
																																					"v",
																																				},
																																			},
																																			Left: &ast.Var{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(735),
																																							Column: int(47),
																																						},
																																						End: ast.Location{
																																							Line: int(735),
																																							Column: int(48),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p7812,
																																					FreeVars: ast.Identifiers{
																																						"v",
																																					},
																																				},
																																				Id: "v",
																																			},
																																			OpFodder: ast.Fodder{},
																																			Op: ast.BinaryOp(3),
																																			Right: &ast.Var{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(735),
																																							Column: int(51),
																																						},
																																						End: ast.Location{
																																							Line: int(735),
																																							Column: int(55),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p7812,
																																					FreeVars: ast.Identifiers{
																																						"code",
																																					},
																																				},
																																				Id: "code",
																																			},
																																		},
																																		CommaFodder: nil,
																																	},
																																},
																																Named: nil,
																															},
																															TrailingComma: false,
																															TailStrict: true,
																															FodderRight: ast.Fodder{},
																															TailStrictFodder: ast.Fodder{},
																														},
																														ElseFodder: ast.Fodder{
																															ast.FodderElement{
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(8),
																																Comment: []string{},
																															},
																														},
																														BranchFalse: &ast.Local{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(737),
																																		Column: int(11),
																																	},
																																	End: ast.Location{
																																		Line: int(767),
																																		Column: int(60),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{
																																	ast.FodderElement{
																																		Kind: ast.FodderKind(0),
																																		Blanks: int(0),
																																		Indent: int(10),
																																		Comment: []string{},
																																	},
																																},
																																Ctx: p7756,
																																FreeVars: ast.Identifiers{
																																	"code",
																																	"codes",
																																	"format_code",
																																	"format_codes_obj",
																																	"i",
																																	"obj",
																																	"pad_left",
																																	"pad_right",
																																	"std",
																																	"v",
																																},
																															},
																															Binds: ast.LocalBinds{
																																ast.LocalBind{
																																	VarFodder: ast.Fodder{},
																																	Variable: "f",
																																	EqFodder: ast.Fodder{},
																																	Body: &ast.Conditional{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(738),
																																					Column: int(13),
																																				},
																																				End: ast.Location{
																																					Line: int(741),
																																					Column: int(24),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{
																																				ast.FodderElement{
																																					Kind: ast.FodderKind(0),
																																					Blanks: int(0),
																																					Indent: int(12),
																																					Comment: []string{},
																																				},
																																			},
																																			Ctx: p7834,
																																			FreeVars: ast.Identifiers{
																																				"code",
																																			},
																																		},
																																		Cond: &ast.Binary{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(738),
																																						Column: int(16),
																																					},
																																					End: ast.Location{
																																						Line: int(738),
																																						Column: int(33),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p7834,
																																				FreeVars: ast.Identifiers{
																																					"code",
																																				},
																																			},
																																			Left: &ast.Index{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(738),
																																							Column: int(16),
																																						},
																																						End: ast.Location{
																																							Line: int(738),
																																							Column: int(25),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p7834,
																																					FreeVars: ast.Identifiers{
																																						"code",
																																					},
																																				},
																																				Target: &ast.Var{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(738),
																																								Column: int(16),
																																							},
																																							End: ast.Location{
																																								Line: int(738),
																																								Column: int(20),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{
																																							"code",
																																						},
																																					},
																																					Id: "code",
																																				},
																																				LeftBracketFodder: ast.Fodder{},
																																				Index: &ast.LiteralString{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							End: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							File: nil,
																																						},
																																						Fodder: nil,
																																						Ctx: nil,
																																						FreeVars: nil,
																																					},
																																					Value: "mkey",
																																					Kind: ast.LiteralStringKind(1),
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																				},
																																				RightBracketFodder: ast.Fodder{},
																																				Id: nil,
																																			},
																																			OpFodder: ast.Fodder{},
																																			Op: ast.BinaryOp(12),
																																			Right: &ast.LiteralNull{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(738),
																																							Column: int(29),
																																						},
																																						End: ast.Location{
																																							Line: int(738),
																																							Column: int(33),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p7834,
																																					FreeVars: nil,
																																				},
																																			},
																																		},
																																		ThenFodder: ast.Fodder{},
																																		BranchTrue: &ast.Error{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(739),
																																						Column: int(15),
																																					},
																																					End: ast.Location{
																																						Line: int(739),
																																						Column: int(45),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{
																																					ast.FodderElement{
																																						Kind: ast.FodderKind(0),
																																						Blanks: int(0),
																																						Indent: int(14),
																																						Comment: []string{},
																																					},
																																				},
																																				Ctx: p7834,
																																				FreeVars: nil,
																																			},
																																			Expr: &ast.LiteralString{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(739),
																																							Column: int(21),
																																						},
																																						End: ast.Location{
																																							Line: int(739),
																																							Column: int(45),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p7834,
																																					FreeVars: nil,
																																				},
																																				Value: "Mapping keys required.",
																																				Kind: ast.LiteralStringKind(1),
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																			},
																																		},
																																		ElseFodder: ast.Fodder{
																																			ast.FodderElement{
																																				Kind: ast.FodderKind(0),
																																				Blanks: int(0),
																																				Indent: int(12),
																																				Comment: []string{},
																																			},
																																		},
																																		BranchFalse: &ast.Index{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(741),
																																						Column: int(15),
																																					},
																																					End: ast.Location{
																																						Line: int(741),
																																						Column: int(24),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p7834,
																																				FreeVars: ast.Identifiers{
																																					"code",
																																				},
																																			},
																																			Target: &ast.Var{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(741),
																																							Column: int(15),
																																						},
																																						End: ast.Location{
																																							Line: int(741),
																																							Column: int(19),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{
																																						ast.FodderElement{
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(14),
																																							Comment: []string{},
																																						},
																																					},
																																					Ctx: nil,
																																					FreeVars: ast.Identifiers{
																																						"code",
																																					},
																																				},
																																				Id: "code",
																																			},
																																			LeftBracketFodder: ast.Fodder{},
																																			Index: &ast.LiteralString{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "",
																																						Begin: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						End: ast.Location{
																																							Line: int(0),
																																							Column: int(0),
																																						},
																																						File: nil,
																																					},
																																					Fodder: nil,
																																					Ctx: nil,
																																					FreeVars: nil,
																																				},
																																				Value: "mkey",
																																				Kind: ast.LiteralStringKind(1),
																																				BlockIndent: "",
																																				BlockTermIndent: "",
																																			},
																																			RightBracketFodder: ast.Fodder{},
																																			Id: nil,
																																		},
																																	},
																																	Fun: nil,
																																	CloseFodder: ast.Fodder{},
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(737),
																																			Column: int(17),
																																		},
																																		End: ast.Location{
																																			Line: int(741),
																																			Column: int(24),
																																		},
																																		File: p1,
																																	},
																																},
																															},
																															Body: &ast.Local{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(742),
																																			Column: int(11),
																																		},
																																		End: ast.Location{
																																			Line: int(767),
																																			Column: int(60),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{
																																		ast.FodderElement{
																																			Kind: ast.FodderKind(0),
																																			Blanks: int(0),
																																			Indent: int(10),
																																			Comment: []string{},
																																		},
																																	},
																																	Ctx: p7756,
																																	FreeVars: ast.Identifiers{
																																		"code",
																																		"codes",
																																		"f",
																																		"format_code",
																																		"format_codes_obj",
																																		"i",
																																		"obj",
																																		"pad_left",
																																		"pad_right",
																																		"std",
																																		"v",
																																	},
																																},
																																Binds: ast.LocalBinds{
																																	ast.LocalBind{
																																		VarFodder: ast.Fodder{},
																																		Variable: "fw",
																																		EqFodder: ast.Fodder{},
																																		Body: &ast.Conditional{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(743),
																																						Column: int(13),
																																					},
																																					End: ast.Location{
																																						Line: int(746),
																																						Column: int(22),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{
																																					ast.FodderElement{
																																						Kind: ast.FodderKind(0),
																																						Blanks: int(0),
																																						Indent: int(12),
																																						Comment: []string{},
																																					},
																																				},
																																				Ctx: p7860,
																																				FreeVars: ast.Identifiers{
																																					"code",
																																				},
																																			},
																																			Cond: &ast.Binary{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(743),
																																							Column: int(16),
																																						},
																																						End: ast.Location{
																																							Line: int(743),
																																							Column: int(30),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p7860,
																																					FreeVars: ast.Identifiers{
																																						"code",
																																					},
																																				},
																																				Left: &ast.Index{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(743),
																																								Column: int(16),
																																							},
																																							End: ast.Location{
																																								Line: int(743),
																																								Column: int(23),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p7860,
																																						FreeVars: ast.Identifiers{
																																							"code",
																																						},
																																					},
																																					Target: &ast.Var{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(743),
																																									Column: int(16),
																																								},
																																								End: ast.Location{
																																									Line: int(743),
																																									Column: int(20),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{
																																								"code",
																																							},
																																						},
																																						Id: "code",
																																					},
																																					LeftBracketFodder: ast.Fodder{},
																																					Index: &ast.LiteralString{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								End: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								File: nil,
																																							},
																																							Fodder: nil,
																																							Ctx: nil,
																																							FreeVars: nil,
																																						},
																																						Value: "fw",
																																						Kind: ast.LiteralStringKind(1),
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																					},
																																					RightBracketFodder: ast.Fodder{},
																																					Id: nil,
																																				},
																																				OpFodder: ast.Fodder{},
																																				Op: ast.BinaryOp(12),
																																				Right: &ast.LiteralString{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(743),
																																								Column: int(27),
																																							},
																																							End: ast.Location{
																																								Line: int(743),
																																								Column: int(30),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p7860,
																																						FreeVars: nil,
																																					},
																																					Value: "*",
																																					Kind: ast.LiteralStringKind(1),
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																				},
																																			},
																																			ThenFodder: ast.Fodder{},
																																			BranchTrue: &ast.Error{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(744),
																																							Column: int(15),
																																						},
																																						End: ast.Location{
																																							Line: int(744),
																																							Column: int(60),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{
																																						ast.FodderElement{
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(14),
																																							Comment: []string{},
																																						},
																																					},
																																					Ctx: p7860,
																																					FreeVars: nil,
																																				},
																																				Expr: &ast.LiteralString{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(744),
																																								Column: int(21),
																																							},
																																							End: ast.Location{
																																								Line: int(744),
																																								Column: int(60),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p7860,
																																						FreeVars: nil,
																																					},
																																					Value: "Cannot use * field width with object.",
																																					Kind: ast.LiteralStringKind(1),
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																				},
																																			},
																																			ElseFodder: ast.Fodder{
																																				ast.FodderElement{
																																					Kind: ast.FodderKind(0),
																																					Blanks: int(0),
																																					Indent: int(12),
																																					Comment: []string{},
																																				},
																																			},
																																			BranchFalse: &ast.Index{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(746),
																																							Column: int(15),
																																						},
																																						End: ast.Location{
																																							Line: int(746),
																																							Column: int(22),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p7860,
																																					FreeVars: ast.Identifiers{
																																						"code",
																																					},
																																				},
																																				Target: &ast.Var{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(746),
																																								Column: int(15),
																																							},
																																							End: ast.Location{
																																								Line: int(746),
																																								Column: int(19),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{
																																							ast.FodderElement{
																																								Kind: ast.FodderKind(0),
																																								Blanks: int(0),
																																								Indent: int(14),
																																								Comment: []string{},
																																							},
																																						},
																																						Ctx: nil,
																																						FreeVars: ast.Identifiers{
																																							"code",
																																						},
																																					},
																																					Id: "code",
																																				},
																																				LeftBracketFodder: ast.Fodder{},
																																				Index: &ast.LiteralString{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "",
																																							Begin: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							End: ast.Location{
																																								Line: int(0),
																																								Column: int(0),
																																							},
																																							File: nil,
																																						},
																																						Fodder: nil,
																																						Ctx: nil,
																																						FreeVars: nil,
																																					},
																																					Value: "fw",
																																					Kind: ast.LiteralStringKind(1),
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																				},
																																				RightBracketFodder: ast.Fodder{},
																																				Id: nil,
																																			},
																																		},
																																		Fun: nil,
																																		CloseFodder: ast.Fodder{},
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(742),
																																				Column: int(17),
																																			},
																																			End: ast.Location{
																																				Line: int(746),
																																				Column: int(22),
																																			},
																																			File: p1,
																																		},
																																	},
																																},
																																Body: &ast.Local{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(747),
																																				Column: int(11),
																																			},
																																			End: ast.Location{
																																				Line: int(767),
																																				Column: int(60),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{
																																			ast.FodderElement{
																																				Kind: ast.FodderKind(0),
																																				Blanks: int(0),
																																				Indent: int(10),
																																				Comment: []string{},
																																			},
																																		},
																																		Ctx: p7756,
																																		FreeVars: ast.Identifiers{
																																			"code",
																																			"codes",
																																			"f",
																																			"format_code",
																																			"format_codes_obj",
																																			"fw",
																																			"i",
																																			"obj",
																																			"pad_left",
																																			"pad_right",
																																			"std",
																																			"v",
																																		},
																																	},
																																	Binds: ast.LocalBinds{
																																		ast.LocalBind{
																																			VarFodder: ast.Fodder{},
																																			Variable: "prec",
																																			EqFodder: ast.Fodder{},
																																			Body: &ast.Conditional{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(748),
																																							Column: int(13),
																																						},
																																						End: ast.Location{
																																							Line: int(751),
																																							Column: int(24),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{
																																						ast.FodderElement{
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(12),
																																							Comment: []string{},
																																						},
																																					},
																																					Ctx: p7886,
																																					FreeVars: ast.Identifiers{
																																						"code",
																																					},
																																				},
																																				Cond: &ast.Binary{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(748),
																																								Column: int(16),
																																							},
																																							End: ast.Location{
																																								Line: int(748),
																																								Column: int(32),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p7886,
																																						FreeVars: ast.Identifiers{
																																							"code",
																																						},
																																					},
																																					Left: &ast.Index{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(748),
																																									Column: int(16),
																																								},
																																								End: ast.Location{
																																									Line: int(748),
																																									Column: int(25),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p7886,
																																							FreeVars: ast.Identifiers{
																																								"code",
																																							},
																																						},
																																						Target: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(748),
																																										Column: int(16),
																																									},
																																									End: ast.Location{
																																										Line: int(748),
																																										Column: int(20),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: nil,
																																								FreeVars: ast.Identifiers{
																																									"code",
																																								},
																																							},
																																							Id: "code",
																																						},
																																						LeftBracketFodder: ast.Fodder{},
																																						Index: &ast.LiteralString{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "",
																																									Begin: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																									End: ast.Location{
																																										Line: int(0),
																																										Column: int(0),
																																									},
																																									File: nil,
																																								},
																																								Fodder: nil,
																																								Ctx: nil,
																																								FreeVars: nil,
																																							},
																																							Value: "prec",
																																							Kind: ast.LiteralStringKind(1),
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																						},
																																						RightBracketFodder: ast.Fodder{},
																																						Id: nil,
																																					},
																																					OpFodder: ast.Fodder{},
																																					Op: ast.BinaryOp(12),
																																					Right: &ast.LiteralString{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(748),
																																									Column: int(29),
																																								},
																																								End: ast.Location{
																																									Line: int(748),
																																									Column: int(32),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p7886,
																																							FreeVars: nil,
																																						},
																																						Value: "*",
																																						Kind: ast.LiteralStringKind(1),
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																					},
																																				},
																																				ThenFodder: ast.Fodder{},
																																				BranchTrue: &ast.Error{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(749),
																																								Column: int(15),
																																							},
																																							End: ast.Location{
																																								Line: int(749),
																																								Column: int(58),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{
																																							ast.FodderElement{
																																								Kind: ast.FodderKind(0),
																																								Blanks: int(0),
																																								Indent: int(14),
																																								Comment: []string{},
																																							},
																																						},
																																						Ctx: p7886,
																																						FreeVars: nil,
																																					},
																																					Expr: &ast.LiteralString{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(749),
																																									Column: int(21),
																																								},
																																								End: ast.Location{
																																									Line: int(749),
																																									Column: int(58),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p7886,
																																							FreeVars: nil,
																																						},
																																						Value: "Cannot use * precision with object.",
																																						Kind: ast.LiteralStringKind(1),
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																					},
																																				},
																																				ElseFodder: ast.Fodder{
																																					ast.FodderElement{
																																						Kind: ast.FodderKind(0),
																																						Blanks: int(0),
																																						Indent: int(12),
																																						Comment: []string{},
																																					},
																																				},
																																				BranchFalse: &ast.Index{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(751),
																																								Column: int(15),
																																							},
																																							End: ast.Location{
																																								Line: int(751),
																																								Column: int(24),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p7886,
																																						FreeVars: ast.Identifiers{
																																							"code",
																																						},
																																					},
																																					Target: &ast.Var{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(751),
																																									Column: int(15),
																																								},
																																								End: ast.Location{
																																									Line: int(751),
																																									Column: int(19),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(14),
																																									Comment: []string{},
																																								},
																																							},
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{
																																								"code",
																																							},
																																						},
																																						Id: "code",
																																					},
																																					LeftBracketFodder: ast.Fodder{},
																																					Index: &ast.LiteralString{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								End: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								File: nil,
																																							},
																																							Fodder: nil,
																																							Ctx: nil,
																																							FreeVars: nil,
																																						},
																																						Value: "prec",
																																						Kind: ast.LiteralStringKind(1),
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																					},
																																					RightBracketFodder: ast.Fodder{},
																																					Id: nil,
																																				},
																																			},
																																			Fun: nil,
																																			CloseFodder: ast.Fodder{},
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(747),
																																					Column: int(17),
																																				},
																																				End: ast.Location{
																																					Line: int(751),
																																					Column: int(24),
																																				},
																																				File: p1,
																																			},
																																		},
																																	},
																																	Body: &ast.Local{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(752),
																																					Column: int(11),
																																				},
																																				End: ast.Location{
																																					Line: int(767),
																																					Column: int(60),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{
																																				ast.FodderElement{
																																					Kind: ast.FodderKind(0),
																																					Blanks: int(0),
																																					Indent: int(10),
																																					Comment: []string{},
																																				},
																																			},
																																			Ctx: p7756,
																																			FreeVars: ast.Identifiers{
																																				"code",
																																				"codes",
																																				"f",
																																				"format_code",
																																				"format_codes_obj",
																																				"fw",
																																				"i",
																																				"obj",
																																				"pad_left",
																																				"pad_right",
																																				"prec",
																																				"std",
																																				"v",
																																			},
																																		},
																																		Binds: ast.LocalBinds{
																																			ast.LocalBind{
																																				VarFodder: ast.Fodder{},
																																				Variable: "val",
																																				EqFodder: ast.Fodder{},
																																				Body: &ast.Conditional{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(753),
																																								Column: int(13),
																																							},
																																							End: ast.Location{
																																								Line: int(756),
																																								Column: int(42),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{
																																							ast.FodderElement{
																																								Kind: ast.FodderKind(0),
																																								Blanks: int(0),
																																								Indent: int(12),
																																								Comment: []string{},
																																							},
																																						},
																																						Ctx: p7912,
																																						FreeVars: ast.Identifiers{
																																							"f",
																																							"obj",
																																							"std",
																																						},
																																					},
																																					Cond: &ast.Apply{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(753),
																																									Column: int(16),
																																								},
																																								End: ast.Location{
																																									Line: int(753),
																																									Column: int(40),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p7912,
																																							FreeVars: ast.Identifiers{
																																								"f",
																																								"obj",
																																								"std",
																																							},
																																						},
																																						Target: &ast.Index{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(753),
																																										Column: int(16),
																																									},
																																									End: ast.Location{
																																										Line: int(753),
																																										Column: int(32),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p7912,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																								},
																																							},
																																							Target: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(753),
																																											Column: int(16),
																																										},
																																										End: ast.Location{
																																											Line: int(753),
																																											Column: int(19),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																									},
																																								},
																																								Id: "std",
																																							},
																																							LeftBracketFodder: ast.Fodder{},
																																							Index: &ast.LiteralString{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										File: nil,
																																									},
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: nil,
																																								},
																																								Value: "objectHasAll",
																																								Kind: ast.LiteralStringKind(1),
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							Id: nil,
																																						},
																																						FodderLeft: ast.Fodder{},
																																						Arguments: ast.Arguments{
																																							Positional: []ast.CommaSeparatedExpr{
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(753),
																																													Column: int(33),
																																												},
																																												End: ast.Location{
																																													Line: int(753),
																																													Column: int(36),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p7923,
																																											FreeVars: ast.Identifiers{
																																												"obj",
																																											},
																																										},
																																										Id: "obj",
																																									},
																																									CommaFodder: ast.Fodder{},
																																								},
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(753),
																																													Column: int(38),
																																												},
																																												End: ast.Location{
																																													Line: int(753),
																																													Column: int(39),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p7923,
																																											FreeVars: ast.Identifiers{
																																												"f",
																																											},
																																										},
																																										Id: "f",
																																									},
																																									CommaFodder: nil,
																																								},
																																							},
																																							Named: nil,
																																						},
																																						TrailingComma: false,
																																						TailStrict: false,
																																						FodderRight: ast.Fodder{},
																																						TailStrictFodder: nil,
																																					},
																																					ThenFodder: ast.Fodder{},
																																					BranchTrue: &ast.Index{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(754),
																																									Column: int(15),
																																								},
																																								End: ast.Location{
																																									Line: int(754),
																																									Column: int(21),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p7912,
																																							FreeVars: ast.Identifiers{
																																								"f",
																																								"obj",
																																							},
																																						},
																																						Target: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(754),
																																										Column: int(15),
																																									},
																																									End: ast.Location{
																																										Line: int(754),
																																										Column: int(18),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{
																																									ast.FodderElement{
																																										Kind: ast.FodderKind(0),
																																										Blanks: int(0),
																																										Indent: int(14),
																																										Comment: []string{},
																																									},
																																								},
																																								Ctx: p7912,
																																								FreeVars: ast.Identifiers{
																																									"obj",
																																								},
																																							},
																																							Id: "obj",
																																						},
																																						LeftBracketFodder: ast.Fodder{},
																																						Index: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(754),
																																										Column: int(19),
																																									},
																																									End: ast.Location{
																																										Line: int(754),
																																										Column: int(20),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p7912,
																																								FreeVars: ast.Identifiers{
																																									"f",
																																								},
																																							},
																																							Id: "f",
																																						},
																																						RightBracketFodder: ast.Fodder{},
																																						Id: nil,
																																					},
																																					ElseFodder: ast.Fodder{
																																						ast.FodderElement{
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(12),
																																							Comment: []string{},
																																						},
																																					},
																																					BranchFalse: &ast.Error{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(756),
																																									Column: int(15),
																																								},
																																								End: ast.Location{
																																									Line: int(756),
																																									Column: int(42),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(14),
																																									Comment: []string{},
																																								},
																																							},
																																							Ctx: p7912,
																																							FreeVars: ast.Identifiers{
																																								"f",
																																							},
																																						},
																																						Expr: &ast.Binary{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(756),
																																										Column: int(21),
																																									},
																																									End: ast.Location{
																																										Line: int(756),
																																										Column: int(42),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p7912,
																																								FreeVars: ast.Identifiers{
																																									"f",
																																								},
																																							},
																																							Left: &ast.LiteralString{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(756),
																																											Column: int(21),
																																										},
																																										End: ast.Location{
																																											Line: int(756),
																																											Column: int(38),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p7912,
																																									FreeVars: nil,
																																								},
																																								Value: "No such field: ",
																																								Kind: ast.LiteralStringKind(1),
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																							},
																																							OpFodder: ast.Fodder{},
																																							Op: ast.BinaryOp(3),
																																							Right: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(756),
																																											Column: int(41),
																																										},
																																										End: ast.Location{
																																											Line: int(756),
																																											Column: int(42),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p7912,
																																									FreeVars: ast.Identifiers{
																																										"f",
																																									},
																																								},
																																								Id: "f",
																																							},
																																						},
																																					},
																																				},
																																				Fun: nil,
																																				CloseFodder: ast.Fodder{},
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(752),
																																						Column: int(17),
																																					},
																																					End: ast.Location{
																																						Line: int(756),
																																						Column: int(42),
																																					},
																																					File: p1,
																																				},
																																			},
																																		},
																																		Body: &ast.Local{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(757),
																																						Column: int(11),
																																					},
																																					End: ast.Location{
																																						Line: int(767),
																																						Column: int(60),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{
																																					ast.FodderElement{
																																						Kind: ast.FodderKind(0),
																																						Blanks: int(0),
																																						Indent: int(10),
																																						Comment: []string{},
																																					},
																																				},
																																				Ctx: p7756,
																																				FreeVars: ast.Identifiers{
																																					"code",
																																					"codes",
																																					"f",
																																					"format_code",
																																					"format_codes_obj",
																																					"fw",
																																					"i",
																																					"obj",
																																					"pad_left",
																																					"pad_right",
																																					"prec",
																																					"v",
																																					"val",
																																				},
																																			},
																																			Binds: ast.LocalBinds{
																																				ast.LocalBind{
																																					VarFodder: ast.Fodder{},
																																					Variable: "s",
																																					EqFodder: ast.Fodder{},
																																					Body: &ast.Conditional{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(758),
																																									Column: int(13),
																																								},
																																								End: ast.Location{
																																									Line: int(761),
																																									Column: int(50),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(12),
																																									Comment: []string{},
																																								},
																																							},
																																							Ctx: p7949,
																																							FreeVars: ast.Identifiers{
																																								"code",
																																								"f",
																																								"format_code",
																																								"fw",
																																								"prec",
																																								"val",
																																							},
																																						},
																																						Cond: &ast.Binary{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(758),
																																										Column: int(16),
																																									},
																																									End: ast.Location{
																																										Line: int(758),
																																										Column: int(33),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p7949,
																																								FreeVars: ast.Identifiers{
																																									"code",
																																								},
																																							},
																																							Left: &ast.Index{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(758),
																																											Column: int(16),
																																										},
																																										End: ast.Location{
																																											Line: int(758),
																																											Column: int(26),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p7949,
																																									FreeVars: ast.Identifiers{
																																										"code",
																																									},
																																								},
																																								Target: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(758),
																																												Column: int(16),
																																											},
																																											End: ast.Location{
																																												Line: int(758),
																																												Column: int(20),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"code",
																																										},
																																									},
																																									Id: "code",
																																								},
																																								LeftBracketFodder: ast.Fodder{},
																																								Index: &ast.LiteralString{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											File: nil,
																																										},
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: nil,
																																									},
																																									Value: "ctype",
																																									Kind: ast.LiteralStringKind(1),
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								Id: nil,
																																							},
																																							OpFodder: ast.Fodder{},
																																							Op: ast.BinaryOp(12),
																																							Right: &ast.LiteralString{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(758),
																																											Column: int(30),
																																										},
																																										End: ast.Location{
																																											Line: int(758),
																																											Column: int(33),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p7949,
																																									FreeVars: nil,
																																								},
																																								Value: "%",
																																								Kind: ast.LiteralStringKind(1),
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																							},
																																						},
																																						ThenFodder: ast.Fodder{},
																																						BranchTrue: &ast.LiteralString{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(759),
																																										Column: int(15),
																																									},
																																									End: ast.Location{
																																										Line: int(759),
																																										Column: int(18),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{
																																									ast.FodderElement{
																																										Kind: ast.FodderKind(0),
																																										Blanks: int(0),
																																										Indent: int(14),
																																										Comment: []string{},
																																									},
																																								},
																																								Ctx: p7949,
																																								FreeVars: nil,
																																							},
																																							Value: "%",
																																							Kind: ast.LiteralStringKind(1),
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																						},
																																						ElseFodder: ast.Fodder{
																																							ast.FodderElement{
																																								Kind: ast.FodderKind(0),
																																								Blanks: int(0),
																																								Indent: int(12),
																																								Comment: []string{},
																																							},
																																						},
																																						BranchFalse: &ast.Apply{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(761),
																																										Column: int(15),
																																									},
																																									End: ast.Location{
																																										Line: int(761),
																																										Column: int(50),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p7949,
																																								FreeVars: ast.Identifiers{
																																									"code",
																																									"f",
																																									"format_code",
																																									"fw",
																																									"prec",
																																									"val",
																																								},
																																							},
																																							Target: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(761),
																																											Column: int(15),
																																										},
																																										End: ast.Location{
																																											Line: int(761),
																																											Column: int(26),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(14),
																																											Comment: []string{},
																																										},
																																									},
																																									Ctx: p7949,
																																									FreeVars: ast.Identifiers{
																																										"format_code",
																																									},
																																								},
																																								Id: "format_code",
																																							},
																																							FodderLeft: ast.Fodder{},
																																							Arguments: ast.Arguments{
																																								Positional: []ast.CommaSeparatedExpr{
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(761),
																																														Column: int(27),
																																													},
																																													End: ast.Location{
																																														Line: int(761),
																																														Column: int(30),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p7969,
																																												FreeVars: ast.Identifiers{
																																													"val",
																																												},
																																											},
																																											Id: "val",
																																										},
																																										CommaFodder: ast.Fodder{},
																																									},
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(761),
																																														Column: int(32),
																																													},
																																													End: ast.Location{
																																														Line: int(761),
																																														Column: int(36),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p7969,
																																												FreeVars: ast.Identifiers{
																																													"code",
																																												},
																																											},
																																											Id: "code",
																																										},
																																										CommaFodder: ast.Fodder{},
																																									},
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(761),
																																														Column: int(38),
																																													},
																																													End: ast.Location{
																																														Line: int(761),
																																														Column: int(40),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p7969,
																																												FreeVars: ast.Identifiers{
																																													"fw",
																																												},
																																											},
																																											Id: "fw",
																																										},
																																										CommaFodder: ast.Fodder{},
																																									},
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(761),
																																														Column: int(42),
																																													},
																																													End: ast.Location{
																																														Line: int(761),
																																														Column: int(46),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p7969,
																																												FreeVars: ast.Identifiers{
																																													"prec",
																																												},
																																											},
																																											Id: "prec",
																																										},
																																										CommaFodder: ast.Fodder{},
																																									},
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(761),
																																														Column: int(48),
																																													},
																																													End: ast.Location{
																																														Line: int(761),
																																														Column: int(49),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p7969,
																																												FreeVars: ast.Identifiers{
																																													"f",
																																												},
																																											},
																																											Id: "f",
																																										},
																																										CommaFodder: nil,
																																									},
																																								},
																																								Named: nil,
																																							},
																																							TrailingComma: false,
																																							TailStrict: false,
																																							FodderRight: ast.Fodder{},
																																							TailStrictFodder: nil,
																																						},
																																					},
																																					Fun: nil,
																																					CloseFodder: ast.Fodder{},
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(757),
																																							Column: int(17),
																																						},
																																						End: ast.Location{
																																							Line: int(761),
																																							Column: int(50),
																																						},
																																						File: p1,
																																					},
																																				},
																																			},
																																			Body: &ast.Local{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(762),
																																							Column: int(11),
																																						},
																																						End: ast.Location{
																																							Line: int(767),
																																							Column: int(60),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{
																																						ast.FodderElement{
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(10),
																																							Comment: []string{},
																																						},
																																					},
																																					Ctx: p7756,
																																					FreeVars: ast.Identifiers{
																																						"code",
																																						"codes",
																																						"format_codes_obj",
																																						"fw",
																																						"i",
																																						"obj",
																																						"pad_left",
																																						"pad_right",
																																						"s",
																																						"v",
																																					},
																																				},
																																				Binds: ast.LocalBinds{
																																					ast.LocalBind{
																																						VarFodder: ast.Fodder{},
																																						Variable: "s_padded",
																																						EqFodder: ast.Fodder{},
																																						Body: &ast.Conditional{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(763),
																																										Column: int(13),
																																									},
																																									End: ast.Location{
																																										Line: int(766),
																																										Column: int(35),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{
																																									ast.FodderElement{
																																										Kind: ast.FodderKind(0),
																																										Blanks: int(0),
																																										Indent: int(12),
																																										Comment: []string{},
																																									},
																																								},
																																								Ctx: p7985,
																																								FreeVars: ast.Identifiers{
																																									"code",
																																									"fw",
																																									"pad_left",
																																									"pad_right",
																																									"s",
																																								},
																																							},
																																							Cond: &ast.Index{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(763),
																																											Column: int(16),
																																										},
																																										End: ast.Location{
																																											Line: int(763),
																																											Column: int(32),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p7985,
																																									FreeVars: ast.Identifiers{
																																										"code",
																																									},
																																								},
																																								Target: &ast.Index{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(763),
																																												Column: int(16),
																																											},
																																											End: ast.Location{
																																												Line: int(763),
																																												Column: int(27),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"code",
																																										},
																																									},
																																									Target: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(763),
																																													Column: int(16),
																																												},
																																												End: ast.Location{
																																													Line: int(763),
																																													Column: int(20),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{
																																												"code",
																																											},
																																										},
																																										Id: "code",
																																									},
																																									LeftBracketFodder: ast.Fodder{},
																																									Index: &ast.LiteralString{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												End: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												File: nil,
																																											},
																																											Fodder: nil,
																																											Ctx: nil,
																																											FreeVars: nil,
																																										},
																																										Value: "cflags",
																																										Kind: ast.LiteralStringKind(1),
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																									},
																																									RightBracketFodder: ast.Fodder{},
																																									Id: nil,
																																								},
																																								LeftBracketFodder: ast.Fodder{},
																																								Index: &ast.LiteralString{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											File: nil,
																																										},
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: nil,
																																									},
																																									Value: "left",
																																									Kind: ast.LiteralStringKind(1),
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								Id: nil,
																																							},
																																							ThenFodder: ast.Fodder{},
																																							BranchTrue: &ast.Apply{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(764),
																																											Column: int(15),
																																										},
																																										End: ast.Location{
																																											Line: int(764),
																																											Column: int(36),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p7985,
																																									FreeVars: ast.Identifiers{
																																										"fw",
																																										"pad_right",
																																										"s",
																																									},
																																								},
																																								Target: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(764),
																																												Column: int(15),
																																											},
																																											End: ast.Location{
																																												Line: int(764),
																																												Column: int(24),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{
																																											ast.FodderElement{
																																												Kind: ast.FodderKind(0),
																																												Blanks: int(0),
																																												Indent: int(14),
																																												Comment: []string{},
																																											},
																																										},
																																										Ctx: p7985,
																																										FreeVars: ast.Identifiers{
																																											"pad_right",
																																										},
																																									},
																																									Id: "pad_right",
																																								},
																																								FodderLeft: ast.Fodder{},
																																								Arguments: ast.Arguments{
																																									Positional: []ast.CommaSeparatedExpr{
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Var{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(764),
																																															Column: int(25),
																																														},
																																														End: ast.Location{
																																															Line: int(764),
																																															Column: int(26),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: p8002,
																																													FreeVars: ast.Identifiers{
																																														"s",
																																													},
																																												},
																																												Id: "s",
																																											},
																																											CommaFodder: ast.Fodder{},
																																										},
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Var{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(764),
																																															Column: int(28),
																																														},
																																														End: ast.Location{
																																															Line: int(764),
																																															Column: int(30),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: p8002,
																																													FreeVars: ast.Identifiers{
																																														"fw",
																																													},
																																												},
																																												Id: "fw",
																																											},
																																											CommaFodder: ast.Fodder{},
																																										},
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.LiteralString{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(764),
																																															Column: int(32),
																																														},
																																														End: ast.Location{
																																															Line: int(764),
																																															Column: int(35),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: p8002,
																																													FreeVars: nil,
																																												},
																																												Value: " ",
																																												Kind: ast.LiteralStringKind(1),
																																												BlockIndent: "",
																																												BlockTermIndent: "",
																																											},
																																											CommaFodder: nil,
																																										},
																																									},
																																									Named: nil,
																																								},
																																								TrailingComma: false,
																																								TailStrict: false,
																																								FodderRight: ast.Fodder{},
																																								TailStrictFodder: nil,
																																							},
																																							ElseFodder: ast.Fodder{
																																								ast.FodderElement{
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(12),
																																									Comment: []string{},
																																								},
																																							},
																																							BranchFalse: &ast.Apply{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(766),
																																											Column: int(15),
																																										},
																																										End: ast.Location{
																																											Line: int(766),
																																											Column: int(35),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p7985,
																																									FreeVars: ast.Identifiers{
																																										"fw",
																																										"pad_left",
																																										"s",
																																									},
																																								},
																																								Target: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(766),
																																												Column: int(15),
																																											},
																																											End: ast.Location{
																																												Line: int(766),
																																												Column: int(23),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{
																																											ast.FodderElement{
																																												Kind: ast.FodderKind(0),
																																												Blanks: int(0),
																																												Indent: int(14),
																																												Comment: []string{},
																																											},
																																										},
																																										Ctx: p7985,
																																										FreeVars: ast.Identifiers{
																																											"pad_left",
																																										},
																																									},
																																									Id: "pad_left",
																																								},
																																								FodderLeft: ast.Fodder{},
																																								Arguments: ast.Arguments{
																																									Positional: []ast.CommaSeparatedExpr{
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Var{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(766),
																																															Column: int(24),
																																														},
																																														End: ast.Location{
																																															Line: int(766),
																																															Column: int(25),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: p8015,
																																													FreeVars: ast.Identifiers{
																																														"s",
																																													},
																																												},
																																												Id: "s",
																																											},
																																											CommaFodder: ast.Fodder{},
																																										},
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Var{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(766),
																																															Column: int(27),
																																														},
																																														End: ast.Location{
																																															Line: int(766),
																																															Column: int(29),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: p8015,
																																													FreeVars: ast.Identifiers{
																																														"fw",
																																													},
																																												},
																																												Id: "fw",
																																											},
																																											CommaFodder: ast.Fodder{},
																																										},
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.LiteralString{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(766),
																																															Column: int(31),
																																														},
																																														End: ast.Location{
																																															Line: int(766),
																																															Column: int(34),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: p8015,
																																													FreeVars: nil,
																																												},
																																												Value: " ",
																																												Kind: ast.LiteralStringKind(1),
																																												BlockIndent: "",
																																												BlockTermIndent: "",
																																											},
																																											CommaFodder: nil,
																																										},
																																									},
																																									Named: nil,
																																								},
																																								TrailingComma: false,
																																								TailStrict: false,
																																								FodderRight: ast.Fodder{},
																																								TailStrictFodder: nil,
																																							},
																																						},
																																						Fun: nil,
																																						CloseFodder: ast.Fodder{},
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(762),
																																								Column: int(17),
																																							},
																																							End: ast.Location{
																																								Line: int(766),
																																								Column: int(35),
																																							},
																																							File: p1,
																																						},
																																					},
																																				},
																																				Body: &ast.Apply{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(767),
																																								Column: int(11),
																																							},
																																							End: ast.Location{
																																								Line: int(767),
																																								Column: int(60),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p7756,
																																						FreeVars: ast.Identifiers{
																																							"codes",
																																							"format_codes_obj",
																																							"i",
																																							"obj",
																																							"s_padded",
																																							"v",
																																						},
																																					},
																																					Target: &ast.Var{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(767),
																																									Column: int(11),
																																								},
																																								End: ast.Location{
																																									Line: int(767),
																																									Column: int(27),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{
																																								ast.FodderElement{
																																									Kind: ast.FodderKind(0),
																																									Blanks: int(0),
																																									Indent: int(10),
																																									Comment: []string{},
																																								},
																																							},
																																							Ctx: p7756,
																																							FreeVars: ast.Identifiers{
																																								"format_codes_obj",
																																							},
																																						},
																																						Id: "format_codes_obj",
																																					},
																																					FodderLeft: ast.Fodder{},
																																					Arguments: ast.Arguments{
																																						Positional: []ast.CommaSeparatedExpr{
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(767),
																																												Column: int(28),
																																											},
																																											End: ast.Location{
																																												Line: int(767),
																																												Column: int(33),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p8027,
																																										FreeVars: ast.Identifiers{
																																											"codes",
																																										},
																																									},
																																									Id: "codes",
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(767),
																																												Column: int(35),
																																											},
																																											End: ast.Location{
																																												Line: int(767),
																																												Column: int(38),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p8027,
																																										FreeVars: ast.Identifiers{
																																											"obj",
																																										},
																																									},
																																									Id: "obj",
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Binary{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(767),
																																												Column: int(40),
																																											},
																																											End: ast.Location{
																																												Line: int(767),
																																												Column: int(45),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p8027,
																																										FreeVars: ast.Identifiers{
																																											"i",
																																										},
																																									},
																																									Left: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(767),
																																													Column: int(40),
																																												},
																																												End: ast.Location{
																																													Line: int(767),
																																													Column: int(41),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p8027,
																																											FreeVars: ast.Identifiers{
																																												"i",
																																											},
																																										},
																																										Id: "i",
																																									},
																																									OpFodder: ast.Fodder{},
																																									Op: ast.BinaryOp(3),
																																									Right: &ast.LiteralNumber{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(767),
																																													Column: int(44),
																																												},
																																												End: ast.Location{
																																													Line: int(767),
																																													Column: int(45),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p8027,
																																											FreeVars: nil,
																																										},
																																										OriginalString: "1",
																																									},
																																								},
																																								CommaFodder: ast.Fodder{},
																																							},
																																							ast.CommaSeparatedExpr{
																																								Expr: &ast.Binary{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(767),
																																												Column: int(47),
																																											},
																																											End: ast.Location{
																																												Line: int(767),
																																												Column: int(59),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p8027,
																																										FreeVars: ast.Identifiers{
																																											"s_padded",
																																											"v",
																																										},
																																									},
																																									Left: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(767),
																																													Column: int(47),
																																												},
																																												End: ast.Location{
																																													Line: int(767),
																																													Column: int(48),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p8027,
																																											FreeVars: ast.Identifiers{
																																												"v",
																																											},
																																										},
																																										Id: "v",
																																									},
																																									OpFodder: ast.Fodder{},
																																									Op: ast.BinaryOp(3),
																																									Right: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(767),
																																													Column: int(51),
																																												},
																																												End: ast.Location{
																																													Line: int(767),
																																													Column: int(59),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p8027,
																																											FreeVars: ast.Identifiers{
																																												"s_padded",
																																											},
																																										},
																																										Id: "s_padded",
																																									},
																																								},
																																								CommaFodder: nil,
																																							},
																																						},
																																						Named: nil,
																																					},
																																					TrailingComma: false,
																																					TailStrict: true,
																																					FodderRight: ast.Fodder{},
																																					TailStrictFodder: ast.Fodder{},
																																				},
																																			},
																																		},
																																	},
																																},
																															},
																														},
																													},
																												},
																											},
																										},
																										Fun: nil,
																										CloseFodder: nil,
																										LocRange: ast.LocationRange{
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											File: nil,
																										},
																									},
																								},
																								Body: &ast.Conditional{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(769),
																												Column: int(5),
																											},
																											End: ast.Location{
																												Line: int(774),
																												Column: int(48),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{
																											ast.FodderElement{
																												Kind: ast.FodderKind(0),
																												Blanks: int(1),
																												Indent: int(4),
																												Comment: []string{},
																											},
																										},
																										Ctx: p3497,
																										FreeVars: ast.Identifiers{
																											"codes",
																											"format_codes_arr",
																											"format_codes_obj",
																											"std",
																											"vals",
																										},
																									},
																									Cond: &ast.Apply{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(769),
																													Column: int(8),
																												},
																												End: ast.Location{
																													Line: int(769),
																													Column: int(25),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p3497,
																											FreeVars: ast.Identifiers{
																												"std",
																												"vals",
																											},
																										},
																										Target: &ast.Index{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(769),
																														Column: int(8),
																													},
																													End: ast.Location{
																														Line: int(769),
																														Column: int(19),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p3497,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																											},
																											Target: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(769),
																															Column: int(8),
																														},
																														End: ast.Location{
																															Line: int(769),
																															Column: int(11),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																												},
																												Id: "std",
																											},
																											LeftBracketFodder: ast.Fodder{},
																											Index: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: nil,
																												},
																												Value: "isArray",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																											RightBracketFodder: ast.Fodder{},
																											Id: nil,
																										},
																										FodderLeft: ast.Fodder{},
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(769),
																																	Column: int(20),
																																},
																																End: ast.Location{
																																	Line: int(769),
																																	Column: int(24),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p8054,
																															FreeVars: ast.Identifiers{
																																"vals",
																															},
																														},
																														Id: "vals",
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										TrailingComma: false,
																										TailStrict: false,
																										FodderRight: ast.Fodder{},
																										TailStrictFodder: nil,
																									},
																									ThenFodder: ast.Fodder{},
																									BranchTrue: &ast.Apply{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(770),
																													Column: int(7),
																												},
																												End: ast.Location{
																													Line: int(770),
																													Column: int(46),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p3497,
																											FreeVars: ast.Identifiers{
																												"codes",
																												"format_codes_arr",
																												"vals",
																											},
																										},
																										Target: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(770),
																														Column: int(7),
																													},
																													End: ast.Location{
																														Line: int(770),
																														Column: int(23),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{
																													ast.FodderElement{
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(6),
																														Comment: []string{},
																													},
																												},
																												Ctx: p3497,
																												FreeVars: ast.Identifiers{
																													"format_codes_arr",
																												},
																											},
																											Id: "format_codes_arr",
																										},
																										FodderLeft: ast.Fodder{},
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(770),
																																	Column: int(24),
																																},
																																End: ast.Location{
																																	Line: int(770),
																																	Column: int(29),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p8063,
																															FreeVars: ast.Identifiers{
																																"codes",
																															},
																														},
																														Id: "codes",
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(770),
																																	Column: int(31),
																																},
																																End: ast.Location{
																																	Line: int(770),
																																	Column: int(35),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p8063,
																															FreeVars: ast.Identifiers{
																																"vals",
																															},
																														},
																														Id: "vals",
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralNumber{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(770),
																																	Column: int(37),
																																},
																																End: ast.Location{
																																	Line: int(770),
																																	Column: int(38),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p8063,
																															FreeVars: nil,
																														},
																														OriginalString: "0",
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralNumber{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(770),
																																	Column: int(40),
																																},
																																End: ast.Location{
																																	Line: int(770),
																																	Column: int(41),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p8063,
																															FreeVars: nil,
																														},
																														OriginalString: "0",
																													},
																													CommaFodder: ast.Fodder{},
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(770),
																																	Column: int(43),
																																},
																																End: ast.Location{
																																	Line: int(770),
																																	Column: int(45),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p8063,
																															FreeVars: nil,
																														},
																														Value: "",
																														Kind: ast.LiteralStringKind(1),
																														BlockIndent: "",
																														BlockTermIndent: "",
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										TrailingComma: false,
																										TailStrict: false,
																										FodderRight: ast.Fodder{},
																										TailStrictFodder: nil,
																									},
																									ElseFodder: ast.Fodder{
																										ast.FodderElement{
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(4),
																											Comment: []string{},
																										},
																									},
																									BranchFalse: &ast.Conditional{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(771),
																													Column: int(10),
																												},
																												End: ast.Location{
																													Line: int(774),
																													Column: int(48),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p3497,
																											FreeVars: ast.Identifiers{
																												"codes",
																												"format_codes_arr",
																												"format_codes_obj",
																												"std",
																												"vals",
																											},
																										},
																										Cond: &ast.Apply{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(771),
																														Column: int(13),
																													},
																													End: ast.Location{
																														Line: int(771),
																														Column: int(31),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p3497,
																												FreeVars: ast.Identifiers{
																													"std",
																													"vals",
																												},
																											},
																											Target: &ast.Index{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(771),
																															Column: int(13),
																														},
																														End: ast.Location{
																															Line: int(771),
																															Column: int(25),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p3497,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																												},
																												Target: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(771),
																																Column: int(13),
																															},
																															End: ast.Location{
																																Line: int(771),
																																Column: int(16),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"std",
																														},
																													},
																													Id: "std",
																												},
																												LeftBracketFodder: ast.Fodder{},
																												Index: &ast.LiteralString{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															File: nil,
																														},
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: nil,
																													},
																													Value: "isObject",
																													Kind: ast.LiteralStringKind(1),
																													BlockIndent: "",
																													BlockTermIndent: "",
																												},
																												RightBracketFodder: ast.Fodder{},
																												Id: nil,
																											},
																											FodderLeft: ast.Fodder{},
																											Arguments: ast.Arguments{
																												Positional: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(771),
																																		Column: int(26),
																																	},
																																	End: ast.Location{
																																		Line: int(771),
																																		Column: int(30),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p8082,
																																FreeVars: ast.Identifiers{
																																	"vals",
																																},
																															},
																															Id: "vals",
																														},
																														CommaFodder: nil,
																													},
																												},
																												Named: nil,
																											},
																											TrailingComma: false,
																											TailStrict: false,
																											FodderRight: ast.Fodder{},
																											TailStrictFodder: nil,
																										},
																										ThenFodder: ast.Fodder{},
																										BranchTrue: &ast.Apply{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(772),
																														Column: int(7),
																													},
																													End: ast.Location{
																														Line: int(772),
																														Column: int(43),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p3497,
																												FreeVars: ast.Identifiers{
																													"codes",
																													"format_codes_obj",
																													"vals",
																												},
																											},
																											Target: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(772),
																															Column: int(7),
																														},
																														End: ast.Location{
																															Line: int(772),
																															Column: int(23),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(6),
																															Comment: []string{},
																														},
																													},
																													Ctx: p3497,
																													FreeVars: ast.Identifiers{
																														"format_codes_obj",
																													},
																												},
																												Id: "format_codes_obj",
																											},
																											FodderLeft: ast.Fodder{},
																											Arguments: ast.Arguments{
																												Positional: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(772),
																																		Column: int(24),
																																	},
																																	End: ast.Location{
																																		Line: int(772),
																																		Column: int(29),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p8091,
																																FreeVars: ast.Identifiers{
																																	"codes",
																																},
																															},
																															Id: "codes",
																														},
																														CommaFodder: ast.Fodder{},
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(772),
																																		Column: int(31),
																																	},
																																	End: ast.Location{
																																		Line: int(772),
																																		Column: int(35),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p8091,
																																FreeVars: ast.Identifiers{
																																	"vals",
																																},
																															},
																															Id: "vals",
																														},
																														CommaFodder: ast.Fodder{},
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.LiteralNumber{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(772),
																																		Column: int(37),
																																	},
																																	End: ast.Location{
																																		Line: int(772),
																																		Column: int(38),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p8091,
																																FreeVars: nil,
																															},
																															OriginalString: "0",
																														},
																														CommaFodder: ast.Fodder{},
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.LiteralString{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(772),
																																		Column: int(40),
																																	},
																																	End: ast.Location{
																																		Line: int(772),
																																		Column: int(42),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p8091,
																																FreeVars: nil,
																															},
																															Value: "",
																															Kind: ast.LiteralStringKind(1),
																															BlockIndent: "",
																															BlockTermIndent: "",
																														},
																														CommaFodder: nil,
																													},
																												},
																												Named: nil,
																											},
																											TrailingComma: false,
																											TailStrict: false,
																											FodderRight: ast.Fodder{},
																											TailStrictFodder: nil,
																										},
																										ElseFodder: ast.Fodder{
																											ast.FodderElement{
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(4),
																												Comment: []string{},
																											},
																										},
																										BranchFalse: &ast.Apply{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(774),
																														Column: int(7),
																													},
																													End: ast.Location{
																														Line: int(774),
																														Column: int(48),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p3497,
																												FreeVars: ast.Identifiers{
																													"codes",
																													"format_codes_arr",
																													"vals",
																												},
																											},
																											Target: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(774),
																															Column: int(7),
																														},
																														End: ast.Location{
																															Line: int(774),
																															Column: int(23),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{
																														ast.FodderElement{
																															Kind: ast.FodderKind(0),
																															Blanks: int(0),
																															Indent: int(6),
																															Comment: []string{},
																														},
																													},
																													Ctx: p3497,
																													FreeVars: ast.Identifiers{
																														"format_codes_arr",
																													},
																												},
																												Id: "format_codes_arr",
																											},
																											FodderLeft: ast.Fodder{},
																											Arguments: ast.Arguments{
																												Positional: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(774),
																																		Column: int(24),
																																	},
																																	End: ast.Location{
																																		Line: int(774),
																																		Column: int(29),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p8105,
																																FreeVars: ast.Identifiers{
																																	"codes",
																																},
																															},
																															Id: "codes",
																														},
																														CommaFodder: ast.Fodder{},
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Array{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(774),
																																		Column: int(31),
																																	},
																																	End: ast.Location{
																																		Line: int(774),
																																		Column: int(37),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p8105,
																																FreeVars: ast.Identifiers{
																																	"vals",
																																},
																															},
																															Elements: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(774),
																																					Column: int(32),
																																				},
																																				End: ast.Location{
																																					Line: int(774),
																																					Column: int(36),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p8111,
																																			FreeVars: ast.Identifiers{
																																				"vals",
																																			},
																																		},
																																		Id: "vals",
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															TrailingComma: false,
																															CloseFodder: ast.Fodder{},
																														},
																														CommaFodder: ast.Fodder{},
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.LiteralNumber{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(774),
																																		Column: int(39),
																																	},
																																	End: ast.Location{
																																		Line: int(774),
																																		Column: int(40),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p8105,
																																FreeVars: nil,
																															},
																															OriginalString: "0",
																														},
																														CommaFodder: ast.Fodder{},
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.LiteralNumber{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(774),
																																		Column: int(42),
																																	},
																																	End: ast.Location{
																																		Line: int(774),
																																		Column: int(43),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p8105,
																																FreeVars: nil,
																															},
																															OriginalString: "0",
																														},
																														CommaFodder: ast.Fodder{},
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.LiteralString{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(774),
																																		Column: int(45),
																																	},
																																	End: ast.Location{
																																		Line: int(774),
																																		Column: int(47),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p8105,
																																FreeVars: nil,
																															},
																															Value: "",
																															Kind: ast.LiteralStringKind(1),
																															BlockIndent: "",
																															BlockTermIndent: "",
																														},
																														CommaFodder: nil,
																													},
																												},
																												Named: nil,
																											},
																											TrailingComma: false,
																											TailStrict: false,
																											FodderRight: ast.Fodder{},
																											TailStrictFodder: nil,
																										},
																									},
																								},
																							},
																						},
																					},
																				},
																			},
																		},
																	},
																},
															},
														},
													},
												},
											},
										},
									},
								},
							},
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(307),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(774),
					Column: int(48),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "foldr",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "func",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(776),
								Column: int(9),
							},
							End: ast.Location{
								Line: int(776),
								Column: int(13),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(776),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(776),
								Column: int(18),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "init",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(776),
								Column: int(20),
							},
							End: ast.Location{
								Line: int(776),
								Column: int(24),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(777),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(782),
								Column: int(46),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p8122,
						FreeVars: ast.Identifiers{
							"arr",
							"func",
							"init",
							"std",
						},
					},
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: nil,
							Variable: "aux",
							EqFodder: nil,
							Body: &ast.Function{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(777),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(781),
											Column: int(57),
										},
										File: p1,
									},
									Fodder: nil,
									Ctx: p8126,
									FreeVars: ast.Identifiers{
										"aux",
									},
								},
								ParenLeftFodder: ast.Fodder{},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "func",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(777),
												Column: int(15),
											},
											End: ast.Location{
												Line: int(777),
												Column: int(19),
											},
											File: p1,
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "arr",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(777),
												Column: int(21),
											},
											End: ast.Location{
												Line: int(777),
												Column: int(24),
											},
											File: p1,
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "running",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(777),
												Column: int(26),
											},
											End: ast.Location{
												Line: int(777),
												Column: int(33),
											},
											File: p1,
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "idx",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: nil,
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(777),
												Column: int(35),
											},
											End: ast.Location{
												Line: int(777),
												Column: int(38),
											},
											File: p1,
										},
									},
								},
								TrailingComma: false,
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Conditional{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(778),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(781),
												Column: int(57),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										Ctx: p8131,
										FreeVars: ast.Identifiers{
											"arr",
											"aux",
											"func",
											"idx",
											"running",
										},
									},
									Cond: &ast.Binary{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(778),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(778),
													Column: int(17),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p8131,
											FreeVars: ast.Identifiers{
												"idx",
											},
										},
										Left: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(778),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(778),
														Column: int(13),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p8131,
												FreeVars: ast.Identifiers{
													"idx",
												},
											},
											Id: "idx",
										},
										OpFodder: ast.Fodder{},
										Op: ast.BinaryOp(9),
										Right: &ast.LiteralNumber{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(778),
														Column: int(16),
													},
													End: ast.Location{
														Line: int(778),
														Column: int(17),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p8131,
												FreeVars: nil,
											},
											OriginalString: "0",
										},
									},
									ThenFodder: ast.Fodder{},
									BranchTrue: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(779),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(779),
													Column: int(16),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
													Comment: []string{},
												},
											},
											Ctx: p8131,
											FreeVars: ast.Identifiers{
												"running",
											},
										},
										Id: "running",
									},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
											Comment: []string{},
										},
									},
									BranchFalse: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(781),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(781),
													Column: int(57),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p8131,
											FreeVars: ast.Identifiers{
												"arr",
												"aux",
												"func",
												"idx",
												"running",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(781),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(781),
														Column: int(12),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
														Comment: []string{},
													},
												},
												Ctx: p8131,
												FreeVars: ast.Identifiers{
													"aux",
												},
											},
											Id: "aux",
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(781),
																	Column: int(13),
																},
																End: ast.Location{
																	Line: int(781),
																	Column: int(17),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p8149,
															FreeVars: ast.Identifiers{
																"func",
															},
														},
														Id: "func",
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(781),
																	Column: int(19),
																},
																End: ast.Location{
																	Line: int(781),
																	Column: int(22),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p8149,
															FreeVars: ast.Identifiers{
																"arr",
															},
														},
														Id: "arr",
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(781),
																	Column: int(24),
																},
																End: ast.Location{
																	Line: int(781),
																	Column: int(47),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p8149,
															FreeVars: ast.Identifiers{
																"arr",
																"func",
																"idx",
																"running",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(781),
																		Column: int(24),
																	},
																	End: ast.Location{
																		Line: int(781),
																		Column: int(28),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p8149,
																FreeVars: ast.Identifiers{
																	"func",
																},
															},
															Id: "func",
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(781),
																					Column: int(29),
																				},
																				End: ast.Location{
																					Line: int(781),
																					Column: int(37),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p8159,
																			FreeVars: ast.Identifiers{
																				"arr",
																				"idx",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(781),
																						Column: int(29),
																					},
																					End: ast.Location{
																						Line: int(781),
																						Column: int(32),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p8159,
																				FreeVars: ast.Identifiers{
																					"arr",
																				},
																			},
																			Id: "arr",
																		},
																		LeftBracketFodder: ast.Fodder{},
																		Index: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(781),
																						Column: int(33),
																					},
																					End: ast.Location{
																						Line: int(781),
																						Column: int(36),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p8159,
																				FreeVars: ast.Identifiers{
																					"idx",
																				},
																			},
																			Id: "idx",
																		},
																		RightBracketFodder: ast.Fodder{},
																		Id: nil,
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(781),
																					Column: int(39),
																				},
																				End: ast.Location{
																					Line: int(781),
																					Column: int(46),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p8159,
																			FreeVars: ast.Identifiers{
																				"running",
																			},
																		},
																		Id: "running",
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														TrailingComma: false,
														TailStrict: false,
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Binary{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(781),
																	Column: int(49),
																},
																End: ast.Location{
																	Line: int(781),
																	Column: int(56),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p8149,
															FreeVars: ast.Identifiers{
																"idx",
															},
														},
														Left: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(781),
																		Column: int(49),
																	},
																	End: ast.Location{
																		Line: int(781),
																		Column: int(52),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p8149,
																FreeVars: ast.Identifiers{
																	"idx",
																},
															},
															Id: "idx",
														},
														OpFodder: ast.Fodder{},
														Op: ast.BinaryOp(4),
														Right: &ast.LiteralNumber{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(781),
																		Column: int(55),
																	},
																	End: ast.Location{
																		Line: int(781),
																		Column: int(56),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p8149,
																FreeVars: nil,
															},
															OriginalString: "1",
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: true,
										FodderRight: ast.Fodder{},
										TailStrictFodder: ast.Fodder{},
									},
								},
							},
							Fun: nil,
							CloseFodder: nil,
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
						},
					},
					Body: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(782),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(782),
									Column: int(46),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p8122,
							FreeVars: ast.Identifiers{
								"arr",
								"aux",
								"func",
								"init",
								"std",
							},
						},
						Target: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(782),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(782),
										Column: int(8),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
										Comment: []string{},
									},
								},
								Ctx: p8122,
								FreeVars: ast.Identifiers{
									"aux",
								},
							},
							Id: "aux",
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(782),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(782),
													Column: int(13),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p8179,
											FreeVars: ast.Identifiers{
												"func",
											},
										},
										Id: "func",
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(782),
													Column: int(15),
												},
												End: ast.Location{
													Line: int(782),
													Column: int(18),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p8179,
											FreeVars: ast.Identifiers{
												"arr",
											},
										},
										Id: "arr",
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(782),
													Column: int(20),
												},
												End: ast.Location{
													Line: int(782),
													Column: int(24),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p8179,
											FreeVars: ast.Identifiers{
												"init",
											},
										},
										Id: "init",
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Binary{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(782),
													Column: int(26),
												},
												End: ast.Location{
													Line: int(782),
													Column: int(45),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p8179,
											FreeVars: ast.Identifiers{
												"arr",
												"std",
											},
										},
										Left: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(782),
														Column: int(26),
													},
													End: ast.Location{
														Line: int(782),
														Column: int(41),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p8179,
												FreeVars: ast.Identifiers{
													"arr",
													"std",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(782),
															Column: int(26),
														},
														End: ast.Location{
															Line: int(782),
															Column: int(36),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p8179,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(782),
																Column: int(26),
															},
															End: ast.Location{
																Line: int(782),
																Column: int(29),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "length",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(782),
																		Column: int(37),
																	},
																	End: ast.Location{
																		Line: int(782),
																		Column: int(40),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p8196,
																FreeVars: ast.Identifiers{
																	"arr",
																},
															},
															Id: "arr",
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
										OpFodder: ast.Fodder{},
										Op: ast.BinaryOp(4),
										Right: &ast.LiteralNumber{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(782),
														Column: int(44),
													},
													End: ast.Location{
														Line: int(782),
														Column: int(45),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p8179,
												FreeVars: nil,
											},
											OriginalString: "1",
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(776),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(782),
					Column: int(46),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "foldl",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "func",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(784),
								Column: int(9),
							},
							End: ast.Location{
								Line: int(784),
								Column: int(13),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(784),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(784),
								Column: int(18),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "init",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(784),
								Column: int(20),
							},
							End: ast.Location{
								Line: int(784),
								Column: int(24),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(785),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(790),
								Column: int(28),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p8205,
						FreeVars: ast.Identifiers{
							"arr",
							"func",
							"init",
							"std",
						},
					},
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: nil,
							Variable: "aux",
							EqFodder: nil,
							Body: &ast.Function{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(785),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(789),
											Column: int(57),
										},
										File: p1,
									},
									Fodder: nil,
									Ctx: p8209,
									FreeVars: ast.Identifiers{
										"aux",
										"std",
									},
								},
								ParenLeftFodder: ast.Fodder{},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "func",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(785),
												Column: int(15),
											},
											End: ast.Location{
												Line: int(785),
												Column: int(19),
											},
											File: p1,
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "arr",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(785),
												Column: int(21),
											},
											End: ast.Location{
												Line: int(785),
												Column: int(24),
											},
											File: p1,
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "running",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(785),
												Column: int(26),
											},
											End: ast.Location{
												Line: int(785),
												Column: int(33),
											},
											File: p1,
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "idx",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: nil,
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(785),
												Column: int(35),
											},
											End: ast.Location{
												Line: int(785),
												Column: int(38),
											},
											File: p1,
										},
									},
								},
								TrailingComma: false,
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Conditional{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(786),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(789),
												Column: int(57),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										Ctx: p8214,
										FreeVars: ast.Identifiers{
											"arr",
											"aux",
											"func",
											"idx",
											"running",
											"std",
										},
									},
									Cond: &ast.Binary{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(786),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(786),
													Column: int(32),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p8214,
											FreeVars: ast.Identifiers{
												"arr",
												"idx",
												"std",
											},
										},
										Left: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(786),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(786),
														Column: int(13),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p8214,
												FreeVars: ast.Identifiers{
													"idx",
												},
											},
											Id: "idx",
										},
										OpFodder: ast.Fodder{},
										Op: ast.BinaryOp(8),
										Right: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(786),
														Column: int(17),
													},
													End: ast.Location{
														Line: int(786),
														Column: int(32),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p8214,
												FreeVars: ast.Identifiers{
													"arr",
													"std",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(786),
															Column: int(17),
														},
														End: ast.Location{
															Line: int(786),
															Column: int(27),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p8214,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(786),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(786),
																Column: int(20),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "length",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(786),
																		Column: int(28),
																	},
																	End: ast.Location{
																		Line: int(786),
																		Column: int(31),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p8229,
																FreeVars: ast.Identifiers{
																	"arr",
																},
															},
															Id: "arr",
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
									},
									ThenFodder: ast.Fodder{},
									BranchTrue: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(787),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(787),
													Column: int(16),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
													Comment: []string{},
												},
											},
											Ctx: p8214,
											FreeVars: ast.Identifiers{
												"running",
											},
										},
										Id: "running",
									},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
											Comment: []string{},
										},
									},
									BranchFalse: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(789),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(789),
													Column: int(57),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p8214,
											FreeVars: ast.Identifiers{
												"arr",
												"aux",
												"func",
												"idx",
												"running",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(789),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(789),
														Column: int(12),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
														Comment: []string{},
													},
												},
												Ctx: p8214,
												FreeVars: ast.Identifiers{
													"aux",
												},
											},
											Id: "aux",
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(789),
																	Column: int(13),
																},
																End: ast.Location{
																	Line: int(789),
																	Column: int(17),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p8242,
															FreeVars: ast.Identifiers{
																"func",
															},
														},
														Id: "func",
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(789),
																	Column: int(19),
																},
																End: ast.Location{
																	Line: int(789),
																	Column: int(22),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p8242,
															FreeVars: ast.Identifiers{
																"arr",
															},
														},
														Id: "arr",
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(789),
																	Column: int(24),
																},
																End: ast.Location{
																	Line: int(789),
																	Column: int(47),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p8242,
															FreeVars: ast.Identifiers{
																"arr",
																"func",
																"idx",
																"running",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(789),
																		Column: int(24),
																	},
																	End: ast.Location{
																		Line: int(789),
																		Column: int(28),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p8242,
																FreeVars: ast.Identifiers{
																	"func",
																},
															},
															Id: "func",
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(789),
																					Column: int(29),
																				},
																				End: ast.Location{
																					Line: int(789),
																					Column: int(36),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p8252,
																			FreeVars: ast.Identifiers{
																				"running",
																			},
																		},
																		Id: "running",
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(789),
																					Column: int(38),
																				},
																				End: ast.Location{
																					Line: int(789),
																					Column: int(46),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p8252,
																			FreeVars: ast.Identifiers{
																				"arr",
																				"idx",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(789),
																						Column: int(38),
																					},
																					End: ast.Location{
																						Line: int(789),
																						Column: int(41),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p8252,
																				FreeVars: ast.Identifiers{
																					"arr",
																				},
																			},
																			Id: "arr",
																		},
																		LeftBracketFodder: ast.Fodder{},
																		Index: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(789),
																						Column: int(42),
																					},
																					End: ast.Location{
																						Line: int(789),
																						Column: int(45),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p8252,
																				FreeVars: ast.Identifiers{
																					"idx",
																				},
																			},
																			Id: "idx",
																		},
																		RightBracketFodder: ast.Fodder{},
																		Id: nil,
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														TrailingComma: false,
														TailStrict: false,
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Binary{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(789),
																	Column: int(49),
																},
																End: ast.Location{
																	Line: int(789),
																	Column: int(56),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p8242,
															FreeVars: ast.Identifiers{
																"idx",
															},
														},
														Left: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(789),
																		Column: int(49),
																	},
																	End: ast.Location{
																		Line: int(789),
																		Column: int(52),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p8242,
																FreeVars: ast.Identifiers{
																	"idx",
																},
															},
															Id: "idx",
														},
														OpFodder: ast.Fodder{},
														Op: ast.BinaryOp(3),
														Right: &ast.LiteralNumber{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(789),
																		Column: int(55),
																	},
																	End: ast.Location{
																		Line: int(789),
																		Column: int(56),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p8242,
																FreeVars: nil,
															},
															OriginalString: "1",
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: true,
										FodderRight: ast.Fodder{},
										TailStrictFodder: ast.Fodder{},
									},
								},
							},
							Fun: nil,
							CloseFodder: nil,
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
						},
					},
					Body: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(790),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(790),
									Column: int(28),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p8205,
							FreeVars: ast.Identifiers{
								"arr",
								"aux",
								"func",
								"init",
							},
						},
						Target: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(790),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(790),
										Column: int(8),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
										Comment: []string{},
									},
								},
								Ctx: p8205,
								FreeVars: ast.Identifiers{
									"aux",
								},
							},
							Id: "aux",
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(790),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(790),
													Column: int(13),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p8272,
											FreeVars: ast.Identifiers{
												"func",
											},
										},
										Id: "func",
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(790),
													Column: int(15),
												},
												End: ast.Location{
													Line: int(790),
													Column: int(18),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p8272,
											FreeVars: ast.Identifiers{
												"arr",
											},
										},
										Id: "arr",
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(790),
													Column: int(20),
												},
												End: ast.Location{
													Line: int(790),
													Column: int(24),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p8272,
											FreeVars: ast.Identifiers{
												"init",
											},
										},
										Id: "init",
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.LiteralNumber{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(790),
													Column: int(26),
												},
												End: ast.Location{
													Line: int(790),
													Column: int(27),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p8272,
											FreeVars: nil,
										},
										OriginalString: "0",
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(784),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(790),
					Column: int(28),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "filterMap",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "filter_func",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(793),
								Column: int(13),
							},
							End: ast.Location{
								Line: int(793),
								Column: int(24),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "map_func",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(793),
								Column: int(26),
							},
							End: ast.Location{
								Line: int(793),
								Column: int(34),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(793),
								Column: int(36),
							},
							End: ast.Location{
								Line: int(793),
								Column: int(39),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(794),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(801),
								Column: int(54),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p8285,
						FreeVars: ast.Identifiers{
							"arr",
							"filter_func",
							"map_func",
							"std",
						},
					},
					Cond: &ast.Unary{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(794),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(794),
									Column: int(36),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p8285,
							FreeVars: ast.Identifiers{
								"filter_func",
								"std",
							},
						},
						Op: ast.UnaryOp(0),
						Expr: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(794),
										Column: int(9),
									},
									End: ast.Location{
										Line: int(794),
										Column: int(36),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p8285,
								FreeVars: ast.Identifiers{
									"filter_func",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(794),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(794),
											Column: int(23),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8285,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(794),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(794),
												Column: int(12),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "isFunction",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(794),
														Column: int(24),
													},
													End: ast.Location{
														Line: int(794),
														Column: int(35),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p8298,
												FreeVars: ast.Identifiers{
													"filter_func",
												},
											},
											Id: "filter_func",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
					},
					ThenFodder: ast.Fodder{},
					BranchTrue: &ast.Error{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(795),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(795),
									Column: int(89),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
									Comment: []string{},
								},
							},
							Ctx: p8285,
							FreeVars: ast.Identifiers{
								"filter_func",
								"std",
							},
						},
						Expr: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(795),
										Column: int(14),
									},
									End: ast.Location{
										Line: int(795),
										Column: int(88),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p8285,
								FreeVars: ast.Identifiers{
									"filter_func",
									"std",
								},
							},
							Left: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(795),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(795),
											Column: int(64),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8285,
									FreeVars: nil,
								},
								Value: "std.filterMap first param must be function, got ",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(3),
							Right: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(795),
											Column: int(67),
										},
										End: ast.Location{
											Line: int(795),
											Column: int(88),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8285,
									FreeVars: ast.Identifiers{
										"filter_func",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(795),
												Column: int(67),
											},
											End: ast.Location{
												Line: int(795),
												Column: int(75),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8285,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(795),
													Column: int(67),
												},
												End: ast.Location{
													Line: int(795),
													Column: int(70),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "type",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(795),
															Column: int(76),
														},
														End: ast.Location{
															Line: int(795),
															Column: int(87),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p8315,
													FreeVars: ast.Identifiers{
														"filter_func",
													},
												},
												Id: "filter_func",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
						},
					},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
							Comment: []string{},
						},
					},
					BranchFalse: &ast.Conditional{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(796),
									Column: int(10),
								},
								End: ast.Location{
									Line: int(801),
									Column: int(54),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p8285,
							FreeVars: ast.Identifiers{
								"arr",
								"filter_func",
								"map_func",
								"std",
							},
						},
						Cond: &ast.Unary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(796),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(796),
										Column: int(38),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p8285,
								FreeVars: ast.Identifiers{
									"map_func",
									"std",
								},
							},
							Op: ast.UnaryOp(0),
							Expr: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(796),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(796),
											Column: int(38),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8285,
									FreeVars: ast.Identifiers{
										"map_func",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(796),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(796),
												Column: int(28),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8285,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(796),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(796),
													Column: int(17),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "isFunction",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(796),
															Column: int(29),
														},
														End: ast.Location{
															Line: int(796),
															Column: int(37),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p8331,
													FreeVars: ast.Identifiers{
														"map_func",
													},
												},
												Id: "map_func",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
						},
						ThenFodder: ast.Fodder{},
						BranchTrue: &ast.Error{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(797),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(797),
										Column: int(87),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
										Comment: []string{},
									},
								},
								Ctx: p8285,
								FreeVars: ast.Identifiers{
									"map_func",
									"std",
								},
							},
							Expr: &ast.Binary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(797),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(797),
											Column: int(86),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8285,
									FreeVars: ast.Identifiers{
										"map_func",
										"std",
									},
								},
								Left: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(797),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(797),
												Column: int(65),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8285,
										FreeVars: nil,
									},
									Value: "std.filterMap second param must be function, got ",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								OpFodder: ast.Fodder{},
								Op: ast.BinaryOp(3),
								Right: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(797),
												Column: int(68),
											},
											End: ast.Location{
												Line: int(797),
												Column: int(86),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8285,
										FreeVars: ast.Identifiers{
											"map_func",
											"std",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(797),
													Column: int(68),
												},
												End: ast.Location{
													Line: int(797),
													Column: int(76),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p8285,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(797),
														Column: int(68),
													},
													End: ast.Location{
														Line: int(797),
														Column: int(71),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "type",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(797),
																Column: int(77),
															},
															End: ast.Location{
																Line: int(797),
																Column: int(85),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p8348,
														FreeVars: ast.Identifiers{
															"map_func",
														},
													},
													Id: "map_func",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
							},
						},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						BranchFalse: &ast.Conditional{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(798),
										Column: int(10),
									},
									End: ast.Location{
										Line: int(801),
										Column: int(54),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p8285,
								FreeVars: ast.Identifiers{
									"arr",
									"filter_func",
									"map_func",
									"std",
								},
							},
							Cond: &ast.Unary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(798),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(798),
											Column: int(30),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8285,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
								},
								Op: ast.UnaryOp(0),
								Expr: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(798),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(798),
												Column: int(30),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8285,
										FreeVars: ast.Identifiers{
											"arr",
											"std",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(798),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(798),
													Column: int(25),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p8285,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(798),
														Column: int(14),
													},
													End: ast.Location{
														Line: int(798),
														Column: int(17),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "isArray",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(798),
																Column: int(26),
															},
															End: ast.Location{
																Line: int(798),
																Column: int(29),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p8364,
														FreeVars: ast.Identifiers{
															"arr",
														},
													},
													Id: "arr",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
							},
							ThenFodder: ast.Fodder{},
							BranchTrue: &ast.Error{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(799),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(799),
											Column: int(78),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
											Comment: []string{},
										},
									},
									Ctx: p8285,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
								},
								Expr: &ast.Binary{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(799),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(799),
												Column: int(77),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8285,
										FreeVars: ast.Identifiers{
											"arr",
											"std",
										},
									},
									Left: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(799),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(799),
													Column: int(61),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p8285,
											FreeVars: nil,
										},
										Value: "std.filterMap third param must be array, got ",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									OpFodder: ast.Fodder{},
									Op: ast.BinaryOp(3),
									Right: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(799),
													Column: int(64),
												},
												End: ast.Location{
													Line: int(799),
													Column: int(77),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p8285,
											FreeVars: ast.Identifiers{
												"arr",
												"std",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(799),
														Column: int(64),
													},
													End: ast.Location{
														Line: int(799),
														Column: int(72),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p8285,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(799),
															Column: int(64),
														},
														End: ast.Location{
															Line: int(799),
															Column: int(67),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: ast.Fodder{},
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "type",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: ast.Fodder{},
											Id: nil,
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(799),
																	Column: int(73),
																},
																End: ast.Location{
																	Line: int(799),
																	Column: int(76),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p8381,
															FreeVars: ast.Identifiers{
																"arr",
															},
														},
														Id: "arr",
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
								},
							},
							ElseFodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
									Comment: []string{},
								},
							},
							BranchFalse: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(801),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(801),
											Column: int(54),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8285,
									FreeVars: ast.Identifiers{
										"arr",
										"filter_func",
										"map_func",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(801),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(801),
												Column: int(14),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8285,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(801),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(801),
													Column: int(10),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
													Comment: []string{},
												},
											},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "map",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(801),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(801),
															Column: int(23),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p8394,
													FreeVars: ast.Identifiers{
														"map_func",
													},
												},
												Id: "map_func",
											},
											CommaFodder: ast.Fodder{},
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(801),
															Column: int(25),
														},
														End: ast.Location{
															Line: int(801),
															Column: int(53),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p8394,
													FreeVars: ast.Identifiers{
														"arr",
														"filter_func",
														"std",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(801),
																Column: int(25),
															},
															End: ast.Location{
																Line: int(801),
																Column: int(35),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p8394,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(801),
																	Column: int(25),
																},
																End: ast.Location{
																	Line: int(801),
																	Column: int(28),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: ast.Fodder{},
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "filter",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: ast.Fodder{},
													Id: nil,
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(801),
																			Column: int(36),
																		},
																		End: ast.Location{
																			Line: int(801),
																			Column: int(47),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p8405,
																	FreeVars: ast.Identifiers{
																		"filter_func",
																	},
																},
																Id: "filter_func",
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(801),
																			Column: int(49),
																		},
																		End: ast.Location{
																			Line: int(801),
																			Column: int(52),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p8405,
																	FreeVars: ast.Identifiers{
																		"arr",
																	},
																},
																Id: "arr",
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(793),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(801),
					Column: int(54),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "assertEqual",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: nil,
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "a",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(803),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(803),
								Column: int(16),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "b",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(803),
								Column: int(18),
							},
							End: ast.Location{
								Line: int(803),
								Column: int(19),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(804),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(807),
								Column: int(50),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p8414,
						FreeVars: ast.Identifiers{
							"a",
							"b",
						},
					},
					Cond: &ast.Binary{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(804),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(804),
									Column: int(14),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p8414,
							FreeVars: ast.Identifiers{
								"a",
								"b",
							},
						},
						Left: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(804),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(804),
										Column: int(9),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p8414,
								FreeVars: ast.Identifiers{
									"a",
								},
							},
							Id: "a",
						},
						OpFodder: ast.Fodder{},
						Op: ast.BinaryOp(12),
						Right: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(804),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(804),
										Column: int(14),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p8414,
								FreeVars: ast.Identifiers{
									"b",
								},
							},
							Id: "b",
						},
					},
					ThenFodder: ast.Fodder{},
					BranchTrue: &ast.LiteralBoolean{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(805),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(805),
									Column: int(11),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
									Comment: []string{},
								},
							},
							Ctx: p8414,
							FreeVars: nil,
						},
						Value: true,
					},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
							Comment: []string{},
						},
					},
					BranchFalse: &ast.Error{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(807),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(807),
									Column: int(50),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
									Comment: []string{},
								},
							},
							Ctx: p8414,
							FreeVars: ast.Identifiers{
								"a",
								"b",
							},
						},
						Expr: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(807),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(807),
										Column: int(50),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p8414,
								FreeVars: ast.Identifiers{
									"a",
									"b",
								},
							},
							Left: &ast.Binary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(807),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(807),
											Column: int(46),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8414,
									FreeVars: ast.Identifiers{
										"a",
									},
								},
								Left: &ast.Binary{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(807),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(807),
												Column: int(37),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8414,
										FreeVars: ast.Identifiers{
											"a",
										},
									},
									Left: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(807),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(807),
													Column: int(33),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p8414,
											FreeVars: nil,
										},
										Value: "Assertion failed. ",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									OpFodder: ast.Fodder{},
									Op: ast.BinaryOp(3),
									Right: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(807),
													Column: int(36),
												},
												End: ast.Location{
													Line: int(807),
													Column: int(37),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p8414,
											FreeVars: ast.Identifiers{
												"a",
											},
										},
										Id: "a",
									},
								},
								OpFodder: ast.Fodder{},
								Op: ast.BinaryOp(3),
								Right: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(807),
												Column: int(40),
											},
											End: ast.Location{
												Line: int(807),
												Column: int(46),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8414,
										FreeVars: nil,
									},
									Value: " != ",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(3),
							Right: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(807),
											Column: int(49),
										},
										End: ast.Location{
											Line: int(807),
											Column: int(50),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8414,
									FreeVars: ast.Identifiers{
										"b",
									},
								},
								Id: "b",
							},
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(803),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(807),
					Column: int(50),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "abs",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "n",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(809),
								Column: int(7),
							},
							End: ast.Location{
								Line: int(809),
								Column: int(8),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(810),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(813),
								Column: int(30),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p8446,
						FreeVars: ast.Identifiers{
							"n",
							"std",
						},
					},
					Cond: &ast.Unary{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(810),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(810),
									Column: int(24),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p8446,
							FreeVars: ast.Identifiers{
								"n",
								"std",
							},
						},
						Op: ast.UnaryOp(0),
						Expr: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(810),
										Column: int(9),
									},
									End: ast.Location{
										Line: int(810),
										Column: int(24),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p8446,
								FreeVars: ast.Identifiers{
									"n",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(810),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(810),
											Column: int(21),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8446,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(810),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(810),
												Column: int(12),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "isNumber",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(810),
														Column: int(22),
													},
													End: ast.Location{
														Line: int(810),
														Column: int(23),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p8459,
												FreeVars: ast.Identifiers{
													"n",
												},
											},
											Id: "n",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
					},
					ThenFodder: ast.Fodder{},
					BranchTrue: &ast.Error{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(811),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(811),
									Column: int(58),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
									Comment: []string{},
								},
							},
							Ctx: p8446,
							FreeVars: ast.Identifiers{
								"n",
								"std",
							},
						},
						Expr: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(811),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(811),
										Column: int(58),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p8446,
								FreeVars: ast.Identifiers{
									"n",
									"std",
								},
							},
							Left: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(811),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(811),
											Column: int(44),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8446,
									FreeVars: nil,
								},
								Value: "std.abs expected number, got ",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(3),
							Right: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(811),
											Column: int(47),
										},
										End: ast.Location{
											Line: int(811),
											Column: int(58),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8446,
									FreeVars: ast.Identifiers{
										"n",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(811),
												Column: int(47),
											},
											End: ast.Location{
												Line: int(811),
												Column: int(55),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8446,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(811),
													Column: int(47),
												},
												End: ast.Location{
													Line: int(811),
													Column: int(50),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "type",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(811),
															Column: int(56),
														},
														End: ast.Location{
															Line: int(811),
															Column: int(57),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p8476,
													FreeVars: ast.Identifiers{
														"n",
													},
												},
												Id: "n",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
						},
					},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
							Comment: []string{},
						},
					},
					BranchFalse: &ast.Conditional{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(813),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(813),
									Column: int(30),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
									Comment: []string{},
								},
							},
							Ctx: p8446,
							FreeVars: ast.Identifiers{
								"n",
							},
						},
						Cond: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(813),
										Column: int(10),
									},
									End: ast.Location{
										Line: int(813),
										Column: int(15),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p8446,
								FreeVars: ast.Identifiers{
									"n",
								},
							},
							Left: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(813),
											Column: int(10),
										},
										End: ast.Location{
											Line: int(813),
											Column: int(11),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8446,
									FreeVars: ast.Identifiers{
										"n",
									},
								},
								Id: "n",
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(7),
							Right: &ast.LiteralNumber{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(813),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(813),
											Column: int(15),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8446,
									FreeVars: nil,
								},
								OriginalString: "0",
							},
						},
						ThenFodder: ast.Fodder{},
						BranchTrue: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(813),
										Column: int(21),
									},
									End: ast.Location{
										Line: int(813),
										Column: int(22),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p8446,
								FreeVars: ast.Identifiers{
									"n",
								},
							},
							Id: "n",
						},
						ElseFodder: ast.Fodder{},
						BranchFalse: &ast.Unary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(813),
										Column: int(28),
									},
									End: ast.Location{
										Line: int(813),
										Column: int(30),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p8446,
								FreeVars: ast.Identifiers{
									"n",
								},
							},
							Op: ast.UnaryOp(3),
							Expr: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(813),
											Column: int(29),
										},
										End: ast.Location{
											Line: int(813),
											Column: int(30),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8446,
									FreeVars: ast.Identifiers{
										"n",
									},
								},
								Id: "n",
							},
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(809),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(813),
					Column: int(30),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "sign",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "n",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(815),
								Column: int(8),
							},
							End: ast.Location{
								Line: int(815),
								Column: int(9),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(816),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(823),
								Column: int(13),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p8499,
						FreeVars: ast.Identifiers{
							"n",
							"std",
						},
					},
					Cond: &ast.Unary{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(816),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(816),
									Column: int(24),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p8499,
							FreeVars: ast.Identifiers{
								"n",
								"std",
							},
						},
						Op: ast.UnaryOp(0),
						Expr: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(816),
										Column: int(9),
									},
									End: ast.Location{
										Line: int(816),
										Column: int(24),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p8499,
								FreeVars: ast.Identifiers{
									"n",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(816),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(816),
											Column: int(21),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8499,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(816),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(816),
												Column: int(12),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "isNumber",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(816),
														Column: int(22),
													},
													End: ast.Location{
														Line: int(816),
														Column: int(23),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p8512,
												FreeVars: ast.Identifiers{
													"n",
												},
											},
											Id: "n",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
					},
					ThenFodder: ast.Fodder{},
					BranchTrue: &ast.Error{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(817),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(817),
									Column: int(59),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
									Comment: []string{},
								},
							},
							Ctx: p8499,
							FreeVars: ast.Identifiers{
								"n",
								"std",
							},
						},
						Expr: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(817),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(817),
										Column: int(59),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p8499,
								FreeVars: ast.Identifiers{
									"n",
									"std",
								},
							},
							Left: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(817),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(817),
											Column: int(45),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8499,
									FreeVars: nil,
								},
								Value: "std.sign expected number, got ",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(3),
							Right: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(817),
											Column: int(48),
										},
										End: ast.Location{
											Line: int(817),
											Column: int(59),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8499,
									FreeVars: ast.Identifiers{
										"n",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(817),
												Column: int(48),
											},
											End: ast.Location{
												Line: int(817),
												Column: int(56),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8499,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(817),
													Column: int(48),
												},
												End: ast.Location{
													Line: int(817),
													Column: int(51),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "type",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(817),
															Column: int(57),
														},
														End: ast.Location{
															Line: int(817),
															Column: int(58),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p8529,
													FreeVars: ast.Identifiers{
														"n",
													},
												},
												Id: "n",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
						},
					},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
							Comment: []string{},
						},
					},
					BranchFalse: &ast.Conditional{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(819),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(823),
									Column: int(13),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
									Comment: []string{},
								},
							},
							Ctx: p8499,
							FreeVars: ast.Identifiers{
								"n",
							},
						},
						Cond: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(819),
										Column: int(10),
									},
									End: ast.Location{
										Line: int(819),
										Column: int(15),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p8499,
								FreeVars: ast.Identifiers{
									"n",
								},
							},
							Left: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(819),
											Column: int(10),
										},
										End: ast.Location{
											Line: int(819),
											Column: int(11),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8499,
									FreeVars: ast.Identifiers{
										"n",
									},
								},
								Id: "n",
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(7),
							Right: &ast.LiteralNumber{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(819),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(819),
											Column: int(15),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8499,
									FreeVars: nil,
								},
								OriginalString: "0",
							},
						},
						ThenFodder: ast.Fodder{},
						BranchTrue: &ast.LiteralNumber{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(820),
										Column: int(9),
									},
									End: ast.Location{
										Line: int(820),
										Column: int(10),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(8),
										Comment: []string{},
									},
								},
								Ctx: p8499,
								FreeVars: nil,
							},
							OriginalString: "1",
						},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(6),
								Comment: []string{},
							},
						},
						BranchFalse: &ast.Conditional{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(821),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(823),
										Column: int(13),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p8499,
								FreeVars: ast.Identifiers{
									"n",
								},
							},
							Cond: &ast.Binary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(821),
											Column: int(15),
										},
										End: ast.Location{
											Line: int(821),
											Column: int(20),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8499,
									FreeVars: ast.Identifiers{
										"n",
									},
								},
								Left: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(821),
												Column: int(15),
											},
											End: ast.Location{
												Line: int(821),
												Column: int(16),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8499,
										FreeVars: ast.Identifiers{
											"n",
										},
									},
									Id: "n",
								},
								OpFodder: ast.Fodder{},
								Op: ast.BinaryOp(9),
								Right: &ast.LiteralNumber{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(821),
												Column: int(19),
											},
											End: ast.Location{
												Line: int(821),
												Column: int(20),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8499,
										FreeVars: nil,
									},
									OriginalString: "0",
								},
							},
							ThenFodder: ast.Fodder{},
							BranchTrue: &ast.Unary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(822),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(822),
											Column: int(11),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(8),
											Comment: []string{},
										},
									},
									Ctx: p8499,
									FreeVars: nil,
								},
								Op: ast.UnaryOp(3),
								Expr: &ast.LiteralNumber{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(822),
												Column: int(10),
											},
											End: ast.Location{
												Line: int(822),
												Column: int(11),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8499,
										FreeVars: nil,
									},
									OriginalString: "1",
								},
							},
							ElseFodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
									Comment: []string{},
								},
							},
							BranchFalse: &ast.LiteralNumber{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(823),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(823),
											Column: int(13),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8499,
									FreeVars: nil,
								},
								OriginalString: "0",
							},
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(815),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(823),
					Column: int(13),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "max",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "a",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(825),
								Column: int(7),
							},
							End: ast.Location{
								Line: int(825),
								Column: int(8),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "b",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(825),
								Column: int(10),
							},
							End: ast.Location{
								Line: int(825),
								Column: int(11),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(826),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(831),
								Column: int(29),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p8561,
						FreeVars: ast.Identifiers{
							"a",
							"b",
							"std",
						},
					},
					Cond: &ast.Unary{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(826),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(826),
									Column: int(24),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p8561,
							FreeVars: ast.Identifiers{
								"a",
								"std",
							},
						},
						Op: ast.UnaryOp(0),
						Expr: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(826),
										Column: int(9),
									},
									End: ast.Location{
										Line: int(826),
										Column: int(24),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p8561,
								FreeVars: ast.Identifiers{
									"a",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(826),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(826),
											Column: int(21),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8561,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(826),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(826),
												Column: int(12),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "isNumber",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(826),
														Column: int(22),
													},
													End: ast.Location{
														Line: int(826),
														Column: int(23),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p8574,
												FreeVars: ast.Identifiers{
													"a",
												},
											},
											Id: "a",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
					},
					ThenFodder: ast.Fodder{},
					BranchTrue: &ast.Error{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(827),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(827),
									Column: int(70),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
									Comment: []string{},
								},
							},
							Ctx: p8561,
							FreeVars: ast.Identifiers{
								"a",
								"std",
							},
						},
						Expr: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(827),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(827),
										Column: int(70),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p8561,
								FreeVars: ast.Identifiers{
									"a",
									"std",
								},
							},
							Left: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(827),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(827),
											Column: int(56),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8561,
									FreeVars: nil,
								},
								Value: "std.max first param expected number, got ",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(3),
							Right: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(827),
											Column: int(59),
										},
										End: ast.Location{
											Line: int(827),
											Column: int(70),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8561,
									FreeVars: ast.Identifiers{
										"a",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(827),
												Column: int(59),
											},
											End: ast.Location{
												Line: int(827),
												Column: int(67),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8561,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(827),
													Column: int(59),
												},
												End: ast.Location{
													Line: int(827),
													Column: int(62),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "type",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(827),
															Column: int(68),
														},
														End: ast.Location{
															Line: int(827),
															Column: int(69),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p8591,
													FreeVars: ast.Identifiers{
														"a",
													},
												},
												Id: "a",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
						},
					},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
							Comment: []string{},
						},
					},
					BranchFalse: &ast.Conditional{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(828),
									Column: int(10),
								},
								End: ast.Location{
									Line: int(831),
									Column: int(29),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p8561,
							FreeVars: ast.Identifiers{
								"a",
								"b",
								"std",
							},
						},
						Cond: &ast.Unary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(828),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(828),
										Column: int(29),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p8561,
								FreeVars: ast.Identifiers{
									"b",
									"std",
								},
							},
							Op: ast.UnaryOp(0),
							Expr: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(828),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(828),
											Column: int(29),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8561,
									FreeVars: ast.Identifiers{
										"b",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(828),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(828),
												Column: int(26),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8561,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(828),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(828),
													Column: int(17),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "isNumber",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(828),
															Column: int(27),
														},
														End: ast.Location{
															Line: int(828),
															Column: int(28),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p8607,
													FreeVars: ast.Identifiers{
														"b",
													},
												},
												Id: "b",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
						},
						ThenFodder: ast.Fodder{},
						BranchTrue: &ast.Error{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(829),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(829),
										Column: int(71),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
										Comment: []string{},
									},
								},
								Ctx: p8561,
								FreeVars: ast.Identifiers{
									"b",
									"std",
								},
							},
							Expr: &ast.Binary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(829),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(829),
											Column: int(71),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8561,
									FreeVars: ast.Identifiers{
										"b",
										"std",
									},
								},
								Left: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(829),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(829),
												Column: int(57),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8561,
										FreeVars: nil,
									},
									Value: "std.max second param expected number, got ",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								OpFodder: ast.Fodder{},
								Op: ast.BinaryOp(3),
								Right: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(829),
												Column: int(60),
											},
											End: ast.Location{
												Line: int(829),
												Column: int(71),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8561,
										FreeVars: ast.Identifiers{
											"b",
											"std",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(829),
													Column: int(60),
												},
												End: ast.Location{
													Line: int(829),
													Column: int(68),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p8561,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(829),
														Column: int(60),
													},
													End: ast.Location{
														Line: int(829),
														Column: int(63),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "type",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(829),
																Column: int(69),
															},
															End: ast.Location{
																Line: int(829),
																Column: int(70),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p8624,
														FreeVars: ast.Identifiers{
															"b",
														},
													},
													Id: "b",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
							},
						},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						BranchFalse: &ast.Conditional{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(831),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(831),
										Column: int(29),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
										Comment: []string{},
									},
								},
								Ctx: p8561,
								FreeVars: ast.Identifiers{
									"a",
									"b",
								},
							},
							Cond: &ast.Binary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(831),
											Column: int(10),
										},
										End: ast.Location{
											Line: int(831),
											Column: int(15),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8561,
									FreeVars: ast.Identifiers{
										"a",
										"b",
									},
								},
								Left: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(831),
												Column: int(10),
											},
											End: ast.Location{
												Line: int(831),
												Column: int(11),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8561,
										FreeVars: ast.Identifiers{
											"a",
										},
									},
									Id: "a",
								},
								OpFodder: ast.Fodder{},
								Op: ast.BinaryOp(7),
								Right: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(831),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(831),
												Column: int(15),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8561,
										FreeVars: ast.Identifiers{
											"b",
										},
									},
									Id: "b",
								},
							},
							ThenFodder: ast.Fodder{},
							BranchTrue: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(831),
											Column: int(21),
										},
										End: ast.Location{
											Line: int(831),
											Column: int(22),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8561,
									FreeVars: ast.Identifiers{
										"a",
									},
								},
								Id: "a",
							},
							ElseFodder: ast.Fodder{},
							BranchFalse: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(831),
											Column: int(28),
										},
										End: ast.Location{
											Line: int(831),
											Column: int(29),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8561,
									FreeVars: ast.Identifiers{
										"b",
									},
								},
								Id: "b",
							},
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(825),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(831),
					Column: int(29),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "min",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "a",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(833),
								Column: int(7),
							},
							End: ast.Location{
								Line: int(833),
								Column: int(8),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "b",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(833),
								Column: int(10),
							},
							End: ast.Location{
								Line: int(833),
								Column: int(11),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(834),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(839),
								Column: int(29),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p8646,
						FreeVars: ast.Identifiers{
							"a",
							"b",
							"std",
						},
					},
					Cond: &ast.Unary{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(834),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(834),
									Column: int(24),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p8646,
							FreeVars: ast.Identifiers{
								"a",
								"std",
							},
						},
						Op: ast.UnaryOp(0),
						Expr: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(834),
										Column: int(9),
									},
									End: ast.Location{
										Line: int(834),
										Column: int(24),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p8646,
								FreeVars: ast.Identifiers{
									"a",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(834),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(834),
											Column: int(21),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8646,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(834),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(834),
												Column: int(12),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "isNumber",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(834),
														Column: int(22),
													},
													End: ast.Location{
														Line: int(834),
														Column: int(23),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p8659,
												FreeVars: ast.Identifiers{
													"a",
												},
											},
											Id: "a",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
					},
					ThenFodder: ast.Fodder{},
					BranchTrue: &ast.Error{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(835),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(835),
									Column: int(70),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
									Comment: []string{},
								},
							},
							Ctx: p8646,
							FreeVars: ast.Identifiers{
								"a",
								"std",
							},
						},
						Expr: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(835),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(835),
										Column: int(70),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p8646,
								FreeVars: ast.Identifiers{
									"a",
									"std",
								},
							},
							Left: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(835),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(835),
											Column: int(56),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8646,
									FreeVars: nil,
								},
								Value: "std.max first param expected number, got ",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(3),
							Right: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(835),
											Column: int(59),
										},
										End: ast.Location{
											Line: int(835),
											Column: int(70),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8646,
									FreeVars: ast.Identifiers{
										"a",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(835),
												Column: int(59),
											},
											End: ast.Location{
												Line: int(835),
												Column: int(67),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8646,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(835),
													Column: int(59),
												},
												End: ast.Location{
													Line: int(835),
													Column: int(62),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "type",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(835),
															Column: int(68),
														},
														End: ast.Location{
															Line: int(835),
															Column: int(69),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p8676,
													FreeVars: ast.Identifiers{
														"a",
													},
												},
												Id: "a",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
						},
					},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
							Comment: []string{},
						},
					},
					BranchFalse: &ast.Conditional{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(836),
									Column: int(10),
								},
								End: ast.Location{
									Line: int(839),
									Column: int(29),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p8646,
							FreeVars: ast.Identifiers{
								"a",
								"b",
								"std",
							},
						},
						Cond: &ast.Unary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(836),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(836),
										Column: int(29),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p8646,
								FreeVars: ast.Identifiers{
									"b",
									"std",
								},
							},
							Op: ast.UnaryOp(0),
							Expr: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(836),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(836),
											Column: int(29),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8646,
									FreeVars: ast.Identifiers{
										"b",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(836),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(836),
												Column: int(26),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8646,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(836),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(836),
													Column: int(17),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "isNumber",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(836),
															Column: int(27),
														},
														End: ast.Location{
															Line: int(836),
															Column: int(28),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p8692,
													FreeVars: ast.Identifiers{
														"b",
													},
												},
												Id: "b",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
						},
						ThenFodder: ast.Fodder{},
						BranchTrue: &ast.Error{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(837),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(837),
										Column: int(71),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
										Comment: []string{},
									},
								},
								Ctx: p8646,
								FreeVars: ast.Identifiers{
									"b",
									"std",
								},
							},
							Expr: &ast.Binary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(837),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(837),
											Column: int(71),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8646,
									FreeVars: ast.Identifiers{
										"b",
										"std",
									},
								},
								Left: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(837),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(837),
												Column: int(57),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8646,
										FreeVars: nil,
									},
									Value: "std.max second param expected number, got ",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								OpFodder: ast.Fodder{},
								Op: ast.BinaryOp(3),
								Right: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(837),
												Column: int(60),
											},
											End: ast.Location{
												Line: int(837),
												Column: int(71),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8646,
										FreeVars: ast.Identifiers{
											"b",
											"std",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(837),
													Column: int(60),
												},
												End: ast.Location{
													Line: int(837),
													Column: int(68),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p8646,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(837),
														Column: int(60),
													},
													End: ast.Location{
														Line: int(837),
														Column: int(63),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "type",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(837),
																Column: int(69),
															},
															End: ast.Location{
																Line: int(837),
																Column: int(70),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p8709,
														FreeVars: ast.Identifiers{
															"b",
														},
													},
													Id: "b",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
							},
						},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						BranchFalse: &ast.Conditional{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(839),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(839),
										Column: int(29),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
										Comment: []string{},
									},
								},
								Ctx: p8646,
								FreeVars: ast.Identifiers{
									"a",
									"b",
								},
							},
							Cond: &ast.Binary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(839),
											Column: int(10),
										},
										End: ast.Location{
											Line: int(839),
											Column: int(15),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8646,
									FreeVars: ast.Identifiers{
										"a",
										"b",
									},
								},
								Left: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(839),
												Column: int(10),
											},
											End: ast.Location{
												Line: int(839),
												Column: int(11),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8646,
										FreeVars: ast.Identifiers{
											"a",
										},
									},
									Id: "a",
								},
								OpFodder: ast.Fodder{},
								Op: ast.BinaryOp(9),
								Right: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(839),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(839),
												Column: int(15),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8646,
										FreeVars: ast.Identifiers{
											"b",
										},
									},
									Id: "b",
								},
							},
							ThenFodder: ast.Fodder{},
							BranchTrue: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(839),
											Column: int(21),
										},
										End: ast.Location{
											Line: int(839),
											Column: int(22),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8646,
									FreeVars: ast.Identifiers{
										"a",
									},
								},
								Id: "a",
							},
							ElseFodder: ast.Fodder{},
							BranchFalse: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(839),
											Column: int(28),
										},
										End: ast.Location{
											Line: int(839),
											Column: int(29),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8646,
									FreeVars: ast.Identifiers{
										"b",
									},
								},
								Id: "b",
							},
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(833),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(839),
					Column: int(29),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "clamp",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: nil,
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "x",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(841),
								Column: int(9),
							},
							End: ast.Location{
								Line: int(841),
								Column: int(10),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "minVal",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(841),
								Column: int(12),
							},
							End: ast.Location{
								Line: int(841),
								Column: int(18),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "maxVal",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(841),
								Column: int(20),
							},
							End: ast.Location{
								Line: int(841),
								Column: int(26),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(842),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(844),
								Column: int(11),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p8730,
						FreeVars: ast.Identifiers{
							"maxVal",
							"minVal",
							"x",
						},
					},
					Cond: &ast.Binary{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(842),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(842),
									Column: int(19),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p8730,
							FreeVars: ast.Identifiers{
								"minVal",
								"x",
							},
						},
						Left: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(842),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(842),
										Column: int(9),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p8730,
								FreeVars: ast.Identifiers{
									"x",
								},
							},
							Id: "x",
						},
						OpFodder: ast.Fodder{},
						Op: ast.BinaryOp(9),
						Right: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(842),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(842),
										Column: int(19),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p8730,
								FreeVars: ast.Identifiers{
									"minVal",
								},
							},
							Id: "minVal",
						},
					},
					ThenFodder: ast.Fodder{},
					BranchTrue: &ast.Var{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(842),
									Column: int(25),
								},
								End: ast.Location{
									Line: int(842),
									Column: int(31),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p8730,
							FreeVars: ast.Identifiers{
								"minVal",
							},
						},
						Id: "minVal",
					},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
							Comment: []string{},
						},
					},
					BranchFalse: &ast.Conditional{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(843),
									Column: int(10),
								},
								End: ast.Location{
									Line: int(844),
									Column: int(11),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p8730,
							FreeVars: ast.Identifiers{
								"maxVal",
								"x",
							},
						},
						Cond: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(843),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(843),
										Column: int(23),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p8730,
								FreeVars: ast.Identifiers{
									"maxVal",
									"x",
								},
							},
							Left: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(843),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(843),
											Column: int(14),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8730,
									FreeVars: ast.Identifiers{
										"x",
									},
								},
								Id: "x",
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(7),
							Right: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(843),
											Column: int(17),
										},
										End: ast.Location{
											Line: int(843),
											Column: int(23),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8730,
									FreeVars: ast.Identifiers{
										"maxVal",
									},
								},
								Id: "maxVal",
							},
						},
						ThenFodder: ast.Fodder{},
						BranchTrue: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(843),
										Column: int(29),
									},
									End: ast.Location{
										Line: int(843),
										Column: int(35),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p8730,
								FreeVars: ast.Identifiers{
									"maxVal",
								},
							},
							Id: "maxVal",
						},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						BranchFalse: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(844),
										Column: int(10),
									},
									End: ast.Location{
										Line: int(844),
										Column: int(11),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p8730,
								FreeVars: ast.Identifiers{
									"x",
								},
							},
							Id: "x",
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(841),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(844),
					Column: int(11),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "flattenArrays",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arrs",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(846),
								Column: int(17),
							},
							End: ast.Location{
								Line: int(846),
								Column: int(21),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(847),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(847),
								Column: int(46),
							},
							File: p1,
						},
						Fodder: ast.Fodder{},
						Ctx: p8759,
						FreeVars: ast.Identifiers{
							"arrs",
							"std",
						},
					},
					Target: &ast.Index{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(847),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(847),
									Column: int(14),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p8759,
							FreeVars: ast.Identifiers{
								"std",
							},
						},
						Target: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(847),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(847),
										Column: int(8),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
										Comment: []string{},
									},
								},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Id: "std",
						},
						LeftBracketFodder: ast.Fodder{},
						Index: &ast.LiteralString{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: nil,
							},
							Value: "foldl",
							Kind: ast.LiteralStringKind(1),
							BlockIndent: "",
							BlockTermIndent: "",
						},
						RightBracketFodder: ast.Fodder{},
						Id: nil,
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Function{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(847),
												Column: int(15),
											},
											End: ast.Location{
												Line: int(847),
												Column: int(35),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8769,
										FreeVars: nil,
									},
									ParenLeftFodder: ast.Fodder{},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "a",
											EqFodder: nil,
											DefaultArg: nil,
											CommaFodder: ast.Fodder{},
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(847),
													Column: int(24),
												},
												End: ast.Location{
													Line: int(847),
													Column: int(25),
												},
												File: p1,
											},
										},
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "b",
											EqFodder: nil,
											DefaultArg: nil,
											CommaFodder: nil,
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(847),
													Column: int(27),
												},
												End: ast.Location{
													Line: int(847),
													Column: int(28),
												},
												File: p1,
											},
										},
									},
									TrailingComma: false,
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Binary{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(847),
													Column: int(30),
												},
												End: ast.Location{
													Line: int(847),
													Column: int(35),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p8772,
											FreeVars: ast.Identifiers{
												"a",
												"b",
											},
										},
										Left: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(847),
														Column: int(30),
													},
													End: ast.Location{
														Line: int(847),
														Column: int(31),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p8772,
												FreeVars: ast.Identifiers{
													"a",
												},
											},
											Id: "a",
										},
										OpFodder: ast.Fodder{},
										Op: ast.BinaryOp(3),
										Right: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(847),
														Column: int(34),
													},
													End: ast.Location{
														Line: int(847),
														Column: int(35),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p8772,
												FreeVars: ast.Identifiers{
													"b",
												},
											},
											Id: "b",
										},
									},
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(847),
												Column: int(37),
											},
											End: ast.Location{
												Line: int(847),
												Column: int(41),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8769,
										FreeVars: ast.Identifiers{
											"arrs",
										},
									},
									Id: "arrs",
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.Array{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(847),
												Column: int(43),
											},
											End: ast.Location{
												Line: int(847),
												Column: int(45),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8769,
										FreeVars: nil,
									},
									Elements: nil,
									TrailingComma: false,
									CloseFodder: ast.Fodder{},
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					TrailingComma: false,
					TailStrict: false,
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(846),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(847),
					Column: int(46),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "manifestIni",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "ini",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(849),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(849),
								Column: int(18),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(850),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(867),
								Column: int(71),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p8787,
						FreeVars: ast.Identifiers{
							"ini",
							"std",
						},
					},
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: nil,
							Variable: "body_lines",
							EqFodder: nil,
							Body: &ast.Function{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(850),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(859),
											Column: int(9),
										},
										File: p1,
									},
									Fodder: nil,
									Ctx: p8791,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								ParenLeftFodder: ast.Fodder{},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "body",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: nil,
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(850),
												Column: int(22),
											},
											End: ast.Location{
												Line: int(850),
												Column: int(26),
											},
											File: p1,
										},
									},
								},
								TrailingComma: false,
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(851),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(859),
												Column: int(9),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8795,
										FreeVars: ast.Identifiers{
											"body",
											"std",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(851),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(851),
													Column: int(15),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p8795,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(851),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(851),
														Column: int(10),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
														Comment: []string{},
													},
												},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "join",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Array{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(851),
																Column: int(16),
															},
															End: ast.Location{
																Line: int(851),
																Column: int(18),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p8805,
														FreeVars: nil,
													},
													Elements: nil,
													TrailingComma: false,
													CloseFodder: ast.Fodder{},
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"body",
															"std",
														},
													},
													Target: &ast.Index{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Id: "std",
														},
														LeftBracketFodder: nil,
														Index: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
															Value: "flatMap",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														RightBracketFodder: nil,
														Id: nil,
													},
													FodderLeft: nil,
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Function{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"body",
																			"std",
																		},
																	},
																	ParenLeftFodder: nil,
																	Parameters: []ast.Parameter{
																		ast.Parameter{
																			NameFodder: nil,
																			Name: "k",
																			EqFodder: nil,
																			DefaultArg: nil,
																			CommaFodder: nil,
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																		},
																	},
																	TrailingComma: false,
																	ParenRightFodder: nil,
																	Body: &ast.Array{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"body",
																				"k",
																				"std",
																			},
																		},
																		Elements: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Local{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(852),
																								Column: int(9),
																							},
																							End: ast.Location{
																								Line: int(856),
																								Column: int(45),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(8),
																								Comment: []string{},
																							},
																						},
																						Ctx: p8822,
																						FreeVars: ast.Identifiers{
																							"body",
																							"k",
																							"std",
																						},
																					},
																					Binds: ast.LocalBinds{
																						ast.LocalBind{
																							VarFodder: ast.Fodder{},
																							Variable: "value_or_values",
																							EqFodder: ast.Fodder{},
																							Body: &ast.Index{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(852),
																											Column: int(33),
																										},
																										End: ast.Location{
																											Line: int(852),
																											Column: int(40),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p8826,
																									FreeVars: ast.Identifiers{
																										"body",
																										"k",
																									},
																								},
																								Target: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(852),
																												Column: int(33),
																											},
																											End: ast.Location{
																												Line: int(852),
																												Column: int(37),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p8826,
																										FreeVars: ast.Identifiers{
																											"body",
																										},
																									},
																									Id: "body",
																								},
																								LeftBracketFodder: ast.Fodder{},
																								Index: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(852),
																												Column: int(38),
																											},
																											End: ast.Location{
																												Line: int(852),
																												Column: int(39),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p8826,
																										FreeVars: ast.Identifiers{
																											"k",
																										},
																									},
																									Id: "k",
																								},
																								RightBracketFodder: ast.Fodder{},
																								Id: nil,
																							},
																							Fun: nil,
																							CloseFodder: ast.Fodder{},
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(852),
																									Column: int(15),
																								},
																								End: ast.Location{
																									Line: int(852),
																									Column: int(40),
																								},
																								File: p1,
																							},
																						},
																					},
																					Body: &ast.Conditional{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(853),
																									Column: int(9),
																								},
																								End: ast.Location{
																									Line: int(856),
																									Column: int(45),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{
																								ast.FodderElement{
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(8),
																									Comment: []string{},
																								},
																							},
																							Ctx: p8822,
																							FreeVars: ast.Identifiers{
																								"k",
																								"std",
																								"value_or_values",
																							},
																						},
																						Cond: &ast.Apply{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(853),
																										Column: int(12),
																									},
																									End: ast.Location{
																										Line: int(853),
																										Column: int(40),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p8822,
																								FreeVars: ast.Identifiers{
																									"std",
																									"value_or_values",
																								},
																							},
																							Target: &ast.Index{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(853),
																											Column: int(12),
																										},
																										End: ast.Location{
																											Line: int(853),
																											Column: int(23),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p8822,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																								},
																								Target: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(853),
																												Column: int(12),
																											},
																											End: ast.Location{
																												Line: int(853),
																												Column: int(15),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: nil,
																										FreeVars: ast.Identifiers{
																											"std",
																										},
																									},
																									Id: "std",
																								},
																								LeftBracketFodder: ast.Fodder{},
																								Index: &ast.LiteralString{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											File: nil,
																										},
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: nil,
																									},
																									Value: "isArray",
																									Kind: ast.LiteralStringKind(1),
																									BlockIndent: "",
																									BlockTermIndent: "",
																								},
																								RightBracketFodder: ast.Fodder{},
																								Id: nil,
																							},
																							FodderLeft: ast.Fodder{},
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(853),
																														Column: int(24),
																													},
																													End: ast.Location{
																														Line: int(853),
																														Column: int(39),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p8844,
																												FreeVars: ast.Identifiers{
																													"value_or_values",
																												},
																											},
																											Id: "value_or_values",
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							TrailingComma: false,
																							TailStrict: false,
																							FodderRight: ast.Fodder{},
																							TailStrictFodder: nil,
																						},
																						ThenFodder: ast.Fodder{},
																						BranchTrue: &ast.Apply{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									File: nil,
																								},
																								Fodder: nil,
																								Ctx: nil,
																								FreeVars: ast.Identifiers{
																									"k",
																									"std",
																									"value_or_values",
																								},
																							},
																							Target: &ast.Index{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										File: nil,
																									},
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																								},
																								Target: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											File: nil,
																										},
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: ast.Identifiers{
																											"std",
																										},
																									},
																									Id: "std",
																								},
																								LeftBracketFodder: nil,
																								Index: &ast.LiteralString{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											File: nil,
																										},
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: nil,
																									},
																									Value: "flatMap",
																									Kind: ast.LiteralStringKind(1),
																									BlockIndent: "",
																									BlockTermIndent: "",
																								},
																								RightBracketFodder: nil,
																								Id: nil,
																							},
																							FodderLeft: nil,
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Function{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													File: nil,
																												},
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"k",
																													"std",
																												},
																											},
																											ParenLeftFodder: nil,
																											Parameters: []ast.Parameter{
																												ast.Parameter{
																													NameFodder: nil,
																													Name: "value",
																													EqFodder: nil,
																													DefaultArg: nil,
																													CommaFodder: nil,
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																												},
																											},
																											TrailingComma: false,
																											ParenRightFodder: nil,
																											Body: &ast.Array{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"k",
																														"std",
																														"value",
																													},
																												},
																												Elements: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Apply{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	File: nil,
																																},
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"k",
																																	"std",
																																	"value",
																																},
																															},
																															Target: &ast.Index{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		File: nil,
																																	},
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																	},
																																},
																																Target: &ast.Var{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			End: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			File: nil,
																																		},
																																		Fodder: nil,
																																		Ctx: nil,
																																		FreeVars: ast.Identifiers{
																																			"std",
																																		},
																																	},
																																	Id: "std",
																																},
																																LeftBracketFodder: nil,
																																Index: &ast.LiteralString{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "",
																																			Begin: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			End: ast.Location{
																																				Line: int(0),
																																				Column: int(0),
																																			},
																																			File: nil,
																																		},
																																		Fodder: nil,
																																		Ctx: nil,
																																		FreeVars: nil,
																																	},
																																	Value: "mod",
																																	Kind: ast.LiteralStringKind(1),
																																	BlockIndent: "",
																																	BlockTermIndent: "",
																																},
																																RightBracketFodder: nil,
																																Id: nil,
																															},
																															FodderLeft: nil,
																															Arguments: ast.Arguments{
																																Positional: []ast.CommaSeparatedExpr{
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.LiteralString{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(854),
																																						Column: int(12),
																																					},
																																					End: ast.Location{
																																						Line: int(854),
																																						Column: int(21),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p8869,
																																				FreeVars: nil,
																																			},
																																			Value: "%s = %s",
																																			Kind: ast.LiteralStringKind(1),
																																			BlockIndent: "",
																																			BlockTermIndent: "",
																																		},
																																		CommaFodder: nil,
																																	},
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Array{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(854),
																																						Column: int(24),
																																					},
																																					End: ast.Location{
																																						Line: int(854),
																																						Column: int(34),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p8869,
																																				FreeVars: ast.Identifiers{
																																					"k",
																																					"value",
																																				},
																																			},
																																			Elements: []ast.CommaSeparatedExpr{
																																				ast.CommaSeparatedExpr{
																																					Expr: &ast.Var{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(854),
																																									Column: int(25),
																																								},
																																								End: ast.Location{
																																									Line: int(854),
																																									Column: int(26),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p8874,
																																							FreeVars: ast.Identifiers{
																																								"k",
																																							},
																																						},
																																						Id: "k",
																																					},
																																					CommaFodder: ast.Fodder{},
																																				},
																																				ast.CommaSeparatedExpr{
																																					Expr: &ast.Var{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(854),
																																									Column: int(28),
																																								},
																																								End: ast.Location{
																																									Line: int(854),
																																									Column: int(33),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p8874,
																																							FreeVars: ast.Identifiers{
																																								"value",
																																							},
																																						},
																																						Id: "value",
																																					},
																																					CommaFodder: nil,
																																				},
																																			},
																																			TrailingComma: false,
																																			CloseFodder: ast.Fodder{},
																																		},
																																		CommaFodder: nil,
																																	},
																																},
																																Named: nil,
																															},
																															TrailingComma: false,
																															TailStrict: false,
																															FodderRight: nil,
																															TailStrictFodder: nil,
																														},
																														CommaFodder: nil,
																													},
																												},
																												TrailingComma: false,
																												CloseFodder: nil,
																											},
																										},
																										CommaFodder: nil,
																									},
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(854),
																														Column: int(48),
																													},
																													End: ast.Location{
																														Line: int(854),
																														Column: int(63),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p8822,
																												FreeVars: ast.Identifiers{
																													"value_or_values",
																												},
																											},
																											Id: "value_or_values",
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							TrailingComma: false,
																							TailStrict: false,
																							FodderRight: nil,
																							TailStrictFodder: nil,
																						},
																						ElseFodder: ast.Fodder{
																							ast.FodderElement{
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(8),
																								Comment: []string{},
																							},
																						},
																						BranchFalse: &ast.Array{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(856),
																										Column: int(11),
																									},
																									End: ast.Location{
																										Line: int(856),
																										Column: int(45),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{
																									ast.FodderElement{
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(10),
																										Comment: []string{},
																									},
																								},
																								Ctx: p8822,
																								FreeVars: ast.Identifiers{
																									"k",
																									"std",
																									"value_or_values",
																								},
																							},
																							Elements: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Apply{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												File: nil,
																											},
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"k",
																												"std",
																												"value_or_values",
																											},
																										},
																										Target: &ast.Index{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													File: nil,
																												},
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																											},
																											Target: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																												},
																												Id: "std",
																											},
																											LeftBracketFodder: nil,
																											Index: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: nil,
																												},
																												Value: "mod",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																											RightBracketFodder: nil,
																											Id: nil,
																										},
																										FodderLeft: nil,
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(856),
																																	Column: int(12),
																																},
																																End: ast.Location{
																																	Line: int(856),
																																	Column: int(21),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p8894,
																															FreeVars: nil,
																														},
																														Value: "%s = %s",
																														Kind: ast.LiteralStringKind(1),
																														BlockIndent: "",
																														BlockTermIndent: "",
																													},
																													CommaFodder: nil,
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Array{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(856),
																																	Column: int(24),
																																},
																																End: ast.Location{
																																	Line: int(856),
																																	Column: int(44),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p8894,
																															FreeVars: ast.Identifiers{
																																"k",
																																"value_or_values",
																															},
																														},
																														Elements: []ast.CommaSeparatedExpr{
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Var{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(856),
																																				Column: int(25),
																																			},
																																			End: ast.Location{
																																				Line: int(856),
																																				Column: int(26),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p8899,
																																		FreeVars: ast.Identifiers{
																																			"k",
																																		},
																																	},
																																	Id: "k",
																																},
																																CommaFodder: ast.Fodder{},
																															},
																															ast.CommaSeparatedExpr{
																																Expr: &ast.Var{
																																	NodeBase: ast.NodeBase{
																																		LocRange: ast.LocationRange{
																																			FileName: "<std>",
																																			Begin: ast.Location{
																																				Line: int(856),
																																				Column: int(28),
																																			},
																																			End: ast.Location{
																																				Line: int(856),
																																				Column: int(43),
																																			},
																																			File: p1,
																																		},
																																		Fodder: ast.Fodder{},
																																		Ctx: p8899,
																																		FreeVars: ast.Identifiers{
																																			"value_or_values",
																																		},
																																	},
																																	Id: "value_or_values",
																																},
																																CommaFodder: nil,
																															},
																														},
																														TrailingComma: false,
																														CloseFodder: ast.Fodder{},
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										TrailingComma: false,
																										TailStrict: false,
																										FodderRight: nil,
																										TailStrictFodder: nil,
																									},
																									CommaFodder: nil,
																								},
																							},
																							TrailingComma: false,
																							CloseFodder: ast.Fodder{},
																						},
																					},
																				},
																				CommaFodder: nil,
																			},
																		},
																		TrailingComma: false,
																		CloseFodder: nil,
																	},
																},
																CommaFodder: nil,
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Apply{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(858),
																				Column: int(18),
																			},
																			End: ast.Location{
																				Line: int(858),
																				Column: int(40),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p8805,
																		FreeVars: ast.Identifiers{
																			"body",
																			"std",
																		},
																	},
																	Target: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(858),
																					Column: int(18),
																				},
																				End: ast.Location{
																					Line: int(858),
																					Column: int(34),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p8805,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(858),
																						Column: int(18),
																					},
																					End: ast.Location{
																						Line: int(858),
																						Column: int(21),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Id: "std",
																		},
																		LeftBracketFodder: ast.Fodder{},
																		Index: &ast.LiteralString{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: nil,
																			},
																			Value: "objectFields",
																			Kind: ast.LiteralStringKind(1),
																			BlockIndent: "",
																			BlockTermIndent: "",
																		},
																		RightBracketFodder: ast.Fodder{},
																		Id: nil,
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(858),
																								Column: int(35),
																							},
																							End: ast.Location{
																								Line: int(858),
																								Column: int(39),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p8912,
																						FreeVars: ast.Identifiers{
																							"body",
																						},
																					},
																					Id: "body",
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	TrailingComma: false,
																	TailStrict: false,
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: nil,
													TailStrictFodder: nil,
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
							},
							Fun: nil,
							CloseFodder: nil,
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
						},
					},
					Body: &ast.Local{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(861),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(867),
									Column: int(71),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(1),
									Indent: int(4),
									Comment: []string{},
								},
							},
							Ctx: p8787,
							FreeVars: ast.Identifiers{
								"body_lines",
								"ini",
								"std",
							},
						},
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: nil,
								Variable: "section_lines",
								EqFodder: nil,
								Body: &ast.Function{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(861),
												Column: int(11),
											},
											End: ast.Location{
												Line: int(861),
												Column: int(79),
											},
											File: p1,
										},
										Fodder: nil,
										Ctx: p8919,
										FreeVars: ast.Identifiers{
											"body_lines",
											"std",
										},
									},
									ParenLeftFodder: ast.Fodder{},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "sname",
											EqFodder: nil,
											DefaultArg: nil,
											CommaFodder: ast.Fodder{},
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(861),
													Column: int(25),
												},
												End: ast.Location{
													Line: int(861),
													Column: int(30),
												},
												File: p1,
											},
										},
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "sbody",
											EqFodder: nil,
											DefaultArg: nil,
											CommaFodder: nil,
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(861),
													Column: int(32),
												},
												End: ast.Location{
													Line: int(861),
													Column: int(37),
												},
												File: p1,
											},
										},
									},
									TrailingComma: false,
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Binary{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(861),
													Column: int(41),
												},
												End: ast.Location{
													Line: int(861),
													Column: int(79),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p8923,
											FreeVars: ast.Identifiers{
												"body_lines",
												"sbody",
												"sname",
												"std",
											},
										},
										Left: &ast.Array{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(861),
														Column: int(41),
													},
													End: ast.Location{
														Line: int(861),
														Column: int(59),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p8923,
												FreeVars: ast.Identifiers{
													"sname",
													"std",
												},
											},
											Elements: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"sname",
																"std",
															},
														},
														Target: &ast.Index{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Target: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Id: "std",
															},
															LeftBracketFodder: nil,
															Index: &ast.LiteralString{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: nil,
																},
																Value: "mod",
																Kind: ast.LiteralStringKind(1),
																BlockIndent: "",
																BlockTermIndent: "",
															},
															RightBracketFodder: nil,
															Id: nil,
														},
														FodderLeft: nil,
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(861),
																					Column: int(42),
																				},
																				End: ast.Location{
																					Line: int(861),
																					Column: int(48),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p8937,
																			FreeVars: nil,
																		},
																		Value: "[%s]",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	CommaFodder: nil,
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Array{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(861),
																					Column: int(51),
																				},
																				End: ast.Location{
																					Line: int(861),
																					Column: int(58),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p8937,
																			FreeVars: ast.Identifiers{
																				"sname",
																			},
																		},
																		Elements: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(861),
																								Column: int(52),
																							},
																							End: ast.Location{
																								Line: int(861),
																								Column: int(57),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p8942,
																						FreeVars: ast.Identifiers{
																							"sname",
																						},
																					},
																					Id: "sname",
																				},
																				CommaFodder: nil,
																			},
																		},
																		TrailingComma: false,
																		CloseFodder: ast.Fodder{},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														TrailingComma: false,
														TailStrict: false,
														FodderRight: nil,
														TailStrictFodder: nil,
													},
													CommaFodder: nil,
												},
											},
											TrailingComma: false,
											CloseFodder: ast.Fodder{},
										},
										OpFodder: ast.Fodder{},
										Op: ast.BinaryOp(3),
										Right: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(861),
														Column: int(62),
													},
													End: ast.Location{
														Line: int(861),
														Column: int(79),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p8923,
												FreeVars: ast.Identifiers{
													"body_lines",
													"sbody",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(861),
															Column: int(62),
														},
														End: ast.Location{
															Line: int(861),
															Column: int(72),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p8923,
													FreeVars: ast.Identifiers{
														"body_lines",
													},
												},
												Id: "body_lines",
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(861),
																		Column: int(73),
																	},
																	End: ast.Location{
																		Line: int(861),
																		Column: int(78),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p8950,
																FreeVars: ast.Identifiers{
																	"sbody",
																},
															},
															Id: "sbody",
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
									},
								},
								Fun: nil,
								CloseFodder: nil,
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
							},
							ast.LocalBind{
								VarFodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(10),
										Comment: []string{},
									},
								},
								Variable: "main_body",
								EqFodder: ast.Fodder{},
								Body: &ast.Conditional{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(862),
												Column: int(23),
											},
											End: ast.Location{
												Line: int(862),
												Column: int(86),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p8954,
										FreeVars: ast.Identifiers{
											"body_lines",
											"ini",
											"std",
										},
									},
									Cond: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(862),
													Column: int(26),
												},
												End: ast.Location{
													Line: int(862),
													Column: int(52),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p8954,
											FreeVars: ast.Identifiers{
												"ini",
												"std",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(862),
														Column: int(26),
													},
													End: ast.Location{
														Line: int(862),
														Column: int(39),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p8954,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(862),
															Column: int(26),
														},
														End: ast.Location{
															Line: int(862),
															Column: int(29),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: ast.Fodder{},
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "objectHas",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: ast.Fodder{},
											Id: nil,
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(862),
																	Column: int(40),
																},
																End: ast.Location{
																	Line: int(862),
																	Column: int(43),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p8965,
															FreeVars: ast.Identifiers{
																"ini",
															},
														},
														Id: "ini",
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(862),
																	Column: int(45),
																},
																End: ast.Location{
																	Line: int(862),
																	Column: int(51),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p8965,
															FreeVars: nil,
														},
														Value: "main",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
									ThenFodder: ast.Fodder{},
									BranchTrue: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(862),
													Column: int(58),
												},
												End: ast.Location{
													Line: int(862),
													Column: int(78),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p8954,
											FreeVars: ast.Identifiers{
												"body_lines",
												"ini",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(862),
														Column: int(58),
													},
													End: ast.Location{
														Line: int(862),
														Column: int(68),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p8954,
												FreeVars: ast.Identifiers{
													"body_lines",
												},
											},
											Id: "body_lines",
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Index{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(862),
																	Column: int(69),
																},
																End: ast.Location{
																	Line: int(862),
																	Column: int(77),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p8974,
															FreeVars: ast.Identifiers{
																"ini",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(862),
																		Column: int(69),
																	},
																	End: ast.Location{
																		Line: int(862),
																		Column: int(72),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"ini",
																},
															},
															Id: "ini",
														},
														LeftBracketFodder: ast.Fodder{},
														Index: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
															Value: "main",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														RightBracketFodder: ast.Fodder{},
														Id: nil,
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
									ElseFodder: ast.Fodder{},
									BranchFalse: &ast.Array{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(862),
													Column: int(84),
												},
												End: ast.Location{
													Line: int(862),
													Column: int(86),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p8954,
											FreeVars: nil,
										},
										Elements: nil,
										TrailingComma: false,
										CloseFodder: ast.Fodder{},
									},
								},
								Fun: nil,
								CloseFodder: ast.Fodder{},
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(862),
										Column: int(11),
									},
									End: ast.Location{
										Line: int(862),
										Column: int(86),
									},
									File: p1,
								},
							},
							ast.LocalBind{
								VarFodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(10),
										Comment: []string{},
									},
								},
								Variable: "all_sections",
								EqFodder: ast.Fodder{},
								Body: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"ini",
											"section_lines",
											"std",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: nil,
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "flatMap",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: nil,
										Id: nil,
									},
									FodderLeft: nil,
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Function{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"ini",
															"section_lines",
														},
													},
													ParenLeftFodder: nil,
													Parameters: []ast.Parameter{
														ast.Parameter{
															NameFodder: nil,
															Name: "k",
															EqFodder: nil,
															DefaultArg: nil,
															CommaFodder: nil,
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
														},
													},
													TrailingComma: false,
													ParenRightFodder: nil,
													Body: &ast.Array{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"ini",
																"k",
																"section_lines",
															},
														},
														Elements: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Apply{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(864),
																				Column: int(7),
																			},
																			End: ast.Location{
																				Line: int(864),
																				Column: int(40),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p8996,
																		FreeVars: ast.Identifiers{
																			"ini",
																			"k",
																			"section_lines",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(864),
																					Column: int(7),
																				},
																				End: ast.Location{
																					Line: int(864),
																					Column: int(20),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(6),
																					Comment: []string{},
																				},
																			},
																			Ctx: p8996,
																			FreeVars: ast.Identifiers{
																				"section_lines",
																			},
																		},
																		Id: "section_lines",
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(864),
																								Column: int(21),
																							},
																							End: ast.Location{
																								Line: int(864),
																								Column: int(22),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p9003,
																						FreeVars: ast.Identifiers{
																							"k",
																						},
																					},
																					Id: "k",
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Index{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(864),
																								Column: int(24),
																							},
																							End: ast.Location{
																								Line: int(864),
																								Column: int(39),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p9003,
																						FreeVars: ast.Identifiers{
																							"ini",
																							"k",
																						},
																					},
																					Target: &ast.Index{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(864),
																									Column: int(24),
																								},
																								End: ast.Location{
																									Line: int(864),
																									Column: int(36),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p9003,
																							FreeVars: ast.Identifiers{
																								"ini",
																							},
																						},
																						Target: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(864),
																										Column: int(24),
																									},
																									End: ast.Location{
																										Line: int(864),
																										Column: int(27),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: nil,
																								FreeVars: ast.Identifiers{
																									"ini",
																								},
																							},
																							Id: "ini",
																						},
																						LeftBracketFodder: ast.Fodder{},
																						Index: &ast.LiteralString{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									File: nil,
																								},
																								Fodder: nil,
																								Ctx: nil,
																								FreeVars: nil,
																							},
																							Value: "sections",
																							Kind: ast.LiteralStringKind(1),
																							BlockIndent: "",
																							BlockTermIndent: "",
																						},
																						RightBracketFodder: ast.Fodder{},
																						Id: nil,
																					},
																					LeftBracketFodder: ast.Fodder{},
																					Index: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(864),
																									Column: int(37),
																								},
																								End: ast.Location{
																									Line: int(864),
																									Column: int(38),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p9003,
																							FreeVars: ast.Identifiers{
																								"k",
																							},
																						},
																						Id: "k",
																					},
																					RightBracketFodder: ast.Fodder{},
																					Id: nil,
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	TrailingComma: false,
																	TailStrict: false,
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																},
																CommaFodder: nil,
															},
														},
														TrailingComma: false,
														CloseFodder: nil,
													},
												},
												CommaFodder: nil,
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(865),
																Column: int(16),
															},
															End: ast.Location{
																Line: int(865),
																Column: int(46),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p9015,
														FreeVars: ast.Identifiers{
															"ini",
															"std",
														},
													},
													Target: &ast.Index{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(865),
																	Column: int(16),
																},
																End: ast.Location{
																	Line: int(865),
																	Column: int(32),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p9015,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(865),
																		Column: int(16),
																	},
																	End: ast.Location{
																		Line: int(865),
																		Column: int(19),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Id: "std",
														},
														LeftBracketFodder: ast.Fodder{},
														Index: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
															Value: "objectFields",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														RightBracketFodder: ast.Fodder{},
														Id: nil,
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(865),
																				Column: int(33),
																			},
																			End: ast.Location{
																				Line: int(865),
																				Column: int(45),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p9024,
																		FreeVars: ast.Identifiers{
																			"ini",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(865),
																					Column: int(33),
																				},
																				End: ast.Location{
																					Line: int(865),
																					Column: int(36),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"ini",
																			},
																		},
																		Id: "ini",
																	},
																	LeftBracketFodder: ast.Fodder{},
																	Index: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: nil,
																		},
																		Value: "sections",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	RightBracketFodder: ast.Fodder{},
																	Id: nil,
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: nil,
									TailStrictFodder: nil,
								},
								Fun: nil,
								CloseFodder: ast.Fodder{},
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(863),
										Column: int(11),
									},
									End: ast.Location{
										Line: int(866),
										Column: int(6),
									},
									File: p1,
								},
							},
						},
						Body: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(867),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(867),
										Column: int(71),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p8787,
								FreeVars: ast.Identifiers{
									"all_sections",
									"main_body",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(867),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(867),
											Column: int(13),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p8787,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(867),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(867),
												Column: int(8),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(4),
												Comment: []string{},
											},
										},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "join",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(867),
														Column: int(14),
													},
													End: ast.Location{
														Line: int(867),
														Column: int(18),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p9039,
												FreeVars: nil,
											},
											Value: "\n",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Binary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(867),
														Column: int(20),
													},
													End: ast.Location{
														Line: int(867),
														Column: int(70),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p9039,
												FreeVars: ast.Identifiers{
													"all_sections",
													"main_body",
													"std",
												},
											},
											Left: &ast.Binary{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(867),
															Column: int(20),
														},
														End: ast.Location{
															Line: int(867),
															Column: int(63),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p9039,
													FreeVars: ast.Identifiers{
														"all_sections",
														"main_body",
														"std",
													},
												},
												Left: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(867),
																Column: int(20),
															},
															End: ast.Location{
																Line: int(867),
																Column: int(29),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p9039,
														FreeVars: ast.Identifiers{
															"main_body",
														},
													},
													Id: "main_body",
												},
												OpFodder: ast.Fodder{},
												Op: ast.BinaryOp(3),
												Right: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(867),
																Column: int(32),
															},
															End: ast.Location{
																Line: int(867),
																Column: int(63),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p9039,
														FreeVars: ast.Identifiers{
															"all_sections",
															"std",
														},
													},
													Target: &ast.Index{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(867),
																	Column: int(32),
																},
																End: ast.Location{
																	Line: int(867),
																	Column: int(49),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p9039,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(867),
																		Column: int(32),
																	},
																	End: ast.Location{
																		Line: int(867),
																		Column: int(35),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Id: "std",
														},
														LeftBracketFodder: ast.Fodder{},
														Index: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
															Value: "flattenArrays",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														RightBracketFodder: ast.Fodder{},
														Id: nil,
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(867),
																				Column: int(50),
																			},
																			End: ast.Location{
																				Line: int(867),
																				Column: int(62),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p9055,
																		FreeVars: ast.Identifiers{
																			"all_sections",
																		},
																	},
																	Id: "all_sections",
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
												},
											},
											OpFodder: ast.Fodder{},
											Op: ast.BinaryOp(3),
											Right: &ast.Array{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(867),
															Column: int(66),
														},
														End: ast.Location{
															Line: int(867),
															Column: int(70),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p9039,
													FreeVars: nil,
												},
												Elements: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(867),
																		Column: int(67),
																	},
																	End: ast.Location{
																		Line: int(867),
																		Column: int(69),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p9060,
																FreeVars: nil,
															},
															Value: "",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														CommaFodder: nil,
													},
												},
												TrailingComma: false,
												CloseFodder: ast.Fodder{},
											},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(849),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(867),
					Column: int(71),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "escapeStringJson",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str_",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(869),
								Column: int(20),
							},
							End: ast.Location{
								Line: int(869),
								Column: int(24),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(870),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(892),
								Column: int(70),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p9067,
						FreeVars: ast.Identifiers{
							"std",
							"str_",
						},
					},
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Variable: "str",
							EqFodder: ast.Fodder{},
							Body: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(870),
											Column: int(17),
										},
										End: ast.Location{
											Line: int(870),
											Column: int(35),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p9071,
									FreeVars: ast.Identifiers{
										"std",
										"str_",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(870),
												Column: int(17),
											},
											End: ast.Location{
												Line: int(870),
												Column: int(29),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p9071,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(870),
													Column: int(17),
												},
												End: ast.Location{
													Line: int(870),
													Column: int(20),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "toString",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(870),
															Column: int(30),
														},
														End: ast.Location{
															Line: int(870),
															Column: int(34),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p9080,
													FreeVars: ast.Identifiers{
														"str_",
													},
												},
												Id: "str_",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
							Fun: nil,
							CloseFodder: ast.Fodder{},
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(870),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(870),
									Column: int(35),
								},
								File: p1,
							},
						},
					},
					Body: &ast.Local{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(871),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(892),
									Column: int(70),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
									Comment: []string{},
								},
							},
							Ctx: p9067,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
						},
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: nil,
								Variable: "trans",
								EqFodder: nil,
								Body: &ast.Function{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(871),
												Column: int(11),
											},
											End: ast.Location{
												Line: int(891),
												Column: int(13),
											},
											File: p1,
										},
										Fodder: nil,
										Ctx: p9087,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									ParenLeftFodder: ast.Fodder{},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "ch",
											EqFodder: nil,
											DefaultArg: nil,
											CommaFodder: nil,
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(871),
													Column: int(17),
												},
												End: ast.Location{
													Line: int(871),
													Column: int(19),
												},
												File: p1,
											},
										},
									},
									TrailingComma: false,
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Conditional{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(872),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(891),
													Column: int(13),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
													Comment: []string{},
												},
											},
											Ctx: p9092,
											FreeVars: ast.Identifiers{
												"ch",
												"std",
											},
										},
										Cond: &ast.Binary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(872),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(872),
														Column: int(19),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p9092,
												FreeVars: ast.Identifiers{
													"ch",
												},
											},
											Left: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(872),
															Column: int(10),
														},
														End: ast.Location{
															Line: int(872),
															Column: int(12),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p9092,
													FreeVars: ast.Identifiers{
														"ch",
													},
												},
												Id: "ch",
											},
											OpFodder: ast.Fodder{},
											Op: ast.BinaryOp(12),
											Right: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(872),
															Column: int(16),
														},
														End: ast.Location{
															Line: int(872),
															Column: int(19),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p9092,
													FreeVars: nil,
												},
												Value: "\"",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
										},
										ThenFodder: ast.Fodder{},
										BranchTrue: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(873),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(873),
														Column: int(14),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
														Comment: []string{},
													},
												},
												Ctx: p9092,
												FreeVars: nil,
											},
											Value: "\\\"",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										BranchFalse: &ast.Conditional{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(874),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(891),
														Column: int(13),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p9092,
												FreeVars: ast.Identifiers{
													"ch",
													"std",
												},
											},
											Cond: &ast.Binary{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(874),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(874),
															Column: int(25),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p9092,
													FreeVars: ast.Identifiers{
														"ch",
													},
												},
												Left: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(874),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(874),
																Column: int(17),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p9092,
														FreeVars: ast.Identifiers{
															"ch",
														},
													},
													Id: "ch",
												},
												OpFodder: ast.Fodder{},
												Op: ast.BinaryOp(12),
												Right: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(874),
																Column: int(21),
															},
															End: ast.Location{
																Line: int(874),
																Column: int(25),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p9092,
														FreeVars: nil,
													},
													Value: "\\",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
											},
											ThenFodder: ast.Fodder{},
											BranchTrue: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(875),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(875),
															Column: int(15),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
															Comment: []string{},
														},
													},
													Ctx: p9092,
													FreeVars: nil,
												},
												Value: "\\\\",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
													Comment: []string{},
												},
											},
											BranchFalse: &ast.Conditional{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(876),
															Column: int(12),
														},
														End: ast.Location{
															Line: int(891),
															Column: int(13),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p9092,
													FreeVars: ast.Identifiers{
														"ch",
														"std",
													},
												},
												Cond: &ast.Binary{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(876),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(876),
																Column: int(25),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p9092,
														FreeVars: ast.Identifiers{
															"ch",
														},
													},
													Left: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(876),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(876),
																	Column: int(17),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p9092,
															FreeVars: ast.Identifiers{
																"ch",
															},
														},
														Id: "ch",
													},
													OpFodder: ast.Fodder{},
													Op: ast.BinaryOp(12),
													Right: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(876),
																	Column: int(21),
																},
																End: ast.Location{
																	Line: int(876),
																	Column: int(25),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p9092,
															FreeVars: nil,
														},
														Value: "\b",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
												},
												ThenFodder: ast.Fodder{},
												BranchTrue: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(877),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(877),
																Column: int(14),
															},
															File: p1,
														},
														Fodder: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(8),
																Comment: []string{},
															},
														},
														Ctx: p9092,
														FreeVars: nil,
													},
													Value: "\\b",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												ElseFodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
														Comment: []string{},
													},
												},
												BranchFalse: &ast.Conditional{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(878),
																Column: int(12),
															},
															End: ast.Location{
																Line: int(891),
																Column: int(13),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p9092,
														FreeVars: ast.Identifiers{
															"ch",
															"std",
														},
													},
													Cond: &ast.Binary{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(878),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(878),
																	Column: int(25),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p9092,
															FreeVars: ast.Identifiers{
																"ch",
															},
														},
														Left: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(878),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(878),
																		Column: int(17),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p9092,
																FreeVars: ast.Identifiers{
																	"ch",
																},
															},
															Id: "ch",
														},
														OpFodder: ast.Fodder{},
														Op: ast.BinaryOp(12),
														Right: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(878),
																		Column: int(21),
																	},
																	End: ast.Location{
																		Line: int(878),
																		Column: int(25),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p9092,
																FreeVars: nil,
															},
															Value: "\f",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
													},
													ThenFodder: ast.Fodder{},
													BranchTrue: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(879),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(879),
																	Column: int(14),
																},
																File: p1,
															},
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																	Comment: []string{},
																},
															},
															Ctx: p9092,
															FreeVars: nil,
														},
														Value: "\\f",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													ElseFodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(6),
															Comment: []string{},
														},
													},
													BranchFalse: &ast.Conditional{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(880),
																	Column: int(12),
																},
																End: ast.Location{
																	Line: int(891),
																	Column: int(13),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p9092,
															FreeVars: ast.Identifiers{
																"ch",
																"std",
															},
														},
														Cond: &ast.Binary{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(880),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(880),
																		Column: int(25),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p9092,
																FreeVars: ast.Identifiers{
																	"ch",
																},
															},
															Left: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(880),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(880),
																			Column: int(17),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p9092,
																	FreeVars: ast.Identifiers{
																		"ch",
																	},
																},
																Id: "ch",
															},
															OpFodder: ast.Fodder{},
															Op: ast.BinaryOp(12),
															Right: &ast.LiteralString{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(880),
																			Column: int(21),
																		},
																		End: ast.Location{
																			Line: int(880),
																			Column: int(25),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p9092,
																	FreeVars: nil,
																},
																Value: "\n",
																Kind: ast.LiteralStringKind(1),
																BlockIndent: "",
																BlockTermIndent: "",
															},
														},
														ThenFodder: ast.Fodder{},
														BranchTrue: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(881),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(881),
																		Column: int(14),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																		Comment: []string{},
																	},
																},
																Ctx: p9092,
																FreeVars: nil,
															},
															Value: "\\n",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														ElseFodder: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(6),
																Comment: []string{},
															},
														},
														BranchFalse: &ast.Conditional{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(882),
																		Column: int(12),
																	},
																	End: ast.Location{
																		Line: int(891),
																		Column: int(13),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p9092,
																FreeVars: ast.Identifiers{
																	"ch",
																	"std",
																},
															},
															Cond: &ast.Binary{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(882),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(882),
																			Column: int(25),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p9092,
																	FreeVars: ast.Identifiers{
																		"ch",
																	},
																},
																Left: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(882),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(882),
																				Column: int(17),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p9092,
																		FreeVars: ast.Identifiers{
																			"ch",
																		},
																	},
																	Id: "ch",
																},
																OpFodder: ast.Fodder{},
																Op: ast.BinaryOp(12),
																Right: &ast.LiteralString{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(882),
																				Column: int(21),
																			},
																			End: ast.Location{
																				Line: int(882),
																				Column: int(25),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p9092,
																		FreeVars: nil,
																	},
																	Value: "\r",
																	Kind: ast.LiteralStringKind(1),
																	BlockIndent: "",
																	BlockTermIndent: "",
																},
															},
															ThenFodder: ast.Fodder{},
															BranchTrue: &ast.LiteralString{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(883),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(883),
																			Column: int(14),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																			Comment: []string{},
																		},
																	},
																	Ctx: p9092,
																	FreeVars: nil,
																},
																Value: "\\r",
																Kind: ast.LiteralStringKind(1),
																BlockIndent: "",
																BlockTermIndent: "",
															},
															ElseFodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(6),
																	Comment: []string{},
																},
															},
															BranchFalse: &ast.Conditional{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(884),
																			Column: int(12),
																		},
																		End: ast.Location{
																			Line: int(891),
																			Column: int(13),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p9092,
																	FreeVars: ast.Identifiers{
																		"ch",
																		"std",
																	},
																},
																Cond: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(884),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(884),
																				Column: int(25),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p9092,
																		FreeVars: ast.Identifiers{
																			"ch",
																		},
																	},
																	Left: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(884),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(884),
																					Column: int(17),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p9092,
																			FreeVars: ast.Identifiers{
																				"ch",
																			},
																		},
																		Id: "ch",
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(12),
																	Right: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(884),
																					Column: int(21),
																				},
																				End: ast.Location{
																					Line: int(884),
																					Column: int(25),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p9092,
																			FreeVars: nil,
																		},
																		Value: "\t",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																},
																ThenFodder: ast.Fodder{},
																BranchTrue: &ast.LiteralString{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(885),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(885),
																				Column: int(14),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(8),
																				Comment: []string{},
																			},
																		},
																		Ctx: p9092,
																		FreeVars: nil,
																	},
																	Value: "\\t",
																	Kind: ast.LiteralStringKind(1),
																	BlockIndent: "",
																	BlockTermIndent: "",
																},
																ElseFodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(6),
																		Comment: []string{},
																	},
																},
																BranchFalse: &ast.Local{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(887),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(891),
																				Column: int(13),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(8),
																				Comment: []string{},
																			},
																		},
																		Ctx: p9092,
																		FreeVars: ast.Identifiers{
																			"ch",
																			"std",
																		},
																	},
																	Binds: ast.LocalBinds{
																		ast.LocalBind{
																			VarFodder: ast.Fodder{},
																			Variable: "cp",
																			EqFodder: ast.Fodder{},
																			Body: &ast.Apply{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(887),
																							Column: int(20),
																						},
																						End: ast.Location{
																							Line: int(887),
																							Column: int(37),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p9167,
																					FreeVars: ast.Identifiers{
																						"ch",
																						"std",
																					},
																				},
																				Target: &ast.Index{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(887),
																								Column: int(20),
																							},
																							End: ast.Location{
																								Line: int(887),
																								Column: int(33),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p9167,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																					},
																					Target: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(887),
																									Column: int(20),
																								},
																								End: ast.Location{
																									Line: int(887),
																									Column: int(23),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: nil,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																						},
																						Id: "std",
																					},
																					LeftBracketFodder: ast.Fodder{},
																					Index: &ast.LiteralString{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								File: nil,
																							},
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: nil,
																						},
																						Value: "codepoint",
																						Kind: ast.LiteralStringKind(1),
																						BlockIndent: "",
																						BlockTermIndent: "",
																					},
																					RightBracketFodder: ast.Fodder{},
																					Id: nil,
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(887),
																											Column: int(34),
																										},
																										End: ast.Location{
																											Line: int(887),
																											Column: int(36),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p9176,
																									FreeVars: ast.Identifiers{
																										"ch",
																									},
																								},
																								Id: "ch",
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				TrailingComma: false,
																				TailStrict: false,
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: nil,
																			},
																			Fun: nil,
																			CloseFodder: ast.Fodder{},
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(887),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(887),
																					Column: int(37),
																				},
																				File: p1,
																			},
																		},
																	},
																	Body: &ast.Conditional{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(888),
																					Column: int(9),
																				},
																				End: ast.Location{
																					Line: int(891),
																					Column: int(13),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(8),
																					Comment: []string{},
																				},
																			},
																			Ctx: p9092,
																			FreeVars: ast.Identifiers{
																				"ch",
																				"cp",
																				"std",
																			},
																		},
																		Cond: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(888),
																						Column: int(12),
																					},
																					End: ast.Location{
																						Line: int(888),
																						Column: int(47),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p9092,
																				FreeVars: ast.Identifiers{
																					"cp",
																				},
																			},
																			Left: &ast.Binary{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(888),
																							Column: int(12),
																						},
																						End: ast.Location{
																							Line: int(888),
																							Column: int(19),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p9092,
																					FreeVars: ast.Identifiers{
																						"cp",
																					},
																				},
																				Left: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(888),
																								Column: int(12),
																							},
																							End: ast.Location{
																								Line: int(888),
																								Column: int(14),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p9092,
																						FreeVars: ast.Identifiers{
																							"cp",
																						},
																					},
																					Id: "cp",
																				},
																				OpFodder: ast.Fodder{},
																				Op: ast.BinaryOp(9),
																				Right: &ast.LiteralNumber{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(888),
																								Column: int(17),
																							},
																							End: ast.Location{
																								Line: int(888),
																								Column: int(19),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p9092,
																						FreeVars: nil,
																					},
																					OriginalString: "32",
																				},
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(18),
																			Right: &ast.Binary{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(888),
																							Column: int(24),
																						},
																						End: ast.Location{
																							Line: int(888),
																							Column: int(46),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p9092,
																					FreeVars: ast.Identifiers{
																						"cp",
																					},
																				},
																				Left: &ast.Binary{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(888),
																								Column: int(24),
																							},
																							End: ast.Location{
																								Line: int(888),
																								Column: int(33),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p9092,
																						FreeVars: ast.Identifiers{
																							"cp",
																						},
																					},
																					Left: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(888),
																									Column: int(24),
																								},
																								End: ast.Location{
																									Line: int(888),
																									Column: int(26),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p9092,
																							FreeVars: ast.Identifiers{
																								"cp",
																							},
																						},
																						Id: "cp",
																					},
																					OpFodder: ast.Fodder{},
																					Op: ast.BinaryOp(8),
																					Right: &ast.LiteralNumber{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(888),
																									Column: int(30),
																								},
																								End: ast.Location{
																									Line: int(888),
																									Column: int(33),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p9092,
																							FreeVars: nil,
																						},
																						OriginalString: "127",
																					},
																				},
																				OpFodder: ast.Fodder{},
																				Op: ast.BinaryOp(17),
																				Right: &ast.Binary{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(888),
																								Column: int(37),
																							},
																							End: ast.Location{
																								Line: int(888),
																								Column: int(46),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p9092,
																						FreeVars: ast.Identifiers{
																							"cp",
																						},
																					},
																					Left: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(888),
																									Column: int(37),
																								},
																								End: ast.Location{
																									Line: int(888),
																									Column: int(39),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p9092,
																							FreeVars: ast.Identifiers{
																								"cp",
																							},
																						},
																						Id: "cp",
																					},
																					OpFodder: ast.Fodder{},
																					Op: ast.BinaryOp(10),
																					Right: &ast.LiteralNumber{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(888),
																									Column: int(43),
																								},
																								End: ast.Location{
																									Line: int(888),
																									Column: int(46),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p9092,
																							FreeVars: nil,
																						},
																						OriginalString: "159",
																					},
																				},
																			},
																		},
																		ThenFodder: ast.Fodder{},
																		BranchTrue: &ast.Apply{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"cp",
																					"std",
																				},
																			},
																			Target: &ast.Index{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																				},
																				Target: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							File: nil,
																						},
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																					},
																					Id: "std",
																				},
																				LeftBracketFodder: nil,
																				Index: &ast.LiteralString{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							File: nil,
																						},
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: nil,
																					},
																					Value: "mod",
																					Kind: ast.LiteralStringKind(1),
																					BlockIndent: "",
																					BlockTermIndent: "",
																				},
																				RightBracketFodder: nil,
																				Id: nil,
																			},
																			FodderLeft: nil,
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.LiteralString{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(889),
																										Column: int(11),
																									},
																									End: ast.Location{
																										Line: int(889),
																										Column: int(20),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{
																									ast.FodderElement{
																										Kind: ast.FodderKind(0),
																										Blanks: int(0),
																										Indent: int(10),
																										Comment: []string{},
																									},
																								},
																								Ctx: p9092,
																								FreeVars: nil,
																							},
																							Value: "\\u%04x",
																							Kind: ast.LiteralStringKind(1),
																							BlockIndent: "",
																							BlockTermIndent: "",
																						},
																						CommaFodder: nil,
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Array{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(889),
																										Column: int(23),
																									},
																									End: ast.Location{
																										Line: int(889),
																										Column: int(27),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p9092,
																								FreeVars: ast.Identifiers{
																									"cp",
																								},
																							},
																							Elements: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(889),
																													Column: int(24),
																												},
																												End: ast.Location{
																													Line: int(889),
																													Column: int(26),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p9214,
																											FreeVars: ast.Identifiers{
																												"cp",
																											},
																										},
																										Id: "cp",
																									},
																									CommaFodder: nil,
																								},
																							},
																							TrailingComma: false,
																							CloseFodder: ast.Fodder{},
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			TrailingComma: false,
																			TailStrict: false,
																			FodderRight: nil,
																			TailStrictFodder: nil,
																		},
																		ElseFodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(8),
																				Comment: []string{},
																			},
																		},
																		BranchFalse: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(891),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(891),
																						Column: int(13),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(10),
																						Comment: []string{},
																					},
																				},
																				Ctx: p9092,
																				FreeVars: ast.Identifiers{
																					"ch",
																				},
																			},
																			Id: "ch",
																		},
																	},
																},
															},
														},
													},
												},
											},
										},
									},
								},
								Fun: nil,
								CloseFodder: nil,
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
							},
						},
						Body: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
									"str",
									"trans",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: nil,
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "mod",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: nil,
								Id: nil,
							},
							FodderLeft: nil,
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(892),
														Column: int(5),
													},
													End: ast.Location{
														Line: int(892),
														Column: int(11),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(4),
														Comment: []string{},
													},
												},
												Ctx: p9067,
												FreeVars: nil,
											},
											Value: "\"%s\"",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										CommaFodder: nil,
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(892),
														Column: int(14),
													},
													End: ast.Location{
														Line: int(892),
														Column: int(70),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p9067,
												FreeVars: ast.Identifiers{
													"std",
													"str",
													"trans",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(892),
															Column: int(14),
														},
														End: ast.Location{
															Line: int(892),
															Column: int(22),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p9067,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(892),
																Column: int(14),
															},
															End: ast.Location{
																Line: int(892),
																Column: int(17),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "join",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(892),
																		Column: int(23),
																	},
																	End: ast.Location{
																		Line: int(892),
																		Column: int(25),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p9239,
																FreeVars: nil,
															},
															Value: "",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Apply{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																	"str",
																	"trans",
																},
															},
															Target: &ast.Index{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Id: "std",
																},
																LeftBracketFodder: nil,
																Index: &ast.LiteralString{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: nil,
																	},
																	Value: "flatMap",
																	Kind: ast.LiteralStringKind(1),
																	BlockIndent: "",
																	BlockTermIndent: "",
																},
																RightBracketFodder: nil,
																Id: nil,
															},
															FodderLeft: nil,
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Function{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"trans",
																				},
																			},
																			ParenLeftFodder: nil,
																			Parameters: []ast.Parameter{
																				ast.Parameter{
																					NameFodder: nil,
																					Name: "ch",
																					EqFodder: nil,
																					DefaultArg: nil,
																					CommaFodder: nil,
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																				},
																			},
																			TrailingComma: false,
																			ParenRightFodder: nil,
																			Body: &ast.Array{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"ch",
																						"trans",
																					},
																				},
																				Elements: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Apply{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(892),
																										Column: int(28),
																									},
																									End: ast.Location{
																										Line: int(892),
																										Column: int(37),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p9255,
																								FreeVars: ast.Identifiers{
																									"ch",
																									"trans",
																								},
																							},
																							Target: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(892),
																											Column: int(28),
																										},
																										End: ast.Location{
																											Line: int(892),
																											Column: int(33),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p9255,
																									FreeVars: ast.Identifiers{
																										"trans",
																									},
																								},
																								Id: "trans",
																							},
																							FodderLeft: ast.Fodder{},
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(892),
																														Column: int(34),
																													},
																													End: ast.Location{
																														Line: int(892),
																														Column: int(36),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p9261,
																												FreeVars: ast.Identifiers{
																													"ch",
																												},
																											},
																											Id: "ch",
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							TrailingComma: false,
																							TailStrict: false,
																							FodderRight: ast.Fodder{},
																							TailStrictFodder: nil,
																						},
																						CommaFodder: nil,
																					},
																				},
																				TrailingComma: false,
																				CloseFodder: nil,
																			},
																		},
																		CommaFodder: nil,
																	},
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Apply{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(892),
																						Column: int(48),
																					},
																					End: ast.Location{
																						Line: int(892),
																						Column: int(68),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p9239,
																				FreeVars: ast.Identifiers{
																					"std",
																					"str",
																				},
																			},
																			Target: &ast.Index{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(892),
																							Column: int(48),
																						},
																						End: ast.Location{
																							Line: int(892),
																							Column: int(63),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p9239,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																				},
																				Target: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(892),
																								Column: int(48),
																							},
																							End: ast.Location{
																								Line: int(892),
																								Column: int(51),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: nil,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																					},
																					Id: "std",
																				},
																				LeftBracketFodder: ast.Fodder{},
																				Index: &ast.LiteralString{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							File: nil,
																						},
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: nil,
																					},
																					Value: "stringChars",
																					Kind: ast.LiteralStringKind(1),
																					BlockIndent: "",
																					BlockTermIndent: "",
																				},
																				RightBracketFodder: ast.Fodder{},
																				Id: nil,
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(892),
																										Column: int(64),
																									},
																									End: ast.Location{
																										Line: int(892),
																										Column: int(67),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p9272,
																								FreeVars: ast.Identifiers{
																									"str",
																								},
																							},
																							Id: "str",
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			TrailingComma: false,
																			TailStrict: false,
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															TrailingComma: false,
															TailStrict: false,
															FodderRight: nil,
															TailStrictFodder: nil,
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: nil,
							TailStrictFodder: nil,
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(869),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(892),
					Column: int(70),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "escapeStringPython",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(894),
								Column: int(22),
							},
							End: ast.Location{
								Line: int(894),
								Column: int(25),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(895),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(895),
								Column: int(30),
							},
							File: p1,
						},
						Fodder: ast.Fodder{},
						Ctx: p9279,
						FreeVars: ast.Identifiers{
							"std",
							"str",
						},
					},
					Target: &ast.Index{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(895),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(895),
									Column: int(25),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p9279,
							FreeVars: ast.Identifiers{
								"std",
							},
						},
						Target: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(895),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(895),
										Column: int(8),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
										Comment: []string{},
									},
								},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Id: "std",
						},
						LeftBracketFodder: ast.Fodder{},
						Index: &ast.LiteralString{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: nil,
							},
							Value: "escapeStringJson",
							Kind: ast.LiteralStringKind(1),
							BlockIndent: "",
							BlockTermIndent: "",
						},
						RightBracketFodder: ast.Fodder{},
						Id: nil,
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(895),
												Column: int(26),
											},
											End: ast.Location{
												Line: int(895),
												Column: int(29),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p9289,
										FreeVars: ast.Identifiers{
											"str",
										},
									},
									Id: "str",
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					TrailingComma: false,
					TailStrict: false,
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(894),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(895),
					Column: int(30),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "escapeStringBash",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str_",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(897),
								Column: int(20),
							},
							End: ast.Location{
								Line: int(897),
								Column: int(24),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(898),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(904),
								Column: int(70),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p9297,
						FreeVars: ast.Identifiers{
							"std",
							"str_",
						},
					},
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Variable: "str",
							EqFodder: ast.Fodder{},
							Body: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(898),
											Column: int(17),
										},
										End: ast.Location{
											Line: int(898),
											Column: int(35),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p9301,
									FreeVars: ast.Identifiers{
										"std",
										"str_",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(898),
												Column: int(17),
											},
											End: ast.Location{
												Line: int(898),
												Column: int(29),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p9301,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(898),
													Column: int(17),
												},
												End: ast.Location{
													Line: int(898),
													Column: int(20),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "toString",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(898),
															Column: int(30),
														},
														End: ast.Location{
															Line: int(898),
															Column: int(34),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p9310,
													FreeVars: ast.Identifiers{
														"str_",
													},
												},
												Id: "str_",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
							Fun: nil,
							CloseFodder: ast.Fodder{},
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(898),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(898),
									Column: int(35),
								},
								File: p1,
							},
						},
					},
					Body: &ast.Local{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(899),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(904),
									Column: int(70),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
									Comment: []string{},
								},
							},
							Ctx: p9297,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
						},
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: nil,
								Variable: "trans",
								EqFodder: nil,
								Body: &ast.Function{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(899),
												Column: int(11),
											},
											End: ast.Location{
												Line: int(903),
												Column: int(11),
											},
											File: p1,
										},
										Fodder: nil,
										Ctx: p9317,
										FreeVars: nil,
									},
									ParenLeftFodder: ast.Fodder{},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "ch",
											EqFodder: nil,
											DefaultArg: nil,
											CommaFodder: nil,
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(899),
													Column: int(17),
												},
												End: ast.Location{
													Line: int(899),
													Column: int(19),
												},
												File: p1,
											},
										},
									},
									TrailingComma: false,
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Conditional{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(900),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(903),
													Column: int(11),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
													Comment: []string{},
												},
											},
											Ctx: p9321,
											FreeVars: ast.Identifiers{
												"ch",
											},
										},
										Cond: &ast.Binary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(900),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(900),
														Column: int(19),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p9321,
												FreeVars: ast.Identifiers{
													"ch",
												},
											},
											Left: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(900),
															Column: int(10),
														},
														End: ast.Location{
															Line: int(900),
															Column: int(12),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p9321,
													FreeVars: ast.Identifiers{
														"ch",
													},
												},
												Id: "ch",
											},
											OpFodder: ast.Fodder{},
											Op: ast.BinaryOp(12),
											Right: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(900),
															Column: int(16),
														},
														End: ast.Location{
															Line: int(900),
															Column: int(19),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p9321,
													FreeVars: nil,
												},
												Value: "'",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
										},
										ThenFodder: ast.Fodder{},
										BranchTrue: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(901),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(901),
														Column: int(18),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
														Comment: []string{},
													},
												},
												Ctx: p9321,
												FreeVars: nil,
											},
											Value: "'\"'\"'",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										BranchFalse: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(903),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(903),
														Column: int(11),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
														Comment: []string{},
													},
												},
												Ctx: p9321,
												FreeVars: ast.Identifiers{
													"ch",
												},
											},
											Id: "ch",
										},
									},
								},
								Fun: nil,
								CloseFodder: nil,
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
							},
						},
						Body: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
									"str",
									"trans",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: nil,
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "mod",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: nil,
								Id: nil,
							},
							FodderLeft: nil,
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(904),
														Column: int(5),
													},
													End: ast.Location{
														Line: int(904),
														Column: int(11),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(4),
														Comment: []string{},
													},
												},
												Ctx: p9297,
												FreeVars: nil,
											},
											Value: "'%s'",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										CommaFodder: nil,
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(904),
														Column: int(14),
													},
													End: ast.Location{
														Line: int(904),
														Column: int(70),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p9297,
												FreeVars: ast.Identifiers{
													"std",
													"str",
													"trans",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(904),
															Column: int(14),
														},
														End: ast.Location{
															Line: int(904),
															Column: int(22),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p9297,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(904),
																Column: int(14),
															},
															End: ast.Location{
																Line: int(904),
																Column: int(17),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "join",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(904),
																		Column: int(23),
																	},
																	End: ast.Location{
																		Line: int(904),
																		Column: int(25),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p9353,
																FreeVars: nil,
															},
															Value: "",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Apply{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																	"str",
																	"trans",
																},
															},
															Target: &ast.Index{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Id: "std",
																},
																LeftBracketFodder: nil,
																Index: &ast.LiteralString{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: nil,
																	},
																	Value: "flatMap",
																	Kind: ast.LiteralStringKind(1),
																	BlockIndent: "",
																	BlockTermIndent: "",
																},
																RightBracketFodder: nil,
																Id: nil,
															},
															FodderLeft: nil,
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Function{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"trans",
																				},
																			},
																			ParenLeftFodder: nil,
																			Parameters: []ast.Parameter{
																				ast.Parameter{
																					NameFodder: nil,
																					Name: "ch",
																					EqFodder: nil,
																					DefaultArg: nil,
																					CommaFodder: nil,
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																				},
																			},
																			TrailingComma: false,
																			ParenRightFodder: nil,
																			Body: &ast.Array{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"ch",
																						"trans",
																					},
																				},
																				Elements: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Apply{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(904),
																										Column: int(28),
																									},
																									End: ast.Location{
																										Line: int(904),
																										Column: int(37),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p9369,
																								FreeVars: ast.Identifiers{
																									"ch",
																									"trans",
																								},
																							},
																							Target: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(904),
																											Column: int(28),
																										},
																										End: ast.Location{
																											Line: int(904),
																											Column: int(33),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p9369,
																									FreeVars: ast.Identifiers{
																										"trans",
																									},
																								},
																								Id: "trans",
																							},
																							FodderLeft: ast.Fodder{},
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(904),
																														Column: int(34),
																													},
																													End: ast.Location{
																														Line: int(904),
																														Column: int(36),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p9375,
																												FreeVars: ast.Identifiers{
																													"ch",
																												},
																											},
																											Id: "ch",
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							TrailingComma: false,
																							TailStrict: false,
																							FodderRight: ast.Fodder{},
																							TailStrictFodder: nil,
																						},
																						CommaFodder: nil,
																					},
																				},
																				TrailingComma: false,
																				CloseFodder: nil,
																			},
																		},
																		CommaFodder: nil,
																	},
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Apply{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(904),
																						Column: int(48),
																					},
																					End: ast.Location{
																						Line: int(904),
																						Column: int(68),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p9353,
																				FreeVars: ast.Identifiers{
																					"std",
																					"str",
																				},
																			},
																			Target: &ast.Index{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(904),
																							Column: int(48),
																						},
																						End: ast.Location{
																							Line: int(904),
																							Column: int(63),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p9353,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																				},
																				Target: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(904),
																								Column: int(48),
																							},
																							End: ast.Location{
																								Line: int(904),
																								Column: int(51),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: nil,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																					},
																					Id: "std",
																				},
																				LeftBracketFodder: ast.Fodder{},
																				Index: &ast.LiteralString{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							File: nil,
																						},
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: nil,
																					},
																					Value: "stringChars",
																					Kind: ast.LiteralStringKind(1),
																					BlockIndent: "",
																					BlockTermIndent: "",
																				},
																				RightBracketFodder: ast.Fodder{},
																				Id: nil,
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(904),
																										Column: int(64),
																									},
																									End: ast.Location{
																										Line: int(904),
																										Column: int(67),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p9386,
																								FreeVars: ast.Identifiers{
																									"str",
																								},
																							},
																							Id: "str",
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			TrailingComma: false,
																			TailStrict: false,
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															TrailingComma: false,
															TailStrict: false,
															FodderRight: nil,
															TailStrictFodder: nil,
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: nil,
							TailStrictFodder: nil,
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(897),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(904),
					Column: int(70),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "escapeStringDollars",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str_",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(906),
								Column: int(23),
							},
							End: ast.Location{
								Line: int(906),
								Column: int(27),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(907),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(913),
								Column: int(69),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p9394,
						FreeVars: ast.Identifiers{
							"std",
							"str_",
						},
					},
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Variable: "str",
							EqFodder: ast.Fodder{},
							Body: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(907),
											Column: int(17),
										},
										End: ast.Location{
											Line: int(907),
											Column: int(35),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p9398,
									FreeVars: ast.Identifiers{
										"std",
										"str_",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(907),
												Column: int(17),
											},
											End: ast.Location{
												Line: int(907),
												Column: int(29),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p9398,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(907),
													Column: int(17),
												},
												End: ast.Location{
													Line: int(907),
													Column: int(20),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "toString",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(907),
															Column: int(30),
														},
														End: ast.Location{
															Line: int(907),
															Column: int(34),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p9407,
													FreeVars: ast.Identifiers{
														"str_",
													},
												},
												Id: "str_",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
							Fun: nil,
							CloseFodder: ast.Fodder{},
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(907),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(907),
									Column: int(35),
								},
								File: p1,
							},
						},
					},
					Body: &ast.Local{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(908),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(913),
									Column: int(69),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
									Comment: []string{},
								},
							},
							Ctx: p9394,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
						},
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: nil,
								Variable: "trans",
								EqFodder: nil,
								Body: &ast.Function{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(908),
												Column: int(11),
											},
											End: ast.Location{
												Line: int(912),
												Column: int(11),
											},
											File: p1,
										},
										Fodder: nil,
										Ctx: p9414,
										FreeVars: nil,
									},
									ParenLeftFodder: ast.Fodder{},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "ch",
											EqFodder: nil,
											DefaultArg: nil,
											CommaFodder: nil,
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(908),
													Column: int(17),
												},
												End: ast.Location{
													Line: int(908),
													Column: int(19),
												},
												File: p1,
											},
										},
									},
									TrailingComma: false,
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Conditional{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(909),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(912),
													Column: int(11),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
													Comment: []string{},
												},
											},
											Ctx: p9418,
											FreeVars: ast.Identifiers{
												"ch",
											},
										},
										Cond: &ast.Binary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(909),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(909),
														Column: int(19),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p9418,
												FreeVars: ast.Identifiers{
													"ch",
												},
											},
											Left: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(909),
															Column: int(10),
														},
														End: ast.Location{
															Line: int(909),
															Column: int(12),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p9418,
													FreeVars: ast.Identifiers{
														"ch",
													},
												},
												Id: "ch",
											},
											OpFodder: ast.Fodder{},
											Op: ast.BinaryOp(12),
											Right: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(909),
															Column: int(16),
														},
														End: ast.Location{
															Line: int(909),
															Column: int(19),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p9418,
													FreeVars: nil,
												},
												Value: "$",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
										},
										ThenFodder: ast.Fodder{},
										BranchTrue: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(910),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(910),
														Column: int(13),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
														Comment: []string{},
													},
												},
												Ctx: p9418,
												FreeVars: nil,
											},
											Value: "$$",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										BranchFalse: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(912),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(912),
														Column: int(11),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
														Comment: []string{},
													},
												},
												Ctx: p9418,
												FreeVars: ast.Identifiers{
													"ch",
												},
											},
											Id: "ch",
										},
									},
								},
								Fun: nil,
								CloseFodder: nil,
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
							},
						},
						Body: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(913),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(913),
										Column: int(69),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p9394,
								FreeVars: ast.Identifiers{
									"std",
									"str",
									"trans",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(913),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(913),
											Column: int(14),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p9394,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(913),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(913),
												Column: int(8),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(4),
												Comment: []string{},
											},
										},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "foldl",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Function{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(913),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(913),
														Column: int(42),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p9441,
												FreeVars: ast.Identifiers{
													"trans",
												},
											},
											ParenLeftFodder: ast.Fodder{},
											Parameters: []ast.Parameter{
												ast.Parameter{
													NameFodder: ast.Fodder{},
													Name: "a",
													EqFodder: nil,
													DefaultArg: nil,
													CommaFodder: ast.Fodder{},
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(913),
															Column: int(24),
														},
														End: ast.Location{
															Line: int(913),
															Column: int(25),
														},
														File: p1,
													},
												},
												ast.Parameter{
													NameFodder: ast.Fodder{},
													Name: "b",
													EqFodder: nil,
													DefaultArg: nil,
													CommaFodder: nil,
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(913),
															Column: int(27),
														},
														End: ast.Location{
															Line: int(913),
															Column: int(28),
														},
														File: p1,
													},
												},
											},
											TrailingComma: false,
											ParenRightFodder: ast.Fodder{},
											Body: &ast.Binary{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(913),
															Column: int(30),
														},
														End: ast.Location{
															Line: int(913),
															Column: int(42),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p9445,
													FreeVars: ast.Identifiers{
														"a",
														"b",
														"trans",
													},
												},
												Left: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(913),
																Column: int(30),
															},
															End: ast.Location{
																Line: int(913),
																Column: int(31),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p9445,
														FreeVars: ast.Identifiers{
															"a",
														},
													},
													Id: "a",
												},
												OpFodder: ast.Fodder{},
												Op: ast.BinaryOp(3),
												Right: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(913),
																Column: int(34),
															},
															End: ast.Location{
																Line: int(913),
																Column: int(42),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p9445,
														FreeVars: ast.Identifiers{
															"b",
															"trans",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(913),
																	Column: int(34),
																},
																End: ast.Location{
																	Line: int(913),
																	Column: int(39),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p9445,
															FreeVars: ast.Identifiers{
																"trans",
															},
														},
														Id: "trans",
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(913),
																				Column: int(40),
																			},
																			End: ast.Location{
																				Line: int(913),
																				Column: int(41),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p9455,
																		FreeVars: ast.Identifiers{
																			"b",
																		},
																	},
																	Id: "b",
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
												},
											},
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(913),
														Column: int(44),
													},
													End: ast.Location{
														Line: int(913),
														Column: int(64),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p9441,
												FreeVars: ast.Identifiers{
													"std",
													"str",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(913),
															Column: int(44),
														},
														End: ast.Location{
															Line: int(913),
															Column: int(59),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p9441,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(913),
																Column: int(44),
															},
															End: ast.Location{
																Line: int(913),
																Column: int(47),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "stringChars",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(913),
																		Column: int(60),
																	},
																	End: ast.Location{
																		Line: int(913),
																		Column: int(63),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p9466,
																FreeVars: ast.Identifiers{
																	"str",
																},
															},
															Id: "str",
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(913),
														Column: int(66),
													},
													End: ast.Location{
														Line: int(913),
														Column: int(68),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p9441,
												FreeVars: nil,
											},
											Value: "",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(906),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(913),
					Column: int(69),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "manifestJson",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "value",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(915),
								Column: int(16),
							},
							End: ast.Location{
								Line: int(915),
								Column: int(21),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(915),
								Column: int(25),
							},
							End: ast.Location{
								Line: int(915),
								Column: int(58),
							},
							File: p1,
						},
						Fodder: ast.Fodder{},
						Ctx: p9474,
						FreeVars: ast.Identifiers{
							"std",
							"value",
						},
					},
					Target: &ast.Index{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(915),
									Column: int(25),
								},
								End: ast.Location{
									Line: int(915),
									Column: int(43),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p9474,
							FreeVars: ast.Identifiers{
								"std",
							},
						},
						Target: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(915),
										Column: int(25),
									},
									End: ast.Location{
										Line: int(915),
										Column: int(28),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Id: "std",
						},
						LeftBracketFodder: ast.Fodder{},
						Index: &ast.LiteralString{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: nil,
							},
							Value: "manifestJsonEx",
							Kind: ast.LiteralStringKind(1),
							BlockIndent: "",
							BlockTermIndent: "",
						},
						RightBracketFodder: ast.Fodder{},
						Id: nil,
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(915),
												Column: int(44),
											},
											End: ast.Location{
												Line: int(915),
												Column: int(49),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p9483,
										FreeVars: ast.Identifiers{
											"value",
										},
									},
									Id: "value",
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(915),
												Column: int(51),
											},
											End: ast.Location{
												Line: int(915),
												Column: int(57),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p9483,
										FreeVars: nil,
									},
									Value: "    ",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					TrailingComma: false,
					TailStrict: false,
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(915),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(915),
					Column: int(58),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "manifestJsonEx",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "value",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(917),
								Column: int(18),
							},
							End: ast.Location{
								Line: int(917),
								Column: int(23),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "indent",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(917),
								Column: int(25),
							},
							End: ast.Location{
								Line: int(917),
								Column: int(31),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(918),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(952),
								Column: int(23),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p9492,
						FreeVars: ast.Identifiers{
							"indent",
							"std",
							"value",
						},
					},
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: nil,
							Variable: "aux",
							EqFodder: nil,
							Body: &ast.Function{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(918),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(951),
											Column: int(28),
										},
										File: p1,
									},
									Fodder: nil,
									Ctx: p9496,
									FreeVars: ast.Identifiers{
										"aux",
										"indent",
										"std",
									},
								},
								ParenLeftFodder: ast.Fodder{},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "v",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(918),
												Column: int(15),
											},
											End: ast.Location{
												Line: int(918),
												Column: int(16),
											},
											File: p1,
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "path",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(918),
												Column: int(18),
											},
											End: ast.Location{
												Line: int(918),
												Column: int(22),
											},
											File: p1,
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "cindent",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: nil,
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(918),
												Column: int(24),
											},
											End: ast.Location{
												Line: int(918),
												Column: int(31),
											},
											File: p1,
										},
									},
								},
								TrailingComma: false,
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Conditional{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(919),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(951),
												Column: int(28),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										Ctx: p9501,
										FreeVars: ast.Identifiers{
											"aux",
											"cindent",
											"indent",
											"path",
											"std",
											"v",
										},
									},
									Cond: &ast.Binary{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(919),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(919),
													Column: int(19),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p9501,
											FreeVars: ast.Identifiers{
												"v",
											},
										},
										Left: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(919),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(919),
														Column: int(11),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p9501,
												FreeVars: ast.Identifiers{
													"v",
												},
											},
											Id: "v",
										},
										OpFodder: ast.Fodder{},
										Op: ast.BinaryOp(12),
										Right: &ast.LiteralBoolean{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(919),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(919),
														Column: int(19),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p9501,
												FreeVars: nil,
											},
											Value: true,
										},
									},
									ThenFodder: ast.Fodder{},
									BranchTrue: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(920),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(920),
													Column: int(15),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
													Comment: []string{},
												},
											},
											Ctx: p9501,
											FreeVars: nil,
										},
										Value: "true",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
											Comment: []string{},
										},
									},
									BranchFalse: &ast.Conditional{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(921),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(951),
													Column: int(28),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p9501,
											FreeVars: ast.Identifiers{
												"aux",
												"cindent",
												"indent",
												"path",
												"std",
												"v",
											},
										},
										Cond: &ast.Binary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(921),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(921),
														Column: int(25),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p9501,
												FreeVars: ast.Identifiers{
													"v",
												},
											},
											Left: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(921),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(921),
															Column: int(16),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p9501,
													FreeVars: ast.Identifiers{
														"v",
													},
												},
												Id: "v",
											},
											OpFodder: ast.Fodder{},
											Op: ast.BinaryOp(12),
											Right: &ast.LiteralBoolean{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(921),
															Column: int(20),
														},
														End: ast.Location{
															Line: int(921),
															Column: int(25),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p9501,
													FreeVars: nil,
												},
												Value: false,
											},
										},
										ThenFodder: ast.Fodder{},
										BranchTrue: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(922),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(922),
														Column: int(16),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
														Comment: []string{},
													},
												},
												Ctx: p9501,
												FreeVars: nil,
											},
											Value: "false",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										BranchFalse: &ast.Conditional{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(923),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(951),
														Column: int(28),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p9501,
												FreeVars: ast.Identifiers{
													"aux",
													"cindent",
													"indent",
													"path",
													"std",
													"v",
												},
											},
											Cond: &ast.Binary{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(923),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(923),
															Column: int(24),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p9501,
													FreeVars: ast.Identifiers{
														"v",
													},
												},
												Left: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(923),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(923),
																Column: int(16),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p9501,
														FreeVars: ast.Identifiers{
															"v",
														},
													},
													Id: "v",
												},
												OpFodder: ast.Fodder{},
												Op: ast.BinaryOp(12),
												Right: &ast.LiteralNull{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(923),
																Column: int(20),
															},
															End: ast.Location{
																Line: int(923),
																Column: int(24),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p9501,
														FreeVars: nil,
													},
												},
											},
											ThenFodder: ast.Fodder{},
											BranchTrue: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(924),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(924),
															Column: int(15),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
															Comment: []string{},
														},
													},
													Ctx: p9501,
													FreeVars: nil,
												},
												Value: "null",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
													Comment: []string{},
												},
											},
											BranchFalse: &ast.Conditional{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(925),
															Column: int(12),
														},
														End: ast.Location{
															Line: int(951),
															Column: int(28),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p9501,
													FreeVars: ast.Identifiers{
														"aux",
														"cindent",
														"indent",
														"path",
														"std",
														"v",
													},
												},
												Cond: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(925),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(925),
																Column: int(30),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p9501,
														FreeVars: ast.Identifiers{
															"std",
															"v",
														},
													},
													Target: &ast.Index{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(925),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(925),
																	Column: int(27),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p9501,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(925),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(925),
																		Column: int(18),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Id: "std",
														},
														LeftBracketFodder: ast.Fodder{},
														Index: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
															Value: "isNumber",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														RightBracketFodder: ast.Fodder{},
														Id: nil,
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(925),
																				Column: int(28),
																			},
																			End: ast.Location{
																				Line: int(925),
																				Column: int(29),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p9542,
																		FreeVars: ast.Identifiers{
																			"v",
																		},
																	},
																	Id: "v",
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
												},
												ThenFodder: ast.Fodder{},
												BranchTrue: &ast.Binary{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(926),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(926),
																Column: int(15),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p9501,
														FreeVars: ast.Identifiers{
															"v",
														},
													},
													Left: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(926),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(926),
																	Column: int(11),
																},
																File: p1,
															},
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																	Comment: []string{},
																},
															},
															Ctx: p9501,
															FreeVars: nil,
														},
														Value: "",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													OpFodder: ast.Fodder{},
													Op: ast.BinaryOp(3),
													Right: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(926),
																	Column: int(14),
																},
																End: ast.Location{
																	Line: int(926),
																	Column: int(15),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p9501,
															FreeVars: ast.Identifiers{
																"v",
															},
														},
														Id: "v",
													},
												},
												ElseFodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
														Comment: []string{},
													},
												},
												BranchFalse: &ast.Conditional{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(927),
																Column: int(12),
															},
															End: ast.Location{
																Line: int(951),
																Column: int(28),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p9501,
														FreeVars: ast.Identifiers{
															"aux",
															"cindent",
															"indent",
															"path",
															"std",
															"v",
														},
													},
													Cond: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(927),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(927),
																	Column: int(30),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p9501,
															FreeVars: ast.Identifiers{
																"std",
																"v",
															},
														},
														Target: &ast.Index{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(927),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(927),
																		Column: int(27),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p9501,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Target: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(927),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(927),
																			Column: int(18),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Id: "std",
															},
															LeftBracketFodder: ast.Fodder{},
															Index: &ast.LiteralString{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: nil,
																},
																Value: "isString",
																Kind: ast.LiteralStringKind(1),
																BlockIndent: "",
																BlockTermIndent: "",
															},
															RightBracketFodder: ast.Fodder{},
															Id: nil,
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(927),
																					Column: int(28),
																				},
																				End: ast.Location{
																					Line: int(927),
																					Column: int(29),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p9562,
																			FreeVars: ast.Identifiers{
																				"v",
																			},
																		},
																		Id: "v",
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														TrailingComma: false,
														TailStrict: false,
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
													},
													ThenFodder: ast.Fodder{},
													BranchTrue: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(928),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(928),
																	Column: int(32),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p9501,
															FreeVars: ast.Identifiers{
																"std",
																"v",
															},
														},
														Target: &ast.Index{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(928),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(928),
																		Column: int(29),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p9501,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Target: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(928),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(928),
																			Column: int(12),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																			Comment: []string{},
																		},
																	},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Id: "std",
															},
															LeftBracketFodder: ast.Fodder{},
															Index: &ast.LiteralString{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: nil,
																},
																Value: "escapeStringJson",
																Kind: ast.LiteralStringKind(1),
																BlockIndent: "",
																BlockTermIndent: "",
															},
															RightBracketFodder: ast.Fodder{},
															Id: nil,
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(928),
																					Column: int(30),
																				},
																				End: ast.Location{
																					Line: int(928),
																					Column: int(31),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p9574,
																			FreeVars: ast.Identifiers{
																				"v",
																			},
																		},
																		Id: "v",
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														TrailingComma: false,
														TailStrict: false,
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
													},
													ElseFodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(6),
															Comment: []string{},
														},
													},
													BranchFalse: &ast.Conditional{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(929),
																	Column: int(12),
																},
																End: ast.Location{
																	Line: int(951),
																	Column: int(28),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p9501,
															FreeVars: ast.Identifiers{
																"aux",
																"cindent",
																"indent",
																"path",
																"std",
																"v",
															},
														},
														Cond: &ast.Apply{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(929),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(929),
																		Column: int(32),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p9501,
																FreeVars: ast.Identifiers{
																	"std",
																	"v",
																},
															},
															Target: &ast.Index{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(929),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(929),
																			Column: int(29),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p9501,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(929),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(929),
																				Column: int(18),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Id: "std",
																},
																LeftBracketFodder: ast.Fodder{},
																Index: &ast.LiteralString{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: nil,
																	},
																	Value: "isFunction",
																	Kind: ast.LiteralStringKind(1),
																	BlockIndent: "",
																	BlockTermIndent: "",
																},
																RightBracketFodder: ast.Fodder{},
																Id: nil,
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(929),
																						Column: int(30),
																					},
																					End: ast.Location{
																						Line: int(929),
																						Column: int(31),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p9588,
																				FreeVars: ast.Identifiers{
																					"v",
																				},
																			},
																			Id: "v",
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															TrailingComma: false,
															TailStrict: false,
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
														},
														ThenFodder: ast.Fodder{},
														BranchTrue: &ast.Error{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(930),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(930),
																		Column: int(54),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																		Comment: []string{},
																	},
																},
																Ctx: p9501,
																FreeVars: ast.Identifiers{
																	"path",
																},
															},
															Expr: &ast.Binary{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(930),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(930),
																			Column: int(54),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p9501,
																	FreeVars: ast.Identifiers{
																		"path",
																	},
																},
																Left: &ast.LiteralString{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(930),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(930),
																				Column: int(47),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p9501,
																		FreeVars: nil,
																	},
																	Value: "Tried to manifest function at ",
																	Kind: ast.LiteralStringKind(1),
																	BlockIndent: "",
																	BlockTermIndent: "",
																},
																OpFodder: ast.Fodder{},
																Op: ast.BinaryOp(3),
																Right: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(930),
																				Column: int(50),
																			},
																			End: ast.Location{
																				Line: int(930),
																				Column: int(54),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p9501,
																		FreeVars: ast.Identifiers{
																			"path",
																		},
																	},
																	Id: "path",
																},
															},
														},
														ElseFodder: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(6),
																Comment: []string{},
															},
														},
														BranchFalse: &ast.Conditional{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(931),
																		Column: int(12),
																	},
																	End: ast.Location{
																		Line: int(951),
																		Column: int(28),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p9501,
																FreeVars: ast.Identifiers{
																	"aux",
																	"cindent",
																	"indent",
																	"path",
																	"std",
																	"v",
																},
															},
															Cond: &ast.Apply{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(931),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(931),
																			Column: int(29),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p9501,
																	FreeVars: ast.Identifiers{
																		"std",
																		"v",
																	},
																},
																Target: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(931),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(931),
																				Column: int(26),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p9501,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(931),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(931),
																					Column: int(18),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Id: "std",
																	},
																	LeftBracketFodder: ast.Fodder{},
																	Index: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: nil,
																		},
																		Value: "isArray",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	RightBracketFodder: ast.Fodder{},
																	Id: nil,
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(931),
																							Column: int(27),
																						},
																						End: ast.Location{
																							Line: int(931),
																							Column: int(28),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p9610,
																					FreeVars: ast.Identifiers{
																						"v",
																					},
																				},
																				Id: "v",
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																TrailingComma: false,
																TailStrict: false,
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
															},
															ThenFodder: ast.Fodder{},
															BranchTrue: &ast.Local{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(932),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(941),
																			Column: int(28),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																			Comment: []string{},
																		},
																	},
																	Ctx: p9501,
																	FreeVars: ast.Identifiers{
																		"aux",
																		"cindent",
																		"indent",
																		"path",
																		"std",
																		"v",
																	},
																},
																Binds: ast.LocalBinds{
																	ast.LocalBind{
																		VarFodder: ast.Fodder{},
																		Variable: "range",
																		EqFodder: ast.Fodder{},
																		Body: &ast.Apply{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(932),
																						Column: int(23),
																					},
																					End: ast.Location{
																						Line: int(932),
																						Column: int(54),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p9617,
																				FreeVars: ast.Identifiers{
																					"std",
																					"v",
																				},
																			},
																			Target: &ast.Index{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(932),
																							Column: int(23),
																						},
																						End: ast.Location{
																							Line: int(932),
																							Column: int(32),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p9617,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																				},
																				Target: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(932),
																								Column: int(23),
																							},
																							End: ast.Location{
																								Line: int(932),
																								Column: int(26),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: nil,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																					},
																					Id: "std",
																				},
																				LeftBracketFodder: ast.Fodder{},
																				Index: &ast.LiteralString{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							File: nil,
																						},
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: nil,
																					},
																					Value: "range",
																					Kind: ast.LiteralStringKind(1),
																					BlockIndent: "",
																					BlockTermIndent: "",
																				},
																				RightBracketFodder: ast.Fodder{},
																				Id: nil,
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.LiteralNumber{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(932),
																										Column: int(33),
																									},
																									End: ast.Location{
																										Line: int(932),
																										Column: int(34),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p9626,
																								FreeVars: nil,
																							},
																							OriginalString: "0",
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Binary{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(932),
																										Column: int(36),
																									},
																									End: ast.Location{
																										Line: int(932),
																										Column: int(53),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p9626,
																								FreeVars: ast.Identifiers{
																									"std",
																									"v",
																								},
																							},
																							Left: &ast.Apply{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(932),
																											Column: int(36),
																										},
																										End: ast.Location{
																											Line: int(932),
																											Column: int(49),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p9626,
																									FreeVars: ast.Identifiers{
																										"std",
																										"v",
																									},
																								},
																								Target: &ast.Index{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(932),
																												Column: int(36),
																											},
																											End: ast.Location{
																												Line: int(932),
																												Column: int(46),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p9626,
																										FreeVars: ast.Identifiers{
																											"std",
																										},
																									},
																									Target: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(932),
																													Column: int(36),
																												},
																												End: ast.Location{
																													Line: int(932),
																													Column: int(39),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"std",
																											},
																										},
																										Id: "std",
																									},
																									LeftBracketFodder: ast.Fodder{},
																									Index: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												File: nil,
																											},
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: nil,
																										},
																										Value: "length",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																									RightBracketFodder: ast.Fodder{},
																									Id: nil,
																								},
																								FodderLeft: ast.Fodder{},
																								Arguments: ast.Arguments{
																									Positional: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(932),
																															Column: int(47),
																														},
																														End: ast.Location{
																															Line: int(932),
																															Column: int(48),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p9638,
																													FreeVars: ast.Identifiers{
																														"v",
																													},
																												},
																												Id: "v",
																											},
																											CommaFodder: nil,
																										},
																									},
																									Named: nil,
																								},
																								TrailingComma: false,
																								TailStrict: false,
																								FodderRight: ast.Fodder{},
																								TailStrictFodder: nil,
																							},
																							OpFodder: ast.Fodder{},
																							Op: ast.BinaryOp(4),
																							Right: &ast.LiteralNumber{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(932),
																											Column: int(52),
																										},
																										End: ast.Location{
																											Line: int(932),
																											Column: int(53),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p9626,
																									FreeVars: nil,
																								},
																								OriginalString: "1",
																							},
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			TrailingComma: false,
																			TailStrict: false,
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																		},
																		Fun: nil,
																		CloseFodder: ast.Fodder{},
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(932),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(932),
																				Column: int(54),
																			},
																			File: p1,
																		},
																	},
																},
																Body: &ast.Local{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(933),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(941),
																				Column: int(28),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(8),
																				Comment: []string{},
																			},
																		},
																		Ctx: p9501,
																		FreeVars: ast.Identifiers{
																			"aux",
																			"cindent",
																			"indent",
																			"path",
																			"range",
																			"std",
																			"v",
																		},
																	},
																	Binds: ast.LocalBinds{
																		ast.LocalBind{
																			VarFodder: ast.Fodder{},
																			Variable: "new_indent",
																			EqFodder: ast.Fodder{},
																			Body: &ast.Binary{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(933),
																							Column: int(28),
																						},
																						End: ast.Location{
																							Line: int(933),
																							Column: int(44),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p9646,
																					FreeVars: ast.Identifiers{
																						"cindent",
																						"indent",
																					},
																				},
																				Left: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(933),
																								Column: int(28),
																							},
																							End: ast.Location{
																								Line: int(933),
																								Column: int(35),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p9646,
																						FreeVars: ast.Identifiers{
																							"cindent",
																						},
																					},
																					Id: "cindent",
																				},
																				OpFodder: ast.Fodder{},
																				Op: ast.BinaryOp(3),
																				Right: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(933),
																								Column: int(38),
																							},
																							End: ast.Location{
																								Line: int(933),
																								Column: int(44),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p9646,
																						FreeVars: ast.Identifiers{
																							"indent",
																						},
																					},
																					Id: "indent",
																				},
																			},
																			Fun: nil,
																			CloseFodder: ast.Fodder{},
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(933),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(933),
																					Column: int(44),
																				},
																				File: p1,
																			},
																		},
																	},
																	Body: &ast.Local{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(934),
																					Column: int(9),
																				},
																				End: ast.Location{
																					Line: int(941),
																					Column: int(28),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(8),
																					Comment: []string{},
																				},
																			},
																			Ctx: p9501,
																			FreeVars: ast.Identifiers{
																				"aux",
																				"cindent",
																				"new_indent",
																				"path",
																				"range",
																				"std",
																				"v",
																			},
																		},
																		Binds: ast.LocalBinds{
																			ast.LocalBind{
																				VarFodder: ast.Fodder{},
																				Variable: "lines",
																				EqFodder: ast.Fodder{},
																				Body: &ast.Binary{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(934),
																								Column: int(23),
																							},
																							End: ast.Location{
																								Line: int(940),
																								Column: int(47),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p9657,
																						FreeVars: ast.Identifiers{
																							"aux",
																							"cindent",
																							"new_indent",
																							"path",
																							"range",
																							"std",
																							"v",
																						},
																					},
																					Left: &ast.Binary{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(934),
																									Column: int(23),
																								},
																								End: ast.Location{
																									Line: int(939),
																									Column: int(36),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p9657,
																							FreeVars: ast.Identifiers{
																								"aux",
																								"new_indent",
																								"path",
																								"range",
																								"std",
																								"v",
																							},
																						},
																						Left: &ast.Array{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(934),
																										Column: int(23),
																									},
																									End: ast.Location{
																										Line: int(934),
																										Column: int(30),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p9657,
																								FreeVars: nil,
																							},
																							Elements: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(934),
																													Column: int(24),
																												},
																												End: ast.Location{
																													Line: int(934),
																													Column: int(29),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p9664,
																											FreeVars: nil,
																										},
																										Value: "[\n",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																									CommaFodder: nil,
																								},
																							},
																							TrailingComma: false,
																							CloseFodder: ast.Fodder{},
																						},
																						OpFodder: ast.Fodder{
																							ast.FodderElement{
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(22),
																								Comment: []string{},
																							},
																						},
																						Op: ast.BinaryOp(3),
																						Right: &ast.Apply{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(935),
																										Column: int(25),
																									},
																									End: ast.Location{
																										Line: int(939),
																										Column: int(36),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p9657,
																								FreeVars: ast.Identifiers{
																									"aux",
																									"new_indent",
																									"path",
																									"range",
																									"std",
																									"v",
																								},
																							},
																							Target: &ast.Index{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(935),
																											Column: int(25),
																										},
																										End: ast.Location{
																											Line: int(935),
																											Column: int(33),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p9657,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																								},
																								Target: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(935),
																												Column: int(25),
																											},
																											End: ast.Location{
																												Line: int(935),
																												Column: int(28),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: nil,
																										FreeVars: ast.Identifiers{
																											"std",
																										},
																									},
																									Id: "std",
																								},
																								LeftBracketFodder: ast.Fodder{},
																								Index: &ast.LiteralString{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											File: nil,
																										},
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: nil,
																									},
																									Value: "join",
																									Kind: ast.LiteralStringKind(1),
																									BlockIndent: "",
																									BlockTermIndent: "",
																								},
																								RightBracketFodder: ast.Fodder{},
																								Id: nil,
																							},
																							FodderLeft: ast.Fodder{},
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Array{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(935),
																														Column: int(34),
																													},
																													End: ast.Location{
																														Line: int(935),
																														Column: int(41),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p9675,
																												FreeVars: nil,
																											},
																											Elements: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.LiteralString{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(935),
																																	Column: int(35),
																																},
																																End: ast.Location{
																																	Line: int(935),
																																	Column: int(40),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p9678,
																															FreeVars: nil,
																														},
																														Value: ",\n",
																														Kind: ast.LiteralStringKind(1),
																														BlockIndent: "",
																														BlockTermIndent: "",
																													},
																													CommaFodder: nil,
																												},
																											},
																											TrailingComma: false,
																											CloseFodder: ast.Fodder{},
																										},
																										CommaFodder: ast.Fodder{},
																									},
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Apply{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													File: nil,
																												},
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"aux",
																													"new_indent",
																													"path",
																													"range",
																													"std",
																													"v",
																												},
																											},
																											Target: &ast.Index{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																												},
																												Target: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															File: nil,
																														},
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"std",
																														},
																													},
																													Id: "std",
																												},
																												LeftBracketFodder: nil,
																												Index: &ast.LiteralString{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															File: nil,
																														},
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: nil,
																													},
																													Value: "flatMap",
																													Kind: ast.LiteralStringKind(1),
																													BlockIndent: "",
																													BlockTermIndent: "",
																												},
																												RightBracketFodder: nil,
																												Id: nil,
																											},
																											FodderLeft: nil,
																											Arguments: ast.Arguments{
																												Positional: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Function{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	File: nil,
																																},
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"aux",
																																	"new_indent",
																																	"path",
																																	"v",
																																},
																															},
																															ParenLeftFodder: nil,
																															Parameters: []ast.Parameter{
																																ast.Parameter{
																																	NameFodder: nil,
																																	Name: "i",
																																	EqFodder: nil,
																																	DefaultArg: nil,
																																	CommaFodder: nil,
																																	LocRange: ast.LocationRange{
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		File: nil,
																																	},
																																},
																															},
																															TrailingComma: false,
																															ParenRightFodder: nil,
																															Body: &ast.Array{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		File: nil,
																																	},
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"aux",
																																		"i",
																																		"new_indent",
																																		"path",
																																		"v",
																																	},
																																},
																																Elements: []ast.CommaSeparatedExpr{
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Array{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(937),
																																						Column: int(36),
																																					},
																																					End: ast.Location{
																																						Line: int(937),
																																						Column: int(84),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{
																																					ast.FodderElement{
																																						Kind: ast.FodderKind(0),
																																						Blanks: int(0),
																																						Indent: int(35),
																																						Comment: []string{},
																																					},
																																				},
																																				Ctx: p9695,
																																				FreeVars: ast.Identifiers{
																																					"aux",
																																					"i",
																																					"new_indent",
																																					"path",
																																					"v",
																																				},
																																			},
																																			Elements: []ast.CommaSeparatedExpr{
																																				ast.CommaSeparatedExpr{
																																					Expr: &ast.Binary{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(937),
																																									Column: int(37),
																																								},
																																								End: ast.Location{
																																									Line: int(937),
																																									Column: int(83),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p9699,
																																							FreeVars: ast.Identifiers{
																																								"aux",
																																								"i",
																																								"new_indent",
																																								"path",
																																								"v",
																																							},
																																						},
																																						Left: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(937),
																																										Column: int(37),
																																									},
																																									End: ast.Location{
																																										Line: int(937),
																																										Column: int(47),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p9699,
																																								FreeVars: ast.Identifiers{
																																									"new_indent",
																																								},
																																							},
																																							Id: "new_indent",
																																						},
																																						OpFodder: ast.Fodder{},
																																						Op: ast.BinaryOp(3),
																																						Right: &ast.Apply{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(937),
																																										Column: int(50),
																																									},
																																									End: ast.Location{
																																										Line: int(937),
																																										Column: int(83),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p9699,
																																								FreeVars: ast.Identifiers{
																																									"aux",
																																									"i",
																																									"new_indent",
																																									"path",
																																									"v",
																																								},
																																							},
																																							Target: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(937),
																																											Column: int(50),
																																										},
																																										End: ast.Location{
																																											Line: int(937),
																																											Column: int(53),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p9699,
																																									FreeVars: ast.Identifiers{
																																										"aux",
																																									},
																																								},
																																								Id: "aux",
																																							},
																																							FodderLeft: ast.Fodder{},
																																							Arguments: ast.Arguments{
																																								Positional: []ast.CommaSeparatedExpr{
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Index{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(937),
																																														Column: int(54),
																																													},
																																													End: ast.Location{
																																														Line: int(937),
																																														Column: int(58),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p9709,
																																												FreeVars: ast.Identifiers{
																																													"i",
																																													"v",
																																												},
																																											},
																																											Target: &ast.Var{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(937),
																																															Column: int(54),
																																														},
																																														End: ast.Location{
																																															Line: int(937),
																																															Column: int(55),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: p9709,
																																													FreeVars: ast.Identifiers{
																																														"v",
																																													},
																																												},
																																												Id: "v",
																																											},
																																											LeftBracketFodder: ast.Fodder{},
																																											Index: &ast.Var{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(937),
																																															Column: int(56),
																																														},
																																														End: ast.Location{
																																															Line: int(937),
																																															Column: int(57),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: p9709,
																																													FreeVars: ast.Identifiers{
																																														"i",
																																													},
																																												},
																																												Id: "i",
																																											},
																																											RightBracketFodder: ast.Fodder{},
																																											Id: nil,
																																										},
																																										CommaFodder: ast.Fodder{},
																																									},
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Binary{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(937),
																																														Column: int(60),
																																													},
																																													End: ast.Location{
																																														Line: int(937),
																																														Column: int(70),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p9709,
																																												FreeVars: ast.Identifiers{
																																													"i",
																																													"path",
																																												},
																																											},
																																											Left: &ast.Var{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(937),
																																															Column: int(60),
																																														},
																																														End: ast.Location{
																																															Line: int(937),
																																															Column: int(64),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: p9709,
																																													FreeVars: ast.Identifiers{
																																														"path",
																																													},
																																												},
																																												Id: "path",
																																											},
																																											OpFodder: ast.Fodder{},
																																											Op: ast.BinaryOp(3),
																																											Right: &ast.Array{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(937),
																																															Column: int(67),
																																														},
																																														End: ast.Location{
																																															Line: int(937),
																																															Column: int(70),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: p9709,
																																													FreeVars: ast.Identifiers{
																																														"i",
																																													},
																																												},
																																												Elements: []ast.CommaSeparatedExpr{
																																													ast.CommaSeparatedExpr{
																																														Expr: &ast.Var{
																																															NodeBase: ast.NodeBase{
																																																LocRange: ast.LocationRange{
																																																	FileName: "<std>",
																																																	Begin: ast.Location{
																																																		Line: int(937),
																																																		Column: int(68),
																																																	},
																																																	End: ast.Location{
																																																		Line: int(937),
																																																		Column: int(69),
																																																	},
																																																	File: p1,
																																																},
																																																Fodder: ast.Fodder{},
																																																Ctx: p9723,
																																																FreeVars: ast.Identifiers{
																																																	"i",
																																																},
																																															},
																																															Id: "i",
																																														},
																																														CommaFodder: nil,
																																													},
																																												},
																																												TrailingComma: false,
																																												CloseFodder: ast.Fodder{},
																																											},
																																										},
																																										CommaFodder: ast.Fodder{},
																																									},
																																									ast.CommaSeparatedExpr{
																																										Expr: &ast.Var{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(937),
																																														Column: int(72),
																																													},
																																													End: ast.Location{
																																														Line: int(937),
																																														Column: int(82),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p9709,
																																												FreeVars: ast.Identifiers{
																																													"new_indent",
																																												},
																																											},
																																											Id: "new_indent",
																																										},
																																										CommaFodder: nil,
																																									},
																																								},
																																								Named: nil,
																																							},
																																							TrailingComma: false,
																																							TailStrict: false,
																																							FodderRight: ast.Fodder{},
																																							TailStrictFodder: nil,
																																						},
																																					},
																																					CommaFodder: nil,
																																				},
																																			},
																																			TrailingComma: false,
																																			CloseFodder: ast.Fodder{},
																																		},
																																		CommaFodder: nil,
																																	},
																																},
																																TrailingComma: false,
																																CloseFodder: nil,
																															},
																														},
																														CommaFodder: nil,
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(938),
																																		Column: int(45),
																																	},
																																	End: ast.Location{
																																		Line: int(938),
																																		Column: int(50),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p9675,
																																FreeVars: ast.Identifiers{
																																	"range",
																																},
																															},
																															Id: "range",
																														},
																														CommaFodder: nil,
																													},
																												},
																												Named: nil,
																											},
																											TrailingComma: false,
																											TailStrict: false,
																											FodderRight: nil,
																											TailStrictFodder: nil,
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							TrailingComma: false,
																							TailStrict: false,
																							FodderRight: ast.Fodder{},
																							TailStrictFodder: nil,
																						},
																					},
																					OpFodder: ast.Fodder{
																						ast.FodderElement{
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(22),
																							Comment: []string{},
																						},
																					},
																					Op: ast.BinaryOp(3),
																					Right: &ast.Array{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(940),
																									Column: int(25),
																								},
																								End: ast.Location{
																									Line: int(940),
																									Column: int(47),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p9657,
																							FreeVars: ast.Identifiers{
																								"cindent",
																							},
																						},
																						Elements: []ast.CommaSeparatedExpr{
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Binary{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(940),
																												Column: int(26),
																											},
																											End: ast.Location{
																												Line: int(940),
																												Column: int(46),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p9734,
																										FreeVars: ast.Identifiers{
																											"cindent",
																										},
																									},
																									Left: &ast.Binary{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(940),
																													Column: int(26),
																												},
																												End: ast.Location{
																													Line: int(940),
																													Column: int(40),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p9734,
																											FreeVars: ast.Identifiers{
																												"cindent",
																											},
																										},
																										Left: &ast.LiteralString{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(940),
																														Column: int(26),
																													},
																													End: ast.Location{
																														Line: int(940),
																														Column: int(30),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p9734,
																												FreeVars: nil,
																											},
																											Value: "\n",
																											Kind: ast.LiteralStringKind(1),
																											BlockIndent: "",
																											BlockTermIndent: "",
																										},
																										OpFodder: ast.Fodder{},
																										Op: ast.BinaryOp(3),
																										Right: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(940),
																														Column: int(33),
																													},
																													End: ast.Location{
																														Line: int(940),
																														Column: int(40),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p9734,
																												FreeVars: ast.Identifiers{
																													"cindent",
																												},
																											},
																											Id: "cindent",
																										},
																									},
																									OpFodder: ast.Fodder{},
																									Op: ast.BinaryOp(3),
																									Right: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(940),
																													Column: int(43),
																												},
																												End: ast.Location{
																													Line: int(940),
																													Column: int(46),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p9734,
																											FreeVars: nil,
																										},
																										Value: "]",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																								},
																								CommaFodder: nil,
																							},
																						},
																						TrailingComma: false,
																						CloseFodder: ast.Fodder{},
																					},
																				},
																				Fun: nil,
																				CloseFodder: ast.Fodder{},
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(934),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(940),
																						Column: int(47),
																					},
																					File: p1,
																				},
																			},
																		},
																		Body: &ast.Apply{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(941),
																						Column: int(9),
																					},
																					End: ast.Location{
																						Line: int(941),
																						Column: int(28),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p9501,
																				FreeVars: ast.Identifiers{
																					"lines",
																					"std",
																				},
																			},
																			Target: &ast.Index{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(941),
																							Column: int(9),
																						},
																						End: ast.Location{
																							Line: int(941),
																							Column: int(17),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p9501,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																				},
																				Target: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(941),
																								Column: int(9),
																							},
																							End: ast.Location{
																								Line: int(941),
																								Column: int(12),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(8),
																								Comment: []string{},
																							},
																						},
																						Ctx: nil,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																					},
																					Id: "std",
																				},
																				LeftBracketFodder: ast.Fodder{},
																				Index: &ast.LiteralString{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							File: nil,
																						},
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: nil,
																					},
																					Value: "join",
																					Kind: ast.LiteralStringKind(1),
																					BlockIndent: "",
																					BlockTermIndent: "",
																				},
																				RightBracketFodder: ast.Fodder{},
																				Id: nil,
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.LiteralString{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(941),
																										Column: int(18),
																									},
																									End: ast.Location{
																										Line: int(941),
																										Column: int(20),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p9752,
																								FreeVars: nil,
																							},
																							Value: "",
																							Kind: ast.LiteralStringKind(1),
																							BlockIndent: "",
																							BlockTermIndent: "",
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(941),
																										Column: int(22),
																									},
																									End: ast.Location{
																										Line: int(941),
																										Column: int(27),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p9752,
																								FreeVars: ast.Identifiers{
																									"lines",
																								},
																							},
																							Id: "lines",
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			TrailingComma: false,
																			TailStrict: false,
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																		},
																	},
																},
															},
															ElseFodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(6),
																	Comment: []string{},
																},
															},
															BranchFalse: &ast.Conditional{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(942),
																			Column: int(12),
																		},
																		End: ast.Location{
																			Line: int(951),
																			Column: int(28),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p9501,
																	FreeVars: ast.Identifiers{
																		"aux",
																		"cindent",
																		"indent",
																		"path",
																		"std",
																		"v",
																	},
																},
																Cond: &ast.Apply{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(942),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(942),
																				Column: int(30),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p9501,
																		FreeVars: ast.Identifiers{
																			"std",
																			"v",
																		},
																	},
																	Target: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(942),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(942),
																					Column: int(27),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p9501,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(942),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(942),
																						Column: int(18),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Id: "std",
																		},
																		LeftBracketFodder: ast.Fodder{},
																		Index: &ast.LiteralString{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: nil,
																			},
																			Value: "isObject",
																			Kind: ast.LiteralStringKind(1),
																			BlockIndent: "",
																			BlockTermIndent: "",
																		},
																		RightBracketFodder: ast.Fodder{},
																		Id: nil,
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(942),
																								Column: int(28),
																							},
																							End: ast.Location{
																								Line: int(942),
																								Column: int(29),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p9767,
																						FreeVars: ast.Identifiers{
																							"v",
																						},
																					},
																					Id: "v",
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	TrailingComma: false,
																	TailStrict: false,
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																},
																ThenFodder: ast.Fodder{},
																BranchTrue: &ast.Local{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(943),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(951),
																				Column: int(28),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(8),
																				Comment: []string{},
																			},
																		},
																		Ctx: p9501,
																		FreeVars: ast.Identifiers{
																			"aux",
																			"cindent",
																			"indent",
																			"path",
																			"std",
																			"v",
																		},
																	},
																	Binds: ast.LocalBinds{
																		ast.LocalBind{
																			VarFodder: ast.Fodder{},
																			Variable: "lines",
																			EqFodder: ast.Fodder{},
																			Body: &ast.Binary{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(943),
																							Column: int(23),
																						},
																						End: ast.Location{
																							Line: int(950),
																							Column: int(47),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p9774,
																					FreeVars: ast.Identifiers{
																						"aux",
																						"cindent",
																						"indent",
																						"path",
																						"std",
																						"v",
																					},
																				},
																				Left: &ast.Binary{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(943),
																								Column: int(23),
																							},
																							End: ast.Location{
																								Line: int(949),
																								Column: int(36),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p9774,
																						FreeVars: ast.Identifiers{
																							"aux",
																							"cindent",
																							"indent",
																							"path",
																							"std",
																							"v",
																						},
																					},
																					Left: &ast.Array{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(943),
																									Column: int(23),
																								},
																								End: ast.Location{
																									Line: int(943),
																									Column: int(30),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p9774,
																							FreeVars: nil,
																						},
																						Elements: []ast.CommaSeparatedExpr{
																							ast.CommaSeparatedExpr{
																								Expr: &ast.LiteralString{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(943),
																												Column: int(24),
																											},
																											End: ast.Location{
																												Line: int(943),
																												Column: int(29),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p9781,
																										FreeVars: nil,
																									},
																									Value: "{\n",
																									Kind: ast.LiteralStringKind(1),
																									BlockIndent: "",
																									BlockTermIndent: "",
																								},
																								CommaFodder: nil,
																							},
																						},
																						TrailingComma: false,
																						CloseFodder: ast.Fodder{},
																					},
																					OpFodder: ast.Fodder{
																						ast.FodderElement{
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(22),
																							Comment: []string{},
																						},
																					},
																					Op: ast.BinaryOp(3),
																					Right: &ast.Apply{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(944),
																									Column: int(25),
																								},
																								End: ast.Location{
																									Line: int(949),
																									Column: int(36),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p9774,
																							FreeVars: ast.Identifiers{
																								"aux",
																								"cindent",
																								"indent",
																								"path",
																								"std",
																								"v",
																							},
																						},
																						Target: &ast.Index{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(944),
																										Column: int(25),
																									},
																									End: ast.Location{
																										Line: int(944),
																										Column: int(33),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p9774,
																								FreeVars: ast.Identifiers{
																									"std",
																								},
																							},
																							Target: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(944),
																											Column: int(25),
																										},
																										End: ast.Location{
																											Line: int(944),
																											Column: int(28),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: nil,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																								},
																								Id: "std",
																							},
																							LeftBracketFodder: ast.Fodder{},
																							Index: &ast.LiteralString{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										File: nil,
																									},
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: nil,
																								},
																								Value: "join",
																								Kind: ast.LiteralStringKind(1),
																								BlockIndent: "",
																								BlockTermIndent: "",
																							},
																							RightBracketFodder: ast.Fodder{},
																							Id: nil,
																						},
																						FodderLeft: ast.Fodder{},
																						Arguments: ast.Arguments{
																							Positional: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Array{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(944),
																													Column: int(34),
																												},
																												End: ast.Location{
																													Line: int(944),
																													Column: int(41),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p9792,
																											FreeVars: nil,
																										},
																										Elements: []ast.CommaSeparatedExpr{
																											ast.CommaSeparatedExpr{
																												Expr: &ast.LiteralString{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(944),
																																Column: int(35),
																															},
																															End: ast.Location{
																																Line: int(944),
																																Column: int(40),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p9795,
																														FreeVars: nil,
																													},
																													Value: ",\n",
																													Kind: ast.LiteralStringKind(1),
																													BlockIndent: "",
																													BlockTermIndent: "",
																												},
																												CommaFodder: nil,
																											},
																										},
																										TrailingComma: false,
																										CloseFodder: ast.Fodder{},
																									},
																									CommaFodder: ast.Fodder{},
																								},
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Apply{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												File: nil,
																											},
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"aux",
																												"cindent",
																												"indent",
																												"path",
																												"std",
																												"v",
																											},
																										},
																										Target: &ast.Index{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													File: nil,
																												},
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																											},
																											Target: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																												},
																												Id: "std",
																											},
																											LeftBracketFodder: nil,
																											Index: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: nil,
																												},
																												Value: "flatMap",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																											RightBracketFodder: nil,
																											Id: nil,
																										},
																										FodderLeft: nil,
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Function{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																End: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																File: nil,
																															},
																															Fodder: nil,
																															Ctx: nil,
																															FreeVars: ast.Identifiers{
																																"aux",
																																"cindent",
																																"indent",
																																"path",
																																"std",
																																"v",
																															},
																														},
																														ParenLeftFodder: nil,
																														Parameters: []ast.Parameter{
																															ast.Parameter{
																																NameFodder: nil,
																																Name: "k",
																																EqFodder: nil,
																																DefaultArg: nil,
																																CommaFodder: nil,
																																LocRange: ast.LocationRange{
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	File: nil,
																																},
																															},
																														},
																														TrailingComma: false,
																														ParenRightFodder: nil,
																														Body: &ast.Array{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	File: nil,
																																},
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"aux",
																																	"cindent",
																																	"indent",
																																	"k",
																																	"path",
																																	"std",
																																	"v",
																																},
																															},
																															Elements: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Array{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(946),
																																					Column: int(36),
																																				},
																																				End: ast.Location{
																																					Line: int(947),
																																					Column: int(79),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{
																																				ast.FodderElement{
																																					Kind: ast.FodderKind(0),
																																					Blanks: int(0),
																																					Indent: int(35),
																																					Comment: []string{},
																																				},
																																			},
																																			Ctx: p9812,
																																			FreeVars: ast.Identifiers{
																																				"aux",
																																				"cindent",
																																				"indent",
																																				"k",
																																				"path",
																																				"std",
																																				"v",
																																			},
																																		},
																																		Elements: []ast.CommaSeparatedExpr{
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Binary{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(946),
																																								Column: int(37),
																																							},
																																							End: ast.Location{
																																								Line: int(947),
																																								Column: int(78),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p9816,
																																						FreeVars: ast.Identifiers{
																																							"aux",
																																							"cindent",
																																							"indent",
																																							"k",
																																							"path",
																																							"std",
																																							"v",
																																						},
																																					},
																																					Left: &ast.Binary{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(946),
																																									Column: int(37),
																																								},
																																								End: ast.Location{
																																									Line: int(946),
																																									Column: int(86),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p9816,
																																							FreeVars: ast.Identifiers{
																																								"cindent",
																																								"indent",
																																								"k",
																																								"std",
																																							},
																																						},
																																						Left: &ast.Binary{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(946),
																																										Column: int(37),
																																									},
																																									End: ast.Location{
																																										Line: int(946),
																																										Column: int(79),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p9816,
																																								FreeVars: ast.Identifiers{
																																									"cindent",
																																									"indent",
																																									"k",
																																									"std",
																																								},
																																							},
																																							Left: &ast.Binary{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(946),
																																											Column: int(37),
																																										},
																																										End: ast.Location{
																																											Line: int(946),
																																											Column: int(53),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p9816,
																																									FreeVars: ast.Identifiers{
																																										"cindent",
																																										"indent",
																																									},
																																								},
																																								Left: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(946),
																																												Column: int(37),
																																											},
																																											End: ast.Location{
																																												Line: int(946),
																																												Column: int(44),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p9816,
																																										FreeVars: ast.Identifiers{
																																											"cindent",
																																										},
																																									},
																																									Id: "cindent",
																																								},
																																								OpFodder: ast.Fodder{},
																																								Op: ast.BinaryOp(3),
																																								Right: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(946),
																																												Column: int(47),
																																											},
																																											End: ast.Location{
																																												Line: int(946),
																																												Column: int(53),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p9816,
																																										FreeVars: ast.Identifiers{
																																											"indent",
																																										},
																																									},
																																									Id: "indent",
																																								},
																																							},
																																							OpFodder: ast.Fodder{},
																																							Op: ast.BinaryOp(3),
																																							Right: &ast.Apply{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(946),
																																											Column: int(56),
																																										},
																																										End: ast.Location{
																																											Line: int(946),
																																											Column: int(79),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p9816,
																																									FreeVars: ast.Identifiers{
																																										"k",
																																										"std",
																																									},
																																								},
																																								Target: &ast.Index{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(946),
																																												Column: int(56),
																																											},
																																											End: ast.Location{
																																												Line: int(946),
																																												Column: int(76),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p9816,
																																										FreeVars: ast.Identifiers{
																																											"std",
																																										},
																																									},
																																									Target: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(946),
																																													Column: int(56),
																																												},
																																												End: ast.Location{
																																													Line: int(946),
																																													Column: int(59),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: nil,
																																											FreeVars: ast.Identifiers{
																																												"std",
																																											},
																																										},
																																										Id: "std",
																																									},
																																									LeftBracketFodder: ast.Fodder{},
																																									Index: &ast.LiteralString{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "",
																																												Begin: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												End: ast.Location{
																																													Line: int(0),
																																													Column: int(0),
																																												},
																																												File: nil,
																																											},
																																											Fodder: nil,
																																											Ctx: nil,
																																											FreeVars: nil,
																																										},
																																										Value: "escapeStringJson",
																																										Kind: ast.LiteralStringKind(1),
																																										BlockIndent: "",
																																										BlockTermIndent: "",
																																									},
																																									RightBracketFodder: ast.Fodder{},
																																									Id: nil,
																																								},
																																								FodderLeft: ast.Fodder{},
																																								Arguments: ast.Arguments{
																																									Positional: []ast.CommaSeparatedExpr{
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Var{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(946),
																																															Column: int(77),
																																														},
																																														End: ast.Location{
																																															Line: int(946),
																																															Column: int(78),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: p9837,
																																													FreeVars: ast.Identifiers{
																																														"k",
																																													},
																																												},
																																												Id: "k",
																																											},
																																											CommaFodder: nil,
																																										},
																																									},
																																									Named: nil,
																																								},
																																								TrailingComma: false,
																																								TailStrict: false,
																																								FodderRight: ast.Fodder{},
																																								TailStrictFodder: nil,
																																							},
																																						},
																																						OpFodder: ast.Fodder{},
																																						Op: ast.BinaryOp(3),
																																						Right: &ast.LiteralString{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(946),
																																										Column: int(82),
																																									},
																																									End: ast.Location{
																																										Line: int(946),
																																										Column: int(86),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p9816,
																																								FreeVars: nil,
																																							},
																																							Value: ": ",
																																							Kind: ast.LiteralStringKind(1),
																																							BlockIndent: "",
																																							BlockTermIndent: "",
																																						},
																																					},
																																					OpFodder: ast.Fodder{
																																						ast.FodderElement{
																																							Kind: ast.FodderKind(0),
																																							Blanks: int(0),
																																							Indent: int(36),
																																							Comment: []string{},
																																						},
																																					},
																																					Op: ast.BinaryOp(3),
																																					Right: &ast.Apply{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(947),
																																									Column: int(39),
																																								},
																																								End: ast.Location{
																																									Line: int(947),
																																									Column: int(78),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p9816,
																																							FreeVars: ast.Identifiers{
																																								"aux",
																																								"cindent",
																																								"indent",
																																								"k",
																																								"path",
																																								"v",
																																							},
																																						},
																																						Target: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(947),
																																										Column: int(39),
																																									},
																																									End: ast.Location{
																																										Line: int(947),
																																										Column: int(42),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p9816,
																																								FreeVars: ast.Identifiers{
																																									"aux",
																																								},
																																							},
																																							Id: "aux",
																																						},
																																						FodderLeft: ast.Fodder{},
																																						Arguments: ast.Arguments{
																																							Positional: []ast.CommaSeparatedExpr{
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Index{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(947),
																																													Column: int(43),
																																												},
																																												End: ast.Location{
																																													Line: int(947),
																																													Column: int(47),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p9847,
																																											FreeVars: ast.Identifiers{
																																												"k",
																																												"v",
																																											},
																																										},
																																										Target: &ast.Var{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(947),
																																														Column: int(43),
																																													},
																																													End: ast.Location{
																																														Line: int(947),
																																														Column: int(44),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p9847,
																																												FreeVars: ast.Identifiers{
																																													"v",
																																												},
																																											},
																																											Id: "v",
																																										},
																																										LeftBracketFodder: ast.Fodder{},
																																										Index: &ast.Var{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(947),
																																														Column: int(45),
																																													},
																																													End: ast.Location{
																																														Line: int(947),
																																														Column: int(46),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p9847,
																																												FreeVars: ast.Identifiers{
																																													"k",
																																												},
																																											},
																																											Id: "k",
																																										},
																																										RightBracketFodder: ast.Fodder{},
																																										Id: nil,
																																									},
																																									CommaFodder: ast.Fodder{},
																																								},
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Binary{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(947),
																																													Column: int(49),
																																												},
																																												End: ast.Location{
																																													Line: int(947),
																																													Column: int(59),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p9847,
																																											FreeVars: ast.Identifiers{
																																												"k",
																																												"path",
																																											},
																																										},
																																										Left: &ast.Var{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(947),
																																														Column: int(49),
																																													},
																																													End: ast.Location{
																																														Line: int(947),
																																														Column: int(53),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p9847,
																																												FreeVars: ast.Identifiers{
																																													"path",
																																												},
																																											},
																																											Id: "path",
																																										},
																																										OpFodder: ast.Fodder{},
																																										Op: ast.BinaryOp(3),
																																										Right: &ast.Array{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(947),
																																														Column: int(56),
																																													},
																																													End: ast.Location{
																																														Line: int(947),
																																														Column: int(59),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p9847,
																																												FreeVars: ast.Identifiers{
																																													"k",
																																												},
																																											},
																																											Elements: []ast.CommaSeparatedExpr{
																																												ast.CommaSeparatedExpr{
																																													Expr: &ast.Var{
																																														NodeBase: ast.NodeBase{
																																															LocRange: ast.LocationRange{
																																																FileName: "<std>",
																																																Begin: ast.Location{
																																																	Line: int(947),
																																																	Column: int(57),
																																																},
																																																End: ast.Location{
																																																	Line: int(947),
																																																	Column: int(58),
																																																},
																																																File: p1,
																																															},
																																															Fodder: ast.Fodder{},
																																															Ctx: p9861,
																																															FreeVars: ast.Identifiers{
																																																"k",
																																															},
																																														},
																																														Id: "k",
																																													},
																																													CommaFodder: nil,
																																												},
																																											},
																																											TrailingComma: false,
																																											CloseFodder: ast.Fodder{},
																																										},
																																									},
																																									CommaFodder: ast.Fodder{},
																																								},
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Binary{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(947),
																																													Column: int(61),
																																												},
																																												End: ast.Location{
																																													Line: int(947),
																																													Column: int(77),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p9847,
																																											FreeVars: ast.Identifiers{
																																												"cindent",
																																												"indent",
																																											},
																																										},
																																										Left: &ast.Var{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(947),
																																														Column: int(61),
																																													},
																																													End: ast.Location{
																																														Line: int(947),
																																														Column: int(68),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p9847,
																																												FreeVars: ast.Identifiers{
																																													"cindent",
																																												},
																																											},
																																											Id: "cindent",
																																										},
																																										OpFodder: ast.Fodder{},
																																										Op: ast.BinaryOp(3),
																																										Right: &ast.Var{
																																											NodeBase: ast.NodeBase{
																																												LocRange: ast.LocationRange{
																																													FileName: "<std>",
																																													Begin: ast.Location{
																																														Line: int(947),
																																														Column: int(71),
																																													},
																																													End: ast.Location{
																																														Line: int(947),
																																														Column: int(77),
																																													},
																																													File: p1,
																																												},
																																												Fodder: ast.Fodder{},
																																												Ctx: p9847,
																																												FreeVars: ast.Identifiers{
																																													"indent",
																																												},
																																											},
																																											Id: "indent",
																																										},
																																									},
																																									CommaFodder: nil,
																																								},
																																							},
																																							Named: nil,
																																						},
																																						TrailingComma: false,
																																						TailStrict: false,
																																						FodderRight: ast.Fodder{},
																																						TailStrictFodder: nil,
																																					},
																																				},
																																				CommaFodder: nil,
																																			},
																																		},
																																		TrailingComma: false,
																																		CloseFodder: ast.Fodder{},
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															TrailingComma: false,
																															CloseFodder: nil,
																														},
																													},
																													CommaFodder: nil,
																												},
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Apply{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(948),
																																	Column: int(45),
																																},
																																End: ast.Location{
																																	Line: int(948),
																																	Column: int(64),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p9792,
																															FreeVars: ast.Identifiers{
																																"std",
																																"v",
																															},
																														},
																														Target: &ast.Index{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(948),
																																		Column: int(45),
																																	},
																																	End: ast.Location{
																																		Line: int(948),
																																		Column: int(61),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p9792,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																															},
																															Target: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(948),
																																			Column: int(45),
																																		},
																																		End: ast.Location{
																																			Line: int(948),
																																			Column: int(48),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																	},
																																},
																																Id: "std",
																															},
																															LeftBracketFodder: ast.Fodder{},
																															Index: &ast.LiteralString{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		File: nil,
																																	},
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: nil,
																																},
																																Value: "objectFields",
																																Kind: ast.LiteralStringKind(1),
																																BlockIndent: "",
																																BlockTermIndent: "",
																															},
																															RightBracketFodder: ast.Fodder{},
																															Id: nil,
																														},
																														FodderLeft: ast.Fodder{},
																														Arguments: ast.Arguments{
																															Positional: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Var{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(948),
																																					Column: int(62),
																																				},
																																				End: ast.Location{
																																					Line: int(948),
																																					Column: int(63),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p9878,
																																			FreeVars: ast.Identifiers{
																																				"v",
																																			},
																																		},
																																		Id: "v",
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															Named: nil,
																														},
																														TrailingComma: false,
																														TailStrict: false,
																														FodderRight: ast.Fodder{},
																														TailStrictFodder: nil,
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										TrailingComma: false,
																										TailStrict: false,
																										FodderRight: nil,
																										TailStrictFodder: nil,
																									},
																									CommaFodder: nil,
																								},
																							},
																							Named: nil,
																						},
																						TrailingComma: false,
																						TailStrict: false,
																						FodderRight: ast.Fodder{},
																						TailStrictFodder: nil,
																					},
																				},
																				OpFodder: ast.Fodder{
																					ast.FodderElement{
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(22),
																						Comment: []string{},
																					},
																				},
																				Op: ast.BinaryOp(3),
																				Right: &ast.Array{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(950),
																								Column: int(25),
																							},
																							End: ast.Location{
																								Line: int(950),
																								Column: int(47),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p9774,
																						FreeVars: ast.Identifiers{
																							"cindent",
																						},
																					},
																					Elements: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Binary{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(950),
																											Column: int(26),
																										},
																										End: ast.Location{
																											Line: int(950),
																											Column: int(46),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p9885,
																									FreeVars: ast.Identifiers{
																										"cindent",
																									},
																								},
																								Left: &ast.Binary{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(950),
																												Column: int(26),
																											},
																											End: ast.Location{
																												Line: int(950),
																												Column: int(40),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p9885,
																										FreeVars: ast.Identifiers{
																											"cindent",
																										},
																									},
																									Left: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(950),
																													Column: int(26),
																												},
																												End: ast.Location{
																													Line: int(950),
																													Column: int(30),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p9885,
																											FreeVars: nil,
																										},
																										Value: "\n",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																									OpFodder: ast.Fodder{},
																									Op: ast.BinaryOp(3),
																									Right: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(950),
																													Column: int(33),
																												},
																												End: ast.Location{
																													Line: int(950),
																													Column: int(40),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p9885,
																											FreeVars: ast.Identifiers{
																												"cindent",
																											},
																										},
																										Id: "cindent",
																									},
																								},
																								OpFodder: ast.Fodder{},
																								Op: ast.BinaryOp(3),
																								Right: &ast.LiteralString{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(950),
																												Column: int(43),
																											},
																											End: ast.Location{
																												Line: int(950),
																												Column: int(46),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p9885,
																										FreeVars: nil,
																									},
																									Value: "}",
																									Kind: ast.LiteralStringKind(1),
																									BlockIndent: "",
																									BlockTermIndent: "",
																								},
																							},
																							CommaFodder: nil,
																						},
																					},
																					TrailingComma: false,
																					CloseFodder: ast.Fodder{},
																				},
																			},
																			Fun: nil,
																			CloseFodder: ast.Fodder{},
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(943),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(950),
																					Column: int(47),
																				},
																				File: p1,
																			},
																		},
																	},
																	Body: &ast.Apply{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(951),
																					Column: int(9),
																				},
																				End: ast.Location{
																					Line: int(951),
																					Column: int(28),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p9501,
																			FreeVars: ast.Identifiers{
																				"lines",
																				"std",
																			},
																		},
																		Target: &ast.Index{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(951),
																						Column: int(9),
																					},
																					End: ast.Location{
																						Line: int(951),
																						Column: int(17),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p9501,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(951),
																							Column: int(9),
																						},
																						End: ast.Location{
																							Line: int(951),
																							Column: int(12),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(8),
																							Comment: []string{},
																						},
																					},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																				},
																				Id: "std",
																			},
																			LeftBracketFodder: ast.Fodder{},
																			Index: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: nil,
																				},
																				Value: "join",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																			RightBracketFodder: ast.Fodder{},
																			Id: nil,
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.LiteralString{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(951),
																									Column: int(18),
																								},
																								End: ast.Location{
																									Line: int(951),
																									Column: int(20),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p9903,
																							FreeVars: nil,
																						},
																						Value: "",
																						Kind: ast.LiteralStringKind(1),
																						BlockIndent: "",
																						BlockTermIndent: "",
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(951),
																									Column: int(22),
																								},
																								End: ast.Location{
																									Line: int(951),
																									Column: int(27),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p9903,
																							FreeVars: ast.Identifiers{
																								"lines",
																							},
																						},
																						Id: "lines",
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		TrailingComma: false,
																		TailStrict: false,
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																	},
																},
																ElseFodder: nil,
																BranchFalse: &ast.LiteralNull{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: nil,
																	},
																},
															},
														},
													},
												},
											},
										},
									},
								},
							},
							Fun: nil,
							CloseFodder: nil,
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
						},
					},
					Body: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(952),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(952),
									Column: int(23),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p9492,
							FreeVars: ast.Identifiers{
								"aux",
								"value",
							},
						},
						Target: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(952),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(952),
										Column: int(8),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
										Comment: []string{},
									},
								},
								Ctx: p9492,
								FreeVars: ast.Identifiers{
									"aux",
								},
							},
							Id: "aux",
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(952),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(952),
													Column: int(14),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p9914,
											FreeVars: ast.Identifiers{
												"value",
											},
										},
										Id: "value",
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Array{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(952),
													Column: int(16),
												},
												End: ast.Location{
													Line: int(952),
													Column: int(18),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p9914,
											FreeVars: nil,
										},
										Elements: nil,
										TrailingComma: false,
										CloseFodder: ast.Fodder{},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(952),
													Column: int(20),
												},
												End: ast.Location{
													Line: int(952),
													Column: int(22),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p9914,
											FreeVars: nil,
										},
										Value: "",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(917),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(952),
					Column: int(23),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "manifestYamlDoc",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "value",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(954),
								Column: int(19),
							},
							End: ast.Location{
								Line: int(954),
								Column: int(24),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "indent_array_in_object",
						EqFodder: ast.Fodder{},
						DefaultArg: &ast.LiteralBoolean{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(954),
										Column: int(49),
									},
									End: ast.Location{
										Line: int(954),
										Column: int(54),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p9923,
								FreeVars: nil,
							},
							Value: false,
						},
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(954),
								Column: int(26),
							},
							End: ast.Location{
								Line: int(954),
								Column: int(54),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(955),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1035),
								Column: int(23),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p9923,
						FreeVars: ast.Identifiers{
							"indent_array_in_object",
							"std",
							"value",
						},
					},
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: nil,
							Variable: "aux",
							EqFodder: nil,
							Body: &ast.Function{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(955),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(1034),
											Column: int(42),
										},
										File: p1,
									},
									Fodder: nil,
									Ctx: p9929,
									FreeVars: ast.Identifiers{
										"aux",
										"indent_array_in_object",
										"std",
									},
								},
								ParenLeftFodder: ast.Fodder{},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "v",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(955),
												Column: int(15),
											},
											End: ast.Location{
												Line: int(955),
												Column: int(16),
											},
											File: p1,
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "path",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(955),
												Column: int(18),
											},
											End: ast.Location{
												Line: int(955),
												Column: int(22),
											},
											File: p1,
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "cindent",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: nil,
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(955),
												Column: int(24),
											},
											End: ast.Location{
												Line: int(955),
												Column: int(31),
											},
											File: p1,
										},
									},
								},
								TrailingComma: false,
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Conditional{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(956),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1034),
												Column: int(42),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										Ctx: p9934,
										FreeVars: ast.Identifiers{
											"aux",
											"cindent",
											"indent_array_in_object",
											"path",
											"std",
											"v",
										},
									},
									Cond: &ast.Binary{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(956),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(956),
													Column: int(19),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p9934,
											FreeVars: ast.Identifiers{
												"v",
											},
										},
										Left: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(956),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(956),
														Column: int(11),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p9934,
												FreeVars: ast.Identifiers{
													"v",
												},
											},
											Id: "v",
										},
										OpFodder: ast.Fodder{},
										Op: ast.BinaryOp(12),
										Right: &ast.LiteralBoolean{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(956),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(956),
														Column: int(19),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p9934,
												FreeVars: nil,
											},
											Value: true,
										},
									},
									ThenFodder: ast.Fodder{},
									BranchTrue: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(957),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(957),
													Column: int(15),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
													Comment: []string{},
												},
											},
											Ctx: p9934,
											FreeVars: nil,
										},
										Value: "true",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
											Comment: []string{},
										},
									},
									BranchFalse: &ast.Conditional{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(958),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(1034),
													Column: int(42),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p9934,
											FreeVars: ast.Identifiers{
												"aux",
												"cindent",
												"indent_array_in_object",
												"path",
												"std",
												"v",
											},
										},
										Cond: &ast.Binary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(958),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(958),
														Column: int(25),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p9934,
												FreeVars: ast.Identifiers{
													"v",
												},
											},
											Left: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(958),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(958),
															Column: int(16),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p9934,
													FreeVars: ast.Identifiers{
														"v",
													},
												},
												Id: "v",
											},
											OpFodder: ast.Fodder{},
											Op: ast.BinaryOp(12),
											Right: &ast.LiteralBoolean{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(958),
															Column: int(20),
														},
														End: ast.Location{
															Line: int(958),
															Column: int(25),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p9934,
													FreeVars: nil,
												},
												Value: false,
											},
										},
										ThenFodder: ast.Fodder{},
										BranchTrue: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(959),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(959),
														Column: int(16),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
														Comment: []string{},
													},
												},
												Ctx: p9934,
												FreeVars: nil,
											},
											Value: "false",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										BranchFalse: &ast.Conditional{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(960),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(1034),
														Column: int(42),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p9934,
												FreeVars: ast.Identifiers{
													"aux",
													"cindent",
													"indent_array_in_object",
													"path",
													"std",
													"v",
												},
											},
											Cond: &ast.Binary{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(960),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(960),
															Column: int(24),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p9934,
													FreeVars: ast.Identifiers{
														"v",
													},
												},
												Left: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(960),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(960),
																Column: int(16),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p9934,
														FreeVars: ast.Identifiers{
															"v",
														},
													},
													Id: "v",
												},
												OpFodder: ast.Fodder{},
												Op: ast.BinaryOp(12),
												Right: &ast.LiteralNull{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(960),
																Column: int(20),
															},
															End: ast.Location{
																Line: int(960),
																Column: int(24),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p9934,
														FreeVars: nil,
													},
												},
											},
											ThenFodder: ast.Fodder{},
											BranchTrue: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(961),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(961),
															Column: int(15),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
															Comment: []string{},
														},
													},
													Ctx: p9934,
													FreeVars: nil,
												},
												Value: "null",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
													Comment: []string{},
												},
											},
											BranchFalse: &ast.Conditional{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(962),
															Column: int(12),
														},
														End: ast.Location{
															Line: int(1034),
															Column: int(42),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p9934,
													FreeVars: ast.Identifiers{
														"aux",
														"cindent",
														"indent_array_in_object",
														"path",
														"std",
														"v",
													},
												},
												Cond: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(962),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(962),
																Column: int(30),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p9934,
														FreeVars: ast.Identifiers{
															"std",
															"v",
														},
													},
													Target: &ast.Index{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(962),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(962),
																	Column: int(27),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p9934,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(962),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(962),
																		Column: int(18),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Id: "std",
														},
														LeftBracketFodder: ast.Fodder{},
														Index: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
															Value: "isNumber",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														RightBracketFodder: ast.Fodder{},
														Id: nil,
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(962),
																				Column: int(28),
																			},
																			End: ast.Location{
																				Line: int(962),
																				Column: int(29),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p9975,
																		FreeVars: ast.Identifiers{
																			"v",
																		},
																	},
																	Id: "v",
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
												},
												ThenFodder: ast.Fodder{},
												BranchTrue: &ast.Binary{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(963),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(963),
																Column: int(15),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p9934,
														FreeVars: ast.Identifiers{
															"v",
														},
													},
													Left: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(963),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(963),
																	Column: int(11),
																},
																File: p1,
															},
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																	Comment: []string{},
																},
															},
															Ctx: p9934,
															FreeVars: nil,
														},
														Value: "",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													OpFodder: ast.Fodder{},
													Op: ast.BinaryOp(3),
													Right: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(963),
																	Column: int(14),
																},
																End: ast.Location{
																	Line: int(963),
																	Column: int(15),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p9934,
															FreeVars: ast.Identifiers{
																"v",
															},
														},
														Id: "v",
													},
												},
												ElseFodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
														Comment: []string{},
													},
												},
												BranchFalse: &ast.Conditional{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(964),
																Column: int(12),
															},
															End: ast.Location{
																Line: int(1034),
																Column: int(42),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p9934,
														FreeVars: ast.Identifiers{
															"aux",
															"cindent",
															"indent_array_in_object",
															"path",
															"std",
															"v",
														},
													},
													Cond: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(964),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(964),
																	Column: int(30),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p9934,
															FreeVars: ast.Identifiers{
																"std",
																"v",
															},
														},
														Target: &ast.Index{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(964),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(964),
																		Column: int(27),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p9934,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Target: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(964),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(964),
																			Column: int(18),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Id: "std",
															},
															LeftBracketFodder: ast.Fodder{},
															Index: &ast.LiteralString{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: nil,
																},
																Value: "isString",
																Kind: ast.LiteralStringKind(1),
																BlockIndent: "",
																BlockTermIndent: "",
															},
															RightBracketFodder: ast.Fodder{},
															Id: nil,
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(964),
																					Column: int(28),
																				},
																				End: ast.Location{
																					Line: int(964),
																					Column: int(29),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p9995,
																			FreeVars: ast.Identifiers{
																				"v",
																			},
																		},
																		Id: "v",
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														TrailingComma: false,
														TailStrict: false,
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
													},
													ThenFodder: ast.Fodder{},
													BranchTrue: &ast.Local{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(965),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(972),
																	Column: int(34),
																},
																File: p1,
															},
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																	Comment: []string{},
																},
															},
															Ctx: p9934,
															FreeVars: ast.Identifiers{
																"cindent",
																"std",
																"v",
															},
														},
														Binds: ast.LocalBinds{
															ast.LocalBind{
																VarFodder: ast.Fodder{},
																Variable: "len",
																EqFodder: ast.Fodder{},
																Body: &ast.Apply{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(965),
																				Column: int(21),
																			},
																			End: ast.Location{
																				Line: int(965),
																				Column: int(34),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p10002,
																		FreeVars: ast.Identifiers{
																			"std",
																			"v",
																		},
																	},
																	Target: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(965),
																					Column: int(21),
																				},
																				End: ast.Location{
																					Line: int(965),
																					Column: int(31),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p10002,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(965),
																						Column: int(21),
																					},
																					End: ast.Location{
																						Line: int(965),
																						Column: int(24),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Id: "std",
																		},
																		LeftBracketFodder: ast.Fodder{},
																		Index: &ast.LiteralString{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: nil,
																			},
																			Value: "length",
																			Kind: ast.LiteralStringKind(1),
																			BlockIndent: "",
																			BlockTermIndent: "",
																		},
																		RightBracketFodder: ast.Fodder{},
																		Id: nil,
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(965),
																								Column: int(32),
																							},
																							End: ast.Location{
																								Line: int(965),
																								Column: int(33),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p10011,
																						FreeVars: ast.Identifiers{
																							"v",
																						},
																					},
																					Id: "v",
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	TrailingComma: false,
																	TailStrict: false,
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																},
																Fun: nil,
																CloseFodder: ast.Fodder{},
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(965),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(965),
																		Column: int(34),
																	},
																	File: p1,
																},
															},
														},
														Body: &ast.Conditional{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(966),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(972),
																		Column: int(34),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																		Comment: []string{},
																	},
																},
																Ctx: p9934,
																FreeVars: ast.Identifiers{
																	"cindent",
																	"len",
																	"std",
																	"v",
																},
															},
															Cond: &ast.Binary{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(966),
																			Column: int(12),
																		},
																		End: ast.Location{
																			Line: int(966),
																			Column: int(20),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p9934,
																	FreeVars: ast.Identifiers{
																		"len",
																	},
																},
																Left: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(966),
																				Column: int(12),
																			},
																			End: ast.Location{
																				Line: int(966),
																				Column: int(15),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p9934,
																		FreeVars: ast.Identifiers{
																			"len",
																		},
																	},
																	Id: "len",
																},
																OpFodder: ast.Fodder{},
																Op: ast.BinaryOp(12),
																Right: &ast.LiteralNumber{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(966),
																				Column: int(19),
																			},
																			End: ast.Location{
																				Line: int(966),
																				Column: int(20),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p9934,
																		FreeVars: nil,
																	},
																	OriginalString: "0",
																},
															},
															ThenFodder: ast.Fodder{},
															BranchTrue: &ast.LiteralString{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(967),
																			Column: int(11),
																		},
																		End: ast.Location{
																			Line: int(967),
																			Column: int(15),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(10),
																			Comment: []string{},
																		},
																	},
																	Ctx: p9934,
																	FreeVars: nil,
																},
																Value: "\"\"",
																Kind: ast.LiteralStringKind(1),
																BlockIndent: "",
																BlockTermIndent: "",
															},
															ElseFodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																	Comment: []string{},
																},
															},
															BranchFalse: &ast.Conditional{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(968),
																			Column: int(14),
																		},
																		End: ast.Location{
																			Line: int(972),
																			Column: int(34),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p9934,
																	FreeVars: ast.Identifiers{
																		"cindent",
																		"len",
																		"std",
																		"v",
																	},
																},
																Cond: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(968),
																				Column: int(17),
																			},
																			End: ast.Location{
																				Line: int(968),
																				Column: int(35),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p9934,
																		FreeVars: ast.Identifiers{
																			"len",
																			"v",
																		},
																	},
																	Left: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(968),
																					Column: int(17),
																				},
																				End: ast.Location{
																					Line: int(968),
																					Column: int(27),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p9934,
																			FreeVars: ast.Identifiers{
																				"len",
																				"v",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(968),
																						Column: int(17),
																					},
																					End: ast.Location{
																						Line: int(968),
																						Column: int(18),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p9934,
																				FreeVars: ast.Identifiers{
																					"v",
																				},
																			},
																			Id: "v",
																		},
																		LeftBracketFodder: ast.Fodder{},
																		Index: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(968),
																						Column: int(19),
																					},
																					End: ast.Location{
																						Line: int(968),
																						Column: int(26),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p9934,
																				FreeVars: ast.Identifiers{
																					"len",
																				},
																			},
																			Left: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(968),
																							Column: int(19),
																						},
																						End: ast.Location{
																							Line: int(968),
																							Column: int(22),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p9934,
																					FreeVars: ast.Identifiers{
																						"len",
																					},
																				},
																				Id: "len",
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(4),
																			Right: &ast.LiteralNumber{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(968),
																							Column: int(25),
																						},
																						End: ast.Location{
																							Line: int(968),
																							Column: int(26),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p9934,
																					FreeVars: nil,
																				},
																				OriginalString: "1",
																			},
																		},
																		RightBracketFodder: ast.Fodder{},
																		Id: nil,
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(12),
																	Right: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(968),
																					Column: int(31),
																				},
																				End: ast.Location{
																					Line: int(968),
																					Column: int(35),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p9934,
																			FreeVars: nil,
																		},
																		Value: "\n",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																},
																ThenFodder: ast.Fodder{},
																BranchTrue: &ast.Local{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(969),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(970),
																				Column: int(82),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(10),
																				Comment: []string{},
																			},
																		},
																		Ctx: p9934,
																		FreeVars: ast.Identifiers{
																			"cindent",
																			"std",
																			"v",
																		},
																	},
																	Binds: ast.LocalBinds{
																		ast.LocalBind{
																			VarFodder: ast.Fodder{},
																			Variable: "split",
																			EqFodder: ast.Fodder{},
																			Body: &ast.Apply{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(969),
																							Column: int(25),
																						},
																						End: ast.Location{
																							Line: int(969),
																							Column: int(43),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p10043,
																					FreeVars: ast.Identifiers{
																						"std",
																						"v",
																					},
																				},
																				Target: &ast.Index{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(969),
																								Column: int(25),
																							},
																							End: ast.Location{
																								Line: int(969),
																								Column: int(34),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p10043,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																					},
																					Target: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(969),
																									Column: int(25),
																								},
																								End: ast.Location{
																									Line: int(969),
																									Column: int(28),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: nil,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																						},
																						Id: "std",
																					},
																					LeftBracketFodder: ast.Fodder{},
																					Index: &ast.LiteralString{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								File: nil,
																							},
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: nil,
																						},
																						Value: "split",
																						Kind: ast.LiteralStringKind(1),
																						BlockIndent: "",
																						BlockTermIndent: "",
																					},
																					RightBracketFodder: ast.Fodder{},
																					Id: nil,
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(969),
																											Column: int(35),
																										},
																										End: ast.Location{
																											Line: int(969),
																											Column: int(36),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p10052,
																									FreeVars: ast.Identifiers{
																										"v",
																									},
																								},
																								Id: "v",
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.LiteralString{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(969),
																											Column: int(38),
																										},
																										End: ast.Location{
																											Line: int(969),
																											Column: int(42),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p10052,
																									FreeVars: nil,
																								},
																								Value: "\n",
																								Kind: ast.LiteralStringKind(1),
																								BlockIndent: "",
																								BlockTermIndent: "",
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				TrailingComma: false,
																				TailStrict: false,
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: nil,
																			},
																			Fun: nil,
																			CloseFodder: ast.Fodder{},
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(969),
																					Column: int(17),
																				},
																				End: ast.Location{
																					Line: int(969),
																					Column: int(43),
																				},
																				File: p1,
																			},
																		},
																	},
																	Body: &ast.Apply{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(970),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(970),
																					Column: int(82),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p9934,
																			FreeVars: ast.Identifiers{
																				"cindent",
																				"split",
																				"std",
																			},
																		},
																		Target: &ast.Index{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(970),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(970),
																						Column: int(19),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p9934,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(970),
																							Column: int(11),
																						},
																						End: ast.Location{
																							Line: int(970),
																							Column: int(14),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(10),
																							Comment: []string{},
																						},
																					},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																				},
																				Id: "std",
																			},
																			LeftBracketFodder: ast.Fodder{},
																			Index: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: nil,
																				},
																				Value: "join",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																			RightBracketFodder: ast.Fodder{},
																			Id: nil,
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Binary{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(970),
																									Column: int(20),
																								},
																								End: ast.Location{
																									Line: int(970),
																									Column: int(41),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p10065,
																							FreeVars: ast.Identifiers{
																								"cindent",
																							},
																						},
																						Left: &ast.Binary{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(970),
																										Column: int(20),
																									},
																									End: ast.Location{
																										Line: int(970),
																										Column: int(34),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p10065,
																								FreeVars: ast.Identifiers{
																									"cindent",
																								},
																							},
																							Left: &ast.LiteralString{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(970),
																											Column: int(20),
																										},
																										End: ast.Location{
																											Line: int(970),
																											Column: int(24),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p10065,
																									FreeVars: nil,
																								},
																								Value: "\n",
																								Kind: ast.LiteralStringKind(1),
																								BlockIndent: "",
																								BlockTermIndent: "",
																							},
																							OpFodder: ast.Fodder{},
																							Op: ast.BinaryOp(3),
																							Right: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(970),
																											Column: int(27),
																										},
																										End: ast.Location{
																											Line: int(970),
																											Column: int(34),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p10065,
																									FreeVars: ast.Identifiers{
																										"cindent",
																									},
																								},
																								Id: "cindent",
																							},
																						},
																						OpFodder: ast.Fodder{},
																						Op: ast.BinaryOp(3),
																						Right: &ast.LiteralString{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(970),
																										Column: int(37),
																									},
																									End: ast.Location{
																										Line: int(970),
																										Column: int(41),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p10065,
																								FreeVars: nil,
																							},
																							Value: "  ",
																							Kind: ast.LiteralStringKind(1),
																							BlockIndent: "",
																							BlockTermIndent: "",
																						},
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Binary{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(970),
																									Column: int(43),
																								},
																								End: ast.Location{
																									Line: int(970),
																									Column: int(81),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p10065,
																							FreeVars: ast.Identifiers{
																								"split",
																								"std",
																							},
																						},
																						Left: &ast.Array{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(970),
																										Column: int(43),
																									},
																									End: ast.Location{
																										Line: int(970),
																										Column: int(48),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p10065,
																								FreeVars: nil,
																							},
																							Elements: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(970),
																													Column: int(44),
																												},
																												End: ast.Location{
																													Line: int(970),
																													Column: int(47),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p10078,
																											FreeVars: nil,
																										},
																										Value: "|",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																									CommaFodder: nil,
																								},
																							},
																							TrailingComma: false,
																							CloseFodder: ast.Fodder{},
																						},
																						OpFodder: ast.Fodder{},
																						Op: ast.BinaryOp(3),
																						Right: &ast.Apply{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									File: nil,
																								},
																								Fodder: nil,
																								Ctx: nil,
																								FreeVars: ast.Identifiers{
																									"split",
																									"std",
																								},
																							},
																							Target: &ast.Index{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										File: nil,
																									},
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																								},
																								Target: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											File: nil,
																										},
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: ast.Identifiers{
																											"std",
																										},
																									},
																									Id: "std",
																								},
																								LeftBracketFodder: nil,
																								Index: &ast.LiteralString{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											File: nil,
																										},
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: nil,
																									},
																									Value: "slice",
																									Kind: ast.LiteralStringKind(1),
																									BlockIndent: "",
																									BlockTermIndent: "",
																								},
																								RightBracketFodder: nil,
																								Id: nil,
																							},
																							FodderLeft: nil,
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(970),
																														Column: int(51),
																													},
																													End: ast.Location{
																														Line: int(970),
																														Column: int(56),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p10065,
																												FreeVars: ast.Identifiers{
																													"split",
																												},
																											},
																											Id: "split",
																										},
																										CommaFodder: nil,
																									},
																									ast.CommaSeparatedExpr{
																										Expr: &ast.LiteralNumber{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(970),
																														Column: int(57),
																													},
																													End: ast.Location{
																														Line: int(970),
																														Column: int(58),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p10065,
																												FreeVars: nil,
																											},
																											OriginalString: "0",
																										},
																										CommaFodder: nil,
																									},
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Binary{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(970),
																														Column: int(59),
																													},
																													End: ast.Location{
																														Line: int(970),
																														Column: int(80),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p10065,
																												FreeVars: ast.Identifiers{
																													"split",
																													"std",
																												},
																											},
																											Left: &ast.Apply{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(970),
																															Column: int(59),
																														},
																														End: ast.Location{
																															Line: int(970),
																															Column: int(76),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p10065,
																													FreeVars: ast.Identifiers{
																														"split",
																														"std",
																													},
																												},
																												Target: &ast.Index{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(970),
																																Column: int(59),
																															},
																															End: ast.Location{
																																Line: int(970),
																																Column: int(69),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p10065,
																														FreeVars: ast.Identifiers{
																															"std",
																														},
																													},
																													Target: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(970),
																																	Column: int(59),
																																},
																																End: ast.Location{
																																	Line: int(970),
																																	Column: int(62),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: nil,
																															FreeVars: ast.Identifiers{
																																"std",
																															},
																														},
																														Id: "std",
																													},
																													LeftBracketFodder: ast.Fodder{},
																													Index: &ast.LiteralString{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																End: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																File: nil,
																															},
																															Fodder: nil,
																															Ctx: nil,
																															FreeVars: nil,
																														},
																														Value: "length",
																														Kind: ast.LiteralStringKind(1),
																														BlockIndent: "",
																														BlockTermIndent: "",
																													},
																													RightBracketFodder: ast.Fodder{},
																													Id: nil,
																												},
																												FodderLeft: ast.Fodder{},
																												Arguments: ast.Arguments{
																													Positional: []ast.CommaSeparatedExpr{
																														ast.CommaSeparatedExpr{
																															Expr: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(970),
																																			Column: int(70),
																																		},
																																		End: ast.Location{
																																			Line: int(970),
																																			Column: int(75),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p10101,
																																	FreeVars: ast.Identifiers{
																																		"split",
																																	},
																																},
																																Id: "split",
																															},
																															CommaFodder: nil,
																														},
																													},
																													Named: nil,
																												},
																												TrailingComma: false,
																												TailStrict: false,
																												FodderRight: ast.Fodder{},
																												TailStrictFodder: nil,
																											},
																											OpFodder: ast.Fodder{},
																											Op: ast.BinaryOp(4),
																											Right: &ast.LiteralNumber{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(970),
																															Column: int(79),
																														},
																														End: ast.Location{
																															Line: int(970),
																															Column: int(80),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p10065,
																													FreeVars: nil,
																												},
																												OriginalString: "1",
																											},
																										},
																										CommaFodder: nil,
																									},
																									ast.CommaSeparatedExpr{
																										Expr: &ast.LiteralNull{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													File: nil,
																												},
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: nil,
																											},
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							TrailingComma: false,
																							TailStrict: false,
																							FodderRight: nil,
																							TailStrictFodder: nil,
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		TrailingComma: false,
																		TailStrict: false,
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																	},
																},
																ElseFodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																		Comment: []string{},
																	},
																},
																BranchFalse: &ast.Apply{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(972),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(972),
																				Column: int(34),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p9934,
																		FreeVars: ast.Identifiers{
																			"std",
																			"v",
																		},
																	},
																	Target: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(972),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(972),
																					Column: int(31),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p9934,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(972),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(972),
																						Column: int(14),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(10),
																						Comment: []string{},
																					},
																				},
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Id: "std",
																		},
																		LeftBracketFodder: ast.Fodder{},
																		Index: &ast.LiteralString{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: nil,
																			},
																			Value: "escapeStringJson",
																			Kind: ast.LiteralStringKind(1),
																			BlockIndent: "",
																			BlockTermIndent: "",
																		},
																		RightBracketFodder: ast.Fodder{},
																		Id: nil,
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(972),
																								Column: int(32),
																							},
																							End: ast.Location{
																								Line: int(972),
																								Column: int(33),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p10116,
																						FreeVars: ast.Identifiers{
																							"v",
																						},
																					},
																					Id: "v",
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	TrailingComma: false,
																	TailStrict: false,
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																},
															},
														},
													},
													ElseFodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(6),
															Comment: []string{},
														},
													},
													BranchFalse: &ast.Conditional{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(973),
																	Column: int(12),
																},
																End: ast.Location{
																	Line: int(1034),
																	Column: int(42),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p9934,
															FreeVars: ast.Identifiers{
																"aux",
																"cindent",
																"indent_array_in_object",
																"path",
																"std",
																"v",
															},
														},
														Cond: &ast.Apply{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(973),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(973),
																		Column: int(32),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p9934,
																FreeVars: ast.Identifiers{
																	"std",
																	"v",
																},
															},
															Target: &ast.Index{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(973),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(973),
																			Column: int(29),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p9934,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(973),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(973),
																				Column: int(18),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Id: "std",
																},
																LeftBracketFodder: ast.Fodder{},
																Index: &ast.LiteralString{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: nil,
																	},
																	Value: "isFunction",
																	Kind: ast.LiteralStringKind(1),
																	BlockIndent: "",
																	BlockTermIndent: "",
																},
																RightBracketFodder: ast.Fodder{},
																Id: nil,
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(973),
																						Column: int(30),
																					},
																					End: ast.Location{
																						Line: int(973),
																						Column: int(31),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p10130,
																				FreeVars: ast.Identifiers{
																					"v",
																				},
																			},
																			Id: "v",
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															TrailingComma: false,
															TailStrict: false,
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
														},
														ThenFodder: ast.Fodder{},
														BranchTrue: &ast.Error{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(974),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(974),
																		Column: int(54),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																		Comment: []string{},
																	},
																},
																Ctx: p9934,
																FreeVars: ast.Identifiers{
																	"path",
																},
															},
															Expr: &ast.Binary{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(974),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(974),
																			Column: int(54),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p9934,
																	FreeVars: ast.Identifiers{
																		"path",
																	},
																},
																Left: &ast.LiteralString{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(974),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(974),
																				Column: int(47),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p9934,
																		FreeVars: nil,
																	},
																	Value: "Tried to manifest function at ",
																	Kind: ast.LiteralStringKind(1),
																	BlockIndent: "",
																	BlockTermIndent: "",
																},
																OpFodder: ast.Fodder{},
																Op: ast.BinaryOp(3),
																Right: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(974),
																				Column: int(50),
																			},
																			End: ast.Location{
																				Line: int(974),
																				Column: int(54),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p9934,
																		FreeVars: ast.Identifiers{
																			"path",
																		},
																	},
																	Id: "path",
																},
															},
														},
														ElseFodder: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(6),
																Comment: []string{},
															},
														},
														BranchFalse: &ast.Conditional{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(975),
																		Column: int(12),
																	},
																	End: ast.Location{
																		Line: int(1034),
																		Column: int(42),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p9934,
																FreeVars: ast.Identifiers{
																	"aux",
																	"cindent",
																	"indent_array_in_object",
																	"path",
																	"std",
																	"v",
																},
															},
															Cond: &ast.Apply{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(975),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(975),
																			Column: int(29),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p9934,
																	FreeVars: ast.Identifiers{
																		"std",
																		"v",
																	},
																},
																Target: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(975),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(975),
																				Column: int(26),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p9934,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(975),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(975),
																					Column: int(18),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Id: "std",
																	},
																	LeftBracketFodder: ast.Fodder{},
																	Index: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: nil,
																		},
																		Value: "isArray",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	RightBracketFodder: ast.Fodder{},
																	Id: nil,
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(975),
																							Column: int(27),
																						},
																						End: ast.Location{
																							Line: int(975),
																							Column: int(28),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p10152,
																					FreeVars: ast.Identifiers{
																						"v",
																					},
																				},
																				Id: "v",
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																TrailingComma: false,
																TailStrict: false,
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
															},
															ThenFodder: ast.Fodder{},
															BranchTrue: &ast.Conditional{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(976),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(1006),
																			Column: int(42),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																			Comment: []string{},
																		},
																	},
																	Ctx: p9934,
																	FreeVars: ast.Identifiers{
																		"aux",
																		"cindent",
																		"path",
																		"std",
																		"v",
																	},
																},
																Cond: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(976),
																				Column: int(12),
																			},
																			End: ast.Location{
																				Line: int(976),
																				Column: int(30),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p9934,
																		FreeVars: ast.Identifiers{
																			"std",
																			"v",
																		},
																	},
																	Left: &ast.Apply{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(976),
																					Column: int(12),
																				},
																				End: ast.Location{
																					Line: int(976),
																					Column: int(25),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p9934,
																			FreeVars: ast.Identifiers{
																				"std",
																				"v",
																			},
																		},
																		Target: &ast.Index{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(976),
																						Column: int(12),
																					},
																					End: ast.Location{
																						Line: int(976),
																						Column: int(22),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p9934,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(976),
																							Column: int(12),
																						},
																						End: ast.Location{
																							Line: int(976),
																							Column: int(15),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																				},
																				Id: "std",
																			},
																			LeftBracketFodder: ast.Fodder{},
																			Index: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: nil,
																				},
																				Value: "length",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																			RightBracketFodder: ast.Fodder{},
																			Id: nil,
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(976),
																									Column: int(23),
																								},
																								End: ast.Location{
																									Line: int(976),
																									Column: int(24),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p10168,
																							FreeVars: ast.Identifiers{
																								"v",
																							},
																						},
																						Id: "v",
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		TrailingComma: false,
																		TailStrict: false,
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(12),
																	Right: &ast.LiteralNumber{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(976),
																					Column: int(29),
																				},
																				End: ast.Location{
																					Line: int(976),
																					Column: int(30),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p9934,
																			FreeVars: nil,
																		},
																		OriginalString: "0",
																	},
																},
																ThenFodder: ast.Fodder{},
																BranchTrue: &ast.LiteralString{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(977),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(977),
																				Column: int(15),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(10),
																				Comment: []string{},
																			},
																		},
																		Ctx: p9934,
																		FreeVars: nil,
																	},
																	Value: "[]",
																	Kind: ast.LiteralStringKind(1),
																	BlockIndent: "",
																	BlockTermIndent: "",
																},
																ElseFodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																		Comment: []string{},
																	},
																},
																BranchFalse: &ast.Local{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(979),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(1006),
																				Column: int(42),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(10),
																				Comment: []string{},
																			},
																		},
																		Ctx: p9934,
																		FreeVars: ast.Identifiers{
																			"aux",
																			"cindent",
																			"path",
																			"std",
																			"v",
																		},
																	},
																	Binds: ast.LocalBinds{
																		ast.LocalBind{
																			VarFodder: nil,
																			Variable: "params",
																			EqFodder: nil,
																			Body: &ast.Function{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(979),
																							Column: int(17),
																						},
																						End: ast.Location{
																							Line: int(999),
																							Column: int(14),
																						},
																						File: p1,
																					},
																					Fodder: nil,
																					Ctx: p10179,
																					FreeVars: ast.Identifiers{
																						"cindent",
																						"std",
																					},
																				},
																				ParenLeftFodder: ast.Fodder{},
																				Parameters: []ast.Parameter{
																					ast.Parameter{
																						NameFodder: ast.Fodder{},
																						Name: "value",
																						EqFodder: nil,
																						DefaultArg: nil,
																						CommaFodder: nil,
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(979),
																								Column: int(24),
																							},
																							End: ast.Location{
																								Line: int(979),
																								Column: int(29),
																							},
																							File: p1,
																						},
																					},
																				},
																				TrailingComma: false,
																				ParenRightFodder: ast.Fodder{},
																				Body: &ast.Conditional{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(980),
																								Column: int(13),
																							},
																							End: ast.Location{
																								Line: int(999),
																								Column: int(14),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(12),
																								Comment: []string{},
																							},
																						},
																						Ctx: p10184,
																						FreeVars: ast.Identifiers{
																							"cindent",
																							"std",
																							"value",
																						},
																					},
																					Cond: &ast.Binary{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(980),
																									Column: int(16),
																								},
																								End: ast.Location{
																									Line: int(980),
																									Column: int(59),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p10184,
																							FreeVars: ast.Identifiers{
																								"std",
																								"value",
																							},
																						},
																						Left: &ast.Apply{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(980),
																										Column: int(16),
																									},
																									End: ast.Location{
																										Line: int(980),
																										Column: int(34),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p10184,
																								FreeVars: ast.Identifiers{
																									"std",
																									"value",
																								},
																							},
																							Target: &ast.Index{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(980),
																											Column: int(16),
																										},
																										End: ast.Location{
																											Line: int(980),
																											Column: int(27),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p10184,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																								},
																								Target: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(980),
																												Column: int(16),
																											},
																											End: ast.Location{
																												Line: int(980),
																												Column: int(19),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: nil,
																										FreeVars: ast.Identifiers{
																											"std",
																										},
																									},
																									Id: "std",
																								},
																								LeftBracketFodder: ast.Fodder{},
																								Index: &ast.LiteralString{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											File: nil,
																										},
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: nil,
																									},
																									Value: "isArray",
																									Kind: ast.LiteralStringKind(1),
																									BlockIndent: "",
																									BlockTermIndent: "",
																								},
																								RightBracketFodder: ast.Fodder{},
																								Id: nil,
																							},
																							FodderLeft: ast.Fodder{},
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(980),
																														Column: int(28),
																													},
																													End: ast.Location{
																														Line: int(980),
																														Column: int(33),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p10197,
																												FreeVars: ast.Identifiers{
																													"value",
																												},
																											},
																											Id: "value",
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							TrailingComma: false,
																							TailStrict: false,
																							FodderRight: ast.Fodder{},
																							TailStrictFodder: nil,
																						},
																						OpFodder: ast.Fodder{},
																						Op: ast.BinaryOp(17),
																						Right: &ast.Binary{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(980),
																										Column: int(38),
																									},
																									End: ast.Location{
																										Line: int(980),
																										Column: int(59),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p10184,
																								FreeVars: ast.Identifiers{
																									"std",
																									"value",
																								},
																							},
																							Left: &ast.Apply{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(980),
																											Column: int(38),
																										},
																										End: ast.Location{
																											Line: int(980),
																											Column: int(55),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p10184,
																									FreeVars: ast.Identifiers{
																										"std",
																										"value",
																									},
																								},
																								Target: &ast.Index{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(980),
																												Column: int(38),
																											},
																											End: ast.Location{
																												Line: int(980),
																												Column: int(48),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p10184,
																										FreeVars: ast.Identifiers{
																											"std",
																										},
																									},
																									Target: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(980),
																													Column: int(38),
																												},
																												End: ast.Location{
																													Line: int(980),
																													Column: int(41),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"std",
																											},
																										},
																										Id: "std",
																									},
																									LeftBracketFodder: ast.Fodder{},
																									Index: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												File: nil,
																											},
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: nil,
																										},
																										Value: "length",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																									RightBracketFodder: ast.Fodder{},
																									Id: nil,
																								},
																								FodderLeft: ast.Fodder{},
																								Arguments: ast.Arguments{
																									Positional: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(980),
																															Column: int(49),
																														},
																														End: ast.Location{
																															Line: int(980),
																															Column: int(54),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p10210,
																													FreeVars: ast.Identifiers{
																														"value",
																													},
																												},
																												Id: "value",
																											},
																											CommaFodder: nil,
																										},
																									},
																									Named: nil,
																								},
																								TrailingComma: false,
																								TailStrict: false,
																								FodderRight: ast.Fodder{},
																								TailStrictFodder: nil,
																							},
																							OpFodder: ast.Fodder{},
																							Op: ast.BinaryOp(7),
																							Right: &ast.LiteralNumber{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(980),
																											Column: int(58),
																										},
																										End: ast.Location{
																											Line: int(980),
																											Column: int(59),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p10184,
																									FreeVars: nil,
																								},
																								OriginalString: "0",
																							},
																						},
																					},
																					ThenFodder: ast.Fodder{},
																					BranchTrue: &ast.DesugaredObject{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(980),
																									Column: int(65),
																								},
																								End: ast.Location{
																									Line: int(989),
																									Column: int(14),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p10184,
																							FreeVars: ast.Identifiers{
																								"cindent",
																							},
																						},
																						Asserts: ast.Nodes{},
																						Fields: ast.DesugaredObjectFields{
																							ast.DesugaredObjectField{
																								Hide: ast.ObjectFieldHide(1),
																								Name: &ast.LiteralString{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											File: nil,
																										},
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: nil,
																									},
																									Value: "new_indent",
																									Kind: ast.LiteralStringKind(1),
																									BlockIndent: "",
																									BlockTermIndent: "",
																								},
																								Body: &ast.Binary{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(987),
																												Column: int(27),
																											},
																											End: ast.Location{
																												Line: int(987),
																												Column: int(41),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p10218,
																										FreeVars: ast.Identifiers{
																											"cindent",
																										},
																									},
																									Left: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(987),
																													Column: int(27),
																												},
																												End: ast.Location{
																													Line: int(987),
																													Column: int(34),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p10218,
																											FreeVars: ast.Identifiers{
																												"cindent",
																											},
																										},
																										Id: "cindent",
																									},
																									OpFodder: ast.Fodder{},
																									Op: ast.BinaryOp(3),
																									Right: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(987),
																													Column: int(37),
																												},
																												End: ast.Location{
																													Line: int(987),
																													Column: int(41),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p10218,
																											FreeVars: nil,
																										},
																										Value: "  ",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																								},
																								PlusSuper: false,
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(987),
																										Column: int(15),
																									},
																									End: ast.Location{
																										Line: int(987),
																										Column: int(41),
																									},
																									File: p1,
																								},
																							},
																							ast.DesugaredObjectField{
																								Hide: ast.ObjectFieldHide(1),
																								Name: &ast.LiteralString{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											File: nil,
																										},
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: nil,
																									},
																									Value: "space",
																									Kind: ast.LiteralStringKind(1),
																									BlockIndent: "",
																									BlockTermIndent: "",
																								},
																								Body: &ast.Binary{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(988),
																												Column: int(22),
																											},
																											End: ast.Location{
																												Line: int(988),
																												Column: int(44),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p10218,
																										FreeVars: nil,
																									},
																									Left: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(988),
																													Column: int(22),
																												},
																												End: ast.Location{
																													Line: int(988),
																													Column: int(26),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p10218,
																											FreeVars: nil,
																										},
																										Value: "\n",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																									OpFodder: ast.Fodder{},
																									Op: ast.BinaryOp(3),
																									Right: &ast.Index{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(988),
																													Column: int(29),
																												},
																												End: ast.Location{
																													Line: int(988),
																													Column: int(44),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p10218,
																											FreeVars: nil,
																										},
																										Target: &ast.Self{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(988),
																														Column: int(29),
																													},
																													End: ast.Location{
																														Line: int(988),
																														Column: int(33),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: nil,
																												FreeVars: nil,
																											},
																										},
																										LeftBracketFodder: ast.Fodder{},
																										Index: &ast.LiteralString{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													File: nil,
																												},
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: nil,
																											},
																											Value: "new_indent",
																											Kind: ast.LiteralStringKind(1),
																											BlockIndent: "",
																											BlockTermIndent: "",
																										},
																										RightBracketFodder: ast.Fodder{},
																										Id: nil,
																									},
																								},
																								PlusSuper: false,
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(988),
																										Column: int(15),
																									},
																									End: ast.Location{
																										Line: int(988),
																										Column: int(44),
																									},
																									File: p1,
																								},
																							},
																						},
																						Locals: ast.LocalBinds{},
																					},
																					ElseFodder: ast.Fodder{},
																					BranchFalse: &ast.Conditional{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(989),
																									Column: int(20),
																								},
																								End: ast.Location{
																									Line: int(999),
																									Column: int(14),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p10184,
																							FreeVars: ast.Identifiers{
																								"cindent",
																								"std",
																								"value",
																							},
																						},
																						Cond: &ast.Binary{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(989),
																										Column: int(23),
																									},
																									End: ast.Location{
																										Line: int(989),
																										Column: int(67),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p10184,
																								FreeVars: ast.Identifiers{
																									"std",
																									"value",
																								},
																							},
																							Left: &ast.Apply{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(989),
																											Column: int(23),
																										},
																										End: ast.Location{
																											Line: int(989),
																											Column: int(42),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p10184,
																									FreeVars: ast.Identifiers{
																										"std",
																										"value",
																									},
																								},
																								Target: &ast.Index{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(989),
																												Column: int(23),
																											},
																											End: ast.Location{
																												Line: int(989),
																												Column: int(35),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p10184,
																										FreeVars: ast.Identifiers{
																											"std",
																										},
																									},
																									Target: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(989),
																													Column: int(23),
																												},
																												End: ast.Location{
																													Line: int(989),
																													Column: int(26),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"std",
																											},
																										},
																										Id: "std",
																									},
																									LeftBracketFodder: ast.Fodder{},
																									Index: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												File: nil,
																											},
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: nil,
																										},
																										Value: "isObject",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																									RightBracketFodder: ast.Fodder{},
																									Id: nil,
																								},
																								FodderLeft: ast.Fodder{},
																								Arguments: ast.Arguments{
																									Positional: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(989),
																															Column: int(36),
																														},
																														End: ast.Location{
																															Line: int(989),
																															Column: int(41),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p10242,
																													FreeVars: ast.Identifiers{
																														"value",
																													},
																												},
																												Id: "value",
																											},
																											CommaFodder: nil,
																										},
																									},
																									Named: nil,
																								},
																								TrailingComma: false,
																								TailStrict: false,
																								FodderRight: ast.Fodder{},
																								TailStrictFodder: nil,
																							},
																							OpFodder: ast.Fodder{},
																							Op: ast.BinaryOp(17),
																							Right: &ast.Binary{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(989),
																											Column: int(46),
																										},
																										End: ast.Location{
																											Line: int(989),
																											Column: int(67),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p10184,
																									FreeVars: ast.Identifiers{
																										"std",
																										"value",
																									},
																								},
																								Left: &ast.Apply{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(989),
																												Column: int(46),
																											},
																											End: ast.Location{
																												Line: int(989),
																												Column: int(63),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p10184,
																										FreeVars: ast.Identifiers{
																											"std",
																											"value",
																										},
																									},
																									Target: &ast.Index{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(989),
																													Column: int(46),
																												},
																												End: ast.Location{
																													Line: int(989),
																													Column: int(56),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p10184,
																											FreeVars: ast.Identifiers{
																												"std",
																											},
																										},
																										Target: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(989),
																														Column: int(46),
																													},
																													End: ast.Location{
																														Line: int(989),
																														Column: int(49),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																											},
																											Id: "std",
																										},
																										LeftBracketFodder: ast.Fodder{},
																										Index: &ast.LiteralString{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													File: nil,
																												},
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: nil,
																											},
																											Value: "length",
																											Kind: ast.LiteralStringKind(1),
																											BlockIndent: "",
																											BlockTermIndent: "",
																										},
																										RightBracketFodder: ast.Fodder{},
																										Id: nil,
																									},
																									FodderLeft: ast.Fodder{},
																									Arguments: ast.Arguments{
																										Positional: []ast.CommaSeparatedExpr{
																											ast.CommaSeparatedExpr{
																												Expr: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(989),
																																Column: int(57),
																															},
																															End: ast.Location{
																																Line: int(989),
																																Column: int(62),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p10255,
																														FreeVars: ast.Identifiers{
																															"value",
																														},
																													},
																													Id: "value",
																												},
																												CommaFodder: nil,
																											},
																										},
																										Named: nil,
																									},
																									TrailingComma: false,
																									TailStrict: false,
																									FodderRight: ast.Fodder{},
																									TailStrictFodder: nil,
																								},
																								OpFodder: ast.Fodder{},
																								Op: ast.BinaryOp(7),
																								Right: &ast.LiteralNumber{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(989),
																												Column: int(66),
																											},
																											End: ast.Location{
																												Line: int(989),
																												Column: int(67),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p10184,
																										FreeVars: nil,
																									},
																									OriginalString: "0",
																								},
																							},
																						},
																						ThenFodder: ast.Fodder{},
																						BranchTrue: &ast.DesugaredObject{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(989),
																										Column: int(73),
																									},
																									End: ast.Location{
																										Line: int(995),
																										Column: int(14),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p10184,
																								FreeVars: ast.Identifiers{
																									"cindent",
																								},
																							},
																							Asserts: ast.Nodes{},
																							Fields: ast.DesugaredObjectFields{
																								ast.DesugaredObjectField{
																									Hide: ast.ObjectFieldHide(1),
																									Name: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												File: nil,
																											},
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: nil,
																										},
																										Value: "new_indent",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																									Body: &ast.Binary{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(990),
																													Column: int(27),
																												},
																												End: ast.Location{
																													Line: int(990),
																													Column: int(41),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p10263,
																											FreeVars: ast.Identifiers{
																												"cindent",
																											},
																										},
																										Left: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(990),
																														Column: int(27),
																													},
																													End: ast.Location{
																														Line: int(990),
																														Column: int(34),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p10263,
																												FreeVars: ast.Identifiers{
																													"cindent",
																												},
																											},
																											Id: "cindent",
																										},
																										OpFodder: ast.Fodder{},
																										Op: ast.BinaryOp(3),
																										Right: &ast.LiteralString{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(990),
																														Column: int(37),
																													},
																													End: ast.Location{
																														Line: int(990),
																														Column: int(41),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p10263,
																												FreeVars: nil,
																											},
																											Value: "  ",
																											Kind: ast.LiteralStringKind(1),
																											BlockIndent: "",
																											BlockTermIndent: "",
																										},
																									},
																									PlusSuper: false,
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(990),
																											Column: int(15),
																										},
																										End: ast.Location{
																											Line: int(990),
																											Column: int(41),
																										},
																										File: p1,
																									},
																								},
																								ast.DesugaredObjectField{
																									Hide: ast.ObjectFieldHide(1),
																									Name: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												File: nil,
																											},
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: nil,
																										},
																										Value: "space",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																									Body: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(994),
																													Column: int(22),
																												},
																												End: ast.Location{
																													Line: int(994),
																													Column: int(25),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p10263,
																											FreeVars: nil,
																										},
																										Value: " ",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																									PlusSuper: false,
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(994),
																											Column: int(15),
																										},
																										End: ast.Location{
																											Line: int(994),
																											Column: int(25),
																										},
																										File: p1,
																									},
																								},
																							},
																							Locals: ast.LocalBinds{},
																						},
																						ElseFodder: ast.Fodder{},
																						BranchFalse: &ast.DesugaredObject{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(995),
																										Column: int(20),
																									},
																									End: ast.Location{
																										Line: int(999),
																										Column: int(14),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p10184,
																								FreeVars: ast.Identifiers{
																									"cindent",
																								},
																							},
																							Asserts: ast.Nodes{},
																							Fields: ast.DesugaredObjectFields{
																								ast.DesugaredObjectField{
																									Hide: ast.ObjectFieldHide(1),
																									Name: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												File: nil,
																											},
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: nil,
																										},
																										Value: "new_indent",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																									Body: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(997),
																													Column: int(27),
																												},
																												End: ast.Location{
																													Line: int(997),
																													Column: int(34),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p10275,
																											FreeVars: ast.Identifiers{
																												"cindent",
																											},
																										},
																										Id: "cindent",
																									},
																									PlusSuper: false,
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(997),
																											Column: int(15),
																										},
																										End: ast.Location{
																											Line: int(997),
																											Column: int(34),
																										},
																										File: p1,
																									},
																								},
																								ast.DesugaredObjectField{
																									Hide: ast.ObjectFieldHide(1),
																									Name: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												File: nil,
																											},
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: nil,
																										},
																										Value: "space",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																									Body: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(998),
																													Column: int(22),
																												},
																												End: ast.Location{
																													Line: int(998),
																													Column: int(25),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p10275,
																											FreeVars: nil,
																										},
																										Value: " ",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																									PlusSuper: false,
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(998),
																											Column: int(15),
																										},
																										End: ast.Location{
																											Line: int(998),
																											Column: int(25),
																										},
																										File: p1,
																									},
																								},
																							},
																							Locals: ast.LocalBinds{},
																						},
																					},
																				},
																			},
																			Fun: nil,
																			CloseFodder: nil,
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																		},
																	},
																	Body: &ast.Local{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1000),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(1006),
																					Column: int(42),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(10),
																					Comment: []string{},
																				},
																			},
																			Ctx: p9934,
																			FreeVars: ast.Identifiers{
																				"aux",
																				"cindent",
																				"params",
																				"path",
																				"std",
																				"v",
																			},
																		},
																		Binds: ast.LocalBinds{
																			ast.LocalBind{
																				VarFodder: ast.Fodder{},
																				Variable: "range",
																				EqFodder: ast.Fodder{},
																				Body: &ast.Apply{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1000),
																								Column: int(25),
																							},
																							End: ast.Location{
																								Line: int(1000),
																								Column: int(56),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p10284,
																						FreeVars: ast.Identifiers{
																							"std",
																							"v",
																						},
																					},
																					Target: &ast.Index{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1000),
																									Column: int(25),
																								},
																								End: ast.Location{
																									Line: int(1000),
																									Column: int(34),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p10284,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																						},
																						Target: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1000),
																										Column: int(25),
																									},
																									End: ast.Location{
																										Line: int(1000),
																										Column: int(28),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: nil,
																								FreeVars: ast.Identifiers{
																									"std",
																								},
																							},
																							Id: "std",
																						},
																						LeftBracketFodder: ast.Fodder{},
																						Index: &ast.LiteralString{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									File: nil,
																								},
																								Fodder: nil,
																								Ctx: nil,
																								FreeVars: nil,
																							},
																							Value: "range",
																							Kind: ast.LiteralStringKind(1),
																							BlockIndent: "",
																							BlockTermIndent: "",
																						},
																						RightBracketFodder: ast.Fodder{},
																						Id: nil,
																					},
																					FodderLeft: ast.Fodder{},
																					Arguments: ast.Arguments{
																						Positional: []ast.CommaSeparatedExpr{
																							ast.CommaSeparatedExpr{
																								Expr: &ast.LiteralNumber{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1000),
																												Column: int(35),
																											},
																											End: ast.Location{
																												Line: int(1000),
																												Column: int(36),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p10293,
																										FreeVars: nil,
																									},
																									OriginalString: "0",
																								},
																								CommaFodder: ast.Fodder{},
																							},
																							ast.CommaSeparatedExpr{
																								Expr: &ast.Binary{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1000),
																												Column: int(38),
																											},
																											End: ast.Location{
																												Line: int(1000),
																												Column: int(55),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p10293,
																										FreeVars: ast.Identifiers{
																											"std",
																											"v",
																										},
																									},
																									Left: &ast.Apply{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(1000),
																													Column: int(38),
																												},
																												End: ast.Location{
																													Line: int(1000),
																													Column: int(51),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p10293,
																											FreeVars: ast.Identifiers{
																												"std",
																												"v",
																											},
																										},
																										Target: &ast.Index{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1000),
																														Column: int(38),
																													},
																													End: ast.Location{
																														Line: int(1000),
																														Column: int(48),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p10293,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																											},
																											Target: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(1000),
																															Column: int(38),
																														},
																														End: ast.Location{
																															Line: int(1000),
																															Column: int(41),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																												},
																												Id: "std",
																											},
																											LeftBracketFodder: ast.Fodder{},
																											Index: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: nil,
																												},
																												Value: "length",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																											RightBracketFodder: ast.Fodder{},
																											Id: nil,
																										},
																										FodderLeft: ast.Fodder{},
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(1000),
																																	Column: int(49),
																																},
																																End: ast.Location{
																																	Line: int(1000),
																																	Column: int(50),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p10305,
																															FreeVars: ast.Identifiers{
																																"v",
																															},
																														},
																														Id: "v",
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										TrailingComma: false,
																										TailStrict: false,
																										FodderRight: ast.Fodder{},
																										TailStrictFodder: nil,
																									},
																									OpFodder: ast.Fodder{},
																									Op: ast.BinaryOp(4),
																									Right: &ast.LiteralNumber{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(1000),
																													Column: int(54),
																												},
																												End: ast.Location{
																													Line: int(1000),
																													Column: int(55),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p10293,
																											FreeVars: nil,
																										},
																										OriginalString: "1",
																									},
																								},
																								CommaFodder: nil,
																							},
																						},
																						Named: nil,
																					},
																					TrailingComma: false,
																					TailStrict: false,
																					FodderRight: ast.Fodder{},
																					TailStrictFodder: nil,
																				},
																				Fun: nil,
																				CloseFodder: ast.Fodder{},
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1000),
																						Column: int(17),
																					},
																					End: ast.Location{
																						Line: int(1000),
																						Column: int(56),
																					},
																					File: p1,
																				},
																			},
																		},
																		Body: &ast.Local{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1001),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(1006),
																						Column: int(42),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(10),
																						Comment: []string{},
																					},
																				},
																				Ctx: p9934,
																				FreeVars: ast.Identifiers{
																					"aux",
																					"cindent",
																					"params",
																					"path",
																					"range",
																					"std",
																					"v",
																				},
																			},
																			Binds: ast.LocalBinds{
																				ast.LocalBind{
																					VarFodder: ast.Fodder{},
																					Variable: "parts",
																					EqFodder: ast.Fodder{},
																					Body: &ast.Apply{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								File: nil,
																							},
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{
																								"aux",
																								"params",
																								"path",
																								"range",
																								"std",
																								"v",
																							},
																						},
																						Target: &ast.Index{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									File: nil,
																								},
																								Fodder: nil,
																								Ctx: nil,
																								FreeVars: ast.Identifiers{
																									"std",
																								},
																							},
																							Target: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										File: nil,
																									},
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																								},
																								Id: "std",
																							},
																							LeftBracketFodder: nil,
																							Index: &ast.LiteralString{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										File: nil,
																									},
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: nil,
																								},
																								Value: "flatMap",
																								Kind: ast.LiteralStringKind(1),
																								BlockIndent: "",
																								BlockTermIndent: "",
																							},
																							RightBracketFodder: nil,
																							Id: nil,
																						},
																						FodderLeft: nil,
																						Arguments: ast.Arguments{
																							Positional: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Function{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												File: nil,
																											},
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"aux",
																												"params",
																												"path",
																												"std",
																												"v",
																											},
																										},
																										ParenLeftFodder: nil,
																										Parameters: []ast.Parameter{
																											ast.Parameter{
																												NameFodder: nil,
																												Name: "i",
																												EqFodder: nil,
																												DefaultArg: nil,
																												CommaFodder: nil,
																												LocRange: ast.LocationRange{
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													File: nil,
																												},
																											},
																										},
																										TrailingComma: false,
																										ParenRightFodder: nil,
																										Body: &ast.Apply{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													File: nil,
																												},
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"aux",
																													"i",
																													"params",
																													"path",
																													"std",
																													"v",
																												},
																											},
																											Target: &ast.Index{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																												},
																												Target: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															File: nil,
																														},
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"std",
																														},
																													},
																													Id: "std",
																												},
																												LeftBracketFodder: nil,
																												Index: &ast.LiteralString{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															File: nil,
																														},
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: nil,
																													},
																													Value: "flatMap",
																													Kind: ast.LiteralStringKind(1),
																													BlockIndent: "",
																													BlockTermIndent: "",
																												},
																												RightBracketFodder: nil,
																												Id: nil,
																											},
																											FodderLeft: nil,
																											Arguments: ast.Arguments{
																												Positional: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Function{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	File: nil,
																																},
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"aux",
																																	"i",
																																	"path",
																																	"v",
																																},
																															},
																															ParenLeftFodder: nil,
																															Parameters: []ast.Parameter{
																																ast.Parameter{
																																	NameFodder: nil,
																																	Name: "param",
																																	EqFodder: nil,
																																	DefaultArg: nil,
																																	CommaFodder: nil,
																																	LocRange: ast.LocationRange{
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		File: nil,
																																	},
																																},
																															},
																															TrailingComma: false,
																															ParenRightFodder: nil,
																															Body: &ast.Array{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		File: nil,
																																	},
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"aux",
																																		"i",
																																		"param",
																																		"path",
																																		"v",
																																	},
																																},
																																Elements: []ast.CommaSeparatedExpr{
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Binary{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(1002),
																																						Column: int(13),
																																					},
																																					End: ast.Location{
																																						Line: int(1002),
																																						Column: int(72),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p10338,
																																				FreeVars: ast.Identifiers{
																																					"aux",
																																					"i",
																																					"param",
																																					"path",
																																					"v",
																																				},
																																			},
																																			Left: &ast.Binary{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(1002),
																																							Column: int(13),
																																						},
																																						End: ast.Location{
																																							Line: int(1002),
																																							Column: int(30),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p10338,
																																					FreeVars: ast.Identifiers{
																																						"param",
																																					},
																																				},
																																				Left: &ast.LiteralString{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(1002),
																																								Column: int(13),
																																							},
																																							End: ast.Location{
																																								Line: int(1002),
																																								Column: int(16),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{
																																							ast.FodderElement{
																																								Kind: ast.FodderKind(0),
																																								Blanks: int(0),
																																								Indent: int(12),
																																								Comment: []string{},
																																							},
																																						},
																																						Ctx: p10338,
																																						FreeVars: nil,
																																					},
																																					Value: "-",
																																					Kind: ast.LiteralStringKind(1),
																																					BlockIndent: "",
																																					BlockTermIndent: "",
																																				},
																																				OpFodder: ast.Fodder{},
																																				Op: ast.BinaryOp(3),
																																				Right: &ast.Index{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(1002),
																																								Column: int(19),
																																							},
																																							End: ast.Location{
																																								Line: int(1002),
																																								Column: int(30),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p10338,
																																						FreeVars: ast.Identifiers{
																																							"param",
																																						},
																																					},
																																					Target: &ast.Var{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(1002),
																																									Column: int(19),
																																								},
																																								End: ast.Location{
																																									Line: int(1002),
																																									Column: int(24),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{
																																								"param",
																																							},
																																						},
																																						Id: "param",
																																					},
																																					LeftBracketFodder: ast.Fodder{},
																																					Index: &ast.LiteralString{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								End: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								File: nil,
																																							},
																																							Fodder: nil,
																																							Ctx: nil,
																																							FreeVars: nil,
																																						},
																																						Value: "space",
																																						Kind: ast.LiteralStringKind(1),
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																					},
																																					RightBracketFodder: ast.Fodder{},
																																					Id: nil,
																																				},
																																			},
																																			OpFodder: ast.Fodder{},
																																			Op: ast.BinaryOp(3),
																																			Right: &ast.Apply{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(1002),
																																							Column: int(33),
																																						},
																																						End: ast.Location{
																																							Line: int(1002),
																																							Column: int(72),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p10338,
																																					FreeVars: ast.Identifiers{
																																						"aux",
																																						"i",
																																						"param",
																																						"path",
																																						"v",
																																					},
																																				},
																																				Target: &ast.Var{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(1002),
																																								Column: int(33),
																																							},
																																							End: ast.Location{
																																								Line: int(1002),
																																								Column: int(36),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p10338,
																																						FreeVars: ast.Identifiers{
																																							"aux",
																																						},
																																					},
																																					Id: "aux",
																																				},
																																				FodderLeft: ast.Fodder{},
																																				Arguments: ast.Arguments{
																																					Positional: []ast.CommaSeparatedExpr{
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.Index{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(1002),
																																											Column: int(37),
																																										},
																																										End: ast.Location{
																																											Line: int(1002),
																																											Column: int(41),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p10355,
																																									FreeVars: ast.Identifiers{
																																										"i",
																																										"v",
																																									},
																																								},
																																								Target: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(1002),
																																												Column: int(37),
																																											},
																																											End: ast.Location{
																																												Line: int(1002),
																																												Column: int(38),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p10355,
																																										FreeVars: ast.Identifiers{
																																											"v",
																																										},
																																									},
																																									Id: "v",
																																								},
																																								LeftBracketFodder: ast.Fodder{},
																																								Index: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(1002),
																																												Column: int(39),
																																											},
																																											End: ast.Location{
																																												Line: int(1002),
																																												Column: int(40),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p10355,
																																										FreeVars: ast.Identifiers{
																																											"i",
																																										},
																																									},
																																									Id: "i",
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								Id: nil,
																																							},
																																							CommaFodder: ast.Fodder{},
																																						},
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.Binary{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(1002),
																																											Column: int(43),
																																										},
																																										End: ast.Location{
																																											Line: int(1002),
																																											Column: int(53),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p10355,
																																									FreeVars: ast.Identifiers{
																																										"i",
																																										"path",
																																									},
																																								},
																																								Left: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(1002),
																																												Column: int(43),
																																											},
																																											End: ast.Location{
																																												Line: int(1002),
																																												Column: int(47),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p10355,
																																										FreeVars: ast.Identifiers{
																																											"path",
																																										},
																																									},
																																									Id: "path",
																																								},
																																								OpFodder: ast.Fodder{},
																																								Op: ast.BinaryOp(3),
																																								Right: &ast.Array{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(1002),
																																												Column: int(50),
																																											},
																																											End: ast.Location{
																																												Line: int(1002),
																																												Column: int(53),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p10355,
																																										FreeVars: ast.Identifiers{
																																											"i",
																																										},
																																									},
																																									Elements: []ast.CommaSeparatedExpr{
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Var{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(1002),
																																															Column: int(51),
																																														},
																																														End: ast.Location{
																																															Line: int(1002),
																																															Column: int(52),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: p10369,
																																													FreeVars: ast.Identifiers{
																																														"i",
																																													},
																																												},
																																												Id: "i",
																																											},
																																											CommaFodder: nil,
																																										},
																																									},
																																									TrailingComma: false,
																																									CloseFodder: ast.Fodder{},
																																								},
																																							},
																																							CommaFodder: ast.Fodder{},
																																						},
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.Index{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(1002),
																																											Column: int(55),
																																										},
																																										End: ast.Location{
																																											Line: int(1002),
																																											Column: int(71),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p10355,
																																									FreeVars: ast.Identifiers{
																																										"param",
																																									},
																																								},
																																								Target: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(1002),
																																												Column: int(55),
																																											},
																																											End: ast.Location{
																																												Line: int(1002),
																																												Column: int(60),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"param",
																																										},
																																									},
																																									Id: "param",
																																								},
																																								LeftBracketFodder: ast.Fodder{},
																																								Index: &ast.LiteralString{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											File: nil,
																																										},
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: nil,
																																									},
																																									Value: "new_indent",
																																									Kind: ast.LiteralStringKind(1),
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								Id: nil,
																																							},
																																							CommaFodder: nil,
																																						},
																																					},
																																					Named: nil,
																																				},
																																				TrailingComma: false,
																																				TailStrict: false,
																																				FodderRight: ast.Fodder{},
																																				TailStrictFodder: nil,
																																			},
																																		},
																																		CommaFodder: nil,
																																	},
																																},
																																TrailingComma: false,
																																CloseFodder: nil,
																															},
																														},
																														CommaFodder: nil,
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Array{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(1004),
																																		Column: int(26),
																																	},
																																	End: ast.Location{
																																		Line: int(1004),
																																		Column: int(40),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p10377,
																																FreeVars: ast.Identifiers{
																																	"i",
																																	"params",
																																	"v",
																																},
																															},
																															Elements: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Apply{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(1004),
																																					Column: int(27),
																																				},
																																				End: ast.Location{
																																					Line: int(1004),
																																					Column: int(39),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p10381,
																																			FreeVars: ast.Identifiers{
																																				"i",
																																				"params",
																																				"v",
																																			},
																																		},
																																		Target: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(1004),
																																						Column: int(27),
																																					},
																																					End: ast.Location{
																																						Line: int(1004),
																																						Column: int(33),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p10381,
																																				FreeVars: ast.Identifiers{
																																					"params",
																																				},
																																			},
																																			Id: "params",
																																		},
																																		FodderLeft: ast.Fodder{},
																																		Arguments: ast.Arguments{
																																			Positional: []ast.CommaSeparatedExpr{
																																				ast.CommaSeparatedExpr{
																																					Expr: &ast.Index{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(1004),
																																									Column: int(34),
																																								},
																																								End: ast.Location{
																																									Line: int(1004),
																																									Column: int(38),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p10387,
																																							FreeVars: ast.Identifiers{
																																								"i",
																																								"v",
																																							},
																																						},
																																						Target: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(1004),
																																										Column: int(34),
																																									},
																																									End: ast.Location{
																																										Line: int(1004),
																																										Column: int(35),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p10387,
																																								FreeVars: ast.Identifiers{
																																									"v",
																																								},
																																							},
																																							Id: "v",
																																						},
																																						LeftBracketFodder: ast.Fodder{},
																																						Index: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(1004),
																																										Column: int(36),
																																									},
																																									End: ast.Location{
																																										Line: int(1004),
																																										Column: int(37),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p10387,
																																								FreeVars: ast.Identifiers{
																																									"i",
																																								},
																																							},
																																							Id: "i",
																																						},
																																						RightBracketFodder: ast.Fodder{},
																																						Id: nil,
																																					},
																																					CommaFodder: nil,
																																				},
																																			},
																																			Named: nil,
																																		},
																																		TrailingComma: false,
																																		TailStrict: false,
																																		FodderRight: ast.Fodder{},
																																		TailStrictFodder: nil,
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															TrailingComma: false,
																															CloseFodder: ast.Fodder{},
																														},
																														CommaFodder: nil,
																													},
																												},
																												Named: nil,
																											},
																											TrailingComma: false,
																											TailStrict: false,
																											FodderRight: nil,
																											TailStrictFodder: nil,
																										},
																									},
																									CommaFodder: nil,
																								},
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(1003),
																													Column: int(22),
																												},
																												End: ast.Location{
																													Line: int(1003),
																													Column: int(27),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p10377,
																											FreeVars: ast.Identifiers{
																												"range",
																											},
																										},
																										Id: "range",
																									},
																									CommaFodder: nil,
																								},
																							},
																							Named: nil,
																						},
																						TrailingComma: false,
																						TailStrict: false,
																						FodderRight: nil,
																						TailStrictFodder: nil,
																					},
																					Fun: nil,
																					CloseFodder: ast.Fodder{},
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1001),
																							Column: int(17),
																						},
																						End: ast.Location{
																							Line: int(1005),
																							Column: int(12),
																						},
																						File: p1,
																					},
																				},
																			},
																			Body: &ast.Apply{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1006),
																							Column: int(11),
																						},
																						End: ast.Location{
																							Line: int(1006),
																							Column: int(42),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p9934,
																					FreeVars: ast.Identifiers{
																						"cindent",
																						"parts",
																						"std",
																					},
																				},
																				Target: &ast.Index{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1006),
																								Column: int(11),
																							},
																							End: ast.Location{
																								Line: int(1006),
																								Column: int(19),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p9934,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																					},
																					Target: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1006),
																									Column: int(11),
																								},
																								End: ast.Location{
																									Line: int(1006),
																									Column: int(14),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{
																								ast.FodderElement{
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(10),
																									Comment: []string{},
																								},
																							},
																							Ctx: nil,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																						},
																						Id: "std",
																					},
																					LeftBracketFodder: ast.Fodder{},
																					Index: &ast.LiteralString{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								File: nil,
																							},
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: nil,
																						},
																						Value: "join",
																						Kind: ast.LiteralStringKind(1),
																						BlockIndent: "",
																						BlockTermIndent: "",
																					},
																					RightBracketFodder: ast.Fodder{},
																					Id: nil,
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Binary{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1006),
																											Column: int(20),
																										},
																										End: ast.Location{
																											Line: int(1006),
																											Column: int(34),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p10405,
																									FreeVars: ast.Identifiers{
																										"cindent",
																									},
																								},
																								Left: &ast.LiteralString{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1006),
																												Column: int(20),
																											},
																											End: ast.Location{
																												Line: int(1006),
																												Column: int(24),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p10405,
																										FreeVars: nil,
																									},
																									Value: "\n",
																									Kind: ast.LiteralStringKind(1),
																									BlockIndent: "",
																									BlockTermIndent: "",
																								},
																								OpFodder: ast.Fodder{},
																								Op: ast.BinaryOp(3),
																								Right: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1006),
																												Column: int(27),
																											},
																											End: ast.Location{
																												Line: int(1006),
																												Column: int(34),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p10405,
																										FreeVars: ast.Identifiers{
																											"cindent",
																										},
																									},
																									Id: "cindent",
																								},
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1006),
																											Column: int(36),
																										},
																										End: ast.Location{
																											Line: int(1006),
																											Column: int(41),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p10405,
																									FreeVars: ast.Identifiers{
																										"parts",
																									},
																								},
																								Id: "parts",
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				TrailingComma: false,
																				TailStrict: false,
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: nil,
																			},
																		},
																	},
																},
															},
															ElseFodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(6),
																	Comment: []string{},
																},
															},
															BranchFalse: &ast.Conditional{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1007),
																			Column: int(12),
																		},
																		End: ast.Location{
																			Line: int(1034),
																			Column: int(42),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p9934,
																	FreeVars: ast.Identifiers{
																		"aux",
																		"cindent",
																		"indent_array_in_object",
																		"path",
																		"std",
																		"v",
																	},
																},
																Cond: &ast.Apply{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1007),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(1007),
																				Column: int(30),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p9934,
																		FreeVars: ast.Identifiers{
																			"std",
																			"v",
																		},
																	},
																	Target: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1007),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(1007),
																					Column: int(27),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p9934,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1007),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(1007),
																						Column: int(18),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Id: "std",
																		},
																		LeftBracketFodder: ast.Fodder{},
																		Index: &ast.LiteralString{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: nil,
																			},
																			Value: "isObject",
																			Kind: ast.LiteralStringKind(1),
																			BlockIndent: "",
																			BlockTermIndent: "",
																		},
																		RightBracketFodder: ast.Fodder{},
																		Id: nil,
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1007),
																								Column: int(28),
																							},
																							End: ast.Location{
																								Line: int(1007),
																								Column: int(29),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p10424,
																						FreeVars: ast.Identifiers{
																							"v",
																						},
																					},
																					Id: "v",
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	TrailingComma: false,
																	TailStrict: false,
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																},
																ThenFodder: ast.Fodder{},
																BranchTrue: &ast.Conditional{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1008),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(1034),
																				Column: int(42),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(8),
																				Comment: []string{},
																			},
																		},
																		Ctx: p9934,
																		FreeVars: ast.Identifiers{
																			"aux",
																			"cindent",
																			"indent_array_in_object",
																			"path",
																			"std",
																			"v",
																		},
																	},
																	Cond: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1008),
																					Column: int(12),
																				},
																				End: ast.Location{
																					Line: int(1008),
																					Column: int(30),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p9934,
																			FreeVars: ast.Identifiers{
																				"std",
																				"v",
																			},
																		},
																		Left: &ast.Apply{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1008),
																						Column: int(12),
																					},
																					End: ast.Location{
																						Line: int(1008),
																						Column: int(25),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p9934,
																				FreeVars: ast.Identifiers{
																					"std",
																					"v",
																				},
																			},
																			Target: &ast.Index{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1008),
																							Column: int(12),
																						},
																						End: ast.Location{
																							Line: int(1008),
																							Column: int(22),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p9934,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																				},
																				Target: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1008),
																								Column: int(12),
																							},
																							End: ast.Location{
																								Line: int(1008),
																								Column: int(15),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: nil,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																					},
																					Id: "std",
																				},
																				LeftBracketFodder: ast.Fodder{},
																				Index: &ast.LiteralString{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							File: nil,
																						},
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: nil,
																					},
																					Value: "length",
																					Kind: ast.LiteralStringKind(1),
																					BlockIndent: "",
																					BlockTermIndent: "",
																				},
																				RightBracketFodder: ast.Fodder{},
																				Id: nil,
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1008),
																										Column: int(23),
																									},
																									End: ast.Location{
																										Line: int(1008),
																										Column: int(24),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p10440,
																								FreeVars: ast.Identifiers{
																									"v",
																								},
																							},
																							Id: "v",
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			TrailingComma: false,
																			TailStrict: false,
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(12),
																		Right: &ast.LiteralNumber{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1008),
																						Column: int(29),
																					},
																					End: ast.Location{
																						Line: int(1008),
																						Column: int(30),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p9934,
																				FreeVars: nil,
																			},
																			OriginalString: "0",
																		},
																	},
																	ThenFodder: ast.Fodder{},
																	BranchTrue: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1009),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(1009),
																					Column: int(15),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(10),
																					Comment: []string{},
																				},
																			},
																			Ctx: p9934,
																			FreeVars: nil,
																		},
																		Value: "{}",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	ElseFodder: ast.Fodder{
																		ast.FodderElement{
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(8),
																			Comment: []string{},
																		},
																	},
																	BranchFalse: &ast.Local{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1011),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(1034),
																					Column: int(42),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(10),
																					Comment: []string{},
																				},
																			},
																			Ctx: p9934,
																			FreeVars: ast.Identifiers{
																				"aux",
																				"cindent",
																				"indent_array_in_object",
																				"path",
																				"std",
																				"v",
																			},
																		},
																		Binds: ast.LocalBinds{
																			ast.LocalBind{
																				VarFodder: nil,
																				Variable: "params",
																				EqFodder: nil,
																				Body: &ast.Function{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1011),
																								Column: int(17),
																							},
																							End: ast.Location{
																								Line: int(1028),
																								Column: int(14),
																							},
																							File: p1,
																						},
																						Fodder: nil,
																						Ctx: p10451,
																						FreeVars: ast.Identifiers{
																							"cindent",
																							"indent_array_in_object",
																							"std",
																						},
																					},
																					ParenLeftFodder: ast.Fodder{},
																					Parameters: []ast.Parameter{
																						ast.Parameter{
																							NameFodder: ast.Fodder{},
																							Name: "value",
																							EqFodder: nil,
																							DefaultArg: nil,
																							CommaFodder: nil,
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1011),
																									Column: int(24),
																								},
																								End: ast.Location{
																									Line: int(1011),
																									Column: int(29),
																								},
																								File: p1,
																							},
																						},
																					},
																					TrailingComma: false,
																					ParenRightFodder: ast.Fodder{},
																					Body: &ast.Conditional{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1012),
																									Column: int(13),
																								},
																								End: ast.Location{
																									Line: int(1028),
																									Column: int(14),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{
																								ast.FodderElement{
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(12),
																									Comment: []string{},
																								},
																							},
																							Ctx: p10456,
																							FreeVars: ast.Identifiers{
																								"cindent",
																								"indent_array_in_object",
																								"std",
																								"value",
																							},
																						},
																						Cond: &ast.Binary{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1012),
																										Column: int(16),
																									},
																									End: ast.Location{
																										Line: int(1012),
																										Column: int(59),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p10456,
																								FreeVars: ast.Identifiers{
																									"std",
																									"value",
																								},
																							},
																							Left: &ast.Apply{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1012),
																											Column: int(16),
																										},
																										End: ast.Location{
																											Line: int(1012),
																											Column: int(34),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p10456,
																									FreeVars: ast.Identifiers{
																										"std",
																										"value",
																									},
																								},
																								Target: &ast.Index{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1012),
																												Column: int(16),
																											},
																											End: ast.Location{
																												Line: int(1012),
																												Column: int(27),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p10456,
																										FreeVars: ast.Identifiers{
																											"std",
																										},
																									},
																									Target: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(1012),
																													Column: int(16),
																												},
																												End: ast.Location{
																													Line: int(1012),
																													Column: int(19),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"std",
																											},
																										},
																										Id: "std",
																									},
																									LeftBracketFodder: ast.Fodder{},
																									Index: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												File: nil,
																											},
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: nil,
																										},
																										Value: "isArray",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																									RightBracketFodder: ast.Fodder{},
																									Id: nil,
																								},
																								FodderLeft: ast.Fodder{},
																								Arguments: ast.Arguments{
																									Positional: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(1012),
																															Column: int(28),
																														},
																														End: ast.Location{
																															Line: int(1012),
																															Column: int(33),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p10469,
																													FreeVars: ast.Identifiers{
																														"value",
																													},
																												},
																												Id: "value",
																											},
																											CommaFodder: nil,
																										},
																									},
																									Named: nil,
																								},
																								TrailingComma: false,
																								TailStrict: false,
																								FodderRight: ast.Fodder{},
																								TailStrictFodder: nil,
																							},
																							OpFodder: ast.Fodder{},
																							Op: ast.BinaryOp(17),
																							Right: &ast.Binary{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1012),
																											Column: int(38),
																										},
																										End: ast.Location{
																											Line: int(1012),
																											Column: int(59),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p10456,
																									FreeVars: ast.Identifiers{
																										"std",
																										"value",
																									},
																								},
																								Left: &ast.Apply{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1012),
																												Column: int(38),
																											},
																											End: ast.Location{
																												Line: int(1012),
																												Column: int(55),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p10456,
																										FreeVars: ast.Identifiers{
																											"std",
																											"value",
																										},
																									},
																									Target: &ast.Index{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(1012),
																													Column: int(38),
																												},
																												End: ast.Location{
																													Line: int(1012),
																													Column: int(48),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p10456,
																											FreeVars: ast.Identifiers{
																												"std",
																											},
																										},
																										Target: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1012),
																														Column: int(38),
																													},
																													End: ast.Location{
																														Line: int(1012),
																														Column: int(41),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																											},
																											Id: "std",
																										},
																										LeftBracketFodder: ast.Fodder{},
																										Index: &ast.LiteralString{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													File: nil,
																												},
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: nil,
																											},
																											Value: "length",
																											Kind: ast.LiteralStringKind(1),
																											BlockIndent: "",
																											BlockTermIndent: "",
																										},
																										RightBracketFodder: ast.Fodder{},
																										Id: nil,
																									},
																									FodderLeft: ast.Fodder{},
																									Arguments: ast.Arguments{
																										Positional: []ast.CommaSeparatedExpr{
																											ast.CommaSeparatedExpr{
																												Expr: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(1012),
																																Column: int(49),
																															},
																															End: ast.Location{
																																Line: int(1012),
																																Column: int(54),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p10482,
																														FreeVars: ast.Identifiers{
																															"value",
																														},
																													},
																													Id: "value",
																												},
																												CommaFodder: nil,
																											},
																										},
																										Named: nil,
																									},
																									TrailingComma: false,
																									TailStrict: false,
																									FodderRight: ast.Fodder{},
																									TailStrictFodder: nil,
																								},
																								OpFodder: ast.Fodder{},
																								Op: ast.BinaryOp(7),
																								Right: &ast.LiteralNumber{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1012),
																												Column: int(58),
																											},
																											End: ast.Location{
																												Line: int(1012),
																												Column: int(59),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p10456,
																										FreeVars: nil,
																									},
																									OriginalString: "0",
																								},
																							},
																						},
																						ThenFodder: ast.Fodder{},
																						BranchTrue: &ast.DesugaredObject{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1012),
																										Column: int(65),
																									},
																									End: ast.Location{
																										Line: int(1021),
																										Column: int(14),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p10456,
																								FreeVars: ast.Identifiers{
																									"cindent",
																									"indent_array_in_object",
																								},
																							},
																							Asserts: ast.Nodes{},
																							Fields: ast.DesugaredObjectFields{
																								ast.DesugaredObjectField{
																									Hide: ast.ObjectFieldHide(1),
																									Name: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												File: nil,
																											},
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: nil,
																										},
																										Value: "new_indent",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																									Body: &ast.Conditional{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(1019),
																													Column: int(27),
																												},
																												End: ast.Location{
																													Line: int(1019),
																													Column: int(85),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p10490,
																											FreeVars: ast.Identifiers{
																												"cindent",
																												"indent_array_in_object",
																											},
																										},
																										Cond: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1019),
																														Column: int(30),
																													},
																													End: ast.Location{
																														Line: int(1019),
																														Column: int(52),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p10490,
																												FreeVars: ast.Identifiers{
																													"indent_array_in_object",
																												},
																											},
																											Id: "indent_array_in_object",
																										},
																										ThenFodder: ast.Fodder{},
																										BranchTrue: &ast.Binary{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1019),
																														Column: int(58),
																													},
																													End: ast.Location{
																														Line: int(1019),
																														Column: int(72),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p10490,
																												FreeVars: ast.Identifiers{
																													"cindent",
																												},
																											},
																											Left: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(1019),
																															Column: int(58),
																														},
																														End: ast.Location{
																															Line: int(1019),
																															Column: int(65),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p10490,
																													FreeVars: ast.Identifiers{
																														"cindent",
																													},
																												},
																												Id: "cindent",
																											},
																											OpFodder: ast.Fodder{},
																											Op: ast.BinaryOp(3),
																											Right: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(1019),
																															Column: int(68),
																														},
																														End: ast.Location{
																															Line: int(1019),
																															Column: int(72),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p10490,
																													FreeVars: nil,
																												},
																												Value: "  ",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																										},
																										ElseFodder: ast.Fodder{},
																										BranchFalse: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1019),
																														Column: int(78),
																													},
																													End: ast.Location{
																														Line: int(1019),
																														Column: int(85),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p10490,
																												FreeVars: ast.Identifiers{
																													"cindent",
																												},
																											},
																											Id: "cindent",
																										},
																									},
																									PlusSuper: false,
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1019),
																											Column: int(15),
																										},
																										End: ast.Location{
																											Line: int(1019),
																											Column: int(85),
																										},
																										File: p1,
																									},
																								},
																								ast.DesugaredObjectField{
																									Hide: ast.ObjectFieldHide(1),
																									Name: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												File: nil,
																											},
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: nil,
																										},
																										Value: "space",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																									Body: &ast.Binary{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(1020),
																													Column: int(22),
																												},
																												End: ast.Location{
																													Line: int(1020),
																													Column: int(44),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p10490,
																											FreeVars: nil,
																										},
																										Left: &ast.LiteralString{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1020),
																														Column: int(22),
																													},
																													End: ast.Location{
																														Line: int(1020),
																														Column: int(26),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p10490,
																												FreeVars: nil,
																											},
																											Value: "\n",
																											Kind: ast.LiteralStringKind(1),
																											BlockIndent: "",
																											BlockTermIndent: "",
																										},
																										OpFodder: ast.Fodder{},
																										Op: ast.BinaryOp(3),
																										Right: &ast.Index{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1020),
																														Column: int(29),
																													},
																													End: ast.Location{
																														Line: int(1020),
																														Column: int(44),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p10490,
																												FreeVars: nil,
																											},
																											Target: &ast.Self{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(1020),
																															Column: int(29),
																														},
																														End: ast.Location{
																															Line: int(1020),
																															Column: int(33),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: nil,
																													FreeVars: nil,
																												},
																											},
																											LeftBracketFodder: ast.Fodder{},
																											Index: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: nil,
																												},
																												Value: "new_indent",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																											RightBracketFodder: ast.Fodder{},
																											Id: nil,
																										},
																									},
																									PlusSuper: false,
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1020),
																											Column: int(15),
																										},
																										End: ast.Location{
																											Line: int(1020),
																											Column: int(44),
																										},
																										File: p1,
																									},
																								},
																							},
																							Locals: ast.LocalBinds{},
																						},
																						ElseFodder: ast.Fodder{},
																						BranchFalse: &ast.Conditional{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1021),
																										Column: int(20),
																									},
																									End: ast.Location{
																										Line: int(1028),
																										Column: int(14),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p10456,
																								FreeVars: ast.Identifiers{
																									"cindent",
																									"std",
																									"value",
																								},
																							},
																							Cond: &ast.Binary{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1021),
																											Column: int(23),
																										},
																										End: ast.Location{
																											Line: int(1021),
																											Column: int(67),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p10456,
																									FreeVars: ast.Identifiers{
																										"std",
																										"value",
																									},
																								},
																								Left: &ast.Apply{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1021),
																												Column: int(23),
																											},
																											End: ast.Location{
																												Line: int(1021),
																												Column: int(42),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p10456,
																										FreeVars: ast.Identifiers{
																											"std",
																											"value",
																										},
																									},
																									Target: &ast.Index{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(1021),
																													Column: int(23),
																												},
																												End: ast.Location{
																													Line: int(1021),
																													Column: int(35),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p10456,
																											FreeVars: ast.Identifiers{
																												"std",
																											},
																										},
																										Target: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1021),
																														Column: int(23),
																													},
																													End: ast.Location{
																														Line: int(1021),
																														Column: int(26),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																											},
																											Id: "std",
																										},
																										LeftBracketFodder: ast.Fodder{},
																										Index: &ast.LiteralString{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													File: nil,
																												},
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: nil,
																											},
																											Value: "isObject",
																											Kind: ast.LiteralStringKind(1),
																											BlockIndent: "",
																											BlockTermIndent: "",
																										},
																										RightBracketFodder: ast.Fodder{},
																										Id: nil,
																									},
																									FodderLeft: ast.Fodder{},
																									Arguments: ast.Arguments{
																										Positional: []ast.CommaSeparatedExpr{
																											ast.CommaSeparatedExpr{
																												Expr: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(1021),
																																Column: int(36),
																															},
																															End: ast.Location{
																																Line: int(1021),
																																Column: int(41),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p10520,
																														FreeVars: ast.Identifiers{
																															"value",
																														},
																													},
																													Id: "value",
																												},
																												CommaFodder: nil,
																											},
																										},
																										Named: nil,
																									},
																									TrailingComma: false,
																									TailStrict: false,
																									FodderRight: ast.Fodder{},
																									TailStrictFodder: nil,
																								},
																								OpFodder: ast.Fodder{},
																								Op: ast.BinaryOp(17),
																								Right: &ast.Binary{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1021),
																												Column: int(46),
																											},
																											End: ast.Location{
																												Line: int(1021),
																												Column: int(67),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p10456,
																										FreeVars: ast.Identifiers{
																											"std",
																											"value",
																										},
																									},
																									Left: &ast.Apply{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(1021),
																													Column: int(46),
																												},
																												End: ast.Location{
																													Line: int(1021),
																													Column: int(63),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p10456,
																											FreeVars: ast.Identifiers{
																												"std",
																												"value",
																											},
																										},
																										Target: &ast.Index{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1021),
																														Column: int(46),
																													},
																													End: ast.Location{
																														Line: int(1021),
																														Column: int(56),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p10456,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																											},
																											Target: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(1021),
																															Column: int(46),
																														},
																														End: ast.Location{
																															Line: int(1021),
																															Column: int(49),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																												},
																												Id: "std",
																											},
																											LeftBracketFodder: ast.Fodder{},
																											Index: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: nil,
																												},
																												Value: "length",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																											RightBracketFodder: ast.Fodder{},
																											Id: nil,
																										},
																										FodderLeft: ast.Fodder{},
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(1021),
																																	Column: int(57),
																																},
																																End: ast.Location{
																																	Line: int(1021),
																																	Column: int(62),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p10533,
																															FreeVars: ast.Identifiers{
																																"value",
																															},
																														},
																														Id: "value",
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										TrailingComma: false,
																										TailStrict: false,
																										FodderRight: ast.Fodder{},
																										TailStrictFodder: nil,
																									},
																									OpFodder: ast.Fodder{},
																									Op: ast.BinaryOp(7),
																									Right: &ast.LiteralNumber{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(1021),
																													Column: int(66),
																												},
																												End: ast.Location{
																													Line: int(1021),
																													Column: int(67),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p10456,
																											FreeVars: nil,
																										},
																										OriginalString: "0",
																									},
																								},
																							},
																							ThenFodder: ast.Fodder{},
																							BranchTrue: &ast.DesugaredObject{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1021),
																											Column: int(73),
																										},
																										End: ast.Location{
																											Line: int(1024),
																											Column: int(14),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p10456,
																									FreeVars: ast.Identifiers{
																										"cindent",
																									},
																								},
																								Asserts: ast.Nodes{},
																								Fields: ast.DesugaredObjectFields{
																									ast.DesugaredObjectField{
																										Hide: ast.ObjectFieldHide(1),
																										Name: &ast.LiteralString{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													File: nil,
																												},
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: nil,
																											},
																											Value: "new_indent",
																											Kind: ast.LiteralStringKind(1),
																											BlockIndent: "",
																											BlockTermIndent: "",
																										},
																										Body: &ast.Binary{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1022),
																														Column: int(27),
																													},
																													End: ast.Location{
																														Line: int(1022),
																														Column: int(41),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p10541,
																												FreeVars: ast.Identifiers{
																													"cindent",
																												},
																											},
																											Left: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(1022),
																															Column: int(27),
																														},
																														End: ast.Location{
																															Line: int(1022),
																															Column: int(34),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p10541,
																													FreeVars: ast.Identifiers{
																														"cindent",
																													},
																												},
																												Id: "cindent",
																											},
																											OpFodder: ast.Fodder{},
																											Op: ast.BinaryOp(3),
																											Right: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(1022),
																															Column: int(37),
																														},
																														End: ast.Location{
																															Line: int(1022),
																															Column: int(41),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p10541,
																													FreeVars: nil,
																												},
																												Value: "  ",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																										},
																										PlusSuper: false,
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1022),
																												Column: int(15),
																											},
																											End: ast.Location{
																												Line: int(1022),
																												Column: int(41),
																											},
																											File: p1,
																										},
																									},
																									ast.DesugaredObjectField{
																										Hide: ast.ObjectFieldHide(1),
																										Name: &ast.LiteralString{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													File: nil,
																												},
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: nil,
																											},
																											Value: "space",
																											Kind: ast.LiteralStringKind(1),
																											BlockIndent: "",
																											BlockTermIndent: "",
																										},
																										Body: &ast.Binary{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1023),
																														Column: int(22),
																													},
																													End: ast.Location{
																														Line: int(1023),
																														Column: int(44),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p10541,
																												FreeVars: nil,
																											},
																											Left: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(1023),
																															Column: int(22),
																														},
																														End: ast.Location{
																															Line: int(1023),
																															Column: int(26),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p10541,
																													FreeVars: nil,
																												},
																												Value: "\n",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																											OpFodder: ast.Fodder{},
																											Op: ast.BinaryOp(3),
																											Right: &ast.Index{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(1023),
																															Column: int(29),
																														},
																														End: ast.Location{
																															Line: int(1023),
																															Column: int(44),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p10541,
																													FreeVars: nil,
																												},
																												Target: &ast.Self{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(1023),
																																Column: int(29),
																															},
																															End: ast.Location{
																																Line: int(1023),
																																Column: int(33),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: nil,
																														FreeVars: nil,
																													},
																												},
																												LeftBracketFodder: ast.Fodder{},
																												Index: &ast.LiteralString{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															File: nil,
																														},
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: nil,
																													},
																													Value: "new_indent",
																													Kind: ast.LiteralStringKind(1),
																													BlockIndent: "",
																													BlockTermIndent: "",
																												},
																												RightBracketFodder: ast.Fodder{},
																												Id: nil,
																											},
																										},
																										PlusSuper: false,
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1023),
																												Column: int(15),
																											},
																											End: ast.Location{
																												Line: int(1023),
																												Column: int(44),
																											},
																											File: p1,
																										},
																									},
																								},
																								Locals: ast.LocalBinds{},
																							},
																							ElseFodder: ast.Fodder{},
																							BranchFalse: &ast.DesugaredObject{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1024),
																											Column: int(20),
																										},
																										End: ast.Location{
																											Line: int(1028),
																											Column: int(14),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p10456,
																									FreeVars: ast.Identifiers{
																										"cindent",
																									},
																								},
																								Asserts: ast.Nodes{},
																								Fields: ast.DesugaredObjectFields{
																									ast.DesugaredObjectField{
																										Hide: ast.ObjectFieldHide(1),
																										Name: &ast.LiteralString{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													File: nil,
																												},
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: nil,
																											},
																											Value: "new_indent",
																											Kind: ast.LiteralStringKind(1),
																											BlockIndent: "",
																											BlockTermIndent: "",
																										},
																										Body: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1026),
																														Column: int(27),
																													},
																													End: ast.Location{
																														Line: int(1026),
																														Column: int(34),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p10557,
																												FreeVars: ast.Identifiers{
																													"cindent",
																												},
																											},
																											Id: "cindent",
																										},
																										PlusSuper: false,
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1026),
																												Column: int(15),
																											},
																											End: ast.Location{
																												Line: int(1026),
																												Column: int(34),
																											},
																											File: p1,
																										},
																									},
																									ast.DesugaredObjectField{
																										Hide: ast.ObjectFieldHide(1),
																										Name: &ast.LiteralString{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													File: nil,
																												},
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: nil,
																											},
																											Value: "space",
																											Kind: ast.LiteralStringKind(1),
																											BlockIndent: "",
																											BlockTermIndent: "",
																										},
																										Body: &ast.LiteralString{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1027),
																														Column: int(22),
																													},
																													End: ast.Location{
																														Line: int(1027),
																														Column: int(25),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p10557,
																												FreeVars: nil,
																											},
																											Value: " ",
																											Kind: ast.LiteralStringKind(1),
																											BlockIndent: "",
																											BlockTermIndent: "",
																										},
																										PlusSuper: false,
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1027),
																												Column: int(15),
																											},
																											End: ast.Location{
																												Line: int(1027),
																												Column: int(25),
																											},
																											File: p1,
																										},
																									},
																								},
																								Locals: ast.LocalBinds{},
																							},
																						},
																					},
																				},
																				Fun: nil,
																				CloseFodder: nil,
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																			},
																		},
																		Body: &ast.Local{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1029),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(1034),
																						Column: int(42),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(10),
																						Comment: []string{},
																					},
																				},
																				Ctx: p9934,
																				FreeVars: ast.Identifiers{
																					"aux",
																					"cindent",
																					"params",
																					"path",
																					"std",
																					"v",
																				},
																			},
																			Binds: ast.LocalBinds{
																				ast.LocalBind{
																					VarFodder: ast.Fodder{},
																					Variable: "lines",
																					EqFodder: ast.Fodder{},
																					Body: &ast.Apply{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								File: nil,
																							},
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{
																								"aux",
																								"params",
																								"path",
																								"std",
																								"v",
																							},
																						},
																						Target: &ast.Index{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									File: nil,
																								},
																								Fodder: nil,
																								Ctx: nil,
																								FreeVars: ast.Identifiers{
																									"std",
																								},
																							},
																							Target: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										File: nil,
																									},
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																								},
																								Id: "std",
																							},
																							LeftBracketFodder: nil,
																							Index: &ast.LiteralString{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										File: nil,
																									},
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: nil,
																								},
																								Value: "flatMap",
																								Kind: ast.LiteralStringKind(1),
																								BlockIndent: "",
																								BlockTermIndent: "",
																							},
																							RightBracketFodder: nil,
																							Id: nil,
																						},
																						FodderLeft: nil,
																						Arguments: ast.Arguments{
																							Positional: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Function{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												File: nil,
																											},
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"aux",
																												"params",
																												"path",
																												"std",
																												"v",
																											},
																										},
																										ParenLeftFodder: nil,
																										Parameters: []ast.Parameter{
																											ast.Parameter{
																												NameFodder: nil,
																												Name: "k",
																												EqFodder: nil,
																												DefaultArg: nil,
																												CommaFodder: nil,
																												LocRange: ast.LocationRange{
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													File: nil,
																												},
																											},
																										},
																										TrailingComma: false,
																										ParenRightFodder: nil,
																										Body: &ast.Apply{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													File: nil,
																												},
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"aux",
																													"k",
																													"params",
																													"path",
																													"std",
																													"v",
																												},
																											},
																											Target: &ast.Index{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																												},
																												Target: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															File: nil,
																														},
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"std",
																														},
																													},
																													Id: "std",
																												},
																												LeftBracketFodder: nil,
																												Index: &ast.LiteralString{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															File: nil,
																														},
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: nil,
																													},
																													Value: "flatMap",
																													Kind: ast.LiteralStringKind(1),
																													BlockIndent: "",
																													BlockTermIndent: "",
																												},
																												RightBracketFodder: nil,
																												Id: nil,
																											},
																											FodderLeft: nil,
																											Arguments: ast.Arguments{
																												Positional: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Function{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	File: nil,
																																},
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"aux",
																																	"k",
																																	"path",
																																	"std",
																																	"v",
																																},
																															},
																															ParenLeftFodder: nil,
																															Parameters: []ast.Parameter{
																																ast.Parameter{
																																	NameFodder: nil,
																																	Name: "param",
																																	EqFodder: nil,
																																	DefaultArg: nil,
																																	CommaFodder: nil,
																																	LocRange: ast.LocationRange{
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		File: nil,
																																	},
																																},
																															},
																															TrailingComma: false,
																															ParenRightFodder: nil,
																															Body: &ast.Array{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		File: nil,
																																	},
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"aux",
																																		"k",
																																		"param",
																																		"path",
																																		"std",
																																		"v",
																																	},
																																},
																																Elements: []ast.CommaSeparatedExpr{
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Binary{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(1030),
																																						Column: int(13),
																																					},
																																					End: ast.Location{
																																						Line: int(1030),
																																						Column: int(98),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p10591,
																																				FreeVars: ast.Identifiers{
																																					"aux",
																																					"k",
																																					"param",
																																					"path",
																																					"std",
																																					"v",
																																				},
																																			},
																																			Left: &ast.Binary{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(1030),
																																							Column: int(13),
																																						},
																																						End: ast.Location{
																																							Line: int(1030),
																																							Column: int(56),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p10591,
																																					FreeVars: ast.Identifiers{
																																						"k",
																																						"param",
																																						"std",
																																					},
																																				},
																																				Left: &ast.Binary{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(1030),
																																								Column: int(13),
																																							},
																																							End: ast.Location{
																																								Line: int(1030),
																																								Column: int(42),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p10591,
																																						FreeVars: ast.Identifiers{
																																							"k",
																																							"std",
																																						},
																																					},
																																					Left: &ast.Apply{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(1030),
																																									Column: int(13),
																																								},
																																								End: ast.Location{
																																									Line: int(1030),
																																									Column: int(36),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p10591,
																																							FreeVars: ast.Identifiers{
																																								"k",
																																								"std",
																																							},
																																						},
																																						Target: &ast.Index{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(1030),
																																										Column: int(13),
																																									},
																																									End: ast.Location{
																																										Line: int(1030),
																																										Column: int(33),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p10591,
																																								FreeVars: ast.Identifiers{
																																									"std",
																																								},
																																							},
																																							Target: &ast.Var{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(1030),
																																											Column: int(13),
																																										},
																																										End: ast.Location{
																																											Line: int(1030),
																																											Column: int(16),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{
																																										ast.FodderElement{
																																											Kind: ast.FodderKind(0),
																																											Blanks: int(0),
																																											Indent: int(12),
																																											Comment: []string{},
																																										},
																																									},
																																									Ctx: nil,
																																									FreeVars: ast.Identifiers{
																																										"std",
																																									},
																																								},
																																								Id: "std",
																																							},
																																							LeftBracketFodder: ast.Fodder{},
																																							Index: &ast.LiteralString{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "",
																																										Begin: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										End: ast.Location{
																																											Line: int(0),
																																											Column: int(0),
																																										},
																																										File: nil,
																																									},
																																									Fodder: nil,
																																									Ctx: nil,
																																									FreeVars: nil,
																																								},
																																								Value: "escapeStringJson",
																																								Kind: ast.LiteralStringKind(1),
																																								BlockIndent: "",
																																								BlockTermIndent: "",
																																							},
																																							RightBracketFodder: ast.Fodder{},
																																							Id: nil,
																																						},
																																						FodderLeft: ast.Fodder{},
																																						Arguments: ast.Arguments{
																																							Positional: []ast.CommaSeparatedExpr{
																																								ast.CommaSeparatedExpr{
																																									Expr: &ast.Var{
																																										NodeBase: ast.NodeBase{
																																											LocRange: ast.LocationRange{
																																												FileName: "<std>",
																																												Begin: ast.Location{
																																													Line: int(1030),
																																													Column: int(34),
																																												},
																																												End: ast.Location{
																																													Line: int(1030),
																																													Column: int(35),
																																												},
																																												File: p1,
																																											},
																																											Fodder: ast.Fodder{},
																																											Ctx: p10607,
																																											FreeVars: ast.Identifiers{
																																												"k",
																																											},
																																										},
																																										Id: "k",
																																									},
																																									CommaFodder: nil,
																																								},
																																							},
																																							Named: nil,
																																						},
																																						TrailingComma: false,
																																						TailStrict: false,
																																						FodderRight: ast.Fodder{},
																																						TailStrictFodder: nil,
																																					},
																																					OpFodder: ast.Fodder{},
																																					Op: ast.BinaryOp(3),
																																					Right: &ast.LiteralString{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(1030),
																																									Column: int(39),
																																								},
																																								End: ast.Location{
																																									Line: int(1030),
																																									Column: int(42),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p10591,
																																							FreeVars: nil,
																																						},
																																						Value: ":",
																																						Kind: ast.LiteralStringKind(1),
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																					},
																																				},
																																				OpFodder: ast.Fodder{},
																																				Op: ast.BinaryOp(3),
																																				Right: &ast.Index{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(1030),
																																								Column: int(45),
																																							},
																																							End: ast.Location{
																																								Line: int(1030),
																																								Column: int(56),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p10591,
																																						FreeVars: ast.Identifiers{
																																							"param",
																																						},
																																					},
																																					Target: &ast.Var{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(1030),
																																									Column: int(45),
																																								},
																																								End: ast.Location{
																																									Line: int(1030),
																																									Column: int(50),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: nil,
																																							FreeVars: ast.Identifiers{
																																								"param",
																																							},
																																						},
																																						Id: "param",
																																					},
																																					LeftBracketFodder: ast.Fodder{},
																																					Index: &ast.LiteralString{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "",
																																								Begin: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								End: ast.Location{
																																									Line: int(0),
																																									Column: int(0),
																																								},
																																								File: nil,
																																							},
																																							Fodder: nil,
																																							Ctx: nil,
																																							FreeVars: nil,
																																						},
																																						Value: "space",
																																						Kind: ast.LiteralStringKind(1),
																																						BlockIndent: "",
																																						BlockTermIndent: "",
																																					},
																																					RightBracketFodder: ast.Fodder{},
																																					Id: nil,
																																				},
																																			},
																																			OpFodder: ast.Fodder{},
																																			Op: ast.BinaryOp(3),
																																			Right: &ast.Apply{
																																				NodeBase: ast.NodeBase{
																																					LocRange: ast.LocationRange{
																																						FileName: "<std>",
																																						Begin: ast.Location{
																																							Line: int(1030),
																																							Column: int(59),
																																						},
																																						End: ast.Location{
																																							Line: int(1030),
																																							Column: int(98),
																																						},
																																						File: p1,
																																					},
																																					Fodder: ast.Fodder{},
																																					Ctx: p10591,
																																					FreeVars: ast.Identifiers{
																																						"aux",
																																						"k",
																																						"param",
																																						"path",
																																						"v",
																																					},
																																				},
																																				Target: &ast.Var{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(1030),
																																								Column: int(59),
																																							},
																																							End: ast.Location{
																																								Line: int(1030),
																																								Column: int(62),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p10591,
																																						FreeVars: ast.Identifiers{
																																							"aux",
																																						},
																																					},
																																					Id: "aux",
																																				},
																																				FodderLeft: ast.Fodder{},
																																				Arguments: ast.Arguments{
																																					Positional: []ast.CommaSeparatedExpr{
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.Index{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(1030),
																																											Column: int(63),
																																										},
																																										End: ast.Location{
																																											Line: int(1030),
																																											Column: int(67),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p10621,
																																									FreeVars: ast.Identifiers{
																																										"k",
																																										"v",
																																									},
																																								},
																																								Target: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(1030),
																																												Column: int(63),
																																											},
																																											End: ast.Location{
																																												Line: int(1030),
																																												Column: int(64),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p10621,
																																										FreeVars: ast.Identifiers{
																																											"v",
																																										},
																																									},
																																									Id: "v",
																																								},
																																								LeftBracketFodder: ast.Fodder{},
																																								Index: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(1030),
																																												Column: int(65),
																																											},
																																											End: ast.Location{
																																												Line: int(1030),
																																												Column: int(66),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p10621,
																																										FreeVars: ast.Identifiers{
																																											"k",
																																										},
																																									},
																																									Id: "k",
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								Id: nil,
																																							},
																																							CommaFodder: ast.Fodder{},
																																						},
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.Binary{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(1030),
																																											Column: int(69),
																																										},
																																										End: ast.Location{
																																											Line: int(1030),
																																											Column: int(79),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p10621,
																																									FreeVars: ast.Identifiers{
																																										"k",
																																										"path",
																																									},
																																								},
																																								Left: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(1030),
																																												Column: int(69),
																																											},
																																											End: ast.Location{
																																												Line: int(1030),
																																												Column: int(73),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p10621,
																																										FreeVars: ast.Identifiers{
																																											"path",
																																										},
																																									},
																																									Id: "path",
																																								},
																																								OpFodder: ast.Fodder{},
																																								Op: ast.BinaryOp(3),
																																								Right: &ast.Array{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(1030),
																																												Column: int(76),
																																											},
																																											End: ast.Location{
																																												Line: int(1030),
																																												Column: int(79),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: p10621,
																																										FreeVars: ast.Identifiers{
																																											"k",
																																										},
																																									},
																																									Elements: []ast.CommaSeparatedExpr{
																																										ast.CommaSeparatedExpr{
																																											Expr: &ast.Var{
																																												NodeBase: ast.NodeBase{
																																													LocRange: ast.LocationRange{
																																														FileName: "<std>",
																																														Begin: ast.Location{
																																															Line: int(1030),
																																															Column: int(77),
																																														},
																																														End: ast.Location{
																																															Line: int(1030),
																																															Column: int(78),
																																														},
																																														File: p1,
																																													},
																																													Fodder: ast.Fodder{},
																																													Ctx: p10635,
																																													FreeVars: ast.Identifiers{
																																														"k",
																																													},
																																												},
																																												Id: "k",
																																											},
																																											CommaFodder: nil,
																																										},
																																									},
																																									TrailingComma: false,
																																									CloseFodder: ast.Fodder{},
																																								},
																																							},
																																							CommaFodder: ast.Fodder{},
																																						},
																																						ast.CommaSeparatedExpr{
																																							Expr: &ast.Index{
																																								NodeBase: ast.NodeBase{
																																									LocRange: ast.LocationRange{
																																										FileName: "<std>",
																																										Begin: ast.Location{
																																											Line: int(1030),
																																											Column: int(81),
																																										},
																																										End: ast.Location{
																																											Line: int(1030),
																																											Column: int(97),
																																										},
																																										File: p1,
																																									},
																																									Fodder: ast.Fodder{},
																																									Ctx: p10621,
																																									FreeVars: ast.Identifiers{
																																										"param",
																																									},
																																								},
																																								Target: &ast.Var{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "<std>",
																																											Begin: ast.Location{
																																												Line: int(1030),
																																												Column: int(81),
																																											},
																																											End: ast.Location{
																																												Line: int(1030),
																																												Column: int(86),
																																											},
																																											File: p1,
																																										},
																																										Fodder: ast.Fodder{},
																																										Ctx: nil,
																																										FreeVars: ast.Identifiers{
																																											"param",
																																										},
																																									},
																																									Id: "param",
																																								},
																																								LeftBracketFodder: ast.Fodder{},
																																								Index: &ast.LiteralString{
																																									NodeBase: ast.NodeBase{
																																										LocRange: ast.LocationRange{
																																											FileName: "",
																																											Begin: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											End: ast.Location{
																																												Line: int(0),
																																												Column: int(0),
																																											},
																																											File: nil,
																																										},
																																										Fodder: nil,
																																										Ctx: nil,
																																										FreeVars: nil,
																																									},
																																									Value: "new_indent",
																																									Kind: ast.LiteralStringKind(1),
																																									BlockIndent: "",
																																									BlockTermIndent: "",
																																								},
																																								RightBracketFodder: ast.Fodder{},
																																								Id: nil,
																																							},
																																							CommaFodder: nil,
																																						},
																																					},
																																					Named: nil,
																																				},
																																				TrailingComma: false,
																																				TailStrict: false,
																																				FodderRight: ast.Fodder{},
																																				TailStrictFodder: nil,
																																			},
																																		},
																																		CommaFodder: nil,
																																	},
																																},
																																TrailingComma: false,
																																CloseFodder: nil,
																															},
																														},
																														CommaFodder: nil,
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Array{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(1032),
																																		Column: int(26),
																																	},
																																	End: ast.Location{
																																		Line: int(1032),
																																		Column: int(40),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p10643,
																																FreeVars: ast.Identifiers{
																																	"k",
																																	"params",
																																	"v",
																																},
																															},
																															Elements: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Apply{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(1032),
																																					Column: int(27),
																																				},
																																				End: ast.Location{
																																					Line: int(1032),
																																					Column: int(39),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p10647,
																																			FreeVars: ast.Identifiers{
																																				"k",
																																				"params",
																																				"v",
																																			},
																																		},
																																		Target: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(1032),
																																						Column: int(27),
																																					},
																																					End: ast.Location{
																																						Line: int(1032),
																																						Column: int(33),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p10647,
																																				FreeVars: ast.Identifiers{
																																					"params",
																																				},
																																			},
																																			Id: "params",
																																		},
																																		FodderLeft: ast.Fodder{},
																																		Arguments: ast.Arguments{
																																			Positional: []ast.CommaSeparatedExpr{
																																				ast.CommaSeparatedExpr{
																																					Expr: &ast.Index{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(1032),
																																									Column: int(34),
																																								},
																																								End: ast.Location{
																																									Line: int(1032),
																																									Column: int(38),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p10653,
																																							FreeVars: ast.Identifiers{
																																								"k",
																																								"v",
																																							},
																																						},
																																						Target: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(1032),
																																										Column: int(34),
																																									},
																																									End: ast.Location{
																																										Line: int(1032),
																																										Column: int(35),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p10653,
																																								FreeVars: ast.Identifiers{
																																									"v",
																																								},
																																							},
																																							Id: "v",
																																						},
																																						LeftBracketFodder: ast.Fodder{},
																																						Index: &ast.Var{
																																							NodeBase: ast.NodeBase{
																																								LocRange: ast.LocationRange{
																																									FileName: "<std>",
																																									Begin: ast.Location{
																																										Line: int(1032),
																																										Column: int(36),
																																									},
																																									End: ast.Location{
																																										Line: int(1032),
																																										Column: int(37),
																																									},
																																									File: p1,
																																								},
																																								Fodder: ast.Fodder{},
																																								Ctx: p10653,
																																								FreeVars: ast.Identifiers{
																																									"k",
																																								},
																																							},
																																							Id: "k",
																																						},
																																						RightBracketFodder: ast.Fodder{},
																																						Id: nil,
																																					},
																																					CommaFodder: nil,
																																				},
																																			},
																																			Named: nil,
																																		},
																																		TrailingComma: false,
																																		TailStrict: false,
																																		FodderRight: ast.Fodder{},
																																		TailStrictFodder: nil,
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															TrailingComma: false,
																															CloseFodder: ast.Fodder{},
																														},
																														CommaFodder: nil,
																													},
																												},
																												Named: nil,
																											},
																											TrailingComma: false,
																											TailStrict: false,
																											FodderRight: nil,
																											TailStrictFodder: nil,
																										},
																									},
																									CommaFodder: nil,
																								},
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Apply{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(1031),
																													Column: int(22),
																												},
																												End: ast.Location{
																													Line: int(1031),
																													Column: int(41),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p10643,
																											FreeVars: ast.Identifiers{
																												"std",
																												"v",
																											},
																										},
																										Target: &ast.Index{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1031),
																														Column: int(22),
																													},
																													End: ast.Location{
																														Line: int(1031),
																														Column: int(38),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p10643,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																											},
																											Target: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(1031),
																															Column: int(22),
																														},
																														End: ast.Location{
																															Line: int(1031),
																															Column: int(25),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																												},
																												Id: "std",
																											},
																											LeftBracketFodder: ast.Fodder{},
																											Index: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: nil,
																												},
																												Value: "objectFields",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																											RightBracketFodder: ast.Fodder{},
																											Id: nil,
																										},
																										FodderLeft: ast.Fodder{},
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(1031),
																																	Column: int(39),
																																},
																																End: ast.Location{
																																	Line: int(1031),
																																	Column: int(40),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p10668,
																															FreeVars: ast.Identifiers{
																																"v",
																															},
																														},
																														Id: "v",
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										TrailingComma: false,
																										TailStrict: false,
																										FodderRight: ast.Fodder{},
																										TailStrictFodder: nil,
																									},
																									CommaFodder: nil,
																								},
																							},
																							Named: nil,
																						},
																						TrailingComma: false,
																						TailStrict: false,
																						FodderRight: nil,
																						TailStrictFodder: nil,
																					},
																					Fun: nil,
																					CloseFodder: ast.Fodder{},
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1029),
																							Column: int(17),
																						},
																						End: ast.Location{
																							Line: int(1033),
																							Column: int(12),
																						},
																						File: p1,
																					},
																				},
																			},
																			Body: &ast.Apply{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1034),
																							Column: int(11),
																						},
																						End: ast.Location{
																							Line: int(1034),
																							Column: int(42),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p9934,
																					FreeVars: ast.Identifiers{
																						"cindent",
																						"lines",
																						"std",
																					},
																				},
																				Target: &ast.Index{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1034),
																								Column: int(11),
																							},
																							End: ast.Location{
																								Line: int(1034),
																								Column: int(19),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p9934,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																					},
																					Target: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1034),
																									Column: int(11),
																								},
																								End: ast.Location{
																									Line: int(1034),
																									Column: int(14),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{
																								ast.FodderElement{
																									Kind: ast.FodderKind(0),
																									Blanks: int(0),
																									Indent: int(10),
																									Comment: []string{},
																								},
																							},
																							Ctx: nil,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																						},
																						Id: "std",
																					},
																					LeftBracketFodder: ast.Fodder{},
																					Index: &ast.LiteralString{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								File: nil,
																							},
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: nil,
																						},
																						Value: "join",
																						Kind: ast.LiteralStringKind(1),
																						BlockIndent: "",
																						BlockTermIndent: "",
																					},
																					RightBracketFodder: ast.Fodder{},
																					Id: nil,
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Binary{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1034),
																											Column: int(20),
																										},
																										End: ast.Location{
																											Line: int(1034),
																											Column: int(34),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p10680,
																									FreeVars: ast.Identifiers{
																										"cindent",
																									},
																								},
																								Left: &ast.LiteralString{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1034),
																												Column: int(20),
																											},
																											End: ast.Location{
																												Line: int(1034),
																												Column: int(24),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p10680,
																										FreeVars: nil,
																									},
																									Value: "\n",
																									Kind: ast.LiteralStringKind(1),
																									BlockIndent: "",
																									BlockTermIndent: "",
																								},
																								OpFodder: ast.Fodder{},
																								Op: ast.BinaryOp(3),
																								Right: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1034),
																												Column: int(27),
																											},
																											End: ast.Location{
																												Line: int(1034),
																												Column: int(34),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p10680,
																										FreeVars: ast.Identifiers{
																											"cindent",
																										},
																									},
																									Id: "cindent",
																								},
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1034),
																											Column: int(36),
																										},
																										End: ast.Location{
																											Line: int(1034),
																											Column: int(41),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p10680,
																									FreeVars: ast.Identifiers{
																										"lines",
																									},
																								},
																								Id: "lines",
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				TrailingComma: false,
																				TailStrict: false,
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: nil,
																			},
																		},
																	},
																},
																ElseFodder: nil,
																BranchFalse: &ast.LiteralNull{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: nil,
																	},
																},
															},
														},
													},
												},
											},
										},
									},
								},
							},
							Fun: nil,
							CloseFodder: nil,
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
						},
					},
					Body: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1035),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1035),
									Column: int(23),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p9923,
							FreeVars: ast.Identifiers{
								"aux",
								"value",
							},
						},
						Target: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1035),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1035),
										Column: int(8),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
										Comment: []string{},
									},
								},
								Ctx: p9923,
								FreeVars: ast.Identifiers{
									"aux",
								},
							},
							Id: "aux",
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1035),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1035),
													Column: int(14),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p10695,
											FreeVars: ast.Identifiers{
												"value",
											},
										},
										Id: "value",
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Array{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1035),
													Column: int(16),
												},
												End: ast.Location{
													Line: int(1035),
													Column: int(18),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p10695,
											FreeVars: nil,
										},
										Elements: nil,
										TrailingComma: false,
										CloseFodder: ast.Fodder{},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1035),
													Column: int(20),
												},
												End: ast.Location{
													Line: int(1035),
													Column: int(22),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p10695,
											FreeVars: nil,
										},
										Value: "",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(954),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1035),
					Column: int(23),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "manifestYamlStream",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "value",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1037),
								Column: int(22),
							},
							End: ast.Location{
								Line: int(1037),
								Column: int(27),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "indent_array_in_object",
						EqFodder: ast.Fodder{},
						DefaultArg: &ast.LiteralBoolean{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1037),
										Column: int(52),
									},
									End: ast.Location{
										Line: int(1037),
										Column: int(57),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p10704,
								FreeVars: nil,
							},
							Value: false,
						},
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1037),
								Column: int(29),
							},
							End: ast.Location{
								Line: int(1037),
								Column: int(57),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "c_document_end",
						EqFodder: ast.Fodder{},
						DefaultArg: &ast.LiteralBoolean{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1037),
										Column: int(74),
									},
									End: ast.Location{
										Line: int(1037),
										Column: int(78),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p10704,
								FreeVars: nil,
							},
							Value: true,
						},
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1037),
								Column: int(59),
							},
							End: ast.Location{
								Line: int(1037),
								Column: int(78),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1038),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1043),
								Column: int(53),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p10704,
						FreeVars: ast.Identifiers{
							"c_document_end",
							"indent_array_in_object",
							"std",
							"value",
						},
					},
					Cond: &ast.Unary{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1038),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(1038),
									Column: int(27),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p10704,
							FreeVars: ast.Identifiers{
								"std",
								"value",
							},
						},
						Op: ast.UnaryOp(0),
						Expr: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1038),
										Column: int(9),
									},
									End: ast.Location{
										Line: int(1038),
										Column: int(27),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p10704,
								FreeVars: ast.Identifiers{
									"std",
									"value",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1038),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(1038),
											Column: int(20),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p10704,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1038),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(1038),
												Column: int(12),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "isArray",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1038),
														Column: int(21),
													},
													End: ast.Location{
														Line: int(1038),
														Column: int(26),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p10720,
												FreeVars: ast.Identifiers{
													"value",
												},
											},
											Id: "value",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
					},
					ThenFodder: ast.Fodder{},
					BranchTrue: &ast.Error{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1039),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(1039),
									Column: int(75),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
									Comment: []string{},
								},
							},
							Ctx: p10704,
							FreeVars: ast.Identifiers{
								"std",
								"value",
							},
						},
						Expr: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1039),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(1039),
										Column: int(75),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p10704,
								FreeVars: ast.Identifiers{
									"std",
									"value",
								},
							},
							Left: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1039),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(1039),
											Column: int(57),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p10704,
									FreeVars: nil,
								},
								Value: "manifestYamlStream only takes arrays, got ",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(3),
							Right: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1039),
											Column: int(60),
										},
										End: ast.Location{
											Line: int(1039),
											Column: int(75),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p10704,
									FreeVars: ast.Identifiers{
										"std",
										"value",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1039),
												Column: int(60),
											},
											End: ast.Location{
												Line: int(1039),
												Column: int(68),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p10704,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1039),
													Column: int(60),
												},
												End: ast.Location{
													Line: int(1039),
													Column: int(63),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "type",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1039),
															Column: int(69),
														},
														End: ast.Location{
															Line: int(1039),
															Column: int(74),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p10737,
													FreeVars: ast.Identifiers{
														"value",
													},
												},
												Id: "value",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
						},
					},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
							Comment: []string{},
						},
					},
					BranchFalse: &ast.Binary{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1041),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(1043),
									Column: int(53),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p10704,
							FreeVars: ast.Identifiers{
								"c_document_end",
								"indent_array_in_object",
								"std",
								"value",
							},
						},
						Left: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1041),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(1043),
										Column: int(8),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p10704,
								FreeVars: ast.Identifiers{
									"indent_array_in_object",
									"std",
									"value",
								},
							},
							Left: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1041),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(1041),
											Column: int(14),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
											Comment: []string{},
										},
									},
									Ctx: p10704,
									FreeVars: nil,
								},
								Value: "---\n",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(3),
							Right: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1041),
											Column: int(17),
										},
										End: ast.Location{
											Line: int(1043),
											Column: int(8),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p10704,
									FreeVars: ast.Identifiers{
										"indent_array_in_object",
										"std",
										"value",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1041),
												Column: int(17),
											},
											End: ast.Location{
												Line: int(1041),
												Column: int(25),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p10704,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1041),
													Column: int(17),
												},
												End: ast.Location{
													Line: int(1041),
													Column: int(20),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "join",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1042),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1042),
															Column: int(18),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
															Comment: []string{},
														},
													},
													Ctx: p10756,
													FreeVars: nil,
												},
												Value: "\n---\n",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											CommaFodder: ast.Fodder{},
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"indent_array_in_object",
														"std",
														"value",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: nil,
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "flatMap",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: nil,
													Id: nil,
												},
												FodderLeft: nil,
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Function{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"indent_array_in_object",
																		"std",
																	},
																},
																ParenLeftFodder: nil,
																Parameters: []ast.Parameter{
																	ast.Parameter{
																		NameFodder: nil,
																		Name: "e",
																		EqFodder: nil,
																		DefaultArg: nil,
																		CommaFodder: nil,
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																	},
																},
																TrailingComma: false,
																ParenRightFodder: nil,
																Body: &ast.Array{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"e",
																			"indent_array_in_object",
																			"std",
																		},
																	},
																	Elements: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Apply{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1042),
																							Column: int(21),
																						},
																						End: ast.Location{
																							Line: int(1042),
																							Column: int(67),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p10772,
																					FreeVars: ast.Identifiers{
																						"e",
																						"indent_array_in_object",
																						"std",
																					},
																				},
																				Target: &ast.Index{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1042),
																								Column: int(21),
																							},
																							End: ast.Location{
																								Line: int(1042),
																								Column: int(40),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p10772,
																						FreeVars: ast.Identifiers{
																							"std",
																						},
																					},
																					Target: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1042),
																									Column: int(21),
																								},
																								End: ast.Location{
																									Line: int(1042),
																									Column: int(24),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: nil,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																						},
																						Id: "std",
																					},
																					LeftBracketFodder: ast.Fodder{},
																					Index: &ast.LiteralString{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								File: nil,
																							},
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: nil,
																						},
																						Value: "manifestYamlDoc",
																						Kind: ast.LiteralStringKind(1),
																						BlockIndent: "",
																						BlockTermIndent: "",
																					},
																					RightBracketFodder: ast.Fodder{},
																					Id: nil,
																				},
																				FodderLeft: ast.Fodder{},
																				Arguments: ast.Arguments{
																					Positional: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1042),
																											Column: int(41),
																										},
																										End: ast.Location{
																											Line: int(1042),
																											Column: int(42),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p10781,
																									FreeVars: ast.Identifiers{
																										"e",
																									},
																								},
																								Id: "e",
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1042),
																											Column: int(44),
																										},
																										End: ast.Location{
																											Line: int(1042),
																											Column: int(66),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p10781,
																									FreeVars: ast.Identifiers{
																										"indent_array_in_object",
																									},
																								},
																								Id: "indent_array_in_object",
																							},
																							CommaFodder: nil,
																						},
																					},
																					Named: nil,
																				},
																				TrailingComma: false,
																				TailStrict: false,
																				FodderRight: ast.Fodder{},
																				TailStrictFodder: nil,
																			},
																			CommaFodder: nil,
																		},
																	},
																	TrailingComma: false,
																	CloseFodder: nil,
																},
															},
															CommaFodder: nil,
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1042),
																			Column: int(77),
																		},
																		End: ast.Location{
																			Line: int(1042),
																			Column: int(82),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p10756,
																	FreeVars: ast.Identifiers{
																		"value",
																	},
																},
																Id: "value",
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: nil,
												TailStrictFodder: nil,
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
										Comment: []string{},
									},
								},
								TailStrictFodder: nil,
							},
						},
						OpFodder: ast.Fodder{},
						Op: ast.BinaryOp(3),
						Right: &ast.Conditional{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1043),
										Column: int(11),
									},
									End: ast.Location{
										Line: int(1043),
										Column: int(53),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p10704,
								FreeVars: ast.Identifiers{
									"c_document_end",
								},
							},
							Cond: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1043),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(1043),
											Column: int(28),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p10704,
									FreeVars: ast.Identifiers{
										"c_document_end",
									},
								},
								Id: "c_document_end",
							},
							ThenFodder: ast.Fodder{},
							BranchTrue: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1043),
											Column: int(34),
										},
										End: ast.Location{
											Line: int(1043),
											Column: int(43),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p10704,
									FreeVars: nil,
								},
								Value: "\n...\n",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							ElseFodder: ast.Fodder{},
							BranchFalse: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1043),
											Column: int(49),
										},
										End: ast.Location{
											Line: int(1043),
											Column: int(53),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p10704,
									FreeVars: nil,
								},
								Value: "\n",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(1037),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1043),
					Column: int(53),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "manifestPython",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "v",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1046),
								Column: int(18),
							},
							End: ast.Location{
								Line: int(1046),
								Column: int(19),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1047),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1066),
								Column: int(13),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p10800,
						FreeVars: ast.Identifiers{
							"std",
							"v",
						},
					},
					Cond: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1047),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(1047),
									Column: int(23),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p10800,
							FreeVars: ast.Identifiers{
								"std",
								"v",
							},
						},
						Target: &ast.Index{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1047),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(1047),
										Column: int(20),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p10800,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1047),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(1047),
											Column: int(11),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Id: "std",
							},
							LeftBracketFodder: ast.Fodder{},
							Index: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Value: "isObject",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							RightBracketFodder: ast.Fodder{},
							Id: nil,
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1047),
													Column: int(21),
												},
												End: ast.Location{
													Line: int(1047),
													Column: int(22),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p10811,
											FreeVars: ast.Identifiers{
												"v",
											},
										},
										Id: "v",
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
					ThenFodder: ast.Fodder{},
					BranchTrue: &ast.Local{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1048),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(1052),
									Column: int(40),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
									Comment: []string{},
								},
							},
							Ctx: p10800,
							FreeVars: ast.Identifiers{
								"std",
								"v",
							},
						},
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: ast.Fodder{},
								Variable: "fields",
								EqFodder: ast.Fodder{},
								Body: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
											"v",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: nil,
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "flatMap",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: nil,
										Id: nil,
									},
									FodderLeft: nil,
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Function{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
															"v",
														},
													},
													ParenLeftFodder: nil,
													Parameters: []ast.Parameter{
														ast.Parameter{
															NameFodder: nil,
															Name: "k",
															EqFodder: nil,
															DefaultArg: nil,
															CommaFodder: nil,
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
														},
													},
													TrailingComma: false,
													ParenRightFodder: nil,
													Body: &ast.Array{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"k",
																"std",
																"v",
															},
														},
														Elements: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Apply{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"k",
																			"std",
																			"v",
																		},
																	},
																	Target: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Id: "std",
																		},
																		LeftBracketFodder: nil,
																		Index: &ast.LiteralString{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: nil,
																			},
																			Value: "mod",
																			Kind: ast.LiteralStringKind(1),
																			BlockIndent: "",
																			BlockTermIndent: "",
																		},
																		RightBracketFodder: nil,
																		Id: nil,
																	},
																	FodderLeft: nil,
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.LiteralString{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1049),
																								Column: int(9),
																							},
																							End: ast.Location{
																								Line: int(1049),
																								Column: int(17),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(8),
																								Comment: []string{},
																							},
																						},
																						Ctx: p10841,
																						FreeVars: nil,
																					},
																					Value: "%s: %s",
																					Kind: ast.LiteralStringKind(1),
																					BlockIndent: "",
																					BlockTermIndent: "",
																				},
																				CommaFodder: nil,
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Array{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1049),
																								Column: int(20),
																							},
																							End: ast.Location{
																								Line: int(1049),
																								Column: int(73),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p10841,
																						FreeVars: ast.Identifiers{
																							"k",
																							"std",
																							"v",
																						},
																					},
																					Elements: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Apply{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1049),
																											Column: int(21),
																										},
																										End: ast.Location{
																											Line: int(1049),
																											Column: int(46),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p10846,
																									FreeVars: ast.Identifiers{
																										"k",
																										"std",
																									},
																								},
																								Target: &ast.Index{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1049),
																												Column: int(21),
																											},
																											End: ast.Location{
																												Line: int(1049),
																												Column: int(43),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p10846,
																										FreeVars: ast.Identifiers{
																											"std",
																										},
																									},
																									Target: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(1049),
																													Column: int(21),
																												},
																												End: ast.Location{
																													Line: int(1049),
																													Column: int(24),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"std",
																											},
																										},
																										Id: "std",
																									},
																									LeftBracketFodder: ast.Fodder{},
																									Index: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												File: nil,
																											},
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: nil,
																										},
																										Value: "escapeStringPython",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																									RightBracketFodder: ast.Fodder{},
																									Id: nil,
																								},
																								FodderLeft: ast.Fodder{},
																								Arguments: ast.Arguments{
																									Positional: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(1049),
																															Column: int(44),
																														},
																														End: ast.Location{
																															Line: int(1049),
																															Column: int(45),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p10855,
																													FreeVars: ast.Identifiers{
																														"k",
																													},
																												},
																												Id: "k",
																											},
																											CommaFodder: nil,
																										},
																									},
																									Named: nil,
																								},
																								TrailingComma: false,
																								TailStrict: false,
																								FodderRight: ast.Fodder{},
																								TailStrictFodder: nil,
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Apply{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1049),
																											Column: int(48),
																										},
																										End: ast.Location{
																											Line: int(1049),
																											Column: int(72),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p10846,
																									FreeVars: ast.Identifiers{
																										"k",
																										"std",
																										"v",
																									},
																								},
																								Target: &ast.Index{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1049),
																												Column: int(48),
																											},
																											End: ast.Location{
																												Line: int(1049),
																												Column: int(66),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p10846,
																										FreeVars: ast.Identifiers{
																											"std",
																										},
																									},
																									Target: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(1049),
																													Column: int(48),
																												},
																												End: ast.Location{
																													Line: int(1049),
																													Column: int(51),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"std",
																											},
																										},
																										Id: "std",
																									},
																									LeftBracketFodder: ast.Fodder{},
																									Index: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												File: nil,
																											},
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: nil,
																										},
																										Value: "manifestPython",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																									RightBracketFodder: ast.Fodder{},
																									Id: nil,
																								},
																								FodderLeft: ast.Fodder{},
																								Arguments: ast.Arguments{
																									Positional: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Index{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(1049),
																															Column: int(67),
																														},
																														End: ast.Location{
																															Line: int(1049),
																															Column: int(71),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p10866,
																													FreeVars: ast.Identifiers{
																														"k",
																														"v",
																													},
																												},
																												Target: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(1049),
																																Column: int(67),
																															},
																															End: ast.Location{
																																Line: int(1049),
																																Column: int(68),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p10866,
																														FreeVars: ast.Identifiers{
																															"v",
																														},
																													},
																													Id: "v",
																												},
																												LeftBracketFodder: ast.Fodder{},
																												Index: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(1049),
																																Column: int(69),
																															},
																															End: ast.Location{
																																Line: int(1049),
																																Column: int(70),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p10866,
																														FreeVars: ast.Identifiers{
																															"k",
																														},
																													},
																													Id: "k",
																												},
																												RightBracketFodder: ast.Fodder{},
																												Id: nil,
																											},
																											CommaFodder: nil,
																										},
																									},
																									Named: nil,
																								},
																								TrailingComma: false,
																								TailStrict: false,
																								FodderRight: ast.Fodder{},
																								TailStrictFodder: nil,
																							},
																							CommaFodder: nil,
																						},
																					},
																					TrailingComma: false,
																					CloseFodder: ast.Fodder{},
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	TrailingComma: false,
																	TailStrict: false,
																	FodderRight: nil,
																	TailStrictFodder: nil,
																},
																CommaFodder: nil,
															},
														},
														TrailingComma: false,
														CloseFodder: nil,
													},
												},
												CommaFodder: nil,
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1050),
																Column: int(18),
															},
															End: ast.Location{
																Line: int(1050),
																Column: int(37),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p10873,
														FreeVars: ast.Identifiers{
															"std",
															"v",
														},
													},
													Target: &ast.Index{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1050),
																	Column: int(18),
																},
																End: ast.Location{
																	Line: int(1050),
																	Column: int(34),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p10873,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1050),
																		Column: int(18),
																	},
																	End: ast.Location{
																		Line: int(1050),
																		Column: int(21),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Id: "std",
														},
														LeftBracketFodder: ast.Fodder{},
														Index: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
															Value: "objectFields",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														RightBracketFodder: ast.Fodder{},
														Id: nil,
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1050),
																				Column: int(35),
																			},
																			End: ast.Location{
																				Line: int(1050),
																				Column: int(36),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p10882,
																		FreeVars: ast.Identifiers{
																			"v",
																		},
																	},
																	Id: "v",
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: nil,
									TailStrictFodder: nil,
								},
								Fun: nil,
								CloseFodder: ast.Fodder{},
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1048),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(1051),
										Column: int(8),
									},
									File: p1,
								},
							},
						},
						Body: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"fields",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: nil,
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "mod",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: nil,
								Id: nil,
							},
							FodderLeft: nil,
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1052),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(1052),
														Column: int(13),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
														Comment: []string{},
													},
												},
												Ctx: p10800,
												FreeVars: nil,
											},
											Value: "{%s}",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										CommaFodder: nil,
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Array{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1052),
														Column: int(16),
													},
													End: ast.Location{
														Line: int(1052),
														Column: int(40),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p10800,
												FreeVars: ast.Identifiers{
													"fields",
													"std",
												},
											},
											Elements: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1052),
																	Column: int(17),
																},
																End: ast.Location{
																	Line: int(1052),
																	Column: int(39),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p10898,
															FreeVars: ast.Identifiers{
																"fields",
																"std",
															},
														},
														Target: &ast.Index{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1052),
																		Column: int(17),
																	},
																	End: ast.Location{
																		Line: int(1052),
																		Column: int(25),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p10898,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Target: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1052),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(1052),
																			Column: int(20),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Id: "std",
															},
															LeftBracketFodder: ast.Fodder{},
															Index: &ast.LiteralString{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: nil,
																},
																Value: "join",
																Kind: ast.LiteralStringKind(1),
																BlockIndent: "",
																BlockTermIndent: "",
															},
															RightBracketFodder: ast.Fodder{},
															Id: nil,
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1052),
																					Column: int(26),
																				},
																				End: ast.Location{
																					Line: int(1052),
																					Column: int(30),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p10907,
																			FreeVars: nil,
																		},
																		Value: ", ",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1052),
																					Column: int(32),
																				},
																				End: ast.Location{
																					Line: int(1052),
																					Column: int(38),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p10907,
																			FreeVars: ast.Identifiers{
																				"fields",
																			},
																		},
																		Id: "fields",
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														TrailingComma: false,
														TailStrict: false,
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
													},
													CommaFodder: nil,
												},
											},
											TrailingComma: false,
											CloseFodder: ast.Fodder{},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: nil,
							TailStrictFodder: nil,
						},
					},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
							Comment: []string{},
						},
					},
					BranchFalse: &ast.Conditional{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1053),
									Column: int(10),
								},
								End: ast.Location{
									Line: int(1066),
									Column: int(13),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p10800,
							FreeVars: ast.Identifiers{
								"std",
								"v",
							},
						},
						Cond: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1053),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(1053),
										Column: int(27),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p10800,
								FreeVars: ast.Identifiers{
									"std",
									"v",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1053),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(1053),
											Column: int(24),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p10800,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1053),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(1053),
												Column: int(16),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "isArray",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1053),
														Column: int(25),
													},
													End: ast.Location{
														Line: int(1053),
														Column: int(26),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p10922,
												FreeVars: ast.Identifiers{
													"v",
												},
											},
											Id: "v",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
						ThenFodder: ast.Fodder{},
						BranchTrue: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
									"v",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: nil,
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "mod",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: nil,
								Id: nil,
							},
							FodderLeft: nil,
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1054),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(1054),
														Column: int(13),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
														Comment: []string{},
													},
												},
												Ctx: p10800,
												FreeVars: nil,
											},
											Value: "[%s]",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										CommaFodder: nil,
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Array{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1054),
														Column: int(16),
													},
													End: ast.Location{
														Line: int(1054),
														Column: int(70),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p10800,
												FreeVars: ast.Identifiers{
													"std",
													"v",
												},
											},
											Elements: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1054),
																	Column: int(17),
																},
																End: ast.Location{
																	Line: int(1054),
																	Column: int(69),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p10938,
															FreeVars: ast.Identifiers{
																"std",
																"v",
															},
														},
														Target: &ast.Index{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1054),
																		Column: int(17),
																	},
																	End: ast.Location{
																		Line: int(1054),
																		Column: int(25),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p10938,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Target: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1054),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(1054),
																			Column: int(20),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Id: "std",
															},
															LeftBracketFodder: ast.Fodder{},
															Index: &ast.LiteralString{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: nil,
																},
																Value: "join",
																Kind: ast.LiteralStringKind(1),
																BlockIndent: "",
																BlockTermIndent: "",
															},
															RightBracketFodder: ast.Fodder{},
															Id: nil,
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1054),
																					Column: int(26),
																				},
																				End: ast.Location{
																					Line: int(1054),
																					Column: int(30),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p10947,
																			FreeVars: nil,
																		},
																		Value: ", ",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Apply{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																				"v",
																			},
																		},
																		Target: &ast.Index{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																				},
																				Id: "std",
																			},
																			LeftBracketFodder: nil,
																			Index: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: nil,
																				},
																				Value: "flatMap",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																			RightBracketFodder: nil,
																			Id: nil,
																		},
																		FodderLeft: nil,
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Function{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								File: nil,
																							},
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{
																								"std",
																							},
																						},
																						ParenLeftFodder: nil,
																						Parameters: []ast.Parameter{
																							ast.Parameter{
																								NameFodder: nil,
																								Name: "v2",
																								EqFodder: nil,
																								DefaultArg: nil,
																								CommaFodder: nil,
																								LocRange: ast.LocationRange{
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									File: nil,
																								},
																							},
																						},
																						TrailingComma: false,
																						ParenRightFodder: nil,
																						Body: &ast.Array{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									File: nil,
																								},
																								Fodder: nil,
																								Ctx: nil,
																								FreeVars: ast.Identifiers{
																									"std",
																									"v2",
																								},
																							},
																							Elements: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Apply{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(1054),
																													Column: int(33),
																												},
																												End: ast.Location{
																													Line: int(1054),
																													Column: int(55),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p10963,
																											FreeVars: ast.Identifiers{
																												"std",
																												"v2",
																											},
																										},
																										Target: &ast.Index{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1054),
																														Column: int(33),
																													},
																													End: ast.Location{
																														Line: int(1054),
																														Column: int(51),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p10963,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																											},
																											Target: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(1054),
																															Column: int(33),
																														},
																														End: ast.Location{
																															Line: int(1054),
																															Column: int(36),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																												},
																												Id: "std",
																											},
																											LeftBracketFodder: ast.Fodder{},
																											Index: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: nil,
																												},
																												Value: "manifestPython",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																											RightBracketFodder: ast.Fodder{},
																											Id: nil,
																										},
																										FodderLeft: ast.Fodder{},
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(1054),
																																	Column: int(52),
																																},
																																End: ast.Location{
																																	Line: int(1054),
																																	Column: int(54),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p10972,
																															FreeVars: ast.Identifiers{
																																"v2",
																															},
																														},
																														Id: "v2",
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										TrailingComma: false,
																										TailStrict: false,
																										FodderRight: ast.Fodder{},
																										TailStrictFodder: nil,
																									},
																									CommaFodder: nil,
																								},
																							},
																							TrailingComma: false,
																							CloseFodder: nil,
																						},
																					},
																					CommaFodder: nil,
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1054),
																									Column: int(66),
																								},
																								End: ast.Location{
																									Line: int(1054),
																									Column: int(67),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p10947,
																							FreeVars: ast.Identifiers{
																								"v",
																							},
																						},
																						Id: "v",
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		TrailingComma: false,
																		TailStrict: false,
																		FodderRight: nil,
																		TailStrictFodder: nil,
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														TrailingComma: false,
														TailStrict: false,
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
													},
													CommaFodder: nil,
												},
											},
											TrailingComma: false,
											CloseFodder: ast.Fodder{},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: nil,
							TailStrictFodder: nil,
						},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						BranchFalse: &ast.Conditional{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1055),
										Column: int(10),
									},
									End: ast.Location{
										Line: int(1066),
										Column: int(13),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p10800,
								FreeVars: ast.Identifiers{
									"std",
									"v",
								},
							},
							Cond: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1055),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(1055),
											Column: int(28),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p10800,
									FreeVars: ast.Identifiers{
										"std",
										"v",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1055),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(1055),
												Column: int(25),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p10800,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1055),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(1055),
													Column: int(16),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "isString",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1055),
															Column: int(26),
														},
														End: ast.Location{
															Line: int(1055),
															Column: int(27),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p10988,
													FreeVars: ast.Identifiers{
														"v",
													},
												},
												Id: "v",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
							ThenFodder: ast.Fodder{},
							BranchTrue: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
										"v",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: nil,
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "mod",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: nil,
									Id: nil,
								},
								FodderLeft: nil,
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1056),
															Column: int(7),
														},
														End: ast.Location{
															Line: int(1056),
															Column: int(11),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(6),
															Comment: []string{},
														},
													},
													Ctx: p10800,
													FreeVars: nil,
												},
												Value: "%s",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											CommaFodder: nil,
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.Array{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1056),
															Column: int(14),
														},
														End: ast.Location{
															Line: int(1056),
															Column: int(41),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p10800,
													FreeVars: ast.Identifiers{
														"std",
														"v",
													},
												},
												Elements: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Apply{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1056),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(1056),
																		Column: int(40),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p11004,
																FreeVars: ast.Identifiers{
																	"std",
																	"v",
																},
															},
															Target: &ast.Index{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1056),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(1056),
																			Column: int(37),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p11004,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1056),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(1056),
																				Column: int(18),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Id: "std",
																},
																LeftBracketFodder: ast.Fodder{},
																Index: &ast.LiteralString{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: nil,
																	},
																	Value: "escapeStringPython",
																	Kind: ast.LiteralStringKind(1),
																	BlockIndent: "",
																	BlockTermIndent: "",
																},
																RightBracketFodder: ast.Fodder{},
																Id: nil,
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1056),
																						Column: int(38),
																					},
																					End: ast.Location{
																						Line: int(1056),
																						Column: int(39),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p11013,
																				FreeVars: ast.Identifiers{
																					"v",
																				},
																			},
																			Id: "v",
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															TrailingComma: false,
															TailStrict: false,
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
														},
														CommaFodder: nil,
													},
												},
												TrailingComma: false,
												CloseFodder: ast.Fodder{},
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: nil,
								TailStrictFodder: nil,
							},
							ElseFodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
									Comment: []string{},
								},
							},
							BranchFalse: &ast.Conditional{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1057),
											Column: int(10),
										},
										End: ast.Location{
											Line: int(1066),
											Column: int(13),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p10800,
									FreeVars: ast.Identifiers{
										"std",
										"v",
									},
								},
								Cond: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1057),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(1057),
												Column: int(30),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p10800,
										FreeVars: ast.Identifiers{
											"std",
											"v",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1057),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(1057),
													Column: int(27),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p10800,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1057),
														Column: int(13),
													},
													End: ast.Location{
														Line: int(1057),
														Column: int(16),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "isFunction",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1057),
																Column: int(28),
															},
															End: ast.Location{
																Line: int(1057),
																Column: int(29),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p11027,
														FreeVars: ast.Identifiers{
															"v",
														},
													},
													Id: "v",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
								ThenFodder: ast.Fodder{},
								BranchTrue: &ast.Error{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1058),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1058),
												Column: int(39),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										Ctx: p10800,
										FreeVars: nil,
									},
									Expr: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1058),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(1058),
													Column: int(39),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p10800,
											FreeVars: nil,
										},
										Value: "cannot manifest function",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
								},
								ElseFodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
										Comment: []string{},
									},
								},
								BranchFalse: &ast.Conditional{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1059),
												Column: int(10),
											},
											End: ast.Location{
												Line: int(1066),
												Column: int(13),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p10800,
										FreeVars: ast.Identifiers{
											"std",
											"v",
										},
									},
									Cond: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1059),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(1059),
													Column: int(28),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p10800,
											FreeVars: ast.Identifiers{
												"std",
												"v",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1059),
														Column: int(13),
													},
													End: ast.Location{
														Line: int(1059),
														Column: int(25),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p10800,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1059),
															Column: int(13),
														},
														End: ast.Location{
															Line: int(1059),
															Column: int(16),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: ast.Fodder{},
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "isNumber",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: ast.Fodder{},
											Id: nil,
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1059),
																	Column: int(26),
																},
																End: ast.Location{
																	Line: int(1059),
																	Column: int(27),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p11044,
															FreeVars: ast.Identifiers{
																"v",
															},
														},
														Id: "v",
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
									ThenFodder: ast.Fodder{},
									BranchTrue: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1060),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(1060),
													Column: int(22),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p10800,
											FreeVars: ast.Identifiers{
												"std",
												"v",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1060),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(1060),
														Column: int(19),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p10800,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1060),
															Column: int(7),
														},
														End: ast.Location{
															Line: int(1060),
															Column: int(10),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(6),
															Comment: []string{},
														},
													},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: ast.Fodder{},
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "toString",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: ast.Fodder{},
											Id: nil,
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1060),
																	Column: int(20),
																},
																End: ast.Location{
																	Line: int(1060),
																	Column: int(21),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p11056,
															FreeVars: ast.Identifiers{
																"v",
															},
														},
														Id: "v",
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
											Comment: []string{},
										},
									},
									BranchFalse: &ast.Conditional{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1061),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(1066),
													Column: int(13),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p10800,
											FreeVars: ast.Identifiers{
												"v",
											},
										},
										Cond: &ast.Binary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1061),
														Column: int(13),
													},
													End: ast.Location{
														Line: int(1061),
														Column: int(22),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p10800,
												FreeVars: ast.Identifiers{
													"v",
												},
											},
											Left: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1061),
															Column: int(13),
														},
														End: ast.Location{
															Line: int(1061),
															Column: int(14),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p10800,
													FreeVars: ast.Identifiers{
														"v",
													},
												},
												Id: "v",
											},
											OpFodder: ast.Fodder{},
											Op: ast.BinaryOp(12),
											Right: &ast.LiteralBoolean{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1061),
															Column: int(18),
														},
														End: ast.Location{
															Line: int(1061),
															Column: int(22),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p10800,
													FreeVars: nil,
												},
												Value: true,
											},
										},
										ThenFodder: ast.Fodder{},
										BranchTrue: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1062),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(1062),
														Column: int(13),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
														Comment: []string{},
													},
												},
												Ctx: p10800,
												FreeVars: nil,
											},
											Value: "True",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(4),
												Comment: []string{},
											},
										},
										BranchFalse: &ast.Conditional{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1063),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(1066),
														Column: int(13),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p10800,
												FreeVars: ast.Identifiers{
													"v",
												},
											},
											Cond: &ast.Binary{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1063),
															Column: int(13),
														},
														End: ast.Location{
															Line: int(1063),
															Column: int(23),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p10800,
													FreeVars: ast.Identifiers{
														"v",
													},
												},
												Left: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1063),
																Column: int(13),
															},
															End: ast.Location{
																Line: int(1063),
																Column: int(14),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p10800,
														FreeVars: ast.Identifiers{
															"v",
														},
													},
													Id: "v",
												},
												OpFodder: ast.Fodder{},
												Op: ast.BinaryOp(12),
												Right: &ast.LiteralBoolean{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1063),
																Column: int(18),
															},
															End: ast.Location{
																Line: int(1063),
																Column: int(23),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p10800,
														FreeVars: nil,
													},
													Value: false,
												},
											},
											ThenFodder: ast.Fodder{},
											BranchTrue: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1064),
															Column: int(7),
														},
														End: ast.Location{
															Line: int(1064),
															Column: int(14),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(6),
															Comment: []string{},
														},
													},
													Ctx: p10800,
													FreeVars: nil,
												},
												Value: "False",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(4),
													Comment: []string{},
												},
											},
											BranchFalse: &ast.Conditional{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1065),
															Column: int(10),
														},
														End: ast.Location{
															Line: int(1066),
															Column: int(13),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p10800,
													FreeVars: ast.Identifiers{
														"v",
													},
												},
												Cond: &ast.Binary{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1065),
																Column: int(13),
															},
															End: ast.Location{
																Line: int(1065),
																Column: int(22),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p10800,
														FreeVars: ast.Identifiers{
															"v",
														},
													},
													Left: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1065),
																	Column: int(13),
																},
																End: ast.Location{
																	Line: int(1065),
																	Column: int(14),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p10800,
															FreeVars: ast.Identifiers{
																"v",
															},
														},
														Id: "v",
													},
													OpFodder: ast.Fodder{},
													Op: ast.BinaryOp(12),
													Right: &ast.LiteralNull{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1065),
																	Column: int(18),
																},
																End: ast.Location{
																	Line: int(1065),
																	Column: int(22),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p10800,
															FreeVars: nil,
														},
													},
												},
												ThenFodder: ast.Fodder{},
												BranchTrue: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1066),
																Column: int(7),
															},
															End: ast.Location{
																Line: int(1066),
																Column: int(13),
															},
															File: p1,
														},
														Fodder: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(6),
																Comment: []string{},
															},
														},
														Ctx: p10800,
														FreeVars: nil,
													},
													Value: "None",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												ElseFodder: nil,
												BranchFalse: &ast.LiteralNull{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
												},
											},
										},
									},
								},
							},
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(1046),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1066),
					Column: int(13),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "manifestPythonVars",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "conf",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1068),
								Column: int(22),
							},
							End: ast.Location{
								Line: int(1068),
								Column: int(26),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1069),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1070),
								Column: int(32),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p11095,
						FreeVars: ast.Identifiers{
							"conf",
							"std",
						},
					},
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Variable: "vars",
							EqFodder: ast.Fodder{},
							Body: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"conf",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: nil,
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "flatMap",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: nil,
									Id: nil,
								},
								FodderLeft: nil,
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Function{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"conf",
														"std",
													},
												},
												ParenLeftFodder: nil,
												Parameters: []ast.Parameter{
													ast.Parameter{
														NameFodder: nil,
														Name: "k",
														EqFodder: nil,
														DefaultArg: nil,
														CommaFodder: nil,
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
													},
												},
												TrailingComma: false,
												ParenRightFodder: nil,
												Body: &ast.Array{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"conf",
															"k",
															"std",
														},
													},
													Elements: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Apply{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"conf",
																		"k",
																		"std",
																	},
																},
																Target: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Id: "std",
																	},
																	LeftBracketFodder: nil,
																	Index: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: nil,
																		},
																		Value: "mod",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	RightBracketFodder: nil,
																	Id: nil,
																},
																FodderLeft: nil,
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1069),
																							Column: int(19),
																						},
																						End: ast.Location{
																							Line: int(1069),
																							Column: int(28),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p11121,
																					FreeVars: nil,
																				},
																				Value: "%s = %s",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																			CommaFodder: nil,
																		},
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Array{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1069),
																							Column: int(31),
																						},
																						End: ast.Location{
																							Line: int(1069),
																							Column: int(63),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p11121,
																					FreeVars: ast.Identifiers{
																						"conf",
																						"k",
																						"std",
																					},
																				},
																				Elements: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1069),
																										Column: int(32),
																									},
																									End: ast.Location{
																										Line: int(1069),
																										Column: int(33),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p11126,
																								FreeVars: ast.Identifiers{
																									"k",
																								},
																							},
																							Id: "k",
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Apply{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1069),
																										Column: int(35),
																									},
																									End: ast.Location{
																										Line: int(1069),
																										Column: int(62),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p11126,
																								FreeVars: ast.Identifiers{
																									"conf",
																									"k",
																									"std",
																								},
																							},
																							Target: &ast.Index{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1069),
																											Column: int(35),
																										},
																										End: ast.Location{
																											Line: int(1069),
																											Column: int(53),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p11126,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																								},
																								Target: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1069),
																												Column: int(35),
																											},
																											End: ast.Location{
																												Line: int(1069),
																												Column: int(38),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: nil,
																										FreeVars: ast.Identifiers{
																											"std",
																										},
																									},
																									Id: "std",
																								},
																								LeftBracketFodder: ast.Fodder{},
																								Index: &ast.LiteralString{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											File: nil,
																										},
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: nil,
																									},
																									Value: "manifestPython",
																									Kind: ast.LiteralStringKind(1),
																									BlockIndent: "",
																									BlockTermIndent: "",
																								},
																								RightBracketFodder: ast.Fodder{},
																								Id: nil,
																							},
																							FodderLeft: ast.Fodder{},
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Index{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1069),
																														Column: int(54),
																													},
																													End: ast.Location{
																														Line: int(1069),
																														Column: int(61),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p11137,
																												FreeVars: ast.Identifiers{
																													"conf",
																													"k",
																												},
																											},
																											Target: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(1069),
																															Column: int(54),
																														},
																														End: ast.Location{
																															Line: int(1069),
																															Column: int(58),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p11137,
																													FreeVars: ast.Identifiers{
																														"conf",
																													},
																												},
																												Id: "conf",
																											},
																											LeftBracketFodder: ast.Fodder{},
																											Index: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(1069),
																															Column: int(59),
																														},
																														End: ast.Location{
																															Line: int(1069),
																															Column: int(60),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p11137,
																													FreeVars: ast.Identifiers{
																														"k",
																													},
																												},
																												Id: "k",
																											},
																											RightBracketFodder: ast.Fodder{},
																											Id: nil,
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							TrailingComma: false,
																							TailStrict: false,
																							FodderRight: ast.Fodder{},
																							TailStrictFodder: nil,
																						},
																						CommaFodder: nil,
																					},
																				},
																				TrailingComma: false,
																				CloseFodder: ast.Fodder{},
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																TrailingComma: false,
																TailStrict: false,
																FodderRight: nil,
																TailStrictFodder: nil,
															},
															CommaFodder: nil,
														},
													},
													TrailingComma: false,
													CloseFodder: nil,
												},
											},
											CommaFodder: nil,
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1069),
															Column: int(73),
														},
														End: ast.Location{
															Line: int(1069),
															Column: int(95),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p11144,
													FreeVars: ast.Identifiers{
														"conf",
														"std",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1069),
																Column: int(73),
															},
															End: ast.Location{
																Line: int(1069),
																Column: int(89),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p11144,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1069),
																	Column: int(73),
																},
																End: ast.Location{
																	Line: int(1069),
																	Column: int(76),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: ast.Fodder{},
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "objectFields",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: ast.Fodder{},
													Id: nil,
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1069),
																			Column: int(90),
																		},
																		End: ast.Location{
																			Line: int(1069),
																			Column: int(94),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p11153,
																	FreeVars: ast.Identifiers{
																		"conf",
																	},
																},
																Id: "conf",
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: nil,
								TailStrictFodder: nil,
							},
							Fun: nil,
							CloseFodder: ast.Fodder{},
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1069),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(1069),
									Column: int(96),
								},
								File: p1,
							},
						},
					},
					Body: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1070),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1070),
									Column: int(32),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p11095,
							FreeVars: ast.Identifiers{
								"std",
								"vars",
							},
						},
						Target: &ast.Index{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1070),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1070),
										Column: int(13),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p11095,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1070),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(1070),
											Column: int(8),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
											Comment: []string{},
										},
									},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Id: "std",
							},
							LeftBracketFodder: ast.Fodder{},
							Index: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Value: "join",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							RightBracketFodder: ast.Fodder{},
							Id: nil,
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1070),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(1070),
													Column: int(18),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p11165,
											FreeVars: nil,
										},
										Value: "\n",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Binary{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1070),
													Column: int(20),
												},
												End: ast.Location{
													Line: int(1070),
													Column: int(31),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p11165,
											FreeVars: ast.Identifiers{
												"vars",
											},
										},
										Left: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1070),
														Column: int(20),
													},
													End: ast.Location{
														Line: int(1070),
														Column: int(24),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p11165,
												FreeVars: ast.Identifiers{
													"vars",
												},
											},
											Id: "vars",
										},
										OpFodder: ast.Fodder{},
										Op: ast.BinaryOp(3),
										Right: &ast.Array{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1070),
														Column: int(27),
													},
													End: ast.Location{
														Line: int(1070),
														Column: int(31),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p11165,
												FreeVars: nil,
											},
											Elements: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1070),
																	Column: int(28),
																},
																End: ast.Location{
																	Line: int(1070),
																	Column: int(30),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p11173,
															FreeVars: nil,
														},
														Value: "",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													CommaFodder: nil,
												},
											},
											TrailingComma: false,
											CloseFodder: ast.Fodder{},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(1068),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1070),
					Column: int(32),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "manifestXmlJsonml",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "value",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1072),
								Column: int(21),
							},
							End: ast.Location{
								Line: int(1072),
								Column: int(26),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1073),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1088),
								Column: int(17),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p11180,
						FreeVars: ast.Identifiers{
							"std",
							"value",
						},
					},
					Cond: &ast.Unary{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1073),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(1073),
									Column: int(27),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p11180,
							FreeVars: ast.Identifiers{
								"std",
								"value",
							},
						},
						Op: ast.UnaryOp(0),
						Expr: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1073),
										Column: int(9),
									},
									End: ast.Location{
										Line: int(1073),
										Column: int(27),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p11180,
								FreeVars: ast.Identifiers{
									"std",
									"value",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1073),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(1073),
											Column: int(20),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p11180,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1073),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(1073),
												Column: int(12),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "isArray",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1073),
														Column: int(21),
													},
													End: ast.Location{
														Line: int(1073),
														Column: int(26),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p11193,
												FreeVars: ast.Identifiers{
													"value",
												},
											},
											Id: "value",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
					},
					ThenFodder: ast.Fodder{},
					BranchTrue: &ast.Error{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1074),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(1074),
									Column: int(75),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
									Comment: []string{},
								},
							},
							Ctx: p11180,
							FreeVars: ast.Identifiers{
								"std",
								"value",
							},
						},
						Expr: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
									"value",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: nil,
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "mod",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: nil,
								Id: nil,
							},
							FodderLeft: nil,
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1074),
														Column: int(13),
													},
													End: ast.Location{
														Line: int(1074),
														Column: int(57),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p11180,
												FreeVars: nil,
											},
											Value: "Expected a JSONML value (an array), got %s",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										CommaFodder: nil,
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1074),
														Column: int(60),
													},
													End: ast.Location{
														Line: int(1074),
														Column: int(75),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p11180,
												FreeVars: ast.Identifiers{
													"std",
													"value",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1074),
															Column: int(60),
														},
														End: ast.Location{
															Line: int(1074),
															Column: int(68),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p11180,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1074),
																Column: int(60),
															},
															End: ast.Location{
																Line: int(1074),
																Column: int(63),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "type",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1074),
																		Column: int(69),
																	},
																	End: ast.Location{
																		Line: int(1074),
																		Column: int(74),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p11216,
																FreeVars: ast.Identifiers{
																	"value",
																},
															},
															Id: "value",
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: nil,
							TailStrictFodder: nil,
						},
					},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
							Comment: []string{},
						},
					},
					BranchFalse: &ast.Local{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1076),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(1088),
									Column: int(17),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
									Comment: []string{},
								},
							},
							Ctx: p11180,
							FreeVars: ast.Identifiers{
								"std",
								"value",
							},
						},
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: nil,
								Variable: "aux",
								EqFodder: nil,
								Body: &ast.Function{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1076),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(1086),
												Column: int(95),
											},
											File: p1,
										},
										Fodder: nil,
										Ctx: p11224,
										FreeVars: ast.Identifiers{
											"aux",
											"std",
										},
									},
									ParenLeftFodder: ast.Fodder{},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "v",
											EqFodder: nil,
											DefaultArg: nil,
											CommaFodder: nil,
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1076),
													Column: int(17),
												},
												End: ast.Location{
													Line: int(1076),
													Column: int(18),
												},
												File: p1,
											},
										},
									},
									TrailingComma: false,
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Conditional{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1077),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1086),
													Column: int(95),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
													Comment: []string{},
												},
											},
											Ctx: p11229,
											FreeVars: ast.Identifiers{
												"aux",
												"std",
												"v",
											},
										},
										Cond: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1077),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(1077),
														Column: int(27),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p11229,
												FreeVars: ast.Identifiers{
													"std",
													"v",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1077),
															Column: int(12),
														},
														End: ast.Location{
															Line: int(1077),
															Column: int(24),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p11229,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1077),
																Column: int(12),
															},
															End: ast.Location{
																Line: int(1077),
																Column: int(15),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "isString",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1077),
																		Column: int(25),
																	},
																	End: ast.Location{
																		Line: int(1077),
																		Column: int(26),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p11240,
																FreeVars: ast.Identifiers{
																	"v",
																},
															},
															Id: "v",
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
										ThenFodder: ast.Fodder{},
										BranchTrue: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1078),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(1078),
														Column: int(12),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(10),
														Comment: []string{},
													},
												},
												Ctx: p11229,
												FreeVars: ast.Identifiers{
													"v",
												},
											},
											Id: "v",
										},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(8),
												Comment: []string{},
											},
										},
										BranchFalse: &ast.Local{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1080),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(1086),
														Column: int(95),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(10),
														Comment: []string{},
													},
												},
												Ctx: p11229,
												FreeVars: ast.Identifiers{
													"aux",
													"std",
													"v",
												},
											},
											Binds: ast.LocalBinds{
												ast.LocalBind{
													VarFodder: ast.Fodder{},
													Variable: "tag",
													EqFodder: ast.Fodder{},
													Body: &ast.Index{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1080),
																	Column: int(23),
																},
																End: ast.Location{
																	Line: int(1080),
																	Column: int(27),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p11251,
															FreeVars: ast.Identifiers{
																"v",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1080),
																		Column: int(23),
																	},
																	End: ast.Location{
																		Line: int(1080),
																		Column: int(24),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p11251,
																FreeVars: ast.Identifiers{
																	"v",
																},
															},
															Id: "v",
														},
														LeftBracketFodder: ast.Fodder{},
														Index: &ast.LiteralNumber{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1080),
																		Column: int(25),
																	},
																	End: ast.Location{
																		Line: int(1080),
																		Column: int(26),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p11251,
																FreeVars: nil,
															},
															OriginalString: "0",
														},
														RightBracketFodder: ast.Fodder{},
														Id: nil,
													},
													Fun: nil,
													CloseFodder: ast.Fodder{},
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1080),
															Column: int(17),
														},
														End: ast.Location{
															Line: int(1080),
															Column: int(27),
														},
														File: p1,
													},
												},
											},
											Body: &ast.Local{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1081),
															Column: int(11),
														},
														End: ast.Location{
															Line: int(1086),
															Column: int(95),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(10),
															Comment: []string{},
														},
													},
													Ctx: p11229,
													FreeVars: ast.Identifiers{
														"aux",
														"std",
														"tag",
														"v",
													},
												},
												Binds: ast.LocalBinds{
													ast.LocalBind{
														VarFodder: ast.Fodder{},
														Variable: "has_attrs",
														EqFodder: ast.Fodder{},
														Body: &ast.Binary{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1081),
																		Column: int(29),
																	},
																	End: ast.Location{
																		Line: int(1081),
																		Column: int(68),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p11261,
																FreeVars: ast.Identifiers{
																	"std",
																	"v",
																},
															},
															Left: &ast.Binary{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1081),
																			Column: int(29),
																		},
																		End: ast.Location{
																			Line: int(1081),
																			Column: int(46),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p11261,
																	FreeVars: ast.Identifiers{
																		"std",
																		"v",
																	},
																},
																Left: &ast.Apply{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1081),
																				Column: int(29),
																			},
																			End: ast.Location{
																				Line: int(1081),
																				Column: int(42),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p11261,
																		FreeVars: ast.Identifiers{
																			"std",
																			"v",
																		},
																	},
																	Target: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1081),
																					Column: int(29),
																				},
																				End: ast.Location{
																					Line: int(1081),
																					Column: int(39),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p11261,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1081),
																						Column: int(29),
																					},
																					End: ast.Location{
																						Line: int(1081),
																						Column: int(32),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Id: "std",
																		},
																		LeftBracketFodder: ast.Fodder{},
																		Index: &ast.LiteralString{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: nil,
																			},
																			Value: "length",
																			Kind: ast.LiteralStringKind(1),
																			BlockIndent: "",
																			BlockTermIndent: "",
																		},
																		RightBracketFodder: ast.Fodder{},
																		Id: nil,
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1081),
																								Column: int(40),
																							},
																							End: ast.Location{
																								Line: int(1081),
																								Column: int(41),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p11274,
																						FreeVars: ast.Identifiers{
																							"v",
																						},
																					},
																					Id: "v",
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	TrailingComma: false,
																	TailStrict: false,
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																},
																OpFodder: ast.Fodder{},
																Op: ast.BinaryOp(7),
																Right: &ast.LiteralNumber{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1081),
																				Column: int(45),
																			},
																			End: ast.Location{
																				Line: int(1081),
																				Column: int(46),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p11261,
																		FreeVars: nil,
																	},
																	OriginalString: "1",
																},
															},
															OpFodder: ast.Fodder{},
															Op: ast.BinaryOp(17),
															Right: &ast.Apply{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1081),
																			Column: int(50),
																		},
																		End: ast.Location{
																			Line: int(1081),
																			Column: int(68),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p11261,
																	FreeVars: ast.Identifiers{
																		"std",
																		"v",
																	},
																},
																Target: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1081),
																				Column: int(50),
																			},
																			End: ast.Location{
																				Line: int(1081),
																				Column: int(62),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p11261,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1081),
																					Column: int(50),
																				},
																				End: ast.Location{
																					Line: int(1081),
																					Column: int(53),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Id: "std",
																	},
																	LeftBracketFodder: ast.Fodder{},
																	Index: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: nil,
																		},
																		Value: "isObject",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	RightBracketFodder: ast.Fodder{},
																	Id: nil,
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Index{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1081),
																							Column: int(63),
																						},
																						End: ast.Location{
																							Line: int(1081),
																							Column: int(67),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p11286,
																					FreeVars: ast.Identifiers{
																						"v",
																					},
																				},
																				Target: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1081),
																								Column: int(63),
																							},
																							End: ast.Location{
																								Line: int(1081),
																								Column: int(64),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p11286,
																						FreeVars: ast.Identifiers{
																							"v",
																						},
																					},
																					Id: "v",
																				},
																				LeftBracketFodder: ast.Fodder{},
																				Index: &ast.LiteralNumber{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1081),
																								Column: int(65),
																							},
																							End: ast.Location{
																								Line: int(1081),
																								Column: int(66),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p11286,
																						FreeVars: nil,
																					},
																					OriginalString: "1",
																				},
																				RightBracketFodder: ast.Fodder{},
																				Id: nil,
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																TrailingComma: false,
																TailStrict: false,
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
															},
														},
														Fun: nil,
														CloseFodder: ast.Fodder{},
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1081),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(1081),
																Column: int(68),
															},
															File: p1,
														},
													},
												},
												Body: &ast.Local{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1082),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(1086),
																Column: int(95),
															},
															File: p1,
														},
														Fodder: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(10),
																Comment: []string{},
															},
														},
														Ctx: p11229,
														FreeVars: ast.Identifiers{
															"aux",
															"has_attrs",
															"std",
															"tag",
															"v",
														},
													},
													Binds: ast.LocalBinds{
														ast.LocalBind{
															VarFodder: ast.Fodder{},
															Variable: "attrs",
															EqFodder: ast.Fodder{},
															Body: &ast.Conditional{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1082),
																			Column: int(25),
																		},
																		End: ast.Location{
																			Line: int(1082),
																			Column: int(55),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p11296,
																	FreeVars: ast.Identifiers{
																		"has_attrs",
																		"v",
																	},
																},
																Cond: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1082),
																				Column: int(28),
																			},
																			End: ast.Location{
																				Line: int(1082),
																				Column: int(37),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p11296,
																		FreeVars: ast.Identifiers{
																			"has_attrs",
																		},
																	},
																	Id: "has_attrs",
																},
																ThenFodder: ast.Fodder{},
																BranchTrue: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1082),
																				Column: int(43),
																			},
																			End: ast.Location{
																				Line: int(1082),
																				Column: int(47),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p11296,
																		FreeVars: ast.Identifiers{
																			"v",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1082),
																					Column: int(43),
																				},
																				End: ast.Location{
																					Line: int(1082),
																					Column: int(44),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p11296,
																			FreeVars: ast.Identifiers{
																				"v",
																			},
																		},
																		Id: "v",
																	},
																	LeftBracketFodder: ast.Fodder{},
																	Index: &ast.LiteralNumber{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1082),
																					Column: int(45),
																				},
																				End: ast.Location{
																					Line: int(1082),
																					Column: int(46),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p11296,
																			FreeVars: nil,
																		},
																		OriginalString: "1",
																	},
																	RightBracketFodder: ast.Fodder{},
																	Id: nil,
																},
																ElseFodder: ast.Fodder{},
																BranchFalse: &ast.DesugaredObject{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1082),
																				Column: int(53),
																			},
																			End: ast.Location{
																				Line: int(1082),
																				Column: int(55),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p11296,
																		FreeVars: nil,
																	},
																	Asserts: ast.Nodes{},
																	Fields: ast.DesugaredObjectFields{},
																	Locals: ast.LocalBinds{},
																},
															},
															Fun: nil,
															CloseFodder: ast.Fodder{},
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1082),
																	Column: int(17),
																},
																End: ast.Location{
																	Line: int(1082),
																	Column: int(55),
																},
																File: p1,
															},
														},
													},
													Body: &ast.Local{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1083),
																	Column: int(11),
																},
																End: ast.Location{
																	Line: int(1086),
																	Column: int(95),
																},
																File: p1,
															},
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(10),
																	Comment: []string{},
																},
															},
															Ctx: p11229,
															FreeVars: ast.Identifiers{
																"attrs",
																"aux",
																"has_attrs",
																"std",
																"tag",
																"v",
															},
														},
														Binds: ast.LocalBinds{
															ast.LocalBind{
																VarFodder: ast.Fodder{},
																Variable: "children",
																EqFodder: ast.Fodder{},
																Body: &ast.Conditional{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1083),
																				Column: int(28),
																			},
																			End: ast.Location{
																				Line: int(1083),
																				Column: int(62),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p11311,
																		FreeVars: ast.Identifiers{
																			"has_attrs",
																			"std",
																			"v",
																		},
																	},
																	Cond: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1083),
																					Column: int(31),
																				},
																				End: ast.Location{
																					Line: int(1083),
																					Column: int(40),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p11311,
																			FreeVars: ast.Identifiers{
																				"has_attrs",
																			},
																		},
																		Id: "has_attrs",
																	},
																	ThenFodder: ast.Fodder{},
																	BranchTrue: &ast.Apply{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																				"v",
																			},
																		},
																		Target: &ast.Index{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																				},
																				Id: "std",
																			},
																			LeftBracketFodder: nil,
																			Index: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: nil,
																				},
																				Value: "slice",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																			RightBracketFodder: nil,
																			Id: nil,
																		},
																		FodderLeft: nil,
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1083),
																									Column: int(46),
																								},
																								End: ast.Location{
																									Line: int(1083),
																									Column: int(47),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p11311,
																							FreeVars: ast.Identifiers{
																								"v",
																							},
																						},
																						Id: "v",
																					},
																					CommaFodder: nil,
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.LiteralNumber{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1083),
																									Column: int(48),
																								},
																								End: ast.Location{
																									Line: int(1083),
																									Column: int(49),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p11311,
																							FreeVars: nil,
																						},
																						OriginalString: "2",
																					},
																					CommaFodder: nil,
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.LiteralNull{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								File: nil,
																							},
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: nil,
																						},
																					},
																					CommaFodder: nil,
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.LiteralNull{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								File: nil,
																							},
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: nil,
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		TrailingComma: false,
																		TailStrict: false,
																		FodderRight: nil,
																		TailStrictFodder: nil,
																	},
																	ElseFodder: ast.Fodder{},
																	BranchFalse: &ast.Apply{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																				"v",
																			},
																		},
																		Target: &ast.Index{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																				},
																				Id: "std",
																			},
																			LeftBracketFodder: nil,
																			Index: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: nil,
																				},
																				Value: "slice",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																			RightBracketFodder: nil,
																			Id: nil,
																		},
																		FodderLeft: nil,
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1083),
																									Column: int(57),
																								},
																								End: ast.Location{
																									Line: int(1083),
																									Column: int(58),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p11311,
																							FreeVars: ast.Identifiers{
																								"v",
																							},
																						},
																						Id: "v",
																					},
																					CommaFodder: nil,
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.LiteralNumber{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1083),
																									Column: int(59),
																								},
																								End: ast.Location{
																									Line: int(1083),
																									Column: int(60),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p11311,
																							FreeVars: nil,
																						},
																						OriginalString: "1",
																					},
																					CommaFodder: nil,
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.LiteralNull{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								File: nil,
																							},
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: nil,
																						},
																					},
																					CommaFodder: nil,
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.LiteralNull{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								File: nil,
																							},
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: nil,
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		TrailingComma: false,
																		TailStrict: false,
																		FodderRight: nil,
																		TailStrictFodder: nil,
																	},
																},
																Fun: nil,
																CloseFodder: ast.Fodder{},
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1083),
																		Column: int(17),
																	},
																	End: ast.Location{
																		Line: int(1083),
																		Column: int(62),
																	},
																	File: p1,
																},
															},
														},
														Body: &ast.Local{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1084),
																		Column: int(11),
																	},
																	End: ast.Location{
																		Line: int(1086),
																		Column: int(95),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(10),
																		Comment: []string{},
																	},
																},
																Ctx: p11229,
																FreeVars: ast.Identifiers{
																	"attrs",
																	"aux",
																	"children",
																	"std",
																	"tag",
																},
															},
															Binds: ast.LocalBinds{
																ast.LocalBind{
																	VarFodder: ast.Fodder{},
																	Variable: "attrs_str",
																	EqFodder: ast.Fodder{},
																	Body: &ast.Apply{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1085),
																					Column: int(13),
																				},
																				End: ast.Location{
																					Line: int(1085),
																					Column: int(88),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p11346,
																			FreeVars: ast.Identifiers{
																				"attrs",
																				"std",
																			},
																		},
																		Target: &ast.Index{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1085),
																						Column: int(13),
																					},
																					End: ast.Location{
																						Line: int(1085),
																						Column: int(21),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p11346,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1085),
																							Column: int(13),
																						},
																						End: ast.Location{
																							Line: int(1085),
																							Column: int(16),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(12),
																							Comment: []string{},
																						},
																					},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																				},
																				Id: "std",
																			},
																			LeftBracketFodder: ast.Fodder{},
																			Index: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: nil,
																				},
																				Value: "join",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																			RightBracketFodder: ast.Fodder{},
																			Id: nil,
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.LiteralString{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1085),
																									Column: int(22),
																								},
																								End: ast.Location{
																									Line: int(1085),
																									Column: int(24),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p11356,
																							FreeVars: nil,
																						},
																						Value: "",
																						Kind: ast.LiteralStringKind(1),
																						BlockIndent: "",
																						BlockTermIndent: "",
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Apply{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								File: nil,
																							},
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: ast.Identifiers{
																								"attrs",
																								"std",
																							},
																						},
																						Target: &ast.Index{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									File: nil,
																								},
																								Fodder: nil,
																								Ctx: nil,
																								FreeVars: ast.Identifiers{
																									"std",
																								},
																							},
																							Target: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										File: nil,
																									},
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																								},
																								Id: "std",
																							},
																							LeftBracketFodder: nil,
																							Index: &ast.LiteralString{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										File: nil,
																									},
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: nil,
																								},
																								Value: "flatMap",
																								Kind: ast.LiteralStringKind(1),
																								BlockIndent: "",
																								BlockTermIndent: "",
																							},
																							RightBracketFodder: nil,
																							Id: nil,
																						},
																						FodderLeft: nil,
																						Arguments: ast.Arguments{
																							Positional: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Function{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												File: nil,
																											},
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"attrs",
																												"std",
																											},
																										},
																										ParenLeftFodder: nil,
																										Parameters: []ast.Parameter{
																											ast.Parameter{
																												NameFodder: nil,
																												Name: "k",
																												EqFodder: nil,
																												DefaultArg: nil,
																												CommaFodder: nil,
																												LocRange: ast.LocationRange{
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													File: nil,
																												},
																											},
																										},
																										TrailingComma: false,
																										ParenRightFodder: nil,
																										Body: &ast.Array{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													File: nil,
																												},
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"attrs",
																													"k",
																													"std",
																												},
																											},
																											Elements: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Apply{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																End: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																File: nil,
																															},
																															Fodder: nil,
																															Ctx: nil,
																															FreeVars: ast.Identifiers{
																																"attrs",
																																"k",
																																"std",
																															},
																														},
																														Target: &ast.Index{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "",
																																	Begin: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	End: ast.Location{
																																		Line: int(0),
																																		Column: int(0),
																																	},
																																	File: nil,
																																},
																																Fodder: nil,
																																Ctx: nil,
																																FreeVars: ast.Identifiers{
																																	"std",
																																},
																															},
																															Target: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		File: nil,
																																	},
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: ast.Identifiers{
																																		"std",
																																	},
																																},
																																Id: "std",
																															},
																															LeftBracketFodder: nil,
																															Index: &ast.LiteralString{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "",
																																		Begin: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		End: ast.Location{
																																			Line: int(0),
																																			Column: int(0),
																																		},
																																		File: nil,
																																	},
																																	Fodder: nil,
																																	Ctx: nil,
																																	FreeVars: nil,
																																},
																																Value: "mod",
																																Kind: ast.LiteralStringKind(1),
																																BlockIndent: "",
																																BlockTermIndent: "",
																															},
																															RightBracketFodder: nil,
																															Id: nil,
																														},
																														FodderLeft: nil,
																														Arguments: ast.Arguments{
																															Positional: []ast.CommaSeparatedExpr{
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.LiteralString{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(1085),
																																					Column: int(27),
																																				},
																																				End: ast.Location{
																																					Line: int(1085),
																																					Column: int(37),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p11380,
																																			FreeVars: nil,
																																		},
																																		Value: " %s=\"%s\"",
																																		Kind: ast.LiteralStringKind(1),
																																		BlockIndent: "",
																																		BlockTermIndent: "",
																																	},
																																	CommaFodder: nil,
																																},
																																ast.CommaSeparatedExpr{
																																	Expr: &ast.Array{
																																		NodeBase: ast.NodeBase{
																																			LocRange: ast.LocationRange{
																																				FileName: "<std>",
																																				Begin: ast.Location{
																																					Line: int(1085),
																																					Column: int(40),
																																				},
																																				End: ast.Location{
																																					Line: int(1085),
																																					Column: int(53),
																																				},
																																				File: p1,
																																			},
																																			Fodder: ast.Fodder{},
																																			Ctx: p11380,
																																			FreeVars: ast.Identifiers{
																																				"attrs",
																																				"k",
																																			},
																																		},
																																		Elements: []ast.CommaSeparatedExpr{
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Var{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(1085),
																																								Column: int(41),
																																							},
																																							End: ast.Location{
																																								Line: int(1085),
																																								Column: int(42),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p11385,
																																						FreeVars: ast.Identifiers{
																																							"k",
																																						},
																																					},
																																					Id: "k",
																																				},
																																				CommaFodder: ast.Fodder{},
																																			},
																																			ast.CommaSeparatedExpr{
																																				Expr: &ast.Index{
																																					NodeBase: ast.NodeBase{
																																						LocRange: ast.LocationRange{
																																							FileName: "<std>",
																																							Begin: ast.Location{
																																								Line: int(1085),
																																								Column: int(44),
																																							},
																																							End: ast.Location{
																																								Line: int(1085),
																																								Column: int(52),
																																							},
																																							File: p1,
																																						},
																																						Fodder: ast.Fodder{},
																																						Ctx: p11385,
																																						FreeVars: ast.Identifiers{
																																							"attrs",
																																							"k",
																																						},
																																					},
																																					Target: &ast.Var{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(1085),
																																									Column: int(44),
																																								},
																																								End: ast.Location{
																																									Line: int(1085),
																																									Column: int(49),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p11385,
																																							FreeVars: ast.Identifiers{
																																								"attrs",
																																							},
																																						},
																																						Id: "attrs",
																																					},
																																					LeftBracketFodder: ast.Fodder{},
																																					Index: &ast.Var{
																																						NodeBase: ast.NodeBase{
																																							LocRange: ast.LocationRange{
																																								FileName: "<std>",
																																								Begin: ast.Location{
																																									Line: int(1085),
																																									Column: int(50),
																																								},
																																								End: ast.Location{
																																									Line: int(1085),
																																									Column: int(51),
																																								},
																																								File: p1,
																																							},
																																							Fodder: ast.Fodder{},
																																							Ctx: p11385,
																																							FreeVars: ast.Identifiers{
																																								"k",
																																							},
																																						},
																																						Id: "k",
																																					},
																																					RightBracketFodder: ast.Fodder{},
																																					Id: nil,
																																				},
																																				CommaFodder: nil,
																																			},
																																		},
																																		TrailingComma: false,
																																		CloseFodder: ast.Fodder{},
																																	},
																																	CommaFodder: nil,
																																},
																															},
																															Named: nil,
																														},
																														TrailingComma: false,
																														TailStrict: false,
																														FodderRight: nil,
																														TailStrictFodder: nil,
																													},
																													CommaFodder: nil,
																												},
																											},
																											TrailingComma: false,
																											CloseFodder: nil,
																										},
																									},
																									CommaFodder: nil,
																								},
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Apply{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(1085),
																													Column: int(63),
																												},
																												End: ast.Location{
																													Line: int(1085),
																													Column: int(86),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p11356,
																											FreeVars: ast.Identifiers{
																												"attrs",
																												"std",
																											},
																										},
																										Target: &ast.Index{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1085),
																														Column: int(63),
																													},
																													End: ast.Location{
																														Line: int(1085),
																														Column: int(79),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p11356,
																												FreeVars: ast.Identifiers{
																													"std",
																												},
																											},
																											Target: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(1085),
																															Column: int(63),
																														},
																														End: ast.Location{
																															Line: int(1085),
																															Column: int(66),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																												},
																												Id: "std",
																											},
																											LeftBracketFodder: ast.Fodder{},
																											Index: &ast.LiteralString{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: nil,
																												},
																												Value: "objectFields",
																												Kind: ast.LiteralStringKind(1),
																												BlockIndent: "",
																												BlockTermIndent: "",
																											},
																											RightBracketFodder: ast.Fodder{},
																											Id: nil,
																										},
																										FodderLeft: ast.Fodder{},
																										Arguments: ast.Arguments{
																											Positional: []ast.CommaSeparatedExpr{
																												ast.CommaSeparatedExpr{
																													Expr: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(1085),
																																	Column: int(80),
																																},
																																End: ast.Location{
																																	Line: int(1085),
																																	Column: int(85),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: p11402,
																															FreeVars: ast.Identifiers{
																																"attrs",
																															},
																														},
																														Id: "attrs",
																													},
																													CommaFodder: nil,
																												},
																											},
																											Named: nil,
																										},
																										TrailingComma: false,
																										TailStrict: false,
																										FodderRight: ast.Fodder{},
																										TailStrictFodder: nil,
																									},
																									CommaFodder: nil,
																								},
																							},
																							Named: nil,
																						},
																						TrailingComma: false,
																						TailStrict: false,
																						FodderRight: nil,
																						TailStrictFodder: nil,
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		TrailingComma: false,
																		TailStrict: false,
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																	},
																	Fun: nil,
																	CloseFodder: ast.Fodder{},
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1084),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(1085),
																			Column: int(88),
																		},
																		File: p1,
																	},
																},
															},
															Body: &ast.Apply{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1086),
																			Column: int(11),
																		},
																		End: ast.Location{
																			Line: int(1086),
																			Column: int(95),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p11229,
																	FreeVars: ast.Identifiers{
																		"attrs_str",
																		"aux",
																		"children",
																		"std",
																		"tag",
																	},
																},
																Target: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1086),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(1086),
																				Column: int(23),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p11229,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1086),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(1086),
																					Column: int(14),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(10),
																					Comment: []string{},
																				},
																			},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Id: "std",
																	},
																	LeftBracketFodder: ast.Fodder{},
																	Index: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: nil,
																		},
																		Value: "deepJoin",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	RightBracketFodder: ast.Fodder{},
																	Id: nil,
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Array{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1086),
																							Column: int(24),
																						},
																						End: ast.Location{
																							Line: int(1086),
																							Column: int(94),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p11414,
																					FreeVars: ast.Identifiers{
																						"attrs_str",
																						"aux",
																						"children",
																						"std",
																						"tag",
																					},
																				},
																				Elements: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.LiteralString{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1086),
																										Column: int(25),
																									},
																									End: ast.Location{
																										Line: int(1086),
																										Column: int(28),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p11418,
																								FreeVars: nil,
																							},
																							Value: "<",
																							Kind: ast.LiteralStringKind(1),
																							BlockIndent: "",
																							BlockTermIndent: "",
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1086),
																										Column: int(30),
																									},
																									End: ast.Location{
																										Line: int(1086),
																										Column: int(33),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p11418,
																								FreeVars: ast.Identifiers{
																									"tag",
																								},
																							},
																							Id: "tag",
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1086),
																										Column: int(35),
																									},
																									End: ast.Location{
																										Line: int(1086),
																										Column: int(44),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p11418,
																								FreeVars: ast.Identifiers{
																									"attrs_str",
																								},
																							},
																							Id: "attrs_str",
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.LiteralString{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1086),
																										Column: int(46),
																									},
																									End: ast.Location{
																										Line: int(1086),
																										Column: int(49),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p11418,
																								FreeVars: nil,
																							},
																							Value: ">",
																							Kind: ast.LiteralStringKind(1),
																							BlockIndent: "",
																							BlockTermIndent: "",
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Apply{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "",
																									Begin: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									End: ast.Location{
																										Line: int(0),
																										Column: int(0),
																									},
																									File: nil,
																								},
																								Fodder: nil,
																								Ctx: nil,
																								FreeVars: ast.Identifiers{
																									"aux",
																									"children",
																									"std",
																								},
																							},
																							Target: &ast.Index{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										File: nil,
																									},
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																								},
																								Target: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											File: nil,
																										},
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: ast.Identifiers{
																											"std",
																										},
																									},
																									Id: "std",
																								},
																								LeftBracketFodder: nil,
																								Index: &ast.LiteralString{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "",
																											Begin: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											End: ast.Location{
																												Line: int(0),
																												Column: int(0),
																											},
																											File: nil,
																										},
																										Fodder: nil,
																										Ctx: nil,
																										FreeVars: nil,
																									},
																									Value: "flatMap",
																									Kind: ast.LiteralStringKind(1),
																									BlockIndent: "",
																									BlockTermIndent: "",
																								},
																								RightBracketFodder: nil,
																								Id: nil,
																							},
																							FodderLeft: nil,
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Function{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "",
																													Begin: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													End: ast.Location{
																														Line: int(0),
																														Column: int(0),
																													},
																													File: nil,
																												},
																												Fodder: nil,
																												Ctx: nil,
																												FreeVars: ast.Identifiers{
																													"aux",
																												},
																											},
																											ParenLeftFodder: nil,
																											Parameters: []ast.Parameter{
																												ast.Parameter{
																													NameFodder: nil,
																													Name: "x",
																													EqFodder: nil,
																													DefaultArg: nil,
																													CommaFodder: nil,
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																												},
																											},
																											TrailingComma: false,
																											ParenRightFodder: nil,
																											Body: &ast.Array{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "",
																														Begin: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														End: ast.Location{
																															Line: int(0),
																															Column: int(0),
																														},
																														File: nil,
																													},
																													Fodder: nil,
																													Ctx: nil,
																													FreeVars: ast.Identifiers{
																														"aux",
																														"x",
																													},
																												},
																												Elements: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Apply{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(1086),
																																		Column: int(52),
																																	},
																																	End: ast.Location{
																																		Line: int(1086),
																																		Column: int(58),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p11439,
																																FreeVars: ast.Identifiers{
																																	"aux",
																																	"x",
																																},
																															},
																															Target: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(1086),
																																			Column: int(52),
																																		},
																																		End: ast.Location{
																																			Line: int(1086),
																																			Column: int(55),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p11439,
																																	FreeVars: ast.Identifiers{
																																		"aux",
																																	},
																																},
																																Id: "aux",
																															},
																															FodderLeft: ast.Fodder{},
																															Arguments: ast.Arguments{
																																Positional: []ast.CommaSeparatedExpr{
																																	ast.CommaSeparatedExpr{
																																		Expr: &ast.Var{
																																			NodeBase: ast.NodeBase{
																																				LocRange: ast.LocationRange{
																																					FileName: "<std>",
																																					Begin: ast.Location{
																																						Line: int(1086),
																																						Column: int(56),
																																					},
																																					End: ast.Location{
																																						Line: int(1086),
																																						Column: int(57),
																																					},
																																					File: p1,
																																				},
																																				Fodder: ast.Fodder{},
																																				Ctx: p11445,
																																				FreeVars: ast.Identifiers{
																																					"x",
																																				},
																																			},
																																			Id: "x",
																																		},
																																		CommaFodder: nil,
																																	},
																																},
																																Named: nil,
																															},
																															TrailingComma: false,
																															TailStrict: false,
																															FodderRight: ast.Fodder{},
																															TailStrictFodder: nil,
																														},
																														CommaFodder: nil,
																													},
																												},
																												TrailingComma: false,
																												CloseFodder: nil,
																											},
																										},
																										CommaFodder: nil,
																									},
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1086),
																														Column: int(68),
																													},
																													End: ast.Location{
																														Line: int(1086),
																														Column: int(76),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p11418,
																												FreeVars: ast.Identifiers{
																													"children",
																												},
																											},
																											Id: "children",
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							TrailingComma: false,
																							TailStrict: false,
																							FodderRight: nil,
																							TailStrictFodder: nil,
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.LiteralString{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1086),
																										Column: int(79),
																									},
																									End: ast.Location{
																										Line: int(1086),
																										Column: int(83),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p11418,
																								FreeVars: nil,
																							},
																							Value: "</",
																							Kind: ast.LiteralStringKind(1),
																							BlockIndent: "",
																							BlockTermIndent: "",
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1086),
																										Column: int(85),
																									},
																									End: ast.Location{
																										Line: int(1086),
																										Column: int(88),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p11418,
																								FreeVars: ast.Identifiers{
																									"tag",
																								},
																							},
																							Id: "tag",
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.LiteralString{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1086),
																										Column: int(90),
																									},
																									End: ast.Location{
																										Line: int(1086),
																										Column: int(93),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p11418,
																								FreeVars: nil,
																							},
																							Value: ">",
																							Kind: ast.LiteralStringKind(1),
																							BlockIndent: "",
																							BlockTermIndent: "",
																						},
																						CommaFodder: nil,
																					},
																				},
																				TrailingComma: false,
																				CloseFodder: ast.Fodder{},
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																TrailingComma: false,
																TailStrict: false,
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
															},
														},
													},
												},
											},
										},
									},
								},
								Fun: nil,
								CloseFodder: nil,
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
							},
						},
						Body: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1088),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(1088),
										Column: int(17),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p11180,
								FreeVars: ast.Identifiers{
									"aux",
									"value",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1088),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(1088),
											Column: int(10),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(1),
											Indent: int(6),
											Comment: []string{},
										},
									},
									Ctx: p11180,
									FreeVars: ast.Identifiers{
										"aux",
									},
								},
								Id: "aux",
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1088),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(1088),
														Column: int(16),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p11460,
												FreeVars: ast.Identifiers{
													"value",
												},
											},
											Id: "value",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(1072),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1088),
					Column: int(17),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "base64",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"base64_table",
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "input",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1093),
								Column: int(10),
							},
							End: ast.Location{
								Line: int(1093),
								Column: int(15),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1094),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1137),
								Column: int(24),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p11468,
						FreeVars: ast.Identifiers{
							"base64_table",
							"input",
							"std",
						},
					},
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Variable: "bytes",
							EqFodder: ast.Fodder{},
							Body: &ast.Conditional{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1095),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(1098),
											Column: int(14),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
											Comment: []string{},
										},
									},
									Ctx: p11473,
									FreeVars: ast.Identifiers{
										"input",
										"std",
									},
								},
								Cond: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1095),
												Column: int(10),
											},
											End: ast.Location{
												Line: int(1095),
												Column: int(29),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p11473,
										FreeVars: ast.Identifiers{
											"input",
											"std",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1095),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(1095),
													Column: int(22),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p11473,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1095),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(1095),
														Column: int(13),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "isString",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1095),
																Column: int(23),
															},
															End: ast.Location{
																Line: int(1095),
																Column: int(28),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p11484,
														FreeVars: ast.Identifiers{
															"input",
														},
													},
													Id: "input",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
								ThenFodder: ast.Fodder{},
								BranchTrue: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1096),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(1096),
												Column: int(53),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p11473,
										FreeVars: ast.Identifiers{
											"input",
											"std",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1096),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1096),
													Column: int(16),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p11473,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1096),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1096),
														Column: int(12),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
														Comment: []string{},
													},
												},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "map",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Function{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1096),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(1096),
																Column: int(45),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p11496,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													ParenLeftFodder: ast.Fodder{},
													Parameters: []ast.Parameter{
														ast.Parameter{
															NameFodder: ast.Fodder{},
															Name: "c",
															EqFodder: nil,
															DefaultArg: nil,
															CommaFodder: nil,
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1096),
																	Column: int(26),
																},
																End: ast.Location{
																	Line: int(1096),
																	Column: int(27),
																},
																File: p1,
															},
														},
													},
													TrailingComma: false,
													ParenRightFodder: ast.Fodder{},
													Body: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1096),
																	Column: int(29),
																},
																End: ast.Location{
																	Line: int(1096),
																	Column: int(45),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p11500,
															FreeVars: ast.Identifiers{
																"c",
																"std",
															},
														},
														Target: &ast.Index{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1096),
																		Column: int(29),
																	},
																	End: ast.Location{
																		Line: int(1096),
																		Column: int(42),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p11500,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Target: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1096),
																			Column: int(29),
																		},
																		End: ast.Location{
																			Line: int(1096),
																			Column: int(32),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Id: "std",
															},
															LeftBracketFodder: ast.Fodder{},
															Index: &ast.LiteralString{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: nil,
																},
																Value: "codepoint",
																Kind: ast.LiteralStringKind(1),
																BlockIndent: "",
																BlockTermIndent: "",
															},
															RightBracketFodder: ast.Fodder{},
															Id: nil,
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1096),
																					Column: int(43),
																				},
																				End: ast.Location{
																					Line: int(1096),
																					Column: int(44),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p11509,
																			FreeVars: ast.Identifiers{
																				"c",
																			},
																		},
																		Id: "c",
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														TrailingComma: false,
														TailStrict: false,
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
													},
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1096),
																Column: int(47),
															},
															End: ast.Location{
																Line: int(1096),
																Column: int(52),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p11496,
														FreeVars: ast.Identifiers{
															"input",
														},
													},
													Id: "input",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
								ElseFodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
										Comment: []string{},
									},
								},
								BranchFalse: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1098),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(1098),
												Column: int(14),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(8),
												Comment: []string{},
											},
										},
										Ctx: p11473,
										FreeVars: ast.Identifiers{
											"input",
										},
									},
									Id: "input",
								},
							},
							Fun: nil,
							CloseFodder: ast.Fodder{},
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1094),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(1098),
									Column: int(14),
								},
								File: p1,
							},
						},
					},
					Body: &ast.Local{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1100),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1137),
									Column: int(24),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(1),
									Indent: int(4),
									Comment: []string{},
								},
							},
							Ctx: p11468,
							FreeVars: ast.Identifiers{
								"base64_table",
								"bytes",
								"std",
							},
						},
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: nil,
								Variable: "aux",
								EqFodder: nil,
								Body: &ast.Function{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1100),
												Column: int(11),
											},
											End: ast.Location{
												Line: int(1131),
												Column: int(33),
											},
											File: p1,
										},
										Fodder: nil,
										Ctx: p11522,
										FreeVars: ast.Identifiers{
											"aux",
											"base64_table",
											"std",
										},
									},
									ParenLeftFodder: ast.Fodder{},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "arr",
											EqFodder: nil,
											DefaultArg: nil,
											CommaFodder: ast.Fodder{},
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1100),
													Column: int(15),
												},
												End: ast.Location{
													Line: int(1100),
													Column: int(18),
												},
												File: p1,
											},
										},
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "i",
											EqFodder: nil,
											DefaultArg: nil,
											CommaFodder: ast.Fodder{},
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1100),
													Column: int(20),
												},
												End: ast.Location{
													Line: int(1100),
													Column: int(21),
												},
												File: p1,
											},
										},
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "r",
											EqFodder: nil,
											DefaultArg: nil,
											CommaFodder: nil,
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1100),
													Column: int(23),
												},
												End: ast.Location{
													Line: int(1100),
													Column: int(24),
												},
												File: p1,
											},
										},
									},
									TrailingComma: false,
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Conditional{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1101),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(1131),
													Column: int(33),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
													Comment: []string{},
												},
											},
											Ctx: p11527,
											FreeVars: ast.Identifiers{
												"arr",
												"aux",
												"base64_table",
												"i",
												"r",
												"std",
											},
										},
										Cond: &ast.Binary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1101),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(1101),
														Column: int(30),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p11527,
												FreeVars: ast.Identifiers{
													"arr",
													"i",
													"std",
												},
											},
											Left: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1101),
															Column: int(10),
														},
														End: ast.Location{
															Line: int(1101),
															Column: int(11),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p11527,
													FreeVars: ast.Identifiers{
														"i",
													},
												},
												Id: "i",
											},
											OpFodder: ast.Fodder{},
											Op: ast.BinaryOp(8),
											Right: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1101),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1101),
															Column: int(30),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p11527,
													FreeVars: ast.Identifiers{
														"arr",
														"std",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1101),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1101),
																Column: int(25),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p11527,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1101),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(1101),
																	Column: int(18),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: ast.Fodder{},
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "length",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: ast.Fodder{},
													Id: nil,
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1101),
																			Column: int(26),
																		},
																		End: ast.Location{
																			Line: int(1101),
																			Column: int(29),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p11542,
																	FreeVars: ast.Identifiers{
																		"arr",
																	},
																},
																Id: "arr",
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
										},
										ThenFodder: ast.Fodder{},
										BranchTrue: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1102),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1102),
														Column: int(10),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
														Comment: []string{},
													},
												},
												Ctx: p11527,
												FreeVars: ast.Identifiers{
													"r",
												},
											},
											Id: "r",
										},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										BranchFalse: &ast.Conditional{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1103),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(1131),
														Column: int(33),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p11527,
												FreeVars: ast.Identifiers{
													"arr",
													"aux",
													"base64_table",
													"i",
													"r",
													"std",
												},
											},
											Cond: &ast.Binary{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1103),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1103),
															Column: int(39),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p11527,
													FreeVars: ast.Identifiers{
														"arr",
														"i",
														"std",
													},
												},
												Left: &ast.Binary{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1103),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1103),
																Column: int(20),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p11527,
														FreeVars: ast.Identifiers{
															"i",
														},
													},
													Left: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1103),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(1103),
																	Column: int(16),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p11527,
															FreeVars: ast.Identifiers{
																"i",
															},
														},
														Id: "i",
													},
													OpFodder: ast.Fodder{},
													Op: ast.BinaryOp(3),
													Right: &ast.LiteralNumber{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1103),
																	Column: int(19),
																},
																End: ast.Location{
																	Line: int(1103),
																	Column: int(20),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p11527,
															FreeVars: nil,
														},
														OriginalString: "1",
													},
												},
												OpFodder: ast.Fodder{},
												Op: ast.BinaryOp(8),
												Right: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1103),
																Column: int(24),
															},
															End: ast.Location{
																Line: int(1103),
																Column: int(39),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p11527,
														FreeVars: ast.Identifiers{
															"arr",
															"std",
														},
													},
													Target: &ast.Index{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1103),
																	Column: int(24),
																},
																End: ast.Location{
																	Line: int(1103),
																	Column: int(34),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p11527,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1103),
																		Column: int(24),
																	},
																	End: ast.Location{
																		Line: int(1103),
																		Column: int(27),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Id: "std",
														},
														LeftBracketFodder: ast.Fodder{},
														Index: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
															Value: "length",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														RightBracketFodder: ast.Fodder{},
														Id: nil,
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1103),
																				Column: int(35),
																			},
																			End: ast.Location{
																				Line: int(1103),
																				Column: int(38),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p11566,
																		FreeVars: ast.Identifiers{
																			"arr",
																		},
																	},
																	Id: "arr",
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
												},
											},
											ThenFodder: ast.Fodder{},
											BranchTrue: &ast.Local{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1104),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1110),
															Column: int(33),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
															Comment: []string{},
														},
													},
													Ctx: p11527,
													FreeVars: ast.Identifiers{
														"arr",
														"aux",
														"base64_table",
														"i",
														"r",
													},
												},
												Binds: ast.LocalBinds{
													ast.LocalBind{
														VarFodder: ast.Fodder{},
														Variable: "str",
														EqFodder: ast.Fodder{},
														Body: &ast.Binary{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1106),
																		Column: int(11),
																	},
																	End: ast.Location{
																		Line: int(1109),
																		Column: int(15),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p11573,
																FreeVars: ast.Identifiers{
																	"arr",
																	"base64_table",
																	"i",
																},
															},
															Left: &ast.Binary{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1106),
																			Column: int(11),
																		},
																		End: ast.Location{
																			Line: int(1108),
																			Column: int(42),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p11573,
																	FreeVars: ast.Identifiers{
																		"arr",
																		"base64_table",
																		"i",
																	},
																},
																Left: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1106),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(1106),
																				Column: int(44),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p11573,
																		FreeVars: ast.Identifiers{
																			"arr",
																			"base64_table",
																			"i",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1106),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(1106),
																					Column: int(23),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(10),
																					Comment: []string{},
																				},
																				ast.FodderElement{
																					Kind: ast.FodderKind(2),
																					Blanks: int(0),
																					Indent: int(10),
																					Comment: []string{
																						"// 6 MSB of i",
																					},
																				},
																			},
																			Ctx: p11573,
																			FreeVars: ast.Identifiers{
																				"base64_table",
																			},
																		},
																		Id: "base64_table",
																	},
																	LeftBracketFodder: ast.Fodder{},
																	Index: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1106),
																					Column: int(24),
																				},
																				End: ast.Location{
																					Line: int(1106),
																					Column: int(43),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p11573,
																			FreeVars: ast.Identifiers{
																				"arr",
																				"i",
																			},
																		},
																		Left: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1106),
																						Column: int(25),
																					},
																					End: ast.Location{
																						Line: int(1106),
																						Column: int(37),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p11573,
																				FreeVars: ast.Identifiers{
																					"arr",
																					"i",
																				},
																			},
																			Left: &ast.Index{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1106),
																							Column: int(25),
																						},
																						End: ast.Location{
																							Line: int(1106),
																							Column: int(31),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p11573,
																					FreeVars: ast.Identifiers{
																						"arr",
																						"i",
																					},
																				},
																				Target: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1106),
																								Column: int(25),
																							},
																							End: ast.Location{
																								Line: int(1106),
																								Column: int(28),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p11573,
																						FreeVars: ast.Identifiers{
																							"arr",
																						},
																					},
																					Id: "arr",
																				},
																				LeftBracketFodder: ast.Fodder{},
																				Index: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1106),
																								Column: int(29),
																							},
																							End: ast.Location{
																								Line: int(1106),
																								Column: int(30),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p11573,
																						FreeVars: ast.Identifiers{
																							"i",
																						},
																					},
																					Id: "i",
																				},
																				RightBracketFodder: ast.Fodder{},
																				Id: nil,
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(14),
																			Right: &ast.LiteralNumber{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1106),
																							Column: int(34),
																						},
																						End: ast.Location{
																							Line: int(1106),
																							Column: int(37),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p11573,
																					FreeVars: nil,
																				},
																				OriginalString: "252",
																			},
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(6),
																		Right: &ast.LiteralNumber{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1106),
																						Column: int(42),
																					},
																					End: ast.Location{
																						Line: int(1106),
																						Column: int(43),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p11573,
																				FreeVars: nil,
																			},
																			OriginalString: "2",
																		},
																	},
																	RightBracketFodder: ast.Fodder{},
																	Id: nil,
																},
																OpFodder: ast.Fodder{},
																Op: ast.BinaryOp(3),
																Right: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1108),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(1108),
																				Column: int(42),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p11573,
																		FreeVars: ast.Identifiers{
																			"arr",
																			"base64_table",
																			"i",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1108),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(1108),
																					Column: int(23),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(10),
																					Comment: []string{},
																				},
																				ast.FodderElement{
																					Kind: ast.FodderKind(2),
																					Blanks: int(0),
																					Indent: int(10),
																					Comment: []string{
																						"// 2 LSB of i",
																					},
																				},
																			},
																			Ctx: p11573,
																			FreeVars: ast.Identifiers{
																				"base64_table",
																			},
																		},
																		Id: "base64_table",
																	},
																	LeftBracketFodder: ast.Fodder{},
																	Index: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1108),
																					Column: int(24),
																				},
																				End: ast.Location{
																					Line: int(1108),
																					Column: int(41),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p11573,
																			FreeVars: ast.Identifiers{
																				"arr",
																				"i",
																			},
																		},
																		Left: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1108),
																						Column: int(25),
																					},
																					End: ast.Location{
																						Line: int(1108),
																						Column: int(35),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p11573,
																				FreeVars: ast.Identifiers{
																					"arr",
																					"i",
																				},
																			},
																			Left: &ast.Index{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1108),
																							Column: int(25),
																						},
																						End: ast.Location{
																							Line: int(1108),
																							Column: int(31),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p11573,
																					FreeVars: ast.Identifiers{
																						"arr",
																						"i",
																					},
																				},
																				Target: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1108),
																								Column: int(25),
																							},
																							End: ast.Location{
																								Line: int(1108),
																								Column: int(28),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p11573,
																						FreeVars: ast.Identifiers{
																							"arr",
																						},
																					},
																					Id: "arr",
																				},
																				LeftBracketFodder: ast.Fodder{},
																				Index: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1108),
																								Column: int(29),
																							},
																							End: ast.Location{
																								Line: int(1108),
																								Column: int(30),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p11573,
																						FreeVars: ast.Identifiers{
																							"i",
																						},
																					},
																					Id: "i",
																				},
																				RightBracketFodder: ast.Fodder{},
																				Id: nil,
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(14),
																			Right: &ast.LiteralNumber{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1108),
																							Column: int(34),
																						},
																						End: ast.Location{
																							Line: int(1108),
																							Column: int(35),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p11573,
																					FreeVars: nil,
																				},
																				OriginalString: "3",
																			},
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(5),
																		Right: &ast.LiteralNumber{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1108),
																						Column: int(40),
																					},
																					End: ast.Location{
																						Line: int(1108),
																						Column: int(41),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p11573,
																				FreeVars: nil,
																			},
																			OriginalString: "4",
																		},
																	},
																	RightBracketFodder: ast.Fodder{},
																	Id: nil,
																},
															},
															OpFodder: ast.Fodder{},
															Op: ast.BinaryOp(3),
															Right: &ast.LiteralString{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1109),
																			Column: int(11),
																		},
																		End: ast.Location{
																			Line: int(1109),
																			Column: int(15),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(10),
																			Comment: []string{},
																		},
																	},
																	Ctx: p11573,
																	FreeVars: nil,
																},
																Value: "==",
																Kind: ast.LiteralStringKind(1),
																BlockIndent: "",
																BlockTermIndent: "",
															},
														},
														Fun: nil,
														CloseFodder: ast.Fodder{},
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1104),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1109),
																Column: int(15),
															},
															File: p1,
														},
													},
												},
												Body: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1110),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(1110),
																Column: int(33),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p11527,
														FreeVars: ast.Identifiers{
															"arr",
															"aux",
															"i",
															"r",
															"str",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1110),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(1110),
																	Column: int(12),
																},
																File: p1,
															},
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																	Comment: []string{},
																},
															},
															Ctx: p11527,
															FreeVars: ast.Identifiers{
																"aux",
															},
														},
														Id: "aux",
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1110),
																				Column: int(13),
																			},
																			End: ast.Location{
																				Line: int(1110),
																				Column: int(16),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p11622,
																		FreeVars: ast.Identifiers{
																			"arr",
																		},
																	},
																	Id: "arr",
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1110),
																				Column: int(18),
																			},
																			End: ast.Location{
																				Line: int(1110),
																				Column: int(23),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p11622,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																	},
																	Left: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1110),
																					Column: int(18),
																				},
																				End: ast.Location{
																					Line: int(1110),
																					Column: int(19),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p11622,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																		},
																		Id: "i",
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(3),
																	Right: &ast.LiteralNumber{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1110),
																					Column: int(22),
																				},
																				End: ast.Location{
																					Line: int(1110),
																					Column: int(23),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p11622,
																			FreeVars: nil,
																		},
																		OriginalString: "3",
																	},
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1110),
																				Column: int(25),
																			},
																			End: ast.Location{
																				Line: int(1110),
																				Column: int(32),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p11622,
																		FreeVars: ast.Identifiers{
																			"r",
																			"str",
																		},
																	},
																	Left: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1110),
																					Column: int(25),
																				},
																				End: ast.Location{
																					Line: int(1110),
																					Column: int(26),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p11622,
																			FreeVars: ast.Identifiers{
																				"r",
																			},
																		},
																		Id: "r",
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(3),
																	Right: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1110),
																					Column: int(29),
																				},
																				End: ast.Location{
																					Line: int(1110),
																					Column: int(32),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p11622,
																			FreeVars: ast.Identifiers{
																				"str",
																			},
																		},
																		Id: "str",
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: true,
													FodderRight: ast.Fodder{},
													TailStrictFodder: ast.Fodder{},
												},
											},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
													Comment: []string{},
												},
											},
											BranchFalse: &ast.Conditional{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1111),
															Column: int(12),
														},
														End: ast.Location{
															Line: int(1131),
															Column: int(33),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p11527,
													FreeVars: ast.Identifiers{
														"arr",
														"aux",
														"base64_table",
														"i",
														"r",
														"std",
													},
												},
												Cond: &ast.Binary{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1111),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1111),
																Column: int(39),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p11527,
														FreeVars: ast.Identifiers{
															"arr",
															"i",
															"std",
														},
													},
													Left: &ast.Binary{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1111),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(1111),
																	Column: int(20),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p11527,
															FreeVars: ast.Identifiers{
																"i",
															},
														},
														Left: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1111),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(1111),
																		Column: int(16),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p11527,
																FreeVars: ast.Identifiers{
																	"i",
																},
															},
															Id: "i",
														},
														OpFodder: ast.Fodder{},
														Op: ast.BinaryOp(3),
														Right: &ast.LiteralNumber{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1111),
																		Column: int(19),
																	},
																	End: ast.Location{
																		Line: int(1111),
																		Column: int(20),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p11527,
																FreeVars: nil,
															},
															OriginalString: "2",
														},
													},
													OpFodder: ast.Fodder{},
													Op: ast.BinaryOp(8),
													Right: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1111),
																	Column: int(24),
																},
																End: ast.Location{
																	Line: int(1111),
																	Column: int(39),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p11527,
															FreeVars: ast.Identifiers{
																"arr",
																"std",
															},
														},
														Target: &ast.Index{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1111),
																		Column: int(24),
																	},
																	End: ast.Location{
																		Line: int(1111),
																		Column: int(34),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p11527,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Target: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1111),
																			Column: int(24),
																		},
																		End: ast.Location{
																			Line: int(1111),
																			Column: int(27),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Id: "std",
															},
															LeftBracketFodder: ast.Fodder{},
															Index: &ast.LiteralString{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: nil,
																},
																Value: "length",
																Kind: ast.LiteralStringKind(1),
																BlockIndent: "",
																BlockTermIndent: "",
															},
															RightBracketFodder: ast.Fodder{},
															Id: nil,
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1111),
																					Column: int(35),
																				},
																				End: ast.Location{
																					Line: int(1111),
																					Column: int(38),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p11654,
																			FreeVars: ast.Identifiers{
																				"arr",
																			},
																		},
																		Id: "arr",
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														TrailingComma: false,
														TailStrict: false,
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
													},
												},
												ThenFodder: ast.Fodder{},
												BranchTrue: &ast.Local{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1112),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(1120),
																Column: int(33),
															},
															File: p1,
														},
														Fodder: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(8),
																Comment: []string{},
															},
														},
														Ctx: p11527,
														FreeVars: ast.Identifiers{
															"arr",
															"aux",
															"base64_table",
															"i",
															"r",
														},
													},
													Binds: ast.LocalBinds{
														ast.LocalBind{
															VarFodder: ast.Fodder{},
															Variable: "str",
															EqFodder: ast.Fodder{},
															Body: &ast.Binary{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1114),
																			Column: int(11),
																		},
																		End: ast.Location{
																			Line: int(1119),
																			Column: int(14),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p11661,
																	FreeVars: ast.Identifiers{
																		"arr",
																		"base64_table",
																		"i",
																	},
																},
																Left: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1114),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(1118),
																				Column: int(47),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p11661,
																		FreeVars: ast.Identifiers{
																			"arr",
																			"base64_table",
																			"i",
																		},
																	},
																	Left: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1114),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(1116),
																					Column: int(68),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p11661,
																			FreeVars: ast.Identifiers{
																				"arr",
																				"base64_table",
																				"i",
																			},
																		},
																		Left: &ast.Index{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1114),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(1114),
																						Column: int(44),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p11661,
																				FreeVars: ast.Identifiers{
																					"arr",
																					"base64_table",
																					"i",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1114),
																							Column: int(11),
																						},
																						End: ast.Location{
																							Line: int(1114),
																							Column: int(23),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(10),
																							Comment: []string{},
																						},
																						ast.FodderElement{
																							Kind: ast.FodderKind(2),
																							Blanks: int(0),
																							Indent: int(10),
																							Comment: []string{
																								"// 6 MSB of i",
																							},
																						},
																					},
																					Ctx: p11661,
																					FreeVars: ast.Identifiers{
																						"base64_table",
																					},
																				},
																				Id: "base64_table",
																			},
																			LeftBracketFodder: ast.Fodder{},
																			Index: &ast.Binary{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1114),
																							Column: int(24),
																						},
																						End: ast.Location{
																							Line: int(1114),
																							Column: int(43),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p11661,
																					FreeVars: ast.Identifiers{
																						"arr",
																						"i",
																					},
																				},
																				Left: &ast.Binary{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1114),
																								Column: int(25),
																							},
																							End: ast.Location{
																								Line: int(1114),
																								Column: int(37),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p11661,
																						FreeVars: ast.Identifiers{
																							"arr",
																							"i",
																						},
																					},
																					Left: &ast.Index{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1114),
																									Column: int(25),
																								},
																								End: ast.Location{
																									Line: int(1114),
																									Column: int(31),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p11661,
																							FreeVars: ast.Identifiers{
																								"arr",
																								"i",
																							},
																						},
																						Target: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1114),
																										Column: int(25),
																									},
																									End: ast.Location{
																										Line: int(1114),
																										Column: int(28),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p11661,
																								FreeVars: ast.Identifiers{
																									"arr",
																								},
																							},
																							Id: "arr",
																						},
																						LeftBracketFodder: ast.Fodder{},
																						Index: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1114),
																										Column: int(29),
																									},
																									End: ast.Location{
																										Line: int(1114),
																										Column: int(30),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p11661,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																							},
																							Id: "i",
																						},
																						RightBracketFodder: ast.Fodder{},
																						Id: nil,
																					},
																					OpFodder: ast.Fodder{},
																					Op: ast.BinaryOp(14),
																					Right: &ast.LiteralNumber{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1114),
																									Column: int(34),
																								},
																								End: ast.Location{
																									Line: int(1114),
																									Column: int(37),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p11661,
																							FreeVars: nil,
																						},
																						OriginalString: "252",
																					},
																				},
																				OpFodder: ast.Fodder{},
																				Op: ast.BinaryOp(6),
																				Right: &ast.LiteralNumber{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1114),
																								Column: int(42),
																							},
																							End: ast.Location{
																								Line: int(1114),
																								Column: int(43),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p11661,
																						FreeVars: nil,
																					},
																					OriginalString: "2",
																				},
																			},
																			RightBracketFodder: ast.Fodder{},
																			Id: nil,
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(3),
																		Right: &ast.Index{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1116),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(1116),
																						Column: int(68),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p11661,
																				FreeVars: ast.Identifiers{
																					"arr",
																					"base64_table",
																					"i",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1116),
																							Column: int(11),
																						},
																						End: ast.Location{
																							Line: int(1116),
																							Column: int(23),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(10),
																							Comment: []string{},
																						},
																						ast.FodderElement{
																							Kind: ast.FodderKind(2),
																							Blanks: int(0),
																							Indent: int(10),
																							Comment: []string{
																								"// 2 LSB of i, 4 MSB of i+1",
																							},
																						},
																					},
																					Ctx: p11661,
																					FreeVars: ast.Identifiers{
																						"base64_table",
																					},
																				},
																				Id: "base64_table",
																			},
																			LeftBracketFodder: ast.Fodder{},
																			Index: &ast.Binary{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1116),
																							Column: int(24),
																						},
																						End: ast.Location{
																							Line: int(1116),
																							Column: int(67),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p11661,
																					FreeVars: ast.Identifiers{
																						"arr",
																						"i",
																					},
																				},
																				Left: &ast.Binary{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1116),
																								Column: int(24),
																							},
																							End: ast.Location{
																								Line: int(1116),
																								Column: int(41),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p11661,
																						FreeVars: ast.Identifiers{
																							"arr",
																							"i",
																						},
																					},
																					Left: &ast.Binary{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1116),
																									Column: int(25),
																								},
																								End: ast.Location{
																									Line: int(1116),
																									Column: int(35),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p11661,
																							FreeVars: ast.Identifiers{
																								"arr",
																								"i",
																							},
																						},
																						Left: &ast.Index{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1116),
																										Column: int(25),
																									},
																									End: ast.Location{
																										Line: int(1116),
																										Column: int(31),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p11661,
																								FreeVars: ast.Identifiers{
																									"arr",
																									"i",
																								},
																							},
																							Target: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1116),
																											Column: int(25),
																										},
																										End: ast.Location{
																											Line: int(1116),
																											Column: int(28),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p11661,
																									FreeVars: ast.Identifiers{
																										"arr",
																									},
																								},
																								Id: "arr",
																							},
																							LeftBracketFodder: ast.Fodder{},
																							Index: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1116),
																											Column: int(29),
																										},
																										End: ast.Location{
																											Line: int(1116),
																											Column: int(30),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p11661,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																								},
																								Id: "i",
																							},
																							RightBracketFodder: ast.Fodder{},
																							Id: nil,
																						},
																						OpFodder: ast.Fodder{},
																						Op: ast.BinaryOp(14),
																						Right: &ast.LiteralNumber{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1116),
																										Column: int(34),
																									},
																									End: ast.Location{
																										Line: int(1116),
																										Column: int(35),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p11661,
																								FreeVars: nil,
																							},
																							OriginalString: "3",
																						},
																					},
																					OpFodder: ast.Fodder{},
																					Op: ast.BinaryOp(5),
																					Right: &ast.LiteralNumber{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1116),
																									Column: int(40),
																								},
																								End: ast.Location{
																									Line: int(1116),
																									Column: int(41),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p11661,
																							FreeVars: nil,
																						},
																						OriginalString: "4",
																					},
																				},
																				OpFodder: ast.Fodder{},
																				Op: ast.BinaryOp(16),
																				Right: &ast.Binary{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1116),
																								Column: int(44),
																							},
																							End: ast.Location{
																								Line: int(1116),
																								Column: int(67),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p11661,
																						FreeVars: ast.Identifiers{
																							"arr",
																							"i",
																						},
																					},
																					Left: &ast.Binary{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1116),
																									Column: int(45),
																								},
																								End: ast.Location{
																									Line: int(1116),
																									Column: int(61),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p11661,
																							FreeVars: ast.Identifiers{
																								"arr",
																								"i",
																							},
																						},
																						Left: &ast.Index{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1116),
																										Column: int(45),
																									},
																									End: ast.Location{
																										Line: int(1116),
																										Column: int(55),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p11661,
																								FreeVars: ast.Identifiers{
																									"arr",
																									"i",
																								},
																							},
																							Target: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1116),
																											Column: int(45),
																										},
																										End: ast.Location{
																											Line: int(1116),
																											Column: int(48),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p11661,
																									FreeVars: ast.Identifiers{
																										"arr",
																									},
																								},
																								Id: "arr",
																							},
																							LeftBracketFodder: ast.Fodder{},
																							Index: &ast.Binary{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1116),
																											Column: int(49),
																										},
																										End: ast.Location{
																											Line: int(1116),
																											Column: int(54),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p11661,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																								},
																								Left: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1116),
																												Column: int(49),
																											},
																											End: ast.Location{
																												Line: int(1116),
																												Column: int(50),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p11661,
																										FreeVars: ast.Identifiers{
																											"i",
																										},
																									},
																									Id: "i",
																								},
																								OpFodder: ast.Fodder{},
																								Op: ast.BinaryOp(3),
																								Right: &ast.LiteralNumber{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1116),
																												Column: int(53),
																											},
																											End: ast.Location{
																												Line: int(1116),
																												Column: int(54),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p11661,
																										FreeVars: nil,
																									},
																									OriginalString: "1",
																								},
																							},
																							RightBracketFodder: ast.Fodder{},
																							Id: nil,
																						},
																						OpFodder: ast.Fodder{},
																						Op: ast.BinaryOp(14),
																						Right: &ast.LiteralNumber{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1116),
																										Column: int(58),
																									},
																									End: ast.Location{
																										Line: int(1116),
																										Column: int(61),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p11661,
																								FreeVars: nil,
																							},
																							OriginalString: "240",
																						},
																					},
																					OpFodder: ast.Fodder{},
																					Op: ast.BinaryOp(6),
																					Right: &ast.LiteralNumber{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1116),
																									Column: int(66),
																								},
																								End: ast.Location{
																									Line: int(1116),
																									Column: int(67),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p11661,
																							FreeVars: nil,
																						},
																						OriginalString: "4",
																					},
																				},
																			},
																			RightBracketFodder: ast.Fodder{},
																			Id: nil,
																		},
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(3),
																	Right: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1118),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(1118),
																					Column: int(47),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p11661,
																			FreeVars: ast.Identifiers{
																				"arr",
																				"base64_table",
																				"i",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1118),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(1118),
																						Column: int(23),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(10),
																						Comment: []string{},
																					},
																					ast.FodderElement{
																						Kind: ast.FodderKind(2),
																						Blanks: int(0),
																						Indent: int(10),
																						Comment: []string{
																							"// 4 LSB of i+1",
																						},
																					},
																				},
																				Ctx: p11661,
																				FreeVars: ast.Identifiers{
																					"base64_table",
																				},
																			},
																			Id: "base64_table",
																		},
																		LeftBracketFodder: ast.Fodder{},
																		Index: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1118),
																						Column: int(24),
																					},
																					End: ast.Location{
																						Line: int(1118),
																						Column: int(46),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p11661,
																				FreeVars: ast.Identifiers{
																					"arr",
																					"i",
																				},
																			},
																			Left: &ast.Binary{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1118),
																							Column: int(25),
																						},
																						End: ast.Location{
																							Line: int(1118),
																							Column: int(40),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p11661,
																					FreeVars: ast.Identifiers{
																						"arr",
																						"i",
																					},
																				},
																				Left: &ast.Index{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1118),
																								Column: int(25),
																							},
																							End: ast.Location{
																								Line: int(1118),
																								Column: int(35),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p11661,
																						FreeVars: ast.Identifiers{
																							"arr",
																							"i",
																						},
																					},
																					Target: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1118),
																									Column: int(25),
																								},
																								End: ast.Location{
																									Line: int(1118),
																									Column: int(28),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p11661,
																							FreeVars: ast.Identifiers{
																								"arr",
																							},
																						},
																						Id: "arr",
																					},
																					LeftBracketFodder: ast.Fodder{},
																					Index: &ast.Binary{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1118),
																									Column: int(29),
																								},
																								End: ast.Location{
																									Line: int(1118),
																									Column: int(34),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p11661,
																							FreeVars: ast.Identifiers{
																								"i",
																							},
																						},
																						Left: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1118),
																										Column: int(29),
																									},
																									End: ast.Location{
																										Line: int(1118),
																										Column: int(30),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p11661,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																							},
																							Id: "i",
																						},
																						OpFodder: ast.Fodder{},
																						Op: ast.BinaryOp(3),
																						Right: &ast.LiteralNumber{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1118),
																										Column: int(33),
																									},
																									End: ast.Location{
																										Line: int(1118),
																										Column: int(34),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p11661,
																								FreeVars: nil,
																							},
																							OriginalString: "1",
																						},
																					},
																					RightBracketFodder: ast.Fodder{},
																					Id: nil,
																				},
																				OpFodder: ast.Fodder{},
																				Op: ast.BinaryOp(14),
																				Right: &ast.LiteralNumber{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1118),
																								Column: int(38),
																							},
																							End: ast.Location{
																								Line: int(1118),
																								Column: int(40),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p11661,
																						FreeVars: nil,
																					},
																					OriginalString: "15",
																				},
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(5),
																			Right: &ast.LiteralNumber{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1118),
																							Column: int(45),
																						},
																						End: ast.Location{
																							Line: int(1118),
																							Column: int(46),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p11661,
																					FreeVars: nil,
																				},
																				OriginalString: "2",
																			},
																		},
																		RightBracketFodder: ast.Fodder{},
																		Id: nil,
																	},
																},
																OpFodder: ast.Fodder{},
																Op: ast.BinaryOp(3),
																Right: &ast.LiteralString{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1119),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(1119),
																				Column: int(14),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(10),
																				Comment: []string{},
																			},
																		},
																		Ctx: p11661,
																		FreeVars: nil,
																	},
																	Value: "=",
																	Kind: ast.LiteralStringKind(1),
																	BlockIndent: "",
																	BlockTermIndent: "",
																},
															},
															Fun: nil,
															CloseFodder: ast.Fodder{},
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1112),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(1119),
																	Column: int(14),
																},
																File: p1,
															},
														},
													},
													Body: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1120),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(1120),
																	Column: int(33),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p11527,
															FreeVars: ast.Identifiers{
																"arr",
																"aux",
																"i",
																"r",
																"str",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1120),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(1120),
																		Column: int(12),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																		Comment: []string{},
																	},
																},
																Ctx: p11527,
																FreeVars: ast.Identifiers{
																	"aux",
																},
															},
															Id: "aux",
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1120),
																					Column: int(13),
																				},
																				End: ast.Location{
																					Line: int(1120),
																					Column: int(16),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p11750,
																			FreeVars: ast.Identifiers{
																				"arr",
																			},
																		},
																		Id: "arr",
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1120),
																					Column: int(18),
																				},
																				End: ast.Location{
																					Line: int(1120),
																					Column: int(23),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p11750,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																		},
																		Left: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1120),
																						Column: int(18),
																					},
																					End: ast.Location{
																						Line: int(1120),
																						Column: int(19),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p11750,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																			},
																			Id: "i",
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(3),
																		Right: &ast.LiteralNumber{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1120),
																						Column: int(22),
																					},
																					End: ast.Location{
																						Line: int(1120),
																						Column: int(23),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p11750,
																				FreeVars: nil,
																			},
																			OriginalString: "3",
																		},
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1120),
																					Column: int(25),
																				},
																				End: ast.Location{
																					Line: int(1120),
																					Column: int(32),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p11750,
																			FreeVars: ast.Identifiers{
																				"r",
																				"str",
																			},
																		},
																		Left: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1120),
																						Column: int(25),
																					},
																					End: ast.Location{
																						Line: int(1120),
																						Column: int(26),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p11750,
																				FreeVars: ast.Identifiers{
																					"r",
																				},
																			},
																			Id: "r",
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(3),
																		Right: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1120),
																						Column: int(29),
																					},
																					End: ast.Location{
																						Line: int(1120),
																						Column: int(32),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p11750,
																				FreeVars: ast.Identifiers{
																					"str",
																				},
																			},
																			Id: "str",
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														TrailingComma: false,
														TailStrict: true,
														FodderRight: ast.Fodder{},
														TailStrictFodder: ast.Fodder{},
													},
												},
												ElseFodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
														Comment: []string{},
													},
												},
												BranchFalse: &ast.Local{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1122),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(1131),
																Column: int(33),
															},
															File: p1,
														},
														Fodder: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(8),
																Comment: []string{},
															},
														},
														Ctx: p11527,
														FreeVars: ast.Identifiers{
															"arr",
															"aux",
															"base64_table",
															"i",
															"r",
														},
													},
													Binds: ast.LocalBinds{
														ast.LocalBind{
															VarFodder: ast.Fodder{},
															Variable: "str",
															EqFodder: ast.Fodder{},
															Body: &ast.Binary{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1124),
																			Column: int(11),
																		},
																		End: ast.Location{
																			Line: int(1130),
																			Column: int(42),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p11769,
																	FreeVars: ast.Identifiers{
																		"arr",
																		"base64_table",
																		"i",
																	},
																},
																Left: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1124),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(1128),
																				Column: int(73),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p11769,
																		FreeVars: ast.Identifiers{
																			"arr",
																			"base64_table",
																			"i",
																		},
																	},
																	Left: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1124),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(1126),
																					Column: int(68),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p11769,
																			FreeVars: ast.Identifiers{
																				"arr",
																				"base64_table",
																				"i",
																			},
																		},
																		Left: &ast.Index{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1124),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(1124),
																						Column: int(44),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p11769,
																				FreeVars: ast.Identifiers{
																					"arr",
																					"base64_table",
																					"i",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1124),
																							Column: int(11),
																						},
																						End: ast.Location{
																							Line: int(1124),
																							Column: int(23),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(10),
																							Comment: []string{},
																						},
																						ast.FodderElement{
																							Kind: ast.FodderKind(2),
																							Blanks: int(0),
																							Indent: int(10),
																							Comment: []string{
																								"// 6 MSB of i",
																							},
																						},
																					},
																					Ctx: p11769,
																					FreeVars: ast.Identifiers{
																						"base64_table",
																					},
																				},
																				Id: "base64_table",
																			},
																			LeftBracketFodder: ast.Fodder{},
																			Index: &ast.Binary{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1124),
																							Column: int(24),
																						},
																						End: ast.Location{
																							Line: int(1124),
																							Column: int(43),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p11769,
																					FreeVars: ast.Identifiers{
																						"arr",
																						"i",
																					},
																				},
																				Left: &ast.Binary{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1124),
																								Column: int(25),
																							},
																							End: ast.Location{
																								Line: int(1124),
																								Column: int(37),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p11769,
																						FreeVars: ast.Identifiers{
																							"arr",
																							"i",
																						},
																					},
																					Left: &ast.Index{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1124),
																									Column: int(25),
																								},
																								End: ast.Location{
																									Line: int(1124),
																									Column: int(31),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p11769,
																							FreeVars: ast.Identifiers{
																								"arr",
																								"i",
																							},
																						},
																						Target: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1124),
																										Column: int(25),
																									},
																									End: ast.Location{
																										Line: int(1124),
																										Column: int(28),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p11769,
																								FreeVars: ast.Identifiers{
																									"arr",
																								},
																							},
																							Id: "arr",
																						},
																						LeftBracketFodder: ast.Fodder{},
																						Index: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1124),
																										Column: int(29),
																									},
																									End: ast.Location{
																										Line: int(1124),
																										Column: int(30),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p11769,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																							},
																							Id: "i",
																						},
																						RightBracketFodder: ast.Fodder{},
																						Id: nil,
																					},
																					OpFodder: ast.Fodder{},
																					Op: ast.BinaryOp(14),
																					Right: &ast.LiteralNumber{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1124),
																									Column: int(34),
																								},
																								End: ast.Location{
																									Line: int(1124),
																									Column: int(37),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p11769,
																							FreeVars: nil,
																						},
																						OriginalString: "252",
																					},
																				},
																				OpFodder: ast.Fodder{},
																				Op: ast.BinaryOp(6),
																				Right: &ast.LiteralNumber{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1124),
																								Column: int(42),
																							},
																							End: ast.Location{
																								Line: int(1124),
																								Column: int(43),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p11769,
																						FreeVars: nil,
																					},
																					OriginalString: "2",
																				},
																			},
																			RightBracketFodder: ast.Fodder{},
																			Id: nil,
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(3),
																		Right: &ast.Index{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1126),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(1126),
																						Column: int(68),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p11769,
																				FreeVars: ast.Identifiers{
																					"arr",
																					"base64_table",
																					"i",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1126),
																							Column: int(11),
																						},
																						End: ast.Location{
																							Line: int(1126),
																							Column: int(23),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(10),
																							Comment: []string{},
																						},
																						ast.FodderElement{
																							Kind: ast.FodderKind(2),
																							Blanks: int(0),
																							Indent: int(10),
																							Comment: []string{
																								"// 2 LSB of i, 4 MSB of i+1",
																							},
																						},
																					},
																					Ctx: p11769,
																					FreeVars: ast.Identifiers{
																						"base64_table",
																					},
																				},
																				Id: "base64_table",
																			},
																			LeftBracketFodder: ast.Fodder{},
																			Index: &ast.Binary{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1126),
																							Column: int(24),
																						},
																						End: ast.Location{
																							Line: int(1126),
																							Column: int(67),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p11769,
																					FreeVars: ast.Identifiers{
																						"arr",
																						"i",
																					},
																				},
																				Left: &ast.Binary{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1126),
																								Column: int(24),
																							},
																							End: ast.Location{
																								Line: int(1126),
																								Column: int(41),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p11769,
																						FreeVars: ast.Identifiers{
																							"arr",
																							"i",
																						},
																					},
																					Left: &ast.Binary{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1126),
																									Column: int(25),
																								},
																								End: ast.Location{
																									Line: int(1126),
																									Column: int(35),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p11769,
																							FreeVars: ast.Identifiers{
																								"arr",
																								"i",
																							},
																						},
																						Left: &ast.Index{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1126),
																										Column: int(25),
																									},
																									End: ast.Location{
																										Line: int(1126),
																										Column: int(31),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p11769,
																								FreeVars: ast.Identifiers{
																									"arr",
																									"i",
																								},
																							},
																							Target: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1126),
																											Column: int(25),
																										},
																										End: ast.Location{
																											Line: int(1126),
																											Column: int(28),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p11769,
																									FreeVars: ast.Identifiers{
																										"arr",
																									},
																								},
																								Id: "arr",
																							},
																							LeftBracketFodder: ast.Fodder{},
																							Index: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1126),
																											Column: int(29),
																										},
																										End: ast.Location{
																											Line: int(1126),
																											Column: int(30),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p11769,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																								},
																								Id: "i",
																							},
																							RightBracketFodder: ast.Fodder{},
																							Id: nil,
																						},
																						OpFodder: ast.Fodder{},
																						Op: ast.BinaryOp(14),
																						Right: &ast.LiteralNumber{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1126),
																										Column: int(34),
																									},
																									End: ast.Location{
																										Line: int(1126),
																										Column: int(35),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p11769,
																								FreeVars: nil,
																							},
																							OriginalString: "3",
																						},
																					},
																					OpFodder: ast.Fodder{},
																					Op: ast.BinaryOp(5),
																					Right: &ast.LiteralNumber{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1126),
																									Column: int(40),
																								},
																								End: ast.Location{
																									Line: int(1126),
																									Column: int(41),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p11769,
																							FreeVars: nil,
																						},
																						OriginalString: "4",
																					},
																				},
																				OpFodder: ast.Fodder{},
																				Op: ast.BinaryOp(16),
																				Right: &ast.Binary{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1126),
																								Column: int(44),
																							},
																							End: ast.Location{
																								Line: int(1126),
																								Column: int(67),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p11769,
																						FreeVars: ast.Identifiers{
																							"arr",
																							"i",
																						},
																					},
																					Left: &ast.Binary{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1126),
																									Column: int(45),
																								},
																								End: ast.Location{
																									Line: int(1126),
																									Column: int(61),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p11769,
																							FreeVars: ast.Identifiers{
																								"arr",
																								"i",
																							},
																						},
																						Left: &ast.Index{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1126),
																										Column: int(45),
																									},
																									End: ast.Location{
																										Line: int(1126),
																										Column: int(55),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p11769,
																								FreeVars: ast.Identifiers{
																									"arr",
																									"i",
																								},
																							},
																							Target: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1126),
																											Column: int(45),
																										},
																										End: ast.Location{
																											Line: int(1126),
																											Column: int(48),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p11769,
																									FreeVars: ast.Identifiers{
																										"arr",
																									},
																								},
																								Id: "arr",
																							},
																							LeftBracketFodder: ast.Fodder{},
																							Index: &ast.Binary{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1126),
																											Column: int(49),
																										},
																										End: ast.Location{
																											Line: int(1126),
																											Column: int(54),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p11769,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																								},
																								Left: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1126),
																												Column: int(49),
																											},
																											End: ast.Location{
																												Line: int(1126),
																												Column: int(50),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p11769,
																										FreeVars: ast.Identifiers{
																											"i",
																										},
																									},
																									Id: "i",
																								},
																								OpFodder: ast.Fodder{},
																								Op: ast.BinaryOp(3),
																								Right: &ast.LiteralNumber{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1126),
																												Column: int(53),
																											},
																											End: ast.Location{
																												Line: int(1126),
																												Column: int(54),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p11769,
																										FreeVars: nil,
																									},
																									OriginalString: "1",
																								},
																							},
																							RightBracketFodder: ast.Fodder{},
																							Id: nil,
																						},
																						OpFodder: ast.Fodder{},
																						Op: ast.BinaryOp(14),
																						Right: &ast.LiteralNumber{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1126),
																										Column: int(58),
																									},
																									End: ast.Location{
																										Line: int(1126),
																										Column: int(61),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p11769,
																								FreeVars: nil,
																							},
																							OriginalString: "240",
																						},
																					},
																					OpFodder: ast.Fodder{},
																					Op: ast.BinaryOp(6),
																					Right: &ast.LiteralNumber{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1126),
																									Column: int(66),
																								},
																								End: ast.Location{
																									Line: int(1126),
																									Column: int(67),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p11769,
																							FreeVars: nil,
																						},
																						OriginalString: "4",
																					},
																				},
																			},
																			RightBracketFodder: ast.Fodder{},
																			Id: nil,
																		},
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(3),
																	Right: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1128),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(1128),
																					Column: int(73),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p11769,
																			FreeVars: ast.Identifiers{
																				"arr",
																				"base64_table",
																				"i",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1128),
																						Column: int(11),
																					},
																					End: ast.Location{
																						Line: int(1128),
																						Column: int(23),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(10),
																						Comment: []string{},
																					},
																					ast.FodderElement{
																						Kind: ast.FodderKind(2),
																						Blanks: int(0),
																						Indent: int(10),
																						Comment: []string{
																							"// 4 LSB of i+1, 2 MSB of i+2",
																						},
																					},
																				},
																				Ctx: p11769,
																				FreeVars: ast.Identifiers{
																					"base64_table",
																				},
																			},
																			Id: "base64_table",
																		},
																		LeftBracketFodder: ast.Fodder{},
																		Index: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1128),
																						Column: int(24),
																					},
																					End: ast.Location{
																						Line: int(1128),
																						Column: int(72),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p11769,
																				FreeVars: ast.Identifiers{
																					"arr",
																					"i",
																				},
																			},
																			Left: &ast.Binary{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1128),
																							Column: int(24),
																						},
																						End: ast.Location{
																							Line: int(1128),
																							Column: int(46),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p11769,
																					FreeVars: ast.Identifiers{
																						"arr",
																						"i",
																					},
																				},
																				Left: &ast.Binary{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1128),
																								Column: int(25),
																							},
																							End: ast.Location{
																								Line: int(1128),
																								Column: int(40),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p11769,
																						FreeVars: ast.Identifiers{
																							"arr",
																							"i",
																						},
																					},
																					Left: &ast.Index{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1128),
																									Column: int(25),
																								},
																								End: ast.Location{
																									Line: int(1128),
																									Column: int(35),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p11769,
																							FreeVars: ast.Identifiers{
																								"arr",
																								"i",
																							},
																						},
																						Target: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1128),
																										Column: int(25),
																									},
																									End: ast.Location{
																										Line: int(1128),
																										Column: int(28),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p11769,
																								FreeVars: ast.Identifiers{
																									"arr",
																								},
																							},
																							Id: "arr",
																						},
																						LeftBracketFodder: ast.Fodder{},
																						Index: &ast.Binary{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1128),
																										Column: int(29),
																									},
																									End: ast.Location{
																										Line: int(1128),
																										Column: int(34),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p11769,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																							},
																							Left: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1128),
																											Column: int(29),
																										},
																										End: ast.Location{
																											Line: int(1128),
																											Column: int(30),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p11769,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																								},
																								Id: "i",
																							},
																							OpFodder: ast.Fodder{},
																							Op: ast.BinaryOp(3),
																							Right: &ast.LiteralNumber{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1128),
																											Column: int(33),
																										},
																										End: ast.Location{
																											Line: int(1128),
																											Column: int(34),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p11769,
																									FreeVars: nil,
																								},
																								OriginalString: "1",
																							},
																						},
																						RightBracketFodder: ast.Fodder{},
																						Id: nil,
																					},
																					OpFodder: ast.Fodder{},
																					Op: ast.BinaryOp(14),
																					Right: &ast.LiteralNumber{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1128),
																									Column: int(38),
																								},
																								End: ast.Location{
																									Line: int(1128),
																									Column: int(40),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p11769,
																							FreeVars: nil,
																						},
																						OriginalString: "15",
																					},
																				},
																				OpFodder: ast.Fodder{},
																				Op: ast.BinaryOp(5),
																				Right: &ast.LiteralNumber{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1128),
																								Column: int(45),
																							},
																							End: ast.Location{
																								Line: int(1128),
																								Column: int(46),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p11769,
																						FreeVars: nil,
																					},
																					OriginalString: "2",
																				},
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(16),
																			Right: &ast.Binary{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1128),
																							Column: int(49),
																						},
																						End: ast.Location{
																							Line: int(1128),
																							Column: int(72),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p11769,
																					FreeVars: ast.Identifiers{
																						"arr",
																						"i",
																					},
																				},
																				Left: &ast.Binary{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1128),
																								Column: int(50),
																							},
																							End: ast.Location{
																								Line: int(1128),
																								Column: int(66),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p11769,
																						FreeVars: ast.Identifiers{
																							"arr",
																							"i",
																						},
																					},
																					Left: &ast.Index{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1128),
																									Column: int(50),
																								},
																								End: ast.Location{
																									Line: int(1128),
																									Column: int(60),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p11769,
																							FreeVars: ast.Identifiers{
																								"arr",
																								"i",
																							},
																						},
																						Target: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1128),
																										Column: int(50),
																									},
																									End: ast.Location{
																										Line: int(1128),
																										Column: int(53),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p11769,
																								FreeVars: ast.Identifiers{
																									"arr",
																								},
																							},
																							Id: "arr",
																						},
																						LeftBracketFodder: ast.Fodder{},
																						Index: &ast.Binary{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1128),
																										Column: int(54),
																									},
																									End: ast.Location{
																										Line: int(1128),
																										Column: int(59),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p11769,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																							},
																							Left: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1128),
																											Column: int(54),
																										},
																										End: ast.Location{
																											Line: int(1128),
																											Column: int(55),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p11769,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																								},
																								Id: "i",
																							},
																							OpFodder: ast.Fodder{},
																							Op: ast.BinaryOp(3),
																							Right: &ast.LiteralNumber{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1128),
																											Column: int(58),
																										},
																										End: ast.Location{
																											Line: int(1128),
																											Column: int(59),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p11769,
																									FreeVars: nil,
																								},
																								OriginalString: "2",
																							},
																						},
																						RightBracketFodder: ast.Fodder{},
																						Id: nil,
																					},
																					OpFodder: ast.Fodder{},
																					Op: ast.BinaryOp(14),
																					Right: &ast.LiteralNumber{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1128),
																									Column: int(63),
																								},
																								End: ast.Location{
																									Line: int(1128),
																									Column: int(66),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p11769,
																							FreeVars: nil,
																						},
																						OriginalString: "192",
																					},
																				},
																				OpFodder: ast.Fodder{},
																				Op: ast.BinaryOp(6),
																				Right: &ast.LiteralNumber{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1128),
																								Column: int(71),
																							},
																							End: ast.Location{
																								Line: int(1128),
																								Column: int(72),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p11769,
																						FreeVars: nil,
																					},
																					OriginalString: "6",
																				},
																			},
																		},
																		RightBracketFodder: ast.Fodder{},
																		Id: nil,
																	},
																},
																OpFodder: ast.Fodder{},
																Op: ast.BinaryOp(3),
																Right: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1130),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(1130),
																				Column: int(42),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p11769,
																		FreeVars: ast.Identifiers{
																			"arr",
																			"base64_table",
																			"i",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1130),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(1130),
																					Column: int(23),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(10),
																					Comment: []string{},
																				},
																				ast.FodderElement{
																					Kind: ast.FodderKind(2),
																					Blanks: int(0),
																					Indent: int(10),
																					Comment: []string{
																						"// 6 LSB of i+2",
																					},
																				},
																			},
																			Ctx: p11769,
																			FreeVars: ast.Identifiers{
																				"base64_table",
																			},
																		},
																		Id: "base64_table",
																	},
																	LeftBracketFodder: ast.Fodder{},
																	Index: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1130),
																					Column: int(25),
																				},
																				End: ast.Location{
																					Line: int(1130),
																					Column: int(40),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p11769,
																			FreeVars: ast.Identifiers{
																				"arr",
																				"i",
																			},
																		},
																		Left: &ast.Index{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1130),
																						Column: int(25),
																					},
																					End: ast.Location{
																						Line: int(1130),
																						Column: int(35),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p11769,
																				FreeVars: ast.Identifiers{
																					"arr",
																					"i",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1130),
																							Column: int(25),
																						},
																						End: ast.Location{
																							Line: int(1130),
																							Column: int(28),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p11769,
																					FreeVars: ast.Identifiers{
																						"arr",
																					},
																				},
																				Id: "arr",
																			},
																			LeftBracketFodder: ast.Fodder{},
																			Index: &ast.Binary{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1130),
																							Column: int(29),
																						},
																						End: ast.Location{
																							Line: int(1130),
																							Column: int(34),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p11769,
																					FreeVars: ast.Identifiers{
																						"i",
																					},
																				},
																				Left: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1130),
																								Column: int(29),
																							},
																							End: ast.Location{
																								Line: int(1130),
																								Column: int(30),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p11769,
																						FreeVars: ast.Identifiers{
																							"i",
																						},
																					},
																					Id: "i",
																				},
																				OpFodder: ast.Fodder{},
																				Op: ast.BinaryOp(3),
																				Right: &ast.LiteralNumber{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1130),
																								Column: int(33),
																							},
																							End: ast.Location{
																								Line: int(1130),
																								Column: int(34),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p11769,
																						FreeVars: nil,
																					},
																					OriginalString: "2",
																				},
																			},
																			RightBracketFodder: ast.Fodder{},
																			Id: nil,
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(14),
																		Right: &ast.LiteralNumber{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1130),
																						Column: int(38),
																					},
																					End: ast.Location{
																						Line: int(1130),
																						Column: int(40),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p11769,
																				FreeVars: nil,
																			},
																			OriginalString: "63",
																		},
																	},
																	RightBracketFodder: ast.Fodder{},
																	Id: nil,
																},
															},
															Fun: nil,
															CloseFodder: ast.Fodder{},
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1122),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(1130),
																	Column: int(42),
																},
																File: p1,
															},
														},
													},
													Body: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1131),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(1131),
																	Column: int(33),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p11527,
															FreeVars: ast.Identifiers{
																"arr",
																"aux",
																"i",
																"r",
																"str",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1131),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(1131),
																		Column: int(12),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																		Comment: []string{},
																	},
																},
																Ctx: p11527,
																FreeVars: ast.Identifiers{
																	"aux",
																},
															},
															Id: "aux",
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1131),
																					Column: int(13),
																				},
																				End: ast.Location{
																					Line: int(1131),
																					Column: int(16),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p11891,
																			FreeVars: ast.Identifiers{
																				"arr",
																			},
																		},
																		Id: "arr",
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1131),
																					Column: int(18),
																				},
																				End: ast.Location{
																					Line: int(1131),
																					Column: int(23),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p11891,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																		},
																		Left: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1131),
																						Column: int(18),
																					},
																					End: ast.Location{
																						Line: int(1131),
																						Column: int(19),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p11891,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																			},
																			Id: "i",
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(3),
																		Right: &ast.LiteralNumber{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1131),
																						Column: int(22),
																					},
																					End: ast.Location{
																						Line: int(1131),
																						Column: int(23),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p11891,
																				FreeVars: nil,
																			},
																			OriginalString: "3",
																		},
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1131),
																					Column: int(25),
																				},
																				End: ast.Location{
																					Line: int(1131),
																					Column: int(32),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p11891,
																			FreeVars: ast.Identifiers{
																				"r",
																				"str",
																			},
																		},
																		Left: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1131),
																						Column: int(25),
																					},
																					End: ast.Location{
																						Line: int(1131),
																						Column: int(26),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p11891,
																				FreeVars: ast.Identifiers{
																					"r",
																				},
																			},
																			Id: "r",
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(3),
																		Right: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1131),
																						Column: int(29),
																					},
																					End: ast.Location{
																						Line: int(1131),
																						Column: int(32),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p11891,
																				FreeVars: ast.Identifiers{
																					"str",
																				},
																			},
																			Id: "str",
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														TrailingComma: false,
														TailStrict: true,
														FodderRight: ast.Fodder{},
														TailStrictFodder: ast.Fodder{},
													},
												},
											},
										},
									},
								},
								Fun: nil,
								CloseFodder: nil,
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
							},
						},
						Body: &ast.Local{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1133),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1137),
										Column: int(24),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(1),
										Indent: int(4),
										Comment: []string{},
									},
								},
								Ctx: p11468,
								FreeVars: ast.Identifiers{
									"aux",
									"bytes",
									"std",
								},
							},
							Binds: ast.LocalBinds{
								ast.LocalBind{
									VarFodder: ast.Fodder{},
									Variable: "sanity",
									EqFodder: ast.Fodder{},
									Body: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1133),
													Column: int(20),
												},
												End: ast.Location{
													Line: int(1133),
													Column: int(73),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p11909,
											FreeVars: ast.Identifiers{
												"bytes",
												"std",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1133),
														Column: int(20),
													},
													End: ast.Location{
														Line: int(1133),
														Column: int(29),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p11909,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1133),
															Column: int(20),
														},
														End: ast.Location{
															Line: int(1133),
															Column: int(23),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: ast.Fodder{},
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "foldl",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: ast.Fodder{},
											Id: nil,
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Function{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1133),
																	Column: int(30),
																},
																End: ast.Location{
																	Line: int(1133),
																	Column: int(59),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p11918,
															FreeVars: nil,
														},
														ParenLeftFodder: ast.Fodder{},
														Parameters: []ast.Parameter{
															ast.Parameter{
																NameFodder: ast.Fodder{},
																Name: "r",
																EqFodder: nil,
																DefaultArg: nil,
																CommaFodder: ast.Fodder{},
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1133),
																		Column: int(39),
																	},
																	End: ast.Location{
																		Line: int(1133),
																		Column: int(40),
																	},
																	File: p1,
																},
															},
															ast.Parameter{
																NameFodder: ast.Fodder{},
																Name: "a",
																EqFodder: nil,
																DefaultArg: nil,
																CommaFodder: nil,
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1133),
																		Column: int(42),
																	},
																	End: ast.Location{
																		Line: int(1133),
																		Column: int(43),
																	},
																	File: p1,
																},
															},
														},
														TrailingComma: false,
														ParenRightFodder: ast.Fodder{},
														Body: &ast.Binary{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1133),
																		Column: int(45),
																	},
																	End: ast.Location{
																		Line: int(1133),
																		Column: int(59),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p11921,
																FreeVars: ast.Identifiers{
																	"a",
																	"r",
																},
															},
															Left: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1133),
																			Column: int(45),
																		},
																		End: ast.Location{
																			Line: int(1133),
																			Column: int(46),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p11921,
																	FreeVars: ast.Identifiers{
																		"r",
																	},
																},
																Id: "r",
															},
															OpFodder: ast.Fodder{},
															Op: ast.BinaryOp(17),
															Right: &ast.Binary{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1133),
																			Column: int(51),
																		},
																		End: ast.Location{
																			Line: int(1133),
																			Column: int(58),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p11921,
																	FreeVars: ast.Identifiers{
																		"a",
																	},
																},
																Left: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1133),
																				Column: int(51),
																			},
																			End: ast.Location{
																				Line: int(1133),
																				Column: int(52),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p11921,
																		FreeVars: ast.Identifiers{
																			"a",
																		},
																	},
																	Id: "a",
																},
																OpFodder: ast.Fodder{},
																Op: ast.BinaryOp(9),
																Right: &ast.LiteralNumber{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1133),
																				Column: int(55),
																			},
																			End: ast.Location{
																				Line: int(1133),
																				Column: int(58),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p11921,
																		FreeVars: nil,
																	},
																	OriginalString: "256",
																},
															},
														},
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1133),
																	Column: int(61),
																},
																End: ast.Location{
																	Line: int(1133),
																	Column: int(66),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p11918,
															FreeVars: ast.Identifiers{
																"bytes",
															},
														},
														Id: "bytes",
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.LiteralBoolean{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1133),
																	Column: int(68),
																},
																End: ast.Location{
																	Line: int(1133),
																	Column: int(72),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p11918,
															FreeVars: nil,
														},
														Value: true,
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
									Fun: nil,
									CloseFodder: ast.Fodder{},
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1133),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(1133),
											Column: int(73),
										},
										File: p1,
									},
								},
							},
							Body: &ast.Conditional{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1134),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(1137),
											Column: int(24),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
											Comment: []string{},
										},
									},
									Ctx: p11468,
									FreeVars: ast.Identifiers{
										"aux",
										"bytes",
										"sanity",
									},
								},
								Cond: &ast.Unary{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1134),
												Column: int(8),
											},
											End: ast.Location{
												Line: int(1134),
												Column: int(15),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p11468,
										FreeVars: ast.Identifiers{
											"sanity",
										},
									},
									Op: ast.UnaryOp(0),
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1134),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1134),
													Column: int(15),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p11468,
											FreeVars: ast.Identifiers{
												"sanity",
											},
										},
										Id: "sanity",
									},
								},
								ThenFodder: ast.Fodder{},
								BranchTrue: &ast.Error{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1135),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1135),
												Column: int(71),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										Ctx: p11468,
										FreeVars: nil,
									},
									Expr: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1135),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(1135),
													Column: int(71),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p11468,
											FreeVars: nil,
										},
										Value: "Can only base64 encode strings / arrays of single bytes.",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
								},
								ElseFodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
										Comment: []string{},
									},
								},
								BranchFalse: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1137),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1137),
												Column: int(24),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p11468,
										FreeVars: ast.Identifiers{
											"aux",
											"bytes",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1137),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(1137),
													Column: int(10),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
													Comment: []string{},
												},
											},
											Ctx: p11468,
											FreeVars: ast.Identifiers{
												"aux",
											},
										},
										Id: "aux",
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1137),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(1137),
																Column: int(16),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p11951,
														FreeVars: ast.Identifiers{
															"bytes",
														},
													},
													Id: "bytes",
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.LiteralNumber{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1137),
																Column: int(18),
															},
															End: ast.Location{
																Line: int(1137),
																Column: int(19),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p11951,
														FreeVars: nil,
													},
													OriginalString: "0",
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1137),
																Column: int(21),
															},
															End: ast.Location{
																Line: int(1137),
																Column: int(23),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p11951,
														FreeVars: nil,
													},
													Value: "",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
							},
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(1093),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1137),
					Column: int(24),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "base64DecodeBytes",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"base64_inv",
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1140),
								Column: int(21),
							},
							End: ast.Location{
								Line: int(1140),
								Column: int(24),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1141),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1159),
								Column: int(22),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p11961,
						FreeVars: ast.Identifiers{
							"base64_inv",
							"std",
							"str",
						},
					},
					Cond: &ast.Binary{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1141),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(1141),
									Column: int(32),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p11961,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
						},
						Left: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
									"str",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: nil,
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "mod",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: nil,
								Id: nil,
							},
							FodderLeft: nil,
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1141),
														Column: int(8),
													},
													End: ast.Location{
														Line: int(1141),
														Column: int(23),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p11961,
												FreeVars: ast.Identifiers{
													"std",
													"str",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1141),
															Column: int(8),
														},
														End: ast.Location{
															Line: int(1141),
															Column: int(18),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p11961,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1141),
																Column: int(8),
															},
															End: ast.Location{
																Line: int(1141),
																Column: int(11),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "length",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1141),
																		Column: int(19),
																	},
																	End: ast.Location{
																		Line: int(1141),
																		Column: int(22),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p11982,
																FreeVars: ast.Identifiers{
																	"str",
																},
															},
															Id: "str",
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
										CommaFodder: nil,
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralNumber{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1141),
														Column: int(26),
													},
													End: ast.Location{
														Line: int(1141),
														Column: int(27),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p11961,
												FreeVars: nil,
											},
											OriginalString: "4",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: nil,
							TailStrictFodder: nil,
						},
						OpFodder: ast.Fodder{},
						Op: ast.BinaryOp(13),
						Right: &ast.LiteralNumber{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1141),
										Column: int(31),
									},
									End: ast.Location{
										Line: int(1141),
										Column: int(32),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p11961,
								FreeVars: nil,
							},
							OriginalString: "0",
						},
					},
					ThenFodder: ast.Fodder{},
					BranchTrue: &ast.Error{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1142),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(1142),
									Column: int(53),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
									Comment: []string{},
								},
							},
							Ctx: p11961,
							FreeVars: ast.Identifiers{
								"std",
								"str",
							},
						},
						Expr: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
									"str",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: nil,
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "mod",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: nil,
								Id: nil,
							},
							FodderLeft: nil,
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1142),
														Column: int(13),
													},
													End: ast.Location{
														Line: int(1142),
														Column: int(47),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p11961,
												FreeVars: nil,
											},
											Value: "Not a base64 encoded string \"%s\"",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										CommaFodder: nil,
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1142),
														Column: int(50),
													},
													End: ast.Location{
														Line: int(1142),
														Column: int(53),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p11961,
												FreeVars: ast.Identifiers{
													"str",
												},
											},
											Id: "str",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: nil,
							TailStrictFodder: nil,
						},
					},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
							Comment: []string{},
						},
					},
					BranchFalse: &ast.Local{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1144),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(1159),
									Column: int(22),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
									Comment: []string{},
								},
							},
							Ctx: p11961,
							FreeVars: ast.Identifiers{
								"base64_inv",
								"std",
								"str",
							},
						},
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: nil,
								Variable: "aux",
								EqFodder: nil,
								Body: &ast.Function{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1144),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(1158),
												Column: int(44),
											},
											File: p1,
										},
										Fodder: nil,
										Ctx: p12006,
										FreeVars: ast.Identifiers{
											"aux",
											"base64_inv",
											"std",
										},
									},
									ParenLeftFodder: ast.Fodder{},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "str",
											EqFodder: nil,
											DefaultArg: nil,
											CommaFodder: ast.Fodder{},
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1144),
													Column: int(17),
												},
												End: ast.Location{
													Line: int(1144),
													Column: int(20),
												},
												File: p1,
											},
										},
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "i",
											EqFodder: nil,
											DefaultArg: nil,
											CommaFodder: ast.Fodder{},
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1144),
													Column: int(22),
												},
												End: ast.Location{
													Line: int(1144),
													Column: int(23),
												},
												File: p1,
											},
										},
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "r",
											EqFodder: nil,
											DefaultArg: nil,
											CommaFodder: nil,
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1144),
													Column: int(25),
												},
												End: ast.Location{
													Line: int(1144),
													Column: int(26),
												},
												File: p1,
											},
										},
									},
									TrailingComma: false,
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Conditional{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1145),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1158),
													Column: int(44),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
													Comment: []string{},
												},
											},
											Ctx: p12011,
											FreeVars: ast.Identifiers{
												"aux",
												"base64_inv",
												"i",
												"r",
												"std",
												"str",
											},
										},
										Cond: &ast.Binary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1145),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(1145),
														Column: int(32),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p12011,
												FreeVars: ast.Identifiers{
													"i",
													"std",
													"str",
												},
											},
											Left: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1145),
															Column: int(12),
														},
														End: ast.Location{
															Line: int(1145),
															Column: int(13),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p12011,
													FreeVars: ast.Identifiers{
														"i",
													},
												},
												Id: "i",
											},
											OpFodder: ast.Fodder{},
											Op: ast.BinaryOp(8),
											Right: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1145),
															Column: int(17),
														},
														End: ast.Location{
															Line: int(1145),
															Column: int(32),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p12011,
													FreeVars: ast.Identifiers{
														"std",
														"str",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1145),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(1145),
																Column: int(27),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p12011,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1145),
																	Column: int(17),
																},
																End: ast.Location{
																	Line: int(1145),
																	Column: int(20),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: ast.Fodder{},
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "length",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: ast.Fodder{},
													Id: nil,
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1145),
																			Column: int(28),
																		},
																		End: ast.Location{
																			Line: int(1145),
																			Column: int(31),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p12026,
																	FreeVars: ast.Identifiers{
																		"str",
																	},
																},
																Id: "str",
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
										},
										ThenFodder: ast.Fodder{},
										BranchTrue: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1146),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(1146),
														Column: int(12),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(10),
														Comment: []string{},
													},
												},
												Ctx: p12011,
												FreeVars: ast.Identifiers{
													"r",
												},
											},
											Id: "r",
										},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(8),
												Comment: []string{},
											},
										},
										BranchFalse: &ast.Local{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1149),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(1158),
														Column: int(44),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(10),
														Comment: []string{},
													},
													ast.FodderElement{
														Kind: ast.FodderKind(2),
														Blanks: int(0),
														Indent: int(10),
														Comment: []string{
															"// all 6 bits of i, 2 MSB of i+1",
														},
													},
												},
												Ctx: p12011,
												FreeVars: ast.Identifiers{
													"aux",
													"base64_inv",
													"i",
													"r",
													"str",
												},
											},
											Binds: ast.LocalBinds{
												ast.LocalBind{
													VarFodder: ast.Fodder{},
													Variable: "n1",
													EqFodder: ast.Fodder{},
													Body: &ast.Array{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1149),
																	Column: int(22),
																},
																End: ast.Location{
																	Line: int(1149),
																	Column: int(79),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p12038,
															FreeVars: ast.Identifiers{
																"base64_inv",
																"i",
																"str",
															},
														},
														Elements: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1149),
																				Column: int(23),
																			},
																			End: ast.Location{
																				Line: int(1149),
																				Column: int(78),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p12042,
																		FreeVars: ast.Identifiers{
																			"base64_inv",
																			"i",
																			"str",
																		},
																	},
																	Left: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1149),
																					Column: int(23),
																				},
																				End: ast.Location{
																					Line: int(1149),
																					Column: int(46),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p12042,
																			FreeVars: ast.Identifiers{
																				"base64_inv",
																				"i",
																				"str",
																			},
																		},
																		Left: &ast.Index{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1149),
																						Column: int(23),
																					},
																					End: ast.Location{
																						Line: int(1149),
																						Column: int(41),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p12042,
																				FreeVars: ast.Identifiers{
																					"base64_inv",
																					"i",
																					"str",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1149),
																							Column: int(23),
																						},
																						End: ast.Location{
																							Line: int(1149),
																							Column: int(33),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p12042,
																					FreeVars: ast.Identifiers{
																						"base64_inv",
																					},
																				},
																				Id: "base64_inv",
																			},
																			LeftBracketFodder: ast.Fodder{},
																			Index: &ast.Index{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1149),
																							Column: int(34),
																						},
																						End: ast.Location{
																							Line: int(1149),
																							Column: int(40),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p12042,
																					FreeVars: ast.Identifiers{
																						"i",
																						"str",
																					},
																				},
																				Target: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1149),
																								Column: int(34),
																							},
																							End: ast.Location{
																								Line: int(1149),
																								Column: int(37),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p12042,
																						FreeVars: ast.Identifiers{
																							"str",
																						},
																					},
																					Id: "str",
																				},
																				LeftBracketFodder: ast.Fodder{},
																				Index: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1149),
																								Column: int(38),
																							},
																							End: ast.Location{
																								Line: int(1149),
																								Column: int(39),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p12042,
																						FreeVars: ast.Identifiers{
																							"i",
																						},
																					},
																					Id: "i",
																				},
																				RightBracketFodder: ast.Fodder{},
																				Id: nil,
																			},
																			RightBracketFodder: ast.Fodder{},
																			Id: nil,
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(5),
																		Right: &ast.LiteralNumber{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1149),
																						Column: int(45),
																					},
																					End: ast.Location{
																						Line: int(1149),
																						Column: int(46),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p12042,
																				FreeVars: nil,
																			},
																			OriginalString: "2",
																		},
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(16),
																	Right: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1149),
																					Column: int(50),
																				},
																				End: ast.Location{
																					Line: int(1149),
																					Column: int(77),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p12042,
																			FreeVars: ast.Identifiers{
																				"base64_inv",
																				"i",
																				"str",
																			},
																		},
																		Left: &ast.Index{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1149),
																						Column: int(50),
																					},
																					End: ast.Location{
																						Line: int(1149),
																						Column: int(72),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p12042,
																				FreeVars: ast.Identifiers{
																					"base64_inv",
																					"i",
																					"str",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1149),
																							Column: int(50),
																						},
																						End: ast.Location{
																							Line: int(1149),
																							Column: int(60),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p12042,
																					FreeVars: ast.Identifiers{
																						"base64_inv",
																					},
																				},
																				Id: "base64_inv",
																			},
																			LeftBracketFodder: ast.Fodder{},
																			Index: &ast.Index{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1149),
																							Column: int(61),
																						},
																						End: ast.Location{
																							Line: int(1149),
																							Column: int(71),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p12042,
																					FreeVars: ast.Identifiers{
																						"i",
																						"str",
																					},
																				},
																				Target: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1149),
																								Column: int(61),
																							},
																							End: ast.Location{
																								Line: int(1149),
																								Column: int(64),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p12042,
																						FreeVars: ast.Identifiers{
																							"str",
																						},
																					},
																					Id: "str",
																				},
																				LeftBracketFodder: ast.Fodder{},
																				Index: &ast.Binary{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1149),
																								Column: int(65),
																							},
																							End: ast.Location{
																								Line: int(1149),
																								Column: int(70),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p12042,
																						FreeVars: ast.Identifiers{
																							"i",
																						},
																					},
																					Left: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1149),
																									Column: int(65),
																								},
																								End: ast.Location{
																									Line: int(1149),
																									Column: int(66),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12042,
																							FreeVars: ast.Identifiers{
																								"i",
																							},
																						},
																						Id: "i",
																					},
																					OpFodder: ast.Fodder{},
																					Op: ast.BinaryOp(3),
																					Right: &ast.LiteralNumber{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1149),
																									Column: int(69),
																								},
																								End: ast.Location{
																									Line: int(1149),
																									Column: int(70),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12042,
																							FreeVars: nil,
																						},
																						OriginalString: "1",
																					},
																				},
																				RightBracketFodder: ast.Fodder{},
																				Id: nil,
																			},
																			RightBracketFodder: ast.Fodder{},
																			Id: nil,
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(6),
																		Right: &ast.LiteralNumber{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1149),
																						Column: int(76),
																					},
																					End: ast.Location{
																						Line: int(1149),
																						Column: int(77),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p12042,
																				FreeVars: nil,
																			},
																			OriginalString: "4",
																		},
																	},
																},
																CommaFodder: nil,
															},
														},
														TrailingComma: false,
														CloseFodder: ast.Fodder{},
													},
													Fun: nil,
													CloseFodder: ast.Fodder{},
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1149),
															Column: int(17),
														},
														End: ast.Location{
															Line: int(1149),
															Column: int(79),
														},
														File: p1,
													},
												},
											},
											Body: &ast.Local{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1151),
															Column: int(11),
														},
														End: ast.Location{
															Line: int(1158),
															Column: int(44),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(10),
															Comment: []string{},
														},
														ast.FodderElement{
															Kind: ast.FodderKind(2),
															Blanks: int(0),
															Indent: int(10),
															Comment: []string{
																"// 4 LSB of i+1, 4MSB of i+2",
															},
														},
													},
													Ctx: p12011,
													FreeVars: ast.Identifiers{
														"aux",
														"base64_inv",
														"i",
														"n1",
														"r",
														"str",
													},
												},
												Binds: ast.LocalBinds{
													ast.LocalBind{
														VarFodder: ast.Fodder{},
														Variable: "n2",
														EqFodder: ast.Fodder{},
														Body: &ast.Conditional{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1152),
																		Column: int(13),
																	},
																	End: ast.Location{
																		Line: int(1153),
																		Column: int(86),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(12),
																		Comment: []string{},
																	},
																},
																Ctx: p12080,
																FreeVars: ast.Identifiers{
																	"base64_inv",
																	"i",
																	"str",
																},
															},
															Cond: &ast.Binary{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1152),
																			Column: int(16),
																		},
																		End: ast.Location{
																			Line: int(1152),
																			Column: int(33),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p12080,
																	FreeVars: ast.Identifiers{
																		"i",
																		"str",
																	},
																},
																Left: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1152),
																				Column: int(16),
																			},
																			End: ast.Location{
																				Line: int(1152),
																				Column: int(26),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p12080,
																		FreeVars: ast.Identifiers{
																			"i",
																			"str",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1152),
																					Column: int(16),
																				},
																				End: ast.Location{
																					Line: int(1152),
																					Column: int(19),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p12080,
																			FreeVars: ast.Identifiers{
																				"str",
																			},
																		},
																		Id: "str",
																	},
																	LeftBracketFodder: ast.Fodder{},
																	Index: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1152),
																					Column: int(20),
																				},
																				End: ast.Location{
																					Line: int(1152),
																					Column: int(25),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p12080,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																		},
																		Left: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1152),
																						Column: int(20),
																					},
																					End: ast.Location{
																						Line: int(1152),
																						Column: int(21),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p12080,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																			},
																			Id: "i",
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(3),
																		Right: &ast.LiteralNumber{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1152),
																						Column: int(24),
																					},
																					End: ast.Location{
																						Line: int(1152),
																						Column: int(25),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p12080,
																				FreeVars: nil,
																			},
																			OriginalString: "2",
																		},
																	},
																	RightBracketFodder: ast.Fodder{},
																	Id: nil,
																},
																OpFodder: ast.Fodder{},
																Op: ast.BinaryOp(12),
																Right: &ast.LiteralString{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1152),
																				Column: int(30),
																			},
																			End: ast.Location{
																				Line: int(1152),
																				Column: int(33),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p12080,
																		FreeVars: nil,
																	},
																	Value: "=",
																	Kind: ast.LiteralStringKind(1),
																	BlockIndent: "",
																	BlockTermIndent: "",
																},
															},
															ThenFodder: ast.Fodder{},
															BranchTrue: &ast.Array{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1152),
																			Column: int(39),
																		},
																		End: ast.Location{
																			Line: int(1152),
																			Column: int(41),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p12080,
																	FreeVars: nil,
																},
																Elements: nil,
																TrailingComma: false,
																CloseFodder: ast.Fodder{},
															},
															ElseFodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(12),
																	Comment: []string{},
																},
															},
															BranchFalse: &ast.Array{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1153),
																			Column: int(18),
																		},
																		End: ast.Location{
																			Line: int(1153),
																			Column: int(86),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p12080,
																	FreeVars: ast.Identifiers{
																		"base64_inv",
																		"i",
																		"str",
																	},
																},
																Elements: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1153),
																						Column: int(19),
																					},
																					End: ast.Location{
																						Line: int(1153),
																						Column: int(85),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p12100,
																				FreeVars: ast.Identifiers{
																					"base64_inv",
																					"i",
																					"str",
																				},
																			},
																			Left: &ast.Binary{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1153),
																							Column: int(19),
																						},
																						End: ast.Location{
																							Line: int(1153),
																							Column: int(53),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p12100,
																					FreeVars: ast.Identifiers{
																						"base64_inv",
																						"i",
																						"str",
																					},
																				},
																				Left: &ast.Binary{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1153),
																								Column: int(20),
																							},
																							End: ast.Location{
																								Line: int(1153),
																								Column: int(47),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p12100,
																						FreeVars: ast.Identifiers{
																							"base64_inv",
																							"i",
																							"str",
																						},
																					},
																					Left: &ast.Index{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1153),
																									Column: int(20),
																								},
																								End: ast.Location{
																									Line: int(1153),
																									Column: int(42),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12100,
																							FreeVars: ast.Identifiers{
																								"base64_inv",
																								"i",
																								"str",
																							},
																						},
																						Target: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1153),
																										Column: int(20),
																									},
																									End: ast.Location{
																										Line: int(1153),
																										Column: int(30),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p12100,
																								FreeVars: ast.Identifiers{
																									"base64_inv",
																								},
																							},
																							Id: "base64_inv",
																						},
																						LeftBracketFodder: ast.Fodder{},
																						Index: &ast.Index{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1153),
																										Column: int(31),
																									},
																									End: ast.Location{
																										Line: int(1153),
																										Column: int(41),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p12100,
																								FreeVars: ast.Identifiers{
																									"i",
																									"str",
																								},
																							},
																							Target: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1153),
																											Column: int(31),
																										},
																										End: ast.Location{
																											Line: int(1153),
																											Column: int(34),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p12100,
																									FreeVars: ast.Identifiers{
																										"str",
																									},
																								},
																								Id: "str",
																							},
																							LeftBracketFodder: ast.Fodder{},
																							Index: &ast.Binary{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1153),
																											Column: int(35),
																										},
																										End: ast.Location{
																											Line: int(1153),
																											Column: int(40),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p12100,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																								},
																								Left: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1153),
																												Column: int(35),
																											},
																											End: ast.Location{
																												Line: int(1153),
																												Column: int(36),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p12100,
																										FreeVars: ast.Identifiers{
																											"i",
																										},
																									},
																									Id: "i",
																								},
																								OpFodder: ast.Fodder{},
																								Op: ast.BinaryOp(3),
																								Right: &ast.LiteralNumber{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1153),
																												Column: int(39),
																											},
																											End: ast.Location{
																												Line: int(1153),
																												Column: int(40),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p12100,
																										FreeVars: nil,
																									},
																									OriginalString: "1",
																								},
																							},
																							RightBracketFodder: ast.Fodder{},
																							Id: nil,
																						},
																						RightBracketFodder: ast.Fodder{},
																						Id: nil,
																					},
																					OpFodder: ast.Fodder{},
																					Op: ast.BinaryOp(14),
																					Right: &ast.LiteralNumber{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1153),
																									Column: int(45),
																								},
																								End: ast.Location{
																									Line: int(1153),
																									Column: int(47),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12100,
																							FreeVars: nil,
																						},
																						OriginalString: "15",
																					},
																				},
																				OpFodder: ast.Fodder{},
																				Op: ast.BinaryOp(5),
																				Right: &ast.LiteralNumber{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1153),
																								Column: int(52),
																							},
																							End: ast.Location{
																								Line: int(1153),
																								Column: int(53),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p12100,
																						FreeVars: nil,
																					},
																					OriginalString: "4",
																				},
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(16),
																			Right: &ast.Binary{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1153),
																							Column: int(57),
																						},
																						End: ast.Location{
																							Line: int(1153),
																							Column: int(84),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p12100,
																					FreeVars: ast.Identifiers{
																						"base64_inv",
																						"i",
																						"str",
																					},
																				},
																				Left: &ast.Index{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1153),
																								Column: int(57),
																							},
																							End: ast.Location{
																								Line: int(1153),
																								Column: int(79),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p12100,
																						FreeVars: ast.Identifiers{
																							"base64_inv",
																							"i",
																							"str",
																						},
																					},
																					Target: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1153),
																									Column: int(57),
																								},
																								End: ast.Location{
																									Line: int(1153),
																									Column: int(67),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12100,
																							FreeVars: ast.Identifiers{
																								"base64_inv",
																							},
																						},
																						Id: "base64_inv",
																					},
																					LeftBracketFodder: ast.Fodder{},
																					Index: &ast.Index{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1153),
																									Column: int(68),
																								},
																								End: ast.Location{
																									Line: int(1153),
																									Column: int(78),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12100,
																							FreeVars: ast.Identifiers{
																								"i",
																								"str",
																							},
																						},
																						Target: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1153),
																										Column: int(68),
																									},
																									End: ast.Location{
																										Line: int(1153),
																										Column: int(71),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p12100,
																								FreeVars: ast.Identifiers{
																									"str",
																								},
																							},
																							Id: "str",
																						},
																						LeftBracketFodder: ast.Fodder{},
																						Index: &ast.Binary{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1153),
																										Column: int(72),
																									},
																									End: ast.Location{
																										Line: int(1153),
																										Column: int(77),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p12100,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																							},
																							Left: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1153),
																											Column: int(72),
																										},
																										End: ast.Location{
																											Line: int(1153),
																											Column: int(73),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p12100,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																								},
																								Id: "i",
																							},
																							OpFodder: ast.Fodder{},
																							Op: ast.BinaryOp(3),
																							Right: &ast.LiteralNumber{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1153),
																											Column: int(76),
																										},
																										End: ast.Location{
																											Line: int(1153),
																											Column: int(77),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p12100,
																									FreeVars: nil,
																								},
																								OriginalString: "2",
																							},
																						},
																						RightBracketFodder: ast.Fodder{},
																						Id: nil,
																					},
																					RightBracketFodder: ast.Fodder{},
																					Id: nil,
																				},
																				OpFodder: ast.Fodder{},
																				Op: ast.BinaryOp(6),
																				Right: &ast.LiteralNumber{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1153),
																								Column: int(83),
																							},
																							End: ast.Location{
																								Line: int(1153),
																								Column: int(84),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p12100,
																						FreeVars: nil,
																					},
																					OriginalString: "2",
																				},
																			},
																		},
																		CommaFodder: nil,
																	},
																},
																TrailingComma: false,
																CloseFodder: ast.Fodder{},
															},
														},
														Fun: nil,
														CloseFodder: ast.Fodder{},
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1151),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(1153),
																Column: int(86),
															},
															File: p1,
														},
													},
												},
												Body: &ast.Local{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1155),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(1158),
																Column: int(44),
															},
															File: p1,
														},
														Fodder: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(10),
																Comment: []string{},
															},
															ast.FodderElement{
																Kind: ast.FodderKind(2),
																Blanks: int(0),
																Indent: int(10),
																Comment: []string{
																	"// 2 LSB of i+2, all 6 bits of i+3",
																},
															},
														},
														Ctx: p12011,
														FreeVars: ast.Identifiers{
															"aux",
															"base64_inv",
															"i",
															"n1",
															"n2",
															"r",
															"str",
														},
													},
													Binds: ast.LocalBinds{
														ast.LocalBind{
															VarFodder: ast.Fodder{},
															Variable: "n3",
															EqFodder: ast.Fodder{},
															Body: &ast.Conditional{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1156),
																			Column: int(13),
																		},
																		End: ast.Location{
																			Line: int(1157),
																			Column: int(78),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(12),
																			Comment: []string{},
																		},
																	},
																	Ctx: p12144,
																	FreeVars: ast.Identifiers{
																		"base64_inv",
																		"i",
																		"str",
																	},
																},
																Cond: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1156),
																				Column: int(16),
																			},
																			End: ast.Location{
																				Line: int(1156),
																				Column: int(33),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p12144,
																		FreeVars: ast.Identifiers{
																			"i",
																			"str",
																		},
																	},
																	Left: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1156),
																					Column: int(16),
																				},
																				End: ast.Location{
																					Line: int(1156),
																					Column: int(26),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p12144,
																			FreeVars: ast.Identifiers{
																				"i",
																				"str",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1156),
																						Column: int(16),
																					},
																					End: ast.Location{
																						Line: int(1156),
																						Column: int(19),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p12144,
																				FreeVars: ast.Identifiers{
																					"str",
																				},
																			},
																			Id: "str",
																		},
																		LeftBracketFodder: ast.Fodder{},
																		Index: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1156),
																						Column: int(20),
																					},
																					End: ast.Location{
																						Line: int(1156),
																						Column: int(25),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p12144,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																			},
																			Left: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1156),
																							Column: int(20),
																						},
																						End: ast.Location{
																							Line: int(1156),
																							Column: int(21),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p12144,
																					FreeVars: ast.Identifiers{
																						"i",
																					},
																				},
																				Id: "i",
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(3),
																			Right: &ast.LiteralNumber{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1156),
																							Column: int(24),
																						},
																						End: ast.Location{
																							Line: int(1156),
																							Column: int(25),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p12144,
																					FreeVars: nil,
																				},
																				OriginalString: "3",
																			},
																		},
																		RightBracketFodder: ast.Fodder{},
																		Id: nil,
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(12),
																	Right: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1156),
																					Column: int(30),
																				},
																				End: ast.Location{
																					Line: int(1156),
																					Column: int(33),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p12144,
																			FreeVars: nil,
																		},
																		Value: "=",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																},
																ThenFodder: ast.Fodder{},
																BranchTrue: &ast.Array{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1156),
																				Column: int(39),
																			},
																			End: ast.Location{
																				Line: int(1156),
																				Column: int(41),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p12144,
																		FreeVars: nil,
																	},
																	Elements: nil,
																	TrailingComma: false,
																	CloseFodder: ast.Fodder{},
																},
																ElseFodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(12),
																		Comment: []string{},
																	},
																},
																BranchFalse: &ast.Array{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1157),
																				Column: int(18),
																			},
																			End: ast.Location{
																				Line: int(1157),
																				Column: int(78),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p12144,
																		FreeVars: ast.Identifiers{
																			"base64_inv",
																			"i",
																			"str",
																		},
																	},
																	Elements: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Binary{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1157),
																							Column: int(19),
																						},
																						End: ast.Location{
																							Line: int(1157),
																							Column: int(77),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p12164,
																					FreeVars: ast.Identifiers{
																						"base64_inv",
																						"i",
																						"str",
																					},
																				},
																				Left: &ast.Binary{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1157),
																								Column: int(19),
																							},
																							End: ast.Location{
																								Line: int(1157),
																								Column: int(52),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p12164,
																						FreeVars: ast.Identifiers{
																							"base64_inv",
																							"i",
																							"str",
																						},
																					},
																					Left: &ast.Binary{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1157),
																									Column: int(20),
																								},
																								End: ast.Location{
																									Line: int(1157),
																									Column: int(46),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12164,
																							FreeVars: ast.Identifiers{
																								"base64_inv",
																								"i",
																								"str",
																							},
																						},
																						Left: &ast.Index{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1157),
																										Column: int(20),
																									},
																									End: ast.Location{
																										Line: int(1157),
																										Column: int(42),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p12164,
																								FreeVars: ast.Identifiers{
																									"base64_inv",
																									"i",
																									"str",
																								},
																							},
																							Target: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1157),
																											Column: int(20),
																										},
																										End: ast.Location{
																											Line: int(1157),
																											Column: int(30),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p12164,
																									FreeVars: ast.Identifiers{
																										"base64_inv",
																									},
																								},
																								Id: "base64_inv",
																							},
																							LeftBracketFodder: ast.Fodder{},
																							Index: &ast.Index{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1157),
																											Column: int(31),
																										},
																										End: ast.Location{
																											Line: int(1157),
																											Column: int(41),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p12164,
																									FreeVars: ast.Identifiers{
																										"i",
																										"str",
																									},
																								},
																								Target: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1157),
																												Column: int(31),
																											},
																											End: ast.Location{
																												Line: int(1157),
																												Column: int(34),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p12164,
																										FreeVars: ast.Identifiers{
																											"str",
																										},
																									},
																									Id: "str",
																								},
																								LeftBracketFodder: ast.Fodder{},
																								Index: &ast.Binary{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1157),
																												Column: int(35),
																											},
																											End: ast.Location{
																												Line: int(1157),
																												Column: int(40),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p12164,
																										FreeVars: ast.Identifiers{
																											"i",
																										},
																									},
																									Left: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(1157),
																													Column: int(35),
																												},
																												End: ast.Location{
																													Line: int(1157),
																													Column: int(36),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p12164,
																											FreeVars: ast.Identifiers{
																												"i",
																											},
																										},
																										Id: "i",
																									},
																									OpFodder: ast.Fodder{},
																									Op: ast.BinaryOp(3),
																									Right: &ast.LiteralNumber{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(1157),
																													Column: int(39),
																												},
																												End: ast.Location{
																													Line: int(1157),
																													Column: int(40),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p12164,
																											FreeVars: nil,
																										},
																										OriginalString: "2",
																									},
																								},
																								RightBracketFodder: ast.Fodder{},
																								Id: nil,
																							},
																							RightBracketFodder: ast.Fodder{},
																							Id: nil,
																						},
																						OpFodder: ast.Fodder{},
																						Op: ast.BinaryOp(14),
																						Right: &ast.LiteralNumber{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1157),
																										Column: int(45),
																									},
																									End: ast.Location{
																										Line: int(1157),
																										Column: int(46),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p12164,
																								FreeVars: nil,
																							},
																							OriginalString: "3",
																						},
																					},
																					OpFodder: ast.Fodder{},
																					Op: ast.BinaryOp(5),
																					Right: &ast.LiteralNumber{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1157),
																									Column: int(51),
																								},
																								End: ast.Location{
																									Line: int(1157),
																									Column: int(52),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12164,
																							FreeVars: nil,
																						},
																						OriginalString: "6",
																					},
																				},
																				OpFodder: ast.Fodder{},
																				Op: ast.BinaryOp(16),
																				Right: &ast.Index{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1157),
																								Column: int(55),
																							},
																							End: ast.Location{
																								Line: int(1157),
																								Column: int(77),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p12164,
																						FreeVars: ast.Identifiers{
																							"base64_inv",
																							"i",
																							"str",
																						},
																					},
																					Target: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1157),
																									Column: int(55),
																								},
																								End: ast.Location{
																									Line: int(1157),
																									Column: int(65),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12164,
																							FreeVars: ast.Identifiers{
																								"base64_inv",
																							},
																						},
																						Id: "base64_inv",
																					},
																					LeftBracketFodder: ast.Fodder{},
																					Index: &ast.Index{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1157),
																									Column: int(66),
																								},
																								End: ast.Location{
																									Line: int(1157),
																									Column: int(76),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12164,
																							FreeVars: ast.Identifiers{
																								"i",
																								"str",
																							},
																						},
																						Target: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1157),
																										Column: int(66),
																									},
																									End: ast.Location{
																										Line: int(1157),
																										Column: int(69),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p12164,
																								FreeVars: ast.Identifiers{
																									"str",
																								},
																							},
																							Id: "str",
																						},
																						LeftBracketFodder: ast.Fodder{},
																						Index: &ast.Binary{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1157),
																										Column: int(70),
																									},
																									End: ast.Location{
																										Line: int(1157),
																										Column: int(75),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p12164,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																							},
																							Left: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1157),
																											Column: int(70),
																										},
																										End: ast.Location{
																											Line: int(1157),
																											Column: int(71),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p12164,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																								},
																								Id: "i",
																							},
																							OpFodder: ast.Fodder{},
																							Op: ast.BinaryOp(3),
																							Right: &ast.LiteralNumber{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1157),
																											Column: int(74),
																										},
																										End: ast.Location{
																											Line: int(1157),
																											Column: int(75),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p12164,
																									FreeVars: nil,
																								},
																								OriginalString: "3",
																							},
																						},
																						RightBracketFodder: ast.Fodder{},
																						Id: nil,
																					},
																					RightBracketFodder: ast.Fodder{},
																					Id: nil,
																				},
																			},
																			CommaFodder: nil,
																		},
																	},
																	TrailingComma: false,
																	CloseFodder: ast.Fodder{},
																},
															},
															Fun: nil,
															CloseFodder: ast.Fodder{},
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1155),
																	Column: int(17),
																},
																End: ast.Location{
																	Line: int(1157),
																	Column: int(78),
																},
																File: p1,
															},
														},
													},
													Body: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1158),
																	Column: int(11),
																},
																End: ast.Location{
																	Line: int(1158),
																	Column: int(44),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p12011,
															FreeVars: ast.Identifiers{
																"aux",
																"i",
																"n1",
																"n2",
																"n3",
																"r",
																"str",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1158),
																		Column: int(11),
																	},
																	End: ast.Location{
																		Line: int(1158),
																		Column: int(14),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(10),
																		Comment: []string{},
																	},
																},
																Ctx: p12011,
																FreeVars: ast.Identifiers{
																	"aux",
																},
															},
															Id: "aux",
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1158),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(1158),
																					Column: int(18),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p12205,
																			FreeVars: ast.Identifiers{
																				"str",
																			},
																		},
																		Id: "str",
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1158),
																					Column: int(20),
																				},
																				End: ast.Location{
																					Line: int(1158),
																					Column: int(25),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p12205,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																		},
																		Left: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1158),
																						Column: int(20),
																					},
																					End: ast.Location{
																						Line: int(1158),
																						Column: int(21),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p12205,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																			},
																			Id: "i",
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(3),
																		Right: &ast.LiteralNumber{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1158),
																						Column: int(24),
																					},
																					End: ast.Location{
																						Line: int(1158),
																						Column: int(25),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p12205,
																				FreeVars: nil,
																			},
																			OriginalString: "4",
																		},
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1158),
																					Column: int(27),
																				},
																				End: ast.Location{
																					Line: int(1158),
																					Column: int(43),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p12205,
																			FreeVars: ast.Identifiers{
																				"n1",
																				"n2",
																				"n3",
																				"r",
																			},
																		},
																		Left: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1158),
																						Column: int(27),
																					},
																					End: ast.Location{
																						Line: int(1158),
																						Column: int(38),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p12205,
																				FreeVars: ast.Identifiers{
																					"n1",
																					"n2",
																					"r",
																				},
																			},
																			Left: &ast.Binary{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1158),
																							Column: int(27),
																						},
																						End: ast.Location{
																							Line: int(1158),
																							Column: int(33),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p12205,
																					FreeVars: ast.Identifiers{
																						"n1",
																						"r",
																					},
																				},
																				Left: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1158),
																								Column: int(27),
																							},
																							End: ast.Location{
																								Line: int(1158),
																								Column: int(28),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p12205,
																						FreeVars: ast.Identifiers{
																							"r",
																						},
																					},
																					Id: "r",
																				},
																				OpFodder: ast.Fodder{},
																				Op: ast.BinaryOp(3),
																				Right: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1158),
																								Column: int(31),
																							},
																							End: ast.Location{
																								Line: int(1158),
																								Column: int(33),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p12205,
																						FreeVars: ast.Identifiers{
																							"n1",
																						},
																					},
																					Id: "n1",
																				},
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(3),
																			Right: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1158),
																							Column: int(36),
																						},
																						End: ast.Location{
																							Line: int(1158),
																							Column: int(38),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p12205,
																					FreeVars: ast.Identifiers{
																						"n2",
																					},
																				},
																				Id: "n2",
																			},
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(3),
																		Right: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1158),
																						Column: int(41),
																					},
																					End: ast.Location{
																						Line: int(1158),
																						Column: int(43),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p12205,
																				FreeVars: ast.Identifiers{
																					"n3",
																				},
																			},
																			Id: "n3",
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														TrailingComma: false,
														TailStrict: true,
														FodderRight: ast.Fodder{},
														TailStrictFodder: ast.Fodder{},
													},
												},
											},
										},
									},
								},
								Fun: nil,
								CloseFodder: nil,
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
							},
						},
						Body: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1159),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(1159),
										Column: int(22),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p11961,
								FreeVars: ast.Identifiers{
									"aux",
									"str",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1159),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(1159),
											Column: int(10),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
											Comment: []string{},
										},
									},
									Ctx: p11961,
									FreeVars: ast.Identifiers{
										"aux",
									},
								},
								Id: "aux",
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1159),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(1159),
														Column: int(14),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p12233,
												FreeVars: ast.Identifiers{
													"str",
												},
											},
											Id: "str",
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralNumber{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1159),
														Column: int(16),
													},
													End: ast.Location{
														Line: int(1159),
														Column: int(17),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p12233,
												FreeVars: nil,
											},
											OriginalString: "0",
										},
										CommaFodder: ast.Fodder{},
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Array{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1159),
														Column: int(19),
													},
													End: ast.Location{
														Line: int(1159),
														Column: int(21),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p12233,
												FreeVars: nil,
											},
											Elements: nil,
											TrailingComma: false,
											CloseFodder: ast.Fodder{},
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(1140),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1159),
					Column: int(22),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "base64Decode",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1161),
								Column: int(16),
							},
							End: ast.Location{
								Line: int(1161),
								Column: int(19),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1162),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1163),
								Column: int(58),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p12243,
						FreeVars: ast.Identifiers{
							"std",
							"str",
						},
					},
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Variable: "bytes",
							EqFodder: ast.Fodder{},
							Body: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1162),
											Column: int(19),
										},
										End: ast.Location{
											Line: int(1162),
											Column: int(45),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p12247,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1162),
												Column: int(19),
											},
											End: ast.Location{
												Line: int(1162),
												Column: int(40),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p12247,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1162),
													Column: int(19),
												},
												End: ast.Location{
													Line: int(1162),
													Column: int(22),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "base64DecodeBytes",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1162),
															Column: int(41),
														},
														End: ast.Location{
															Line: int(1162),
															Column: int(44),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p12256,
													FreeVars: ast.Identifiers{
														"str",
													},
												},
												Id: "str",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
							Fun: nil,
							CloseFodder: ast.Fodder{},
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1162),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(1162),
									Column: int(45),
								},
								File: p1,
							},
						},
					},
					Body: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1163),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1163),
									Column: int(58),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p12243,
							FreeVars: ast.Identifiers{
								"bytes",
								"std",
							},
						},
						Target: &ast.Index{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1163),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1163),
										Column: int(13),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p12243,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1163),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(1163),
											Column: int(8),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
											Comment: []string{},
										},
									},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Id: "std",
							},
							LeftBracketFodder: ast.Fodder{},
							Index: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Value: "join",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							RightBracketFodder: ast.Fodder{},
							Id: nil,
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1163),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(1163),
													Column: int(16),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p12268,
											FreeVars: nil,
										},
										Value: "",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1163),
													Column: int(18),
												},
												End: ast.Location{
													Line: int(1163),
													Column: int(57),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p12268,
											FreeVars: ast.Identifiers{
												"bytes",
												"std",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1163),
														Column: int(18),
													},
													End: ast.Location{
														Line: int(1163),
														Column: int(25),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p12268,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1163),
															Column: int(18),
														},
														End: ast.Location{
															Line: int(1163),
															Column: int(21),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: ast.Fodder{},
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "map",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: ast.Fodder{},
											Id: nil,
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Function{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1163),
																	Column: int(26),
																},
																End: ast.Location{
																	Line: int(1163),
																	Column: int(49),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p12278,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														ParenLeftFodder: ast.Fodder{},
														Parameters: []ast.Parameter{
															ast.Parameter{
																NameFodder: ast.Fodder{},
																Name: "b",
																EqFodder: nil,
																DefaultArg: nil,
																CommaFodder: nil,
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1163),
																		Column: int(35),
																	},
																	End: ast.Location{
																		Line: int(1163),
																		Column: int(36),
																	},
																	File: p1,
																},
															},
														},
														TrailingComma: false,
														ParenRightFodder: ast.Fodder{},
														Body: &ast.Apply{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1163),
																		Column: int(38),
																	},
																	End: ast.Location{
																		Line: int(1163),
																		Column: int(49),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p12282,
																FreeVars: ast.Identifiers{
																	"b",
																	"std",
																},
															},
															Target: &ast.Index{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1163),
																			Column: int(38),
																		},
																		End: ast.Location{
																			Line: int(1163),
																			Column: int(46),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p12282,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1163),
																				Column: int(38),
																			},
																			End: ast.Location{
																				Line: int(1163),
																				Column: int(41),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Id: "std",
																},
																LeftBracketFodder: ast.Fodder{},
																Index: &ast.LiteralString{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: nil,
																	},
																	Value: "char",
																	Kind: ast.LiteralStringKind(1),
																	BlockIndent: "",
																	BlockTermIndent: "",
																},
																RightBracketFodder: ast.Fodder{},
																Id: nil,
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1163),
																						Column: int(47),
																					},
																					End: ast.Location{
																						Line: int(1163),
																						Column: int(48),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p12291,
																				FreeVars: ast.Identifiers{
																					"b",
																				},
																			},
																			Id: "b",
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															TrailingComma: false,
															TailStrict: false,
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
														},
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1163),
																	Column: int(51),
																},
																End: ast.Location{
																	Line: int(1163),
																	Column: int(56),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p12278,
															FreeVars: ast.Identifiers{
																"bytes",
															},
														},
														Id: "bytes",
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(1161),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1163),
					Column: int(58),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "reverse",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1165),
								Column: int(11),
							},
							End: ast.Location{
								Line: int(1165),
								Column: int(14),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1166),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1167),
								Column: int(49),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p12301,
						FreeVars: ast.Identifiers{
							"arr",
							"std",
						},
					},
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Variable: "l",
							EqFodder: ast.Fodder{},
							Body: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1166),
											Column: int(15),
										},
										End: ast.Location{
											Line: int(1166),
											Column: int(30),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p12305,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1166),
												Column: int(15),
											},
											End: ast.Location{
												Line: int(1166),
												Column: int(25),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p12305,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1166),
													Column: int(15),
												},
												End: ast.Location{
													Line: int(1166),
													Column: int(18),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "length",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1166),
															Column: int(26),
														},
														End: ast.Location{
															Line: int(1166),
															Column: int(29),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p12314,
													FreeVars: ast.Identifiers{
														"arr",
													},
												},
												Id: "arr",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
							Fun: nil,
							CloseFodder: ast.Fodder{},
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1166),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(1166),
									Column: int(30),
								},
								File: p1,
							},
						},
					},
					Body: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1167),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1167),
									Column: int(49),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p12301,
							FreeVars: ast.Identifiers{
								"arr",
								"l",
								"std",
							},
						},
						Target: &ast.Index{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1167),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1167),
										Column: int(18),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p12301,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1167),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(1167),
											Column: int(8),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
											Comment: []string{},
										},
									},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Id: "std",
							},
							LeftBracketFodder: ast.Fodder{},
							Index: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Value: "makeArray",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							RightBracketFodder: ast.Fodder{},
							Id: nil,
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1167),
													Column: int(19),
												},
												End: ast.Location{
													Line: int(1167),
													Column: int(20),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p12326,
											FreeVars: ast.Identifiers{
												"l",
											},
										},
										Id: "l",
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Function{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1167),
													Column: int(22),
												},
												End: ast.Location{
													Line: int(1167),
													Column: int(48),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p12326,
											FreeVars: ast.Identifiers{
												"arr",
												"l",
											},
										},
										ParenLeftFodder: ast.Fodder{},
										Parameters: []ast.Parameter{
											ast.Parameter{
												NameFodder: ast.Fodder{},
												Name: "i",
												EqFodder: nil,
												DefaultArg: nil,
												CommaFodder: nil,
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1167),
														Column: int(31),
													},
													End: ast.Location{
														Line: int(1167),
														Column: int(32),
													},
													File: p1,
												},
											},
										},
										TrailingComma: false,
										ParenRightFodder: ast.Fodder{},
										Body: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1167),
														Column: int(34),
													},
													End: ast.Location{
														Line: int(1167),
														Column: int(48),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p12332,
												FreeVars: ast.Identifiers{
													"arr",
													"i",
													"l",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1167),
															Column: int(34),
														},
														End: ast.Location{
															Line: int(1167),
															Column: int(37),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p12332,
													FreeVars: ast.Identifiers{
														"arr",
													},
												},
												Id: "arr",
											},
											LeftBracketFodder: ast.Fodder{},
											Index: &ast.Binary{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1167),
															Column: int(38),
														},
														End: ast.Location{
															Line: int(1167),
															Column: int(47),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p12332,
													FreeVars: ast.Identifiers{
														"i",
														"l",
													},
												},
												Left: &ast.Binary{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1167),
																Column: int(38),
															},
															End: ast.Location{
																Line: int(1167),
																Column: int(43),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p12332,
														FreeVars: ast.Identifiers{
															"i",
															"l",
														},
													},
													Left: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1167),
																	Column: int(38),
																},
																End: ast.Location{
																	Line: int(1167),
																	Column: int(39),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p12332,
															FreeVars: ast.Identifiers{
																"l",
															},
														},
														Id: "l",
													},
													OpFodder: ast.Fodder{},
													Op: ast.BinaryOp(4),
													Right: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1167),
																	Column: int(42),
																},
																End: ast.Location{
																	Line: int(1167),
																	Column: int(43),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p12332,
															FreeVars: ast.Identifiers{
																"i",
															},
														},
														Id: "i",
													},
												},
												OpFodder: ast.Fodder{},
												Op: ast.BinaryOp(4),
												Right: &ast.LiteralNumber{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1167),
																Column: int(46),
															},
															End: ast.Location{
																Line: int(1167),
																Column: int(47),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p12332,
														FreeVars: nil,
													},
													OriginalString: "1",
												},
											},
											RightBracketFodder: ast.Fodder{},
											Id: nil,
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(1165),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1167),
					Column: int(49),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "sort",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"id",
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1170),
								Column: int(8),
							},
							End: ast.Location{
								Line: int(1170),
								Column: int(11),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "keyF",
						EqFodder: ast.Fodder{},
						DefaultArg: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1170),
										Column: int(18),
									},
									End: ast.Location{
										Line: int(1170),
										Column: int(20),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p12350,
								FreeVars: ast.Identifiers{
									"id",
								},
							},
							Id: "id",
						},
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1170),
								Column: int(13),
							},
							End: ast.Location{
								Line: int(1170),
								Column: int(20),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1171),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1203),
								Column: int(67),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p12350,
						FreeVars: ast.Identifiers{
							"arr",
							"id",
							"keyF",
							"std",
						},
					},
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: nil,
							Variable: "quickSort",
							EqFodder: nil,
							Body: &ast.Function{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1171),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(1181),
											Column: int(68),
										},
										File: p1,
									},
									Fodder: nil,
									Ctx: p12357,
									FreeVars: ast.Identifiers{
										"id",
										"quickSort",
										"std",
									},
								},
								ParenLeftFodder: ast.Fodder{},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "arr",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1171),
												Column: int(21),
											},
											End: ast.Location{
												Line: int(1171),
												Column: int(24),
											},
											File: p1,
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "keyF",
										EqFodder: ast.Fodder{},
										DefaultArg: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1171),
														Column: int(31),
													},
													End: ast.Location{
														Line: int(1171),
														Column: int(33),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p12361,
												FreeVars: ast.Identifiers{
													"id",
												},
											},
											Id: "id",
										},
										CommaFodder: nil,
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1171),
												Column: int(26),
											},
											End: ast.Location{
												Line: int(1171),
												Column: int(33),
											},
											File: p1,
										},
									},
								},
								TrailingComma: false,
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Local{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1172),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1181),
												Column: int(68),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										Ctx: p12361,
										FreeVars: ast.Identifiers{
											"arr",
											"keyF",
											"quickSort",
											"std",
										},
									},
									Binds: ast.LocalBinds{
										ast.LocalBind{
											VarFodder: ast.Fodder{},
											Variable: "l",
											EqFodder: ast.Fodder{},
											Body: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1172),
															Column: int(17),
														},
														End: ast.Location{
															Line: int(1172),
															Column: int(32),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p12368,
													FreeVars: ast.Identifiers{
														"arr",
														"std",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1172),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(1172),
																Column: int(27),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p12368,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1172),
																	Column: int(17),
																},
																End: ast.Location{
																	Line: int(1172),
																	Column: int(20),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: ast.Fodder{},
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "length",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: ast.Fodder{},
													Id: nil,
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1172),
																			Column: int(28),
																		},
																		End: ast.Location{
																			Line: int(1172),
																			Column: int(31),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p12377,
																	FreeVars: ast.Identifiers{
																		"arr",
																	},
																},
																Id: "arr",
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
											Fun: nil,
											CloseFodder: ast.Fodder{},
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1172),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(1172),
													Column: int(32),
												},
												File: p1,
											},
										},
									},
									Body: &ast.Conditional{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1173),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(1181),
													Column: int(68),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
													Comment: []string{},
												},
											},
											Ctx: p12361,
											FreeVars: ast.Identifiers{
												"arr",
												"keyF",
												"l",
												"quickSort",
												"std",
											},
										},
										Cond: &ast.Binary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1173),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(1173),
														Column: int(30),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p12361,
												FreeVars: ast.Identifiers{
													"arr",
													"std",
												},
											},
											Left: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1173),
															Column: int(10),
														},
														End: ast.Location{
															Line: int(1173),
															Column: int(25),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p12361,
													FreeVars: ast.Identifiers{
														"arr",
														"std",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1173),
																Column: int(10),
															},
															End: ast.Location{
																Line: int(1173),
																Column: int(20),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p12361,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1173),
																	Column: int(10),
																},
																End: ast.Location{
																	Line: int(1173),
																	Column: int(13),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: ast.Fodder{},
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "length",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: ast.Fodder{},
													Id: nil,
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1173),
																			Column: int(21),
																		},
																		End: ast.Location{
																			Line: int(1173),
																			Column: int(24),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p12393,
																	FreeVars: ast.Identifiers{
																		"arr",
																	},
																},
																Id: "arr",
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
											OpFodder: ast.Fodder{},
											Op: ast.BinaryOp(10),
											Right: &ast.LiteralNumber{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1173),
															Column: int(29),
														},
														End: ast.Location{
															Line: int(1173),
															Column: int(30),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p12361,
													FreeVars: nil,
												},
												OriginalString: "1",
											},
										},
										ThenFodder: ast.Fodder{},
										BranchTrue: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1174),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1174),
														Column: int(12),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
														Comment: []string{},
													},
												},
												Ctx: p12361,
												FreeVars: ast.Identifiers{
													"arr",
												},
											},
											Id: "arr",
										},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										BranchFalse: &ast.Local{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1176),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1181),
														Column: int(68),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
														Comment: []string{},
													},
												},
												Ctx: p12361,
												FreeVars: ast.Identifiers{
													"arr",
													"keyF",
													"l",
													"quickSort",
													"std",
												},
											},
											Binds: ast.LocalBinds{
												ast.LocalBind{
													VarFodder: ast.Fodder{},
													Variable: "pos",
													EqFodder: ast.Fodder{},
													Body: &ast.LiteralNumber{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1176),
																	Column: int(21),
																},
																End: ast.Location{
																	Line: int(1176),
																	Column: int(22),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p12405,
															FreeVars: nil,
														},
														OriginalString: "0",
													},
													Fun: nil,
													CloseFodder: ast.Fodder{},
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1176),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1176),
															Column: int(22),
														},
														File: p1,
													},
												},
											},
											Body: &ast.Local{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1177),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1181),
															Column: int(68),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
															Comment: []string{},
														},
													},
													Ctx: p12361,
													FreeVars: ast.Identifiers{
														"arr",
														"keyF",
														"l",
														"pos",
														"quickSort",
														"std",
													},
												},
												Binds: ast.LocalBinds{
													ast.LocalBind{
														VarFodder: ast.Fodder{},
														Variable: "pivot",
														EqFodder: ast.Fodder{},
														Body: &ast.Apply{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1177),
																		Column: int(23),
																	},
																	End: ast.Location{
																		Line: int(1177),
																		Column: int(37),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p12411,
																FreeVars: ast.Identifiers{
																	"arr",
																	"keyF",
																	"pos",
																},
															},
															Target: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1177),
																			Column: int(23),
																		},
																		End: ast.Location{
																			Line: int(1177),
																			Column: int(27),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p12411,
																	FreeVars: ast.Identifiers{
																		"keyF",
																	},
																},
																Id: "keyF",
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Index{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1177),
																						Column: int(28),
																					},
																					End: ast.Location{
																						Line: int(1177),
																						Column: int(36),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p12417,
																				FreeVars: ast.Identifiers{
																					"arr",
																					"pos",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1177),
																							Column: int(28),
																						},
																						End: ast.Location{
																							Line: int(1177),
																							Column: int(31),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p12417,
																					FreeVars: ast.Identifiers{
																						"arr",
																					},
																				},
																				Id: "arr",
																			},
																			LeftBracketFodder: ast.Fodder{},
																			Index: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1177),
																							Column: int(32),
																						},
																						End: ast.Location{
																							Line: int(1177),
																							Column: int(35),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p12417,
																					FreeVars: ast.Identifiers{
																						"pos",
																					},
																				},
																				Id: "pos",
																			},
																			RightBracketFodder: ast.Fodder{},
																			Id: nil,
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															TrailingComma: false,
															TailStrict: false,
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
														},
														Fun: nil,
														CloseFodder: ast.Fodder{},
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1177),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1177),
																Column: int(37),
															},
															File: p1,
														},
													},
												},
												Body: &ast.Local{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1178),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(1181),
																Column: int(68),
															},
															File: p1,
														},
														Fodder: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(8),
																Comment: []string{},
															},
														},
														Ctx: p12361,
														FreeVars: ast.Identifiers{
															"arr",
															"keyF",
															"l",
															"pivot",
															"pos",
															"quickSort",
															"std",
														},
													},
													Binds: ast.LocalBinds{
														ast.LocalBind{
															VarFodder: ast.Fodder{},
															Variable: "rest",
															EqFodder: ast.Fodder{},
															Body: &ast.Apply{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1178),
																			Column: int(22),
																		},
																		End: ast.Location{
																			Line: int(1178),
																			Column: int(94),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p12428,
																	FreeVars: ast.Identifiers{
																		"arr",
																		"l",
																		"pos",
																		"std",
																	},
																},
																Target: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1178),
																				Column: int(22),
																			},
																			End: ast.Location{
																				Line: int(1178),
																				Column: int(35),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p12428,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1178),
																					Column: int(22),
																				},
																				End: ast.Location{
																					Line: int(1178),
																					Column: int(25),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Id: "std",
																	},
																	LeftBracketFodder: ast.Fodder{},
																	Index: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: nil,
																		},
																		Value: "makeArray",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	RightBracketFodder: ast.Fodder{},
																	Id: nil,
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Binary{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1178),
																							Column: int(36),
																						},
																						End: ast.Location{
																							Line: int(1178),
																							Column: int(41),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p12437,
																					FreeVars: ast.Identifiers{
																						"l",
																					},
																				},
																				Left: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1178),
																								Column: int(36),
																							},
																							End: ast.Location{
																								Line: int(1178),
																								Column: int(37),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p12437,
																						FreeVars: ast.Identifiers{
																							"l",
																						},
																					},
																					Id: "l",
																				},
																				OpFodder: ast.Fodder{},
																				Op: ast.BinaryOp(4),
																				Right: &ast.LiteralNumber{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1178),
																								Column: int(40),
																							},
																							End: ast.Location{
																								Line: int(1178),
																								Column: int(41),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p12437,
																						FreeVars: nil,
																					},
																					OriginalString: "1",
																				},
																			},
																			CommaFodder: ast.Fodder{},
																		},
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Function{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1178),
																							Column: int(43),
																						},
																						End: ast.Location{
																							Line: int(1178),
																							Column: int(93),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p12437,
																					FreeVars: ast.Identifiers{
																						"arr",
																						"pos",
																					},
																				},
																				ParenLeftFodder: ast.Fodder{},
																				Parameters: []ast.Parameter{
																					ast.Parameter{
																						NameFodder: ast.Fodder{},
																						Name: "i",
																						EqFodder: nil,
																						DefaultArg: nil,
																						CommaFodder: nil,
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1178),
																								Column: int(52),
																							},
																							End: ast.Location{
																								Line: int(1178),
																								Column: int(53),
																							},
																							File: p1,
																						},
																					},
																				},
																				TrailingComma: false,
																				ParenRightFodder: ast.Fodder{},
																				Body: &ast.Conditional{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1178),
																								Column: int(55),
																							},
																							End: ast.Location{
																								Line: int(1178),
																								Column: int(93),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p12446,
																						FreeVars: ast.Identifiers{
																							"arr",
																							"i",
																							"pos",
																						},
																					},
																					Cond: &ast.Binary{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1178),
																									Column: int(58),
																								},
																								End: ast.Location{
																									Line: int(1178),
																									Column: int(65),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12446,
																							FreeVars: ast.Identifiers{
																								"i",
																								"pos",
																							},
																						},
																						Left: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1178),
																										Column: int(58),
																									},
																									End: ast.Location{
																										Line: int(1178),
																										Column: int(59),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p12446,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																							},
																							Id: "i",
																						},
																						OpFodder: ast.Fodder{},
																						Op: ast.BinaryOp(9),
																						Right: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1178),
																										Column: int(62),
																									},
																									End: ast.Location{
																										Line: int(1178),
																										Column: int(65),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p12446,
																								FreeVars: ast.Identifiers{
																									"pos",
																								},
																							},
																							Id: "pos",
																						},
																					},
																					ThenFodder: ast.Fodder{},
																					BranchTrue: &ast.Index{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1178),
																									Column: int(71),
																								},
																								End: ast.Location{
																									Line: int(1178),
																									Column: int(77),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12446,
																							FreeVars: ast.Identifiers{
																								"arr",
																								"i",
																							},
																						},
																						Target: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1178),
																										Column: int(71),
																									},
																									End: ast.Location{
																										Line: int(1178),
																										Column: int(74),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p12446,
																								FreeVars: ast.Identifiers{
																									"arr",
																								},
																							},
																							Id: "arr",
																						},
																						LeftBracketFodder: ast.Fodder{},
																						Index: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1178),
																										Column: int(75),
																									},
																									End: ast.Location{
																										Line: int(1178),
																										Column: int(76),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p12446,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																							},
																							Id: "i",
																						},
																						RightBracketFodder: ast.Fodder{},
																						Id: nil,
																					},
																					ElseFodder: ast.Fodder{},
																					BranchFalse: &ast.Index{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1178),
																									Column: int(83),
																								},
																								End: ast.Location{
																									Line: int(1178),
																									Column: int(93),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12446,
																							FreeVars: ast.Identifiers{
																								"arr",
																								"i",
																							},
																						},
																						Target: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1178),
																										Column: int(83),
																									},
																									End: ast.Location{
																										Line: int(1178),
																										Column: int(86),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p12446,
																								FreeVars: ast.Identifiers{
																									"arr",
																								},
																							},
																							Id: "arr",
																						},
																						LeftBracketFodder: ast.Fodder{},
																						Index: &ast.Binary{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1178),
																										Column: int(87),
																									},
																									End: ast.Location{
																										Line: int(1178),
																										Column: int(92),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p12446,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																							},
																							Left: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1178),
																											Column: int(87),
																										},
																										End: ast.Location{
																											Line: int(1178),
																											Column: int(88),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p12446,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																								},
																								Id: "i",
																							},
																							OpFodder: ast.Fodder{},
																							Op: ast.BinaryOp(3),
																							Right: &ast.LiteralNumber{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1178),
																											Column: int(91),
																										},
																										End: ast.Location{
																											Line: int(1178),
																											Column: int(92),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p12446,
																									FreeVars: nil,
																								},
																								OriginalString: "1",
																							},
																						},
																						RightBracketFodder: ast.Fodder{},
																						Id: nil,
																					},
																				},
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																TrailingComma: false,
																TailStrict: false,
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
															},
															Fun: nil,
															CloseFodder: ast.Fodder{},
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1178),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(1178),
																	Column: int(94),
																},
																File: p1,
															},
														},
													},
													Body: &ast.Local{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1179),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(1181),
																	Column: int(68),
																},
																File: p1,
															},
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																	Comment: []string{},
																},
															},
															Ctx: p12361,
															FreeVars: ast.Identifiers{
																"arr",
																"keyF",
																"pivot",
																"pos",
																"quickSort",
																"rest",
																"std",
															},
														},
														Binds: ast.LocalBinds{
															ast.LocalBind{
																VarFodder: ast.Fodder{},
																Variable: "left",
																EqFodder: ast.Fodder{},
																Body: &ast.Apply{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1179),
																				Column: int(22),
																			},
																			End: ast.Location{
																				Line: int(1179),
																				Column: int(67),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p12474,
																		FreeVars: ast.Identifiers{
																			"keyF",
																			"pivot",
																			"rest",
																			"std",
																		},
																	},
																	Target: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1179),
																					Column: int(22),
																				},
																				End: ast.Location{
																					Line: int(1179),
																					Column: int(32),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p12474,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1179),
																						Column: int(22),
																					},
																					End: ast.Location{
																						Line: int(1179),
																						Column: int(25),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Id: "std",
																		},
																		LeftBracketFodder: ast.Fodder{},
																		Index: &ast.LiteralString{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: nil,
																			},
																			Value: "filter",
																			Kind: ast.LiteralStringKind(1),
																			BlockIndent: "",
																			BlockTermIndent: "",
																		},
																		RightBracketFodder: ast.Fodder{},
																		Id: nil,
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Function{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1179),
																								Column: int(33),
																							},
																							End: ast.Location{
																								Line: int(1179),
																								Column: int(60),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p12483,
																						FreeVars: ast.Identifiers{
																							"keyF",
																							"pivot",
																						},
																					},
																					ParenLeftFodder: ast.Fodder{},
																					Parameters: []ast.Parameter{
																						ast.Parameter{
																							NameFodder: ast.Fodder{},
																							Name: "x",
																							EqFodder: nil,
																							DefaultArg: nil,
																							CommaFodder: nil,
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1179),
																									Column: int(42),
																								},
																								End: ast.Location{
																									Line: int(1179),
																									Column: int(43),
																								},
																								File: p1,
																							},
																						},
																					},
																					TrailingComma: false,
																					ParenRightFodder: ast.Fodder{},
																					Body: &ast.Binary{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1179),
																									Column: int(45),
																								},
																								End: ast.Location{
																									Line: int(1179),
																									Column: int(60),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12487,
																							FreeVars: ast.Identifiers{
																								"keyF",
																								"pivot",
																								"x",
																							},
																						},
																						Left: &ast.Apply{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1179),
																										Column: int(45),
																									},
																									End: ast.Location{
																										Line: int(1179),
																										Column: int(52),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p12487,
																								FreeVars: ast.Identifiers{
																									"keyF",
																									"x",
																								},
																							},
																							Target: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1179),
																											Column: int(45),
																										},
																										End: ast.Location{
																											Line: int(1179),
																											Column: int(49),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p12487,
																									FreeVars: ast.Identifiers{
																										"keyF",
																									},
																								},
																								Id: "keyF",
																							},
																							FodderLeft: ast.Fodder{},
																							Arguments: ast.Arguments{
																								Positional: []ast.CommaSeparatedExpr{
																									ast.CommaSeparatedExpr{
																										Expr: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1179),
																														Column: int(50),
																													},
																													End: ast.Location{
																														Line: int(1179),
																														Column: int(51),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p12495,
																												FreeVars: ast.Identifiers{
																													"x",
																												},
																											},
																											Id: "x",
																										},
																										CommaFodder: nil,
																									},
																								},
																								Named: nil,
																							},
																							TrailingComma: false,
																							TailStrict: false,
																							FodderRight: ast.Fodder{},
																							TailStrictFodder: nil,
																						},
																						OpFodder: ast.Fodder{},
																						Op: ast.BinaryOp(9),
																						Right: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1179),
																										Column: int(55),
																									},
																									End: ast.Location{
																										Line: int(1179),
																										Column: int(60),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p12487,
																								FreeVars: ast.Identifiers{
																									"pivot",
																								},
																							},
																							Id: "pivot",
																						},
																					},
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1179),
																								Column: int(62),
																							},
																							End: ast.Location{
																								Line: int(1179),
																								Column: int(66),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p12483,
																						FreeVars: ast.Identifiers{
																							"rest",
																						},
																					},
																					Id: "rest",
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	TrailingComma: false,
																	TailStrict: false,
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																},
																Fun: nil,
																CloseFodder: ast.Fodder{},
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1179),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(1179),
																		Column: int(67),
																	},
																	File: p1,
																},
															},
														},
														Body: &ast.Local{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1180),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(1181),
																		Column: int(68),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																		Comment: []string{},
																	},
																},
																Ctx: p12361,
																FreeVars: ast.Identifiers{
																	"arr",
																	"keyF",
																	"left",
																	"pivot",
																	"pos",
																	"quickSort",
																	"rest",
																	"std",
																},
															},
															Binds: ast.LocalBinds{
																ast.LocalBind{
																	VarFodder: ast.Fodder{},
																	Variable: "right",
																	EqFodder: ast.Fodder{},
																	Body: &ast.Apply{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1180),
																					Column: int(23),
																				},
																				End: ast.Location{
																					Line: int(1180),
																					Column: int(69),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p12506,
																			FreeVars: ast.Identifiers{
																				"keyF",
																				"pivot",
																				"rest",
																				"std",
																			},
																		},
																		Target: &ast.Index{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1180),
																						Column: int(23),
																					},
																					End: ast.Location{
																						Line: int(1180),
																						Column: int(33),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p12506,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1180),
																							Column: int(23),
																						},
																						End: ast.Location{
																							Line: int(1180),
																							Column: int(26),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																				},
																				Id: "std",
																			},
																			LeftBracketFodder: ast.Fodder{},
																			Index: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: nil,
																				},
																				Value: "filter",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																			RightBracketFodder: ast.Fodder{},
																			Id: nil,
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Function{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1180),
																									Column: int(34),
																								},
																								End: ast.Location{
																									Line: int(1180),
																									Column: int(62),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12515,
																							FreeVars: ast.Identifiers{
																								"keyF",
																								"pivot",
																							},
																						},
																						ParenLeftFodder: ast.Fodder{},
																						Parameters: []ast.Parameter{
																							ast.Parameter{
																								NameFodder: ast.Fodder{},
																								Name: "x",
																								EqFodder: nil,
																								DefaultArg: nil,
																								CommaFodder: nil,
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1180),
																										Column: int(43),
																									},
																									End: ast.Location{
																										Line: int(1180),
																										Column: int(44),
																									},
																									File: p1,
																								},
																							},
																						},
																						TrailingComma: false,
																						ParenRightFodder: ast.Fodder{},
																						Body: &ast.Binary{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1180),
																										Column: int(46),
																									},
																									End: ast.Location{
																										Line: int(1180),
																										Column: int(62),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p12519,
																								FreeVars: ast.Identifiers{
																									"keyF",
																									"pivot",
																									"x",
																								},
																							},
																							Left: &ast.Apply{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1180),
																											Column: int(46),
																										},
																										End: ast.Location{
																											Line: int(1180),
																											Column: int(53),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p12519,
																									FreeVars: ast.Identifiers{
																										"keyF",
																										"x",
																									},
																								},
																								Target: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1180),
																												Column: int(46),
																											},
																											End: ast.Location{
																												Line: int(1180),
																												Column: int(50),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p12519,
																										FreeVars: ast.Identifiers{
																											"keyF",
																										},
																									},
																									Id: "keyF",
																								},
																								FodderLeft: ast.Fodder{},
																								Arguments: ast.Arguments{
																									Positional: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Var{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(1180),
																															Column: int(51),
																														},
																														End: ast.Location{
																															Line: int(1180),
																															Column: int(52),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p12527,
																													FreeVars: ast.Identifiers{
																														"x",
																													},
																												},
																												Id: "x",
																											},
																											CommaFodder: nil,
																										},
																									},
																									Named: nil,
																								},
																								TrailingComma: false,
																								TailStrict: false,
																								FodderRight: ast.Fodder{},
																								TailStrictFodder: nil,
																							},
																							OpFodder: ast.Fodder{},
																							Op: ast.BinaryOp(8),
																							Right: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1180),
																											Column: int(57),
																										},
																										End: ast.Location{
																											Line: int(1180),
																											Column: int(62),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p12519,
																									FreeVars: ast.Identifiers{
																										"pivot",
																									},
																								},
																								Id: "pivot",
																							},
																						},
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1180),
																									Column: int(64),
																								},
																								End: ast.Location{
																									Line: int(1180),
																									Column: int(68),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12515,
																							FreeVars: ast.Identifiers{
																								"rest",
																							},
																						},
																						Id: "rest",
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		TrailingComma: false,
																		TailStrict: false,
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																	},
																	Fun: nil,
																	CloseFodder: ast.Fodder{},
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1180),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(1180),
																			Column: int(69),
																		},
																		File: p1,
																	},
																},
															},
															Body: &ast.Binary{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1181),
																			Column: int(9),
																		},
																		End: ast.Location{
																			Line: int(1181),
																			Column: int(68),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p12361,
																	FreeVars: ast.Identifiers{
																		"arr",
																		"keyF",
																		"left",
																		"pos",
																		"quickSort",
																		"right",
																	},
																},
																Left: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1181),
																				Column: int(9),
																			},
																			End: ast.Location{
																				Line: int(1181),
																				Column: int(43),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p12361,
																		FreeVars: ast.Identifiers{
																			"arr",
																			"keyF",
																			"left",
																			"pos",
																			"quickSort",
																		},
																	},
																	Left: &ast.Apply{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1181),
																					Column: int(9),
																				},
																				End: ast.Location{
																					Line: int(1181),
																					Column: int(30),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p12361,
																			FreeVars: ast.Identifiers{
																				"keyF",
																				"left",
																				"quickSort",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1181),
																						Column: int(9),
																					},
																					End: ast.Location{
																						Line: int(1181),
																						Column: int(18),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(8),
																						Comment: []string{},
																					},
																				},
																				Ctx: p12361,
																				FreeVars: ast.Identifiers{
																					"quickSort",
																				},
																			},
																			Id: "quickSort",
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1181),
																									Column: int(19),
																								},
																								End: ast.Location{
																									Line: int(1181),
																									Column: int(23),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12544,
																							FreeVars: ast.Identifiers{
																								"left",
																							},
																						},
																						Id: "left",
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1181),
																									Column: int(25),
																								},
																								End: ast.Location{
																									Line: int(1181),
																									Column: int(29),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12544,
																							FreeVars: ast.Identifiers{
																								"keyF",
																							},
																						},
																						Id: "keyF",
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		TrailingComma: false,
																		TailStrict: false,
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(3),
																	Right: &ast.Array{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1181),
																					Column: int(33),
																				},
																				End: ast.Location{
																					Line: int(1181),
																					Column: int(43),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p12361,
																			FreeVars: ast.Identifiers{
																				"arr",
																				"pos",
																			},
																		},
																		Elements: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Index{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1181),
																								Column: int(34),
																							},
																							End: ast.Location{
																								Line: int(1181),
																								Column: int(42),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p12552,
																						FreeVars: ast.Identifiers{
																							"arr",
																							"pos",
																						},
																					},
																					Target: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1181),
																									Column: int(34),
																								},
																								End: ast.Location{
																									Line: int(1181),
																									Column: int(37),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12552,
																							FreeVars: ast.Identifiers{
																								"arr",
																							},
																						},
																						Id: "arr",
																					},
																					LeftBracketFodder: ast.Fodder{},
																					Index: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1181),
																									Column: int(38),
																								},
																								End: ast.Location{
																									Line: int(1181),
																									Column: int(41),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12552,
																							FreeVars: ast.Identifiers{
																								"pos",
																							},
																						},
																						Id: "pos",
																					},
																					RightBracketFodder: ast.Fodder{},
																					Id: nil,
																				},
																				CommaFodder: nil,
																			},
																		},
																		TrailingComma: false,
																		CloseFodder: ast.Fodder{},
																	},
																},
																OpFodder: ast.Fodder{},
																Op: ast.BinaryOp(3),
																Right: &ast.Apply{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1181),
																				Column: int(46),
																			},
																			End: ast.Location{
																				Line: int(1181),
																				Column: int(68),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p12361,
																		FreeVars: ast.Identifiers{
																			"keyF",
																			"quickSort",
																			"right",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1181),
																					Column: int(46),
																				},
																				End: ast.Location{
																					Line: int(1181),
																					Column: int(55),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p12361,
																			FreeVars: ast.Identifiers{
																				"quickSort",
																			},
																		},
																		Id: "quickSort",
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1181),
																								Column: int(56),
																							},
																							End: ast.Location{
																								Line: int(1181),
																								Column: int(61),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p12564,
																						FreeVars: ast.Identifiers{
																							"right",
																						},
																					},
																					Id: "right",
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1181),
																								Column: int(63),
																							},
																							End: ast.Location{
																								Line: int(1181),
																								Column: int(67),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p12564,
																						FreeVars: ast.Identifiers{
																							"keyF",
																						},
																					},
																					Id: "keyF",
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	TrailingComma: false,
																	TailStrict: false,
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																},
															},
														},
													},
												},
											},
										},
									},
								},
							},
							Fun: nil,
							CloseFodder: nil,
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
						},
					},
					Body: &ast.Local{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1183),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1203),
									Column: int(67),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(1),
									Indent: int(4),
									Comment: []string{},
								},
							},
							Ctx: p12350,
							FreeVars: ast.Identifiers{
								"arr",
								"keyF",
								"quickSort",
								"std",
							},
						},
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: nil,
								Variable: "merge",
								EqFodder: nil,
								Body: &ast.Function{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1183),
												Column: int(11),
											},
											End: ast.Location{
												Line: int(1195),
												Column: int(20),
											},
											File: p1,
										},
										Fodder: nil,
										Ctx: p12573,
										FreeVars: ast.Identifiers{
											"keyF",
											"std",
										},
									},
									ParenLeftFodder: ast.Fodder{},
									Parameters: []ast.Parameter{
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "a",
											EqFodder: nil,
											DefaultArg: nil,
											CommaFodder: ast.Fodder{},
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1183),
													Column: int(17),
												},
												End: ast.Location{
													Line: int(1183),
													Column: int(18),
												},
												File: p1,
											},
										},
										ast.Parameter{
											NameFodder: ast.Fodder{},
											Name: "b",
											EqFodder: nil,
											DefaultArg: nil,
											CommaFodder: nil,
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1183),
													Column: int(20),
												},
												End: ast.Location{
													Line: int(1183),
													Column: int(21),
												},
												File: p1,
											},
										},
									},
									TrailingComma: false,
									ParenRightFodder: ast.Fodder{},
									Body: &ast.Local{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1184),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(1195),
													Column: int(20),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
													Comment: []string{},
												},
											},
											Ctx: p12578,
											FreeVars: ast.Identifiers{
												"a",
												"b",
												"keyF",
												"std",
											},
										},
										Binds: ast.LocalBinds{
											ast.LocalBind{
												VarFodder: ast.Fodder{},
												Variable: "la",
												EqFodder: ast.Fodder{},
												Body: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1184),
																Column: int(18),
															},
															End: ast.Location{
																Line: int(1184),
																Column: int(31),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p12582,
														FreeVars: ast.Identifiers{
															"a",
															"std",
														},
													},
													Target: &ast.Index{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1184),
																	Column: int(18),
																},
																End: ast.Location{
																	Line: int(1184),
																	Column: int(28),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p12582,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1184),
																		Column: int(18),
																	},
																	End: ast.Location{
																		Line: int(1184),
																		Column: int(21),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Id: "std",
														},
														LeftBracketFodder: ast.Fodder{},
														Index: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
															Value: "length",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														RightBracketFodder: ast.Fodder{},
														Id: nil,
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1184),
																				Column: int(29),
																			},
																			End: ast.Location{
																				Line: int(1184),
																				Column: int(30),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p12591,
																		FreeVars: ast.Identifiers{
																			"a",
																		},
																	},
																	Id: "a",
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
												},
												Fun: nil,
												CloseFodder: ast.Fodder{},
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1184),
														Column: int(13),
													},
													End: ast.Location{
														Line: int(1184),
														Column: int(31),
													},
													File: p1,
												},
											},
											ast.LocalBind{
												VarFodder: ast.Fodder{},
												Variable: "lb",
												EqFodder: ast.Fodder{},
												Body: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1184),
																Column: int(38),
															},
															End: ast.Location{
																Line: int(1184),
																Column: int(51),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p12594,
														FreeVars: ast.Identifiers{
															"b",
															"std",
														},
													},
													Target: &ast.Index{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1184),
																	Column: int(38),
																},
																End: ast.Location{
																	Line: int(1184),
																	Column: int(48),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p12594,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1184),
																		Column: int(38),
																	},
																	End: ast.Location{
																		Line: int(1184),
																		Column: int(41),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Id: "std",
														},
														LeftBracketFodder: ast.Fodder{},
														Index: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
															Value: "length",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														RightBracketFodder: ast.Fodder{},
														Id: nil,
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1184),
																				Column: int(49),
																			},
																			End: ast.Location{
																				Line: int(1184),
																				Column: int(50),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p12603,
																		FreeVars: ast.Identifiers{
																			"b",
																		},
																	},
																	Id: "b",
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
												},
												Fun: nil,
												CloseFodder: ast.Fodder{},
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1184),
														Column: int(33),
													},
													End: ast.Location{
														Line: int(1184),
														Column: int(51),
													},
													File: p1,
												},
											},
										},
										Body: &ast.Local{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1185),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(1195),
														Column: int(20),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(6),
														Comment: []string{},
													},
												},
												Ctx: p12578,
												FreeVars: ast.Identifiers{
													"a",
													"b",
													"keyF",
													"la",
													"lb",
													"std",
												},
											},
											Binds: ast.LocalBinds{
												ast.LocalBind{
													VarFodder: nil,
													Variable: "aux",
													EqFodder: nil,
													Body: &ast.Function{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1185),
																	Column: int(13),
																},
																End: ast.Location{
																	Line: int(1194),
																	Column: int(43),
																},
																File: p1,
															},
															Fodder: nil,
															Ctx: p12610,
															FreeVars: ast.Identifiers{
																"a",
																"aux",
																"b",
																"keyF",
																"la",
																"lb",
																"std",
															},
														},
														ParenLeftFodder: ast.Fodder{},
														Parameters: []ast.Parameter{
															ast.Parameter{
																NameFodder: ast.Fodder{},
																Name: "i",
																EqFodder: nil,
																DefaultArg: nil,
																CommaFodder: ast.Fodder{},
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1185),
																		Column: int(17),
																	},
																	End: ast.Location{
																		Line: int(1185),
																		Column: int(18),
																	},
																	File: p1,
																},
															},
															ast.Parameter{
																NameFodder: ast.Fodder{},
																Name: "j",
																EqFodder: nil,
																DefaultArg: nil,
																CommaFodder: ast.Fodder{},
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1185),
																		Column: int(20),
																	},
																	End: ast.Location{
																		Line: int(1185),
																		Column: int(21),
																	},
																	File: p1,
																},
															},
															ast.Parameter{
																NameFodder: ast.Fodder{},
																Name: "prefix",
																EqFodder: nil,
																DefaultArg: nil,
																CommaFodder: nil,
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1185),
																		Column: int(23),
																	},
																	End: ast.Location{
																		Line: int(1185),
																		Column: int(29),
																	},
																	File: p1,
																},
															},
														},
														TrailingComma: false,
														ParenRightFodder: ast.Fodder{},
														Body: &ast.Conditional{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1186),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(1194),
																		Column: int(43),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																		Comment: []string{},
																	},
																},
																Ctx: p12615,
																FreeVars: ast.Identifiers{
																	"a",
																	"aux",
																	"b",
																	"i",
																	"j",
																	"keyF",
																	"la",
																	"lb",
																	"prefix",
																	"std",
																},
															},
															Cond: &ast.Binary{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1186),
																			Column: int(12),
																		},
																		End: ast.Location{
																			Line: int(1186),
																			Column: int(19),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p12615,
																	FreeVars: ast.Identifiers{
																		"i",
																		"la",
																	},
																},
																Left: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1186),
																				Column: int(12),
																			},
																			End: ast.Location{
																				Line: int(1186),
																				Column: int(13),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p12615,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																	},
																	Id: "i",
																},
																OpFodder: ast.Fodder{},
																Op: ast.BinaryOp(12),
																Right: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1186),
																				Column: int(17),
																			},
																			End: ast.Location{
																				Line: int(1186),
																				Column: int(19),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p12615,
																		FreeVars: ast.Identifiers{
																			"la",
																		},
																	},
																	Id: "la",
																},
															},
															ThenFodder: ast.Fodder{},
															BranchTrue: &ast.Binary{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1187),
																			Column: int(11),
																		},
																		End: ast.Location{
																			Line: int(1187),
																			Column: int(25),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p12615,
																	FreeVars: ast.Identifiers{
																		"b",
																		"j",
																		"prefix",
																		"std",
																	},
																},
																Left: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1187),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(1187),
																				Column: int(17),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(10),
																				Comment: []string{},
																			},
																		},
																		Ctx: p12615,
																		FreeVars: ast.Identifiers{
																			"prefix",
																		},
																	},
																	Id: "prefix",
																},
																OpFodder: ast.Fodder{},
																Op: ast.BinaryOp(3),
																Right: &ast.Apply{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"b",
																			"j",
																			"std",
																		},
																	},
																	Target: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Id: "std",
																		},
																		LeftBracketFodder: nil,
																		Index: &ast.LiteralString{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: nil,
																			},
																			Value: "slice",
																			Kind: ast.LiteralStringKind(1),
																			BlockIndent: "",
																			BlockTermIndent: "",
																		},
																		RightBracketFodder: nil,
																		Id: nil,
																	},
																	FodderLeft: nil,
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1187),
																								Column: int(20),
																							},
																							End: ast.Location{
																								Line: int(1187),
																								Column: int(21),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p12615,
																						FreeVars: ast.Identifiers{
																							"b",
																						},
																					},
																					Id: "b",
																				},
																				CommaFodder: nil,
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1187),
																								Column: int(22),
																							},
																							End: ast.Location{
																								Line: int(1187),
																								Column: int(23),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p12615,
																						FreeVars: ast.Identifiers{
																							"j",
																						},
																					},
																					Id: "j",
																				},
																				CommaFodder: nil,
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.LiteralNull{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							File: nil,
																						},
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: nil,
																					},
																				},
																				CommaFodder: nil,
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.LiteralNull{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "",
																							Begin: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							End: ast.Location{
																								Line: int(0),
																								Column: int(0),
																							},
																							File: nil,
																						},
																						Fodder: nil,
																						Ctx: nil,
																						FreeVars: nil,
																					},
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	TrailingComma: false,
																	TailStrict: false,
																	FodderRight: nil,
																	TailStrictFodder: nil,
																},
															},
															ElseFodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																	Comment: []string{},
																},
															},
															BranchFalse: &ast.Conditional{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1188),
																			Column: int(14),
																		},
																		End: ast.Location{
																			Line: int(1194),
																			Column: int(43),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p12615,
																	FreeVars: ast.Identifiers{
																		"a",
																		"aux",
																		"b",
																		"i",
																		"j",
																		"keyF",
																		"lb",
																		"prefix",
																		"std",
																	},
																},
																Cond: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1188),
																				Column: int(17),
																			},
																			End: ast.Location{
																				Line: int(1188),
																				Column: int(24),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p12615,
																		FreeVars: ast.Identifiers{
																			"j",
																			"lb",
																		},
																	},
																	Left: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1188),
																					Column: int(17),
																				},
																				End: ast.Location{
																					Line: int(1188),
																					Column: int(18),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p12615,
																			FreeVars: ast.Identifiers{
																				"j",
																			},
																		},
																		Id: "j",
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(12),
																	Right: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1188),
																					Column: int(22),
																				},
																				End: ast.Location{
																					Line: int(1188),
																					Column: int(24),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p12615,
																			FreeVars: ast.Identifiers{
																				"lb",
																			},
																		},
																		Id: "lb",
																	},
																},
																ThenFodder: ast.Fodder{},
																BranchTrue: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1189),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(1189),
																				Column: int(25),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p12615,
																		FreeVars: ast.Identifiers{
																			"a",
																			"i",
																			"prefix",
																			"std",
																		},
																	},
																	Left: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1189),
																					Column: int(11),
																				},
																				End: ast.Location{
																					Line: int(1189),
																					Column: int(17),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(10),
																					Comment: []string{},
																				},
																			},
																			Ctx: p12615,
																			FreeVars: ast.Identifiers{
																				"prefix",
																			},
																		},
																		Id: "prefix",
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(3),
																	Right: &ast.Apply{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"a",
																				"i",
																				"std",
																			},
																		},
																		Target: &ast.Index{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																				},
																				Id: "std",
																			},
																			LeftBracketFodder: nil,
																			Index: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: nil,
																				},
																				Value: "slice",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																			RightBracketFodder: nil,
																			Id: nil,
																		},
																		FodderLeft: nil,
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1189),
																									Column: int(20),
																								},
																								End: ast.Location{
																									Line: int(1189),
																									Column: int(21),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12615,
																							FreeVars: ast.Identifiers{
																								"a",
																							},
																						},
																						Id: "a",
																					},
																					CommaFodder: nil,
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1189),
																									Column: int(22),
																								},
																								End: ast.Location{
																									Line: int(1189),
																									Column: int(23),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12615,
																							FreeVars: ast.Identifiers{
																								"i",
																							},
																						},
																						Id: "i",
																					},
																					CommaFodder: nil,
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.LiteralNull{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								File: nil,
																							},
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: nil,
																						},
																					},
																					CommaFodder: nil,
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.LiteralNull{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "",
																								Begin: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								End: ast.Location{
																									Line: int(0),
																									Column: int(0),
																								},
																								File: nil,
																							},
																							Fodder: nil,
																							Ctx: nil,
																							FreeVars: nil,
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		TrailingComma: false,
																		TailStrict: false,
																		FodderRight: nil,
																		TailStrictFodder: nil,
																	},
																},
																ElseFodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																		Comment: []string{},
																	},
																},
																BranchFalse: &ast.Conditional{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1191),
																				Column: int(11),
																			},
																			End: ast.Location{
																				Line: int(1194),
																				Column: int(43),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(10),
																				Comment: []string{},
																			},
																		},
																		Ctx: p12615,
																		FreeVars: ast.Identifiers{
																			"a",
																			"aux",
																			"b",
																			"i",
																			"j",
																			"keyF",
																			"prefix",
																		},
																	},
																	Cond: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1191),
																					Column: int(14),
																				},
																				End: ast.Location{
																					Line: int(1191),
																					Column: int(38),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p12615,
																			FreeVars: ast.Identifiers{
																				"a",
																				"b",
																				"i",
																				"j",
																				"keyF",
																			},
																		},
																		Left: &ast.Apply{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1191),
																						Column: int(14),
																					},
																					End: ast.Location{
																						Line: int(1191),
																						Column: int(24),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p12615,
																				FreeVars: ast.Identifiers{
																					"a",
																					"i",
																					"keyF",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1191),
																							Column: int(14),
																						},
																						End: ast.Location{
																							Line: int(1191),
																							Column: int(18),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p12615,
																					FreeVars: ast.Identifiers{
																						"keyF",
																					},
																				},
																				Id: "keyF",
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Index{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1191),
																										Column: int(19),
																									},
																									End: ast.Location{
																										Line: int(1191),
																										Column: int(23),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p12682,
																								FreeVars: ast.Identifiers{
																									"a",
																									"i",
																								},
																							},
																							Target: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1191),
																											Column: int(19),
																										},
																										End: ast.Location{
																											Line: int(1191),
																											Column: int(20),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p12682,
																									FreeVars: ast.Identifiers{
																										"a",
																									},
																								},
																								Id: "a",
																							},
																							LeftBracketFodder: ast.Fodder{},
																							Index: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1191),
																											Column: int(21),
																										},
																										End: ast.Location{
																											Line: int(1191),
																											Column: int(22),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p12682,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																								},
																								Id: "i",
																							},
																							RightBracketFodder: ast.Fodder{},
																							Id: nil,
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			TrailingComma: false,
																			TailStrict: false,
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(10),
																		Right: &ast.Apply{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1191),
																						Column: int(28),
																					},
																					End: ast.Location{
																						Line: int(1191),
																						Column: int(38),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p12615,
																				FreeVars: ast.Identifiers{
																					"b",
																					"j",
																					"keyF",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1191),
																							Column: int(28),
																						},
																						End: ast.Location{
																							Line: int(1191),
																							Column: int(32),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p12615,
																					FreeVars: ast.Identifiers{
																						"keyF",
																					},
																				},
																				Id: "keyF",
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Index{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1191),
																										Column: int(33),
																									},
																									End: ast.Location{
																										Line: int(1191),
																										Column: int(37),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p12694,
																								FreeVars: ast.Identifiers{
																									"b",
																									"j",
																								},
																							},
																							Target: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1191),
																											Column: int(33),
																										},
																										End: ast.Location{
																											Line: int(1191),
																											Column: int(34),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p12694,
																									FreeVars: ast.Identifiers{
																										"b",
																									},
																								},
																								Id: "b",
																							},
																							LeftBracketFodder: ast.Fodder{},
																							Index: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1191),
																											Column: int(35),
																										},
																										End: ast.Location{
																											Line: int(1191),
																											Column: int(36),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p12694,
																									FreeVars: ast.Identifiers{
																										"j",
																									},
																								},
																								Id: "j",
																							},
																							RightBracketFodder: ast.Fodder{},
																							Id: nil,
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			TrailingComma: false,
																			TailStrict: false,
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: nil,
																		},
																	},
																	ThenFodder: ast.Fodder{},
																	BranchTrue: &ast.Apply{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1192),
																					Column: int(13),
																				},
																				End: ast.Location{
																					Line: int(1192),
																					Column: int(43),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p12615,
																			FreeVars: ast.Identifiers{
																				"a",
																				"aux",
																				"i",
																				"j",
																				"prefix",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1192),
																						Column: int(13),
																					},
																					End: ast.Location{
																						Line: int(1192),
																						Column: int(16),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(12),
																						Comment: []string{},
																					},
																				},
																				Ctx: p12615,
																				FreeVars: ast.Identifiers{
																					"aux",
																				},
																			},
																			Id: "aux",
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Binary{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1192),
																									Column: int(17),
																								},
																								End: ast.Location{
																									Line: int(1192),
																									Column: int(22),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12707,
																							FreeVars: ast.Identifiers{
																								"i",
																							},
																						},
																						Left: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1192),
																										Column: int(17),
																									},
																									End: ast.Location{
																										Line: int(1192),
																										Column: int(18),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p12707,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																							},
																							Id: "i",
																						},
																						OpFodder: ast.Fodder{},
																						Op: ast.BinaryOp(3),
																						Right: &ast.LiteralNumber{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1192),
																										Column: int(21),
																									},
																									End: ast.Location{
																										Line: int(1192),
																										Column: int(22),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p12707,
																								FreeVars: nil,
																							},
																							OriginalString: "1",
																						},
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1192),
																									Column: int(24),
																								},
																								End: ast.Location{
																									Line: int(1192),
																									Column: int(25),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12707,
																							FreeVars: ast.Identifiers{
																								"j",
																							},
																						},
																						Id: "j",
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Binary{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1192),
																									Column: int(27),
																								},
																								End: ast.Location{
																									Line: int(1192),
																									Column: int(42),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12707,
																							FreeVars: ast.Identifiers{
																								"a",
																								"i",
																								"prefix",
																							},
																						},
																						Left: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1192),
																										Column: int(27),
																									},
																									End: ast.Location{
																										Line: int(1192),
																										Column: int(33),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p12707,
																								FreeVars: ast.Identifiers{
																									"prefix",
																								},
																							},
																							Id: "prefix",
																						},
																						OpFodder: ast.Fodder{},
																						Op: ast.BinaryOp(3),
																						Right: &ast.Array{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1192),
																										Column: int(36),
																									},
																									End: ast.Location{
																										Line: int(1192),
																										Column: int(42),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p12707,
																								FreeVars: ast.Identifiers{
																									"a",
																									"i",
																								},
																							},
																							Elements: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Index{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(1192),
																													Column: int(37),
																												},
																												End: ast.Location{
																													Line: int(1192),
																													Column: int(41),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p12722,
																											FreeVars: ast.Identifiers{
																												"a",
																												"i",
																											},
																										},
																										Target: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1192),
																														Column: int(37),
																													},
																													End: ast.Location{
																														Line: int(1192),
																														Column: int(38),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p12722,
																												FreeVars: ast.Identifiers{
																													"a",
																												},
																											},
																											Id: "a",
																										},
																										LeftBracketFodder: ast.Fodder{},
																										Index: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1192),
																														Column: int(39),
																													},
																													End: ast.Location{
																														Line: int(1192),
																														Column: int(40),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p12722,
																												FreeVars: ast.Identifiers{
																													"i",
																												},
																											},
																											Id: "i",
																										},
																										RightBracketFodder: ast.Fodder{},
																										Id: nil,
																									},
																									CommaFodder: nil,
																								},
																							},
																							TrailingComma: false,
																							CloseFodder: ast.Fodder{},
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		TrailingComma: false,
																		TailStrict: true,
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: ast.Fodder{},
																	},
																	ElseFodder: ast.Fodder{
																		ast.FodderElement{
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(10),
																			Comment: []string{},
																		},
																	},
																	BranchFalse: &ast.Apply{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1194),
																					Column: int(13),
																				},
																				End: ast.Location{
																					Line: int(1194),
																					Column: int(43),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p12615,
																			FreeVars: ast.Identifiers{
																				"aux",
																				"b",
																				"i",
																				"j",
																				"prefix",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1194),
																						Column: int(13),
																					},
																					End: ast.Location{
																						Line: int(1194),
																						Column: int(16),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(12),
																						Comment: []string{},
																					},
																				},
																				Ctx: p12615,
																				FreeVars: ast.Identifiers{
																					"aux",
																				},
																			},
																			Id: "aux",
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1194),
																									Column: int(17),
																								},
																								End: ast.Location{
																									Line: int(1194),
																									Column: int(18),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12736,
																							FreeVars: ast.Identifiers{
																								"i",
																							},
																						},
																						Id: "i",
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Binary{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1194),
																									Column: int(20),
																								},
																								End: ast.Location{
																									Line: int(1194),
																									Column: int(25),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12736,
																							FreeVars: ast.Identifiers{
																								"j",
																							},
																						},
																						Left: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1194),
																										Column: int(20),
																									},
																									End: ast.Location{
																										Line: int(1194),
																										Column: int(21),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p12736,
																								FreeVars: ast.Identifiers{
																									"j",
																								},
																							},
																							Id: "j",
																						},
																						OpFodder: ast.Fodder{},
																						Op: ast.BinaryOp(3),
																						Right: &ast.LiteralNumber{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1194),
																										Column: int(24),
																									},
																									End: ast.Location{
																										Line: int(1194),
																										Column: int(25),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p12736,
																								FreeVars: nil,
																							},
																							OriginalString: "1",
																						},
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Binary{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1194),
																									Column: int(27),
																								},
																								End: ast.Location{
																									Line: int(1194),
																									Column: int(42),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12736,
																							FreeVars: ast.Identifiers{
																								"b",
																								"j",
																								"prefix",
																							},
																						},
																						Left: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1194),
																										Column: int(27),
																									},
																									End: ast.Location{
																										Line: int(1194),
																										Column: int(33),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p12736,
																								FreeVars: ast.Identifiers{
																									"prefix",
																								},
																							},
																							Id: "prefix",
																						},
																						OpFodder: ast.Fodder{},
																						Op: ast.BinaryOp(3),
																						Right: &ast.Array{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1194),
																										Column: int(36),
																									},
																									End: ast.Location{
																										Line: int(1194),
																										Column: int(42),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p12736,
																								FreeVars: ast.Identifiers{
																									"b",
																									"j",
																								},
																							},
																							Elements: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Index{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(1194),
																													Column: int(37),
																												},
																												End: ast.Location{
																													Line: int(1194),
																													Column: int(41),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p12751,
																											FreeVars: ast.Identifiers{
																												"b",
																												"j",
																											},
																										},
																										Target: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1194),
																														Column: int(37),
																													},
																													End: ast.Location{
																														Line: int(1194),
																														Column: int(38),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p12751,
																												FreeVars: ast.Identifiers{
																													"b",
																												},
																											},
																											Id: "b",
																										},
																										LeftBracketFodder: ast.Fodder{},
																										Index: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1194),
																														Column: int(39),
																													},
																													End: ast.Location{
																														Line: int(1194),
																														Column: int(40),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p12751,
																												FreeVars: ast.Identifiers{
																													"j",
																												},
																											},
																											Id: "j",
																										},
																										RightBracketFodder: ast.Fodder{},
																										Id: nil,
																									},
																									CommaFodder: nil,
																								},
																							},
																							TrailingComma: false,
																							CloseFodder: ast.Fodder{},
																						},
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		TrailingComma: false,
																		TailStrict: true,
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: ast.Fodder{},
																	},
																},
															},
														},
													},
													Fun: nil,
													CloseFodder: nil,
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
												},
											},
											Body: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1195),
															Column: int(7),
														},
														End: ast.Location{
															Line: int(1195),
															Column: int(20),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p12578,
													FreeVars: ast.Identifiers{
														"aux",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1195),
																Column: int(7),
															},
															End: ast.Location{
																Line: int(1195),
																Column: int(10),
															},
															File: p1,
														},
														Fodder: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(6),
																Comment: []string{},
															},
														},
														Ctx: p12578,
														FreeVars: ast.Identifiers{
															"aux",
														},
													},
													Id: "aux",
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralNumber{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1195),
																			Column: int(11),
																		},
																		End: ast.Location{
																			Line: int(1195),
																			Column: int(12),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p12764,
																	FreeVars: nil,
																},
																OriginalString: "0",
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralNumber{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1195),
																			Column: int(14),
																		},
																		End: ast.Location{
																			Line: int(1195),
																			Column: int(15),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p12764,
																	FreeVars: nil,
																},
																OriginalString: "0",
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Array{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1195),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(1195),
																			Column: int(19),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p12764,
																	FreeVars: nil,
																},
																Elements: nil,
																TrailingComma: false,
																CloseFodder: ast.Fodder{},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
										},
									},
								},
								Fun: nil,
								CloseFodder: nil,
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
							},
						},
						Body: &ast.Local{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1197),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1203),
										Column: int(67),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(1),
										Indent: int(4),
										Comment: []string{},
									},
								},
								Ctx: p12350,
								FreeVars: ast.Identifiers{
									"arr",
									"keyF",
									"merge",
									"quickSort",
									"std",
								},
							},
							Binds: ast.LocalBinds{
								ast.LocalBind{
									VarFodder: ast.Fodder{},
									Variable: "l",
									EqFodder: ast.Fodder{},
									Body: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1197),
													Column: int(15),
												},
												End: ast.Location{
													Line: int(1197),
													Column: int(30),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p12772,
											FreeVars: ast.Identifiers{
												"arr",
												"std",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1197),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(1197),
														Column: int(25),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p12772,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1197),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1197),
															Column: int(18),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: ast.Fodder{},
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "length",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: ast.Fodder{},
											Id: nil,
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1197),
																	Column: int(26),
																},
																End: ast.Location{
																	Line: int(1197),
																	Column: int(29),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p12781,
															FreeVars: ast.Identifiers{
																"arr",
															},
														},
														Id: "arr",
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
									Fun: nil,
									CloseFodder: ast.Fodder{},
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1197),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(1197),
											Column: int(30),
										},
										File: p1,
									},
								},
							},
							Body: &ast.Conditional{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1198),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(1203),
											Column: int(67),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
											Comment: []string{},
										},
									},
									Ctx: p12350,
									FreeVars: ast.Identifiers{
										"arr",
										"keyF",
										"l",
										"merge",
										"quickSort",
										"std",
									},
								},
								Cond: &ast.Binary{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1198),
												Column: int(8),
											},
											End: ast.Location{
												Line: int(1198),
												Column: int(29),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p12350,
										FreeVars: ast.Identifiers{
											"arr",
											"std",
										},
									},
									Left: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1198),
													Column: int(8),
												},
												End: ast.Location{
													Line: int(1198),
													Column: int(23),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p12350,
											FreeVars: ast.Identifiers{
												"arr",
												"std",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1198),
														Column: int(8),
													},
													End: ast.Location{
														Line: int(1198),
														Column: int(18),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p12350,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1198),
															Column: int(8),
														},
														End: ast.Location{
															Line: int(1198),
															Column: int(11),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: ast.Fodder{},
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "length",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: ast.Fodder{},
											Id: nil,
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1198),
																	Column: int(19),
																},
																End: ast.Location{
																	Line: int(1198),
																	Column: int(22),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p12797,
															FreeVars: ast.Identifiers{
																"arr",
															},
														},
														Id: "arr",
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
									OpFodder: ast.Fodder{},
									Op: ast.BinaryOp(10),
									Right: &ast.LiteralNumber{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1198),
													Column: int(27),
												},
												End: ast.Location{
													Line: int(1198),
													Column: int(29),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p12350,
											FreeVars: nil,
										},
										OriginalString: "30",
									},
								},
								ThenFodder: ast.Fodder{},
								BranchTrue: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1199),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1199),
												Column: int(32),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p12350,
										FreeVars: ast.Identifiers{
											"arr",
											"keyF",
											"quickSort",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1199),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(1199),
													Column: int(16),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
													Comment: []string{},
												},
											},
											Ctx: p12350,
											FreeVars: ast.Identifiers{
												"quickSort",
											},
										},
										Id: "quickSort",
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1199),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(1199),
																Column: int(20),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p12807,
														FreeVars: ast.Identifiers{
															"arr",
														},
													},
													Id: "arr",
												},
												CommaFodder: ast.Fodder{},
											},
										},
										Named: []ast.NamedArgument{
											ast.NamedArgument{
												NameFodder: ast.Fodder{},
												Name: "keyF",
												EqFodder: ast.Fodder{},
												Arg: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1199),
																Column: int(27),
															},
															End: ast.Location{
																Line: int(1199),
																Column: int(31),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p12807,
														FreeVars: ast.Identifiers{
															"keyF",
														},
													},
													Id: "keyF",
												},
												CommaFodder: ast.Fodder{},
											},
										},
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
								ElseFodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
										Comment: []string{},
									},
								},
								BranchFalse: &ast.Local{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1201),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1203),
												Column: int(67),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										Ctx: p12350,
										FreeVars: ast.Identifiers{
											"arr",
											"keyF",
											"l",
											"merge",
											"std",
										},
									},
									Binds: ast.LocalBinds{
										ast.LocalBind{
											VarFodder: ast.Fodder{},
											Variable: "mid",
											EqFodder: ast.Fodder{},
											Body: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1201),
															Column: int(19),
														},
														End: ast.Location{
															Line: int(1201),
															Column: int(35),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p12818,
													FreeVars: ast.Identifiers{
														"l",
														"std",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1201),
																Column: int(19),
															},
															End: ast.Location{
																Line: int(1201),
																Column: int(28),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p12818,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1201),
																	Column: int(19),
																},
																End: ast.Location{
																	Line: int(1201),
																	Column: int(22),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: ast.Fodder{},
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "floor",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: ast.Fodder{},
													Id: nil,
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Binary{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1201),
																			Column: int(29),
																		},
																		End: ast.Location{
																			Line: int(1201),
																			Column: int(34),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p12827,
																	FreeVars: ast.Identifiers{
																		"l",
																	},
																},
																Left: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1201),
																				Column: int(29),
																			},
																			End: ast.Location{
																				Line: int(1201),
																				Column: int(30),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p12827,
																		FreeVars: ast.Identifiers{
																			"l",
																		},
																	},
																	Id: "l",
																},
																OpFodder: ast.Fodder{},
																Op: ast.BinaryOp(1),
																Right: &ast.LiteralNumber{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1201),
																				Column: int(33),
																			},
																			End: ast.Location{
																				Line: int(1201),
																				Column: int(34),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p12827,
																		FreeVars: nil,
																	},
																	OriginalString: "2",
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
											Fun: nil,
											CloseFodder: ast.Fodder{},
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1201),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(1201),
													Column: int(35),
												},
												File: p1,
											},
										},
									},
									Body: &ast.Local{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1202),
													Column: int(7),
												},
												End: ast.Location{
													Line: int(1203),
													Column: int(67),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
													Comment: []string{},
												},
											},
											Ctx: p12350,
											FreeVars: ast.Identifiers{
												"arr",
												"keyF",
												"merge",
												"mid",
												"std",
											},
										},
										Binds: ast.LocalBinds{
											ast.LocalBind{
												VarFodder: ast.Fodder{},
												Variable: "left",
												EqFodder: ast.Fodder{},
												Body: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"arr",
															"mid",
															"std",
														},
													},
													Target: &ast.Index{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Id: "std",
														},
														LeftBracketFodder: nil,
														Index: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
															Value: "slice",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														RightBracketFodder: nil,
														Id: nil,
													},
													FodderLeft: nil,
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1202),
																				Column: int(20),
																			},
																			End: ast.Location{
																				Line: int(1202),
																				Column: int(23),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p12845,
																		FreeVars: ast.Identifiers{
																			"arr",
																		},
																	},
																	Id: "arr",
																},
																CommaFodder: nil,
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.LiteralNull{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: nil,
																	},
																},
																CommaFodder: nil,
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1202),
																				Column: int(25),
																			},
																			End: ast.Location{
																				Line: int(1202),
																				Column: int(28),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p12845,
																		FreeVars: ast.Identifiers{
																			"mid",
																		},
																	},
																	Id: "mid",
																},
																CommaFodder: nil,
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.LiteralNull{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: nil,
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: nil,
													TailStrictFodder: nil,
												},
												Fun: nil,
												CloseFodder: ast.Fodder{},
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1202),
														Column: int(13),
													},
													End: ast.Location{
														Line: int(1202),
														Column: int(29),
													},
													File: p1,
												},
											},
											ast.LocalBind{
												VarFodder: ast.Fodder{},
												Variable: "right",
												EqFodder: ast.Fodder{},
												Body: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"arr",
															"mid",
															"std",
														},
													},
													Target: &ast.Index{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Id: "std",
														},
														LeftBracketFodder: nil,
														Index: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
															Value: "slice",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														RightBracketFodder: nil,
														Id: nil,
													},
													FodderLeft: nil,
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1202),
																				Column: int(39),
																			},
																			End: ast.Location{
																				Line: int(1202),
																				Column: int(42),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p12860,
																		FreeVars: ast.Identifiers{
																			"arr",
																		},
																	},
																	Id: "arr",
																},
																CommaFodder: nil,
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1202),
																				Column: int(43),
																			},
																			End: ast.Location{
																				Line: int(1202),
																				Column: int(46),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p12860,
																		FreeVars: ast.Identifiers{
																			"mid",
																		},
																	},
																	Id: "mid",
																},
																CommaFodder: nil,
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.LiteralNull{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: nil,
																	},
																},
																CommaFodder: nil,
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.LiteralNull{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: nil,
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: nil,
													TailStrictFodder: nil,
												},
												Fun: nil,
												CloseFodder: ast.Fodder{},
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1202),
														Column: int(31),
													},
													End: ast.Location{
														Line: int(1202),
														Column: int(48),
													},
													File: p1,
												},
											},
										},
										Body: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1203),
														Column: int(7),
													},
													End: ast.Location{
														Line: int(1203),
														Column: int(67),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p12350,
												FreeVars: ast.Identifiers{
													"keyF",
													"left",
													"merge",
													"right",
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1203),
															Column: int(7),
														},
														End: ast.Location{
															Line: int(1203),
															Column: int(12),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(6),
															Comment: []string{},
														},
													},
													Ctx: p12350,
													FreeVars: ast.Identifiers{
														"merge",
													},
												},
												Id: "merge",
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Apply{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1203),
																		Column: int(13),
																	},
																	End: ast.Location{
																		Line: int(1203),
																		Column: int(38),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p12873,
																FreeVars: ast.Identifiers{
																	"keyF",
																	"left",
																	"std",
																},
															},
															Target: &ast.Index{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1203),
																			Column: int(13),
																		},
																		End: ast.Location{
																			Line: int(1203),
																			Column: int(21),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p12873,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1203),
																				Column: int(13),
																			},
																			End: ast.Location{
																				Line: int(1203),
																				Column: int(16),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Id: "std",
																},
																LeftBracketFodder: ast.Fodder{},
																Index: &ast.LiteralString{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: nil,
																	},
																	Value: "sort",
																	Kind: ast.LiteralStringKind(1),
																	BlockIndent: "",
																	BlockTermIndent: "",
																},
																RightBracketFodder: ast.Fodder{},
																Id: nil,
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1203),
																						Column: int(22),
																					},
																					End: ast.Location{
																						Line: int(1203),
																						Column: int(26),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p12882,
																				FreeVars: ast.Identifiers{
																					"left",
																				},
																			},
																			Id: "left",
																		},
																		CommaFodder: ast.Fodder{},
																	},
																},
																Named: []ast.NamedArgument{
																	ast.NamedArgument{
																		NameFodder: ast.Fodder{},
																		Name: "keyF",
																		EqFodder: ast.Fodder{},
																		Arg: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1203),
																						Column: int(33),
																					},
																					End: ast.Location{
																						Line: int(1203),
																						Column: int(37),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p12882,
																				FreeVars: ast.Identifiers{
																					"keyF",
																				},
																			},
																			Id: "keyF",
																		},
																		CommaFodder: ast.Fodder{},
																	},
																},
															},
															TrailingComma: false,
															TailStrict: false,
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Apply{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1203),
																		Column: int(40),
																	},
																	End: ast.Location{
																		Line: int(1203),
																		Column: int(66),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p12873,
																FreeVars: ast.Identifiers{
																	"keyF",
																	"right",
																	"std",
																},
															},
															Target: &ast.Index{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1203),
																			Column: int(40),
																		},
																		End: ast.Location{
																			Line: int(1203),
																			Column: int(48),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p12873,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1203),
																				Column: int(40),
																			},
																			End: ast.Location{
																				Line: int(1203),
																				Column: int(43),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Id: "std",
																},
																LeftBracketFodder: ast.Fodder{},
																Index: &ast.LiteralString{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: nil,
																	},
																	Value: "sort",
																	Kind: ast.LiteralStringKind(1),
																	BlockIndent: "",
																	BlockTermIndent: "",
																},
																RightBracketFodder: ast.Fodder{},
																Id: nil,
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1203),
																						Column: int(49),
																					},
																					End: ast.Location{
																						Line: int(1203),
																						Column: int(54),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p12896,
																				FreeVars: ast.Identifiers{
																					"right",
																				},
																			},
																			Id: "right",
																		},
																		CommaFodder: ast.Fodder{},
																	},
																},
																Named: []ast.NamedArgument{
																	ast.NamedArgument{
																		NameFodder: ast.Fodder{},
																		Name: "keyF",
																		EqFodder: ast.Fodder{},
																		Arg: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1203),
																						Column: int(61),
																					},
																					End: ast.Location{
																						Line: int(1203),
																						Column: int(65),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p12896,
																				FreeVars: ast.Identifiers{
																					"keyF",
																				},
																			},
																			Id: "keyF",
																		},
																		CommaFodder: ast.Fodder{},
																	},
																},
															},
															TrailingComma: false,
															TailStrict: false,
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
									},
								},
							},
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(1170),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1203),
					Column: int(67),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "uniq",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"id",
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1205),
								Column: int(8),
							},
							End: ast.Location{
								Line: int(1205),
								Column: int(11),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "keyF",
						EqFodder: ast.Fodder{},
						DefaultArg: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1205),
										Column: int(18),
									},
									End: ast.Location{
										Line: int(1205),
										Column: int(20),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p12906,
								FreeVars: ast.Identifiers{
									"id",
								},
							},
							Id: "id",
						},
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1205),
								Column: int(13),
							},
							End: ast.Location{
								Line: int(1205),
								Column: int(20),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1206),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1213),
								Column: int(26),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p12906,
						FreeVars: ast.Identifiers{
							"arr",
							"keyF",
							"std",
						},
					},
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: nil,
							Variable: "f",
							EqFodder: nil,
							Body: &ast.Function{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1206),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(1212),
											Column: int(16),
										},
										File: p1,
									},
									Fodder: nil,
									Ctx: p12913,
									FreeVars: ast.Identifiers{
										"keyF",
										"std",
									},
								},
								ParenLeftFodder: ast.Fodder{},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "a",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1206),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(1206),
												Column: int(14),
											},
											File: p1,
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "b",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: nil,
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1206),
												Column: int(16),
											},
											End: ast.Location{
												Line: int(1206),
												Column: int(17),
											},
											File: p1,
										},
									},
								},
								TrailingComma: false,
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Conditional{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1207),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1212),
												Column: int(16),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										Ctx: p12918,
										FreeVars: ast.Identifiers{
											"a",
											"b",
											"keyF",
											"std",
										},
									},
									Cond: &ast.Binary{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1207),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(1207),
													Column: int(28),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p12918,
											FreeVars: ast.Identifiers{
												"a",
												"std",
											},
										},
										Left: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1207),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(1207),
														Column: int(23),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p12918,
												FreeVars: ast.Identifiers{
													"a",
													"std",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1207),
															Column: int(10),
														},
														End: ast.Location{
															Line: int(1207),
															Column: int(20),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p12918,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1207),
																Column: int(10),
															},
															End: ast.Location{
																Line: int(1207),
																Column: int(13),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "length",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1207),
																		Column: int(21),
																	},
																	End: ast.Location{
																		Line: int(1207),
																		Column: int(22),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p12931,
																FreeVars: ast.Identifiers{
																	"a",
																},
															},
															Id: "a",
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
										OpFodder: ast.Fodder{},
										Op: ast.BinaryOp(12),
										Right: &ast.LiteralNumber{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1207),
														Column: int(27),
													},
													End: ast.Location{
														Line: int(1207),
														Column: int(28),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p12918,
												FreeVars: nil,
											},
											OriginalString: "0",
										},
									},
									ThenFodder: ast.Fodder{},
									BranchTrue: &ast.Array{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1208),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1208),
													Column: int(12),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
													Comment: []string{},
												},
											},
											Ctx: p12918,
											FreeVars: ast.Identifiers{
												"b",
											},
										},
										Elements: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1208),
																Column: int(10),
															},
															End: ast.Location{
																Line: int(1208),
																Column: int(11),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p12939,
														FreeVars: ast.Identifiers{
															"b",
														},
													},
													Id: "b",
												},
												CommaFodder: nil,
											},
										},
										TrailingComma: false,
										CloseFodder: ast.Fodder{},
									},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
											Comment: []string{},
										},
									},
									BranchFalse: &ast.Conditional{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1209),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(1212),
													Column: int(16),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p12918,
											FreeVars: ast.Identifiers{
												"a",
												"b",
												"keyF",
												"std",
											},
										},
										Cond: &ast.Binary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1209),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(1209),
														Column: int(52),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p12918,
												FreeVars: ast.Identifiers{
													"a",
													"b",
													"keyF",
													"std",
												},
											},
											Left: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1209),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1209),
															Column: int(41),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p12918,
													FreeVars: ast.Identifiers{
														"a",
														"keyF",
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1209),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1209),
																Column: int(19),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p12918,
														FreeVars: ast.Identifiers{
															"keyF",
														},
													},
													Id: "keyF",
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Index{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1209),
																			Column: int(20),
																		},
																		End: ast.Location{
																			Line: int(1209),
																			Column: int(40),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p12952,
																	FreeVars: ast.Identifiers{
																		"a",
																		"std",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1209),
																				Column: int(20),
																			},
																			End: ast.Location{
																				Line: int(1209),
																				Column: int(21),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p12952,
																		FreeVars: ast.Identifiers{
																			"a",
																		},
																	},
																	Id: "a",
																},
																LeftBracketFodder: ast.Fodder{},
																Index: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1209),
																				Column: int(22),
																			},
																			End: ast.Location{
																				Line: int(1209),
																				Column: int(39),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p12952,
																		FreeVars: ast.Identifiers{
																			"a",
																			"std",
																		},
																	},
																	Left: &ast.Apply{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1209),
																					Column: int(22),
																				},
																				End: ast.Location{
																					Line: int(1209),
																					Column: int(35),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p12952,
																			FreeVars: ast.Identifiers{
																				"a",
																				"std",
																			},
																		},
																		Target: &ast.Index{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1209),
																						Column: int(22),
																					},
																					End: ast.Location{
																						Line: int(1209),
																						Column: int(32),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p12952,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1209),
																							Column: int(22),
																						},
																						End: ast.Location{
																							Line: int(1209),
																							Column: int(25),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																				},
																				Id: "std",
																			},
																			LeftBracketFodder: ast.Fodder{},
																			Index: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: nil,
																				},
																				Value: "length",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																			RightBracketFodder: ast.Fodder{},
																			Id: nil,
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1209),
																									Column: int(33),
																								},
																								End: ast.Location{
																									Line: int(1209),
																									Column: int(34),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p12967,
																							FreeVars: ast.Identifiers{
																								"a",
																							},
																						},
																						Id: "a",
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		TrailingComma: false,
																		TailStrict: false,
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(4),
																	Right: &ast.LiteralNumber{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1209),
																					Column: int(38),
																				},
																				End: ast.Location{
																					Line: int(1209),
																					Column: int(39),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p12952,
																			FreeVars: nil,
																		},
																		OriginalString: "1",
																	},
																},
																RightBracketFodder: ast.Fodder{},
																Id: nil,
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
											OpFodder: ast.Fodder{},
											Op: ast.BinaryOp(12),
											Right: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1209),
															Column: int(45),
														},
														End: ast.Location{
															Line: int(1209),
															Column: int(52),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p12918,
													FreeVars: ast.Identifiers{
														"b",
														"keyF",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1209),
																Column: int(45),
															},
															End: ast.Location{
																Line: int(1209),
																Column: int(49),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p12918,
														FreeVars: ast.Identifiers{
															"keyF",
														},
													},
													Id: "keyF",
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1209),
																			Column: int(50),
																		},
																		End: ast.Location{
																			Line: int(1209),
																			Column: int(51),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p12976,
																	FreeVars: ast.Identifiers{
																		"b",
																	},
																},
																Id: "b",
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
										},
										ThenFodder: ast.Fodder{},
										BranchTrue: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1210),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1210),
														Column: int(10),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
														Comment: []string{},
													},
												},
												Ctx: p12918,
												FreeVars: ast.Identifiers{
													"a",
												},
											},
											Id: "a",
										},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										BranchFalse: &ast.Binary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1212),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1212),
														Column: int(16),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p12918,
												FreeVars: ast.Identifiers{
													"a",
													"b",
												},
											},
											Left: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1212),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1212),
															Column: int(10),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
															Comment: []string{},
														},
													},
													Ctx: p12918,
													FreeVars: ast.Identifiers{
														"a",
													},
												},
												Id: "a",
											},
											OpFodder: ast.Fodder{},
											Op: ast.BinaryOp(3),
											Right: &ast.Array{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1212),
															Column: int(13),
														},
														End: ast.Location{
															Line: int(1212),
															Column: int(16),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p12918,
													FreeVars: ast.Identifiers{
														"b",
													},
												},
												Elements: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1212),
																		Column: int(14),
																	},
																	End: ast.Location{
																		Line: int(1212),
																		Column: int(15),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p12991,
																FreeVars: ast.Identifiers{
																	"b",
																},
															},
															Id: "b",
														},
														CommaFodder: nil,
													},
												},
												TrailingComma: false,
												CloseFodder: ast.Fodder{},
											},
										},
									},
								},
							},
							Fun: nil,
							CloseFodder: nil,
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
						},
					},
					Body: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1213),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1213),
									Column: int(26),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p12906,
							FreeVars: ast.Identifiers{
								"arr",
								"f",
								"std",
							},
						},
						Target: &ast.Index{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1213),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1213),
										Column: int(14),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p12906,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1213),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(1213),
											Column: int(8),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
											Comment: []string{},
										},
									},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Id: "std",
							},
							LeftBracketFodder: ast.Fodder{},
							Index: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Value: "foldl",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							RightBracketFodder: ast.Fodder{},
							Id: nil,
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1213),
													Column: int(15),
												},
												End: ast.Location{
													Line: int(1213),
													Column: int(16),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13003,
											FreeVars: ast.Identifiers{
												"f",
											},
										},
										Id: "f",
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1213),
													Column: int(18),
												},
												End: ast.Location{
													Line: int(1213),
													Column: int(21),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13003,
											FreeVars: ast.Identifiers{
												"arr",
											},
										},
										Id: "arr",
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Array{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1213),
													Column: int(23),
												},
												End: ast.Location{
													Line: int(1213),
													Column: int(25),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13003,
											FreeVars: nil,
										},
										Elements: nil,
										TrailingComma: false,
										CloseFodder: ast.Fodder{},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(1205),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1213),
					Column: int(26),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "set",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"id",
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1215),
								Column: int(7),
							},
							End: ast.Location{
								Line: int(1215),
								Column: int(10),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "keyF",
						EqFodder: ast.Fodder{},
						DefaultArg: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1215),
										Column: int(17),
									},
									End: ast.Location{
										Line: int(1215),
										Column: int(19),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p13013,
								FreeVars: ast.Identifiers{
									"id",
								},
							},
							Id: "id",
						},
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1215),
								Column: int(12),
							},
							End: ast.Location{
								Line: int(1215),
								Column: int(19),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1216),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1216),
								Column: int(40),
							},
							File: p1,
						},
						Fodder: ast.Fodder{},
						Ctx: p13013,
						FreeVars: ast.Identifiers{
							"arr",
							"keyF",
							"std",
						},
					},
					Target: &ast.Index{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1216),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1216),
									Column: int(13),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p13013,
							FreeVars: ast.Identifiers{
								"std",
							},
						},
						Target: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1216),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1216),
										Column: int(8),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
										Comment: []string{},
									},
								},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Id: "std",
						},
						LeftBracketFodder: ast.Fodder{},
						Index: &ast.LiteralString{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: nil,
							},
							Value: "uniq",
							Kind: ast.LiteralStringKind(1),
							BlockIndent: "",
							BlockTermIndent: "",
						},
						RightBracketFodder: ast.Fodder{},
						Id: nil,
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1216),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(1216),
												Column: int(33),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p13025,
										FreeVars: ast.Identifiers{
											"arr",
											"keyF",
											"std",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1216),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(1216),
													Column: int(22),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13025,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1216),
														Column: int(14),
													},
													End: ast.Location{
														Line: int(1216),
														Column: int(17),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "sort",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1216),
																Column: int(23),
															},
															End: ast.Location{
																Line: int(1216),
																Column: int(26),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p13034,
														FreeVars: ast.Identifiers{
															"arr",
														},
													},
													Id: "arr",
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1216),
																Column: int(28),
															},
															End: ast.Location{
																Line: int(1216),
																Column: int(32),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p13034,
														FreeVars: ast.Identifiers{
															"keyF",
														},
													},
													Id: "keyF",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1216),
												Column: int(35),
											},
											End: ast.Location{
												Line: int(1216),
												Column: int(39),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p13025,
										FreeVars: ast.Identifiers{
											"keyF",
										},
									},
									Id: "keyF",
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					TrailingComma: false,
					TailStrict: false,
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(1215),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1216),
					Column: int(40),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "setMember",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"id",
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "x",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1218),
								Column: int(13),
							},
							End: ast.Location{
								Line: int(1218),
								Column: int(14),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1218),
								Column: int(16),
							},
							End: ast.Location{
								Line: int(1218),
								Column: int(19),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "keyF",
						EqFodder: ast.Fodder{},
						DefaultArg: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1218),
										Column: int(26),
									},
									End: ast.Location{
										Line: int(1218),
										Column: int(28),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p13045,
								FreeVars: ast.Identifiers{
									"id",
								},
							},
							Id: "id",
						},
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1218),
								Column: int(21),
							},
							End: ast.Location{
								Line: int(1218),
								Column: int(28),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Binary{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1220),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1220),
								Column: int(49),
							},
							File: p1,
						},
						Fodder: ast.Fodder{},
						Ctx: p13045,
						FreeVars: ast.Identifiers{
							"arr",
							"keyF",
							"std",
							"x",
						},
					},
					Left: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1220),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1220),
									Column: int(45),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p13045,
							FreeVars: ast.Identifiers{
								"arr",
								"keyF",
								"std",
								"x",
							},
						},
						Target: &ast.Index{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1220),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1220),
										Column: int(15),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p13045,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1220),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(1220),
											Column: int(8),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
											Comment: []string{},
										},
										ast.FodderElement{
											Kind: ast.FodderKind(2),
											Blanks: int(0),
											Indent: int(4),
											Comment: []string{
												"// TODO(dcunnin): Binary chop for O(log n) complexity",
											},
										},
									},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Id: "std",
							},
							LeftBracketFodder: ast.Fodder{},
							Index: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Value: "length",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							RightBracketFodder: ast.Fodder{},
							Id: nil,
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1220),
													Column: int(16),
												},
												End: ast.Location{
													Line: int(1220),
													Column: int(44),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13060,
											FreeVars: ast.Identifiers{
												"arr",
												"keyF",
												"std",
												"x",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1220),
														Column: int(16),
													},
													End: ast.Location{
														Line: int(1220),
														Column: int(28),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p13060,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1220),
															Column: int(16),
														},
														End: ast.Location{
															Line: int(1220),
															Column: int(19),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: ast.Fodder{},
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "setInter",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: ast.Fodder{},
											Id: nil,
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Array{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1220),
																	Column: int(29),
																},
																End: ast.Location{
																	Line: int(1220),
																	Column: int(32),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p13069,
															FreeVars: ast.Identifiers{
																"x",
															},
														},
														Elements: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1220),
																				Column: int(30),
																			},
																			End: ast.Location{
																				Line: int(1220),
																				Column: int(31),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13073,
																		FreeVars: ast.Identifiers{
																			"x",
																		},
																	},
																	Id: "x",
																},
																CommaFodder: nil,
															},
														},
														TrailingComma: false,
														CloseFodder: ast.Fodder{},
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1220),
																	Column: int(34),
																},
																End: ast.Location{
																	Line: int(1220),
																	Column: int(37),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p13069,
															FreeVars: ast.Identifiers{
																"arr",
															},
														},
														Id: "arr",
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1220),
																	Column: int(39),
																},
																End: ast.Location{
																	Line: int(1220),
																	Column: int(43),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p13069,
															FreeVars: ast.Identifiers{
																"keyF",
															},
														},
														Id: "keyF",
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
					OpFodder: ast.Fodder{},
					Op: ast.BinaryOp(7),
					Right: &ast.LiteralNumber{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1220),
									Column: int(48),
								},
								End: ast.Location{
									Line: int(1220),
									Column: int(49),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p13045,
							FreeVars: nil,
						},
						OriginalString: "0",
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(1218),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1220),
					Column: int(49),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "setUnion",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"id",
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "a",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1222),
								Column: int(12),
							},
							End: ast.Location{
								Line: int(1222),
								Column: int(13),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "b",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1222),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(1222),
								Column: int(16),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "keyF",
						EqFodder: ast.Fodder{},
						DefaultArg: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1222),
										Column: int(23),
									},
									End: ast.Location{
										Line: int(1222),
										Column: int(25),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p13085,
								FreeVars: ast.Identifiers{
									"id",
								},
							},
							Id: "id",
						},
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1222),
								Column: int(18),
							},
							End: ast.Location{
								Line: int(1222),
								Column: int(25),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1224),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1238),
								Column: int(24),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
							ast.FodderElement{
								Kind: ast.FodderKind(2),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{
									"// NOTE: order matters, values in `a` win",
								},
							},
						},
						Ctx: p13085,
						FreeVars: ast.Identifiers{
							"a",
							"b",
							"keyF",
							"std",
						},
					},
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: nil,
							Variable: "aux",
							EqFodder: nil,
							Body: &ast.Function{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1224),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(1237),
											Column: int(44),
										},
										File: p1,
									},
									Fodder: nil,
									Ctx: p13093,
									FreeVars: ast.Identifiers{
										"aux",
										"keyF",
										"std",
									},
								},
								ParenLeftFodder: ast.Fodder{},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "a",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1224),
												Column: int(15),
											},
											End: ast.Location{
												Line: int(1224),
												Column: int(16),
											},
											File: p1,
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "b",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1224),
												Column: int(18),
											},
											End: ast.Location{
												Line: int(1224),
												Column: int(19),
											},
											File: p1,
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "i",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1224),
												Column: int(21),
											},
											End: ast.Location{
												Line: int(1224),
												Column: int(22),
											},
											File: p1,
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "j",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1224),
												Column: int(24),
											},
											End: ast.Location{
												Line: int(1224),
												Column: int(25),
											},
											File: p1,
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "acc",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: nil,
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1224),
												Column: int(27),
											},
											End: ast.Location{
												Line: int(1224),
												Column: int(30),
											},
											File: p1,
										},
									},
								},
								TrailingComma: false,
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Conditional{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1225),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1237),
												Column: int(44),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										Ctx: p13098,
										FreeVars: ast.Identifiers{
											"a",
											"acc",
											"aux",
											"b",
											"i",
											"j",
											"keyF",
											"std",
										},
									},
									Cond: &ast.Binary{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1225),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(1225),
													Column: int(28),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13098,
											FreeVars: ast.Identifiers{
												"a",
												"i",
												"std",
											},
										},
										Left: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1225),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(1225),
														Column: int(11),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p13098,
												FreeVars: ast.Identifiers{
													"i",
												},
											},
											Id: "i",
										},
										OpFodder: ast.Fodder{},
										Op: ast.BinaryOp(8),
										Right: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1225),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(1225),
														Column: int(28),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p13098,
												FreeVars: ast.Identifiers{
													"a",
													"std",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1225),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1225),
															Column: int(25),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p13098,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1225),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1225),
																Column: int(18),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "length",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1225),
																		Column: int(26),
																	},
																	End: ast.Location{
																		Line: int(1225),
																		Column: int(27),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p13113,
																FreeVars: ast.Identifiers{
																	"a",
																},
															},
															Id: "a",
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
									},
									ThenFodder: ast.Fodder{},
									BranchTrue: &ast.Binary{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1226),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1226),
													Column: int(20),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13098,
											FreeVars: ast.Identifiers{
												"acc",
												"b",
												"j",
												"std",
											},
										},
										Left: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1226),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1226),
														Column: int(12),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
														Comment: []string{},
													},
												},
												Ctx: p13098,
												FreeVars: ast.Identifiers{
													"acc",
												},
											},
											Id: "acc",
										},
										OpFodder: ast.Fodder{},
										Op: ast.BinaryOp(3),
										Right: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"b",
													"j",
													"std",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: nil,
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "slice",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: nil,
												Id: nil,
											},
											FodderLeft: nil,
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1226),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(1226),
																		Column: int(16),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p13098,
																FreeVars: ast.Identifiers{
																	"b",
																},
															},
															Id: "b",
														},
														CommaFodder: nil,
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1226),
																		Column: int(17),
																	},
																	End: ast.Location{
																		Line: int(1226),
																		Column: int(18),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p13098,
																FreeVars: ast.Identifiers{
																	"j",
																},
															},
															Id: "j",
														},
														CommaFodder: nil,
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.LiteralNull{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
														},
														CommaFodder: nil,
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.LiteralNull{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: nil,
											TailStrictFodder: nil,
										},
									},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
											Comment: []string{},
										},
									},
									BranchFalse: &ast.Conditional{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1227),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(1237),
													Column: int(44),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13098,
											FreeVars: ast.Identifiers{
												"a",
												"acc",
												"aux",
												"b",
												"i",
												"j",
												"keyF",
												"std",
											},
										},
										Cond: &ast.Binary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1227),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(1227),
														Column: int(33),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p13098,
												FreeVars: ast.Identifiers{
													"b",
													"j",
													"std",
												},
											},
											Left: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1227),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1227),
															Column: int(16),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p13098,
													FreeVars: ast.Identifiers{
														"j",
													},
												},
												Id: "j",
											},
											OpFodder: ast.Fodder{},
											Op: ast.BinaryOp(8),
											Right: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1227),
															Column: int(20),
														},
														End: ast.Location{
															Line: int(1227),
															Column: int(33),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p13098,
													FreeVars: ast.Identifiers{
														"b",
														"std",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1227),
																Column: int(20),
															},
															End: ast.Location{
																Line: int(1227),
																Column: int(30),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p13098,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1227),
																	Column: int(20),
																},
																End: ast.Location{
																	Line: int(1227),
																	Column: int(23),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: ast.Fodder{},
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "length",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: ast.Fodder{},
													Id: nil,
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1227),
																			Column: int(31),
																		},
																		End: ast.Location{
																			Line: int(1227),
																			Column: int(32),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13150,
																	FreeVars: ast.Identifiers{
																		"b",
																	},
																},
																Id: "b",
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
										},
										ThenFodder: ast.Fodder{},
										BranchTrue: &ast.Binary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1228),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1228),
														Column: int(20),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p13098,
												FreeVars: ast.Identifiers{
													"a",
													"acc",
													"i",
													"std",
												},
											},
											Left: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1228),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1228),
															Column: int(12),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
															Comment: []string{},
														},
													},
													Ctx: p13098,
													FreeVars: ast.Identifiers{
														"acc",
													},
												},
												Id: "acc",
											},
											OpFodder: ast.Fodder{},
											Op: ast.BinaryOp(3),
											Right: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"a",
														"i",
														"std",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: nil,
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "slice",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: nil,
													Id: nil,
												},
												FodderLeft: nil,
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1228),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(1228),
																			Column: int(16),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13098,
																	FreeVars: ast.Identifiers{
																		"a",
																	},
																},
																Id: "a",
															},
															CommaFodder: nil,
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1228),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(1228),
																			Column: int(18),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13098,
																	FreeVars: ast.Identifiers{
																		"i",
																	},
																},
																Id: "i",
															},
															CommaFodder: nil,
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralNull{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: nil,
																},
															},
															CommaFodder: nil,
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralNull{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: nil,
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: nil,
												TailStrictFodder: nil,
											},
										},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										BranchFalse: &ast.Local{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1230),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1237),
														Column: int(44),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
														Comment: []string{},
													},
												},
												Ctx: p13098,
												FreeVars: ast.Identifiers{
													"a",
													"acc",
													"aux",
													"b",
													"i",
													"j",
													"keyF",
												},
											},
											Binds: ast.LocalBinds{
												ast.LocalBind{
													VarFodder: ast.Fodder{},
													Variable: "ak",
													EqFodder: ast.Fodder{},
													Body: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1230),
																	Column: int(20),
																},
																End: ast.Location{
																	Line: int(1230),
																	Column: int(30),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p13177,
															FreeVars: ast.Identifiers{
																"a",
																"i",
																"keyF",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1230),
																		Column: int(20),
																	},
																	End: ast.Location{
																		Line: int(1230),
																		Column: int(24),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p13177,
																FreeVars: ast.Identifiers{
																	"keyF",
																},
															},
															Id: "keyF",
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1230),
																					Column: int(25),
																				},
																				End: ast.Location{
																					Line: int(1230),
																					Column: int(29),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p13183,
																			FreeVars: ast.Identifiers{
																				"a",
																				"i",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1230),
																						Column: int(25),
																					},
																					End: ast.Location{
																						Line: int(1230),
																						Column: int(26),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p13183,
																				FreeVars: ast.Identifiers{
																					"a",
																				},
																			},
																			Id: "a",
																		},
																		LeftBracketFodder: ast.Fodder{},
																		Index: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1230),
																						Column: int(27),
																					},
																					End: ast.Location{
																						Line: int(1230),
																						Column: int(28),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p13183,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																			},
																			Id: "i",
																		},
																		RightBracketFodder: ast.Fodder{},
																		Id: nil,
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														TrailingComma: false,
														TailStrict: false,
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
													},
													Fun: nil,
													CloseFodder: ast.Fodder{},
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1230),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1230),
															Column: int(30),
														},
														File: p1,
													},
												},
											},
											Body: &ast.Local{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1231),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1237),
															Column: int(44),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
															Comment: []string{},
														},
													},
													Ctx: p13098,
													FreeVars: ast.Identifiers{
														"a",
														"acc",
														"ak",
														"aux",
														"b",
														"i",
														"j",
														"keyF",
													},
												},
												Binds: ast.LocalBinds{
													ast.LocalBind{
														VarFodder: ast.Fodder{},
														Variable: "bk",
														EqFodder: ast.Fodder{},
														Body: &ast.Apply{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1231),
																		Column: int(20),
																	},
																	End: ast.Location{
																		Line: int(1231),
																		Column: int(30),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p13194,
																FreeVars: ast.Identifiers{
																	"b",
																	"j",
																	"keyF",
																},
															},
															Target: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1231),
																			Column: int(20),
																		},
																		End: ast.Location{
																			Line: int(1231),
																			Column: int(24),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13194,
																	FreeVars: ast.Identifiers{
																		"keyF",
																	},
																},
																Id: "keyF",
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Index{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1231),
																						Column: int(25),
																					},
																					End: ast.Location{
																						Line: int(1231),
																						Column: int(29),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p13200,
																				FreeVars: ast.Identifiers{
																					"b",
																					"j",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1231),
																							Column: int(25),
																						},
																						End: ast.Location{
																							Line: int(1231),
																							Column: int(26),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p13200,
																					FreeVars: ast.Identifiers{
																						"b",
																					},
																				},
																				Id: "b",
																			},
																			LeftBracketFodder: ast.Fodder{},
																			Index: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1231),
																							Column: int(27),
																						},
																						End: ast.Location{
																							Line: int(1231),
																							Column: int(28),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p13200,
																					FreeVars: ast.Identifiers{
																						"j",
																					},
																				},
																				Id: "j",
																			},
																			RightBracketFodder: ast.Fodder{},
																			Id: nil,
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															TrailingComma: false,
															TailStrict: false,
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
														},
														Fun: nil,
														CloseFodder: ast.Fodder{},
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1231),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1231),
																Column: int(30),
															},
															File: p1,
														},
													},
												},
												Body: &ast.Conditional{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1232),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(1237),
																Column: int(44),
															},
															File: p1,
														},
														Fodder: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(8),
																Comment: []string{},
															},
														},
														Ctx: p13098,
														FreeVars: ast.Identifiers{
															"a",
															"acc",
															"ak",
															"aux",
															"b",
															"bk",
															"i",
															"j",
														},
													},
													Cond: &ast.Binary{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1232),
																	Column: int(12),
																},
																End: ast.Location{
																	Line: int(1232),
																	Column: int(20),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p13098,
															FreeVars: ast.Identifiers{
																"ak",
																"bk",
															},
														},
														Left: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1232),
																		Column: int(12),
																	},
																	End: ast.Location{
																		Line: int(1232),
																		Column: int(14),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p13098,
																FreeVars: ast.Identifiers{
																	"ak",
																},
															},
															Id: "ak",
														},
														OpFodder: ast.Fodder{},
														Op: ast.BinaryOp(12),
														Right: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1232),
																		Column: int(18),
																	},
																	End: ast.Location{
																		Line: int(1232),
																		Column: int(20),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p13098,
																FreeVars: ast.Identifiers{
																	"bk",
																},
															},
															Id: "bk",
														},
													},
													ThenFodder: ast.Fodder{},
													BranchTrue: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1233),
																	Column: int(11),
																},
																End: ast.Location{
																	Line: int(1233),
																	Column: int(48),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p13098,
															FreeVars: ast.Identifiers{
																"a",
																"acc",
																"aux",
																"b",
																"i",
																"j",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1233),
																		Column: int(11),
																	},
																	End: ast.Location{
																		Line: int(1233),
																		Column: int(14),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(10),
																		Comment: []string{},
																	},
																},
																Ctx: p13098,
																FreeVars: ast.Identifiers{
																	"aux",
																},
															},
															Id: "aux",
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1233),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(1233),
																					Column: int(16),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p13222,
																			FreeVars: ast.Identifiers{
																				"a",
																			},
																		},
																		Id: "a",
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1233),
																					Column: int(18),
																				},
																				End: ast.Location{
																					Line: int(1233),
																					Column: int(19),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p13222,
																			FreeVars: ast.Identifiers{
																				"b",
																			},
																		},
																		Id: "b",
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1233),
																					Column: int(21),
																				},
																				End: ast.Location{
																					Line: int(1233),
																					Column: int(26),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p13222,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																		},
																		Left: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1233),
																						Column: int(21),
																					},
																					End: ast.Location{
																						Line: int(1233),
																						Column: int(22),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p13222,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																			},
																			Id: "i",
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(3),
																		Right: &ast.LiteralNumber{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1233),
																						Column: int(25),
																					},
																					End: ast.Location{
																						Line: int(1233),
																						Column: int(26),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p13222,
																				FreeVars: nil,
																			},
																			OriginalString: "1",
																		},
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1233),
																					Column: int(28),
																				},
																				End: ast.Location{
																					Line: int(1233),
																					Column: int(33),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p13222,
																			FreeVars: ast.Identifiers{
																				"j",
																			},
																		},
																		Left: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1233),
																						Column: int(28),
																					},
																					End: ast.Location{
																						Line: int(1233),
																						Column: int(29),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p13222,
																				FreeVars: ast.Identifiers{
																					"j",
																				},
																			},
																			Id: "j",
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(3),
																		Right: &ast.LiteralNumber{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1233),
																						Column: int(32),
																					},
																					End: ast.Location{
																						Line: int(1233),
																						Column: int(33),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p13222,
																				FreeVars: nil,
																			},
																			OriginalString: "1",
																		},
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1233),
																					Column: int(35),
																				},
																				End: ast.Location{
																					Line: int(1233),
																					Column: int(47),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p13222,
																			FreeVars: ast.Identifiers{
																				"a",
																				"acc",
																				"i",
																			},
																		},
																		Left: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1233),
																						Column: int(35),
																					},
																					End: ast.Location{
																						Line: int(1233),
																						Column: int(38),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p13222,
																				FreeVars: ast.Identifiers{
																					"acc",
																				},
																			},
																			Id: "acc",
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(3),
																		Right: &ast.Array{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1233),
																						Column: int(41),
																					},
																					End: ast.Location{
																						Line: int(1233),
																						Column: int(47),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p13222,
																				FreeVars: ast.Identifiers{
																					"a",
																					"i",
																				},
																			},
																			Elements: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Index{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1233),
																									Column: int(42),
																								},
																								End: ast.Location{
																									Line: int(1233),
																									Column: int(46),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p13244,
																							FreeVars: ast.Identifiers{
																								"a",
																								"i",
																							},
																						},
																						Target: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1233),
																										Column: int(42),
																									},
																									End: ast.Location{
																										Line: int(1233),
																										Column: int(43),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p13244,
																								FreeVars: ast.Identifiers{
																									"a",
																								},
																							},
																							Id: "a",
																						},
																						LeftBracketFodder: ast.Fodder{},
																						Index: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1233),
																										Column: int(44),
																									},
																									End: ast.Location{
																										Line: int(1233),
																										Column: int(45),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p13244,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																							},
																							Id: "i",
																						},
																						RightBracketFodder: ast.Fodder{},
																						Id: nil,
																					},
																					CommaFodder: nil,
																				},
																			},
																			TrailingComma: false,
																			CloseFodder: ast.Fodder{},
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														TrailingComma: false,
														TailStrict: true,
														FodderRight: ast.Fodder{},
														TailStrictFodder: ast.Fodder{},
													},
													ElseFodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
															Comment: []string{},
														},
													},
													BranchFalse: &ast.Conditional{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1234),
																	Column: int(14),
																},
																End: ast.Location{
																	Line: int(1237),
																	Column: int(44),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p13098,
															FreeVars: ast.Identifiers{
																"a",
																"acc",
																"ak",
																"aux",
																"b",
																"bk",
																"i",
																"j",
															},
														},
														Cond: &ast.Binary{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1234),
																		Column: int(17),
																	},
																	End: ast.Location{
																		Line: int(1234),
																		Column: int(24),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p13098,
																FreeVars: ast.Identifiers{
																	"ak",
																	"bk",
																},
															},
															Left: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1234),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(1234),
																			Column: int(19),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13098,
																	FreeVars: ast.Identifiers{
																		"ak",
																	},
																},
																Id: "ak",
															},
															OpFodder: ast.Fodder{},
															Op: ast.BinaryOp(9),
															Right: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1234),
																			Column: int(22),
																		},
																		End: ast.Location{
																			Line: int(1234),
																			Column: int(24),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13098,
																	FreeVars: ast.Identifiers{
																		"bk",
																	},
																},
																Id: "bk",
															},
														},
														ThenFodder: ast.Fodder{},
														BranchTrue: &ast.Apply{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1235),
																		Column: int(11),
																	},
																	End: ast.Location{
																		Line: int(1235),
																		Column: int(44),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p13098,
																FreeVars: ast.Identifiers{
																	"a",
																	"acc",
																	"aux",
																	"b",
																	"i",
																	"j",
																},
															},
															Target: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1235),
																			Column: int(11),
																		},
																		End: ast.Location{
																			Line: int(1235),
																			Column: int(14),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(10),
																			Comment: []string{},
																		},
																	},
																	Ctx: p13098,
																	FreeVars: ast.Identifiers{
																		"aux",
																	},
																},
																Id: "aux",
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1235),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(1235),
																						Column: int(16),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p13266,
																				FreeVars: ast.Identifiers{
																					"a",
																				},
																			},
																			Id: "a",
																		},
																		CommaFodder: ast.Fodder{},
																	},
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1235),
																						Column: int(18),
																					},
																					End: ast.Location{
																						Line: int(1235),
																						Column: int(19),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p13266,
																				FreeVars: ast.Identifiers{
																					"b",
																				},
																			},
																			Id: "b",
																		},
																		CommaFodder: ast.Fodder{},
																	},
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1235),
																						Column: int(21),
																					},
																					End: ast.Location{
																						Line: int(1235),
																						Column: int(26),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p13266,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																			},
																			Left: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1235),
																							Column: int(21),
																						},
																						End: ast.Location{
																							Line: int(1235),
																							Column: int(22),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p13266,
																					FreeVars: ast.Identifiers{
																						"i",
																					},
																				},
																				Id: "i",
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(3),
																			Right: &ast.LiteralNumber{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1235),
																							Column: int(25),
																						},
																						End: ast.Location{
																							Line: int(1235),
																							Column: int(26),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p13266,
																					FreeVars: nil,
																				},
																				OriginalString: "1",
																			},
																		},
																		CommaFodder: ast.Fodder{},
																	},
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1235),
																						Column: int(28),
																					},
																					End: ast.Location{
																						Line: int(1235),
																						Column: int(29),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p13266,
																				FreeVars: ast.Identifiers{
																					"j",
																				},
																			},
																			Id: "j",
																		},
																		CommaFodder: ast.Fodder{},
																	},
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1235),
																						Column: int(31),
																					},
																					End: ast.Location{
																						Line: int(1235),
																						Column: int(43),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p13266,
																				FreeVars: ast.Identifiers{
																					"a",
																					"acc",
																					"i",
																				},
																			},
																			Left: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1235),
																							Column: int(31),
																						},
																						End: ast.Location{
																							Line: int(1235),
																							Column: int(34),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p13266,
																					FreeVars: ast.Identifiers{
																						"acc",
																					},
																				},
																				Id: "acc",
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(3),
																			Right: &ast.Array{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1235),
																							Column: int(37),
																						},
																						End: ast.Location{
																							Line: int(1235),
																							Column: int(43),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p13266,
																					FreeVars: ast.Identifiers{
																						"a",
																						"i",
																					},
																				},
																				Elements: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Index{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1235),
																										Column: int(38),
																									},
																									End: ast.Location{
																										Line: int(1235),
																										Column: int(42),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p13285,
																								FreeVars: ast.Identifiers{
																									"a",
																									"i",
																								},
																							},
																							Target: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1235),
																											Column: int(38),
																										},
																										End: ast.Location{
																											Line: int(1235),
																											Column: int(39),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p13285,
																									FreeVars: ast.Identifiers{
																										"a",
																									},
																								},
																								Id: "a",
																							},
																							LeftBracketFodder: ast.Fodder{},
																							Index: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1235),
																											Column: int(40),
																										},
																										End: ast.Location{
																											Line: int(1235),
																											Column: int(41),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p13285,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																								},
																								Id: "i",
																							},
																							RightBracketFodder: ast.Fodder{},
																							Id: nil,
																						},
																						CommaFodder: nil,
																					},
																				},
																				TrailingComma: false,
																				CloseFodder: ast.Fodder{},
																			},
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															TrailingComma: false,
															TailStrict: true,
															FodderRight: ast.Fodder{},
															TailStrictFodder: ast.Fodder{},
														},
														ElseFodder: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(8),
																Comment: []string{},
															},
														},
														BranchFalse: &ast.Apply{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1237),
																		Column: int(11),
																	},
																	End: ast.Location{
																		Line: int(1237),
																		Column: int(44),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p13098,
																FreeVars: ast.Identifiers{
																	"a",
																	"acc",
																	"aux",
																	"b",
																	"i",
																	"j",
																},
															},
															Target: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1237),
																			Column: int(11),
																		},
																		End: ast.Location{
																			Line: int(1237),
																			Column: int(14),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(10),
																			Comment: []string{},
																		},
																	},
																	Ctx: p13098,
																	FreeVars: ast.Identifiers{
																		"aux",
																	},
																},
																Id: "aux",
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1237),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(1237),
																						Column: int(16),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p13299,
																				FreeVars: ast.Identifiers{
																					"a",
																				},
																			},
																			Id: "a",
																		},
																		CommaFodder: ast.Fodder{},
																	},
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1237),
																						Column: int(18),
																					},
																					End: ast.Location{
																						Line: int(1237),
																						Column: int(19),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p13299,
																				FreeVars: ast.Identifiers{
																					"b",
																				},
																			},
																			Id: "b",
																		},
																		CommaFodder: ast.Fodder{},
																	},
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1237),
																						Column: int(21),
																					},
																					End: ast.Location{
																						Line: int(1237),
																						Column: int(22),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p13299,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																			},
																			Id: "i",
																		},
																		CommaFodder: ast.Fodder{},
																	},
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1237),
																						Column: int(24),
																					},
																					End: ast.Location{
																						Line: int(1237),
																						Column: int(29),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p13299,
																				FreeVars: ast.Identifiers{
																					"j",
																				},
																			},
																			Left: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1237),
																							Column: int(24),
																						},
																						End: ast.Location{
																							Line: int(1237),
																							Column: int(25),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p13299,
																					FreeVars: ast.Identifiers{
																						"j",
																					},
																				},
																				Id: "j",
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(3),
																			Right: &ast.LiteralNumber{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1237),
																							Column: int(28),
																						},
																						End: ast.Location{
																							Line: int(1237),
																							Column: int(29),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p13299,
																					FreeVars: nil,
																				},
																				OriginalString: "1",
																			},
																		},
																		CommaFodder: ast.Fodder{},
																	},
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1237),
																						Column: int(31),
																					},
																					End: ast.Location{
																						Line: int(1237),
																						Column: int(43),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p13299,
																				FreeVars: ast.Identifiers{
																					"acc",
																					"b",
																					"j",
																				},
																			},
																			Left: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1237),
																							Column: int(31),
																						},
																						End: ast.Location{
																							Line: int(1237),
																							Column: int(34),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p13299,
																					FreeVars: ast.Identifiers{
																						"acc",
																					},
																				},
																				Id: "acc",
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(3),
																			Right: &ast.Array{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1237),
																							Column: int(37),
																						},
																						End: ast.Location{
																							Line: int(1237),
																							Column: int(43),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p13299,
																					FreeVars: ast.Identifiers{
																						"b",
																						"j",
																					},
																				},
																				Elements: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Index{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1237),
																										Column: int(38),
																									},
																									End: ast.Location{
																										Line: int(1237),
																										Column: int(42),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p13318,
																								FreeVars: ast.Identifiers{
																									"b",
																									"j",
																								},
																							},
																							Target: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1237),
																											Column: int(38),
																										},
																										End: ast.Location{
																											Line: int(1237),
																											Column: int(39),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p13318,
																									FreeVars: ast.Identifiers{
																										"b",
																									},
																								},
																								Id: "b",
																							},
																							LeftBracketFodder: ast.Fodder{},
																							Index: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1237),
																											Column: int(40),
																										},
																										End: ast.Location{
																											Line: int(1237),
																											Column: int(41),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p13318,
																									FreeVars: ast.Identifiers{
																										"j",
																									},
																								},
																								Id: "j",
																							},
																							RightBracketFodder: ast.Fodder{},
																							Id: nil,
																						},
																						CommaFodder: nil,
																					},
																				},
																				TrailingComma: false,
																				CloseFodder: ast.Fodder{},
																			},
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															TrailingComma: false,
															TailStrict: true,
															FodderRight: ast.Fodder{},
															TailStrictFodder: ast.Fodder{},
														},
													},
												},
											},
										},
									},
								},
							},
							Fun: nil,
							CloseFodder: nil,
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
						},
					},
					Body: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1238),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1238),
									Column: int(24),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p13085,
							FreeVars: ast.Identifiers{
								"a",
								"aux",
								"b",
							},
						},
						Target: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1238),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1238),
										Column: int(8),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
										Comment: []string{},
									},
								},
								Ctx: p13085,
								FreeVars: ast.Identifiers{
									"aux",
								},
							},
							Id: "aux",
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1238),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1238),
													Column: int(10),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13331,
											FreeVars: ast.Identifiers{
												"a",
											},
										},
										Id: "a",
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1238),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(1238),
													Column: int(13),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13331,
											FreeVars: ast.Identifiers{
												"b",
											},
										},
										Id: "b",
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.LiteralNumber{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1238),
													Column: int(15),
												},
												End: ast.Location{
													Line: int(1238),
													Column: int(16),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13331,
											FreeVars: nil,
										},
										OriginalString: "0",
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.LiteralNumber{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1238),
													Column: int(18),
												},
												End: ast.Location{
													Line: int(1238),
													Column: int(19),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13331,
											FreeVars: nil,
										},
										OriginalString: "0",
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Array{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1238),
													Column: int(21),
												},
												End: ast.Location{
													Line: int(1238),
													Column: int(23),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13331,
											FreeVars: nil,
										},
										Elements: nil,
										TrailingComma: false,
										CloseFodder: ast.Fodder{},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(1222),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1238),
					Column: int(24),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "setInter",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"id",
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "a",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1240),
								Column: int(12),
							},
							End: ast.Location{
								Line: int(1240),
								Column: int(13),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "b",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1240),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(1240),
								Column: int(16),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "keyF",
						EqFodder: ast.Fodder{},
						DefaultArg: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1240),
										Column: int(23),
									},
									End: ast.Location{
										Line: int(1240),
										Column: int(25),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p13343,
								FreeVars: ast.Identifiers{
									"id",
								},
							},
							Id: "id",
						},
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1240),
								Column: int(18),
							},
							End: ast.Location{
								Line: int(1240),
								Column: int(25),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1241),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1251),
								Column: int(24),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p13343,
						FreeVars: ast.Identifiers{
							"a",
							"b",
							"keyF",
							"std",
						},
					},
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: nil,
							Variable: "aux",
							EqFodder: nil,
							Body: &ast.Function{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1241),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(1250),
											Column: int(35),
										},
										File: p1,
									},
									Fodder: nil,
									Ctx: p13350,
									FreeVars: ast.Identifiers{
										"aux",
										"keyF",
										"std",
									},
								},
								ParenLeftFodder: ast.Fodder{},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "a",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1241),
												Column: int(15),
											},
											End: ast.Location{
												Line: int(1241),
												Column: int(16),
											},
											File: p1,
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "b",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1241),
												Column: int(18),
											},
											End: ast.Location{
												Line: int(1241),
												Column: int(19),
											},
											File: p1,
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "i",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1241),
												Column: int(21),
											},
											End: ast.Location{
												Line: int(1241),
												Column: int(22),
											},
											File: p1,
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "j",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1241),
												Column: int(24),
											},
											End: ast.Location{
												Line: int(1241),
												Column: int(25),
											},
											File: p1,
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "acc",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: nil,
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1241),
												Column: int(27),
											},
											End: ast.Location{
												Line: int(1241),
												Column: int(30),
											},
											File: p1,
										},
									},
								},
								TrailingComma: false,
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Conditional{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1242),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1250),
												Column: int(35),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										Ctx: p13355,
										FreeVars: ast.Identifiers{
											"a",
											"acc",
											"aux",
											"b",
											"i",
											"j",
											"keyF",
											"std",
										},
									},
									Cond: &ast.Binary{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1242),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(1242),
													Column: int(50),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13355,
											FreeVars: ast.Identifiers{
												"a",
												"b",
												"i",
												"j",
												"std",
											},
										},
										Left: &ast.Binary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1242),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(1242),
														Column: int(28),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p13355,
												FreeVars: ast.Identifiers{
													"a",
													"i",
													"std",
												},
											},
											Left: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1242),
															Column: int(10),
														},
														End: ast.Location{
															Line: int(1242),
															Column: int(11),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p13355,
													FreeVars: ast.Identifiers{
														"i",
													},
												},
												Id: "i",
											},
											OpFodder: ast.Fodder{},
											Op: ast.BinaryOp(8),
											Right: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1242),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1242),
															Column: int(28),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p13355,
													FreeVars: ast.Identifiers{
														"a",
														"std",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1242),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1242),
																Column: int(25),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p13355,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1242),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(1242),
																	Column: int(18),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: ast.Fodder{},
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "length",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: ast.Fodder{},
													Id: nil,
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1242),
																			Column: int(26),
																		},
																		End: ast.Location{
																			Line: int(1242),
																			Column: int(27),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13372,
																	FreeVars: ast.Identifiers{
																		"a",
																	},
																},
																Id: "a",
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
										},
										OpFodder: ast.Fodder{},
										Op: ast.BinaryOp(18),
										Right: &ast.Binary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1242),
														Column: int(32),
													},
													End: ast.Location{
														Line: int(1242),
														Column: int(50),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p13355,
												FreeVars: ast.Identifiers{
													"b",
													"j",
													"std",
												},
											},
											Left: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1242),
															Column: int(32),
														},
														End: ast.Location{
															Line: int(1242),
															Column: int(33),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p13355,
													FreeVars: ast.Identifiers{
														"j",
													},
												},
												Id: "j",
											},
											OpFodder: ast.Fodder{},
											Op: ast.BinaryOp(8),
											Right: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1242),
															Column: int(37),
														},
														End: ast.Location{
															Line: int(1242),
															Column: int(50),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p13355,
													FreeVars: ast.Identifiers{
														"b",
														"std",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1242),
																Column: int(37),
															},
															End: ast.Location{
																Line: int(1242),
																Column: int(47),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p13355,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1242),
																	Column: int(37),
																},
																End: ast.Location{
																	Line: int(1242),
																	Column: int(40),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: ast.Fodder{},
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "length",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: ast.Fodder{},
													Id: nil,
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1242),
																			Column: int(48),
																		},
																		End: ast.Location{
																			Line: int(1242),
																			Column: int(49),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13387,
																	FreeVars: ast.Identifiers{
																		"b",
																	},
																},
																Id: "b",
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
										},
									},
									ThenFodder: ast.Fodder{},
									BranchTrue: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1243),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1243),
													Column: int(12),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
													Comment: []string{},
												},
											},
											Ctx: p13355,
											FreeVars: ast.Identifiers{
												"acc",
											},
										},
										Id: "acc",
									},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
											Comment: []string{},
										},
									},
									BranchFalse: &ast.Conditional{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1245),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1250),
													Column: int(35),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
													Comment: []string{},
												},
											},
											Ctx: p13355,
											FreeVars: ast.Identifiers{
												"a",
												"acc",
												"aux",
												"b",
												"i",
												"j",
												"keyF",
											},
										},
										Cond: &ast.Binary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1245),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(1245),
														Column: int(36),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p13355,
												FreeVars: ast.Identifiers{
													"a",
													"b",
													"i",
													"j",
													"keyF",
												},
											},
											Left: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1245),
															Column: int(12),
														},
														End: ast.Location{
															Line: int(1245),
															Column: int(22),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p13355,
													FreeVars: ast.Identifiers{
														"a",
														"i",
														"keyF",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1245),
																Column: int(12),
															},
															End: ast.Location{
																Line: int(1245),
																Column: int(16),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p13355,
														FreeVars: ast.Identifiers{
															"keyF",
														},
													},
													Id: "keyF",
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Index{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1245),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(1245),
																			Column: int(21),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13404,
																	FreeVars: ast.Identifiers{
																		"a",
																		"i",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1245),
																				Column: int(17),
																			},
																			End: ast.Location{
																				Line: int(1245),
																				Column: int(18),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13404,
																		FreeVars: ast.Identifiers{
																			"a",
																		},
																	},
																	Id: "a",
																},
																LeftBracketFodder: ast.Fodder{},
																Index: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1245),
																				Column: int(19),
																			},
																			End: ast.Location{
																				Line: int(1245),
																				Column: int(20),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13404,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																	},
																	Id: "i",
																},
																RightBracketFodder: ast.Fodder{},
																Id: nil,
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
											OpFodder: ast.Fodder{},
											Op: ast.BinaryOp(12),
											Right: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1245),
															Column: int(26),
														},
														End: ast.Location{
															Line: int(1245),
															Column: int(36),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p13355,
													FreeVars: ast.Identifiers{
														"b",
														"j",
														"keyF",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1245),
																Column: int(26),
															},
															End: ast.Location{
																Line: int(1245),
																Column: int(30),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p13355,
														FreeVars: ast.Identifiers{
															"keyF",
														},
													},
													Id: "keyF",
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Index{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1245),
																			Column: int(31),
																		},
																		End: ast.Location{
																			Line: int(1245),
																			Column: int(35),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13416,
																	FreeVars: ast.Identifiers{
																		"b",
																		"j",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1245),
																				Column: int(31),
																			},
																			End: ast.Location{
																				Line: int(1245),
																				Column: int(32),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13416,
																		FreeVars: ast.Identifiers{
																			"b",
																		},
																	},
																	Id: "b",
																},
																LeftBracketFodder: ast.Fodder{},
																Index: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1245),
																				Column: int(33),
																			},
																			End: ast.Location{
																				Line: int(1245),
																				Column: int(34),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13416,
																		FreeVars: ast.Identifiers{
																			"j",
																		},
																	},
																	Id: "j",
																},
																RightBracketFodder: ast.Fodder{},
																Id: nil,
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
										},
										ThenFodder: ast.Fodder{},
										BranchTrue: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1246),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(1246),
														Column: int(48),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p13355,
												FreeVars: ast.Identifiers{
													"a",
													"acc",
													"aux",
													"b",
													"i",
													"j",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1246),
															Column: int(11),
														},
														End: ast.Location{
															Line: int(1246),
															Column: int(14),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(10),
															Comment: []string{},
														},
													},
													Ctx: p13355,
													FreeVars: ast.Identifiers{
														"aux",
													},
												},
												Id: "aux",
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1246),
																		Column: int(15),
																	},
																	End: ast.Location{
																		Line: int(1246),
																		Column: int(16),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p13429,
																FreeVars: ast.Identifiers{
																	"a",
																},
															},
															Id: "a",
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1246),
																		Column: int(18),
																	},
																	End: ast.Location{
																		Line: int(1246),
																		Column: int(19),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p13429,
																FreeVars: ast.Identifiers{
																	"b",
																},
															},
															Id: "b",
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Binary{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1246),
																		Column: int(21),
																	},
																	End: ast.Location{
																		Line: int(1246),
																		Column: int(26),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p13429,
																FreeVars: ast.Identifiers{
																	"i",
																},
															},
															Left: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1246),
																			Column: int(21),
																		},
																		End: ast.Location{
																			Line: int(1246),
																			Column: int(22),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13429,
																	FreeVars: ast.Identifiers{
																		"i",
																	},
																},
																Id: "i",
															},
															OpFodder: ast.Fodder{},
															Op: ast.BinaryOp(3),
															Right: &ast.LiteralNumber{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1246),
																			Column: int(25),
																		},
																		End: ast.Location{
																			Line: int(1246),
																			Column: int(26),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13429,
																	FreeVars: nil,
																},
																OriginalString: "1",
															},
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Binary{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1246),
																		Column: int(28),
																	},
																	End: ast.Location{
																		Line: int(1246),
																		Column: int(33),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p13429,
																FreeVars: ast.Identifiers{
																	"j",
																},
															},
															Left: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1246),
																			Column: int(28),
																		},
																		End: ast.Location{
																			Line: int(1246),
																			Column: int(29),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13429,
																	FreeVars: ast.Identifiers{
																		"j",
																	},
																},
																Id: "j",
															},
															OpFodder: ast.Fodder{},
															Op: ast.BinaryOp(3),
															Right: &ast.LiteralNumber{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1246),
																			Column: int(32),
																		},
																		End: ast.Location{
																			Line: int(1246),
																			Column: int(33),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13429,
																	FreeVars: nil,
																},
																OriginalString: "1",
															},
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Binary{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1246),
																		Column: int(35),
																	},
																	End: ast.Location{
																		Line: int(1246),
																		Column: int(47),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p13429,
																FreeVars: ast.Identifiers{
																	"a",
																	"acc",
																	"i",
																},
															},
															Left: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1246),
																			Column: int(35),
																		},
																		End: ast.Location{
																			Line: int(1246),
																			Column: int(38),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13429,
																	FreeVars: ast.Identifiers{
																		"acc",
																	},
																},
																Id: "acc",
															},
															OpFodder: ast.Fodder{},
															Op: ast.BinaryOp(3),
															Right: &ast.Array{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1246),
																			Column: int(41),
																		},
																		End: ast.Location{
																			Line: int(1246),
																			Column: int(47),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13429,
																	FreeVars: ast.Identifiers{
																		"a",
																		"i",
																	},
																},
																Elements: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Index{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1246),
																						Column: int(42),
																					},
																					End: ast.Location{
																						Line: int(1246),
																						Column: int(46),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p13451,
																				FreeVars: ast.Identifiers{
																					"a",
																					"i",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1246),
																							Column: int(42),
																						},
																						End: ast.Location{
																							Line: int(1246),
																							Column: int(43),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p13451,
																					FreeVars: ast.Identifiers{
																						"a",
																					},
																				},
																				Id: "a",
																			},
																			LeftBracketFodder: ast.Fodder{},
																			Index: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1246),
																							Column: int(44),
																						},
																						End: ast.Location{
																							Line: int(1246),
																							Column: int(45),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p13451,
																					FreeVars: ast.Identifiers{
																						"i",
																					},
																				},
																				Id: "i",
																			},
																			RightBracketFodder: ast.Fodder{},
																			Id: nil,
																		},
																		CommaFodder: nil,
																	},
																},
																TrailingComma: false,
																CloseFodder: ast.Fodder{},
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: true,
											FodderRight: ast.Fodder{},
											TailStrictFodder: ast.Fodder{},
										},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(8),
												Comment: []string{},
											},
										},
										BranchFalse: &ast.Conditional{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1247),
														Column: int(14),
													},
													End: ast.Location{
														Line: int(1250),
														Column: int(35),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p13355,
												FreeVars: ast.Identifiers{
													"a",
													"acc",
													"aux",
													"b",
													"i",
													"j",
													"keyF",
												},
											},
											Cond: &ast.Binary{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1247),
															Column: int(17),
														},
														End: ast.Location{
															Line: int(1247),
															Column: int(40),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p13355,
													FreeVars: ast.Identifiers{
														"a",
														"b",
														"i",
														"j",
														"keyF",
													},
												},
												Left: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1247),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(1247),
																Column: int(27),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p13355,
														FreeVars: ast.Identifiers{
															"a",
															"i",
															"keyF",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1247),
																	Column: int(17),
																},
																End: ast.Location{
																	Line: int(1247),
																	Column: int(21),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p13355,
															FreeVars: ast.Identifiers{
																"keyF",
															},
														},
														Id: "keyF",
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1247),
																				Column: int(22),
																			},
																			End: ast.Location{
																				Line: int(1247),
																				Column: int(26),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13468,
																		FreeVars: ast.Identifiers{
																			"a",
																			"i",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1247),
																					Column: int(22),
																				},
																				End: ast.Location{
																					Line: int(1247),
																					Column: int(23),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p13468,
																			FreeVars: ast.Identifiers{
																				"a",
																			},
																		},
																		Id: "a",
																	},
																	LeftBracketFodder: ast.Fodder{},
																	Index: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1247),
																					Column: int(24),
																				},
																				End: ast.Location{
																					Line: int(1247),
																					Column: int(25),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p13468,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																		},
																		Id: "i",
																	},
																	RightBracketFodder: ast.Fodder{},
																	Id: nil,
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
												},
												OpFodder: ast.Fodder{},
												Op: ast.BinaryOp(9),
												Right: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1247),
																Column: int(30),
															},
															End: ast.Location{
																Line: int(1247),
																Column: int(40),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p13355,
														FreeVars: ast.Identifiers{
															"b",
															"j",
															"keyF",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1247),
																	Column: int(30),
																},
																End: ast.Location{
																	Line: int(1247),
																	Column: int(34),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p13355,
															FreeVars: ast.Identifiers{
																"keyF",
															},
														},
														Id: "keyF",
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1247),
																				Column: int(35),
																			},
																			End: ast.Location{
																				Line: int(1247),
																				Column: int(39),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13480,
																		FreeVars: ast.Identifiers{
																			"b",
																			"j",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1247),
																					Column: int(35),
																				},
																				End: ast.Location{
																					Line: int(1247),
																					Column: int(36),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p13480,
																			FreeVars: ast.Identifiers{
																				"b",
																			},
																		},
																		Id: "b",
																	},
																	LeftBracketFodder: ast.Fodder{},
																	Index: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1247),
																					Column: int(37),
																				},
																				End: ast.Location{
																					Line: int(1247),
																					Column: int(38),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p13480,
																			FreeVars: ast.Identifiers{
																				"j",
																			},
																		},
																		Id: "j",
																	},
																	RightBracketFodder: ast.Fodder{},
																	Id: nil,
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
												},
											},
											ThenFodder: ast.Fodder{},
											BranchTrue: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1248),
															Column: int(11),
														},
														End: ast.Location{
															Line: int(1248),
															Column: int(35),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p13355,
													FreeVars: ast.Identifiers{
														"a",
														"acc",
														"aux",
														"b",
														"i",
														"j",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1248),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(1248),
																Column: int(14),
															},
															File: p1,
														},
														Fodder: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(10),
																Comment: []string{},
															},
														},
														Ctx: p13355,
														FreeVars: ast.Identifiers{
															"aux",
														},
													},
													Id: "aux",
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1248),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(1248),
																			Column: int(16),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13493,
																	FreeVars: ast.Identifiers{
																		"a",
																	},
																},
																Id: "a",
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1248),
																			Column: int(18),
																		},
																		End: ast.Location{
																			Line: int(1248),
																			Column: int(19),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13493,
																	FreeVars: ast.Identifiers{
																		"b",
																	},
																},
																Id: "b",
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Binary{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1248),
																			Column: int(21),
																		},
																		End: ast.Location{
																			Line: int(1248),
																			Column: int(26),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13493,
																	FreeVars: ast.Identifiers{
																		"i",
																	},
																},
																Left: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1248),
																				Column: int(21),
																			},
																			End: ast.Location{
																				Line: int(1248),
																				Column: int(22),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13493,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																	},
																	Id: "i",
																},
																OpFodder: ast.Fodder{},
																Op: ast.BinaryOp(3),
																Right: &ast.LiteralNumber{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1248),
																				Column: int(25),
																			},
																			End: ast.Location{
																				Line: int(1248),
																				Column: int(26),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13493,
																		FreeVars: nil,
																	},
																	OriginalString: "1",
																},
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1248),
																			Column: int(28),
																		},
																		End: ast.Location{
																			Line: int(1248),
																			Column: int(29),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13493,
																	FreeVars: ast.Identifiers{
																		"j",
																	},
																},
																Id: "j",
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1248),
																			Column: int(31),
																		},
																		End: ast.Location{
																			Line: int(1248),
																			Column: int(34),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13493,
																	FreeVars: ast.Identifiers{
																		"acc",
																	},
																},
																Id: "acc",
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: true,
												FodderRight: ast.Fodder{},
												TailStrictFodder: ast.Fodder{},
											},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
													Comment: []string{},
												},
											},
											BranchFalse: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1250),
															Column: int(11),
														},
														End: ast.Location{
															Line: int(1250),
															Column: int(35),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p13355,
													FreeVars: ast.Identifiers{
														"a",
														"acc",
														"aux",
														"b",
														"i",
														"j",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1250),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(1250),
																Column: int(14),
															},
															File: p1,
														},
														Fodder: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(10),
																Comment: []string{},
															},
														},
														Ctx: p13355,
														FreeVars: ast.Identifiers{
															"aux",
														},
													},
													Id: "aux",
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1250),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(1250),
																			Column: int(16),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13514,
																	FreeVars: ast.Identifiers{
																		"a",
																	},
																},
																Id: "a",
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1250),
																			Column: int(18),
																		},
																		End: ast.Location{
																			Line: int(1250),
																			Column: int(19),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13514,
																	FreeVars: ast.Identifiers{
																		"b",
																	},
																},
																Id: "b",
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1250),
																			Column: int(21),
																		},
																		End: ast.Location{
																			Line: int(1250),
																			Column: int(22),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13514,
																	FreeVars: ast.Identifiers{
																		"i",
																	},
																},
																Id: "i",
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Binary{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1250),
																			Column: int(24),
																		},
																		End: ast.Location{
																			Line: int(1250),
																			Column: int(29),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13514,
																	FreeVars: ast.Identifiers{
																		"j",
																	},
																},
																Left: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1250),
																				Column: int(24),
																			},
																			End: ast.Location{
																				Line: int(1250),
																				Column: int(25),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13514,
																		FreeVars: ast.Identifiers{
																			"j",
																		},
																	},
																	Id: "j",
																},
																OpFodder: ast.Fodder{},
																Op: ast.BinaryOp(3),
																Right: &ast.LiteralNumber{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1250),
																				Column: int(28),
																			},
																			End: ast.Location{
																				Line: int(1250),
																				Column: int(29),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13514,
																		FreeVars: nil,
																	},
																	OriginalString: "1",
																},
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1250),
																			Column: int(31),
																		},
																		End: ast.Location{
																			Line: int(1250),
																			Column: int(34),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13514,
																	FreeVars: ast.Identifiers{
																		"acc",
																	},
																},
																Id: "acc",
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: true,
												FodderRight: ast.Fodder{},
												TailStrictFodder: ast.Fodder{},
											},
										},
									},
								},
							},
							Fun: nil,
							CloseFodder: nil,
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
						},
					},
					Body: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1251),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1251),
									Column: int(24),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p13343,
							FreeVars: ast.Identifiers{
								"a",
								"aux",
								"b",
							},
						},
						Target: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1251),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1251),
										Column: int(8),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
										Comment: []string{},
									},
								},
								Ctx: p13343,
								FreeVars: ast.Identifiers{
									"aux",
								},
							},
							Id: "aux",
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1251),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1251),
													Column: int(10),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13534,
											FreeVars: ast.Identifiers{
												"a",
											},
										},
										Id: "a",
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1251),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(1251),
													Column: int(13),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13534,
											FreeVars: ast.Identifiers{
												"b",
											},
										},
										Id: "b",
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.LiteralNumber{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1251),
													Column: int(15),
												},
												End: ast.Location{
													Line: int(1251),
													Column: int(16),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13534,
											FreeVars: nil,
										},
										OriginalString: "0",
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.LiteralNumber{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1251),
													Column: int(18),
												},
												End: ast.Location{
													Line: int(1251),
													Column: int(19),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13534,
											FreeVars: nil,
										},
										OriginalString: "0",
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Array{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1251),
													Column: int(21),
												},
												End: ast.Location{
													Line: int(1251),
													Column: int(23),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13534,
											FreeVars: nil,
										},
										Elements: nil,
										TrailingComma: false,
										CloseFodder: ast.Fodder{},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: true,
						FodderRight: ast.Fodder{},
						TailStrictFodder: ast.Fodder{},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(1240),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1251),
					Column: int(24),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "setDiff",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"id",
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "a",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1253),
								Column: int(11),
							},
							End: ast.Location{
								Line: int(1253),
								Column: int(12),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "b",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1253),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(1253),
								Column: int(15),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "keyF",
						EqFodder: ast.Fodder{},
						DefaultArg: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1253),
										Column: int(22),
									},
									End: ast.Location{
										Line: int(1253),
										Column: int(24),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p13546,
								FreeVars: ast.Identifiers{
									"id",
								},
							},
							Id: "id",
						},
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1253),
								Column: int(17),
							},
							End: ast.Location{
								Line: int(1253),
								Column: int(24),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1254),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1266),
								Column: int(24),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p13546,
						FreeVars: ast.Identifiers{
							"a",
							"b",
							"keyF",
							"std",
						},
					},
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: nil,
							Variable: "aux",
							EqFodder: nil,
							Body: &ast.Function{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1254),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(1265),
											Column: int(35),
										},
										File: p1,
									},
									Fodder: nil,
									Ctx: p13553,
									FreeVars: ast.Identifiers{
										"aux",
										"keyF",
										"std",
									},
								},
								ParenLeftFodder: ast.Fodder{},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "a",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1254),
												Column: int(15),
											},
											End: ast.Location{
												Line: int(1254),
												Column: int(16),
											},
											File: p1,
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "b",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1254),
												Column: int(18),
											},
											End: ast.Location{
												Line: int(1254),
												Column: int(19),
											},
											File: p1,
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "i",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1254),
												Column: int(21),
											},
											End: ast.Location{
												Line: int(1254),
												Column: int(22),
											},
											File: p1,
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "j",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1254),
												Column: int(24),
											},
											End: ast.Location{
												Line: int(1254),
												Column: int(25),
											},
											File: p1,
										},
									},
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "acc",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: nil,
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1254),
												Column: int(27),
											},
											End: ast.Location{
												Line: int(1254),
												Column: int(30),
											},
											File: p1,
										},
									},
								},
								TrailingComma: false,
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Conditional{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1255),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1265),
												Column: int(35),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										Ctx: p13558,
										FreeVars: ast.Identifiers{
											"a",
											"acc",
											"aux",
											"b",
											"i",
											"j",
											"keyF",
											"std",
										},
									},
									Cond: &ast.Binary{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1255),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(1255),
													Column: int(28),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13558,
											FreeVars: ast.Identifiers{
												"a",
												"i",
												"std",
											},
										},
										Left: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1255),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(1255),
														Column: int(11),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p13558,
												FreeVars: ast.Identifiers{
													"i",
												},
											},
											Id: "i",
										},
										OpFodder: ast.Fodder{},
										Op: ast.BinaryOp(8),
										Right: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1255),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(1255),
														Column: int(28),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p13558,
												FreeVars: ast.Identifiers{
													"a",
													"std",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1255),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1255),
															Column: int(25),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p13558,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1255),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1255),
																Column: int(18),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "length",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1255),
																		Column: int(26),
																	},
																	End: ast.Location{
																		Line: int(1255),
																		Column: int(27),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p13573,
																FreeVars: ast.Identifiers{
																	"a",
																},
															},
															Id: "a",
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
									},
									ThenFodder: ast.Fodder{},
									BranchTrue: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1256),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1256),
													Column: int(12),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
													Comment: []string{},
												},
											},
											Ctx: p13558,
											FreeVars: ast.Identifiers{
												"acc",
											},
										},
										Id: "acc",
									},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
											Comment: []string{},
										},
									},
									BranchFalse: &ast.Conditional{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1257),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(1265),
													Column: int(35),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13558,
											FreeVars: ast.Identifiers{
												"a",
												"acc",
												"aux",
												"b",
												"i",
												"j",
												"keyF",
												"std",
											},
										},
										Cond: &ast.Binary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1257),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(1257),
														Column: int(33),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p13558,
												FreeVars: ast.Identifiers{
													"b",
													"j",
													"std",
												},
											},
											Left: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1257),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1257),
															Column: int(16),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p13558,
													FreeVars: ast.Identifiers{
														"j",
													},
												},
												Id: "j",
											},
											OpFodder: ast.Fodder{},
											Op: ast.BinaryOp(8),
											Right: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1257),
															Column: int(20),
														},
														End: ast.Location{
															Line: int(1257),
															Column: int(33),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p13558,
													FreeVars: ast.Identifiers{
														"b",
														"std",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1257),
																Column: int(20),
															},
															End: ast.Location{
																Line: int(1257),
																Column: int(30),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p13558,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1257),
																	Column: int(20),
																},
																End: ast.Location{
																	Line: int(1257),
																	Column: int(23),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: ast.Fodder{},
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "length",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: ast.Fodder{},
													Id: nil,
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1257),
																			Column: int(31),
																		},
																		End: ast.Location{
																			Line: int(1257),
																			Column: int(32),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13594,
																	FreeVars: ast.Identifiers{
																		"b",
																	},
																},
																Id: "b",
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
										},
										ThenFodder: ast.Fodder{},
										BranchTrue: &ast.Binary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1258),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1258),
														Column: int(20),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p13558,
												FreeVars: ast.Identifiers{
													"a",
													"acc",
													"i",
													"std",
												},
											},
											Left: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1258),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1258),
															Column: int(12),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
															Comment: []string{},
														},
													},
													Ctx: p13558,
													FreeVars: ast.Identifiers{
														"acc",
													},
												},
												Id: "acc",
											},
											OpFodder: ast.Fodder{},
											Op: ast.BinaryOp(3),
											Right: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"a",
														"i",
														"std",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: nil,
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "slice",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: nil,
													Id: nil,
												},
												FodderLeft: nil,
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1258),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(1258),
																			Column: int(16),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13558,
																	FreeVars: ast.Identifiers{
																		"a",
																	},
																},
																Id: "a",
															},
															CommaFodder: nil,
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1258),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(1258),
																			Column: int(18),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13558,
																	FreeVars: ast.Identifiers{
																		"i",
																	},
																},
																Id: "i",
															},
															CommaFodder: nil,
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralNull{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: nil,
																},
															},
															CommaFodder: nil,
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralNull{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: nil,
																},
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: nil,
												TailStrictFodder: nil,
											},
										},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										BranchFalse: &ast.Conditional{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1260),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1265),
														Column: int(35),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
														Comment: []string{},
													},
												},
												Ctx: p13558,
												FreeVars: ast.Identifiers{
													"a",
													"acc",
													"aux",
													"b",
													"i",
													"j",
													"keyF",
												},
											},
											Cond: &ast.Binary{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1260),
															Column: int(12),
														},
														End: ast.Location{
															Line: int(1260),
															Column: int(36),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p13558,
													FreeVars: ast.Identifiers{
														"a",
														"b",
														"i",
														"j",
														"keyF",
													},
												},
												Left: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1260),
																Column: int(12),
															},
															End: ast.Location{
																Line: int(1260),
																Column: int(22),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p13558,
														FreeVars: ast.Identifiers{
															"a",
															"i",
															"keyF",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1260),
																	Column: int(12),
																},
																End: ast.Location{
																	Line: int(1260),
																	Column: int(16),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p13558,
															FreeVars: ast.Identifiers{
																"keyF",
															},
														},
														Id: "keyF",
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1260),
																				Column: int(17),
																			},
																			End: ast.Location{
																				Line: int(1260),
																				Column: int(21),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13627,
																		FreeVars: ast.Identifiers{
																			"a",
																			"i",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1260),
																					Column: int(17),
																				},
																				End: ast.Location{
																					Line: int(1260),
																					Column: int(18),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p13627,
																			FreeVars: ast.Identifiers{
																				"a",
																			},
																		},
																		Id: "a",
																	},
																	LeftBracketFodder: ast.Fodder{},
																	Index: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1260),
																					Column: int(19),
																				},
																				End: ast.Location{
																					Line: int(1260),
																					Column: int(20),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p13627,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																		},
																		Id: "i",
																	},
																	RightBracketFodder: ast.Fodder{},
																	Id: nil,
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
												},
												OpFodder: ast.Fodder{},
												Op: ast.BinaryOp(12),
												Right: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1260),
																Column: int(26),
															},
															End: ast.Location{
																Line: int(1260),
																Column: int(36),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p13558,
														FreeVars: ast.Identifiers{
															"b",
															"j",
															"keyF",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1260),
																	Column: int(26),
																},
																End: ast.Location{
																	Line: int(1260),
																	Column: int(30),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p13558,
															FreeVars: ast.Identifiers{
																"keyF",
															},
														},
														Id: "keyF",
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1260),
																				Column: int(31),
																			},
																			End: ast.Location{
																				Line: int(1260),
																				Column: int(35),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13639,
																		FreeVars: ast.Identifiers{
																			"b",
																			"j",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1260),
																					Column: int(31),
																				},
																				End: ast.Location{
																					Line: int(1260),
																					Column: int(32),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p13639,
																			FreeVars: ast.Identifiers{
																				"b",
																			},
																		},
																		Id: "b",
																	},
																	LeftBracketFodder: ast.Fodder{},
																	Index: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1260),
																					Column: int(33),
																				},
																				End: ast.Location{
																					Line: int(1260),
																					Column: int(34),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p13639,
																			FreeVars: ast.Identifiers{
																				"j",
																			},
																		},
																		Id: "j",
																	},
																	RightBracketFodder: ast.Fodder{},
																	Id: nil,
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
												},
											},
											ThenFodder: ast.Fodder{},
											BranchTrue: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1261),
															Column: int(11),
														},
														End: ast.Location{
															Line: int(1261),
															Column: int(39),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p13558,
													FreeVars: ast.Identifiers{
														"a",
														"acc",
														"aux",
														"b",
														"i",
														"j",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1261),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(1261),
																Column: int(14),
															},
															File: p1,
														},
														Fodder: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(10),
																Comment: []string{},
															},
														},
														Ctx: p13558,
														FreeVars: ast.Identifiers{
															"aux",
														},
													},
													Id: "aux",
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1261),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(1261),
																			Column: int(16),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13652,
																	FreeVars: ast.Identifiers{
																		"a",
																	},
																},
																Id: "a",
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1261),
																			Column: int(18),
																		},
																		End: ast.Location{
																			Line: int(1261),
																			Column: int(19),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13652,
																	FreeVars: ast.Identifiers{
																		"b",
																	},
																},
																Id: "b",
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Binary{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1261),
																			Column: int(21),
																		},
																		End: ast.Location{
																			Line: int(1261),
																			Column: int(26),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13652,
																	FreeVars: ast.Identifiers{
																		"i",
																	},
																},
																Left: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1261),
																				Column: int(21),
																			},
																			End: ast.Location{
																				Line: int(1261),
																				Column: int(22),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13652,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																	},
																	Id: "i",
																},
																OpFodder: ast.Fodder{},
																Op: ast.BinaryOp(3),
																Right: &ast.LiteralNumber{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1261),
																				Column: int(25),
																			},
																			End: ast.Location{
																				Line: int(1261),
																				Column: int(26),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13652,
																		FreeVars: nil,
																	},
																	OriginalString: "1",
																},
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Binary{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1261),
																			Column: int(28),
																		},
																		End: ast.Location{
																			Line: int(1261),
																			Column: int(33),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13652,
																	FreeVars: ast.Identifiers{
																		"j",
																	},
																},
																Left: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1261),
																				Column: int(28),
																			},
																			End: ast.Location{
																				Line: int(1261),
																				Column: int(29),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13652,
																		FreeVars: ast.Identifiers{
																			"j",
																		},
																	},
																	Id: "j",
																},
																OpFodder: ast.Fodder{},
																Op: ast.BinaryOp(3),
																Right: &ast.LiteralNumber{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1261),
																				Column: int(32),
																			},
																			End: ast.Location{
																				Line: int(1261),
																				Column: int(33),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13652,
																		FreeVars: nil,
																	},
																	OriginalString: "1",
																},
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1261),
																			Column: int(35),
																		},
																		End: ast.Location{
																			Line: int(1261),
																			Column: int(38),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13652,
																	FreeVars: ast.Identifiers{
																		"acc",
																	},
																},
																Id: "acc",
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: true,
												FodderRight: ast.Fodder{},
												TailStrictFodder: ast.Fodder{},
											},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
													Comment: []string{},
												},
											},
											BranchFalse: &ast.Conditional{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1262),
															Column: int(14),
														},
														End: ast.Location{
															Line: int(1265),
															Column: int(35),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p13558,
													FreeVars: ast.Identifiers{
														"a",
														"acc",
														"aux",
														"b",
														"i",
														"j",
														"keyF",
													},
												},
												Cond: &ast.Binary{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1262),
																Column: int(17),
															},
															End: ast.Location{
																Line: int(1262),
																Column: int(40),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p13558,
														FreeVars: ast.Identifiers{
															"a",
															"b",
															"i",
															"j",
															"keyF",
														},
													},
													Left: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1262),
																	Column: int(17),
																},
																End: ast.Location{
																	Line: int(1262),
																	Column: int(27),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p13558,
															FreeVars: ast.Identifiers{
																"a",
																"i",
																"keyF",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1262),
																		Column: int(17),
																	},
																	End: ast.Location{
																		Line: int(1262),
																		Column: int(21),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p13558,
																FreeVars: ast.Identifiers{
																	"keyF",
																},
															},
															Id: "keyF",
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1262),
																					Column: int(22),
																				},
																				End: ast.Location{
																					Line: int(1262),
																					Column: int(26),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p13679,
																			FreeVars: ast.Identifiers{
																				"a",
																				"i",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1262),
																						Column: int(22),
																					},
																					End: ast.Location{
																						Line: int(1262),
																						Column: int(23),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p13679,
																				FreeVars: ast.Identifiers{
																					"a",
																				},
																			},
																			Id: "a",
																		},
																		LeftBracketFodder: ast.Fodder{},
																		Index: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1262),
																						Column: int(24),
																					},
																					End: ast.Location{
																						Line: int(1262),
																						Column: int(25),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p13679,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																			},
																			Id: "i",
																		},
																		RightBracketFodder: ast.Fodder{},
																		Id: nil,
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														TrailingComma: false,
														TailStrict: false,
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
													},
													OpFodder: ast.Fodder{},
													Op: ast.BinaryOp(9),
													Right: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1262),
																	Column: int(30),
																},
																End: ast.Location{
																	Line: int(1262),
																	Column: int(40),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p13558,
															FreeVars: ast.Identifiers{
																"b",
																"j",
																"keyF",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1262),
																		Column: int(30),
																	},
																	End: ast.Location{
																		Line: int(1262),
																		Column: int(34),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p13558,
																FreeVars: ast.Identifiers{
																	"keyF",
																},
															},
															Id: "keyF",
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1262),
																					Column: int(35),
																				},
																				End: ast.Location{
																					Line: int(1262),
																					Column: int(39),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p13691,
																			FreeVars: ast.Identifiers{
																				"b",
																				"j",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1262),
																						Column: int(35),
																					},
																					End: ast.Location{
																						Line: int(1262),
																						Column: int(36),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p13691,
																				FreeVars: ast.Identifiers{
																					"b",
																				},
																			},
																			Id: "b",
																		},
																		LeftBracketFodder: ast.Fodder{},
																		Index: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1262),
																						Column: int(37),
																					},
																					End: ast.Location{
																						Line: int(1262),
																						Column: int(38),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p13691,
																				FreeVars: ast.Identifiers{
																					"j",
																				},
																			},
																			Id: "j",
																		},
																		RightBracketFodder: ast.Fodder{},
																		Id: nil,
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														TrailingComma: false,
														TailStrict: false,
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
													},
												},
												ThenFodder: ast.Fodder{},
												BranchTrue: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1263),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(1263),
																Column: int(44),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p13558,
														FreeVars: ast.Identifiers{
															"a",
															"acc",
															"aux",
															"b",
															"i",
															"j",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1263),
																	Column: int(11),
																},
																End: ast.Location{
																	Line: int(1263),
																	Column: int(14),
																},
																File: p1,
															},
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(10),
																	Comment: []string{},
																},
															},
															Ctx: p13558,
															FreeVars: ast.Identifiers{
																"aux",
															},
														},
														Id: "aux",
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1263),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(1263),
																				Column: int(16),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13704,
																		FreeVars: ast.Identifiers{
																			"a",
																		},
																	},
																	Id: "a",
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1263),
																				Column: int(18),
																			},
																			End: ast.Location{
																				Line: int(1263),
																				Column: int(19),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13704,
																		FreeVars: ast.Identifiers{
																			"b",
																		},
																	},
																	Id: "b",
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1263),
																				Column: int(21),
																			},
																			End: ast.Location{
																				Line: int(1263),
																				Column: int(26),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13704,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																	},
																	Left: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1263),
																					Column: int(21),
																				},
																				End: ast.Location{
																					Line: int(1263),
																					Column: int(22),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p13704,
																			FreeVars: ast.Identifiers{
																				"i",
																			},
																		},
																		Id: "i",
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(3),
																	Right: &ast.LiteralNumber{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1263),
																					Column: int(25),
																				},
																				End: ast.Location{
																					Line: int(1263),
																					Column: int(26),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p13704,
																			FreeVars: nil,
																		},
																		OriginalString: "1",
																	},
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1263),
																				Column: int(28),
																			},
																			End: ast.Location{
																				Line: int(1263),
																				Column: int(29),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13704,
																		FreeVars: ast.Identifiers{
																			"j",
																		},
																	},
																	Id: "j",
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1263),
																				Column: int(31),
																			},
																			End: ast.Location{
																				Line: int(1263),
																				Column: int(43),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13704,
																		FreeVars: ast.Identifiers{
																			"a",
																			"acc",
																			"i",
																		},
																	},
																	Left: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1263),
																					Column: int(31),
																				},
																				End: ast.Location{
																					Line: int(1263),
																					Column: int(34),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p13704,
																			FreeVars: ast.Identifiers{
																				"acc",
																			},
																		},
																		Id: "acc",
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(3),
																	Right: &ast.Array{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1263),
																					Column: int(37),
																				},
																				End: ast.Location{
																					Line: int(1263),
																					Column: int(43),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p13704,
																			FreeVars: ast.Identifiers{
																				"a",
																				"i",
																			},
																		},
																		Elements: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Index{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1263),
																								Column: int(38),
																							},
																							End: ast.Location{
																								Line: int(1263),
																								Column: int(42),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p13723,
																						FreeVars: ast.Identifiers{
																							"a",
																							"i",
																						},
																					},
																					Target: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1263),
																									Column: int(38),
																								},
																								End: ast.Location{
																									Line: int(1263),
																									Column: int(39),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p13723,
																							FreeVars: ast.Identifiers{
																								"a",
																							},
																						},
																						Id: "a",
																					},
																					LeftBracketFodder: ast.Fodder{},
																					Index: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1263),
																									Column: int(40),
																								},
																								End: ast.Location{
																									Line: int(1263),
																									Column: int(41),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p13723,
																							FreeVars: ast.Identifiers{
																								"i",
																							},
																						},
																						Id: "i",
																					},
																					RightBracketFodder: ast.Fodder{},
																					Id: nil,
																				},
																				CommaFodder: nil,
																			},
																		},
																		TrailingComma: false,
																		CloseFodder: ast.Fodder{},
																	},
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: true,
													FodderRight: ast.Fodder{},
													TailStrictFodder: ast.Fodder{},
												},
												ElseFodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
														Comment: []string{},
													},
												},
												BranchFalse: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1265),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(1265),
																Column: int(35),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p13558,
														FreeVars: ast.Identifiers{
															"a",
															"acc",
															"aux",
															"b",
															"i",
															"j",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1265),
																	Column: int(11),
																},
																End: ast.Location{
																	Line: int(1265),
																	Column: int(14),
																},
																File: p1,
															},
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(10),
																	Comment: []string{},
																},
															},
															Ctx: p13558,
															FreeVars: ast.Identifiers{
																"aux",
															},
														},
														Id: "aux",
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1265),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(1265),
																				Column: int(16),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13737,
																		FreeVars: ast.Identifiers{
																			"a",
																		},
																	},
																	Id: "a",
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1265),
																				Column: int(18),
																			},
																			End: ast.Location{
																				Line: int(1265),
																				Column: int(19),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13737,
																		FreeVars: ast.Identifiers{
																			"b",
																		},
																	},
																	Id: "b",
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1265),
																				Column: int(21),
																			},
																			End: ast.Location{
																				Line: int(1265),
																				Column: int(22),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13737,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																	},
																	Id: "i",
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1265),
																				Column: int(24),
																			},
																			End: ast.Location{
																				Line: int(1265),
																				Column: int(29),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13737,
																		FreeVars: ast.Identifiers{
																			"j",
																		},
																	},
																	Left: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1265),
																					Column: int(24),
																				},
																				End: ast.Location{
																					Line: int(1265),
																					Column: int(25),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p13737,
																			FreeVars: ast.Identifiers{
																				"j",
																			},
																		},
																		Id: "j",
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(3),
																	Right: &ast.LiteralNumber{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1265),
																					Column: int(28),
																				},
																				End: ast.Location{
																					Line: int(1265),
																					Column: int(29),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p13737,
																			FreeVars: nil,
																		},
																		OriginalString: "1",
																	},
																},
																CommaFodder: ast.Fodder{},
															},
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1265),
																				Column: int(31),
																			},
																			End: ast.Location{
																				Line: int(1265),
																				Column: int(34),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13737,
																		FreeVars: ast.Identifiers{
																			"acc",
																		},
																	},
																	Id: "acc",
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: true,
													FodderRight: ast.Fodder{},
													TailStrictFodder: ast.Fodder{},
												},
											},
										},
									},
								},
							},
							Fun: nil,
							CloseFodder: nil,
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
						},
					},
					Body: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1266),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1266),
									Column: int(24),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p13546,
							FreeVars: ast.Identifiers{
								"a",
								"aux",
								"b",
							},
						},
						Target: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1266),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1266),
										Column: int(8),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
										Comment: []string{},
									},
								},
								Ctx: p13546,
								FreeVars: ast.Identifiers{
									"aux",
								},
							},
							Id: "aux",
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1266),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1266),
													Column: int(10),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13757,
											FreeVars: ast.Identifiers{
												"a",
											},
										},
										Id: "a",
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1266),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(1266),
													Column: int(13),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13757,
											FreeVars: ast.Identifiers{
												"b",
											},
										},
										Id: "b",
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.LiteralNumber{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1266),
													Column: int(15),
												},
												End: ast.Location{
													Line: int(1266),
													Column: int(16),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13757,
											FreeVars: nil,
										},
										OriginalString: "0",
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.LiteralNumber{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1266),
													Column: int(18),
												},
												End: ast.Location{
													Line: int(1266),
													Column: int(19),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13757,
											FreeVars: nil,
										},
										OriginalString: "0",
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Array{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1266),
													Column: int(21),
												},
												End: ast.Location{
													Line: int(1266),
													Column: int(23),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13757,
											FreeVars: nil,
										},
										Elements: nil,
										TrailingComma: false,
										CloseFodder: ast.Fodder{},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: true,
						FodderRight: ast.Fodder{},
						TailStrictFodder: ast.Fodder{},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(1253),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1266),
					Column: int(24),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "mergePatch",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "target",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1268),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(1268),
								Column: int(20),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "patch",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1268),
								Column: int(22),
							},
							End: ast.Location{
								Line: int(1268),
								Column: int(27),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1269),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1290),
								Column: int(12),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p13770,
						FreeVars: ast.Identifiers{
							"patch",
							"std",
							"target",
						},
					},
					Cond: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1269),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(1269),
									Column: int(27),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p13770,
							FreeVars: ast.Identifiers{
								"patch",
								"std",
							},
						},
						Target: &ast.Index{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1269),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(1269),
										Column: int(20),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p13770,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1269),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(1269),
											Column: int(11),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Id: "std",
							},
							LeftBracketFodder: ast.Fodder{},
							Index: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Value: "isObject",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							RightBracketFodder: ast.Fodder{},
							Id: nil,
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1269),
													Column: int(21),
												},
												End: ast.Location{
													Line: int(1269),
													Column: int(26),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13781,
											FreeVars: ast.Identifiers{
												"patch",
											},
										},
										Id: "patch",
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
					ThenFodder: ast.Fodder{},
					BranchTrue: &ast.Local{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1270),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(1288),
									Column: int(8),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
									Comment: []string{},
								},
							},
							Ctx: p13770,
							FreeVars: ast.Identifiers{
								"patch",
								"std",
								"target",
							},
						},
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: ast.Fodder{},
								Variable: "target_object",
								EqFodder: ast.Fodder{},
								Body: &ast.Conditional{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1271),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(1271),
												Column: int(52),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(8),
												Comment: []string{},
											},
										},
										Ctx: p13789,
										FreeVars: ast.Identifiers{
											"std",
											"target",
										},
									},
									Cond: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1271),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(1271),
													Column: int(32),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13789,
											FreeVars: ast.Identifiers{
												"std",
												"target",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1271),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(1271),
														Column: int(24),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p13789,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1271),
															Column: int(12),
														},
														End: ast.Location{
															Line: int(1271),
															Column: int(15),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: ast.Fodder{},
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "isObject",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: ast.Fodder{},
											Id: nil,
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1271),
																	Column: int(25),
																},
																End: ast.Location{
																	Line: int(1271),
																	Column: int(31),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p13800,
															FreeVars: ast.Identifiers{
																"target",
															},
														},
														Id: "target",
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
									ThenFodder: ast.Fodder{},
									BranchTrue: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1271),
													Column: int(38),
												},
												End: ast.Location{
													Line: int(1271),
													Column: int(44),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13789,
											FreeVars: ast.Identifiers{
												"target",
											},
										},
										Id: "target",
									},
									ElseFodder: ast.Fodder{},
									BranchFalse: &ast.DesugaredObject{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1271),
													Column: int(50),
												},
												End: ast.Location{
													Line: int(1271),
													Column: int(52),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p13789,
											FreeVars: nil,
										},
										Asserts: ast.Nodes{},
										Fields: ast.DesugaredObjectFields{},
										Locals: ast.LocalBinds{},
									},
								},
								Fun: nil,
								CloseFodder: ast.Fodder{},
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1270),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(1271),
										Column: int(52),
									},
									File: p1,
								},
							},
						},
						Body: &ast.Local{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1273),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(1288),
										Column: int(8),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(1),
										Indent: int(6),
										Comment: []string{},
									},
								},
								Ctx: p13770,
								FreeVars: ast.Identifiers{
									"patch",
									"std",
									"target_object",
								},
							},
							Binds: ast.LocalBinds{
								ast.LocalBind{
									VarFodder: ast.Fodder{},
									Variable: "target_fields",
									EqFodder: ast.Fodder{},
									Body: &ast.Conditional{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1274),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1274),
													Column: int(84),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
													Comment: []string{},
												},
											},
											Ctx: p13811,
											FreeVars: ast.Identifiers{
												"std",
												"target_object",
											},
										},
										Cond: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1274),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(1274),
														Column: int(39),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p13811,
												FreeVars: ast.Identifiers{
													"std",
													"target_object",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1274),
															Column: int(12),
														},
														End: ast.Location{
															Line: int(1274),
															Column: int(24),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p13811,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1274),
																Column: int(12),
															},
															End: ast.Location{
																Line: int(1274),
																Column: int(15),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "isObject",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1274),
																		Column: int(25),
																	},
																	End: ast.Location{
																		Line: int(1274),
																		Column: int(38),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p13822,
																FreeVars: ast.Identifiers{
																	"target_object",
																},
															},
															Id: "target_object",
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
										ThenFodder: ast.Fodder{},
										BranchTrue: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1274),
														Column: int(45),
													},
													End: ast.Location{
														Line: int(1274),
														Column: int(76),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p13811,
												FreeVars: ast.Identifiers{
													"std",
													"target_object",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1274),
															Column: int(45),
														},
														End: ast.Location{
															Line: int(1274),
															Column: int(61),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p13811,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1274),
																Column: int(45),
															},
															End: ast.Location{
																Line: int(1274),
																Column: int(48),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "objectFields",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1274),
																		Column: int(62),
																	},
																	End: ast.Location{
																		Line: int(1274),
																		Column: int(75),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p13833,
																FreeVars: ast.Identifiers{
																	"target_object",
																},
															},
															Id: "target_object",
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
										ElseFodder: ast.Fodder{},
										BranchFalse: &ast.Array{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1274),
														Column: int(82),
													},
													End: ast.Location{
														Line: int(1274),
														Column: int(84),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p13811,
												FreeVars: nil,
											},
											Elements: nil,
											TrailingComma: false,
											CloseFodder: ast.Fodder{},
										},
									},
									Fun: nil,
									CloseFodder: ast.Fodder{},
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1273),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(1274),
											Column: int(84),
										},
										File: p1,
									},
								},
							},
							Body: &ast.Local{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1276),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(1288),
											Column: int(8),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(1),
											Indent: int(6),
											Comment: []string{},
										},
									},
									Ctx: p13770,
									FreeVars: ast.Identifiers{
										"patch",
										"std",
										"target_fields",
										"target_object",
									},
								},
								Binds: ast.LocalBinds{
									ast.LocalBind{
										VarFodder: ast.Fodder{},
										Variable: "null_fields",
										EqFodder: ast.Fodder{},
										Body: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"patch",
													"std",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: nil,
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "flatMap",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: nil,
												Id: nil,
											},
											FodderLeft: nil,
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Function{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"patch",
																},
															},
															ParenLeftFodder: nil,
															Parameters: []ast.Parameter{
																ast.Parameter{
																	NameFodder: nil,
																	Name: "k",
																	EqFodder: nil,
																	DefaultArg: nil,
																	CommaFodder: nil,
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																},
															},
															TrailingComma: false,
															ParenRightFodder: nil,
															Body: &ast.Conditional{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"k",
																		"patch",
																	},
																},
																Cond: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1276),
																				Column: int(66),
																			},
																			End: ast.Location{
																				Line: int(1276),
																				Column: int(82),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13854,
																		FreeVars: ast.Identifiers{
																			"k",
																			"patch",
																		},
																	},
																	Left: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1276),
																					Column: int(66),
																				},
																				End: ast.Location{
																					Line: int(1276),
																					Column: int(74),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p13854,
																			FreeVars: ast.Identifiers{
																				"k",
																				"patch",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1276),
																						Column: int(66),
																					},
																					End: ast.Location{
																						Line: int(1276),
																						Column: int(71),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p13854,
																				FreeVars: ast.Identifiers{
																					"patch",
																				},
																			},
																			Id: "patch",
																		},
																		LeftBracketFodder: ast.Fodder{},
																		Index: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1276),
																						Column: int(72),
																					},
																					End: ast.Location{
																						Line: int(1276),
																						Column: int(73),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p13854,
																				FreeVars: ast.Identifiers{
																					"k",
																				},
																			},
																			Id: "k",
																		},
																		RightBracketFodder: ast.Fodder{},
																		Id: nil,
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(12),
																	Right: &ast.LiteralNull{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1276),
																					Column: int(78),
																				},
																				End: ast.Location{
																					Line: int(1276),
																					Column: int(82),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p13854,
																			FreeVars: nil,
																		},
																	},
																},
																ThenFodder: nil,
																BranchTrue: &ast.Array{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"k",
																		},
																	},
																	Elements: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1276),
																							Column: int(28),
																						},
																						End: ast.Location{
																							Line: int(1276),
																							Column: int(29),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p13867,
																					FreeVars: ast.Identifiers{
																						"k",
																					},
																				},
																				Id: "k",
																			},
																			CommaFodder: nil,
																		},
																	},
																	TrailingComma: false,
																	CloseFodder: nil,
																},
																ElseFodder: nil,
																BranchFalse: &ast.Array{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: nil,
																	},
																	Elements: nil,
																	TrailingComma: false,
																	CloseFodder: nil,
																},
															},
														},
														CommaFodder: nil,
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Apply{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1276),
																		Column: int(39),
																	},
																	End: ast.Location{
																		Line: int(1276),
																		Column: int(62),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p13854,
																FreeVars: ast.Identifiers{
																	"patch",
																	"std",
																},
															},
															Target: &ast.Index{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1276),
																			Column: int(39),
																		},
																		End: ast.Location{
																			Line: int(1276),
																			Column: int(55),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13854,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1276),
																				Column: int(39),
																			},
																			End: ast.Location{
																				Line: int(1276),
																				Column: int(42),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Id: "std",
																},
																LeftBracketFodder: ast.Fodder{},
																Index: &ast.LiteralString{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: nil,
																	},
																	Value: "objectFields",
																	Kind: ast.LiteralStringKind(1),
																	BlockIndent: "",
																	BlockTermIndent: "",
																},
																RightBracketFodder: ast.Fodder{},
																Id: nil,
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1276),
																						Column: int(56),
																					},
																					End: ast.Location{
																						Line: int(1276),
																						Column: int(61),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p13879,
																				FreeVars: ast.Identifiers{
																					"patch",
																				},
																			},
																			Id: "patch",
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															TrailingComma: false,
															TailStrict: false,
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: nil,
											TailStrictFodder: nil,
										},
										Fun: nil,
										CloseFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1276),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(1276),
												Column: int(83),
											},
											File: p1,
										},
									},
								},
								Body: &ast.Local{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1277),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1288),
												Column: int(8),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										Ctx: p13770,
										FreeVars: ast.Identifiers{
											"null_fields",
											"patch",
											"std",
											"target_fields",
											"target_object",
										},
									},
									Binds: ast.LocalBinds{
										ast.LocalBind{
											VarFodder: ast.Fodder{},
											Variable: "both_fields",
											EqFodder: ast.Fodder{},
											Body: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1277),
															Column: int(27),
														},
														End: ast.Location{
															Line: int(1277),
															Column: int(79),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p13886,
													FreeVars: ast.Identifiers{
														"patch",
														"std",
														"target_fields",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1277),
																Column: int(27),
															},
															End: ast.Location{
																Line: int(1277),
																Column: int(39),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p13886,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1277),
																	Column: int(27),
																},
																End: ast.Location{
																	Line: int(1277),
																	Column: int(30),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: ast.Fodder{},
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "setUnion",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: ast.Fodder{},
													Id: nil,
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1277),
																			Column: int(40),
																		},
																		End: ast.Location{
																			Line: int(1277),
																			Column: int(53),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13895,
																	FreeVars: ast.Identifiers{
																		"target_fields",
																	},
																},
																Id: "target_fields",
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Apply{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1277),
																			Column: int(55),
																		},
																		End: ast.Location{
																			Line: int(1277),
																			Column: int(78),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p13895,
																	FreeVars: ast.Identifiers{
																		"patch",
																		"std",
																	},
																},
																Target: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1277),
																				Column: int(55),
																			},
																			End: ast.Location{
																				Line: int(1277),
																				Column: int(71),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p13895,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1277),
																					Column: int(55),
																				},
																				End: ast.Location{
																					Line: int(1277),
																					Column: int(58),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Id: "std",
																	},
																	LeftBracketFodder: ast.Fodder{},
																	Index: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: nil,
																		},
																		Value: "objectFields",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	RightBracketFodder: ast.Fodder{},
																	Id: nil,
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1277),
																							Column: int(72),
																						},
																						End: ast.Location{
																							Line: int(1277),
																							Column: int(77),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p13906,
																					FreeVars: ast.Identifiers{
																						"patch",
																					},
																				},
																				Id: "patch",
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																TrailingComma: false,
																TailStrict: false,
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
											Fun: nil,
											CloseFodder: ast.Fodder{},
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1277),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(1277),
													Column: int(79),
												},
												File: p1,
											},
										},
									},
									Body: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"both_fields",
												"null_fields",
												"patch",
												"std",
												"target_object",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: nil,
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "$objectFlatMerge",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: nil,
											Id: nil,
										},
										FodderLeft: nil,
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"both_fields",
																"null_fields",
																"patch",
																"std",
																"target_object",
															},
														},
														Target: &ast.Index{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Target: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Id: "std",
															},
															LeftBracketFodder: nil,
															Index: &ast.LiteralString{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: nil,
																},
																Value: "flatMap",
																Kind: ast.LiteralStringKind(1),
																BlockIndent: "",
																BlockTermIndent: "",
															},
															RightBracketFodder: nil,
															Id: nil,
														},
														FodderLeft: nil,
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Function{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"patch",
																				"std",
																				"target_object",
																			},
																		},
																		ParenLeftFodder: nil,
																		Parameters: []ast.Parameter{
																			ast.Parameter{
																				NameFodder: nil,
																				Name: "k",
																				EqFodder: nil,
																				DefaultArg: nil,
																				CommaFodder: nil,
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																			},
																		},
																		TrailingComma: false,
																		ParenRightFodder: nil,
																		Body: &ast.Array{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"k",
																					"patch",
																					"std",
																					"target_object",
																				},
																			},
																			Elements: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.DesugaredObject{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1279),
																									Column: int(7),
																								},
																								End: ast.Location{
																									Line: int(1288),
																									Column: int(8),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{
																								ast.FodderElement{
																									Kind: ast.FodderKind(0),
																									Blanks: int(1),
																									Indent: int(6),
																									Comment: []string{},
																								},
																							},
																							Ctx: p13770,
																							FreeVars: ast.Identifiers{
																								"k",
																								"patch",
																								"std",
																								"target_object",
																							},
																						},
																						Asserts: ast.Nodes{},
																						Fields: ast.DesugaredObjectFields{
																							ast.DesugaredObjectField{
																								Hide: ast.ObjectFieldHide(1),
																								Name: &ast.Var{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1280),
																												Column: int(10),
																											},
																											End: ast.Location{
																												Line: int(1280),
																												Column: int(11),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p13770,
																										FreeVars: ast.Identifiers{
																											"k",
																										},
																									},
																									Id: "k",
																								},
																								Body: &ast.Conditional{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1281),
																												Column: int(11),
																											},
																											End: ast.Location{
																												Line: int(1286),
																												Column: int(55),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{
																											ast.FodderElement{
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(10),
																												Comment: []string{},
																											},
																										},
																										Ctx: p13938,
																										FreeVars: ast.Identifiers{
																											"k",
																											"patch",
																											"std",
																											"target_object",
																										},
																									},
																									Cond: &ast.Unary{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(1281),
																													Column: int(14),
																												},
																												End: ast.Location{
																													Line: int(1281),
																													Column: int(38),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p13938,
																											FreeVars: ast.Identifiers{
																												"k",
																												"patch",
																												"std",
																											},
																										},
																										Op: ast.UnaryOp(0),
																										Expr: &ast.Apply{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1281),
																														Column: int(15),
																													},
																													End: ast.Location{
																														Line: int(1281),
																														Column: int(38),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p13938,
																												FreeVars: ast.Identifiers{
																													"k",
																													"patch",
																													"std",
																												},
																											},
																											Target: &ast.Index{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(1281),
																															Column: int(15),
																														},
																														End: ast.Location{
																															Line: int(1281),
																															Column: int(28),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p13938,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																												},
																												Target: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(1281),
																																Column: int(15),
																															},
																															End: ast.Location{
																																Line: int(1281),
																																Column: int(18),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"std",
																														},
																													},
																													Id: "std",
																												},
																												LeftBracketFodder: ast.Fodder{},
																												Index: &ast.LiteralString{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															File: nil,
																														},
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: nil,
																													},
																													Value: "objectHas",
																													Kind: ast.LiteralStringKind(1),
																													BlockIndent: "",
																													BlockTermIndent: "",
																												},
																												RightBracketFodder: ast.Fodder{},
																												Id: nil,
																											},
																											FodderLeft: ast.Fodder{},
																											Arguments: ast.Arguments{
																												Positional: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(1281),
																																		Column: int(29),
																																	},
																																	End: ast.Location{
																																		Line: int(1281),
																																		Column: int(34),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p13951,
																																FreeVars: ast.Identifiers{
																																	"patch",
																																},
																															},
																															Id: "patch",
																														},
																														CommaFodder: ast.Fodder{},
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Var{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(1281),
																																		Column: int(36),
																																	},
																																	End: ast.Location{
																																		Line: int(1281),
																																		Column: int(37),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p13951,
																																FreeVars: ast.Identifiers{
																																	"k",
																																},
																															},
																															Id: "k",
																														},
																														CommaFodder: nil,
																													},
																												},
																												Named: nil,
																											},
																											TrailingComma: false,
																											TailStrict: false,
																											FodderRight: ast.Fodder{},
																											TailStrictFodder: nil,
																										},
																									},
																									ThenFodder: ast.Fodder{},
																									BranchTrue: &ast.Index{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(1282),
																													Column: int(13),
																												},
																												End: ast.Location{
																													Line: int(1282),
																													Column: int(29),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p13938,
																											FreeVars: ast.Identifiers{
																												"k",
																												"target_object",
																											},
																										},
																										Target: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1282),
																														Column: int(13),
																													},
																													End: ast.Location{
																														Line: int(1282),
																														Column: int(26),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{
																													ast.FodderElement{
																														Kind: ast.FodderKind(0),
																														Blanks: int(0),
																														Indent: int(12),
																														Comment: []string{},
																													},
																												},
																												Ctx: p13938,
																												FreeVars: ast.Identifiers{
																													"target_object",
																												},
																											},
																											Id: "target_object",
																										},
																										LeftBracketFodder: ast.Fodder{},
																										Index: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1282),
																														Column: int(27),
																													},
																													End: ast.Location{
																														Line: int(1282),
																														Column: int(28),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p13938,
																												FreeVars: ast.Identifiers{
																													"k",
																												},
																											},
																											Id: "k",
																										},
																										RightBracketFodder: ast.Fodder{},
																										Id: nil,
																									},
																									ElseFodder: ast.Fodder{
																										ast.FodderElement{
																											Kind: ast.FodderKind(0),
																											Blanks: int(0),
																											Indent: int(10),
																											Comment: []string{},
																										},
																									},
																									BranchFalse: &ast.Conditional{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(1283),
																													Column: int(16),
																												},
																												End: ast.Location{
																													Line: int(1286),
																													Column: int(55),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p13938,
																											FreeVars: ast.Identifiers{
																												"k",
																												"patch",
																												"std",
																												"target_object",
																											},
																										},
																										Cond: &ast.Unary{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1283),
																														Column: int(19),
																													},
																													End: ast.Location{
																														Line: int(1283),
																														Column: int(51),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p13938,
																												FreeVars: ast.Identifiers{
																													"k",
																													"std",
																													"target_object",
																												},
																											},
																											Op: ast.UnaryOp(0),
																											Expr: &ast.Apply{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(1283),
																															Column: int(20),
																														},
																														End: ast.Location{
																															Line: int(1283),
																															Column: int(51),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p13938,
																													FreeVars: ast.Identifiers{
																														"k",
																														"std",
																														"target_object",
																													},
																												},
																												Target: &ast.Index{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(1283),
																																Column: int(20),
																															},
																															End: ast.Location{
																																Line: int(1283),
																																Column: int(33),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p13938,
																														FreeVars: ast.Identifiers{
																															"std",
																														},
																													},
																													Target: &ast.Var{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "<std>",
																																Begin: ast.Location{
																																	Line: int(1283),
																																	Column: int(20),
																																},
																																End: ast.Location{
																																	Line: int(1283),
																																	Column: int(23),
																																},
																																File: p1,
																															},
																															Fodder: ast.Fodder{},
																															Ctx: nil,
																															FreeVars: ast.Identifiers{
																																"std",
																															},
																														},
																														Id: "std",
																													},
																													LeftBracketFodder: ast.Fodder{},
																													Index: &ast.LiteralString{
																														NodeBase: ast.NodeBase{
																															LocRange: ast.LocationRange{
																																FileName: "",
																																Begin: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																End: ast.Location{
																																	Line: int(0),
																																	Column: int(0),
																																},
																																File: nil,
																															},
																															Fodder: nil,
																															Ctx: nil,
																															FreeVars: nil,
																														},
																														Value: "objectHas",
																														Kind: ast.LiteralStringKind(1),
																														BlockIndent: "",
																														BlockTermIndent: "",
																													},
																													RightBracketFodder: ast.Fodder{},
																													Id: nil,
																												},
																												FodderLeft: ast.Fodder{},
																												Arguments: ast.Arguments{
																													Positional: []ast.CommaSeparatedExpr{
																														ast.CommaSeparatedExpr{
																															Expr: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(1283),
																																			Column: int(34),
																																		},
																																		End: ast.Location{
																																			Line: int(1283),
																																			Column: int(47),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p13976,
																																	FreeVars: ast.Identifiers{
																																		"target_object",
																																	},
																																},
																																Id: "target_object",
																															},
																															CommaFodder: ast.Fodder{},
																														},
																														ast.CommaSeparatedExpr{
																															Expr: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(1283),
																																			Column: int(49),
																																		},
																																		End: ast.Location{
																																			Line: int(1283),
																																			Column: int(50),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p13976,
																																	FreeVars: ast.Identifiers{
																																		"k",
																																	},
																																},
																																Id: "k",
																															},
																															CommaFodder: nil,
																														},
																													},
																													Named: nil,
																												},
																												TrailingComma: false,
																												TailStrict: false,
																												FodderRight: ast.Fodder{},
																												TailStrictFodder: nil,
																											},
																										},
																										ThenFodder: ast.Fodder{},
																										BranchTrue: &ast.Apply{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1284),
																														Column: int(13),
																													},
																													End: ast.Location{
																														Line: int(1284),
																														Column: int(43),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p13938,
																												FreeVars: ast.Identifiers{
																													"k",
																													"patch",
																													"std",
																												},
																											},
																											Target: &ast.Index{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(1284),
																															Column: int(13),
																														},
																														End: ast.Location{
																															Line: int(1284),
																															Column: int(27),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p13938,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																												},
																												Target: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(1284),
																																Column: int(13),
																															},
																															End: ast.Location{
																																Line: int(1284),
																																Column: int(16),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(12),
																																Comment: []string{},
																															},
																														},
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"std",
																														},
																													},
																													Id: "std",
																												},
																												LeftBracketFodder: ast.Fodder{},
																												Index: &ast.LiteralString{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															File: nil,
																														},
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: nil,
																													},
																													Value: "mergePatch",
																													Kind: ast.LiteralStringKind(1),
																													BlockIndent: "",
																													BlockTermIndent: "",
																												},
																												RightBracketFodder: ast.Fodder{},
																												Id: nil,
																											},
																											FodderLeft: ast.Fodder{},
																											Arguments: ast.Arguments{
																												Positional: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.LiteralNull{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(1284),
																																		Column: int(28),
																																	},
																																	End: ast.Location{
																																		Line: int(1284),
																																		Column: int(32),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p13990,
																																FreeVars: nil,
																															},
																														},
																														CommaFodder: ast.Fodder{},
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Index{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(1284),
																																		Column: int(34),
																																	},
																																	End: ast.Location{
																																		Line: int(1284),
																																		Column: int(42),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p13990,
																																FreeVars: ast.Identifiers{
																																	"k",
																																	"patch",
																																},
																															},
																															Target: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(1284),
																																			Column: int(34),
																																		},
																																		End: ast.Location{
																																			Line: int(1284),
																																			Column: int(39),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p13990,
																																	FreeVars: ast.Identifiers{
																																		"patch",
																																	},
																																},
																																Id: "patch",
																															},
																															LeftBracketFodder: ast.Fodder{},
																															Index: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(1284),
																																			Column: int(40),
																																		},
																																		End: ast.Location{
																																			Line: int(1284),
																																			Column: int(41),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p13990,
																																	FreeVars: ast.Identifiers{
																																		"k",
																																	},
																																},
																																Id: "k",
																															},
																															RightBracketFodder: ast.Fodder{},
																															Id: nil,
																														},
																														CommaFodder: nil,
																													},
																												},
																												Named: nil,
																											},
																											TrailingComma: false,
																											TailStrict: true,
																											FodderRight: ast.Fodder{},
																											TailStrictFodder: ast.Fodder{},
																										},
																										ElseFodder: ast.Fodder{
																											ast.FodderElement{
																												Kind: ast.FodderKind(0),
																												Blanks: int(0),
																												Indent: int(10),
																												Comment: []string{},
																											},
																										},
																										BranchFalse: &ast.Apply{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1286),
																														Column: int(13),
																													},
																													End: ast.Location{
																														Line: int(1286),
																														Column: int(55),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p13938,
																												FreeVars: ast.Identifiers{
																													"k",
																													"patch",
																													"std",
																													"target_object",
																												},
																											},
																											Target: &ast.Index{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(1286),
																															Column: int(13),
																														},
																														End: ast.Location{
																															Line: int(1286),
																															Column: int(27),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p13938,
																													FreeVars: ast.Identifiers{
																														"std",
																													},
																												},
																												Target: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(1286),
																																Column: int(13),
																															},
																															End: ast.Location{
																																Line: int(1286),
																																Column: int(16),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{
																															ast.FodderElement{
																																Kind: ast.FodderKind(0),
																																Blanks: int(0),
																																Indent: int(12),
																																Comment: []string{},
																															},
																														},
																														Ctx: nil,
																														FreeVars: ast.Identifiers{
																															"std",
																														},
																													},
																													Id: "std",
																												},
																												LeftBracketFodder: ast.Fodder{},
																												Index: &ast.LiteralString{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "",
																															Begin: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															End: ast.Location{
																																Line: int(0),
																																Column: int(0),
																															},
																															File: nil,
																														},
																														Fodder: nil,
																														Ctx: nil,
																														FreeVars: nil,
																													},
																													Value: "mergePatch",
																													Kind: ast.LiteralStringKind(1),
																													BlockIndent: "",
																													BlockTermIndent: "",
																												},
																												RightBracketFodder: ast.Fodder{},
																												Id: nil,
																											},
																											FodderLeft: ast.Fodder{},
																											Arguments: ast.Arguments{
																												Positional: []ast.CommaSeparatedExpr{
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Index{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(1286),
																																		Column: int(28),
																																	},
																																	End: ast.Location{
																																		Line: int(1286),
																																		Column: int(44),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p14008,
																																FreeVars: ast.Identifiers{
																																	"k",
																																	"target_object",
																																},
																															},
																															Target: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(1286),
																																			Column: int(28),
																																		},
																																		End: ast.Location{
																																			Line: int(1286),
																																			Column: int(41),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p14008,
																																	FreeVars: ast.Identifiers{
																																		"target_object",
																																	},
																																},
																																Id: "target_object",
																															},
																															LeftBracketFodder: ast.Fodder{},
																															Index: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(1286),
																																			Column: int(42),
																																		},
																																		End: ast.Location{
																																			Line: int(1286),
																																			Column: int(43),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p14008,
																																	FreeVars: ast.Identifiers{
																																		"k",
																																	},
																																},
																																Id: "k",
																															},
																															RightBracketFodder: ast.Fodder{},
																															Id: nil,
																														},
																														CommaFodder: ast.Fodder{},
																													},
																													ast.CommaSeparatedExpr{
																														Expr: &ast.Index{
																															NodeBase: ast.NodeBase{
																																LocRange: ast.LocationRange{
																																	FileName: "<std>",
																																	Begin: ast.Location{
																																		Line: int(1286),
																																		Column: int(46),
																																	},
																																	End: ast.Location{
																																		Line: int(1286),
																																		Column: int(54),
																																	},
																																	File: p1,
																																},
																																Fodder: ast.Fodder{},
																																Ctx: p14008,
																																FreeVars: ast.Identifiers{
																																	"k",
																																	"patch",
																																},
																															},
																															Target: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(1286),
																																			Column: int(46),
																																		},
																																		End: ast.Location{
																																			Line: int(1286),
																																			Column: int(51),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p14008,
																																	FreeVars: ast.Identifiers{
																																		"patch",
																																	},
																																},
																																Id: "patch",
																															},
																															LeftBracketFodder: ast.Fodder{},
																															Index: &ast.Var{
																																NodeBase: ast.NodeBase{
																																	LocRange: ast.LocationRange{
																																		FileName: "<std>",
																																		Begin: ast.Location{
																																			Line: int(1286),
																																			Column: int(52),
																																		},
																																		End: ast.Location{
																																			Line: int(1286),
																																			Column: int(53),
																																		},
																																		File: p1,
																																	},
																																	Fodder: ast.Fodder{},
																																	Ctx: p14008,
																																	FreeVars: ast.Identifiers{
																																		"k",
																																	},
																																},
																																Id: "k",
																															},
																															RightBracketFodder: ast.Fodder{},
																															Id: nil,
																														},
																														CommaFodder: nil,
																													},
																												},
																												Named: nil,
																											},
																											TrailingComma: false,
																											TailStrict: true,
																											FodderRight: ast.Fodder{},
																											TailStrictFodder: ast.Fodder{},
																										},
																									},
																								},
																								PlusSuper: false,
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1280),
																										Column: int(9),
																									},
																									End: ast.Location{
																										Line: int(1286),
																										Column: int(55),
																									},
																									File: p1,
																								},
																							},
																						},
																						Locals: ast.LocalBinds{},
																					},
																					CommaFodder: nil,
																				},
																			},
																			TrailingComma: false,
																			CloseFodder: nil,
																		},
																	},
																	CommaFodder: nil,
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Apply{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1287),
																					Column: int(18),
																				},
																				End: ast.Location{
																					Line: int(1287),
																					Column: int(55),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p13770,
																			FreeVars: ast.Identifiers{
																				"both_fields",
																				"null_fields",
																				"std",
																			},
																		},
																		Target: &ast.Index{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1287),
																						Column: int(18),
																					},
																					End: ast.Location{
																						Line: int(1287),
																						Column: int(29),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p13770,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1287),
																							Column: int(18),
																						},
																						End: ast.Location{
																							Line: int(1287),
																							Column: int(21),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: nil,
																					FreeVars: ast.Identifiers{
																						"std",
																					},
																				},
																				Id: "std",
																			},
																			LeftBracketFodder: ast.Fodder{},
																			Index: &ast.LiteralString{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: nil,
																				},
																				Value: "setDiff",
																				Kind: ast.LiteralStringKind(1),
																				BlockIndent: "",
																				BlockTermIndent: "",
																			},
																			RightBracketFodder: ast.Fodder{},
																			Id: nil,
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1287),
																									Column: int(30),
																								},
																								End: ast.Location{
																									Line: int(1287),
																									Column: int(41),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p14029,
																							FreeVars: ast.Identifiers{
																								"both_fields",
																							},
																						},
																						Id: "both_fields",
																					},
																					CommaFodder: ast.Fodder{},
																				},
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1287),
																									Column: int(43),
																								},
																								End: ast.Location{
																									Line: int(1287),
																									Column: int(54),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p14029,
																							FreeVars: ast.Identifiers{
																								"null_fields",
																							},
																						},
																						Id: "null_fields",
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		TrailingComma: false,
																		TailStrict: false,
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														TrailingComma: false,
														TailStrict: false,
														FodderRight: nil,
														TailStrictFodder: nil,
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: nil,
										TailStrictFodder: nil,
									},
								},
							},
						},
					},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
							Comment: []string{},
						},
					},
					BranchFalse: &ast.Var{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1290),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(1290),
									Column: int(12),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
									Comment: []string{},
								},
							},
							Ctx: p13770,
							FreeVars: ast.Identifiers{
								"patch",
							},
						},
						Id: "patch",
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(1268),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1290),
					Column: int(12),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "objectFields",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "o",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1292),
								Column: int(16),
							},
							End: ast.Location{
								Line: int(1292),
								Column: int(17),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1293),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1293),
								Column: int(33),
							},
							File: p1,
						},
						Fodder: ast.Fodder{},
						Ctx: p14042,
						FreeVars: ast.Identifiers{
							"o",
							"std",
						},
					},
					Target: &ast.Index{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1293),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1293),
									Column: int(23),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p14042,
							FreeVars: ast.Identifiers{
								"std",
							},
						},
						Target: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1293),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1293),
										Column: int(8),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
										Comment: []string{},
									},
								},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Id: "std",
						},
						LeftBracketFodder: ast.Fodder{},
						Index: &ast.LiteralString{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: nil,
							},
							Value: "objectFieldsEx",
							Kind: ast.LiteralStringKind(1),
							BlockIndent: "",
							BlockTermIndent: "",
						},
						RightBracketFodder: ast.Fodder{},
						Id: nil,
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1293),
												Column: int(24),
											},
											End: ast.Location{
												Line: int(1293),
												Column: int(25),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p14052,
										FreeVars: ast.Identifiers{
											"o",
										},
									},
									Id: "o",
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.LiteralBoolean{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1293),
												Column: int(27),
											},
											End: ast.Location{
												Line: int(1293),
												Column: int(32),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p14052,
										FreeVars: nil,
									},
									Value: false,
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					TrailingComma: false,
					TailStrict: false,
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(1292),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1293),
					Column: int(33),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "objectFieldsAll",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "o",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1295),
								Column: int(19),
							},
							End: ast.Location{
								Line: int(1295),
								Column: int(20),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1296),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1296),
								Column: int(32),
							},
							File: p1,
						},
						Fodder: ast.Fodder{},
						Ctx: p14060,
						FreeVars: ast.Identifiers{
							"o",
							"std",
						},
					},
					Target: &ast.Index{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1296),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1296),
									Column: int(23),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p14060,
							FreeVars: ast.Identifiers{
								"std",
							},
						},
						Target: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1296),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1296),
										Column: int(8),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
										Comment: []string{},
									},
								},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Id: "std",
						},
						LeftBracketFodder: ast.Fodder{},
						Index: &ast.LiteralString{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: nil,
							},
							Value: "objectFieldsEx",
							Kind: ast.LiteralStringKind(1),
							BlockIndent: "",
							BlockTermIndent: "",
						},
						RightBracketFodder: ast.Fodder{},
						Id: nil,
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1296),
												Column: int(24),
											},
											End: ast.Location{
												Line: int(1296),
												Column: int(25),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p14070,
										FreeVars: ast.Identifiers{
											"o",
										},
									},
									Id: "o",
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.LiteralBoolean{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1296),
												Column: int(27),
											},
											End: ast.Location{
												Line: int(1296),
												Column: int(31),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p14070,
										FreeVars: nil,
									},
									Value: true,
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					TrailingComma: false,
					TailStrict: false,
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(1295),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1296),
					Column: int(32),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "objectHas",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "o",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1298),
								Column: int(13),
							},
							End: ast.Location{
								Line: int(1298),
								Column: int(14),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "f",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1298),
								Column: int(16),
							},
							End: ast.Location{
								Line: int(1298),
								Column: int(17),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1299),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1299),
								Column: int(33),
							},
							File: p1,
						},
						Fodder: ast.Fodder{},
						Ctx: p14078,
						FreeVars: ast.Identifiers{
							"f",
							"o",
							"std",
						},
					},
					Target: &ast.Index{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1299),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1299),
									Column: int(20),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p14078,
							FreeVars: ast.Identifiers{
								"std",
							},
						},
						Target: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1299),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1299),
										Column: int(8),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
										Comment: []string{},
									},
								},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Id: "std",
						},
						LeftBracketFodder: ast.Fodder{},
						Index: &ast.LiteralString{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: nil,
							},
							Value: "objectHasEx",
							Kind: ast.LiteralStringKind(1),
							BlockIndent: "",
							BlockTermIndent: "",
						},
						RightBracketFodder: ast.Fodder{},
						Id: nil,
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1299),
												Column: int(21),
											},
											End: ast.Location{
												Line: int(1299),
												Column: int(22),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p14088,
										FreeVars: ast.Identifiers{
											"o",
										},
									},
									Id: "o",
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1299),
												Column: int(24),
											},
											End: ast.Location{
												Line: int(1299),
												Column: int(25),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p14088,
										FreeVars: ast.Identifiers{
											"f",
										},
									},
									Id: "f",
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.LiteralBoolean{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1299),
												Column: int(27),
											},
											End: ast.Location{
												Line: int(1299),
												Column: int(32),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p14088,
										FreeVars: nil,
									},
									Value: false,
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					TrailingComma: false,
					TailStrict: false,
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(1298),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1299),
					Column: int(33),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "objectHasAll",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "o",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1301),
								Column: int(16),
							},
							End: ast.Location{
								Line: int(1301),
								Column: int(17),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "f",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1301),
								Column: int(19),
							},
							End: ast.Location{
								Line: int(1301),
								Column: int(20),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Apply{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1302),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1302),
								Column: int(32),
							},
							File: p1,
						},
						Fodder: ast.Fodder{},
						Ctx: p14098,
						FreeVars: ast.Identifiers{
							"f",
							"o",
							"std",
						},
					},
					Target: &ast.Index{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1302),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1302),
									Column: int(20),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p14098,
							FreeVars: ast.Identifiers{
								"std",
							},
						},
						Target: &ast.Var{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1302),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1302),
										Column: int(8),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
										Comment: []string{},
									},
								},
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Id: "std",
						},
						LeftBracketFodder: ast.Fodder{},
						Index: &ast.LiteralString{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: nil,
							},
							Value: "objectHasEx",
							Kind: ast.LiteralStringKind(1),
							BlockIndent: "",
							BlockTermIndent: "",
						},
						RightBracketFodder: ast.Fodder{},
						Id: nil,
					},
					FodderLeft: ast.Fodder{},
					Arguments: ast.Arguments{
						Positional: []ast.CommaSeparatedExpr{
							ast.CommaSeparatedExpr{
								Expr: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1302),
												Column: int(21),
											},
											End: ast.Location{
												Line: int(1302),
												Column: int(22),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p14108,
										FreeVars: ast.Identifiers{
											"o",
										},
									},
									Id: "o",
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1302),
												Column: int(24),
											},
											End: ast.Location{
												Line: int(1302),
												Column: int(25),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p14108,
										FreeVars: ast.Identifiers{
											"f",
										},
									},
									Id: "f",
								},
								CommaFodder: ast.Fodder{},
							},
							ast.CommaSeparatedExpr{
								Expr: &ast.LiteralBoolean{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1302),
												Column: int(27),
											},
											End: ast.Location{
												Line: int(1302),
												Column: int(31),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p14108,
										FreeVars: nil,
									},
									Value: true,
								},
								CommaFodder: nil,
							},
						},
						Named: nil,
					},
					TrailingComma: false,
					TailStrict: false,
					FodderRight: ast.Fodder{},
					TailStrictFodder: nil,
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(1301),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1302),
					Column: int(32),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "equals",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "a",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1304),
								Column: int(10),
							},
							End: ast.Location{
								Line: int(1304),
								Column: int(11),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "b",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1304),
								Column: int(13),
							},
							End: ast.Location{
								Line: int(1304),
								Column: int(14),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1305),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1338),
								Column: int(34),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p14119,
						FreeVars: ast.Identifiers{
							"a",
							"b",
							"std",
						},
					},
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Variable: "ta",
							EqFodder: ast.Fodder{},
							Body: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1305),
											Column: int(16),
										},
										End: ast.Location{
											Line: int(1305),
											Column: int(27),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p14123,
									FreeVars: ast.Identifiers{
										"a",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1305),
												Column: int(16),
											},
											End: ast.Location{
												Line: int(1305),
												Column: int(24),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p14123,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1305),
													Column: int(16),
												},
												End: ast.Location{
													Line: int(1305),
													Column: int(19),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "type",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1305),
															Column: int(25),
														},
														End: ast.Location{
															Line: int(1305),
															Column: int(26),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p14132,
													FreeVars: ast.Identifiers{
														"a",
													},
												},
												Id: "a",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
							Fun: nil,
							CloseFodder: ast.Fodder{},
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1305),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(1305),
									Column: int(27),
								},
								File: p1,
							},
						},
					},
					Body: &ast.Local{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1306),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1338),
									Column: int(34),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
									Comment: []string{},
								},
							},
							Ctx: p14119,
							FreeVars: ast.Identifiers{
								"a",
								"b",
								"std",
								"ta",
							},
						},
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: ast.Fodder{},
								Variable: "tb",
								EqFodder: ast.Fodder{},
								Body: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1306),
												Column: int(16),
											},
											End: ast.Location{
												Line: int(1306),
												Column: int(27),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p14139,
										FreeVars: ast.Identifiers{
											"b",
											"std",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1306),
													Column: int(16),
												},
												End: ast.Location{
													Line: int(1306),
													Column: int(24),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p14139,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1306),
														Column: int(16),
													},
													End: ast.Location{
														Line: int(1306),
														Column: int(19),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "type",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1306),
																Column: int(25),
															},
															End: ast.Location{
																Line: int(1306),
																Column: int(26),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p14148,
														FreeVars: ast.Identifiers{
															"b",
														},
													},
													Id: "b",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
								Fun: nil,
								CloseFodder: ast.Fodder{},
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1306),
										Column: int(11),
									},
									End: ast.Location{
										Line: int(1306),
										Column: int(27),
									},
									File: p1,
								},
							},
						},
						Body: &ast.Conditional{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1307),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1338),
										Column: int(34),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
										Comment: []string{},
									},
								},
								Ctx: p14119,
								FreeVars: ast.Identifiers{
									"a",
									"b",
									"std",
									"ta",
									"tb",
								},
							},
							Cond: &ast.Unary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1307),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(1307),
											Column: int(36),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p14119,
									FreeVars: ast.Identifiers{
										"std",
										"ta",
										"tb",
									},
								},
								Op: ast.UnaryOp(0),
								Expr: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1307),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(1307),
												Column: int(36),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p14119,
										FreeVars: ast.Identifiers{
											"std",
											"ta",
											"tb",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1307),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1307),
													Column: int(28),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p14119,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1307),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1307),
														Column: int(12),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "primitiveEquals",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1307),
																Column: int(29),
															},
															End: ast.Location{
																Line: int(1307),
																Column: int(31),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p14164,
														FreeVars: ast.Identifiers{
															"ta",
														},
													},
													Id: "ta",
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1307),
																Column: int(33),
															},
															End: ast.Location{
																Line: int(1307),
																Column: int(35),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p14164,
														FreeVars: ast.Identifiers{
															"tb",
														},
													},
													Id: "tb",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
							},
							ThenFodder: ast.Fodder{},
							BranchTrue: &ast.LiteralBoolean{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1308),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(1308),
											Column: int(12),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
											Comment: []string{},
										},
									},
									Ctx: p14119,
									FreeVars: nil,
								},
								Value: false,
							},
							ElseFodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
									Comment: []string{},
								},
							},
							BranchFalse: &ast.Conditional{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1310),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(1338),
											Column: int(34),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
											Comment: []string{},
										},
									},
									Ctx: p14119,
									FreeVars: ast.Identifiers{
										"a",
										"b",
										"std",
										"ta",
									},
								},
								Cond: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1310),
												Column: int(10),
											},
											End: ast.Location{
												Line: int(1310),
												Column: int(42),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p14119,
										FreeVars: ast.Identifiers{
											"std",
											"ta",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1310),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(1310),
													Column: int(29),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p14119,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1310),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(1310),
														Column: int(13),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "primitiveEquals",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1310),
																Column: int(30),
															},
															End: ast.Location{
																Line: int(1310),
																Column: int(32),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p14183,
														FreeVars: ast.Identifiers{
															"ta",
														},
													},
													Id: "ta",
												},
												CommaFodder: ast.Fodder{},
											},
											ast.CommaSeparatedExpr{
												Expr: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1310),
																Column: int(34),
															},
															End: ast.Location{
																Line: int(1310),
																Column: int(41),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p14183,
														FreeVars: nil,
													},
													Value: "array",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
								ThenFodder: ast.Fodder{},
								BranchTrue: &ast.Local{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1311),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(1322),
												Column: int(23),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(8),
												Comment: []string{},
											},
										},
										Ctx: p14119,
										FreeVars: ast.Identifiers{
											"a",
											"b",
											"std",
										},
									},
									Binds: ast.LocalBinds{
										ast.LocalBind{
											VarFodder: ast.Fodder{},
											Variable: "la",
											EqFodder: ast.Fodder{},
											Body: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1311),
															Column: int(20),
														},
														End: ast.Location{
															Line: int(1311),
															Column: int(33),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p14191,
													FreeVars: ast.Identifiers{
														"a",
														"std",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1311),
																Column: int(20),
															},
															End: ast.Location{
																Line: int(1311),
																Column: int(30),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p14191,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1311),
																	Column: int(20),
																},
																End: ast.Location{
																	Line: int(1311),
																	Column: int(23),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: ast.Fodder{},
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "length",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: ast.Fodder{},
													Id: nil,
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1311),
																			Column: int(31),
																		},
																		End: ast.Location{
																			Line: int(1311),
																			Column: int(32),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p14200,
																	FreeVars: ast.Identifiers{
																		"a",
																	},
																},
																Id: "a",
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
											Fun: nil,
											CloseFodder: ast.Fodder{},
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1311),
													Column: int(15),
												},
												End: ast.Location{
													Line: int(1311),
													Column: int(33),
												},
												File: p1,
											},
										},
									},
									Body: &ast.Conditional{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1312),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1322),
													Column: int(23),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
													Comment: []string{},
												},
											},
											Ctx: p14119,
											FreeVars: ast.Identifiers{
												"a",
												"b",
												"la",
												"std",
											},
										},
										Cond: &ast.Unary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1312),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(1312),
														Column: int(51),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p14119,
												FreeVars: ast.Identifiers{
													"b",
													"la",
													"std",
												},
											},
											Op: ast.UnaryOp(0),
											Expr: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1312),
															Column: int(13),
														},
														End: ast.Location{
															Line: int(1312),
															Column: int(51),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p14119,
													FreeVars: ast.Identifiers{
														"b",
														"la",
														"std",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1312),
																Column: int(13),
															},
															End: ast.Location{
																Line: int(1312),
																Column: int(32),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p14119,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1312),
																	Column: int(13),
																},
																End: ast.Location{
																	Line: int(1312),
																	Column: int(16),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: ast.Fodder{},
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "primitiveEquals",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: ast.Fodder{},
													Id: nil,
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1312),
																			Column: int(33),
																		},
																		End: ast.Location{
																			Line: int(1312),
																			Column: int(35),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p14216,
																	FreeVars: ast.Identifiers{
																		"la",
																	},
																},
																Id: "la",
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Apply{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1312),
																			Column: int(37),
																		},
																		End: ast.Location{
																			Line: int(1312),
																			Column: int(50),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p14216,
																	FreeVars: ast.Identifiers{
																		"b",
																		"std",
																	},
																},
																Target: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1312),
																				Column: int(37),
																			},
																			End: ast.Location{
																				Line: int(1312),
																				Column: int(47),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p14216,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1312),
																					Column: int(37),
																				},
																				End: ast.Location{
																					Line: int(1312),
																					Column: int(40),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Id: "std",
																	},
																	LeftBracketFodder: ast.Fodder{},
																	Index: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: nil,
																		},
																		Value: "length",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	RightBracketFodder: ast.Fodder{},
																	Id: nil,
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1312),
																							Column: int(48),
																						},
																						End: ast.Location{
																							Line: int(1312),
																							Column: int(49),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p14227,
																					FreeVars: ast.Identifiers{
																						"b",
																					},
																				},
																				Id: "b",
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																TrailingComma: false,
																TailStrict: false,
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
										},
										ThenFodder: ast.Fodder{},
										BranchTrue: &ast.LiteralBoolean{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1313),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(1313),
														Column: int(16),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(10),
														Comment: []string{},
													},
												},
												Ctx: p14119,
												FreeVars: nil,
											},
											Value: false,
										},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(8),
												Comment: []string{},
											},
										},
										BranchFalse: &ast.Local{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1315),
														Column: int(11),
													},
													End: ast.Location{
														Line: int(1322),
														Column: int(23),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(10),
														Comment: []string{},
													},
												},
												Ctx: p14119,
												FreeVars: ast.Identifiers{
													"a",
													"b",
													"la",
												},
											},
											Binds: ast.LocalBinds{
												ast.LocalBind{
													VarFodder: nil,
													Variable: "aux",
													EqFodder: nil,
													Body: &ast.Function{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1315),
																	Column: int(17),
																},
																End: ast.Location{
																	Line: int(1321),
																	Column: int(31),
																},
																File: p1,
															},
															Fodder: nil,
															Ctx: p14237,
															FreeVars: ast.Identifiers{
																"aux",
																"la",
															},
														},
														ParenLeftFodder: ast.Fodder{},
														Parameters: []ast.Parameter{
															ast.Parameter{
																NameFodder: ast.Fodder{},
																Name: "a",
																EqFodder: nil,
																DefaultArg: nil,
																CommaFodder: ast.Fodder{},
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1315),
																		Column: int(21),
																	},
																	End: ast.Location{
																		Line: int(1315),
																		Column: int(22),
																	},
																	File: p1,
																},
															},
															ast.Parameter{
																NameFodder: ast.Fodder{},
																Name: "b",
																EqFodder: nil,
																DefaultArg: nil,
																CommaFodder: ast.Fodder{},
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1315),
																		Column: int(24),
																	},
																	End: ast.Location{
																		Line: int(1315),
																		Column: int(25),
																	},
																	File: p1,
																},
															},
															ast.Parameter{
																NameFodder: ast.Fodder{},
																Name: "i",
																EqFodder: nil,
																DefaultArg: nil,
																CommaFodder: nil,
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1315),
																		Column: int(27),
																	},
																	End: ast.Location{
																		Line: int(1315),
																		Column: int(28),
																	},
																	File: p1,
																},
															},
														},
														TrailingComma: false,
														ParenRightFodder: ast.Fodder{},
														Body: &ast.Conditional{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1316),
																		Column: int(13),
																	},
																	End: ast.Location{
																		Line: int(1321),
																		Column: int(31),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(12),
																		Comment: []string{},
																	},
																},
																Ctx: p14242,
																FreeVars: ast.Identifiers{
																	"a",
																	"aux",
																	"b",
																	"i",
																	"la",
																},
															},
															Cond: &ast.Binary{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1316),
																			Column: int(16),
																		},
																		End: ast.Location{
																			Line: int(1316),
																			Column: int(23),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p14242,
																	FreeVars: ast.Identifiers{
																		"i",
																		"la",
																	},
																},
																Left: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1316),
																				Column: int(16),
																			},
																			End: ast.Location{
																				Line: int(1316),
																				Column: int(17),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p14242,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																	},
																	Id: "i",
																},
																OpFodder: ast.Fodder{},
																Op: ast.BinaryOp(8),
																Right: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1316),
																				Column: int(21),
																			},
																			End: ast.Location{
																				Line: int(1316),
																				Column: int(23),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p14242,
																		FreeVars: ast.Identifiers{
																			"la",
																		},
																	},
																	Id: "la",
																},
															},
															ThenFodder: ast.Fodder{},
															BranchTrue: &ast.LiteralBoolean{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1317),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(1317),
																			Column: int(19),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{
																		ast.FodderElement{
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(14),
																			Comment: []string{},
																		},
																	},
																	Ctx: p14242,
																	FreeVars: nil,
																},
																Value: true,
															},
															ElseFodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(12),
																	Comment: []string{},
																},
															},
															BranchFalse: &ast.Conditional{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1318),
																			Column: int(18),
																		},
																		End: ast.Location{
																			Line: int(1321),
																			Column: int(31),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p14242,
																	FreeVars: ast.Identifiers{
																		"a",
																		"aux",
																		"b",
																		"i",
																	},
																},
																Cond: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1318),
																				Column: int(21),
																			},
																			End: ast.Location{
																				Line: int(1318),
																				Column: int(33),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p14242,
																		FreeVars: ast.Identifiers{
																			"a",
																			"b",
																			"i",
																		},
																	},
																	Left: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1318),
																					Column: int(21),
																				},
																				End: ast.Location{
																					Line: int(1318),
																					Column: int(25),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p14242,
																			FreeVars: ast.Identifiers{
																				"a",
																				"i",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1318),
																						Column: int(21),
																					},
																					End: ast.Location{
																						Line: int(1318),
																						Column: int(22),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p14242,
																				FreeVars: ast.Identifiers{
																					"a",
																				},
																			},
																			Id: "a",
																		},
																		LeftBracketFodder: ast.Fodder{},
																		Index: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1318),
																						Column: int(23),
																					},
																					End: ast.Location{
																						Line: int(1318),
																						Column: int(24),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p14242,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																			},
																			Id: "i",
																		},
																		RightBracketFodder: ast.Fodder{},
																		Id: nil,
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(13),
																	Right: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1318),
																					Column: int(29),
																				},
																				End: ast.Location{
																					Line: int(1318),
																					Column: int(33),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p14242,
																			FreeVars: ast.Identifiers{
																				"b",
																				"i",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1318),
																						Column: int(29),
																					},
																					End: ast.Location{
																						Line: int(1318),
																						Column: int(30),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p14242,
																				FreeVars: ast.Identifiers{
																					"b",
																				},
																			},
																			Id: "b",
																		},
																		LeftBracketFodder: ast.Fodder{},
																		Index: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1318),
																						Column: int(31),
																					},
																					End: ast.Location{
																						Line: int(1318),
																						Column: int(32),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p14242,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																			},
																			Id: "i",
																		},
																		RightBracketFodder: ast.Fodder{},
																		Id: nil,
																	},
																},
																ThenFodder: ast.Fodder{},
																BranchTrue: &ast.LiteralBoolean{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1319),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(1319),
																				Column: int(20),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(14),
																				Comment: []string{},
																			},
																		},
																		Ctx: p14242,
																		FreeVars: nil,
																	},
																	Value: false,
																},
																ElseFodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(12),
																		Comment: []string{},
																	},
																},
																BranchFalse: &ast.Apply{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1321),
																				Column: int(15),
																			},
																			End: ast.Location{
																				Line: int(1321),
																				Column: int(31),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p14242,
																		FreeVars: ast.Identifiers{
																			"a",
																			"aux",
																			"b",
																			"i",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1321),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(1321),
																					Column: int(18),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(14),
																					Comment: []string{},
																				},
																			},
																			Ctx: p14242,
																			FreeVars: ast.Identifiers{
																				"aux",
																			},
																		},
																		Id: "aux",
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1321),
																								Column: int(19),
																							},
																							End: ast.Location{
																								Line: int(1321),
																								Column: int(20),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p14279,
																						FreeVars: ast.Identifiers{
																							"a",
																						},
																					},
																					Id: "a",
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1321),
																								Column: int(22),
																							},
																							End: ast.Location{
																								Line: int(1321),
																								Column: int(23),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p14279,
																						FreeVars: ast.Identifiers{
																							"b",
																						},
																					},
																					Id: "b",
																				},
																				CommaFodder: ast.Fodder{},
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Binary{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1321),
																								Column: int(25),
																							},
																							End: ast.Location{
																								Line: int(1321),
																								Column: int(30),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p14279,
																						FreeVars: ast.Identifiers{
																							"i",
																						},
																					},
																					Left: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1321),
																									Column: int(25),
																								},
																								End: ast.Location{
																									Line: int(1321),
																									Column: int(26),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p14279,
																							FreeVars: ast.Identifiers{
																								"i",
																							},
																						},
																						Id: "i",
																					},
																					OpFodder: ast.Fodder{},
																					Op: ast.BinaryOp(3),
																					Right: &ast.LiteralNumber{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1321),
																									Column: int(29),
																								},
																								End: ast.Location{
																									Line: int(1321),
																									Column: int(30),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p14279,
																							FreeVars: nil,
																						},
																						OriginalString: "1",
																					},
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	TrailingComma: false,
																	TailStrict: true,
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: ast.Fodder{},
																},
															},
														},
													},
													Fun: nil,
													CloseFodder: nil,
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
												},
											},
											Body: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1322),
															Column: int(11),
														},
														End: ast.Location{
															Line: int(1322),
															Column: int(23),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p14119,
													FreeVars: ast.Identifiers{
														"a",
														"aux",
														"b",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1322),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(1322),
																Column: int(14),
															},
															File: p1,
														},
														Fodder: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(10),
																Comment: []string{},
															},
														},
														Ctx: p14119,
														FreeVars: ast.Identifiers{
															"aux",
														},
													},
													Id: "aux",
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1322),
																			Column: int(15),
																		},
																		End: ast.Location{
																			Line: int(1322),
																			Column: int(16),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p14295,
																	FreeVars: ast.Identifiers{
																		"a",
																	},
																},
																Id: "a",
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1322),
																			Column: int(18),
																		},
																		End: ast.Location{
																			Line: int(1322),
																			Column: int(19),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p14295,
																	FreeVars: ast.Identifiers{
																		"b",
																	},
																},
																Id: "b",
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralNumber{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1322),
																			Column: int(21),
																		},
																		End: ast.Location{
																			Line: int(1322),
																			Column: int(22),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p14295,
																	FreeVars: nil,
																},
																OriginalString: "0",
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
										},
									},
								},
								ElseFodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
										Comment: []string{},
									},
								},
								BranchFalse: &ast.Conditional{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1323),
												Column: int(12),
											},
											End: ast.Location{
												Line: int(1338),
												Column: int(34),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p14119,
										FreeVars: ast.Identifiers{
											"a",
											"b",
											"std",
											"ta",
										},
									},
									Cond: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1323),
													Column: int(15),
												},
												End: ast.Location{
													Line: int(1323),
													Column: int(48),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p14119,
											FreeVars: ast.Identifiers{
												"std",
												"ta",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1323),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(1323),
														Column: int(34),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p14119,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1323),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1323),
															Column: int(18),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: ast.Fodder{},
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "primitiveEquals",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: ast.Fodder{},
											Id: nil,
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1323),
																	Column: int(35),
																},
																End: ast.Location{
																	Line: int(1323),
																	Column: int(37),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p14312,
															FreeVars: ast.Identifiers{
																"ta",
															},
														},
														Id: "ta",
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1323),
																	Column: int(39),
																},
																End: ast.Location{
																	Line: int(1323),
																	Column: int(47),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p14312,
															FreeVars: nil,
														},
														Value: "object",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
									ThenFodder: ast.Fodder{},
									BranchTrue: &ast.Local{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1324),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1336),
													Column: int(23),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
													Comment: []string{},
												},
											},
											Ctx: p14119,
											FreeVars: ast.Identifiers{
												"a",
												"b",
												"std",
											},
										},
										Binds: ast.LocalBinds{
											ast.LocalBind{
												VarFodder: ast.Fodder{},
												Variable: "fields",
												EqFodder: ast.Fodder{},
												Body: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1324),
																Column: int(24),
															},
															End: ast.Location{
																Line: int(1324),
																Column: int(43),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p14320,
														FreeVars: ast.Identifiers{
															"a",
															"std",
														},
													},
													Target: &ast.Index{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1324),
																	Column: int(24),
																},
																End: ast.Location{
																	Line: int(1324),
																	Column: int(40),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p14320,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1324),
																		Column: int(24),
																	},
																	End: ast.Location{
																		Line: int(1324),
																		Column: int(27),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Id: "std",
														},
														LeftBracketFodder: ast.Fodder{},
														Index: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
															Value: "objectFields",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														RightBracketFodder: ast.Fodder{},
														Id: nil,
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1324),
																				Column: int(41),
																			},
																			End: ast.Location{
																				Line: int(1324),
																				Column: int(42),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p14329,
																		FreeVars: ast.Identifiers{
																			"a",
																		},
																	},
																	Id: "a",
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
												},
												Fun: nil,
												CloseFodder: ast.Fodder{},
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1324),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(1324),
														Column: int(43),
													},
													File: p1,
												},
											},
										},
										Body: &ast.Local{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1325),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1336),
														Column: int(23),
													},
													File: p1,
												},
												Fodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
														Comment: []string{},
													},
												},
												Ctx: p14119,
												FreeVars: ast.Identifiers{
													"a",
													"b",
													"fields",
													"std",
												},
											},
											Binds: ast.LocalBinds{
												ast.LocalBind{
													VarFodder: ast.Fodder{},
													Variable: "lfields",
													EqFodder: ast.Fodder{},
													Body: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1325),
																	Column: int(25),
																},
																End: ast.Location{
																	Line: int(1325),
																	Column: int(43),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p14336,
															FreeVars: ast.Identifiers{
																"fields",
																"std",
															},
														},
														Target: &ast.Index{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1325),
																		Column: int(25),
																	},
																	End: ast.Location{
																		Line: int(1325),
																		Column: int(35),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p14336,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Target: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1325),
																			Column: int(25),
																		},
																		End: ast.Location{
																			Line: int(1325),
																			Column: int(28),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Id: "std",
															},
															LeftBracketFodder: ast.Fodder{},
															Index: &ast.LiteralString{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: nil,
																},
																Value: "length",
																Kind: ast.LiteralStringKind(1),
																BlockIndent: "",
																BlockTermIndent: "",
															},
															RightBracketFodder: ast.Fodder{},
															Id: nil,
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1325),
																					Column: int(36),
																				},
																				End: ast.Location{
																					Line: int(1325),
																					Column: int(42),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p14345,
																			FreeVars: ast.Identifiers{
																				"fields",
																			},
																		},
																		Id: "fields",
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														TrailingComma: false,
														TailStrict: false,
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
													},
													Fun: nil,
													CloseFodder: ast.Fodder{},
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1325),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1325),
															Column: int(43),
														},
														File: p1,
													},
												},
											},
											Body: &ast.Conditional{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1326),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1336),
															Column: int(23),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
															Comment: []string{},
														},
													},
													Ctx: p14119,
													FreeVars: ast.Identifiers{
														"a",
														"b",
														"fields",
														"lfields",
														"std",
													},
												},
												Cond: &ast.Binary{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1326),
																Column: int(12),
															},
															End: ast.Location{
																Line: int(1326),
																Column: int(41),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p14119,
														FreeVars: ast.Identifiers{
															"b",
															"fields",
															"std",
														},
													},
													Left: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1326),
																	Column: int(12),
																},
																End: ast.Location{
																	Line: int(1326),
																	Column: int(18),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p14119,
															FreeVars: ast.Identifiers{
																"fields",
															},
														},
														Id: "fields",
													},
													OpFodder: ast.Fodder{},
													Op: ast.BinaryOp(13),
													Right: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1326),
																	Column: int(22),
																},
																End: ast.Location{
																	Line: int(1326),
																	Column: int(41),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p14119,
															FreeVars: ast.Identifiers{
																"b",
																"std",
															},
														},
														Target: &ast.Index{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1326),
																		Column: int(22),
																	},
																	End: ast.Location{
																		Line: int(1326),
																		Column: int(38),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p14119,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Target: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1326),
																			Column: int(22),
																		},
																		End: ast.Location{
																			Line: int(1326),
																			Column: int(25),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Id: "std",
															},
															LeftBracketFodder: ast.Fodder{},
															Index: &ast.LiteralString{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: nil,
																},
																Value: "objectFields",
																Kind: ast.LiteralStringKind(1),
																BlockIndent: "",
																BlockTermIndent: "",
															},
															RightBracketFodder: ast.Fodder{},
															Id: nil,
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1326),
																					Column: int(39),
																				},
																				End: ast.Location{
																					Line: int(1326),
																					Column: int(40),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p14363,
																			FreeVars: ast.Identifiers{
																				"b",
																			},
																		},
																		Id: "b",
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														TrailingComma: false,
														TailStrict: false,
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
													},
												},
												ThenFodder: ast.Fodder{},
												BranchTrue: &ast.LiteralBoolean{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1327),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(1327),
																Column: int(16),
															},
															File: p1,
														},
														Fodder: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(10),
																Comment: []string{},
															},
														},
														Ctx: p14119,
														FreeVars: nil,
													},
													Value: false,
												},
												ElseFodder: ast.Fodder{
													ast.FodderElement{
														Kind: ast.FodderKind(0),
														Blanks: int(0),
														Indent: int(8),
														Comment: []string{},
													},
												},
												BranchFalse: &ast.Local{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1329),
																Column: int(11),
															},
															End: ast.Location{
																Line: int(1336),
																Column: int(23),
															},
															File: p1,
														},
														Fodder: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(10),
																Comment: []string{},
															},
														},
														Ctx: p14119,
														FreeVars: ast.Identifiers{
															"a",
															"b",
															"fields",
															"lfields",
														},
													},
													Binds: ast.LocalBinds{
														ast.LocalBind{
															VarFodder: nil,
															Variable: "aux",
															EqFodder: nil,
															Body: &ast.Function{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1329),
																			Column: int(17),
																		},
																		End: ast.Location{
																			Line: int(1335),
																			Column: int(31),
																		},
																		File: p1,
																	},
																	Fodder: nil,
																	Ctx: p14373,
																	FreeVars: ast.Identifiers{
																		"aux",
																		"fields",
																		"lfields",
																	},
																},
																ParenLeftFodder: ast.Fodder{},
																Parameters: []ast.Parameter{
																	ast.Parameter{
																		NameFodder: ast.Fodder{},
																		Name: "a",
																		EqFodder: nil,
																		DefaultArg: nil,
																		CommaFodder: ast.Fodder{},
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1329),
																				Column: int(21),
																			},
																			End: ast.Location{
																				Line: int(1329),
																				Column: int(22),
																			},
																			File: p1,
																		},
																	},
																	ast.Parameter{
																		NameFodder: ast.Fodder{},
																		Name: "b",
																		EqFodder: nil,
																		DefaultArg: nil,
																		CommaFodder: ast.Fodder{},
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1329),
																				Column: int(24),
																			},
																			End: ast.Location{
																				Line: int(1329),
																				Column: int(25),
																			},
																			File: p1,
																		},
																	},
																	ast.Parameter{
																		NameFodder: ast.Fodder{},
																		Name: "i",
																		EqFodder: nil,
																		DefaultArg: nil,
																		CommaFodder: nil,
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1329),
																				Column: int(27),
																			},
																			End: ast.Location{
																				Line: int(1329),
																				Column: int(28),
																			},
																			File: p1,
																		},
																	},
																},
																TrailingComma: false,
																ParenRightFodder: ast.Fodder{},
																Body: &ast.Conditional{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1330),
																				Column: int(13),
																			},
																			End: ast.Location{
																				Line: int(1335),
																				Column: int(31),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(12),
																				Comment: []string{},
																			},
																		},
																		Ctx: p14378,
																		FreeVars: ast.Identifiers{
																			"a",
																			"aux",
																			"b",
																			"fields",
																			"i",
																			"lfields",
																		},
																	},
																	Cond: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1330),
																					Column: int(16),
																				},
																				End: ast.Location{
																					Line: int(1330),
																					Column: int(28),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p14378,
																			FreeVars: ast.Identifiers{
																				"i",
																				"lfields",
																			},
																		},
																		Left: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1330),
																						Column: int(16),
																					},
																					End: ast.Location{
																						Line: int(1330),
																						Column: int(17),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p14378,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																			},
																			Id: "i",
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(8),
																		Right: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1330),
																						Column: int(21),
																					},
																					End: ast.Location{
																						Line: int(1330),
																						Column: int(28),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p14378,
																				FreeVars: ast.Identifiers{
																					"lfields",
																				},
																			},
																			Id: "lfields",
																		},
																	},
																	ThenFodder: ast.Fodder{},
																	BranchTrue: &ast.LiteralBoolean{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1331),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(1331),
																					Column: int(19),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(14),
																					Comment: []string{},
																				},
																			},
																			Ctx: p14378,
																			FreeVars: nil,
																		},
																		Value: true,
																	},
																	ElseFodder: ast.Fodder{
																		ast.FodderElement{
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(12),
																			Comment: []string{},
																		},
																	},
																	BranchFalse: &ast.Conditional{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1332),
																					Column: int(18),
																				},
																				End: ast.Location{
																					Line: int(1335),
																					Column: int(31),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p14378,
																			FreeVars: ast.Identifiers{
																				"a",
																				"aux",
																				"b",
																				"fields",
																				"i",
																			},
																		},
																		Cond: &ast.Local{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1332),
																						Column: int(21),
																					},
																					End: ast.Location{
																						Line: int(1332),
																						Column: int(54),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p14378,
																				FreeVars: ast.Identifiers{
																					"a",
																					"b",
																					"fields",
																					"i",
																				},
																			},
																			Binds: ast.LocalBinds{
																				ast.LocalBind{
																					VarFodder: ast.Fodder{},
																					Variable: "f",
																					EqFodder: ast.Fodder{},
																					Body: &ast.Index{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1332),
																									Column: int(31),
																								},
																								End: ast.Location{
																									Line: int(1332),
																									Column: int(40),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p14395,
																							FreeVars: ast.Identifiers{
																								"fields",
																								"i",
																							},
																						},
																						Target: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1332),
																										Column: int(31),
																									},
																									End: ast.Location{
																										Line: int(1332),
																										Column: int(37),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p14395,
																								FreeVars: ast.Identifiers{
																									"fields",
																								},
																							},
																							Id: "fields",
																						},
																						LeftBracketFodder: ast.Fodder{},
																						Index: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1332),
																										Column: int(38),
																									},
																									End: ast.Location{
																										Line: int(1332),
																										Column: int(39),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p14395,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																							},
																							Id: "i",
																						},
																						RightBracketFodder: ast.Fodder{},
																						Id: nil,
																					},
																					Fun: nil,
																					CloseFodder: ast.Fodder{},
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1332),
																							Column: int(27),
																						},
																						End: ast.Location{
																							Line: int(1332),
																							Column: int(40),
																						},
																						File: p1,
																					},
																				},
																			},
																			Body: &ast.Binary{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1332),
																							Column: int(42),
																						},
																						End: ast.Location{
																							Line: int(1332),
																							Column: int(54),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p14378,
																					FreeVars: ast.Identifiers{
																						"a",
																						"b",
																						"f",
																					},
																				},
																				Left: &ast.Index{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1332),
																								Column: int(42),
																							},
																							End: ast.Location{
																								Line: int(1332),
																								Column: int(46),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p14378,
																						FreeVars: ast.Identifiers{
																							"a",
																							"f",
																						},
																					},
																					Target: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1332),
																									Column: int(42),
																								},
																								End: ast.Location{
																									Line: int(1332),
																									Column: int(43),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p14378,
																							FreeVars: ast.Identifiers{
																								"a",
																							},
																						},
																						Id: "a",
																					},
																					LeftBracketFodder: ast.Fodder{},
																					Index: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1332),
																									Column: int(44),
																								},
																								End: ast.Location{
																									Line: int(1332),
																									Column: int(45),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p14378,
																							FreeVars: ast.Identifiers{
																								"f",
																							},
																						},
																						Id: "f",
																					},
																					RightBracketFodder: ast.Fodder{},
																					Id: nil,
																				},
																				OpFodder: ast.Fodder{},
																				Op: ast.BinaryOp(13),
																				Right: &ast.Index{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1332),
																								Column: int(50),
																							},
																							End: ast.Location{
																								Line: int(1332),
																								Column: int(54),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p14378,
																						FreeVars: ast.Identifiers{
																							"b",
																							"f",
																						},
																					},
																					Target: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1332),
																									Column: int(50),
																								},
																								End: ast.Location{
																									Line: int(1332),
																									Column: int(51),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p14378,
																							FreeVars: ast.Identifiers{
																								"b",
																							},
																						},
																						Id: "b",
																					},
																					LeftBracketFodder: ast.Fodder{},
																					Index: &ast.Var{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1332),
																									Column: int(52),
																								},
																								End: ast.Location{
																									Line: int(1332),
																									Column: int(53),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p14378,
																							FreeVars: ast.Identifiers{
																								"f",
																							},
																						},
																						Id: "f",
																					},
																					RightBracketFodder: ast.Fodder{},
																					Id: nil,
																				},
																			},
																		},
																		ThenFodder: ast.Fodder{},
																		BranchTrue: &ast.LiteralBoolean{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1333),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(1333),
																						Column: int(20),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{
																					ast.FodderElement{
																						Kind: ast.FodderKind(0),
																						Blanks: int(0),
																						Indent: int(14),
																						Comment: []string{},
																					},
																				},
																				Ctx: p14378,
																				FreeVars: nil,
																			},
																			Value: false,
																		},
																		ElseFodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(12),
																				Comment: []string{},
																			},
																		},
																		BranchFalse: &ast.Apply{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1335),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(1335),
																						Column: int(31),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p14378,
																				FreeVars: ast.Identifiers{
																					"a",
																					"aux",
																					"b",
																					"i",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1335),
																							Column: int(15),
																						},
																						End: ast.Location{
																							Line: int(1335),
																							Column: int(18),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(14),
																							Comment: []string{},
																						},
																					},
																					Ctx: p14378,
																					FreeVars: ast.Identifiers{
																						"aux",
																					},
																				},
																				Id: "aux",
																			},
																			FodderLeft: ast.Fodder{},
																			Arguments: ast.Arguments{
																				Positional: []ast.CommaSeparatedExpr{
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1335),
																										Column: int(19),
																									},
																									End: ast.Location{
																										Line: int(1335),
																										Column: int(20),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p14425,
																								FreeVars: ast.Identifiers{
																									"a",
																								},
																							},
																							Id: "a",
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Var{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1335),
																										Column: int(22),
																									},
																									End: ast.Location{
																										Line: int(1335),
																										Column: int(23),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p14425,
																								FreeVars: ast.Identifiers{
																									"b",
																								},
																							},
																							Id: "b",
																						},
																						CommaFodder: ast.Fodder{},
																					},
																					ast.CommaSeparatedExpr{
																						Expr: &ast.Binary{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1335),
																										Column: int(25),
																									},
																									End: ast.Location{
																										Line: int(1335),
																										Column: int(30),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p14425,
																								FreeVars: ast.Identifiers{
																									"i",
																								},
																							},
																							Left: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1335),
																											Column: int(25),
																										},
																										End: ast.Location{
																											Line: int(1335),
																											Column: int(26),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p14425,
																									FreeVars: ast.Identifiers{
																										"i",
																									},
																								},
																								Id: "i",
																							},
																							OpFodder: ast.Fodder{},
																							Op: ast.BinaryOp(3),
																							Right: &ast.LiteralNumber{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1335),
																											Column: int(29),
																										},
																										End: ast.Location{
																											Line: int(1335),
																											Column: int(30),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p14425,
																									FreeVars: nil,
																								},
																								OriginalString: "1",
																							},
																						},
																						CommaFodder: nil,
																					},
																				},
																				Named: nil,
																			},
																			TrailingComma: false,
																			TailStrict: true,
																			FodderRight: ast.Fodder{},
																			TailStrictFodder: ast.Fodder{},
																		},
																	},
																},
															},
															Fun: nil,
															CloseFodder: nil,
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
														},
													},
													Body: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1336),
																	Column: int(11),
																},
																End: ast.Location{
																	Line: int(1336),
																	Column: int(23),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p14119,
															FreeVars: ast.Identifiers{
																"a",
																"aux",
																"b",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1336),
																		Column: int(11),
																	},
																	End: ast.Location{
																		Line: int(1336),
																		Column: int(14),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(10),
																		Comment: []string{},
																	},
																},
																Ctx: p14119,
																FreeVars: ast.Identifiers{
																	"aux",
																},
															},
															Id: "aux",
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1336),
																					Column: int(15),
																				},
																				End: ast.Location{
																					Line: int(1336),
																					Column: int(16),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p14441,
																			FreeVars: ast.Identifiers{
																				"a",
																			},
																		},
																		Id: "a",
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1336),
																					Column: int(18),
																				},
																				End: ast.Location{
																					Line: int(1336),
																					Column: int(19),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p14441,
																			FreeVars: ast.Identifiers{
																				"b",
																			},
																		},
																		Id: "b",
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.LiteralNumber{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1336),
																					Column: int(21),
																				},
																				End: ast.Location{
																					Line: int(1336),
																					Column: int(22),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p14441,
																			FreeVars: nil,
																		},
																		OriginalString: "0",
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														TrailingComma: false,
														TailStrict: false,
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
													},
												},
											},
										},
									},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
											Comment: []string{},
										},
									},
									BranchFalse: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1338),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1338),
													Column: int(34),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p14119,
											FreeVars: ast.Identifiers{
												"a",
												"b",
												"std",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1338),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1338),
														Column: int(28),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p14119,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1338),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1338),
															Column: int(12),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
															Comment: []string{},
														},
													},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: ast.Fodder{},
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "primitiveEquals",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: ast.Fodder{},
											Id: nil,
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1338),
																	Column: int(29),
																},
																End: ast.Location{
																	Line: int(1338),
																	Column: int(30),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p14457,
															FreeVars: ast.Identifiers{
																"a",
															},
														},
														Id: "a",
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1338),
																	Column: int(32),
																},
																End: ast.Location{
																	Line: int(1338),
																	Column: int(33),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p14457,
															FreeVars: ast.Identifiers{
																"b",
															},
														},
														Id: "b",
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
								},
							},
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(1304),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1338),
					Column: int(34),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "resolvePath",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "f",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1341),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(1341),
								Column: int(16),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "r",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1341),
								Column: int(18),
							},
							End: ast.Location{
								Line: int(1341),
								Column: int(19),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1342),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1343),
								Column: int(80),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p14467,
						FreeVars: ast.Identifiers{
							"f",
							"r",
							"std",
						},
					},
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: ast.Fodder{},
							Variable: "arr",
							EqFodder: ast.Fodder{},
							Body: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1342),
											Column: int(17),
										},
										End: ast.Location{
											Line: int(1342),
											Column: int(34),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p14471,
									FreeVars: ast.Identifiers{
										"f",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1342),
												Column: int(17),
											},
											End: ast.Location{
												Line: int(1342),
												Column: int(26),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p14471,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1342),
													Column: int(17),
												},
												End: ast.Location{
													Line: int(1342),
													Column: int(20),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "split",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1342),
															Column: int(27),
														},
														End: ast.Location{
															Line: int(1342),
															Column: int(28),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p14480,
													FreeVars: ast.Identifiers{
														"f",
													},
												},
												Id: "f",
											},
											CommaFodder: ast.Fodder{},
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1342),
															Column: int(30),
														},
														End: ast.Location{
															Line: int(1342),
															Column: int(33),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p14480,
													FreeVars: nil,
												},
												Value: "/",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
							Fun: nil,
							CloseFodder: ast.Fodder{},
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1342),
									Column: int(11),
								},
								End: ast.Location{
									Line: int(1342),
									Column: int(34),
								},
								File: p1,
							},
						},
					},
					Body: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1343),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1343),
									Column: int(80),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p14467,
							FreeVars: ast.Identifiers{
								"arr",
								"r",
								"std",
							},
						},
						Target: &ast.Index{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1343),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(1343),
										Column: int(13),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p14467,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1343),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(1343),
											Column: int(8),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
											Comment: []string{},
										},
									},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Id: "std",
							},
							LeftBracketFodder: ast.Fodder{},
							Index: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Value: "join",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							RightBracketFodder: ast.Fodder{},
							Id: nil,
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1343),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(1343),
													Column: int(17),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p14493,
											FreeVars: nil,
										},
										Value: "/",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Binary{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1343),
													Column: int(19),
												},
												End: ast.Location{
													Line: int(1343),
													Column: int(79),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p14493,
											FreeVars: ast.Identifiers{
												"arr",
												"r",
												"std",
											},
										},
										Left: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1343),
														Column: int(19),
													},
													End: ast.Location{
														Line: int(1343),
														Column: int(73),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p14493,
												FreeVars: ast.Identifiers{
													"arr",
													"std",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1343),
															Column: int(19),
														},
														End: ast.Location{
															Line: int(1343),
															Column: int(32),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p14493,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1343),
																Column: int(19),
															},
															End: ast.Location{
																Line: int(1343),
																Column: int(22),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "makeArray",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Binary{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1343),
																		Column: int(33),
																	},
																	End: ast.Location{
																		Line: int(1343),
																		Column: int(52),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p14505,
																FreeVars: ast.Identifiers{
																	"arr",
																	"std",
																},
															},
															Left: &ast.Apply{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1343),
																			Column: int(33),
																		},
																		End: ast.Location{
																			Line: int(1343),
																			Column: int(48),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p14505,
																	FreeVars: ast.Identifiers{
																		"arr",
																		"std",
																	},
																},
																Target: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1343),
																				Column: int(33),
																			},
																			End: ast.Location{
																				Line: int(1343),
																				Column: int(43),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p14505,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1343),
																					Column: int(33),
																				},
																				End: ast.Location{
																					Line: int(1343),
																					Column: int(36),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Id: "std",
																	},
																	LeftBracketFodder: ast.Fodder{},
																	Index: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: nil,
																		},
																		Value: "length",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	RightBracketFodder: ast.Fodder{},
																	Id: nil,
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1343),
																							Column: int(44),
																						},
																						End: ast.Location{
																							Line: int(1343),
																							Column: int(47),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p14516,
																					FreeVars: ast.Identifiers{
																						"arr",
																					},
																				},
																				Id: "arr",
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																TrailingComma: false,
																TailStrict: false,
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
															},
															OpFodder: ast.Fodder{},
															Op: ast.BinaryOp(4),
															Right: &ast.LiteralNumber{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1343),
																			Column: int(51),
																		},
																		End: ast.Location{
																			Line: int(1343),
																			Column: int(52),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p14505,
																	FreeVars: nil,
																},
																OriginalString: "1",
															},
														},
														CommaFodder: ast.Fodder{},
													},
													ast.CommaSeparatedExpr{
														Expr: &ast.Function{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1343),
																		Column: int(54),
																	},
																	End: ast.Location{
																		Line: int(1343),
																		Column: int(72),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p14505,
																FreeVars: ast.Identifiers{
																	"arr",
																},
															},
															ParenLeftFodder: ast.Fodder{},
															Parameters: []ast.Parameter{
																ast.Parameter{
																	NameFodder: ast.Fodder{},
																	Name: "i",
																	EqFodder: nil,
																	DefaultArg: nil,
																	CommaFodder: nil,
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1343),
																			Column: int(63),
																		},
																		End: ast.Location{
																			Line: int(1343),
																			Column: int(64),
																		},
																		File: p1,
																	},
																},
															},
															TrailingComma: false,
															ParenRightFodder: ast.Fodder{},
															Body: &ast.Index{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1343),
																			Column: int(66),
																		},
																		End: ast.Location{
																			Line: int(1343),
																			Column: int(72),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p14523,
																	FreeVars: ast.Identifiers{
																		"arr",
																		"i",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1343),
																				Column: int(66),
																			},
																			End: ast.Location{
																				Line: int(1343),
																				Column: int(69),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p14523,
																		FreeVars: ast.Identifiers{
																			"arr",
																		},
																	},
																	Id: "arr",
																},
																LeftBracketFodder: ast.Fodder{},
																Index: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1343),
																				Column: int(70),
																			},
																			End: ast.Location{
																				Line: int(1343),
																				Column: int(71),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p14523,
																		FreeVars: ast.Identifiers{
																			"i",
																		},
																	},
																	Id: "i",
																},
																RightBracketFodder: ast.Fodder{},
																Id: nil,
															},
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
										OpFodder: ast.Fodder{},
										Op: ast.BinaryOp(3),
										Right: &ast.Array{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1343),
														Column: int(76),
													},
													End: ast.Location{
														Line: int(1343),
														Column: int(79),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p14493,
												FreeVars: ast.Identifiers{
													"r",
												},
											},
											Elements: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1343),
																	Column: int(77),
																},
																End: ast.Location{
																	Line: int(1343),
																	Column: int(78),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p14533,
															FreeVars: ast.Identifiers{
																"r",
															},
														},
														Id: "r",
													},
													CommaFodder: nil,
												},
											},
											TrailingComma: false,
											CloseFodder: ast.Fodder{},
										},
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(1341),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1343),
					Column: int(80),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "prune",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"$",
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "a",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1345),
								Column: int(9),
							},
							End: ast.Location{
								Line: int(1345),
								Column: int(10),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Local{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1346),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1362),
								Column: int(8),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p14541,
						FreeVars: ast.Identifiers{
							"$",
							"a",
							"std",
						},
					},
					Binds: ast.LocalBinds{
						ast.LocalBind{
							VarFodder: nil,
							Variable: "isContent",
							EqFodder: nil,
							Body: &ast.Function{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1346),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(1354),
											Column: int(13),
										},
										File: p1,
									},
									Fodder: nil,
									Ctx: p14545,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								ParenLeftFodder: ast.Fodder{},
								Parameters: []ast.Parameter{
									ast.Parameter{
										NameFodder: ast.Fodder{},
										Name: "b",
										EqFodder: nil,
										DefaultArg: nil,
										CommaFodder: nil,
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1346),
												Column: int(21),
											},
											End: ast.Location{
												Line: int(1346),
												Column: int(22),
											},
											File: p1,
										},
									},
								},
								TrailingComma: false,
								ParenRightFodder: ast.Fodder{},
								Body: &ast.Conditional{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1347),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1354),
												Column: int(13),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										Ctx: p14550,
										FreeVars: ast.Identifiers{
											"b",
											"std",
										},
									},
									Cond: &ast.Binary{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1347),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(1347),
													Column: int(19),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p14550,
											FreeVars: ast.Identifiers{
												"b",
											},
										},
										Left: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1347),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(1347),
														Column: int(11),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p14550,
												FreeVars: ast.Identifiers{
													"b",
												},
											},
											Id: "b",
										},
										OpFodder: ast.Fodder{},
										Op: ast.BinaryOp(12),
										Right: &ast.LiteralNull{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1347),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(1347),
														Column: int(19),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p14550,
												FreeVars: nil,
											},
										},
									},
									ThenFodder: ast.Fodder{},
									BranchTrue: &ast.LiteralBoolean{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1348),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1348),
													Column: int(14),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
													Comment: []string{},
												},
											},
											Ctx: p14550,
											FreeVars: nil,
										},
										Value: false,
									},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
											Comment: []string{},
										},
									},
									BranchFalse: &ast.Conditional{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1349),
													Column: int(12),
												},
												End: ast.Location{
													Line: int(1354),
													Column: int(13),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p14550,
											FreeVars: ast.Identifiers{
												"b",
												"std",
											},
										},
										Cond: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1349),
														Column: int(15),
													},
													End: ast.Location{
														Line: int(1349),
														Column: int(29),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p14550,
												FreeVars: ast.Identifiers{
													"b",
													"std",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1349),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1349),
															Column: int(26),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p14550,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1349),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1349),
																Column: int(18),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "isArray",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1349),
																		Column: int(27),
																	},
																	End: ast.Location{
																		Line: int(1349),
																		Column: int(28),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p14571,
																FreeVars: ast.Identifiers{
																	"b",
																},
															},
															Id: "b",
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
										ThenFodder: ast.Fodder{},
										BranchTrue: &ast.Binary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1350),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1350),
														Column: int(26),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p14550,
												FreeVars: ast.Identifiers{
													"b",
													"std",
												},
											},
											Left: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1350),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1350),
															Column: int(22),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p14550,
													FreeVars: ast.Identifiers{
														"b",
														"std",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1350),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(1350),
																Column: int(19),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p14550,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1350),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(1350),
																	Column: int(12),
																},
																File: p1,
															},
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(8),
																	Comment: []string{},
																},
															},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: ast.Fodder{},
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "length",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: ast.Fodder{},
													Id: nil,
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1350),
																			Column: int(20),
																		},
																		End: ast.Location{
																			Line: int(1350),
																			Column: int(21),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p14585,
																	FreeVars: ast.Identifiers{
																		"b",
																	},
																},
																Id: "b",
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
											OpFodder: ast.Fodder{},
											Op: ast.BinaryOp(7),
											Right: &ast.LiteralNumber{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1350),
															Column: int(25),
														},
														End: ast.Location{
															Line: int(1350),
															Column: int(26),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p14550,
													FreeVars: nil,
												},
												OriginalString: "0",
											},
										},
										ElseFodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										BranchFalse: &ast.Conditional{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1351),
														Column: int(12),
													},
													End: ast.Location{
														Line: int(1354),
														Column: int(13),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p14550,
												FreeVars: ast.Identifiers{
													"b",
													"std",
												},
											},
											Cond: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1351),
															Column: int(15),
														},
														End: ast.Location{
															Line: int(1351),
															Column: int(30),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p14550,
													FreeVars: ast.Identifiers{
														"b",
														"std",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1351),
																Column: int(15),
															},
															End: ast.Location{
																Line: int(1351),
																Column: int(27),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p14550,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1351),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(1351),
																	Column: int(18),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: ast.Fodder{},
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "isObject",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: ast.Fodder{},
													Id: nil,
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1351),
																			Column: int(28),
																		},
																		End: ast.Location{
																			Line: int(1351),
																			Column: int(29),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p14600,
																	FreeVars: ast.Identifiers{
																		"b",
																	},
																},
																Id: "b",
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
											ThenFodder: ast.Fodder{},
											BranchTrue: &ast.Binary{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1352),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1352),
															Column: int(26),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p14550,
													FreeVars: ast.Identifiers{
														"b",
														"std",
													},
												},
												Left: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1352),
																Column: int(9),
															},
															End: ast.Location{
																Line: int(1352),
																Column: int(22),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p14550,
														FreeVars: ast.Identifiers{
															"b",
															"std",
														},
													},
													Target: &ast.Index{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1352),
																	Column: int(9),
																},
																End: ast.Location{
																	Line: int(1352),
																	Column: int(19),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p14550,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Target: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1352),
																		Column: int(9),
																	},
																	End: ast.Location{
																		Line: int(1352),
																		Column: int(12),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{
																	ast.FodderElement{
																		Kind: ast.FodderKind(0),
																		Blanks: int(0),
																		Indent: int(8),
																		Comment: []string{},
																	},
																},
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Id: "std",
														},
														LeftBracketFodder: ast.Fodder{},
														Index: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: nil,
															},
															Value: "length",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														RightBracketFodder: ast.Fodder{},
														Id: nil,
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1352),
																				Column: int(20),
																			},
																			End: ast.Location{
																				Line: int(1352),
																				Column: int(21),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p14614,
																		FreeVars: ast.Identifiers{
																			"b",
																		},
																	},
																	Id: "b",
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
												},
												OpFodder: ast.Fodder{},
												Op: ast.BinaryOp(7),
												Right: &ast.LiteralNumber{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1352),
																Column: int(25),
															},
															End: ast.Location{
																Line: int(1352),
																Column: int(26),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p14550,
														FreeVars: nil,
													},
													OriginalString: "0",
												},
											},
											ElseFodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(6),
													Comment: []string{},
												},
											},
											BranchFalse: &ast.LiteralBoolean{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1354),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1354),
															Column: int(13),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
															Comment: []string{},
														},
													},
													Ctx: p14550,
													FreeVars: nil,
												},
												Value: true,
											},
										},
									},
								},
							},
							Fun: nil,
							CloseFodder: nil,
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
						},
					},
					Body: &ast.Conditional{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1355),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(1362),
									Column: int(8),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
									Comment: []string{},
								},
							},
							Ctx: p14541,
							FreeVars: ast.Identifiers{
								"$",
								"a",
								"isContent",
								"std",
							},
						},
						Cond: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1355),
										Column: int(8),
									},
									End: ast.Location{
										Line: int(1355),
										Column: int(22),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p14541,
								FreeVars: ast.Identifiers{
									"a",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1355),
											Column: int(8),
										},
										End: ast.Location{
											Line: int(1355),
											Column: int(19),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p14541,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1355),
												Column: int(8),
											},
											End: ast.Location{
												Line: int(1355),
												Column: int(11),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "isArray",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1355),
														Column: int(20),
													},
													End: ast.Location{
														Line: int(1355),
														Column: int(21),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p14632,
												FreeVars: ast.Identifiers{
													"a",
												},
											},
											Id: "a",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
						ThenFodder: ast.Fodder{},
						BranchTrue: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"$",
									"a",
									"isContent",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: nil,
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "flatMap",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: nil,
								Id: nil,
							},
							FodderLeft: nil,
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Function{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"$",
													"isContent",
													"std",
												},
											},
											ParenLeftFodder: nil,
											Parameters: []ast.Parameter{
												ast.Parameter{
													NameFodder: nil,
													Name: "x",
													EqFodder: nil,
													DefaultArg: nil,
													CommaFodder: nil,
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
												},
											},
											TrailingComma: false,
											ParenRightFodder: nil,
											Body: &ast.Conditional{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"$",
														"isContent",
														"std",
														"x",
													},
												},
												Cond: &ast.Apply{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1356),
																Column: int(35),
															},
															End: ast.Location{
																Line: int(1356),
																Column: int(56),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p14541,
														FreeVars: ast.Identifiers{
															"$",
															"isContent",
															"x",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1356),
																	Column: int(35),
																},
																End: ast.Location{
																	Line: int(1356),
																	Column: int(44),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p14541,
															FreeVars: ast.Identifiers{
																"isContent",
															},
														},
														Id: "isContent",
													},
													FodderLeft: ast.Fodder{},
													Arguments: ast.Arguments{
														Positional: []ast.CommaSeparatedExpr{
															ast.CommaSeparatedExpr{
																Expr: &ast.Apply{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1356),
																				Column: int(45),
																			},
																			End: ast.Location{
																				Line: int(1356),
																				Column: int(55),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p14653,
																		FreeVars: ast.Identifiers{
																			"$",
																			"x",
																		},
																	},
																	Target: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1356),
																					Column: int(45),
																				},
																				End: ast.Location{
																					Line: int(1356),
																					Column: int(52),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p14653,
																			FreeVars: ast.Identifiers{
																				"$",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1356),
																						Column: int(45),
																					},
																					End: ast.Location{
																						Line: int(1356),
																						Column: int(46),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"$",
																				},
																			},
																			Id: "$",
																		},
																		LeftBracketFodder: ast.Fodder{},
																		Index: &ast.LiteralString{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: nil,
																			},
																			Value: "prune",
																			Kind: ast.LiteralStringKind(1),
																			BlockIndent: "",
																			BlockTermIndent: "",
																		},
																		RightBracketFodder: ast.Fodder{},
																		Id: nil,
																	},
																	FodderLeft: ast.Fodder{},
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1356),
																								Column: int(53),
																							},
																							End: ast.Location{
																								Line: int(1356),
																								Column: int(54),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p14662,
																						FreeVars: ast.Identifiers{
																							"x",
																						},
																					},
																					Id: "x",
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	TrailingComma: false,
																	TailStrict: false,
																	FodderRight: ast.Fodder{},
																	TailStrictFodder: nil,
																},
																CommaFodder: nil,
															},
														},
														Named: nil,
													},
													TrailingComma: false,
													TailStrict: false,
													FodderRight: ast.Fodder{},
													TailStrictFodder: nil,
												},
												ThenFodder: nil,
												BranchTrue: &ast.Array{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
															"x",
														},
													},
													Elements: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Apply{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1356),
																			Column: int(8),
																		},
																		End: ast.Location{
																			Line: int(1356),
																			Column: int(20),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p14668,
																	FreeVars: ast.Identifiers{
																		"std",
																		"x",
																	},
																},
																Target: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1356),
																				Column: int(8),
																			},
																			End: ast.Location{
																				Line: int(1356),
																				Column: int(17),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p14668,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1356),
																					Column: int(8),
																				},
																				End: ast.Location{
																					Line: int(1356),
																					Column: int(11),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Id: "std",
																	},
																	LeftBracketFodder: ast.Fodder{},
																	Index: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: nil,
																		},
																		Value: "prune",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	RightBracketFodder: ast.Fodder{},
																	Id: nil,
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1356),
																							Column: int(18),
																						},
																						End: ast.Location{
																							Line: int(1356),
																							Column: int(19),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p14677,
																					FreeVars: ast.Identifiers{
																						"x",
																					},
																				},
																				Id: "x",
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																TrailingComma: false,
																TailStrict: false,
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
															},
															CommaFodder: nil,
														},
													},
													TrailingComma: false,
													CloseFodder: nil,
												},
												ElseFodder: nil,
												BranchFalse: &ast.Array{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Elements: nil,
													TrailingComma: false,
													CloseFodder: nil,
												},
											},
										},
										CommaFodder: nil,
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1356),
														Column: int(30),
													},
													End: ast.Location{
														Line: int(1356),
														Column: int(31),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p14541,
												FreeVars: ast.Identifiers{
													"a",
												},
											},
											Id: "a",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: nil,
							TailStrictFodder: nil,
						},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						BranchFalse: &ast.Conditional{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1357),
										Column: int(10),
									},
									End: ast.Location{
										Line: int(1362),
										Column: int(8),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p14541,
								FreeVars: ast.Identifiers{
									"$",
									"a",
									"isContent",
									"std",
								},
							},
							Cond: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1357),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(1357),
											Column: int(28),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p14541,
									FreeVars: ast.Identifiers{
										"a",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1357),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(1357),
												Column: int(25),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p14541,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1357),
													Column: int(13),
												},
												End: ast.Location{
													Line: int(1357),
													Column: int(16),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "isObject",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1357),
															Column: int(26),
														},
														End: ast.Location{
															Line: int(1357),
															Column: int(27),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p14694,
													FreeVars: ast.Identifiers{
														"a",
													},
												},
												Id: "a",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
							ThenFodder: ast.Fodder{},
							BranchTrue: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"$",
										"a",
										"isContent",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: nil,
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "$objectFlatMerge",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: nil,
									Id: nil,
								},
								FodderLeft: nil,
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"$",
														"a",
														"isContent",
														"std",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: nil,
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "flatMap",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: nil,
													Id: nil,
												},
												FodderLeft: nil,
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.Function{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"$",
																		"a",
																		"isContent",
																		"std",
																	},
																},
																ParenLeftFodder: nil,
																Parameters: []ast.Parameter{
																	ast.Parameter{
																		NameFodder: nil,
																		Name: "x",
																		EqFodder: nil,
																		DefaultArg: nil,
																		CommaFodder: nil,
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																	},
																},
																TrailingComma: false,
																ParenRightFodder: nil,
																Body: &ast.Conditional{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"$",
																			"a",
																			"isContent",
																			"std",
																			"x",
																		},
																	},
																	Cond: &ast.Apply{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1360),
																					Column: int(10),
																				},
																				End: ast.Location{
																					Line: int(1360),
																					Column: int(36),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p14541,
																			FreeVars: ast.Identifiers{
																				"a",
																				"isContent",
																				"std",
																				"x",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1360),
																						Column: int(10),
																					},
																					End: ast.Location{
																						Line: int(1360),
																						Column: int(19),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p14541,
																				FreeVars: ast.Identifiers{
																					"isContent",
																				},
																			},
																			Id: "isContent",
																		},
																		FodderLeft: ast.Fodder{},
																		Arguments: ast.Arguments{
																			Positional: []ast.CommaSeparatedExpr{
																				ast.CommaSeparatedExpr{
																					Expr: &ast.Apply{
																						NodeBase: ast.NodeBase{
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1360),
																									Column: int(20),
																								},
																								End: ast.Location{
																									Line: int(1360),
																									Column: int(35),
																								},
																								File: p1,
																							},
																							Fodder: ast.Fodder{},
																							Ctx: p14723,
																							FreeVars: ast.Identifiers{
																								"a",
																								"std",
																								"x",
																							},
																						},
																						Target: &ast.Index{
																							NodeBase: ast.NodeBase{
																								LocRange: ast.LocationRange{
																									FileName: "<std>",
																									Begin: ast.Location{
																										Line: int(1360),
																										Column: int(20),
																									},
																									End: ast.Location{
																										Line: int(1360),
																										Column: int(29),
																									},
																									File: p1,
																								},
																								Fodder: ast.Fodder{},
																								Ctx: p14723,
																								FreeVars: ast.Identifiers{
																									"std",
																								},
																							},
																							Target: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1360),
																											Column: int(20),
																										},
																										End: ast.Location{
																											Line: int(1360),
																											Column: int(23),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: nil,
																									FreeVars: ast.Identifiers{
																										"std",
																									},
																								},
																								Id: "std",
																							},
																							LeftBracketFodder: ast.Fodder{},
																							Index: &ast.LiteralString{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "",
																										Begin: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										End: ast.Location{
																											Line: int(0),
																											Column: int(0),
																										},
																										File: nil,
																									},
																									Fodder: nil,
																									Ctx: nil,
																									FreeVars: nil,
																								},
																								Value: "prune",
																								Kind: ast.LiteralStringKind(1),
																								BlockIndent: "",
																								BlockTermIndent: "",
																							},
																							RightBracketFodder: ast.Fodder{},
																							Id: nil,
																						},
																						FodderLeft: ast.Fodder{},
																						Arguments: ast.Arguments{
																							Positional: []ast.CommaSeparatedExpr{
																								ast.CommaSeparatedExpr{
																									Expr: &ast.Index{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(1360),
																													Column: int(30),
																												},
																												End: ast.Location{
																													Line: int(1360),
																													Column: int(34),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: p14732,
																											FreeVars: ast.Identifiers{
																												"a",
																												"x",
																											},
																										},
																										Target: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1360),
																														Column: int(30),
																													},
																													End: ast.Location{
																														Line: int(1360),
																														Column: int(31),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p14732,
																												FreeVars: ast.Identifiers{
																													"a",
																												},
																											},
																											Id: "a",
																										},
																										LeftBracketFodder: ast.Fodder{},
																										Index: &ast.Var{
																											NodeBase: ast.NodeBase{
																												LocRange: ast.LocationRange{
																													FileName: "<std>",
																													Begin: ast.Location{
																														Line: int(1360),
																														Column: int(32),
																													},
																													End: ast.Location{
																														Line: int(1360),
																														Column: int(33),
																													},
																													File: p1,
																												},
																												Fodder: ast.Fodder{},
																												Ctx: p14732,
																												FreeVars: ast.Identifiers{
																													"x",
																												},
																											},
																											Id: "x",
																										},
																										RightBracketFodder: ast.Fodder{},
																										Id: nil,
																									},
																									CommaFodder: nil,
																								},
																							},
																							Named: nil,
																						},
																						TrailingComma: false,
																						TailStrict: false,
																						FodderRight: ast.Fodder{},
																						TailStrictFodder: nil,
																					},
																					CommaFodder: nil,
																				},
																			},
																			Named: nil,
																		},
																		TrailingComma: false,
																		TailStrict: false,
																		FodderRight: ast.Fodder{},
																		TailStrictFodder: nil,
																	},
																	ThenFodder: nil,
																	BranchTrue: &ast.Array{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"$",
																				"a",
																				"x",
																			},
																		},
																		Elements: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.DesugaredObject{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1357),
																								Column: int(34),
																							},
																							End: ast.Location{
																								Line: int(1361),
																								Column: int(6),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p14541,
																						FreeVars: ast.Identifiers{
																							"$",
																							"a",
																							"x",
																						},
																					},
																					Asserts: ast.Nodes{},
																					Fields: ast.DesugaredObjectFields{
																						ast.DesugaredObjectField{
																							Hide: ast.ObjectFieldHide(1),
																							Name: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1358),
																											Column: int(8),
																										},
																										End: ast.Location{
																											Line: int(1358),
																											Column: int(9),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p14541,
																									FreeVars: ast.Identifiers{
																										"x",
																									},
																								},
																								Id: "x",
																							},
																							Body: &ast.Apply{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(1358),
																											Column: int(12),
																										},
																										End: ast.Location{
																											Line: int(1358),
																											Column: int(25),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p14747,
																									FreeVars: ast.Identifiers{
																										"$",
																										"a",
																										"x",
																									},
																								},
																								Target: &ast.Index{
																									NodeBase: ast.NodeBase{
																										LocRange: ast.LocationRange{
																											FileName: "<std>",
																											Begin: ast.Location{
																												Line: int(1358),
																												Column: int(12),
																											},
																											End: ast.Location{
																												Line: int(1358),
																												Column: int(19),
																											},
																											File: p1,
																										},
																										Fodder: ast.Fodder{},
																										Ctx: p14747,
																										FreeVars: ast.Identifiers{
																											"$",
																										},
																									},
																									Target: &ast.Var{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "<std>",
																												Begin: ast.Location{
																													Line: int(1358),
																													Column: int(12),
																												},
																												End: ast.Location{
																													Line: int(1358),
																													Column: int(13),
																												},
																												File: p1,
																											},
																											Fodder: ast.Fodder{},
																											Ctx: nil,
																											FreeVars: ast.Identifiers{
																												"$",
																											},
																										},
																										Id: "$",
																									},
																									LeftBracketFodder: ast.Fodder{},
																									Index: &ast.LiteralString{
																										NodeBase: ast.NodeBase{
																											LocRange: ast.LocationRange{
																												FileName: "",
																												Begin: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												End: ast.Location{
																													Line: int(0),
																													Column: int(0),
																												},
																												File: nil,
																											},
																											Fodder: nil,
																											Ctx: nil,
																											FreeVars: nil,
																										},
																										Value: "prune",
																										Kind: ast.LiteralStringKind(1),
																										BlockIndent: "",
																										BlockTermIndent: "",
																									},
																									RightBracketFodder: ast.Fodder{},
																									Id: nil,
																								},
																								FodderLeft: ast.Fodder{},
																								Arguments: ast.Arguments{
																									Positional: []ast.CommaSeparatedExpr{
																										ast.CommaSeparatedExpr{
																											Expr: &ast.Index{
																												NodeBase: ast.NodeBase{
																													LocRange: ast.LocationRange{
																														FileName: "<std>",
																														Begin: ast.Location{
																															Line: int(1358),
																															Column: int(20),
																														},
																														End: ast.Location{
																															Line: int(1358),
																															Column: int(24),
																														},
																														File: p1,
																													},
																													Fodder: ast.Fodder{},
																													Ctx: p14756,
																													FreeVars: ast.Identifiers{
																														"a",
																														"x",
																													},
																												},
																												Target: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(1358),
																																Column: int(20),
																															},
																															End: ast.Location{
																																Line: int(1358),
																																Column: int(21),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p14756,
																														FreeVars: ast.Identifiers{
																															"a",
																														},
																													},
																													Id: "a",
																												},
																												LeftBracketFodder: ast.Fodder{},
																												Index: &ast.Var{
																													NodeBase: ast.NodeBase{
																														LocRange: ast.LocationRange{
																															FileName: "<std>",
																															Begin: ast.Location{
																																Line: int(1358),
																																Column: int(22),
																															},
																															End: ast.Location{
																																Line: int(1358),
																																Column: int(23),
																															},
																															File: p1,
																														},
																														Fodder: ast.Fodder{},
																														Ctx: p14756,
																														FreeVars: ast.Identifiers{
																															"x",
																														},
																													},
																													Id: "x",
																												},
																												RightBracketFodder: ast.Fodder{},
																												Id: nil,
																											},
																											CommaFodder: nil,
																										},
																									},
																									Named: nil,
																								},
																								TrailingComma: false,
																								TailStrict: false,
																								FodderRight: ast.Fodder{},
																								TailStrictFodder: nil,
																							},
																							PlusSuper: false,
																							LocRange: ast.LocationRange{
																								FileName: "<std>",
																								Begin: ast.Location{
																									Line: int(1358),
																									Column: int(7),
																								},
																								End: ast.Location{
																									Line: int(1358),
																									Column: int(25),
																								},
																								File: p1,
																							},
																						},
																					},
																					Locals: ast.LocalBinds{},
																				},
																				CommaFodder: nil,
																			},
																		},
																		TrailingComma: false,
																		CloseFodder: nil,
																	},
																	ElseFodder: nil,
																	BranchFalse: &ast.Array{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: nil,
																		},
																		Elements: nil,
																		TrailingComma: false,
																		CloseFodder: nil,
																	},
																},
															},
															CommaFodder: nil,
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.Apply{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1359),
																			Column: int(16),
																		},
																		End: ast.Location{
																			Line: int(1359),
																			Column: int(35),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p14541,
																	FreeVars: ast.Identifiers{
																		"a",
																		"std",
																	},
																},
																Target: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1359),
																				Column: int(16),
																			},
																			End: ast.Location{
																				Line: int(1359),
																				Column: int(32),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p14541,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1359),
																					Column: int(16),
																				},
																				End: ast.Location{
																					Line: int(1359),
																					Column: int(19),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Id: "std",
																	},
																	LeftBracketFodder: ast.Fodder{},
																	Index: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: nil,
																		},
																		Value: "objectFields",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	RightBracketFodder: ast.Fodder{},
																	Id: nil,
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1359),
																							Column: int(33),
																						},
																						End: ast.Location{
																							Line: int(1359),
																							Column: int(34),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p14772,
																					FreeVars: ast.Identifiers{
																						"a",
																					},
																				},
																				Id: "a",
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																TrailingComma: false,
																TailStrict: false,
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
															},
															CommaFodder: nil,
														},
													},
													Named: nil,
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: nil,
												TailStrictFodder: nil,
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: nil,
								TailStrictFodder: nil,
							},
							ElseFodder: ast.Fodder{},
							BranchFalse: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1362),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(1362),
											Column: int(8),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
											Comment: []string{},
										},
									},
									Ctx: p14541,
									FreeVars: ast.Identifiers{
										"a",
									},
								},
								Id: "a",
							},
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(1345),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1362),
					Column: int(8),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "findSubstr",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "pat",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1364),
								Column: int(14),
							},
							End: ast.Location{
								Line: int(1364),
								Column: int(17),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1364),
								Column: int(19),
							},
							End: ast.Location{
								Line: int(1364),
								Column: int(22),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1365),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1375),
								Column: int(91),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p14783,
						FreeVars: ast.Identifiers{
							"pat",
							"std",
							"str",
						},
					},
					Cond: &ast.Unary{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1365),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(1365),
									Column: int(26),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p14783,
							FreeVars: ast.Identifiers{
								"pat",
								"std",
							},
						},
						Op: ast.UnaryOp(0),
						Expr: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1365),
										Column: int(9),
									},
									End: ast.Location{
										Line: int(1365),
										Column: int(26),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p14783,
								FreeVars: ast.Identifiers{
									"pat",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1365),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(1365),
											Column: int(21),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p14783,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1365),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(1365),
												Column: int(12),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "isString",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1365),
														Column: int(22),
													},
													End: ast.Location{
														Line: int(1365),
														Column: int(25),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p14796,
												FreeVars: ast.Identifiers{
													"pat",
												},
											},
											Id: "pat",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
					},
					ThenFodder: ast.Fodder{},
					BranchTrue: &ast.Error{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1366),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(1366),
									Column: int(82),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
									Comment: []string{},
								},
							},
							Ctx: p14783,
							FreeVars: ast.Identifiers{
								"pat",
								"std",
							},
						},
						Expr: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1366),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(1366),
										Column: int(82),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p14783,
								FreeVars: ast.Identifiers{
									"pat",
									"std",
								},
							},
							Left: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1366),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(1366),
											Column: int(66),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p14783,
									FreeVars: nil,
								},
								Value: "findSubstr first parameter should be a string, got ",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(3),
							Right: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1366),
											Column: int(69),
										},
										End: ast.Location{
											Line: int(1366),
											Column: int(82),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p14783,
									FreeVars: ast.Identifiers{
										"pat",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1366),
												Column: int(69),
											},
											End: ast.Location{
												Line: int(1366),
												Column: int(77),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p14783,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1366),
													Column: int(69),
												},
												End: ast.Location{
													Line: int(1366),
													Column: int(72),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "type",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1366),
															Column: int(78),
														},
														End: ast.Location{
															Line: int(1366),
															Column: int(81),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p14813,
													FreeVars: ast.Identifiers{
														"pat",
													},
												},
												Id: "pat",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
						},
					},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
							Comment: []string{},
						},
					},
					BranchFalse: &ast.Conditional{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1367),
									Column: int(10),
								},
								End: ast.Location{
									Line: int(1375),
									Column: int(91),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p14783,
							FreeVars: ast.Identifiers{
								"pat",
								"std",
								"str",
							},
						},
						Cond: &ast.Unary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1367),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(1367),
										Column: int(31),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p14783,
								FreeVars: ast.Identifiers{
									"std",
									"str",
								},
							},
							Op: ast.UnaryOp(0),
							Expr: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1367),
											Column: int(14),
										},
										End: ast.Location{
											Line: int(1367),
											Column: int(31),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p14783,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1367),
												Column: int(14),
											},
											End: ast.Location{
												Line: int(1367),
												Column: int(26),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p14783,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1367),
													Column: int(14),
												},
												End: ast.Location{
													Line: int(1367),
													Column: int(17),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "isString",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1367),
															Column: int(27),
														},
														End: ast.Location{
															Line: int(1367),
															Column: int(30),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p14829,
													FreeVars: ast.Identifiers{
														"str",
													},
												},
												Id: "str",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
						},
						ThenFodder: ast.Fodder{},
						BranchTrue: &ast.Error{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1368),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(1368),
										Column: int(83),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
										Comment: []string{},
									},
								},
								Ctx: p14783,
								FreeVars: ast.Identifiers{
									"std",
									"str",
								},
							},
							Expr: &ast.Binary{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1368),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(1368),
											Column: int(83),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p14783,
									FreeVars: ast.Identifiers{
										"std",
										"str",
									},
								},
								Left: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1368),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(1368),
												Column: int(67),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p14783,
										FreeVars: nil,
									},
									Value: "findSubstr second parameter should be a string, got ",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								OpFodder: ast.Fodder{},
								Op: ast.BinaryOp(3),
								Right: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1368),
												Column: int(70),
											},
											End: ast.Location{
												Line: int(1368),
												Column: int(83),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p14783,
										FreeVars: ast.Identifiers{
											"std",
											"str",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1368),
													Column: int(70),
												},
												End: ast.Location{
													Line: int(1368),
													Column: int(78),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p14783,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1368),
														Column: int(70),
													},
													End: ast.Location{
														Line: int(1368),
														Column: int(73),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "type",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1368),
																Column: int(79),
															},
															End: ast.Location{
																Line: int(1368),
																Column: int(82),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p14846,
														FreeVars: ast.Identifiers{
															"str",
														},
													},
													Id: "str",
												},
												CommaFodder: nil,
											},
										},
										Named: nil,
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
							},
						},
						ElseFodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						BranchFalse: &ast.Local{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1370),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(1375),
										Column: int(91),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(6),
										Comment: []string{},
									},
								},
								Ctx: p14783,
								FreeVars: ast.Identifiers{
									"pat",
									"std",
									"str",
								},
							},
							Binds: ast.LocalBinds{
								ast.LocalBind{
									VarFodder: ast.Fodder{},
									Variable: "pat_len",
									EqFodder: ast.Fodder{},
									Body: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1370),
													Column: int(23),
												},
												End: ast.Location{
													Line: int(1370),
													Column: int(38),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p14854,
											FreeVars: ast.Identifiers{
												"pat",
												"std",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1370),
														Column: int(23),
													},
													End: ast.Location{
														Line: int(1370),
														Column: int(33),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p14854,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1370),
															Column: int(23),
														},
														End: ast.Location{
															Line: int(1370),
															Column: int(26),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: ast.Fodder{},
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "length",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: ast.Fodder{},
											Id: nil,
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1370),
																	Column: int(34),
																},
																End: ast.Location{
																	Line: int(1370),
																	Column: int(37),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p14863,
															FreeVars: ast.Identifiers{
																"pat",
															},
														},
														Id: "pat",
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
									Fun: nil,
									CloseFodder: ast.Fodder{},
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1370),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(1370),
											Column: int(38),
										},
										File: p1,
									},
								},
							},
							Body: &ast.Local{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1371),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(1375),
											Column: int(91),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
											Comment: []string{},
										},
									},
									Ctx: p14783,
									FreeVars: ast.Identifiers{
										"pat",
										"pat_len",
										"std",
										"str",
									},
								},
								Binds: ast.LocalBinds{
									ast.LocalBind{
										VarFodder: ast.Fodder{},
										Variable: "str_len",
										EqFodder: ast.Fodder{},
										Body: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1371),
														Column: int(23),
													},
													End: ast.Location{
														Line: int(1371),
														Column: int(38),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p14870,
												FreeVars: ast.Identifiers{
													"std",
													"str",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1371),
															Column: int(23),
														},
														End: ast.Location{
															Line: int(1371),
															Column: int(33),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p14870,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1371),
																Column: int(23),
															},
															End: ast.Location{
																Line: int(1371),
																Column: int(26),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "length",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.Var{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1371),
																		Column: int(34),
																	},
																	End: ast.Location{
																		Line: int(1371),
																		Column: int(37),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p14879,
																FreeVars: ast.Identifiers{
																	"str",
																},
															},
															Id: "str",
														},
														CommaFodder: nil,
													},
												},
												Named: nil,
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
										Fun: nil,
										CloseFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1371),
												Column: int(13),
											},
											End: ast.Location{
												Line: int(1371),
												Column: int(38),
											},
											File: p1,
										},
									},
								},
								Body: &ast.Conditional{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1372),
												Column: int(7),
											},
											End: ast.Location{
												Line: int(1375),
												Column: int(91),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(6),
												Comment: []string{},
											},
										},
										Ctx: p14783,
										FreeVars: ast.Identifiers{
											"pat",
											"pat_len",
											"std",
											"str",
											"str_len",
										},
									},
									Cond: &ast.Binary{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1372),
													Column: int(10),
												},
												End: ast.Location{
													Line: int(1372),
													Column: int(59),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p14783,
											FreeVars: ast.Identifiers{
												"pat_len",
												"str_len",
											},
										},
										Left: &ast.Binary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1372),
														Column: int(10),
													},
													End: ast.Location{
														Line: int(1372),
														Column: int(38),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p14783,
												FreeVars: ast.Identifiers{
													"pat_len",
													"str_len",
												},
											},
											Left: &ast.Binary{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1372),
															Column: int(10),
														},
														End: ast.Location{
															Line: int(1372),
															Column: int(22),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p14783,
													FreeVars: ast.Identifiers{
														"pat_len",
													},
												},
												Left: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1372),
																Column: int(10),
															},
															End: ast.Location{
																Line: int(1372),
																Column: int(17),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p14783,
														FreeVars: ast.Identifiers{
															"pat_len",
														},
													},
													Id: "pat_len",
												},
												OpFodder: ast.Fodder{},
												Op: ast.BinaryOp(12),
												Right: &ast.LiteralNumber{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1372),
																Column: int(21),
															},
															End: ast.Location{
																Line: int(1372),
																Column: int(22),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p14783,
														FreeVars: nil,
													},
													OriginalString: "0",
												},
											},
											OpFodder: ast.Fodder{},
											Op: ast.BinaryOp(18),
											Right: &ast.Binary{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1372),
															Column: int(26),
														},
														End: ast.Location{
															Line: int(1372),
															Column: int(38),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p14783,
													FreeVars: ast.Identifiers{
														"str_len",
													},
												},
												Left: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1372),
																Column: int(26),
															},
															End: ast.Location{
																Line: int(1372),
																Column: int(33),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p14783,
														FreeVars: ast.Identifiers{
															"str_len",
														},
													},
													Id: "str_len",
												},
												OpFodder: ast.Fodder{},
												Op: ast.BinaryOp(12),
												Right: &ast.LiteralNumber{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1372),
																Column: int(37),
															},
															End: ast.Location{
																Line: int(1372),
																Column: int(38),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p14783,
														FreeVars: nil,
													},
													OriginalString: "0",
												},
											},
										},
										OpFodder: ast.Fodder{},
										Op: ast.BinaryOp(18),
										Right: &ast.Binary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1372),
														Column: int(42),
													},
													End: ast.Location{
														Line: int(1372),
														Column: int(59),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p14783,
												FreeVars: ast.Identifiers{
													"pat_len",
													"str_len",
												},
											},
											Left: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1372),
															Column: int(42),
														},
														End: ast.Location{
															Line: int(1372),
															Column: int(49),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p14783,
													FreeVars: ast.Identifiers{
														"pat_len",
													},
												},
												Id: "pat_len",
											},
											OpFodder: ast.Fodder{},
											Op: ast.BinaryOp(7),
											Right: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1372),
															Column: int(52),
														},
														End: ast.Location{
															Line: int(1372),
															Column: int(59),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p14783,
													FreeVars: ast.Identifiers{
														"str_len",
													},
												},
												Id: "str_len",
											},
										},
									},
									ThenFodder: ast.Fodder{},
									BranchTrue: &ast.Array{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1373),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1373),
													Column: int(11),
												},
												File: p1,
											},
											Fodder: ast.Fodder{
												ast.FodderElement{
													Kind: ast.FodderKind(0),
													Blanks: int(0),
													Indent: int(8),
													Comment: []string{},
												},
											},
											Ctx: p14783,
											FreeVars: nil,
										},
										Elements: nil,
										TrailingComma: false,
										CloseFodder: ast.Fodder{},
									},
									ElseFodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
											Comment: []string{},
										},
									},
									BranchFalse: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1375),
													Column: int(9),
												},
												End: ast.Location{
													Line: int(1375),
													Column: int(91),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p14783,
											FreeVars: ast.Identifiers{
												"pat",
												"pat_len",
												"std",
												"str",
												"str_len",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1375),
														Column: int(9),
													},
													End: ast.Location{
														Line: int(1375),
														Column: int(19),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p14783,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1375),
															Column: int(9),
														},
														End: ast.Location{
															Line: int(1375),
															Column: int(12),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(8),
															Comment: []string{},
														},
													},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: ast.Fodder{},
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "filter",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: ast.Fodder{},
											Id: nil,
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Function{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1375),
																	Column: int(20),
																},
																End: ast.Location{
																	Line: int(1375),
																	Column: int(57),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p14917,
															FreeVars: ast.Identifiers{
																"pat",
																"pat_len",
																"std",
																"str",
															},
														},
														ParenLeftFodder: ast.Fodder{},
														Parameters: []ast.Parameter{
															ast.Parameter{
																NameFodder: ast.Fodder{},
																Name: "i",
																EqFodder: nil,
																DefaultArg: nil,
																CommaFodder: nil,
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1375),
																		Column: int(29),
																	},
																	End: ast.Location{
																		Line: int(1375),
																		Column: int(30),
																	},
																	File: p1,
																},
															},
														},
														TrailingComma: false,
														ParenRightFodder: ast.Fodder{},
														Body: &ast.Binary{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1375),
																		Column: int(32),
																	},
																	End: ast.Location{
																		Line: int(1375),
																		Column: int(57),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p14921,
																FreeVars: ast.Identifiers{
																	"i",
																	"pat",
																	"pat_len",
																	"std",
																	"str",
																},
															},
															Left: &ast.Apply{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"i",
																		"pat_len",
																		"std",
																		"str",
																	},
																},
																Target: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Id: "std",
																	},
																	LeftBracketFodder: nil,
																	Index: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: nil,
																		},
																		Value: "slice",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	RightBracketFodder: nil,
																	Id: nil,
																},
																FodderLeft: nil,
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1375),
																							Column: int(32),
																						},
																						End: ast.Location{
																							Line: int(1375),
																							Column: int(35),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p14921,
																					FreeVars: ast.Identifiers{
																						"str",
																					},
																				},
																				Id: "str",
																			},
																			CommaFodder: nil,
																		},
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1375),
																							Column: int(36),
																						},
																						End: ast.Location{
																							Line: int(1375),
																							Column: int(37),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p14921,
																					FreeVars: ast.Identifiers{
																						"i",
																					},
																				},
																				Id: "i",
																			},
																			CommaFodder: nil,
																		},
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Binary{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1375),
																							Column: int(38),
																						},
																						End: ast.Location{
																							Line: int(1375),
																							Column: int(49),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p14921,
																					FreeVars: ast.Identifiers{
																						"i",
																						"pat_len",
																					},
																				},
																				Left: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1375),
																								Column: int(38),
																							},
																							End: ast.Location{
																								Line: int(1375),
																								Column: int(39),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p14921,
																						FreeVars: ast.Identifiers{
																							"i",
																						},
																					},
																					Id: "i",
																				},
																				OpFodder: ast.Fodder{},
																				Op: ast.BinaryOp(3),
																				Right: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(1375),
																								Column: int(42),
																							},
																							End: ast.Location{
																								Line: int(1375),
																								Column: int(49),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p14921,
																						FreeVars: ast.Identifiers{
																							"pat_len",
																						},
																					},
																					Id: "pat_len",
																				},
																			},
																			CommaFodder: nil,
																		},
																		ast.CommaSeparatedExpr{
																			Expr: &ast.LiteralNull{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "",
																						Begin: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						End: ast.Location{
																							Line: int(0),
																							Column: int(0),
																						},
																						File: nil,
																					},
																					Fodder: nil,
																					Ctx: nil,
																					FreeVars: nil,
																				},
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: nil,
																},
																TrailingComma: false,
																TailStrict: false,
																FodderRight: nil,
																TailStrictFodder: nil,
															},
															OpFodder: ast.Fodder{},
															Op: ast.BinaryOp(12),
															Right: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1375),
																			Column: int(54),
																		},
																		End: ast.Location{
																			Line: int(1375),
																			Column: int(57),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p14921,
																	FreeVars: ast.Identifiers{
																		"pat",
																	},
																},
																Id: "pat",
															},
														},
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1375),
																	Column: int(59),
																},
																End: ast.Location{
																	Line: int(1375),
																	Column: int(90),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p14917,
															FreeVars: ast.Identifiers{
																"pat_len",
																"std",
																"str_len",
															},
														},
														Target: &ast.Index{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1375),
																		Column: int(59),
																	},
																	End: ast.Location{
																		Line: int(1375),
																		Column: int(68),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p14917,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Target: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1375),
																			Column: int(59),
																		},
																		End: ast.Location{
																			Line: int(1375),
																			Column: int(62),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Id: "std",
															},
															LeftBracketFodder: ast.Fodder{},
															Index: &ast.LiteralString{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: nil,
																},
																Value: "range",
																Kind: ast.LiteralStringKind(1),
																BlockIndent: "",
																BlockTermIndent: "",
															},
															RightBracketFodder: ast.Fodder{},
															Id: nil,
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.LiteralNumber{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1375),
																					Column: int(69),
																				},
																				End: ast.Location{
																					Line: int(1375),
																					Column: int(70),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p14953,
																			FreeVars: nil,
																		},
																		OriginalString: "0",
																	},
																	CommaFodder: ast.Fodder{},
																},
																ast.CommaSeparatedExpr{
																	Expr: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(1375),
																					Column: int(72),
																				},
																				End: ast.Location{
																					Line: int(1375),
																					Column: int(89),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p14953,
																			FreeVars: ast.Identifiers{
																				"pat_len",
																				"str_len",
																			},
																		},
																		Left: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1375),
																						Column: int(72),
																					},
																					End: ast.Location{
																						Line: int(1375),
																						Column: int(79),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p14953,
																				FreeVars: ast.Identifiers{
																					"str_len",
																				},
																			},
																			Id: "str_len",
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(4),
																		Right: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1375),
																						Column: int(82),
																					},
																					End: ast.Location{
																						Line: int(1375),
																						Column: int(89),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p14953,
																				FreeVars: ast.Identifiers{
																					"pat_len",
																				},
																			},
																			Id: "pat_len",
																		},
																	},
																	CommaFodder: nil,
																},
															},
															Named: nil,
														},
														TrailingComma: false,
														TailStrict: false,
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
								},
							},
						},
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(1364),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1375),
					Column: int(91),
				},
				File: p1,
			},
		},
		ast.DesugaredObjectField{
			Hide: ast.ObjectFieldHide(0),
			Name: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
				Value: "find",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p11,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "value",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1377),
								Column: int(8),
							},
							End: ast.Location{
								Line: int(1377),
								Column: int(13),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "arr",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1377),
								Column: int(15),
							},
							End: ast.Location{
								Line: int(1377),
								Column: int(18),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(1378),
								Column: int(5),
							},
							End: ast.Location{
								Line: int(1381),
								Column: int(81),
							},
							File: p1,
						},
						Fodder: ast.Fodder{
							ast.FodderElement{
								Kind: ast.FodderKind(0),
								Blanks: int(0),
								Indent: int(4),
								Comment: []string{},
							},
						},
						Ctx: p14966,
						FreeVars: ast.Identifiers{
							"arr",
							"std",
							"value",
						},
					},
					Cond: &ast.Unary{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1378),
									Column: int(8),
								},
								End: ast.Location{
									Line: int(1378),
									Column: int(25),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p14966,
							FreeVars: ast.Identifiers{
								"arr",
								"std",
							},
						},
						Op: ast.UnaryOp(0),
						Expr: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1378),
										Column: int(9),
									},
									End: ast.Location{
										Line: int(1378),
										Column: int(25),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p14966,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1378),
											Column: int(9),
										},
										End: ast.Location{
											Line: int(1378),
											Column: int(20),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p14966,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1378),
												Column: int(9),
											},
											End: ast.Location{
												Line: int(1378),
												Column: int(12),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: ast.Fodder{},
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "isArray",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: ast.Fodder{},
								Id: nil,
							},
							FodderLeft: ast.Fodder{},
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1378),
														Column: int(21),
													},
													End: ast.Location{
														Line: int(1378),
														Column: int(24),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p14979,
												FreeVars: ast.Identifiers{
													"arr",
												},
											},
											Id: "arr",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: ast.Fodder{},
							TailStrictFodder: nil,
						},
					},
					ThenFodder: ast.Fodder{},
					BranchTrue: &ast.Error{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1379),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(1379),
									Column: int(77),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(6),
									Comment: []string{},
								},
							},
							Ctx: p14966,
							FreeVars: ast.Identifiers{
								"arr",
								"std",
							},
						},
						Expr: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1379),
										Column: int(13),
									},
									End: ast.Location{
										Line: int(1379),
										Column: int(77),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p14966,
								FreeVars: ast.Identifiers{
									"arr",
									"std",
								},
							},
							Left: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1379),
											Column: int(13),
										},
										End: ast.Location{
											Line: int(1379),
											Column: int(61),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p14966,
									FreeVars: nil,
								},
								Value: "find second parameter should be an array, got ",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(3),
							Right: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1379),
											Column: int(64),
										},
										End: ast.Location{
											Line: int(1379),
											Column: int(77),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p14966,
									FreeVars: ast.Identifiers{
										"arr",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(1379),
												Column: int(64),
											},
											End: ast.Location{
												Line: int(1379),
												Column: int(72),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p14966,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1379),
													Column: int(64),
												},
												End: ast.Location{
													Line: int(1379),
													Column: int(67),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: ast.Fodder{},
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "type",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: ast.Fodder{},
									Id: nil,
								},
								FodderLeft: ast.Fodder{},
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1379),
															Column: int(73),
														},
														End: ast.Location{
															Line: int(1379),
															Column: int(76),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p14996,
													FreeVars: ast.Identifiers{
														"arr",
													},
												},
												Id: "arr",
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: ast.Fodder{},
								TailStrictFodder: nil,
							},
						},
					},
					ElseFodder: ast.Fodder{
						ast.FodderElement{
							Kind: ast.FodderKind(0),
							Blanks: int(0),
							Indent: int(4),
							Comment: []string{},
						},
					},
					BranchFalse: &ast.Apply{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(1381),
									Column: int(7),
								},
								End: ast.Location{
									Line: int(1381),
									Column: int(81),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p14966,
							FreeVars: ast.Identifiers{
								"arr",
								"std",
								"value",
							},
						},
						Target: &ast.Index{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(1381),
										Column: int(7),
									},
									End: ast.Location{
										Line: int(1381),
										Column: int(17),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p14966,
								FreeVars: ast.Identifiers{
									"std",
								},
							},
							Target: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(1381),
											Column: int(7),
										},
										End: ast.Location{
											Line: int(1381),
											Column: int(10),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(6),
											Comment: []string{},
										},
									},
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Id: "std",
							},
							LeftBracketFodder: ast.Fodder{},
							Index: &ast.LiteralString{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: nil,
								},
								Value: "filter",
								Kind: ast.LiteralStringKind(1),
								BlockIndent: "",
								BlockTermIndent: "",
							},
							RightBracketFodder: ast.Fodder{},
							Id: nil,
						},
						FodderLeft: ast.Fodder{},
						Arguments: ast.Arguments{
							Positional: []ast.CommaSeparatedExpr{
								ast.CommaSeparatedExpr{
									Expr: &ast.Function{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1381),
													Column: int(18),
												},
												End: ast.Location{
													Line: int(1381),
													Column: int(45),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p15009,
											FreeVars: ast.Identifiers{
												"arr",
												"value",
											},
										},
										ParenLeftFodder: ast.Fodder{},
										Parameters: []ast.Parameter{
											ast.Parameter{
												NameFodder: ast.Fodder{},
												Name: "i",
												EqFodder: nil,
												DefaultArg: nil,
												CommaFodder: nil,
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1381),
														Column: int(27),
													},
													End: ast.Location{
														Line: int(1381),
														Column: int(28),
													},
													File: p1,
												},
											},
										},
										TrailingComma: false,
										ParenRightFodder: ast.Fodder{},
										Body: &ast.Binary{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1381),
														Column: int(30),
													},
													End: ast.Location{
														Line: int(1381),
														Column: int(45),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p15013,
												FreeVars: ast.Identifiers{
													"arr",
													"i",
													"value",
												},
											},
											Left: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1381),
															Column: int(30),
														},
														End: ast.Location{
															Line: int(1381),
															Column: int(36),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p15013,
													FreeVars: ast.Identifiers{
														"arr",
														"i",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1381),
																Column: int(30),
															},
															End: ast.Location{
																Line: int(1381),
																Column: int(33),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p15013,
														FreeVars: ast.Identifiers{
															"arr",
														},
													},
													Id: "arr",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1381),
																Column: int(34),
															},
															End: ast.Location{
																Line: int(1381),
																Column: int(35),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p15013,
														FreeVars: ast.Identifiers{
															"i",
														},
													},
													Id: "i",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											OpFodder: ast.Fodder{},
											Op: ast.BinaryOp(12),
											Right: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1381),
															Column: int(40),
														},
														End: ast.Location{
															Line: int(1381),
															Column: int(45),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p15013,
													FreeVars: ast.Identifiers{
														"value",
													},
												},
												Id: "value",
											},
										},
									},
									CommaFodder: ast.Fodder{},
								},
								ast.CommaSeparatedExpr{
									Expr: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(1381),
													Column: int(47),
												},
												End: ast.Location{
													Line: int(1381),
													Column: int(80),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p15009,
											FreeVars: ast.Identifiers{
												"arr",
												"std",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(1381),
														Column: int(47),
													},
													End: ast.Location{
														Line: int(1381),
														Column: int(56),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p15009,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1381),
															Column: int(47),
														},
														End: ast.Location{
															Line: int(1381),
															Column: int(50),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: ast.Fodder{},
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "range",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: ast.Fodder{},
											Id: nil,
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.LiteralNumber{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1381),
																	Column: int(57),
																},
																End: ast.Location{
																	Line: int(1381),
																	Column: int(58),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p15032,
															FreeVars: nil,
														},
														OriginalString: "0",
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Binary{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1381),
																	Column: int(60),
																},
																End: ast.Location{
																	Line: int(1381),
																	Column: int(79),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p15032,
															FreeVars: ast.Identifiers{
																"arr",
																"std",
															},
														},
														Left: &ast.Apply{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1381),
																		Column: int(60),
																	},
																	End: ast.Location{
																		Line: int(1381),
																		Column: int(75),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p15032,
																FreeVars: ast.Identifiers{
																	"arr",
																	"std",
																},
															},
															Target: &ast.Index{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1381),
																			Column: int(60),
																		},
																		End: ast.Location{
																			Line: int(1381),
																			Column: int(70),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p15032,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Target: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1381),
																				Column: int(60),
																			},
																			End: ast.Location{
																				Line: int(1381),
																				Column: int(63),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Id: "std",
																},
																LeftBracketFodder: ast.Fodder{},
																Index: &ast.LiteralString{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: nil,
																	},
																	Value: "length",
																	Kind: ast.LiteralStringKind(1),
																	BlockIndent: "",
																	BlockTermIndent: "",
																},
																RightBracketFodder: ast.Fodder{},
																Id: nil,
															},
															FodderLeft: ast.Fodder{},
															Arguments: ast.Arguments{
																Positional: []ast.CommaSeparatedExpr{
																	ast.CommaSeparatedExpr{
																		Expr: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1381),
																						Column: int(71),
																					},
																					End: ast.Location{
																						Line: int(1381),
																						Column: int(74),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p15044,
																				FreeVars: ast.Identifiers{
																					"arr",
																				},
																			},
																			Id: "arr",
																		},
																		CommaFodder: nil,
																	},
																},
																Named: nil,
															},
															TrailingComma: false,
															TailStrict: false,
															FodderRight: ast.Fodder{},
															TailStrictFodder: nil,
														},
														OpFodder: ast.Fodder{},
														Op: ast.BinaryOp(4),
														Right: &ast.LiteralNumber{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(1381),
																		Column: int(78),
																	},
																	End: ast.Location{
																		Line: int(1381),
																		Column: int(79),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p15032,
																FreeVars: nil,
															},
															OriginalString: "1",
														},
													},
													CommaFodder: nil,
												},
											},
											Named: nil,
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
									CommaFodder: nil,
								},
							},
							Named: nil,
						},
						TrailingComma: false,
						TailStrict: false,
						FodderRight: ast.Fodder{},
						TailStrictFodder: nil,
					},
				},
			},
			PlusSuper: false,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(1377),
					Column: int(3),
				},
				End: ast.Location{
					Line: int(1381),
					Column: int(81),
				},
				File: p1,
			},
		},
	},
	Locals: ast.LocalBinds{
		ast.LocalBind{
			VarFodder: nil,
			Variable: "std",
			EqFodder: nil,
			Body: &ast.Self{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "<std>",
						Begin: ast.Location{
							Line: int(25),
							Column: int(15),
						},
						End: ast.Location{
							Line: int(25),
							Column: int(19),
						},
						File: p1,
					},
					Fodder: ast.Fodder{},
					Ctx: p15049,
					FreeVars: nil,
				},
			},
			Fun: nil,
			CloseFodder: nil,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(25),
					Column: int(9),
				},
				End: ast.Location{
					Line: int(25),
					Column: int(19),
				},
				File: p1,
			},
		},
		ast.LocalBind{
			VarFodder: nil,
			Variable: "id",
			EqFodder: nil,
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "<std>",
						Begin: ast.Location{
							Line: int(26),
							Column: int(14),
						},
						End: ast.Location{
							Line: int(26),
							Column: int(27),
						},
						File: p1,
					},
					Fodder: ast.Fodder{},
					Ctx: p15051,
					FreeVars: nil,
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "x",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(26),
								Column: int(23),
							},
							End: ast.Location{
								Line: int(26),
								Column: int(24),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Var{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(26),
								Column: int(26),
							},
							End: ast.Location{
								Line: int(26),
								Column: int(27),
							},
							File: p1,
						},
						Fodder: ast.Fodder{},
						Ctx: p15054,
						FreeVars: ast.Identifiers{
							"x",
						},
					},
					Id: "x",
				},
			},
			Fun: nil,
			CloseFodder: nil,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(26),
					Column: int(9),
				},
				End: ast.Location{
					Line: int(26),
					Column: int(27),
				},
				File: p1,
			},
		},
		ast.LocalBind{
			VarFodder: nil,
			Variable: "parse_nat",
			EqFodder: nil,
			Body: &ast.Function{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: p15057,
					FreeVars: ast.Identifiers{
						"std",
					},
				},
				ParenLeftFodder: ast.Fodder{},
				Parameters: []ast.Parameter{
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "str",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: ast.Fodder{},
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(76),
								Column: int(19),
							},
							End: ast.Location{
								Line: int(76),
								Column: int(22),
							},
							File: p1,
						},
					},
					ast.Parameter{
						NameFodder: ast.Fodder{},
						Name: "base",
						EqFodder: nil,
						DefaultArg: nil,
						CommaFodder: nil,
						LocRange: ast.LocationRange{
							FileName: "<std>",
							Begin: ast.Location{
								Line: int(76),
								Column: int(24),
							},
							End: ast.Location{
								Line: int(76),
								Column: int(28),
							},
							File: p1,
						},
					},
				},
				TrailingComma: false,
				ParenRightFodder: ast.Fodder{},
				Body: &ast.Conditional{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "",
							Begin: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							End: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							File: nil,
						},
						Fodder: nil,
						Ctx: nil,
						FreeVars: ast.Identifiers{
							"base",
							"std",
							"str",
						},
					},
					Cond: &ast.Binary{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(77),
									Column: int(12),
								},
								End: ast.Location{
									Line: int(77),
									Column: int(34),
								},
								File: p1,
							},
							Fodder: ast.Fodder{},
							Ctx: p15063,
							FreeVars: ast.Identifiers{
								"base",
							},
						},
						Left: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(77),
										Column: int(12),
									},
									End: ast.Location{
										Line: int(77),
										Column: int(20),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p15066,
								FreeVars: ast.Identifiers{
									"base",
								},
							},
							Left: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(77),
											Column: int(12),
										},
										End: ast.Location{
											Line: int(77),
											Column: int(16),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p15069,
									FreeVars: ast.Identifiers{
										"base",
									},
								},
								Id: "base",
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(7),
							Right: &ast.LiteralNumber{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(77),
											Column: int(19),
										},
										End: ast.Location{
											Line: int(77),
											Column: int(20),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p15072,
									FreeVars: nil,
								},
								OriginalString: "0",
							},
						},
						OpFodder: ast.Fodder{},
						Op: ast.BinaryOp(17),
						Right: &ast.Binary{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(77),
										Column: int(24),
									},
									End: ast.Location{
										Line: int(77),
										Column: int(34),
									},
									File: p1,
								},
								Fodder: ast.Fodder{},
								Ctx: p15074,
								FreeVars: ast.Identifiers{
									"base",
								},
							},
							Left: &ast.Var{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(77),
											Column: int(24),
										},
										End: ast.Location{
											Line: int(77),
											Column: int(28),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p15077,
									FreeVars: ast.Identifiers{
										"base",
									},
								},
								Id: "base",
							},
							OpFodder: ast.Fodder{},
							Op: ast.BinaryOp(10),
							Right: &ast.LiteralNumber{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(77),
											Column: int(32),
										},
										End: ast.Location{
											Line: int(77),
											Column: int(34),
										},
										File: p1,
									},
									Fodder: ast.Fodder{},
									Ctx: p15080,
									FreeVars: nil,
								},
								OriginalString: "16",
							},
						},
					},
					ThenFodder: nil,
					BranchTrue: &ast.Local{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "<std>",
								Begin: ast.Location{
									Line: int(79),
									Column: int(5),
								},
								End: ast.Location{
									Line: int(92),
									Column: int(49),
								},
								File: p1,
							},
							Fodder: ast.Fodder{
								ast.FodderElement{
									Kind: ast.FodderKind(0),
									Blanks: int(0),
									Indent: int(4),
									Comment: []string{},
								},
								ast.FodderElement{
									Kind: ast.FodderKind(2),
									Blanks: int(0),
									Indent: int(4),
									Comment: []string{
										"// These codepoints are in ascending order:",
									},
								},
							},
							Ctx: p15084,
							FreeVars: ast.Identifiers{
								"base",
								"std",
								"str",
							},
						},
						Binds: ast.LocalBinds{
							ast.LocalBind{
								VarFodder: ast.Fodder{},
								Variable: "zero_code",
								EqFodder: ast.Fodder{},
								Body: &ast.Apply{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(79),
												Column: int(23),
											},
											End: ast.Location{
												Line: int(79),
												Column: int(41),
											},
											File: p1,
										},
										Fodder: ast.Fodder{},
										Ctx: p15088,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Index{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(79),
													Column: int(23),
												},
												End: ast.Location{
													Line: int(79),
													Column: int(36),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p15091,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(79),
														Column: int(23),
													},
													End: ast.Location{
														Line: int(79),
														Column: int(26),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: nil,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Id: "std",
										},
										LeftBracketFodder: ast.Fodder{},
										Index: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "",
													Begin: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													End: ast.Location{
														Line: int(0),
														Column: int(0),
													},
													File: nil,
												},
												Fodder: nil,
												Ctx: nil,
												FreeVars: nil,
											},
											Value: "codepoint",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										RightBracketFodder: ast.Fodder{},
										Id: nil,
									},
									FodderLeft: ast.Fodder{},
									Arguments: ast.Arguments{
										Positional: []ast.CommaSeparatedExpr{
											ast.CommaSeparatedExpr{
												Expr: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(79),
																Column: int(37),
															},
															End: ast.Location{
																Line: int(79),
																Column: int(40),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p15098,
														FreeVars: nil,
													},
													Value: "0",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												CommaFodder: nil,
											},
										},
										Named: []ast.NamedArgument{},
									},
									TrailingComma: false,
									TailStrict: false,
									FodderRight: ast.Fodder{},
									TailStrictFodder: nil,
								},
								Fun: nil,
								CloseFodder: ast.Fodder{},
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(79),
										Column: int(11),
									},
									End: ast.Location{
										Line: int(79),
										Column: int(41),
									},
									File: p1,
								},
							},
						},
						Body: &ast.Local{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "<std>",
									Begin: ast.Location{
										Line: int(80),
										Column: int(5),
									},
									End: ast.Location{
										Line: int(92),
										Column: int(49),
									},
									File: p1,
								},
								Fodder: ast.Fodder{
									ast.FodderElement{
										Kind: ast.FodderKind(0),
										Blanks: int(0),
										Indent: int(4),
										Comment: []string{},
									},
								},
								Ctx: p15101,
								FreeVars: ast.Identifiers{
									"base",
									"std",
									"str",
									"zero_code",
								},
							},
							Binds: ast.LocalBinds{
								ast.LocalBind{
									VarFodder: ast.Fodder{},
									Variable: "upper_a_code",
									EqFodder: ast.Fodder{},
									Body: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(80),
													Column: int(26),
												},
												End: ast.Location{
													Line: int(80),
													Column: int(44),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p15105,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(80),
														Column: int(26),
													},
													End: ast.Location{
														Line: int(80),
														Column: int(39),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p15108,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(80),
															Column: int(26),
														},
														End: ast.Location{
															Line: int(80),
															Column: int(29),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: ast.Fodder{},
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "codepoint",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: ast.Fodder{},
											Id: nil,
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(80),
																	Column: int(40),
																},
																End: ast.Location{
																	Line: int(80),
																	Column: int(43),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p15115,
															FreeVars: nil,
														},
														Value: "A",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													CommaFodder: nil,
												},
											},
											Named: []ast.NamedArgument{},
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
									Fun: nil,
									CloseFodder: ast.Fodder{},
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(80),
											Column: int(11),
										},
										End: ast.Location{
											Line: int(80),
											Column: int(44),
										},
										File: p1,
									},
								},
							},
							Body: &ast.Local{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "<std>",
										Begin: ast.Location{
											Line: int(81),
											Column: int(5),
										},
										End: ast.Location{
											Line: int(92),
											Column: int(49),
										},
										File: p1,
									},
									Fodder: ast.Fodder{
										ast.FodderElement{
											Kind: ast.FodderKind(0),
											Blanks: int(0),
											Indent: int(4),
											Comment: []string{},
										},
									},
									Ctx: p15118,
									FreeVars: ast.Identifiers{
										"base",
										"std",
										"str",
										"upper_a_code",
										"zero_code",
									},
								},
								Binds: ast.LocalBinds{
									ast.LocalBind{
										VarFodder: ast.Fodder{},
										Variable: "lower_a_code",
										EqFodder: ast.Fodder{},
										Body: &ast.Apply{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(81),
														Column: int(26),
													},
													End: ast.Location{
														Line: int(81),
														Column: int(44),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p15122,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Index{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(81),
															Column: int(26),
														},
														End: ast.Location{
															Line: int(81),
															Column: int(39),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p15125,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Var{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(81),
																Column: int(26),
															},
															End: ast.Location{
																Line: int(81),
																Column: int(29),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Id: "std",
												},
												LeftBracketFodder: ast.Fodder{},
												Index: &ast.LiteralString{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: nil,
													},
													Value: "codepoint",
													Kind: ast.LiteralStringKind(1),
													BlockIndent: "",
													BlockTermIndent: "",
												},
												RightBracketFodder: ast.Fodder{},
												Id: nil,
											},
											FodderLeft: ast.Fodder{},
											Arguments: ast.Arguments{
												Positional: []ast.CommaSeparatedExpr{
													ast.CommaSeparatedExpr{
														Expr: &ast.LiteralString{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(81),
																		Column: int(40),
																	},
																	End: ast.Location{
																		Line: int(81),
																		Column: int(43),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p15132,
																FreeVars: nil,
															},
															Value: "a",
															Kind: ast.LiteralStringKind(1),
															BlockIndent: "",
															BlockTermIndent: "",
														},
														CommaFodder: nil,
													},
												},
												Named: []ast.NamedArgument{},
											},
											TrailingComma: false,
											TailStrict: false,
											FodderRight: ast.Fodder{},
											TailStrictFodder: nil,
										},
										Fun: nil,
										CloseFodder: ast.Fodder{},
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(81),
												Column: int(11),
											},
											End: ast.Location{
												Line: int(81),
												Column: int(44),
											},
											File: p1,
										},
									},
								},
								Body: &ast.Local{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "<std>",
											Begin: ast.Location{
												Line: int(82),
												Column: int(5),
											},
											End: ast.Location{
												Line: int(92),
												Column: int(49),
											},
											File: p1,
										},
										Fodder: ast.Fodder{
											ast.FodderElement{
												Kind: ast.FodderKind(0),
												Blanks: int(0),
												Indent: int(4),
												Comment: []string{},
											},
										},
										Ctx: p15135,
										FreeVars: ast.Identifiers{
											"base",
											"lower_a_code",
											"std",
											"str",
											"upper_a_code",
											"zero_code",
										},
									},
									Binds: ast.LocalBinds{
										ast.LocalBind{
											VarFodder: nil,
											Variable: "addDigit",
											EqFodder: nil,
											Body: &ast.Function{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(82),
															Column: int(11),
														},
														End: ast.Location{
															Line: int(91),
															Column: int(31),
														},
														File: p1,
													},
													Fodder: nil,
													Ctx: p15139,
													FreeVars: ast.Identifiers{
														"base",
														"lower_a_code",
														"std",
														"str",
														"upper_a_code",
														"zero_code",
													},
												},
												ParenLeftFodder: ast.Fodder{},
												Parameters: []ast.Parameter{
													ast.Parameter{
														NameFodder: ast.Fodder{},
														Name: "aggregate",
														EqFodder: nil,
														DefaultArg: nil,
														CommaFodder: ast.Fodder{},
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(82),
																Column: int(20),
															},
															End: ast.Location{
																Line: int(82),
																Column: int(29),
															},
															File: p1,
														},
													},
													ast.Parameter{
														NameFodder: ast.Fodder{},
														Name: "char",
														EqFodder: nil,
														DefaultArg: nil,
														CommaFodder: nil,
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(82),
																Column: int(31),
															},
															End: ast.Location{
																Line: int(82),
																Column: int(35),
															},
															File: p1,
														},
													},
												},
												TrailingComma: false,
												ParenRightFodder: ast.Fodder{},
												Body: &ast.Local{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(83),
																Column: int(7),
															},
															End: ast.Location{
																Line: int(91),
																Column: int(31),
															},
															File: p1,
														},
														Fodder: ast.Fodder{
															ast.FodderElement{
																Kind: ast.FodderKind(0),
																Blanks: int(0),
																Indent: int(6),
																Comment: []string{},
															},
														},
														Ctx: p15144,
														FreeVars: ast.Identifiers{
															"aggregate",
															"base",
															"char",
															"lower_a_code",
															"std",
															"str",
															"upper_a_code",
															"zero_code",
														},
													},
													Binds: ast.LocalBinds{
														ast.LocalBind{
															VarFodder: ast.Fodder{},
															Variable: "code",
															EqFodder: ast.Fodder{},
															Body: &ast.Apply{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(83),
																			Column: int(20),
																		},
																		End: ast.Location{
																			Line: int(83),
																			Column: int(39),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p15148,
																	FreeVars: ast.Identifiers{
																		"char",
																		"std",
																	},
																},
																Target: &ast.Index{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(83),
																				Column: int(20),
																			},
																			End: ast.Location{
																				Line: int(83),
																				Column: int(33),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p15151,
																		FreeVars: ast.Identifiers{
																			"std",
																		},
																	},
																	Target: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(83),
																					Column: int(20),
																				},
																				End: ast.Location{
																					Line: int(83),
																					Column: int(23),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Id: "std",
																	},
																	LeftBracketFodder: ast.Fodder{},
																	Index: &ast.LiteralString{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: nil,
																		},
																		Value: "codepoint",
																		Kind: ast.LiteralStringKind(1),
																		BlockIndent: "",
																		BlockTermIndent: "",
																	},
																	RightBracketFodder: ast.Fodder{},
																	Id: nil,
																},
																FodderLeft: ast.Fodder{},
																Arguments: ast.Arguments{
																	Positional: []ast.CommaSeparatedExpr{
																		ast.CommaSeparatedExpr{
																			Expr: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(83),
																							Column: int(34),
																						},
																						End: ast.Location{
																							Line: int(83),
																							Column: int(38),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p15158,
																					FreeVars: ast.Identifiers{
																						"char",
																					},
																				},
																				Id: "char",
																			},
																			CommaFodder: nil,
																		},
																	},
																	Named: []ast.NamedArgument{},
																},
																TrailingComma: false,
																TailStrict: false,
																FodderRight: ast.Fodder{},
																TailStrictFodder: nil,
															},
															Fun: nil,
															CloseFodder: ast.Fodder{},
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(83),
																	Column: int(13),
																},
																End: ast.Location{
																	Line: int(83),
																	Column: int(39),
																},
																File: p1,
															},
														},
													},
													Body: &ast.Local{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(84),
																	Column: int(7),
																},
																End: ast.Location{
																	Line: int(91),
																	Column: int(31),
																},
																File: p1,
															},
															Fodder: ast.Fodder{
																ast.FodderElement{
																	Kind: ast.FodderKind(0),
																	Blanks: int(0),
																	Indent: int(6),
																	Comment: []string{},
																},
															},
															Ctx: p15162,
															FreeVars: ast.Identifiers{
																"aggregate",
																"base",
																"code",
																"lower_a_code",
																"std",
																"str",
																"upper_a_code",
																"zero_code",
															},
														},
														Binds: ast.LocalBinds{
															ast.LocalBind{
																VarFodder: ast.Fodder{},
																Variable: "digit",
																EqFodder: ast.Fodder{},
																Body: &ast.Conditional{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(84),
																				Column: int(21),
																			},
																			End: ast.Location{
																				Line: int(89),
																				Column: int(25),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p15166,
																		FreeVars: ast.Identifiers{
																			"code",
																			"lower_a_code",
																			"upper_a_code",
																			"zero_code",
																		},
																	},
																	Cond: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(84),
																					Column: int(24),
																				},
																				End: ast.Location{
																					Line: int(84),
																					Column: int(44),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p15169,
																			FreeVars: ast.Identifiers{
																				"code",
																				"lower_a_code",
																			},
																		},
																		Left: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(84),
																						Column: int(24),
																					},
																					End: ast.Location{
																						Line: int(84),
																						Column: int(28),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p15172,
																				FreeVars: ast.Identifiers{
																					"code",
																				},
																			},
																			Id: "code",
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(8),
																		Right: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(84),
																						Column: int(32),
																					},
																					End: ast.Location{
																						Line: int(84),
																						Column: int(44),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p15175,
																				FreeVars: ast.Identifiers{
																					"lower_a_code",
																				},
																			},
																			Id: "lower_a_code",
																		},
																	},
																	ThenFodder: ast.Fodder{},
																	BranchTrue: &ast.Binary{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(85),
																					Column: int(9),
																				},
																				End: ast.Location{
																					Line: int(85),
																					Column: int(33),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p15178,
																			FreeVars: ast.Identifiers{
																				"code",
																				"lower_a_code",
																			},
																		},
																		Left: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(85),
																						Column: int(9),
																					},
																					End: ast.Location{
																						Line: int(85),
																						Column: int(28),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p15181,
																				FreeVars: ast.Identifiers{
																					"code",
																					"lower_a_code",
																				},
																			},
																			Left: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(85),
																							Column: int(9),
																						},
																						End: ast.Location{
																							Line: int(85),
																							Column: int(13),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(8),
																							Comment: []string{},
																						},
																					},
																					Ctx: p15185,
																					FreeVars: ast.Identifiers{
																						"code",
																					},
																				},
																				Id: "code",
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(4),
																			Right: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(85),
																							Column: int(16),
																						},
																						End: ast.Location{
																							Line: int(85),
																							Column: int(28),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p15188,
																					FreeVars: ast.Identifiers{
																						"lower_a_code",
																					},
																				},
																				Id: "lower_a_code",
																			},
																		},
																		OpFodder: ast.Fodder{},
																		Op: ast.BinaryOp(3),
																		Right: &ast.LiteralNumber{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(85),
																						Column: int(31),
																					},
																					End: ast.Location{
																						Line: int(85),
																						Column: int(33),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p15191,
																				FreeVars: nil,
																			},
																			OriginalString: "10",
																		},
																	},
																	ElseFodder: ast.Fodder{
																		ast.FodderElement{
																			Kind: ast.FodderKind(0),
																			Blanks: int(0),
																			Indent: int(6),
																			Comment: []string{},
																		},
																	},
																	BranchFalse: &ast.Conditional{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(86),
																					Column: int(12),
																				},
																				End: ast.Location{
																					Line: int(89),
																					Column: int(25),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p15194,
																			FreeVars: ast.Identifiers{
																				"code",
																				"upper_a_code",
																				"zero_code",
																			},
																		},
																		Cond: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(86),
																						Column: int(15),
																					},
																					End: ast.Location{
																						Line: int(86),
																						Column: int(35),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p15197,
																				FreeVars: ast.Identifiers{
																					"code",
																					"upper_a_code",
																				},
																			},
																			Left: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(86),
																							Column: int(15),
																						},
																						End: ast.Location{
																							Line: int(86),
																							Column: int(19),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p15200,
																					FreeVars: ast.Identifiers{
																						"code",
																					},
																				},
																				Id: "code",
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(8),
																			Right: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(86),
																							Column: int(23),
																						},
																						End: ast.Location{
																							Line: int(86),
																							Column: int(35),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p15203,
																					FreeVars: ast.Identifiers{
																						"upper_a_code",
																					},
																				},
																				Id: "upper_a_code",
																			},
																		},
																		ThenFodder: ast.Fodder{},
																		BranchTrue: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(87),
																						Column: int(9),
																					},
																					End: ast.Location{
																						Line: int(87),
																						Column: int(33),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p15206,
																				FreeVars: ast.Identifiers{
																					"code",
																					"upper_a_code",
																				},
																			},
																			Left: &ast.Binary{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(87),
																							Column: int(9),
																						},
																						End: ast.Location{
																							Line: int(87),
																							Column: int(28),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p15209,
																					FreeVars: ast.Identifiers{
																						"code",
																						"upper_a_code",
																					},
																				},
																				Left: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(87),
																								Column: int(9),
																							},
																							End: ast.Location{
																								Line: int(87),
																								Column: int(13),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{
																							ast.FodderElement{
																								Kind: ast.FodderKind(0),
																								Blanks: int(0),
																								Indent: int(8),
																								Comment: []string{},
																							},
																						},
																						Ctx: p15213,
																						FreeVars: ast.Identifiers{
																							"code",
																						},
																					},
																					Id: "code",
																				},
																				OpFodder: ast.Fodder{},
																				Op: ast.BinaryOp(4),
																				Right: &ast.Var{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(87),
																								Column: int(16),
																							},
																							End: ast.Location{
																								Line: int(87),
																								Column: int(28),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p15216,
																						FreeVars: ast.Identifiers{
																							"upper_a_code",
																						},
																					},
																					Id: "upper_a_code",
																				},
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(3),
																			Right: &ast.LiteralNumber{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(87),
																							Column: int(31),
																						},
																						End: ast.Location{
																							Line: int(87),
																							Column: int(33),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p15219,
																					FreeVars: nil,
																				},
																				OriginalString: "10",
																			},
																		},
																		ElseFodder: ast.Fodder{
																			ast.FodderElement{
																				Kind: ast.FodderKind(0),
																				Blanks: int(0),
																				Indent: int(6),
																				Comment: []string{},
																			},
																		},
																		BranchFalse: &ast.Binary{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(89),
																						Column: int(9),
																					},
																					End: ast.Location{
																						Line: int(89),
																						Column: int(25),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p15222,
																				FreeVars: ast.Identifiers{
																					"code",
																					"zero_code",
																				},
																			},
																			Left: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(89),
																							Column: int(9),
																						},
																						End: ast.Location{
																							Line: int(89),
																							Column: int(13),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{
																						ast.FodderElement{
																							Kind: ast.FodderKind(0),
																							Blanks: int(0),
																							Indent: int(8),
																							Comment: []string{},
																						},
																					},
																					Ctx: p15226,
																					FreeVars: ast.Identifiers{
																						"code",
																					},
																				},
																				Id: "code",
																			},
																			OpFodder: ast.Fodder{},
																			Op: ast.BinaryOp(4),
																			Right: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(89),
																							Column: int(16),
																						},
																						End: ast.Location{
																							Line: int(89),
																							Column: int(25),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p15229,
																					FreeVars: ast.Identifiers{
																						"zero_code",
																					},
																				},
																				Id: "zero_code",
																			},
																		},
																	},
																},
																Fun: nil,
																CloseFodder: ast.Fodder{},
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(84),
																		Column: int(13),
																	},
																	End: ast.Location{
																		Line: int(89),
																		Column: int(25),
																	},
																	File: p1,
																},
															},
														},
														Body: &ast.Conditional{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "",
																	Begin: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	End: ast.Location{
																		Line: int(0),
																		Column: int(0),
																	},
																	File: nil,
																},
																Fodder: nil,
																Ctx: nil,
																FreeVars: ast.Identifiers{
																	"aggregate",
																	"base",
																	"digit",
																	"std",
																	"str",
																},
															},
															Cond: &ast.Binary{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(90),
																			Column: int(14),
																		},
																		End: ast.Location{
																			Line: int(90),
																			Column: int(40),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p15234,
																	FreeVars: ast.Identifiers{
																		"base",
																		"digit",
																	},
																},
																Left: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(90),
																				Column: int(14),
																			},
																			End: ast.Location{
																				Line: int(90),
																				Column: int(24),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p15237,
																		FreeVars: ast.Identifiers{
																			"digit",
																		},
																	},
																	Left: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(90),
																					Column: int(14),
																				},
																				End: ast.Location{
																					Line: int(90),
																					Column: int(19),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p15240,
																			FreeVars: ast.Identifiers{
																				"digit",
																			},
																		},
																		Id: "digit",
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(8),
																	Right: &ast.LiteralNumber{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(90),
																					Column: int(23),
																				},
																				End: ast.Location{
																					Line: int(90),
																					Column: int(24),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p15243,
																			FreeVars: nil,
																		},
																		OriginalString: "0",
																	},
																},
																OpFodder: ast.Fodder{},
																Op: ast.BinaryOp(17),
																Right: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(90),
																				Column: int(28),
																			},
																			End: ast.Location{
																				Line: int(90),
																				Column: int(40),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p15245,
																		FreeVars: ast.Identifiers{
																			"base",
																			"digit",
																		},
																	},
																	Left: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(90),
																					Column: int(28),
																				},
																				End: ast.Location{
																					Line: int(90),
																					Column: int(33),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p15248,
																			FreeVars: ast.Identifiers{
																				"digit",
																			},
																		},
																		Id: "digit",
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(9),
																	Right: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(90),
																					Column: int(36),
																				},
																				End: ast.Location{
																					Line: int(90),
																					Column: int(40),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p15251,
																			FreeVars: ast.Identifiers{
																				"base",
																			},
																		},
																		Id: "base",
																	},
																},
															},
															ThenFodder: nil,
															BranchTrue: &ast.Binary{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(91),
																			Column: int(7),
																		},
																		End: ast.Location{
																			Line: int(91),
																			Column: int(31),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p15254,
																	FreeVars: ast.Identifiers{
																		"aggregate",
																		"base",
																		"digit",
																	},
																},
																Left: &ast.Binary{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(91),
																				Column: int(7),
																			},
																			End: ast.Location{
																				Line: int(91),
																				Column: int(23),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p15257,
																		FreeVars: ast.Identifiers{
																			"aggregate",
																			"base",
																		},
																	},
																	Left: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(91),
																					Column: int(7),
																				},
																				End: ast.Location{
																					Line: int(91),
																					Column: int(11),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{
																				ast.FodderElement{
																					Kind: ast.FodderKind(0),
																					Blanks: int(0),
																					Indent: int(6),
																					Comment: []string{},
																				},
																			},
																			Ctx: p15261,
																			FreeVars: ast.Identifiers{
																				"base",
																			},
																		},
																		Id: "base",
																	},
																	OpFodder: ast.Fodder{},
																	Op: ast.BinaryOp(0),
																	Right: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(91),
																					Column: int(14),
																				},
																				End: ast.Location{
																					Line: int(91),
																					Column: int(23),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p15264,
																			FreeVars: ast.Identifiers{
																				"aggregate",
																			},
																		},
																		Id: "aggregate",
																	},
																},
																OpFodder: ast.Fodder{},
																Op: ast.BinaryOp(3),
																Right: &ast.Var{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(91),
																				Column: int(26),
																			},
																			End: ast.Location{
																				Line: int(91),
																				Column: int(31),
																			},
																			File: p1,
																		},
																		Fodder: ast.Fodder{},
																		Ctx: p15267,
																		FreeVars: ast.Identifiers{
																			"digit",
																		},
																	},
																	Id: "digit",
																},
															},
															ElseFodder: nil,
															BranchFalse: &ast.Error{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"base",
																		"std",
																		"str",
																	},
																},
																Expr: &ast.Apply{
																	NodeBase: ast.NodeBase{
																		LocRange: ast.LocationRange{
																			FileName: "",
																			Begin: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			End: ast.Location{
																				Line: int(0),
																				Column: int(0),
																			},
																			File: nil,
																		},
																		Fodder: nil,
																		Ctx: nil,
																		FreeVars: ast.Identifiers{
																			"base",
																			"std",
																			"str",
																		},
																	},
																	Target: &ast.Index{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "",
																				Begin: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				End: ast.Location{
																					Line: int(0),
																					Column: int(0),
																				},
																				File: nil,
																			},
																			Fodder: nil,
																			Ctx: nil,
																			FreeVars: ast.Identifiers{
																				"std",
																			},
																		},
																		Target: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: ast.Identifiers{
																					"std",
																				},
																			},
																			Id: "std",
																		},
																		LeftBracketFodder: nil,
																		Index: &ast.LiteralString{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "",
																					Begin: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					End: ast.Location{
																						Line: int(0),
																						Column: int(0),
																					},
																					File: nil,
																				},
																				Fodder: nil,
																				Ctx: nil,
																				FreeVars: nil,
																			},
																			Value: "mod",
																			Kind: ast.LiteralStringKind(1),
																			BlockIndent: "",
																			BlockTermIndent: "",
																		},
																		RightBracketFodder: nil,
																		Id: nil,
																	},
																	FodderLeft: nil,
																	Arguments: ast.Arguments{
																		Positional: []ast.CommaSeparatedExpr{
																			ast.CommaSeparatedExpr{
																				Expr: &ast.LiteralString{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(90),
																								Column: int(43),
																							},
																							End: ast.Location{
																								Line: int(90),
																								Column: int(72),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p15280,
																						FreeVars: nil,
																					},
																					Value: "%s is not a base %d integer",
																					Kind: ast.LiteralStringKind(1),
																					BlockIndent: "",
																					BlockTermIndent: "",
																				},
																				CommaFodder: nil,
																			},
																			ast.CommaSeparatedExpr{
																				Expr: &ast.Array{
																					NodeBase: ast.NodeBase{
																						LocRange: ast.LocationRange{
																							FileName: "<std>",
																							Begin: ast.Location{
																								Line: int(90),
																								Column: int(75),
																							},
																							End: ast.Location{
																								Line: int(90),
																								Column: int(86),
																							},
																							File: p1,
																						},
																						Fodder: ast.Fodder{},
																						Ctx: p15282,
																						FreeVars: ast.Identifiers{
																							"base",
																							"str",
																						},
																					},
																					Elements: []ast.CommaSeparatedExpr{
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(90),
																											Column: int(76),
																										},
																										End: ast.Location{
																											Line: int(90),
																											Column: int(79),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p15286,
																									FreeVars: ast.Identifiers{
																										"str",
																									},
																								},
																								Id: "str",
																							},
																							CommaFodder: ast.Fodder{},
																						},
																						ast.CommaSeparatedExpr{
																							Expr: &ast.Var{
																								NodeBase: ast.NodeBase{
																									LocRange: ast.LocationRange{
																										FileName: "<std>",
																										Begin: ast.Location{
																											Line: int(90),
																											Column: int(81),
																										},
																										End: ast.Location{
																											Line: int(90),
																											Column: int(85),
																										},
																										File: p1,
																									},
																									Fodder: ast.Fodder{},
																									Ctx: p15289,
																									FreeVars: ast.Identifiers{
																										"base",
																									},
																								},
																								Id: "base",
																							},
																							CommaFodder: nil,
																						},
																					},
																					TrailingComma: false,
																					CloseFodder: ast.Fodder{},
																				},
																				CommaFodder: nil,
																			},
																		},
																		Named: nil,
																	},
																	TrailingComma: false,
																	TailStrict: false,
																	FodderRight: nil,
																	TailStrictFodder: nil,
																},
															},
														},
													},
												},
											},
											Fun: nil,
											CloseFodder: nil,
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
										},
									},
									Body: &ast.Apply{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "<std>",
												Begin: ast.Location{
													Line: int(92),
													Column: int(5),
												},
												End: ast.Location{
													Line: int(92),
													Column: int(49),
												},
												File: p1,
											},
											Fodder: ast.Fodder{},
											Ctx: p15292,
											FreeVars: ast.Identifiers{
												"addDigit",
												"std",
												"str",
											},
										},
										Target: &ast.Index{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(92),
														Column: int(5),
													},
													End: ast.Location{
														Line: int(92),
														Column: int(14),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p15295,
												FreeVars: ast.Identifiers{
													"std",
												},
											},
											Target: &ast.Var{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(92),
															Column: int(5),
														},
														End: ast.Location{
															Line: int(92),
															Column: int(8),
														},
														File: p1,
													},
													Fodder: ast.Fodder{
														ast.FodderElement{
															Kind: ast.FodderKind(0),
															Blanks: int(0),
															Indent: int(4),
															Comment: []string{},
														},
													},
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Id: "std",
											},
											LeftBracketFodder: ast.Fodder{},
											Index: &ast.LiteralString{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: nil,
												},
												Value: "foldl",
												Kind: ast.LiteralStringKind(1),
												BlockIndent: "",
												BlockTermIndent: "",
											},
											RightBracketFodder: ast.Fodder{},
											Id: nil,
										},
										FodderLeft: ast.Fodder{},
										Arguments: ast.Arguments{
											Positional: []ast.CommaSeparatedExpr{
												ast.CommaSeparatedExpr{
													Expr: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(92),
																	Column: int(15),
																},
																End: ast.Location{
																	Line: int(92),
																	Column: int(23),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p15303,
															FreeVars: ast.Identifiers{
																"addDigit",
															},
														},
														Id: "addDigit",
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.Apply{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(92),
																	Column: int(25),
																},
																End: ast.Location{
																	Line: int(92),
																	Column: int(45),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p15306,
															FreeVars: ast.Identifiers{
																"std",
																"str",
															},
														},
														Target: &ast.Index{
															NodeBase: ast.NodeBase{
																LocRange: ast.LocationRange{
																	FileName: "<std>",
																	Begin: ast.Location{
																		Line: int(92),
																		Column: int(25),
																	},
																	End: ast.Location{
																		Line: int(92),
																		Column: int(40),
																	},
																	File: p1,
																},
																Fodder: ast.Fodder{},
																Ctx: p15309,
																FreeVars: ast.Identifiers{
																	"std",
																},
															},
															Target: &ast.Var{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(92),
																			Column: int(25),
																		},
																		End: ast.Location{
																			Line: int(92),
																			Column: int(28),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: nil,
																	FreeVars: ast.Identifiers{
																		"std",
																	},
																},
																Id: "std",
															},
															LeftBracketFodder: ast.Fodder{},
															Index: &ast.LiteralString{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "",
																		Begin: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		End: ast.Location{
																			Line: int(0),
																			Column: int(0),
																		},
																		File: nil,
																	},
																	Fodder: nil,
																	Ctx: nil,
																	FreeVars: nil,
																},
																Value: "stringChars",
																Kind: ast.LiteralStringKind(1),
																BlockIndent: "",
																BlockTermIndent: "",
															},
															RightBracketFodder: ast.Fodder{},
															Id: nil,
														},
														FodderLeft: ast.Fodder{},
														Arguments: ast.Arguments{
															Positional: []ast.CommaSeparatedExpr{
																ast.CommaSeparatedExpr{
																	Expr: &ast.Var{
																		NodeBase: ast.NodeBase{
																			LocRange: ast.LocationRange{
																				FileName: "<std>",
																				Begin: ast.Location{
																					Line: int(92),
																					Column: int(41),
																				},
																				End: ast.Location{
																					Line: int(92),
																					Column: int(44),
																				},
																				File: p1,
																			},
																			Fodder: ast.Fodder{},
																			Ctx: p15316,
																			FreeVars: ast.Identifiers{
																				"str",
																			},
																		},
																		Id: "str",
																	},
																	CommaFodder: nil,
																},
															},
															Named: []ast.NamedArgument{},
														},
														TrailingComma: false,
														TailStrict: false,
														FodderRight: ast.Fodder{},
														TailStrictFodder: nil,
													},
													CommaFodder: ast.Fodder{},
												},
												ast.CommaSeparatedExpr{
													Expr: &ast.LiteralNumber{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(92),
																	Column: int(47),
																},
																End: ast.Location{
																	Line: int(92),
																	Column: int(48),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: p15319,
															FreeVars: nil,
														},
														OriginalString: "0",
													},
													CommaFodder: nil,
												},
											},
											Named: []ast.NamedArgument{},
										},
										TrailingComma: false,
										TailStrict: false,
										FodderRight: ast.Fodder{},
										TailStrictFodder: nil,
									},
								},
							},
						},
					},
					ElseFodder: nil,
					BranchFalse: &ast.Error{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"base",
								"std",
							},
						},
						Expr: &ast.Apply{
							NodeBase: ast.NodeBase{
								LocRange: ast.LocationRange{
									FileName: "",
									Begin: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									End: ast.Location{
										Line: int(0),
										Column: int(0),
									},
									File: nil,
								},
								Fodder: nil,
								Ctx: nil,
								FreeVars: ast.Identifiers{
									"base",
									"std",
								},
							},
							Target: &ast.Index{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"std",
									},
								},
								Target: &ast.Var{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Id: "std",
								},
								LeftBracketFodder: nil,
								Index: &ast.LiteralString{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: nil,
									},
									Value: "mod",
									Kind: ast.LiteralStringKind(1),
									BlockIndent: "",
									BlockTermIndent: "",
								},
								RightBracketFodder: nil,
								Id: nil,
							},
							FodderLeft: nil,
							Arguments: ast.Arguments{
								Positional: []ast.CommaSeparatedExpr{
									ast.CommaSeparatedExpr{
										Expr: &ast.LiteralString{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(77),
														Column: int(37),
													},
													End: ast.Location{
														Line: int(77),
														Column: int(62),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p15331,
												FreeVars: nil,
											},
											Value: "integer base %d invalid",
											Kind: ast.LiteralStringKind(1),
											BlockIndent: "",
											BlockTermIndent: "",
										},
										CommaFodder: nil,
									},
									ast.CommaSeparatedExpr{
										Expr: &ast.Var{
											NodeBase: ast.NodeBase{
												LocRange: ast.LocationRange{
													FileName: "<std>",
													Begin: ast.Location{
														Line: int(77),
														Column: int(65),
													},
													End: ast.Location{
														Line: int(77),
														Column: int(69),
													},
													File: p1,
												},
												Fodder: ast.Fodder{},
												Ctx: p15333,
												FreeVars: ast.Identifiers{
													"base",
												},
											},
											Id: "base",
										},
										CommaFodder: nil,
									},
								},
								Named: nil,
							},
							TrailingComma: false,
							TailStrict: false,
							FodderRight: nil,
							TailStrictFodder: nil,
						},
					},
				},
			},
			Fun: nil,
			CloseFodder: nil,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(76),
					Column: int(9),
				},
				End: ast.Location{
					Line: int(92),
					Column: int(49),
				},
				File: p1,
			},
		},
		ast.LocalBind{
			VarFodder: nil,
			Variable: "base64_table",
			EqFodder: nil,
			Body: &ast.LiteralString{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "<std>",
						Begin: ast.Location{
							Line: int(1090),
							Column: int(24),
						},
						End: ast.Location{
							Line: int(1090),
							Column: int(90),
						},
						File: p1,
					},
					Fodder: ast.Fodder{},
					Ctx: p15336,
					FreeVars: nil,
				},
				Value: "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/",
				Kind: ast.LiteralStringKind(1),
				BlockIndent: "",
				BlockTermIndent: "",
			},
			Fun: nil,
			CloseFodder: nil,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(1090),
					Column: int(9),
				},
				End: ast.Location{
					Line: int(1090),
					Column: int(90),
				},
				File: p1,
			},
		},
		ast.LocalBind{
			VarFodder: nil,
			Variable: "base64_inv",
			EqFodder: nil,
			Body: &ast.Apply{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: ast.Identifiers{
						"base64_table",
						"std",
					},
				},
				Target: &ast.Index{
					NodeBase: ast.NodeBase{
						LocRange: ast.LocationRange{
							FileName: "",
							Begin: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							End: ast.Location{
								Line: int(0),
								Column: int(0),
							},
							File: nil,
						},
						Fodder: nil,
						Ctx: nil,
						FreeVars: ast.Identifiers{
							"std",
						},
					},
					Target: &ast.Var{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
							Fodder: nil,
							Ctx: nil,
							FreeVars: ast.Identifiers{
								"std",
							},
						},
						Id: "std",
					},
					LeftBracketFodder: nil,
					Index: &ast.LiteralString{
						NodeBase: ast.NodeBase{
							LocRange: ast.LocationRange{
								FileName: "",
								Begin: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								End: ast.Location{
									Line: int(0),
									Column: int(0),
								},
								File: nil,
							},
							Fodder: nil,
							Ctx: nil,
							FreeVars: nil,
						},
						Value: "$objectFlatMerge",
						Kind: ast.LiteralStringKind(1),
						BlockIndent: "",
						BlockTermIndent: "",
					},
					RightBracketFodder: nil,
					Id: nil,
				},
				FodderLeft: nil,
				Arguments: ast.Arguments{
					Positional: []ast.CommaSeparatedExpr{
						ast.CommaSeparatedExpr{
							Expr: &ast.Apply{
								NodeBase: ast.NodeBase{
									LocRange: ast.LocationRange{
										FileName: "",
										Begin: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										End: ast.Location{
											Line: int(0),
											Column: int(0),
										},
										File: nil,
									},
									Fodder: nil,
									Ctx: nil,
									FreeVars: ast.Identifiers{
										"base64_table",
										"std",
									},
								},
								Target: &ast.Index{
									NodeBase: ast.NodeBase{
										LocRange: ast.LocationRange{
											FileName: "",
											Begin: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											End: ast.Location{
												Line: int(0),
												Column: int(0),
											},
											File: nil,
										},
										Fodder: nil,
										Ctx: nil,
										FreeVars: ast.Identifiers{
											"std",
										},
									},
									Target: &ast.Var{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: ast.Identifiers{
												"std",
											},
										},
										Id: "std",
									},
									LeftBracketFodder: nil,
									Index: &ast.LiteralString{
										NodeBase: ast.NodeBase{
											LocRange: ast.LocationRange{
												FileName: "",
												Begin: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												End: ast.Location{
													Line: int(0),
													Column: int(0),
												},
												File: nil,
											},
											Fodder: nil,
											Ctx: nil,
											FreeVars: nil,
										},
										Value: "flatMap",
										Kind: ast.LiteralStringKind(1),
										BlockIndent: "",
										BlockTermIndent: "",
									},
									RightBracketFodder: nil,
									Id: nil,
								},
								FodderLeft: nil,
								Arguments: ast.Arguments{
									Positional: []ast.CommaSeparatedExpr{
										ast.CommaSeparatedExpr{
											Expr: &ast.Function{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "",
														Begin: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														End: ast.Location{
															Line: int(0),
															Column: int(0),
														},
														File: nil,
													},
													Fodder: nil,
													Ctx: nil,
													FreeVars: ast.Identifiers{
														"base64_table",
													},
												},
												ParenLeftFodder: nil,
												Parameters: []ast.Parameter{
													ast.Parameter{
														NameFodder: nil,
														Name: "i",
														EqFodder: nil,
														DefaultArg: nil,
														CommaFodder: nil,
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
													},
												},
												TrailingComma: false,
												ParenRightFodder: nil,
												Body: &ast.Array{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "",
															Begin: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															End: ast.Location{
																Line: int(0),
																Column: int(0),
															},
															File: nil,
														},
														Fodder: nil,
														Ctx: nil,
														FreeVars: ast.Identifiers{
															"base64_table",
															"i",
														},
													},
													Elements: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.DesugaredObject{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1091),
																			Column: int(22),
																		},
																		End: ast.Location{
																			Line: int(1091),
																			Column: int(72),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p15360,
																	FreeVars: ast.Identifiers{
																		"base64_table",
																		"i",
																	},
																},
																Asserts: ast.Nodes{},
																Fields: ast.DesugaredObjectFields{
																	ast.DesugaredObjectField{
																		Hide: ast.ObjectFieldHide(1),
																		Name: &ast.Index{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1091),
																						Column: int(25),
																					},
																					End: ast.Location{
																						Line: int(1091),
																						Column: int(40),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p15364,
																				FreeVars: ast.Identifiers{
																					"base64_table",
																					"i",
																				},
																			},
																			Target: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1091),
																							Column: int(25),
																						},
																						End: ast.Location{
																							Line: int(1091),
																							Column: int(37),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p15367,
																					FreeVars: ast.Identifiers{
																						"base64_table",
																					},
																				},
																				Id: "base64_table",
																			},
																			LeftBracketFodder: ast.Fodder{},
																			Index: &ast.Var{
																				NodeBase: ast.NodeBase{
																					LocRange: ast.LocationRange{
																						FileName: "<std>",
																						Begin: ast.Location{
																							Line: int(1091),
																							Column: int(38),
																						},
																						End: ast.Location{
																							Line: int(1091),
																							Column: int(39),
																						},
																						File: p1,
																					},
																					Fodder: ast.Fodder{},
																					Ctx: p15370,
																					FreeVars: ast.Identifiers{
																						"i",
																					},
																				},
																				Id: "i",
																			},
																			RightBracketFodder: ast.Fodder{},
																			Id: nil,
																		},
																		Body: &ast.Var{
																			NodeBase: ast.NodeBase{
																				LocRange: ast.LocationRange{
																					FileName: "<std>",
																					Begin: ast.Location{
																						Line: int(1091),
																						Column: int(43),
																					},
																					End: ast.Location{
																						Line: int(1091),
																						Column: int(44),
																					},
																					File: p1,
																				},
																				Fodder: ast.Fodder{},
																				Ctx: p15373,
																				FreeVars: ast.Identifiers{
																					"i",
																				},
																			},
																			Id: "i",
																		},
																		PlusSuper: false,
																		LocRange: ast.LocationRange{
																			FileName: "<std>",
																			Begin: ast.Location{
																				Line: int(1091),
																				Column: int(24),
																			},
																			End: ast.Location{
																				Line: int(1091),
																				Column: int(44),
																			},
																			File: p1,
																		},
																	},
																},
																Locals: ast.LocalBinds{},
															},
															CommaFodder: nil,
														},
													},
													TrailingComma: false,
													CloseFodder: nil,
												},
											},
											CommaFodder: nil,
										},
										ast.CommaSeparatedExpr{
											Expr: &ast.Apply{
												NodeBase: ast.NodeBase{
													LocRange: ast.LocationRange{
														FileName: "<std>",
														Begin: ast.Location{
															Line: int(1091),
															Column: int(54),
														},
														End: ast.Location{
															Line: int(1091),
															Column: int(70),
														},
														File: p1,
													},
													Fodder: ast.Fodder{},
													Ctx: p15376,
													FreeVars: ast.Identifiers{
														"std",
													},
												},
												Target: &ast.Index{
													NodeBase: ast.NodeBase{
														LocRange: ast.LocationRange{
															FileName: "<std>",
															Begin: ast.Location{
																Line: int(1091),
																Column: int(54),
															},
															End: ast.Location{
																Line: int(1091),
																Column: int(63),
															},
															File: p1,
														},
														Fodder: ast.Fodder{},
														Ctx: p15379,
														FreeVars: ast.Identifiers{
															"std",
														},
													},
													Target: &ast.Var{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "<std>",
																Begin: ast.Location{
																	Line: int(1091),
																	Column: int(54),
																},
																End: ast.Location{
																	Line: int(1091),
																	Column: int(57),
																},
																File: p1,
															},
															Fodder: ast.Fodder{},
															Ctx: nil,
															FreeVars: ast.Identifiers{
																"std",
															},
														},
														Id: "std",
													},
													LeftBracketFodder: ast.Fodder{},
													Index: &ast.LiteralString{
														NodeBase: ast.NodeBase{
															LocRange: ast.LocationRange{
																FileName: "",
																Begin: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																End: ast.Location{
																	Line: int(0),
																	Column: int(0),
																},
																File: nil,
															},
															Fodder: nil,
															Ctx: nil,
															FreeVars: nil,
														},
														Value: "range",
														Kind: ast.LiteralStringKind(1),
														BlockIndent: "",
														BlockTermIndent: "",
													},
													RightBracketFodder: ast.Fodder{},
													Id: nil,
												},
												FodderLeft: ast.Fodder{},
												Arguments: ast.Arguments{
													Positional: []ast.CommaSeparatedExpr{
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralNumber{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1091),
																			Column: int(64),
																		},
																		End: ast.Location{
																			Line: int(1091),
																			Column: int(65),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p15386,
																	FreeVars: nil,
																},
																OriginalString: "0",
															},
															CommaFodder: ast.Fodder{},
														},
														ast.CommaSeparatedExpr{
															Expr: &ast.LiteralNumber{
																NodeBase: ast.NodeBase{
																	LocRange: ast.LocationRange{
																		FileName: "<std>",
																		Begin: ast.Location{
																			Line: int(1091),
																			Column: int(67),
																		},
																		End: ast.Location{
																			Line: int(1091),
																			Column: int(69),
																		},
																		File: p1,
																	},
																	Fodder: ast.Fodder{},
																	Ctx: p15388,
																	FreeVars: nil,
																},
																OriginalString: "63",
															},
															CommaFodder: nil,
														},
													},
													Named: []ast.NamedArgument{},
												},
												TrailingComma: false,
												TailStrict: false,
												FodderRight: ast.Fodder{},
												TailStrictFodder: nil,
											},
											CommaFodder: nil,
										},
									},
									Named: nil,
								},
								TrailingComma: false,
								TailStrict: false,
								FodderRight: nil,
								TailStrictFodder: nil,
							},
							CommaFodder: nil,
						},
					},
					Named: nil,
				},
				TrailingComma: false,
				TailStrict: false,
				FodderRight: nil,
				TailStrictFodder: nil,
			},
			Fun: nil,
			CloseFodder: nil,
			LocRange: ast.LocationRange{
				FileName: "<std>",
				Begin: ast.Location{
					Line: int(1091),
					Column: int(9),
				},
				End: ast.Location{
					Line: int(1091),
					Column: int(72),
				},
				File: p1,
			},
		},
		ast.LocalBind{
			VarFodder: nil,
			Variable: "$",
			EqFodder: nil,
			Body: &ast.Self{
				NodeBase: ast.NodeBase{
					LocRange: ast.LocationRange{
						FileName: "",
						Begin: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						End: ast.Location{
							Line: int(0),
							Column: int(0),
						},
						File: nil,
					},
					Fodder: nil,
					Ctx: nil,
					FreeVars: nil,
				},
			},
			Fun: nil,
			CloseFodder: nil,
			LocRange: ast.LocationRange{
				FileName: "",
				Begin: ast.Location{
					Line: int(0),
					Column: int(0),
				},
				End: ast.Location{
					Line: int(0),
					Column: int(0),
				},
				File: nil,
			},
		},
	},
}
